/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "BaumerImageSequence.h"
#include "SeariseLogger.hpp"
#include <boost/format.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/condition_variable.hpp>
#include <boost/filesystem.hpp>
#include <sys/wait.h>
#include "get_secs.hpp"
#include "time_logger.hpp"
#include <fstream>

using namespace BGAPI;
using namespace boost;

BGAPI::System* BaumerImageSequence::g_pSystem = NULL;
int BaumerImageSequence::g_active_users_count = 0;
int BaumerImageSequence::g_system_count = -1;
std::vector< BaumerImageSequence::singletonCamera > BaumerImageSequence::g_pCameras = std::vector< BaumerImageSequence::singletonCamera >();
std::string BaumerImageSequence::g_library_path = "../libs/Baumer/Bin";

class CameraBuffer
{
public:

  typedef CircularFifo< BGAPI::Image * > image_fifo_t;
  void  set_constant_fps( float fps ){ constant_fps = fps; }
  float get_constant_fps( ) const {return constant_fps; }
private:	
	
  image_fifo_t image_fifo;
  size_t header_size;
  size_t width, height, depth;
  HostImage::ColorSpace space;
  BGAPI_RESULT res;
  BGAPI::System * system;
  bool is_new;
  size_t dropped_cnt, caught_cnt;
  bool verbose;
  int camera_number;  
  size_t total_cnt;
  int prev_time_s;
  int prev_time_h;
  TimeLogger logger;
  TimeLogger time_tic;
  float      constant_fps;
  BGAPI::Camera* my_pCamera;

public:
  int hw_start_time;
  int hw_start_num;
  double start_time;
  double hw_fps;
  double fps_ratio;
  double dt;
  double last_hw_time;
public:  
  
  CameraBuffer( const CameraBuffer & )
	: header_size(0)
	, width(0)
	, height(0)
	, depth(0)
	, system(0)
	, res(BGAPI_RESULT_OK)
	, dropped_cnt(0)
	, caught_cnt(0)
	, verbose(false)
	, camera_number(-1)
	, total_cnt(0)
    , prev_time_s(0)
    , prev_time_h(0)
    , time_tic( "", false )
    , constant_fps( 0  )
    , my_pCamera(NULL)
    , hw_start_time(0)
    , hw_start_num(0)
    , start_time(0)
    , hw_fps(0)
    , fps_ratio(1)
    , dt(0)
    , last_hw_time(0)
  {
  }

  image_fifo_t & fifo() { return image_fifo; }

  CameraBuffer()
	: header_size(0)
	, width(0)
	, height(0)
	, depth(0)
	, system(0)
	, res(BGAPI_RESULT_OK)
	, dropped_cnt(0)
	, caught_cnt(0)
	, verbose(true)
	, camera_number(-1)
	, total_cnt(0)
    , prev_time_s(0)
    , prev_time_h(0)
    , time_tic( "", false )
    , constant_fps( 0 )
    , my_pCamera(NULL)
    , hw_start_time(0)
    , hw_start_num(0)
    , start_time(0)
    , hw_fps(0)
    , fps_ratio(1)
    , dt(0)
    , last_hw_time(0)
  {
  }


	bool keep() { return true; }
  
  void caught() { ++caught_cnt; }
  void dropped() { ++dropped_cnt; }
  float efficiency() { return float(caught_cnt) / std::max( 1.0f, float(dropped_cnt + caught_cnt) ); }
  
  ~CameraBuffer()
  {
		BOOST_FOREACH( BGAPI::Image * image, image_fifo )
		{
			/// delete ImageBuffer
			unsigned char* tmp = NULL;
			//res = image->getBuffer( &tmp );
			//if( res != BGAPI_RESULT_OK )
			//	throw std::runtime_error(str( boost::format("getBuffer errorcode: %d\n") % res));
			//delete[] tmp;				
			/// RELEASING IMAGE
			res = system->releaseImage( image );
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("releaseImage errorcode: %d\n") % res));
		}
  }
  
  void setDestinationPixelFormat( int fmt )
  {
	BOOST_FOREACH( BGAPI::Image * image, image_fifo )
	{
		res = image->setDestinationPixelFormat(fmt);
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("BGAPI::Camera::setDestinationPixelFormat Errorcode: %d\n") % res));
	}
  } 

  void getDestinationPixelFormat( int sourceFormat, BGAPI_FeatureState & state, BGAPIX_TypeListINT & destpixellist )
  {
	res = image_fifo.get_head()->getDestinationPixelFormat( sourceFormat, &state, &destpixellist );
	if( res != BGAPI_RESULT_OK )
	  throw std::runtime_error(str( boost::format("BGAPI::Camera::getPixelFormat Errorcode: %d\n") % res));
  }
  
  HostImage copy_current_image( bool wait = true )
  {
		logger.start();
        
        double after_lock = get_secs();
        int times = 0;
		
        
        if ( constant_fps != 0 && (image_fifo.size() - image_fifo.space_available() <= 1) )
        {
          double tic_time = time_tic.tic();
          double sleeptime = ((1.f/constant_fps) - tic_time)* 1000000;
          if (sleeptime < 0 )
          {
            //log().info << "WARNING: computation speed is not sufficient to ensure "<< constant_fps << " fps - loop took "<< (tic_time*1000) << " ms, while "<< ((1.f/constant_fps)*1000) << " ms are maximum"<<std::endl;
            sleeptime = 0;
          }
          //log().info << "sleeping for " << (sleeptime/1000.f) << " ms to get constant fps of " << constant_fps << std::endl;
          usleep( sleeptime );
          time_tic.tic();
        }
        
#ifdef USE_TRIGGER
        /// trigger next image capture
        BGAPI_RESULT res = my_pCamera->doTrigger();
        if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->doTrigger errorcode: %d camera index: %d\n") % res % camera_number));
#endif

		if ( wait && image_fifo.empty() )
		{
#if 1
            //std::cout << "  CONSUMER: baumer waiting " << std::flush;
			while ( image_fifo.empty() )
			{
				//std::cout << "." << std::flush;
				usleep( 1000 );
			}
			//std::cout << std::endl;
#else
			image_fifo.consumer_busy_wait( 1000 );
#endif
		}
		if ( image_fifo.empty() )
        { 
            logger.end();
			return HostImage();
        }
		//std::cout << "  CONSUMER: SPACE AVAILABLE " << image_fifo.space_available()  << std::endl;

		int num_s=-1, num_h=-1, time_s=-1, time_h=-1;
		int rows=0, cols=0;	

		double after_wait = get_secs();
		//if ( (after_wait-after_lock) > 0.2 )  log().info << "WARNING! camera receiver " <<camera_number<<"  waited "<< (after_wait-after_lock) <<" s ("<< times<<" loops) for new frame, check network connection chain for bottlenecks! "<< std::endl;  
		
		BGAPI::Image * image = image_fifo.get_head();
		
		//get time and sequence info ( and check size )
		image->getSize( &rows, &cols );
		image->getNumber( &num_s, &num_h );
		image->getTimeStamp( &time_s, &time_h );
        //std::cout << "   time btw frames time_s = " << (double( time_s - prev_time_s )/1000.f) << "   time_h = " << (double( time_h - prev_time_h )/1000.f) << std::endl;
        prev_time_h = time_h; prev_time_s = time_s;
        
		if ( cols != height || rows != width)
			throw std::runtime_error(str( boost::format("retrieved image size [%d x %d] does not match the expected"
														" [%d x %d]\n")%cols %rows %height %width ));			
		
		// get the data
		unsigned char * datap;
		res = image->get( &datap );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("BGAPI::Image::get Errorcode: %d\n") % res));
		CPUData<unsigned char> data( height, width, depth );
		memcpy( data.data(), datap, data.size() );
		
		image_fifo.pop();   

		// deliver a set up host image with unique data
		HostImage result( data, space );
		result.setFrameNumber( num_s );
		result.setTimeStamp( time_h );
        
        logger.end();
		return result;
  }
  
  

  void init( size_t datasize, size_t w, size_t h, size_t d, HostImage::ColorSpace s, BGAPI::System * sys, int camera_number_ = -1, size_t bufsize = 5, bool write_log = false )
  {		
		image_fifo.init( bufsize );

		system = sys;
		width = w;
		height = h;
		depth = d;
		space = s;
		size_t size_of_buffer = datasize + 1;
		header_size = datasize - d * w * h + 1;
		camera_number = camera_number_;
        my_pCamera = BaumerImageSequence::g_pCameras[camera_number].pCamera;
        
        if (write_log)
          logger.init( str( boost::format("log_camerabuffer_%d.txt")%camera_number ), true );

		typedef BGAPI::Image * image_p;
		BOOST_FOREACH( image_p & image, image_fifo )
		{
			std::cout << " creating image of size ["<< w << " x "<< h <<" x " << d << "]" << std::endl;
            image = NULL;
			res = createImage( &image );
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("BGAPI::createImage, Errorcode: %d\n") % res));
#ifdef USE_EXTERNAL_BUFFER
            image->setBuffer( new unsigned char[size_of_buffer], size_of_buffer );
#endif
		}
  }

  int get_camera_number() const 
  {
	  return camera_number;
  }

  void get_timings( double & current_hw_time, double & current_sw_time )
  {
    int time_h, time_s, num_h, num_s;
    image_fifo.get_tail()->getTimeStamp( &time_s, &time_h );
    image_fifo.get_tail()->getNumber( &num_s, &num_h );
    if ( hw_start_time == 0 )
    {
        hw_start_time = time_h;
        hw_start_num = num_h;
        start_time = get_secs();
        last_hw_time = current_hw_time = 0;
    }
    else
    {
        current_hw_time = (time_h - hw_start_time ) * HW_TIMER_FACTOR;
        current_sw_time = get_secs() - start_time;
        if ( hw_fps == 0 )
        {
            hw_fps = 1.f / (float(current_hw_time)/float(num_h-hw_start_num) );
            if ( constant_fps == 0 )
                constant_fps = hw_fps;
            fps_ratio = hw_fps / constant_fps;
            dt = 1.f / constant_fps;
            std::cout << "Camera "<<camera_number<<": camera hardware fps = "<<hw_fps << "  desired fps = "<< constant_fps<< "  ratio = "<<fps_ratio<< std::endl;
            if ( fabs( fps_ratio - round(fps_ratio) ) > 1e-2 )
                std::cout << "WARNING! fps ratio for cam "<<camera_number<<" is not int, so odd frameskipping is unavoidable!"<<std::endl;
        }
        if ( current_hw_time < 0.f )
        {
            hw_start_time = time_h;
            hw_start_num = num_h; 
            start_time = get_secs();
            current_hw_time = current_sw_time = last_hw_time = 0.f;
        }
    }
  }

};

static std::map<BGAPI::Camera *,CameraBuffer> g_camera_buffers;
static std::map<BGAPI::Camera *,TimeLogger  > g_camera_loggers;
static std::map<BGAPI::Camera *,TimeLogger  > g_callback_loggers;


/// global callback function
BGAPI_RESULT BGAPI_CALLBACK imageCallback( void * callBackOwner, BGAPI::Image * pImage )
{
	/// get objects
    BGAPI::Camera * cam = (BGAPI::Camera*)callBackOwner;
	BGAPI_RESULT res = BGAPI_RESULT_OK;
	CameraBuffer & buffer = g_camera_buffers[cam];
	
    /// initialize temporary variables
    int time_s, time_h, num_s, num_h;
    double current_hw_time = 0, current_sw_time = 0;
    
    /// get timings and timers
    buffer.get_timings( current_hw_time, current_sw_time );
    
    /// if desired fps is equal or higher than hardware fps, push everything into buffer as long as it is not full
    //std::cout << "imagecallback: starting to access buffer"<<std::endl;
    if ( buffer.fps_ratio < 1 )
    {
        if ( buffer.fifo().space_available() >= 2 )
        {
          buffer.fifo().push();
          buffer.caught();
        }
        else
          buffer.dropped();
    }
    else
    {
    /// else: check if image can be stored:
    /// buffer space must be free & image timing has to be right
      if (  buffer.fifo().space_available() >= 2                         &&  /// enough buffer space free?
            current_hw_time  >  buffer.last_hw_time + 0.5f * buffer.dt       /// check if this frame is closest to the desired frame time
         )
      {
        //if ( current_hw_time  > buffer.last_hw_time + 1.5f * buffer.dt )
        //  std::cout << "WARNING: producer "<< buffer.get_camera_number() <<" lags behind desired frame rate. expected receive time = "<< (buffer.last_hw_time + 1.0f * buffer.dt) << "  real received time = "<< current_hw_time << "  efficency = "<< buffer.efficiency()<< std::endl;
        buffer.fifo().push();
        buffer.caught();
        buffer.last_hw_time = current_hw_time;
        //std::cout << " cam " << g_camera_buffers[cam].get_camera_number() << " space available " << buffer.fifo().space_available() << std::endl;
      }
      else
      {
        buffer.dropped();
        //std::cout << "DROPPED! current_hw_time = "<< current_hw_time << "  last_time = "<< buffer.last_hw_time << "  dt = " << buffer.dt <<std::endl;
        //std::cout << "  PRODUCER "<<g_camera_buffers[cam].get_camera_number()<<": warning: dropping image from cam " << g_camera_buffers[cam].get_camera_number() << " eficency = "<< buffer.efficiency()<<std::endl;
      }
    }
    //std::cout << "imagecallback: done accessing buffer, filling"<<std::endl;
	double tic = g_callback_loggers[ cam ].tic();

    g_camera_loggers[ cam ].start();
    res = cam->setImage( buffer.fifo().get_tail() );
	g_camera_loggers[ cam ].end();
  
    //std::cout << "imagecallback: done filling"<<std::endl;



/*
    buffer.fifo().get_tail()->getTimeStamp( &time_s, &time_h );
    buffer.fifo().get_tail()->getNumber( &num_s, &num_h );
    if ( hw_start_time == 0 )
    {
        hw_start_time = time_h;
        hw_start_num = num_h;
        start_time = get_secs();
    }
    else
    {
        current_hw_time = (time_h - hw_start_time ) * HW_TIMER_FACTOR;
        current_sw_time = get_secs() - start_time;
    
        if ( hw_fps == 0 )
        {
            hw_fps = 1.f / (float(current_hw_time)/float(num_h-hw_start_num) );
            std::cout << "hw fps = "<<hw_fps<<std::endl;
        }
        if ( current_hw_time < 0.f )
        {
            hw_start_time = time_h;
            hw_start_num = num_h; 
            start_time = get_secs();
            current_hw_time = current_sw_time = 0.f;
        }
    }
    
    std::cout << " sw_time = " << current_sw_time << "  hw_time = " << current_hw_time << "  hw_fps = "<< (1.f / (float(current_hw_time)/float(num_h-hw_start_num) )) << "  desired fps = "<< buffer.get_constant_fps() <<"  err = "<< fabs(current_hw_time - current_sw_time )<<std::endl;
        //std::cout << " callback: time btw frames = "<< ( double(time_h - g_time) * HW_TIMER_FACTOR ) << "  tic = "<< tic <<  std::endl;
    */
    
    if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("setImage failed with %d\n") % res ));

	//g_callback_log << ( get_secs() - start_time ) << " " << g_camera_buffers[cam].get_camera_number() << " " << -1<<std::endl;
		
	return res;
}


/// create env. and connect to first free camera
BaumerImageSequence::BaumerImageSequence( HostImage::ColorSpace C, int imageformat, int pixelformat, int gain, float exposure, int bufferlength_, bool write_log_ )
: _space( C )
, _frame_number(0)
, _time_stamp(0)
, _camera_running(false)
, _camera_number(0)
, pSystem(NULL)
, pCamera(NULL)
, res(BGAPI_RESULT_FAIL)
, ImageFormatIndex(DEFAULT_IMAGEFORMAT)
, PixelFormatIndex(DEFAULT_PIXELFORMAT)
, exposurevalue( exposure == -1 ? DEFAULT_EXPOSURE : exposure )
, gainvalue( gain == -1 ? DEFAULT_GAIN : exposure )
, bufferlength( bufferlength_ == -1 ? DEFAULT_BUFFERLENGTH : bufferlength_ )
, write_log( write_log_ )
{
    /// try to init the system 
	pSystem = initBGAPI();
	
	init( 0, imageformat, pixelformat );
}


/// create env. and connect to first free camera
BaumerImageSequence::BaumerImageSequence( std::string NameOrSerial, HostImage::ColorSpace C, int imageformat, int pixelformat, int gain, float exposure, int bufferlength_, bool write_log_ )
: _space( C )
, _frame_number(0)
, _time_stamp(0)
, _camera_running(false)
, _camera_number(0)
, _camera_name( NameOrSerial )
, pSystem(NULL)
, pCamera( NULL )
, res(BGAPI_RESULT_FAIL)
, ImageFormatIndex(DEFAULT_IMAGEFORMAT)
, PixelFormatIndex(DEFAULT_PIXELFORMAT)
, exposurevalue( exposure == -1 ? DEFAULT_EXPOSURE : exposure )
, gainvalue( gain == -1 ? DEFAULT_GAIN : exposure )
, bufferlength( bufferlength_ == -1 ? DEFAULT_BUFFERLENGTH : bufferlength_ )
, write_log( write_log_ )
{
    
    /// try to init the system 
	pSystem = initBGAPI();
	
	init( getCameraByString(NameOrSerial), imageformat, pixelformat );
	
}


/// create env. and connect to first free camera
BaumerImageSequence::BaumerImageSequence( int camera_num, HostImage::ColorSpace C, int imageformat, int pixelformat, int gain, float exposure, int bufferlength_, bool write_log_ )
: _space( C )
, _frame_number(0)
, _time_stamp(0)
, _camera_running(false)
, _camera_number(0)
, pSystem(NULL)
, pCamera(NULL)
, res(BGAPI_RESULT_FAIL)
, ImageFormatIndex( DEFAULT_IMAGEFORMAT )
, PixelFormatIndex( DEFAULT_PIXELFORMAT )
, exposurevalue( exposure == -1 ? DEFAULT_EXPOSURE : exposure )
, gainvalue( gain == -1 ? DEFAULT_GAIN : exposure )
, bufferlength( bufferlength_ == -1 ? DEFAULT_BUFFERLENGTH : bufferlength_ )
, write_log( write_log_ )
{
    /// try to init the system 
	pSystem = initBGAPI();
	
	init( camera_num, imageformat, pixelformat );
}




HostImage BaumerImageSequence::getNextImage()
{	
	startCamera();
	CameraBuffer & cbuf = g_camera_buffers[pCamera];
	_current_image = cbuf.copy_current_image( true );
	return _current_image;
}


HostImage BaumerImageSequence::getCurrentImage()
{
	if ( _frame_number < 0 )
	  return getNextImage();
	else
	  return _current_image;
}

bool BaumerImageSequence::hasNextImage()
{
	return startCamera();
}

int BaumerImageSequence::getWidth()
{
	return _width;
}

int BaumerImageSequence::getHeight()
{
	return _height;
}

void BaumerImageSequence::reset()
{
	_frame_number = 0;
	_time_stamp = 0;
	stopCamera();
}

bool BaumerImageSequence::has_file( std::string filename )
{
	boost::filesystem::path dir_path = ".";
	
	if ( !boost::filesystem::exists( dir_path ) ) return false;
  boost::filesystem::directory_iterator end_itr; // default construction yields past-the-end
  for ( boost::filesystem::directory_iterator itr( dir_path );
        itr != end_itr;
        ++itr )
  {
    if ( is_directory(itr->status()) )
    {
      continue;
    }
    else if ( itr->path().filename() == filename ) // see below
    {
      return true;
    }
  }
  return false;
}



BaumerImageSequence::~BaumerImageSequence()
{
  clear();
  
	/// try to release the system
  releaseBGAPI();
}

void BaumerImageSequence::init( int camera_number, int imageformat, int pixelformat )
{

	
	
	state.cbSize = sizeof( BGAPI_FeatureState );
	exposurerange.cbSize = sizeof( BGAPIX_TypeRangeINT );
	gainrange.cbSize = sizeof( BGAPIX_TypeRangeFLOAT );
	
	/// get some gobal variables
	int camera_count = getCameraCount();
	if (camera_number >= camera_count || camera_number<0 )
			throw std::runtime_error(str( boost::format("Wrong camera index: %d available cameras: %d (%d)\n") 
										  % camera_number % camera_count% (g_pCameras[camera_number].isFree==false ? "free": "used") ));
	if ( g_pCameras[camera_number].isFree == false )
	  throw std::runtime_error(str( boost::format("camera %d busy\n") 
									% camera_number ) );
	
	/// bind camera
	_camera_number = camera_number;
	pCamera = g_pCameras[_camera_number].pCamera;
	g_pCameras[_camera_number].isFree = false;
	
	/// OPENING CREATED CAMERA
	res = pCamera->open();
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("Camera->open errorcode: %d camera index: %d\n") % res % _camera_number));
	
#if 0
	/// set ip
	unsigned char ip[4]         = { 168, 192,   0, 12  };
	unsigned char subnetmask[4] = { 255, 255, 255,  0  };
	unsigned char gateway[4]    = { 168, 192,   0,  1  };
	setIP( ip, subnetmask, gateway );
#endif
	
	/// NORMAL MODE / USING EXTERNAL BUFFER
#ifdef USE_EXTERNAL_BUFFER
    res = pCamera->setDataAccessMode( BGAPI_DATAACCESSMODE_NORMALEXTERN, bufferlength );
#else
    res = pCamera->setDataAccessMode( BGAPI_DATAACCESSMODE_NORMALINTERN, bufferlength );
#endif

#ifdef USE_TRIGGER
    /// activate trigger
    res = pCamera->setTriggerSource( BGAPI_TRIGGERSOURCE_SOFTWARE );
    if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->setTriggerSource errorcode: %d camera index: %d\n") % res % _camera_number));
    //res = pCamera->setTriggerActivation( BGAPI_ACTIVATION_STATEHIGH );
    //if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->setTriggerActivation errorcode: %d camera index: %d\n") % res % _camera_number));
    res = pCamera->setTrigger( true );
    if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->setTrigger errorcode: %d camera index: %d\n") % res % _camera_number));
    BGAPIX_TypeRangeINT delay;
    delay.cbSize = sizeof ( BGAPIX_TypeRangeINT );
    res = pCamera->getTriggerDelay( &state, &delay );
    int old_delay, wanted_delay = 0, new_delay;
    if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->getTriggerDelay errorcode: %d camera index: %d\n") % res % _camera_number));
    old_delay = delay.current;
    res = pCamera->setTriggerDelay( wanted_delay );
    if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->śetTriggerDelay errorcode: %d camera index: %d\n") % res % _camera_number));
    res = pCamera->getTriggerDelay( &state, &delay );
    if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->getTriggerDelay errorcode: %d camera index: %d\n") % res % _camera_number));
    new_delay = delay.current;
    std::cout << "TriggerDelay: old = "<< old_delay << "  wanted = "<< wanted_delay << "  new = "<< new_delay<<std::endl;
#else
    res = pCamera->setTriggerSource( BGAPI_TRIGGERSOURCE_OFF );
    if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->setTriggerSource errorcode: %d camera index: %d\n") % res % _camera_number));
    res = pCamera->setTrigger( false );
    if( res != BGAPI_RESULT_OK )        throw std::runtime_error(str( boost::format("Camera->setTrigger errorcode: %d camera index: %d\n") % res % _camera_number));
#endif


	/// grab and store available image/pixel formats
	//setImageFormat( ImageFormatIndex );
	switch ( _space )
	{
		case HostImage::LUM:
								ImageFormatIndex = imageformat == -1 ? 1 : imageformat;
                                setImageFormat( ImageFormatIndex );
                                if ( pixelformat == -1 )
                                    setPixelFormat( "Mono8" );
                                else
                                    setPixelFormat( pixelformat );
                                _depth = 1;
								break;
		case HostImage::RGB:
                                
								ImageFormatIndex = imageformat == -1 ? 2 : imageformat;
                                setImageFormat( ImageFormatIndex );
                                if ( pixelformat == -1 )
                                    setPixelFormat( "RGB8Packed" );
                                else
                                    setPixelFormat( pixelformat );
								_depth = 3;
								break;
		case HostImage::YCbCr:
								ImageFormatIndex = imageformat == -1 ? 2 : imageformat;
                                setImageFormat( ImageFormatIndex );
                                if ( pixelformat == -1 )
                                    setPixelFormat( "YUV444Packed" );
                                else
                                    setPixelFormat( pixelformat );
								_depth = 3;
								break;
								
	}
    
	getPixelFormats( true );
	std::cout << getImageFormatStr() << std::endl;
    std::cout << getPixelFormatStr() << std::endl;
    
	/// allocate unsigned char buffer
	BGAPIX_CameraImageFormat my_format = getCurrentImageFormat( true );
	_height = my_format.iSizeY;
	_width = my_format.iSizeX;
	Datalength.cbSize = sizeof ( BGAPIX_TypeRangeINT );
	pCamera->getDataLength( &state, &Datalength );

	g_camera_buffers[pCamera].init( Datalength.current, _width, _height, _depth, _space, pSystem, camera_number, bufferlength, write_log);
    if ( write_log )
    {
        g_camera_loggers[pCamera].init( str(boost::format("log_callback_%d_%s.txt")% camera_number %_camera_name), true );
        g_callback_loggers[pCamera].init( str(boost::format("log_callback_dt_%d_%s.txt")% camera_number % _camera_name), false );
    }
    res = pCamera->setImage( g_camera_buffers[pCamera].fifo().get_tail() );
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("pCamera->setImage errorcode: %d\n") % res));
	

	res = pCamera->registerNotifyCallback( pCamera, (BGAPI::BGAPI_NOTIFY_CALLBACK) &imageCallback );
	
	
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("Register NotifyCallback errorcode: %d\n") % res));


    
  getExposure( true );
  setExposure( exposurevalue );
  getGain( true );
  setGain( gainvalue );
}


void BaumerImageSequence::clear()
{
	if (pSystem)
	{
		if (pCamera)
		{
			/// stop camera
			stopCamera();
			
			g_pCameras[_camera_number].isFree = true;

			pCamera = NULL;

		}		
		
	}
}

bool BaumerImageSequence::startCamera()
{
	if (!pSystem || !pCamera ) throw std::runtime_error("pSystem or pCamera or pImage is not initialized!");
	if ( _camera_running == true ) return _camera_running;
	
	/// STARTING CAMERA
	res = pCamera->setStart( true );
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("pCamera->setStart errorcode: %d\n")% res));
		
	_camera_running = true;
	return _camera_running;	
}

bool BaumerImageSequence::stopCamera()
{
	if (!pSystem || !pCamera ) throw std::runtime_error("pSystem or pCamera or pImage is not initialized!");
	if ( _camera_running == false ) return _camera_running;
	
	/// STOPPING CAMERA
	res = pCamera->setStart( false );
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("pCamera->setStart errorcode: %d\n") % res));

	_camera_running = false;
	return _camera_running;	
}

/// ********* Baumer enviroment *************

void BaumerImageSequence::setIP( unsigned char* ip, unsigned char* subnetmask, unsigned char* gateway )
{
	/// init
	BGAPIX_TypeBOOL enable;
	enable.cbSize = sizeof( BGAPIX_TypeBOOL );
	BGAPIX_TypeArrayBYTE my_ip;
	my_ip.cbSize = sizeof( BGAPIX_TypeArrayBYTE );
	BGAPIX_TypeArrayBYTE my_subnetmask;
	my_subnetmask.cbSize = sizeof( BGAPIX_TypeArrayBYTE );
	BGAPIX_TypeArrayBYTE my_gateway;
	my_gateway.cbSize = sizeof( BGAPIX_TypeArrayBYTE );

	/// get ip info and print
	res = pCamera->getGVSPersistentIP( &state, &enable, &my_ip, &my_subnetmask, &my_gateway );
	if( res != BGAPI_RESULT_OK )
			 throw std::runtime_error(str( boost::format("pCamera->getGVSPersistentIP errorcode: %d\n") % res));
			 
	printf( "=== Old IP settings:\n");
	printf( "  = Persistent IP: %s\n", enable.bcur == true ? "Enabled": "Disabled");
	printf( "  = IP: %d.%d.%d.%d \n", my_ip.array[0], my_ip.array[1], my_ip.array[2],my_ip.array[3] );
	printf( "  = Subnetmask %d.%d.%d.%d \n", my_subnetmask.array[0], my_subnetmask.array[1], my_subnetmask.array[2], my_subnetmask.array[3] );
	printf( "  = Gateway %d.%d.%d.%d \n", my_gateway.array[0], my_gateway.array[1], my_gateway.array[2], my_gateway.array[3] );

	/// set IP
	printf( "=== desired IP settings:\n");
	printf( "  = IP: %d.%d.%d.%d \n", ip[0], ip[1], ip[2],ip[3] );
	printf( "  = Subnetmask %d.%d.%d.%d \n", subnetmask[0], subnetmask[1], subnetmask[2], subnetmask[3] );
	printf( "  = Gateway %d.%d.%d.%d \n", gateway[0], gateway[1], gateway[2], gateway[3] );

	res = pCamera->setGVSPersistentIP( true, ip, subnetmask, gateway, 4);
	if( res != BGAPI_RESULT_OK )
			 throw std::runtime_error(str( boost::format("pCamera->setGVSPersistentIP errorcode: %d\n") % res));
	
	/// recheck data
	res = pCamera->getGVSPersistentIP( &state, &enable, &my_ip, &my_subnetmask, &my_gateway );
	if( res != BGAPI_RESULT_OK )
			 throw std::runtime_error(str( boost::format("pCamera->getGVSPersistentIP errorcode: %d\n") % res));
			 
	printf( "=== current IP settings:\n");
	printf( "  = Persistent IP: %s\n", enable.bcur == true ? "Enabled": "Disabled");
	printf( "  = IP: %d.%d.%d.%d \n", my_ip.array[0], my_ip.array[1], my_ip.array[2],my_ip.array[3] );
	printf( "  = Subnetmask %d.%d.%d.%d \n", my_subnetmask.array[0], my_subnetmask.array[1], my_subnetmask.array[2], my_subnetmask.array[3] );
	printf( "  = Gateway %d.%d.%d.%d \n", my_gateway.array[0], my_gateway.array[1], my_gateway.array[2], my_gateway.array[3] );



}




void BaumerImageSequence::setImageFormat( int index )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if (ImageFormats.empty())
		getImageFormats( true );
	
	if (index < 0 || index >= getImageFormats().size() )
		throw std::runtime_error(str( boost::format("setImageFormat: index (%d) out of bounds (%d) Errorcode: %d\n")% index % getImageFormats().size() % res));

	if (index==ImageFormatIndex)	return;
	
	/// SET IMAGE FORMAT
	res = pCamera->setImageFormat( index );
	ImageFormatIndex = index;
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setImageFormat Errorcode: %d\n") % res));
}

void BaumerImageSequence::setImageFormat( std::string name )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if (ImageFormats.empty())
		getImageFormats( true );
	
	int available_at=-1;
	for (int i=0; i<ImageFormats.size(); i++)
	{
		std::string current_name( (char*)ImageFormats[i].sName );
		if ( name.compare( current_name ) == 0 )
		{
			available_at = i;
			break;
		}
	}
	if (available_at < 0)
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setImageFormat: could not find format of name %s, possible formats are:\n%s\n ") % name.c_str() %getImageFormatStr() ));
	
	/// SET IMAGE FORMAT
	res = pCamera->setImageFormat( available_at );
	ImageFormatIndex = available_at;
    
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setImageFormat Errorcode: %d\n") % res));
}


std::vector<BGAPIX_CameraImageFormat> const& BaumerImageSequence::getImageFormats( bool force_check )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if ( force_check || ImageFormats.empty() )
	{
		ImageFormats.clear();
		BGAPIX_TypeListINT listint;
		listint.cbSize = sizeof( BGAPIX_TypeListINT );
		
		/// GET IMAGE FORMAT
		res = pCamera->getImageFormat( &state, &listint );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("BGAPI::Camera::getImageFormat Errorcode: %d\n") % res));

		/// GET IMAGE FORMAT DESCRIPTION FOR ALL AVAILABLE IMAGE FORMATS AND MARK CURRENT
		ImageFormatIndex = listint.current;
		for( int i = 0; i<listint.length; i++ )
		{
			BGAPIX_CameraImageFormat format;
			format.cbSize = sizeof ( BGAPIX_CameraImageFormat );
			res = pCamera->getImageFormatDescription( i, &format );
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("BGAPI::Camera::getImageFormatDesc Errorcode: %d\n") % res));
			ImageFormats.push_back( format );
			// printf("          %cImageFormat %d: %s | %d x %d\n", i==listint.current?'*': ' ', format.iFormat, (char*)format.sName, format.iSizeX, format.iSizeY);
		}
		
	}
	return ImageFormats;
}

std::string BaumerImageSequence::getImageFormatStr()
{
    getImageFormats( true );
    char tmp[256];
    sprintf( tmp, "=== Image Formats for camera %d (%s): ===\n", _camera_number, _camera_name.c_str() );
    std::string res = tmp;
    
    for ( int i = 0; i < ImageFormats.size(); ++i )
    {
        char tmp[256];
        sprintf( tmp, "  %cImageFormat %d: %s | %d x %d offset: %d x %d roi: %d x %d \n", i==ImageFormatIndex?'*': ' ', ImageFormats[i].iFormat, 
				 (char*)ImageFormats[i].sName, 
				 ImageFormats[i].iSizeX, ImageFormats[i].iSizeY,
				 ImageFormats[i].iOffsetX,ImageFormats[i].iOffsetY,
				 ImageFormats[i].iScaleRoiX,ImageFormats[i].iScaleRoiY
				 );
        res.append( tmp );
    }
    res.append( "=== Image Formats end ===\n" );
    return res;
}


int BaumerImageSequence::getCurrentImageFormatIndex( bool force_check )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if ( force_check || ImageFormats.empty() )
		getImageFormats( force_check );
	
	return ImageFormatIndex;
}	

BGAPIX_CameraImageFormat const& BaumerImageSequence::getCurrentImageFormat( bool force_check )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if ( force_check || ImageFormats.empty() )
		return getImageFormats( true )[ImageFormatIndex];
	else
		return getImageFormats( false )[ImageFormatIndex];
}

/**	
 *  TODO: these functions have to be implemented someday...
 * 	float getGain( bool force_check=false ) const;
	void setGain( float gain );
	int getExposure( bool force_check=false ) const;
	void setExposure( int exposure );
**/




void BaumerImageSequence::setExposure( int exposurevalue )
{
  if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	
	/// CHECK IF EXPOSURE TIME IS WITHIN LIMITS
  if( exposurevalue < exposurerange.minimum ) exposurevalue = exposurerange.minimum;
	if( exposurevalue > exposurerange.maximum ) exposurevalue = exposurerange.maximum;
	
	/// SET EXPOSURE TIME
	res = pCamera->setExposure( exposurevalue );
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setExposure Errorcode: %d\n") % res));

}

int BaumerImageSequence::getExposure( bool force_check )
{
  if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if ( force_check )
  {
    res = pCamera->getExposure( &state, &exposurerange );
    if( res != BGAPI_RESULT_OK )
      throw std::runtime_error(str( boost::format("BGAPI::Camera::getExposure Errorcode: %d\n") % res));
    return exposurerange.current;
  }
  else
  {
    return exposurerange.current;
  }
}

void BaumerImageSequence::setGain( float gainvalue )
{
  if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	
	/// CHECK IF EXPOSURE TIME IS WITHIN LIMITS
  if( gainvalue < gainrange.minimum ) gainvalue = gainrange.minimum;
	if( gainvalue > gainrange.maximum ) gainvalue = gainrange.maximum;
	
	/// SET EXPOSURE TIME
	res = pCamera->setGain( gainvalue );
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setGain Errorcode: %d\n") % res));

}

float BaumerImageSequence::getGain( bool force_check )
{
  if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if ( force_check )
  {
    res = pCamera->getGain( &state, &gainrange );
    if( res != BGAPI_RESULT_OK )
      throw std::runtime_error(str( boost::format("BGAPI::Camera::getGain Errorcode: %d\n") % res));
    return gainrange.current;
  }
  else
  {
    return gainrange.current;
  }
}



void BaumerImageSequence::setPacketDelay( int Ipg )
{
    if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
    BGAPIX_TypeINT old_delay, new_delay;
    old_delay.cbSize = sizeof( BGAPIX_TypeINT );
    new_delay.cbSize = sizeof( BGAPIX_TypeINT );
    
    res = pCamera->getGVSPacketDelay( &state, &old_delay );
    if( res != BGAPI_RESULT_OK )
      throw std::runtime_error(str( boost::format("BGAPI::Camera::getGVSPacketDelay Errorcode: %d\n") % res));
    
    if ( old_delay.current != Ipg )
    {
        res = pCamera->setGVSPacketDelay( Ipg );
        if( res != BGAPI_RESULT_OK )
            throw std::runtime_error(str( boost::format("BGAPI::Camera::setGVSPacketDelay Errorcode: %d\n") % res));
    }
    
    res = pCamera->getGVSPacketDelay( &state, &new_delay );
    if( res != BGAPI_RESULT_OK )
      throw std::runtime_error(str( boost::format("BGAPI::Camera::getGVSPacketDelay Errorcode: %d\n") % res));
    
    std::cout << "SET_PACKETDELAY: old = "<< old_delay.current << "  wanted = " << Ipg << "  new = " << new_delay.current << std::endl;
}

void BaumerImageSequence::setPacketSize( int packetsize )
{
    if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
    BGAPIX_TypeRangeINT old_size, new_size;
    old_size.cbSize = sizeof( BGAPIX_TypeRangeINT );
    new_size.cbSize = sizeof( BGAPIX_TypeRangeINT );
    
    res = pCamera->getPacketSize( &state, &old_size );
    if( res != BGAPI_RESULT_OK )
      throw std::runtime_error(str( boost::format("BGAPI::Camera::getPacketSize Errorcode: %d\n") % res));
    
    if ( old_size.current != packetsize )
    {
        res = pCamera->setPacketSize( packetsize );
        if( res != BGAPI_RESULT_OK )
            throw std::runtime_error(str( boost::format("BGAPI::Camera::setPacketSize Errorcode: %d\n") % res));
    }
    
    res = pCamera->getPacketSize( &state, &new_size );
    if( res != BGAPI_RESULT_OK )
      throw std::runtime_error(str( boost::format("BGAPI::Camera::getPacketSize Errorcode: %d\n") % res));
    
    std::cout << "SET_PACKET_SIZE: old = "<< old_size.current << "  wanted = " << packetsize << "  new = " << new_size.current << std::endl;
}

void BaumerImageSequence::setConstantFps( float fps )
{
    g_camera_buffers[ g_pCameras[_camera_number].pCamera ].set_constant_fps(fps);
}

void BaumerImageSequence::setPixelFormat( int index )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if (PixelFormats.empty())
		getPixelFormats( true );
	
	int available_at=index;

	if (available_at < 0 || available_at >= PixelFormats.size())
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setPixelFormat: could not find format of index %d, possible formats are\n%s\n") % index % getPixelFormatStr() ));
	
	/// SET IMAGE FORMAT
	res = pCamera->setPixelFormat( PixelFormats[available_at].iPixelFormat );
	PixelFormatIndex = available_at;
    
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setPixelFormat Errorcode: %d\n") % res));
}

void BaumerImageSequence::setPixelFormat( std::string name )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if (PixelFormats.empty())
		getPixelFormats( true );
	
	int available_at=-1;
	for (int i=0; i<PixelFormats.size(); i++)
	{
		std::string current_name( (char*)PixelFormats[i].sName );
		if ( name.compare( current_name ) == 0 )
		{
			available_at = i;
			break;
		}
	}
	if (available_at < 0)
	{	std::string possible_names("\"");
		for (int i=0; i<PixelFormats.size(); i++)
		{
			possible_names.append( (char*)PixelFormats[i].sName ); 
			possible_names.append( "\",\"" );
		}
		possible_names.append("\"");
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setPixelFormat: could not find format of name \"%s\". possible names are:%s\n") % name.c_str() %possible_names.c_str() ));
	}
	/// SET IMAGE FORMAT
	res = pCamera->setPixelFormat( PixelFormats[available_at].iPixelFormat );
	PixelFormatIndex = available_at;
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setPixelFormat Errorcode: %d\n") % res));
}

int BaumerImageSequence::getPixelFormatByName( std::string name )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");

	getPixelFormats( true );
	
	int available_at=-1;
	for (int i=0; i<PixelFormats.size(); i++)
	{
		std::string current_name( (char*)PixelFormats[i].sName );
		if ( name.compare( current_name ) == 0 )
		{
			available_at = i;
			break;
		}
	}
	if (available_at < 0)
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setPixelFormat: could not find format of name %d\n") % name.c_str()));
	
	/// SET IMAGE FORMAT
	return available_at;
}

std::vector<BGAPIX_CameraPixelFormat> const& BaumerImageSequence::getPixelFormats( bool force_check )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	int imageformatindex = getCurrentImageFormatIndex();
	
	if ( force_check || PixelFormats.empty() )
	{
		PixelFormats.clear();
		/// Variables for PixelFormat (chapter 4.3.2.)
		BGAPIX_TypeListINT pixellist;
		pixellist.cbSize = sizeof( BGAPIX_TypeListINT );
	
		/// GET PIXEL FORMAT
		res = pCamera->getPixelFormat( imageformatindex, &state, &pixellist );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("BGAPI::Camera::getPixelFormat Errorcode: %d\n") % res));
	
		PixelFormatIndex = pixellist.current;
		/// GET PIXEL FORMAT DESCRIPTION FOR ALL AVAILABLE PIXEL FORMATS AND MARK CURRENT
		for( int p = 0; p<pixellist.length; p++ )
		{
			BGAPIX_CameraPixelFormat pformat;
			pformat.cbSize = sizeof ( BGAPIX_CameraPixelFormat );
		
			res = pCamera->getPixelFormatDescription( imageformatindex, pixellist.array[p], &pformat );
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("BGAPI::Camera::getPixelFormatDesc Errorcode: %d\n") % res));
			
			//printf("          %cPixelFormat %x: %s | %d \n", p==pixellist.current?'*': ' ', pformat.iPixelFormat, (char*)pformat.sName, pformat.iPixelBits);
			PixelFormats.push_back( pformat );
		}
	}
	return PixelFormats;
}


std::string BaumerImageSequence::getPixelFormatStr()
{
    getPixelFormats( true );
    char tmp[256];
    sprintf( tmp, "=== Pixel Formats for camera %d (%s): ===\n", _camera_number, _camera_name.c_str() );
    std::string res = tmp;
    
    for ( int i = 0; i < PixelFormats.size(); ++i )
    {
        char tmp[256];
        sprintf( tmp, "  %cPixelFormat %x: %s | %d \n", i==PixelFormatIndex?'*': ' ', PixelFormats[i].iPixelFormat, (char*)PixelFormats[i].sName, PixelFormats[i].iPixelBits);
			
        res.append( tmp );
    }
    res.append( "=== Pixel Formats end ===\n" );
    return res;
}


int BaumerImageSequence::getCurrentPixelFormatIndex( bool force_check )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if ( force_check || PixelFormats.empty() )
		getPixelFormats( force_check );
	
	return PixelFormatIndex;
}	

BGAPIX_CameraPixelFormat const& BaumerImageSequence::getCurrentPixelFormat( bool force_check )
{
	if (!pSystem || !pCamera) throw std::runtime_error("pSystem is not initialized!");
	if ( force_check || ImageFormats.empty() )
		return getPixelFormats( true )[PixelFormatIndex];
	else
		return getPixelFormats( false )[PixelFormatIndex];
	
}


void BaumerImageSequence::setDestinationPixelFormat( int index )
{
	if ( !pSystem || !pCamera ) throw std::runtime_error("pSystem or pCamera or pImage is not initialized!");
	if (DestinationPixelFormats.empty())
		getDestinationPixelFormats( true );
	
	int available_at=-1;
	for (int i=0; i<DestinationPixelFormats.size(); i++)
	{
		if ( index == DestinationPixelFormats[i].iPixelFormat )
		{
			available_at = i;
			break;
		}
	}
	if (available_at < 0)
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setDestinationPixelFormat: could not find format of index %d\n") % index));
	
	/// SET IMAGE FORMAT
	g_camera_buffers[pCamera].setDestinationPixelFormat( DestinationPixelFormats[available_at].iPixelFormat );
	DestinationPixelFormatIndex = available_at;
}

void BaumerImageSequence::setDestinationPixelFormat( std::string name )
{
	if (!pSystem || !pCamera ) throw std::runtime_error("pSystem or pCamera or is not initialized!");
	if (DestinationPixelFormats.empty())
		getDestinationPixelFormats( true );
	
	int available_at=-1;
	for (int i=0; i<DestinationPixelFormats.size(); i++)
	{
		std::string current_name( (char*)DestinationPixelFormats[i].sName );
		if ( name.compare( current_name ) == 0 )
		{
			available_at = i;
			break;
		}
	}
	if (available_at < 0)
		throw std::runtime_error(str( boost::format("BGAPI::Camera::setDestinationPixelFormat: could not find format of name %d\n") % name.c_str()));
	
	/// SET IMAGE FORMAT
	g_camera_buffers[pCamera].setDestinationPixelFormat( DestinationPixelFormats[available_at].iPixelFormat );
	DestinationPixelFormatIndex = available_at;
}


std::vector<BGAPIX_CameraPixelFormat> const& BaumerImageSequence::getDestinationPixelFormats( bool force_check )
{
	if (!pSystem || !pCamera ) throw std::runtime_error("pSystem or pCamera or is not initialized!");
	//int imageformatindex = getCurrentImageFormatIndex();
	int sourceindex = getCurrentPixelFormatIndex();
	int sourceFormat = getPixelFormats()[sourceindex].iPixelFormat;
	
	if ( force_check || DestinationPixelFormats.empty() )
	{
		DestinationPixelFormats.clear();
		/// Variables for PixelFormat (chapter 4.3.2.)
		BGAPIX_TypeListINT destpixellist;
		destpixellist.cbSize = sizeof (BGAPIX_TypeListINT);
	
		/// GET PIXEL FORMAT
		g_camera_buffers[pCamera].getDestinationPixelFormat( sourceFormat, state, destpixellist );
	
		
		/// GET PIXEL FORMAT DESCRIPTION FOR ALL AVAILABLE PIXEL FORMATS AND MARK CURRENT
		DestinationPixelFormatIndex = destpixellist.current;
		for (int p = 0; p < destpixellist.length; p++)
		{
			char destpixel_str[32];
			switch( destpixellist.array[p] )
			{
				case BGAPI_PIXTYPE_MONO8:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_MONO8");
					break;

				case BGAPI_PIXTYPE_MONO8_SIGNED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_MONO8_SIGNED");
					break;

				case BGAPI_PIXTYPE_MONO10:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_MONO10");
					break;

				case BGAPI_PIXTYPE_MONO10_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_MONO10_PACKED");
					break;

				case BGAPI_PIXTYPE_MONO12:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_MONO12");
					break;

				case BGAPI_PIXTYPE_MONO12_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_MONO12PACKED");
					break;

				case BGAPI_PIXTYPE_MONO16:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_MONO16");
					break;

				case BGAPI_PIXTYPE_BAYGR8:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYGR8");
					break;

				case BGAPI_PIXTYPE_BAYRG8:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYRG8");
					break;

				case BGAPI_PIXTYPE_BAYGB8:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYGB8");
					break;

				case BGAPI_PIXTYPE_BAYBG8:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYBG8");
					break;

				case BGAPI_PIXTYPE_BAYGR10:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYGR10");
					break;

				case BGAPI_PIXTYPE_BAYRG10:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYRG10");
					break;

				case BGAPI_PIXTYPE_BAYGB10:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYGB10");
					break;

				case BGAPI_PIXTYPE_BAYBG10:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYBG10");
					break;

				case BGAPI_PIXTYPE_BAYGR12:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYGR12");
					break;

				case BGAPI_PIXTYPE_BAYRG12:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYRG12");
					break;

				case BGAPI_PIXTYPE_BAYGB12:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYGB12");
					break;

				case BGAPI_PIXTYPE_BAYBG12:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BAYBG12");
					break;

				case BGAPI_PIXTYPE_RGB8_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB8_PACKED");
					break;

				case BGAPI_PIXTYPE_BGR8_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BGR8_PACKED");
					break;

				case BGAPI_PIXTYPE_RGBA8_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGBA8_PACKED");
					break;

				case BGAPI_PIXTYPE_BGRA8_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BGRA8_PACKED");
					break;

				case BGAPI_PIXTYPE_RGB10_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB10_PACKED");
					break;

				case BGAPI_PIXTYPE_BGR10_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BGR10_PACKED");
					break;

				case BGAPI_PIXTYPE_RGB12_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB12_PACKED");
					break;

				case BGAPI_PIXTYPE_BGR12_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BGR12_PACKED");
					break;

				case BGAPI_PIXTYPE_RGB10V1_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB10V1_PACKED");
					break;

				case BGAPI_PIXTYPE_RGB10V2_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_BGR10V2_PACKED");
					break;

				case BGAPI_PIXTYPE_YUV411_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_YUV411_PACKED");
					break;

				case BGAPI_PIXTYPE_YUV422_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_YUV422_PACKED");
					break;

				case BGAPI_PIXTYPE_YUV444_PACKED:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_YUV444_PACKED");
					break;

				case BGAPI_PIXTYPE_RGB8_PLANAR:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB8_PLANAR");
					break;

				case BGAPI_PIXTYPE_RGB10_PLANAR:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB10_PLANAR");
					break;

				case BGAPI_PIXTYPE_RGB12_PLANAR:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB12_PLANAR");
					break;

				case BGAPI_PIXTYPE_RGB16_PLANAR:
					sprintf( (char*)destpixel_str, "BGAPI_PIXTYPE_RGB16_PLANAR");
					break;

				default:
					sprintf(destpixel_str, "ERROR");
			}
			BGAPIX_CameraPixelFormat format;
			format.cbSize = sizeof ( BGAPIX_CameraPixelFormat );
			strcpy(format.sName, destpixel_str);
			format.iPixelFormat = destpixellist.array[p];
			DestinationPixelFormats.push_back( format );
			//printf("          %cDestinationPixelFormat: %d %s\n", p==destpixellist.current?'*': ' ',format.iPixelFormat, (char*)format.sName);
		}

	}
	return DestinationPixelFormats;
}


int BaumerImageSequence::getCurrentDestinationPixelFormatIndex( bool force_check )
{
	if (!pSystem || !pCamera ) throw std::runtime_error("pSystem or pCamera or is not initialized!");
	if ( force_check || DestinationPixelFormats.empty() )
		getDestinationPixelFormats( force_check );
	
	return DestinationPixelFormatIndex;
}	

BGAPIX_CameraPixelFormat const& BaumerImageSequence::getCurrentDestinationPixelFormat( bool force_check )
{
	if (!pSystem || !pCamera ) throw std::runtime_error("pSystem or pCamera or is not initialized!");
	if ( force_check || DestinationPixelFormats.empty() )
		return getDestinationPixelFormats( true )[DestinationPixelFormatIndex];
	else
		return getDestinationPixelFormats( false )[DestinationPixelFormatIndex];
}	


///************* static Baumer functions *************************************
BGAPI::System* BaumerImageSequence::initBGAPI(  )
{

  log().info << " init BGAPI " << g_active_users_count << std::endl;

	if ( !has_file("sys_gige.so.16" ) ||
	     !has_file("img_proc.so.16" ) ||
	     !has_file("libbgapi.so"    ) )
	     throw std::runtime_error("missing one of the following files: sys_gige.so.16 img_proc.so.16 libbgapi.so in the directory you invoked this program from (Note: It does not matter if the the binaries exist in the directory of the executable, they MUST be located in the folder you invoked the program from! (Thanks for this one, Baumer...))");
	
	
	///  if no enviroment is active: initialize
	if ( g_active_users_count == 0 )
	{
	  
	  log().info << " initializing Baumer api..." << std::endl;

		g_active_users_count++;
		BGAPI_RESULT res = BGAPI_RESULT_FAIL;
		
		/// set library path
		char* lib = new char[g_library_path.length()];
		strcpy( lib, g_library_path.data() );
        //BGAPI::setModuleSearchPath (lib, g_library_path.length());
        delete lib;
		
		int camera_count = 0;
		res = BGAPI::countSystems( &g_system_count );
		if (g_system_count > 1)
			throw std::runtime_error( str(boost::format( "BGAPI::countSystems has found %d systems, this case is not implemented!") % g_system_count ));
		if (g_system_count == 0)
			throw std::runtime_error( str(boost::format( "BGAPI::countSystems has found %d systems, cameras are not detected!") % g_system_count ));
		
        
		log().info << g_system_count << " systems..." << std::endl;

		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI_CountSystems Errorcode: %d system_count %d\n")% res % g_system_count) );
		
		res = BGAPI::createSystem( 0, &g_pSystem );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format("BGAPI::CreateSystem Errorcode: %d\n") % res   ));

		log().info << " created system..." << std::endl;
		
		res = g_pSystem->open();		
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::open  Errorcode: %d\n")  % res ));
		
		
		
		/*
		char baumerlogfile[256];
		sprintf( baumerlogfile, "logfile_baumer.txt" );
		log().info <<" opened system, setting logfile \""<< baumerlogfile << std::endl;
		
		
		BGAPI::TraceObject * pTrace = NULL;
		res = g_pSystem->createTraceObject( &pTrace );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::createTraceObject  Errorcode: %d\n")  % res ));
		
		res = pTrace->setTarget( BGAPI_TRACETARGET_FILE );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::setTarget  Errorcode: %d\n")  % res ));
		
		res = pTrace->setTraceFileName( baumerlogfile, 256);
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::setTraceFileName  Errorcode: %d\n")  % res ));
		
		res = pTrace->setOutputOptions( BGAPI_TRACEOUTPUTOPTION_NEWLINE | BGAPI_TRACEOUTPUTOPTION_TIMESTAMP | BGAPI_TRACEOUTPUTOPTION_TIMESTAMP_DIFF | BGAPI_TRACEOUTPUTOPTION_PREFIX | BGAPI_TRACEOUTPUTOPTION_TRACELEVEL );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::setOutputOptions  Errorcode: %d\n")  % res ));
		
		res = pTrace->setLevel( BGAPI_TRACELEVEL_ALL );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::setLevel  Errorcode: %d\n")  % res ));
	
		res = pTrace->setMask(  BGAPI_TRACEMASK_ERROR | BGAPI_TRACEMASK_WARNING | BGAPI_TRACEMASK_INFORMATION );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::setMask  Errorcode: %d\n")  % res ));
	
		res = pTrace->enableLogging( true );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::enableLogging  Errorcode: %d\n")  % res ));
	*/



		res = g_pSystem->countCameras( &camera_count );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::countCameras  Errorcode: %d\n")  % res ));
		
		log().info << " no. of cameras in system = "<< camera_count << std::endl;
		if ( camera_count == 0 )
            throw std::runtime_error( "initBGAPI: no cameras found!" );
		/// SYSTEM DEVICE INFORMATIONEN

		BGAPI_FeatureState state;
		state.cbSize = sizeof( BGAPI_FeatureState );
		BGAPIX_SystemInfo sysdeviceinfo;
		sysdeviceinfo.cbSize = sizeof( BGAPIX_SystemInfo );
		res = g_pSystem->getDeviceInformation( &state, &sysdeviceinfo );
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("SysDeviceInfo: ErrorCode: %d\n") % res));

		log().info << "system opened: " << std::endl;
		log().info << "   cbSize        = " << sysdeviceinfo.cbSize << std::endl;	
		log().info << "   vendorName    = " << sysdeviceinfo.vendorName << std::endl;
		log().info << "   vendor        = " << sysdeviceinfo.vendor << std::endl;
		log().info << "   serialNumber  = " << sysdeviceinfo.serialNumber << std::endl;
		log().info << "   versionStr    = " << sysdeviceinfo.versionStr << std::endl;
		log().info << "   payload_send  = " << sysdeviceinfo.payload_send << std::endl;
		log().info << "   payload_recv  = " << sysdeviceinfo.payload_recv << std::endl;
		log().info << "starting to create cameras" << std::endl;

		/// CREATING ALL CAMERAS
		for ( int i = 0; i < camera_count; i++ )
		{
			singletonCamera newCamera;
			newCamera.isFree =  true;
			newCamera.pCamera = NULL;
			res = g_pSystem->createCamera( i, &newCamera.pCamera);
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("Create camera object errorcode: %d camera index: %d\n") % res % i));

			
			/// CAMERA DEVICE INFORMATION
			BGAPIX_CameraInfo camdeviceinfo;
			camdeviceinfo.cbSize = sizeof( BGAPIX_CameraInfo );
			res = newCamera.pCamera->getDeviceInformation( &state, &camdeviceinfo);
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("BGAPI::Camera::getDeviceInformation Errorcode: %d Camera index: %d\n") % res % i));

			/// save modelName and Serial
			newCamera.modelName = (char*)camdeviceinfo.modelName;
			newCamera.serialNumber = (char*)camdeviceinfo.serialNumber;
			
			/*
			/// OPENING CREATED CAMERA
			res = newCamera.pCamera->open();
			if( res != BGAPI_RESULT_OK )
				if ( res == -36 )
				{
					log().info << str( boost::format("Camera->open: camera %d is already busy\n") % i) << std::endl;
					continue; /// if camera already busy
				}
				else
					throw std::runtime_error(str( boost::format("Camera->open errorcode: %d camera index: %d\n") % res % i));
			*/
			g_pCameras.push_back( newCamera );

			log().info << "camera " << i << " found..." << std::endl;

		}

		log().info << "done." << std::endl;
	}
	else
	{
		g_active_users_count++;
	}
	
	//createCamera( camera_name );
	
	return g_pSystem;
} 

void BaumerImageSequence::createCamera( std::string camera_name )
{
	BGAPI_RESULT res;
	BGAPI_FeatureState state;
	if ( camera_name == "" )
	{
		int camera_count = getGlobalCameraCount();
		if ( camera_count == 0 )
			throw std::runtime_error("createCamera(\"\"): no cameras found, cannot create");
		
		singletonCamera newCamera;
		newCamera.isFree =  true;
		newCamera.pCamera = NULL;
		res = g_pSystem->createCamera( 0, &newCamera.pCamera);
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("Create camera object errorcode: %d camera index: 0\n") % res));
		
		/// OPENING CREATED CAMERA
		res = newCamera.pCamera->open();
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("Camera->open errorcode: %d camera index: 0\n") % res ));
		g_pCameras.push_back( newCamera );
	}
	else
	{
		int camera_count = getGlobalCameraCount();
		int available_at = -1;
		for ( int i = 0; i < camera_count; i++ )
		{
			singletonCamera newCamera;
			newCamera.isFree =  true;
			newCamera.pCamera = NULL;
			res = g_pSystem->createCamera( i, &newCamera.pCamera);
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("Create camera object errorcode: %d camera index: %d\n") % res % i));

			/// CAMERA DEVICE INFORMATION
			BGAPIX_CameraInfo camdeviceinfo;
			camdeviceinfo.cbSize = sizeof( BGAPIX_CameraInfo );
			res = newCamera.pCamera->getDeviceInformation( &state, &camdeviceinfo);
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("BGAPI::createCamera getDeviceInformation Errorcode: %d Camera index: %d  name: %s\n") % res % i % camera_name));

			/// try to resolve camera_name as modelName or Serial
			std::string current_name( (char*)camdeviceinfo.modelName );
			
			if ( camera_name.compare( current_name ) == 0 )
			{
				available_at = i;
				break;
			}
			current_name = (char*)camdeviceinfo.serialNumber;
			if ( camera_name.compare( current_name ) == 0 )
			{
				available_at = i;
				break;
			}
		}

		if (available_at < 0)
		{	
			std::string possible_names("\"");
			for (int i=0; i<camera_count; i++)
			{
				singletonCamera newCamera;
				newCamera.isFree =  true;
				newCamera.pCamera = NULL;
				res = g_pSystem->createCamera( i, &newCamera.pCamera);
				if( res != BGAPI_RESULT_OK )
					throw std::runtime_error(str( boost::format("Create camera object errorcode: %d camera index: %d\n") % res % available_at));

				/// CAMERA DEVICE INFORMATION
				BGAPIX_CameraInfo camdeviceinfo;
				camdeviceinfo.cbSize = sizeof( BGAPIX_CameraInfo );
				res = newCamera.pCamera->getDeviceInformation( &state, &camdeviceinfo);
			
				possible_names.append( (char*)camdeviceinfo.modelName ); 
				possible_names.append( " | " );
				possible_names.append( (char*)camdeviceinfo.serialNumber ); 
				possible_names.append( "\",\"" );
			}
			possible_names.append("\"");
			throw std::runtime_error(str( boost::format("BGAPI::Camera::createCamera: could not find camera with name or serial \"%s\". possible names are:%s\n") % camera_name.c_str() %possible_names.c_str() ));
		}
		else
		{
			/// actually create camera at found index
			singletonCamera newCamera;
			newCamera.isFree =  true;
			newCamera.pCamera = NULL;
			res = g_pSystem->createCamera( available_at, &newCamera.pCamera);
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("Create camera object errorcode: %d camera index: %d\n") % res % available_at));
			
			/// OPENING CREATED CAMERA
			res = newCamera.pCamera->open();
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("Camera->open errorcode: %d camera index: %d\n") % res % available_at));
			g_pCameras.push_back( newCamera );

			log().info << "camera " << available_at << " created!" << std::endl;

		}
	}
}

void BaumerImageSequence::releaseBGAPI()
{
	if (!g_pSystem && g_active_users_count>0) throw std::runtime_error("something went wrong, no system available, but user registered!");
	
	if (g_pSystem && g_active_users_count==1 )
	{

		BOOST_FOREACH( singletonCamera & camera, g_pCameras )
		  g_pSystem->releaseCamera( camera.pCamera );
		g_pCameras.clear();

		///  if this is the last active user
		BGAPI_RESULT res = g_pSystem->release();
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error( str(boost::format( "BGAPI::System::release Errorcode: %d\n") % (int)res ));
		
		///  we trust that the api cleans its objects correctly
		g_pSystem = NULL;		
		g_active_users_count = 0;
	}
	else
	{
		g_active_users_count--;
	}

	//std::cout << " released BGAPI user count " << g_active_users_count << std::endl;
}




int BaumerImageSequence::getSystemCount() 
{ 
	if (g_pSystem)
		return g_system_count;
	else
		return -1; 
}

int BaumerImageSequence::getCameraCount() 
{ 
	if (!g_pSystem) throw std::runtime_error("getCameraCount: pSystem is not initialized!");
	return g_pCameras.size();
	 
}

int BaumerImageSequence::getGlobalCameraCount() 
{ 
	if (!g_pSystem) throw std::runtime_error("getCameraCount: pSystem is not initialized!");
	
	BGAPI_RESULT res;
	int camera_count = 0;
	res = g_pSystem->countCameras( &camera_count );
	if( res != BGAPI_RESULT_OK )
		throw std::runtime_error( str(boost::format( "BGAPI::System::countCameras  Errorcode: %d\n")  % res ));
	return camera_count;
	 
}

BGAPI::System* BaumerImageSequence::getSystem()
{
	return g_pSystem;
}



int BaumerImageSequence::getCameraByString( std::string name )
{
	if (!g_pSystem || g_pCameras.empty()) throw std::runtime_error("pSystem is not initialized!");
	
	BGAPI_RESULT res = BGAPI_RESULT_OK;
	int available_at = -1;
	
	for ( int i=0; i < g_pCameras.size(); i++ )
	{
		if (  name.compare( g_pCameras[i].modelName ) == 0 || name.compare( g_pCameras[i].serialNumber ) == 0 )
		{
			available_at = i;
			break;
		}
	}
	
	/*log().info << "getCameraByString: found camera \""<< name<<"\" at "<< available_at<<", sleeping a while..."<<std::endl;
	usleep(10000000 * (available_at+1));
	log().info << "getCameraByString: done sleeping."<<std::endl;
	*/
	  
	/*
	BGAPI_FeatureState state;
	state.cbSize = sizeof( BGAPI_FeatureState );
	BGAPIX_CameraInfo camdeviceinfo;
	camdeviceinfo.cbSize = sizeof( BGAPIX_CameraInfo );
	BGAPI_RESULT res = BGAPI_RESULT_OK;
	int available_at = -1;
	
	///  try to resolve string as ModelName
	for ( int i=0; i < g_pCameras.size(); i++ )
	{
		res = g_pCameras[i].pCamera->getDeviceInformation( &state, &camdeviceinfo);
		if( res != BGAPI_RESULT_OK )
			throw std::runtime_error(str( boost::format("BGAPI::Camera::getDeviceInformation Errorcode: %d Camera index: %d\n") % res % i));

		std::string current_name( (char*)camdeviceinfo.modelName );
		if ( name.compare( current_name ) == 0 )
		{
			available_at = i;
			break;
		}
	}

	if (available_at < 0)
	{
		/// if no suitable name could be found, try to resolve as SerialNumber
		for ( int i=0; i < g_pCameras.size(); i++ )
		{
			res = g_pCameras[i].pCamera->getDeviceInformation( &state, &camdeviceinfo);
			if( res != BGAPI_RESULT_OK )
				throw std::runtime_error(str( boost::format("BGAPI::Camera::getDeviceInformation Errorcode: %d Camera index: %d\n") % res % i));

			std::string current_name( (char*)camdeviceinfo.serialNumber );
			if ( name.compare( current_name ) == 0 )
			{
				available_at = i;
				break;
			}
		}
	*/
	
		if (available_at < 0)
		{ 
			BGAPI_FeatureState state;
			state.cbSize = sizeof( BGAPI_FeatureState );
			BGAPIX_CameraInfo camdeviceinfo;
			camdeviceinfo.cbSize = sizeof( BGAPIX_CameraInfo );
			std::string possible_names("\"");
			for (int i=0; i<g_pCameras.size(); i++)
			{
				res = g_pCameras[i].pCamera->getDeviceInformation( &state, &camdeviceinfo);
				possible_names.append( (char*)camdeviceinfo.modelName ); 
				possible_names.append( " | " );
				possible_names.append( (char*)camdeviceinfo.serialNumber ); 
				possible_names.append( "\",\"" );
			}
			possible_names.append("\"");
			throw std::runtime_error(str( boost::format("BGAPI::Camera::getCameraByString: could not find camera with name or serial \"%s\". possible names are:%s\n") % name.c_str() %possible_names.c_str() ));
		}
	
	return available_at;
}


/// *****************************************
