/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef CPUDATA_H
#define CPUDATA_H

#include <CPUDataIterators.hpp>

#include <set>
#include <algorithm>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>



#include <boost/preprocessor/enum_params.hpp>
#include <boost/preprocessor/iteration/local.hpp>

#include "CPUDataBlock.hpp"
#ifdef HAVE_MPI
  #include <mpi.h>
  #include "Header.hpp"
#endif

// TODO: support slicing properly (i.e. as a cpudata object)

template<typename T = float>
class CPUData : public CPUDataBlockReference<T>
{

public:

  typedef CPUDataBlock<T> DataBlockT;
  typedef CPUDataBlockReference<T> DataBlockReferenceT;
  typedef CPUDataViewIterator<T> iterator;
  typedef CPUDataViewIterator<T const> const_iterator;
  typedef CPUDataIterator<T> fast_iterator;
  typedef CPUDataIterator<T const> fast_const_iterator;
  typedef CPUDataFeatureColumn<T> ColumnT;
  typedef CPUDataFeatureColumn<const T> ConstColumnT;
  typedef CPUData<T> This;
  typedef std::vector<size_t> ColumnShapeT;
  typedef T DataT;

  CPUData()
  : _offset(0),
    _rowstride(0)
  {
    _shape[0] = _shape[1] = _shape[2] = 0;
  }

  CPUData( const CPUData & other, bool weak = false )
	: DataBlockReferenceT(0)
  {
	reference( other, weak );
  }

  template<typename T2>
  CPUData( const CPUData<T2>& other )
  {
    size_t n1 = other.rows(), n2 = other.columns(), n3 = other.depth();
    _offset = 0;
    _rowstride = n2 * n3;
    _shape[0] = n1;
    _shape[1] = n2;
    _shape[2] = n3;
    if ( n1 * n2 * n3 > 0 )
    {
      DataBlockReferenceT::reference( new DataBlockT( n1 * n2 * n3 ) );
      std::copy( other.begin(), other.end(), fast_begin() );
    }
  }

  // make a container referencing a rectangular sub-region in another container
  // r: start row
  // nr: number of rows in the region
  // c: start columns
  // nc: number of columns in the region
  CPUData( const CPUData& other, size_t r, size_t nr, size_t c, size_t nc, bool weak = false )
  {
    reference( other, r, nr, c, nc, weak );
  }

  CPUData(size_t n1, size_t n2, size_t n3)
    : DataBlockReferenceT( ( ( n1>0 ) && ( n2>0 ) && ( n3>0 ) ) ? ( new DataBlockT( n1 * n2 * n3 ) ) : 0 )
	, _offset(0)
	, _rowstride( n2 * n3 )
  {
    _shape[0] = n1;
    _shape[1] = n2;
    _shape[2] = n3;
  }

  CPUData( T * data, size_t n1, size_t n2, size_t n3, bool weak = true )
    : DataBlockReferenceT( ( (n1>0) && (n2>0) && (n3>0) ) ? ( new DataBlockT( data, n1 * n2 * n3, weak ) ) : 0  )
	, _offset(0)
	, _rowstride( n2 * n3 )
  {
    _shape[0] = n1;
	_shape[1] = n2;
	_shape[2] = n3;
  }

  template <class Keeper>
  CPUData( T * data, size_t n1, size_t n2, size_t n3, const Keeper & keeper )
    : DataBlockReferenceT( ( (n1>0) && (n2>0) && (n3>0) ) ? ( new DataBlockT( data, n1 * n2 * n3, keeper ) ) : 0  )
	, _offset(0)
	, _rowstride( n2 * n3 )
  {
    _shape[0] = n1;
	_shape[1] = n2;
	_shape[2] = n3;
  }

  CPUData( DataBlockT * block, size_t n1, size_t n2, size_t n3 )
    : DataBlockReferenceT( ( (n1>0) && (n2>0) && (n3>0) ) ? ( block ) : 0  )
	, _offset(0)
	, _rowstride( n2 * n3 )
  {
    _shape[0] = n1;
	_shape[1] = n2;
	_shape[2] = n3;
  }

  template <class Keeper>
  CPUData( T * data, size_t n1, size_t n2, size_t n3, ptrdiff_t rowstride, const Keeper & keeper )
    : DataBlockReferenceT( ( (n1>0) && (n2>0) && (n3>0) ) ? ( new DataBlockT( data, n1 * n2 * n3, keeper ) ) : 0  )
	, _offset(0)
	, _rowstride( rowstride )
  {
    _shape[0] = n1;
	_shape[1] = n2;
	_shape[2] = n3;
  }

  bool is_valid() const
  {
  	if ( empty() ) return false;
	size_t endoff = _offset + row_stride() * ( rows() - 1 ) + ( columns() - 1 ) * col_stride();
	return endoff < DataBlockReferenceT::size();
  }

  /// checks for nans and infs and gives iterator of first occurence, if none found, gives back end iterator
  const_iterator find_invalid_values() const
  {
    const_iterator iend = end(), iter = begin();
    
#ifndef WIN32
    for ( ; iter != iend; ++iter )
    {
      for (int i = 0; i< (*iter).size(); ++i )
      if (  isnan( (*iter)[i] ) || isinf( (*iter)[i] ) )
        return iter;

    }
#endif

	return iend;
  }
  
  bool has_invalid_values() const
  {
    return ( find_invalid_values() != end() );
  }

  // change the offset within the underlying memory block (hazardous low-level operation, use wisely!)
  void rebase( size_t offset )
  {
  	_offset = offset;
  }

  template< typename U >
  double manhattan_distance( const CPUData<U> & other ) const
  {
#ifdef DEBUG
	assert( rows() == other.rows() && columns() == other.columns() && depth() == other.depth() );
#endif
	double d = 0.0;
	const_iterator ithis = begin(), iend = end();
	typename CPUData<U>::const_iterator iother = other.begin();
	for( ; ithis != iend; ++ithis, ++iother )
	  d += ithis->manhattan_distance( *iother );
	return d;
  }

  void normalize()
  {
	BOOST_FOREACH( ColumnT & column, *this )
	  column.normalize();
  }

  // returns the offset within the underlying memory block
  size_t offset() const
  {
	return _offset;
  }

  // equivalent to reference(other)
  This & operator= ( const This & other )
  {
	reference( other );
    return *this;
  }

  // return a container referencing a rectangular subrange of this container
  // r: start row
  // nr: number of rows in the region
  // c: start columns
  // nc: number of columns in the region
  This view( size_t r, size_t nr, size_t c, size_t nc )
  {
	return This( *this, r, nr, c, nc );	
  }
  
  // make this data a full copy of other, new data is contiguous, old data is lost
  void copy_from(const CPUData & other)
  {
    resize( other.rows(), other.columns(), other.depth() );
    
    if ( contiguous() && other.contiguous() )
      memcpy( data(), other.data(), size() * sizeof(T) );
    else
      std::copy( other.begin(), other.end(), begin() );
      
    _column_shape = other.column_shape();
  }

  // create a full copy of this data, new data is contiguous
  This copy() const
  {
    This result;
    result.copy_from( *this );
    return result;
  }
  
  // make this container a reference to another container, looses data stored in this container
  void reference( const This& other, bool weak = false )
  {
	if ( this == &other ) return;
    DataBlockReferenceT::reference( other, weak );
	std::copy( other._shape, other._shape + 3, _shape );
	_rowstride = other._rowstride;
	_offset = other._offset;
	_column_shape = other.column_shape();
  }
  
  // make this container reference a rectangular sub-region in another container
  // r: start row
  // nr: number of rows in the region
  // c: start columns
  // nc: number of columns in the region
  void reference( const This& other, size_t r, size_t nr, size_t c, size_t nc, bool weak = false )
  {
#ifdef DEBUG
	  assert( r + nr <= other.rows() && c + nc <= other.columns() );
#endif
	  reference( other, weak );
	  _offset += r * row_stride() + c * col_stride();
	  _shape[0] = nr;
	  _shape[1] = nc;
  }

  /// lazily resizes this data, data is unique and contiguous afterwards in any case
  void resize(size_t n1, size_t n2, size_t n3)
  {
	if ( unique() && contiguous() && rows() == n1 && columns() == n2 && depth() == n3 )
	  return;
	reference( This( n1, n2, n3 ) );
  }
  void resize( const This & other )
  {
	resize( other.rows(), other.columns(), other.depth() );
  }


  void flip( bool horizontal, bool vertical )
  {
	if ( horizontal && vertical )
	  {
		for ( size_t i1 = 0, i2 = rows() - 1; i1 < rows() / 2; ++i1, --i2 )
		  for ( size_t j1 = 0, j2 = columns() - 1; j1 < columns(); ++j1, --j2 )
			{
			  ColumnT c1 = (*this)( i1, j1 );
			  ColumnT c2 = (*this)( i2, j2 );
			  c1.swap( c2 );
			}
	  }
	else if ( horizontal )
	  {
		for ( size_t i = 0; i < rows(); ++i )
		  for ( size_t j1 = 0, j2 = columns() - 1; j1 < columns() / 2; ++j1, --j2 )
			{
			  ColumnT c1 = (*this)( i, j1 );
			  ColumnT c2 = (*this)( i, j2 );
			  c1.swap( c2 );
			}
	  }
	else if ( vertical )
	  {
		for ( size_t i1 = 0, i2 = rows() - 1; i1 < rows() / 2; ++i1, --i2 )
		  for ( size_t j = 0; j < columns(); ++j )
			{
			  ColumnT c1 = (*this)( i1, j );
			  ColumnT c2 = (*this)( i2, j );
			  c1.swap( c2 );
			}
	  }

  }

  // release the data reference held by this container
  void clear()
  {
    resize( 0, 0, 0 );
  }

  // number of elements visible through this container
  size_t size() const
  {
	return rows() * columns() * depth();
  }
  
  bool empty() const
  {
    return depth() == 0 || columns() == 0 || rows() == 0;
  }
  
  T * data()
  {
    return DataBlockReferenceT::data() + offset();
  }

  const T * data() const
  {
    return DataBlockReferenceT::data() + offset();
  }
	
  T * data_end()
  {
    return empty() ? data() : data() + depth() * columns() + row_stride() * ( rows() - 1 );
  }

  const T * data_end() const
  {
    return empty() ? data() : data() + depth() * columns() + row_stride() * ( rows() - 1 );
  }
  
  std::pair<T*,T*> data_range()
  {
    return std::make_pair( data(), data_end() );
  }
  
  std::pair<const T*,const T*> data_range() const
  {
    return std::make_pair( data(), data_end() );
  }
 
 
  This& operator+=( const CPUData & other )
  {
#ifdef DEBUG
	assert( rows() == other.rows() && columns() == other.columns() && depth() == other.depth() );
#endif
	iterator ithis = begin(), iend = end();
	const_iterator iother = other.begin();
	for( ; ithis != iend; ++ithis, ++iother )
	  (*ithis) += (*iother);
	return *this;
  }
 
  This& operator*=( const CPUData & other )
  {
#ifdef DEBUG
	assert( rows() == other.rows() && columns() == other.columns() && depth() == other.depth() );
#endif
	iterator ithis = begin(), iend = end();
	const_iterator iother = other.begin();
	for( ; ithis != iend; ++ithis, ++iother )
	  (*ithis) *= (*iother);
	return *this;
  }

  This& operator+=( CPUData & other )
  {
#ifdef DEBUG
	assert( rows() == other.rows() && columns() == other.columns() && depth() == other.depth() );
#endif
	iterator ithis = begin(), iend = end();
	iterator iother = other.begin();
	for( ; ithis != iend; ++ithis, ++iother )
	  (*ithis) += (*iother);
	return *this;
  }
 
  This& operator*=( CPUData & other )
  {
#ifdef DEBUG
	assert( rows() == other.rows() && columns() == other.columns() && depth() == other.depth() );
#endif
	iterator ithis = begin(), iend = end();
	iterator iother = other.begin();
	for( ; ithis != iend; ++ithis, ++iother )
	  (*ithis) *= (*iother);
	return *this;
  }
  
  This& add_scaled( CPUData & other, T s )
  {
#ifdef DEBUG
	assert( rows() == other.rows() && columns() == other.columns() && depth() == other.depth() );
#endif
	iterator ithis = begin(), iend = end();
	iterator iother = other.begin();
	for( ; ithis != iend; ++ithis, ++iother )
	  ithis->add_scaled( *iother, s );
	return *this;
  }
  
  This& operator=( T value )
  {
	std::fill( begin(), end(), value );
	//iterator ithis = begin(), iend = end();
	//for( ; ithis != iend; ++ithis )
	//  (*ithis) = value;
	return *this;
  }

  This& operator+=( T value )
  {
	iterator ithis = begin(), iend = end();
	for( ; ithis != iend; ++ithis )
	  (*ithis) += value;
	return *this;
  }
 
  This& operator*=( T value )
  {
	iterator ithis = begin(), iend = end();
	for( ; ithis != iend; ++ithis )
	  (*ithis) *= value;
	return *this;
  }

  This& operator/=( T value )
  {
	iterator ithis = begin(), iend = end();
	for( ; ithis != iend; ++ithis )
	  (*ithis) /= value;
	return *this;
  }
 
  template<typename S>
  This& clamp( S low, S high )
  {
	iterator ithis = begin(), iend = end();
	for( ; ithis != iend; ++ithis )
	  (*ithis).clamp( low, high );
	return *this;
  }
  
  T max( ) const
  {
  	T maximum=std::numeric_limits<T>::min();
  	const_iterator ithis = begin(), iend = end();
	for( ; ithis != iend; ++ithis )
		maximum = std::max( maximum, (*ithis).max() );
	return maximum;
  }
  
  T min( ) const
  {
  	T minimum=std::numeric_limits<T>::max();
  	const_iterator ithis = begin(), iend = end();
	for( ; ithis != iend; ++ithis )
		minimum = std::min( minimum, (*ithis).min() );
	return minimum;
  }
  
  
  ColumnT operator()(size_t row, size_t column)
  {
#ifdef DEBUG
    assert( row < rows() && column < columns() );
#endif
    T * off = data() + depth() * column + row_stride() * row;
    return CPUDataFeatureColumn<T>( off, off + depth() );
  }

  ConstColumnT operator()( size_t row, size_t column ) const
  {
#ifdef DEBUG
    assert( row < rows() && column < columns() );
#endif
    const T * off = data() + depth() * column + row_stride() * row;
    return ConstColumnT( off, off + depth() );
  }
  
  T & operator()(size_t row, size_t column, size_t _depth)
  {
#ifdef DEBUG
    assert( row < rows() && column < columns() && _depth < depth() );
#endif
    size_t off = depth() * column + row_stride() * row + _depth;
	return *( data() + off );
  }
  
  const T & operator()(size_t row, size_t column, size_t _depth) const
  {
#ifdef DEBUG
    assert( row < rows() && column < columns() && _depth < depth() );
#endif
    size_t off = depth() * column + row_stride() * row + _depth;
	return *( data() + off );
  }
  
  size_t row_stride() const
  {
	return _rowstride;
  }
  

  size_t column_row_stride() const
  {
	return depth() > 0 ? row_stride() / depth() : 0;
  }

  iterator begin()
  {
    T * p0 = data();
    return iterator( p0, depth(), columns(), column_row_stride() );
  }
  const_iterator begin() const
  {
    const T * p0 = data();
    return const_iterator( p0, depth(), columns(), column_row_stride() );
  }
  iterator end()
  {
    T * p0 = data() + rows() * row_stride();
    return iterator( p0, depth(), columns(), column_row_stride() );
  }
  const_iterator end() const
  {
    const T * p0 = data() + rows() * row_stride();
    return const_iterator( p0, depth(), columns(), column_row_stride() );
  }      
  
  ///returns an iterator that may be much faster, but only on contiguous data
  fast_iterator fast_begin()
  {
#ifdef DEBUG
	assert( contiguous() );
#endif
    T * p0 = data();
    return fast_iterator( p0, depth() );
  }
  
  ///returns an iterator that may be much faster, but only on contiguous data
  fast_iterator fast_end()
  {
#ifdef DEBUG
	assert( contiguous() );
#endif
    T * p0 = data() + rows() * row_stride();
    return fast_iterator( p0, depth() );
  }
  
  ///returns an iterator that may be much faster, but only on contiguous data
  fast_const_iterator fast_begin() const
  {
#ifdef DEBUG
	assert( contiguous() );
#endif
    T * p0 = data();
    return fast_const_iterator( p0, depth() );
  }
  
  ///returns an iterator that may be much faster, but only on contiguous data
  fast_const_iterator fast_end() const
  {
#ifdef DEBUG
	assert( contiguous() );
#endif
    T * p0 = data() + rows() * row_stride();
    return fast_const_iterator( p0, depth() );
  }
  
  std::pair<iterator,iterator> range()
  {
    return std::make_pair( begin(), end() );
  }
  
  std::pair<const_iterator,const_iterator> range() const
  {
    return std::make_pair( begin(), end() );
  }
  
  std::pair<iterator,iterator> fast_range()
  {
    return std::make_pair( fast_begin(), fast_end() );
  }
  
  std::pair<const_iterator,const_iterator> fast_range() const
  {
    return std::make_pair( fast_begin(), fast_end() );
  }
  
  size_t rows() const {return _shape[0];}
  size_t columns() const {return _shape[1];}
  size_t depth() const {return _shape[2];}
  size_t col_stride() const {return depth();}
 
  /// returns a formatted string which shows the three dimensions
  std::string dump_dims() const
  {
	std::string result("[ ");
	result.append(boost::lexical_cast<std::string>(_shape[0])+" x " );
	result.append(boost::lexical_cast<std::string>(_shape[1])+" x " );
	result.append(boost::lexical_cast<std::string>(_shape[2])+" ]" );
	if ( _column_shape.size() > 0 )
	  {
		result.append("(");
		result.append(boost::lexical_cast<std::string>(_shape[0])+" x " );
		result.append(boost::lexical_cast<std::string>(_shape[1])+" x " );
		for ( size_t i = 0; i < _column_shape.size(); ++i )
		  {
			result.append(boost::lexical_cast<std::string>(_column_shape[i]));
			if ( i < _column_shape.size() - 1 )
			  result.append(" x ");
		  }
		result.append(")");
	  }
  result.append("r=");
  result.append(boost::lexical_cast<std::string>( DataBlockReferenceT::refcount() ));
	return result;	
  }

  /// creates a histogram of size bins x 1 x depth
  This make_histogram( size_t bins, size_t skip=0, bool is_normalized = true ) const
  {
    
    This out( bins, 1, depth() );
    if ( empty() ) 
    {
        out.clear();
        return out;
    }
    out = T(0);
    size_t bin = 0;
    T minimum, maximum;
    if ( is_normalized )
    {
      maximum = 1;      /// expect values in [0,1]
      minimum = 0;
    }
    else
    {
      minimum = min();  /// make histogram relative to observed data
      maximum = max();  /// make histogram relative to observed data
    }
    
    if ( maximum == minimum ) 
    {
    	out = 0;
    	out( 0, 0 ) = 1;
    	return out;
    }
    
    for ( size_t i = 0; i < rows(); i += skip + 1 )
    for ( size_t j = 0; j < columns(); j += skip + 1 )
    for ( size_t k = 0; k < depth(); ++k )
    {
      bin = size_t((bins-1) * ( ( (*this)( i, j, k ) - minimum ) / ( maximum - minimum ) ) + float(0.5));
      bin = std::min( bins - 1, bin );
	  ++out( bin, 0, k );
    }
    return out;
  }

  /// nearest neighbpour upsampling by scale factor > 1
  void upsample( size_t scale_factor )
  {
    if ( scale_factor == 0 ) throw std::runtime_error("CPUData::upsample: scale_factor mustn't be 0!");
    if ( scale_factor == 1 ) return;
    This result;
    result.resize( rows() * scale_factor, columns() * scale_factor, depth() );
    
    result = 0;
    
    for ( int i = 0; i < rows(); ++i )
    for ( int j = 0; j < columns(); ++j )
    {
      ColumnT col = operator()( i, j );
      for ( int s1 = 0; s1 < scale_factor; ++s1 )
      for ( int s2 = 0; s2 < scale_factor; ++s2 )
      {
        result( scale_factor * i + s1, scale_factor * j + s2 ) = col;
      }
    }
    reference( result );
  }

  /// tolerant nearest neighbour upsampling with padding
  void upsample ( size_t target_rows, size_t target_cols, T init_val = T(0) )
  {
    if ( target_rows/float(rows()) < 1 || target_cols/float(columns()) < 1 ) throw std::runtime_error( str( boost::format("upsample: cannot upsample %s to size [ %d, %d ]")%dump_dims() %target_rows %target_cols ) );
    This result;
    result.resize( target_rows, target_cols, depth() );
    result = init_val;
    size_t even_factor_r = target_rows / rows();
    size_t even_factor_c = target_cols / columns();
    size_t off_r = ( target_rows - rows()    * even_factor_r ) / 2;
    size_t off_c = ( target_cols - columns() * even_factor_c ) / 2;
    
    //std::cout << "UPSCALE ( "<< target_rows << " , "<< target_cols<<" ): going from "<< dump_dims() << " to "<< result.dump_dims() << " with offset ("<<off_r<<","<<off_c<<") and even factor ("<<even_factor_r<<","<<even_factor_c<<")"<<std::endl;
    
    for ( int i = 0; i < rows(); ++i )
    for ( int j = 0; j < columns(); ++j )
    {
      ColumnT col = operator()( i, j );
      for ( int s1 = 0; s1 < even_factor_r; ++s1 )
      for ( int s2 = 0; s2 < even_factor_c; ++s2 )
      {
        result( even_factor_r * i + s1 + off_r, even_factor_c * j + s2 + off_c ) = col;
      }
    }
    reference( result );
  }
  
  /// tolerant nearest neighbour upsampling with fractional row/column duplication, at least to given sizes
  void upsample_atleast ( size_t target_rows, size_t target_cols )
  {
    target_cols=std::max(target_cols,columns());
    target_rows=std::max(target_rows,rows());
    
    if ( target_rows/float(rows()) < 1 || target_cols/float(columns()) < 1 ) throw std::runtime_error( str( boost::format("upsample: cannot upsample %s to size [ %d, %d ]")%dump_dims() %target_rows %target_cols ) );
    This result;
    result.resize( target_rows, target_cols, depth() );
    result = 0;
    
    const float rowfac=1.0/target_rows;
    const float colfac=1.0/target_cols;
    
    
    for( unsigned i = 0; i < result.rows(); ++i)
    for( unsigned j = 0; j < result.columns(); ++j)
    {
      const int srow=floor(i*rowfac*rows());
      const int scol=floor(j*colfac*columns());
      
      ColumnT col=operator()(srow,scol);
      result(i,j)=col;
      
    }
    
    reference( result );
  }
  
  

  /// false whenever the data referenced by this container is also referenced by other containers
  bool unique() const
  {
	return DataBlockReferenceT::unique();
  }
  
  /// if this container is not unique make it reference a unique copy of the data, else make it contiguous, so that unnessesary data can be dumped
  void make_unique()
  {
	if ( !unique() )
	  reference( copy() );
  else
    make_contiguous();
  }

  /// if this container is not contiguous make it reference a contigouous copy of the data
  void make_contiguous()
  {
	if ( !contiguous() )
	  reference( copy() );
  }

  /// whether this data is organized contiguously in memory
  /// or there are gaps between rows
  bool contiguous() const
  {
	return row_stride() == columns() * depth();
  }

  ~CPUData()
  {
  }
		
  const size_t* shape() const
  {
	return _shape;
  }
  
  const ColumnShapeT & column_shape() const
  {
	return _column_shape;
  }

  void set_column_shape( const ColumnShapeT & sizes )
  {	
	_column_shape = sizes;
  }
	
  T sum() const
  {
    T result = T(0);
    BOOST_FOREACH( ConstColumnT & column, *this )
      result += column.sum();
    return result;
  }

  void find_max( size_t & row, size_t & col, size_t & depth )
  {
    size_t pos = 0;
    T value = std::numeric_limits<T>::min();
    assert(contiguous());
    for ( T * i = data(); i != data_end(); ++i )
      if ( *i >= value )
      {
        value = *i;
        pos = i - data();
      }
    row = pos / row_stride();
    col = pos - ( row_stride() * row );
    depth = pos - ( row_stride() * row ) - col * this->depth();
  }

  T kl( CPUData & other )
  {
	T result = T(0);
	for( iterator i = begin(), iend = end(), j = other.begin(); i != iend; ++i, ++j )
	  result += i->kl( *j );
	return result;
  }

  T mean() const
  {
	return sum() / size();
  }

  /// generates set_column_shape( n1, n2, ... , nD ) for up to six dimensions
#define BOOST_PP_LOCAL_LIMITS (1,6)
#define BOOST_PP_LOCAL_MACRO(I)									\
  void set_column_shape( BOOST_PP_ENUM_PARAMS(I, size_t n) )	\
  {																\
    _column_shape.resize(I);									\
    size_t N[I] = { BOOST_PP_ENUM_PARAMS(I,n) };				\
	std::copy( N, N + I, _column_shape );						\
  }
#include BOOST_PP_LOCAL_ITERATE()

#ifdef MPI_INCLUDED
  CPUData( const Header& src )
  : _offset(0),
    _rowstride(0)
  {
    _shape[0] = _shape[1] = _shape[2] = 0;
    resize( src );
  }

  Header get_header( bool is_io_msg = false ) const
  {
    Header result;
    std::copy( _shape, _shape+3, result.shape() );
    result.set_column_shape( _column_shape );
    result.set_data_type( Header::CPUDATA_HEADER );
    if ( is_io_msg )
      result.set_msg_type( Header::IO_MSG );
    //std::cout << "CPUData::get_header:   original="<< dump_dims() << "  header="<< result.dump_dims() << std::endl;
    return result;
  }

  void resize( const Header& src )
  {
    if ( src.get_data_type() != Header::CPUDATA_HEADER )
      throw std::runtime_error( str( boost::format("Error in CPUData::resize(Header): header is not of type \"CPUDATA_HEADER\", but \"%s\"")%src.get_data_type_str() ) );
    
    resize( src.shape()[0], src.shape()[1], src.shape()[2] );
    set_column_shape( src.get_column_shape_as_vector() );
  }
#endif

private:
	
  bool check_column_shape( const ColumnShapeT& shape ) const
  {
    if ( shape.empty() ) return true;
    bool is_valid = true;
    size_t product = 1.f;
    for (int i = 0; i < shape.size(); i++)
      product *= shape[i];
      
    return (product==_shape[2]);
  }
  
  size_t _shape[3];
  size_t _offset;
  size_t _rowstride;
  ColumnShapeT _column_shape;
		
};


template< class ostream, class T > ostream & operator<<( ostream & os, const CPUData<T> d )
{
  os << "CPUData<" << typeid(T).name() << ">(" << d.rows() << "x" << d.columns() << "x" << d.depth() << ")" << std::endl;
  for ( size_t k = 0; k < d.depth(); ++k )
	{
	  os << "feature " << k << ":" << std::endl;
	  for ( size_t i = 0; i < d.rows(); ++i )
		{
		  for ( size_t j = 0; j < d.columns(); ++j )
		  {
			os << " " << ( boost::format("%5d") % d( i, j, k ) );
		  }
		  os << std::endl;
		}
	}
  return os;
}


///a 2d view into a larger array
///this is a weak reference, the user is responsible
///for making sure the parent array stays around
///as long as the view is used!
///FIXME: currently this a strong ref, making it potentially slow in loops
template<typename T>
class CPUDataLayer
{
public:
  
  typedef CPUData<T> CPUDataT;
  typedef CPUDataViewLayerIterator<T> iterator;
  typedef CPUDataViewLayerIterator<T const> const_iterator;

  CPUDataLayer( CPUDataT & parent, int layer ) : _data(parent), _layer(layer) {}
  
  CPUDataLayer( int n1 = 0, int n2 = 0 ) : _data( n1, n2, 1 ), _layer(0) {}

  iterator begin() 
  {
	T * p0 = data();
    return iterator( p0, columns(), cpu_data().row_stride() / cpu_data().depth(), cpu_data().depth() );
  }

  iterator end() 
  {
	T * p0 = data() + rows() * cpu_data().row_stride();
    return iterator( p0, columns(), cpu_data().row_stride() / cpu_data().depth(), cpu_data().depth() );
  }

  const_iterator begin() const
  {
	const T * p0 = data();
    return const_iterator( p0, columns(), cpu_data().row_stride() / cpu_data().depth(), cpu_data().depth() );
  }

  const_iterator end() const
  {
	const T * p0 = data() + rows() * cpu_data().row_stride();
    return const_iterator( p0, columns(), cpu_data().row_stride() / cpu_data().depth(), cpu_data().depth() );
  }

  std::pair<iterator,iterator> range()
  {
    return std::make_pair(begin(),end());
  }
		
  T & operator()(int i, int j)
  {
    return _data(i,j,_layer);
  }

  T operator()(int i, int j) const
  {
    return _data(i,j,_layer);
  }

  bool empty() const
  {
    return rows() == 0 || columns() == 0;
  }

  void clear()
  {
    _data.clear();
    _layer = 0;
  }

  T min( ) const
  {
    T minimum=std::numeric_limits<T>::max();
    const_iterator ithis = begin(), iend = end();
    for( ; ithis != iend; ++ithis )
      minimum = std::min( minimum, (*ithis) );
    return minimum;
  }
  
  T max( ) const
  {
    T maximum=std::numeric_limits<T>::min();
    const_iterator ithis = begin(), iend = end();
    for( ; ithis != iend; ++ithis )
      maximum = std::max( maximum, (*ithis) );
    return maximum;
  }

  T sum()
  {
    T result = T(0);
    const_iterator ithis = begin(), iend = end();
    for( ; ithis != iend; ++ithis )
      result += (*ithis);
    return result;
  }

  ///resize storage, warning: makes data unique (i.e. if it was a view before now it is independent
  void resize( int n1, int n2 )
  {
    _data.resize( n1, n2, 1 );
    _layer = 0;
  }

  // make this container reference a rectangular subregion in another container
  // r: start row
  // nr: number of rows in the region
  // c: start columns
  // nc: number of columns in the region
  void reference( const CPUDataLayer & other, size_t r, size_t nr, size_t c, size_t nc )
  {
    _data.reference( other.cpu_data(), r, nr, c, nc );
    _layer = other.layer_index();
  }

  CPUDataT & cpu_data() { return _data; }
  const CPUDataT & cpu_data() const { return _data; }
  T * data() { return _data.data() + _layer; }
  const T * data() const { return _data.data() + _layer; }

  size_t size() const { return columns() * rows(); }
  size_t rows() const {return _data.rows();}
  size_t columns() const {return _data.columns();}
  int layer_index() const {return _layer;}

private:
  
  CPUDataT _data;
  int _layer;

};

#endif
  
