/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _CPUDATABLOCK_HPP_
#define _CPUDATABLOCK_HPP_

#include <boost/any.hpp>
#include <iostream>

#ifdef HAVE_AO
//having this defined will make refcount atomic and thus threadsafe
//does _not_ provide data locking, just makes sure handling the same mamory block
//in multiple threads will not cause double frees or lost memory
#include <atomic_ops.h>
#endif

//just a slab of memory with known elementsize and blocksize
//and a refcounter (not managed by this class)
template<typename T>
class CPUDataBlock
{
  
public:
  
  typedef T DataType;
  
  CPUDataBlock()
	: _data(0), _refcount(0), _n(0), _foreign(false)
  {
  }

  explicit CPUDataBlock( size_t n )
	: _data(0), _refcount(0), _n(0), _foreign(false)
  {
	allocate( n );
  }

  ///create from existing data, optionally own it
  CPUDataBlock( size_t n, T * data, bool weak = true )
	: _data(data), _refcount(0), _n(n), _foreign(weak)
  {
	if ( !weak )
	  keep_array( _data );
  }

  ///create from existing data with separate keeper, don't own this
  template<class Keeper>
  CPUDataBlock( T * data, size_t n, const Keeper & keeper, bool weak = true )
	: _keeper(keeper), _data(data), _refcount(0), _n(n), _foreign(true)
  {
  }

  ~CPUDataBlock()
  {
  }
  
  size_t legth() const { return size(); }

  size_t size() const
  {
    return _n;
  }
  
  T * data()
  {
    return _data;
  }

  const T * data() const
  {
    return _data;
  }
  
  int refcount() const
  {
#ifdef HAVE_AO
    return AO_load_full(&_refcount);
#else
    return _refcount;
#endif
  }
  
  void addReference()
  {
	increase_refcount();
  }

  int removeReference()
  {
	return decrease_refcount();
  }

  int increase_refcount()
  {
#ifdef HAVE_AO
    return AO_fetch_and_add1_full(&_refcount) + 1;
#else
    return ++_refcount;
#endif
  }
  
  int decrease_refcount()
  {
#ifdef HAVE_AO
    return AO_fetch_and_sub1_full(&_refcount) - 1;
#else
    return --_refcount;
#endif
  }

  int references() const
  {
	return refcount();
  }

  bool foreign() { return _foreign; }
  
protected:

  inline void allocate( size_t n )
  {
    if (n > 0)
	  {
		keep_array( new DataType[n] );
	  }
    _n = n;
   }

private:

  void keep_array( DataType * data_ )
  {
	_data = data_;
	if ( _data )
	  {
		_keeper = array_deleter();
		boost::any_cast<array_deleter>( &_keeper )->set_data( _data );
	  }
	else
	  _keeper = boost::any();
  }

  struct array_deleter
  {
	DataType * _data;
	array_deleter( ) : _data(0) {}
	void set_data( DataType * data ) { _data = data; }
	~array_deleter() { if ( _data ) { delete[] _data; } }
  };

  boost::any _keeper;
  DataType * _data;
  volatile size_t _refcount;
  size_t _n;
  bool _foreign;
  
};

///a reference to a datablock
///each strong reference adds one to the refcount
///when a strong reference dies it takes one from the refcount
///and when the refcount becomes zero the block is deleted
template<typename T> 
class CPUDataBlockReference
{
public:
  typedef T DataType;
  typedef CPUDataBlock<T> DataBlockT;
  
  CPUDataBlockReference( DataBlockT * block = 0 )
    : _block(0), _weak(false)
  {
    construct(block);
  }
  
  CPUDataBlockReference( const CPUDataBlockReference & other )
    : _block(0), _weak(false)
  {
    reference(other);
  }
		
  void reference( const CPUDataBlockReference & other, bool weak = false )
  {
    release();
	construct( other._block, weak || other.weak() );
  }
		
  bool weak() const
  {
	return _weak;
  }

  ~CPUDataBlockReference()
  {
    release();
  }
		
  T * data()
  {
    if (_block == 0) 
      return 0;
    else
      return _block->data();
  }

  const T * data() const
  {
    if (_block == 0) 
      return 0;
    else
      return _block->data();
  }
  
  void release()
  {
    if ( _block && !_weak )
	  if ( _block->decrease_refcount() == 0 )
		delete _block;
    _block = 0;
  }
		
  size_t size() const
  {
    return _block ? _block->size() : 0;
  }
  
  size_t refcount() const
  {
    return _block ? _block->refcount() : 0;
  }

  bool unique() const
  {
	return _block == NULL || ( !_weak && _block->refcount() <= 1 && !_block->foreign() );
  }

  DataBlockT * block() { return _block; }

  const DataBlockT * block() const { return _block; }
		
protected:

  void construct( DataBlockT * block, bool weak = false )
  {
    _block = block;
    if ( _block )
	  {
		_weak = weak;
		if ( !_weak )		  
		  _block->increase_refcount();
	  }
  }
		
  DataBlockT * _block;
  bool _weak;
};


#endif /* _CPUDATABLOCK_HPP_ */


