/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef CPUDATAITERATORS_H
#define CPUDATAITERATORS_H

#include <boost/iterator/iterator_facade.hpp>
#include <boost/iterator/zip_iterator.hpp>
#include <boost/range/functions.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/foreach.hpp>
#include <boost/type_traits.hpp>
#include <boost/utility/enable_if.hpp>
#include <boost/mpl/not.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/type_traits.hpp>
#include "strided_iterator.hpp"

#include <utility>
#include <cassert>
#include <cmath>
#include <stdexcept>
#include <limits>

template<typename Index, typename Value>
class
CPUDataHypothesis
{
  typedef CPUDataHypothesis<Index,Value> This;
  Index index;
  Value value;
  CPUDataHypothesis(Index i = 0, Value v = 0)
	: index(i), value(v) {}
  bool operator<( const This & other )
  {
	return value < other.value;
  }
};

template<class Value> class CPUDataIterator;
template<class Value> class CPUDataViewIterator;

//feature columns are assumed to be packed dense in memory
//in this case this class is effectively an iterator
//and the facade class below is just that: a facade to 
//provide all the bells and whistles of a real iterator
//this is also an iterator range, and probably also a container
//that means you can do: 
// foreach(CPUDataFeatureColumn & column, data)
//		foreach(Value & value, column)
//			do_something(value);
//or:
// foreach(CPUDataFeatureColumn & column, data)
//		for(int i = 0; i < data.depth(); ++i)
//			do_something(column[i]);
//or any iterator base algorithms

#ifdef HAVE_SSE
#include "sse_templates.hpp"
#endif

#include <iostream>

template<class Value> class CPUDataFeatureColumn
{
public:
  CPUDataFeatureColumn()
  : _begin(NULL), _end(NULL)
  {}

  CPUDataFeatureColumn(Value * b, Value * e)
	: _begin(b), _end(e)
  {}
  
  typedef Value * iterator;
  typedef Value const * const_iterator;
  typedef CPUDataFeatureColumn<Value> This;
  typedef typename boost::add_const<Value>::type ConstValue;
  typedef typename boost::remove_const<Value>::type NonConstValue;
  typedef CPUDataFeatureColumn< ConstValue > ConstThis;
  typedef CPUDataFeatureColumn< NonConstValue  > NonConstThis;
  
  This & operator= ( Value v )
  {
    std::fill( begin(), end(), v );
	return *this;
  }

  //FIXME: for sparse notation these need to operate on the values only, also need to distinguish between scalars and index/value pairs

  template<typename S>
  This & add_scaled( const ConstThis & other, S s )
  {
	size_t i = 0, n = size();
#ifdef HAVE_SSE
	for ( size_t i4 = 0, n4 = n / 4; i4 < n4; ++i4, i += 4 )
	  ( vecf4( data() + i ) + vecf4( other.data() + i ) * s ).store( data() + i );
#endif
	for ( ; i < n; ++i )
	  (*this)[i] += other[i] * s;
	return *this;
  }
  
  This & add( const ConstThis & other )
  {
	size_t i = 0, n = size();
#ifdef HAVE_SSE
	for ( size_t i4 = 0, n4 = n / 4; i4 < n4; ++i4, i += 4 )
	  ( vecf4( data() + i ) + vecf4( other.data() + i ) ).store( data() + i );
#endif
	for ( ; i < n; ++i )
	  (*this)[i] += other[i];
	return *this;
  }

  This & add( const NonConstThis & other )
  {
	size_t i = 0, n = size();
#ifdef HAVE_SSE
	for ( size_t i4 = 0, n4 = n / 4; i4 < n4; ++i4, i += 4 )
	  ( vecf4( data() + i ) + vecf4( other.data() + i ) ).store( data() + i );
#endif
	for ( ; i < n; ++i )
	  (*this)[i] += other[i];
	  return *this;
  }

  This & sub( const ConstThis & other )
  {
	size_t i = 0, n = size();
#ifdef HAVE_SSE
	for ( size_t i4 = 0, n4 = n / 4; i4 < n4; ++i4, i += 4 )
	  ( vecf4( data() + i ) - vecf4( other.data() + i ) ).store( data() + i );
#endif
	for ( ; i < n; ++i )
	  (*this)[i] -= other[i];
	return *this;
  }

  This & sub( const NonConstThis & other )
  {
	size_t i = 0, n = size();
#ifdef HAVE_SSE
	for ( size_t i4 = 0, n4 = n / 4; i4 < n4; ++i4, i += 4 )
	  ( vecf4( data() + i ) - vecf4( other.data() + i ) ).store( data() + i );
#endif
	for ( ; i < n; ++i )
	  (*this)[i] -= other[i];
	  return *this;
  }

  This & swap( This & other )
  {
	for ( iterator i1 = begin(), i2 = other.begin(), iend = end(); i1 != iend; ++i1, ++i2 )
	  std::swap( *i1, *i2 );
	return *this;
  }

  Value kl( const NonConstThis & other )
  {
	Value result = Value(0);
	for ( size_t i = 0, n = size(); i < n; ++i )
	  {
		Value x = (*this)[i];
		Value y = other[i];
		if ( x > Value(0.00000000001) )
		  {
			result += x * ( std::log( x ) - std::log( y ) );
		  }
	  }
	return result;
  }

  This & operator+= ( const ConstThis & other )
  {
	  return add( other );
  }

  This & operator+= ( const NonConstThis & other )
  {
	  return add( other );
  }
  
  This & operator-= ( const ConstThis & other )
  {
	  return sub( other );
  }

  This & operator-= ( const NonConstThis & other )
  {
	  return sub( other );
  }
  
  template<typename S>
  This & add_scaled( This & other, S s )
  {
	for ( size_t i = 0, n = size(); i < n; ++i )
	  (*this)[i] += other[i] * s;
	return *this;
  }
  
  This & add( This & other )
  {
	for ( size_t i = 0, n = size(); i < n; ++i )
	  (*this)[i] += other[i];
	return *this;
  }

  This & operator+= ( This & other )
  {
	return add( other );
  }
  
  template<typename S>
  This & mult_scaled( const ConstThis & other, S s )
  {
	for ( size_t i = 0, n = size(); i < n; ++i )
	  (*this)[i] *= other[i] * s;
	return *this;
  }

  This & mult( const ConstThis & other )
  {
	for ( size_t i = 0, n = size(); i < n; ++i )
	  (*this)[i] *= other[i];
	return *this;
  }
  
  This & operator*= ( const ConstThis & other )
  {
	return mult( other );
  }
  
  template<typename S>
  This & mult_scaled( This & other, S s )
  {
	for ( size_t i = 0, n = size(); i < n; ++i )
	  (*this)[i] *= other[i] * s;
	return *this;
  }

  This & mult( This & other )
  {
	for ( size_t i = 0, n = size(); i < n; ++i )
	  (*this)[i] *= other[i];
	return *this;
  }
  
  This & operator*= ( This & other )
  {
	return mult( other );
  }
  
  
  template<typename S>
  This & clamp( S low, S high )
  {
	for ( size_t i = 0, n = size(); i < n; ++i )
	  (*this)[i] = std::min(  Value(high), std::max( Value(low), (*this)[i] )  );
	return *this;
  }
  
  This & operator*= ( Value v )
  {
	BOOST_FOREACH( Value & vr, *this )
	  vr *= v;
	return *this;
  }

  This & operator/= ( Value v )
  {
      return (*this) *= 1 / v;
  }

  This & operator+= ( Value v )
  {
	BOOST_FOREACH( Value & vr, *this )
	  vr += v;
	return *this;
  }

  This & operator-= ( Value v )
  {
	BOOST_FOREACH( Value & vr, *this )
	  vr -= v;
	return *this;
  }


  unsigned max_index() const
  {
	if ( empty() ) throw std::runtime_error("called max_index on empty column!");
	unsigned max_index = 0;
	Value max_value = (*this)[0];
	for ( size_t i = 1, n = size(); i < n; ++i )
	  {
		Value v = (*this)[i];
		if ( v > max_value )
		  {
			max_index = i;
			max_value = v;
		  }
	  }
	return max_index;
  }

  bool empty() const { return begin() == end(); }

  Value * data() { return begin(); }
  const Value * data() const { return begin(); }

  NonConstValue sum() const
  {
	NonConstValue result = 0;
	BOOST_FOREACH( Value v, *this )
	  result += v;
	return result;
  }



  NonConstValue max() const
  {
    NonConstValue result = std::numeric_limits<Value>::min();
    BOOST_FOREACH( Value v, *this )
      result = std::max( result, v );
    return result;
  }

  NonConstValue min() const
  {
    NonConstValue result = std::numeric_limits<Value>::max();
    BOOST_FOREACH( Value v, *this )
      result = std::min( result, v );
    return result;
  }


  NonConstValue entropy() const
  {
	Value Z = sum();
	if ( Z == 0 ) return std::log( Value( size() ) );
	else
	  {
		Value result = 0;
		BOOST_FOREACH( Value v, *this )
		  {
			Value p = v / Z;
			result -= p > 0 ? ( p * std::log(p) ) : 0;
		  }
		return result;
	  }
  }

  template<typename U>
  double
  manhattan_distance( const CPUDataFeatureColumn<U> & other ) const
  {
#ifdef DEBUG
	assert( other.size() == size() );
#endif
	double result = 0.0;
	const_iterator ithis = begin(), iend = end();
	typename CPUDataFeatureColumn<U>::const_iterator iother = other.begin();
	for ( ; ithis != iend; ++ithis, ++iother )
	  result += double( std::abs( *ithis - *iother ) );
	return result;
  }

  template<typename U>
  double
  euklidian_distance_squared( const CPUDataFeatureColumn<U> & other ) const
  {
#ifdef DEBUG
	assert( other.size() == size() );
#endif
	double result = 0.0;
	const_iterator ithis = begin(), iend = end();
	typename CPUDataFeatureColumn<U>::const_iterator iother = other.begin();
	for ( ; ithis != iend; ++ithis, ++iother )
	  {
		double tmp = *ithis - *iother;
		result += tmp * tmp;
	  }
	return result;
  }
  
  template<typename U>
  double
  maximum_distance( const CPUDataFeatureColumn<U> & other ) const
  {
#ifdef DEBUG
  assert( other.size() == size() );
#endif
  double maximum = 0.0;
  const_iterator ithis = begin(), iend = end();
  typename CPUDataFeatureColumn<U>::const_iterator iother = other.begin();
  for ( ; ithis != iend; ++ithis, ++iother )
    {
    maximum = std::max( double(std::abs(*ithis - *iother)) , maximum );
    }
  return maximum;
  }




  This & operator= ( const ConstThis & other )
  {
#ifdef DEBUG
	assert( other.size() == size() );
#endif	
	std::copy( other.begin(), other.end(), begin() );
	return *this;
  }

  This & operator= ( const NonConstThis & other )
  {
#ifdef DEBUG
	assert( other.size() == size() );
#endif	
	std::copy( other.begin(), other.end(), begin() );
	return *this;
  }

  template<typename U>
  This & operator=( const CPUDataFeatureColumn<U> & other )
  {
	copy_from( other );
	return *this;
  }

  template<typename U>
  This & copy_from ( const CPUDataFeatureColumn<U> & other )
  {
#ifdef DEBUG
	assert( other.size() == size() );
#endif	
	std::copy( other.begin(), other.end(), begin() );
	return *this;
  }

  Value & operator[](size_t i)
  {
	return *(_begin + i);
  }

  typename boost::remove_const<Value>::type 
  operator[](size_t i) const
  {
	return *(_begin + i);
  }
  
  Value * begin()
  {
	return _begin;
  }
  
  Value * end()
  {
	return _end;	
  }
  
  Value const * begin() const
  {
	return _begin;
  }
  
  Value const * end() const
  {
	return _end;	
  }
  
  size_t size() const
  {
	return end() - begin();
  }

  Value normalize()
  {
	Value s = sum();
	if ( s > 0 )
	  {
	  	(*this) *= Value(1) / s;
	  	return s;
	  }
	else
	  {
	  	(*this) = Value(1) / Value( size() );
	  	return size();
	  }

  }
	

  Value normalize_minmaxeps( Value eps )
  {
    Value my_min = min();
    Value my_max = max();
    Value diff = my_max - my_min;

    BOOST_FOREACH( Value& v, *this )
      v = (  v - my_min  )  /  (  eps + diff  );
      
    return (my_max-my_min)/(eps + diff);
  }


private:
	
  void increment()
  {
	do_advance(size());
  }
  
  void decrement()
  {
	do_advance(-size());
  }
  
  void advance( ptrdiff_t n )
  {
	do_advance(n * size());
  }
  
  void do_advance( ptrdiff_t delta )
  {
	_begin += delta;
	_end += delta;
  }
  
  friend class CPUDataIterator<Value>;
  friend class CPUDataViewIterator<Value>;
  
  Value * _begin, * _end;
  
};

template<class Value> class CPUDataIterator
  : public boost::iterator_facade<
        CPUDataIterator<Value>
      , CPUDataFeatureColumn<Value>
      , boost::random_access_traversal_tag
			, CPUDataFeatureColumn<Value> &
			, int
    >
{

public:
  
  typedef CPUDataIterator<Value> This;

  CPUDataIterator( Value * cbegin = 0, size_t depth = 0 ) 
	: _column( cbegin, cbegin + depth ) 
  {}
  
  explicit CPUDataIterator(const CPUDataFeatureColumn<Value> & c) 
	: _column(c.begin(),c.end()) 
  {}
  
  template <class OtherValue>
  CPUDataIterator(CPUDataIterator<OtherValue> const& other)
    : _column(other._column) 
  {}

  This & operator=(const This & other)
  {
	_column.begin = other._column.begin;
	_column.end = other._column.end;
	return *this;
  }

private:
  
  friend class boost::iterator_core_access;
  template <class> friend class CPUDataIterator;
  
  void increment() 
  { 
	_column.increment(); 
  }
  
  void decrement() 
  { 
	_column.decrement(); 
  }
  
  void advance(int n)
  {
	_column.advance(n);
  }
  
  int distance_to(CPUDataIterator<Value> const& other) const
  {
	return (other._column.begin() - _column.begin()) / _column.size();
  }

  template <class OtherValue>
  bool equal(CPUDataIterator<OtherValue> const& other) const
  {
    return this->_column.begin() == other._column.begin();
  }

  CPUDataFeatureColumn<Value>& dereference() const 
  { 
    return const_cast<CPUDataFeatureColumn<Value>&>(_column);
  }
  
  CPUDataFeatureColumn<Value>& dereference() { return _column; }
	
  CPUDataFeatureColumn<Value> _column;
};

//FIXME: this iterator is not among the fastest, could be much faster
template<class Value> class CPUDataViewIterator
  : public boost::iterator_facade<   CPUDataViewIterator<Value>
								   , CPUDataFeatureColumn<Value>
								   , boost::bidirectional_traversal_tag
								   , CPUDataFeatureColumn<Value> &
								   , int
								   >
{

public:

  typedef CPUDataViewIterator<Value> This;

  CPUDataViewIterator(Value * cbegin = 0, size_t depth = 0, int ni = 0, int di = 0)
	: _column( cbegin, cbegin + depth ), _i(0), _j(0), _ni(ni), _di(di)
	{
	}
	
	template <class OtherValue>
  CPUDataViewIterator(CPUDataIterator<OtherValue> const& other)
	: _column(other._column), _i(other._i), _j(other._j), _ni(other._ni), _di(other._di)
	{}

  This & operator=(const This & other)
  {
	_column._begin = other._column._begin;
	_column._end = other._column._end;
	_i = other._i;
	_ni = other._ni;
	_di = other._di;
	return *this;
  }

  size_t column() { return _i; }
  size_t row() { return _j; }

private:
	
  friend class boost::iterator_core_access;
  template <class> friend class CPUDataViewIterator;

  void increment() 
  { 
	if (++_i == _ni)
	  {		
		_i = 0;
		_column.advance( _di - _ni + 1 );
		++_j;
	  } 
	else
	  _column.increment();
  }
  
  void decrement() 
  { 
	if (--_i == -1)
	  {
		_i = _ni-1;
		_column.advance( _ni - _di - 1 );
		++_j;
	  } 
	else
	  _column.decrement();
  }
  /*  
  void advance(int n)
  {
	int N = (n + _ni - 1) / _ni; //divide with upwards rounding
	n -= N * _ni;
	_column.advance(N * _di + n); //move N rows and n columns (n can be negative if we fell of an end)
  }
  //not sure if this is right...what is it needed for?
  int distance_to(CPUDataViewIterator<Value> const& other) const
  {
	int d = (other._column.begin() - _column.begin()) / _column.size();
	int nrows = d / (_di + _ni - 1);
	if (d > 0)
	  return _ni * nrows + (_i > other._i ? _ni - (_i - other._i) : other._i - _i);
	else
	  return _ni * nrows + (_i < other._i ? - _ni - (_i - other._i) : other._i - _i);
  }
  */
  
  template <class OtherValue>
  bool equal(CPUDataViewIterator<OtherValue> const& other) const
  {
    return this->_column.begin() == other._column.begin();
  }
  CPUDataFeatureColumn<Value>& dereference() const 
  { 
	return const_cast<CPUDataFeatureColumn<Value>&>(_column); 
  }
  CPUDataFeatureColumn<Value>& dereference() 
  { 
	return _column;
  }
	
  CPUDataFeatureColumn<Value> _column;
  int _i, _j, _ni, _di;
};

#if 0
template<class Iterator> class SegmentedRowsIterator
  : public boost::iterator_adaptor
<     SegmentedRowsIterator<Iterator>
	, Iterator 
	>
{

  typedef SegmentedRowsIterator<Iterator> This;
  typedef boost::iterator_adaptor< This, Iterator > iterator_adaptor;

  SegmentedRowsIterator( const Iterator & base, ptrdiff_t rowstride, ptrdiff_t di )
	: iterator_adaptor( base ), _i(0), _ni(ni), _di( rowstride + 1 - ni ) {}

private:

  friend class boost::iterator_core_access;

  void increment() 
  { 
	if (++_i == _ni)
	  {
		_i = 0;
		base_reference().advance( _di );
	  } 
	else
	  base_reference().increment();
  }

  void decrement()
  {
	if (--_i == 0)
	  {
		_i = _ni-1;
		base_reference().advance( -_di );
	  } 
	else
	  base_reference().decrement();
  }

  void advance(typename iterator_adaptor::difference_type n)
  {
	ptrdiff_t n_rows = 0;
	if ( n > 0 )
	  while ( (-n) > _i ) { n += _ni; --n_rows; }
	else
	  while (   n  >= (_ni - _i) ) { n -= _ni; ++n_rows; }
	ptrdiff_t rowstride = _ni + _di - 1;
	base_reference().advance( n_rows * rowstride + n );
  }

  ptrdiff_t _i, _ni, _di;

};
#endif

template<class Value> class CPUDataViewLayerIterator
  : public boost::iterator_facade< CPUDataViewLayerIterator<Value>
								   , Value *
								   , boost::bidirectional_traversal_tag
								   , Value &
								   , int
								   >
{
  
public:

  CPUDataViewLayerIterator(Value * p = 0, int ni = 0, int di = 0, int stride = 0)
	: _p(p), _i(0), _ni(ni), _di(di), _j(0), _stride(stride)
	{}
	
	template <class OtherValue>
  CPUDataViewLayerIterator(CPUDataIterator<OtherValue> const& other)
	: _p(other._p), _i(other._i), _ni(other._ni), _di(other._di), _j(other._j), _stride(other._stride)
	{}

  int row() { return _j; }
  int column() { return _i; }

private:
	
  friend class boost::iterator_core_access;
  template <class> friend class CPUDataViewLayerIterator;

  // advance as usual, but whenever a step goes out of the view (_i >= _ni or _i < 0)
  // advance by the view's inter-row-stride instead (+/- _di)

  void increment() 
	{ 
		if (++_i == _ni)
		  {
			_i  = 0;
			_p += (_di - _ni + 1) * _stride;
			++_j;
		  } 
		else
		  _p += _stride;
	}

  void decrement() 
	{ 
	  if (--_i == 0)
		{
		  _i  = ( _ni - _di - 1 ) * _stride;
		  _p -=   _di;
		  --_j;
		} 
	  else
		_p -= _stride;
	}

  /*
	void advance(int n)
	{
	  int N = n / _di;
	  n -= N;
	  _p += N * _di + n;
	}
  */
	
  //not sure if this is right...what is it needed for?
  int distance_to(CPUDataViewLayerIterator<Value> const& other) const
  {
    int d = other._p - _p;
    int nrows = d / (_di + _ni - 1);
    if (d > 0)
      return _ni * nrows + (_i > other._i ? _ni - (_i - other._i) : other._i - _i);
    else
      return _ni * nrows + (_i < other._i ? - _ni - (_i - other._i) : other._i - _i);
  }
  
  template <class OtherValue>
  bool equal(CPUDataViewLayerIterator<OtherValue> const& other) const
  {
    return this->_p == other._p;
  }

  Value & dereference() const { return *_p; }
  
  Value * _p;
  int _i, _ni, _di, _j, _stride;
};

//FIXME: the above two should be merged into one and instantiated with different base iterators (T* and ColumnT)

//this is just a utility function to use zip_iterator under 100 loc :)
//take care: it does not do any concept check as it should
template<class Range1, class Range2> 
std::pair<
	boost::zip_iterator<
		boost::tuple<typename boost::range_iterator<Range1>::type, typename boost::range_iterator<Range2>::type> 
	>,
	boost::zip_iterator<
		boost::tuple<typename boost::range_iterator<Range1>::type, typename boost::range_iterator<Range2>::type> 
	>
>
zip2(Range1 r1, Range2 r2)
{
	return std::make_pair(
												boost::make_zip_iterator(boost::make_tuple(boost::begin(r1), boost::begin(r2))),
												boost::make_zip_iterator(boost::make_tuple(boost::end(r1), 	 boost::end(r2))) 
											 );
																				
}


//this is just a utility function to use zip_iterator under 100 loc :)
//take care: it does not do any concept check as it should
template<class Range1, class Range2, class Range3> 
std::pair<
	boost::zip_iterator<
		boost::tuple< typename boost::range_iterator<Range1>::type, 
		              typename boost::range_iterator<Range2>::type,
                  typename boost::range_iterator<Range3>::type
                  > 
	>,
	boost::zip_iterator<
		boost::tuple< typename boost::range_iterator<Range1>::type, 
		              typename boost::range_iterator<Range2>::type,
                  typename boost::range_iterator<Range3>::type
                  > 
	>
>
zip3(Range1 r1, Range2 r2, Range3 r3)
{
	return std::make_pair(
			boost::make_zip_iterator(boost::make_tuple(boost::begin(r1), boost::begin(r2), boost::begin(r3))),
			boost::make_zip_iterator(boost::make_tuple(boost::end(r1), 	 boost::end(r2), 	 boost::end(r3)))
											 );
																				
}

//this is just a utility function to use zip_iterator under 100 loc :)
//take care: it does not do any concept check as it should
template<class Range1, class Range2, class Range3, class Range4> 
std::pair<
	boost::zip_iterator<
		boost::tuple< typename boost::range_iterator<Range1>::type, 
		              typename boost::range_iterator<Range2>::type,
                  typename boost::range_iterator<Range3>::type,
                  typename boost::range_iterator<Range4>::type
                  > 
	>,
	boost::zip_iterator<
		boost::tuple< typename boost::range_iterator<Range1>::type, 
		              typename boost::range_iterator<Range2>::type,
                  typename boost::range_iterator<Range3>::type,
                  typename boost::range_iterator<Range4>::type
                  > 
	>
>
zip4(Range1 r1, Range2 r2, Range3 r3, Range4 r4)
{
	return std::make_pair(
			boost::make_zip_iterator(boost::make_tuple(boost::begin(r1), boost::begin(r2), boost::begin(r3), boost::begin(r4))),
			boost::make_zip_iterator(boost::make_tuple(boost::end(r1), 	 boost::end(r2), 	 boost::end(r3), 	 boost::end(r4)))
											 );
																				
}

#endif
