/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _CPUDATANDBASE_HPP_
#define _CPUDATANDBASE_HPP_

#include <stdexcept>
#include <vector>
#include <typeinfo>
#include <boost/proto/proto.hpp>

#define USE_TWO_LEVEL_PROTO 0


template<typename T> struct CPUDataND_
  : private CPUDataBlockReference<T>
{

  typedef T num_type;
  typedef CPUDataBlock<T> DataBlockT;
  typedef CPUDataBlockReference<T> DataBlockReferenceT;

  struct Dim
  {
    ptrdiff_t stride;
    size_t extent;
    Dim( ptrdiff_t s = 0, ptrdiff_t e = 0 ) : stride(s), extent(e) {}
  };

  CPUDataND_()
  {
  }

  std::vector<Dim> & dims() { return _dims; }
  const std::vector<Dim> & dims() const { return _dims; }
  Dim & dim(size_t i) { return _dims.at(i); }
  const Dim & dim(size_t i) const { return _dims.at(i); }
  size_t extent(size_t index) const { return dim(index).extent; }
  size_t stride(size_t index) const { return dim(index).stride; }
  ptrdiff_t offset() const { return _offset; }
  void rebase( ptrdiff_t offset ) { _offset = offset; }
  T * data() { return DataBlockReferenceT::data(); }
  const T * data() const { return DataBlockReferenceT::data(); }
  void clear() { DataBlockReferenceT::release(); _dims.clear(); }
  size_t rank() const { return _dims.size(); }
  size_t size() const { return DataBlockReferenceT::size(); }

  void init( size_t rank = 0, size_t * shape = 0 )
  {
    clear();
    _dims.resize(rank);
	size_t n = 1;
	if ( rank > 0 )
	  {
	    //iterate backwards, update n multiplicatively, and use intermediate values for n as strides
	    for ( int i = rank - 1; i >= 0; --i )
	      {
			_dims[i] = Dim( n, shape[i] );
			n *= shape[i];
	      }
	  }
	  if ( n > 0 )
	    {
		    DataBlockReferenceT::construct( new DataBlockT( n ) );
	    }
  }

  void slice( size_t dim, size_t index  )
  {

#ifdef DEBUG
    if ( dim >= rank() ) throw std::runtime_error("invalid dimension");
#endif

    // adjust parameters
    _contiguous = _contiguous && ( stride(dim) == 0 || extent(dim) == 1 );
    _offset += index * stride(dim);
    
    // finally really erase that dimension
    _dims.erase( _dims.begin() + dim );

  }

 friend std::ostream &operator <<(std::ostream &sout, const CPUDataND_<T> & data )
 {
   sout << "CPUDataND_" << "<" << typeid(T).name() << ">";
   sout << "(";
   for( size_t i = 0; i < data.rank(); ++i )
     {
       sout << data.extent(i);
       if ( i + 1 < data.rank() ) sout << "x";
     }
   sout << ")";
   return sout;
 }

private:
  
  std::vector<Dim> _dims;
  ptrdiff_t _offset;
  bool _contiguous;
  
};



#endif /* _CPUDATANDBASE_HPP_ */

