/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _CPUDATANDEXPRESSIONITERATORS_HPP_
#define _CPUDATANDEXPRESSIONITERATORS_HPP_

#include "CPUDataNDExpressions.hpp"
#include "IndexExpressions.hpp"

#include <boost/proto/proto.hpp>
#include <boost/fusion/algorithm.hpp>
#include <boost/fusion/include/algorithm.hpp>

#include <boost/mpl/vector.hpp>
#include <boost/mpl/int.hpp>
#include <boost/format.hpp>


#include "CPUDataNDGrammars.hpp"

namespace proto = boost::proto;
namespace fusion = boost::fusion;
namespace mpl = boost::mpl;

// TODO: improve end check (cache in flag)
// TODO: compile time eliminate representations of indices not represented in base iters
template<typename T, int N, class LeftIter, class RightIter>
struct CPUDatataNDExpressionIterator
{
  struct Dim 
  {
    Dim( ptrdiff_t p = 0, ptrdiff_t lb = 0, ptrdiff_t ub = 0 )
      : pos(p), lbound(lb), ubound(ub)
      {}
    ptrdiff_t pos;
    ptrdiff_t lbound;
    ptrdiff_t ubound;
    bool atend() const { return pos > ubound; }
    void reset() { pos = lbound; }
  };

  Dim dims[N];

  bool atend() const { return dims[N-1].atend(); };
  unsigned rank() const { return N; }

  void increment() 
  {
    increment(0);
    if ( dims[0].atend() )
    {
      for ( unsigned i = 1;  i < rank(); ++i )
      {
        reset(i-1);
        increment(i);
        if ( !dims[i].atend() )
          break;
      }
    }
        
  }
  
  void reset(size_t i)
  {
    left.advance( i, -dims[i].pos );
    right.advance( i, -dims[i].pos );
    dims[i].reset();
  }

  void increment(size_t i)
  {
    left.increment(i);
    right.increment(i);
    ++dims[i].pos;
  }

  ptrdiff_t & pos(size_t i)           { return dims[i].pos;    }
  ptrdiff_t   pos(size_t i)    const  { return dims[i].pos;    }
  ptrdiff_t & lbound(size_t i)        { return dims[i].lbound; }
  ptrdiff_t   lbound(size_t i) const  { return dims[i].lbound; }
  ptrdiff_t & ubound(size_t i)        { return dims[i].ubound; }
  ptrdiff_t   ubound(size_t i) const  { return dims[i].ubound; }

  LeftIter left;
  RightIter right;

};

template<typename T, int N>
struct CPUDataNDWalker
{
  CPUDataNDWalker(T * _p = 0) : p(_p) 
  {
    memset(strides,0,sizeof(strides));
  }  

  ptrdiff_t strides[N];
  T * p;

  unsigned rank() const { return N; }
  ptrdiff_t stride(size_t i) const { return strides[i]; }
  ptrdiff_t & stride(size_t i) { return strides[i]; }

  T& operator*()
  {
    return *p;
  }
  
  T operator*() const
  {
    return *p;
  }
  
  void advance(size_t i, size_t n)
  {
    p += stride(i) * n;
  }

  void increment(size_t i)
  {
    p += stride(i);
  }

};

template<typename T>
struct CPUDataNDConstIterator
{
  T value;
  CPUDataNDConstIterator( T v = T(0) ) : value(v) {}
  T operator*() const { return value; }
  void advance(size_t, size_t ) {}
  void increment(size_t) {}
  unsigned rank() const { return 0; }
  ptrdiff_t stride(size_t i) const { return 0; }
  ptrdiff_t & stride(size_t i) { return 0; }
};

template<typename T,typename OP,typename IT1>
struct CPUDataNDUnaryOpWalker
{
  IT1 iter1;
  OP op;
  CPUDataNDUnaryOpWalker(const OP & o, const IT1 & it1)
    : op(o), iter1(it1)
    {}
    
  T operator*() const
  {
    return op(*iter1);
  } 
  
  void increment()
  {
    iter1.increment();
  }
  
  bool atend()
  {
    return iter1.atend();
  }
  
};

template<typename T,typename OP,typename IT1,typename IT2>
struct CPUDataNDBinaryOpWalker
{
  OP op;
  IT1 iter1;
  IT2 iter2;
  CPUDataNDBinaryOpWalker() {} 
  CPUDataNDBinaryOpWalker(const OP & o, const IT1 & it1, const IT2 & it2)
    : op(o), iter1(it1), iter2(it2)
    {}
    
  T operator*() const
  {
    return op(T(*iter1),T(*iter2));
  } 
  
  void increment(size_t i)
  {
    iter1.increment(i);
    iter2.increment(i);
  }
  
  void advance(size_t i, size_t n)
  {
	iter1.advance( i, n );
	iter2.advance( i, n );
  }
  
};




// callable transform that builds an interator from indexed array "terminals"
template<typename T, typename IndexSet, typename Callable = proto::callable>
struct TerminalIteratorBuilder : proto::callable
{
  typedef typename mpl::size<IndexSet>::type Arity;
  typedef CPUDataNDWalker<const T,Arity::value> result_type;
  typedef CPUDataND_<T> array_t;

  template<typename Expr>
  result_type operator()(Expr & expr)
  {
	//boost::proto::display_expr( expr );
    const array_t & array = proto::value( fusion::front(expr) );
    result_type iterator(array.data());
	//std::cout << std::endl;
    indexing_decoder decode(iterator,array);
	//std::cout << std::endl;
    fusion::for_each( fusion::pop_front( expr ), decode );
    return iterator;
  }

  // a functor to decode index-strides and put them into the iterator
  struct indexing_decoder
  {
    int dim;
    const result_type & iterator;
    const array_t & array;
    indexing_decoder(const result_type & it, const array_t & a)
      : dim(0), iterator(it), array(a)
    {}

    template<typename index_expr>
    struct indexing_decoder_helper
    {
      int index_pos;
      const result_type & iterator;
      const index_expr & expr;
      ptrdiff_t astride;
      indexing_decoder_helper(const index_expr & e, const result_type & it, ptrdiff_t s)
       : index_pos(0), iterator(it), expr(e), astride(s)
      {}
      template<int I> void operator()(mpl::int_<I>) const
      {
        ptrdiff_t index_stride = IndexStride<I>()(expr);
        const_cast<result_type&>(iterator).stride(index_pos) += astride * index_stride;
		//std::cout << " index pos " << index_pos << " index " << I << " stride " << const_cast<result_type&>(iterator).stride(index_pos) << std::endl;
        ++const_cast<int&>(index_pos);
      }
    };
    
    template<typename index_expr> void operator()(index_expr const &expr) const
    {
      // modify strides according to array stride and index expression
      ptrdiff_t astride = array.stride(dim);
      indexing_decoder_helper<index_expr> decode_index( expr, iterator, astride );
      fusion::for_each( IndexSet(), decode_index );
      const_cast<result_type&>(iterator).p += astride * IndexOffset()(expr);
	  //std::cout << " dim " << dim << " astride " << astride << std::endl;
      ++const_cast<int&>(dim);
    }
        
  };
      
};

// callable transform that builds an interator from indexed array "terminals"
template<typename T, typename IndexSet, typename RightIter, typename Callable = proto::callable>
struct LValueIteratorBuilder : proto::callable
{
  typedef typename mpl::size<IndexSet>::type Arity;
  typedef CPUDataNDWalker<T,Arity::value> left_base_iter;
  typedef CPUDatataNDExpressionIterator< T, Arity::value, left_base_iter, RightIter > result_type;
  typedef CPUDataND_<T> array_t;

  template<typename Expr>
  result_type operator()(Expr & expr)
  {
    array_t & array = proto::value( fusion::front(expr) );
    result_type iterator;
	iterator.left.p = array.data();
    indexing_decoder decode(iterator,array);
	//std::cout << std::endl;
    fusion::for_each( fusion::pop_front( expr ), decode );
	//std::cout << std::endl;
    return iterator;
  }

  struct indexing_decoder
  {
    int dim;
    const result_type & iterator;
    const array_t & array;
    indexing_decoder(const result_type & it, const array_t & a)
      : dim(0), iterator(it), array(a)
    {}

	template<typename index_expr> void operator()( index_expr const & ) const
    {
	  typedef typename proto::result_of::left< index_expr >::type placeholder;
	  const int I = placeholder::value;
      const_cast<result_type&>(iterator).left.stride(dim) = array.stride(I);
	  const_cast<result_type&>(iterator).dims[dim].ubound = array.extent(I) - 1;
	  //std::cout << I << "->" << dim << " stride " << array.stride(I) << " extent " << array.extent(I) << std::endl;
      ++const_cast<int&>(dim);
    }
        
  };
      
};

template<typename T, typename OP, typename IT1, typename Callable = proto::callable>
struct UnaryOpIteratorBuilder : proto::callable
{
  typedef CPUDataNDUnaryOpWalker<T,OP,IT1> result_type;
  result_type operator()(const OP & op, const IT1 & it1)
  {
    return result_type(op,it1);
  }
};

template<typename T, typename IndexSet> struct ExpressionIteratorBuilder;

#include <boost/type_traits/remove_reference.hpp>
#include <boost/type_traits/remove_const.hpp>

struct ConstIteratorBuilder : proto::transform< ConstIteratorBuilder >
{
  template<typename Expr, typename State, typename Data>
  struct impl : proto::transform_impl<Expr, State, Data>
  {
	typedef typename boost::remove_reference<Expr>::type Expr_noref;
	//typedef typename proto::result_of::value<Expr_noref> const_ref_t;
	typedef typename Expr_noref::proto_child0 const_ref_t;
	typedef typename boost::remove_reference<const_ref_t>::type const_t;
	typedef typename boost::remove_const<const_t>::type value_t;
	typedef CPUDataNDConstIterator<value_t> result_type;
	result_type operator ()
	(
	 typename impl::expr_param expr,
	 typename impl::state_param state,
	 typename impl::data_param data
	 ) const
	{
	  return result_type( proto::value( expr ) );
	}
  };
};

template<typename T, typename IndexSet, typename OP>
struct BinaryOpIteratorBuilder : proto::transform< BinaryOpIteratorBuilder< T, IndexSet, OP > >
{
  template<typename Expr, typename State, typename Data>
  struct impl : proto::transform_impl<Expr, State, Data>
  {
	typedef typename proto::result_of::left < Expr >::type Left;
	typedef typename proto::result_of::right< Expr >::type Right;
	typedef ExpressionIteratorBuilder<T,IndexSet> Builder;
	typedef typename boost::result_of< Builder(Left)  >::type left_iterator;
	typedef typename boost::result_of< Builder(Right) >::type right_iterator;
	typedef CPUDataNDBinaryOpWalker<T,OP,left_iterator,right_iterator> result_type;
	result_type operator ()
	(
	 typename impl::expr_param expr,
	 typename impl::state_param state,
	 typename impl::data_param data
	 ) const
	{
	  return result_type( OP(), Builder()( proto::left( expr ) ), Builder()( proto::right( expr ) ) );
	}
  };
};

#define HANDLE_BINARY_OPERATOR(OP)                                                  \
proto::when<															            \
  proto::OP< cpudata_grammar, cpudata_grammar > ,                                   \
  proto::call< BinaryOpIteratorBuilder<T, IndexSet, std::OP<T> >(  proto::_ ) > \
>

#define HANDLE_UNARY_OPERATOR(OP)                                                    \
proto::when<							                                             \
 proto::OP< cpudata_grammar > ,                                                      \
  UnaryOpIteratorBuilder<T, std::OP<T>,                                              \
     ExpressionIteratorBuilder<T, IndexSet>( proto::_left ) >                        \
   (                                                                                 \
	std::OP<T>(),														\
    ExpressionIteratorBuilder<T, IndexSet>( proto::_left  )                          \
   )                                                                                 \
>


template<typename T, typename IndexSet>
struct ExpressionIteratorBuilder
  : proto::or_<
      proto::when< 
        cpudata_indexing_grammar< proto::_ > ,
        TerminalIteratorBuilder<T, IndexSet>( proto::_ )
	    > 
      ,proto::when<literal_number_grammar,ConstIteratorBuilder( proto::_ )>
      ,HANDLE_BINARY_OPERATOR(plus)
      ,HANDLE_BINARY_OPERATOR(minus)
      ,HANDLE_BINARY_OPERATOR(multiplies)
      ,HANDLE_BINARY_OPERATOR(divides)
      ,HANDLE_UNARY_OPERATOR(negate)
      // there is no unary_plus in std, just recurse to the child (i.e. pass-through)
      ,proto::when<															             
        proto::unary_plus< cpudata_grammar > ,                                                      
		 ExpressionIteratorBuilder<T, IndexSet>( proto::_left )
      >
  >
{};


#undef HANDLE_BINARY_OPERATOR
#undef HANDLE_UNARY_OPERATOR

#endif /* _CPUDATANDEXPRESSIONITERATORS_HPP_ */
