/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef NEWCPUDATAEXPRESSIONS_HPP_Y2DKSX99
#define NEWCPUDATAEXPRESSIONS_HPP_Y2DKSX99

#include <boost/proto/proto.hpp>
#include <boost/fusion/algorithm.hpp>
#include <boost/fusion/include/algorithm.hpp>

#include "CPUDataNDBase.hpp"
#include "IndexExpressions.hpp"
#include "CPUDataNDExpressionIterators.hpp"

#include "CPUDataNDGrammars.hpp"

namespace proto = boost::proto;
namespace fusion = boost::fusion;
namespace mpl = boost::mpl;

struct cpudata_domain;

template<int N, typename T> struct cpudata_indexing_ctx;

#include <iostream>
template<typename Expr>
struct cpudata_expr
: proto::extends<Expr, cpudata_expr<Expr>, cpudata_domain>
{
  typedef proto::extends<Expr, cpudata_expr<Expr>, cpudata_domain> base_type;

  typedef typename boost::result_of<CPUDataNDExprNumType(Expr)>::type num_type;
  typedef typename mpl::if_< proto::matches<Expr,simple_cpudata_indexing_grammar< num_type > >
							 , num_type &, num_type >::type value_type;

  template<typename Expr2>
  cpudata_expr & operator=( Expr2 & right ) const
  {
	cpudata_expr & left = *const_cast<cpudata_expr*>(this);

	typedef typename boost::result_of<CPUDataNDExprIndexSet(Expr)>::type IndexSet;
	
	typedef ExpressionIteratorBuilder<num_type,IndexSet> RightBuilder;
	typedef typename boost::result_of<RightBuilder(Expr2)>::type RightIterator;
	
	RightIterator right_iterator = RightBuilder()( right );
	
	typedef LValueIteratorBuilder<num_type,IndexSet,RightIterator> LeftBuilder;
	typedef typename LeftBuilder::result_type LeftIterator;
	
	LeftIterator iterator = LeftBuilder()( left );
	iterator.right = right_iterator;
	
	while ( !iterator.atend() )
	  {
		*iterator.left = *iterator.right;
		iterator.increment();
	  }

	return left;
  }  

  value_type at(size_t i1, size_t i2 ) const
  {
	cpudata_indexing_ctx<2,value_type> ctx(i1,i2);
	cpudata_expr & self = *const_cast<cpudata_expr*>(this);
	return proto::eval( self, ctx );
  }

  cpudata_expr( Expr const & expr = Expr() )
	: base_type( expr )
  {}
};

struct cpudata_domain
  : proto::domain< proto::generator<cpudata_expr>, cpudata_grammar >
{};

template<int N, typename T>
struct cpudata_indexing_ctx
{
  
  typedef index_context<N> index_context_t;
  typedef cpudata_indexing_ctx<N,T> cpudata_indexing_ctx_t;

  index_context_t _indices;
  
#define BOOST_PP_LOCAL_LIMITS (0, INDEX_PLACEHOLDER_MAX )
#define BOOST_PP_LOCAL_MACRO(I)                                  \
    cpudata_indexing_ctx( BOOST_PP_ENUM_PARAMS(I, ptrdiff_t a) ) \
      : _indices( BOOST_PP_ENUM_PARAMS(I, a) ) {}
#include BOOST_PP_LOCAL_ITERATE()

  //default
  template<typename Expr, typename EnableIf = void>
    struct eval
      : proto::default_eval<Expr, cpudata_indexing_ctx_t const>
    {};

  //evaluate array indexing
  template<typename Expr>
    struct eval<
        Expr
      , typename boost::enable_if<
            proto::matches< Expr, cpudata_indexing_grammar< proto::_ > >
        >::type
    >
    { 

      // a functor to perform indexing
	  template<typename array_num_type>
      struct do_indexing
      {
		typedef CPUDataND_<array_num_type> array_t;
        int dim;
        const array_t & array;
        const array_num_type* & p;
        const index_context_t & ctx;
        
        do_indexing( const array_t & a, const array_num_type* & _p, const index_context_t & c )
        : dim(0), array(a), p(_p), ctx(c) {}
        
        template<typename index_expr> void operator()(index_expr const &e) const
        {
          ptrdiff_t stride = array.stride(dim);
          ptrdiff_t index = proto::eval(e,ctx);
          p += index * stride;
          // hack because fusion only takes const& functors
          ++const_cast<int&>(dim);
        }
        
      };
      
      typedef T result_type;
      result_type operator ()(Expr &expr, cpudata_indexing_ctx_t const & ctx) const
      {
		typedef typename terminal_expr_num_type<Expr>::type array_num_type;
		typedef CPUDataND_< array_num_type > array_t;
        const array_t & array = proto::value( fusion::front(expr) );
        const array_num_type * p = array.data();
        do_indexing<array_num_type> f( array, p, ctx._indices );
        fusion::for_each( fusion::pop_front( expr ), f );
        return *const_cast<array_num_type*>(p);
      }
      
    };
};




#endif /* end of include guard: NEWCPUDATAEXPRESSIONS_HPP_Y2DKSX99 */
