/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __CPUDATANDGRAMMARS_HPP__
#define __CPUDATANDGRAMMARS_HPP__

#include <boost/proto/proto.hpp>
#include <boost/mpl/comparison.hpp>
#include <boost/mpl/sizeof.hpp>

namespace proto = boost::proto;
namespace mpl = boost::mpl;

template<int I>
struct cpudata_placeholder
{
   friend std::ostream &operator <<(std::ostream &sout, cpudata_placeholder)
   {
       return sout << "cpudata_placeholder<" << I << ">";
   }
};


// Here are the cases used by the switch_ above.

struct literal_number_grammar
  : proto::or_<
  proto::terminal< float       >,
  proto::terminal< int         >,
  proto::terminal< const int & >,
  proto::terminal< double      >
  > 
{};

// Define the grammar of cpudata expressions
struct cpudata_grammar;

// this is for array terminals, indexed by index placeholder expressions
template<typename T> struct cpudata_indexing_grammar
  : proto::function< proto::terminal< CPUDataND_< T > >, proto::vararg< index_grammar > >
{};

// this is for array terminals, indexed by index placeholder expressions
template<typename T> struct simple_cpudata_indexing_grammar
: proto::function< proto::terminal< CPUDataND_< T > >, proto::vararg< index_terminal_grammar > >
{};

struct AssignOpsCases
{
    template<typename Tag, int D = 0> struct case_  : proto::not_<proto::_> {};

#define ASSIGN_OP(op)                                                        \
    template<int D> struct case_< proto::tag:: op, D >                       \
      : proto:: op < cpudata_indexing_grammar< proto::_ >, cpudata_grammar > \
    {}
    
  ASSIGN_OP(assign);
  ASSIGN_OP(plus_assign);
  ASSIGN_OP(minus_assign);
  ASSIGN_OP(multiplies_assign);
  ASSIGN_OP(divides_assign);
    
#undef ASSIGN_OP

};  

struct cpudata_grammar_cases
{
    template<typename Tag>
    struct case_ : proto::not_<proto::_>  {};
};
template<> struct cpudata_grammar_cases::case_<proto::tag::terminal>
: proto::or_< proto::terminal< CPUDataND_< proto::_ > >, literal_number_grammar > {};
template<> struct cpudata_grammar_cases::case_<proto::tag::plus>
: proto::plus<cpudata_grammar,cpudata_grammar> {};
template<> struct cpudata_grammar_cases::case_<proto::tag::minus>
: proto::minus<cpudata_grammar,cpudata_grammar> {};
template<> struct cpudata_grammar_cases::case_<proto::tag::multiplies>
: proto::multiplies<cpudata_grammar,cpudata_grammar> {};
template<> struct cpudata_grammar_cases::case_<proto::tag::divides>
: proto::divides<cpudata_grammar,cpudata_grammar> {};
template<> struct cpudata_grammar_cases::case_<proto::tag::function>
: cpudata_indexing_grammar< proto::_ > {};

struct cpudata_grammar
  : proto::or_<
      proto::switch_< cpudata_grammar_cases >,
      proto::switch_< AssignOpsCases >
    >
{};
  
struct CPUDataNDExprIndexSet
  : proto::or_<
      proto::when< index_grammar, IndexExprIndexSet(proto::_) >,
      proto::when< proto::terminal< CPUDataND_< proto::_ > >, mpl::set<>() >,
      proto::when< 
        proto::nary_expr<proto::_, proto::vararg<proto::_> >,
        proto::fold<
          proto::_,
          mpl::set<>(),
	      mpl::joint_view< proto::_state, CPUDataNDExprIndexSet >()
	    >
	  >
    >
                  
{  
};

template<typename Expr> struct terminal_expr_num_type
{
  typedef typename proto::result_of::left< Expr >::type left_terminal_expression;
  typedef typename proto::result_of::value< left_terminal_expression >::type array_t;
  typedef typename array_t::num_type type;
};

struct CPUDataNDTerminalExprNumType : proto::transform<CPUDataNDTerminalExprNumType>
{
  template<typename Expr, typename State, typename Data>
  struct impl
  {
	typedef typename proto::result_of::value< Expr >::type array_t;
	typedef typename array_t::num_type result_type;
  };
};

template<typename T1, typename T2>
struct promote
{
  typedef typename mpl::if_< mpl::less_equal< mpl::sizeof_<T1>, mpl::sizeof_<T2> >, T2, T1 >::type type;
};

struct CPUDataNDExprNumType
  : proto::or_<
  proto::when< literal_number_grammar, proto::_value >,
  proto::when< proto::terminal< CPUDataND_< proto::_ > >, CPUDataNDTerminalExprNumType() >,
  proto::when< cpudata_grammar, 
			   proto::fold< proto::_, char(), 
							promote < proto::_state, CPUDataNDExprNumType >()
							>
			   >,
  proto::otherwise< char() >
  >
{
};


// bounds calculator
// go down to indexed arrays
// use index stride and index offset of index expression for each dimension together with array bounds to calc index bounds
// shrinking combine bounds along the expression tree
// index expression leaves ( i.e. non-array leaves ) are unbounded
struct CPUDataNDExprIndexBounds;

#endif
