/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _CPUDATARFITERATOR_HPP
#define _CPUDATARFITERATOR_HPP

#include "CPUData.hpp"
#include <boost/format.hpp>

#define RF_ASSERTS


template<typename T>
class CPUDataRFIterator
{

public:

  CPUDataRFIterator()
  : _i(0), _j(0), _ni(0), _nj(0), _rf( NULL,0,0,0,0),_step1(0),_step2(0),_offset(0)
  {

  }

  CPUDataRFIterator( const CPUData<T> & data, size_t d, size_t step )
	: _i(0)
	, _j(0)
	, _ni(0)
	, _nj(0)
	, _rf( data, 0, d, 0, d )
	, _step1(step)
	, _step2(step)
	, _offset(data.offset())
  {
  	
  	if ( step == 0 )
    {
  		throw(std::runtime_error(str(boost::format("CPUDataRFIterator: wrong stepsize: %d" ) % _step1 ) ) );
        _rf.clear();
    }
    else
    {
        if ( data.rows()    >= d ) _ni = ( data.rows()    - (d) ) / step + 1;
        if ( data.columns() >= d ) _nj = ( data.columns() - (d) ) / step + 1;
    }
  }

  CPUDataRFIterator( const CPUData<T> & data, size_t d1, size_t step1, size_t d2, size_t step2 )
	: _i(0)
	, _j(0)
	, _ni(0)
	, _nj(0)
	, _rf( data, 0, d1, 0, d2 )
	, _step1(step1)
	, _step2(step2)
	, _offset(data.offset())
  {
  	
  	if ( _step1 <= 0 )
  		throw(std::runtime_error(str(boost::format("CPUDataRFIterator: wrong stepsize (1): %d ( data: %s )" ) % _step1 % data.dump_dims() ) ) );
  	if ( _step2 <= 0 )
  		throw(std::runtime_error(str(boost::format("CPUDataRFIterator: wrong stepsize (2): %d ( data: %s )" ) % _step2  % data.dump_dims() ) ) );

    if ( data.rows()    >= d1 ) _ni = ( data.rows()    - (d1) ) / step1 + 1;
    if ( data.columns() >= d2 ) _nj = ( data.columns() - (d2) ) / step2 + 1;
    
  }

  CPUDataRFIterator( const CPUData<T> & data, size_t d, size_t step, size_t start_row )
	: _i(0),_j(0), _ni(0), _nj(0), _rf( data, start_row, d, 0, d ), _step1(step), _step2(step), _offset(data.offset())
  {
  	
  	if ( step <= 0 )
  		throw(std::runtime_error(str(boost::format("CPUDataRFIterator: wrong stepsize: %d ( data: %s )" ) % step % data.dump_dims() ) ) );
  	if ( data.rows()    >= d+start_row ) _i = start_row;
  	if ( data.rows()    >= d ) _ni = ( data.rows()    - d ) / step + 1;
    if ( data.columns() >= d ) _nj = ( data.columns() - d ) / step + 1;
    
  	//std::cout << data.dump_dims() << ", d="<<d<<"  step="<<step<<std::flush;
	//std::cout << "  _i="<<_i<<" _ni="<<_ni<<std::flush;
	//std::cout << "  _j="<<_j<<" _nj="<<_nj<<std::endl;
	
  }

  void advance( ptrdiff_t nrows, ptrdiff_t ncols )
  {
  	ptrdiff_t rowmove = nrows * _rf.row_stride() * _step1;
  	ptrdiff_t colmove = ncols * _rf.col_stride() * _step2;
  	_rf.rebase( size_t( ptrdiff_t ( _rf.offset() ) + rowmove + colmove ) );
  }

  void move_to( size_t nrows, size_t ncols )
  {
  	size_t rowmove = nrows * _rf.row_stride() * _step1;
  	size_t colmove = ncols * _rf.col_stride() * _step2;
  	_rf.rebase( _offset + rowmove + colmove );
  }

  CPUDataRFIterator & increment()
  {
	++_j;
	
	if ( _j > _nj ) throw std::runtime_error("Error: j out of bounds");

	if ( _j == _nj )
	  { 
		// go to the beginning of the next row
		_j = 0;		
		++_i;
	  }

	move_to( _i, _j );

	return *this;

  }

  void reset()
  {
	_i = _j = 0;
  }

  bool more()
  {
	bool has_more = _i < _ni;
  	#ifdef DEBUG
	if ( has_more && !_rf.is_valid() )
		throw std::runtime_error(str(boost::format("rf iterator wants to keep going with invalid pointers..: i=%d, ni=%d, j=%d, nj=%d, step1=%d, step2=%d, dim=%d")%_i%_ni%_j%_nj%_step1%_step2%_rf.rows()));
	#endif
	return has_more;
  }



  CPUData<T>  substract( const CPUDataRFIterator & other ) const
  {
  	#ifdef DEBUG
  	if (!_rf.is_valid()) throw std::runtime_error("I'm not valid!");
  	if (!other->is_valid()) throw std::runtime_error("He's not valid!");
	/*
	// check if adresses are ok
	T* my_adr = (T*)_rf.data() - _offset;
	T* his_adr = (T*)other->data() - other->offset();
	if ( my_adr != his_adr )
		throw std::runtime_error(str(boost::format("cannot make diff because iterators point to different sources! this_adr=%d, other_adr=%d")%my_adr%his_adr));
	*/  	
  	// check if sizes are ok
  	if ( _rf.rows() != other->rows() || _rf.columns() != other->columns() || rows() != other.rows() || columns() != other.columns() || step() != other.step() || other->offset() > _rf.offset() )
  		throw std::runtime_error(str(boost::format("cannot make diff because iterators are differently sized: src: rows= %d, cols=%d ni=%d, nj=%d, step=%d, offset=%d   other: rows= %d, cols=%d ni=%d, nj=%d, step=%d, offset=%d ") %_rf.rows() %_rf.columns() %rows() %columns() %step() %_rf.offset() %other->rows() %other->columns() %other.rows() %other.columns() %other.step() %other->offset() ) ); 
  	#endif
  	
	CPUData<T> result; 
	// option 1: row-wise dislocation  	
  	if ( row() == other.row() )
  	{
		// check if diff is possible
		size_t diff_offset = _rf.offset();
		size_t diff_rows   = _rf.rows() - ( _rf.offset() - other->offset())/_rf.col_stride();
		size_t diff_cols   = _rf.columns();
		//std::cout << "ohter.offset="<<other->offset()<< ", this.offset="<<_rf.offset()<<", diff_rows="<<diff_rows<< ", diff_cols="<<diff_cols<<std::endl;
		result.reference( _rf, 0, diff_rows, 0, diff_cols );
	}
	// option 2: column-wise dislocation
	else if ( column() == other.column() )
	{
		throw std::runtime_error("substract: column mode not yet implemented, sorry.");	
	}
	else
		throw std::runtime_error(str(boost::format("dislacation btw iterator is not correct: this: ni=%d, nj=%d   other: ni=%d, nj=%d")%row()%column() %other.row() %other.column() ));
  	
  	return result;
  }  


  CPUData<T>  intersect( const CPUDataRFIterator & other ) const
  {
  	throw std::runtime_error("intersect not yet implemented, sorry.");
  }
  
  CPUData<T>  unite( const CPUDataRFIterator & other ) const
  {
  	throw std::runtime_error("unite not yet implemented, sorry.");
  }
  
  
  CPUData<T> & operator*()
  {
  	#ifdef DEBUG
  		if (!_rf.is_valid() )
			throw std::runtime_error(str(boost::format("trying to reference invalid rf: %s")%_rf.dump_dims() ));
	#endif
	return _rf;
  }

  const CPUData<T> & operator*() const
  {
   	#ifdef DEBUG
  		if (!_rf.is_valid() )
			throw std::runtime_error(str(boost::format("trying to reference invalid rf: %s")%_rf.dump_dims() ));
	#endif
	return _rf;
  }
  
  CPUData<T> * operator->()
  {  	
  	#ifdef DEBUG
  		if (!_rf.is_valid() )
			throw std::runtime_error(str(boost::format("trying to reference invalid rf: %s")%_rf.dump_dims() ));
	#endif
	return &_rf;
  }

  const CPUData<T> * operator->() const
  {
	#ifdef DEBUG
  		if (!_rf.is_valid() )
			throw std::runtime_error(str(boost::format("trying to reference invalid rf: %s")%_rf.dump_dims() ));
	#endif
	return &_rf;
  }

  size_t rows() const
  {
	return _ni;
  }

  size_t columns() const
  {
	return _nj;
  }

  size_t row() const
  {
	return _i;
  }

  size_t column() const
  {
	return _j;
  }

  size_t offset() const
  {
  	return _offset;	
  }
  
  size_t step() const
  {
	return _step1;	
  }

  size_t step1() const
  {
	return _step1;	
  }

  size_t step2() const
  {
	return _step2;	
  }

  CPUDataRFIterator & operator++()
  {
    return increment();
  }

private:

  size_t _i, _j, _ni, _nj;
  CPUData<T> _rf;
  size_t _step1;
  size_t _step2;
  size_t _offset;

};


#endif
