/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "Header.hpp"
#include <boost/lexical_cast.hpp>
#include <iostream>

MPI::Datatype Header::header_type = MPI_DATATYPE_NULL;


void Header::set_column_shape( const ColumnShapeT & sizes )
  {	
    size_t old_dim = 1;
    for (size_t i = 0; i < MAX_ADDITIONAL_DIMS; ++i )
      old_dim *= _column_shape[i];
    if (old_dim == 1) old_dim = _shape[2];
    size_t new_dim = 1;
    for (size_t i = 0; i < sizes.size(); ++i )
      new_dim *= sizes[i];
#if 0
    if ( old_dim != new_dim )
    {
      std::cerr << "Warning in CPUDataHeader: changing old column_shape : \n["<<std::flush;
      for (int i = 0; i < MAX_ADDITIONAL_DIMS; ++i)
        std::cerr << _column_shape[i]<<" ";
      
      std::cerr << "]\ninto new header column_shape :\n["<<std::flush;
      for (int i = 0; i < sizes.size(); ++i)
        std::cerr << sizes[i]<<" ";
      std::cerr << "]"<<std::endl;
    }
#endif
  
  if ( sizes.size() > MAX_ADDITIONAL_DIMS)
    {
      reset_column_shape();
      for (size_t i = 0; i < MAX_ADDITIONAL_DIMS-1; ++i)
        _column_shape[i] = sizes[i];
      
      for(size_t i = MAX_ADDITIONAL_DIMS-1; i < sizes.size(); ++i )
        _column_shape[MAX_ADDITIONAL_DIMS-1] *= sizes[i];

      std::cerr << "Warning in CPUDataHeader: reducing original column_shape : \n["<<std::flush;
      for (size_t i = 0; i < sizes.size(); ++i)
        std::cerr << sizes[i]<<" ";
      std::cerr << "]\nto new header column_shape :\n["<<std::flush;
      for (size_t i = 0; i < MAX_ADDITIONAL_DIMS; ++i)
        std::cerr << _column_shape[i]<<" ";
      std::cerr << "]"<<std::endl;
    }
  else
    {
      reset_column_shape();
      for (size_t i=0; i < sizes.size(); ++i)
        _column_shape[i] = sizes[i];
    }
  }
  
  
  
Header::ColumnShapeT Header::get_column_shape_as_vector() const
  {
    ColumnShapeT result;
    
    size_t deepest_dim = 0;
    for (size_t i = 0; i < MAX_ADDITIONAL_DIMS; ++i )
    {
      if (_column_shape[i] > 1 ) deepest_dim = i;
    }
    
    if (deepest_dim > 0)
    {
      result.resize( deepest_dim + 1 );
      //std::copy( _column_shape, _column_shape + deepest_dim + 1, result.begin() );
      for ( size_t i = 0; i <= deepest_dim; ++i )
        result[i] = _column_shape[i];
    }
    return result;
  }
  
void Header::commit( )
  {
    if ( (MPI_Datatype)header_type == MPI_DATATYPE_NULL )
    {
      Header* dummy =  new Header();
      MPI::Aint start_adr     = MPI::Get_address( dummy );
      MPI::Aint type_adr      = MPI::Get_address( dummy->data_type() );
      MPI::Aint msg_adr       = MPI::Get_address( dummy->msg_type()  );
      MPI::Aint id_adr        = MPI::Get_address( dummy->id() );
      MPI::Aint per_adr       = MPI::Get_address( dummy->persistent() );
      MPI::Aint shape_adr     = MPI::Get_address( dummy->shape() );
      MPI::Aint colshape_adr  = MPI::Get_address( dummy->column_shape() );
      
      /*
      std::cout << "---- create_cpudata_header, adresses and values:"<<std::endl;
      std::cout << "  -- start_adr    = "<< start_adr <<std::endl;
      std::cout << "  -- type_adt     = "<< type_adr <<" byteoffset = "<<(type_adr-start_adr) <<std::endl;
      std::cout << "  -- msg_adr      = "<< msg_adr <<" byteoffset = "<<(msg_adr-start_adr) <<std::endl;
      std::cout << "  -- id_adr       = "<< id_adr <<" byteoffset = "<<(id_adr-start_adr) <<std::endl;
      std::cout << "  -- shape_adr    = "<< shape_adr <<" byteoffset = "<<(shape_adr-start_adr) <<std::endl;
      std::cout << "  -- shape_adr[0] = "<< (shape_adr-start_adr) <<std::endl;
      std::cout << "  -- shape_adr[1] = "<< (shape_adr-start_adr + MPI::UNSIGNED_LONG.Get_size()) <<std::endl;
      std::cout << "  -- shape_adr[2] = "<< (shape_adr-start_adr + MPI::UNSIGNED_LONG.Get_size()*2) <<std::endl;
      std::cout << "  -- colshape_adr = "<< (colshape_adr) <<" byteoffset = "<<(colshape_adr-start_adr) <<std::endl;
      std::cout << "---- sizeof( size_t )="<<sizeof( size_t )<<"  sizeof( MPI::UNSIGNED_LONG )="<< MPI::UNSIGNED_LONG.Get_size() <<std::endl;
      */
      int count                           =   6;
      int array_of_blocklengths[6]        = { 1                      ,   1                      ,   1                      ,   1                      ,   3                      ,   MAX_ADDITIONAL_DIMS    };
      MPI::Aint array_of_displacements[6] = { type_adr-start_adr     ,   msg_adr-start_adr      ,   id_adr-start_adr       ,   per_adr-start_adr      ,   shape_adr-start_adr    ,   colshape_adr-start_adr };
      MPI::Datatype array_of_types[6]     = { MPI::UNSIGNED          ,   MPI::UNSIGNED          ,   MPI::UNSIGNED_LONG     ,   MPI::BOOL              ,   MPI::UNSIGNED_LONG     ,   MPI::UNSIGNED_LONG     };
      
      header_type = MPI::Datatype::Create_struct(
                                    count                                   /// number of blocks
                                  , array_of_blocklengths                   /// number of elements in each block
                                  , array_of_displacements                  /// byte displacements of each block
                                  , array_of_types                          /// type of elements in each block
                                  );
      header_type.Commit();
      delete dummy;
    }
  }

std::string Header::get_data_type_str() const
{
  switch (_data_type)
  {
    case UNDEFINED_HEADER : return std::string( "UNDEFINED_HEADER" );
    case BOOL_HEADER :      return std::string( "BOOL_HEADER" );
    case INT_HEADER :       return std::string( "INT_HEADER" );
    case FLOAT_HEADER :     return std::string( "FLOAT_HEADER" );
    case STRING_HEADER :    return std::string( "STRING_HEADER" );
    case CPUDATA_HEADER :   return std::string( "CPUDATA_HEADER" );
    default:                return std::string( "UNKNOWN_HEADER" );
  }
}

std::string Header::get_msg_type_str() const
{
  switch (_msg_type)
  {
    case UNDEFINED_MSG : return std::string( "UNDEFINED_MSG" );
    case GET_MSG :       return std::string( "GET_MSG" );
    case SET_MSG :       return std::string( "SET_MSG" );
    case IO_MSG  :       return std::string( "IO_MSG" );
    case STATUS_MSG :    return std::string( "STATUS_MSG" );
    case STEP_MSG:       return std::string( "STEP_MSG" );
    case OUTPUT_MSG:     return std::string( "OUTPUT_MSG" );
    default:             return std::string( "UNKNOWN_MSG" );
  }
}

std::string Header::dump_dims( bool dump_addr ) const
{
	std::string result(" data_type = ");
	result.append( get_data_type_str());
	
	result.append( " , msg_type = " );
	result.append( get_msg_type_str() );
	
	result.append( " , id = " );
	result.append( boost::lexical_cast<std::string>( get_id() ) );
	
  result.append( " , persistent = " );
	result.append( boost::lexical_cast<std::string>( is_persistent() ) );
	result.append( " , [ " );
	
	result.append(boost::lexical_cast<std::string>(_shape[0])+" x " );
	result.append(boost::lexical_cast<std::string>(_shape[1])+" x " );
	result.append(boost::lexical_cast<std::string>(_shape[2])+" ]" );
  std::vector<size_t> col = get_column_shape_as_vector();
	if ( col.size() > 0 )
	  {
		result.append("(");
		result.append(boost::lexical_cast<std::string>(_shape[0])+" x " );
		result.append(boost::lexical_cast<std::string>(_shape[1])+" x " );
		for ( size_t i = 0; i < col.size(); ++i )
		  {
			result.append(boost::lexical_cast<std::string>(col[i]));
			if ( i < col.size() - 1 )
			  result.append(" x ");
		  }
		result.append(")");
	  }
      if ( dump_addr )
      {
          result.append(" ( addr = ");
          result.append(boost::lexical_cast<std::string>( this ) );
          result.append(" )");
      }
	return result;
}
