/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef HEADER_H
#define HEADER_H


#include <mpi.h>
#include <vector>
#include <string>

#include <boost/preprocessor/enum_params.hpp>
#include <boost/preprocessor/iteration/local.hpp>

#define MAX_ADDITIONAL_DIMS   6

class Header
{
public:
  typedef std::vector<size_t> ColumnShapeT;
  enum DATA_TYPE	{ UNDEFINED_HEADER  /// default non-initialized header
                  , BOOL_HEADER       /// bool value
                  , INT_HEADER        /// int value
                  , FLOAT_HEADER      /// float value
                  , STRING_HEADER     /// string value
                  , CPUDATA_HEADER    /// cpudata<float> value
                  };
                  
  enum MSG_TYPE   { UNDEFINED_MSG     /// default non-initialized header
                  , IO_MSG            /// standard inter-module data transfer
                  , GET_MSG           /// get_property request, algorihtm und property index are filled
                  , SET_MSG           /// set_property command, algorihtm und property index are filled
                  , STATUS_MSG        /// status update, usually data_type string
                  , STEP_MSG          /// used to wait for x frames to finish in main thread
                  , OUTPUT_MSG        /// reserved for gathering outputs
                  };
  
  Header()
  {
    clear();
  }
  Header( const Header& other )
  {
    _data_type = other.get_data_type();
    _msg_type  = other.get_msg_type();
    _id = other.get_id();
    _persistent = other.is_persistent();
    std::copy( other.shape(), other.shape()+3, _shape );
    std::copy( other.column_shape(), other.column_shape()+MAX_ADDITIONAL_DIMS, _column_shape );
  }
  ~Header( )
  {
    clear();
  }
  
  
  /// this one is unsave, expects sizes to be right size
  void set_column_shape( const size_t* sizes )
  {
    std::copy( sizes, sizes + MAX_ADDITIONAL_DIMS, _column_shape );
  }
  
  /// sets parameters to zero
  void clear()
  {
    _data_type = UNDEFINED_HEADER;
    _msg_type  = UNDEFINED_MSG;
    _id = 0;
    _persistent = false;
    _shape[0] = _shape[1] = _shape[2] = 0;
    reset_column_shape();
  }
  
  void set_column_shape( const ColumnShapeT & sizes );
  
  ColumnShapeT get_column_shape_as_vector() const;

  size_t* shape()
  {
	return _shape;
  }
  
  const size_t* shape() const
  {
	return _shape;
  }
  
  size_t* column_shape()
  {
	return _column_shape;
  }
  
  const size_t* column_shape() const
  {
	return _column_shape;
  }

  void       set_data_type( DATA_TYPE type ) { _data_type = type; };
  DATA_TYPE  get_data_type() const { return _data_type; };
  DATA_TYPE* data_type(){ return &_data_type; };
  std::string get_data_type_str() const;

  MSG_TYPE   get_msg_type() const { return _msg_type; };
  MSG_TYPE*  msg_type(){ return &_msg_type; };
  void       set_msg_type( MSG_TYPE type ) { _msg_type = type; };
  std::string get_msg_type_str() const;
  
  size_t*    id(){ return &_id; };
  size_t     get_id() const { return _id; };
  void       set_id( size_t id ){ _id = id; };
  
  bool*      persistent(){ return &_persistent; };
  bool       is_persistent() const { return _persistent; };
  void       set_persistent( bool persistent = true ){ _persistent = persistent; };
  
  
  static MPI::Datatype get_type()
  {
    commit();
    return header_type;
  }
  
  std::string dump_dims( bool dump_addr = false ) const;
  
    /// generates set_column_shape( n1, n2, ... , nD ) for up to six dimensions
#define BOOST_PP_LOCAL_LIMITS (1,MAX_ADDITIONAL_DIMS)
#define BOOST_PP_LOCAL_MACRO(I)									\
  void set_column_shape( BOOST_PP_ENUM_PARAMS(I, size_t n) )	\
  {																\
    reset_column_shape();                           \
    size_t N[I] = { BOOST_PP_ENUM_PARAMS(I,n) };		\
	  std::copy( N, N + I, _column_shape );						\
  }
#include BOOST_PP_LOCAL_ITERATE()
  
  
protected:

  void reset_column_shape()
  {
    for ( size_t i = 0; i < MAX_ADDITIONAL_DIMS; ++i)
      _column_shape[i] = 1;
  }


  static void commit( );
  static MPI::Datatype header_type;
  
  
  DATA_TYPE _data_type;
  MSG_TYPE  _msg_type;
  size_t    _id;
  bool      _persistent;
  size_t    _shape[3];
  size_t    _column_shape[MAX_ADDITIONAL_DIMS];
  
};


#endif
