/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
//-*-  tab-width:4  -*-
#include "HostImage.h"

#include <cstdio>
#include <iostream>
#include <algorithm>

#include <png.h>
//#include <Magick++.h>

#include <boost/format.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <IL/il.h>
#include <IL/ilu.h>
//#include <IL/ilut.h>

//#include <GL/glut.h>

#ifdef _WIN32
const float HostImage::Kb = 0.114;
const float HostImage::Kr = 0.299;
#endif
#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

boost::recursive_mutex _devil_mutex;
typedef boost::lock_guard< boost::recursive_mutex > devil_lock_t;

void HostImage::initWithZeros()
{
  if ( !_data_float.empty() )   std::fill( _data_float.begin(), _data_float.end(), float(0) );
  if ( !_data_ubyte.empty() )   std::fill( _data_ubyte.begin(), _data_ubyte.end(), ubyte(0) );
}

void HostImage::make_unique()
{
  _data_float.make_unique();
  _data_ubyte.make_unique();
}

HostImage::HostImage( const BData & data, ColorSpace S )
  :  _data_ubyte(data), _type(UBYTE), _space(S), _frame_number(-1), _time_stamp(-1)
{  
  if ( data.depth() == 1 )
	_space = LUM;
  else if ( data.depth() == 4 ) { 
	_space = RGBA;  
  } else if ( data.depth() != 3 && !data.empty() )
	throw std::runtime_error(str(boost::format("invalid depth for hostimage: %i")%data.depth()));
	
}

HostImage::HostImage( const FData & data, ColorSpace S )
  : _data_float(data), _type(FLOAT), _space(S), _frame_number(-1), _time_stamp(-1)
{  
  if ( data.depth() == 1 )
	_space = LUM;
  else if ( data.depth() == 4 ) { 
	_space = RGBA;  
  } else if ( data.depth() != 3 && !data.empty() )
	throw std::runtime_error(str(boost::format("invalid depth for hostimage: %i")%data.depth()));	
}

HostImage::HostImage(const char* fileName, DataType T, ColorSpace S )
 :_type(T)
 , _space(S), _frame_number(-1), _time_stamp(-1)
{
  read(fileName);
  if ( T == FLOAT )
	{
	  convert_to_type( T );
	  convert_to_space( S );
	}
  else
	{
	  convert_to_space( S );
	  convert_to_type( T );
	}
}

HostImage::HostImage(const char* fileName, DataType T )
 :_type(T)
 ,_space(HostImage::DEFAULT_SPACE), _frame_number(-1), _time_stamp(-1)
{
  read(fileName);
  convert_to_type( T );
}

HostImage::HostImage(const char* fileName, ColorSpace S )
:_type(HostImage::DEFAULT_TYPE), _frame_number(-1), _time_stamp(-1)
{
  read(fileName);
  convert_to_space( S );
}

HostImage::HostImage( const char* fileName )
:_type(HostImage::DEFAULT_TYPE)
,_space(HostImage::DEFAULT_SPACE), _frame_number(-1), _time_stamp(-1)
{
  read(fileName);
}

HostImage::HostImage( const std::string& fileName )
:_type(HostImage::DEFAULT_TYPE)
,_space(HostImage::DEFAULT_SPACE)
{
  read(fileName.c_str());
}

HostImage::HostImage( DataType T, ColorSpace S)
 :_type(T)
 , _space(S), _frame_number(-1), _time_stamp(-1)
{
	;
}

HostImage::HostImage(int wA, int hA, DataType T, ColorSpace S)
  : _type(T)
  , _space(S), _frame_number(-1), _time_stamp(-1)
{
	resize(wA,hA);
}

const float* HostImage::getPointerToImage() const
{
  return _data_float.data();
}

float* HostImage::getPointerToImage()
{
  return _data_float.data();
}

const unsigned char* HostImage::getBytePointerToImage() const
{
  return _data_ubyte.data();
}

unsigned char* HostImage::getBytePointerToImage()
{
  return _data_ubyte.data();
}

#define CHECK_DATA_RANGE(data,n)										\
  {																		\
	size_t _n = data.rows() * data.columns();							\
	if ( _n <= n )														\
	  throw std::runtime_error(											\
		str( boost::format( "Tried to access %s element %d of %d!" )	\
			 % BOOST_PP_STRINGIZE(data) % n % _n ) );					\
  }

float& HostImage::getF( size_t n )
{
  CHECK_DATA_RANGE( _data_float, n );
  return _data_float.data()[n];
}

const float& HostImage::getF( size_t n ) const
{
  CHECK_DATA_RANGE( _data_float, n );
  return _data_float.data()[n];
}

HostImage::ubyte& HostImage::getU( size_t n )
{
  CHECK_DATA_RANGE( _data_ubyte, n );
  return _data_ubyte.data()[n];
}

const HostImage::ubyte& HostImage::getU( size_t n ) const
{
  CHECK_DATA_RANGE( _data_ubyte, n );
  return _data_ubyte.data()[n];
}


CPUData<float>& HostImage::cpu_data_float( bool convert, bool flip )
{
    if ( convert )
	  convert_to_type( FLOAT, flip );
	else if ( _type != FLOAT )
	  throw std::runtime_error(str(boost::format("Error! Tried to access empty datafield \"_data_float\" file: \"%s\".\n")%_fileName));
	return _data_float;
}

const CPUData<float>& HostImage::cpu_data_float() const
{
	if ( _type != FLOAT ) 	 
	  throw std::runtime_error(str(boost::format("Error! Tried to access empty datafield \"_data_float\" file: \"%s\".\n")%_fileName));
	return _data_float;
}

CPUData<HostImage::ubyte>& HostImage::cpu_data_ubyte( bool convert, bool flip )
{
    if ( convert )
	  convert_to_type( UBYTE, flip );
	else if ( _type != UBYTE )
	  throw std::runtime_error(str(boost::format("Error! Tried to access empty datafield \"_data_ubyte\" file: \"%s\".\n")%_fileName));
	return _data_ubyte;
}
const CPUData<HostImage::ubyte>& HostImage::cpu_data_ubyte() const
{
	if ( _type != UBYTE )
	  throw std::runtime_error(str(boost::format("Error! Tried to access empty datafield \"_data_ubyte\" file: \"%s\".\n")%_fileName));
	return _data_ubyte;
}

bool HostImage::isImagePresent() const
{
  switch (_type)
	{
	case FLOAT: return !_data_float.empty();
	case UBYTE: return !_data_ubyte.empty();
	default: throw std::runtime_error( str( boost::format( "invalid data type %d in HostImage instance" ) % _type ) );
	}
}

void HostImage::crop( size_t row_number, size_t row_count, size_t col_number, size_t col_count  )
{
  if (_type==HostImage::FLOAT)
  {
    CPUData<float> tmp( _data_float );
    _data_float.reference( tmp, row_number, row_count, col_number, col_count );
  }
  else
  {
    CPUData<HostImage::ubyte> tmp( _data_ubyte );
    _data_ubyte.reference( tmp, row_number, row_count, col_number, col_count );
  }
}


void HostImage::freeImage()
{
	_data_float.clear();
	_data_ubyte.clear();
}

unsigned HostImage::getWidth() const
{
  	switch (_type)
	{
	case FLOAT: return _data_float.columns();
	case UBYTE: return _data_ubyte.columns();
	default: throw std::runtime_error( str( boost::format( "invalid data type %d in HostImage instance" ) % _type ) );
	}
}

unsigned HostImage::getHeight() const
{
  	switch (_type)
	{
	case FLOAT: return _data_float.rows();
	case UBYTE: return _data_ubyte.rows();
	default: throw std::runtime_error( str( boost::format( "invalid data type %d in HostImage instance" ) % _type ) );
	}
}

void HostImage::getSize(int &w, int &h) const
{
  w = getWidth();
  h = getHeight();
}


void check_il_error( const std::string & msg )
{
  devil_lock_t lock_guard( _devil_mutex );

  ILenum err = ilGetError();
  if ( err != IL_NO_ERROR )
	{
	  const char * errstr = iluErrorString( err );
	  throw std::runtime_error( str( boost::format( "error using DevIL: \"%s\" msg: \"%s\"" )
									 % errstr % msg ) );
	}
}

void HostImage::check_prepare()
{
  static bool il_initialized = false;
  if ( !il_initialized )
	{
	  devil_lock_t lock_guard( _devil_mutex );
	  //std::cout << " initializing DevIL " << std::endl;
	  ilInit();
	  iluInit();
	  check_il_error( "initilizing" );
	  il_initialized = true;
	}
}

void HostImage::read(const char* fileName)
{
  devil_lock_t lock_guard( _devil_mutex );

  check_prepare();

  freeImage();

  // init devIL
  ILuint texid;


  // create a new Image
  ilGenImages(1, &texid);
  check_il_error( "requesting image id" );
  ilBindImage(texid);

  try
  {

    // try to load image
	ilLoadImage( (char*) fileName );
    check_il_error( "loading image" );

    if ( std::string( fileName ).find( ".bmp" ) != std::string::npos )
        iluFlipImage();

    // copy the image to an IL image
    int il_space = (int)ilGetInteger(IL_IMAGE_FORMAT);

   // std::cout << "il_space="<<il_space<<std::endl;
    switch (il_space)
    {
    case IL_LUMINANCE:
      {
		resize( (int)ilGetInteger(IL_IMAGE_WIDTH), (int)ilGetInteger(IL_IMAGE_HEIGHT), UBYTE, LUM );
		ilCopyPixels(           0, //ILuint XOff,
								0, //ILuint YOff,
								0, //ILuint ZOff,
								getWidth(), //ILuint Width,
								getHeight(), //ILuint Height,
								1, //ILuint Depth,
								IL_LUMINANCE, //ILenum Format,
								IL_UNSIGNED_BYTE, //ILenum Type,
								_data_ubyte.data());
		break;
      }
    case IL_BGR:
    case IL_BGRA:
    case IL_RGBA:
	  {
		  resize( (int)ilGetInteger(IL_IMAGE_WIDTH), (int)ilGetInteger(IL_IMAGE_HEIGHT), UBYTE, RGBA );
		  ilCopyPixels(           0, //ILuint XOff,			
		  							0, //ILuint YOff,
		  							0, //ILuint ZOff,
		  							getWidth(), //ILuint Width,
		  							getHeight(), //ILuint Height,
		  							1, //ILuint Depth,
		  							IL_RGBA, //ILenum Format,
		  							IL_UNSIGNED_BYTE, //ILenum Type,
		  							_data_ubyte.data());
	  }
    case IL_RGB:
      {
		resize( (int)ilGetInteger(IL_IMAGE_WIDTH), (int)ilGetInteger(IL_IMAGE_HEIGHT), UBYTE, RGB );
		ilCopyPixels(           0, //ILuint XOff,
								0, //ILuint YOff,
								0, //ILuint ZOff,
								getWidth(), //ILuint Width,
								getHeight(), //ILuint Height,
								1, //ILuint Depth,
								IL_RGB, //ILenum Format,
								IL_UNSIGNED_BYTE, //ILenum Type,
								_data_ubyte.data());
		break;
      }
    default:
      throw std::runtime_error( str( boost::format( "invalid color space %d in HostImage instance" ) % il_space ) );
    }


  }
  catch ( const std::exception & e )
  {
    throw std::runtime_error( str(boost::format("Tried to load image \"%s\" and failed: \"%s\".")%fileName%e.what()) );
  }

   ilDeleteImages(1, &texid);
}




void HostImage::resize(int wA, int hA, int type, int space )
{
  _type = type != -1 ? DataType(type) : _type;
  _space = space != -1 ? ColorSpace(space) : _space;
  switch (_space)
	{
	case LUM:
	  {
		if (_type==FLOAT) { _data_ubyte.clear(); _data_float.resize(hA,wA,1); }
		else			  { _data_float.clear(); _data_ubyte.resize(hA,wA,1); }
		break;
	  }
	case YCbCr:
	case RGB:
	  {
		if (_type==FLOAT) { _data_ubyte.clear(); _data_float.resize(hA,wA,3); }
		else			  { _data_float.clear(); _data_ubyte.resize(hA,wA,3); }
		break;
	  }
	case RGBA:
	  {
		if (_type==FLOAT) { _data_ubyte.clear(); _data_float.resize(hA,wA,4); }
		else			  { _data_float.clear(); _data_ubyte.resize(hA,wA,4); }
		break;
	  }

	}
  make_unique();
}


void HostImage::convert_to_type( DataType T, bool flip )
{
	if (T==_type) 
		if (!flip)
			return;
		else
		{
			_data_float.flip( false, true );
			_data_ubyte.flip( false, true );
			return;
		}


	switch (T)
	{
	case UBYTE:
	  {
		if ( !empty() )
		  {

			if ( _type != FLOAT )
			  throw(std::runtime_error("ERROR! Something's wrong here; Type is neither UBYTE nor FLOAT !?!\n") );

			int hA = getHeight();
			int wA = getWidth();
			
			FData fdata = cpu_data_float();

			resize( wA, hA, UBYTE, _space );
      
      int i1 = 0;
			int i1_end = hA;
			int i2 = flip ? hA - 1 : 0;
			int i1_step = 1;
			int i2_step = flip ? -1 : 1;
			for ( ; i1 != i1_end; i1 += i1_step, i2 += i2_step )
		  	for ( size_t j = 0; j < wA; ++j )
				{
			  	FData::ColumnT fData = fdata( i1, j );
			  	BData::ColumnT uData = _data_ubyte( i2, j );
			  	if (_space==LUM)
						uData[0] = float( fData[0] *255 );
					else if (_space==RGB || _space==YCbCr) {
						uData[0] = float( fData[0] *255);
						uData[1] = float( fData[1] *255);
						uData[2] = float( fData[2] *255);
					} else {
						uData[0] = float( fData[0] *255);
						uData[1] = float( fData[1] *255);
						uData[2] = float( fData[2] *255);
						uData[3] = float( fData[3] *255);					
					}
				}
	  	
      
      /*
			BData::iterator uData = _data_ubyte.begin(), uEnd = _data_ubyte.end();
			FData::iterator fData = fdata.begin(), fEnd = fdata.end();
			for (; uData != uEnd && fData != fEnd; ++uData, ++fData )
			  {				
				if (_space==LUM)
				  (*uData)[0] = ubyte( (*fData)[0] *255 );
				else if (_space==RGB || _space==YCbCr) {
				  (*uData)[0] = ubyte( (*fData)[0] *255);
				  (*uData)[1] = ubyte( (*fData)[1] *255);
				  (*uData)[2] = ubyte( (*fData)[2] *255);
				} else {
				  (*uData)[0] = ubyte( (*fData)[0] *255);
				  (*uData)[1] = ubyte( (*fData)[1] *255);
				  (*uData)[2] = ubyte( (*fData)[2] *255);	
				  (*uData)[3] = ubyte( (*fData)[3] *255);					
				}
			  }
			*/

		  }
		break;
	  }

	case FLOAT:
	  {
		
		if ( !empty() )
		  {
			
			if ( _type != UBYTE )	
			  throw(std::runtime_error("ERROR! Something's wrong here; Type is neither UBYTE nor FLOAT !?!\n") );

			int hA = getHeight();
			int wA = getWidth();

			BData bdata = cpu_data_ubyte();
			
			resize( wA, hA, FLOAT, _space );
			
			int i1 = 0;
			int i1_end = hA;
			int i2 = flip ? hA - 1 : 0;
			int i1_step = 1;
			int i2_step = flip ? -1 : 1;
			const float scale = 1.f / 255.f;
			
			switch ( _space )
			{
				case LUM:
				{
					//BData::ColumnT uData;
					//FData::ColumnT fData;
					for ( ; i1 != i1_end; i1 += i1_step, i2 += i2_step )
					for ( size_t j = 0; j < wA; ++j )
					{
						_data_float( i2, j, 0 ) = static_cast<float>( bdata( i1, j, 0 ) ) * scale;
						//BData::ColumnT uData = bdata( i1, j );
						//FData::ColumnT fData = _data_float( i2, j );
						//fData[0] = static_cast<float>( uData[0] ) * scale;
					}
					break;
				}
				case RGB:
				case YCbCr:
				{
					for ( ; i1 != i1_end; i1 += i1_step, i2 += i2_step )
					for ( size_t j = 0; j < wA; ++j )
					{
						BData::ColumnT uData = bdata( i1, j );
						FData::ColumnT fData = _data_float( i2, j );
						fData[0] = float( uData[0] * scale);
						fData[1] = float( uData[1] * scale);
						fData[2] = float( uData[2] * scale);
					}
					break;
				}
				default:
				{
					for ( ; i1 != i1_end; i1 += i1_step, i2 += i2_step )
					for ( size_t j = 0; j < wA; ++j )
					{
						BData::ColumnT uData = bdata( i1, j );
						FData::ColumnT fData = _data_float( i2, j );
						fData[0] = float( uData[0] * scale);
						fData[1] = float( uData[1] * scale);
						fData[2] = float( uData[2] * scale);
						fData[3] = float( uData[3] * scale);
					}
					break;
				}
			}
			
	  	
		  }

		break;
	  }
	default: throw std::runtime_error( str( boost::format( "HostImage invalid data type %s" ) % T ) );
	}

	_type = T;
}


void HostImage::convert_to_space( ColorSpace S )
{

	if (S == _space) return;

	if ( _type != FLOAT )
	  convert_to_type(FLOAT);


	int hA = getHeight();
	int wA = getWidth();
	switch (S)
	  {

	  case LUM:
		{

		  FData data_tmp( hA, wA, 1 );
		  FData::iterator itmp = data_tmp.begin(), iEnd = data_tmp.end();
		  FData::iterator gData = _data_float.begin();
		  for (; itmp != iEnd; ++itmp, ++gData )
			{
			  //  RGB -> LUM
			  if (_space==RGB || _space==RGBA) (*itmp)[0] = 0.3f * (*gData)[0] + 0.59f * (*gData)[1] + 0.11f * (*gData)[2];
			  // YCbCr -> LUM
			  else 			 (*itmp)[0] = (*gData)[0];
			}

		  _data_float.reference( data_tmp );
		  break;
		}

	  case HSV:
		{
		  FData data_tmp( hA, wA, 3 );
		  FData::iterator itmp = data_tmp.begin(), iEnd = data_tmp.end();
		  FData::iterator gData = _data_float.begin();
		  if ( _space == LUM )
			{
			  for (; itmp != iEnd; ++itmp, ++gData )
				{
				  float & H = (*itmp)[0], & S = (*itmp)[1], & V = (*itmp)[2];
				  H = 0;
				  S = 0;
				  V = (*gData)[0];
				}
			}
		  else if ( _space == RGB || _space == RGBA )
			{
			  for (; itmp != iEnd; ++itmp, ++gData )
				{
				  float & H = (*itmp)[0], & S = (*itmp)[1], & V = (*itmp)[2];
				  float   R = (*gData)[0],  G = (*gData)[1],  B = (*gData)[2];
				  float mx = std::max( std::max( R, G ), B );
				  float mn = std::min( std::min( R, G ), B );
				  float d = mx - mn;
				  if ( d == 0 )
					H = 0;
				  else if ( mx == R )
					H = M_PI / 3.0f * ( 0 + ( G - B ) / d );
				  else if ( mx == G )
					H = M_PI / 3.0f * ( 2 + ( B - R ) / d );
				  else if ( mx == B )
					H = M_PI / 3.0f * ( 4 + ( R - G ) / d );
				  if ( H < 0 )
					H += 2 * M_PI;
				  if ( mx == 0 )
					S = 0;
				  else
					S = d / mx;
				  V = mx;
				}			  
			}
		  else
			throw std::runtime_error("can only convert to hsv from luminance or rgb");
		  _data_float.reference(data_tmp);
		  break;
		}

	  case RGB:
		{
		  FData data_tmp( hA, wA, 3 );
		  FData::iterator itmp = data_tmp.begin(), iEnd = data_tmp.end();
		  FData::iterator gData = _data_float.begin();
		  for (; itmp != iEnd; ++itmp, ++gData )
			{

			  float & R = (*itmp)[0], & G = (*itmp)[1], & B = (*itmp)[2];

			  // LUM -> RGB
			  if (_space==LUM)
				{
				  float lum = (*gData)[0];
				  R = lum; //0.30f * lum;
				  G = lum; //0.59f * lum;
				  B = lum; //0.11f * lum;
				}
			  // YCbCr -> RGB
			  else if (_space==YCbCr)
				{
				  // 1  -1.2189e-06  1.402
				  // 1  -3.4414e-01  -7.1414e-01
				  // 1  1.772        4.0630e-07
				  float  Y = (*gData)[0], Cb = (*gData)[1], Cr = (*gData)[2];
				  R = Y - 1.2189e-06 * Cb + 1.4020e-00  * Cr;
				  G = Y - 3.4414e-01 * Cb - 7.1414e-01  * Cr;
				  B = Y + 1.7720e-00 * Cb + 4.0630e-07  * Cr;
				}
			  // RGBA -> RGB
			  else if (_space==RGBA)
                {
				  R = (*gData)[0];
				  G = (*gData)[1];
				  B = (*gData)[2];
			    }
               else if (_space==YUV422)
               {
                  
                  
               }
               else
                throw std::runtime_error( "cannot convert types" );
			}

		  _data_float.reference( data_tmp );
		  break;
		}

	  case RGBA:
		{
		  FData data_tmp( hA, wA, 4 );
		  FData::iterator itmp = data_tmp.begin(), iEnd = data_tmp.end();
		  FData::iterator gData = _data_float.begin();
		  for (; itmp != iEnd; ++itmp, ++gData )
			{

			  float & R = (*itmp)[0], & G = (*itmp)[1], & B = (*itmp)[2];
			  (*itmp)[3] = 1.0f;

			  // LUM -> RGB(A)
			  if (_space==LUM)
				{
				  float lum = (*gData)[0];
				  R = 0.30f * lum;
				  G = 0.59f * lum;
				  B = 0.11f * lum;
				}
			  // YCbCr -> RGB(A)
			  else if (_space==YCbCr)
				{
				  // 1  -1.2189e-06  1.402
				  // 1  -3.4414e-01  -7.1414e-01
				  // 1  1.772        4.0630e-07
				  float  Y = (*gData)[0], Cb = (*gData)[1], Cr = (*gData)[2];
				  R = Y - 1.2189e-06 * Cb + 1.4020e-00  * Cr;
				  G = Y - 3.4414e-01 * Cb - 7.1414e-01  * Cr;
				  B = Y + 1.7720e-00 * Cb + 4.0630e-07  * Cr;
				}
			  // RGB -> RGB(A)
			  else {
				  R = (*gData)[0];
				  G = (*gData)[1];
				  B = (*gData)[2];
			  }
			}

		  _data_float.reference( data_tmp );
		  break;
		}

	  case YCbCr:
		{
		  FData data_tmp;
		  data_tmp.resize(hA,wA,3);
		  FData::iterator itmp = data_tmp.begin(), iEnd = data_tmp.end();
		  FData::iterator gData = _data_float.begin();

		  for ( ; itmp != iEnd; ++itmp, ++gData )
			{
			  float & Y = (*itmp)[0], & Cb = (*itmp)[1], & Cr = (*itmp)[2];
			  // LUM -> YCbCr
			  if (_space==LUM)
				{
				  float lum = (*gData)[0];
				  Y = lum;
				  Cb = Cr = 0.0f;
				}
			  // (RGB | RGBA) -> YCbCr
			  else
				{// 0.299, 0.587, 0.114;   -0.168736 , -0.331264, 0.5;   0.5, -0.418688, -0.081312
				  float  R = (*gData)[0], G = (*gData)[1], B = (*gData)[2];
				  Y  =  0.299000 * R + 0.587000 * G + 0.114000 * B;
				  Cb = -0.168736 * R - 0.331264 * G + 0.500000 * B;
				  Cr =  0.500000 * R - 0.418688 * G - 0.081312 * B;
				  //Y =     Kr  * R   + (1 - Kr - Kb) * G   +  Kb    * B
				  //Cb =   -Kr  * R   - (1 - Kr - Kb) * G   + (1-Kb) * B
				  //Cr = (1-Kr) * R   - (1 - Kr - Kb) * G   -  Kb    * B

				}
			}
		  _data_float.reference(data_tmp);
		  break;
		}
	  }

	_space=S;
}


void HostImage::writePNG(const char* fileName, bool flip)
{
  std::cout << "Warning in HostImage::writePNG(): this function is depreciated, please use writeImage() instead."<<std::endl;
  writeImage(fileName, flip);
}

void HostImage::writeImage(const char* fileName, bool flip)
{
  if ( _type != UBYTE )
	{
	  HostImage tmp(*this);
	  tmp.convert_to_type( UBYTE );
	  tmp.writeImage( fileName );
	  return;
	}


  devil_lock_t lock_guard( _devil_mutex );

  check_prepare();

  ILuint texid;
  ILboolean success = false;

  // create a new Image
#define CHECK_IL_ERROR													\
  {																		\
	  ILenum err = ilGetError();										\
	  if (err != IL_NO_ERROR)											\
	  {																	\
		  throw std::runtime_error( str( boost::format("%s(%i) DevIL error (%i): %s") \
		                                % __FILE__ % __LINE__ % err % iluErrorString(err) ) ); \
	  }																	\
  }																		\

  ilGenImages(1, &texid);
  
	CHECK_IL_ERROR;
	
  ilBindImage(texid);
  
	CHECK_IL_ERROR;

  try
	{

	 // std::cout << " texing image width " << ILuint(getWidth()) << " height " << ILuint(getHeight()) << std::endl;
	  // copy the image to an IL image
	  switch (_space)
		{
		case LUM:
		  {
			success = ilTexImage( ILuint(getWidth()), ILuint(getHeight()),
								  1, //depth
								  1, //bytes per pixel
								  IL_LUMINANCE, // data format
								  IL_UNSIGNED_BYTE, // data type
								  NULL//_data_ubyte.data() // already available data
								  );
			break;
		  }
		case (RGB):
		  {
			success = ilTexImage( ILuint(getWidth()),(getHeight()),
								  1, //depth
								  3, //bytes per pixel
								  IL_RGB, // data format
								  IL_UNSIGNED_BYTE, // data type
								  NULL//_data_ubyte.data() // already available data
								  );
			break;
		  }
		case (RGBA):
		  {
			success = ilTexImage( ILuint(getWidth()),(getHeight()),
								  1, //depth
								  4, //bytes per pixel
								  IL_RGBA, // data format
								  IL_UNSIGNED_BYTE, // data type
								  NULL//_data_ubyte.data() // already available data
								  );
			break;
		  }
		case (YCbCr):
		  {
			throw(std::runtime_error("Writing of YCbCr not implemented. convert to RGB and retry.\n"));
			break;
		  }
		default:
		  throw std::runtime_error( str( boost::format( "invalid color space %d in HostImage instance" ) % _space ) );
		}

	CHECK_IL_ERROR;

    success = ilSetData( _data_ubyte.data() );
    if (flip) iluFlipImage();
    
	CHECK_IL_ERROR;


	  //write the Image to disk
	  if (success)
		{
		  //std::cout << "\n saving image to " << fileName << " .\n";
		  ilEnable(IL_FILE_OVERWRITE);
		  ilSaveImage((char*)fileName);
		}
	  else
		{
			throw std::runtime_error( str( boost::format("Could not save image to: \"%s\"") % fileName ) );
		}

	}
  catch ( std::runtime_error & e )
	{
	  throw std::runtime_error( str(boost::format("Tried to write image to \"%s\" and failed. what: %s ") % fileName % e.what() )  );
	}
	
	CHECK_IL_ERROR;
  
  ilDeleteImages(1, &texid);
  
	CHECK_IL_ERROR;


#undef CHECK_IL_ERROR


}


void HostImage::deleteShadows()
{
	// make sure its FLOAT and YCbCr
	if (_type!=HostImage::FLOAT) convert_to_type(HostImage::FLOAT);
	if (_space==HostImage::LUM)	throw(std::runtime_error("Cannot delete shadows from grayscale image!"));
	if (_space!=HostImage::YCbCr) convert_to_space(HostImage::YCbCr);
	
	int h, w;
	getSize(h,w);

	CPUData<float> temp;
	temp.resize(w,h,1);

	CPUData<float>::iterator iter1 = cpu_data_float().begin();
	CPUData<float>::iterator iter2 = temp.begin();
	for (;iter1!=cpu_data_float().end() && iter2!=temp.end(); iter1++, iter2++)
	{
		(*iter2)[0] = ( (*iter1)[1] + (*iter1)[2] ) / 2.f;
	}


	cpu_data_float().copy_from( temp );
	_space = HostImage::LUM;
}

