/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef HOSTIMAGE_H_
#define HOSTIMAGE_H_

#include <string>
#include <CPUData.hpp>

/**
 * This class stores an Image of CPUData. In Addition, the DataType
 * (either UBYTE or FLOAT) and the color space (LUM, RGB or YCbCr)
 * is being stored. Images can be written and loaded according to
 * the specified Type and Space. Additionally, conversion is possible
 * (Note that conversion to LUM is lossy, of course).
 * For convenience the default space is grayscale (LUM) and type is FLOAT.
 */



class
#ifdef _WIN32
__declspec(dllexport)
#endif
HostImage
{
public:
	typedef unsigned char ubyte;

	enum DataType	{ UBYTE, FLOAT };
	enum ColorSpace { LUM, RGB, YCbCr, RGBA, HSV, YUV422 };
	static const DataType   DEFAULT_TYPE  = UBYTE;
	static const ColorSpace DEFAULT_SPACE = RGB;
	// conversion parameters for YCbCr, according to standard ITU-R BT.601, for st. 709 set Kr = 0.2126; Kb = 0.0722;
	// Note that conversions btw RBG and YCbCr are hardcoded, so a change of parameters will not influence conversion.
#ifndef _WIN32
	static const float Kb=0.114; // generates C2864 in visual studio
	static const float Kr=0.299; // generates C2864 in visual studio
#else
	static const float Kb;	// initialize is now done in HostImage.cpp
	static const float Kr;	// initialize is now done in HostImage.cpp
#endif

	typedef CPUData<float> FData;
	typedef CPUData<ubyte> BData;

protected:

	FData _data_float;
	BData _data_ubyte;
	std::string _fileName;

public:

	DataType _type;
	ColorSpace _space;

	/**
	 * free all allocated memory and sets the image size (w & h) to 0
	 */
	void freeImage();

public:

	/// makes sure one-time initializations are performed
	/// do this before you start using this class in a multithreaded context!
	static void check_prepare();

	/**
	 * returns true if an image has been loaded into this object or at least
	 * the memory for the image is allocated.
	 */
	bool isImagePresent() const;

	/**
	 * returns the width of this image (or 0 if non image is present)
	 */
	unsigned getWidth() const;

	/**
	 * returns the heigth of this image (or 0 if non image is present)
	 */
	unsigned getHeight() const;

	/**
	 * returns the size (width and heigth) of this image
	 */
	void getSize(int &w, int &h) const;

	/**
	 * returns the type of the stored data. Possiblilities are: UBYTE, FLOAT
	 */
	DataType getType() const {return _type;};

	/**
	 * returns the type of the stored data. Possiblilities are: LUM, RGB, YCbCr
	 */
	ColorSpace getSpace() const {return _space;};

	/**
	 * Returns the pointer to the first element of the image,
	 * which is stored in a row-major way (i = x+y*sizeX).
	 * const version
	 */
	const float* getPointerToImage() const;

	/**
	 * Returns the pointer to the first element of the image,
	 * which is stored in a row-major way (i = x+y*sizeX).
	 */
	float* getPointerToImage();
	
	/**
	 * Returns the pointer to the first element of the image,
	 * which is stored in a row-major way (i = x+y*sizeX).
	 * const version
	 */
	const unsigned char* getBytePointerToImage() const;

	/**
	 * Returns the pointer to the first element of the image,
	 * which is stored in a row-major way (i = x+y*sizeX).
	 */
	unsigned char* getBytePointerToImage();

	void convert_to_type( DataType T, bool flip = false );
	void convert_to_space( ColorSpace S );

	float& getF( size_t n );
	const float& getF( size_t n )const;
	ubyte& getU( size_t n );
	const ubyte& getU( size_t n )const;

	CPUData<float>& cpu_data(bool convert = false, bool flip = false){return cpu_data_float(convert);}
	const CPUData<float>& cpu_data() const{return cpu_data_float();}
	CPUData<float>& cpu_data_float(bool convert = false, bool flip = false);
	const CPUData<float>& cpu_data_float() const;
	CPUData<ubyte>& cpu_data_ubyte(bool convert = false, bool flip = false);
	const CPUData<ubyte>& cpu_data_ubyte() const;

	size_t numElements() const
	{
	  return _type == UBYTE ? _data_ubyte.size() : _data_float.size();
	}

	size_t numPixels() const
	{
	  return getWidth() * getHeight();
	}

  void crop( size_t row_number, size_t row_count, size_t col_number, size_t col_count  );

  HostImage( const BData & data, ColorSpace S=RGB );
  HostImage( const FData & data, ColorSpace S=RGB );

	/**
	 * default constructor, creates an empty image with no memory allocated.
	 */
	HostImage( DataType T=FLOAT, ColorSpace S=LUM );

	/**
	 * uses read to load the given image and enforce the given type/colorspace settings
	 */
	HostImage(const char* filename, DataType T, ColorSpace S);

	/**
	 * uses read to load the given image and enforce the given type/colorspace settings
	 */
	HostImage(const char* filename, DataType T);

	/**
	 * uses read to load the given image and enforce the given type/colorspace settings
	 */
	HostImage(const char* filename, ColorSpace S);

	/**
	 * uses read to load the given image using optimal colrspayce datatype settings
	 */
	explicit HostImage(const char* filename);
	
	/**
	 * uses read to load the given image using optimal colrspayce datatype settings
	 */
	explicit HostImage(const std::string& filename);

	/**
	 * creates an image of given size with uninitializes content but allocated
	 * memory.
	 */
	HostImage(int w, int h, DataType T=FLOAT, ColorSpace S=LUM);

	HostImage copy()
	{
	  HostImage result = *this;
	  result.make_unique();
	  return result;
	}

	/// ensures that the data referenced is not shared
	void make_unique();

	bool empty() const { return _type == FLOAT ? cpu_data_float().empty() : cpu_data_ubyte().empty(); }

	/**
	 * reads a png file with the given name using libpng.
	 * This function can handle only some simple color formats
	 * at the moment.
	 */
	void readPNG(const char* fileName);

  /**
  * arbitrary reader based on devIL
  */
  void read(const char* fileName);

  /**
  * discontinued function, use writeImage instead (writes all formats)
  */
  void writePNG(const char* fileName, bool flip = true);
  void writeImage(const char* fileName, bool flip = false);

	void initWithZeros();

	/**
	 * this function just reallocates enough memory, without initializing
	 * or preserving of image data.
	 */
	void resize(int w, int h, int type = -1, int space = -1 );

	void clear(){ freeImage(); }

	/**
	* Deletes the luminance infromation from Image (e.g for shadow cancellation)
	* Note that deletion form grayscale image is not possible.
	* Resulting image is stored as a luminance image (average of Cb and Cr)
	*/
	void deleteShadows();

	/**
	* prints information about this instance to stdout
	*/
	void dumpInfo();

	const std::string & fileName()const { return _fileName; }
	
	void setFrameNumber( int number ){ _frame_number=number; }
	int getFrameNumber() const { return _frame_number; }
	void setTimeStamp( int time ){ _time_stamp=time; }
	int getTimeStamp() const { return _time_stamp; }


protected:
	int size, sizeW, sizeH; // for fast boundschecking
	int _frame_number;
	int _time_stamp;


};

template<class ostream> ostream & operator<<( ostream & os, const HostImage & image )
{
  os << "\n==== HostImage \"" << image.fileName().c_str() << "\" ====\n";
  os << "    Width=" << image.getWidth() << "  Height="<< image.getHeight() << std::endl;
  os << "    Colorspace=";
  if (image.getSpace()==HostImage::RGB) os << "RGB";
  else if (image.getSpace()==HostImage::YCbCr) os << "YCbCr";
  else os << "LUM";
  os << std::endl;
  os << "    DataType=";
  if (image.getType()==HostImage::FLOAT) os << "FLOAT";
  else if (image.getType()==HostImage::UBYTE) os << "UBYTE";
  os << std::endl;  
  os << "================================" <<std::endl;
  return os;
}

#endif /*HOSTIMAGE_H_*/
