/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _IDXBYTEFILE_HPP_
#define _IDXBYTEFILE_HPP_

#include <fstream> 
#include <stdexcept> 
#include <algorithm>
#include <vector>
#include <boost/cstdint.hpp>

#ifdef WIN32
using boost::int8_t;
using boost::int16_t;
using boost::uint8_t;
using boost::uint32_t;
#endif

#include <boost/format.hpp>

template<typename T>
inline void byte_swap(T& x)
{
  char * px = reinterpret_cast<char*>(&x);
  size_t n = sizeof(T);
  std::reverse( px, px + n );
}

// no endian detection, assumes big endian file and little endian architecture
struct IDXFile
{

  enum TYPE { UBYTE = 8, BYTE = 9, INT16 = 0xb, INT32 = 0xc, FLOAT32 = 0xd, FLOAT64 = 0xe };

  IDXFile(std::string filename)
  : _file(filename.c_str(),std::ios_base::binary), _needs_byte_swap(true)
  {
    uint8_t byte=112;
	read_in( byte );
	if ( byte != 0 ) 
	  throw(std::runtime_error(str(boost::format("\"%s\" is not an idx file, byte[0] == %i != 0")%filename%int(byte))));
	read_in( byte );
	if ( byte != 0 )
	  throw(std::runtime_error(str(boost::format("\"%s\" is not an idx file, byte[1] == %i != 0")%filename%int(byte))));
	read_in( byte );
	if ( byte != UBYTE &&
		 byte != BYTE &&
		 byte != INT16 &&
		 byte != INT32 &&
		 byte != FLOAT32 &&
		 byte != FLOAT64  ) 
	  throw(std::runtime_error(str(boost::format("\"%s\" is not an idx file of known datatype")%filename)));
	_type = TYPE(byte);
	read_in( _rank );
	//std::cout << " rank = " << int(_rank) << " shape: ";
	_size = 1;
	if ( rank() > 0 )
	  {
		_shape.resize( _rank );
		for ( int i = 0; i < _rank; ++i )
		  {
			read_in( _shape[i] );
			_size *= _shape[i];
			//std::cout << _shape[i] << " ";
		  }
	  }
	//std::cout << std::endl;
	_data_start = _file.tellg();
  }

  void reset()
  {
	_file.seekg( _data_start );
  }
  
  size_t rank()
  {
	return _rank;
  }

  size_t size()
  {
	return _size;
  }

  TYPE type() { return _type; }

  size_t extent(size_t i)
  {
	return i < _rank ? _shape[i] : size() > 0 ? 1 : 0;
  }
  
  template<typename T> void read_elements(T*v,size_t n)
  {
	for ( size_t i = 0; i < n; ++i )
	  {
		read_element( v[i] );
	  }
  }

  template<typename T> void read_element(T&v)
  {
	switch (_type)
	  {
#define read_from(TI,T1) case TI: { T1 val; read_in(val); v = val; break; }
		read_from( UBYTE  , uint8_t );
		read_from( BYTE   , int8_t  );
		read_from( INT16  , int16_t );
		read_from( INT32  , int32_t );
		read_from( FLOAT32, float   );
		read_from( FLOAT64, double  );
#undef read_from
	  }
  }
  
  ~IDXFile()
  {
  }
  
private:  
  
  template<typename T> 
  void read_in( T & v )
  {
	if ( _file.eof() )
	  throw std::runtime_error("idx: tried to read beyond eof");
	char * pv = reinterpret_cast<char*>(&v);
	_file.read( pv, sizeof(T) );
	if ( _file.bad() )
	  throw std::runtime_error("idx: error reading element from file");
	if ( _needs_byte_swap ) byte_swap(v);
  }

  std::ifstream _file;
  uint8_t _rank;
  std::vector<uint32_t> _shape;
  TYPE _type;
  size_t _size;
  std::streampos _data_start;
  bool _needs_byte_swap;
};


#endif /* _IDXBYTEFILE_HPP_ */
