/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "SimpleHostImageSequence.h"
#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <boost/thread.hpp>
#include <boost/filesystem.hpp>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <algorithm>

#include "concurrent_queue.hpp"

class image_fetcher
{

public:

  image_fetcher( const std::string & pattern, int start, int end, float step = 1, size_t max_load = 10 )
	: _pattern( pattern )
	, _start( start )
	, _end( end )
	, _image_queue( max_load )
	, _done( false )
	, _step(step)
  {
	HostImage::check_prepare(); // do this before going multithreaded
	if ( _pattern.find("*") == std::string::npos )
	  {
      if ( _step == 0 )
      {
         _filepaths.push_back( _pattern );
      }
      else
      {
        for ( float index = _start; index <= _end; index += _step )
          {
            std::string filepath = str( boost::format( _pattern ) % int(index) );
            _filepaths.push_back( filepath );
          }
      }
	  }
	else // glob emulation
	  {
      boost::filesystem::path patternpath( _pattern );
      boost::filesystem::path dir_path = patternpath.parent_path();
      boost::filesystem::directory_iterator end_itr; // default construction yields past-the-end
      std::string regex_str( patternpath.filename() );
      boost::replace_all( regex_str, ".", "\\." );
      boost::replace_all( regex_str, "*", ".*" );
      boost::regex regex( regex_str );
	  int i = 0;
      for ( boost::filesystem::directory_iterator itr( dir_path );
          itr != end_itr;
			++itr  )
        {
		  if ( !boost::filesystem::is_directory( itr->status() ) )
			{
			  std::string filename = itr->filename();
			  if ( boost::regex_match( filename, regex ) )
				{
				  ++i;
				  if ( _step == 0 || i % int(_step) == 0 )
					_filepaths.push_back( itr->path().file_string() );
				  if ( _step == 0 ) break;
				}
			}
        }	
	  _filepaths.sort();
	  std::cout << "pattern " << _pattern << " globbed " << _filepaths.size() << " files " << std::endl;
      }
    _thread.reset( new boost::thread(boost::ref(*this)) );
  }

  // thread entry point
  void operator()()
  {
    int index = 0;
    if ( _step == 0 )
    {
      HostImage image( _filepaths.begin()->c_str() );
      while(true)
      {
        image.setFrameNumber( index++ );
        if ( !_image_queue.push( image ) )
          break;
      }
    }
    else
    {
      BOOST_FOREACH( const std::string & filepath, _filepaths )
        {
        HostImage image( filepath.c_str() );
        image.setFrameNumber( index++ );
        if ( !_image_queue.push( image ) )
          break;
        }
      _image_queue.terminate_out( );
    }
  }

  ~image_fetcher()
  {
	/*
	std::cout << "destroying image fetcher for " << _pattern 
			  << "[" << _start << "," << _step << "," << _end << "]"
			  << std::endl;
	*/
	_image_queue.terminate_in();
	_thread->join();
  }

  bool get_one( HostImage & image )
  {
	return _image_queue.pop( image );
  }

private:

  std::string _pattern;
  int _start, _end;
  concurrent_queue<HostImage> _image_queue;
  boost::scoped_ptr<boost::thread> _thread;
  std::list<std::string> _filepaths;
  bool _done;
  float _step;
};

SimpleHostImageSequence::SimpleHostImageSequence(const char* fileExpr, int start, int end, int numbersToLoadA, float frame_stepsizeA)
  : _pattern( fileExpr )
  , _start(start)
  , _end(end)
  , _n_prefetch(numbersToLoadA)
  , _step_size(frame_stepsizeA)
  , _image_available(false)
{
  reset();
}

void SimpleHostImageSequence::reset()
{
  _image_fetcher.reset();
  _image_fetcher.reset( new image_fetcher( _pattern, _start, _end, _step_size, _n_prefetch ) );  
}

SimpleHostImageSequence::~SimpleHostImageSequence()
{
}

bool SimpleHostImageSequence::hasNextImage() 
{
  if ( _image_available )
	return true;
  else
	return ( _image_available = _image_fetcher->get_one( _current_image ) );
}

HostImage SimpleHostImageSequence::getNextImage()
{
  if ( !hasNextImage() )
    throw std::runtime_error("SimpleHostImageSequence::getNextImage(): No Images left!"
							 " Please watch out not to access unavailable images next time ;-)");

  _image_available = false;
  return getCurrentImage();
}

HostImage SimpleHostImageSequence::getCurrentImage()
{
  return _current_image;
}

int SimpleHostImageSequence::getHeight() 
{
  if ( _current_image.empty() && !hasNextImage() ) return 0;
  return _current_image.getHeight();
}

int SimpleHostImageSequence::getWidth() 
{
  if ( _current_image.empty() && !hasNextImage() ) return 0;
  return _current_image.getWidth();
}

/*
void SimpleHostImageSequence::copyNextImage(GPUImage& gData)
{
	HostImage temp(getNextImage());
	gData.resize(temp.getWidth(),temp.getHeight());
	if ( temp.getType()==HostImage::FLOAT &&
		 temp.getSpace()==HostImage::LUM )
	{
		gData.copyFromHostMemory(temp.getPointerToImage());
	} else {
		temp.make_unique();
		temp.convert_to_type(HostImage::FLOAT);
		temp.convert_to_space(HostImage::LUM);
		gData.copyFromHostMemory(temp.getPointerToImage());
	}
}
*/
