/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "VideoImageSequence.h"
#include <string>
#include <boost/format.hpp>
//#include <iostream>

extern "C"
{
#ifdef HAVE_FFMPEG_LIB_PREFIXES
	#include <libavcodec/avcodec.h>
	#include <libavformat/avformat.h>
	#include <libswscale/swscale.h>
#else
	#include <avcodec.h>
	#include <avformat.h>
	#include <swscale.h>
#endif
}

VideoImageSequence::VideoImageSequence( const char* filename )
  : pFormatCtx(NULL)
  ,videoStream(0)
  ,pCodecCtx(NULL)
  ,pCodec(NULL)
  ,pFrame(NULL)
  ,pFrameRGB(NULL)
  ,numBytes(0)
  ,buffer(NULL)
  ,hasImageInBuffer(false)
  ,frame(0)
  ,frame_next(0)
  ,_path(filename)  
  ,_img_convert_ctx(NULL)
{

  frame_stepsize=1.f;
    // Register all formats and codecs
    av_register_all();

    // Open video file
	int err = 0;
    if( ( err = av_open_input_file(&pFormatCtx, filename, NULL, 0, NULL) ) != 0 )
        throw std::runtime_error(str(boost::format("Couldn't open file %s error: %i")%filename%err));

    // Retrieve stream information
    if(av_find_stream_info(pFormatCtx)<0)
        throw std::runtime_error(str(boost::format("Couldn't find stream information")));

    // Dump information about file onto standard error
    // dump_format(pFormatCtx, 0, filename, false);

    // Find the first video stream
    videoStream=-1;
    for( size_t i = 0; i < pFormatCtx->nb_streams; i++ )
	  {
        if ( pFormatCtx->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO )
        {
            videoStream = i;
            break;
        }
	  }

    if( videoStream == -1 )
        throw std::runtime_error( str( boost::format("Didn't find a video stream") ) );

    // Get a pointer to the codec context for the video stream
    // pCodecCtx = &pFormatCtx->streams[videoStream]->codec;
	pCodecCtx=pFormatCtx->streams[videoStream]->codec;

    // allocate a converter
	_img_convert_ctx = sws_getContext( pCodecCtx->width, pCodecCtx->height, 
									   pCodecCtx->pix_fmt, 
									   pCodecCtx->width, pCodecCtx->height, PIX_FMT_RGB24, SWS_BICUBIC, 
									   NULL, NULL, NULL);

	// Find the decoder for the video stream
    pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
    if(pCodec==NULL)
       throw std::runtime_error(str(boost::format("Codec not found")));


    // Inform the codec that we can handle truncated bitstreams -- i.e.,
    // bitstreams where frame boundaries can fall in the middle of packets
    if(pCodec->capabilities & CODEC_CAP_TRUNCATED)
        pCodecCtx->flags|=CODEC_FLAG_TRUNCATED;

    // Open codec
    if(avcodec_open(pCodecCtx, pCodec)<0)
        throw std::runtime_error(str(boost::format("Could not open codec")));

    // Hack to correct wrong frame rates that seem to be generated by some 
    // codecs
    //if(pCodecCtx->frame_rate>1000 && pCodecCtx->frame_rate_base==1)
    //   pCodecCtx->frame_rate_base=1000;

    // Allocate video frame
    pFrame=avcodec_alloc_frame();

    // Allocate an AVFrame structure
    pFrameRGB=avcodec_alloc_frame();
    if(pFrameRGB==NULL)
        throw std::runtime_error(str(boost::format("Could not allocate Frame")));

    // Determine required buffer size and allocate buffer
    numBytes=avpicture_get_size(PIX_FMT_RGB24, pCodecCtx->width,
        pCodecCtx->height);
    buffer=new uint8_t[numBytes];

    // Assign appropriate parts of buffer to image planes in pFrameRGB
    avpicture_fill((AVPicture *)pFrameRGB, buffer, PIX_FMT_RGB24,
        pCodecCtx->width, pCodecCtx->height);

    
}

VideoImageSequence::~VideoImageSequence()
{

    // Free the YUV frame
    av_free(pFrame);

    // Close the codec
    avcodec_close(pCodecCtx);

    // Close the video file
    av_close_input_file(pFormatCtx);

	if ( _img_convert_ctx )
	  sws_freeContext( _img_convert_ctx );

}


bool VideoImageSequence::GetNextFrame()
{
    static AVPacket packet;
    static int      bytesRemaining=0;
    static uint8_t  *rawData;
    static bool     fFirstTime=true;
    int             bytesDecoded;
    int             frameFinished;

    // First time we're called, set packet.data to NULL to indicate it
    // doesn't have to be freed
    if(fFirstTime)
    {
        fFirstTime=false;
        packet.data=NULL;
    }

    // Decode packets until we have decoded a complete frame
    while(true)
    {
        // Work on the current packet until we have decoded all of it
        while(bytesRemaining > 0)
        {
            // Decode the next chunk of data
            bytesDecoded=avcodec_decode_video(pCodecCtx, pFrame,
                &frameFinished, rawData, bytesRemaining);

			//fprintf(stdout, "rawData: %d %d %d\n", int(rawData[0]),int(rawData[1]),int(rawData[2]));
            // Was there an error?
            if(bytesDecoded < 0)
            {
                fprintf(stderr, "Error while decoding frame\n");
                return false;
            }

            bytesRemaining-=bytesDecoded;
            rawData+=bytesDecoded;

            // Did we finish the current frame? Then we can return
            if(frameFinished)
			  {
				return true;
			  } 
       }

        // Read the next packet, skipping all packets that aren't for this
        // stream
        do
        {
            // Free old packet
            if(packet.data!=NULL)
                av_free_packet(&packet);

            // Read new packet
            if(av_read_packet(pFormatCtx, &packet)<0)
                goto loop_exit;
        } while(packet.stream_index!=videoStream);

        bytesRemaining=packet.size;
        rawData=packet.data;
    }

loop_exit:

    // Decode the rest of the last frame
    bytesDecoded=avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, 
        rawData, bytesRemaining);

    // Free last packet
    if(packet.data!=NULL)
        av_free_packet(&packet);

    return frameFinished!=0;
    
}

HostImage VideoImageSequence::getCurrentImage()
{
  // use getNextImage but leave the Image in the buffer
  HostImage result = getNextImage();
  hasImageInBuffer = true;
  return result;
}


bool VideoImageSequence::hasNextImage()
{
  // check if an image is already buffered
  if (hasImageInBuffer) return true;

  // check if a codec is initialized and ready to use
  if (!pCodecCtx) throw std::runtime_error("VideoImageSequence::hasNextImage(): cannot find avCodecContext. avcodec not initialized?");

  // try to load the image.
  hasImageInBuffer = GetNextFrame();

  // if successful: lucky you! via getNextImage it can be fetched
  return hasImageInBuffer;
}

int VideoImageSequence::getWidth()
{
  if (pCodecCtx) return pCodecCtx->width;
  else  throw std::runtime_error("VideoImageSequence::getWidth(): cannot find avCodecContext. avcodec not initialized?");
}

int VideoImageSequence::getHeight()
{
  if (pCodecCtx) return pCodecCtx->height;
  else  throw std::runtime_error("VideoImageSequence::getHeight(): cannot find avCodecContext. avcodec not initialized?");
}



HostImage VideoImageSequence::getNextImage()
{
  frame = floor(frame_next + 0.5);
  frame_next += frame_stepsize;
  int frames_to_go = floor(frame_next - frame + 0.5);
  if (frames_to_go<=0) throw std::runtime_error("frames_to_go==0? something went wrong...");



  HostImage result;
  while ( frames_to_go > 0 )
  {
    // hasNextImage()
    if ( !hasNextImage() ) throw std::runtime_error("getNextImage(): could not find next image"); 
    
	sws_scale(	_img_convert_ctx, pFrame->data, 
				pFrame->linesize, 0, 
				pCodecCtx->height, 
				pFrameRGB->data, pFrameRGB->linesize);
	
	// Insert the extracted data into a new HostImage
	result = HostImage(  pCodecCtx->width,pCodecCtx->height, HostImage::UBYTE, HostImage::RGB );
	
	CPUData<HostImage::ubyte> & target = result.cpu_data_ubyte();
	
  	for ( int i = 0; i <   pCodecCtx->height; ++i)
	  {
        const unsigned char * rgb = pFrameRGB->data[0] + i * pFrameRGB->linesize[0];
        std::copy( rgb, rgb + pCodecCtx->width * 3, &target(i,0,0) );
	  }
  
    // reset the buffer flag
    hasImageInBuffer = false;
    frames_to_go--;
  }

  return result;
}

