/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "VideoOutputFile.hpp"

#include <limits>

static AVFrame *alloc_picture(PixelFormat pix_fmt, int width, int height)
{
    AVFrame *picture;
    uint8_t *picture_buf;
    int size;

    picture = avcodec_alloc_frame();
    if (!picture)
        return NULL;
    size = avpicture_get_size(pix_fmt, width, height);
    picture_buf = (uint8_t *)av_malloc(size);
    if (!picture_buf) {
        av_free(picture);
        return NULL;
    }
    avpicture_fill((AVPicture *)picture, picture_buf,
                   pix_fmt, width, height);
    return picture;
}

VideoOutputFile::VideoOutputFile( const std::string & sfilename, size_t width, size_t height, 
                                  HostImage::ColorSpace space, float fps, size_t rate, CodecID codec_id )
  : _output_queue( 5 )
{
  av_register_all();
  open( sfilename, width, height, space, fps, rate, codec_id );
}  

VideoOutputFile::~VideoOutputFile()
{
  close();   
}


void VideoOutputFile::open( const std::string & sfilename, size_t width, size_t height, HostImage::ColorSpace colorspace,
                            float fps, size_t rate, CodecID codec_id )
{
    const char *filename = sfilename.c_str();
    AVCodecContext *c;
    AVCodec *codec;

    switch ( colorspace )
    {
      case HostImage::RGB:   _pixel_format = PIX_FMT_RGB24; break;
      case HostImage::LUM:   _pixel_format = PIX_FMT_GRAY8; break;
      case HostImage::YCbCr: throw std::runtime_error("cannot encode YCbCr 8:8:8");
    }
    _color_space = colorspace;

    /* auto detect the output format from the name. default is
       mpeg. */
    _av_format = guess_format(NULL, filename, NULL);
    if (!_av_format) {
        printf("VideoOutputFile: Could not deduce output format from file extension: using MPEG.\n");
        _av_format = guess_format("mpeg", NULL, NULL);
    }
    if (!_av_format) throw std::runtime_error("Could not find suitable output format\n");

    /* allocate the output media context */
    _av_format_context = av_alloc_format_context();
    if (!_av_format_context) throw std::runtime_error("Memory error\n");

    _av_format_context->oformat = _av_format;
    snprintf( _av_format_context->filename, sizeof(_av_format_context->filename), "%s", filename);

    _av_stream = av_new_stream( _av_format_context, 0);
    if (!_av_stream) throw std::runtime_error("Could not alloc stream\n");

    c = _av_stream->codec;
    c->codec_id = codec_id;
    c->codec_type = CODEC_TYPE_VIDEO;

    /* put sample parameters */
    c->bit_rate = rate;
    /* resolution must be a multiple of two */
    c->width  = ( ( width  + 1 ) / 2 ) * 2;
    c->height = ( ( height + 1 ) / 2 ) * 2;
    /* time base: this is the fundamental unit of time (in seconds) in terms
       of which frame timestamps are represented. for fixed-fps content,
       timebase should be 1/framerate and timestamp increments should be
       identically 1. */
    c->time_base.den = fps;
    c->time_base.num = 1;
    c->gop_size = 12; /* emit one intra frame every twelve frames at most */
	c->pix_fmt = PIX_FMT_YUV420P;
    // if (c->codec_id == CODEC_ID_MPEG2VIDEO) {
    //     /* just for testing, we also add B frames */
    //     c->max_b_frames = 2;
    // }
    if (c->codec_id == CODEC_ID_MPEG1VIDEO){
        /* Needed to avoid using macroblocks in which some coeffs overflow.
           This does not happen with normal video, it just happens here as
           the motion of the chroma plane does not match the luma plane. */
        c->mb_decision=2;
    }
    
    // some formats want stream headers to be separate
    if( _av_format_context->oformat->flags & AVFMT_GLOBALHEADER )
        c->flags |= CODEC_FLAG_GLOBAL_HEADER;
        
    /* set the output parameters (must be done even if no
       parameters). */
    if (av_set_parameters( _av_format_context, NULL ) < 0)
        throw std::runtime_error("Invalid output format parameters\n");

    // dump info
    dump_format( _av_format_context, 0, filename, 1);

    /* find the video encoder */
    codec = avcodec_find_encoder( c->codec_id );
    if ( !codec )
      throw std::runtime_error("codec not found\n");

    /* open the codec */
    if (avcodec_open(c, codec) < 0)
      throw std::runtime_error("codec could not be opened\n");

    _video_outbuf = NULL;
    if (!(_av_format_context->oformat->flags & AVFMT_RAWPICTURE)) {
        /* allocate output buffer */
        /* XXX: API change will be done */
        /* buffers passed into lav* can be allocated any way you prefer,
           as long as they're aligned enough for the architecture, and
           they're freed appropriately (such as using av_free for buffers
           allocated with av_malloc) */
        _video_outbuf_size = 2000000;
        _video_outbuf = (uint8_t*) av_malloc( _video_outbuf_size );
    }

    /* allocate the encoded raw picture */
    _picture = alloc_picture(c->pix_fmt, c->width, c->height);
    if (!_picture) 
      throw std::runtime_error("Could not allocate picture\n");

	// allocate the input picture buffer
    _picture_buf = alloc_picture( _pixel_format, c->width, c->height);
    if (!_picture_buf) 
      throw std::runtime_error("Could not allocate picture buffer\n");


    /* open the output file, if needed */
    if (!(_av_format->flags & AVFMT_NOFILE)) {
        if ( url_fopen( &_av_format_context->pb, filename, URL_WRONLY ) < 0 ) 
          throw std::runtime_error( str( boost::format("Could not open '%s'\n") % filename) );
    }

    /* write the stream header, if any */
    av_write_header( _av_format_context );

	// allocate the converter from the input format to the video format
	_img_convert_ctx = sws_getContext(c->width, c->height,
									  _pixel_format,
									  c->width, c->height,
									  c->pix_fmt,
									  SWS_BICUBIC, NULL, NULL, NULL);

	if ( _img_convert_ctx == NULL )
	  throw std::runtime_error("could not allocate colorspace converter");

    _frame_count = 0;
    _open = true;

  _thread.reset( new boost::thread(boost::ref(*this)) );

}


void VideoOutputFile::write_video_frame( HostImage image )
{
  //do_write_video_frame( image );
  image.make_unique();  
  _output_queue.push( image );  
}

void VideoOutputFile::do_write_video_frame( HostImage image )
{
  
    int out_size, ret;
    AVCodecContext *c;

    c = _av_stream->codec;

        // FIXME: the one extra copy could be avoided by directly converting into the picture
	//	if ( false )
        {
          image.convert_to_space( _color_space );
          image.convert_to_type( HostImage::UBYTE );
          HostImage::BData & bdata = image.cpu_data_ubyte();
          size_t height = std::min( size_t(c->height), size_t(image.getHeight()) );
          size_t width  = std::min( size_t(c->width ), size_t(image.getWidth ()) );
          // copy line-wise
          for ( size_t i = 0; i < height; ++i )
          {
            uint8_t * src_line = &bdata(i,0,0);
            uint8_t * tgt_line = _picture_buf->data[0] + i * _picture_buf->linesize[0];
            size_t bpp = bdata.depth();
            std::copy( src_line, src_line + bpp * width, tgt_line );
          }
		  int h_out = sws_scale( _img_convert_ctx, _picture_buf->data, _picture_buf->linesize,
								 0, c->height, _picture->data, _picture->linesize);
		  //		  std::cout << " converted " << h_out << " lines " << std::endl;
        }


    if ( _av_format_context->oformat->flags & AVFMT_RAWPICTURE ) 
    {
        /* raw video case. The API will change slightly in the near
           futur for that */
        AVPacket pkt;
        av_init_packet(&pkt);

        pkt.flags |= PKT_FLAG_KEY;
        pkt.stream_index= _av_stream->index;
        pkt.data= (uint8_t *)_picture;
        pkt.size= sizeof(AVPicture);

        ret = av_write_frame( _av_format_context, &pkt );
    } else {
        /* encode the image */
        out_size = avcodec_encode_video(c, _video_outbuf, _video_outbuf_size, _picture);
        /* if zero size, it means the image was buffered */
        if (out_size > 0) 
        {

            AVPacket pkt;
            av_init_packet(&pkt);

// broken:            if ( c->coded_frame->pts != int64_t( AV_NOPTS_VALUE ) )
            if ( c->coded_frame->pts != std::numeric_limits<int64_t>::min() )
                pkt.pts = av_rescale_q( c->coded_frame->pts, c->time_base, _av_stream->time_base );
                
            if(c->coded_frame->key_frame)
                pkt.flags |= PKT_FLAG_KEY;
                
            pkt.stream_index = _av_stream->index;
            pkt.data         = _video_outbuf;
            pkt.size         = out_size;

            /* write the compressed frame in the media file */
            ret = av_write_frame( _av_format_context, & pkt );
            
        } 
        else 
        {
            ret = 0;
        }
    }

    if (ret != 0) 
      throw std::runtime_error("Error while writing video frame\n");

    _frame_count++;

}

void VideoOutputFile::close()
{
  if ( !_open ) return;

  _output_queue.terminate_out();
  _thread->join();  

  av_write_trailer( _av_format_context );
  
  if ( _av_stream )
	avcodec_close( _av_stream->codec );
  
  if ( _picture )
    {
      av_free( _picture->data[0] );
      av_free( _picture          );
      _picture = 0;
    }
  
  if ( _picture_buf )
    {
      av_free( _picture_buf->data[0] );
      av_free( _picture_buf          );
      _picture_buf = 0;
    }
  
  if ( _video_outbuf )
	av_free( _video_outbuf );
  
  /* free the streams */
  for ( size_t i = 0; i < _av_format_context->nb_streams; i++ ) 
    {
      AVStream * stream = _av_format_context->streams[i];
	  // this crashes:
      // av_freep( stream->codec );
      av_freep( stream        );
    }
  
  if (!(_av_format->flags & AVFMT_NOFILE)) 
    {
	  /* close the output file */
#if LIBAVFORMAT_VERSION_INT < ((52<<16)+(7<<8)+0)
	  url_fclose(&_av_format_context->pb); // old
#else
	  url_fclose(_av_format_context->pb); // new
#endif

    }

  if ( _img_convert_ctx )
	sws_freeContext( _img_convert_ctx );

  _open = false;

}
