/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "connected_components.hpp"

#include <boost/foreach.hpp>
#include <algorithm>

bool connected_components::find_one( const CPUDataLayer<bool> & data, const CPUDataLayer<bool> & mask, point_t & result )
{
  for ( CPUDataLayer<bool>::const_iterator i = data.begin(), im = mask.begin(), end = data.end();
		i != end; ++i, ++im )
	{
	  if ( *i && *im ) 
		{
		  result.row    = i.row();
		  result.column = i.column();
		  return true;
		}
	}
  return false;
}

bool connected_components::find_max( const CPUDataLayer<float> & data
									 , const CPUDataLayer<bool> & mask
									 , point_t & result, float threshold, size_t region )
{
  float maxval = 1e-10f;
  CPUDataLayer<bool>::const_iterator im = mask.begin();
  for ( CPUDataLayer<float>::const_iterator iter = data.begin(), end = data.end(); iter != end; ++iter, ++im )
	{
	  if ( *im && *iter > maxval ) 
		{

		  if ( region > 0 )
			{
			  CPUDataLayer<float> R;
			  size_t i0 = iter.row()    >= region ? iter.row()    - region : 0;
			  size_t j0 = iter.column() >= region ? iter.column() - region : 0;
			  size_t i1 = iter.row()    < ( data.rows()    - region ) ? iter.row()     + region : ( data.rows()    - 1 );
			  size_t j1 = iter.column() < ( data.columns() - region ) ? iter.column()  + region : ( data.columns() - 1 );
			  bool ispeak = true;
			  for ( size_t i = i0; ispeak && i <= i1; ++i )
				for ( size_t j = j0; ispeak && j <= j1; ++j )
				  if ( *iter < data( i, j ) )
					ispeak = false;
			  if ( !ispeak )
				continue;
			}
		  result.row    = iter.row();
		  result.column = iter.column();
		  maxval = *iter;
		}
	}
  return maxval > threshold;
}

struct filter
{
  filter ( size_t t ) : threshold(t) {}
  size_t threshold;
  bool operator() ( const connected_components::component_t & component )
  {
	return component.size() < threshold;
  }
};

struct sum_filter
{
  sum_filter ( float t ) : threshold(t) {}
  float threshold;
  bool operator() ( const connected_components::component_t & component )
  {
	return component.wsum < threshold;
  }
};

struct kurtosis_filter
{
  kurtosis_filter ( float t ) : threshold(t) {}
  float threshold;
  bool operator() ( const connected_components::component_t & component )
  {
	return component.kurtosis < threshold;
  }
};

void connected_components::sum_filter_components( float threshold )
{
  _components.remove_if( sum_filter(threshold) );
}

void connected_components::remove_platykurtic_components( float threshold )
{
  _components.remove_if( kurtosis_filter(threshold) ); 
}

void connected_components::remove_small_components( size_t threshold )
{
  _components.remove_if( filter(threshold) );
}

struct is_true
{
  bool operator()( bool b, bool bc ) const
  {
	return b;
  }
};

void connected_components::component_t::analyze()
{
  imean = var_i = var_j = var_ij = 0;
  imin = imax = points.front().row;
  jmin = jmax = points.front().column;
  float wmax = -1e10f;
  wsum = 0;
  //std::cout << " [ " << imin << ", " << imax << " ] x [ " << jmin << ", " << jmax << "] " << std::endl;
  BOOST_FOREACH( point_t & p, points )
	{	  
	  float i = p.row, j = p.column;
	  imin = std::min( imin, i );
	  imax = std::max( imax, i );
	  jmin = std::min( jmin, j );
	  jmax = std::max( jmax, j );
	  //std::cout << i << ", " << j << " [ " << imin << ", " << imax << " ] x [ " << jmin << ", " << jmax << "] " << std::endl;
	  imean += i;
	  jmean += j;
	  var_i += i * i;
	  var_j += j * j;
	  var_ij += i * j;
	  wmax = std::max( p.w, wmax );
	  wsum += p.w;
	}
  score = wmax;
  imean /= points.size();
  jmean /= points.size();
  var_i /= points.size();
  var_j /= points.size();
  var_ij /= points.size();
  var_i = std::max( 0.25f, var_i - imean * imean );
  var_j = std::max( 0.25f, var_j - jmean * jmean );
  var_ij = var_ij - imean * jmean;
  double m4i = 0, m4j = 0;
  BOOST_FOREACH( point_t & p, points )
	{
	  float di = p.row - imean, dj = p.column - jmean;
	  m4i += di * di * di * di;
	  m4j += dj * dj * dj * dj;
	}
  float kurt_i = m4i / size() / std::max( 1.0f, var_i * var_i ) - 3.0f;
  float kurt_j = m4j / size() / std::max( 1.0f, var_j * var_j ) - 3.0f;
  kurtosis = std::max( kurt_i, kurt_j );
}

inline float euklidian_hausdorff_distance( const connected_components::component_t & c1,  const connected_components::component_t & c2 )
{
  float d = 1e10f;
  BOOST_FOREACH( const connected_components::point_t & p1, c1.border )
	BOOST_FOREACH( const connected_components::point_t & p2, c2.border )
	{
	  float di = p1.row - p2.row;
	  float dj = p1.column - p2.column;
	  d = std::min( d, std::sqrt( di * di + dj * dj ) );
	}
  return d;
}

void connected_components::group_components( float dmax, float size_max )
{
  bool todo = true;
  while( todo )	;
  //  with each component that is smaller than size_max
  //     find nearest other component
  //     if nearer than dmax 
  //         merge with this component and remove from list
  //         set todo = true
}

struct is_larger
{
  is_larger( float t
			 , float dmin = std::numeric_limits<float>::min()
			 , float dmax = std::numeric_limits<float>::max() ) 
	: _t(t), _dmin(dmin), _dmax(dmax) {}
  bool operator()( float x, float xc ) const
  {
	float dx = x - xc;
	return x > _t && dx >= _dmin && dx <= _dmax;
  }
  float _t, _dmin, _dmax;
};

connected_components::connected_components( CPUDataLayer<float> data, float threshold, float margin, float dmin, float dmax, bool relative, size_t max_count, size_t max_size, size_t region )
{
  point_t seed(0,0);
  // always start with finding a seed point in the image
  CPUDataLayer<bool> mask( data.rows(), data.columns() );
  BOOST_FOREACH( bool & x, mask )
	x = true;
  while ( find_max( data, mask, seed, threshold, region ) )
	{
	  float maxval = data( seed.row, seed.column );
	  float t = maxval - ( relative ? std::abs(maxval) * margin : margin );
	  find_component_at( data, mask, seed, is_larger( t, dmin, dmax ), max_count, max_size );
	  _components.back().score = maxval;
	}
}

connected_components::connected_components( CPUDataLayer<bool> data, size_t max_count, size_t max_size )
{
  point_t seed(0,0);
  CPUDataLayer<bool> mask( data.rows(), data.columns() );
  BOOST_FOREACH( bool & x, mask )
	x = true;
  while ( find_one( data, mask, seed ) )
	{
	  find_component_at( data, mask, seed, is_true(), max_count, max_size );
	}
}

