/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __CONNECTED_COMPONENTS_HPP__
#define __CONNECTED_COMPONENTS_HPP__

#include "CPUData.hpp"

#include <vector>
#include <list>
#include <limits>

class connected_components
{

public:
    
  struct point_t
  {
	point_t ( size_t in_row, size_t in_column, float in_w = 0 ) 
	  : row(in_row), column(in_column), w(in_w), border(false) {}
	size_t row;
	size_t column;
	float w;
	bool border;
	bool operator==( const point_t & other ) const
	{
	  return row == other.row && column == other.column;
	}
	bool operator>( const point_t & other ) const
	{
	  return w > other.w;
	}
  };
  
  typedef std::vector<point_t> points_t;

  struct component_t
  {
	component_t() : score(0) {}
	void analyze();
	size_t size() const { return points.size(); }
	points_t points;
	points_t border;
	float score;
	float wsum;
	float imean, jmean, var_i, var_j, var_ij, imin, imax, jmin, jmax, kurtosis;
  };

  typedef std::list< component_t > components_t;

  connected_components( CPUDataLayer<bool> data, size_t max_count, size_t max_size );

  connected_components( CPUDataLayer<float> data, float threshold, float margin, float dmin, float dmax, bool relative, size_t max_count, size_t max_size, size_t region );

  const components_t & components()
  {
	return _components;
  }

  void prune( size_t nmax )
  {
	while ( _components.size() >= nmax )
	  _components.pop_back();
  }

  template< typename T >
  void to_masks( CPUData<T> & out )
  {
	
	size_t c_index = 0;
	
	std::fill( out.begin(), out.end(), T(false) );
	
	BOOST_FOREACH ( const component_t & component, _components )
	  {
		
		BOOST_FOREACH ( const point_t & point, component.points )
		  {
			out( point.row, point.column, c_index ) = T(true);
		  }
		
		++c_index;
		
	  }
	
  }  

  void remove_small_components( size_t threshold );

  void sum_filter_components( float threshold );

  void remove_platykurtic_components( float threshold );

  void group_components( float dmax, float size_max );

  size_t size() { return _components.size(); }

  template<typename T, typename F>
  void find_component_at( CPUDataLayer<T> & data, CPUDataLayer<bool> & mask, point_t seed, const F & condition
						  , size_t max_count = std::numeric_limits<size_t>::max() 
						  , size_t max_size  = std::numeric_limits<size_t>::max() 
						  )
  {

	// make a new component
	_components.push_back( component_t() );
	component_t & component = _components.back();
	
	// buffer for new points
	points_t front, new_front;
	
	// move the seed point from the image to the front list, initialize bounds
	front.push_back( seed );
	mask( seed.row, seed.column ) = false;
	size_t imin, imax, jmin, jmax;
	imin = imax = seed.row;
	jmin = jmax = seed.column;
   
	while ( !front.empty() )
	  {

		// consider each front point
		BOOST_FOREACH( const point_t & front_point, front )
		  {			

			T center_value = data( front_point.row, front_point.column );

			// bail if limit was reached
			if ( component.points.size() >= max_count ) break;

			size_t fi = front_point.row, fj = front_point.column;
			size_t di = fi > imax ? ( fi - imin ) : ( fi < imin ? ( imax - fi ) : ( imax - imin ) );
			size_t dj = fj > jmax ? ( fj - jmin ) : ( fj < jmin ? ( jmax - fj ) : ( jmax - jmin ) );

			if ( std::max( di, dj ) > max_size ) continue;
			
			// add the front point to the component, update bounds
			component.points.push_back( front_point );
			imin = std::min( imin, front_point.row );
			imax = std::max( imax, front_point.row );
			jmin = std::min( jmin, front_point.column );
			jmax = std::max( jmax, front_point.column );

			// consider the 8-way neighborhood
			bool is_border = false;
			for ( ptrdiff_t i = std::max( ptrdiff_t(0), ptrdiff_t( front_point.row ) - 1 ),
					i_max = std::min( ptrdiff_t( data.rows() ) - 1,  ptrdiff_t( front_point.row ) + 1 );
				  i <= i_max;
				  ++i )
			  for ( ptrdiff_t j = std::max( ptrdiff_t(0), ptrdiff_t( front_point.column ) - 1 ),
					  j_max = std::min( ptrdiff_t( data.columns() ) - 1,  ptrdiff_t( front_point.column ) + 1 );
					j <= j_max;
					++j )
				{
				  if ( condition( data( i, j ), center_value ) )
					{
					  if ( mask( i, j ) )
						{
						  new_front.push_back( point_t( i, j, data(i,j) ) );
						  mask( i, j ) = false;
						}
					  else if ( std::find( component.points.begin(), component.points.end()
										   , point_t( i, j ) ) == component.points.end() )
						is_border = true;						
					}
				  else // border points are component points neighboring non-component or foreign component points
					{
					  is_border = true;
					}
				} // neighborhood scan				
			// annotate border points
			component.points.back().border = is_border;
		  }
		

		// make the next front the current one
		front.clear();
		std::swap( front, new_front );
		
	  }

	_components.back().analyze();
	
  }

private:

  bool find_max( const CPUDataLayer<float> & data, const CPUDataLayer<bool> & mask, point_t & result, float threshold, size_t region );
  bool find_one( const CPUDataLayer<bool> & data, const CPUDataLayer<bool> & mask, point_t & result );
  
  components_t _components;

};

#endif
