/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __SLIDING_WINDOW_ITERATOR__
#define __SLIDING_WINDOW_ITERATOR__

#include "CPUData.hpp"

struct sw_base
{
  sw_base( int depth )
  : _buffer( 1, 1, depth )
  , _sum( _buffer( 0, 0 ) )
  {}
  
  void init( CPUData<float> & window ){};
  void update( CPUData<float> & left_edge, CPUData<float> & right_edge ){};
  CPUDataFeatureColumn<float>& get()
  {
    return _sum;
  }
  
protected:
  // remembered sum
  CPUData<float> _buffer;
  CPUDataFeatureColumn<float> _sum;
};

struct sw_accum : sw_base
{
  sw_accum( int depth )
    : sw_base( depth )
  {}
  
  void init( CPUData<float> & window )
  {
    std::fill( _sum.begin(), _sum.end(), 0.0f );
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, window )
       _sum += col;    
  }
  
  void update( CPUData<float> & left_edge, CPUData<float> & right_edge )
  {    
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, left_edge )
      _sum -= col;
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, right_edge )
      _sum += col;
  }


};

struct sw_sparse_vel_accum : sw_base
{
  
  sw_sparse_vel_accum( int depth ) 
    : sw_base(depth)
    {
#ifndef _WIN32
		_velocity = floor( 0.5f * (floor( sqrt( depth ) ) - 1));
#else
		_velocity = floor( 0.5f * (floor( sqrt( (float)depth ) ) - 1));
#endif
    }
  
  void init( CPUData<float> & window )
  {
    std::fill( _sum.begin(), _sum.end(), 0.0f );
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, window )
    {
      const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
      const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
      const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
      _sum[ pos ] += col[2];
    }
  }
  
  void update( CPUData<float> & left_edge, CPUData<float> & right_edge )
  {    
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, left_edge )
    {
      const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
      const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
      const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
      _sum[ pos ] -= col[2];
    }
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, right_edge )
    {
      const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
      const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
      const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
      _sum[ pos ] += col[2];
    }
  }

private:

  inline int clamp( int val, int const& val_min, int const& val_max ) const
  {
    if ( val > val_max ) return val_max;
    if ( val < val_min ) return val_min;
    return val;
  }
  
  int _velocity;

};

/// this iterator  computes the mean over a window but ignores values smaller than selection_threshold
struct sw_sparse_vel_selective_mean : sw_base
{
  
  sw_sparse_vel_selective_mean( int depth, float selection_threshold = 1e-03f ) 
  : sw_base( depth )
  , _threshold(selection_threshold) 
  , _countbuffer( 1, 1, _buffer.depth() )
  , _resultbuffer( 1, 1, _buffer.depth() )
  , _count( _countbuffer( 0, 0 ) )
  , _result( _resultbuffer( 0, 0 ) )
  , _has_result( false )
  {
#ifndef _WIN32
	  _velocity = floor( 0.5f * (floor( sqrt( depth ) ) - 1));
#else
	  _velocity = floor( 0.5f * (floor( sqrt( (float)depth ) ) - 1));
#endif
  }
  
  void init( CPUData<float> & window )
  {
    _has_result = false;
    std::fill( _sum.begin(), _sum.end(), 0.0f );
    std::fill( _count.begin(), _count.end(), 0.0f );
    float conf = 0.f;
    /*for ( CPUData<float>::iterator irf = window.begin(), iend = window.end(); irf != iend; ++irf )
      {
        conf = (*irf)[2];
        if ( conf > _threshold )
        {
          const int i_dx = clamp( _velocity + (*irf)[0] + 0.5f, 0, 2 * _velocity );
          const int i_dy = clamp( _velocity + (*irf)[1] + 0.5f, 0, 2 * _velocity );
          const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
          _sum[ pos ] += conf;
          _count[pos] += 1;
        }
      }*/
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, window )
    {
      conf = col[2];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] += conf;
        _count[pos] += 1;
      }
    }
  }
  
  void update(  CPUData<float> & left_edge, CPUData<float> & right_edge )
  { 
    _has_result = false;
    float conf = 0.f;
    for ( CPUData<float>::iterator irf = left_edge.begin(), iend = left_edge.end(); irf != iend; ++irf )
      {
        conf = (*irf)[2];
        if ( conf > _threshold )
        {
          const int i_dx = clamp( _velocity + (*irf)[0] + 0.5f, 0, 2 * _velocity );
          const int i_dy = clamp( _velocity + (*irf)[1] + 0.5f, 0, 2 * _velocity );
          const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
          _sum[ pos ] -= conf;
          _count[pos] -= 1;
        }
      }
    for ( CPUData<float>::iterator irf = right_edge.begin(), iend = right_edge.end(); irf != iend; ++irf )
      {
        conf = (*irf)[2];
        if ( conf > _threshold )
        {
          const int i_dx = clamp( _velocity + (*irf)[0] + 0.5f, 0, 2 * _velocity );
          const int i_dy = clamp( _velocity + (*irf)[1] + 0.5f, 0, 2 * _velocity );
          const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
          _sum[ pos ] += conf;
          _count[pos] += 1;
        }
    }
    
    /*
    // I have no idea why this version is not working
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, left_edge )
    {
      conf = col[2];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] -= conf;
        _count[pos] -= 1;
      }
    }
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, right_edge )
    {
      conf = col[2];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] += conf;
        _count[pos] += 1;
      }
    }*/
  }

  CPUDataFeatureColumn<float> & get()
  {
    // compute mean for each bin and return it
    if ( !_has_result )
    {
      std::fill( _result.begin(), _result.end(), 0.0f );
      for ( unsigned i = 0; i < _sum.size(); ++i )
      {
        if ( _count[i] != 0 )
          _result[i] = _sum[i] / _count[i];
      }
      _has_result = true;
    }
    
    return _result;
  }

private:

  inline int clamp( int val, int const& val_min, int const& val_max ) const
  {
    if ( val > val_max ) return val_max;
    if ( val < val_min ) return val_min;
    return val;
  }
  
  int _velocity;
  float _threshold;
  CPUData<float> _countbuffer, _resultbuffer;
  CPUDataFeatureColumn<float> _count, _result;
  bool _has_result;

};

/// this iterator  computes the mean over a window but ignores values smaller than selection_threshold
struct sw_selective_mean : sw_base
{
  
  sw_selective_mean( int depth, float selection_threshold = 1e-03f ) 
  : sw_base( depth )
  , _threshold(selection_threshold) 
  , _countbuffer( 1, 1, _buffer.depth() )
  , _resultbuffer( 1, 1, _buffer.depth() )
  , _count( _countbuffer( 0, 0 ) )
  , _result( _resultbuffer( 0, 0 ) )
  , _has_result( false )
  {
  }
  
  void init( CPUData<float> & window )
  {
    _has_result = false;
    std::fill( _sum.begin(), _sum.end(), 0.0f );
    std::fill( _count.begin(), _count.end(), 0.0f );
    float conf = 0.f;

    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, window )
    {
      for (size_t i=0; i<col.size(); i++)
      {
        conf = col[i];
        if ( conf > _threshold )
        {
          _sum[ i ] += conf;
          _count[i] += 1;
        }
      }
    }
  }
  
  void update(  CPUData<float> & left_edge, CPUData<float> & right_edge )
  { 
    _has_result = false;
    float conf = 0.f;
    for ( CPUData<float>::iterator irf = left_edge.begin(), iend = left_edge.end(); irf != iend; ++irf )
      {
        for (size_t i=0; i<(*irf).size(); i++)
        {
          conf = (*irf)[i];
          if ( conf > _threshold )
          {
            _sum[ i ] -= conf;
            _count[i] -= 1;
          }
        }
      }
    for ( CPUData<float>::iterator irf = right_edge.begin(), iend = right_edge.end(); irf != iend; ++irf )
      {
        for (size_t i=0; i<(*irf).size(); i++)
        {
          conf = (*irf)[i];
          if ( conf > _threshold )
          {
            _sum[ i ] += conf;
            _count[i] += 1;
          }
        }
      }
    
    /*
    // I have no idea why this version is not working
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, left_edge )
    {
      conf = col[2];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] -= conf;
        _count[pos] -= 1;
      }
    }
    BOOST_FOREACH( CPUDataFeatureColumn<float> & col, right_edge )
    {
      conf = col[2];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + col[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + col[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] += conf;
        _count[pos] += 1;
      }
    }*/
  }

  CPUDataFeatureColumn<float> & get()
  {
    // compute mean for each bin and return it
    if ( !_has_result )
    {
      std::fill( _result.begin(), _result.end(), 0.0f );
      for ( unsigned i = 0; i < _sum.size(); ++i )
      {
        if ( _count[i] != 0 )
          _result[i] = _sum[i] / _count[i];
      }
      _has_result = true;
    }
    
    return _result;
  }

private:

  float _threshold;
  CPUData<float> _countbuffer, _resultbuffer;
  CPUDataFeatureColumn<float> _count, _result;
  bool _has_result;

};

struct sw_accum_w : sw_base
{
  sw_accum_w( int depth )
    : sw_base( depth )
  {}
  sw_accum_w( int depth, const CPUData<float>& weights )
    : sw_base( depth )
  {
    set_weights( weights );
  }
  
  void init( CPUData<float> & window )
  {
    std::fill( _sum.begin(), _sum.end(), 0.0f );
    if ( _weights.empty() )
    {
      _weights.resize( window.rows(), window.columns(), window.depth() );
      _weights = 1.f / window.size();
    }
    
    CPUData<float>::iterator iwindow = window.begin(), iend = window.end();
    CPUData<float>::iterator iweight = _weights.begin();
    for ( ; iwindow != iend; ++iwindow, ++iweight )
      _sum.add_scaled( (*iwindow), (*iweight)[0] );
  }
  
  
  void update( CPUData<float> & left_edge, CPUData<float> & right_edge )
  {    
    CPUData<float> left_edge_w( _weights, 0, _weights.rows(), 0, 1 );
    CPUData<float> right_edge_w( _weights, 0, _weights.rows(), _weights.columns() - 1, 1 );
    
    CPUData<float>::iterator iwindow = left_edge.begin(), iend = left_edge.end();
    CPUData<float>::iterator iweight = left_edge_w.begin();
    for ( ; iwindow != iend; ++iwindow, ++iweight )
      _sum.add_scaled( (*iwindow), -(*iweight)[0] );
      
    iwindow = right_edge.begin(), iend = right_edge.end();
    iweight = right_edge_w.begin();
    for ( ; iwindow != iend; ++iwindow, ++iweight )
      _sum.add_scaled( (*iwindow), (*iweight)[0] );
      
  }

  void set_weights( const CPUData<float>& w )
  {
    _weights.copy_from( w );
  }
  
  CPUData<float> _weights;
};

template< typename F = sw_accum >  
struct sliding_window_iterator
{

  sliding_window_iterator( const CPUData<float> & data, size_t window_size, size_t depth, F f )
    : _data( data )
    , _window_size( window_size )
    , _done(false)
    , _f(f)
  {
    init();
  }    
  
  sliding_window_iterator( const CPUData<float> & data, size_t window_size, size_t depth )
  : _data( data )
    , _window_size( window_size )
    ,_row(0)
    ,_column(0)
    , _done(false)
    , _f( depth )
  {
    init();
  } 

  
  virtual void increment()
  {
    _column += 1;
    if ( _column > _data.columns() - _window_size )
    {
      _column = 0;
      _row += 1;
      if ( _row > _data.rows() - _window_size )
      {
        _done = true;
        return;
      }
      start_line( _row, _window_size );
    }
    else
    {
      move_right();
    }
  }
  
  bool more() const { return !_done; }
  
  virtual CPUDataFeatureColumn<float> & operator*()
  {
    return _f.get();
  }
  virtual CPUDataFeatureColumn<float> * operator->()
  {
    return &(_f.get());
  }
  
  virtual size_t window_size() const { return _window_size;};
  
  virtual F& core(){ return _f;};
  
  size_t row() const
  {
    return _row;
  }

  size_t column() const
  {
    return _column;
  }
private:

  void init( )
  {
    if ( _window_size >= std::min( _data.rows(), _data.columns() ) )
    {
      _done = true;
    }
    else
      start_line(0,_window_size);
  }

  void start_line( size_t row, size_t window_size )
  {
    _window.reference( _data, row, window_size, 0, window_size );
    _f.init( _window );
  }
  
  void move_right( )
  {
    CPUData<float> left_edge( _window, 0, _window.rows(), 0, 1 );
    _window.rebase( _window.offset() + _window.depth() );
    CPUData<float> right_edge( _window, 0, _window.rows(), _window.columns() - 1, 1 );
    _f.update( left_edge, right_edge );
  }
  

  
  size_t _window_size;
  // the actual input data
  CPUData<float> _data;
  // the current window
  CPUData<float> _window;
  // current window position
  size_t _row, _column;

  // flag
  bool _done;
  // functor
  F _f;
};
  
#endif
