/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __SLIDING_WINDOW_ITERATOR_SPLITTED__
#define __SLIDING_WINDOW_ITERATOR_SPLITTED__

#include "CPUData.hpp"

struct sws_base
{
  sws_base( int depth )
  : _buffer( 1, 1, depth )
  , _sum( _buffer( 0, 0 ) )
  {}
  
  void init( CPUData<float> & window_c, CPUData<float> & window_xy ){};
  void update( CPUData<float> & left_edge_c, CPUData<float> & left_edge_xy, CPUData<float> & right_edge_c, CPUData<float> & right_edge_xy ){};
  CPUDataFeatureColumn<float>& get()
  {
    return _sum;
  }
  
protected:
  // remembered sum
  CPUData<float>              _buffer;
  CPUDataFeatureColumn<float> _sum;
};


struct sws_sparse_vel_accum : sws_base
{
  
  sws_sparse_vel_accum( int depth ) 
    : sws_base(depth)
    {
		_velocity = floor( 0.5f * (floor( std::sqrt( float(depth) ) ) - 1));
    }
  
  void init( CPUData<float> & window_c, CPUData<float> & window_xy )
  {
    std::fill( _sum.begin(), _sum.end(), 0.0f );
    
    for ( CPUData<float>::iterator i_c = window_c.begin(), i_xy = window_xy.begin(), iend = window_c.end();
          i_c != iend;
          ++i_c, ++i_xy )
    {
      const int i_dx = clamp( _velocity + (*i_xy)[0] + 0.5f, 0, 2 * _velocity );
      const int i_dy = clamp( _velocity + (*i_xy)[1] + 0.5f, 0, 2 * _velocity );
      const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
      _sum[ pos ] += (*i_c)[0];
    }
    
  }
  
  void update( CPUData<float> & left_edge_c, CPUData<float> & left_edge_xy, CPUData<float> & right_edge_c, CPUData<float> & right_edge_xy )
  {
    for ( CPUData<float>::iterator i_c = left_edge_c.begin(), i_xy = left_edge_xy.begin(), iend = left_edge_c.end();
          i_c != iend;
          ++i_c, ++i_xy )
    {
      const int i_dx = clamp( _velocity + (*i_xy)[0] + 0.5f, 0, 2 * _velocity );
      const int i_dy = clamp( _velocity + (*i_xy)[1] + 0.5f, 0, 2 * _velocity );
      const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
      _sum[ pos ] -= (*i_c)[0];
    }
    
    for ( CPUData<float>::iterator i_c = right_edge_c.begin(), i_xy = right_edge_xy.begin(), iend = right_edge_c.end();
          i_c != iend;
          ++i_c, ++i_xy )
    {
      const int i_dx = clamp( _velocity + (*i_xy)[0] + 0.5f, 0, 2 * _velocity );
      const int i_dy = clamp( _velocity + (*i_xy)[1] + 0.5f, 0, 2 * _velocity );
      const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
      _sum[ pos ] += (*i_c)[0];
    }

  }

private:

  inline int clamp( int val, int const& val_min, int const& val_max ) const
  {
    if ( val > val_max ) return val_max;
    if ( val < val_min ) return val_min;
    return val;
  }
  
  int _velocity;

};

/// this iterator  computes the mean over a window but ignores values smaller than selection_threshold
struct sws_sparse_vel_selective_mean : sws_base
{
  
  sws_sparse_vel_selective_mean( int depth, float selection_threshold = 1e-03f ) 
  : sws_base( depth )
  , _threshold(selection_threshold) 
  , _countbuffer( 1, 1, _buffer.depth() )
  , _resultbuffer( 1, 1, _buffer.depth() )
  , _count( _countbuffer( 0, 0 ) )
  , _result( _resultbuffer( 0, 0 ) )
  , _has_result( false )
  {
	  _velocity = floor( 0.5f * (floor( std::sqrt( float(depth) ) ) - 1));
  }
  
  void init( CPUData<float> & window_c, CPUData<float> & window_xy )
  {
    _has_result = false;
    std::fill( _sum.begin(), _sum.end(), 0.0f );
    std::fill( _count.begin(), _count.end(), 0.0f );
    float conf = 0.f;
    for ( CPUData<float>::iterator i_c = window_c.begin(), i_xy = window_xy.begin(), iend = window_c.end();
          i_c != iend;
          ++i_c, ++i_xy )
    {
      conf = (*i_c)[0];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + (*i_xy)[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + (*i_xy)[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] += conf;
        _count[pos] += 1;
      }
    }

  }
  
  void update(  CPUData<float> & left_edge_c, CPUData<float> & left_edge_xy, CPUData<float> & right_edge_c, CPUData<float> & right_edge_xy )
  { 
    _has_result = false;
    float conf = 0.f;
    for ( CPUData<float>::iterator i_c = left_edge_c.begin(), i_xy = left_edge_xy.begin(), iend = left_edge_c.end();
          i_c != iend;
          ++i_c, ++i_xy )
    {
      conf = (*i_c)[0];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + (*i_xy)[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + (*i_xy)[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] -= conf;
        _count[pos] -= 1;
      }
    }
    
    for ( CPUData<float>::iterator i_c = right_edge_c.begin(), i_xy = right_edge_xy.begin(), iend = right_edge_c.end();
          i_c != iend;
          ++i_c, ++i_xy )
    {
      conf = (*i_c)[0];
      if ( conf > _threshold )
      {
        const int i_dx = clamp( _velocity + (*i_xy)[0] + 0.5f, 0, 2 * _velocity );
        const int i_dy = clamp( _velocity + (*i_xy)[1] + 0.5f, 0, 2 * _velocity );
        const int pos  = i_dy + i_dx * ( 2 * _velocity + 1 ) ;
        _sum[ pos ] += conf;
        _count[pos] += 1;
      }
    }
  
  }

  CPUDataFeatureColumn<float> & get()
  {
    // compute mean for each bin and return it
    if ( !_has_result )
    {
      std::fill( _result.begin(), _result.end(), 0.0f );
      for ( unsigned i = 0; i < _sum.size(); ++i )
      {
        if ( _count[i] != 0 )
          _result[i] = _sum[i] / _count[i];
      }
      _has_result = true;
    }
    
    return _result;
  }

private:

  inline int clamp( int val, int const& val_min, int const& val_max ) const
  {
    if ( val > val_max ) return val_max;
    if ( val < val_min ) return val_min;
    return val;
  }
  
  int _velocity;
  float _threshold;
  CPUData<float> _countbuffer, _resultbuffer;
  CPUDataFeatureColumn<float> _count, _result;
  bool _has_result;

};


template< typename F = sws_sparse_vel_accum >
struct sliding_window_iterator_splitted
{

  
  sliding_window_iterator_splitted( const CPUData<float> & data_c, const CPUData<float> & data_xy, size_t window_size, size_t depth, F f )
    : _data_c( data_c )
    , _data_xy( data_xy )
    , _window_size( window_size )
    , _done(false)
    , _f(f)
  {
    //std::cout << "creating sliding_window_iterator_splitted for data_c = " << _data_c.dump_dims() << " data_xy =" << _data_xy.dump_dims() << " window_size = " << _window_size << std::endl;
    init();
  }    
  
  sliding_window_iterator_splitted( const CPUData<float> & data_c, const CPUData<float> & data_xy, size_t window_size, size_t depth )
  : _data_c( data_c )
    , _data_xy( data_xy )
    , _window_size( window_size )
    ,_row(0)
    ,_column(0)
    , _done(false)
    , _f( depth )
  {
    //std::cout << "creating sliding_window_iterator_splitted for data_c = " << _data_c.dump_dims() << " data_xy =" << _data_xy.dump_dims() << " window_size = " << _window_size << " depth = "<< depth<< std::endl;
    init();
  } 

  
  virtual void increment()
  {
    _column += 1;
    if ( _column > _data_c.columns() - _window_size )
    {
      _column = 0;
      _row += 1;
      if ( _row > _data_c.rows() - _window_size )
      {
        _done = true;
        return;
      }
      start_line( _row, _window_size );
    }
    else
    {
      move_right();
    }
  }
  
  virtual bool more() const { return !_done; }
  
  virtual CPUDataFeatureColumn<float> & operator*()
  {
    return _f.get();
  }
  
  virtual size_t window_size() const { return _window_size;};
  
  virtual F& core(){ return _f;};
  
private:

  void init( )
  {
    if (_data_c.rows() != _data_xy.rows() || _data_c.columns() != _data_xy.columns() )
      throw std::runtime_error(str( boost::format("Error in sliding window iterator splitted: data_c (%s) != data (%s) !")%_data_c.dump_dims()%_data_xy.dump_dims() ));
    
    if ( _window_size >= std::min( _data_c.rows(), _data_c.columns() ) )
    {
      _done = true;
    }
    else
      start_line(0,_window_size);
  }

  void start_line( size_t row, size_t window_size )
  {
    _window_c.reference( _data_c, row, window_size, 0, window_size );
    _window_xy.reference( _data_xy, row, window_size, 0, window_size );
    _f.init( _window_c, _window_xy );
  }
  
  void move_right( )
  {
    CPUData<float> left_edge_c( _window_c, 0, _window_c.rows(), 0, 1 );
    CPUData<float> left_edge_xy( _window_xy, 0, _window_xy.rows(), 0, 1 );
    
    _window_c.rebase( _window_c.offset() + _window_c.depth() );
    _window_xy.rebase( _window_xy.offset() + _window_xy.depth() );
    
    CPUData<float> right_edge_c( _window_c, 0, _window_c.rows(), _window_c.columns() - 1, 1 );
    CPUData<float> right_edge_xy( _window_xy, 0, _window_xy.rows(), _window_xy.columns() - 1, 1 );
    
    
    _f.update( left_edge_c, left_edge_xy, right_edge_c, right_edge_xy );
  }
  
  size_t _window_size;
  // the actual input data
  CPUData<float> _data_c;
  CPUData<float> _data_xy;
  
  // the current window
  CPUData<float> _window_c;
  CPUData<float> _window_xy;
  
  // current window position
  size_t _row, _column;

  // flag
  bool _done;
  // functor
  F _f;
};
  


#endif
