/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/// FIXME: error handling is very weak, in the work thread errors are completely
/// silent beyond the thread exiting

#include "boost/optional.hpp"
#include "boost/format.hpp"

#include "uEyeImageSequence.hpp"
#include "SeariseLogger.hpp"

/// if this is on the fifo mirrors the uEye image sequence and images
///  in the sequence are locked until consumed
/// dont know what happens in the uEye code if all the seq is locked...
#define LOCK_SEQ_IN_FIFO 1

void uEyeImageSequence::dump_cam_info( const UEYE_CAMERA_INFO & cam, size_t i ) const
{
      log().info << " camera " << ( i + 1 ) /*<< "/" << _available_cameras.size() */<< ": " << std::endl;
	  log().info << "    camera id: " << cam.dwCameraID << std::endl;
	  log().info << "    device id: " << cam.dwDeviceID << std::endl;
	  log().info << "    sensor id: " << cam.dwSensorID << std::endl;
	  log().info << "       in use: " << cam.dwInUse << std::endl;
	  log().info << "       serial: " << std::string( cam.SerNo, 16 ) << std::endl;
	  log().info << "        model: " << std::string( cam.Model, 16 ) << std::endl;
}

uEyeImageSequence::uEyeImageSequence( size_t id, size_t buffer_size )
  : _camera_handle( IS_INVALID_HIDS )
  , _images( buffer_size )
  , _image_fifo( buffer_size )
  , _transfer_failures(0)
  , _colorspace( HostImage::LUM )
  , _width(0)
  , _height(0)
{
  _trigger_enum
	.add_choice( IS_SET_TRIGGER_OFF       , "off" )
	.add_choice( IS_SET_TRIGGER_HI_LO     , "hi_lo" )
	.add_choice( IS_SET_TRIGGER_LO_HI     , "lo_hi" )
	.add_choice( IS_SET_TRIGGER_HI_LO_SYNC, "hi_lo_sync" )
	.add_choice( IS_SET_TRIGGER_LO_HI_SYNC, "lo_hi_sync" )
	.add_choice( IS_SET_TRIGGER_SOFTWARE  , "lo_hi_sync" );
  // select the first available camera
  _available_cameras.update();
  log().info() << "uEye: found " << _available_cameras.size() << " cameras" << std::endl;
  std::vector < int > camera_ids;
  for ( int i = 0; i < _available_cameras.size(); ++i )
	{
      
	  UEYE_CAMERA_INFO & cam = _available_cameras[i];
	  /*
      log().info << " camera " << ( i + 1 ) << "/" << _available_cameras.size() << ": " << std::endl;
	  log().info << "    camera id: " << cam.dwCameraID << std::endl;
	  log().info << "    device id: " << cam.dwDeviceID << std::endl;
	  log().info << "    sensor id: " << cam.dwSensorID << std::endl;
	  log().info << "       in use: " << cam.dwInUse << std::endl;
	  log().info << "       serial: " << std::string( cam.SerNo, 16 ) << std::endl;
	  log().info << "        model: " << std::string( cam.Model, 16 ) << std::endl;
	  */
      dump_cam_info( cam, i );
      camera_ids.push_back( int(cam.dwCameraID) );
      /*
		DWORD     dwCameraID;   // this is the user definable camera ID
		DWORD     dwDeviceID;   // this is the systems enumeration ID
		DWORD     dwSensorID;   // this is the sensor ID e.g. IS_SENSOR_UI141X_M
		DWORD     dwInUse;      // flag, whether the camera is in use or not
		IS_CHAR   SerNo[16];    // serial number of the camera
		IS_CHAR   Model[16];    // model name of the camera
		DWORD     dwStatus;     // various flags with camera status
		DWORD     dwReserved[15];
	  */
	}
  //if ( id >= _available_cameras.size() )
  //  throw std::runtime_error("wrong camera index!");
  int cam_number = 0;
  bool found = false;
  for ( ; cam_number < camera_ids.size(); ++cam_number )
  {
      if ( camera_ids[cam_number] == id )
      {
          found = true;
          break;
      }
  }
  if ( !found )
  {
      std::cout << "camera id "<< id << " not found in camera list:"<<std::endl;
      for ( int i = 0; i < _available_cameras.size(); ++i)
        dump_cam_info( _available_cameras[i], i );
      throw std::runtime_error("wrong camera id");
  }

  if ( _available_cameras.size() > 0 )
	_selected_camera = _available_cameras[id];
}

HostImage uEyeImageSequence::getNextImage()
{
  if ( hasNextImage() )
	{
	  _image_fifo.consumer_busy_wait( 5000 ); // poll freq of 5ms
	  _current_image = _image_fifo.get_head().copy();
	  if ( LOCK_SEQ_IN_FIFO )
		{
		  if ( is_UnlockSeqBuf( _camera_handle, IS_IGNORE_PARAMETER, 
								(char*)_image_fifo.get_head().cpu_data_ubyte().data() )
			   != IS_SUCCESS )
			throw std::runtime_error( "could not unlock image buffer" );
		}
	  _image_fifo.pop();
	}
  else
	_current_image.clear();
  return _current_image;
}

HostImage uEyeImageSequence::getCurrentImage()
{
  return _current_image;
}

bool uEyeImageSequence::hasNextImage()
{
  return _camera_thread;
}

int uEyeImageSequence::getWidth()
{
  return _width;
}

int uEyeImageSequence::getHeight()
{
  return _height;
}

void uEyeImageSequence::set_triggermode( int mode )
{
  if ( _camera_handle == IS_INVALID_HIDS )
	throw std::runtime_error( "uEye camera not open wehn setting triggermode" );
  is_SetExternalTrigger( _camera_handle, mode );
}

void uEyeImageSequence::set_triggermode( const std::string & mode )
{
  set_triggermode( _trigger_enum.str2enum( mode ) );
}

void uEyeImageSequence::set_colorspace( HostImage::ColorSpace colorspace )
{
}

void uEyeImageSequence::select_any_camera( )
{
  _selected_camera = boost::none;
  _available_cameras.update();
  if ( _available_cameras.size() > 0 )
	{
	  _selected_camera = _available_cameras[0];  
	  log().info << " uEye: selected first camera " << std::endl;
	}
  else
	throw std::runtime_error("could not find any uEye camera");

  open_camera();
}

void uEyeImageSequence::select_camera( unsigned camera_id )
{
  _available_cameras.update();
  _selected_camera = boost::none;
  
  for ( size_t i = 0; i < _available_cameras.size(); ++i )
	if ( _available_cameras[i].dwCameraID == camera_id )
	  {
		_selected_camera = _available_cameras[i];
		log().info << " uEye: selected camera with id " << camera_id << std::endl;
		break;
	  }
  
  if ( !_selected_camera )
	throw std::runtime_error("could not find requested uEye camera");

  open_camera();
}

void uEyeImageSequence::select_camera( const std::string & serial_number )
{  
  _available_cameras.update();
  _selected_camera = boost::none;

  for ( size_t i = 0; i < _available_cameras.size(); ++i )
	{
	  if ( strncmp( _available_cameras[i].SerNo, serial_number.c_str(), 16 ) == 0  )
		{
		  _selected_camera = _available_cameras[i];
		  log().info << " uEye: selected camera with serial " << serial_number << std::endl;
		  break;
		}
	}

  if ( !_selected_camera )
	throw std::runtime_error("could not find requested uEye camera");

  open_camera();
}

double uEyeImageSequence::set_exposure( double exposure )
{
  if ( _camera_handle == IS_INVALID_HIDS )
	throw std::runtime_error( "uEye camera not open wehn setting exposure" );
  double dbl_auto = exposure >= 0 ? 0.0 : 1.0;
  if ( is_SetAutoParameter ( _camera_handle, IS_SET_ENABLE_AUTO_SHUTTER, &dbl_auto, NULL) != IS_SUCCESS)
	{
	  throw std::runtime_error("uEye SetAutoShutter failed");
	}
  else
  {
      std::cout << "uEye: set auto exposure successfull!"<<std::endl;
  }
  if ( exposure >= 0 )
	{
	  double real_exposure;
	  is_SetExposureTime ( _camera_handle, exposure, &real_exposure );
	  return real_exposure;
	}
  else
	{
	  return -1;
	}
}
 
void uEyeImageSequence::set_gain( int gain )
{
  if ( _camera_handle == IS_INVALID_HIDS )
	throw std::runtime_error( "uEye camera not open when setting gain" );
  double dbl_auto = gain >= 0 ? 0.0 : 1.0;
  if ( is_SetAutoParameter ( _camera_handle, IS_SET_ENABLE_AUTO_GAIN, &dbl_auto, NULL) != IS_SUCCESS)
	{
	  throw std::runtime_error("uEye SetAutoGain failed");
	}
  if ( gain >= 0 )
	{
	  // no rgb setting
	  is_SetHardwareGain ( _camera_handle, gain, IS_IGNORE_PARAMETER
						                       , IS_IGNORE_PARAMETER
						                       , IS_IGNORE_PARAMETER );
	}
}
 
double uEyeImageSequence::set_framerate( double framerate )
{
  if ( _camera_handle == IS_INVALID_HIDS )
	throw std::runtime_error( "uEye camera not open wehn setting framerate" );
  double dbl_auto = framerate >= 0 ? 0.0 : 1.0;
  if ( is_SetAutoParameter ( _camera_handle, IS_SET_ENABLE_AUTO_FRAMERATE, &dbl_auto, NULL) != IS_SUCCESS)
	{
	  throw std::runtime_error("uEye SetAutoFPS failed");
	}
  if ( framerate >= 0 )
	{
	  double real_framerate;
	  is_SetFrameRate ( _camera_handle, framerate, &real_framerate );
	  std::cout << "intended framerate = "<< framerate << "  ral framerate = "<< real_framerate<<std::endl;
      return real_framerate;
	}
  else
	{
	  return -1;
	}
}
 
void uEyeImageSequence::stop( )
{
  if ( _camera_thread )
	{
	  _keep_going = false;
	  _camera_thread->join();
	  _camera_thread.reset();
	}
}

void uEyeImageSequence::close_camera()
{
  int ret;
  if ( _camera_handle != IS_INVALID_HIDS )
	{
	  ret = is_ExitCamera( _camera_handle );
	  if ( ret != 0 )
		throw std::runtime_error( "error closing uEye camera" );
	  _camera_handle = IS_INVALID_HIDS;
	}
}

void uEyeImageSequence::open_camera()
{
  int ret;

  close_camera();

  if ( !_selected_camera )
	throw std::runtime_error( "cannot open uEye capture with no selected camera" );

  // open the camera
  _camera_handle = (HIDS) ( _selected_camera->dwDeviceID | IS_USE_DEVICE_ID );
  ret = is_InitCamera( &_camera_handle, 0 );
  if ( ret != 0 )
	throw std::runtime_error( str( boost::format( "error initializing uEye camera, error: %i" ) % ret ) );

  // reset to defaults
  ret = is_ResetToDefault ( _camera_handle );
  if ( ret != 0 )
	throw std::runtime_error( str( boost::format( "error reseting uEye camera, error: %i" ) % ret ) );

  // collect some info
  if ((ret=is_GetCameraInfo (_camera_handle, &_camera_info)) != IS_SUCCESS)
	throw std::runtime_error( str( boost::format( "error getting uEye camera info, error: %i" ) % ret ) );

  if ((ret=is_GetSensorInfo (_camera_handle, &_sensor_info)) != IS_SUCCESS)
	throw std::runtime_error( str( boost::format( "error getting uEye sensor info, error: %i" ) % ret ) );

  log().info << "opened camera: " << _selected_camera->dwDeviceID << std::endl;
  log().info << "        width: " << _sensor_info.nMaxWidth << std::endl;
  log().info << "       height: " << _sensor_info.nMaxHeight << std::endl;

  INT xpos = 0, ypos = 0;
  _width = _sensor_info.nMaxWidth;
  _height = _sensor_info.nMaxHeight;
  is_SetAOI( _camera_handle, IS_SET_IMAGE_AOI, &xpos, &ypos, &_width, &_height);

}

void uEyeImageSequence::start()
{
  if ( !_selected_camera )
	throw std::runtime_error( "cannot start uEye capture with no selected camera" );

  stop( );

  // start the capture thread
  _keep_going = true;
  _camera_thread.reset( new boost::thread( boost::ref( *this ) ) );
}

int uEyeImageSequence::colormode()
{
  switch ( _colorspace )
	{
	case HostImage::LUM:
	  return IS_CM_MONO8;
	case HostImage::RGB:
	  return IS_CM_RGB8_PACKED;
	default: 
	  return IS_CM_MONO8; // FIXME: exception ?
	}
}

int uEyeImageSequence::bpp()
{
  switch ( _colorspace )
	{
	case HostImage::LUM:
	  return 8;
	case HostImage::RGB:
	  return 24;
	default: 
	  return 8; // FIXME: exception ?
	}
}

void uEyeImageSequence::operator()()
{
  // FIXME: actually report the errors somehow

  int ret;

  // set colormode
  if ( is_SetColorMode ( _camera_handle, colormode() ) != IS_SUCCESS )
	return;
  
  // alloc images
  alloc_images();

  // enable frame events
#ifndef __LINUX__
  ret = is_InitEvent( _camera_handle, _event_handle, _event );
  if ( ret != 0 )
	return;
#endif
  ret = is_EnableEvent( _camera_handle, _event );
  if ( ret != 0 )
	return;

  // start capture mode
  ret = is_CaptureVideo ( _camera_handle, IS_DONT_WAIT );
  if ( ret != 0 )
	return;

  log().info << " camera thread event loop enter " << std::endl;

  while ( _keep_going )
	{
#ifdef __LINUX__
        if (is_WaitEvent ( _camera_handle, _event, 1) == IS_SUCCESS)
#else
        if (WaitForSingleObject( _event_handle, 1000) == WAIT_OBJECT_0)
#endif
        {
		  handle_event( _event );
        }
    }

  log().info << " camera thread event loop exit " << std::endl;

  // stop capture mode
  is_StopLiveVideo ( _camera_handle, IS_WAIT );

  // disable frame event
  is_DisableEvent ( _camera_handle, _event );
#ifndef __LINUX__
  is_ExitEvent( _camera_handle, _event );
#endif
}

void uEyeImageSequence::process_frame()
{
  ++_framecount;

  if ( _image_fifo.full() )
	{
	  ++_dropcount;
      //log().info << " uEye: dropped frame " << std::endl;
	  return;
	}

  INT dummy = 0;
  char *pLast = NULL, *pMem = NULL;
  int ret = IS_SUCCESS;
  
  ret = is_GetActSeqBuf ( _camera_handle, &dummy, &pMem, &pLast );
  ret = is_LockSeqBuf( _camera_handle, IS_IGNORE_PARAMETER, pLast );

  // push out the data
  HostImage & out = _image_fifo.get_tail();
  if ( LOCK_SEQ_IN_FIFO )
	{
	  out = HostImage( HostImage::BData( (unsigned char*)pLast, _height, _width, ( bpp() + 1 ) / 8, /* weak */ true ) );
	}
  else
	{
	  out.resize( _width, _height, HostImage::UBYTE, _colorspace );
	  std::copy( pLast, pLast + out.numElements(), out.cpu_data_ubyte().data() );
	  ret |= is_UnlockSeqBuf( _camera_handle, IS_IGNORE_PARAMETER, pLast );;
	}
  _image_fifo.push();
}

void uEyeImageSequence::alloc_images( )
{
  INT ret = IS_SUCCESS;  
  ret = is_ClearSequence ( _camera_handle );
  for ( size_t i = 0; i < _images.size(); i++ )
    {
	  UEYE_IMAGE & image = _images[i];

	  if ( image.pBuf)
        {
		  ret = is_FreeImageMem ( _camera_handle, image.pBuf, image.nImageID);
		  image.clear();
        }
	  
	  int xpos, ypos;
	  is_SetAOI( _camera_handle, IS_GET_IMAGE_AOI, &xpos, &ypos, &_width, &_height);
	  
	  if ( is_AllocImageMem ( _camera_handle, _width, _height, bpp(), &image.pBuf, &image.nImageID) != IS_SUCCESS)
		throw std::runtime_error( "could not allocate uEye image buffer" );

	  if ( is_AddToSequence ( _camera_handle, image.pBuf, image.nImageID) != IS_SUCCESS)
		throw std::runtime_error( "could not enqueue uEye image buffer" );
	  
	  // leave room for adjust space
	  if ( _width % 4 != 0 )
		_width += 4 - _width % 4;

	  image.nImageSeqNum = i + 1;
	  image.nBufferSize = _width * _height * ( ( bpp() + 1 ) / 8 );
    }
}

void uEyeImageSequence::handle_event( int event )
{
    bool bUpdateCameraList = false;
    switch (event)
    {

    case IS_SET_EVENT_REMOVAL:
	  //printf ("device removed\n");
	  bUpdateCameraList = true;
	  break;

    case IS_SET_EVENT_NEW_DEVICE:
      //printf ("device added\n");
	  bUpdateCameraList = true;
	  break;

    case IS_SET_EVENT_STATUS_CHANGED:
	  //printf ("cameralist changed\n");
	  bUpdateCameraList = true;
	  break;

    case IS_SET_EVENT_TRANSFER_FAILED:
	  ++_transfer_failures; // could query more detailed information...
	  break;

    case IS_SET_EVENT_FRAME:
        //printf ("New Frame received\n");
        if (!_camera_handle )
		  break;
        process_frame ();
        break;

    default:
        break;
    }
}

int uEyeImageSequence::set_pixelclock( int clock )
{
  if ( _camera_handle == IS_INVALID_HIDS )
	throw std::runtime_error( "uEye camera not open wehn setting pixelclock" );
  if ( is_SetPixelClock( _camera_handle, clock ) != IS_SUCCESS )
	throw std::runtime_error( "uEye could not set pixelclock" );
  // what does ret mean?
  return is_SetPixelClock ( _camera_handle, IS_GET_PIXEL_CLOCK );
}
