/*

CPUDATA

  THIS MODULE PROVIDES MANAGEMENT OF 2+N-DIMENSIONAL DATA IN CPU MEMORY
  THIS INCLUDES DEALING WITH DIFFERENT KINDS OF IMAGE REPRESENTATIONS
  AND STREAMING OF IMAGES FROM VARIOUS SOURCES (VIDEO FILES, IMAGE FILES,
  HDF5 FILES, AND VARIOUS CAMERAS) AND TO VIDEO FILES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include <vector>
#include <string>

#include "boost/optional.hpp"
#include "boost/thread.hpp"
#include "boost/scoped_ptr.hpp"

#include <uEye.h> 

#include "HostImageSequence.h"
#include "simple_fifo.hpp"
#include "smart_enum.hpp"

/*!
 * \brief uEye Image parameter structure
 */
struct UEYE_IMAGE
{
    char *pBuf;
    INT nImageID;
    INT nImageSeqNum;
    INT nBufferSize;
  UEYE_IMAGE()
  {
	clear();
  }
  void clear()
  {
	pBuf = 0;
	nImageID = 0;
	nImageSeqNum = 0;
	nBufferSize = 0;
  }
};

/// wrapper for old school c-style extended array
struct uEye_camera_list_wrapper
{

  uEye_camera_list_wrapper()
  {
	resize( 0 );
  }

  size_t size()
  {
	return ( _data.size() - sizeof(DWORD) ) / sizeof(UEYE_CAMERA_INFO);
  }

  void resize( size_t n )
  {
	_data.resize( sizeof(DWORD) + n * sizeof(UEYE_CAMERA_INFO) );
	get_ptr()->dwCount = n;
  }

  UEYE_CAMERA_LIST * get_ptr()
  {
	return ((UEYE_CAMERA_LIST*)&_data.at(0));
  }

  UEYE_CAMERA_LIST & operator*()
  {
	return *get_ptr();
  }

  UEYE_CAMERA_LIST * operator->()
  {
	return get_ptr();
  }

  UEYE_CAMERA_INFO & operator[]( size_t i )
  {
	return get_ptr()->uci[i];
  }

  void update()
  {
	resize( 0 );
  
	if ( is_GetCameraList( get_ptr() ) != IS_SUCCESS )
	  {
		throw std::runtime_error( "error pre-querying list of uEye cameras" );
	  }
	
	resize( get_ptr()->dwCount ); 
  
	if ( is_GetCameraList( get_ptr() ) != IS_SUCCESS )
	  {
		resize( 0 );
		throw std::runtime_error( "error querying list of uEye cameras" );
	  } 
  }

private:

  std::vector<char> _data;

};

class uEyeImageSequence : public HostImageSequence
{
public:


  uEyeImageSequence( size_t id = 0, size_t buffer_size = 2 );

  void select_any_camera( );
  void select_camera( unsigned camera_id );
  void select_camera( const std::string & serial_number );

  /// returns the exposure that was actually set
  /// zero exposure is special and means max
  /// negative values mean auto exposure
  double set_exposure( double exposure );

  /// returns the rate that was actually set
  /// negative values mean auto fps
  double set_framerate( double exposure );

  /// returns the clock that was actually set
  int set_pixelclock( int clock  );

  void set_gain( int gain );

  /// only allowed now: HostImage::LUM and HostImage::RGB
  void set_colorspace( HostImage::ColorSpace colorspace );

  void start();
  void stop();
  void set_triggermode( int mode );  
  void set_triggermode( const std::string & mode );  

  void operator()();

  virtual HostImage getNextImage();

  virtual HostImage getCurrentImage();

  virtual bool hasNextImage();

  virtual int getWidth();

  virtual int getHeight();

private:

  void process_frame();
  void handle_event( int event );
  void open_camera();
  void close_camera();
  int colormode();
  int bpp();
  void alloc_images();
  void dump_cam_info( const UEYE_CAMERA_INFO & cam, size_t i ) const;

  boost::scoped_ptr<boost::thread> _camera_thread;

  smart_enum<int> _trigger_enum;

  uEye_camera_list_wrapper _available_cameras;
  boost::optional<UEYE_CAMERA_INFO> _selected_camera;
  HIDS _camera_handle;
  CAMINFO _camera_info; // Camera (Board)info
  SENSORINFO _sensor_info; // SensorInfo
  std::vector<UEYE_IMAGE> _images;
  CircularFifo<HostImage> _image_fifo;
  
  static const int _event = IS_SET_EVENT_FRAME;
  size_t _transfer_failures;

  volatile bool _keep_going;

  HostImage::ColorSpace _colorspace;
  int _width, _height;

  volatile size_t _framecount;
  volatile size_t _dropcount;

  HostImage _current_image;

#ifndef __LINUX__                                                                                                                    
    /* on windows we need an Event handle member */                                                                                  
    HANDLE _event_handle;
#endif                                                                                                                               

};
