#define DEBUG

#include <algorithm>
#include <numeric>
#include <cmath>

#include <boost/foreach.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>

#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/ui/text/TestRunner.h>

#include <CPUData.hpp>
#include <CPUDataAlgorithms.hpp>
#include <CPUDataRFIterator.hpp>

class CPUDataTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE(CPUDataTest);
	CPPUNIT_TEST(testCreation);
	CPPUNIT_TEST(testDeviceExchanges);
	CPPUNIT_TEST(testView);
	CPPUNIT_TEST(testFill);
	CPPUNIT_TEST(testMul);
	CPPUNIT_TEST(testDot);
	CPPUNIT_TEST(testSum);
	CPPUNIT_TEST(testSubrange);
//	CPPUNIT_TEST(testMinMax);
	CPPUNIT_TEST(testAddMult);
	CPPUNIT_TEST(testClamp);	
	CPPUNIT_TEST(testLayer);
	CPPUNIT_TEST(testIteratorRF);
	CPPUNIT_TEST(testSlidingWindow);
	CPPUNIT_TEST(testSlidingWindow2);
	CPPUNIT_TEST_SUITE_END();
	
public:
	
	void testCreation() 
	{
		CPUData<float> a(12,23,10);
		CPPUNIT_ASSERT( !a.empty() );
	}

	void testLayer()
	{
		CPUDataLayer<float> layer(10,10);
		for ( int i = 0; i < layer.rows(); ++i )
			for ( int j = 0; j < layer.columns(); ++j )
				layer( i, j ) = i + j;

		//BOOST_FOREACH( float v, layer )
			//std::cout << v << ", " << std::flush;
		for ( int i = 0; i < layer.rows(); ++i )
			for ( int j = 0; j < layer.columns(); ++j )
				CPPUNIT_ASSERT( layer( i, j ) == i + j );
	}

	void testSubrange()
	{
		CPUData<float> a(4,4,3);
		std::fill( a.begin(), a.end(), 0.0f );
		// set some values
		a(1,1,1) = 1;
		a(1,2,1) = 2;
		a(2,1,1) = 3;
		a(2,2,1) = 4;
		CPUData<float> b( a, 1, 2, 1, 2 );
		// look at the values from the view
		CPPUNIT_ASSERT( b(0,0,1) == 1 );
		CPPUNIT_ASSERT( b(0,1,1) == 2 );
		CPPUNIT_ASSERT( b(1,0,1) == 3 );
		CPPUNIT_ASSERT( b(1,1,1) == 4 );
		// fill the view with some values
		std::fill( b.begin(), b.end(), -1.0f );
		b(0,0,1) = -2.0f;
		// look at the original container
		CPPUNIT_ASSERT( a(1,1,0) == -1.0f );
		CPPUNIT_ASSERT( a(1,1,1) == -2.0f );
		CPPUNIT_ASSERT( a(1,1,2) == -1.0f );
		for ( int i = 0; i < 3; ++i )
			{
				for ( int j = 0; j < 4; ++j )
					{
						CPPUNIT_ASSERT( a(0,j,i) == 0 );
						CPPUNIT_ASSERT( a(3,j,i) == 0 );
						CPPUNIT_ASSERT( a(j,0,i) == 0 );
						CPPUNIT_ASSERT( a(j,3,i) == 0 );
					}
				CPPUNIT_ASSERT( a(1,2,i) == -1.0f );
				CPPUNIT_ASSERT( a(2,2,i) == -1.0f );
				CPPUNIT_ASSERT( a(2,1,i) == -1.0f );
			}
	}

	void testDeviceExchanges() {
		CPUData<float> a(12,23,10);
		std::fill( a.begin(), a.end(), 0.0f );
		a(1,1,1) = 1;
		a(2,2,3) = 2;
		CPPUNIT_ASSERT(a(1,1,1) == 1);		
		CPPUNIT_ASSERT(a(2,2,3) == 2);		
	}
	
	void testView() 
	{
		CPUData<float> d(2, 3, 4);
		std::fill( d.begin(), d.end(), 1 );
		CPUData<float> e( d, 1, 1, 1, 2 );
		std::fill( e.begin(), e.end(), 2 );
		for ( unsigned i = 0; i < d.rows(); ++i )
			for ( unsigned j = 0; j < d.columns(); ++j )
				for ( unsigned k = 0; k < d.depth(); ++k )
					CPPUNIT_ASSERT( d(i,j,k) == ( i < 1 || j < 1 ) ? 1 : 2 );
		CPUData<float> f( 2, 3, 4 );
		std::fill( f.begin(), f.end(), 3 );
		CPUData<float> g( f, 0, 1, 0, 2 );
		std::copy( e.begin(), e.end(), g.begin() );
		for ( unsigned i = 0; i < d.rows(); ++i )
			for ( unsigned j = 0; j < d.columns(); ++j )
				for ( unsigned k = 0; k < d.depth(); ++k )
					CPPUNIT_ASSERT( f(i,j,k) == ( i > 0 || j > 2 ) ? 3 : 2 );
	}
	
	void testFill() 
	{
		int dim1DSize = 11*11*11;

		//manual fill
		CPUData<float> a(11,11,11);
		for (int i=0; i<dim1DSize; ++i) {
			a.data()[i] = 0.0f;
		}
		for (int l = 0; l<11; l++) {
			for (int k = 0; k<11; k++) {
				for (int j = 0; j<11; j++) {
					CPPUNIT_ASSERT(a(l,k,j) == 0);
				}
			}
		}
		
		//data iterator fill
		std::fill( a.data(), a.data_end(), 2.345f );
		for (int l = 0; l<11; l++) {
			for (int k = 0; k<11; k++) {
				for (int j = 0; j<11; j++) {
					CPPUNIT_ASSERT(a(l,k,j) == 2.345f);
				}
			}
		}
		
		//feature-column iterator fill
		std::fill( a.begin(), a.end(), 2 );
		for (int l = 0; l<11; l++) {
			for (int k = 0; k<11; k++) {
				for (int j = 0; j<11; j++) {
					CPPUNIT_ASSERT(a(l,k,j) == 2);
				}
			}
		}
	}
	
	void testMul() 
	{
		CPUData<float> a(5,4,3);
		CPUData<float> b(5,4,3);
		std::fill(a.data(),a.data_end(),2.25f);
		std::copy(a.data(),a.data_end(),b.data());
		b(2,0,0) = 1.0f;
		typedef boost::tuple<float&,float&> reftuple;
		BOOST_FOREACH(reftuple ab, zip2(a.data_range(),b.data_range()))
			ab.get<0>() *= ab.get<1>();
		
		CPPUNIT_ASSERT(a(0,0,0) == 2.25f*2.25f);
		CPPUNIT_ASSERT(a(2,0,0) == 1.00f*2.25f);
		
		BOOST_FOREACH(reftuple ab, zip2(a.data_range(),b.data_range()))
			ab.get<0>() *= ab.get<1>() * 33.0f;
		CPPUNIT_ASSERT(a(0,0,0) == 2.25f*2.25f*2.25f*33.0f);
		CPPUNIT_ASSERT(a(2,0,0) == 2.25f*1.00f*1.00f*33.0f);
		
		BOOST_FOREACH(float& aa, a.data_range())
			aa *= 0.4f;
		
		CPPUNIT_ASSERT(a(0,0,0) == 2.25f*2.25f*2.25f*33.0f*0.4f);
		CPPUNIT_ASSERT(a(2,0,0) == 2.25f*1.00f*1.00f*33.0f*0.4f);
	}
		
	typedef float(*floatfun1)(float); //need to extract overloaded function
	void testSum() {
		using namespace boost::lambda;
		CPUData<float> a(30,30,1);
		std::fill(a.data(),a.data_end(),3.0f);
		a(2,2,0)=-1.5f;
		CPPUNIT_ASSERT(std::accumulate(a.data(),a.data_end(),float(0))==30*30*3-3-1.5f);
		CPPUNIT_ASSERT(std::accumulate(a.data(),a.data_end(),float(0),_1 + bind((floatfun1)std::fabs,_2))==30*30*3-3+1.5f);
	}

	void testDot() {
		CPUData<float> a(5,4,3);
		CPUData<float> b(5,4,3);
		CPUData<float> c(5,4,1);
		std::fill(a.data(),a.data_end(),2.25f);
    a(2,2,2) = 2.0f;
		std::copy(a.data(),a.data_end(),b.data());
    typedef CPUData<float>::ColumnT Column;
		typedef boost::tuple<Column&,Column&,Column&> reftuple3;
    BOOST_FOREACH(reftuple3 abc, zip3(a.range(),b.range(),c.range()))
    {
      abc.get<2>()[0] = dot(abc.get<0>(),abc.get<1>());
    }
		CPPUNIT_ASSERT(std::accumulate(c.data(),c.data_end(),float(0))==(5*4*3-1)*2.25f*2.25f+2.0f*2.0f);
		CPPUNIT_ASSERT(c(2,2,0)==2*2.25f*2.25f+2.0f*2.0f);
  }
  
#if 0
	void testMinMax() {
		int size = 2*16*1024+4;
		GPUData a(1,&size);
		a.fill(0.1f);
		a[0] = -2.0f;
		a[2] =  111.0f;
		a[2*16*1024+4-1] = -2.1f;
		a[1*16*1024+2]=-222.0f;
		CPPUNIT_ASSERT(a.max()==111.0f);
		CPPUNIT_ASSERT(a.absMax()==222.0f);
		
		CPPUNIT_ASSERT(a.min()==-222.0f);
		CPPUNIT_ASSERT(a.absMin()==0.1f);				
	}
#endif

  void testIteratorRF()
  {


    size_t rows = 15;
    size_t columns = 20;
    size_t depth = 3;

    float value = 1.f;
    int bin_size = 5;

    CPUData<float> A(bin_size*rows,bin_size*columns,depth);
    std::fill( A.begin(), A.end(), 0.f );
    


    CPUDataRFIterator<float> iter( A, bin_size, bin_size );

    while ( iter.more() )
    {
      CPUData<float> B = *iter;
      for (unsigned i=0; i < B.rows(); ++i)
      {
        for (unsigned j=0; j < B.columns(); ++j)
        {
          for (unsigned k=0; k < B.depth(); ++k)
          {
            B(i,j,k) += value;
          }
        }
      }
      iter.increment();
    }

    for  (unsigned i = 0; i < A.rows(); i++)
      for  (unsigned j = 0; j < A.columns(); j++)
        CPPUNIT_ASSERT( A(i,j,0) == value );

  }

  void testSlidingWindow()
  {
  	// init with 0
  	int binsize	= 5;
  	int rows	= 30;
  	int cols	= 20;
  	int depth	= 3;  
  	CPUData<float> A( rows, cols, depth );
  	std::fill( A.begin(), A.end(), 1.f );
  	CPUData<float> sums( rows - (binsize-1), cols - (binsize-1), 1 );
  	std::fill( sums.begin(), sums.end(), 0.f );
  	float sum = 0.f;


 
  	// slide over all columns
  	for (int i = 0; i < A.columns() - (binsize-1); ++i )
  	{
  		CPUDataRFIterator<float> iter( A, binsize, 1, i );
  		CPUDataRFIterator<float> prev_iter = iter;
  		CPUData<float> currentWindow = *iter;
  		
 
		
		// first in line: take all into account
		CPUData<float>::iterator sub_iter = currentWindow.begin();
		for ( ; sub_iter != currentWindow.end(); ++sub_iter )
		{
			/***
			 * here the first window is accessed 
			***/
			// sum over the depth
			for (int k = 0; k < (*sub_iter).size(); k++)
				sum += (*sub_iter)[k];
			sums( 0, i, 0 ) = sum;
		}
		iter.increment();
		
		// run over complete column
		for (int l = 1; l < A.rows() - (binsize-1); l++ )  
		{
		
			CPUData<float> currentWindow = *iter;
			
			// remove old
			CPUData<float> lastWindow = *prev_iter;
			
			for (int j = 0; j < lastWindow.rows(); j++ )
			{
				/***
				 * here you can acces the "left over" row 
				***/
				for (int k = 0; k < lastWindow.depth(); k++ )
				{
					 sum -= lastWindow( j, 0, k );
				}
			}
				
			// add new ones
  			for (int j = 0; j < currentWindow.rows(); j++ )
			{
				/***
				 * here you can acces the new row 
				***/
				for (int k = 0; k < currentWindow.depth(); k++ )
				{		
					sum += currentWindow( j, binsize-1, k );
				}
			}
			sums( l, i, 0 ) = sum;
			
			// set iterators
			prev_iter = iter;
			iter.increment();	
		}
		// reset the sum!
		sum = 0;
	}
	CPUData<float>::iterator test_iter = sums.begin();
	for ( ; test_iter != sums.end(); ++test_iter )
		CPPUNIT_ASSERT( (*test_iter)[0] == binsize*binsize*depth  );
  		
  }
  
  void testSlidingWindow2()
  {
  	//TODO: replace hand made substrations and adding by iterator functions
  	
  	// init with 0
  	int binsize	= 5;
  	int rows	= 30;
  	int cols	= 20;
  	int depth	= 3;  
  	CPUData<float> A( rows, cols, depth );
  	std::fill( A.begin(), A.end(), 1.f );
  	CPUData<float> sums( rows - (binsize-1), cols - (binsize-1), 1 );
  	std::fill( sums.begin(), sums.end(), 0.f );
  	float sum = 0.f;


 
  	// slide over all columns
  	for (int i = 0; i < A.columns() - (binsize-1); ++i )
  	{
  		CPUDataRFIterator<float> iter( A, binsize, 1, i );
  		CPUDataRFIterator<float> prev_iter = iter;
  		CPUData<float> currentWindow = *iter;
  		
  		
		
		// first in line: take all into account
		CPUData<float>::iterator sub_iter = currentWindow.begin();
		for ( ; sub_iter != currentWindow.end(); ++sub_iter )
		{
			/***
			 * here the first window is accessed 
			***/
			// sum over the depth
			for (int k = 0; k < (*sub_iter).size(); k++)
				sum += (*sub_iter)[k];
			sums( 0, i, 0 ) = sum;
		}
		iter.increment();
		CPUData<float> testWindow =   iter.substract( prev_iter );
		
		// run over complete column
		for (int l = 1; l < A.rows() - (binsize-1); l++ )  
		{
		
			CPUData<float> currentWindow = *iter;
			
			// remove old
			CPUData<float> lastWindow = *prev_iter;
			
			for (int j = 0; j < lastWindow.rows(); j++ )
			{
				/***
				 * here you can acces the "left over" row 
				***/
				for (int k = 0; k < lastWindow.depth(); k++ )
				{
					 sum -= lastWindow( j, 0, k );
				}
			}
				
			// add new ones
  			for (int j = 0; j < currentWindow.rows(); j++ )
			{
				/***
				 * here you can acces the new row 
				***/
				for (int k = 0; k < currentWindow.depth(); k++ )
				{		
					sum += currentWindow( j, binsize-1, k );
				}
			}
			sums( l, i, 0 ) = sum;
			
			// set iterators
			prev_iter = iter;
			iter.increment();	
		}
		// reset the sum!
		sum = 0;
	}
	CPUData<float>::iterator test_iter = sums.begin();
	for ( ; test_iter != sums.end(); ++test_iter )
		CPPUNIT_ASSERT( (*test_iter)[0] == binsize*binsize*depth  );
  		
  }
  

  void testAddMult()
    {
		CPUData<float> a(5,4,3);
		CPUData<float> b(5,4,3);
		CPUData<float> c(5,4,3);
		std::fill(a.begin(),a.end(),3.f);
		std::fill(b.begin(),b.end(),3.f);
		
		std::fill(c.begin(),c.end(),3.f);
		
		b += c;
		a *= c;
		
		
		for  (unsigned i = 0; i < a.rows(); i++)
      		for  (unsigned j = 0; j < a.columns(); j++)
      			for  (unsigned k = 0; k < a.depth(); k++)
					CPPUNIT_ASSERT( a(i,j,k)== 9.f );
		
		for  (unsigned i = 0; i < b.rows(); i++)
      		for  (unsigned j = 0; j < b.columns(); j++)
      			for  (unsigned k = 0; k < b.depth(); k++)
					CPPUNIT_ASSERT( b(i,j,k)== 6.f );

    }
	
  void testClamp()
    {
 		CPUData<float> a(5,4,3);
 		std::fill(a.begin(),a.end(),0.f);
 		a(1,1,1) = -100.f;
 		a(1,1,1) = 100.f;
 		
 		a.clamp( 0.f, 1.f );
 		
 		for  (unsigned i = 0; i < a.rows(); i++)
      		for  (unsigned j = 0; j < a.columns(); j++)
      			for  (unsigned k = 0; k < a.depth(); k++)
					CPPUNIT_ASSERT( a(i,j,k)>= 0.f && a(i,j,k)<= 1.f );
    }

};

CPPUNIT_TEST_SUITE_REGISTRATION(CPUDataTest);

int main() {
	CppUnit::TextUi::TestRunner runner;
	CppUnit::TestFactoryRegistry& registry = CppUnit::TestFactoryRegistry::getRegistry();
	runner.addTest(registry.makeTest());
	return ! runner.run("", false);
}

// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
