#define DEBUG

#include <algorithm>
#include <numeric>
#include <cmath>

#include <boost/foreach.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>

#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/ui/text/TestRunner.h>

#include "CPUDataND.hpp"
#include "CPUDataNDExpressions.hpp"
#include "CPUDataNDExpressionIterators.hpp"
#include "IndexExpressions.hpp"

class CPUDataNDTest : public CppUnit::TestFixture {
  CPPUNIT_TEST_SUITE(CPUDataNDTest);
  CPPUNIT_TEST(testIndexExpressions);
  CPPUNIT_TEST(test_CPUDataNDExpressions);
  CPPUNIT_TEST(test_CPUDataNDIndexing);
  CPPUNIT_TEST(testIndexStrideEvaluation);
  CPPUNIT_TEST(testIndexOffsetEvaluation);
  //CPPUNIT_TEST(testTerminalIteratorBuilder);
  CPPUNIT_TEST(testExpressionIteratorBuilder);
  CPPUNIT_TEST(testBinaryExpressionAssignment);
  CPPUNIT_TEST_SUITE_END();
	
public:
	
	typedef CPUDataND<float> DataF;
	
	template<typename Expr>  
	const void test_index_expr(Expr const &expr)
	{
		//boost::proto::display_expr( expr );
#if HAVE_INDEX_GRAMMAR
		CPPUNIT_ASSERT((boost::proto::matches<Expr, index_grammar>::value));
#endif
	}

	template<typename Expr, typename Builder>
	void doTestExpressionIteratorBuilder( Expr & expr, Builder & builder )
	{
	  typedef typename boost::result_of<Builder(Expr)>::type ExprIterator;
	  ExprIterator iterator = builder( expr );	  
	}

	void testExpressionIteratorBuilder()
	{
	  typedef mpl::vector< mpl::int_<0>, mpl::int_<1> > IndexSet01;
	  typedef ExpressionIteratorBuilder<float,IndexSet01> Builder;
	  float f1[] = {1,2,3,4};
	  float f2[] = {-1,-2,-3,-4};
	  DataF A(2,2);
	  DataF B(2,2);
	  std::copy(f1,f1+4,A.data());
	  std::copy(f2,f2+4,B.data());

	  CPPUNIT_ASSERT( A(_i,_j).at(0,0) ==  1 );
	  CPPUNIT_ASSERT( A(_i,_j).at(0,1) ==  2 );
	  CPPUNIT_ASSERT( A(_i,_j).at(1,0) ==  3 );
	  CPPUNIT_ASSERT( A(_i,_j).at(1,1) ==  4 );
	  
	  A(_i,_j) = B(_j,_i) + 1;
	  
	  CPPUNIT_ASSERT( ( 1 + A(_i,_j) ).at(0,0) == 1 );
	  CPPUNIT_ASSERT( ( 1 + A(_i,_j) ).at(1,0) == 0 );

	  CPPUNIT_ASSERT( A(_i,_j).at(0,0) ==  0 );
	  CPPUNIT_ASSERT( A(_i,_j).at(1,0) == -1 );
	  CPPUNIT_ASSERT( A(_i,_j).at(0,1) == -2 );
	  CPPUNIT_ASSERT( A(_i,_j).at(1,1) == -3 );
	  
	  Builder builder;
	  doTestExpressionIteratorBuilder( A(_i,_j), builder  );
	  doTestExpressionIteratorBuilder( A(_i,_j), builder  );
	}

	void testBinaryExpressionAssignment()
	{
	  typedef mpl::vector< mpl::int_<0>, mpl::int_<1> > IndexSet01;
	  typedef ExpressionIteratorBuilder<float,IndexSet01> Builder;
	  float f1[] = {1,2,3,4};
	  float f2[] = {-1,-2,-3,-4};
	  DataF A(2,2);
	  DataF B(2,2);
	  std::copy(f1,f1+4,A.data());
	  std::copy(f2,f2+4,B.data());

	  A(_i,_j) = B(_j,_i) * B(_i,_j);
	  
	  CPPUNIT_ASSERT( A(_i,_j).at(0,0) ==  1 );
	  CPPUNIT_ASSERT( A(_i,_j).at(1,0) ==  6 );
	  CPPUNIT_ASSERT( A(_i,_j).at(0,1) ==  6 );
	  CPPUNIT_ASSERT( A(_i,_j).at(1,1) == 16 );

	  A(_i,_j).at(1,0) = 10;

	  CPPUNIT_ASSERT( A(_i,_j).at(1,0) ==  10 );

	  CPPUNIT_ASSERT( ( B(_j,_i) * B(_i,_j) ).at(1,1) == 16 );
	  
	}

  void testIndexOffsetEvaluation()
  {
    IndexOffset evaluate;
    CPPUNIT_ASSERT( evaluate( _i ) == 0 );
    CPPUNIT_ASSERT( evaluate( _i + 1 ) == 1 );
    CPPUNIT_ASSERT( evaluate( 1 + _i ) == 1 );
    CPPUNIT_ASSERT( evaluate( proto::lit(1) ) == 1 );
    CPPUNIT_ASSERT( evaluate( 2 * (_i + 1) ) == 2 );
    CPPUNIT_ASSERT( evaluate( 2 * (_i - 1) ) == -2 );
    CPPUNIT_ASSERT( evaluate( 2 * (_i - 1) + 2 ) == 0 );
    CPPUNIT_ASSERT( evaluate( 2 * (_i - 1) + 2 * (_j + 1) ) == 0 );
  }
  
  void testIndexStrideEvaluation()
  {
    IndexStride<0> evaluate0;
    IndexStride<1> evaluate1;
    CPPUNIT_ASSERT( evaluate0( _i ) == 1);
    CPPUNIT_ASSERT( evaluate0( 2 * _i ) == 2);
    CPPUNIT_ASSERT( evaluate0( _i * 2 ) == 2);
    CPPUNIT_ASSERT( evaluate0( _i + 1 ) == 1);
    CPPUNIT_ASSERT( evaluate1( _i ) == 0);
    CPPUNIT_ASSERT( evaluate1( 2 * _i ) == 0);
    CPPUNIT_ASSERT( evaluate1( _i * 2 ) == 0);
    CPPUNIT_ASSERT( evaluate1( _i * 2 + _j ) == 1);
    CPPUNIT_ASSERT( evaluate1( 2 * ( _i * 2 + _j ) ) == 2);
  }
  
	void testIndexExpressions()
	{
		ptrdiff_t data[] = { 1, 2, 3, 4 };
		std::transform( data, data + 4, data, _i * _i );
		for( ptrdiff_t i = 0; i < 4; ++i )
			{
				CPPUNIT_ASSERT( data[i] == (i+1) * (i+1) );
			}
		test_index_expr( _i + _i );
		CPPUNIT_ASSERT( ( _i + _i )(1) == 2 );
		test_index_expr( _i + 1 );
		CPPUNIT_ASSERT( ( _i + 1 )(1) == 2 );
		test_index_expr( ( _i + 1 + _j ) * _k );
		CPPUNIT_ASSERT( ( ( _i + 1 + _j ) * _k )( 0, 1, 2 ) == 4 );
	}
	
	template<typename Expr>  
	void test_cpudata_expr(Expr const &expr)
	{
		boost::proto::display_expr( expr );
		CPPUNIT_ASSERT((boost::proto::matches<Expr, cpudata_grammar>::value));
	}
	
	void test_CPUDataNDExpressions()
	{
		using boost::proto::lit;
		DataF A(2,2);
		test_cpudata_expr( A ( _i, _j ) + A );
	}

  void test_CPUDataNDIndexing()
  {
    cpudata_indexing_ctx<2,float> ctx(1,1);
		DataF A(2,2);
    for ( unsigned i = 0; i < A.size(); ++i )
      A.data()[i] = i+1;
    CPPUNIT_ASSERT( proto::eval( A(_i,_j) + A(_j,_i-1), ctx ) == 7 );
    CPPUNIT_ASSERT( proto::eval( A(_i-1,_j*2-_i), ctx ) == 2 );
  }
  
};

CPPUNIT_TEST_SUITE_REGISTRATION(CPUDataNDTest);

int main() {
	CppUnit::TextUi::TestRunner runner;
	CppUnit::TestFactoryRegistry& registry = CppUnit::TestFactoryRegistry::getRegistry();
	runner.addTest(registry.makeTest());
	return ! runner.run("", false);
}

