/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"
#include <stdio.h>
__global__ void d_linFirstOrderDGLEulerStep(
								float* data,
								const int w,
								const int h,
								const float a,
								const float* A,
								const float b,
								const float* B,
								const float C,
								const float d,
								const float* D,
								const float e,
								const float* E,
								const float F,
								const float dt,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float b_;
		__shared__ float e_;
		if (threadIdx.x==0) b_ = b*B[bIndex];
		if (threadIdx.x==2) e_ = e*E[bIndex];

		__syncthreads();

		const float x_ = data[ij];
		const float a_ = a*A[ij];
		const float d_ = d*D[ij];

		data[ij] = x_ + dt * ( x_ * (a_+b_+C) + (d_+e_+F) );

}
__global__ void d_linFirstOrderDGLImplicitEulerStep(
								float* data,
								const int w,
								const int h,
								const float a,
								const float* A,
								const float b,
								const float* B,
								const float C,
								const float d,
								const float* D,
								const float e,
								const float* E,
								const float F,
								const float dt,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float b_;
		__shared__ float e_;
		if (threadIdx.x==0) b_ = b*B[bIndex];
		if (threadIdx.x==2) e_ = e*E[bIndex];

		__syncthreads();

		const float x_ = data[ij];
		const float a_ = a*A[ij];
		const float d_ = d*D[ij];

		data[ij] = ( x_ + dt * (d_+e_+F) ) / ( 1 - dt * (a_+b_+C) );

}

__global__ void d_linFirstOrderDGLSteadyStateSolution(
								float* data,
								const int w,
								const int h,
								const float a,
								const float* A,
								const float b,
								const float* B,
								const float C,
								const float d,
								const float* D,
								const float e,
								const float* E,
								const float F,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float b_;
		__shared__ float e_;
		if (threadIdx.x==0) b_ = b*B[bIndex];
		if (threadIdx.x==2) e_ = e*E[bIndex];

		__syncthreads();

		const float a_ = a*A[ij];
		const float d_ = d*D[ij];

		data[ij] = -1.0f * ( (d_+e_+F) / (a_+b_+C) );

}

__global__ void d_linFirstOrderDGL2EulerStep(
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const float dt,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float c_;
		if (threadIdx.x==2) c_ = c*C[bIndex];

		__syncthreads();

		const float x_ = data[ij];
		const float b_ = b*B[ij];

		data[ij] = x_ + dt * ( -a*x_ + (1.0f-x_) * (b_ + c_) );

}
__global__ void d_linFirstOrderDGL2EulerStepRect(
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const float dt,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float c_;
		if (threadIdx.x==2) c_ = c*C[bIndex];

		__syncthreads();

		const float x_ = data[ij];
		const float b_ = b*B[ij];

		data[ij] = x_ + dt * ( -a*x_ + (1.0f-x_) * fmax(0.0f,(b_ + c_)) );
}

__global__ void d_linFirstOrderDGL2ImplicitEulerStep(
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const float dt,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float c_;
		if (threadIdx.x==2) c_ = c*C[bIndex];

		__syncthreads();

		const float x_ = data[ij];
		const float b_ = b*B[ij];

		data[ij] = ( x_ + dt * (b_+c_) ) / ( 1 + dt * (a +b_+c_) );

}

__global__ void d_linFirstOrderDGL2ImplicitEulerStepRect(
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const float dt,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float c_;
		if (threadIdx.x==2) c_ = c*C[bIndex];

		__syncthreads();

		const float x_ = data[ij];
		const float b_c_ = fmax(0.0f,b*B[ij]+c_);

		data[ij] = ( x_ + dt * (b_c_) ) / ( 1 + dt * (a +b_c_) );
}

__global__ void d_linFirstOrderDGL2SteadyStateSolution(
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float c_;
		if (threadIdx.x==2) c_ = c*C[bIndex];

		__syncthreads();

		const float b_ = b*B[ij];

		data[ij] = (b_+c_) / (a +b_+c_) ;
}

__global__ void d_linFirstOrderDGL2SteadyStateSolutionRect(
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const bool colMajor
) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// index Offset
		const int ij = colMajor ? bIndex + threadIdx.x * h  : bIndex * blockDim.x + threadIdx.x ;

		__shared__ float c_;
		if (threadIdx.x==2) c_ = c*C[bIndex];

		__syncthreads();

		const float b_c_ = fmax(0.0f,b*B[ij]+c_);

		data[ij] = (b_c_) / (a +b_c_) ;
}

extern "C" void do_linFirstOrderDGL(const int type,
								float* data,
								const int w,
								const int h,
								const float a,
								const float* A,
								const float b,
								const float* B,
								const float C,
								const float d,
								const float* D,
								const float e,
								const float* E,
								const float F,
								const float dt,
								const bool colMajor
) {
	if (w>512) {
		fprintf(stderr,"errorn in do_linFirstOrderDGL()\n : width > 512 ; this is not yet supported\n");
		return;
	}
	dim3 dimBlock(w);
//	int s1 = h;
//	int s2 = 1;
//	while (s1>=65535) {
//		s2 *= 2;
//		s1  = iDivUp(s1,2);
//	}
//	dim3 dimGrid(s1,s2);
	dim3 dimGrid = make_2d_gridDim_for_1d_blockDim(1,h);
	if (type==1) {
		d_linFirstOrderDGLEulerStep<<<dimGrid,dimBlock>>>(
									data,
									 w,
									 h,
									 a, A, b, B,
									 C,
									 d, D, e, E,
									 F,
									 dt,colMajor);
	} else if (type==2) {
		d_linFirstOrderDGLImplicitEulerStep<<<dimGrid,dimBlock>>>(
								data,
								 w,
								 h,
								 a, A, b, B,
								 C,
								 d, D, e, E,
								 F,
								 dt,colMajor);
	} else if (type==3) {
		d_linFirstOrderDGLSteadyStateSolution<<<dimGrid,dimBlock>>>(
								data,
								 w,
								 h,
								 a, A, b, B,
								 C,
								 d, D, e, E,
								 F,colMajor);
	}
}

extern "C" void do_linFirstOrderDGL2(const int type,
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const float dt,
								const bool rect_input,
								const bool colMajor
) {
	if (w>512) {
		fprintf(stderr,"errorn in do_linFirstOrderDGL2()\n : width > 512 ; this is not yet supported\n");
		return;
	}
	dim3 dimBlock(w);
//	int s1 = h;
//	int s2 = 1;
//	while (s1>=65535) {
//		s2 *= 2;
//		s1  = iDivUp(s1,2);
//	}
//	dim3 dimGrid(s1,s2);
	dim3 dimGrid = make_2d_gridDim_for_1d_blockDim(1,h);
	
	if (rect_input) {
		if (type==1) {
			d_linFirstOrderDGL2EulerStepRect<<<dimGrid,dimBlock>>>(
									data,
									 w,
									 h,
									 a, b, B,
									 c, C,
									 dt,colMajor);
		} else if (type==2) {
			d_linFirstOrderDGL2ImplicitEulerStepRect<<<dimGrid,dimBlock>>>(
									data,
									 w,
									 h,
									 a, b, B,
									 c, C,
									 dt,colMajor);
		} else if (type==3) {
			d_linFirstOrderDGL2SteadyStateSolutionRect<<<dimGrid,dimBlock>>>(
									data,
									 w,
									 h,
									 a, b, B,
									 c, C,colMajor);
		}
	} else {
		if (type==1) {
			d_linFirstOrderDGL2EulerStep<<<dimGrid,dimBlock>>>(
									data,
									 w,
									 h,
									 a, b, B,
									 c, C,
									 dt,colMajor);
		} else if (type==2) {
			d_linFirstOrderDGL2ImplicitEulerStep<<<dimGrid,dimBlock>>>(
									data,
									 w,
									 h,
									 a, b, B,
									 c, C,
									 dt,colMajor);
		} else if (type==3) {
			d_linFirstOrderDGL2SteadyStateSolution<<<dimGrid,dimBlock>>>(
									data,
									 w,
									 h,
									 a, b, B,
									 c, C,colMajor);
		}
	}
}
