/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include <cutil.h>
#include <stdio.h>

#include "GPUTextureData2D.hpp"

static texture<float, 1, cudaReadModeElementType> tex;
static texture<float, 2, cudaReadModeElementType> tex2;
static texture<float, 2, cudaReadModeElementType> tex2_2;

//__global__ void texture1D(float *d_data, int size ){
//
//	int idx = blockIdx.x*blockDim.x + threadIdx.x;
//
//	const float u = (idx+0.5f) / (float) size;
//
//	d_data[idx] = tex1D(tex, u);
//	__syncthreads();
//}
//__global__ void texturing2D(float *d_data, int width, int height ){
//
//	int idx = blockIdx.x*blockDim.x + threadIdx.x;
//	int idy = blockIdx.y*blockDim.y + threadIdx.y;
//
//	const float u = (idx+0.5f) / (float) width;
//	const float v = (idy+0.5f) / (float) height;
//
//	d_data[idy*width + idx] = tex2D(tex2, u, v);
//	__syncthreads();
//}

__global__ void texture1D(float *d_data, const int size )
{
  const int idx = blockIdx.x*blockDim.x + threadIdx.x;
  
  if ( idx < size ) 
	d_data[idx] = tex1D(tex, idx);	
}

__global__ void texturing2D(float *d_data, const int width, const int height )
{
  const int idx = blockIdx.x*blockDim.x + threadIdx.x;
  const int idy = blockIdx.y*blockDim.y + threadIdx.y;
  
  if ( idx < width && idy < height )
	d_data[ idy * width + idx ] = tex2D( tex2, idx, idy );
}

struct InnerTexture1DData 
{

  InnerTexture1DData();
  InnerTexture1DData(int size);
  ~InnerTexture1DData();
  int increaseCounter();
  int decreaseCounter();
  void fill(const float val);
  void read(float* d_data);
  void bindTexture() const;
  void writeGPUData( const float* val, size_t size );

  cudaArray *cuArray;
  cudaChannelFormatDesc channelDesc;
  int size;
  int refCount;

};

struct InnerTexture3DData 
{
	InnerTexture3DData();
	InnerTexture3DData(int height, int width, int depth);
	~InnerTexture3DData();
	cudaArray *cuArray;
	cudaChannelFormatDesc channelDesc;
	int size;
	int height;
	int width;
	int depth;
	int refCount;
	int increaseCounter();
	int decreaseCounter();
	void fill(const float val);
	void read(float* d_data);
	void bindTexture() const;
	void writeGPUData(const float* val, const size_t* size);
} ;


InnerTexture1DData::InnerTexture1DData() {
  size=0;
  //std::cout << "construct empty TexInnerData with size " << size << std::endl;
  refCount=1;
}


InnerTexture1DData::InnerTexture1DData(int n) 
{  
  size = n;
  
  channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
  cudaMallocArray(&cuArray, &channelDesc, size, 1);
  
  refCount=1;
}


InnerTexture1DData::~InnerTexture1DData() {
  //remove memory from GPU
  //	std::cout << "destroy TexInnerData with size " << size << std::endl;
}

int InnerTexture1DData::increaseCounter() {
  //	std::cout << "increase TexInnerData coutner with size " << size;
  //	std::cout << " to " << refCount+1 << std::endl;
  return ++refCount;
}

int InnerTexture1DData::decreaseCounter() {
  //	std::cout << "decrease TexInnerData coutner with size " << size;
  //	std::cout << " to " << refCount-1 << std::endl;
  return --refCount;
}

void InnerTexture1DData::fill(const float val)
{
  float *data;
  data = new float[size];

  for( int i = 0; i < size; i++ )
	{
	  data[i] = val;
	}

  cudaMemcpyToArray( cuArray, 0, 0, data, size * sizeof(float), cudaMemcpyHostToDevice);
  delete[] data;
}

void InnerTexture1DData::writeGPUData(const float* val, size_t n )
{
  
  if( size != n )
	{
	  size = n;
	  channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
	  cudaMallocArray(&cuArray, &channelDesc, size, 1);
	}
  
  cudaMemcpyToArray(cuArray, 0, 0, val, size * sizeof(float), cudaMemcpyDeviceToDevice);
  
}

void InnerTexture1DData::read( float* d_data )
{
  tex.addressMode[0] = cudaAddressModeClamp;
  tex.addressMode[1] = cudaAddressModeClamp;
  tex.filterMode = cudaFilterModePoint;
  tex.normalized = false;
  cudaBindTextureToArray( tex, cuArray, channelDesc);
  
  dim3 dimBlock1d = GPUData_dimBlock1d;
  dim3 dimGrid1d  = ( size+dimBlock1d.x - 1 ) / dimBlock1d.x;
  texture1D<<< dimGrid1d, dimBlock1d >>>( d_data, size );
}

void InnerTexture1DData::bindTexture() const
{
  tex.addressMode[0] = cudaAddressModeClamp;
  tex.addressMode[1] = cudaAddressModeClamp;
  tex.filterMode = cudaFilterModePoint;
  tex.normalized = false;
  cudaBindTextureToArray( tex, cuArray, channelDesc);
}

// 2D

InnerTexture2DData::InnerTexture2DData() 
{
  size     = 0;
  cuArray  = 0;
  refCount = 1;
}

InnerTexture2DData::InnerTexture2DData( int h, int w ) 
{
  cuArray  = 0;
  refCount = 1;
  size = h * w;
  height = h;
  width = w;
  channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
  cudaMallocArray( &cuArray, &channelDesc, width, height );
}

InnerTexture2DData::~InnerTexture2DData() 
{
  if ( cuArray )
	cudaFreeArray( cuArray );
}

int InnerTexture2DData::increaseCounter()
{
  return ++refCount;
}

int InnerTexture2DData::decreaseCounter() 
{
  return --refCount;
}

void InnerTexture2DData::fill(const float val)
{
  float *data;
  data = new float [size];
  for(int y=0; y<height; y++)
	{
	  for(int x=0; x<width; x++)
		{
		  data[x+y*width] = val;
		}
	}
  cudaMemcpyToArray( cuArray, 0, 0, data, size * sizeof(float), cudaMemcpyHostToDevice);
  delete[] data;
}

void InnerTexture2DData::read(float* d_data)
{
  tex2.addressMode[0] = cudaAddressModeClamp;
  tex2.addressMode[1] = cudaAddressModeClamp;
  tex2.filterMode = cudaFilterModePoint;
  tex2.normalized = false;
  cudaBindTextureToArray( tex2, cuArray, channelDesc);
  
  dim3 dimBlock(16,16);
  dim3 dimGrid (((width+dimBlock.x-1)/dimBlock.x),((height+dimBlock.y-1)/dimBlock.y));
  texturing2D<<< dimGrid, dimBlock >>>( d_data, width, height );
  
}

void InnerTexture2DData::bindTexture( cudaTextureFilterMode filtermode ) const
{
  tex2.addressMode[0] = cudaAddressModeClamp;
  tex2.addressMode[1] = cudaAddressModeClamp;
  tex2.filterMode = filtermode;
  tex2.normalized = false;
  cudaBindTextureToArray( tex2, cuArray, channelDesc);
}

void InnerTexture2DData::bindTexture2() const
{
  tex2_2.addressMode[0] = cudaAddressModeClamp;
  tex2_2.addressMode[1] = cudaAddressModeClamp;
  tex2_2.filterMode = cudaFilterModePoint;
  tex2_2.normalized = false;
  cudaBindTextureToArray( tex2_2, cuArray, channelDesc);
}

void InnerTexture2DData::writeGPUData(const float* val, const size_t* dims )
{
  size_t n = dims[0] * dims[1];
  if( size != n )
	{
	  if ( cuArray )
		cudaFreeArray( cuArray );
	  cuArray = 0;
	  size = n;
	  width = dims[1];
	  height = dims[0];
	  channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
	  cudaMallocArray(&cuArray, &channelDesc, width, height );
	}
  cudaMemcpyToArray(cuArray, 0, 0, val, size * sizeof(float), cudaMemcpyDeviceToDevice);
}

// 3D
/*
  
  InnerTexture3DData::InnerTexture3DData() 
  {
  size=0;
  refCount=1;
  }
  
  InnerTexture3DData::InnerTexture3DData(int height, int width, int depth) 
  {
  size=height * width;
  height=height;
  width=width;
  depth=depth;
  
  // Daten in CudaArray
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0,
  cudaChannelFormatKindFloat);
  cudaArray* cuArray;
  
  // Tiefe korrigieren!
  cudaMallocArray(&cuArray, &channelDesc, height, width*depth);
  
  texCounter = tex3DCounter++;
  
  tex3DArray[texCounter].addressMode[0] = cudaAddressModeWrap;
  tex3DArray[texCounter].addressMode[1] = cudaAddressModeWrap;
  tex3DArray[texCounter].filterMode = cudaFilterModeLinear;
  tex3DArray[texCounter].normalized = true; // access with normalized texture coordinates
  
  // Bind the array to the texture
  cudaBindTextureToArray(tex3DArray[texCounter], cuArray, channelDesc);
  
  refCount=1;
  }
  InnerTexture3DData::~InnerTexture3DData() {
  }
  
  int InnerTexture3DData::increaseCounter() {
  return ++refCount;
  }
  
  int InnerTexture3DData::decreaseCounter() {
  return --refCount;
  }*/


__global__ void con2DsharedTexture( int inx, int iny, float *d_inkernel, int inkernelx, int inkernely, float *d_erg )
{
  
  extern __shared__ float s_kernel[];

  const size_t ny = ( inkernely + blockDim.y - 1 ) / blockDim.y;

  for(int y = 0; y < ny; y++)
	{
	  const ptrdiff_t offset = threadIdx.x + ( inkernelx * ( threadIdx.y + y * blockDim.y ) );
	  s_kernel[ offset ] = d_inkernel[ offset  ];
	}
  
  // Load one element per thread from device memory and store it
  const int idx = blockIdx.x * blockDim.x + threadIdx.x;
  const int idy = blockIdx.y * blockDim.y + threadIdx.y;
  
  if( idx >= inx || idy >= iny )
	return;
  
  const int L = inkernelx / 2, K = inkernely / 2;
  
  float sum = 0;
  
  for ( int i = -L; i <= L; i++ )
	{
	  int t;
	  t = idx+i;
	  if(t<0) 
		t = 0;
	  if( t >= inx ) 
		t = inx - 1;
	  if( t >= 0 && t < inx )
		{
		  for (int j=-K; j<=K; j++) 
			{
			  int u;
			  u = idy+j;
			  if(u<0) 
				u = 0; 
			  if( u >= iny ) 
				u = iny-1;
			  if( u >= 0 && u < iny )				
				sum +=  tex2D( tex2, t, u ) * s_kernel[ L + inkernelx * ( K - j ) - i ];
			}
		}
	}

  // Block until all threads in the block have written their data to shared mem
  __syncthreads();
  
  //d_erg[idx+(idy*inx)] = s_data[threadIdx.x+(blockDim.x*threadIdx.y)];
  d_erg[idx+(idy*inx)] = sum;
}

__device__ int clampi( const int x, const int x1, const int x2 )
{
  return x < x1 ? x1 : ( x > x2 ? x2 : x );
}

__global__ void con2DTexture(int width, int height, float *d_inkernel, int inkernelx, int inkernely, float *d_erg)
{

  const int x = blockIdx.x*blockDim.x + threadIdx.x;
  const int y = blockIdx.y*blockDim.y + threadIdx.y;
  
  if ( x >= width || y >= height ) 
	return;
  
  const int L = inkernelx / 2, K = inkernely / 2;
  
  float sum = 0.0f;
  
  
  for (int j = -K; j <= K; ++j ) 
	for ( int i = -L; i <= L; ++i )
	  sum += tex2D( tex2, x + i, y + j ) * d_inkernel[ L - i + inkernelx * ( K - j ) ];
  
  d_erg[ x + y * width ] = sum;

}

__global__ void con2DTexture2(int width, int height, int inkernelx, int inkernely, float *d_erg)
{

  const int x = blockIdx.x*blockDim.x + threadIdx.x;
  const int y = blockIdx.y*blockDim.y + threadIdx.y;
  
  if ( x >= width || y >= height ) 
	return;
  
  const int L = inkernelx / 2, K = inkernely / 2;
  
  float sum = 0.0f;
  
  for ( int j = -K; j <= K; ++j ) 
	for (int i = - L; i <= L; ++i )
	  sum += tex2D( tex2, x + i, y + j ) * tex2D( tex2_2, L - i, K - j );
  
  d_erg[ x + y * width ] = sum;
  
}

__global__ void con2DTexture3(int width, int height, int depth, int inkernelx, int inkernely, float *d_erg)
{

  const int x = blockIdx.x*blockDim.x + threadIdx.x;
  const int y = blockIdx.y*blockDim.y + threadIdx.y;
  
  if ( x >= width || y >= height ) 
	return;
  
  const int L = inkernelx / 2, K = inkernely / 2;
  
  float sum = 0.0f;
  
  for ( int k = 0; k < depth; ++k ) 
	{
	  const int image_offset  = width * k;
	  const int kernel_offset = inkernely * k;
	  for ( int j = -K; j <= K; ++j ) 
		for (int i = - L; i <= L; ++i )
		  sum += tex2D( tex2, x + i, image_offset + y + j ) * tex2D( tex2_2, L - i, kernel_offset + K - j );
	}
  d_erg[ x + y * width ] = sum;  
}

__global__ void d_scale_from_texture( float xscale, float yscale
									  , int in_width, int in_height
									  , float * target, int width, int height, int pitch0, int pitch1 )
{
  const int xout = blockIdx.x * blockDim.x + threadIdx.x;
  const int yout = blockIdx.y * blockDim.y + threadIdx.y;
  const int iout = yout * pitch0 + xout * pitch1;
  if ( xout >= width || yout >= height ) return;
  const float x0 = width  / 2.0f;
  const float y0 = height / 2.0f;
  const float ix0 = in_width  / 2.0f;
  const float iy0 = in_height / 2.0f;
  const float xin = ( xout - x0 ) / xscale + ix0;
  const float yin = ( yout - y0 ) / yscale + iy0;
  if ( xin < 0 || yin < 0 || xin >= in_width || yin >= in_height )
	target[iout] = 0;
  else
	target[iout] = tex2D( tex2,  xin, yin );
}

void GPUTextureData2D::scale( const GPUDataBase & target, float xscale, float yscale ) const
{
  _innerData->bindTexture( cudaFilterModePoint );

  int width = target.dims[1];
  int height = target.dims[0];
  //dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimBlock2d( 8, 4 );
  dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
  d_scale_from_texture<<<dimGrid2d, dimBlock2d>>>( xscale, yscale
												   , this->width() , this->height()
												   , target._data , width , height , target.strides[0], target.strides[1] );
}

struct transmat
{
  float a, b, c;
  float d, e, f;
  float g, h, i;
};

__global__ void d_affine_transform_from_texture( const transmat A
												 , int in_width, int in_height
												 , float * target, int width, int height, int pitch0, int pitch1 )
{
  const int xout = blockIdx.x * blockDim.x + threadIdx.x;
  const int yout = blockIdx.y * blockDim.y + threadIdx.y;
  const int iout = yout * pitch0 + xout * pitch1;
  if ( xout >= width || yout >= height ) return;
  const float x0 = width  / 2.0f;
  const float y0 = height / 2.0f;
  const float ix0 = in_width  / 2.0f;
  const float iy0 = in_height / 2.0f;
  const float win =   A.g * ( xout - x0 ) + A.h * ( yout - y0 ) + A.i;
  const float xin = ( A.a * ( xout - x0 ) + A.b * ( yout - y0 ) + A.c ) / win + ix0;
  const float yin = ( A.d * ( xout - x0 ) + A.e * ( yout - y0 ) + A.f ) / win + iy0;
  if ( xin < 0 || yin < 0 || xin >= in_width || yin >= in_height )
	target[iout] = 0;
  else
	target[iout] = tex2D( tex2,  xin, yin );
}


void GPUTextureData2D::affineTransform( const GPUDataBase & target, const float A[3][3] ) const
{
  _innerData->bindTexture( cudaFilterModePoint );

  int width = target.dims[1];
  int height = target.dims[0];
  //dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimBlock2d( 8, 4 );
  dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
  
  transmat B = { A[0][0], A[0][1], A[0][2],
				 A[1][0], A[1][1], A[1][2],
				 A[2][0], A[2][1], A[2][2] };
  d_affine_transform_from_texture<<<dimGrid2d, dimBlock2d>>>( B
															  , this->width() , this->height()
															  , target._data , width , height , target.strides[0], target.strides[1] );

}

__global__ void d_lens_distort_from_texture( const lens_distortion_params P
											 , int in_width, int in_height
											 , float * target, int width, int height, int pitch0, int pitch1 )
{
  const int xout = blockIdx.x * blockDim.x + threadIdx.x;
  const int yout = blockIdx.y * blockDim.y + threadIdx.y;
  const int iout = yout * pitch0 + xout * pitch1;
  if ( xout >= width || yout >= height ) return;
  const float x_n = ( xout - width  * P.x_c ) / ( width  * P.f_x );
  const float y_n = ( yout - height * P.y_c ) / ( height * P.f_y );
  const float x_d = x_n - P.alpha_c * y_n;
  const float y_d = y_n;
  float x_in = x_d;
  float y_in = y_d;
  for ( int i = 0; i < 20; ++i )
	{
	  const float r2 = x_in * x_in + y_in * y_in;
	  const float k_radial = 1 + P.K1 * r2 + P.K2 * r2 * r2 + P.K3 * r2 * r2 * r2;
	  const float delta_x = 2 * P.P1 * x_in * y_in + P.P2 * ( r2 + 2 * x_in * x_in );
	  const float delta_y = 2 * P.P2 * x_in * y_in + P.P1 * ( r2 + 2 * y_in * y_in );
	  x_in = ( x_d - delta_x ) / k_radial;
	  y_in = ( y_d - delta_y ) / k_radial;
	}
  x_in = in_width  * ( x_in * P.f_x + P.x_c );
  y_in = in_height * ( y_in * P.f_y + P.y_c );
  if ( x_in < 0 || y_in < 0 || x_in >= in_width || y_in >= in_height )
	target[iout] = 0;
  else
	target[iout] = tex2D( tex2,  x_in, y_in );
}

void GPUTextureData2D::lensDistort( const GPUDataBase & target, const lens_distortion_params P ) const
{
  _innerData->bindTexture( cudaFilterModePoint );

  int width = target.dims[1];
  int height = target.dims[0];
  //dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimBlock2d( 8, 4 );
  dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
  d_lens_distort_from_texture<<<dimGrid2d, dimBlock2d>>>
	( P
	  , this->width() , this->height()
	  , target._data , width , height , target.strides[0], target.strides[1] );  
}

extern "C" void conv2DFromTexture(InnerTexture2DData *d_in, float *d_kernel, int kernel_x, int kernel_y, float *d_erg) {
  
  const int in_x = (int) d_in->width;
  const int in_y = (int) d_in->height;
  
  const int blockDimensions_x = kernel_x;
  const int blockDimensions_y = min(512/kernel_x,kernel_y);
  const int gridDimension_x = (in_x+blockDimensions_x-1)/blockDimensions_x;
  const int gridDimension_y = (in_y+blockDimensions_y-1)/blockDimensions_y;
  int memSize = blockDimensions_x * blockDimensions_y * sizeof(float);
  
  
  d_in->bindTexture();
  
  dim3 dimBlock(blockDimensions_x,blockDimensions_y); // threadIdx
  dim3 dimGrid(gridDimension_x,gridDimension_y);		// blockIdx
  
  if (blockDimensions_x*blockDimensions_y>512) {
	con2DTexture<<< dimGrid, dimBlock >>>(in_x, in_y, d_kernel, kernel_x, kernel_y, d_erg);
  } else {
	con2DsharedTexture<<< dimGrid, dimBlock, memSize >>>(in_x, in_y, d_kernel, kernel_x, kernel_y, d_erg);
  }
  
  
  //	unsigned int timer2 = 0;
  //	CUT_SAFE_CALL( cutCreateTimer( &timer2 ) );
  //	CUT_SAFE_CALL( cutStartTimer( timer2 ) );
  //
  //	con2D<<< dimGrid, dimBlock >>>(d_in, in_x, in_y, d_kernel, kernel_x, kernel_y, d_erg, neumann);
  //
  //	cudaThreadSynchronize();
  //	CUT_SAFE_CALL( cutStopTimer( timer2 ) );
  //	printf( "Processing time: %f (ms)\n", cutGetTimerValue( timer2 ) );
  //	CUT_SAFE_CALL( cutDeleteTimer( timer2 ) );
  
  
}

extern "C" void conv2DFromTexture2(InnerTexture2DData *d_in, InnerTexture2DData *d_kernel, float *d_erg) 
{
  
  const int kernel_x = (int) d_kernel->width;
  const int kernel_y = (int) d_kernel->height;
  
  const int in_x = (int) d_in->width;
  const int in_y = (int) d_in->height;
  
  const int blockDimensions_x = kernel_x;
  const int blockDimensions_y = min(512/kernel_x,kernel_y);
  const int gridDimension_x = (in_x+blockDimensions_x-1)/blockDimensions_x;
  const int gridDimension_y = (in_y+blockDimensions_y-1)/blockDimensions_y;
  
  d_in->bindTexture();
  d_kernel->bindTexture2();
  
  dim3 dimBlock(blockDimensions_x,blockDimensions_y); // threadIdx
  dim3 dimGrid(gridDimension_x,gridDimension_y);		// blockIdx
  
  con2DTexture2<<< dimGrid, dimBlock >>>(in_x, in_y, kernel_x, kernel_y, d_erg);
  
}

extern "C" void conv2DFromTexture3(InnerTexture2DData *d_in, InnerTexture2DData *d_kernel, int depth, float *d_erg) 
{
  
  const int kernel_x = (int) d_kernel->width / depth;
  const int kernel_y = (int) d_kernel->height;
  
  const int in_x = (int) d_in->width / depth;
  const int in_y = (int) d_in->height;
  
  const int blockDimensions_x = kernel_x;
  const int blockDimensions_y = min(512/kernel_x,kernel_y);
  const int gridDimension_x = (in_x+blockDimensions_x-1)/blockDimensions_x;
  const int gridDimension_y = (in_y+blockDimensions_y-1)/blockDimensions_y;
  
  d_in->bindTexture();
  d_kernel->bindTexture2();
  
  dim3 dimBlock(blockDimensions_x,blockDimensions_y); // threadIdx
  dim3 dimGrid(gridDimension_x,gridDimension_y);		// blockIdx
  
  con2DTexture3<<< dimGrid, dimBlock >>>(in_x, in_y, depth, kernel_x, kernel_y, d_erg);
}
