/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cublas.h>
#include <cutil.h>

#include <cstdio>

#include "cu_utils.h"

//24-bit multiplication is faster on G80,
//but we must be sure to multiply integers
//only within [-8M, 8M - 1] range
#define IMUL(a, b) __mul24(a, b)

__device__ __constant__ float dd_Kernel[10*1024];

__global__ void convolutionGeneral1Din4DPaddZeroes(
    float *d_Data,
    int KernelIdx,
	const int rk,
	const int rrk,
	float *d_Result,
     const int incX,
    const int sizeX,
    const int incY,
    const int sizeY,
    const int incU,
    const int sizeU,
    const int incV,
    const int sizeV
){
    //Data cache
    extern __shared__ float data[];

    //Current tile and apron limits, in dim coordinates
    const int         xStart = blockIdx.x * (blockDim.x-rk-rk);
    const int         xEnd   = xStart + blockDim.x - rk - rk - 1;
    const int         xApronStart = xStart - rk;
    const int         xApronEnd   = xEnd   + rk;

    //Clamp tile and apron limits by image borders
    const int    xEndClamped = min(xEnd, sizeX - 1);
    const int xApronStartClamped = max(xApronStart, 0);
    const int   xApronEndClamped = min(xApronEnd, sizeX - 1);

    //Current index
    const int     x = xApronStart + threadIdx.x;

    const int 	y = blockIdx.y;
    const int 	gYOff = y*incY;

    int v = 0;
    int u = 0;

    for (v=0; v<sizeV; v++) {
    for (u=0; u<sizeU; u++) {

    	const int gIndex = (x*incX) + gYOff + (u*incU) + (v*incV);

	    data[threadIdx.x] =
	    			        ( (x>=xApronStartClamped) && (x<=xApronEndClamped)) ?
	    			        d_Data[ gIndex ] : 0;


//	    	( (x>=xApronStartClamped) && (x<=xApronEndClamped)) ?
//	    	gIndex : 0;



	    __syncthreads();

	    if (x>=xStart && x<=xEndClamped) {
	    	float sum = 0;
//	    	sum = data[threadIdx.x]*dd_Kernel[KernelIdx+0];
	    	for(int k = -rrk; k <= rrk; k++) sum += data[threadIdx.x + k] * dd_Kernel[KernelIdx+ rrk - k];

	    	d_Result[ gIndex ] = sum;
//	    	d_Result[ gIndex ] = dd_Kernel[KernelIdx+threadIdx.x-5];
//	    	d_Result[ gIndex ] = data[threadIdx.x-6];
//	    	d_Result[ gIndex ] = threadIdx.x-5;
//	    	d_Result[ gIndex ] = KernelIdx+threadIdx.x-5;
//	    	if ( threadIdx.x%2==0)
//	    		d_Result[ gIndex ] = threadIdx.x;
//	    	else
//	    		d_Result[ gIndex ] = 2*rk;
	    }
	    __syncthreads();
    }
    }
}

__global__ void convolutionGeneral1Din4DCyclic(
    float *d_Data,
    const int KernelIdx,
    const int rk,
    const int rrk,
    float *d_Result,
    const int incX,
    const int sizeX,
    const int incY,
    const int sizeY,
    const int incU,
    const int sizeU,
    const int incV,
    const int sizeV
) {
	    //Data cache
	    extern __shared__ float data[];

	    //Current tile and apron limits, in dim coordinates
	    const int         xStart = (blockIdx.x* (blockDim.x-rk-rk) );
	    const int         xEnd   = xStart + blockDim.x - rk - rk - 1;
	    const int         xApronStart = xStart - rk;
	    const int         xApronEnd   = xEnd   + rk;

	    //Clamp tile and apron limits by image borders
	    const int    xEndClamped = min(xEnd, sizeX - 1);
	    const int xApronStartClamped = max(xApronStart, 0);
	    const int   xApronEndClamped = min(xApronEnd, sizeX - 1);

	    //Current index
	    const int     x = xApronStart + threadIdx.x;

	    const int 	y = blockIdx.y;
	    const int 	gYOff = (y*incY);

	    int v = 0;
	    int u = 0;

	    for (v=0; v<sizeV; v++) {
	    for (u=0; u<sizeU; u++) {

	    	int gIndex = gYOff + (u*incU) + (v*incV);

	    	if (x>=0) {
	    		if (x<sizeX) {
	    			gIndex += (x*incX);
	    		} else {
	    			gIndex += ((x-sizeX)*incX);
	    		}
	    	} else {
	    		gIndex += ((x+sizeX)*incX);
	    	}

			data[threadIdx.x] = d_Data[ gIndex ];

		    __syncthreads();

		    if (x>=xStart && x<=xEndClamped) {
		    	float sum = 0;

		    	for(int k = -rrk; k <= rrk; k++) sum += data[threadIdx.x + k] * dd_Kernel[KernelIdx+rrk - k];

		    	d_Result[ gIndex ] = sum;
		    }
		    __syncthreads();
	    }
	    }
}

__global__ void convolutionGeneral1Din4DReplicate(
    float *d_Data,
    const int KernelIdx,
    const int rk,
    const int rrk,
    float *d_Result,
    const int incX,
    const int sizeX,
    const int incY,
    const int sizeY,
    const int incU,
    const int sizeU,
    const int incV,
    const int sizeV
) {
	    //Data cache
	    extern __shared__ float data[];

	    //Current tile and apron limits, in dim coordinates
	    const int         xStart = (blockIdx.x* (blockDim.x-rk-rk) );
	    const int         xEnd   = xStart + blockDim.x - rk - rk - 1;
	    const int         xApronStart = xStart - rk;
	    const int         xApronEnd   = xEnd   + rk;

	    //Clamp tile and apron limits by image borders
	    const int    xEndClamped = min(xEnd, sizeX - 1);
	    const int xApronStartClamped = max(xApronStart, 0);
	    const int   xApronEndClamped = min(xApronEnd, sizeX - 1);

	    //Current index
	    const int     x = xApronStart + threadIdx.x;

	    const int 	y = blockIdx.y;
	    const int 	gYOff = (y*incY);

	    int v = 0;
	    int u = 0;

	    for (v=0; v<sizeV; v++) {
	    for (u=0; u<sizeU; u++) {

	    	int gIndex = gYOff + (u*incU) + (v*incV);

	    	if (x>=0) {
	    		if (x<sizeX) {
	    			gIndex += x*incX;
	    		} else {
	    			gIndex += (sizeX-1)*incX;
	    		}
	    	} else {
//	    		gIndex += 0;
	    	}

			data[threadIdx.x] = d_Data[ gIndex ];

		    __syncthreads();

		    if (x>=xStart && x<=xEndClamped) {
		    	float sum = 0;

		    	for(int k = -rrk; k <= rrk; k++) sum += data[threadIdx.x + k] * dd_Kernel[KernelIdx+rrk - k];

		    	d_Result[ gIndex ] = sum;
		    }
		    __syncthreads();
	    }
	    }
}

__global__ void convolutionRowGPU(
    float *d_Data,
	int    kernelNo,
	int    rk,
	float *d_Result,
    int dataW,
    int dataH
){
    //Data cache
    extern __shared__ float data[];

    //Current tile and apron limits, relative to row start
    const int         tileStart = (blockIdx.x * 128);
    const int           tileEnd = tileStart + 128 - 1;
    const int        apronStart = tileStart - rk;
    const int          apronEnd = tileEnd   + rk;

    //Clamp tile and apron limits by image borders
    const int    tileEndClamped = min(tileEnd, dataW - 1);
    const int apronStartClamped = max(apronStart, 0);
    const int   apronEndClamped = min(apronEnd, dataW - 1);

    //Row start index in d_Data[]
    const int          rowStart = (blockIdx.y * dataW);

    //Aligned apron start. Assuming dataW and 128 are multiples
    //of half-warp size, rowStart + apronStartAligned is also a
    //multiple of half-warp size, thus having proper alignment
    //for coalesced d_Data[] read.
    const int apronStartAligned = tileStart - 16;

    const int loadPos = apronStartAligned + threadIdx.x;
    //Set the entire data cache contents
    //Load global memory values, if indices are within the image borders,
    //or initialize with zeroes otherwise
    if(loadPos >= apronStart){
        const int smemPos = loadPos - apronStart;


        data[smemPos] =
            ((loadPos >= apronStartClamped) && (loadPos <= apronEndClamped)) ?
            d_Data[rowStart + loadPos] : 0;

		//data[smemPos] = threadIdx.x;

    }

    //Ensure the completness of the loading stage
    //because results, emitted by each thread depend on the data,
    //loaded by another threads
    __syncthreads();
    const int writePos = tileStart + threadIdx.x;
    //Assuming dataW and 128 are multiples of half-warp size,
    //rowStart + tileStart is also a multiple of half-warp size,
    //thus having proper alignment for coalesced d_Result[] write.
    if(writePos <= tileEndClamped){
        const int smemPos = writePos - apronStart;
        float sum = 0;

        //sum = convolutionRowKernel5<2 * kr,kr>(data + smemPos,kernelNo);
		/*for(int k = -2; k <= 2; k++)
            sum += data[smemPos + k] * d_Kernel5[2 - k];*/

		for(int k = -rk; k <= rk; k++) sum += data[smemPos + k] * dd_Kernel[rk - k+kernelNo];

        d_Result[rowStart + writePos] = sum;
    }
}

__global__ void convolutionRowGPU_BIG(
    float *d_Data,
	int    kernelNo,
	int    rk,
	float *d_Result,
    int dataW,
    int dataH
){
    //Data cache
    extern __shared__ float data[];

    //Current tile and apron limits, relative to row start
    const int         tileStart = (blockIdx.x * 128);
    const int           tileEnd = tileStart + 128 - 1;
    const int        apronStart = tileStart - rk;
    const int          apronEnd = tileEnd   + rk;

    //Clamp tile and apron limits by image borders
    const int    tileEndClamped = min(tileEnd, dataW - 1);
    const int apronStartClamped = max(apronStart, 0);
    const int   apronEndClamped = min(apronEnd, dataW - 1);

    //Row start index in d_Data[]
    const int          rowStart = (blockIdx.y * dataW);

    //Aligned apron start. Assuming dataW and 128 are multiples
    //of half-warp size, rowStart + apronStartAligned is also a
    //multiple of half-warp size, thus having proper alignment
    //for coalesced d_Data[] read.
    const int apronStartAligned = tileStart - 32;

    const int loadPos = apronStartAligned + threadIdx.x;
    //Set the entire data cache contents
    //Load global memory values, if indices are within the image borders,
    //or initialize with zeroes otherwise
    if(loadPos >= apronStart){
        const int smemPos = loadPos - apronStart;


        data[smemPos] =
            ((loadPos >= apronStartClamped) && (loadPos <= apronEndClamped)) ?
            d_Data[rowStart + loadPos] : 0;

		//data[smemPos] = threadIdx.x;

    }

    //Ensure the completness of the loading stage
    //because results, emitted by each thread depend on the data,
    //loaded by another threads
    __syncthreads();
    const int writePos = tileStart + threadIdx.x;
    //Assuming dataW and 128 are multiples of half-warp size,
    //rowStart + tileStart is also a multiple of half-warp size,
    //thus having proper alignment for coalesced d_Result[] write.
    if(writePos <= tileEndClamped){
        const int smemPos = writePos - apronStart;
        float sum = 0;

        //sum = convolutionRowKernel5<2 * kr,kr>(data + smemPos,kernelNo);
		/*for(int k = -2; k <= 2; k++)
            sum += data[smemPos + k] * d_Kernel5[2 - k];*/

		for(int k = -rk; k <= rk; k++) sum += data[smemPos + k] * dd_Kernel[rk - k+kernelNo];

        d_Result[rowStart + writePos] = sum;
    }
}



////////////////////////////////////////////////////////////////////////////////
// Column convolution filter
////////////////////////////////////////////////////////////////////////////////
__global__ void convolutionColumnGPU(
    float *d_Data,
	int kernelNo,
	int rk,
	float *d_Result,
    int dataW,
    int dataH,
    int smemStride,
    int gmemStride
){
    //Data cache
    extern __shared__ float data[];

    //Current tile and apron limits, in rows
    const int         tileStart = (blockIdx.y * 48);
    const int           tileEnd = tileStart + 48 - 1;
    const int        apronStart = tileStart - rk;
    const int          apronEnd = tileEnd   + rk;

    //Clamp tile and apron limits by image borders
    const int    tileEndClamped = min(tileEnd, dataH - 1);
    const int apronStartClamped = max(apronStart, 0);
    const int   apronEndClamped = min(apronEnd, dataH - 1);

    //Current column index
    const int       columnStart = (blockIdx.x * 16) + threadIdx.x;

    //Shared and global memory indices for current column
    int smemPos = (threadIdx.y* 16) + threadIdx.x;
    int gmemPos = ((apronStart + threadIdx.y) * dataW) + columnStart;
    //Cycle through the entire data cache
    //Load global memory values, if indices are within the image borders,
    //or initialize with zero otherwise
    if (columnStart<dataW)
    for(int y = apronStart + threadIdx.y; y <= apronEnd; y += blockDim.y){

        data[smemPos] =
	        ((y >= apronStartClamped) && (y <= apronEndClamped)) ?
	        d_Data[gmemPos] : 0;
//        data[smemPos] =
//	        ((y >= apronStartClamped) && (y <= apronEndClamped)) ?
//	        		1+y/100 : 0;

        smemPos += smemStride;
        gmemPos += gmemStride;
    }

    //Ensure the completness of the loading stage
    //because results, emitted by each thread depend on the data,
    //loaded by another threads
    __syncthreads();
    //Shared and global memory indices for current column
    smemPos = ((threadIdx.y + rk) * 16) + threadIdx.x;
    gmemPos = ((tileStart + threadIdx.y) * dataW) + columnStart;
    //Cycle through the tile body, clamped by image borders
    //Calculate and output the results
    if (columnStart<dataW)
    for(int y = tileStart + threadIdx.y; y <= tileEndClamped; y += blockDim.y){
	        float sum = 0;


	        //sum = convolutionColumnKernel5<2 * kr,kr>(data + smemPos,kernelNo);
			for(int k = -rk; k <= rk; k++) sum += data[smemPos + IMUL(k, 16)] * dd_Kernel[rk - k+kernelNo];

			d_Result[gmemPos] = sum;//(d_Result[gmemPos]*2+sum*1)/3;
			//d_Result[gmemPos] = data[smemPos + IMUL(-rk, 16)];
			//d_Result[gmemPos] = data[smemPos];
	        smemPos += smemStride;
	        gmemPos += gmemStride;

    }
}


extern "C" {

  /**
   * This function convolutes over the dimension x
   * type == 0 : (padded with 0)
   * type == 1 : (cyclic)
   * type == 2 : (replicate)
   */
  void convolute1DInContiguous4D( float *d_Data, int kernelIdx, int rk, float *d_Result,
								  int incX, int sizeX,
								  int incY, int sizeY,
								  int incU, int sizeU,
								  int incV, int sizeV,
								  int type)
  {
	// if type ist zeor, the border is filled with zeroes.


	int ark = rk;
	//if (ark<16) ark = 16;

	dim3 block( min( 64, sizeX + 2 * ark ) ); // to prevent the kernel to be more than one times larger than input
													// if this is not applied, in case of e.g. cyclic convolution, wrong
													// memory is accessed
	dim3 grid( iDivUp( sizeX, ( block.x - 2 * ark ) ), sizeY );

	if (type==0)
	  convolutionGeneral1Din4DPaddZeroes<<<grid,block,sizeof(float)*block.x>>>
		( d_Data, kernelIdx, ark,rk, d_Result,
		  incX, sizeX, incY,
		  sizeY, incU, sizeU, incV, sizeV
		  );
	else if (type==1)
	  convolutionGeneral1Din4DCyclic<<<grid,block,sizeof(float)*block.x>>>(d_Data, kernelIdx, ark,rk, d_Result, incX, sizeX, incY, sizeY, incU, sizeU, incV, sizeV);
	else
	  convolutionGeneral1Din4DReplicate<<<grid,block,sizeof(float)*block.x>>>(d_Data, kernelIdx, ark,rk, d_Result, incX, sizeX, incY, sizeY, incU, sizeU, incV, sizeV);


	/*
	cudaError_t err = cudaGetLastError();
	if ( err != cudaSuccess )
	  fprintf( stderr, "error in convolute1DInContiguous4D %i %s args: %x %i %i %x %i %i %i %i %i %i %i %i %i block: %i %i %i grid: %i %i %i \n"
			   , err, cudaGetErrorString( err )
			   , d_Data, kernelIdx, rk, d_Result, incX, sizeX, incY, sizeY, incU, sizeU, incV, sizeV, type
			   , block.x, block.y, block.z, grid.x, grid.y, grid.z );

	*/
  }
  
	/*
  	 * The following to function could onyl use border values filled with 0 
  	 */
  void doConvoluteDeepestOfTwoDims(float* data, float* dest, int N, int dimSize, int kernel, int kRadius) {

	dim3 blockGridRows(iDivUp(dimSize, 128), N/dimSize);

    if (kRadius>16) {
	  dim3 threadBlockRows(32 + 128 + kRadius);
	  if (kRadius>32)
		fprintf(stderr,"Kernel radius to large!! %d", kRadius);

	  convolutionRowGPU_BIG<<<blockGridRows, threadBlockRows,sizeof(float)*(kRadius + 128 + kRadius)>>>
		(
		 data,
		 kernel,
		 kRadius,
		 dest,
		 dimSize,
		 blockGridRows.y
		 );
    } else {
	  dim3 threadBlockRows(16 + 128 + kRadius);
	  convolutionRowGPU<<<blockGridRows, threadBlockRows,sizeof(float)*(kRadius + 128 + kRadius)>>>
		(
		 data,
		 kernel,
		 kRadius,
		 dest,
		 dimSize,
		 blockGridRows.y
		 );
    }

	/*
	cudaError_t err = cudaGetLastError();
	if ( err != cudaSuccess )
	  fprintf( stderr, "error in doConvoluteDeepestOfTwoDims %i %s args: %x %x %i %i %i %i"
			   , err, cudaGetErrorString( err )
			   , data, dest, N, dimSize, kernel, kRadius );
	*/
  }

  void doConvoluteFirstOfTwoDims(float* data, float* dest, int N, int dimSize, int kernel, int kRadius)
  {

	const int DATA_W = N/dimSize;
	const int DATA_H = dimSize;

	dim3 blockGridColumns(iDivUp(DATA_W, 16), iDivUp(DATA_H, 48));
	dim3 threadBlockColumns(16, 8);

	//	printf("DATA_W %d ",DATA_W);
	//	printf("DATA_H %d \n",DATA_H);
	//
	//	printf("Grid.x %d ",blockGridColumns.x);
	//	printf("Grid.y %d \n",blockGridColumns.y);
	//
	//	printf("Block.x %d ",threadBlockColumns.x);
	//	printf("Block.y %d \n", threadBlockColumns.y);
	//
	//	printf("Kernelradius %d \n", kRadius);
	//	printf("Kernel storage offs %d \n", kernel);
	convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, sizeof(float)*16 * (kRadius + 48 + kRadius)>>>
	  (
	   data,
	   kernel,
	   kRadius,
	   dest,
	   DATA_W,
	   DATA_H,
	   16 * threadBlockColumns.y,
	   DATA_W * threadBlockColumns.y
	   );

	/*
	cudaError_t err = cudaGetLastError();
	if ( err != cudaSuccess )
	  fprintf( stderr, "error in doConvoluteFirstOfTwoDims %i %s args: %x %x %i %i %i %i"
			   , err, cudaGetErrorString( err )
			   , data, dest, N, dimSize, kernel, kRadius );
	*/

  }

}
