/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"

__global__ void dCopyFromSpecial4D( const int tgt_stride_x , const int nx,
                                    const int tgt_stride_y , const int ny,
                                    const int tgt_stride_u , const int nu,
                                    const int tgt_stride_v , const int nv,
									float* tgt_data,
                                    const int src_stride_x,
                                    const int src_stride_y,
                                    const int src_stride_u,
                                    const int src_stride_v,
									float* src_data )
{
        // Block index
        const int bIndexX = blockIdx.x;
        const int bIndexY = blockIdx.y;

        // no of threads
        const int noOfThreadsX = blockDim.x;
        const int noOfThreadsY = blockDim.y;

        // Thread index
        const int tIndexX = threadIdx.x;
        const int tIndexY = threadIdx.y;

        // index Offset
        const int x   = ( bIndexX * noOfThreadsX) + tIndexX; // origin
        const int y   = ( bIndexY * noOfThreadsY) + tIndexY;  // origin
        const int itgt = ( y * tgt_stride_y ) + ( x * tgt_stride_x ); // destination
        const int isrc = ( y * src_stride_y ) + ( x * src_stride_x ); // origin


        if (x < nx && y < ny )
		  {

			for ( int u = 0, isrc_u = isrc, itgt_u = itgt;
				  u < nu;
				  ++u, isrc_u += src_stride_u, itgt_u += tgt_stride_u )
			  {
				for ( int v = 0, isrc_v = isrc_u, itgt_v = itgt_u;
					  v < nv;
					  ++v, isrc_v += src_stride_v, itgt_v += tgt_stride_v )
				  {
					tgt_data[itgt_v] = src_data[isrc_v];
				  }
			  }
		  }
}


extern "C" void doCopyFromSpecial4D( const int tgt_stride_x , const int nx,
									 const int tgt_stride_y , const int ny,
									 const int tgt_stride_u , const int nu,
									 const int tgt_stride_v , const int nv,
									 float* tgt_data,
									 const int src_stride_x,
									 const int src_stride_y,
									 const int src_stride_u,
									 const int src_stride_v,
									 float* src_data )
{
	/* the special treatment of ny == 1 seems to be necessary because of
	 * a bug in CUDA????
	 */
	int bx = ny==1 ? 512 : 64;  // max block size assumed to be 512
	int by = ny==1 ? 1   : 8;
	while (iDivUp(nx,bx)>65535) // max grid dim size assumed to be 65535
	{
		bx*=2;
		by/=2;
	}
	dim3 block(bx,by);
	dim3 grid( iDivUp( nx, block.x ), iDivUp( ny, block.y ) );
	dCopyFromSpecial4D<<<grid,block>>>( tgt_stride_x, nx,
										tgt_stride_y, ny,
										tgt_stride_u, nu,
										tgt_stride_v, nv, tgt_data,
										src_stride_x,
										src_stride_y,
										src_stride_u,
										src_stride_v, src_data );
}
