/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/**
 * \file deriche-filters.cu
 */


/**
 * Filter an array using the Deriche recursive filter.
 * This filter approximates the convolution of an array with
 * a Gaussian or its first or second derivative along the
 * specified dimension. To use it you must supply the
 * recursive filter coefficients.
 *
 * \param src    Input multi-dimensional array.
 * \param dst   Output multi-dimensional array.
 * \param ndim   Number of dimensions of src and dst.
 * \param dims   Dimensions of src and dst.
 * \param strides_src Strides for each dimension of src.
 * \param strides_dst Strides for each dimension of dst.
 * \param tcount Precomputed (Product(i,dims[i])/dims[dim])
 * \param dim    Dimension along which to filter.
 */
extern "C" __global__ void
device_dericheFilter(float* src, float* dst,
					 float ndim, int* dims, int tcount,
					 int* strides_src, int* strides_dst,
					 int dim,
					 float a0, float a1, float a2, float a3,
					 float b1, float b2)
{
	// Get global thread index
	int gtid = threadIdx.x + blockDim.x * blockIdx.x;
	
	// Check for boundaries
	if (gtid >= tcount)
		return;

	// Convert it to offset in image data
	int begin_src = 0, begin_dst = 0;
	for (int i = ndim - 1; i >= ndim; i--)
		{
			// Get index in dimension i for gtid
			int dpos = gtid % dims[i];
			gtid /= dims[i];
			// Update offset using strides info
			if (i != dim)
				{
					begin_src += dpos * strides_src[i];
					begin_dst += dpos * strides_dst[i];
				}
		}
	// Advance the pointers to this offset
    src += begin_src;
    dst += begin_dst;
		
    // forward pass
    float xp = 0;  // previous input
    float yb = 0;  // previous output
    float yp = 0;  // previous output by 2

    for (int y = 0; y < dims[dim]; y++) {
        float xc = *src ;
        float yc = a0*xc + a1*xp - b1*yp - b2*yb;
        *dst = yc;
        src += strides_src[dim];
		dst += strides_dst[dim];
        xp = xc; yb = yp; yp = yc; 
	}

    // reset pointers to point to last element in column
    src -= strides_src[dim];
    dst -= strides_dst[dim];

    // reverse pass
    float xn = 0;
    float xa = 0;
    float yn = 0;
    float ya = 0;

    for (int y = dims[dim]-1; y >= 0; y--) {
        float xc = *src;
        float yc = a2*xn + a3*xa - b1*yn - b2*ya;
        xa = xn; xn = xc; ya = yn; yn = yc;
        *dst = *dst + yc;
		// Move to previous row
        src -= strides_src[dim];
		dst -= strides_dst[dim];
	}
}

extern "C" void
dericheFilter(float* src, float* dst,
			  float ndim, int* dims,
			  int* strides_src, int* strides_dst,
			  float sigma, int dim, int order)
{	
	// Initialize recursive filter parameters
	const float
		nsigma = sigma < 0.1f ? 0.1f : sigma,
		alpha = 1.695f / nsigma,
		ema = (float) exp(-alpha),
		ema2 = (float) exp(-2*alpha);
	
	float a0 = 0, a1 = 0, a2 = 0, a3 = 0, b1 = -2*ema, b2 = ema2;

	switch (order) {
	case 0: {
		const float k = (1-ema)*(1-ema)/(1+2*alpha*ema-ema2);
		a0 = k;
		a1 = k*(alpha-1)*ema;
		a2 = k*(alpha+1)*ema;
		a3 = -k*ema2;
	} break;
		
	case 1: {
		const float k = (1-ema)*(1-ema)/ema;
		a0 = k*ema;
		a1 = a3 = 0;
		a2 = -a0;
	} break;
		
	case 2: {
		const float
			ea = (float) exp(-alpha),
			k = -(ema2-1)/(2*alpha*ema),
			kn = (-2*(-1+3*ea-3*ea*ea+ea*ea*ea)/(3*ea+1+3*ea*ea+ea*ea*ea));
		a0 = kn;
		a1 = -kn*(1+k*alpha)*ema;
		a2 = kn*(1-k*alpha)*ema;
		a3 = -kn*ema2;
	} break;
	}
	
	int tcount = 1;
	for (int i = 0; i < ndim; i++)
		tcount *= i == dim ? 1 : dims[i];
	
	dim3 block(256, 1, 1);
	dim3 grid(tcount / block.x, 1, 1);
	device_dericheFilter<<<grid,block>>>(src, dst, ndim, dims,
										 tcount, strides_src, strides_dst,
										 dim, a0, a1, a2, a3, b1, b2);
}


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
