/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "cutil.h"
#include "Exceptions.hpp"
#include <stdio.h>

/*
   This version adds multiple elements per thread sequentially.  This reduces the overall
   cost of the algorithm while keeping the work complexity O(n) and the step complexity O(log n).
   (Brent's Theorem optimization)
*/

#define define_reduction_kernel(name,op)								\
  template < unsigned int blockSize >									\
  __global__ void reduce_ ## name ## _kernel(float *g_odata, float *g_idata, unsigned int n) \
  {																		\
	__shared__ float sdata[blockSize];									\
																		\
	const unsigned int tid = threadIdx.x;								\
	unsigned int i = blockIdx.x*blockSize + threadIdx.x;				\
	const unsigned int gridSize = blockSize*gridDim.x;					\
																		\
	float thMax = op(g_idata[i], g_idata[i + gridSize]);				\
	i += 2 * gridSize;													\
	while (i < n)														\
	  {																	\
		thMax = op(thMax, g_idata[i]);									\
		i += gridSize;													\
	  }																	\
	sdata[tid] = thMax;													\
	__syncthreads();													\
																		\
	if (blockSize >= 512) { if (tid < 256) { sdata[tid] = op(sdata[tid], sdata[tid + 256]); } __syncthreads(); } \
	if (blockSize >= 256) { if (tid < 128) { sdata[tid] = op(sdata[tid], sdata[tid + 128]); } __syncthreads(); } \
	if (blockSize >= 128) { if (tid <  64) { sdata[tid] = op(sdata[tid], sdata[tid +  64]); } __syncthreads(); } \
																		\
	if (tid < 32)														\
	  {																	\
		if (blockSize >=  64) { sdata[tid] = op(sdata[tid], sdata[tid + 32]); } \
		if (blockSize >=  32) { sdata[tid] = op(sdata[tid], sdata[tid + 16]); } \
		if (blockSize >=  16) { sdata[tid] = op(sdata[tid], sdata[tid +  8]); } \
		if (blockSize >=   8) { sdata[tid] = op(sdata[tid], sdata[tid +  4]); } \
		if (blockSize >=   4) { sdata[tid] = op(sdata[tid], sdata[tid +  2]); } \
		if (blockSize >=   2) { sdata[tid] = op(sdata[tid], sdata[tid +  1]); } \
	  }																	\
																		\
	if (tid == 0) g_odata[blockIdx.x] = sdata[0];						\
  }																		

int greatest_power_of_two( int x )
{
  if ( x < 2 ) return 0;
  x /= 2;
  int n = 1;
  while ( x > 0 )
	{
	  x /= 2;
	  n *= 2;
	}
  return n;
}

////////////////////////////////////////////////////////////////////////////////
// Wrapper function for kernel launch
////////////////////////////////////////////////////////////////////////////////
// should result in two elements per thread
#define define_reduction_call(name,op)									\
  extern "C" float reduce_ ## name ( float *d_idata, int size )			\
  {																		\
	const int maxThreads = min( 512, greatest_power_of_two( GPUData_max_block_dim_size ) ); \
	const int maxBlocks  = GPUData_max_grid_dim_size;					\
	int threads = min( maxThreads, greatest_power_of_two( size / 2 ) ); \
	float * h_odata;													\
	size_t n_odata;														\
																		\
	if ( threads >= 1 )													\
	  {																	\
		int blocks = size / (threads * 2);								\
		n_odata = blocks = min(maxBlocks, blocks);						\
																		\
		dim3 dimBlock(threads, 1, 1);									\
		dim3 dimGrid(blocks, 1, 1);										\
		int smemSize = threads * sizeof(float);							\
																		\
		float * d_odata;												\
		size_t tmp_size = n_odata * sizeof(float);						\
		GPUDataException::check_error( "malloc in accelerated cuda reduction" ,cudaMalloc( (void**) &d_odata, tmp_size ) ); \
																		\
		switch (threads)												\
		  {																\
		  case 512:														\
			reduce_ ## name ## _kernel<512><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case 256:														\
			reduce_ ## name ## _kernel<256><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case 128:														\
			reduce_ ## name ## _kernel<128><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case 64:														\
			reduce_ ## name ## _kernel< 64><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case 32:														\
			reduce_ ## name ## _kernel< 32><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case 16:														\
			reduce_ ## name ## _kernel< 16><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case  8:														\
			reduce_ ## name ## _kernel<  8><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case  4:														\
			reduce_ ## name ## _kernel<  4><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case  2:														\
			reduce_ ## name ## _kernel<  2><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  case  1:														\
			reduce_ ## name ## _kernel<  1><<< dimGrid, dimBlock, smemSize >>>(d_odata, d_idata, size); break; \
		  default:														\
			exit(1);													\
		  }																\
																		\
		char tmp[1024];													\
		sprintf(tmp,"error after reduction with %i threads %i blocks",threads,blocks); \
		GPUDataException::check_error(tmp);								\
																		\
		h_odata = new float[n_odata];									\
		sprintf(tmp,"memcopy in accelerated cuda reduction from device %x to host %x %i bytes",d_odata, h_odata, tmp_size); \
		GPUDataException::check_error( tmp, cudaMemcpy( h_odata, d_odata, tmp_size, cudaMemcpyDeviceToHost ) ); \
		GPUDataException::check_error( "free in accelerated cuda reduction" ,cudaFree( d_odata ) ); \
	  }																	\
	else																\
	  {																	\
		h_odata = new float[size];										\
		GPUDataException::check_error( "free in non-accelerated cuda reduction", cudaMemcpy( h_odata, d_idata, size*sizeof(float), cudaMemcpyDeviceToHost) ); \
		n_odata = size;													\
	  }																	\
																		\
																		\
	float result = h_odata[0];											\
	for( int i = 1; i < n_odata; i++ )									\
	  result = op(result, h_odata[i]);									\
	 																	\
	delete[] h_odata;													\
  																		\
	return result;														\
  }


#define define_reduction(name,gpu_op,cpu_op)	\
  define_reduction_kernel( name, gpu_op )		\
  define_reduction_call  ( name, cpu_op )

__device__ float gpu_add( float a, float b ) { return a + b; }
__device__ float gpu_mul( float a, float b ) { return a * b; }
float cpu_add( float a, float b ) { return a + b; }
float cpu_mul( float a, float b ) { return a * b; }

define_reduction(min,fminf,fminf)
define_reduction(max,fmaxf,fmaxf)
define_reduction(sum,gpu_add,cpu_add)
define_reduction(product,gpu_mul,cpu_mul)
