/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"

__global__ void dsetIfBetween(float* a, const float min, const float max, const float val, int N) {
	// no of blocks
	//const int noOfBlocks = gridDim.x;
	// Block index
	const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

	// no of threads
	const int noOfThreads = blockDim.x;
	// Thread index
	const int tIndex = threadIdx.x;

	// index Offset
	const int i = bIndex* noOfThreads + tIndex ;

	if (i<N) {
		float va = a[i];
		if (va>=min && va<=max) va = val;
		a[i] = va;
	}
}

__global__ void dsetAnotherIfBetween(float* a, float* res, const float min, const float max, const float val, int N) {
	// no of blocks
	//const int noOfBlocks = gridDim.x;
	// Block index
	const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

	// no of threads
	const int noOfThreads = blockDim.x;
	// Thread index
	const int tIndex = threadIdx.x;

	// index Offset
	const int i = bIndex* noOfThreads + tIndex ;

	if (i<N) {
//		float va = a[i];
//		float re = res[i];
//		if (va>=min && va<=max) re = val;
//		res[i] = re;
		float va = a[i];
		if (va>=min && va<=max) res[i] = val;
	}
}

__global__ void dHeaviside(float* a, int N, float v) {
	// no of blocks
	//const int noOfBlocks = gridDim.x;
	// Block index
	const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

	// no of threads
	const int noOfThreads = blockDim.x;
	// Thread index
	const int tIndex = threadIdx.x;

	// index Offset
	const int i = bIndex* noOfThreads + tIndex ;

	if (i<N) {
		float va = a[i];
		a[i] = (va>v) ? (1.0f) : ( (va<v) ? (0.0f) : (0.5f) );
	}
}

__global__ void dHeaviside2(float* a, int N, float v, const float centerValue) {
	// no of blocks
	//const int noOfBlocks = gridDim.x;
	// Block index
	const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

	// no of threads
	const int noOfThreads = blockDim.x;
	// Thread index
	const int tIndex = threadIdx.x;

	// index Offset
	const int i = bIndex* noOfThreads + tIndex ;

	if (i<N) {
		float va = a[i];
		a[i] = (va>v) ? (1.0f) : ( (va<v) ? (0.0f) : (centerValue) );
	}
}

extern "C" void doSetIfBetween(float* data, const float min, const float max, const float val, int N)
{
	dim3 dimBlock(128);
	int s  = iDivUp(N,dimBlock.x);
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);
	dsetIfBetween<<<dimGrid, dimBlock>>>(data,min,max, val,N);
}
extern "C" void doSetAnotherIfBetween(float* data, float* res, const float min, const float max, const float val, int N)
{
	dim3 dimBlock(128);
	int s  = iDivUp(N,dimBlock.x);
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);
	dsetAnotherIfBetween<<<dimGrid, dimBlock>>>(data,res,min,max, val,N);
}

extern "C" void doHeaviside(float* data, float f, int N) {
	dim3 dimBlock(128);
	int s  = iDivUp(N,dimBlock.x);
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);
    dHeaviside<<<dimGrid, dimBlock>>>(data, N, f);
}
extern "C" void doHeaviside2(float* data, float f, float centerValue, int N) {
	dim3 dimBlock(128);
	int s  = iDivUp(N,dimBlock.x);
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);
    dHeaviside2<<<dimGrid, dimBlock>>>(data, N, f, centerValue);
}
