/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "GPUDataBase.h"
#include "cutil.h"

/// performs regionized linear feature transform
/// transform is expected unit-stride in last dim
/// input is expected unit-stride in last dim
/// output is expected unit-stride in last dim
__global__ void d_pooled_linear_feature_transform( const float * featuremap, int fs1, int fs2, int fs3, int Kin, int Kout,
                                                   const float * ptrans, int ss1, 
												   const float * pmeans,
                                                   float * target, int td1, int td2, int ts1, int ts2, int ts3,
                                                   int d, int step )
{
  const int idx = blockIdx.x*blockDim.x + threadIdx.x;
  const int idy = blockIdx.y*blockDim.y + threadIdx.y;
  if ( idx < 0 || idy < 0 || idx >= td1 || idy >= td2 ) return;

        float * pout   = target     + idx        * ts1 + idy *        ts2;
  const float * pin    = featuremap + idx * step * fs1 + idy * step * fs2; 

  for ( int kout = 0; kout < Kout; ++kout, ptrans += ss1, pout += ts3 )
  {

    float value = 0;	
	
    const float * pptrans = ptrans;
    const float * ppmeans = pmeans;
    const float * ppin = pin;

#if 1

    for ( int x = 0; x < d; ++x, ppin += fs1 )
	  {

		const float * pppin = ppin;

		for ( int y = 0; y < d; ++y, pppin += fs2 )
		  {

			const float * ppppin = pppin;
			
			for ( int kin = 0 ; kin < Kin; ++kin, ++pptrans, ++ppmeans, ppppin += fs3 )
			  value += ( (*ppppin) - (*ppmeans) ) * (*pptrans);

		  }
	  }

#else

	// broken

	for ( int kin = 0 ; kin < Kin; ++kin, ++pptrans, ++ppmeans, ppin += fs3 )
	{

	  const float * pppin = pin;
	  const float * ppptrans = pptrans;
	  const float * pppmeans = ppmeans;
	  
	  for ( int x = 0; x < d; ++x, pppin += fs1 )
		{
		  
		  const float * ppppin = pppin;
		  
		  for ( int y = 0; y < d; ++y )
			{
			  
			  value += ( (*ppppin) - (*pppmeans) ) * (*ppptrans);
			  ppppin += fs2;
			  pppmeans += Kin;
			  ppptrans += Kin;

			}

	  	}
	}
	

#endif

    (*pout) = value;
	
  }
  
}

extern "C" {

void do_pooled_linear_feature_transform( const GPUDataBase * featuremap, 
										 const GPUDataBase * transform, 
										 const GPUDataBase * means, 
										 GPUDataBase * target, 
										 size_t d, size_t step )
{
  size_t N1 = target->dims[0];
  size_t N2 = target->dims[1];
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(N1,dimBlock2d.x),iDivUp(N2,dimBlock2d.y));
  d_pooled_linear_feature_transform<<<dimGrid2d, dimBlock2d>>>( featuremap->_data, featuremap->strides[0],
																featuremap->strides[1], featuremap->strides[2],
																featuremap->dims[2], transform->dims[0], // kin is featuremap depth
																transform->_data, transform->strides[0], // not transform, in-dim !!!
																means->_data,
																target->_data, target->dims[0], target->dims[1],
																target->strides[0], target->strides[1], target->strides[2],
																d, step );
}

}
