/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"

#define KERNEL2_A0_CONT(name,func)											\
  __global__ void name( float a0, float *d_in, int N, const float * d_op)				\
  {																		\
	const int bx = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x;					\
	const int i  = (bx + threadIdx.x);									\
	if (i<N)															\
	  {																	\
		d_in[i] = func( d_in[i], d_op[i], a0 );								\
	  }																	\
  }

#define KERNEL2_A0_1D(name,func)											\
  __global__ void name( float a0, float *d_in, int inc1,int N1, const float * d_op, int inc1o) \
  {																		\
	const int bx = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x;					\
	const int i1  = (bx + threadIdx.x);									\
	const int i = i1 * inc1;											\
	const int j = i1 * inc1o;											\
	if (i1<N1)															\
	  {																	\
		d_in[i] = func( d_in[i], d_op[j], a0 );								\
	  }																	\
  }

#define KERNEL2_A0_2D(name,func)											\
  __global__ void name( float a0, float *d_in, int inc1, int inc2, int N1, int N2,	\
					   const float *d_op, int inc1o, int inc2o )				\
  {																		\
	const int bx = (blockIdx.x * blockDim.x);							\
	const int by = (blockIdx.y * blockDim.y);							\
	const int i1  = (bx + threadIdx.x);									\
	const int i2  = (by + threadIdx.y);									\
	int i = i1 * inc1  + i2 * inc2;										\
	int j = i1 * inc1o + i2 * inc2o;									\
	if (i1<N1 && i2<N2)													\
	  {																	\
		d_in[i] = func( d_in[i], d_op[j], a0 );								\
	  }																	\
  }

#define KERNEL2_A0_3D(name,func)											\
  __global__ void name( float a0, float *d_in,int inc1, int inc2, int inc3, int N1, int N2, int N3, \
					   const float *d_op,int inc1o, int inc2o, int inc3o )	\
  {																		\
	const int bx = (blockIdx.x * blockDim.x);							\
	const int by = (blockIdx.y * blockDim.y);							\
	const int i1  = (bx + threadIdx.x);									\
	const int i2  = (by + threadIdx.y);									\
	int i = i1 * inc1  + i2 * inc2;										\
	int j = i1 * inc1o + i2 * inc2o;									\
	if ( i1 < N1 && i2 < N2 )											\
	  for ( int i3 = 0; i3 < N3; i3++ )									\
		{																\
		  d_in[i] = func( d_in[i], d_op[j], a0 );							\
		  i += inc3;													\
		  j += inc3o;													\
		}																\
  }


#define KERNEL2_A0_4D(name,func)											\
  __global__ void name( float a0,  float *d_in,int inc1, int inc2, int inc3, int inc4, \
						int N1, int N2, int N3, int N4,					\
						const float *d_op,int inc1o, int inc2o, int inc3o, int inc4o ) \
  {																		\
	const int bx = (blockIdx.x * blockDim.x);							\
	const int by = (blockIdx.y * blockDim.y);							\
	const int i1  = (bx + threadIdx.x);									\
	const int i2  = (by + threadIdx.y);									\
	int i = i1 * inc1  + i2 * inc2;										\
	int j = i1 * inc1o + i2 * inc2o;									\
	if (i1<N1 && i2<N2)													\
	  {																	\
		int index1  = i;												\
		int index1o = j;												\
		for ( int i4 = 0; i4 < N4; i4++ )								\
		  {																\
			int index2  = index1;										\
			int index2o = index1o;										\
			for( int i3 = 0; i3 < N3; i3++ )							\
			  {															\
				d_in[index2] = func( d_in[index2], d_op[index2o], a0 );		\
				index2  += inc3;										\
				index2o += inc3o;										\
			  }															\
			index1  += inc4;											\
			index1o += inc4o;											\
		  }																\
	  }																	\
  }

#define KERNEL_ALL(name,function)			\
  KERNEL2_A0_CONT(name##_cont,function)		\
  KERNEL2_A0_1D(name##_1d,function)			\
  KERNEL2_A0_2D(name##_2d,function)			\
  KERNEL2_A0_3D(name##_3d,function)			\
  KERNEL2_A0_4D(name##_4d,function)

#define EXTERN_C(name, kernelname)										\
  extern "C" void name( float a0,  float *d_in,int inc1, int inc2, int inc3, int inc4, \
						int N1, int N2, int N3, int N4,					\
						const float *d_op,int inc1o, int inc2o, int inc3o, int inc4o ) \
  {																		\
	/* Decision which kernel is used by size of strides	*/				\
	if(N2==1 && N3==1 && N4==1 && inc1==1 && inc1o == 1){				\
	  dim3 dimBlock1d = GPUData_dimBlock1d;								\
	  dim3 dimGrid1d  = make_2d_gridDim_for_1d_blockDim(dimBlock1d.x,N1); \
	  kernelname##_cont<<<dimGrid1d, dimBlock1d>>>( a0, d_in, N1, d_op);		\
	} else if(N2==1 && N3==1 && N4==1){									\
	  dim3 dimBlock1d = GPUData_dimBlock1d;								\
	  dim3 dimGrid1d  = make_2d_gridDim_for_1d_blockDim(dimBlock1d.x,N1); \
	  kernelname##_1d<<<dimGrid1d, dimBlock1d>>>( a0, d_in, inc1, N1, d_op, inc1o); \
	} else if(N3==1 && N4==1) {											\
		dim3 dimBlock2d = GPUData_dimBlock2d;											\
		dim3 dimGrid2d ( iDivUp( N1, dimBlock2d.x ), iDivUp( N2, dimBlock2d.y ) ); \
		kernelname##_2d<<<dimGrid2d, dimBlock2d>>>( a0, d_in, inc1, inc2, N1, N2, d_op, inc1o, inc2o); \
	} else if(N4==1){													\
		dim3 dimBlock2d = GPUData_dimBlock2d;											\
		dim3 dimGrid2d ( iDivUp( N1, dimBlock2d.x ), iDivUp( N2, dimBlock2d.y ) ); \
		kernelname##_3d<<<dimGrid2d, dimBlock2d>>>( a0, d_in, inc1, inc2, inc3, N1, N2, N3, d_op, inc1o, inc2o, inc3o); \
	} else {															\
		dim3 dimBlock2d = GPUData_dimBlock2d;											\
		dim3 dimGrid2d ( iDivUp( N1, dimBlock2d.x ), iDivUp( N2, dimBlock2d.y ) ); \
		kernelname##_4d<<<dimGrid2d, dimBlock2d>>>( a0, d_in, inc1, inc2, inc3, inc4, N1, N2, N3, N4, d_op, inc1o, inc2o, inc3o, inc4o); \
	}																	\
  }


  __device__ float mul_scaled( float a, float b, float s ) { return a * b * s; }
  __device__ float fb_mod( float a, float b, float c ) { return a * ( 1 + b * c ); }
  __device__ float add_scaled( float a, float b, float s ) { return a + b * s; }
  __device__ float add_squared( float a, float b, float s ) { return a + b * b * s; }
  __device__ float add_abs( float a, float b, float s ) { return a + fabsf(b) * s; }

#define KERNEL_AND_EXTERN_C(name,fun)			\
  KERNEL_ALL(d ## name ## _binary_a0, fun)					\
   EXTERN_C( do_local_binary_a0_ ## name, d ## name ## _binary_a0 )

KERNEL_AND_EXTERN_C(Mul, mul_scaled)
KERNEL_AND_EXTERN_C(Add, add_scaled)
KERNEL_AND_EXTERN_C(AddSquared, add_squared)
KERNEL_AND_EXTERN_C(AddAbs, add_abs)
KERNEL_AND_EXTERN_C(FbMod, fb_mod)

