/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
 /*
 * local_contigous_array.cu
 *
 *  Created on: 12-Feb-2009
 *      Author: feisenkeil
 */

#include <cuda.h>
#include "cutil.h"

#define KERNEL_A0_CONT(name,func)  													\
__global__ void name ## _a0(float a0, float *d_in, int N){							\
	const int bx = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x;				\
	const int i  = (bx + threadIdx.x); 												\
	if (i<N) {																		\
		d_in[i] = func ## _a0(d_in[i],  a0);													\
	}																				\
}

#define KERNEL_A0_1D(name,func)  														\
__global__ void name ## _a0(float a0, float *d_in, int inc1,int N1){				\
	const int bx = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x;					\
	const int i1  = (bx + threadIdx.x); 												\
	const int i  = 	i1*inc1; 															\
	if (i1<N1) {																	\
		d_in[i] = func ## _a0(d_in[i],  a0);													\
	}																				\
}

#define KERNEL_A0_2D(name,func)  														\
__global__ void name ## _a0(float a0, float *d_in, int inc1, int inc2, int N1, int N2)\
{				\
const int bx = (blockIdx.x * blockDim.x); 											\
const int by = (blockIdx.y * blockDim.y); 											\
const int i1  = (bx + threadIdx.x); 												\
const int i2  = (by + threadIdx.y); 												\
const int i  = 	i1*inc1+i2*inc2; 													\
	if (i1<N1 && i2<N2) {															\
		d_in[i] = func ## _a0(d_in[i],  a0);													\
	}																				\
}

#define KERNEL_A0_3D(name,func)  														\
__global__ void name ## _a0(float a0, float *d_in,int inc1, int inc2, int inc3, int N1, int N2, int N3){	\
const int bx = (blockIdx.x * blockDim.x); 											\
const int by = (blockIdx.y * blockDim.y); 											\
const int i1  = (bx + threadIdx.x); 												\
const int i2  = (by + threadIdx.y); 												\
int i = i1*inc1+i2*inc2;	 														\
if (i1<N1 && i2<N2)																	\
for (int i3=0; i3<N3; i3++){														\
	d_in[i] = func ## _a0(d_in[i],  a0);														\
	i += inc3;																		\
}																					\
}


#define KERNEL_A0_4D(name,func)  														\
__global__ void name ## _a0(float a0, float *d_in,int inc1, int inc2, int inc3, int inc4, int N1, int N2, int N3, int N4){ \
const int bx = (blockIdx.x * blockDim.x);											\
const int by = (blockIdx.y * blockDim.y);											\
const int i1  = (bx + threadIdx.x);													\
const int i2  = (by + threadIdx.y);													\
int i = i1*inc1+i2*inc2;															\
if (i1<N1 && i2<N2){																\
	int index1 = i;																	\
	for (int i4=0; i4<N4; i4++){													\
		int index2 = index1;														\
		for(int i3=0; i3<N3; i3++){													\
			d_in[index2] = func ## _a0(d_in[index2], a0);										\
			index2 += inc3;															\
		}																			\
		index1 += inc4;																\
	}																				\
}																					\
}

#define KERNEL_A0_ALL(name,function)        \
KERNEL_A0_CONT(name##_cont,function)       \
KERNEL_A0_1D(name##_1d,function)           \
KERNEL_A0_2D(name##_2d,function)           \
KERNEL_A0_3D(name##_3d,function)           \
KERNEL_A0_4D(name##_4d,function)

#define EXTERN_C(name, kernelname)													\
extern "C" void name ## _a0(float a0, float* data, int inc1, int inc2, int inc3, int inc4, int N1, int N2, int N3, int N4) {\
	/* Decision which kernel is used by size of strides	*/							\
	if(N2==1 && N3==1 && N4==1 && inc1==1){											\
		dim3 dimBlock1d = GPUData_dimBlock1d;										\
		dim3 dimGrid1d  = make_2d_gridDim_for_1d_blockDim(dimBlock1d.x,N1);		\
		kernelname##_cont_a0<<<dimGrid1d, dimBlock1d>>>( a0, data, N1);				\
	} else if(N2==1 && N3==1 && N4==1){												\
		dim3 dimBlock1d = GPUData_dimBlock1d;										\
		dim3 dimGrid1d  = make_2d_gridDim_for_1d_blockDim(dimBlock1d.x,N1);		\
		kernelname##_1d_a0<<<dimGrid1d, dimBlock1d>>>( a0, data, inc1, N1);			\
	} else if(N3==1 && N4==1) {														\
		dim3 dimBlock2d = GPUData_dimBlock2d;											\
		dim3 dimGrid2d(iDivUp(N1,dimBlock2d.x),iDivUp(N2,dimBlock2d.y));				\
		kernelname##_2d_a0<<<dimGrid2d, dimBlock2d>>>( a0, data, inc1, inc2, N1, N2);			\
	} else if(N4==1){																\
		dim3 dimBlock2d = GPUData_dimBlock2d;											\
		dim3 dimGrid2d(iDivUp(N1,dimBlock2d.x),iDivUp(N2,dimBlock2d.y));				\
		kernelname##_3d_a0<<<dimGrid2d, dimBlock2d>>>( a0, data, inc1, inc2, inc3, N1, N2, N3);	\
	} else {																		\
		dim3 dimBlock2d = GPUData_dimBlock2d;											\
		dim3 dimGrid2d(iDivUp(N1,dimBlock2d.x),iDivUp(N2,dimBlock2d.y));				\
		kernelname##_4d_a0<<<dimGrid2d, dimBlock2d>>>( a0, data, inc1, inc2, inc3, inc4, N1, N2, N3, N4);	\
	}																	\
}


#define KERNEL_AND_EXTERN_C(name,fun)			\
  KERNEL_A0_ALL(d ## name, fun)					\
  EXTERN_C( do_local_ ## name, d ## name )


__device__ float mod_a0(float a, float a0) {
  	return fmodf(a,a0);
}

__device__ float pow_a0(float a, float a0) {
  	return powf(a,a0);
}

__device__ float safe_log_a0(float a, float epsilon) {
  	return logf( max(a, epsilon) );
}

__device__ float rect_a0(float a, float a0) {
  if ( a >= a0 ) return a;
  else return 0.0f;
}

__device__ float compress_a0(float a, float a0) {
  float tmp = a*a0;
  return tmp < 1.0f ? tmp : 1.0f;
}


__device__ float max_a0(float a, float a0) {
  return a > a0 ? a : a0;
}
__device__ float min_a0(float a, float a0) {
  return a > a0 ? a0 : a;
}

__device__ float sigmoid_a0(float a, float a0) {
  float tmp = a*a0*2.0f;
  return 0.5f * ( 1.0f + tmp * rsqrtf( 1 + tmp*tmp ) );
}

/* 
 * The following lines generates extern "C" functions.
 * Eg.: KERNEL_AND_EXTERN_C(Max,max) evaluates to 
 *
 * void do_local_Max_a0(float a0, float* dataPtr, int inc1, int inc2, int inc3, int inc4, int N1, int N2, int N3, int N4);
 * 
 * inc1 should be the smalles dimension stride, inc4 the largest.
 * N1 determines the size of the deepst dimension (or more exact, the size of the
 * dimension with stride inc1).
 * ...
 * 
 */

KERNEL_AND_EXTERN_C(Mod,mod)
KERNEL_AND_EXTERN_C(Pow,pow)
KERNEL_AND_EXTERN_C(Safe_log,safe_log)
KERNEL_AND_EXTERN_C(Rect,rect)
KERNEL_AND_EXTERN_C(Compress,compress)
KERNEL_AND_EXTERN_C(Max,max)
KERNEL_AND_EXTERN_C(Min,min)
KERNEL_AND_EXTERN_C(Sigmoid,sigmoid)


#define GEN_OP_FUN(name,op)									\
  __device__ float name ## _a0(float a, float b) { return a op b; }

#define OP_KERNEL_AND_EXTERN_C( name, op )		\
  GEN_OP_FUN(name, op)							\
  KERNEL_AND_EXTERN_C(name,name)

/* 
 * The following lines generates extern "C" functions.
 * Eg.: OP_KERNEL_AND_EXTERN_C(Add, + ) evaluates to 
 *
 * void do_local_Add_a0(float a0, float* dataPtr, int inc1, int inc2, int inc3, int inc4, int N1, int N2, int N3, int N4);
 * 
 * details see above.
 * 
 */
OP_KERNEL_AND_EXTERN_C( Add, + )
OP_KERNEL_AND_EXTERN_C( Sub, - )
OP_KERNEL_AND_EXTERN_C( Mul, * )
OP_KERNEL_AND_EXTERN_C( Div, / )
