/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include <cmath>

// forward declarations
__device__ void compare_tuple (float*, float*, float, float);
__device__ void compare_triple (float*, float*, float*, float, float, float);
__device__ float get_median (float, float, float, float, float, float, float, float, float);
__global__ void median3x3_kernel (float*, const float*, const int, const int);


#define BLOCKDIM_X 32		//! blocksize X
#define BLOCKDIM_Y 12		//! blocksize Y
#define PPT        4		//! how many points a thread will take care of when calculating results
#define APRON      1		//! apron size. 3x3 median filter, so it's 1 around each point


/**
 * median3x3 function for external calling from within a C/C++ project
 */
extern "C" void
do_median3x3 (
		float *res,			//<! memory to store result to. has to be allocated on CUDA device
		const float *src,	//<! source from which to calculate the filtered image. has to be CUDA allocated memory
		int width,			//<! input data width
		int height)			//<! input data height
{
	int x = ceil( (float)width/(BLOCKDIM_X*PPT) );
	int y = ceil( (float)height/(BLOCKDIM_Y) );

	dim3 blocks(x, y);
	dim3 threads(BLOCKDIM_X, BLOCKDIM_Y);
	median3x3_kernel<<<blocks, threads>>>(res, src, width, height);
}


/**
 * median3x3 Kernel for computing, as the name says, a 3x3 median on some input
 * data. The result will be stored in res, source is coming from src. width and
 * height have to be set appropriatly.
 */
__global__ void
median3x3_kernel (
		float *res,			//<! memory to store result to
		const float *src,	//<! source from which to calculate the filtered image
		const int width,	//<! input data width
		const int height)	//<! input data height
{
	/*
	 * Kernel Implementation specific notes. 
	 * 
	 * As they are not relevant to the user of the cuda median3x3 kernel, these 
	 * notes here are not included into doxygen material. Assumption is that 
	 * they are only relevant to one who wants to alter the behaviour.
	 *
	 * Loading the source into shared memory happens in 5 "steps", where as many
	 * threads a possible will have something to do. In the first step every
	 * thread will have to work, then most of the threads will do some work in
	 * either of the steps 2-5, a few threads are idle meanwhile.
	 *
	 * Assuming PPT=4, BLOCKDIM_X=32, BLOCKDIM_Y=12, APRON=1, the shared memory 
	 * layout looks like the following image:
	 *
	 * 
	 *    x 0 1 2 3 ...                                    ... 129
	 *  y  ______________________________________ ... ___________
	 *  0  |_|_|_|_|_|___________________________ ... _________|_|
	 *  1  |_|_|_|_|_|___________________________ ... _________|_|
	 *  2  |_|_|_|_|_|___________________________ ... _________|_|
	 *  .   .
	 *  .   .
	 *  .   .
	 * 12  |_|_|_|_|_|___________________________ ... _________|_|
	 * 13  |_|_|_|_|_|___________________________ ... _________|_|
	 *
	 * 
	 * Step 1: main data is loaded into shared memory. It will occupy memory
	 * from index 1/1 to 128/13. As the memory is loaded in chunks (each thread
	 * will load 4 pixels, consecutively), one half warp will access 16 floats.
	 * This ensures that even on compute capability 1.0 and 1.1 cards, memory
	 * coalescing will happen for the whole main data. Take note that there are
	 * 32 * 12 threads, so thread with thread-index 0/0 will load data into
	 * items 1/1, 1/32, 1/64, 1/96. other threads respectively. When the image
	 * size is not a multiple of BLOCKDIM_X or _Y, the image border will
	 * automatically be expanded, though this might break coalescing at the
	 * right image edge.
	 *
	 * Step 2: The upper apron. As there are enough threads for work, the
	 * threads with an y-index of 0 <= y < 4 will be taken to load the upper
	 * apron. The thread with an index of 0/0 loads memory to 1/0, the thread
	 * with an index of 0/15 loads memory to 0/15 and a thread with index 2/13
	 * will load memory to 2*BLOCKDIM_X + 13 = 0/77.
	 * This again will be coalesced, as consecutive threads will load
	 * consecutive chunks from host memory.
	 * When apron-loading reaches an image edge, the border will
	 *
	 * Step 3: The same as step 2, but for threads with an y-index of 4 <= y <
	 * 8. They will write input data to the lower apron (y-index 13).
	 *
	 * Step 4: The threads of y-Index 8 and with a lower x-index than
	 * BLOCKDIM_Y+2*APRON (=14) will be taken to load the left side apron. This
	 * is not coalesced, as coalescing won't speed up anything here. We might
	 * load chunks of 16 floats every time, but as the values we need (1 of 16
	 * for each row) are not consecutive, loading 16 * 16 values or 16 * 1 value
	 * will need same time. Image border is automatically expanded.
	 *
	 * Step 5: Same as step 4, but for y-Index 9 and right apron.
	 *
	 * Finally, each thread will calculate "its" corresponding median values.
	 * Median calculation is, at the moment, not split up. It would require more
	 * shared memory to store results to, what might break up current memory
	 * layout - which uses about 7.5k shared mem, 16 registers and 28 bytes of 
	 * constant memory already.
	 */


	// setting up shared memory
	__shared__ float sMem [BLOCKDIM_Y       + 2*APRON] 
	                      [BLOCKDIM_X * PPT + 2*APRON];
	
	// offsets into global memory
	const int blockbaseX = blockIdx.x * (BLOCKDIM_X * PPT);		//<! x-offset into global memory of current block
	const int blockbaseY = blockIdx.y * (BLOCKDIM_Y);			//<! y-offset into global memory of current block
	const int baseX = blockbaseX + threadIdx.x;					//<! x-offset into global memory of current thread
	const int baseY = blockbaseY + threadIdx.y;					//<! y-offset into global memory of current thread
	const int maxIdx = width * height;							//<! maximal index of input data


	// ------------------------------------------------------------------------
	//
	// main data
	//
	#define Y threadIdx.y + APRON
	#define X threadIdx.x + APRON + i*BLOCKDIM_X
	#pragma unroll
	for (int i = 0; i < PPT; i++) {
		//
		// expand automatically. this might break coalescing in some
		// circumstances, for example when the image width is not a multiple of
		// BLOCKDIM_X. It should not affect coalescing when the lower end is
		// trimmed to o = height-1, though.
		// It's made here so one has not to think about the right edge in
		// loading the apron if the image edge "cuts" right through a shared
		// memory block
		//
		int x = baseX + i*BLOCKDIM_X;
		if (x >= width) 
			x = width-1;

		int o = baseY;
		if (o >= height) 
			o = height-1;

		o = o * width + x;
		sMem[Y][X] = src[o];
	}
	#undef X
	#undef Y
	

	// ------------------------------------------------------------------------
	//
	// upper apron
	//
	#define Y 0
	#define X threadIdx.x + (BLOCKDIM_X * threadIdx.y)
	if (threadIdx.y < 4)
	{
		int x = blockbaseX + X;
		if (x >= width)
			x = width-1;

		int o = blockbaseY - APRON;
		if (o < 0)
			o = 0;

		o = o * width + x;
		sMem[Y][X + APRON] = src[o];
	}
	#undef X
	#undef Y


	// ------------------------------------------------------------------------
	//
	// lower apron
	//
	#define Y BLOCKDIM_Y + APRON
	#define X threadIdx.x + (BLOCKDIM_X * (threadIdx.y-4))
	if (threadIdx.y >= 4 && threadIdx.y < 8) 
	{
		int x = blockbaseX + X;
		if (x >= width)
			x = width - 1;

		int o = blockbaseY + BLOCKDIM_Y;
		if (o >= height)
			o = height-1;

		o = o * width + x;
		sMem[Y][X + APRON] = src[o];
	}
	#undef X
	#undef Y


	// ------------------------------------------------------------------------
	//
	// left apron
	//
	#define Y threadIdx.x
	#define X 0
	if (threadIdx.y == 8 && threadIdx.x < (BLOCKDIM_Y + 2*APRON))
	{
		int x = blockbaseX - APRON;
		if (x < 0)
			x = 0;

		int o = blockbaseY - APRON + Y;
		if (o < 0)
			o = 0;
		if (o >= height)
			o = height-1;

		o = o * width + x;
		sMem[Y][X] = src[o];
	}
	#undef X
	#undef Y


	// ------------------------------------------------------------------------
	//
	// right apron
	//
	#define Y threadIdx.x
	#define X BLOCKDIM_X*PPT + APRON
	if (threadIdx.y == 9 && threadIdx.x < (BLOCKDIM_Y + 2*APRON))
	{
		int x = blockbaseX + BLOCKDIM_X*PPT;
		if (x >= width)
			x = width - 1;

		int o = blockbaseY - APRON + Y;
		if (o < 0)
			o = 0;
		if (o >= height)
			o = height - 1;

		o = o * width + x;
		sMem[Y][X] = src[o];
	}
	#undef X
	#undef Y


	// ------------------------------------------------------------------------
	//
	// calculate and store resulst
	//
	__syncthreads();
	#define Y threadIdx.y + APRON
	#define X threadIdx.x + APRON + i*BLOCKDIM_X
	#pragma unroll
	for (int i = 0; i < PPT; i++) {
		int x = baseX + i * BLOCKDIM_X;
		int o = baseY * width + x;

		// make sure to don't write over the allocated memory space
		if (x < width && o < maxIdx)
			/*
			 * parameters for get_median are taken the following
			 * way around sMem[Y][X]
			 *
			 * f1 f2 f3
			 * f4 f5 f6
			 * f7 f8 f9
			 *
			 */
			res[o] = get_median
				( sMem[Y-1][X-1]
				, sMem[Y-1][X  ]
				, sMem[Y-1][X+1]

				, sMem[Y  ][X-1]
				, sMem[Y  ][X  ]
				, sMem[Y  ][X+1]
				
				, sMem[Y+1][X-1]
				, sMem[Y+1][X  ]
				, sMem[Y+1][X+1]
				);
	}
	#undef X
	#undef Y
}	


/**
 * Compare two floats. The greater value will be stored in r0, the greater value
 * will be stored in r1
 *
 * \param r0	result/out parameter 0
 * \param r1	result/out parameter 1
 * \param f1	first float to examine
 * \param f2	second float to examine
 */
__device__ void
compare_tuple (float *r0, float *r1, float f1, float f2)
{
	if (f1 < f2) {
		*r0 = f2;
		*r1 = f1;
	}
	else {
		*r0 = f1;
		*r1 = f2;
	}
}


/**
 *
 * compare three floats and store the values into r0-r2 in ascending order.
 *
 * \param r0	lower result value
 * \param r1	middle result value
 * \param r2	bigger result value
 * \param f1	first float to examine
 * \param f2	second float to examine
 * \param f3	third float to examine
 */
__device__ void
compare_triple (float *r0, float *r1, float *r2,
                float f1, float f2, float f3)
{
	float c10, c11;
	float c20, c21;
	float c30, c31;

	compare_tuple(&c10, &c11, 
	              f2, f3);

	compare_tuple(&c20, &c21, 
	              f1, c11);

	compare_tuple(&c30, &c31, 
	              c20, c10);

	*r0 = c21;
	*r1 = c31;
	*r2 = c30;
}

/**
 * As proposed in [TODO: INSERT REFERENCE HERE], on a 3x3 median filter it's
 * faster to calculate the median value with as few comparisons as possible
 * instead of having a sorted list.
 */
__device__ float
get_median (float f1, float f2, float f3, 
            float f4, float f5, float f6, 
            float f7, float f8, float f9)
{
	float tmp1, tmp2;
	float t10, t11, t12;
	float t20, t21, t22;
	float t30, t31, t32;
	float d0, d1;

	compare_triple(&t10, &t11, &t12, 
			       f1, f2, f3);

	compare_triple(&t20, &t21, &t22,
	               f4, f5, f6);

	compare_triple(&t30, &t31, &t32, 
	               f7, f8, f9);


	compare_tuple(&d0, &d1, 
	              t10, t20);
	tmp1 = d0;
	compare_tuple(&d0, &d1, 
	              tmp1, t30);
	tmp1 = d0;

	compare_tuple(&d0, &d1, 
	              t22, t32);
	tmp2 = d1;
	compare_tuple(&d0, &d1, 
	              t12, tmp2);
	tmp2 = d1;

	compare_triple(&t10, &t11, &t12, 
	               t11, t21, t31);
	compare_triple(&t10, &t11, &t12,
	               tmp1, t11, tmp2);
	return t11;
}

