/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cutil.h>
#include <stdio.h>


__global__ void dMaxOfTwoVectors(float* a,float *b, int N) {
	
	// index Offset
	const int i = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (i<N) {
		a[i] = fmaxf(a[i],b[i]);
	}
}
__global__ void dMinOfTwoVectors(float* a,float *b, int N) {
	
	// index Offset
	const int i = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (i<N) {
		a[i] = fminf(a[i],b[i]);
	}
}


//24-bit multiplication is faster on G80,
//but we must be sure to multiply integers
//only within [-8M, 8M - 1] range
#define IMUL(a, b) __mul24(a, b)

__device__ float rtv;
__device__ int rtvInt;
__device__ float rtvTemp[1024];
__device__ int   rtvIntTemp[1024];

__global__ void dev_Sum(float* data, const int N, const int off) {
	
	__shared__ float temp[16];
	
	const int pos = IMUL(blockIdx.x,16)+threadIdx.x;
	
	if (pos<N) {
		temp[threadIdx.x] = data[pos+off];
	} else {
		temp[threadIdx.x] = 0;
	}
	
	__syncthreads();
	
	if (threadIdx.x%2==0) {
		temp[threadIdx.x] = temp[threadIdx.x] + temp[threadIdx.x+1];
	} else ;;
	__syncthreads();
	if (threadIdx.x%4==0) {
		temp[threadIdx.x] = temp[threadIdx.x] + temp[threadIdx.x+2];
	} else ;;
	__syncthreads();
	if (threadIdx.x%8==0) {
		temp[threadIdx.x] = temp[threadIdx.x] + temp[threadIdx.x+4];
	} else ;;
	__syncthreads();
	if (threadIdx.x%16==0) {
		rtvTemp[blockIdx.x] = temp[threadIdx.x] + temp[threadIdx.x+8];
		if (gridDim.x==1) rtv = temp[threadIdx.x] + temp[threadIdx.x+8];
	}
}

__global__ void dev_MaxOnExtern(float* data, const int N, const int off) {
	
	__shared__ float temp[16];
	__shared__ int    tempInt[16];
	
	const int pos = IMUL(blockIdx.x,16)+threadIdx.x;
	const int posShure = IMUL(blockIdx.x,16);
	const int tIndex = threadIdx.x;
	const int noOfThreads = blockDim.x;
	
	if (pos<N) {
		temp[threadIdx.x] = data[pos+off];
		tempInt[threadIdx.x] = pos+off;
	} else {
		temp[threadIdx.x] = data[posShure+off]-1.0f;
		tempInt[threadIdx.x] = -1;
	}
	
	__syncthreads();
	
	if (threadIdx.x%2==0 && (tIndex+2/2)<noOfThreads) {
		if (temp[threadIdx.x] < temp[threadIdx.x+1]) {
			temp[threadIdx.x] = temp[threadIdx.x+1];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+1];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%4==0 && (tIndex+4/2)<noOfThreads) {
		if (temp[threadIdx.x] < temp[threadIdx.x+2]) {
			temp[threadIdx.x] = temp[threadIdx.x+2];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+2];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%8==0 && (tIndex+8/2)<noOfThreads) {
		if (temp[threadIdx.x] < temp[threadIdx.x+4]) {
			temp[threadIdx.x] = temp[threadIdx.x+4];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+4];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%16==0) {
		if (temp[threadIdx.x] < temp[threadIdx.x+8]) {
			rtvTemp[blockIdx.x] = temp[threadIdx.x+8];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x+8];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x+8];
				rtvInt = tempInt[threadIdx.x+8];
			}
		} else {
			rtvTemp[blockIdx.x] = temp[threadIdx.x];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x];
				rtvInt = tempInt[threadIdx.x];
			}
		}
	}
}

__global__ void dev_MinOnExtern(float* data, const int N, const int off) {
	
	__shared__ float temp[16];
	__shared__ int    tempInt[16];
	
	const int pos = IMUL(blockIdx.x,16)+threadIdx.x;
	const int posShure = IMUL(blockIdx.x,16);
	const int tIndex = threadIdx.x;
	const int noOfThreads = blockDim.x;
	
	if (pos<N) {
		temp[threadIdx.x] = data[pos+off];
		tempInt[threadIdx.x] = pos+off;
	} else {
		temp[threadIdx.x] = data[posShure+off]+1.0f;
		tempInt[threadIdx.x] = -1;
	}
	
	__syncthreads();
	
	if (threadIdx.x%2==0 && (tIndex+2/2)<noOfThreads) {
		if (temp[threadIdx.x] > temp[threadIdx.x+1]) {
			temp[threadIdx.x] = temp[threadIdx.x+1];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+1];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%4==0 && (tIndex+4/2)<noOfThreads) {
		if (temp[threadIdx.x] > temp[threadIdx.x+2]) {
			temp[threadIdx.x] = temp[threadIdx.x+2];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+2];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%8==0 && (tIndex+8/2)<noOfThreads) {
		if (temp[threadIdx.x] > temp[threadIdx.x+4]) {
			temp[threadIdx.x] = temp[threadIdx.x+4];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+4];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%16==0) {
		if (temp[threadIdx.x] > temp[threadIdx.x+8]) {
			rtvTemp[blockIdx.x] = temp[threadIdx.x+8];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x+8];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x+8];
				rtvInt = tempInt[threadIdx.x+8];
			}
		} else {
			rtvTemp[blockIdx.x] = temp[threadIdx.x];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x];
				rtvInt = tempInt[threadIdx.x];
			}
		}
	}
}


__global__ void dev_MinOnIntern(const int N) {
	
	__shared__ float temp[16];
	__shared__ int   tempInt[16];
	
	const int pos = IMUL(blockIdx.x,16)+threadIdx.x;
	const int posShure = IMUL(blockIdx.x,16);
	const int noOfThreads = blockDim.x;
	const int tIndex = threadIdx.x;
	
	if (pos<N) {
		temp[threadIdx.x] = rtvTemp[pos];
		tempInt[threadIdx.x] = rtvIntTemp[pos];
	} else {
		temp[threadIdx.x] = rtvTemp[posShure]+1.0f;
		tempInt[threadIdx.x] = -1;
	}
	
	__syncthreads();
	
	if (threadIdx.x%2==0 && (tIndex+2/2)<noOfThreads) {
		if (temp[threadIdx.x] > temp[threadIdx.x+1]) {
			temp[threadIdx.x] = temp[threadIdx.x+1];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+1];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%4==0 && (tIndex+4/2)<noOfThreads) {
		if (temp[threadIdx.x] > temp[threadIdx.x+2]) {
			temp[threadIdx.x] = temp[threadIdx.x+2];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+2];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%8==0 && (tIndex+8/2)<noOfThreads) {
		if (temp[threadIdx.x] > temp[threadIdx.x+4]) {
			temp[threadIdx.x] = temp[threadIdx.x+4];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+4];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%16==0) {
		if (temp[threadIdx.x] > temp[threadIdx.x+8]) {
			rtvTemp[blockIdx.x] = temp[threadIdx.x+8];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x+8];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x+8];
				rtvInt = tempInt[threadIdx.x+8];
			}
		} else {
			rtvTemp[blockIdx.x] = temp[threadIdx.x];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x];
				rtvInt = tempInt[threadIdx.x];
			}
		}
	}
}

__global__ void dev_MaxOnIntern(const int N) {
	
	__shared__ float temp[16];
	__shared__ int   tempInt[16];
	
	const int pos = IMUL(blockIdx.x,16)+threadIdx.x;
	const int posShure = IMUL(blockIdx.x,16);
	const int noOfThreads = blockDim.x;
	const int tIndex = threadIdx.x;
	
	if (pos<N) {
		temp[threadIdx.x] = rtvTemp[pos];
		tempInt[threadIdx.x] = rtvIntTemp[pos];
	} else {
		temp[threadIdx.x] = rtvTemp[posShure]-1.0f;
		tempInt[threadIdx.x] = -1;
	}
	
	__syncthreads();
	
	if (threadIdx.x%2==0 && (tIndex+2/2)<noOfThreads) {
		if (temp[threadIdx.x] < temp[threadIdx.x+1]) {
			temp[threadIdx.x] = temp[threadIdx.x+1];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+1];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%4==0 && (tIndex+4/2)<noOfThreads) {
		if (temp[threadIdx.x] < temp[threadIdx.x+2]) {
			temp[threadIdx.x] = temp[threadIdx.x+2];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+2];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%8==0 && (tIndex+8/2)<noOfThreads) {
		if (temp[threadIdx.x] < temp[threadIdx.x+4]) {
			temp[threadIdx.x] = temp[threadIdx.x+4];
			tempInt[threadIdx.x] = tempInt[threadIdx.x+4];
		}
	} else ;;
	__syncthreads();
	if (threadIdx.x%16==0) {
		if (temp[threadIdx.x] < temp[threadIdx.x+8]) {
			rtvTemp[blockIdx.x] = temp[threadIdx.x+8];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x+8];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x+8];
				rtvInt = tempInt[threadIdx.x+8];
			}
		} else {
			rtvTemp[blockIdx.x] = temp[threadIdx.x];
			rtvIntTemp[blockIdx.x] = tempInt[threadIdx.x];
			if (gridDim.x==1) {
				rtv = temp[threadIdx.x];
				rtvInt = tempInt[threadIdx.x];
			}
		}
	}
}


extern "C" {
	
	float doSumMax1024x16(float* dData, int N, int offs) {
		float  retVal = 0;
		float* tempAdress;
		cudaGetSymbolAddress((void**) &tempAdress,rtvTemp);
		int     nN = N;
		do {
			dim3 block(16);
			dim3 grid(iDivUp(nN,block.x));
			
			//		printf("n %d \n",nN);
			//
			//		printf("block.x %d \n",block.x);
			//
			//		printf("grid.x %d \n",grid.x);
			
			if (N==nN) {
				dev_Sum<<< grid, block >>>(dData,nN,offs);
			} else {
				dev_Sum<<< grid, block >>>(tempAdress,nN,0);
			}
			
			nN = grid.x;
		} while (nN!=1);
		
		cudaGetSymbolAddress((void**) &tempAdress,rtv);
		cudaMemcpy(&retVal, tempAdress, sizeof(float)*1, cudaMemcpyDeviceToHost);
		return retVal;
	}
	float doSumMax1024x16AndCopyTo(float* dData, int N, int offs, float* dest) {
		float  retVal = 0;
		float* tempAdress;
		cudaGetSymbolAddress((void**) &tempAdress,rtvTemp);
		int     nN = N;
		do {
			dim3 block(16);
			dim3 grid(iDivUp(nN,block.x));
			
			//		printf("n %d \n",nN);
			//
			//		printf("block.x %d \n",block.x);
			//
			//		printf("grid.x %d \n",grid.x);
			
			if (N==nN) {
				dev_Sum<<< grid, block >>>(dData,nN,offs);
			} else {
				dev_Sum<<< grid, block >>>(tempAdress,nN,0);
			}
			
			nN = grid.x;
		} while (nN!=1);
		
		cudaGetSymbolAddress((void**) &tempAdress,rtv);
		cudaMemcpy(dest, tempAdress, sizeof(float)*1, cudaMemcpyDeviceToDevice);
		return retVal;
	}
	
	void doSumOverLastDimension(float* origin, const int n_x, const int n_y, float* dest) {
		if (n_x>1024*16) {
			fprintf(stderr,"error in doSumOverLastDimension(): dimension_x is too large");
			return;
		}
		for (int i=0; i<n_y; ++i) {
			doSumMax1024x16AndCopyTo(origin,n_x,i*n_x,dest);
			dest+=1;
		}
	}
	
	float doSum(float* dData, int N, int offs) {
		float retValAccu = 0;
		float retVal = 0;
		
		if (N<=1024*16) return doSumMax1024x16(dData,N,offs);
		
		int restN = N;
		while (restN>0) {
			retVal = doSumMax1024x16(dData,min(1024*16,restN),offs);
			retValAccu += retVal;
			//		printf("accu = %f ; val = %f\noffs = %d ; restN = %d\n",retValAccu, retVal,offs,restN);
			offs += 1024*16;
			restN-= 1024*16;
		}
		
		return retValAccu;
	}
	
	float doGetMaxOnMax1024x16(float* dData, int N, int offs, int& id) {
		float  retVal = 0;
		float* tempAdress;
		int*    tempAdress4id;
		cudaGetSymbolAddress((void**) &tempAdress,rtvTemp);
		//	CUDA_SAFE_CALL( cudaGetSymbolAddress((void**) &tempAdress4id,rtvIntTemp) );
		int     nN = N;
		do {
			dim3 block(16);
			dim3 grid(iDivUp(nN,block.x));
			
			//		printf("n %d \n",nN);
			//
			//		printf("block.x %d \n",block.x);
			//
			//		printf("grid.x %d \n",grid.x);
			
			if (N==nN) {
				dev_MaxOnExtern<<< grid, block >>>(dData,nN,offs);
			} else {
				dev_MaxOnIntern<<< grid, block >>>(nN);
			}
			
			nN = grid.x;
		} while (nN!=1);
		
		cudaGetSymbolAddress((void**) &tempAdress,rtv) ;
		cudaGetSymbolAddress((void**) &tempAdress4id,rtvInt) ;
		cudaMemcpy(&retVal, tempAdress, sizeof(float)*1, cudaMemcpyDeviceToHost) ;
		cudaMemcpy(&id, tempAdress4id, sizeof(int)*1, cudaMemcpyDeviceToHost) ;
		return retVal;
	}
	
	float doGetMinOnMax1024x16(float* dData, int N, int offs) {
		float  retVal = 0;
		float* tempAdress;
		cudaGetSymbolAddress((void**) &tempAdress,rtvTemp);
		int     nN = N;
		do {
			dim3 block(16);
			dim3 grid(iDivUp(nN,block.x));
			//
			//		printf("n %d \n",nN);
			//
			//		printf("block.x %d \n",block.x);
			//
			//		printf("grid.x %d \n",grid.x);
			
			if (N==nN) {
				dev_MinOnExtern<<< grid, block >>>(dData,nN,offs);
			} else {
				dev_MinOnIntern<<< grid, block >>>(nN);
			}
			
			nN = grid.x;
		} while (nN!=1);
		
		cudaGetSymbolAddress((void**) &tempAdress,rtv) ;
		cudaMemcpy(&retVal, tempAdress, sizeof(float)*1, cudaMemcpyDeviceToHost) ;
		return retVal;
	}
	
	float doGetMin(float* dData, int N, int offs) {
		float retVal = +1000000000000000000.0f;
		
		if (N<=1024*16) return doGetMinOnMax1024x16(dData,N,offs);
		
		int restN = N;
		
		retVal = doGetMinOnMax1024x16(dData,1024*16,offs);
		offs += 1024*16;
		restN-= 1024*16;
		while (restN>0) {
			retVal = min(doGetMinOnMax1024x16(dData,min(1024*16,restN),offs),retVal);
			//		printf("val = %f\noffs = %d ; restN = %d\n", retVal,offs,restN);
			offs += 1024*16;
			restN-= 1024*16;
		}
		
		return retVal;
	}
	float doGetMax(float* dData, int N, int offs, int* idA) {
		float retVal;
		float lastMax;
		
		int id,id2;
		
		if (N<=1024*16) return doGetMaxOnMax1024x16(dData,N,offs,id);
		
		int restN = N;
		retVal = doGetMaxOnMax1024x16(dData,1024*16,offs,id);
		lastMax=retVal;
		offs += 1024*16;
		restN-= 1024*16;
		while (restN>0) {
			retVal=doGetMaxOnMax1024x16(dData,min(1024*16,restN),offs,id2);
			if (retVal>lastMax) {
				id = id2;
				lastMax = retVal;
			} else {
				;
			}
			//		printf("val = %f\noffs = %d ; restN = %d\n", retVal,offs,restN);
			offs += 1024*16;
			restN-= 1024*16;
		}
		
		(*idA) = id;
		
		return lastMax;
	}
	
	void doGetMaxOfTwoVectors(float* destAndOrgi, float* orig, int N) {
		dim3 dimBlock(128);
		dim3 dimGrid(iDivUp(N,dimBlock.x));
		// Launch the device computation
		//printf("add %f\n",f);
		dMaxOfTwoVectors<<<dimGrid, dimBlock>>>(destAndOrgi, orig, N);
	}
	
	void doGetMinOfTwoVectors(float* destAndOrgi, float* orig, int N) {
		dim3 dimBlock(128);
		dim3 dimGrid(iDivUp(N,dimBlock.x));
		// Launch the device computation
		//printf("add %f\n",f);
		dMinOfTwoVectors<<<dimGrid, dimBlock>>>(destAndOrgi, orig, N);
	}
	
}


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
