/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __CUDA_DISPATCHER_HPP__
#define __CUDA_DISPATCHER_HPP__

#include <boost/preprocessor/enum_params.hpp>
#include <boost/preprocessor/iteration/local.hpp>

#include <boost/fusion/sequence.hpp>
#include <boost/fusion/container.hpp>
#include <boost/fusion/algorithm/iteration/for_each.hpp>
#include <boost/fusion/view/zip_view.hpp>
#include <boost/fusion/include/zip_view.hpp>
#include <boost/fusion/view/zip_view.hpp>
#include <boost/fusion/include/zip_view.hpp>
#include <boost/fusion/include/for_each.hpp>

#include <boost/function_types/parameter_types.hpp>
#include <boost/function_types/function_arity.hpp>

#include <boost/type_traits/remove_reference.hpp>

#include <boost/format.hpp>

#include <iostream>
#include <typeinfo>

// includes, CUDA
#include <cuda_runtime.h>

#define CALL_CUDA_KERNEL( grid, block, fun, ... ) \
  {												  \
	cudaConfigureCall( grid, block );			  \
	push_cuda_args( fun (), __VA_ARGS__ );		  \
	cudaLaunch( BOOST_PP_STRINGIZE( fun ) );	  \
  }
  
/// this is a functor which takes a pair of a value and a target type
/// converts the value to the target type
/// and pushes the converted value on the cuda argument stack
/// it is used by push_cuda_args
struct cuda_args_pusher
{
  size_t offset;
  cuda_args_pusher() : offset(0) {}

  template<typename ArgAndType>
  void operator()( const ArgAndType & arg_and_type ) const
  {
	typedef typename boost::fusion::result_of::at_c<ArgAndType, 0>::type InType;
	typedef typename boost::fusion::result_of::at_c<ArgAndType, 1>::type OutRefType;
	typedef typename boost::remove_reference<OutRefType>::type OutType;

	const InType in_arg = boost::fusion::at_c<0,ArgAndType>( arg_and_type );
	OutType out_arg = in_arg;
	std::cout << " pushing arg from type <" << typeid(InType).name() << "> value: " << in_arg
			  << " to offset " << offset << " type <" << typeid(OutType).name() << "> value: " << out_arg
			  << std::endl;
	cudaSetupArgument(out_arg,offset);
	const_cast<size_t&>(offset) += sizeof(OutType);
  }
};

///
/// this is a combination of black preprocessor and template magic
/// to do typesafe argument passing to cuda kernels
///
/// the template magic 
/// - takes a function type and arguments
/// - iterates the arguments together with the functions argument types
/// - for each argument/type pair each time calls the argument pusher,
///   which converts the arguments to what the function type expects
///
/// the preprocessor magic just generates overloaded versions of this
/// function with one to ten arguments
///
#define BOOST_PP_LOCAL_LIMITS (1, 10)
#define BOOST_PP_LOCAL_MACRO(I)											\
  template< typename F, BOOST_PP_ENUM_PARAMS(I, typename T) >			\
  void push_cuda_args(F, BOOST_PP_ENUM_BINARY_PARAMS(I, T, a) )			\
  {																		\
	typedef typename boost::function_types::function_arity<F> Arity;	\
	const size_t arity = Arity::value;									\
	typedef typename boost::fusion::vector< BOOST_PP_ENUM_PARAMS(I,T) > Args; \
	Args args( BOOST_PP_ENUM_PARAMS(I,a) );								\
	assert( boost::fusion::size(args) == arity );						\
	typedef typename boost::function_types::parameter_types<F> ParameterTypes; \
	ParameterTypes types;												\
	typedef boost::fusion::vector< Args&, ParameterTypes& > ArgsAndTypes; \
	ArgsAndTypes args_and_types( args, types );							\
	boost::fusion::zip_view<ArgsAndTypes> args_and_types_zipped(args_and_types); \
	boost::fusion::for_each( args_and_types_zipped, cuda_args_pusher() ); \
  }
#include BOOST_PP_LOCAL_ITERATE()

#endif
