/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <algorithm>
#include <cuda_runtime_api.h>
#include <cuda_runtime.h>
#include <cuda.h>
#include <cublas.h>
#include "GPUData.h"

#include <cassert>

#include <list>
#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <map>

#include <cstring>

#include <boost/preprocessor/stringize.hpp>

#include <boost/ptr_container/ptr_map.hpp>

#include "cuda_context_handler.hpp"

typedef std::pair<int,int> oentry;
bool order_compare ( const oentry & p1, const oentry & p2) { return p1.first < p2.first; }

struct CudaMemBlock 
{
	void * data;
	size_t size;
	CudaMemBlock( void * d = 0, size_t n = 0 ) : data(d), size(n) {}
};


class CudaMemPool
{
	
	typedef std::map<void*,CudaMemBlock> CudaBlockMap;
	
public:
	
	size_t size() const { return _blocks.size(); }

	size_t usage() const { return _total; }
	
	void * alloc(size_t size) throw(GPUDataException)
	{
		void * data;
		GPUDataException::check_error( str( boost::format( "allocating %d bytes") % size ), cudaMalloc( &data, size ) );
		_blocks.insert( CudaBlockMap::value_type(data, CudaMemBlock( data, size ) ) );
		_total += size;
		_max = std::max( _max, _total );
		double percentage = double(_total) / double(_devprops.totalGlobalMem) * 100.0;
		if ( percentage > 90.0 )
			std::cout << " *** warning ** using > 90% of the device's global memory (" << percentage << "%)" << std::endl;
		//log().info << " allocated " << data << " with " << size << " bytes on context " << cuda_context_handler::get().current() << std::endl;
		return data;
	}
	
	CudaMemBlock find_block( void * address )
	{
		for ( CudaBlockMap::iterator ientry = _blocks.begin(), iend = _blocks.end(); ientry != iend; ++ientry )
			if ( ientry->second.data <= address && ( (size_t) ientry->second.data + ientry->second.size ) > (size_t) address )

				return ientry->second;
		return CudaMemBlock();
	}

	void free( void * data )
	{
		CudaBlockMap::iterator iblock = _blocks.find( data );
		if ( iblock == _blocks.end() )
			{
				std::cerr << str( boost::format( "freeing unmapped block at %x") % data) << std::endl;
			}
		else
			{
				size_t size = iblock->second.size;
				_total -= size;
				_blocks.erase( iblock );
				try
					{
						GPUDataException::check_error( str( boost::format( "freeing cuda memory @ %x " ) % data ), cudaFree((void*) data) );
					} 
				catch ( GPUDataException & e )
					{
						std::cerr << e.what() << std::endl;
					}
			}
	}

	~CudaMemPool()
	{
	}

	size_t total() { return _total; }

	static CudaMemPool & get()
	{
		CudaMemPool * instance = _instances.get();
		if ( instance == 0 )
			_instances.set( instance = new CudaMemPool() );
		return *instance;
	}

private:

	CudaMemPool() : _total(0), _max(0)
	{
#if 0
		GPUDataException::check_error( "initializing driver api", cuInit(0) );
#endif
		GPUDataException::check_error( "getting device index", cudaGetDevice( &_device_index ) );
		GPUDataException::check_error( "getting device info",  cudaGetDeviceProperties( &_devprops, _device_index ) );
		/*
			log().info 
			<< "constructing mem pool, total const mem: " << _devprops.totalConstMem
			<< " total global mem " << _devprops.totalGlobalMem
			<< " context " << cuda_context_handler::get().current()
			<< std::endl << std::flush;
		*/
	}


	size_t _total;
	size_t _max;
	CudaBlockMap _blocks;
	cudaDeviceProp _devprops;
	int _device_index;
	static cuda_context_specific_ptr<CudaMemPool> _instances;
};

cuda_context_specific_ptr<CudaMemPool> CudaMemPool::_instances;

typedef boost::unordered_map<plan_key, cufftHandle*> cufft_plan_map_t;
typedef cuda_context_specific_ptr<cufft_plan_map_t> cufft_plan_maps_t;

bool GPUData::has_plan( const plan_key & key )
{
  size_t val = hash_value(key);
  cufft_plan_map_t::iterator b = getPlanMap().begin(), e = getPlanMap().end();
  for (int i=0; b != e; ++b, ++i)
    if ( val == hash_value(b->first) )
        return true;
  return false;
}

cufft_plan_map_t & GPUData::getPlanMap()
{
	static cufft_plan_maps_t cufft_plan_maps;
	cufft_plan_map_t * instance = cufft_plan_maps.get();
	if ( instance == 0 )
		cufft_plan_maps.set( instance = new cufft_plan_map_t() );
	return * instance;
}

cufftHandle * GPUData::get_fft_plan( plan_key p_key ) const
{
  cufftHandle*  plan;
  cufft_plan_map_t::iterator b = getPlanMap().begin(), e = getPlanMap().end();
  //std::cout << "---- get_fft_plan ------------------------"<<std::endl;
  //for (int i=0; b != e; ++b, ++i)
  //  std::cout << "  -- plan["<<i<<"]: key="<< hash_value(b->first) <<" c="<< getPlanMap().count( b->first ) <<std::endl;
  
	if (getPlanMap().count(p_key)>0) 
  {
        //std::cout << " --- get_fft_plan: recalling plan for key "<<hash_value( p_key )<< " c="<<getPlanMap().count( p_key ) <<std::endl;
		plan = GPUData::getPlanMap()[p_key];
	} 
  else 
  {
        //std::cout << " --- get_fft_plan: creating new plan for key "<<hash_value( p_key) <<std::endl;
		plan = p_key.create_plan();
		GPUData::getPlanMap()[p_key] = plan;
	}
  //std::cout << "---- done get_fft_plan -------------------"<<std::endl;
  return plan;
}

size_t GPUData::cuda_global_mem_usage()
{
	return CudaMemPool::get().usage();
}


GPUData::InnerData::InnerData()
: refcount( 0 )
, data( 0 )
, size( 0 )
, array( 0 )
{
	//	CudaMemPool::retain();
}

GPUData::InnerData::InnerData(size_t n)
: refcount( 1 )
, size( n )
, array( 0 )
{
	//	CudaMemPool::retain();
	data = (float*) CudaMemPool::get().alloc( n );
}

int GPUData::InnerData::retain()
{
	return ++refcount;
}

int GPUData::InnerData::release()
{
	return --refcount;
}

GPUData::InnerData::~InnerData()
{
	if ( data )
		CudaMemPool::get().free( data );
	if ( array )
	{
			cudaFreeArray( array );
			array = 0;
	}
	//	CudaMemPool::release();
}


GPUData::GPUData()
{
	contiguous = true;
    ndim = 0;
    dims = NULL;
    inner = NULL;
    strides = NULL;
    _data = NULL;
	constData = -1;
}

GPUData::GPUData(const GPUData& src)
{
  dims = NULL;
  strides = NULL;
  inner = NULL;
  reference(src);
}

GPUData::GPUData(size_t size) 
{
    dims = NULL;
    inner = NULL;
    strides = NULL;
	init (1,&size);
}
GPUData::GPUData(size_t row, size_t col) 
{
    dims = NULL;
    inner = NULL;
    strides = NULL;
	size_t sizes[2] = {row,col};
	init(2,sizes);
}
GPUData::GPUData(size_t size_3, size_t size_2, size_t size_1) 
{
    dims = NULL;
    inner = NULL;
    strides = NULL;
	size_t sizes[3] = {size_3,size_2,size_1};
	init(3,sizes);
}
GPUData::GPUData(size_t size_4, size_t size_3, size_t size_2, size_t size_1) 
{
    dims = NULL;
    inner = NULL;
    strides = NULL;
	size_t sizes[4] = {size_4,size_3,size_2,size_1};
	init(4,sizes);
}
GPUData::GPUData(size_t size_5, size_t size_4, size_t size_3, size_t size_2, size_t size_1) 
{
    dims = NULL;
    inner = NULL;
    strides = NULL;
	size_t sizes[5] = {size_5,size_4,size_3,size_2,size_1};
	init(5,sizes);
}

GPUData::GPUData( size_t n, const size_t* d ) 
{
    dims = NULL;
    inner = NULL;
    strides = NULL;
	init ( n, d );
}

void GPUData::resize ( size_t d1, size_t d2, size_t d3, size_t d4, size_t d5 )
{
	size_t dims[] = { d1, d2, d3, d4, d5 };
	resize( 5, dims );
}

void GPUData::resize ( size_t d1, size_t d2, size_t d3, size_t d4 )
{
	size_t dims[] = { d1, d2, d3, d4 };
	resize( 4, dims );
}

void GPUData::resize ( size_t d1, size_t d2, size_t d3 )
{
	size_t dims[] = { d1, d2, d3 };
	resize( 3, dims );
}

void GPUData::resize ( size_t d1, size_t d2 )
{
	size_t dims[] = { d1, d2 };
	resize( 2, dims );
}

void GPUData::resize ( size_t d1 )
{
	size_t dims[] = { d1 };
	resize( 1, dims );
}


void GPUData::resize( size_t ndims, const size_t* d )
{
	if ( ndim != ndims || 0 != memcmp( dims, d, ndims * sizeof( size_t ) ) )
		{
			clear();
			init( ndims, d );
		}
}

void GPUData::resize( const GPUData& other )
{
  resize( other.getNoOfDimensions(), other.getDimensions() );
}

void GPUData::resize( size_t ndims, const size_t* d, const std::vector<size_t> & o )
{
	if ( rank() != ndims 
		 || 0 != memcmp( dims, d, ndims * sizeof( size_t ) ) 
		 || o != order() )
		{
			clear();
			init( ndims, d, o );
		}
}


void GPUData::init( size_t ndimA, const size_t* dimsA )
{
	clear();

	if ( ndimA > 0 )
		{

			ndim = ndimA;
			dims = new size_t[ndim];
			std::copy( dimsA, dimsA + ndim, dims);

			// TODO: If 2D alloc with a pitch?
			size_t n = sizeof( *data() );
			for ( int i = 0; i < ndim; ++i ) n *= dimsA[i];
			inner = new InnerData(n);
			_data = inner->data;
			constData = -1;

			// Compute the strides
			strides = new int[ndim];
			strides[ndim - 1] = 1;
			for ( size_t i = ndim - 1; i > 0; --i )
				strides[i - 1] = dims[i] * strides[i];

			contiguous = true;

		}
}

void GPUData::init( size_t ndimA, const size_t* dimsA, const std::vector<size_t> & o )
{
	clear();
	if ( ndimA > 0 )
		{

			ndim = ndimA;
			dims = new size_t[ndim];
			std::copy( dimsA, dimsA + ndim, dims);

			size_t n = 1;
			for ( size_t i = 0; i < ndim; ++i ) n *= dims[i];
			inner = new InnerData(n);
			_data = inner->data;
			constData = -1;

			// Compute the strides
			strides = new int[ndim];
			strides[ o.back() ] = 1;
			for ( size_t i = ndim - 1; i > 0; --i )
				strides[ o[i - 1] ] = dims[ o[i] ] * strides[ o[i] ];

			contiguous = true;
		}
}

GPUData& GPUData::operator=(const GPUData& src)
{
	return reference( src );
}

GPUData& GPUData::reference(const GPUData& src)
{
	if ( this != &src ) // prevent copy to self (which destroys everything)
		{
			clear();
			ndim = src.ndim;
			dims     = new size_t[src.ndim]; std::copy( src.dims   , src.dims    + src.ndim, dims );
			strides  = new int[src.ndim]; std::copy( src.strides, src.strides + src.ndim, strides );
			contiguous = src.contiguous;
			inner = src.inner;
			if ( inner ) inner->retain();
			_data = src.data();
			constData = src.constData;
		}
    return *this;
}

GPUData& GPUData::operator=(const float & val) {

	cudaError_t err;

	if (this->size() != 1)
		throw GPUDataException("Cannot index-cast a non-singleton");

	float v = val;
	err = cudaMemcpy(data(), &v, sizeof(float), cudaMemcpyHostToDevice);
	if (err != cudaSuccess)
		{
			CudaMemBlock block = CudaMemPool::get().find_block( data() );
			ptrdiff_t offset = block.data != 0 ? ((ptrdiff_t)data()) - ((ptrdiff_t)block.data) : 0;
			std::string msg = str( boost::format( "error in operator=%f from: %x to: %x n: %d size: %d block: ( %x(+%d), %d )" ) 
														 %val % ( &v ) % data() % sizeof(float) % ( size() * sizeof(float) ) % block.data % offset % block.size );
			throw GPUDataException(msg, err);
		}

	return *this;
}

void GPUData::clear()
{
    if (inner && inner->release() == 0)
		delete inner;
    if (strides)
		delete [] strides;
    if (dims)
		delete [] dims;

		contiguous = true;
    ndim = 0;
    dims = NULL;
    inner = NULL;
    strides = NULL;
    _data = NULL;
		constData = -1;
}

GPUData::~GPUData()
{
	clear();
}

//
//bool GPUData::operator==(const GPUData& data) {
//	if (ndim!=data.ndim)
//		return false;
//	if (memcmp(dims,data.dims,sizeof(int)*ndim)!=0)
//		return false;
//	GPUData temp(data.copy());
//	temp.sub(*this);
//	float mean = temp.sum()/size();
//	if (mean!=0)
//		return false;
//	if (temp.max()!=0.0f && temp.min()!=0.0f)
//		return false;
//	// TODO: is this sufficient??
//	return true;
//}

GPUData GPUData::operator[](size_t index) const
{
	if ( ndim < 1 ) throw GPUDataException("slicing a singleton array");
	GPUData a(*this);

    a._data = _data + index * strides[0];
    a.ndim--;

	for ( size_t i = 0; i + 1 < ndim; ++i )
		{
			a.dims   [i] = a.dims   [i+1];
			a.strides[i] = a.strides[i+1];
		}

	a.contiguous = a.isContiguous(true);

    return a;
}

GPUData GPUData::range( size_t dim, size_t startindex, size_t endindex, size_t step ) const
{
	if ( startindex > endindex ) std::swap( startindex, endindex );
	if ( endindex >= extent(dim) )
		throw GPUDataException(str( boost::format("range out of range: dim=%d  startindex=%d endindex=%d extent(dim)=%d")
									% dim % startindex % endindex % extent(dim) ));
	GPUData result(*this);
	if ( dim < rank() )
		{
			result._data += startindex * stride(dim);
			result.strides[dim] *= step;
			result.dims[dim] = 1 + ( endindex - startindex ) / step;			
		}
	return result;
}

GPUData GPUData::axisMax( size_t dim1 ) const
{
	GPUData result = slice( dim1, 0 ).copy();
	for ( size_t i = 1; i < extent( dim1 ); ++i )
		result.max( slice( dim1, i ) );
	return result;
}

GPUData GPUData::sumContract( size_t dim1, size_t dim2 ) const
{
	if ( dim1 > dim2 ) std::swap( dim1, dim2 );
	GPUData stacked = reduceDimensionality( dim1, dim2 );
	if ( stacked.rank() > 3 ) throw GPUDataException("sumContract: could not reduce dimensions to three or less");
	size_t dimr = dim1 == 1 ? 0 : ( dim2 == 1 ? 2 : 1 );
	size_t n1 = stacked.extent(dim1), n2 = stacked.extent(dim2), nr = stacked.extent(dimr);
	size_t idims[] = {n1,n2};
	if ( nr <= 1 ) return stacked;
	GPUData s( 2, idims );
	s.copyFrom( stacked.slice( dimr, 0 ), true );
	for ( size_t i = 1; i < nr; ++i )
		{
			s.add( stacked.slice( dimr, i ) );
		}
	return s;
}

GPUData GPUData::maxContract( size_t dim1, size_t dim2 ) const
{
	if ( dim1 > dim2 ) std::swap( dim1, dim2 );
	GPUData stacked = reduceDimensionality( dim1, dim2 );
	if ( stacked.rank() > 3 ) throw GPUDataException("sumContract: could not reduce dimensions to three or less");
	size_t dimr = dim1 == 1 ? 0 : ( dim2 == 1 ? 2 : 1 );
	size_t n1 = stacked.extent(dim1), n2 = stacked.extent(dim2), nr = stacked.extent(dimr);
	size_t idims[] = {n1,n2};
	if ( nr <= 1 ) return stacked;
	GPUData s( 2, idims );
	s.copyFrom( stacked.slice( dimr, 0 ), true );
	for ( size_t i = 1; i < nr; ++i )
		{
			s.max( stacked.slice( dimr, i ) );
		}
	return s;
}

GPUData GPUData::entropyContract( size_t dim1, size_t dim2 ) const
{
	// convert to probability, limiting small values to just above zero
	GPUData tmp( this->copy() );
	tmp.max( 1e-10f );
	tmp.normalize( dim1, dim2 );
	tmp.clamp(0.0000000001f,1.0f);

	GPUData tmp2( tmp.copy() );
	tmp2.log();
	tmp2.mul( tmp );

	return tmp2.sumContract( dim1, dim2 );
}

void GPUData::normalize( size_t dim1, size_t dim2 )
{
	if ( dim1 > dim2 ) std::swap( dim1, dim2 );
	GPUData stacked = reduceDimensionality( dim1, dim2 );
	size_t dimr = dim1 == 1 ? 0 : ( dim2 == 1 ? 2 : 1 );
	size_t nr = stacked.extent(dimr);
	if ( nr <= 1 ) return;
	GPUData s = sumContract( dim1, dim2 );
	s.max( 0.0001 );
	for ( size_t i = 0; i < nr; ++i )
		{
			stacked.slice( dimr, i ).div( s );
		}
}

GPUData GPUData::slice( size_t dim, size_t index ) const
{
	if ( index == 0 && dim >= ndim )
		return *this;
	else if ( index >= extent(dim) || dim >= ndim )
		throw GPUDataException(str( boost::format("slice out of range: dim=%d  index=%d  extent(dim)=%d rank=%d")
																% dim % index % extent(dim) % rank() ));

	// start with an identical reference
	GPUData a(*this);

	// move data pointer to the start of the slice
	a._data += index * stride(dim);

	// reduce dimensionality and move dimensions and strides down
	--a.ndim;
	for ( size_t i = dim + 1; i < ndim; ++i )
		{
			a.dims[i-1]    = extent(i);
			a.strides[i-1] = stride(i);
		}

	// slightly overcautious, contiguous would also be preserved if we sliced the dim with the largest stride
	a.contiguous = contiguous && extent(dim) == 1;

    return a;
}


float GPUData::as_float() 
{

	cudaError_t err;
	float dst;

    if ( size() != 1)
		throw GPUDataException("Cannot index-cast a non-singleton");

	err = cudaMemcpy(&dst, data(), sizeof(dst), cudaMemcpyDeviceToHost);
	if (err != cudaSuccess)
		throw GPUDataException("Could not fetch back GPU memory in GPUData::operator float()", err);

	return dst;
}


int GPUData::index(const int* pos) const
{
    int ans = 0;
    for (size_t i = 0; i < ndim; i++)
		ans += pos[i] * strides[i];
    return ans;
}

GPUData GPUData::view(const size_t* dims) const
{
	return view( 0, dims );
}

GPUData GPUData::view(const int* offset, const size_t* dims) const {

    GPUData a(*this);
	a.contiguous = false;

	std::copy( dims, dims + ndim, a.dims );

	if ( offset )
		a._data += index(offset);

    return a;
}


GPUData GPUData::view(const int* offsetA, size_t ndimA, const size_t* dimsA, const int* stridesA) const
{
    GPUData a(*this);
	if ( a.ndim != ndimA )
		{
			if ( a.dims ) delete[] a.dims;
			if ( a.strides ) delete[] a.strides;
			a.ndim	  = ndimA;
			a.dims    = new size_t[ndimA];
			a.strides = new int[ndimA];
		}
	std::copy( dimsA    , dimsA    + ndimA, a.dims    );
	std::copy( stridesA , stridesA + ndimA, a.strides );
	if ( offsetA )
		a._data += index( offsetA );
	a.contiguous = a.isContiguous(true);
    return a;
}

GPUData GPUData::view(size_t ndimA, const size_t* dimsA, const int* stridesA) const
{
	return view( 0, ndimA, dimsA, stridesA );
}

std::list<int>::iterator operator-(const std::list<int>::iterator& it, int n) {
	std::list<int>::iterator rtv = it;
	for (int i=0; i<n; i++) rtv--;
	return rtv;
}

std::list<int>::iterator operator+(const std::list<int>::iterator& it, int n) {
	std::list<int>::iterator rtv = it;
	for (int i=0; i<n; i++) rtv++;
	return rtv;
}

//GPUData GPUData::shiftDimToRight(size_t in) const {
//	GPUData rtv(*this);
//	for (int i=0; i<ndim; ++i) {
//		rtv.dims[(i+in)%ndim] 	 = dims[i];
//		rtv.strides[(i+in)%ndim] = strides[i];
//	}
//	return rtv;
//}

GPUData GPUData::shiftDim(int in) const
{
	GPUData rtv(*this);
	if (in>=0)
	for ( size_t i=0; i<ndim; ++i )
		{
			rtv.dims[(i+in)%ndim] 	 = dims[i];
			rtv.strides[(i+in)%ndim] = strides[i];
		}
	else
	for ( size_t i=0; i<ndim; ++i )
		{
			rtv.dims[i]    = dims[(i-in)%ndim];
			rtv.strides[i] = strides[(i-in)%ndim];
		}
	rtv.contiguous = rtv.isContiguous(true);
	return rtv;
}

GPUData GPUData::reduceDimensionality(size_t& dim) const
{
	if (ndim==1) return GPUData(*this); // nothing to do

	std::list<int> strideList;//(&(strides[0]),&(strides[ndim]));
	std::list<int> sizeList;//(&(dims[0]),&(dims[ndim]));
	for (size_t i=0; i<ndim; ++i)
		{
			sizeList.push_back(dims[i]);
			strideList.push_back(strides[i]);
		}

	std::list<int>::iterator protectIt = sizeList.begin();
	if (dim>=0 && dim<ndim) {
		protectIt = protectIt + dim;
	}

	std::list<int>::iterator striIt = strideList.end();
	std::list<int>::iterator sizeIt = sizeList.end();
	striIt--;
	sizeIt--;
	do {
		if (	(sizeIt!=protectIt) // the protected dimension should not absorb a higher one
			&&  ((sizeIt-1)!=protectIt)  // and should not be absorbed
			&&  (*striIt) * (*sizeIt) == (*(striIt-1))) {
			(*sizeIt) = (*sizeIt) * (*(sizeIt-1));
			striIt = strideList.erase(striIt-1);
			sizeIt = sizeList.erase(sizeIt-1);
		} else {
			sizeIt--;
			striIt--;
		}
	} while (sizeIt!=sizeList.begin());

	sizeIt = sizeList.begin();
	for (size_t i=0; i<sizeList.size(); ++i) {
		if (sizeIt==protectIt) {
			dim = int(i);
			break;
		}
		sizeIt++;
	}

	const int new_ndim 	= (int) sizeList.size();
	size_t* new_dims    = new size_t[new_ndim];
	int   * new_strides = new    int[new_ndim];

	striIt = strideList.begin();
	sizeIt = sizeList.begin();
	for (int i=0; i<new_ndim; ++i) {
		new_dims[i] 	= (*sizeIt);
		new_strides[i] 	= (*striIt);
		sizeIt++;
		striIt++;
	}

	int* new_offs = new int[ndim];
	for (size_t i=0; i<ndim; ++i) new_offs[i] = 0;

	GPUData rtv = view(new_offs,new_ndim,new_dims,new_strides);

	delete [] new_offs;
	delete [] new_dims;
	delete [] new_strides;

	return rtv;
}

GPUData GPUData::removeSingletonDimensions() const
{
	for ( size_t i = 0; i < ndim; ++i )
		{
			if(extent(i) == 1)
			{
				return slice( i, 0 ).removeSingletonDimensions();
			}
		}

	return *this;
}

GPUData GPUData::reduceDimensionality( size_t& dim1, size_t& dim2 ) const
{
	if (ndim==1) return GPUData(*this); // nothing to do

	std::list<int> strideList;//(&(strides[0]),&(strides[ndim]));
	std::list<int> sizeList;//(&(dims[0]),&(dims[ndim]));
	for (size_t i=0; i<ndim; ++i) {
		sizeList.push_back(dims[i]);
		strideList.push_back(strides[i]);
	}

	std::list<int>::iterator protectIt1 = sizeList.begin();
	if (dim1>0 && dim1<=ndim) {
		protectIt1 = protectIt1 + dim1;
	}
	std::list<int>::iterator protectIt2 = sizeList.begin();
	if (dim2>0 && dim2<=ndim) {
		protectIt2 = protectIt2 + dim2;
	}

	std::list<int>::iterator striIt = strideList.end();
	std::list<int>::iterator sizeIt = sizeList.end();
	striIt--;
	sizeIt--;
	do {
		if (	(sizeIt!=protectIt1 && sizeIt!=protectIt2) // the protected dimension should not absorb a higher one
			&&  ((sizeIt-1)!=protectIt1 && (sizeIt-1)!=protectIt2)  // and should not be absorbed
			&&  (*striIt) * (*sizeIt) == (*(striIt-1))) {
			(*sizeIt) = (*sizeIt) * (*(sizeIt-1));
			striIt = strideList.erase(striIt-1);
			sizeIt = sizeList.erase(sizeIt-1);
		} else {
			sizeIt--;
			striIt--;
		}
	} while (sizeIt!=sizeList.begin());

	sizeIt = sizeList.begin();
	for (size_t i=0; i<sizeList.size(); ++i) {
		if (sizeIt==protectIt1) {
			dim1 = int(i);
			break;
		}
		sizeIt++;
	}
	sizeIt = sizeList.begin();
	for (size_t i=0; i<sizeList.size(); ++i) {
		if (sizeIt==protectIt2) {
			dim2 = (int)i;
			break;
		}
		sizeIt++;
	}

	const size_t new_ndim 	= sizeList.size();
	size_t* new_dims	= new size_t[new_ndim];
	int*    new_strides = new int[new_ndim];

	striIt = strideList.begin();
	sizeIt = sizeList.begin();
	for ( size_t i = 0; i < new_ndim; ++i )
		{
			new_dims[i] 	= (*sizeIt);
			new_strides[i] 	= (*striIt);
			sizeIt++;
			striIt++;
		}

	int* new_offs = new int[ndim];
	for ( size_t i = 0; i < ndim; ++i )
		new_offs[i] = 0;

	GPUData rtv = view( new_offs, new_ndim, new_dims, new_strides );

	delete[] new_offs;
	delete[] new_dims;
	delete[] new_strides;

	return rtv;
}

GPUData GPUData::insertDimension( size_t index ) const
{
	GPUData a(*this);
	size_t new_index = std::min(rank(),index+1);
	delete[] a.dims;
	delete[] a.strides;
	++a.ndim;
	a.dims = new size_t[a.ndim];
	a.strides = new int[a.ndim];
	std::copy( dims   , dims    + new_index, a.dims    );
	std::copy( strides, strides + new_index, a.strides );
	std::copy( dims    + new_index,dims     + ndim, a.dims    + new_index + 1 );
	std::copy( strides + new_index, strides + ndim, a.strides + new_index + 1 );
	a.dims[new_index] = 1;
	a.strides[new_index] = 1; // doesn't really matter as it's singleton, should be maxstride * dim_of_maxstride
	if ( new_index != index+1 )
		a.reference( a.insertDimension(index) );
	return a;
}

GPUData GPUData::extendDimensionLeft() const
{
	const size_t n = rank() + 1;
	size_t* dims_new = new size_t[n];
	int*    stri_new = new int[n];
	int*    offs     = new int[n-1];

	dims_new[0] = 1;
	stri_new[0] = stride(0) * extent(0);

	std::copy( dims   , dims    + ndim, dims_new + 1 );
	std::copy( strides, strides + ndim, stri_new + 1 );
	std::fill( offs, offs + ndim, 0 );

	GPUData rtv( view( offs, n, dims_new, stri_new ) );

	delete[] dims_new;
	delete[] stri_new;
	delete[] offs;

	return rtv;
}

GPUData GPUData::swapDimensions(size_t dim1, size_t dim2) const {
	if (dim1>=ndim || dim2>=ndim)
		throw GPUDataException("dimension indiex out of range");
	if (dim1==dim2)
		return *this;
	GPUData rtv(*this);
	rtv.dims[dim1]   = dims[dim2];
	rtv.dims[dim2]   = dims[dim1];
	rtv.strides[dim1] = strides[dim2];
	rtv.strides[dim2] = strides[dim1];
	return rtv;
}

GPUData GPUData::reduceDimensionality() const {
	if (ndim==1) return GPUData(*this); // nothing to do

	std::list<int> strideList;//(&(strides[0]),&(strides[ndim]));
	std::list<int> sizeList;//(&(dims[0]),&(dims[ndim]));
	for (size_t i=0; i<ndim; ++i)
		{
			sizeList.push_back(dims[i]);
			strideList.push_back(strides[i]);
		}

	std::list<int>::iterator striIt = strideList.end();
	std::list<int>::iterator sizeIt = sizeList.end();
	striIt--;
	sizeIt--;
	do {
		if ((*striIt) * (*sizeIt) == (*(striIt-1))) {
			(*sizeIt) = (*sizeIt) * (*(sizeIt-1));
			striIt = strideList.erase(striIt-1);
			sizeIt = sizeList.erase(sizeIt-1);
		} else {
			sizeIt--;
			striIt--;
		}
	} while (sizeIt!=sizeList.begin());

	const size_t new_ndim 	= (int) sizeList.size();
	size_t* new_dims 	    = new size_t[new_ndim];
	int* new_strides	    = new int[new_ndim];

	striIt = strideList.begin();
	sizeIt = sizeList.begin();
	for ( size_t i=0; i<new_ndim; ++i, sizeIt++, striIt++ )
		{
			new_dims[i] 	= (*sizeIt);
			new_strides[i] 	= (*striIt);
		}

	int* new_offs 		= new int[ndim];
	for ( size_t i = 0; i < ndim; ++i ) new_offs[i] = 0;

	GPUData rtv = view(new_offs,new_ndim,new_dims,new_strides);

	delete [] new_offs;
	delete [] new_dims;
	delete [] new_strides;

	return rtv;
//	printf("no of dimensions = %d\nlist size = %d\n",ndim,strideList.size());
//	for (std::list<int>::iterator i=strideList.begin(); i!=strideList.end(); i++)
//		printf("list = %d\n",(*i));
}

//union memSetTyp_u {
//	float f;
//	float* fPtr;
//	size_t i;
//	CUdeviceptr dPtr;
//};

extern "C" void doFillSpecial4D(const int dix , const int dsx,
				const int diy , const int dsy,
				const int diu , const int dsu,
				const int div , const int dsv,  float* dData,
				const float val);

void GPUData::fill(const float val)
{
	cudaError_t err;

	if (isContiguous() && val==0.0f)
		{
			err = cudaMemset( data(), 0, size() * sizeof(float) );
		}
	else if (ndim<=4)
		{
			int dsx = dims[ndim-1];
			int dsy = 1;
			int dsu = 1;
			int dsv = 1;
			int dix = strides[ndim-1];
			int diy = 1;
			int diu = 1;
			int div = 1;
			switch (ndim) {
			case 4 :
				dsy = dims[2]; diy = strides[2];
				dsu = dims[1]; diu = strides[1];
				dsv = dims[0]; div = strides[0];
				break;
			case 3 :
				dsy = dims[1]; diy = strides[1];
				dsu = dims[0]; diu = strides[0];
				break;
			case 2 :
				dsy = dims[0]; diy = strides[0];
				break;
			case 1 :
				break;
			}
			doFillSpecial4D(dix , dsx,
							diy , dsy,
							diu , dsu,
							div , dsv,
							data(), val);
		}
	/* this does not seem like a very good idea... daniel
	else if (isContiguous())
		{
			const size_t sizeL = size();
			float* hData;
			err = cudaMallocHost((void**)&hData,sizeL*sizeof(float));
			for ( size_t i = 0; i < sizeL; ++i )
				{
					hData[i]=val;
				}
			this->copyFromHostMemory(hData);
			err = cudaFreeHost(hData);
			// TODO: check for error code!

		}
	*/
	else
		{

			// TODO: recurse

			int* idx = new int[ndim];
			size_t  j;
			std::fill( idx, idx + ndim, 0 );

			do
				{
					ptrdiff_t offset = index(idx);

					err = cudaMemcpy( data() + offset, &val, sizeof(*data()), cudaMemcpyHostToDevice );

					if (err != cudaSuccess)
						throw GPUDataException("Could not write back GPU memory in GPUData::fill()", err);

					for ( j = 0; j < ndim && ++idx[j] >= int(dims[j]); ++j )
						{
							idx[j] = 0;
						}
				}
			while( j < ndim );

			delete[] idx;
		}
}


extern "C" bool doAddRepeatedSimple(float* a, float* b, int Na, int Nb, const float f);
void GPUData::add_repeated(const GPUData& values, const float f) {
	const int size = this->size();
	if (this->isContiguous() && values.isContiguous()) {
		if (size%values.size()!=0) throw GPUDataException("add_repeated only possible if divisor of both sizes is a natural number");
		if (!doAddRepeatedSimple(data(), values.data(), size, values.size(),f))
			throw GPUDataException("error in kernel of doAddRepeatedSimple()");;
	} else throw GPUDataNYIException("add_repeated of non contiguous data not yet implemented");
}


extern "C" bool doMulRepeatedSimple(float* a, float* b, int Na, int Nb);
void GPUData::mulRepeated(const GPUData& values) {
	const int size = this->size();
	if (this->isContiguous() && values.isContiguous()) {
		if (size%values.size()!=0) throw GPUDataException("mulRepeated only possible if divisor of both sizes is a natural number");
		if (!doMulRepeatedSimple(data(), values.data(), size, values.size()))
			throw GPUDataException("error in kernel of doMulRepeatedSimple()");;
	} else throw GPUDataNYIException("mulRepeated of non contiguous data not yet implemented");
}

float GPUData::absSum() const
{
	if (!isContiguous() && ndim!=1) {
		GPUData own = reduceDimensionality();
		if (own.getDimensionsCount()!=1)
			throw GPUDataNYIException("absMin on non contiguous data not yet implemented");
		return own.absSum();
	}

	float rtv = cublasSasum (size(), data(), strides[ndim-1]);
	cublasStatus status = cublasGetError();
	if (status != CUBLAS_STATUS_SUCCESS) {
        fprintf (stderr, "!!!! error during cublasSasum in function absSum(..) \n");
	}

	return rtv;
}

extern "C" void doSumOverLastDimension(float* origin, const int n_x, const int n_y, float* dest);

void GPUData::sumOverDeepestDimension(GPUData& dest) const
{
	if (!isContiguous())
		throw GPUDataNYIException("sumOverDeepestDimension on non contiguous data not yet implemented");

	const size_t n_x = dims[ndim-1];
	const size_t new_size = size()/n_x;
	if (dest.size()<new_size)
		throw GPUDataException("GPUData::sumOverDeepestDimension(): sizes does not fit");

	doSumOverLastDimension(data(), n_x, new_size, dest.data());
}


// impelemted in heavyside.cu
extern "C" void doHeaviside(float* data, float f, int N);

void GPUData::heavisideStepFunction(const float f)
{
	if (!isContiguous()) throw GPUDataNYIException("heavisideStepFunction on non contiguous data not yet implemented");
	doHeaviside(data(),f,size());
}

// impelemted in heavyside.cu
extern "C" void doHeaviside2(float* data, float f, float centerValue, int N);

void GPUData::heavisideStepFunction(const float f, const float centerValue)
{
	if (!isContiguous()) throw GPUDataNYIException("heavisideStepFunction on non contiguous data not yet implemented");
	doHeaviside2(data(),f,centerValue,size());
}

// impelemted in heavyside.cu
extern "C" void doSetIfBetween(float* data, const float min, const float max, const float val, int N);
extern "C" void doSetAnotherIfBetween(float* data,float* res, const float min, const float max, const float val, int N);
void GPUData::setIfBetween(const float min, const float max, const float val) {
	if (!isContiguous()) throw GPUDataNYIException("setIfBetween on non contiguous data not yet implemented");
	doSetIfBetween(data(),min,max,val,size());
}
void GPUData::setIfBetween(const float min, const float max, const float val, GPUData& res) const {
	size_t sizeL = size();
	if (sizeL!=res.size()) throw GPUDataException("setIfBetween not defined for data with different sizes");
	if (!isContiguous()) throw GPUDataNYIException("setIfBetween on non contiguous data not yet implemented");
	doSetAnotherIfBetween(data(),res.data(),min,max,val,size());
}

float GPUData::absMax() const {
	if (!isContiguous() && ndim!=1)
		throw GPUDataNYIException("absMax on non contiguous data not yet implemented");
	int rtv = cublasIsamax ( size(), data(), 1 ) -1;
	cublasStatus status = cublasGetError();
	if (status != CUBLAS_STATUS_SUCCESS) {
        fprintf (stderr, "!!!! error during cublasIsamax in function sumOfAbs(..) \n");
	}

	float rtv2 = 0;
	cudaError_t err = cudaMemcpy(&rtv2,data()+rtv, sizeof(*data()), cudaMemcpyDeviceToHost);
	if (err != cudaSuccess)
		throw GPUDataException("Could fetch back GPU memory in GPUData::absMax()", err);

	return fabs(rtv2);
}

float GPUData::absMin() const {
	if (!isContiguous() && ndim!=1)
		throw GPUDataNYIException("absMin on non contiguous data not yet implemented");
	int rtv = cublasIsamin ( size(), data(), 1 ) -1;
	cublasStatus status = cublasGetError();
	if (status != CUBLAS_STATUS_SUCCESS) {
        fprintf (stderr, "!!!! error during cublasIsamin in function sumOfAbs(..) \n");
	}

	float rtv2 = 0;
	cudaError_t err = cudaMemcpy(&rtv2,data()+rtv, sizeof(*data()), cudaMemcpyDeviceToHost);
	if (err != cudaSuccess)
		throw GPUDataException("Could fetch back GPU memory within GPUData::absMin()", err);

	return fabs(rtv2);
}

extern "C" void doClampSimple(float* a, int N, float min, float max);

void GPUData::clamp(float min_val, float max_val)
{

  if (isContiguous())
    doClampSimple(data(),size(), min_val, max_val);
  else
      {
        max( max_val );
        min( min_val );
      }//throw GPUDataNYIException("clamp on on noncontigous not yet implemented");
}

extern "C" void do_atan2_simple(float* res, const float* y, const float* x, int N);

void GPUData::atan2(const GPUData& y, const GPUData& x) {
	const size_t sizeL = this->size();
	if (y.size()!=sizeL || sizeL!=x.size()) {
		throw GPUDataException("atan2 not definde for input with different sizes");
	}
	if (!isContiguous() || !y.isContiguous() || !x.isContiguous()) {
		throw GPUDataNYIException("atan2 not jet implemented for non contiguous data");
	}
	do_atan2_simple(data(),y.data(),x.data(),sizeL);
}

//extern "C" void do_medianOnBinary(float* res, const float* src, int range, int sizeL);
extern "C" void do_medianOnBinary2D(float* res, const float* src, int range_x, int range_y, int size_x, int size_y);

void GPUData::medianOnBinary(const GPUData& src, int range) {
	const size_t sizeL = this->size();
	if (range<1 || range>10) {
		throw GPUDataException("medianOnBinary only works in limited ranges: 1-10 ");
	}
	if (src.size()!=sizeL ) {
		throw GPUDataException("medianOnBinary not defined for input with different sizes");
	}
	if (src.ndim!=2) {
		throw GPUDataException("medianOnBinary not defined for n!=2 dim input");
	}
	if (!isContiguous() || !src.isContiguous() ) {
		throw GPUDataNYIException("medianOnBinary not yet implemented for non-contiguous data");
	}
	do_medianOnBinary2D(data(), src.data(), range,range,(int)extent(1),(int)extent(0));
}


extern "C" void do_median3x3 (float *res, const float *src, const int width, const int height);
void GPUData:: 
median3x3 (const GPUData &src)
{
	if (this->size() != src.size()) {
		throw GPUDataException("median3x3 not defined for input with different sizes");
	}
	if (src.ndim!=2) {
		throw GPUDataException("median3x3 not defined for n!=2 dim input");
	}
	if (!isContiguous() || !src.isContiguous()) {
		throw GPUDataNYIException("median3x3 not implemented for non-contiguous data");
	}
	do_median3x3(data(), src.data(), (int)extent(1), (int)extent(0));
}


// implemented in pythagoras.cu
extern "C" void do_pyth_simple(float* res, const float* a, const float* b, int N);
extern "C" void do_pyth_1D(float* res, const float* a, int inca, const float* b, int incb, int N);

void GPUData::pythagoras(const GPUData& a, const GPUData& b) {
	const size_t sizeL = this->size();
	if (a.size()!=sizeL || sizeL!=b.size()) {
		throw GPUDataException("pythagoras not defined for input with different sizes");
	}
	if (!isContiguous()) {
		throw GPUDataNYIException("pythagoras not jet implemented for non contiguous data");
	}
	if (!a.isContiguous() || !b.isContiguous()) {
		if (a.getDimensionsCount()!=1 || b.getDimensionsCount()!=1) {
			GPUData a_new(a.reduceDimensionality());
			GPUData b_new(b.reduceDimensionality());
			if (a_new.getDimensionsCount()!=1 || b_new.getDimensionsCount()!=1) {
				throw GPUDataNYIException("pythagoras not jet implemented for non contiguous data");
			}
			int inca = a_new.strides[0];
			int incb = b_new.strides[0];
			do_pyth_1D(data(),a.data(),inca,b.data(),incb,sizeL);
		} else {
			int inca = a.strides[0];
			int incb = b.strides[0];
			do_pyth_1D(data(),a.data(),inca,b.data(),incb,sizeL);
		}
	} else {
		do_pyth_simple(data(),a.data(),b.data(),sizeL);
	}
}

extern "C" void do_nakarushton2PreservingSign(const int N, float* res, float* input,  const float eps, const float maxim);

void GPUData::nakarushton2(const float n50, const float resultFaktor) {
	nakarushton2( *this, n50, resultFaktor );
}

void GPUData::nakarushton2(GPUData& dest, const float n50, const float resultFaktor) const {
	const size_t sizeL = size();
	if (sizeL!=dest.size()) {
		throw GPUDataException("sizes of parameter and result does not fit in nakarushton2(..)");
	}
	if (!isContiguous() || !dest.isContiguous()) {
		throw GPUDataNYIException("nakarushton2 not jet implemented for non contiguous data");
	}
	do_nakarushton2PreservingSign(sizeL,dest.data(),data(),n50,resultFaktor);
	GPUDataException::check_error(str(boost::format("GPUData::nakarushton2(%s,%s,%f,%f")%dump_dims()%dest.dump_dims()%n50%resultFaktor));
}


bool GPUData::isContiguous(bool check) const
{
	if ( check )
		{
			// sort strides and dims by strides
			std::vector< oentry > order(getNoOfDimensions());
			for ( size_t i = 0; i < order.size(); ++i )
				{
					order[i].first = stride(i);
					order[i].second = extent(i);
				}
			// we use lexicographic sort, because dims with smaller extents also must come first
			std::sort( order.begin(), order.end() );
			int estride = 1;
			BOOST_FOREACH( const oentry & oe, order )
				{
					if ( estride != oe.first && oe.second != 1 )
						{
							return false;
						}
					estride *= oe.second;
				}
			return true;
		}
	return contiguous;
}

/// row major means with descending strides
bool GPUData::isRowMajor() const
{
	int ndims = getNoOfDimensions();
	for ( int i = 1; i < ndims; ++i )
		if ( stride(i) > stride(i-1) )
			return false;
	return true;
}

/// row major means with descending strides
bool GPUData::isColumnMajor() const
{
	int ndims = getNoOfDimensions();
	for ( int i = 1; i < ndims; ++i )
		if ( stride(i) < stride(i-1) )
			return false;
	return true;
}

void GPUData::makeContiguous( bool rowMajor )
{
	if ( !isContiguous(true) || ( rowMajor && !isRowMajor() ) )
		{
			GPUData tmp( getNoOfDimensions(), getDimensions() );
			tmp.copyFrom( *this );
			reference( tmp );
		}
}

// suggests a dimension for recursion of algorithms that cannot run on the given number of dimensions
size_t GPUData::suggest_recursion_dim(size_t exclude1, size_t exclude2) const
{
	// find the smalles extent by sorting
	int smallest_extent = -1;
	int smallest_dim = -1;
	for ( size_t i = 0; i < getNoOfDimensions(); ++i )
		{
			if ( exclude1 == i || exclude2 == i ) continue;
			if ( smallest_extent >= 0 || extent(i) < size_t(smallest_extent) )
				{
					smallest_extent = extent(i);
					smallest_dim = i;
				}
		}
	return smallest_dim;
}

extern "C" void doCopyFromSpecial4D( const int tgt_stride_x , const int nx,
									 const int tgt_stride_y , const int ny,
									 const int tgt_stride_u , const int nu,
									 const int tgt_stride_v , const int nv,
									 float* tgt_data,
									 const int src_stride_x,
									 const int src_stride_y,
									 const int src_stride_u,
									 const int src_stride_v,
									 float* src_data );

void GPUData::copyFrom( const GPUData& origin, bool may_resize )
{

	if ( ( empty() && !origin.empty() ) && !may_resize )
		{
      throw std::runtime_error("GPUData::copyFrom() this instance is empty and may_resize is false, thus copyFrom is not possible.");
			//clear();
			//return;
		}


	size_t ndims = getNoOfDimensions();
	size_t ondims = origin.getNoOfDimensions();

	// fast path if both arrays are contiguous and have the same memory layout
	if ( isContiguous() && origin.isContiguous()
		 && rank() == origin.rank()
		 && memcmp( getSizes(), origin.getSizes(), sizeof(size_t) * rank() ) == 0
		 && memcmp( getStrides(), origin.getStrides(), sizeof(int) * rank() ) == 0 )
		{
			GPUDataException::check_error( "Could not copy GPU memory",
										   cudaMemcpy( data(), origin.data(), size() * sizeof(float), cudaMemcpyDeviceToDevice)
										   );
			return;
		}

	// only check now to support direct contiguous to contiguous copy
	if ( ndims != ondims || 0 != memcmp(dims, origin.dims, ndims * sizeof(*dims) ) )
		{
			if ( may_resize )
				resize( origin.getNoOfDimensions(), origin.getDimensions() );
			else
					throw GPUDataException( str( boost::format("copyFrom sizes don't match and resize not allowed: origin=%s, dest=%s") %origin.dump_dims() %dump_dims() ) );
		}

	if ( origin.ndim <=4 && 0 == memcmp(dims, origin.dims, ndims * sizeof(*dims) ))
		{
			// generate an order that puts the largest extents up front
#if 1
			std::vector< oentry > order(4);
			for ( int i = 0; i < 4; ++i )
				{ order[i].first = extent(i); order[i].second = i; }
			std::sort( order.begin(), order.end(), order_compare );
			// std::sort sorts ascending, so we put the last elements first
	   		doCopyFromSpecial4D( stride(order[3].second), extent(order[3].second),
								 stride(order[2].second), extent(order[2].second),
								 stride(order[0].second), extent(order[0].second),
								 stride(order[1].second), extent(order[1].second),
								 data(),
								 origin.stride(order[3].second),
								 origin.stride(order[2].second),
								 origin.stride(order[0].second),
								 origin.stride(order[1].second),
								 origin.data()
								 );
#else
			doCopyFromSpecial4D( stride(3), extent(3),
								 stride(2), extent(2),
								 stride(1), extent(1),
								 stride(0), extent(0),
								 data(),
								 origin.stride(3),
								 origin.stride(2),
								 origin.stride(1),
								 origin.stride(0),
								 origin.data()
								 );
#endif
		}
	else //recursive
		{
			int dim_recurse = suggest_recursion_dim();
			for ( size_t i = 0; i < extent(dim_recurse); ++i )
				{
					slice( dim_recurse, i ).copyFrom( origin.slice( dim_recurse, i ) );
				}
		}

	GPUDataException::check_error( "at end of GPUData::copyFrom(GPUData &)" );
}

void GPUData::reshape( size_t ndim, const size_t * dims, const int * strides  )
{

	size_t n = ndim > 0 ? 1 : 0;
	for ( size_t i = 0; i < ndim; ++i )
		n *= dims[i];
	if ( n != size() )
		throw GPUDataException("reshape to different number of elements");

	size_t * newdims = new size_t[ndim];
	int    * newstrides = new int[ndim];

	if ( GPUDataBase::dims ) delete[] GPUDataBase::dims;
	if ( GPUDataBase::strides ) delete[] GPUDataBase::strides;
	
	std::copy( dims   , dims    + ndim, newdims    );
	std::copy( strides, strides + ndim, newstrides );

	GPUDataBase::dims = newdims;
	GPUDataBase::strides = newstrides;
	GPUDataBase::ndim = ndim;

}

struct compare_by_stride_lookup
{
	compare_by_stride_lookup( const int * m ) : lut(m) {}
	bool operator() ( size_t i1, size_t i2 ) const
	{
		return lut[i1] > lut[i2];
	}
	const int * lut;
};

std::vector<size_t> GPUData::order()
{
	std::vector<size_t> entries;
	for ( size_t i = 0; i < rank(); ++i )
		entries.push_back( i );
	std::sort( entries.begin(), entries.end(), compare_by_stride_lookup( strides ) );
	return entries;
}

GPUData GPUData::makeAnother() const
{
	GPUData rtv( size() );
	rtv.reshape( ndim, dims, strides );
	rtv.copyFrom(*this);
	return rtv;
}

GPUData GPUData::copy( bool keepshape ) const
{
	GPUData rtv;

	if ( isContiguous() && keepshape )
		rtv.reference( makeAnother() );
	else
		rtv.resize( ndim, dims );

	rtv.copyFrom(*this);
	return rtv;
}


void GPUData::copyFrom( const CPUData<float> & cdata, bool may_resize )
{
	if ( cdata.empty() )
		{
			clear();
			return;
		}


	// respect column shape information if available
	const std::vector<size_t> & column_shape = cdata.column_shape();
    size_t ndim = cdata.depth() > 1 ? ( column_shape.size() ? 2 + column_shape.size() : 3 ) : 2;
#ifdef _WIN32
	size_t* dims = new size_t[ndim]; // fix to compile under VCC2005
#else
	size_t dims[ndim];
#endif // ifdef _WIN32
	dims[0] = cdata.rows();
	dims[1] = cdata.columns();
	if ( ndim > 2 )
		{
			if ( column_shape.size() )
				std::copy( column_shape.begin(), column_shape.end(), dims + 2 );
			else
				dims[2] = cdata.depth();
		}

	if ( ! isContiguous() || getDimensionsCount() != ndim || 0 != memcmp( getDimensions(), dims, ndim * sizeof(size_t) ) )
		{
			if ( may_resize )
				{
					clear();
					init( ndim, dims );
				}
			else
				throw GPUDataException("copyFrom: size mismatch and resize not allowed");
		}
	if ( cdata.contiguous() )
		copyFromHostMemory( cdata.data() );
	else
		{
			// rows are contiguous, so we copy a row at a time
			float * datap = data();
			size_t n_row = cdata.depth() * cdata.columns();
			for ( size_t row = 0; row < cdata.rows(); ++row )
				{
					GPUDataException::check_error( "copying from host",
												   cudaMemcpy( datap, &cdata(row,0,0),
															   n_row * sizeof(float),
															   cudaMemcpyHostToDevice )
												   );
					datap += n_row;
				}
		}
#ifdef _WIN32
	delete [] dims;
#endif
}

// implemented in sincos.cu
extern "C" void do_sin_simple(float* data, int N);
extern "C" void do_sin_simple2(float* data, const float* param, int N);

void GPUData::sin(const GPUData& a) {
	const size_t sizeL = this->size();
	if (sizeL!=a.size()) throw GPUDataException("sizes of parameter and result does not fit  in sin(..)");
	if (!isContiguous() || !a.isContiguous()) {
		throw GPUDataNYIException("sin not jet implemented for non contiguous data");
	}
	do_sin_simple2(data(), a.data() ,this->size());
}

// implemented in sincos.cu
extern "C" void do_cos_simple(float* data, int N);
extern "C" void do_cos_simple2(float* data, const float* param, int N);

void GPUData::cos(const GPUData& a) {
	const size_t sizeL = this->size();
	if (sizeL!=a.size()) throw GPUDataException("sizes of parameter and result does not fit  in cos(..)");
	if (!isContiguous() || !a.isContiguous()) {
		throw GPUDataNYIException("cos not jet implemented for non contiguous data");
	}
	do_cos_simple2(data(),a.data(),this->size());
}



void GPUData::copyFrom( CPUDataLayer<float> & cdatalayer, bool may_resize )
{
	throw std::runtime_error("GPUData::copyFrom( CPUDataLayer ) not implemented");
}

void GPUData::copyTo( CPUData<float> & cdata,  bool may_resize ) const
{
	if ( empty() )
		{
			if ( may_resize )
				cdata.clear();
			else if ( !cdata.empty() )
				throw GPUDataException("copyTo data empty, target not empty, and resize not allowed");
			return;
		}

	int rank = getNoOfDimensions();
	size_t rows = extent(0), columns = extent(1), depth = extent(2);
	for ( int dim = 3; dim < rank; ++dim )
		depth *= extent(dim);
	if ( may_resize )
		cdata.resize( rows, columns, depth );
	else if ( cdata.columns() != rows || cdata.rows() != rows || cdata.depth() != depth )
		throw GPUDataException("copyTo data size mismatch and resize not allowed");

	if ( !isContiguous() || !isRowMajor()  )
		{
			// make a flat copy and transfer that
			// this is basically an impedance mismatch
			// in that GPUData is more flexible than CPUData currently
			GPUData tmp;
			tmp.copyFrom( *this, true );
			tmp.copyTo( cdata, may_resize );
			return;
		}

	// preserve column_shape information
	std::vector<size_t> column_shape( std::max( 0, rank - 2 ) );
	if ( rank > 2 )
		std::copy( dims + 2, dims + rank, column_shape.begin() );
	cdata.set_column_shape( column_shape );

	if ( cdata.contiguous() )
		copyToHostMemory( cdata.data() );
	else
		{
			float * gdatap = data();
			size_t n_row = cdata.depth() * cdata.columns();
			for ( size_t row = 0; row < cdata.rows(); ++row )
				{
					float * startp = &cdata( row, 0, 0 );
					GPUDataException::check_error( "copying to host",
												   cudaMemcpy( startp, gdatap,
															   n_row * sizeof(float),
															   cudaMemcpyDeviceToHost )
												   );
					gdatap += n_row;
				}
		}
}

#include "GPUDataDebugging.hpp"

void GPUData::copyFromHostMemory(const float* host_data)
{

	if ( empty() ) return;

	cudaError_t err;

	if (! isContiguous ())
		{
			dump_info( *this );
			throw GPUDataNYIException("Copy to non contiguous not yet implemented");
		}

	GPUDataException::check_error("before copyFromHost");

	size_t n = this->size() * sizeof(float);
	err = cudaMemcpy( data(), host_data, n, cudaMemcpyHostToDevice );

	if (err != cudaSuccess)
		{
			CudaMemBlock block = CudaMemPool::get().find_block( data() );
			ptrdiff_t offset = block.data != 0 ? ((ptrdiff_t)data()) - ((ptrdiff_t)block.data) : 0;
			std::string msg = str( boost::format( "error in copyFromHostMemory from: %x to: %x n: %d size: %d block: ( %x(+%d), %d )" ) 
														 % host_data % data() % n % ( size() * sizeof(float) ) % block.data % offset % block.size );
			throw GPUDataException( msg, err );
		}

}

void GPUData::copyToHostMemory(float* host_data) const {

	
	cudaError_t err;

	if (! isContiguous ())
		{
			dump_info( *this );
			throw GPUDataNYIException("Copy from non contiguous not yet implemented");
		}
	
	if ( is_array() )
	{
		GPUData tmp( *this );
		tmp.move_to_data();
		err = cudaMemcpy(host_data, tmp.data(), tmp.size() * sizeof(*host_data), cudaMemcpyDeviceToHost);
		if (err != cudaSuccess)
			throw GPUDataException("Could not fetch back from GPU memory (array)", err);
	}
	else
	{
		err = cudaMemcpy(host_data, data(), this->size() * sizeof(*host_data), cudaMemcpyDeviceToHost);
		if (err != cudaSuccess)
			throw GPUDataException(str( boost::format("Could not fetch back from GPU memory, for %s")%dump_dims() ), err);
	}
}

extern "C" int copyDeviceDataToConstMemory(const float* dData, int N);

int GPUData::copyToConstMemory( bool overwrite )
{
	if ( ! isContiguous() )
		{
			throw GPUDataNYIException("copyToConstMemory for non contiguouse memory not yet implemented!");
		}
	if (constData==-1 || overwrite)
		{
			constData = copyDeviceDataToConstMemory(data(),size());
		}
	return constData;
}

extern "C" void convolute1DInContiguous4D(
	float *d_Data, int kernelIdx, int rk, float *d_Result,
	int incX, int sizeX,
	int incY, int sizeY,
	int incU, int sizeU,
	int incV, int sizeV,
	int type);

extern "C" void doConvoluteDeepestOfTwoDims(float* data, float* dest, int N, int otherDimSize, int kernel, int kRadius);
extern "C" void doConvoluteFirstOfTwoDims(float* data, float* dest, int N, int otherDimSize, int kernel, int kRadius);
extern "C" void doConvolute1DStrided( float * d_in, int n_in, int s_in, float * d_k, int n_k, float * d_out, int s_out, bool neumann );
extern "C" void do_conv1din4d( int n1, int n2, int n3, int n4, int nk,
							   const float *  src, int ss1, int ss2, int ss3, int ss4,
							   float * dst, int ds1, int ds2, int ds3, int ds4,
							   const float * krn, int type );

void GPUData::convolute1D(GPUData& k, size_t dim, GPUData& dst, int typeA) const
{
	bool paddedWithZero = typeA==0;

	if ( extent(dim) == 1 )
		//throw std::runtime_error( str( boost::format("conv1d over singleton dimension %i") % dim ) );
		dst.copyFrom(*this);

	if ( rank() != dst.rank() )
		throw GPUDataException( str( boost::format( "conv1d: rank mismatch this: %i other: %i" )
									 % rank() % dst.rank() ) );

	for ( size_t i = 0; i < rank(); ++i )
		{
			if ( extent(i) != dst.extent(i) ||  stride(i) != dst.stride(i) )
				throw GPUDataException( str( boost::format( "conv1d: incompatible arrangement along dim %i:"
															" extents %i,%i strides %i,%i" )
											 % i % extent(i) % dst.extent(i) % stride(i) % dst.stride(i) ) );
		}

	if ( empty() )
		return;

	if ( rank() > 4 )
		{
			{ /* if the number of dimensions could be reduced ... */
				size_t dim_new = dim;
				GPUData reduced_src ( reduceDimensionality(dim_new) );
				if ( reduced_src.rank() < rank() )
					{
						GPUData reduced_dst ( dst.reduceDimensionality(dim) );
						reduced_src.convolute1D(k,dim_new,reduced_dst,typeA);
						return;
					}
			}
			// recurse
			size_t dim_recurse = suggest_recursion_dim( dim ); // finds a recursion candidate beside the convolution dim
			if ( dim_recurse < dim ) --dim; // corrects for dimension taken away
			for ( size_t i = 0; i < extent(dim_recurse); ++i )
				{
					GPUData srcslice(     slice( dim_recurse, i ) );
					GPUData dstslice( dst.slice( dim_recurse, i ) );
					srcslice.convolute1D( k, dim, dstslice );
				}
		}
	else if ( rank() == 2 && isContiguous() && paddedWithZero )
		{
			if ( stride(dim) == 1 )
				{
					doConvoluteDeepestOfTwoDims( data(), dst.data(), size(), extent(dim), k.copyToConstMemory(), k.size() / 2 );
				}
			else
				{
					doConvoluteFirstOfTwoDims  ( data(), dst.data(), size(), extent(dim), k.copyToConstMemory(), k.size() / 2 );
				}
		}
	else // rank is < 4, by first condition
		{
			size_t kernelSize = k.size();

			size_t order[4] = { 0, 1, 2, 3 };

			// first dim must be conv dim
			std::swap( order[dim], order[0] );

			// second dim should be cont. and at least 4 layers deep
			size_t second_dim = 1;
			for ( size_t i = 2; i < 4; ++i )
				{
					size_t d_i = order[i];
					if ( (    stride(d_i) < stride( order[second_dim] ) && extent(d_i) >= 4 )
						 || ( extent( order[second_dim] ) < 4 && extent( d_i ) >= 4 ) )
						{
							second_dim = i;
						}
				}
			std::swap( order[second_dim], order[1] );

#if 1
			convolute1DInContiguous4D( data(), k.copyToConstMemory(), kernelSize/2, dst.data()
									   , stride( order[0] ), extent( order[0] )
									   , stride( order[1] ), extent( order[1] )
									   , stride( order[2] ), extent( order[2] )
									   , stride( order[3] ), extent( order[3] )
									   , typeA
									   );
#else
			do_conv1din4d(                 extent( order[0] ),     extent( order[1] ),     extent( order[2] ),     extent( order[3] ),
										   k.size(),
						       data(),     stride( order[0] ),     stride( order[1] ),     stride( order[2] ),     stride( order[3] ),
						   dst.data(), dst.stride( order[0] ), dst.stride( order[1] ), dst.stride( order[2] ), dst.stride( order[3] ),
						   k.data(), typeA );
#endif
		}

	GPUDataException::check_error ( "end of DPUData::convolute1d" );

}

extern "C" void
dericheFilter(float* src, float* dst,
			  float ndim, int* dims,
			  int* strides_src, int* strides_dst,
			  float sigma, int dim, int order);

GPUData GPUData::dericheFilter(float sigma, size_t dim, int order, GPUData& dst) const
{
	if (order < 0 || order > 2)
		throw GPUDataException("Invalid order for Deriche filter");

	::dericheFilter(data(), dst.data(),
					this->getDimensionsCount(), (int*) this->getDimensions(),
					(int*) this->getStrides(), (int*) dst.getStrides(),
					sigma, dim, order);
	return dst;
}

GPUData GPUData::dericheFilter(float sigma, size_t dim, int order) const
{

	if (order < 0 || order > 2)
		throw GPUDataException("Invalid order for Deriche filter");

	GPUData out(this->ndim, this->dims);
	::dericheFilter(data(), out.data(),
					this->getDimensionsCount(), (int*) this->getDimensions(),
					(int*) this->getStrides(), (int*) out.getStrides(),
					sigma, dim, order);
	return out;
}


bool GPUDataIsEqualTo::operator()( const GPUData& a, const GPUData& b ) const {
	if (a.getNoOfDimensions()!=b.getNoOfDimensions())
		return false;
	const size_t size = a.size();
	if (size!=b.size())
		return false;
	if (memcmp(a.getSizes(),b.getSizes(),sizeof(int)*a.getNoOfDimensions())!=0)
		return false;
	if (a.data() == b.data() && memcmp(a.getStrides(),b.getStrides(),sizeof(int)*a.getNoOfDimensions())!=0)
		return true;
	GPUData temp(a.copy());
	temp.sub(b);
	float mean = temp.sum()/a.size();
	if (mean!=0)
		return false;
	if (temp.max()!=0.0f && temp.min()!=0.0f)
		return false;
	// TODO: is this sufficient??
	return true;
}

extern "C" void do_linFirstOrderDGL( const int type,
								float* data,
								const int w,
								const int h,
								const float a,
								const float* A,
								const float b,
								const float* B,
								const float C,
								const float d,
								const float* D,
								const float e,
								const float* E,
								const float F,
								const float dt,
								const bool colMaj);

void GPUData::linFirstOrderODEEulerStep(
								const float a,
								const GPUData& A,
								const float b,
								const GPUData& B,
								const float C,
								const float d,
								const GPUData& D,
								const float e,
								const GPUData& E,
								const float F,
								const float dt)
{
	if (!B.isContiguous() || !E.isContiguous())
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderDGLEulerStep() not yet supported");
	}

	if ( ndim!=2 || A.ndim!=2 || B.ndim!=1 ||
			        D.ndim!=2 || E.ndim!=1
	) {
		throw GPUDataNYIException("GPUData::linFirstOrderDGLEulerStep() input has wrong dimensionaltiy");
	}
	
	bool isColMajor = stride(0)==1;
	
	if (A.stride(isColMajor?0:1)!=1 || D.stride(isColMajor?0:1)!=1) {
		throw GPUDataNYIException("all input to GPUData::linFirstOrderDGLEulerStep() must have the same memory layout");
	}
	if ( A.stride( isColMajor ? 1 : 0 ) != int( A.extent( isColMajor ? 0 : 1 ) ) ||
			 D.stride( isColMajor ? 1 : 0 ) != int( D.extent( isColMajor ? 0 : 1 ) ) ||
			 stride  ( isColMajor ? 1 : 0 ) != int(   extent( isColMajor ? 0 : 1 ) )
			 )
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderDGLEulerStep() not yet supported");
	}

	const size_t w = dims[1];
	const size_t h = dims[0];

	if (A.dims[1]!=w || A.dims[0]!=h ||
		B.size()!=h ||
		D.dims[1]!=w || D.dims[0]!=h ||
		E.size()!=h)
	{
		throw GPUDataException("sizes do not match");
	}

	do_linFirstOrderDGL(	1, 		data(),
									w,
									h,
									a,A.data(), b, B.data(),
									C,
									d,D.data(), e, E.data(),
									F,
									dt,isColMajor);

	GPUDataException::check_error("GPUData::linFirstOrderODEEulerStep()");
}

void GPUData::linFirstOrderODEImplicitEulerStep(
								const float a,
								const GPUData& A,
								const float b,
								const GPUData& B,
								const float C,
								const float d,
								const GPUData& D,
								const float e,
								const GPUData& E,
								const float F,
								const float dt)
{
	if (!B.isContiguous() || !E.isContiguous())
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderODEImplicitEulerStep() not yet supported");
	}

	if ( ndim!=2 || A.ndim!=2 || B.ndim!=1 ||
					D.ndim!=2 || E.ndim!=1
	) {
		throw GPUDataNYIException("GPUData::linFirstOrderODEImplicitEulerStep() input has wrong dimensionaltiy");
	}
	
	bool isColMajor = stride(0)==1;
	
	if (A.stride(isColMajor?0:1)!=1 || D.stride(isColMajor?0:1)!=1) {
		throw GPUDataNYIException("all input to GPUData::linFirstOrderODEImplicitEulerStep() must have the same memory layout");
	}

	if ( A.stride( isColMajor ? 1 : 0 ) != int( A.extent( isColMajor ? 0 : 1 ) ) ||
			 D.stride( isColMajor ? 1 : 0 ) != int( D.extent( isColMajor ? 0 : 1 ) ) ||
			 stride  ( isColMajor ? 1 : 0 ) != int(   extent( isColMajor ? 0 : 1 ) )
			 )
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderODEImplicitEulerStep() not yet supported");
	}
	const size_t w = dims[1];
	const size_t h = dims[0];

	if (A.dims[1]!=w || A.dims[0]!=h ||
		B.size()!=h ||
		D.dims[1]!=w || D.dims[0]!=h ||
		E.size()!=h)
	{
		throw GPUDataException("sizes do not match");
	}

	do_linFirstOrderDGL(	2, 		data(),
									w,
									h,
									a,A.data(), b, B.data(),
									C,
									d,D.data(), e, E.data(),
									F,
									dt,isColMajor);
	GPUDataException::check_error("GPUData::linFirstOrderODEImplicitEulerStep()");
}

void GPUData::linFirstOrderODESteayStateSolution(
								const float a,
								const GPUData& A,
								const float b,
								const GPUData& B,
								const float C,
								const float d,
								const GPUData& D,
								const float e,
								const GPUData& E,
								const float F)
{
	if (!B.isContiguous() || !E.isContiguous())
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderODESteayStateSolution() not yet supported");
	}

	if ( ndim!=2 || A.ndim!=2 || B.ndim!=1 ||
					D.ndim!=2 || E.ndim!=1
	) {
		throw GPUDataNYIException("GPUData::linFirstOrderODESteayStateSolution() input has wrong dimensionaltiy");
	}
	
	bool isColMajor = stride(0)==1;
	
	if (A.stride(isColMajor?0:1)!=1 || D.stride(isColMajor?0:1)!=1) {
		throw GPUDataNYIException("all input to GPUData::linFirstOrderODESteayStateSolution() must have the same memory layout");
	}
	if ( A.stride( isColMajor ? 1 : 0 ) != int( A.extent( isColMajor ? 0 : 1 ) ) ||
			 D.stride( isColMajor ? 1 : 0 ) != int( D.extent( isColMajor ? 0 : 1 ) ) ||
			 stride  ( isColMajor ? 1 : 0 ) != int(   extent( isColMajor ? 0 : 1 ) )
			 )
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderDGLEulerStep() not yet supported");
	}

	const size_t w = dims[1];
	const size_t h = dims[0];

	if (A.dims[1]!=w || A.dims[0]!=h ||
		B.size()!=h ||
		D.dims[1]!=w || D.dims[0]!=h ||
		E.size()!=h)
	{
		throw GPUDataException("sizes do not match");
	}

	do_linFirstOrderDGL( 3,					data(),
											w,
											h,
											a,A.data(), b, B.data(),
											C,
											d,D.data(), e, E.data(),
											F, 10000000000.0f,isColMajor);
	GPUDataException::check_error("GPUData::linFirstOrderODESteayStateSolution()");
}


extern "C" void do_linFirstOrderDGL2(const int type,
								float* data,
								const int w,
								const int h,
								const float a,
								const float b,
								const float* B,
								const float c,
								const float* C,
								const float dt,
								const bool rect_input,
								const bool colMaj);


void GPUData::linFirstOrderODESteayStateSolution(
		const float a,
		const float b, const GPUData& B,
		const float c, const GPUData& C, bool rect_input)
{
//	throw GPUDataNYIException("do_linFirstOrderDGL2 is not yet tested");
	if (!C.isContiguous())
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderDGLSteayStateSolution() not yet supported");
	}

	if ( ndim!=2 || B.ndim!=2 || C.ndim!=1)
	{
		throw GPUDataNYIException("GPUData::linFirstOrderDGLSteayStateSolution() input has wrong dimensionaltiy");
	}

	bool isColMajor = stride(0)==1;

	if (B.stride(isColMajor?0:1)!=1) {
		throw GPUDataNYIException("all input to GPUData::linFirstOrderDGLSteayStateSolution() must have the same memory layout");
	}
	if ( B.stride( isColMajor ? 1 : 0 ) != int( B.extent(isColMajor?0:1) ) ||
			 stride  ( isColMajor ? 1 : 0 ) != int(  extent(isColMajor?0:1 ) ) 
			 )
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderDGLSteayStateSolution() not yet supported");
	}
	
	const size_t w = dims[1];
	const size_t h = dims[0];

	if (B.dims[1]!=w || B.dims[0]!=h ||
		C.size()!=h)
	{
		throw GPUDataException("sizes do not match");
	}

	do_linFirstOrderDGL2( 3, data(),
						  w,
						  h,
						  a,
						  b, B.data(),
						  c, C.data(),
						  1e10f, rect_input,isColMajor);
	GPUDataException::check_error("GPUData::do_linFirstOrderDGL2()");

}

void GPUData::linFirstOrderODEEulerStep(
		const float a,
		const float b, const GPUData& B,
		const float c, const GPUData& C,
		const float dt, bool rect_input)
{
//	throw GPUDataNYIException("do_linFirstOrderDGL2 is not yet tested");
	if (!C.isContiguous())
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderODEEulerStep() not yet supported");
	}

	if ( ndim!=2 || B.ndim!=2 || C.ndim!=1)
	{
		throw GPUDataNYIException("GPUData::linFirstOrderODEEulerStep() input has wrong dimensionaltiy");
	}

	bool isColMajor = stride(0)==1;

	if (B.stride(isColMajor?0:1)!=1) {
		throw GPUDataNYIException("all input to GPUData::linFirstOrderODEEulerStep() must have the same memory layout");
	}

	if ( B.stride( isColMajor ? 1 : 0 ) != int( B.extent( isColMajor ? 0 : 1 ) ) ||
			 stride  ( isColMajor ? 1 : 0 ) != int(  extent ( isColMajor ? 0 : 1 ) ) 
			 )
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderODEEulerStep() not yet supported");
	}

	if ( B.extent(0) != extent(0) ||
		 B.extent(1) != extent(1) ||
		 C.size()    != extent(0) )
	{
		throw GPUDataException("sizes do not match");
	}

	do_linFirstOrderDGL2( 1, data(), extent(1), extent(0),
						  a,
						  b, B.data(),
						  c, C.data(),
						  dt, rect_input,isColMajor);
	GPUDataException::check_error("GPUData::linFirstOrderDGLSteayStateSolution()");
}

void GPUData::linFirstOrderODEImplicitEulerStep(
		const float a,
		const float b, const GPUData& B,
		const float c, const GPUData& C,
		const float dt, bool rect_input)
{
//	throw GPUDataNYIException("do_linFirstOrderDGL2 is not yet tested");
	if (!C.isContiguous())
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderODEImplicitEulerStep() not yet supported");
	}

	if ( ndim!=2 || B.ndim!=2 || C.ndim!=1)
	{
		throw GPUDataNYIException("GPUData::linFirstOrderODEImplicitEulerStep() input has wrong dimensionaltiy");
	}

	bool isColMajor = stride(0)==1;

	if (B.stride(isColMajor?0:1)!=1) {
		throw GPUDataNYIException("all input to GPUData::linFirstOrderODEImplicitEulerStep() must have the same memory layout");
	}
	if ( B.stride( isColMajor ? 1 : 0 ) != int( B.extent( isColMajor ? 0 : 1 ) ) ||
			 stride  ( isColMajor ? 1 : 0 ) != int(  extent ( isColMajor ? 0 : 1 ) ) 
			 )
	{
		throw GPUDataNYIException("non contiguous data in GPUData::linFirstOrderODEImplicitEulerStep() not yet supported");
	}
	
	if ( B.extent(0) != extent(0) ||
		 B.extent(1) != extent(1) ||
		 C.size()    != extent(0) )
	{
		throw GPUDataException("sizes do not match");
	}

	do_linFirstOrderDGL2( 2, data(), extent(1), extent(0),
						  a,
						  b, B.data(),
						  c, C.data(),
						  dt, rect_input,isColMajor);

	GPUDataException::check_error("GPUData::linFirstOrderDGLSteayStateSolution()");
}

extern "C" void do_pooled_linear_feature_transform( const GPUDataBase * featuremap,
													const GPUDataBase * transform,
													const GPUDataBase * means,
													GPUDataBase * target,
													size_t d, size_t step );

void GPUData::linearPoolingFeatureTransform( const GPUData featuremap, size_t d, size_t step, const GPUData& transform,
											 const GPUData& means, bool columnMajor )
{
	size_t one = 0, two = 1;
	GPUData f_reduced( featuremap.reduceDimensionality( one, two ) );
	if ( f_reduced.rank() > 3 || f_reduced.rank() < 2 )
		{
			throw GPUDataException("linearPoolingFeatureTransform needs 2d or 3d features");
		}
	size_t k_in = d * d * f_reduced.extent(2);
	if ( k_in != transform.extent(1) )
		throw GPUDataException("linearPoolingFeatureTransform transform does not match data");
	if ( transform.rank() > 2 || transform.stride(1) != 1 )
		throw GPUDataException("linearPoolingFeatureTransform needs 2d transform with contiguous feature data");
	if ( featuremap.extent(0) < d || featuremap.extent(1) < d )
		{
			clear();
			return;
		}

	size_t k_out = transform.extent(0);
	size_t d1 = 1 + ( f_reduced.extent(0) - d ) / step;
	size_t d2 = 1 + ( f_reduced.extent(1) - d ) / step;
	makeContiguous( true );
	if ( columnMajor )
		{
			resize( k_out, d1, d2 );
			reference( shiftDim( -1 ) );
		}
	else
		{
			resize( d1, d2, k_out );
		}
	do_pooled_linear_feature_transform( &f_reduced, &transform, &means, this, d, step );

}

void GPUData::linearFeatureTransform( const GPUData featuremap, const GPUData& transform, const GPUData& means, bool columnMajor )
{

	// what I need to know is: are the features and pixels contiguous and do they match the transform?

	size_t ndata = featuremap.extent(0) * featuremap.extent(1);
	size_t kin   = transform .extent(1);
	size_t kout  = transform .extent(0);


	char ftrans;
	int fstride;

	if ( featuremap.stride( featuremap.rank() - 1 ) != 1 )
		{
			ftrans = columnMajor ? 'N' : 'T';
			fstride = featuremap.stride( featuremap.rank() - 1 );
		}
	else
		{
			ftrans = columnMajor ? 'T' : 'N';
			fstride = featuremap.stride( 1 );
		}


	if ( columnMajor )
		{
			resize( kout, featuremap.extent(0), featuremap.extent(1) );
			reference(shiftDim(-1));

			cublasSgemm (ftrans,'N', ndata, kout, kin,
						 1.0f,
						 featuremap.data(), fstride,              // features
						 transform.data(), transform.stride(0),   // transform
						 0.0f, data(), ndata );               // output, no update
		}
	else
		{
			resize( featuremap.extent(0), featuremap.extent(1), kout );

			cublasSgemm ('T', ftrans, kout, ndata, kin,
						 1.0f,
						 transform.data(), transform.stride(0),   // transform
						 featuremap.data(), fstride,              // features
						 0.0f, data(), kout );               // output, no update
		}
	cublasStatus status = cublasGetError();
	if (status != CUBLAS_STATUS_SUCCESS) {
		throw GPUDataException("error performing cublasSgemm in GPUData::lnearFeatureTransform()");
	}

	//makeContiguous();

	// apply means

	GPUData outmeans( kout );
	outmeans.matVecMul( transform, means );
	add_repeated( outmeans, -1.0f );

}

void GPUData::matVecMul(const GPUData& matrixInput, const GPUData& vectorInput, bool allow_resize )
{


	const size_t kin = vectorInput.size();
	size_t kout;
	char mtrans;
	int mstride;

	// orientation of matrix is determined by the unit stride dimension
	// which must be the input dimension
	if ( matrixInput.stride(0) == 1 )
		{
			kout = matrixInput.extent(1);
			mtrans = 'N';
			mstride = matrixInput.stride(1);
			if ( matrixInput.extent(0) != kin )
				throw GPUDataException("matVecMul: input size mismatch");
		}
	else if ( matrixInput.stride(1) == 1 )
		{
			kout = matrixInput.extent(0);
			mtrans = 'T';
			mstride = matrixInput.stride(0);
			if ( matrixInput.extent(1) != kin )
				throw GPUDataException("matVecMul: input size mismatch");
		}
	else
		throw GPUDataException("GPUData::matVecMul(..) transform must be unit stride in input dimension");

	if ( kout != size() )
		{
			if ( allow_resize )
				resize( kout );
			else
				throw GPUDataException("matVecMul output size mismatch");
		}

//	std::cout << "DEBUG in matVecMul A: " << this->max();
//	std::cout << " " << vectorInput.max();
//	std::cout << " " << matrixInput.max();
	// fortran order view!!!
	cublasSgemv (mtrans, matrixInput.extent(1), matrixInput.extent(0), 1.0f,
				 matrixInput.data(), mstride,
				 vectorInput.data(), vectorInput.stride(0),
				 0.0f, data(), stride(0));

//	std::cout << "... B: " << this->max() << std::endl;
	cublasStatus status = cublasGetError();
	if (status != CUBLAS_STATUS_SUCCESS) {
		throw GPUDataException("error performing cublasSgemv in GPUData::matVecMul()");
	}
}


float GPUData::mean() const
{
	return sum()/size();
}

void GPUData::make_unique()
{
  if (inner != NULL && inner->get_refcount() > 1)
  {
      GPUData new_instance;
      new_instance.copyFrom( *this, true );
      reference( new_instance );
  }
}

bool GPUData::is_same_shape_as( const GPUData & other, bool test_strides ) const
{
	if ( rank() != other.rank() )
		return false;

	for ( size_t i = 0; i < rank(); ++i )
		if ( extent(i) != other.extent(i) )
			return false;

	if ( test_strides )
		for ( size_t i = 0; i < rank(); ++i )
			if ( stride(i) != other.stride(i) )
				return false;

	return true;
}

float* GPUData::data(bool move) const
{ 
	if ( is_array() && move ) 
		throw std::runtime_error("cannot move const data");
	return _data; 
}

float* GPUData::data(bool move) 
{ 
	if ( is_array() && move ) 
		move_to_data();
	return _data; 
}


void GPUData::move_to_array( bool features_first )
{
	
	if ( is_array() ) return;
	//		throw GPUDataException("Error in move_to_array: array is already filled");
	
	make_unique();
	makeContiguous( isRowMajor() );
	GPUData three_d_view( *this );
	size_t n = getNoOfDimensions();
	size_t one, two;
	if ( features_first )
	{
		one = n-2;
		two = n-1;
	}
	else
	{
		one = 0;
		two = 1;
	}
	three_d_view.reference( three_d_view.reduceDimensionality( one, two ) );
	
	//allocate
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float>();
	int tex_width, tex_height;
	if ( features_first )
	{
		tex_width = three_d_view.extent(2);
		tex_height = three_d_view.extent(0) * three_d_view.extent(1);
	}
	else
	{
		tex_width = three_d_view.extent(1)* three_d_view.extent(2);
		tex_height = three_d_view.extent(0);
  }
	three_d_view.clear();
  
  GPUDataException::check_error( str( boost::format("failed allocation of [%d,%d] bytes at %0x") %tex_width %tex_height %inner->array )
  															,cudaMallocArray( &(inner->array), &channelDesc, tex_width, tex_height) ); 
	
	//move
	GPUDataException::check_error( str( boost::format("copy from data (%0x) to array (%0x), size=%d in_dims %s 3d %s tex_dims %dx%d")
																			%inner->data %inner->array %inner->size % dump_dims() % three_d_view.dump_dims() % tex_width % tex_height )
								 , cudaMemcpyToArray(inner->array, 0, 0, 
													 inner->data, 
													 inner->size, 
													 cudaMemcpyDeviceToDevice) );


	// delete data
	CudaMemPool::get().free( inner->data );
	_data = 0;
	inner->data = 0;
}

void GPUData::move_to_data()
{
	
	if ( !is_array() )
		throw GPUDataException("Error in move_to_data: array is empty");
	
	// allocate data
	inner->data = (float*) CudaMemPool::get().alloc( inner->size );
	_data = inner->data;
	
	// move
	GPUDataException::check_error( "copy from array to data"
      , cudaMemcpyFromArray( inner->data,
					  inner->array, 0, 0, 
					  inner->size /* * sizeof(float)*/,
					  cudaMemcpyDeviceToDevice) ); 
	
	// delete array
	cudaFreeArray( inner->array );
	inner->array = 0;
	GPUDataException::check_error( "delete array");
}

bool GPUData::is_array( ) const
{
	return ( ( inner != NULL ) && ( inner->array != NULL ) );
}

std::string GPUData::dump_dims() const
{
	std::string result("[ ");
	if (ndim==0) return (result+"]");
	for ( size_t i = 0; i < ndim - 1; i++ )
		result.append(   boost::lexical_cast<std::string>( extent(i) )
					   + " ( " + boost::lexical_cast<std::string>( stride(i) ) + " )"
					   + " x " );
	result.append(     boost::lexical_cast<std::string>( extent( ndim - 1 ) )
					   + " ( " + boost::lexical_cast<std::string>( stride( ndim - 1 ) ) + " )"
					   + " ]" );
	result.append("(r:" + boost::lexical_cast<std::string>( refcount() ) + ") " );
	if ( is_array() )
		result.append(str(boost::format(" @%0x( array: %0x)")%this%array()));
	else
		result.append(str(boost::format(" @%0x(%0x)")%this%data()));
	return result;
}


// oberhoff
#define BINARY_KERNEL_ARGS												\
	float* data, int inc1, int inc2, int inc3, int inc4,				\
		int N1, int N2, int N3, int N4,									\
		const float* op, int inc1o, int inc2o, int inc3o, int inc4o

typedef void (*local_binary_kernel_entry)( BINARY_KERNEL_ARGS );

#define LOCAL_BINARY_FUN(mname,kname)																		\
	extern "C" void do_local_binary_ ## kname( BINARY_KERNEL_ARGS );			\
	GPUData& GPUData:: mname ( const GPUData & op )												\
	{																																			\
		this->localBinaryFunc( do_local_binary_ ## kname, op );							\
		GPUDataException::check_error( str(boost::format("error after local binary GPUData function %s from %s to %s") \
																			 % BOOST_PP_STRINGIZE(mname) % op.dump_dims() % dump_dims() ) ); \
		return *this;																												\
	}

#define BINARY_KERNEL_A0_ARGS												\
	float a0, float* data, int inc1, int inc2, int inc3, int inc4,				\
		int N1, int N2, int N3, int N4,									\
		const float* op, int inc1o, int inc2o, int inc3o, int inc4o

typedef void (*local_binary_kernel_a0_entry)( BINARY_KERNEL_A0_ARGS );

#define LOCAL_BINARY_A0_FUN(mname,kname)																\
	extern "C" void do_local_binary_a0_ ## kname( BINARY_KERNEL_A0_ARGS ); \
	GPUData & GPUData:: mname ( const GPUData & op, float a0 )						\
	{																																			\
		this->localBinaryFunc_a0( do_local_binary_a0_ ## kname, op, a0 );		\
		GPUDataException::check_error( str(boost::format("error after local binary GPUData function %s from %s to %s") \
																			 % BOOST_PP_STRINGIZE(mname) % op.dump_dims() % dump_dims() ) ); \
		return *this;																												\
	}


#define UNARY_KERNEL_ARGS												\
	float* data, int inc1, int inc2, int inc3, int inc4,				\
		int N1, int N2, int N3, int N4

#define LOCAL_UNARY_FUNC(mname, kname)									\
	extern "C" void doLocal_ ## kname(UNARY_KERNEL_ARGS);				\
	GPUData& GPUData:: mname()											\
	{																	\
		this->localFunc(doLocal_ ## kname);								\
		GPUDataException::check_error( str(boost::format("error after local unary GPUData function %s") \
										   % BOOST_PP_STRINGIZE(mname) ) ); \
		return *this;													\
	}



#define UNARY_KERNEL_A0_ARGS												\
	float a0, float* data, int inc1, int inc2, int inc3, int inc4,				\
		int N1, int N2, int N3, int N4

#define LOCAL_UNARY_A0_FUNC(mname, kname)								\
	extern "C" void do_local_ ## kname ## _a0(UNARY_KERNEL_A0_ARGS);	\
	GPUData& GPUData:: mname(float a0) {								\
		this->localFunc_a0( do_local_ ## kname ## _a0, a0);				\
		GPUDataException::check_error( str(boost::format("error after local unary one arg GPUData function %s")	\
										   % BOOST_PP_STRINGIZE(mname) ) ); \
		return *this;													\
	}

GPUData GPUData::dimsSorted()
{
	std::vector< oentry > order(rank());
	//std::cout << " sorting dims, before: ";
	for ( size_t i = 0; i < rank(); ++i )
		{ 
			//std::cout << i << ": " << extent(i) << " ";
			order[i].first = extent(i); 
			order[i].second = i; 
		}
	//std::cout << std::endl << " after: ";
	std::sort( order.begin(), order.end(), order_compare );
  GPUData result(*this);
	for ( size_t i = 0; i < rank(); ++i )
	{
    size_t pos = order[ rank() - i - 1 ].second;
		//std::cout << pos << ": " << extent(pos) << " ";
    result.strides[i] = stride(pos);
    result.dims[i]    = extent(pos);
	}
	//std::cout << std::endl;
  return result;
}

void GPUData::localBinaryFunc( local_binary_kernel_entry entry, const GPUData & op)
{
	if ( rank() != op.rank() )
		throw GPUDataException( str( boost::format( "rank mismatch for binary op r1 = %i r2 = %i ") % rank() % op.rank() ) );

	for ( size_t dim = 0; dim < rank(); ++dim )
		if ( extent(dim) != op.extent(dim) )
			throw GPUDataException( str( boost::format( "extents mismatch for binary op on dim %i n1 = %i n2 = %i (this:%s, other:%s)")
										   % dim % extent(dim) % op.extent(dim) % dump_dims() %op.dump_dims() ) );

	if ( empty() ) return;

	/// sort dims descending
	std::vector< oentry > order( rank() );
	for ( int i = 0; i < rank(); ++i )
		{ order[i].first = extent(i); order[i].second = i; }
	std::sort( order.begin(), order.end(), order_compare );

	if ( rank() > 4 )
		{
			size_t red_dim = order.front().second;
			for ( size_t i = 0; i < extent(red_dim); ++i )
				{
					slice( red_dim, i ).localBinaryFunc( entry, op.slice( red_dim, i ) );
				}
		}
	else
		{
			size_t  dims[4] = {1,1,1,1};
			int istrides[4] = {0,0,0,0};
			int ostrides[4] = {0,0,0,0};
			for ( size_t i = 0; i < rank(); ++i )
				{
					size_t j = order[rank() - i - 1].second;
					dims[i] = extent(j);
					istrides[i] = stride(j);
					ostrides[i] = op.stride(j);					
				}
			entry( data(), istrides[0], istrides[1], istrides[2], istrides[3],
						 dims[0], dims[1], dims[2], dims[3], 
						 op.data(), ostrides[0], ostrides[1], ostrides[2], ostrides[3] );

		}


	GPUDataException::check_error( str(boost::format("error after local binary GPUData function from %s to %s")
																		 % op.dump_dims() % dump_dims() ) );
}

void GPUData::localBinaryFunc_a0( local_binary_kernel_a0_entry entry, const GPUData & op, float a0)
{
	if ( rank() != op.rank() )
		throw GPUDataException( str( boost::format( "rank mismatch for binary op r1 = %i r2 = %i ") % rank() % op.rank() ) );

	for ( size_t dim = 0; dim < rank(); ++dim )
		if ( extent(dim) != op.extent(dim) )
			throw GPUDataException( str( boost::format( "extents mismatch for binary op on dim %i n1 = %i n2 = %i (this:%s, other:%s)")
										   % dim % extent(dim) % op.extent(dim) % dump_dims() %op.dump_dims() ) );
	if ( empty() ) return;

	/// sort dims descending
	std::vector< oentry > order( rank() );
	for ( int i = 0; i < rank(); ++i )
		{ order[i].first = extent(i); order[i].second = i; }
	std::sort( order.begin(), order.end(), order_compare );

	if ( rank() > 3 ) // 4d seems to be broken :(
		{
			size_t red_dim = order.front().second;
			for ( size_t i = 0; i < extent(red_dim); ++i )
				{
					slice( red_dim, i ).localBinaryFunc_a0( entry, op.slice( red_dim, i ), a0 );
				}
		}
	else
		{
			size_t  dims[4] = {1,1,1,1};
			int istrides[4] = {0,0,0,0};
			int ostrides[4] = {0,0,0,0};
			for ( size_t i = 0; i < rank(); ++i )
				{
					size_t j = order[rank() - i - 1].second;
					dims[i] = extent(j);
					istrides[i] = stride(j);
					ostrides[i] = op.stride(j);					
				}
			entry( a0, data(), istrides[0], istrides[1], istrides[2], istrides[3],
						 dims[0], dims[1], dims[2], dims[3], 
						 op.data(), ostrides[0], ostrides[1], ostrides[2], ostrides[3] );

		}

	GPUDataException::check_error( str(boost::format("error after local binary on arg GPUData function from %s to %s")
																		 % op.dump_dims() % dump_dims() ) );

}

void GPUData::localFunc(void (*func_pointer)(UNARY_KERNEL_ARGS))
{

	if ( empty() ) return;

	GPUData own = reduceDimensionality().dimsSorted();
	if ( own.rank() <= 4 )
	{
		(*func_pointer)(own.data(),own.stride(0),own.stride(1),own.stride(2),own.stride(3),
		                own.extent(0),own.extent(1),own.extent(2),own.extent(3));
	} else {
		throw GPUDataNYIException("local functions for more than 4 dimensions not implemented");
	}
}

void GPUData::localFunc_a0(void (*func_pointer)(UNARY_KERNEL_A0_ARGS), float a0)
{

	if ( empty() ) return;

	GPUData own = reduceDimensionality().dimsSorted();
	if ( own.rank() <= 4 )
	{
		(*func_pointer)(a0, own.data(),own.stride(0),own.stride(1),own.stride(2),own.stride(3),
		                own.extent(0),own.extent(1),own.extent(2),own.extent(3));
	}
	else
	{
		throw GPUDataNYIException("local functions for more than 4 dimensions not implemented");
	}
	
}

#define REDUCTION(name)																						\
	extern "C" float reduce_ ## name( float *d_idata, int size );		\
																																	\
	float GPUData:: name () const																					\
	{																																			\
		float result;																												\
		if (isContiguous()) {																								\
			result = reduce_ ## name (data(),size());													\
		} else {																														\
			result = copy().name();																						\
		}																																		\
		GPUDataException::check_error(str(boost::format("error after reduction on %s ")%dump_dims())); \
		return result;																											\
	}

REDUCTION(min)
REDUCTION(max)
REDUCTION(sum)
REDUCTION(product)

LOCAL_BINARY_FUN( add, Add )
LOCAL_BINARY_FUN( assign, Assign )
LOCAL_BINARY_FUN( sub, Sub )
LOCAL_BINARY_FUN( mul, Mul )
LOCAL_BINARY_FUN( div, Div )
LOCAL_BINARY_FUN( add_squared, AddSquared )
LOCAL_BINARY_FUN( add_abs, AddAbs )
LOCAL_BINARY_FUN( max, Max )
LOCAL_BINARY_FUN( min, Min )

LOCAL_BINARY_A0_FUN( fb_mod, FbMod )
LOCAL_BINARY_A0_FUN( mul, Mul )
LOCAL_BINARY_A0_FUN( add, Add )
LOCAL_BINARY_A0_FUN( add_squared, AddSquared )
LOCAL_BINARY_A0_FUN( add_abs, AddAbs )

LOCAL_UNARY_FUNC(sin,sin)
LOCAL_UNARY_FUNC(cos,cos)
LOCAL_UNARY_FUNC(asin,asin)
LOCAL_UNARY_FUNC(acos,acos)
LOCAL_UNARY_FUNC(exp,exp)
LOCAL_UNARY_FUNC(exp2,exp2)
LOCAL_UNARY_FUNC(log,log)
LOCAL_UNARY_FUNC(log2,log2)
LOCAL_UNARY_FUNC(rint,rintf)
LOCAL_UNARY_FUNC(isnan,isnan)
LOCAL_UNARY_FUNC(sqrt,sqrt)
LOCAL_UNARY_FUNC(rsqrt,rsqrt)
LOCAL_UNARY_FUNC(sqr,sqr)
LOCAL_UNARY_FUNC(rect,rect)
LOCAL_UNARY_FUNC(abs,abs)

LOCAL_UNARY_A0_FUNC( add, Add )
LOCAL_UNARY_A0_FUNC( sub, Sub )
LOCAL_UNARY_A0_FUNC( mul, Mul )
LOCAL_UNARY_A0_FUNC( div, Div )
LOCAL_UNARY_A0_FUNC( mod, Mod )
LOCAL_UNARY_A0_FUNC( modulo, Mod )
LOCAL_UNARY_A0_FUNC( pow, Pow )
LOCAL_UNARY_A0_FUNC( safe_log, Safe_log )
LOCAL_UNARY_A0_FUNC( rect, Rect )
LOCAL_UNARY_A0_FUNC( compress, Compress )
LOCAL_UNARY_A0_FUNC( min, Min )
LOCAL_UNARY_A0_FUNC( max, Max )
LOCAL_UNARY_A0_FUNC( sigmoid, Sigmoid )

// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 2
// tab-width: 2
// End:
