/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef GPUDATA_H
#define GPUDATA_H

#include <string>
#include <string.h>
#include <stdexcept>
#include "Exceptions.hpp"
#include <boost/unordered_map.hpp>
#include "cufft-kernel-key.hpp"
#include "CPUData.hpp"
#include "GPUDataBase.h"

class cudaArray;
//struct cufftHandle;

/**
 * Multi-dimensional array class stored on GPU memory.
 */
class GPUData : public GPUDataBase
{

public:
	
	GPUData();
	GPUData(size_t size);
	GPUData(size_t row, size_t col);
	GPUData(size_t size_3, size_t size_2, size_t size_1);
	GPUData(size_t size_4, size_t size_3, size_t size_2, size_t size_1);
	GPUData(size_t size_5, size_t size_4, size_t size_3, size_t size_2, size_t size_1);
	GPUData(size_t ndim, const size_t* dims);
	GPUData(const GPUData& src);
	~GPUData();

	/** Construct a copy of an array. */
	GPUData copy( bool keepshape = false ) const;
	///swaps handles
	void swap( GPUData & other ) { GPUData tmp( other ); other = *this; *this = tmp; }
	///release all resources and references
	void clear();
	///checked resize
	void resize ( size_t ndims, const size_t * d );
	void resize( size_t ndims, const size_t* d, const std::vector<size_t> & o );
	///checked resize
	void resize ( size_t d1, size_t d2, size_t d3, size_t d4, size_t d5 );
	void resize ( size_t d1, size_t d2, size_t d3, size_t d4 );
	void resize ( size_t d1, size_t d2, size_t d3 );
	void resize ( size_t d1, size_t d2 );
	void resize ( size_t d1 );
	void resize ( const GPUData & other );

	bool empty() const { return data(false) == 0 && array() == 0; }
	bool isEmpty() const { return empty(); }
	size_t size() const { return empty() ? 0 : GPUDataBase::size(); }

	static size_t cuda_global_mem_usage();
  
	/** Return the number of dimensions in the array. */
	size_t getNoOfDimensions() const { return unsigned(this->ndim); }
	/** Return the number of dimensions in the array. */
	size_t getDimensionsCount() const { return getNoOfDimensions(); }
	/// same as getNoOfDimensions, but shorter...
	size_t rank() const { return getNoOfDimensions(); }
	/** Check if all the elements are next to the others.*/
    bool isContiguous( bool check = true ) const;
	/** Check if strides are ascending */
	bool isRowMajor() const;
	/** Check if strides are descending */
	bool isColumnMajor() const;
	/// staps the given two dimensions
	GPUData swapDimensions(size_t dim1, size_t dim2) const;


	/** make sure data is contiguous, if it is not new data is allocated and the
	 *  data copied there in a contiguous fashion
	 */
	void makeContiguous(bool rowMajor);

	void linearPoolingFeatureTransform( const GPUData featuremap, size_t d, size_t step, const GPUData& transform
										, const GPUData& means, bool columnMajor = false );
	void linearFeatureTransform( const GPUData featuremap, const GPUData& transform, const GPUData& means, bool columnMajor = false );
	/// make sure this array reference is the only reference to the data, if not it is made
	/// a contiguous, row-major copy of the data it references
	void make_unique();
	/// convenience method to test shape equality of two matrices, including strides if needed
	bool is_same_shape_as( const GPUData & other, bool test_strides = false ) const;
	/// returns a string containing the dimensions
	std::string dump_dims() const;

	GPUData makeAnother() const;

	std::vector<size_t> order();

	void reshape( size_t ndim, const size_t * dims, const int * strides  );

	/**
     * Return a shared view of the array.
     * The view will be of the given size starting at the desired
     * position and using the same strides and number of dimensions
     */
    GPUData view(const int* offset, const size_t* dims) const;
	/**
     * Return a shared view of the array.
     * The view will be of the given size starting at the beginning
     * of this array
     */
    GPUData view(const size_t* dims) const;
	/**
     * Return a shared view of the array and can also change the number
     * of dimenstions.
     * The view will be of the given sizes using the given, new stides and
     * new dimensionality.
     * The view will start at the desired position (in respect to the old
     * configuration and old number of dimensions).
     */
    GPUData view(const int* offset, const size_t ndim, const size_t* dims, const int* strides) const;
	/**
     * like GPUData::view(int* offset, size_t ndim, int* dims, int* strides)
     * but with offset = {0,0,...,0}
     */
    GPUData view(size_t ndim, const size_t* dims, const int* strides) const;
    /**
     * Returns a shared view of the array, trying to reduce
	 * the number of dimensions while preserving all elements and their
	 * ordering.
	 * This is done by joining neighbouring dimensions so that they
	 * share one new index within a new larger range.
     */
    GPUData reduceDimensionality() const;
    /**
     * Same like reduceDimensionality() but protects the given dimension
     * so that this dimension will not be absorbed into another.
     * The dimension number of the protected dimension will probably be
     * different in the new view and, thus, will be given back through
     * the referenc parameter dim.
     *
     * \param dim if this parameter is set to a valid dimension [1;ndim]
     * 				this dimension will be protected from being absorbed
     * 				and the new dimension number will be returned through
     * 				this reference
     */
    GPUData reduceDimensionality(size_t& dim) const;
    /**
	 * Returns a shared view of the array where all singleton dimensions have been removed
	 */
	GPUData removeSingletonDimensions() const;

    /**
     * Same like reduceDimensionality(int& dim) but protects two dimensions
     */
    GPUData reduceDimensionality(size_t& dim1, size_t& dim2) const;
    /**
     * Inserts a singleton dimension to the left
     */
    GPUData extendDimensionLeft() const;
	/**
     * Inserts a singleton dimension to the right
     */
    GPUData extendDimensionRight() const { return insertDimension( rank() - 1 ); }
	/// inserts a dimension after the given one
	GPUData insertDimension( size_t index ) const;
	/// make data available in const memory for optimizations
	int copyToConstMemory(bool overwrite = false);

    /**
     * Returns a view with the dimension shifted (circular).
     *
     * \param i if i>0 the dimension is shifted to the right,
     * 			otherwise it is shifted to the left. i=0 is
     * 			allowed too.
     */
    GPUData shiftDim(int i=1) const;
	/**
     * Return a dimensionality reduced view of the array.
     */
	GPUData operator[](size_t index) const;
	/**
	 * reduce the given dimension by fixing the corresponding offset at the given value
	 */
	GPUData slice( size_t dim, size_t index) const;
	/// return a view in which dimension dim is reduced to the given range, strided with a given step
	GPUData range( size_t dim, size_t startindex, size_t endindex, size_t step = 1 ) const;
    /**
     * Convert a singleton to its value.
	 * Since the data is on GPU it cannot be passed as a reference.
     */
    float as_float();
    /**
     * Change the value at the given position.
     */
    GPUData& operator=(const float & src);

	GPUData& reference(const GPUData& src);

	GPUData& operator=(const GPUData& src);

    /**
     * Convert a multi-dimensional position into an offset from
     * the beginning of the data.
     */
    int index(const int* pos) const;

    /**
     * Fill an entire array with a single value.
     */
    void fill(float val);

	/**
	 * Copy memory from host to the GPU allocated for the array.
	 */
	void copyFromHostMemory( const float* data );

	/**
	 * Copy memory from the GPU allocated array to host memory.
	 */
	void copyToHostMemory(float* host_data) const;

	/**
	 * copies the memory content from the origin to this object
	 */
	void copyFrom( const GPUData& origin, bool may_resize = false );

	/// copy data from cpu, may reallocate the data for this container
	void copyFrom( const CPUData<float> & cdata, bool may_resize = true );
	/// copy data from cpu, may reallocate the data for this container
	void copyFrom( CPUDataLayer<float> & cdatalayer, bool may_resize = true );
	///copy data to cpu
	void copyTo( CPUData<float> & cdata, bool may_resize = true ) const;

	/**
	 * Performs a matrix (2d) vector (1d) multiplication over the deepest
	 * dimension (second in matrix).
	 * The result is stored in this object.
	 */
	void matVecMul(const GPUData& matrixInput, const GPUData& vectorInput, bool allow_resize = false );

	void add_repeated(const GPUData& values, const float f = 1.0f);

	/**
	 * multiplicates the given data to this data this[i]=f*this[i]*values[i%s]
	 * with s=size of given values
	 */
	void mulRepeated(const GPUData& values);


	/**
	 * sum over the deepest dimension.
	 */
	void sumOverDeepestDimension(GPUData& dest) const;

	void setIfBetween(const float min, const float max, const float val);
	void setIfBetween(const float min, const float max, const float val, GPUData& res) const;

	/**
	 * returns
	 * 	1 	if a[i]>f
	 * 	0.5 if a[i]==f
	 * 	0	if a[i]<f
	 */
	void heavisideStepFunction(const float f=0.0f);

	/**
	 * returns
	 * 	1 	if a[i]>f
	 * 	centereValue if a[i]==f
	 * 	0	if a[i]<f
	 */
	void heavisideStepFunction(const float f, const float centerValue);


// ---- reductions --- //

	/**
	 * calculates the sum
	 */
	float sum() const;

	/**
	 * calculates the product
	 */
	float product() const;

	/**
	 * calculates the sum of absolutvalues
	 */
	float absSum() const;

	/**
	 * returns the maximum within this dataset
	 */
	float max() const;

	/**
	 * returns the mean of this dataset
	 */
	float mean() const;
	/**
	 * returns the maximum of abs(x[i])
	 */
	float absMax() const;

	/**
	 * returns the minimum within this dataset
	 */
	float min() const;

	/**
	 * returns the maximum of abs(x[i])
	 */
	float absMin() const;

  /// ------------- partial reductions -------------- ///

	/// sums together all but the given dimensions and returns a 2d image of these sums
	GPUData sumContract( size_t dim1 = 0, size_t dim2  = 1) const;

	/// sums together all but the given dimensions and returns a 2d image of these sums
	GPUData maxContract( size_t dim1 = 0, size_t dim2  = 1) const;

	/// sums together all but the given dimensions and returns a 2d image of these sums
	GPUData axisMax( size_t dim1 ) const;

	/// calculates the entropy along all but the given two dimensions
	GPUData entropyContract( size_t dim1 = 0, size_t dim2  = 1) const;

	/// normalizes this over all but the given 2 dimensions, such that the corresponding sum is all ones
	void normalize( size_t dim1 = 0, size_t dim2 = 0 );

	/**
	 *  clamps all values to min max
	 */
	void clamp(float min, float max);

	/**
	 * computes the cosinus
	 * this[i] = cos(a[i])
	 */
	void cos(const GPUData& a);
	/**
	 * computes the sinus
	 * this[i] = sin(a[i])
	 */
	void sin(const GPUData& a);
	/*
	 * compute the arctan2 and stores the result in this dataset
	 */
	void atan2(const GPUData& y, const GPUData& x);
	/*
	 * compute the length of the hypotenuse
	 * this[i] = sqrt(a[i]**2 + b[i]**2)
	 */

	/*
	 * Binary Median. Only for BW-Images
	 */
	void medianOnBinary(const GPUData& src, int range);

	/*
	 * 3x3 Median filter
	 */
	void median3x3 (const GPUData &src);


	void pythagoras(const GPUData& a, const GPUData& b);
	/**
	 * computes the Naka-Rusthon equation with exponent 2
	 * \f$ a[i] = { a[i]^2 \over a[i]^2 + n_{50}^2} \cdot resultFaktor \f$
	 * while preserving the original sign
	 */
	void nakarushton2(const float n50, const float resultFaktor = 1.0f);
	/**
	 * computes the Naka-Rusthon equation with exponent 2
	 * \f$ res[i] = { a[i]^2 \over a[i]^2 + n_{50}^2} \cdot resultFaktor \f$
	 * while preserving the original sign
	 */
	void nakarushton2(GPUData& res, const float n50, const float resultFaktor = 1.0f) const;
	/**
	 * This method performs a simple one dimensional convolution
	 * over the given dimension.
	 * \param k		The kernel to convolute with (its size have to be odd)
	 * \param dim	The dimension to convolute with (0 = slowest, left most, index)
	 * \param dst	The GPUData where to the result will be stored
	 * \param type  (0:padded with 0, 1:cyclic, 2:replicate)
	 */
	void convolute1D(GPUData& k, size_t dim, GPUData& dst, int type=2) const;
	/**
	 * Deriche recursive filters to compute convolution with
	 * a Gaussian or its derivatives.
	 *
	 * \paran sigma The filter extent.
	 * \param dim   The dimension along which to filter.
	 * \param order Use 0 to blur the image, and 1 or 2 to
	 *              compute its first or second derivative.
	 * \param dst   An array to store the result.
	 * \return A reference to the resulting array.
	 */
	GPUData dericheFilter(float sigma, size_t dim, int order, GPUData& dst) const;
	GPUData dericheFilter(float sigma, size_t dim, int order) const;
	/**
	 * This function computes one explicit euler step for a set of first
	 * order, linear DGLs of the following type:
	 *
	 * \f[ \dot{x}_{i\,j} = x_{i\,j} \cdot
	 * 			\left( a \cdot A_{i\,j} + b \cdot B_i + C \right) +
	 * 			\left( d \cdot D_{i\,j} + e \cdot E_i + F \right) \f]
	 *
	 * Therein is \f$ i \in \left[1..row\right] \f$ and
	 * \f$ j \in \left[1..col\right] \f$.
	 */
	void linFirstOrderODEEulerStep( const float a, const GPUData& A,
									const float b, const GPUData& B,
									const float C,
									const float d, const GPUData& D,
									const float e, const GPUData& E,
									const float F,
									const float dt );
	/**
	 * This function computes one implicit euler step
	 * (\f$ x_{n+1} = x_n+\delta t \; \dot{x}_{n+1}\f$) for a set of first
	 * order, linear DGLs of the following type:
	 *
	 * \f[ \dot{x}_{i\,j} = x_{i\,j} \cdot
	 * 			\left( a \cdot A_{i\,j} + b \cdot B_i + C \right) +
	 * 			\left( d \cdot D_{i\,j} + e \cdot E_i + F \right) \f]
	 *
	 * Therein is \f$ i \in \left[1..row\right] \f$ and
	 * \f$ j \in \left[1..col\right] \f$.
	 */
	void linFirstOrderODEImplicitEulerStep( const float a, const GPUData& A,
											const float b, const GPUData& B,
											const float C,
											const float d, const GPUData& D,
											const float e, const GPUData& E,
											const float F,
											const float dt);
	/**
	 * This function computes the steady state solution
	 * (\f$ \dot{x}=0\f$) for a set of first order, linear
	 * DGLs of the following type:
	 *
	 * \f[ \dot{x}_{i\,j} = x_{i\,j} \cdot
	 * 			\left( a \cdot A_{i\,j} + b \cdot B_i + C \right) +
	 * 			\left( d \cdot D_{i\,j} + e \cdot E_i + F \right) \f]
	 *
	 * Therein is \f$ i \in \left[1..row\right] \f$ and
	 * \f$ j \in \left[1..col\right] \f$.
	 */
	void linFirstOrderODESteayStateSolution(const float a, const GPUData& A,
											const float b, const GPUData& B,
											const float C,
											const float d, const GPUData& D,
											const float e, const GPUData& E,
											const float F);

	/**
	 * This function computes one euler step for a set of first order, linear
	 * DGLs of the following type:
	 *
	 * \f[ \dot{x}_{i\,j} = -a \cdot x_{i\,j} + \left( 1 - x_{i\,j} \right)
	 * 			\left( b \cdot B_{i\,j} + c \cdot C_i \right) \f]
	 *
	 * Therein is \f$ i \in \left[1..row\right] \f$ and
	 * \f$ j \in \left[1..col\right] \f$.
	 *
	 * If parameter \e rect_input is \e true, the whole input sum is rectified.
	 * Thus, the evaluated ODE is:
	 *
	 * \f[ \dot{x}_{i\,j} = -a \cdot x_{i\,j} + \left( 1 - x_{i\,j} \right)
	 * 			\left[ b \cdot B_{i\,j} + c \cdot C_i \right]^+ \f]
	 */
	void linFirstOrderODEEulerStep(const float a,
										const float b, const GPUData& B,
										const float c, const GPUData& C,
										const float dt,
										const bool rect_input = false);
	/**
	 * This function computes one implicit euler step (\f$ x_{n+1} = x_n+\delta t \; \dot{x}_{n+1}\f$)
	 * for a set of first order, linear DGLs of the following type:
	 *
	 * \f[ \dot{x}_{i\,j} = -a \cdot x_{i\,j} + \left( 1 - x_{i\,j} \right)
	 * 			\left( b \cdot B_{i\,j} + c \cdot C_i \right) \f]
	 *
	 * Therein is \f$ i \in \left[1..row\right] \f$ and
	 * \f$ j \in \left[1..col\right] \f$.
	 *
	 * If parameter \e rect_input is \e true, the whole input sum is rectified.
	 * Thus, the evaluated ODE is:
	 *
	 * \f[ \dot{x}_{i\,j} = -a \cdot x_{i\,j} + \left( 1 - x_{i\,j} \right)
	 * 			\left[ b \cdot B_{i\,j} + c \cdot C_i \right]^+ \f]
	 */
	void linFirstOrderODEImplicitEulerStep(const float a,
										const float b, const GPUData& B,
										const float c, const GPUData& C,
										const float dt,
										const bool rect_input = false);
	/**
	 * This function computes the steady state solution
	 * (\f$ \dot{x}=0\f$) for a set of first order, linear
	 * DGLs of the following type:
	 *
	 * \f[ \dot{x}_{i\,j} = -a \cdot x_{i\,j} + \left( 1 - x_{i\,j} \right)
	 * 			\left( b \cdot B_{i\,j} + c \cdot C_i \right) \f]
	 *
	 * Therein is \f$ i \in \left[1..row\right] \f$ and
	 * \f$ j \in \left[1..col\right] \f$.
	 *
	 * If parameter \e rect_input is \e true, the whole input sum is rectified.
	 * Thus, the evaluated ODE is:
	 *
	 * \f[ \dot{x}_{i\,j} = -a \cdot x_{i\,j} + \left( 1 - x_{i\,j} \right)
	 * 			\left[ b \cdot B_{i\,j} + c \cdot C_i \right]^+ \f]
	 */
	void linFirstOrderODESteayStateSolution(const float a,
											const float b, const GPUData& B,
											const float c, const GPUData& C,
											const bool rect_input = false);



	/// binary funs
	/// performs a + b
	GPUData & add( const GPUData & other );
	/// berforms b
	GPUData & assign( const GPUData & other );
	/// performs a * b
	GPUData & mul( const GPUData & other );
	/// performs a / b
	GPUData & div( const GPUData & other );
	/// performs a - b
	GPUData & sub( const GPUData & other );
	/// performs a + b^2
	GPUData & add_squared( const GPUData & other );
	/// performs a + |b|
	GPUData & add_abs( const GPUData & other );
	/// performs max(a,b)
	GPUData & max( const GPUData & other );
	/// performs min(a,b)
	GPUData & min( const GPUData & other );
	/// binary one arg funs
	/// performs a + b
	GPUData & add( const GPUData & other, float s );
	/// performs a * ( 1 + c * b )
	GPUData & fb_mod( const GPUData & other, float c );
	/// performs a * b
	GPUData & mul( const GPUData & other, float s );
	/// performs a + |b|
	GPUData & add_squared( const GPUData & other, float s  );
	/// performs a + |b|
	GPUData & add_abs( const GPUData & other, float s  );
	/// local one arg funs
	/// performs a + b
	GPUData & add( float a0 );
	/// performs a * b
	GPUData & mul( float a0 );
	/// performs a / b
	GPUData & div( float a0 );
	/// performs a - b
	GPUData & sub( float a0 );
	/// performs fpow
	GPUData & pow( float a0 );
	/// performs clipped log
	GPUData & safe_log( float a0 = 1e-20f );
	/// performs fmodf
	GPUData & mod( float a0 );
	/// sets a = a0+rect(a-a0)
	GPUData & rect( float a0 );
  /// does clamp( a*a0, 0, 1 )
	GPUData & compress( float a0 );
  /// does min(a, a0)
	GPUData & min( float a0 );
  /// does max(a, a0)
	GPUData & max( float a0 );
  /// does uses a sigmoid function ( x/sqrt(1+x^2) ) for clamping
	GPUData & sigmoid( float a0 );
	/// performs fmodf
	GPUData & modulo(const float f=3.14159265359f);
	/// local funs
	/// sets a = |a|
	GPUData & abs();
	/// sets a = sin(a)
	GPUData & sin();
	/// sets a = cos(a)
	GPUData & cos();
	/// sets a = asin(a)
	GPUData & asin();
	/// sets a = acos(a)
	GPUData & acos();
	/// sets a = exp(a)
	GPUData & exp();
	/// sets a = exp2(a)
	GPUData & exp2();
	/// sets a = log(a)
	GPUData & log();
	/// sets a = log2(a)
	GPUData & log2();
	/// sets a = rint(a)
	GPUData & rint();
	/// sets a = isnan(a)
	GPUData & isnan();
	/// sets a = sqr(a)
	GPUData & sqr();
	/// sets a = sqrt(a)
	GPUData & sqrt();
	/// sets a = rsqrt(a)
	GPUData & rsqrt();
	/// sets a = rect(a)
	GPUData & rect();

	int refcount() const { return inner ? inner->get_refcount() : -10 ; }

	//void * innerdata() { return (void*)inner; }

	/// copies the whole ( made contigous ) data to a 3d texture and deletes data. remember that you cannot write into array, just read (but thats fast)
	void move_to_array( bool features_first = false );
	/// copies the array back to modifieable data
	void move_to_data();
	/// true if array is filled. If you want to modify data, invoke move_to_data first
	bool is_array( ) const;
	
	cudaArray* array() const 
	{
		return inner ? inner->array : NULL;
	}
	
	float* data(bool move = false) const;
	float* data(bool move = true);
	
	cufftHandle * get_fft_plan( plan_key p_key ) const;

	static void free_pool();

protected:

	/// construction helper
	void init( size_t ndimA, const size_t* dimsA, const std::vector<size_t> & o );
	void init(size_t ndim, const size_t* dims);
	/// heuristic to suggest a dimension to recurse over, two dimensions can be excluded from the search
	size_t suggest_recursion_dim(size_t exclude1 = -1, size_t exclude2 = -1) const;
	/**
	 * Shared reference to allocated GPU memory.
	 * Each instance is reference-counted by the GPUData
	 * instances.
	 */
    struct InnerData
    {
			cudaArray* array;
		  float* data;
		  size_t size; // byte count!
		  int retain();
		  int release();
		  InnerData();
		  InnerData(size_t n);
		  ~InnerData();
      int get_refcount() const{return refcount;}

	  private:

		  int refcount;

    };

	/**
	 * Reference counter pointer to the allocated memory.
	 */
	InnerData* inner;
	/// query the fft plan map, which caches fft-plans

	// static members
	static boost::unordered_map<plan_key,cufftHandle*>& getPlanMap();
    static bool has_plan( const plan_key & key );


private :

  GPUData dimsSorted();


	typedef void (*local_binary_kernel_entry)(float* data, int inc1, int inc2, int inc3, int inc4,
											  int N1, int N2, int N3, int N4,
											  const float* op, int inc1o, int inc2o, int inc3o, int inc4o);

	typedef void (*local_binary_kernel_a0_entry)(float a0, float* data, int inc1, int inc2, int inc3, int inc4,
											  int N1, int N2, int N3, int N4,
											  const float* op, int inc1o, int inc2o, int inc3o, int inc4o);

	void localBinaryFunc( local_binary_kernel_entry e, const GPUData & opx );

  void localBinaryFunc_a0( local_binary_kernel_a0_entry entry, const GPUData & op, float a0);

	void localFunc(void (*func_pointer)(float* data, int inc1, int inc2, int inc3, int inc4, int N1, int N2, int N3, int N4));

  void localFunc_a0(void (*func_pointer)(float a0, float* data, int inc1, int inc2, int inc3, int inc4, int N1, int N2, int N3, int N4), float a0);

};

inline size_t hash_value(const GPUData& d) {
	return    d.size()*2-d.getNoOfDimensions()*3+d.extent(0)+d.extent(d.getNoOfDimensions()-1)
			* size_t(d.sum()) + size_t(d.max()*10);
}

struct GPUDataIsEqualTo : public std::equal_to<GPUData> {
	bool operator()( const GPUData& a, const GPUData& b ) const;
};

/** Duplicate the given array using new. */
template <typename T>
T* memdup(const T* src, int count)
{
    T* dst = new T[count];
	std::copy( src, src + count, dst );
    return dst;
}

#endif


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
