/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/

#include "GPUTextureData2D.hpp"
#include "cutil.h"
#include "Exceptions.hpp"

#include <iostream>
#include "boost/format.hpp"

GPUTextureData2D::GPUTextureData2D() 
  : _innerData( new InnerTexture2DData() ) 
{
}

GPUTextureData2D::GPUTextureData2D( int n1, int n2 ) 
  : _innerData( new InnerTexture2DData( n1, n2 ) ) 
{
}

GPUTextureData2D::GPUTextureData2D( const GPUDataBase & d ) 
  : _innerData( new InnerTexture2DData( d.extent(0), d.extent(1) ) )
{
  writeGPUData( d );
}

GPUTextureData2D::GPUTextureData2D( const GPUTextureData2D& d )
  : _innerData( d._innerData )
{
  d._innerData->increaseCounter();
}

void GPUTextureData2D::clear() 
{
  if ( _innerData ) 
	{
	  if ( _innerData->decreaseCounter() == 0 )
		{
		  delete _innerData;
		  _innerData=0;
		}
	} 
}

GPUTextureData2D::~GPUTextureData2D() 
{
  clear();
}

GPUTextureData2D& GPUTextureData2D::operator=(const GPUTextureData2D& d) 
{
  reference(d);
  return *this;
}

void GPUTextureData2D::reference(const GPUTextureData2D& d) 
{
  if ( this == &d ) 
	return;
  clear();
  _innerData = d._innerData;
  if ( _innerData ) 
	_innerData->increaseCounter();
}

int GPUTextureData2D::getSize() 
{
  if (_innerData!=0) return _innerData->size;
  else return 0;
}

void GPUTextureData2D::fill(const float val)
{
  if( _innerData )
	_innerData->fill(val);
}

/*
GPUData GPUTextureData2D::read()
{
  GPUData dat( height(), width() );
  if( _innerData )
	_innerData->read(dat.getDataPointer());
  return dat;
}
*/
void GPUTextureData2D::bindTexture() const
{
  if( _innerData )
	_innerData->bindTexture();
}

void GPUTextureData2D::writeGPUData(const GPUDataBase& val)
{
  if ( val.ndim != 2 ) 
	throw GPUDataException( str( boost::format("trying to make 2d texture from array of rank %i") % val.ndim ) );
  if( width() != val.extent(1) || height() != val.extent(0) )
	reference( GPUTextureData2D( val.extent(0), val.extent(1) ) );
  _innerData->writeGPUData( val.getDataPointer(), val.getDimensions() );
}

extern "C" void conv2DFromTexture(InnerTexture2DData *d_in, float *d_kernel, int kernel_x, int kernel_y, float *d_erg);
void GPUTextureData2D::conv2D(const GPUDataBase& kernel, GPUDataBase& dest)
{
  const int kernel_w = kernel.extent(1);
  const int kernel_h = kernel.extent(0);
  conv2DFromTexture( _innerData, kernel.getDataPointer(), kernel_w, kernel_h, dest.getDataPointer() );
  GPUDataException::check_error("conv2D");
}

extern "C" void conv2DFromTexture2(InnerTexture2DData *d_in, InnerTexture2DData *d_kernel, float *d_erg);
void GPUTextureData2D::conv2D(const GPUTextureData2D& kernel, GPUDataBase& dest)
{
  conv2DFromTexture2( _innerData, kernel._innerData, dest.getDataPointer() );
  GPUDataException::check_error("conv2D2");
}

extern "C" void conv2DFromTexture3(InnerTexture2DData *d_in, InnerTexture2DData *d_kernel, int depth, float *d_erg);
void GPUTextureData2D::conv2D3(const GPUTextureData2D& kernel, int depth, GPUDataBase& dest)
{
  conv2DFromTexture3( _innerData, kernel._innerData, depth, dest.getDataPointer() );
  GPUDataException::check_error("conv2D3");
}
