/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include <boost/thread.hpp>
#include <boost/format.hpp>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <cuda_runtime.h>
#include <stdexcept>
#include <list>
#include <map>
#include <boost/any.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/locks.hpp>

struct cuda_context_handler
{

  static cuda_context_handler & get()
  {
	cuda_context_handler * instance = _tls_instance.get();
	if ( instance == NULL )
	  _tls_instance.reset( instance = new cuda_context_handler() );
	return *instance;
  }

  void push( CUcontext & ctx )
  {
	_cuda_context_stack.push_back( ctx );
	cuCtxPushCurrent( ctx );
  }

  CUcontext pop()
  {
	CUcontext ctx;
	cuCtxPopCurrent( &ctx );
	_cuda_context_stack.pop_back();
	if ( !_cuda_context_stack.empty() && current() != 0 && ctx != current() )
	  throw std::runtime_error(str( boost::format("cuda context handler out of sync local: %x current: %x ") % current() % ctx ) );
	return ctx;
  }

  CUcontext current()
  {
	if ( _cuda_context_stack.empty() )
	  throw std::runtime_error( "no cuda context!" );
	return _cuda_context_stack.back();
  }

private:

  cuda_context_handler()
  {
	_cuda_context_stack.push_back(0);
  }

  std::list<CUcontext> _cuda_context_stack;
  static boost::thread_specific_ptr<cuda_context_handler> _tls_instance;
  typedef std::list<boost::any> resource_list;
};

// construct to equip any object pointer with a virtual destructor
struct cuda_context_specific_resource_base
{
  cuda_context_specific_resource_base( void * p ) : _p(p) {}
  virtual ~cuda_context_specific_resource_base();
  void * get_ptr() { return _p; }
private:
  void * _p;
};
template<typename T>
struct cuda_context_specific_resource : public cuda_context_specific_resource_base
{
  cuda_context_specific_resource( T * p )
	: cuda_context_specific_resource_base(p)
	,_p(p)
  {}
  virtual ~cuda_context_specific_resource()
  {
	if (_p)
	  {
		delete _p;
	  }
  }
private:
  T * _p;
};

void add_cuda_context_specific_resource_internal( cuda_context_specific_resource_base * p );

template<typename T>
void add_cuda_context_specific_resource( T * p )
{
  add_cuda_context_specific_resource_internal( new cuda_context_specific_resource<T>( p ) );
}

void remove_cuda_context_specific_resource( void * p );

void delete_cuda_context_sepcific_resources();

// todo: make thread-safe
template<typename T>
struct cuda_context_specific_ptr
{

  T * get()
  {
	boost::mutex::scoped_lock lock(_mutex);
	CUcontext ctx = cuda_context_handler::get().current();
	typename instance_map_t::iterator i = _instances.find( ctx );
	if ( i == _instances.end() )
	  return NULL;
	else
	  return i->second;
  }

  void set( T * t = 0, bool autodelete = true )
  {
	void * t_old = get();
	if ( t_old )
	  remove_cuda_context_specific_resource( t_old );
	CUcontext ctx = cuda_context_handler::get().current();
	// avoid storint NULL pointers for beauty...
	if ( t != 0 )
	  {
		{
		  boost::mutex::scoped_lock lock(_mutex);
		  _instances[ctx] = t;
		}
		if ( autodelete )
		  {
			add_cuda_context_specific_resource( t );
		  }
	  }
	else
	  {
		boost::mutex::scoped_lock lock(_mutex);
		_instances.erase( ctx );
	  }
  }

private:

  mutable boost::mutex _mutex;
  typedef std::map<CUcontext,T*> instance_map_t;
  instance_map_t _instances;

};

