/*

GPUDATA

  THIS MODULE PROVIDES MULTI-DIMENSIONAL MATRIX STORAGE 
  AND OPERATIONS ON GRAPHICAL PROCESSING UNITS (GPU)
  BY MAKING USE OF CUDA TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "cufft-kernel-key.hpp"
#include <boost/lexical_cast.hpp>

plan_key::plan_key() {
	dim = 0;
	dims= 0;
	type= CUFFT_C2C;
}
plan_key::~plan_key() {
	if (dims!=0) delete[] dims;
}
plan_key::plan_key(const plan_key& k) {
	dim = k.dim;
	dims= new int[dim];
	type= k.type;
	std::copy( k.dims, k.dims + dim, dims );
}
plan_key::plan_key(cufftType_t type_, int dim_, int* dims_) {
	 dim = dim_;
	 dims= new int[dim];
	 type= type_;
	 std::copy( dims_, dims_ + dim, dims );
}
plan_key::plan_key(cufftType_t type_, int z, int y, int x) {
	 dim = 3;
	 dims= new int[dim];
	 dims[0] = z;
	 dims[1] = y;
	 dims[3] = x;
	 type= type_;
}
plan_key::plan_key(cufftType_t type_, int h, int w) {
	 dim = 2;
	 dims= new int[dim];
	 dims[0] = h;
	 dims[1] = w;
	 type= type_;
}
plan_key::plan_key(cufftType_t type_, int size) {
	 dim = 1;
	 dims= new int[dim];
	 dims[0] = size;
	 type= type_;
}

/**
 * factory method
 */
cufftHandle* plan_key::create_plan() {
	cufftHandle* rtv = new cufftHandle;
	cufftResult  cufft_return;
	switch(dim) {
	case 1: cufft_return = cufftPlan1d(rtv, dims[0], type, 1); break;
	case 2: cufft_return = cufftPlan2d(rtv, dims[0], dims[1], type); break;
	case 3: cufft_return = cufftPlan3d(rtv, dims[0], dims[1], dims[2], type); break;
	default:
		throw std::runtime_error("plan_key::create_plan() couldn't create a plan of this type");
	}
	if (cufft_return!=CUFFT_SUCCESS)
		throw GPUDataException("error setting up the cufft plan in plan_key::create_plan()");
	return rtv;
}
void plan_key::destroy_plan(cufftHandle* p) {
	cufftResult  cufft_return  = cufftDestroy(*p);
	if (cufft_return!=CUFFT_SUCCESS)
		throw GPUDataException("error destroying cufft plan in plan_key::destroy_plan()");
}

std::string plan_key::dump() const
{
    std::string result("[ ");
	if (dim==0) return (result+"]");
	for ( size_t i = 0; i < dim - 1; i++ )
    {
		result.append(   boost::lexical_cast<std::string>( dims[i] )+ " x " );
	}
    result.append(     boost::lexical_cast<std::string>( dims[ dim - 1 ] ) + " ]" );
	result.append("(type:" + boost::lexical_cast<std::string>( type ) + ") " );
	return result;
}

bool plan_key::operator==(const plan_key& k) const 
{
  return type==k.type && dim==k.dim && ( memcmp(dims,k.dims,sizeof(int)*dim) == 0 );
}

plan_key& plan_key::operator=(const plan_key& k) 
{
	this->~plan_key();

	type = k.type;
	dim  = k.dim;
	dims = new int[dim];
	memcpy(dims,k.dims,sizeof(int)*dim);
	return *this;
}


size_t hash_value(const plan_key& k) {
	int size = 1;
	for (int i=0; i<k.dim; ++i) {
		size *= k.dims[i];
	}
	return k.dim*size+size_t(k.type);
}
