/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "cutil.h"
#include <cuda.h>
#include "float.h"
#include "stdio.h"

#ifdef _WIN32
#define M_PI 3.14159265
#endif

static texture<float, 2, cudaReadModeElementType> tex;

__global__ void d_R_HOG_grey_tex( float * out, int n1_cells, int n2_cells
								  , int cell_size, int block_size
								  , int block_overlap, int n_orientations
								  , bool interpolate, float alpha
								  , float xscale, float yscale )
{
  const int i1_block  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2_block  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  const int block_stride = ( block_size - block_overlap ) * cell_size;
  const int n1_blocks = n1_cells / block_size;
  const int n2_blocks = n2_cells / block_size;
  if ( i1_block >= n1_blocks ) return;
  if ( i2_block >= n2_blocks ) return;
  const float dx_pixel = 1.0f / xscale;
  const float dy_pixel = 1.0f / yscale;
  const float x0_pixel = float(i2_block * block_stride) * dx_pixel;
  const float y0_pixel = float(i1_block * block_stride) * dy_pixel;
  const int i01_cell = i1_block * block_size;
  const int i02_cell = i2_block * block_size;

  extern __shared__ float shared_mem[];
  float * histogram_buffer = shared_mem + n_orientations * ( threadIdx.x + blockDim.x * threadIdx.y );

#define cell_index(i1,i2) ( ( i01_cell + i1 ) * n2_cells + ( i02_cell + i2 ) )

  // loop over cells
  float Z = 0;
  for ( int i1_cell = 0; i1_cell < block_size; ++i1_cell )
        for ( int i2_cell = 0; i2_cell < block_size; ++i2_cell )
          {
                float * histogram_out = out + cell_index( i1_cell, i2_cell ) * n_orientations;
                //histogram_buffer = histogram_out;
                for ( int o = 0; o < n_orientations; ++o )
                  {
                        histogram_buffer[o] = 0.0f;
                  }
                // loop over pixels in the cell
                for ( int i1_pixel = 0; i1_pixel < cell_size; ++i1_pixel )
                  for ( int i2_pixel = 0; i2_pixel < cell_size; ++i2_pixel )
                        {
						  float x_pixel = x0_pixel + dx_pixel * ( i2_cell * cell_size + i2_pixel );
						  float y_pixel = y0_pixel + dy_pixel * ( i1_cell * cell_size + i1_pixel );
                          float d1 = tex2D( tex, x_pixel, y_pixel + dy_pixel ) - tex2D( tex, x_pixel, y_pixel - dy_pixel );
						  float d2 = tex2D( tex, x_pixel + dx_pixel, y_pixel ) - tex2D( tex, x_pixel - dx_pixel, y_pixel );
                          float weight = sqrtf( d1 * d1 + d2 * d2 );
						  float angle = atan2( d1, d2 ) + float( M_PI );
						  float f_iangle = angle / float(M_PI) * n_orientations;
						  if ( interpolate )
							{
							  int iangle1 = int( f_iangle        );
							  int iangle2 = int( f_iangle + 0.5f );
							  float d = f_iangle - float( iangle1 );
							  histogram_buffer[iangle1 % n_orientations] += ( 1 - d ) * weight;
							  histogram_buffer[iangle2 % n_orientations] += (     d ) * weight;
							}
						  else
							{
							  int iangle = int( f_iangle ) % n_orientations;
							  histogram_buffer[iangle] += weight;
							}
                        }
                // write histogram
                for ( int o = 0; o < n_orientations; ++o )
                  {
                        float weight = histogram_buffer[o];
                        Z += weight * weight;
                        histogram_out[o] = weight;
                  }
          }

  // block_normalize
  Z = sqrt( Z );
  if ( Z < 1e-5f )
        {
          const float fillval = 1.0f / sqrt( float( n_orientations * block_size * block_size ) );
          for ( int i1_cell = 0; i1_cell < block_size; ++i1_cell )
                for ( int i2_cell = 0; i2_cell < block_size; ++i2_cell )
                  {
                        float * histogram_out = out + cell_index( i1_cell, i2_cell ) * n_orientations;
                        for ( int o = 0; o < n_orientations; ++o )
                          {
                                histogram_out[o] = fillval;
                          }
                  }
        }
  else
        {
          float C = 1.0f / ( alpha * Z );
          for ( int i1_cell = 0; i1_cell < block_size; ++i1_cell )
                for ( int i2_cell = 0; i2_cell < block_size; ++i2_cell )
                  {
                        float * histogram_out = out + cell_index( i1_cell, i2_cell ) * n_orientations;
                        for ( int o = 0; o < n_orientations; ++o )
                          {
							histogram_out[o] = min( 1.0f, histogram_out[o] * C );
                          }
                  }
        }
}

// naive kernel for l2 normalized unsigned gradient R-HOG on gray images
// computes exactly one block
// does not interpolate spatially
__global__ void d_R_HOG_grey( const float * in, int in_n1, int in_n2
							  , float * out, int cell_size, int block_size
							  , int block_overlap, int n_orientations
							  , bool interpolate, float alpha )
{
  const int i1_block  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2_block  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  const int block_stride = ( block_size - block_overlap ) * cell_size;
  const int block_extent = block_size * cell_size;
  const int n1_blocks = 1 + ( in_n1 - block_extent - 2 ) / block_stride;
  const int n2_blocks = 1 + ( in_n2 - block_extent - 2 ) / block_stride;
  const int n1_cells = n1_blocks * block_size;
  const int n2_cells = n2_blocks * block_size;
  if ( i1_block >= n1_blocks ) return;
  if ( i2_block >= n2_blocks ) return;
  const int i01_pixel = i1_block * block_stride;
  const int i02_pixel = i2_block * block_stride;
  const int i01_cell = i1_block * block_size;
  const int i02_cell = i2_block * block_size;

  extern __shared__ float shared_mem[];
  float * histogram_buffer = shared_mem + n_orientations * ( threadIdx.x + blockDim.x * threadIdx.y );

#define cell_index(i1,i2) ( ( i01_cell + i1 ) * n2_cells + ( i02_cell + i2 ) )

  // loop over cells
  float Z = 0;
  for ( int i1_cell = 0; i1_cell < block_size; ++i1_cell )
        for ( int i2_cell = 0; i2_cell < block_size; ++i2_cell )
          {
                float * histogram_out = out + cell_index( i1_cell, i2_cell ) * n_orientations;
                //histogram_buffer = histogram_out;
                for ( int o = 0; o < n_orientations; ++o )
                  {
                        histogram_buffer[o] = 0.0f;
                  }
                // loop over pixels in the cell
                for ( int i1_pixel = 0; i1_pixel < cell_size; ++i1_pixel )
                  for ( int i2_pixel = 0; i2_pixel < cell_size; ++i2_pixel )
                        {
                          int i1_center = 1 + i01_pixel + i1_cell * cell_size + i1_pixel;
                          int i2_center = 1 + i02_pixel + i2_cell * cell_size + i2_pixel;                         
                          const float * center = in + i1_center * in_n2 + i2_center;
                          float d1 = ( center[-in_n2] - center[in_n2] );
                          float d2 = ( center[-1    ] - center[1    ] );
                          float weight = sqrtf( d1 * d1 + d2 * d2 );
						  float angle = atan2( d1, d2 ) + float( M_PI );
						  float f_iangle = angle / float(M_PI) * n_orientations;
						  if ( interpolate )
							{
							  int iangle1 = int( f_iangle        );
							  int iangle2 = int( f_iangle + 0.5f );
							  float d = f_iangle - float( iangle1 );
							  histogram_buffer[iangle1 % n_orientations] += ( 1 - d ) * weight;
							  histogram_buffer[iangle2 % n_orientations] += (     d ) * weight;
							}
						  else
							{
							  int iangle = int( f_iangle ) % n_orientations;
							  histogram_buffer[iangle] += weight;
							}
                        }
                // write histogram
                for ( int o = 0; o < n_orientations; ++o )
                  {
                        float weight = histogram_buffer[o];
                        Z += weight * weight;
                        histogram_out[o] = weight;
                  }
          }

  // block_normalize
  Z = sqrt( Z );
  if ( Z < 1e-5f )
        {
          const float fillval = 1.0f / sqrt( float( n_orientations * block_size * block_size ) );
          for ( int i1_cell = 0; i1_cell < block_size; ++i1_cell )
                for ( int i2_cell = 0; i2_cell < block_size; ++i2_cell )
                  {
                        float * histogram_out = out + cell_index( i1_cell, i2_cell ) * n_orientations;
                        for ( int o = 0; o < n_orientations; ++o )
                          {
                                histogram_out[o] = fillval;
                          }
                  }
        }
  else
        {
          float C = 1.0f / ( alpha * Z );
          for ( int i1_cell = 0; i1_cell < block_size; ++i1_cell )
                for ( int i2_cell = 0; i2_cell < block_size; ++i2_cell )
                  {
                        float * histogram_out = out + cell_index( i1_cell, i2_cell ) * n_orientations;
                        for ( int o = 0; o < n_orientations; ++o )
                          {
							histogram_out[o] = min( 1.0f, histogram_out[o] * C );
                          }
                  }
        }
}


// 1. do your cell, collecting gradient hist and square sum, write the square sums to the right neighbors
// 2. produce the feature including truncation

// felzenszwalb features, based on R_HOG but more compact
__global__ void d_felzenszwalb_RHOG_grey_step1( const float * in, int in_n1, int in_n2, float * out
											   , int cell_size, int n_orientations
											   , bool contrast_sensitive )
{
  const int i1_cell = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2_cell = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  const int n1_cells = ( in_n1 - 2 ) / cell_size;
  const int n2_cells = ( in_n2 - 2 ) / cell_size;
  const int i01_pixel = i1_cell * cell_size + 1;
  const int i02_pixel = i2_cell * cell_size + 1;

  if ( i1_cell >= n1_cells || i2_cell >= n2_cells )
	return;

  extern __shared__ float shared_mem[];
  const size_t histogram_size = n_orientations;
  float * histogram_buffer = shared_mem + histogram_size * ( threadIdx.x + blockDim.x * threadIdx.y );
  for ( int i = 0; i < histogram_size; ++i ) histogram_buffer[i] = 0.0f;

  // step 1: collect histogram and square sum
  float L2 = 0;
  // loop over pixels in the cell
  for ( int i1_pixel = 0; i1_pixel < cell_size; ++i1_pixel )
	for ( int i2_pixel = 0; i2_pixel < cell_size; ++i2_pixel )
	  {
		int i1_center = i01_pixel + i1_pixel;
		int i2_center = i02_pixel + i2_pixel;			  
		const float * center = in + i1_center * in_n2 + i2_center;
		float d1 = ( center[-in_n2] - center[in_n2] );
		float d2 = ( center[-1    ] - center[1    ] );
		float angle = atan2( d1, d2 );
		float weight = sqrtf( d1 * d1 + d2 * d2 );
		float f_iangle;
		if ( contrast_sensitive )
		  f_iangle = ( 1.0f + angle / M_PI ) / 2 * n_orientations;
		else
		  f_iangle = ( 1.0f + angle / M_PI ) * n_orientations;
		float d_iangle = f_iangle - int( f_iangle );
		int iangle1 = int( f_iangle ) % n_orientations; // floor
		int iangle2 = int( f_iangle + 0.5f ) % n_orientations; // ceil
		histogram_buffer[iangle1] += ( 1 - d_iangle ) * weight;
		histogram_buffer[iangle2] += (     d_iangle ) * weight;
		L2 += weight * weight;
	  }

  // write orientation histogram to main mem if not on edge
  int n_features = n_orientations + 4;
  int out_stride1 = ( n2_cells - 2 ) * n_features;
  int out_stride2 = n_features;
  float * out_fvec = out + ( ( i1_cell - 1 ) * ( n2_cells - 2 ) + ( i2_cell - 1 ) ) * n_features;
  if ( i1_cell > 0 && i2_cell > 0 && ( i1_cell + 1 ) < n1_cells && ( i2_cell + 1 ) < n2_cells )
	{
	  for ( int o = 0; o < n_orientations; ++o )
		out_fvec[o] = histogram_buffer[o];
	  out_fvec[n_orientations] = L2;
	}

  // write normalizations for edge cells, only works if min( n1_cells, n2_cells ) >= 4
  bool top    = i1_cell == 0;
  bool bottom = i1_cell == n1_cells - 1;
  bool left   = i2_cell == 0;
  bool right  = i2_cell == n2_cells - 1;
  // l2-vec:
  // 0: self
  // 1: top/bottom
  // 2: left/right
  // 3: corners
  if ( top )
	{
	  if ( left )
		out_fvec[ out_stride1 + out_stride2 + n_orientations + 3 ] = L2; // top left: r+1 c+1 f+3
	  else if ( right )
		out_fvec[ out_stride1 - out_stride2 + n_orientations + 3 ] = L2; // top right: r+1 c-1 f+3
	  else
		out_fvec[ out_stride1 + n_orientations + 1 ] = L2; // top: r+1 c+0 f+1
	}
  else if ( bottom )
	{
	  if ( left )
		out_fvec[ -out_stride1 + out_stride2 + n_orientations + 3 ] = L2; // bottom left : r-1 c+1 f+3
	  else if ( right )
		out_fvec[ -out_stride1 - out_stride2 + n_orientations + 3 ] = L2; // bottom right: r-1 c-1 f+3
	  else
		out_fvec[ -out_stride1 + n_orientations + 1 ] = L2; // bottom: r-1 c+0 f+1
	}
  else if ( left )
	out_fvec[  out_stride2 + n_orientations + 2 ] = L2; // left : r+0 c+1 f+2
  else if ( right )
	out_fvec[ -out_stride2 + n_orientations + 2 ] = L2; // left : r+0 c-1 f+2
}

// collect block normalizations and complete features
__global__ void d_felzenszwalb_RHOG_grey_step2( float * out, int n1_cells, int n2_cells
											   , int n_orientations, float alpha, bool contrast_sensitive )
{
  const int i1_cell = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2_cell = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  if ( i1_cell >= n1_cells || i2_cell >= n2_cells )
	return;
  
  enum { UL = 0, UR = 1, LR = 2, LL = 3 };
  float block_norms[4] = { 0, 0, 0, 0 }; // upper left, upper right, lower right, lower left
  float energy_features[4] = { 0, 0, 0, 0 };
  bool top    = i1_cell == 0;
  bool left   = i2_cell == 0;
  bool bottom = i1_cell == n1_cells - 1;
  bool right  = i2_cell == n2_cells - 1;

  int n_features = n_orientations + 4; // todo: contrast sensitive, the felzenszwalb way
  int out_stride1 = n2_cells * n_features;
  int out_stride2 = n_features;
  float * out_fvec = out + i1_cell * out_stride1 + i2_cell * out_stride2;

  // collect block normalizations
  float topval, leftval, rightval, botval;
  if ( top )
	{
	  if ( left )
		block_norms[UL] += out_fvec[ n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[UL] += out_fvec[ -out_stride2 + n_orientations + 1 ]; // stored left f+1
	  if ( right )
		block_norms[UR] += out_fvec[ n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[UR] += out_fvec[  out_stride2 + n_orientations + 1 ]; // stored right f+1
	  topval =  out_fvec[ n_orientations + 1 ]; // stored locally f+1
	}
  else
	topval =  out_fvec[ -out_stride1 + n_orientations ];

  if ( bottom )
	{
	  if ( left )
		block_norms[LL] += out_fvec[ n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[LL] += out_fvec[ -out_stride2 + n_orientations + 2 ]; // stored left f+2
	  if ( right )
		block_norms[LR] += out_fvec[ n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[LR] += out_fvec[  out_stride2 + n_orientations + 2 ]; // stored right f+2
	}
  else
	botval =  out_fvec[ out_stride1 + n_orientations ];
	
  if ( left )
	leftval =  out_fvec[ n_orientations + 2 ];
  else
	{
	  leftval =  out_fvec[ -out_stride2 + n_orientations ];
	  if ( !top )
		block_norms[UL] += out_fvec[ -out_stride1 - out_stride2 + n_orientations ];
	  if ( !bottom )
		block_norms[LL] += out_fvec[  out_stride1 - out_stride2 + n_orientations ];
	}

  if ( right )
	rightval =  out_fvec[ n_orientations + 2 ];
  else
	{
	  rightval =  out_fvec[ out_stride2 + n_orientations ];
	  if ( !top )
		block_norms[UR] += out_fvec[ -out_stride1 + out_stride2 + n_orientations ];
	  if ( !bottom )
		block_norms[LR] += out_fvec[  out_stride1 + out_stride2 + n_orientations ];		  
	}

  float centerval = out_fvec[n_orientations];
  const float epsilon = 1e-5;
  block_norms[UL] += topval + leftval  + centerval;
  block_norms[LL] += botval + leftval  + centerval;
  block_norms[UR] += topval + rightval + centerval;
  block_norms[LR] += botval + rightval + centerval;
  for ( int b = 0; b < 4; ++b ) // block index
	block_norms[b] = 1.0f / ( alpha * sqrtf( block_norms[b] + epsilon ) );

  for ( int o = 0; o < n_orientations; ++o )
	{
	  float orientation_feature_in = out_fvec[o];
	  float orientation_feature_out = 0;
	  for ( int b = 0; b < 4; ++b ) // block index
		{
		  float T = min( 1.0f, orientation_feature_in * block_norms[b] );
		  orientation_feature_out += T;	  
		  energy_features[b] += T;
		}
	  out_fvec[o] = orientation_feature_out;
	}
  for ( int b = 0; b < 4; ++b )
	out_fvec[n_orientations+b] = energy_features[b];

}


// 1. do your cell, collecting gradient hist and square sum, write the square sums to the right neighbors
// 2. produce the feature including truncation

// felzenszwalb features, based on R_HOG but more compact
__global__ void d_felzenszwalb_HOG_grey_step1( const float * in, int in_n1, int in_n2, float * out
											   , int cell_size, int n_orientations )
{
  const int i1_cell = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2_cell = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  const int n1_cells = ( in_n1 - 2 ) / cell_size;
  const int n2_cells = ( in_n2 - 2 ) / cell_size;
  const int i01_pixel = i1_cell * cell_size + 1;
  const int i02_pixel = i2_cell * cell_size + 1;

  if ( i1_cell >= n1_cells || i2_cell >= n2_cells )
	return;

  extern __shared__ float shared_mem[];
  const size_t histogram_size = 3 * n_orientations;
  float * histogram_buffer = shared_mem + histogram_size * ( threadIdx.x + blockDim.x * threadIdx.y );
  for ( int i = 0; i < histogram_size; ++i ) histogram_buffer[i] = 0.0f;

  // step 1: collect histogram and square sum
  float L2 = 0;
  // loop over pixels in the cell
  for ( int i1_pixel = 0; i1_pixel < cell_size; ++i1_pixel )
	for ( int i2_pixel = 0; i2_pixel < cell_size; ++i2_pixel )
	  {
		int i1_center = i01_pixel + i1_pixel;
		int i2_center = i02_pixel + i2_pixel;			  
		const float * center = in + i1_center * in_n2 + i2_center;
		float d1 = ( center[-in_n2] - center[in_n2] );
		float d2 = ( center[-1    ] - center[1    ] );
		float angle = atan2( d1, d2 );
		float weight = sqrtf( d1 * d1 + d2 * d2 );
		float f_iangle_cs;
		float f_iangle_ci;
		f_iangle_cs = ( 1.0f + angle / M_PI ) * n_orientations;
		f_iangle_ci = ( 1.0f + angle / M_PI ) / 2 * n_orientations;
		float d_iangle_cs = f_iangle_cs - int( f_iangle_cs );
		float d_iangle_ci = f_iangle_ci - int( f_iangle_ci );
		int iangle1_cs = int( f_iangle_cs ) % ( 2 * n_orientations ); // floor
		int iangle2_cs = int( f_iangle_cs + 0.5f ) % ( 2 * n_orientations ); // ceil
		int iangle1_ci = int( f_iangle_ci ) % n_orientations; // floor
		int iangle2_ci = int( f_iangle_ci + 0.5f ) % n_orientations; // ceil
		histogram_buffer[iangle1_cs] += ( 1 - d_iangle_cs ) * weight;
		histogram_buffer[iangle2_cs] += (     d_iangle_cs ) * weight;
		histogram_buffer[ 2 * n_orientations + iangle1_ci ] += ( 1 - d_iangle_ci ) * weight;
		histogram_buffer[ 2 * n_orientations + iangle2_ci ] += (     d_iangle_ci ) * weight;
		L2 += weight * weight;
	  }

  // write orientation histogram to main mem if not on edge
  int n_features = n_orientations + 4;
  int out_stride1 = ( n2_cells - 2 ) * n_features;
  int out_stride2 = n_features;
  float * out_fvec = out + ( ( i1_cell - 1 ) * ( n2_cells - 2 ) + ( i2_cell - 1 ) ) * n_features;
  if ( i1_cell > 0 && i2_cell > 0 && ( i1_cell + 1 ) < n1_cells && ( i2_cell + 1 ) < n2_cells )
	{
	  for ( int o = 0; o < 3 * n_orientations; ++o )
		out_fvec[o] = histogram_buffer[o];
	  out_fvec[ 3 * n_orientations ] = L2;
	}

  // write normalizations for edge cells, only works if min( n1_cells, n2_cells ) >= 4
  bool top    = i1_cell == 0;
  bool bottom = i1_cell == n1_cells - 1;
  bool left   = i2_cell == 0;
  bool right  = i2_cell == n2_cells - 1;
  // l2-vec:
  // 0: self
  // 1: top/bottom
  // 2: left/right
  // 3: corners
  if ( top )
	{
	  if ( left )
		out_fvec[ out_stride1 + out_stride2 + 3 * n_orientations + 3 ] = L2; // top left: r+1 c+1 f+3
	  else if ( right )
		out_fvec[ out_stride1 - out_stride2 + 3 * n_orientations + 3 ] = L2; // top right: r+1 c-1 f+3
	  else
		out_fvec[ out_stride1 + 3 * n_orientations + 1 ] = L2; // top: r+1 c+0 f+1
	}
  else if ( bottom )
	{
	  if ( left )
		out_fvec[ -out_stride1 + out_stride2 + 3 * n_orientations + 3 ] = L2; // bottom left : r-1 c+1 f+3
	  else if ( right )
		out_fvec[ -out_stride1 - out_stride2 + 3 * n_orientations + 3 ] = L2; // bottom right: r-1 c-1 f+3
	  else
		out_fvec[ -out_stride1 + 3 * n_orientations + 1 ] = L2; // bottom: r-1 c+0 f+1
	}
  else if ( left )
	out_fvec[  out_stride2 + 3 * n_orientations + 2 ] = L2; // left : r+0 c+1 f+2
  else if ( right )
	out_fvec[ -out_stride2 + 3 * n_orientations + 2 ] = L2; // left : r+0 c-1 f+2
}

// collect block normalizations and complete features
__global__ void d_felzenszwalb_HOG_grey_step2( float * out, int n1_cells, int n2_cells
											   , int n_orientations, float alpha )
{
  const int i1_cell = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2_cell = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  if ( i1_cell >= n1_cells || i2_cell >= n2_cells )
	return;
  
  enum { UL = 0, UR = 1, LR = 2, LL = 3 };
  float block_norms[4] = { 0, 0, 0, 0 }; // upper left, upper right, lower right, lower left
  float energy_features[4] = { 0, 0, 0, 0 };
  bool top    = i1_cell == 0;
  bool left   = i2_cell == 0;
  bool bottom = i1_cell == n1_cells - 1;
  bool right  = i2_cell == n2_cells - 1;

  int n_features = n_orientations + 4; // todo: contrast sensitive, the felzenszwalb way
  int out_stride1 = n2_cells * n_features;
  int out_stride2 = n_features;
  float * out_fvec = out + i1_cell * out_stride1 + i2_cell * out_stride2;

  // collect block normalizations
  float topval, leftval, rightval, botval;
  if ( top )
	{
	  if ( left )
		block_norms[UL] += out_fvec[ 3 * n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[UL] += out_fvec[ -out_stride2 + 3 * n_orientations + 1 ]; // stored left f+1
	  if ( right )
		block_norms[UR] += out_fvec[ 3 * n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[UR] += out_fvec[  out_stride2 + 3 * n_orientations + 1 ]; // stored right f+1
	  topval =  out_fvec[ 3 * n_orientations + 1 ]; // stored locally f+1
	}
  else
	topval =  out_fvec[ -out_stride1 + 3 * n_orientations ];

  if ( bottom )
	{
	  if ( left )
		block_norms[LL] += out_fvec[ 3 * n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[LL] += out_fvec[ -out_stride2 + 3 * n_orientations + 2 ]; // stored left f+2
	  if ( right )
		block_norms[LR] += out_fvec[ 3 * n_orientations + 3 ]; // stored locally f+3
	  else
		block_norms[LR] += out_fvec[  out_stride2 + 3 * n_orientations + 2 ]; // stored right f+2
	}
  else
	botval =  out_fvec[ out_stride1 + 3 * n_orientations ];
	
  if ( left )
	leftval =  out_fvec[ 3 * n_orientations + 2 ];
  else
	{
	  leftval =  out_fvec[ -out_stride2 + 3 * n_orientations ];
	  if ( !top )
		block_norms[UL] += out_fvec[ -out_stride1 - out_stride2 + 3 * n_orientations ];
	  if ( !bottom )
		block_norms[LL] += out_fvec[  out_stride1 - out_stride2 + 3 * n_orientations ];
	}

  if ( right )
	rightval =  out_fvec[ 3 * n_orientations + 2 ];
  else
	{
	  rightval =  out_fvec[ out_stride2 + 3 * n_orientations ];
	  if ( !top )
		block_norms[UR] += out_fvec[ -out_stride1 + out_stride2 + 3 * n_orientations ];
	  if ( !bottom )
		block_norms[LR] += out_fvec[  out_stride1 + out_stride2 + 3 * n_orientations ];		  
	}

  float centerval = out_fvec[3 * n_orientations];
  const float epsilon = 1e-5;
  block_norms[UL] += topval + leftval  + centerval;
  block_norms[LL] += botval + leftval  + centerval;
  block_norms[UR] += topval + rightval + centerval;
  block_norms[LR] += botval + rightval + centerval;
  for ( int b = 0; b < 4; ++b ) // block index
	block_norms[b] = 1.0f / ( alpha * sqrtf( block_norms[b] + epsilon ) );

  for ( int o = 0; o < 3 * n_orientations; ++o )
	{
	  float orientation_feature_in = out_fvec[o];
	  float orientation_feature_out = 0;
	  for ( int b = 0; b < 4; ++b ) // block index
		{
		  float T = min( 1.0f, orientation_feature_in * block_norms[b] );
		  orientation_feature_out += T;	  
		  energy_features[b] += T;
		}
	  out_fvec[o] = orientation_feature_out;
	}
  for ( int b = 0; b < 4; ++b )
	out_fvec[3 * n_orientations+b] = energy_features[b];

}

extern "C" void do_R_HOG_grey( const float * in, int in_n1, int in_n2
							   , float * out, int cell_size, int block_size
							   , int block_overlap, int n_orientations
							   , bool interpolate, float alpha )
{
  const int block_stride = ( block_size - block_overlap ) * cell_size;
  const int block_extent = block_size * cell_size;
  const int n1_blocks = 1 + ( in_n1 - block_extent - 2 ) / block_stride;
  const int n2_blocks = 1 + ( in_n2 - block_extent - 2 ) / block_stride;
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d( iDivUp( n1_blocks, dimBlock2d.x ), iDivUp( n2_blocks, dimBlock2d.y ) );
  const int n_shared = dimBlock2d.x * dimBlock2d.y * n_orientations * sizeof(float);
  d_R_HOG_grey<<< dimGrid2d, dimBlock2d, n_shared >>>( in, in_n1, in_n2, out, cell_size, block_size, block_overlap, n_orientations, interpolate, alpha );
}

static void array2tex( const cudaArray * array, texture<float, 2> & texture )
{
  // Bind the array to the texture
  texture.normalized = false;
  texture.filterMode = cudaFilterModePoint;
  texture.addressMode[0] = cudaAddressModeClamp;
  texture.addressMode[1] = cudaAddressModeClamp;
  cudaBindTextureToArray( texture, array );
}

extern "C" void do_R_HOG_grey_tex( const cudaArray * in
								   , float * out, int out_n1, int out_n2
								   , int cell_size, int block_size
								   , int block_overlap, int n_orientations
								   , bool interpolate, float alpha
								   , float xscale, float yscale )
{
  array2tex( in, tex );
  const int n1_blocks = out_n1 / block_size;
  const int n2_blocks = out_n2 / block_size;
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d( iDivUp( n1_blocks, dimBlock2d.x ), iDivUp( n2_blocks, dimBlock2d.y ) );
  const int n_shared = dimBlock2d.x * dimBlock2d.y * n_orientations * sizeof(float);
  d_R_HOG_grey_tex<<< dimGrid2d, dimBlock2d, n_shared >>>( out, out_n1, out_n2, cell_size, block_size, block_overlap
														   , n_orientations, interpolate, alpha, xscale, yscale );
}

// n_orientations references halve or full circle, depending on sensitivity
extern "C" void do_felzenszwalb_RHOG_grey( const float * in, int in_n1, int in_n2, float * out
										  , int cell_size, int n_orientations, float alpha, bool contrast_sensitive )
{
  int n1_cells_in = ( in_n1 - 2 ) / cell_size;
  int n2_cells_in = ( in_n2 - 2 ) / cell_size;
  int n1_cells_out = n1_cells_in - 2;
  int n2_cells_out = n2_cells_in - 2;
  dim3 dimBlock2d = GPUData_dimBlock2d;
  // step 1
  {
	dim3 dimGrid2d( iDivUp( n1_cells_in, dimBlock2d.x ), iDivUp( n2_cells_in, dimBlock2d.y ) );
	const int n_shared = dimBlock2d.x * dimBlock2d.y * n_orientations * sizeof(float);
	d_felzenszwalb_RHOG_grey_step1<<< dimGrid2d, dimBlock2d, n_shared >>>( in, in_n1, in_n2, out, cell_size, n_orientations, contrast_sensitive );
  }
  // step 2
  {
	dim3 dimGrid2d( iDivUp( n1_cells_out, dimBlock2d.x ), iDivUp( n2_cells_out, dimBlock2d.y ) );
	d_felzenszwalb_RHOG_grey_step2<<< dimGrid2d, dimBlock2d >>>( out, n1_cells_out, n2_cells_out, n_orientations, alpha, contrast_sensitive );
  }
}

// n_orientations references halve circle
extern "C" void do_felzenszwalb_HOG_grey( const float * in, int in_n1, int in_n2, float * out
										  , int cell_size, int n_orientations, float alpha )
{
  int n1_cells_in = ( in_n1 - 2 ) / cell_size;
  int n2_cells_in = ( in_n2 - 2 ) / cell_size;
  int n1_cells_out = n1_cells_in - 2;
  int n2_cells_out = n2_cells_in - 2;
  dim3 dimBlock2d = GPUData_dimBlock2d;
  // step 1
  {
	dim3 dimGrid2d( iDivUp( n1_cells_in, dimBlock2d.x ), iDivUp( n2_cells_in, dimBlock2d.y ) );
	const int n_shared = dimBlock2d.x * dimBlock2d.y * 3 * n_orientations * sizeof(float);
	d_felzenszwalb_HOG_grey_step1<<< dimGrid2d, dimBlock2d, n_shared >>>( in, in_n1, in_n2, out, cell_size, n_orientations );
  }
  // step 2
  {
	dim3 dimGrid2d( iDivUp( n1_cells_out, dimBlock2d.x ), iDivUp( n2_cells_out, dimBlock2d.y ) );
	d_felzenszwalb_HOG_grey_step2<<< dimGrid2d, dimBlock2d >>>( out, n1_cells_out, n2_cells_out, n_orientations, alpha );
  }
}
