/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
//#include <cutil.h>
//#include <cu_math.h>

#include <cuda_runtime.h>
#include <cutil.h>
#define CB_TILE_W  16
#define CB_TILE_H  16
#define CUDA_SAFE_CALL( x ) x




static texture<float, 2> _ftex;  // texture containing the input features (normally the dir-features from BOF)

/***************************************************************************************/

// calculates the arithmetic mean of the respective dir-features over the receptive field
inline __device__
float
eqmotion_detector( float2 addr, size_t feature_depth, int n_pool_d,
                   int f1Idx )
{
  float ret = 0;

  for (int i = 0; i < n_pool_d; i++)
  {
    for (int j = 0; j < n_pool_d; j++ )
    {
      float2 pos = make_float2( addr.x + i, addr.y + j );
      
      float f = tex2D( _ftex, (pos.x * feature_depth) + f1Idx, pos.y );
      ret += f;
    }
  }
  
  ret /= float(n_pool_d * n_pool_d);
  
  return ret;  // no log in Matthias' version
}

/***************************************************************************************/

inline __device__
float
opmotion_detector( float2 addr, size_t feature_depth, int n_pool_d,
                   int f1Idx, int f2Idx )
	// f1Idx: index of feature 1
	// f2Idx: index of feature 2 (usually opposite direction to feature 2)
{
  float ret = 0;

  for (int i = 0; i < n_pool_d; i++)
  {
    for (int j = 0; j < n_pool_d; j++ )
    {
      float2 pos = make_float2( addr.x + i, addr.y + j );
      
      float f1 = tex2D( _ftex, (pos.x * feature_depth) + f1Idx, pos.y );
      float f2 = tex2D( _ftex, (pos.x * feature_depth) + f2Idx, pos.y );
      
      float of = sqrt( f1 * f2 );
      ret = max( ret, of );
    }
  }

  return ret; // no log in Matthias' version
}

/***************************************************************************************/

inline __device__
float
multiply_with_detectormask( float2 addr, size_t feature_depth, int feature_idx, float shift, int n_pool_d  )
{
  const float rest = float(0.5) * (1 - (n_pool_d % 2)); // 0.5 if n_pool_d is even, 0 if n_pool_d is odd
  const float center = (float(n_pool_d) / 2) - rest;

  float ret = 0;
  float msum = 0;  // sum of mask values

  for (int i = 0; i < n_pool_d; i++)
  {
    float x = i - center;
    for (int j = 0; j < n_pool_d; j++ )
    {
      float y = j - center;
      float phi = atan2( y, x );
      
	  // formula (2) from Searise Deliverable 4.1
      float m = float(0.5) * (float(1) + cos( phi - shift ) );
      m *= m; // ^2
      m *= m; // ^4
      m *= m; // ^8
      
      float2 pos = make_float2( addr.x + i, addr.y + j );
      msum += m;
      ret += m * tex2D( _ftex, (pos.x * feature_depth) + feature_idx, pos.y ); 
    }
  }

  return (ret / msum);  // Matthias' version
}


inline __device__
float
rotation_detector( float2 addr, size_t feature_depth, float gamma, size_t depth, int n_pool_d )
{
  float ret = 0;

  for( int i = 0; i < depth; i++ )
  {
    float alpha = ( 2. * M_PI * i ) / float(depth);  // alpha: prefered direction of i-th input dir feature
    ret += log( max( float(1e-15), 
                     multiply_with_detectormask( addr, feature_depth, i, gamma+alpha, n_pool_d ) ) );
  }
  return exp( ret / depth ); // in Matthias' version, there was an exp function around ret/depth
}


/***************************************************************************************/

#define eqmotion 0x01
#define opmotion 0x02
#define crmotion 0x04
#define rotation 0x08


__global__ void RotationSelectionKernel( float *results,  
                                         int n_pool_d, int n_pool_step,
                                         size_t pitch_y, size_t pitch_x, size_t feature_depth,
                                         unsigned int width, unsigned int height, unsigned int gamma_steps,
                                         float alpha_shift,
                                         char mflag )
	// feature_depth: depth of input features (usually downsampled dir-features from BOF)
{
  float2 addr;
  addr.x = blockIdx.x * blockDim.x + threadIdx.x;
  addr.y = blockIdx.y * blockDim.y + threadIdx.y;
  
  const unsigned int offset = (addr.y * pitch_y) + (addr.x * pitch_x);
  
  addr.x *= n_pool_step;
  addr.y *= n_pool_step;

  if(addr.x < width && addr.y < height && addr.x >= 0.f && addr.y >= 0.f ) 
  {
    int oi = 0;
    
    if (mflag & eqmotion)
    {
      for (int i = 0; i < feature_depth; i++)
      {
        results[offset+oi] = eqmotion_detector( addr, feature_depth, n_pool_step, i);
        oi++;
      }
    }

    if (mflag & opmotion)
    {
      size_t half = feature_depth >> 1;
      for (int i=0; i < half; i++)
      {
        results[offset+oi] = opmotion_detector( addr, feature_depth, n_pool_d, i, i+half );
		// function opmotion_detector from Matthias not used anymore because it calculates opponent motion pixelwise. yields almost 0 usually.
		// well, not pixelwise for image, but pixelwise for the input of RotationSelection. At this point, downsamplers should have been used already.

		// two new lines, multiplies two eqmotion detectors, so not just pixelwise, but after pooling
		/*
		results[offset+oi] = eqmotion_detector( addr, feature_depth, n_pool_step, i);
		results[offset+oi] *= eqmotion_detector( addr, feature_depth, n_pool_step, i+half);
		results[offset+oi] = sqrt( results[offset+oi] );
		*/
        oi++;
      }
    }
    
    if (mflag & crmotion)
    {
      size_t q = feature_depth >> 2;
      //for (int i=0; i < (3*q); i++)  Matthias' version
	  for (int i=0; i<feature_depth; i++)
      {
         // there is a right angle between the direction corresponding to i and the direction corresponding to i2
        int i2 = (i + q) % feature_depth; 
        results[offset+oi] = opmotion_detector( addr, feature_depth, n_pool_d, i, i2 );
		/*
		results[offset+oi] = eqmotion_detector( addr, feature_depth, n_pool_step, i);
		results[offset+oi] *= eqmotion_detector( addr, feature_depth, n_pool_step, i2);
		results[offset+oi] = sqrt( results[offset+oi] );
		*/
        oi++;
      }
    }
    
    if (mflag & rotation)
    {
      for (int k = 0; k < gamma_steps; k++ )
      {
        float gamma = ( 2. * M_PI * k ) / float( gamma_steps );
        gamma += alpha_shift;  // constant shift in gamma can be given as a parameter, don't know what this is good for (Frank)
        
        results[offset+oi] = rotation_detector( addr, feature_depth, gamma, feature_depth, n_pool_d );
        oi++;
      }
    }
  }
}


extern "C" void do_rotation_selection( const cudaArray *feature_map, float *results,
                                       int n_pool_d, int n_pool_step,
                                       size_t pitch_y, size_t pitch_x, size_t feature_depth,
                                       unsigned int rwidth, unsigned int rheight, unsigned int gamma_steps,
                                       float alpha_shift, char mflag)
{
  
  // Setup texture parameters
  _ftex.normalized = false;
  _ftex.filterMode = cudaFilterModePoint;
  _ftex.addressMode[0] = cudaAddressModeClamp;
  _ftex.addressMode[1] = cudaAddressModeClamp;
  
  cudaBindTextureToArray( _ftex, feature_map );  
  
  const unsigned int width  = rwidth  * n_pool_step;
  const unsigned int height = rheight * n_pool_step;
  
  dim3 grid(iDivUp(rwidth, CB_TILE_W), iDivUp(rheight, CB_TILE_H));
  dim3 threads(CB_TILE_W, CB_TILE_H);
  
  //dim3 dimBlock2d( 16 , 8 );
  //dim3 dimGrid2d ( iDivUp( n1, dimBlock2d.x ), iDivUp( n2, dimBlock2d.y ) );
  //  fprintf(stderr,"%d,%d ; %d,%d\n",dimBlock2d.x,dimBlock2d.y,dimGrid2d.x,dimGrid2d.y);
  
  RotationSelectionKernel <<< grid, threads >>> ( results, 
                                                  n_pool_d, n_pool_step,
                                                  pitch_y, pitch_x, feature_depth,
                                                  width, height, gamma_steps,
                                                  alpha_shift, mflag );
  cudaError_t err = cudaGetLastError();
  
  CUDA_SAFE_CALL( cudaUnbindTexture( _ftex ) );

  cudaThreadSynchronize();
}


