/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"
#include "Exceptions.hpp"
#include <stdio.h>

static texture<float, 2> _scan_src_tex;
static texture<float, 2> _scan_kernel1_tex;
static texture<float, 2> _scan_kernel2_tex;

struct diagonal_gaussian_kernel
{
  __device__ float operator()( int kernel_x0, int kernel_y, int src_x0, int src_y, int n3 )
  {
    float lp = 0;
    for ( int m = 0; m < n3; ++m )
      {
		float mu  = tex2D( _scan_kernel1_tex, kernel_x0 + m, kernel_y );
		float tau = tex2D( _scan_kernel2_tex, kernel_x0 + m, kernel_y );
		float x = tex2D( _scan_src_tex, src_x0 + m, src_y );
		float d = x - mu;
		lp -= 0.5f * tau * d * d;
      }
    return lp;
  }
};

struct multinomial_kernel
{
  __device__ float operator()( int kernel_x0, int kernel_y, int src_x0, int src_y, int n3 )
  {
    float l = 0;
    for ( int m = 0; m < n3; ++m )
      {
		float mu  = tex2D( _scan_kernel1_tex, kernel_x0 + m, kernel_y );
		float x = tex2D( _scan_src_tex, src_x0 + m, src_y );
		l += x * mu;
      }
    return log( max( 1e-20f, l ) );
  }
  __device__ float eval_cached( int kernel_x0, int kernel_y, float * src, int src_stride, int n3 )
  {
    float l = 0;
    for ( int m = 0; m < n3; ++m, src += src_stride )
      {
		float mu  = tex2D( _scan_kernel1_tex, kernel_x0 + m, kernel_y );
		float x = *src;
		l += x * mu;
      }
    return log( max( 1e-20f, l ) );
  }
};

template<typename Kernel>
__global__ void d_kernel_scan( float * out, const float * lC, int n1, int n2, int n3,
			       int d1, int s1, int d2, int s2, int nk,
			       Kernel kernel
			       )
{
  // index Offset
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  const int i1  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  if ( i1 >= m1 || i2 >= m2 ) return;

  out += ( i1 * m2 + i2 ) * nk;

  // unnormalized collection pass
  float lpmax = -1e20;
  for ( int k = 0; k < nk; ++k )
    {
      float lp = lC[k];
      for ( int i = 0; i < d1; ++i )
		for ( int j = 0; j < d2; ++j )
		  {
			int kernel_x0 = ( j * nk + k ) * n3;
			int kernel_y = i;
			int src_x0 = ( i2 * s2 + j ) * n3;
			int src_y = i1 * s1 + i;
			lp += kernel( kernel_x0, kernel_y, src_x0, src_y, n3 );
		  }
      lpmax = max( lp, lpmax );
      out[ k ] = lp;
    }

  // normrho passes
  float Z = 0;
  for ( int k = 0; k < nk; ++k )
    {
      float l = exp( out[ k ] - lpmax );
      out[ k ] = l;
      Z += l;
    }
  for ( int k = 0; k < nk; ++k )
    out[ k ] /= Z;
}

template<typename Reduction>
__device__ inline float
block_reduce( int ithread, int nthreads, float * data, float * buffer, Reduction reduce )
{
  float & local_data = data[ithread];
  float & local_buffer = buffer[ithread];
  bool first = true;
  for ( int kmax = ( nthreads + 1 ) / 2; kmax > 0; kmax /= 2 )
  {
    int iother = ithread + kmax;
	if ( ithread < kmax && iother < nthreads )
      local_buffer = reduce( ( first ? local_data : local_buffer ), ( first ? data : buffer )[iother] );
    if ( kmax > 32 ) // is this right?
		__syncthreads();
	first = false;
  }
  __syncthreads();
  return buffer[0];
}

template<typename T>
struct addfun
{
  __device__ T operator()( T a, T b )
  {
	return a + b;
  }
};

template<typename T>
struct maxfun
{
  __device__ T operator()( T a, T b )
  {
	return a > b ? a : b;
  }
};

template<typename Kernel>
__global__ void d_kernel_scan_distributed( float * out, const float * lC, int n1, int n2, int n3,
			       int d1, int s1, int d2, int s2, int nk,
			       Kernel kernel
			       )
{
  // index Offset
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  const int i1  = blockIdx.x; // globalIndex
  const int i2  = blockIdx.y; // globalIndex
  const int k   = threadIdx.x;
  if ( i1 >= m1 || i2 >= m2 ) return;

  out += ( i1 * m2 + i2 ) * nk;

  extern __shared__ float shared[];
  float * belief = shared;
  float * buffer = shared + nk;
  float & lp = belief[k];
  float & p  = belief[k];

  // unnormalized collection pass
  lp = lC[k];
  for ( int i = 0; i < d1; ++i )
    for ( int j = 0; j < d2; ++j )
    {
      int kernel_x0 = ( j * nk + k ) * n3;
      int kernel_y = i;
      int src_x0 = ( i2 * s2 + j ) * n3;
      int src_y = i1 * s1 + i;
      lp += kernel( kernel_x0, kernel_y, src_x0, src_y, n3 );
    }
  // normrho passes
  float lpmax = block_reduce( k, nk, belief, buffer, maxfun<float>() ); 
  p = exp( lp - lpmax );
  float Z = block_reduce( k, nk, belief, buffer, addfun<float>() );
  out[k] = p / Z;
}


template<typename Kernel>
__global__ void d_kernel_scan_cached( float * out, const float * lC, int n1, int n2, int n3,
									  int d1, int s1, int d2, int s2, int nk,
									  Kernel kernel
									  )
{
  // index Offset
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  const int i1  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  if ( i1 >= m1 || i2 >= m2 ) return;

  out += ( i1 * m2 + i2 ) * nk;
  
  const int nthreads = blockDim.y * blockDim.y;
  const int ithread  = threadIdx.x * blockDim.y + threadIdx.y;
  extern __shared__ float buffer[];
  float * outbuf = buffer + ithread;
  float * inbuf  = buffer + nk * nthreads + ithread;
  for ( int k = 0, koff = 0; k < nk; ++k, koff += nthreads )
	outbuf[ koff ] = lC[k];

  // unnormalized collection pass
  for ( int i = 0; i < d1; ++i )
	for ( int j = 0; j < d2; ++j )
	  {
		// load "pixel" into inbuf
		int src_x0 = ( i2 * s2 + j ) * n3;
		int src_y = i1 * s1 + i;
		for ( int m = 0, moff = 0; m < n3; ++m, moff += nthreads )
		  {
			inbuf[ moff ] = tex2D( _scan_src_tex, src_x0 + m, src_y );
		  }
		// apply all kernels
		for ( int k = 0, koff = 0; k < nk; ++k, koff += nthreads )
		  {
			int kernel_x0 = ( j * nk + k ) * n3;
			int kernel_y = i;
			outbuf[ koff ] += kernel.eval_cached( kernel_x0, kernel_y, inbuf, nthreads, n3 );
		  }
	  }

  // normrho passes
  float lpmax = -1e20;
  for ( int k = 0, koff = 0; k < nk; ++k, koff += nthreads )
	lpmax = max( lpmax, outbuf[ koff ] );
  float Z = 0;
  for ( int k = 0, koff = 0; k < nk; ++k, koff += nthreads )
    {
      float l = exp( outbuf[ koff ] - lpmax );
      outbuf[ koff ] = l;
      Z += l;
    }
  for ( int k = 0, koff = 0; k < nk; ++k, koff += nthreads )
    out[ k ] = outbuf[ koff ] / Z;
}

static void array2tex( const cudaArray * array, texture<float, 2> & texture )
{
  // Bind the array to the texture
  texture.normalized = false;
  texture.filterMode = cudaFilterModePoint;
  texture.addressMode[0] = cudaAddressModeClamp;
  texture.addressMode[1] = cudaAddressModeClamp;
  cudaBindTextureToArray( texture, array );
}

extern "C" void h_diagonal_gaussian_kernel_scan( float * out, int n1, int n2, int n3, 
						 const cudaArray * src,  const cudaArray * mu,  
						 const cudaArray * tau, const float * lC,
						 int d1, int s1, int d2, int s2, int nk )
{
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d( iDivUp( m1, dimBlock2d.x ), iDivUp( m2, dimBlock2d.y ) );
  array2tex( src, _scan_src_tex );
  GPUDataException::check_error( "binding src texture" );
  array2tex( mu , _scan_kernel1_tex );
  GPUDataException::check_error( "binding mu texture" );
  array2tex( tau, _scan_kernel2_tex );
  GPUDataException::check_error( "binding tau texture" );
  d_kernel_scan<<< dimGrid2d, dimBlock2d >>>( out, lC, n1, n2, n3, d1, s1, d2, s2, nk, diagonal_gaussian_kernel() );
  GPUDataException::check_error( "calling normal kernel" );
  cudaUnbindTexture( _scan_src_tex );
  GPUDataException::check_error( "unbinding src texture" );
  cudaUnbindTexture( _scan_kernel1_tex );
  GPUDataException::check_error( "unbinding mu texture" );
  cudaUnbindTexture( _scan_kernel2_tex );
  GPUDataException::check_error( "unbinding tau texture" );

}  

extern "C" void h_diagonal_gaussian_kernel_scan_distributed(
             float * out, int n1, int n2, int n3, 
						 const cudaArray * src,  const cudaArray * mu,  
						 const cudaArray * tau, const float * lC,
						 int d1, int s1, int d2, int s2, int nk )
{
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  char msg[255];
  array2tex( src, _scan_src_tex );
  GPUDataException::check_error( "binding src texture" );
  array2tex( mu , _scan_kernel1_tex );
  GPUDataException::check_error( "binding mu texture" );
  array2tex( tau, _scan_kernel2_tex );
  GPUDataException::check_error( "binding tau texture" );
  d_kernel_scan_distributed<<< dim3( m1, m2 ), dim3( nk ), 2 * nk * sizeof(float) >>>( out, lC, n1, n2, n3, d1, s1, d2, s2, nk, diagonal_gaussian_kernel() );
  sprintf( msg, "calling distributed kernel m1 %i m2 %i nk %i", m1, m2, nk );
  GPUDataException::check_error( msg );
  cudaUnbindTexture( _scan_src_tex );
  GPUDataException::check_error( "unbinding src texture" );
  cudaUnbindTexture( _scan_kernel1_tex );
  GPUDataException::check_error( "unbinding mu texture" );
  cudaUnbindTexture( _scan_kernel2_tex );
  GPUDataException::check_error( "unbinding tau texture" );
}  

extern "C" void h_multinomial_kernel_scan_distributed(
             float * out, int n1, int n2, int n3
			 , const cudaArray * src,  const cudaArray * mu
			 , const float * lC
			 , int d1, int s1, int d2, int s2, int nk )
{
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  char msg[255];
  array2tex( src, _scan_src_tex );
  array2tex( mu , _scan_kernel1_tex );
  d_kernel_scan_distributed<<< dim3( m1, m2 ), dim3( nk ), 2 * nk * sizeof(float) >>>( out, lC, n1, n2, n3, d1, s1, d2, s2, nk, multinomial_kernel() );
  sprintf( msg, "calling distributed kernel m1 %i m2 %i nk %i", m1, m2, nk );
  GPUDataException::check_error( msg );
  cudaUnbindTexture( _scan_src_tex );
  cudaUnbindTexture( _scan_kernel1_tex );
}  


extern "C" void h_multinomial_kernel_scan( float * out, int n1, int n2, int n3, 
					   const cudaArray * src,  const cudaArray * mu,  
					   const float * lC,
					   int d1, int s1, int d2, int s2, int nk )
{
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d( iDivUp( m1, dimBlock2d.x ), iDivUp( m2, dimBlock2d.y ) );
  array2tex( src, _scan_src_tex );
  array2tex( mu , _scan_kernel1_tex );
  d_kernel_scan<<< dimGrid2d, dimBlock2d >>>( out, lC, n1, n2, n3, d1, s1, d2, s2, nk, multinomial_kernel() );
  cudaUnbindTexture( _scan_src_tex );
  cudaUnbindTexture( _scan_kernel1_tex );
}  

extern "C" void h_multinomial_kernel_scan_cached( float * out, int n1, int n2, int n3, 
												  const cudaArray * src,  const cudaArray * mu,  
												  const float * lC,
												  int d1, int s1, int d2, int s2, int nk )
{
  const int m1 = 1 + ( n1 - d1 ) / s1;
  const int m2 = 1 + ( n2 - d2 ) / s2;
  dim3 dimBlock2d( 4, 8 );
  dim3 dimGrid2d( iDivUp( m1, dimBlock2d.x ), iDivUp( m2, dimBlock2d.y ) );
  int nthreads = dimBlock2d.x * dimBlock2d.y;
  array2tex( src, _scan_src_tex );
  array2tex( mu , _scan_kernel1_tex );
  int nshared = sizeof(float) * nthreads * ( n3 + nk );
  d_kernel_scan_cached<<< dimGrid2d, dimBlock2d, nshared >>>( out, lC, n1, n2, n3, d1, s1, d2, s2, nk, multinomial_kernel() );
  cudaUnbindTexture( _scan_src_tex );
  cudaUnbindTexture( _scan_kernel1_tex );
}  
