/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include <stdio.h>
//#include <device_functions.h>
#include "cutil.h"


__global__ void d_computeHyposAmbiguous(float* censusOccDev, 
										float* hypoImageV1, 
										float* hypoImageMT, 
										unsigned int* currCensus, 
										unsigned int* pastCensus, 
										const int w, 
										const int NPix, 
										const int fbf, const float fb0, const int df_mt,
										const int VEL_MAX, 
										const int H_MAX_2 )
{

	const unsigned int H_MAX_1 = blockDim.x;

	const unsigned int iPixel= blockIdx.x+blockIdx.y*gridDim.x;
	const unsigned int xPos = iPixel%w;
	const unsigned int yPos = iPixel/w;

	const unsigned int iHypoMemOffset= iPixel*H_MAX_1*3;

	const int h = NPix/w;
	const int xP= iPixel%w,  yP= iPixel/w;

	//use MT resolution for FB
	const unsigned int delta = df_mt / 2;
	const unsigned int iPixelMT=  ( xPos + delta ) / df_mt + ( ( yPos + delta ) / df_mt )  * ( w / df_mt );
	const unsigned int iHypoMemOffsetMT = iPixelMT * H_MAX_1 * 3;
	const unsigned int iHypo = threadIdx.x;

	if ( iPixel>=NPix || iPixelMT > NPix / ( df_mt * df_mt ) ) 
	  return;

	extern __shared__ unsigned int cachI[];

	unsigned int * cachPast = cachI;
	unsigned int * cachCurr = cachI + H_MAX_1;

	const unsigned int census = pastCensus[iPixel];
	//how many identical Census values in next frame
	const unsigned int noOfMa = censusOccDev[iPixel];

	//border treatment
	const unsigned int imax = NPix / ( df_mt * df_mt );
	const unsigned int ioffV1 = iHypoMemOffset   + iHypo * 3;
	const unsigned int ioffMT = iHypoMemOffsetMT + iHypo * 3;
	if( iPixelMT < imax 
		&& xPos < w	&& yPos < h
		&& ( ioffMT + 2 ) < ( imax * iPixelMT * H_MAX_1 * 3 ) )
	{

		const int dx = ( int ) hypoImageMT[ ioffMT + 1 ];
		const int dy = ( int ) hypoImageMT[ ioffMT + 2 ];
		const int d2 = dx * dx + dy * dy;
		const int vmax2 = VEL_MAX * VEL_MAX;
		const int ioffVel = w * dy + dx;
		
		//border treatment
		if ( xP + dx < w && xP + dx >= 0 
			 && yP + dy < h && yP+dy >= 0 
			 && hypoImageMT[ ioffMT ] > 0.00001
			 && noOfMa < H_MAX_2 )
		{
		  //compute hypos for ambiguous census values
		  if ( noOfMa > H_MAX_1  )
			{

			  //conditions: same census value, velocity smaller than limit
			  if ( census == currCensus[ iPixel + ioffVel ] && d2 <= vmax2 ) 
				  { 
					//feedback is instantly multiplied!
					hypoImageV1[ ioffV1 + 0 ] = ( fb0 + fbf * hypoImageMT[ ioffMT ] );
					hypoImageV1[ ioffV1 + 1 ] = dx;
					hypoImageV1[ ioffV1 + 2 ] = dy;
				}
			  else 
				{
				  hypoImageV1[ ioffV1 + 0 ] = 0.0f;
				  hypoImageV1[ ioffV1 + 1 ] = 0.0f;
				  hypoImageV1[ ioffV1 + 2 ] = 0.0f;
				}

			} 
		  //feedback to unambiguous hypos that have been computed before on CPU
		  else
			{
			  //note: we could include the MT prediction shift here! because V1 velocity is known
			  //-->computy cachPast[iHypo] using dy and dx

			  //vel of MT hypos (=prediction)
			  cachPast[ iHypo ] = ioffVel;

			  //value of current V1 values
			  //cachCurr[ iHypo ] = hypoImageV1[ ioffV1 + 2 ] * w + hypoImageV1[ ioffV1 + 1 ];
			  			  			  
			   __syncthreads();

			  //loop to find identical V1 hypos using MT predictions
			  for ( int i = 0; i < H_MAX_1 ; ++i ) 
				{
				  if ( cachCurr[iHypo] == cachPast[i] )
					hypoImageV1[ ioffV1 ] *= ( 1 + fbf * hypoImageMT[ iHypoMemOffsetMT + i * 3 ] );
				}
			}
		}
		
	}
	
}

extern "C" {

	extern "C" void do_computeHyposAmbiguous(
			float* censusOcc,
			float* hypoImageV1,
			float* hypoImageMT,
			float* currCensus,
			float* pastCensus,
			const int w,
			const int NPix,
			const int fbf, const float fb0, const int df_mt,
			const int VEL_MAX,
			const int H_MAX_1,
			const int H_MAX_2
			)
	{
		dim3 dimBlock((int)H_MAX_1);
		int bx  = (int)NPix;
		int by = 1;
		while (bx>=65535) 
		  {
			by *= 2;
			bx  = iDivUp(bx,2);
		  }
		dim3 dimGrid(bx,by);

		d_computeHyposAmbiguous<<<dimGrid, dimBlock,sizeof(int)*H_MAX_1*2>>>
		  ( (float*) censusOcc, 
			hypoImageV1, 
			hypoImageMT, 
			(unsigned int*) currCensus, 
			(unsigned int*) pastCensus, 
			w, NPix, fbf, fb0, df_mt, VEL_MAX, H_MAX_2);

		cudaThreadSynchronize();

	}
}
