/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/

#include <cuda.h>
#include "cutil.h"
#include "compute_hypos.hcu"
#include "modelsetup.h"

// XXX: try to avoid including std-lib-code in CUDA files
#include <stdio.h>

// general subsampling function type. this documents the signature subsampling
// functions have to have
typedef void (*subsampling_function) (
		HypoFeed *feed,					// in: hypothese feed-in
		int n,							// in: number of pixels in result image of compute_hypos_ambiguous
		int x,							// in: x-coordinate of pixel to be subsampled
		int y,							// in: y-coordinate of pixel to be subsampled
		int w,							// in: width of result image
		bool *overflow,					// out: is there an overflow? e.g. subsampled pixel position is out of image constraints
		unsigned int *result			// out: resulting index into feed data
		);



__device__ void
ssf_downsample_common (
		HypoFeed *feed,
		int n, int x, int y, int w,
		bool *overflow, unsigned int *result)
{
	// use local memory -> all threads will access feed->subsampling_factor at the same time, so
	// according to the CUDA ref manual, the value will be 'broadcastet' to all
	// threads simultaneously
	const unsigned int ds_factor = feed->subsampling_factor;
	//const unsigned int delta = ds_factor / 2;

	//*result = (x + delta) / ds_factor + ((y+delta) / ds_factor) * (w/ds_factor);
	*result = x / ds_factor + (y / ds_factor) * ( w / ds_factor);
	*overflow = (*result >= (n / (ds_factor*ds_factor)));
}



__global__ void 
d_compute_hypos_ambiguous (
		float *censusOccDev,
		float *hypoImageV1,
		unsigned int *currCensus,
		unsigned int *pastCensus,
		const int w,
		const int nPix,
		const float fb0,
		const unsigned int nfeeds,
		HypoFeed *feed,
		int H_MAX_1, int H_MAX_2, int VEL_MAX
		)
{
	// image setup calculation
	const unsigned int iPixel = blockIdx.x + blockIdx.y * gridDim.x;
	const unsigned int xPos = iPixel % w;
	const unsigned int yPos = iPixel / w;
	const unsigned h = nPix / w;

	// check image constraints before continuing. hopefully this load of shit
	// called nvcc won't move it somewhere else
	if (iPixel >= nPix || xPos >= w || yPos >= h)
		return;

	const unsigned int iHypo = threadIdx.x;
	const unsigned int iHypoMemOffset = iPixel * H_MAX_1 * 3;
	const unsigned int census = pastCensus[iPixel];
	const unsigned int noOfMa = censusOccDev[iPixel]; // how many identical values in next frame
	const unsigned int ioffV1 = iHypoMemOffset + iHypo * 3;
	const unsigned int vmax2 = VEL_MAX * VEL_MAX;

	extern __shared__ unsigned int cache[];
	unsigned int *cachePast = cache;
	unsigned int *cacheCurr = cache + H_MAX_1;
	
	unsigned int iPixelHF;
	bool overflow;
	float *data_ptr;

	// iterate over all hypothese feeds and apply them accordingly
	for (int fi = 0; fi < nfeeds; fi++)  { 
	
		// because function pointers aren't allow in CUDA code, there's need for
		// a switch statement here
		overflow = false;
		switch (feed[fi].ssf) {
			case SSF_NO_SUBSAMPLING:
				iPixelHF = iPixel;
				overflow = iPixelHF >= nPix;
				break;

			case SSF_DOWNSAMPLE_COMMON:
				ssf_downsample_common(&feed[fi], nPix, xPos, yPos, w, &overflow, &iPixelHF);
				break;
		}
		data_ptr = feed[fi].data_ptr;
		const float weight = feed[fi].weight;


		// don't continue if the subsampling 'threw' an overflow
		if (!overflow) {
			unsigned int iHypoMemOffsetHF = iPixelHF * H_MAX_1 * 3;
			unsigned int ioffHF = iHypoMemOffsetHF + iHypo * 3;

			int dx = (int) data_ptr[ioffHF + 1];
			int dy = (int) data_ptr[ioffHF + 2];
			dx = ceil((float)dx * feed[fi].speed_scaling);
			dy = ceil((float)dy * feed[fi].speed_scaling);

			int d2 = dx*dx + dy*dy;
			int ioffVel = w * dy + dx;


			// constraint checking
			if (xPos + dx < w
			&& (int)xPos + dx >= 0
			&& yPos + dy < h
			&& (int)yPos + dy >= 0
			&& data_ptr[ioffHF] > EPSILON
			&& noOfMa < H_MAX_2
			&& dx <= VEL_MAX
			&& dy <= VEL_MAX
			){
				#define V1(i) hypoImageV1[ioffV1 + (i)]
				// compute hypos for ambiguous census values
				if ((fi == 0) && noOfMa <= H_MAX_2 && noOfMa > H_MAX_1) {
					if (census == currCensus[iPixel + ioffVel] && d2 <= vmax2) {
						V1(0) = fb0 + weight * data_ptr[ioffHF];
						V1(1) = dx;
						V1(2) = dy;
					}
				}

				// feedback to unambiguous hypos that have been computed before
				// on the CPU
				//
				// this is the MT feedback, and only if the hypothese is
				// unambiguous
				else if (fi != 0) {
					// velocity of MT hypos ( = prediction)
					cachePast[iHypo] = ioffVel;

					// value of current V1 values
					cacheCurr[iHypo] = V1(2)* w + V1(1);

					__syncthreads();

					// loop to find identical V1 hypos using MT predictions
					for (int i = 0; i < H_MAX_1; ++i) {
						if (cacheCurr[iHypo] == cachePast[i]) // BANK CONFLICT
							V1(0) *= (fb0 + weight * data_ptr[iHypoMemOffsetHF + i * 3]);
					}
				}
				#undef V1
			}
		}
		if (fi == 0)
			__syncthreads();
	}
}


extern "C" void
do_compute_hypos_ambiguous (
		float *censusOcc,
		float *hypoImageV1,
		float *currCensus,
		float *pastCensus,
		const int w,
		const int nPix,
		const float fb0,
		const unsigned int nfeeds,
		const HypoFeed *feeds,
		int H_MAX_1, int H_MAX_2, int VEL_MAX
		)
{
	dim3 block(H_MAX_1);
	int bx = nPix;
	int by = 1;
	while (bx >= 0xFFFF) {
		by *= 2;
		bx = iDivUp(bx, 2);
	}
	dim3 grid(bx, by);

	size_t s_feeds = sizeof(HypoFeed) * nfeeds;
	HypoFeed *dfeeds;
	cudaMalloc((void**)&dfeeds, s_feeds);
	cudaMemcpy(dfeeds, feeds, s_feeds, cudaMemcpyHostToDevice);

	d_compute_hypos_ambiguous<<<grid, block, sizeof(int) * H_MAX_1 * 2>>>
		( censusOcc,
		  hypoImageV1,
		  (unsigned int*) currCensus,
		  (unsigned int*) pastCensus,
		  w, nPix, fb0, nfeeds, dfeeds,
		  H_MAX_1, H_MAX_2, VEL_MAX);

	cudaThreadSynchronize();
	cudaFree(dfeeds);
}
