/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "cuda.h"
#include <cutil.h>
#include "../src/Inc.h"
#include <cublas.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include "cuda_matrix_kernel.cu"
#include "cuda_warp_kernel.cu"

#undef CUDA_SAFE_CALL
#define CUDA_SAFE_CALL(A) A

//texture<float, 2, cudaReadModeElementType > texData;
__device__ __constant__ float d_Kernel[KERNEL_W];
////////////////////////////////////////////////////////////////////////////////
// Loop unrolling templates, needed for best performance
////////////////////////////////////////////////////////////////////////////////
template<int i> __device__ float convolutionRow(float *data)
{
    return
        data[KERNEL_RADIUS - i] * d_Kernel[i]
        + convolutionRow<i - 1>(data);
}

template<> __device__ float convolutionRow<-1>(float *data)
{
    return 0;
}

template<int i> __device__ float convolutionColumn(float *data)
{
    return 
        data[(KERNEL_RADIUS - i) * COLUMN_TILE_W] * d_Kernel[i]
        + convolutionColumn<i - 1>(data);
}

template<> __device__ float convolutionColumn<-1>(float *data)
{
    return 0;
}



////////////////////////////////////////////////////////////////////////////////
// Row convolution filter
////////////////////////////////////////////////////////////////////////////////
__global__ void convolutionRowGPU(
    float *d_Result,
    float *d_Data,
    int dataW,
    int dataH
){
    //Data cache
    __shared__ float data[KERNEL_RADIUS + ROW_TILE_W + KERNEL_RADIUS];

    //Current tile and apron limits, relative to row start
    const int         tileStart = IMUL(blockIdx.x, ROW_TILE_W);
    const int           tileEnd = tileStart + ROW_TILE_W - 1;
    const int        apronStart = tileStart - KERNEL_RADIUS;
    const int          apronEnd = tileEnd   + KERNEL_RADIUS;

    //Clamp tile and apron limits by image borders
    const int    tileEndClamped = min(tileEnd, dataW - 1);
    const int apronStartClamped = max(apronStart, 0);
    const int   apronEndClamped = min(apronEnd, dataW - 1);

    //Row start index in d_Data[]
    const int          rowStart = IMUL(blockIdx.y, dataW);

    //Aligned apron start. Assuming dataW and ROW_TILE_W are multiples 
    //of half-warp size, rowStart + apronStartAligned is also a 
    //multiple of half-warp size, thus having proper alignment 
    //for coalesced d_Data[] read.
    const int apronStartAligned = tileStart - KERNEL_RADIUS_ALIGNED;

    const int loadPos = apronStartAligned + threadIdx.x;
    //Set the entire data cache contents
    //Load global memory values, if indices are within the image borders,
    //or initialize with zeroes otherwise
    if(loadPos >= apronStart){
        const int smemPos = loadPos - apronStart;

        data[smemPos] = 
            ((loadPos >= apronStartClamped) && (loadPos <= apronEndClamped)) ?
            d_Data[rowStart + loadPos] : 0;
    }


    //Ensure the completness of the loading stage
    //because results, emitted by each thread depend on the data,
    //loaded by another threads
    __syncthreads();
    const int writePos = tileStart + threadIdx.x;
    //Assuming dataW and ROW_TILE_W are multiples of half-warp size,
    //rowStart + tileStart is also a multiple of half-warp size,
    //thus having proper alignment for coalesced d_Result[] write.
    if(writePos <= tileEndClamped){
        const int smemPos = writePos - apronStart;
        float sum = 0;

#ifdef UNROLL_INNER
        sum = convolutionRow<2 * KERNEL_RADIUS>(data + smemPos);
#else
        for(int k = -KERNEL_RADIUS; k <= KERNEL_RADIUS; k++)
            sum += data[smemPos + k] * d_Kernel[KERNEL_RADIUS - k];
#endif

        d_Result[rowStart + writePos] = sum;
    }
}



////////////////////////////////////////////////////////////////////////////////
// Column convolution filter
////////////////////////////////////////////////////////////////////////////////
__global__ void convolutionColumnGPU(
    float *d_Result,
    float *d_Data,
    int dataW,
    int dataH,
    int smemStride,
    int gmemStride
){
    //Data cache
    __shared__ float data[COLUMN_TILE_W * (KERNEL_RADIUS + COLUMN_TILE_H + KERNEL_RADIUS)];

    //Current tile and apron limits, in rows
    const int         tileStart = IMUL(blockIdx.y, COLUMN_TILE_H);
    const int           tileEnd = tileStart + COLUMN_TILE_H - 1;
    const int        apronStart = tileStart - KERNEL_RADIUS;
    const int          apronEnd = tileEnd   + KERNEL_RADIUS;

    //Clamp tile and apron limits by image borders
    const int    tileEndClamped = min(tileEnd, dataH - 1);
    const int apronStartClamped = max(apronStart, 0);
    const int   apronEndClamped = min(apronEnd, dataH - 1);

    //Current column index
    const int       columnStart = IMUL(blockIdx.x, COLUMN_TILE_W) + threadIdx.x;

    //Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, COLUMN_TILE_W) + threadIdx.x;
    int gmemPos = IMUL(apronStart + threadIdx.y, dataW) + columnStart;
    //Cycle through the entire data cache
    //Load global memory values, if indices are within the image borders,
    //or initialize with zero otherwise
    for(int y = apronStart + threadIdx.y; y <= apronEnd; y += blockDim.y){
        data[smemPos] = 
        ((y >= apronStartClamped) && (y <= apronEndClamped)) ? 
        d_Data[gmemPos] : 0;
        smemPos += smemStride;
        gmemPos += gmemStride;
    }

    //Ensure the completness of the loading stage
    //because results, emitted by each thread depend on the data, 
    //loaded by another threads
    __syncthreads();
    //Shared and global memory indices for current column
    smemPos = IMUL(threadIdx.y + KERNEL_RADIUS, COLUMN_TILE_W) + threadIdx.x;
    gmemPos = IMUL(tileStart + threadIdx.y , dataW) + columnStart;
    //Cycle through the tile body, clamped by image borders
    //Calculate and output the results
    for(int y = tileStart + threadIdx.y; y <= tileEndClamped; y += blockDim.y){
        float sum = 0;

#ifdef UNROLL_INNER
        sum = convolutionColumn<2 * KERNEL_RADIUS>(data + smemPos);
#else
        for(int k = -KERNEL_RADIUS; k <= KERNEL_RADIUS; k++)
            sum += 
                data[smemPos + IMUL(k, COLUMN_TILE_W)] *
                d_Kernel[KERNEL_RADIUS - k];
#endif

        d_Result[gmemPos] = sum;
        smemPos += smemStride;
        gmemPos += gmemStride;
    }
}

extern "C" void convolution(float* d_DataA, float* d_DataB, int data_w, int data_h)
{
	dim3 blockGridRows(iDivUp(data_w, ROW_TILE_W), data_h);
	dim3 blockGridColumns(iDivUp(data_w, COLUMN_TILE_W), iDivUp(data_h, COLUMN_TILE_H));
	dim3 threadBlockRows(KERNEL_RADIUS_ALIGNED + ROW_TILE_W + KERNEL_RADIUS);
	dim3 threadBlockColumns(COLUMN_TILE_W, KERNEL_RADIUS_ALIGNED+COLUMN_TILE_H);
	
	//printf("GPU convolution...\n");
							CUDA_SAFE_CALL( cudaThreadSynchronize() );
							convolutionRowGPU<<<blockGridRows, threadBlockRows>>>(
								d_DataB,
								d_DataA,
								data_w,
								data_h
								);
							//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
							//daniel: doesnt compile here...
							CUDA_SAFE_CALL( cudaThreadSynchronize() );
							convolutionColumnGPU<<<blockGridColumns, threadBlockColumns>>>(
								d_DataA,
								d_DataB,
								data_w,
								data_h,
								COLUMN_TILE_W * threadBlockColumns.y,
								data_w * threadBlockColumns.y
								);
							//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
							//daniel: doesnt compile here...
							CUDA_SAFE_CALL( cudaThreadSynchronize() );
	}

	
extern "C" void Disparity(float **I, float *DX, float *DY, int scale, int n_scale, int data_size, int data_w, int data_h, float *DXwarp, float* DYwarp) {
	

	//Gabor separable components f0=1/4 b=0.833 DC=0
	float F[9][11]={{0.08244625, 0.20469046, 0.41072304, 0.67425587, 0.90708149, 0.99982890, 0.90708149, 0.67425587, 0.41072304, 0.20469046, 0.08244625},
	{-0.00594233, 0.19967099, -0.00647302, -0.68142846, -0.00582718, 1.00000000, -0.00582718, -0.68142846, -0.00647302, 0.19967099, -0.00594233},
	{-0.08544031, 0.00053402, 0.41244529, -0.00056667, -0.90934738, 0.00000000, 0.90934738, 0.00056667, -0.41244529, -0.00053402, 0.08544031},
	{0.04485397, -0.07286766, -0.42181225, -0.42640289, 0.38455161, 0.98335443, 0.38455161, -0.42640289, -0.42181225, -0.07286766, 0.04485397 },
	{0.05638481, 0.19853638, 0.07795001, -0.53686286, -0.81335453, -0.00000000, 0.81335453, 0.53686286, -0.07795001, -0.19853638, -0.05638481 },
	{0.04220859, 0.17536045, -0.14991172, -0.66565053, 0.09967660, 0.99663323, 0.09967660, -0.66565053, -0.14991172, 0.17536045, 0.04220859},
	{-0.07029199, 0.09854188, 0.39000786, -0.16080715, -0.90114083, -0.00000000, 0.90114083, 0.16080715, -0.39000786, -0.09854188, 0.07029199 },
	{-0.08650217, -0.15447067, -0.09619099, 0.24252298, 0.74448122, 0.99996745, 0.74448122, 0.24252298, -0.09619099, -0.15447067, -0.08650217 },
	{-0.00824940, -0.13874206, -0.40044315, -0.63163823, -0.51617938, 0.00000344, 0.51617938, 0.63163823, 0.40044315, 0.13874206, 0.00824940}};

	dim3 blockGridRows(iDivUp(data_w, ROW_TILE_W), data_h);
	dim3 blockGridColumns(iDivUp(data_w, COLUMN_TILE_W), iDivUp(data_h, COLUMN_TILE_H));
	dim3 threadBlockRows(KERNEL_RADIUS_ALIGNED + ROW_TILE_W + KERNEL_RADIUS);
	dim3 threadBlockColumns(COLUMN_TILE_W, 8);
	dim3 grid(iDivUp(data_w, SUM_TILE_W), iDivUp(data_h, SUM_TILE_H));
	dim3 block(SUM_TILE_W, SUM_TILE_H);
	dim3 grid1(iDivUp(data_w, TEMP_W), iDivUp(data_h, TEMP_H));
	dim3 block1(TEMP_W, TEMP_H);


	float *even[2], *odd[2], *temp1, *temp2, *temp3, *temp4, *temp5;
	//float *even[0] *odd[0] filtered left image
	//float *even[1] *odd[1] filtered right image
	CUDA_SAFE_CALL( cudaMalloc( (void **)&even[0], data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&odd[0], data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&even[1], data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&odd[1], data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp1, data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp2, data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp3, data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp4, data_size));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp5, data_size));

	float *even_r[n_ph]; //filtered right image (for the phase shifts)
	float *odd_r[n_ph];	 //filtered right image (for the phase shifts)

	float *E[n_ori][n_ph];		//Population energy
	float *E_norm[n_ori][n_ph]; //Normalized population energy

	for(int j=0; j<n_ori; j++)
	{
		for(int i=0; i<n_ph; i++)
		{
			CUDA_SAFE_CALL( cudaMalloc( (void **)&E[j][i], data_size));
			CUDA_SAFE_CALL( cudaMalloc( (void **)&E_norm[j][i], data_size));
			cudaMemset(E_norm[j][i],0,data_size);
		}
	}

	for(int i=0; i<n_ph; i++)
	{
		CUDA_SAFE_CALL( cudaMalloc( (void **)&even_r[i], data_size));
		CUDA_SAFE_CALL( cudaMalloc( (void **)&odd_r[i], data_size));
	}



	float *dx[n_ori]; //disparity along each orientation
	float *dy[n_ori]; //disparity along each orientation

	for(int i=0; i<n_ori; i++)
	{
		CUDA_SAFE_CALL( cudaMalloc( (void **)&dx[i], data_size));
		CUDA_SAFE_CALL( cudaMalloc( (void **)&dy[i], data_size));
	}

	cublasStatus status = cublasInit();
	if (status != CUBLAS_STATUS_SUCCESS) {
		fprintf (stderr, "!!!! CUBLAS initialization error\n");
	}

	
	//Gabor separable filtering


	//BEGIN GABOR FILTERING AND ENERGY

	//Ori = 0
	
	for(int i=0; i<2; i++) //0 left - 1 right
	{
		
		//CUDA_SAFE_CALL( cudaMemcpy(d_Kernel, F[0], KERNEL_SIZE, cudaMemcpyHostToDevice) );
				
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[0], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[1], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(even[i],temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[2], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(odd[i],temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		if(scale!=0 && i==0) //warping for the left image only
		{
			
			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
	
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaThreadSynchronize() );
			
		}
		if(i==1) //phase shifts (right image)
		{

			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}
	}

	//energy for the current orientation
	compute_energy(even[0], odd[0], even_r, odd_r, E[0], n_ph,data_size, data_w, data_h,temp1,temp2);

	//pi/2
	for(int i=0; i<2; i++) //left right
	{
		//CUDA_SAFE_CALL( cudaMemcpy(d_Kernel, F[0], KERNEL_SIZE, cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[0], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp1,I[i],data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[1], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(even[i],temp1,data_w,data_h);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[2], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(odd[i],temp1,data_w,data_h);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		if(scale!=0 && i==0)
		{

			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
		}
		if(i==1)
		{
			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}
	}

	compute_energy(even[0], odd[0], even_r, odd_r, E[4], n_ph,data_size, data_w, data_h,temp1,temp2);




	//%%%%%%%%%%%%
	//% Diagonal %
	//%%%%%%%%%%%%
	for(int i=0; i<2; i++) //left right
	{
		//CUDA_SAFE_CALL( cudaMemcpy(d_Kernel, F[3], KERNEL_SIZE, cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[3], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[3], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp2,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[4], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp3,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );

		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[4], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[3], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp4,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[4], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp5,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );



		// pi/4

		matrixSubGPU<<<grid, block, 0>>>(even[i],temp2,temp5,data_w);

		matrixSumGPU<<<grid, block, 0>>>(odd[i],temp4,temp3, data_w);
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		if(scale!=0 && i==0)
		{


			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
		}
		if(i==1)
		{
			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}
	}

	compute_energy(even[0], odd[0], even_r, odd_r, E[2], n_ph,data_size, data_w, data_h,temp1,temp2);



	// 3pi/4
	for(int i=0; i<2; i++) //left right
	{
		//CUDA_SAFE_CALL( cudaMemcpy(d_Kernel, F[3], KERNEL_SIZE, cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[3], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[3], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp2,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[4], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp3,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[4], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[3], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp4,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[4], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp5,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );

		matrixSumGPU<<<grid, block, 0>>>(even[i],temp2,temp5,data_w);

		matrixSubGPU<<<grid, block, 0>>>(odd[i],temp4,temp3,data_w);
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		if(scale!=0 && i==0)
		{


			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
		}
		if(i==1)
		{
			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}
	}
	compute_energy(even[0], odd[0], even_r, odd_r, E[6], n_ph,data_size, data_w, data_h,temp1,temp2);







	//%%%%%%%%%%%%%%%%%
	//% 'In-betweens' %
	//%%%%%%%%%%%%%%%%%


	for(int i=0; i<2; i++) //left right
	{
		//CUDA_SAFE_CALL( cudaMemcpy(d_Kernel, F[7], KERNEL_SIZE, cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[7], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[5], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp2,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[6], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp3,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaThreadSynchronize() );


		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[8], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[5], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp4,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[6], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp5,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );


		// pi/8

		matrixSubGPU<<<grid, block, 0>>>(even[i],temp2,temp5,data_w);

		matrixSumGPU<<<grid, block, 0>>>(odd[i],temp4,temp3,data_w);
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		if(scale!=0 && i==0)
		{

			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
		}
		if(i==1)
		{
			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}

	}
	compute_energy(even[0], odd[0], even_r, odd_r, E[1], n_ph,data_size, data_w, data_h,temp1,temp2);


	for(int i=0; i<2; i++) //left right
	{
		//CUDA_SAFE_CALL( cudaMemcpy(d_Kernel, F[7], KERNEL_SIZE, cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[7], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[5], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp2,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[6], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp3,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaThreadSynchronize() );

		//CUDA_SAFE_CALL( cudaMemcpy(d_Kernel, F[8], KERNEL_SIZE, cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[8], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[5], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp4,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[6], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp5,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		// 7pi/8

		matrixSumGPU<<<grid, block, 0>>>(even[i],temp2,temp5,data_w);

		matrixSubGPU<<<grid, block, 0>>>(odd[i],temp4,temp3,data_w);
		CUDA_SAFE_CALL( cudaThreadSynchronize() );

		if(scale!=0 && i==0)
		{

			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
		}
		if(i==1)
		{
			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}
	}

	compute_energy(even[0], odd[0], even_r, odd_r, E[7], n_ph,data_size, data_w, data_h,temp1,temp2);




	for(int i=0; i<2; i++) //left right
	{

		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[5], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[7], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp2,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[8], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp3,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );


		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[6], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[7], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp4,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[8], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp5,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );



		// pi/4

		matrixSubGPU<<<grid, block, 0>>>(even[i],temp2,temp5,data_w);

		matrixSumGPU<<<grid, block, 0>>>(odd[i],temp4,temp3,data_w);
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		if(scale!=0 && i==0)
		{


			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
		}
		if(i==1)
		{
			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}
	}

	compute_energy(even[0], odd[0], even_r, odd_r, E[3], n_ph,data_size, data_w, data_h,temp1,temp2);



	for(int i=0; i<2; i++) //left right
	{

		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[5], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[7], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp2,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[8], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp3,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );


		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[6], KERNEL_SIZE) );
		convolutionRowGPU<<<blockGridRows, threadBlockRows, 0>>>(temp1,I[i],data_w,data_h);
		//CUT_CHECK_ERROR("convolutionRowGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[7], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp4,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol(d_Kernel, F[8], KERNEL_SIZE) );
		convolutionColumnGPU<<<blockGridColumns, threadBlockColumns, 0>>>(temp5,temp1,data_w,data_h,COLUMN_TILE_W * (threadBlockColumns).y,data_w * (threadBlockColumns).y);
		//CUT_CHECK_ERROR("convolutionColumnGPU() execution failed\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		// 3pi/4

		matrixSumGPU<<<grid, block, 0>>>(even[i],temp2,temp5,data_w);

		matrixSubGPU<<<grid, block, 0>>>(odd[i],temp4,temp3,data_w);

		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		if(scale!=0 && i==0)
		{

			CudaWarping(even[0], temp1, DYwarp, DXwarp, data_w, data_h, data_size);
			CudaWarping(odd[0], temp2, DYwarp, DXwarp, data_w, data_h, data_size);
			CUDA_SAFE_CALL( cudaMemcpy(even[0], temp1, data_size, cudaMemcpyDeviceToDevice) );
			CUDA_SAFE_CALL( cudaMemcpy(odd[0], temp2, data_size, cudaMemcpyDeviceToDevice) );
		}
		if(i==1)
		{
			shift_in_phase(even[i],odd[i],even_r,odd_r,data_w,data_h,data_size,temp1,temp2);

		}
	}
	compute_energy(even[0], odd[0], even_r, odd_r, E[5], n_ph,data_size, data_w, data_h,temp1,temp2);



	//END GABOR FILTERING AND ENERGY


	//Energy normalization
	int kmax=0, itmp=0, jtmp=0, tmp;
	float Emax=0;
	for(int i=0; i<n_ori; i++)
	{
		for(int j=0; j<n_ph; j++)
		{
			tmp=cublasIsamax(data_w*data_h,E[i][j],1);
			if(tmp>Emax)
			{
				kmax=tmp;
				itmp=i;
				jtmp=j;
			}
		}
	}

	float Em;
	CUDA_SAFE_CALL( cudaMemcpy(&Em, &E[itmp][jtmp][kmax], sizeof(float), cudaMemcpyDeviceToHost) );

	for(int i=0; i<n_ori; i++)
	{
		for(int j=0; j<n_ph; j++)
		{
			cublasSaxpy(data_w*data_h,1.0/Em,E[i][j],1,E_norm[i][j],1);
		}
	}



	float *temp6;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp6, data_size));
	float *temp7;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp7, data_size));
	float *temp8;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp8, data_size));
	float *temp9;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&temp9, data_size));

	float **d_array_E;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&d_array_E, n_ph*sizeof(float*)));

	//Disparity estimation for each orientation
	for(int i=0; i<n_ori; i++)
	{
		CUDA_SAFE_CALL( cudaMemcpy(d_array_E, E_norm[i], n_ph*sizeof(float*), cudaMemcpyHostToDevice) );
		Decode<<<grid, block, 0>>>(temp6, temp7,d_array_E, data_w,i);
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
		CUDA_SAFE_CALL( cudaMemcpy(dx[i], temp6, data_size, cudaMemcpyDeviceToDevice) );
		CUDA_SAFE_CALL( cudaMemcpy(dy[i], temp7, data_size, cudaMemcpyDeviceToDevice) );
		//CUT_CHECK_ERROR("Decode() execution failed\n");

	}


	//2D Disparity estimation
	float **d_array_DX;
	float **d_array_DY;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&d_array_DX, n_ori*sizeof(float*)));
	CUDA_SAFE_CALL( cudaMalloc( (void **)&d_array_DY, n_ori*sizeof(float*)));
	CUDA_SAFE_CALL( cudaMemcpy(d_array_DX, dx, n_ori*sizeof(float*), cudaMemcpyHostToDevice) );
	CUDA_SAFE_CALL( cudaMemcpy(d_array_DY, dy, n_ori*sizeof(float*), cudaMemcpyHostToDevice) );


	/*float *DX;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&DX, data_size));
	float *DY;
	CUDA_SAFE_CALL( cudaMalloc( (void **)&DY, data_size));*/

	FullDisparity<<<grid, block, 0>>>(DX,DY,d_array_DX,d_array_DY,data_w);
	CUDA_SAFE_CALL( cudaThreadSynchronize() );



	if(scale!=0)
	{
		cublasSaxpy(data_w*data_h,1.0,DXwarp,1,DX,1);
		cublasSaxpy(data_w*data_h,1.0,DYwarp,1,DY,1);
	}
	//CUDA_SAFE_CALL( cudaMemcpy(h_DX, DX, data_size, cudaMemcpyDeviceToHost) );
	//CUDA_SAFE_CALL( cudaMemcpy(h_DY, DY, data_size, cudaMemcpyDeviceToHost) );




	//Deallocazione matrici temporanee
	CUDA_SAFE_CALL( cudaFree(even[0]) );
	CUDA_SAFE_CALL( cudaFree(odd[0]) );

	CUDA_SAFE_CALL( cudaFree(even[1]) );
	CUDA_SAFE_CALL( cudaFree(odd[1]) );

	CUDA_SAFE_CALL( cudaFree(temp1) );
	CUDA_SAFE_CALL( cudaFree(temp2) );
	CUDA_SAFE_CALL( cudaFree(temp3) );
	CUDA_SAFE_CALL( cudaFree(temp4) );
	CUDA_SAFE_CALL( cudaFree(temp5) );
	CUDA_SAFE_CALL( cudaFree(temp6) );
	CUDA_SAFE_CALL( cudaFree(temp7) );
	CUDA_SAFE_CALL( cudaFree(temp8) );
	CUDA_SAFE_CALL( cudaFree(temp9) );

	//CUDA_SAFE_CALL( cudaFree(DX) );
	//CUDA_SAFE_CALL( cudaFree(DY) );

	CUDA_SAFE_CALL( cudaFree(d_array_DX) );
	CUDA_SAFE_CALL( cudaFree(d_array_DY) );
	CUDA_SAFE_CALL( cudaFree(d_array_E) );

	for(int i=0; i<n_ori; i++)
	{
		CUDA_SAFE_CALL( cudaFree(dx[i]) );
		CUDA_SAFE_CALL( cudaFree(dy[i]) );
	}

	for(int j=0; j<n_ori; j++)
	{
		for(int i=0; i<n_ph; i++)
		{
			CUDA_SAFE_CALL( cudaFree(E[j][i]) );
			CUDA_SAFE_CALL( cudaFree(E_norm[j][i]) );
		}
	}

	for(int i=0; i<n_ph; i++)
	{
		CUDA_SAFE_CALL( cudaFree(even_r[i]) );
		CUDA_SAFE_CALL( cudaFree(odd_r[i]) );
	}


}//End Of Disparity



extern "C" void shift_in_phase(float *even,float* odd, float** even_r, float** odd_r, int data_w, int data_h,int data_size,float* temp11, float* temp22)
{

	dim3 grid(iDivUp(data_w, TEMP_W), iDivUp(data_h, TEMP_H));
	dim3 block(TEMP_W, TEMP_H);

	//phase shifts
	double ph_shift[n_ph] ={3.14159265358979,	2.35619449019235,	1.57079632679490,	0.785398163397448,	0,	-0.785398163397448,	-1.57079632679490,	-2.35619449019235,	-3.14159265358979};

	for(int ph=0; ph<n_ph; ph++)
	{
		//CUDA_SAFE_CALL( cudaThreadSynchronize() );
		//G_even(:,:,orient,phase)=G_even_tmp*cos(ph)-G_odd_tmp*sin(ph);
		//G_odd(:,:,orient,phase)=G_odd_tmp*cos(ph)+G_even_tmp*sin(ph);
		float c=(float)cos(ph_shift[ph]);
		float s=(float)sin(ph_shift[ph]);

		cudaMemset(temp11,0,data_size);
		cudaMemset(temp22,0,data_size);


		cublasSaxpy(data_w*data_h,c,even,1,temp11,1);

		cublasSaxpy(data_w*data_h,s,odd,1,temp22,1);

		matrixSubGPU_temp<<<grid, block, 0>>>(even_r[ph],temp11,temp22,data_w);
		cudaThreadSynchronize();

		cudaMemset(temp11,0,data_size);
		cudaMemset(temp22,0,data_size);

		cublasSaxpy(data_w*data_h,c,odd,1,temp11,1);
		cublasSaxpy(data_w*data_h,s,even,1,temp22,1);

		matrixSumGPU_temp<<<grid, block, 0>>>(odd_r[ph],temp11,temp22, data_w);

	}

}

extern "C" void compute_energy(float* even, float* odd, float** even_r, float** odd_r, float** E, int n_ph, int data_size, int data_w, int data_h,float* temp11, float* temp22)
{

	dim3 grid1(iDivUp(data_w, TEMP_W), iDivUp(data_h, TEMP_H));
	dim3 block1(TEMP_W, TEMP_H);
	//dim3 grid(iDivUp(data_w, SUM_TILE_W), iDivUp(data_h, SUM_TILE_H));
	//dim3 block(SUM_TILE_W, SUM_TILE_H);
	for(int ph=0; ph<n_ph; ph++)
	{
		//E(:,:,o,ph_n) = sqrt((CL(:,:,o) + squeeze(CR(:,:,o,ph_n))).^2 + (SL(:,:,o) + squeeze(SR(:,:,o,ph_n))).^2);
		cublasSaxpy(data_w*data_h,1,even,1,even_r[ph],1);
		cublasSaxpy(data_w*data_h,1,odd,1,odd_r[ph],1);


		Sqr<<<grid1, block1, 0>>>(temp11, odd_r[ph], odd_r[ph], data_w);

		Sqr<<<grid1, block1, 0>>>(temp22, even_r[ph], even_r[ph], data_w);
		cudaThreadSynchronize();

		matrixSumGPU_temp<<<grid1, block1, 0>>>(E[ph],temp11,temp22, data_w);

		cudaThreadSynchronize();


	}

}


extern "C" void CudaWarping(float* input, float* output, float *d_X, float *d_Y, int data_w, int data_h, int data_size) {

	cudaArray* cu_array;

	//TEXTURE

	// allocate array and copy image data
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
	cudaMallocArray( &cu_array, &channelDesc, data_w, data_h );
	cudaMemcpyToArray( cu_array, 0, 0, input, data_size, cudaMemcpyDeviceToDevice);

	//set texture parameters
	texData.addressMode[0] = cudaAddressModeClamp;	//Wrap
	texData.addressMode[1] = cudaAddressModeClamp;
	texData.filterMode = cudaFilterModeLinear;
	texData.normalized = true;    // access with normalized texture coordinates

	// Bind the array to the texture
	cudaBindTextureToArray( texData, cu_array, channelDesc);

	//END TEXTURE PART

	dim3 threadBlock(WARP_TILE_W, WARP_TILE_H);
	dim3 blockGrid(iDivUp(data_w, threadBlock.x), iDivUp(data_h, threadBlock.y));

	///Main warping function
	
	
	inverseWarping_kernel<<<blockGrid, threadBlock>>>(d_X, d_Y, output, data_w, data_h);
	
	//CUT_CHECK_ERROR("WarpingGPU() execution failed\n");
	cudaThreadSynchronize();


	cudaUnbindTexture(texData);
	cudaFreeArray(cu_array);
	cudaThreadSynchronize();
	
	

}

extern "C" void Expand(float* input, float* output, int data_w, int data_h, int data_size) {


	cudaArray* cu_array;

	//TEXTURE

	// allocate array and copy image data
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
	cudaMallocArray( &cu_array, &channelDesc, data_w, data_h ); //data_w input
	cudaMemcpyToArray( cu_array, 0, 0, input, data_size, cudaMemcpyDeviceToDevice);

	// set texture parameters
	texData.addressMode[0] = cudaAddressModeClamp;	//Wrap
	texData.addressMode[1] = cudaAddressModeClamp;
	texData.filterMode = cudaFilterModeLinear;
	texData.normalized = true;    // access with normalized texture coordinates

	// Bind the array to the texture
	cudaBindTextureToArray( texData, cu_array, channelDesc);

	//END TEXTURE PART

	dim3 threadBlock(WARP_TILE_W, WARP_TILE_H);
	dim3 blockGrid(iDivUp(data_w*2, threadBlock.x), iDivUp(data_h*2, threadBlock.y));

	///Main expand function
	expandOnly_kernel<<<blockGrid, threadBlock>>>(output, data_w*2, data_h*2);
	//CUT_CHECK_ERROR("ExpandOnly() execution failed\n");

	cudaThreadSynchronize();

	cudaUnbindTexture(texData);
	cudaFreeArray(cu_array);

}
