/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __CU_MATH_H__
#define __CU_MATH_H__

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

__device__ inline int clamp( int val, int val_min, int val_max )
{
    if ( val > val_max ) return val_max;
    if ( val < val_min ) return val_min;
    return val;
}

__device__ inline float det_sym( const float& xx, const float& yy, const float& xy )
{
  return xx * yy - xy * xy;
}

__device__ inline float det_sym( const float3& M )
{
  return M.x * M.y - M.z * M.z;
}

__device__ inline 
float gauss_norm( float lxx, float lyy, float lxy )
{
  const float factor = 1.f /  ( 2 * M_PI );
  return rsqrtf( max( 0.00000001f, det_sym( lxx, lyy, lxy ) )) * factor;
}

__device__ inline 
void sym_inv2_inplace( float& dxx, float& dyy, float& dxy )
{
  float det = det_sym( dxx, dyy, dxy );
  if ( det > 0.0001f )
  {
    float tau   = 1.f / det;
    float tmpyy = dyy;
    dyy         = tau * dxx;
    dxx         = tau * tmpyy;
    dxy         = (-tau) * dxy;
  }
  else
  {
    dyy         = 1e10f;
    dxx         = 1e10f;
  }
}

__device__ inline
float3 sym_inv2( const float& Mx, const float& My, const float& Mz  )
{
  float3 N;  
  N.x = Mx;
  N.y = My;
  N.z = Mz;
  sym_inv2_inplace( N.x, N.y, N.z );
  return N;
}


__device__ inline
float3 sym_inv2( const float3& M )
{
  float3 N;
  float invdet = 1.0f / max( 0.000001f,det_sym( M ));
  N.x = M.y * invdet;
  N.y = M.x * invdet;
  N.z = -M.z * invdet;
  return N;
}

__device__ inline
float2 sym_mvp2( const float3& M, const float2& x )
{
  float2 y;
  y.x = M.x * x.x + M.z * x.y;
  y.y = M.y * x.y + M.z * x.x;
  return y;
}

__device__ inline
float2 mul( const float2& x, const float& v )
{
  float2 y;
  y.x = x.x * v;
  y.y = x.y * v;
  return y;
}

__device__ inline
float3 mul( const float3& x, const float& v )
{
  float3 y;
  y.x = x.x * v;
  y.y = x.y * v;
  y.z = x.z * v;
  return y;
}

__device__
inline float gauss2d( const float& x, const float& y, const float& lxx, const float& lyy, const float& lxy)
{ 
  const float factor = 1.f /  ( 2 * M_PI );
  float tau = 1.0f / max( 0.00001f, det_sym( lxx, lyy, lxy ));
  float exp_val = -0.5f * ( lyy * x*x - 2*lxy * x*y + lxx * y*y );  
  return factor * sqrtf(tau) * exp( exp_val );
}

__device__
inline float gauss2dinv( const float& x, const float& y, const float& tauxx, const float& tauyy, const float& tauxy)
{ 
  const float factor = 1.f /  ( 2 * M_PI );
  float tau = det_sym( tauxx, tauyy, tauxy );
  float exp_val = -0.5f * ( tauxx * x*x + 2*tauxy * x*y + tauyy * y*y );  
  
  return factor * sqrtf(tau) * exp( exp_val );
}

/*
__device__ void
combine_gaussians( float s, float3 Tau1, float2 mu1, float3 Tau2, float2 mu2, float3 Tau12, float2 Mu12, bool return_covariance = true )
{
  Tau12 = Tau1 + mul(Tau2, s);
  float3 Lambda12 = sym_inv2( Tau12 );
  mu12 = sym_mvp( Lambda12, sym_mvp2( Tau1, mu1 ) + mul(sym_mvp2( Tau2, mu2 ), s) );
  if ( return_covariance )
    Tau12 = Lambda12;
}
*/

__device__ inline void
combine_gaussians(  float s, 
                    const float& Tau1x, const float& Tau1y, const float& Tau1z,
                    const float& mu1x, const float& mu1y,
                    const float& Tau2x, const float& Tau2y, const float& Tau2z,
                    const float& mu2x, const float& mu2y,
                    float& Tau12x, float& Tau12y, float& Tau12z,
                    float& mu12x, float& mu12y )
{
  Tau12x = Tau1x + Tau2x * s;
  Tau12y = Tau1y + Tau2y * s;
  Tau12z = Tau1z + Tau2z * s;
 
  float3 Lambda12;
  float det = det_sym( Tau12x, Tau12y, Tau12z );
  if ( det > 0.0001f )
  {
    float invdet = 1.0f / det;
    Lambda12.x = Tau12y * invdet;
    Lambda12.y = Tau12x * invdet;
    Lambda12.z = -Tau12z * invdet;

    float2 temp;
  
    /// temp = Tau1*mu1 + s*Tau2*mu2
    temp.x = Tau1x * mu1x + Tau1z * mu1y + s * ( Tau2x * mu2x + Tau2z * mu2y ); 
    temp.y = Tau1z * mu1x + Tau1y * mu1y + s * ( Tau2z * mu2x + Tau2y * mu2y );
  
    /// mu12 = Lambda12 * temp ( matrix-vector-product )
    mu12x = Lambda12.x * temp.x + Lambda12.z * temp.y;
    mu12y = Lambda12.z * temp.x + Lambda12.y * temp.y;
  }
  else
  {
    mu12x = 0.0f;
    mu12y = 0.0f;
  }
}










#endif
