/*Copyright or  or Copr. 
PSPC-lab - Biophysical and Electronic Engineering Department - University of Genoa
contributors: Manuela Chessa, Silvio P. Sabatini, Fabio Solari (February, 2011)
manuela.chessa@unige.it, silvio.sabatini@unige.it, fabio.solari@unige.it

This software is a computer program whose purpose is to compute the horizontal 
and vertical disparity maps from a stereo image pair.

This software is governed by the CeCILL-C  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.*/
//
//This software contains source code provided by NVIDIA Corporation.

#include "cuda.h"
#include "GPUDisparityInc.h"

__global__ void matrixSumGPU(
    float *d_Result,
    float *d_A,
	float *d_B,
    int dataW
){

	 //Data cache
    __shared__ float data_A[SUM_TILE_W * SUM_TILE_H];	
    __shared__ float data_B[SUM_TILE_W * SUM_TILE_H];	

    
    const int         tileStart = IMUL(blockIdx.y, SUM_TILE_H);
    const int       columnStart = IMUL(blockIdx.x, SUM_TILE_W) + threadIdx.x;

	//Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, SUM_TILE_W) + threadIdx.x;
    int gmemPos = IMUL(tileStart + threadIdx.y, dataW) + columnStart;

	
	data_A[smemPos] = d_A[gmemPos];
	data_B[smemPos] = d_B[gmemPos];

    
	d_Result[gmemPos] = data_A[smemPos]+data_B[smemPos];

}

__global__ void matrixSumGPU_temp(
    float *d_Result,
    float *d_A,
	float *d_B,
    int dataW
){

	 //Data cache
    __shared__ float data_A[TEMP_W * TEMP_H];	
    __shared__ float data_B[TEMP_W * TEMP_H];	

    
    const int         tileStart = IMUL(blockIdx.y, TEMP_H);
    const int       columnStart = IMUL(blockIdx.x, TEMP_W) + threadIdx.x;

	//Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, TEMP_W) + threadIdx.x;
    int gmemPos = IMUL(tileStart + threadIdx.y, dataW) + columnStart;

	
	data_A[smemPos] = d_A[gmemPos];
	data_B[smemPos] = d_B[gmemPos];

    
	d_Result[gmemPos] = data_A[smemPos]+data_B[smemPos];

}

__global__ void matrixSubGPU_temp(
    float *d_Result,
    float *d_A,
	float *d_B,
    int dataW
){

	 //Data cache
    __shared__ float data_A[TEMP_W * TEMP_H];	
    __shared__ float data_B[TEMP_W * TEMP_H];	

    
    const int         tileStart = IMUL(blockIdx.y, TEMP_H);
    const int       columnStart = IMUL(blockIdx.x, TEMP_W) + threadIdx.x;

	//Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, TEMP_W) + threadIdx.x;
    int gmemPos = IMUL(tileStart + threadIdx.y, dataW) + columnStart;

	
	data_A[smemPos] = d_A[gmemPos];
	data_B[smemPos] = d_B[gmemPos];

    
	d_Result[gmemPos] = data_A[smemPos]-data_B[smemPos];

}




__global__ void matrixSubGPU(
    float *d_Result,
    float *d_A,
	float *d_B,
    int dataW
){
	    //Data cache
    __shared__ float data_A[SUM_TILE_W * SUM_TILE_H];	
    __shared__ float data_B[SUM_TILE_W * SUM_TILE_H];	

    
    const int         tileStart = IMUL(blockIdx.y, SUM_TILE_H);
    const int       columnStart = IMUL(blockIdx.x, SUM_TILE_W) + threadIdx.x;

	//Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, SUM_TILE_W) + threadIdx.x;
    int gmemPos = IMUL(tileStart + threadIdx.y, dataW) + columnStart;

	
	data_A[smemPos] = d_A[gmemPos];
	data_B[smemPos] = d_B[gmemPos];

    
	d_Result[gmemPos] = data_A[smemPos]-data_B[smemPos];

}

//Element by element multiplication 
__global__ void Sqr(
    float *d_Result,
    float *d_A,
	float *d_B,
    int dataW
){

    //Data cache
    __shared__ float data_A[TEMP_W * TEMP_H];	
    __shared__ float data_B[TEMP_W * TEMP_H];	
	
    
    const int         tileStart = IMUL(blockIdx.y, TEMP_H);
    const int       columnStart = IMUL(blockIdx.x, TEMP_W) + threadIdx.x;

    //Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, TEMP_W) + threadIdx.x;
    int gmemPos = IMUL(tileStart + threadIdx.y, dataW) + columnStart;

	
	data_A[smemPos] = d_A[gmemPos];
	data_B[smemPos] = d_B[gmemPos];


	d_Result[gmemPos] = data_A[smemPos]*data_B[smemPos];
}


//Population decoding
__global__ void Decode(
	float *d_Result_x,
	float *d_Result_y,
	float **d_A,  //orientation fixed (n_ph x data_size)
    int dataW,
	int ori
){

    //disparity values
	__shared__ float disp[n_ph];
	disp[0]=-2;
	disp[1]=-1.5;
	disp[2]=-1;
	disp[3]=-0.5;
	disp[4]=0;
	disp[5]=0.5;
	disp[6]=1;
	disp[7]=1.5;
	disp[8]=2;
	
	__shared__ float m[SUM_TILE_W_D * SUM_TILE_H_D]; //energy minimum
	__shared__ float M[SUM_TILE_W_D * SUM_TILE_H_D]; //energy maximum

	__shared__ float dtmp[n_ph][SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float d_Result[SUM_TILE_W_D * SUM_TILE_H_D]; //decoding result
	__shared__ float sum[SUM_TILE_W_D * SUM_TILE_H_D];

    
    const int         tileStart = IMUL(blockIdx.y, SUM_TILE_H_D);
	const int       columnStart = IMUL(blockIdx.x, SUM_TILE_W_D) + threadIdx.x;

    //Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, SUM_TILE_W_D) + threadIdx.x;
    int gmemPos = IMUL(tileStart + threadIdx.y, dataW) + columnStart;
	

	sum[smemPos]=0;
	d_Result[smemPos]=0;

	m[smemPos]=d_A[0][gmemPos];
	M[smemPos]=d_A[0][gmemPos];
	for(int i=0; i<n_ph; i++)
	{
		if(d_A[i][gmemPos]<m[smemPos])
			m[smemPos]=d_A[i][gmemPos];
		if(d_A[i][gmemPos]>M[smemPos])
			M[smemPos]=d_A[i][gmemPos];
		
	}

	//weighted sum
	for(int i=0; i<n_ph; i++)
	{
		dtmp[i][smemPos]=(d_A[i][gmemPos]-m[smemPos])/(M[smemPos]-m[smemPos]);
		if(dtmp[i][smemPos]<0.7)
			dtmp[i][smemPos]=0;
		sum[smemPos]+=dtmp[i][smemPos];
		d_Result[smemPos] += dtmp[i][smemPos]*disp[i];
	}
	d_Result[smemPos]=d_Result[smemPos]*(1.0/sum[smemPos]);


	if(M[smemPos]<0.000001)
	{
			d_Result[smemPos]=NEW_NAN;
	}
	//x and y disparity component
	d_Result_x[gmemPos]=d_Result[smemPos]*cos(ori*3.14159265358979/8.0);
	d_Result_y[gmemPos]=d_Result[smemPos]*sin(ori*3.14159265358979/8.0);
	
}

//2D disparity - intersection of constraints
__global__ void FullDisparity(
	float *d_Result_x, //horizontal disparity
	float *d_Result_y, //vertical disparity
	float **dx,		   //horizontal disparity along each orientation
	float **dy,		//vertical disparity along each orientation
    int dataW
){



    const int         tileStart = IMUL(blockIdx.y, SUM_TILE_H_D);
    const int       columnStart = IMUL(blockIdx.x, SUM_TILE_W_D) + threadIdx.x;

    //Shared and global memory indices for current column
    int smemPos = IMUL(threadIdx.y, SUM_TILE_W_D) + threadIdx.x;
    int gmemPos = IMUL(tileStart + threadIdx.y, dataW) + columnStart;
	

	__shared__ float sumx[SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float sumy[SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float sumxyl2[SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float sumxxl2[SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float sumyyl2[SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float den[SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float L2[n_ori][SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float x[n_ori][SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ float y[n_ori][SUM_TILE_W_D * SUM_TILE_H_D];
	__shared__ int n[SUM_TILE_W_D * SUM_TILE_H_D];

	n[smemPos]=0;
	sumx[smemPos]=0;
	sumy[smemPos]=0;
	sumxyl2[smemPos]=0;
	sumxxl2[smemPos]=0;
	sumyyl2[smemPos]=0;

	for(int i=0; i<n_ori; i++)
	{

		x[i][smemPos]=dx[i][gmemPos];
		y[i][smemPos]=dy[i][gmemPos];
	
	}

	for(int i=0; i<n_ori; i++)
	{
	
		/*sumX = sum(X);
             sumY = sum(Y);
             sumXYL_2 = sum(X.*Y./L_2);
             sumXXL_2 = sum(X.^2./L_2);
             sumYYL_2 = sum(Y.^2./L_2);
             den = (sumXYL_2^2-sumXXL_2*sumYYL_2);*/
		L2[i][smemPos]=(x[i][smemPos]*x[i][smemPos])+(y[i][smemPos]*y[i][smemPos]);
		sumx[smemPos]+=x[i][smemPos];
		sumy[smemPos]+=y[i][smemPos];
		if(L2[i][smemPos]>0.0001)
		{
			n[smemPos]++;
			sumxyl2[smemPos]+=(x[i][smemPos]*y[i][smemPos])/L2[i][smemPos];
			sumxxl2[smemPos]+=(x[i][smemPos]*x[i][smemPos])/L2[i][smemPos];
			sumyyl2[smemPos]+=(y[i][smemPos]*y[i][smemPos])/L2[i][smemPos];
		}
		
		

	}
			/*xr = -(sumX*sumYYL_2-sumY*sumXYL_2) / den;
                 yr = (sumX*sumXYL_2-sumY*sumXXL_2) / den;
                 O(i,j,:) = [xr yr];*/
		den[smemPos] = ((sumxyl2[smemPos]*sumxyl2[smemPos])-(sumxxl2[smemPos]*sumyyl2[smemPos]));
		if((den[smemPos]>0.0001 || den[smemPos]<-0.0001) && n[smemPos]>4 )
		{
			d_Result_x[gmemPos]=-((sumx[smemPos]*sumyyl2[smemPos])-(sumy[smemPos]*sumxyl2[smemPos]))/den[smemPos];
			d_Result_y[gmemPos]=((sumx[smemPos]*sumxyl2[smemPos])-(sumy[smemPos]*sumxxl2[smemPos]))/den[smemPos];
		}
		else
		{
			d_Result_x[gmemPos]=0;
			d_Result_y[gmemPos]=0;
		}
}

