/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"
#include "stdio.h"

#ifdef _WIN32
	#define PRAGMA(x) __pragma(#x)
#else
	#define PRAGMA(x) _Pragma(#x)
#endif

#define sum_combine(x,y) x += y
#define max_combine(x,y) x = fmaxf( x, y )
#define min_combine(x,y) x = fminf( x, y )

static texture<float, 2> _down_texture;

__global__ void d_downsample_cont_rm( float* res, int n1, int n2, int nk, int down_factor, int shift,
									  const float* src, int sx, int mode )
{
// index Offset
  const int i1  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  
  if ( i1 >= n1 || i2 >= n2 ) return;
  
  const int i_out  = ( ( i1 * n2 ) + i2 ) * nk;
  int i_in  = ( i1 * sx + i2 * nk ) * shift;

  for (int k = 0; k < nk; ++k )
	res[ i_out + k ] = 0.0f;

//#define sum_combine(x,y) x += y
//#define max_combine(x,y) x = fmaxf( x, y )

#define doit(combiner)											\
	  for (int i = 0; i < down_factor; ++i, i_in += sx )		\
		{														\
		int ij_in = i_in;										\
		for (int j = 0; j < down_factor; ++j, ij_in += nk )		\
		  {														\
		  for (int k = 0; k < nk; ++k )							\
			combiner(res[ i_out + k ], src[ ij_in + k ]);		\
		  }														\
	}

#define doit_unroll(combiner,n)									\
	  PRAGMA( unroll n )										\
  for (int i = 0; i < n; ++i, i_in += sx )						\
	{															\
	  int ij_in = i_in;											\
	  PRAGMA( unroll n )										\
	  for (int j = 0; j < n; ++j, ij_in += nk )					\
		{														\
		  for (int k = 0; k < nk; ++k )							\
			combiner(res[ i_out + k ], src[ ij_in + k ]);		\
		}														\
	}


#define doit_dispatch(combiner)								\
  switch(down_factor)										\
	{														\
	case 1: doit_unroll(combiner,1); break;					\
	case 2: doit_unroll(combiner,2); break;					\
	case 3: doit_unroll(combiner,3); break;					\
	case 4: doit_unroll(combiner,4); break;					\
	case 5: doit_unroll(combiner,5); break;					\
	case 6: doit_unroll(combiner,6); break;					\
	case 7: doit_unroll(combiner,7); break;					\
	case 8: doit_unroll(combiner,8); break;					\
	default: doit(combiner);break;							\
	}
  
  if ( mode == 1 || mode == 3 )
	{
	  
	  doit_dispatch( sum_combine );
	  
	  if ( mode == 1 )
		{
		  const float scale = 1.0f / ( down_factor * down_factor );
		  for (int k = 0; k < nk; ++k )
			res[ i_out + k ] *= scale;
		}
	  

	}
  if ( mode == 2 )
	{

	  doit_dispatch( max_combine );

	}

}


__global__ void d_downsample_array( float* dst, int n1, int n2, int shift1, int shift2,
                                    int w, int h, int feature_depth, int mode )
{
  const int i1  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  
  int dst_pos = (i2 * w + i1 ) * feature_depth;
  float res = 0.f;
  float val = 1.f; // mode == "max" or "sum"
  if ( mode == 1 ) // mode == "mean"
    val /= (n1*n2);
  
  
  if ( i1 >= w || i2 >= h ) return;
  
#define do_iteration( combiner, init )  	  \
  for ( int k = 0; k < feature_depth; ++k )   \
  {                                           \
    res = init;                                \
    for ( int i = 0; i < n1; ++i )             \
    for ( int j = 0; j < n2; ++j )             \
    {                                         \
      combiner( res, tex2D(_down_texture, (i1 * shift1 + i)*feature_depth +k, (i2 * shift2 + j) ) );   \
    }                                         \
    res *= val;                               \
    dst[ dst_pos + k ] = res;                 \
  }
  
  if ( mode == 1 || mode == 3 )   // mode == "mean" or "sum"
    do_iteration( sum_combine, 0 );
  if ( mode == 2 )                // mode == "max"
    do_iteration( max_combine, -1e10 );
  if ( mode == 4 )                // mode == "min"
    do_iteration( min_combine, 1e10 );
  
}

__global__ void d_downsample_array_ff( float* dst, int n1, int n2, int shift1, int shift2,
                                       int w, int h, int feature_depth, int mode )
{
	const int i1  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int i2  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  
  int dst_pos = i2 * w + i1;
  float res = 0.f;
  float val = 1.f; // mode == "max" or "sum"
  if ( mode == 1 ) // mode == "mean"
    val /= (n1*n2);
  
  
  if ( i1 >= w || i2 >= h ) return;
  
  


#define do_iteration( combiner, init )			  \
  for ( int k = 0; k < feature_depth; ++k )   \
  {                                           \
    res = init;                                \
    for ( int i = 0; i < n1; ++i )             \
    for ( int j = 0; j < n2; ++j )             \
    {                                         \
      combiner( res, tex2D(_down_texture, (i1 * shift1 + i)*feature_depth +k, (i2 * shift2 + j) ) );   \
    }                                         \
    res *= val;                               \
    dst[ dst_pos * k ] = res;                 \
  }
  
  if ( mode == 1 || mode == 3 )   // mode == "mean" or "sum"
    do_iteration( sum_combine, 0 );
  if ( mode == 2 )                // mode == "max"
    do_iteration( max_combine, -1e10 );
   if ( mode == 4 )                // mode == "min"
	 do_iteration( min_combine, 1e10 );
 
}


extern "C" void h_downsample_cont_rm( float* res, int n1, int n2, int nk, int down_factor, int shift,
									  const float* src, int sx, int mode )
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d( iDivUp( n1, dimBlock2d.x ), iDivUp( n2, dimBlock2d.y ) );
  d_downsample_cont_rm<<< dimGrid2d, dimBlock2d >>>( res, n1, n2, nk, down_factor, shift, src, sx, mode );
}  



extern "C" void downsample_array( float* dst, int n1, int n2, int shift1, int shift2, const cudaArray* src, int w, int h, int feature_depth, int mode, bool features_first )
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d( iDivUp( w, dimBlock2d.x ), iDivUp( h, dimBlock2d.y ) );
  
  // Bind the array to the texture
  _down_texture.normalized = false;
  _down_texture.filterMode = cudaFilterModePoint;
  _down_texture.addressMode[0] = cudaAddressModeClamp;
  _down_texture.addressMode[1] = cudaAddressModeClamp;
  cudaBindTextureToArray(_down_texture, src);
  
  if (!features_first)
    d_downsample_array<<< dimGrid2d, dimBlock2d >>>( dst, n1, n2, shift1, shift2, w, h, feature_depth, mode );
  else
    d_downsample_array_ff<<< dimGrid2d, dimBlock2d >>>( dst, n1, n2, shift1, shift2, w, h, feature_depth, mode );
  // clean up texture
  cudaUnbindTexture(_down_texture);
  cudaThreadSynchronize();
  
}

