/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif
#ifndef SQR
#define SQR(x) ((x) * (x))
#endif
#ifndef NORM_ANGLE
#define NORM_ANGLE(theta) ((theta) < 0.0 ?				\
    			   fmod((theta), (float) (2.0 * M_PI)) + 2.0 * M_PI \
    			   : fmod((theta), (float) (2.0 * M_PI)))
#endif

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

static inline float __device__
sqr_gaussian (float x, float sigma)
{
  return exp (-x/(2*sigma*sigma));
}

extern "C" __global__ void
heth_diffusion(float* out, const float* in, const float* phi,
	       float sigma, int width, int height, int ddim)
{
  int x = threadIdx.x + blockDim.x * blockIdx.x;
  int y = threadIdx.y + blockDim.y * blockIdx.y;

  if (x >= width || y >= height)
    return;

  int nsz = (int) MAX(3., 3. * sigma);
  int a_min = MAX(0, x - nsz);
  int a_max = MIN(width - 1, x + nsz);
  int b_min = MAX(0, y - nsz);
  int b_max = MIN(height - 1, y + nsz);

  float sum = 0.0;
  float dsum = 0.0;
  for (int b = b_min; b <= b_max; b++)
    for (int a = a_min; a <= a_max; a++)
      {
	int dx = a - x;
	int dy = b - y;

	float theta = NORM_ANGLE(atan2f(dy, dx));
	float sqrho = SQR(dx) + SQR(dy);
	int pi_dir = (int) (theta / (2. * M_PI)) * ddim;

	float d = sqr_gaussian(sqrho, sigma)
	  * phi[a + width * (b + height * pi_dir)];
	sum += d * in[a + width * b];
	dsum += d;
      }
  out[x + width * y] = sum / dsum;
  //debug("Writing to out[%d]\n", x + width * y);
}

extern "C" void
do_heth_diffusion(float* out, const float* in, const float* phi,
	       float sigma, int width, int height, int ddim)
{
	dim3 dimBlock(16,16);
	dim3 dimGrid((width+16-1)/16,(height+16-1)/16);

	heth_diffusion<<<dimGrid,dimBlock>>>(out,in,phi,sigma,width,height,ddim);
}


