/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include <stdio.h>
#ifndef _WIN32
  #include <device_functions.h>
#else
  #include <cuda_runtime.h>
#endif
#include "cutil.h"
#include "GPUDataBase.h"

struct kernel
{
  float w, v1, v2;
};

template<typename T> 
__device__ void swap( T & a, T & b )
{
  T & c = a;
  a = b;
  b = c;
}

#include "block_reduce.h"

struct heavier_kernel
{
  __device__ kernel operator()( const kernel & k1, const kernel & k2 ) const
  {
	return k1.w > k2.w ? k1 : k2;
  }
};

__global__ void d_hypos_mean_shift_smooth
( const float * in_hypos, int in_width, int in_height, int in_depth
  , float * out_hypos,  int out_width, int out_height, int out_depth
  , int window_size, int window_shift
  , float v1min, float v1max, int n1
  , float v2min, float v2max, int n2
  , float sigma_space, float sigma1, float sigma2
  , int iterations	
  )
{

  // this is the target pixel
  const int xout = blockIdx.x;
  const int yout = blockIdx.y;

  // these are the upper left of the input window
  const int xin  = xout * window_shift;
  const int yin  = yout * window_shift;
  
  extern __shared__ kernel kernels[];

  const size_t id = threadIdx.y * blockDim.x + threadIdx.x;
  const size_t N  = blockDim.x * blockDim.y;

  kernel current;
  kernel & next    = kernels[ id     ];

  current.v1 = v1min + ( v1max - v1min ) * float( threadIdx.x + 1 ) / float( n1 + 1 );
  current.v2 = v2min + ( v2max - v2min ) * float( threadIdx.y + 1 ) / float( n2 + 1 );

  const kernel * in_kernels = ( ( const kernel * ) in_hypos ) + ( yin * in_width + xin ) * in_depth;
  kernel * out_kernel =  ( ( kernel * ) out_hypos ) + ( yout * out_width + xout ) * out_depth;
  
  const float tau_space = 1.0f / ( sigma_space * sigma_space );
  const float tau1 = 1.0f / ( sigma1 * sigma1 );
  const float tau2 = 1.0f / ( sigma2 * sigma2 );

  for ( size_t iteration = 0; iteration < iterations; ++iteration )
	{	  	  
	  next.w = 0.0001;
	  next.v1 = 0;
	  next.v2 = 0;

	  for ( size_t i = 0; i < window_size; ++i )
		for ( size_t j = 0; j < window_size; ++j )
		  {
			const float dy = ( i - window_size / 2 );
			const float dx = ( j - window_size / 2 );
			const float lp_space = -tau_space * ( dy * dy + dx * dx );
			const kernel * in_col = in_kernels + ( i * in_width + j ) * in_depth;
			for ( size_t k = 0; k < in_depth; ++k )
			  {
				const kernel & in_kernel = in_col[k];
				const float dv1 = in_kernel.v1 - current.v1;
				const float dv2 = in_kernel.v2 - current.v2;
				const float lp = lp_space - tau1 * dv1 * dv1 - tau2 * dv2 * dv2;
				const float w = expf( lp ) * in_kernel.w;
				next.w += w;
				next.v1 += w * in_kernel.v1;
				next.v2 += w * in_kernel.v2;
			  }
		  }
	  next.v1 /= next.w;
	  next.v2 /= next.w;
	  current = next;
	}

  __syncthreads();

  block_reduce( id, N, kernels, 1, heavier_kernel(), out_depth );

  __syncthreads();

  if ( id < out_depth )
	out_kernel[id] = next;

}

struct weight_sum
{
  __device__ kernel operator()( const kernel & k1, const kernel & k2 ) const
  {
	kernel result = k1;
	result.w += k2.w;
	return result;
  }
};


__global__ void d_hypos_gmm_smooth
( const float * in_hypos, int in_width, int in_height, int in_depth
  , float * out_hypos,  int out_width, int out_height, int out_depth
  , int window_size, int window_shift
  , float v1min, float v1max, int n1
  , float v2min, float v2max, int n2
  , float sigma_space, float sigma1, float sigma2
  , int iterations	
  )
{

  // this is the target pixel
  const int xout = blockIdx.x;
  const int yout = blockIdx.y;

  // these are the upper left of the input window
  const int xin  = xout * window_shift;
  const int yin  = yout * window_shift;
  
  extern __shared__ kernel kernels[];

  const size_t id = threadIdx.y * blockDim.x + threadIdx.x;
  const size_t N  = blockDim.x * blockDim.y;

  kernel current;
  kernel & next    = kernels[ id ];

  current.v1 = v1min + ( v1max - v1min ) * float( threadIdx.x + 1 ) / float( n1 + 1 );
  current.v2 = v2min + ( v2max - v2min ) * float( threadIdx.y + 1 ) / float( n2 + 1 );

  const kernel * in_kernels = ( ( const kernel * ) in_hypos ) + ( yin * in_width + xin ) * in_depth;
  kernel * out_kernel =  ( ( kernel * ) out_hypos ) + ( yout * out_width + xout ) * out_depth;
  
  const float tau_space = 1.0f / ( sigma_space * sigma_space );
  const float tau1 = 1.0f / ( sigma1 * sigma1 );
  const float tau2 = 1.0f / ( sigma2 * sigma2 );

  for ( size_t iteration = 0; iteration < iterations; ++iteration )
	{	  	  
	  float wsum = 0.0001;
	  next.v1 = 0;
	  next.v2 = 0;
	  for ( size_t i = 0; i < window_size; ++i )
		for ( size_t j = 0; j < window_size; ++j )
		  {
			const float dy = ( i - window_size / 2 );
			const float dx = ( j - window_size / 2 );
			const float lp_space = -tau_space * ( dy * dy + dx * dx );
			const kernel * in_col = in_kernels + ( i * in_width + j ) * in_depth;
			for ( size_t k = 0; k < in_depth; ++k )
			  {
				const kernel & in_kernel = in_col[k];
				const float dv1 = in_kernel.v1 - current.v1;
				const float dv2 = in_kernel.v2 - current.v2;
				const float lp = lp_space - tau1 * dv1 * dv1 - tau2 * dv2 * dv2;
				float w = max( 0.0000000001f, expf( lp ) * in_kernel.w * current.w );
				next.w = w;
				__syncthreads();
				block_reduce( id, N, kernels, 1, weight_sum(), 1 );
				__syncthreads();
				w /= kernels[0].w;
				wsum += w;
				next.v1 += w * in_kernel.v1;
				next.v2 += w * in_kernel.v2;				
			  }
		  }
	  next.v1 /= wsum;
	  next.v2 /= wsum;
	  next.w = wsum;
	  current = next;
	}


  block_reduce( id, N, kernels, 1, heavier_kernel(), out_depth );

  __syncthreads();

  if ( id < out_depth )
	out_kernel[id] = next;

}

extern "C" void do_hypos_mean_shift_smooth
( const float * in_hypos, int in_width, int in_height, int in_depth
  , float * out_hypos,  int out_width, int out_height, int out_depth
  , int window_size, int window_shift
  , float v1min, float v1max, int n1
  , float v2min, float v2max, int n2
  , float sigma_space, float sigma1, float sigma2
  , int iterations	)
{
  dim3 dimBlock( n1, n2 );
  dim3 dimGrid( out_width, out_height );
  size_t n_shared = n1 * n2 * sizeof( kernel );

  
  d_hypos_mean_shift_smooth<<<dimGrid, dimBlock, n_shared >>>
	(
	 in_hypos, in_width, in_height, in_depth,
	 out_hypos, out_width, out_height, out_depth,
	 window_size, window_shift,
	 v1min, v1max, n1,
	 v2min, v2max, n2,
	 sigma_space, sigma1, sigma2,
	 iterations
	 );
  
}


extern "C" void do_hypos_gmm_smooth
( const float * in_hypos, int in_width, int in_height, int in_depth
  , float * out_hypos,  int out_width, int out_height, int out_depth
  , int window_size, int window_shift
  , float v1min, float v1max, int n1
  , float v2min, float v2max, int n2
  , float sigma_space, float sigma1, float sigma2
  , int iterations	)
{
  dim3 dimBlock( n1, n2 );
  dim3 dimGrid( out_width, out_height );
  size_t n_shared = n1 * n2 * sizeof( kernel );

  
  d_hypos_gmm_smooth<<<dimGrid, dimBlock, n_shared >>>
	(
	 in_hypos, in_width, in_height, in_depth,
	 out_hypos, out_width, out_height, out_depth,
	 window_size, window_shift,
	 v1min, v1max, n1,
	 v2min, v2max, n2,
	 sigma_space, sigma1, sigma2,
	 iterations
	 );
  
}
