/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"

__global__ void inhibAndRect_d(const int N, float* res, float* exita, float* inhib, const float subsInhibFactor, const float divInhibFactor, const float divInhibConstant) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// no of threads
		const int noOfThreads = blockDim.x;

		// Thread index
		const int tIndex = threadIdx.x;

		// index Offset
		const int xb = bIndex * noOfThreads;
		const int x  = xb + tIndex; // globalIndex

		inhib += xb;
		exita += xb;

		res += xb;

		if (x<N) {
			float om = inhib[tIndex];
			float op = exita[tIndex];

			float resL = (op-subsInhibFactor*om)/(divInhibConstant + divInhibFactor*om);

			res[tIndex] = fmaxf(resL,0.0f);
		}
}

__global__ void inhibAndRectSpecial_d(float* res, float* exita, float* inhib, const float subsInhibFactor, const float divInhibFactor, const float divInhibConstant) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// no of threads
		const int noOfThreads = blockDim.x;

		// Thread index
		const int tIndex = threadIdx.x;

		// index Offset
		const int xb = bIndex * noOfThreads;

		const float om = inhib[bIndex];

		exita += xb;

		res += xb;

		{
			float op = exita[tIndex];

			float resL = (op-subsInhibFactor*om)/(divInhibConstant + divInhibFactor*om);

			res[tIndex] = fmaxf(resL,0.0f);
		}
}

__global__ void inhibSpecial_d(float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// no of threads
		const int noOfThreads = blockDim.x;

		// Thread index
		const int tIndex = threadIdx.x;

		// index Offset
		const int xb = bIndex * noOfThreads;

		const float om = inhib[bIndex];

		exita += xb;

		res += xb;

		{
			float op = exita[tIndex];

			float resL = (op-subInhibFactor*om)/(divInhibConstant + divInhibFactor*om);

			res[tIndex] = fmaxf(resL,0.0f);
		}
}

__global__ void inhibSpecial2_d(float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor, const float divSelfInhibFactor) {
		// Block index
		const int bIndex = blockIdx.x+blockIdx.y*gridDim.x;

		// no of threads
		const int noOfThreads = blockDim.x;

		// Thread index
		const int tIndex = threadIdx.x;

		// index Offset
		const int xb = bIndex * noOfThreads;

		const float om = inhib[bIndex];

		exita += xb;

		res += xb;

		{
			float op = exita[tIndex];

			float resL = (op-subInhibFactor*om)/(divInhibConstant + divInhibFactor*om + divSelfInhibFactor*op);

			res[tIndex] = fmaxf(resL,0.0f);
		}
}

extern "C" void inhibitionAndRect(const int N, float* res, float* exita, float* inhib, const float subsInhibFactor, const float divInhibConstant, const float divInhibFactor) {
	dim3 dimBlock(128);
	int s  = iDivUp(N,dimBlock.x);
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);
    inhibAndRect_d<<<dimGrid,dimBlock>>>(N, res, exita, inhib, subsInhibFactor, divInhibFactor, divInhibConstant);
}

extern "C" void inhibitionAndRectSpecial(const int Nx, const int Ny, float* res, float* exita, float* inhib, const float subsInhibFactor, const float divInhibFactor, const float divInhibConstant) {
	dim3 dimBlock(Nx);
    int s  = Ny;
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);

    inhibAndRectSpecial_d<<<dimGrid,dimBlock>>>(res, exita, inhib, subsInhibFactor, divInhibFactor, divInhibConstant);
}

extern "C" void inhibitionSpecial(const int Nx, const int Ny, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor) {
	dim3 dimBlock(Nx);
	int s  = Ny;
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);

    inhibSpecial_d<<<dimGrid,dimBlock>>>(res, exita, inhib, subInhibFactor, divInhibConstant, divInhibFactor);
}

extern "C" void inhibitionSpecial2(const int Nx, const int Ny, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor, const float divSelfInhibFactor) {
	dim3 dimBlock(Nx);
	int s  = Ny;
	int s2 = 1;
	while (s>=65535) {
		s2 *= 2;
		s   = iDivUp(s,2);
	}
	dim3 dimGrid(s,s2);

    inhibSpecial2_d<<<dimGrid,dimBlock>>>(res, exita, inhib, subInhibFactor, divInhibConstant, divInhibFactor, divSelfInhibFactor);
}
