/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"

__device__ float d_d_interp2(const float ul, const float ur, const float dl, const float dr, const float x, const float y) {
	return (ul*(1-x)*(1-y)+ur*(x)*(1-y)+dl*(1-x)*(y)+dr*(x)*(y));
}

__global__ void d_interp2From( float* origin, float* xC, float* yC, const int NxO, const int NyO, 
							   const int pitchO, const int NxD, const int NyD, const int pitchD, float* dest
							   ) 
{

	const int indexX  = blockIdx.x * blockDim.x + threadIdx.x;
	const int indexY  = blockIdx.y * blockDim.y + threadIdx.y;
	const int index   = indexX + indexY * pitchD;

	const float xk = xC[index];
	const float yk = yC[index];

	int xMin = floorf(xk);
	int xMax = ceilf(xk);
	int yMin = floorf(yk);
	int yMax = ceilf(yk);

	if (xMin<0) xMin=0;
	if (xMax<0) xMax=0;
	if (xMin>=NxO) xMin=NxO-1;
	if (xMax>=NxO) xMax=NxO-1;
	if (yMin<0) yMin=0;
	if (yMax<0) yMax=0;
	if (yMin>=NyO) yMin=NyO-1;
	if (yMax>=NyO) yMax=NyO-1;

	float ul = origin[xMin+yMin*pitchO];
	float ur = origin[xMax+yMin*pitchO];
	float dl = origin[xMin+yMax*pitchO];
	float dr = origin[xMax+yMax*pitchO];

	if (indexX<NxD && indexY<NyD)
	  dest[index] = d_d_interp2(ul,ur,dl,dr,xk-xMin,yk-yMin);
	
}

__global__ void d_affineInverseWarpFrom(float* origin, const float p1, const float p2, const float p3, 
										const float p4, const float p5, const float p6, 
										const int NxO, const int NyO, const int pitchO,const int NxD,const int NyD,
										const int pitchD, float* dest, const float xCenter, const float yCenter)
{
	// no of threads per block
	const int noOfThreadsX = blockDim.x;
	const int noOfThreadsY = blockDim.y;

	// block Idx
	const int bIndexX = blockIdx.x;
	const int bIndexY = blockIdx.y;

	// blobal index
	const int indexX  = bIndexX * noOfThreadsX + threadIdx.x;
	const int indexY  = bIndexY * noOfThreadsY + threadIdx.y;
	const int index   = indexX+indexY*pitchD;

	const float xk = (1+p1)*(indexX-xCenter)+p3*(indexY-yCenter)+p5 + xCenter;
	const float yk = p2*(indexX-xCenter)+(1+p4)*(indexY-yCenter)+p6 + yCenter;

	int xMin = floorf(xk);
	int xMax = ceilf(xk);
	int yMin = floorf(yk);
	int yMax = ceilf(yk);

	if (xMin<0) xMin=0;
	if (xMax<0) xMax=0;
	if (xMin>=NxO) xMin=NxO-1;
	if (xMax>=NxO) xMax=NxO-1;
	if (yMin<0) yMin=0;
	if (yMax<0) yMax=0;
	if (yMin>=NyO) yMin=NyO-1;
	if (yMax>=NyO) yMax=NyO-1;

	float ul = origin[xMin+yMin*pitchO];
	float ur = origin[xMax+yMin*pitchO];
	float dl = origin[xMin+yMax*pitchO];
	float dr = origin[xMax+yMax*pitchO];

	if (indexX<NxD && indexY<NyD)
		dest[index] = d_d_interp2(ul,ur,dl,dr,xk-xMin,yk-yMin);
}

__global__ void d_affineInverseWarpFrom2(float* origin, float* p1, float* p2, float* p3, float* p4, float* p5, float* p6, const int NxO, const int NyO, const int pitchO,const int NxD,const int NyD,const int pitchD, float* dest, const float xCenter, const float yCenter)
{
	// no of threads per block
	const int noOfThreadsX = blockDim.x;
	const int noOfThreadsY = blockDim.y;

	// block Idx
	const int bIndexX = blockIdx.x;
	const int bIndexY = blockIdx.y;

	// blobal index
	const int indexX  = bIndexX * noOfThreadsX + threadIdx.x;
	const int indexY  = bIndexY * noOfThreadsY + threadIdx.y;
	const int index   = indexX+indexY*pitchD;

	const float p1L = p1[index];
	const float p2L = p2[index];
	const float p3L = p3[index];
	const float p4L = p4[index];
	const float p5L = p5[index];
	const float p6L = p6[index];

	const float xk = (1+p1L)*(indexX-xCenter)+p3L*(indexY-yCenter)+p5L + xCenter;
	const float yk = p2L*(indexX-xCenter)+(1+p4L)*(indexY-yCenter)+p6L + yCenter;

	int xMin = floorf(xk);
	int xMax = ceilf(xk);
	int yMin = floorf(yk);
	int yMax = ceilf(yk);

	if (xMin<0) xMin=0;
	if (xMax<0) xMax=0;
	if (xMin>=NxO) xMin=NxO-1;
	if (xMax>=NxO) xMax=NxO-1;
	if (yMin<0) yMin=0;
	if (yMax<0) yMax=0;
	if (yMin>=NyO) yMin=NyO-1;
	if (yMax>=NyO) yMax=NyO-1;

	float ul = origin[xMin+yMin*pitchO];
	float ur = origin[xMax+yMin*pitchO];
	float dl = origin[xMin+yMax*pitchO];
	float dr = origin[xMax+yMax*pitchO];

	if (indexX<NxD && indexY<NyD)
		dest[index] = d_d_interp2(ul,ur,dl,dr,xk-xMin,yk-yMin);
}


extern "C" void hinterp2From(float* origin, float* xC, float* yC, const int NxO, const int NyO, 
							 const int pitchO,const int NxD,const int NyD,const int pitchD, float* dest) 
{
    //	dim3 dimBlock(16,8);
    dim3 dimBlock = GPUData_dimBlock2d;
    dim3 dimGrid(iDivUp(NxD,dimBlock.x),iDivUp(NyD,dimBlock.y));

	d_interp2From<<<dimGrid, dimBlock>>>(origin,  xC,  yC, NxO, NyO, pitchO, NxD, NyD, pitchD, dest);
}

extern "C" void haffineInverseWarpFrom(float* origin, const float p1, const float p2, const float p3, 
									   const float p4, const float p5, const float p6, 
									   const int NxO, const int NyO, const int pitchO,const int NxD,const int NyD,
									   const int pitchD, float* dest, const float xCenter, const float yCenter)
{
    // dim3 dimBlock(16,8);
    dim3 dimBlock = GPUData_dimBlock2d;
    dim3 dimGrid(iDivUp(NxD,dimBlock.x),iDivUp(NyD,dimBlock.y));

	d_affineInverseWarpFrom<<<dimGrid, dimBlock>>>(origin, p1,p2,p3,p4,p5,p6, NxO, NyO, pitchO, NxD, NyD, pitchD, dest, xCenter, yCenter);
}

extern "C" void haffineInverseWarpFrom2( float* origin, float* p1, float* p2, float* p3, float* p4, float* p5, float* p6, 
										 const int NxO, const int NyO, const int pitchO,const int NxD,const int NyD,
										 const int pitchD, float* dest, const float xCenter, const float yCenter)
{
    // dim3 dimBlock(16,8);
    dim3 dimBlock = GPUData_dimBlock2d;
    dim3 dimGrid(iDivUp(NxD,dimBlock.x),iDivUp(NyD,dimBlock.y));

	d_affineInverseWarpFrom2<<<dimGrid, dimBlock>>>(origin, p1,p2,p3,p4,p5,p6, NxO, NyO, pitchO, NxD, NyD, pitchD, dest, xCenter, yCenter);
}
