/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"

__global__ void d_hypos_cartesian_log_compress( float * hypos, int width, int height, int depth )
{
  const int xout = blockIdx.x * blockDim.x + threadIdx.x;
  const int yout = blockIdx.y * blockDim.y + threadIdx.y;

  if ( xout < width && yout < height )
	{
	  int ihypos = ( xout + width * yout ) * depth;
	  for ( int i = 0; i < depth; i += 3 )
		{
		  float & vx = hypos[ i + ihypos + 1 ];
		  float & vy = hypos[ i + ihypos + 2 ];
		  const float r = sqrtf( vx * vx + vy * vy );
		  if ( r > 0 )
			{			
			  const float rho = log2f( 1.0f + r );
			  vx *= rho / r;
			  vy *= rho / r;
			}
		}
	}
  
}

__global__ void d_hypos_cartesian_log_expand( float * hypos, int width, int height, int depth )
{
  const int xout = blockIdx.x * blockDim.x + threadIdx.x;
  const int yout = blockIdx.y * blockDim.y + threadIdx.y;

  if ( xout < width && yout < height )
	{
	  int ihypos = ( xout + width * yout ) * depth;
	  for ( int i = 0; i < depth; i += 3 )
		{
		  float & vx = hypos[ i + ihypos + 1 ];
		  float & vy = hypos[ i + ihypos + 2 ];
		  const float rho = sqrtf( vx * vx + vy * vy );
		  if ( rho > 0 )
			{			
			  const float r = expf( rho ) - 1;
			  vx *= r / rho;
			  vy *= r / rho;
			}
		}
	}
  
}

__global__ void d_hypos_cartesian_to_log_polar( float * hypos, int width, int height, int depth )
{
  const int xout = blockIdx.x * blockDim.x + threadIdx.x;
  const int yout = blockIdx.y * blockDim.y + threadIdx.y;

  if ( xout < width && yout < height )
	{
	  int ihypos = ( xout + width * yout ) * depth;
	  for ( int i = 0; i < depth; i += 3 )
		{
		  const float vx = hypos[ i + ihypos + 1 ];
		  const float vy = hypos[ i + ihypos + 2 ];
		  const float r = sqrtf( vx * vx + vy * vy );
		  const float rho = log2f( 1.0f + r );
		  const float phi = atan2f( vx, vy );
		  hypos[ i + ihypos  + 1 ] = rho;
		  hypos[ i + ihypos  + 2 ] = phi;
		}
	}
  
}

__global__ void d_hypos_log_polar_to_cartesian( float * hypos, int width, int height, int depth )
{
  const int xout = blockIdx.x * blockDim.x + threadIdx.x;
  const int yout = blockIdx.y * blockDim.y + threadIdx.y;
  
  if ( xout < width && yout < height )
	{
	  int ihypos = ( xout + width * yout ) * depth;;
	  for ( int i = 0; i < depth; i += 3 )
		{
		  const float rho = hypos[ i + ihypos  + 1 ];
		  const float phi = hypos[ i + ihypos  + 2 ];
		  const float r = exp2f( rho ) - 1.0f;
		  const float vx = cosf( phi ) * r;
		  const float vy = sinf( phi ) * r;
		  hypos[ i + ihypos  + 1 ] = vx;
		  hypos[ i + ihypos  + 2 ] = vy;
		}
	}
  
}


extern "C" void do_log_polar_transform( float * hypos, int width, int height, int depth, int direction )
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
  
  if ( direction == 0 )
	d_hypos_cartesian_to_log_polar<<<dimGrid2d, dimBlock2d>>>( hypos, width, height, depth );
  else
	d_hypos_log_polar_to_cartesian<<<dimGrid2d, dimBlock2d>>>( hypos, width, height, depth );
}

extern "C" void do_log_transform( float * hypos, int width, int height, int depth, int direction )
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
  
  if ( direction == 0 )
	d_hypos_cartesian_log_compress<<<dimGrid2d, dimBlock2d>>>( hypos, width, height, depth );
  else
	d_hypos_cartesian_log_expand<<<dimGrid2d, dimBlock2d>>>( hypos, width, height, depth );
}
