/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cublas.h>
#include <cutil.h>

#ifndef MIN
#  define MIN(x,y) ((x)<(y)?(x):(y))
#endif
#ifndef MAX
#  define MAX(x,y) ((x)>(y)?(x):(y))
#endif

__global__ void modulated_diffusion( float * d_in, int n1, int n2, int s_in1, int s_in2,
									 const float * d_mod         , int s_mod1, int s_mod2,
									 float * d_out               , int s_out1, int s_out2,
									 float gmod, float g0, float sigma         )
{
  const int bx = (blockIdx.x * blockDim.x);
  const int by = (blockIdx.y * blockDim.y);
  const int i1  = (bx + threadIdx.x);
  const int i2  = (by + threadIdx.y);
  if ( i1 < n1 && i2 < n2 )
	{
	  const int range = (int) 2.5f * sigma;
	  const int i_out = i1 * s_out1 + i2 * s_out2 ;
	  const int i_mod = i1 * s_mod1 + i2 * s_mod2 ;
	  const int j1_lo = MAX( 0, i1 - range );
	  const int j2_lo = MAX( 0, i2 - range );
	  const int j1_hi = MIN( n1 - 1, i1 + range );
	  const int j2_hi = MIN( n2 - 1, i2 + range );
	  float outval = 0;
	  const float outmod = g0 + gmod * d_mod[i_mod];
	  const float lZ = logf( 2.0f * 3.14159f * sigma * sigma );
	  //const float c = 1.0f;
	  for ( int j1 = j1_lo; j1 <= j1_hi; ++j1 )
		{
		  const float d1 = j1 - i1;
		  const float d12 = d1 * d1;
		  for ( int j2 = j2_lo; j2 <= j2_hi; ++j2 )
		  {
			const float d2 = j2 - i2;
			const float d22 = d2 * d2;
			const float g  = expf( - ( d12 + d22 ) / ( 2 * sigma )  - lZ );
			//			float g = 1.0f;
			const int j_in  = j1 * s_in1  + j2 * s_in2 ;
			const int j_mod = j1 * s_mod1 + j2 * s_mod2;
			// gaussian envelope * gain * in
			// gain has a constant and a modulated contribution (g0 and gmod resp.)
			outval += g * ( gmod * d_mod[j_mod] + g0 ) * d_in[j_in];
		  }
		}
	  d_out[i_out] = outmod * outval;
	}
}

__global__ void directionally_modulated_diffusion( float * d_in, int n1, int n2   ,int s_in1 , int s_in2,
												   const float * d_mod, int n_mod ,int s_mod1, int s_mod2, int s_mod3,
												   float * d_out                  ,int s_out1, int s_out2,
												   float gmod, float g0, float sigma_long, float sigma_lat           )
{
  const int bx = (blockIdx.x * blockDim.x);
  const int by = (blockIdx.y * blockDim.y);
  const int i1  = (bx + threadIdx.x);
  const int i2  = (by + threadIdx.y);
  if ( i1 < n1 && i2 < n2 )
	{
	  int range = (int) 2.5f * sigma_long;
	  int i_out = i1 * s_out1 + i2 * s_out2 ;
	  int j1_lo = MAX( 0, i1 - range );
	  int j2_lo = MAX( 0, i2 - range );
	  int j1_hi = MIN( n1 - 1, i1 + range );
	  int j2_hi = MIN( n2 - 1, i2 + range );
	  float outval = 0;
	  const float lZ = logf( 2.0f * 3.14159f * sigma_lat * sigma_long );
	  //const float c = 1.0f;
	  for ( int j1 = j1_lo; j1 <= j1_hi; ++j1 )
		{
		  for ( int j2 = j2_lo; j2 <= j2_hi; ++j2 )
		  {
			const int j_in  = j1 * s_in1  + j2 * s_in2 ;
			for ( int jmod = 0; jmod < n_mod; ++jmod )
			  {
				const float mod = d_mod[ j1 * s_mod1 + j2 * s_mod2 + jmod * s_mod3 ];
				/*
				const float d1 = j1 - i1;
				const float d12 = d1 * d1;
				const float d2 = j2 - i2;
				const float d22 = d2 * d2;
				const float g  = expf( - ( d12 + d22 ) / ( 2 * sigma )  - lZ );
				//			float g = 1.0f;
				// gaussian envelope * gain * in
				// gain has a constant and a modulated contribution (g0 and gmod resp.)
				outval += g * ( gmod * d_mod[j_mod] + g0 ) * d_in[j_in];
				*/
			  }
		  }
		}
	  d_out[i_out] = outval;
	}
}

extern "C" void do_modulated_diffusion( float * d_in, int n1, int n2, int s_in1 , int s_in2,
										const float * d_mod         , int s_mod1, int s_mod2,
										float * d_out               , int s_out1, int s_out2,
										float gmod, float g0, float sigma         )
{

  dim3 dimBlock2d( 16 , 8 );
  dim3 dimGrid2d ( iDivUp( n1, dimBlock2d.x ), iDivUp( n2, dimBlock2d.y ) );
//  fprintf(stderr,"%d,%d ; %d,%d\n",dimBlock2d.x,dimBlock2d.y,dimGrid2d.x,dimGrid2d.y);
  modulated_diffusion <<< dimGrid2d, dimBlock2d >>> ( d_in, n1, n2, s_in1, s_in2,
													  d_mod, s_mod1, s_mod2,
													  d_out, s_out1, s_out2,
													  gmod, g0, sigma            );
  cudaError_t err = cudaGetLastError();
}

extern "C" void do_directionally_modulated_diffusion( float * d_in, int n1, int n2  , int s_in1 , int s_in2,
													  const float * d_mod, int n_mod, int s_mod1, int s_mod2, int s_mod3,
													  float * d_out                 , int s_out1, int s_out2,
													  float gmod, float g0, float sigma_long, float sigma_lat         )
{
  dim3 dimBlock2d( 16 , 16 );
  dim3 dimGrid2d ( iDivUp( n1, dimBlock2d.x ), iDivUp( n2, dimBlock2d.y ) );
  directionally_modulated_diffusion <<< dimGrid2d, dimBlock2d >>> ( d_in, n1, n2, s_in1, s_in2,
																	d_mod, n_mod, s_mod1, s_mod2, s_mod3,
																	d_out, s_out1, s_out2,
																	gmod, g0, sigma_long, sigma_lat   );
  cudaError_t err = cudaGetLastError();   
}
