/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "cutil.h"
#include "cu_math.h"

#define SIGN(x) ((x)>=0?1:-1)
#define ABS(x)  ((x)>=0?(x):-(x))

__global__ void quantize_velocities(  float* output, const float* dx, const float* dy,
									                    int width, int height, int max_vel, bool dolog, float scale )
{
		// index Offset
		const int x  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
		const int y  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex

    if ( x >= width || y >= height ) return;

    const int i_in  = y + x * height;
    const int n_vel = ( 2 * max_vel + 1 );
    const int i_out = ( y + x * height ) * n_vel * n_vel;
  
    const float ddx = dolog ? SIGN(dx[i_in]) * log2f( 1.0f + ABS( dx[i_in] ) ) : dx[i_in];
    const float ddy = dolog ? SIGN(dy[i_in]) * log2f( 1.0f + ABS( dy[i_in] ) ) : dy[i_in];
  
    const int i_dx = clamp( max_vel + scale * ddx + 0.5f, 0, 2*max_vel );
    const int i_dy = clamp( max_vel + scale * ddy + 0.5f, 0, 2*max_vel );
  
    output[ i_out + ( i_dy ) +  ( i_dx ) * n_vel ] = 1.0f;
}


__global__ void quantize_velocities_with_confidence(  float* output, const float* dx, const float* dy
													  , const float* dxx, const float* dyy, const float* dxy
													  , int width, int height, int max_vel, bool dolog, float scale  )
{
  // index Offset
  const int x  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int y  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  
  if ( x >= width || y >= height ) return;
  
  const int i_in  = y + x * height;
  const int n_vel = ( 2 * max_vel + 1 );
  const int i_out = ( y + x * height ) * n_vel * n_vel;
  
  const float ddx = dolog ? SIGN(dx[i_in]) * log2f( 1.0f + ABS( dx[i_in] ) ) : dx[i_in];
  const float ddy = dolog ? SIGN(dy[i_in]) * log2f( 1.0f + ABS( dy[i_in] ) ) : dy[i_in];
  
  const int i_dx = clamp( max_vel + scale * ddx + 0.5f, 0, 2*max_vel );
  const int i_dy = clamp( max_vel + scale * ddy + 0.5f, 0, 2*max_vel );
  
  // compute determinant of covariance matrix    
  output[ i_out + i_dy +  i_dx * n_vel ] = gauss_norm( dxx[ i_in ], dyy[ i_in ], dxy[ i_in ] );
}



__global__ void make_sparse_velocities(  float* output, const float* in,
                                        int width, int height, int n_vel )
{
		int fdim = n_vel*n_vel;
    int halfdim = int( float(n_vel) * 0.5f);
		// index Offset
		const int x  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
		const int y  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex

    if ( x >= width || y >= height ) return;

    const int i_out = ( y + x * height ) * 3;
    const int i_in  = ( y + x * height ) * fdim;
  
    float sum_c = 0.0000001f;
    float sum_x = 0.f;
    float sum_y = 0.f;
    
    for ( int i = 0; i < n_vel; ++i )
      {
        for ( int j = 0; j < n_vel; ++j )
          {
            sum_c += in[ i_in + i*n_vel + j ];
            sum_x += (i-halfdim) * in[ i_in + i*n_vel + j ];
            sum_y += (j-halfdim) * in[ i_in + i*n_vel + j ];
          }
      }
    
    sum_x /= sum_c;
    sum_y /= sum_c;
 
    
    output[ i_out + 0 ] = sum_y;
    output[ i_out + 1 ] = sum_x;
    output[ i_out + 2 ] = sum_c;

}



// combines two w x h x 1 arrays into one  w x h x 2  array 
__global__ void concatenate_2D_images( float* output, const float* dx, const float* dy,
									                    int width, int height, bool dolog )
{
  // index Offset
  const int x  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int y  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  
  if ( x >= width || y >= height ) return;
  
  const int i_in  = y + x * height;
  const int i_out = ( y + x * height ) * 2;
  
  const float ddx = dolog ? SIGN(dx[i_in]) * log2f( 1.0f + ABS( dx[i_in] ) ) : dx[i_in];
  const float ddy = dolog ? SIGN(dy[i_in]) * log2f( 1.0f + ABS( dy[i_in] ) ) : dy[i_in];
  
  output[ i_out + 0 ] = ddx;
  output[ i_out + 1 ] = ddy;
    
}

// combines two w x h x 1 arrays and computes determinant dxx, dyy, dxy
__global__ void concatenate_and_compute_sigma(  float* output, const float* dx, const float* dy, 
												const float* dxx, const float* dyy, const float* dxy,
												int width, int height, float confidence_threshold )
{
  // index Offset
  const int x  = blockIdx.x * blockDim.x + threadIdx.x; // globalIndex
  const int y  = blockIdx.y * blockDim.y + threadIdx.y; // globalIndex
  
  if ( x >= width || y >= height ) return;
  
  const int i_in  = y + x * height;
  const int i_out = ( y + x * height ) * 3;
  
  output[ i_out + 0 ] = dx[ i_in ];
  output[ i_out + 1 ] = dy[ i_in ];
  float confidence    = gauss_norm( dxx[ i_in ], dyy[ i_in ], dxy[ i_in ] );
  output[ i_out + 2 ] = confidence - confidence_threshold;
  
  const float& _dxx = dxx[ i_in ];
  const float& _dxy = dxy[ i_in ];
  const float& _dyy = dyy[ i_in ];
  
  //compute largest eigenvalue
  //float lambda0 = 1000.f * factor / ( _dxx + _dyy + sqrt( _dxx*_dxx - 2*_dxx*_dyy + 4*_dxy*_dxy + _dyy*_dyy) );
  //output[ i_out + 2 ] = ( lambda0 >= confidence_threshold ) ? lambda0 : 0.0f; 
  
  
  
  //float lambda0 = ( _dxx + _dyy + sqrt( _dxx*_dxx - 2*_dxx*_dyy + 4*_dxy*_dxy + _dyy*_dyy) );
  //float lambda1 = ( _dxx + _dyy - sqrt( _dxx*_dxx - 2*_dxx*_dyy + 4*_dxy*_dxy + _dyy*_dyy) );
  //output[i_out + 2 ] = 1.f - lambda1 / lambda0;
  
    
}


extern "C" void do_quantize_velocities( float* output, const float* dx, const float* dy,
									                    int width, int height, int max_vel, bool dolog, float scale )
{
    dim3 dimBlock2d = GPUData_dimBlock2d;
	dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
    quantize_velocities<<<dimGrid2d, dimBlock2d>>>( output, dx, dy, width, height, max_vel, dolog, scale );
}

extern "C" void do_quantize_velocities_with_confidence( float* output, const float* dx, const float* dy, const float* dxx, const float* dyy, const float* dxy,  
                                                        int width, int height, int max_vel, bool dolog, float scale )
{
    dim3 dimBlock2d = GPUData_dimBlock2d;
	dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
    quantize_velocities_with_confidence<<<dimGrid2d, dimBlock2d>>>( output, dx, dy, dxx, dyy, dxy, width, height, max_vel, dolog, scale );
}

extern "C" void do_concatenate_2D_images( float* output, const float* dx, const float* dy,
									                    int width, int height, bool dolog )
{
    dim3 dimBlock2d = GPUData_dimBlock2d;
	dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
    concatenate_2D_images<<<dimGrid2d, dimBlock2d>>>( output, dx, dy, width, height, dolog );
}

extern "C" void do_concatenate_and_compute_sigma( float* output, const float* dx, const float* dy, 
												  const float* dxx, const float* dyy, const float* dxy,  
                                                  int width, int height, float confidence_threshold )
{
    dim3 dimBlock2d = GPUData_dimBlock2d;
	dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
    concatenate_and_compute_sigma<<<dimGrid2d, dimBlock2d>>>( output, dx, dy, dxx, dyy, dxy, width, height, confidence_threshold );
}


extern "C" void do_make_sparse_velocities(  float* output, const float* in,
                                            int width, int height, int n_vel )
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(width,dimBlock2d.x),iDivUp(height,dimBlock2d.y));
  make_sparse_velocities<<<dimGrid2d, dimBlock2d>>>( output, in, width, height, n_vel );
}


