/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include "cutil.h"

#define BLOCK_WIDTH 16
#define BLOCK_HEIGHT 8

__global__ void shiftMulAddAll_d( float* toShift, float* toMul, float* addTo,
								  const int shiftXmin, const int shiftXmax, const int shiftYmin, const int shiftYmax,
								  const int NX, const int incX, const int NY, const int incY, const int incSx, const int incSy
							   )
{
		// Block index
		const int bIndexX = blockIdx.x;
		const int bIndexY = blockIdx.y;

		// no of threads
		const int noOfThreadsX = BLOCK_WIDTH;
		const int noOfThreadsY = BLOCK_HEIGHT;

		// Thread index
		const int tIndexX = threadIdx.x;
		const int tIndexY = threadIdx.y;
		const int tIndex  = threadIdx.y*BLOCK_WIDTH+threadIdx.x;

		__shared__ float temp[BLOCK_WIDTH * BLOCK_HEIGHT];

		// index Offset
		const int x  = bIndexX * noOfThreadsX + tIndexX; // destination
		const int y  = bIndexY * noOfThreadsY + tIndexY; // destination

		if (x<NX && y<NY)
		  {

			for ( int dx = shiftXmin; dx <= shiftXmax; ++dx )
			  for ( int dy = shiftYmin; dy <= shiftYmax; ++dy )
				{

				  const int xO = x - dx;
				  const int yO = y - dy;
				  const int o1  = yO * incY + xO;
				  const int o2  = y * incY + x;
				  const int d  = y * incY + x * incX + ( dx - shiftXmin ) * incSx + ( dy - shiftYmin ) * incSy;

				  if ( xO<0 || yO<0 || xO>=NX || yO>=NY )
					  temp[tIndex] = 0.0f;
				  else
					temp[tIndex] = toShift[o1];

				  temp[tIndex] *= toMul[o2];
				  temp[tIndex] += addTo[d];

				  __syncthreads();

				  if (x<NX && y<NY)
					{
					  addTo[d] = temp[tIndex];
					}
				}
		  }
}

__global__ void shiftMulAdd_d( float* toShift, float* toMul, float* addTo,
							   const int shiftX, const int shiftY,
							   const int NX, const int incY, const int NY )
{
		// Block index
		const int bIndexX = blockIdx.x;
		const int bIndexY = blockIdx.y;

		// no of threads
		const int noOfThreadsX = BLOCK_WIDTH;
		const int noOfThreadsY = BLOCK_HEIGHT;

		// Thread index
		const int tIndexX = threadIdx.x;
		const int tIndexY = threadIdx.y;
		const int tIndex  = threadIdx.y*BLOCK_WIDTH+threadIdx.x;

		__shared__ float temp[BLOCK_WIDTH * BLOCK_HEIGHT];

		// index Offset
		const int x  = bIndexX * noOfThreadsX + tIndexX; // destination
		const int y  = bIndexY * noOfThreadsY + tIndexY; // destination
		const int d  = y*incY+x;
		const int xO = x - shiftX;
		const int yO = y - shiftY;
		const int o  = yO*incY+xO;

		if (x<NX && y<NY) {
			if (xO<0 || yO<0 || xO>=NX || yO>=NY) {
				temp[tIndex] = 0.0f;
			} else {
				temp[tIndex] = toShift[o];
			}

			temp[tIndex] *= toMul[d];
			temp[tIndex] += addTo[d];
		}

		__syncthreads();

		if (x<NX && y<NY) {
			addTo[d] = temp[tIndex];
		}
}
__global__ void shiftMulAddSpecial_d(
							   float* toShiftXX,float* toShiftXY, float* toShiftYY,
							   float* toMulXX, float* toMulXY, float* toMulYY, float* addTo,
							   const int shiftX, const int shiftY,
							   const int NX, const int incY, const int NY )
{
		// Block index
		const int bIndexX = blockIdx.x;
		const int bIndexY = blockIdx.y;

		// no of threads
		const int noOfThreadsX = BLOCK_WIDTH;
		const int noOfThreadsY = BLOCK_HEIGHT;

		// Thread index
		const int tIndexX = threadIdx.x;
		const int tIndexY = threadIdx.y;
		const int tIndex  = threadIdx.y*BLOCK_WIDTH+threadIdx.x;

		// index Offset
		const int x  = bIndexX * noOfThreadsX + tIndexX; // destination
		const int y  = bIndexY * noOfThreadsY + tIndexY; // destination
		const int d  = y*incY+x;
		const int xO = x - shiftX;
		const int yO = y - shiftY;
		const int o  = yO*incY+xO;

		float t;

		if (x<NX && y<NY) {
			if (xO<0 || yO<0 || xO>=NX || yO>=NY) {
				t=0.0f;
			} else {
				t = (3.1415927f*(3*toMulYY[d]*toShiftYY[o]
				                  +toMulXX[d]*(toShiftYY[o]+3*toShiftXX[o])
				                  +toMulYY[d]*toShiftXX[o]
				                 +4*toMulXY[d]*toShiftXY[o]
				                 ))/4.0f;
//				(%pi*(3*dyy1*dyy2+dxx1*dyy2+dxx2*dyy1+4*dxy1*dxy2+3*dxx1*dxx2))/4
			}
			t += addTo[d];
		}

		__syncthreads();

		if (x<NX && y<NY) {
			addTo[d] = t;
		}
}

extern "C" void shiftMulAdd(float* toShift, float* toMul, float* addTo, int shiftX, int shiftY, int Nx, int incY, int Ny) {
	dim3 dimBlock(BLOCK_WIDTH,BLOCK_HEIGHT);
    dim3 dimGrid(iDivUp(Nx,dimBlock.x),iDivUp(Ny,dimBlock.y));
    shiftMulAdd_d<<<dimGrid,dimBlock>>>(toShift, toMul, addTo, shiftX, shiftY, Nx, incY,  Ny);
}
extern "C" void shiftMulAddSpecial(float* toShiftXX, float* toShiftXY, float* toShiftYY,
		float* toMulXX, float* toMulXY, float* toMulYY,
										float* addTo, int shiftX, int shiftY, int Nx, int incY, int Ny) {
	dim3 dimBlock(BLOCK_WIDTH,BLOCK_HEIGHT);
    dim3 dimGrid(iDivUp(Nx,dimBlock.x),iDivUp(Ny,dimBlock.y));
    shiftMulAddSpecial_d<<<dimGrid,dimBlock>>>(toShiftXX, toShiftXY, toShiftYY, toMulXX, toMulXY, toMulYY, addTo, shiftX, shiftY, Nx, incY,  Ny);
}


