/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include <stdio.h>
#ifndef _WIN32
  #include <device_functions.h>
#else
  #include <cuda_runtime.h>
#endif

#define NO_OF_LINES_SS 6
#define  NO_OF_LINES_LIST 1
#define DIV_UP_MACRO(a,b) (a+b-1)/b


#define M_PI 3.14159265359
//#define USE_NEW_SUBSAMPLING 1

inline int iDivUp(int a, int b){
  return (a % b != 0) ? (a / b + 1) : (a / b);
}

__device__ inline void ddswapValues(float& a, float& b) {
	float tmp = a;
    a = b;
    b = tmp;
}

__global__ void d_sortHypos_simple(float* hypoImage, const unsigned int NPix, const unsigned int pixOffs, const unsigned int hypoOffset) {

	const unsigned int h_max = blockDim.x;

	const unsigned int iPixel= blockIdx.x+blockIdx.y*gridDim.x;
	const unsigned int iPixelMemOffset= iPixel*pixOffs;

	const unsigned int iHypo = threadIdx.x;

	extern __shared__ float cach[];

	float* values  = (float*) &(cach[0]);
	float* values1 = (float*) &(cach[h_max]);
	float* values2 = (float*) &(cach[2*h_max]);

	if (iPixel<NPix) {

		values [iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset];
		values1[iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset+1];
		values2[iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset+2];

		__syncthreads();

		for (unsigned int k = 2; k <= h_max; k *= 2) {
			for (unsigned int j = k / 2; j>0; j /= 2) {
				unsigned int ixj = iHypo ^ j;
				if (ixj > iHypo) {
					if ((iHypo & k) == 0) {
						if (values[iHypo] < values[ixj]) {
							ddswapValues(values[iHypo], values[ixj]);
							ddswapValues(values1[iHypo], values1[ixj]);
							ddswapValues(values2[iHypo], values2[ixj]);
						}
					} else {
						if (values[iHypo] > values[ixj]) {
							ddswapValues(values[iHypo], values[ixj]);
							ddswapValues(values1[iHypo], values1[ixj]);
							ddswapValues(values2[iHypo], values2[ixj]);
						}
					}
				}
				__syncthreads();
			}
		}

		hypoImage[iPixelMemOffset+iHypo*hypoOffset] = values[iHypo];
		hypoImage[iPixelMemOffset+iHypo*hypoOffset+1] = values1[iHypo];
		hypoImage[iPixelMemOffset+iHypo*hypoOffset+2] = values2[iHypo];
	}
}
__global__ void d_normHypos_simple(float* hypoImage, const unsigned int NPix, const unsigned int pixOffs, float epsilon) {

	const unsigned int h_max = blockDim.x;

	const unsigned int iPixel= blockIdx.x+blockIdx.y*gridDim.x;
	const unsigned int iPixelMemOffset= iPixel*pixOffs;

	const unsigned int iHypo = threadIdx.x;

	extern __shared__ float cach[];

	float* values  = (float*) &(cach[0]);

	if (iPixel<NPix) {
		values[iHypo] = hypoImage[iPixelMemOffset+iHypo*3];

		__syncthreads();

		for (unsigned int k = 1; k <= h_max; k *= 2) {
			if ( (iHypo & (k*2-1)) == 0 && iHypo+k < h_max ) {
				float temp = values[iHypo];
				values[iHypo] = temp + values[iHypo+k]; //values[iHypo] = temp + values[iHypo+k*3]; //WHAT IS CORRECT??
			}
			__syncthreads();
		}
		float sum = values[0] > epsilon ? values[0] : epsilon;
		hypoImage[iPixelMemOffset+iHypo*3] = hypoImage[iPixelMemOffset+iHypo*3] / sum;
	}
}
__global__ void d_getHypSum_simple(float* hypoImage, float* res, const unsigned int NPix, const unsigned int pixOffs) {

	const unsigned int h_max = blockDim.x;

	const unsigned int iPixel= blockIdx.x+blockIdx.y*gridDim.x;
	const unsigned int iPixelMemOffset= iPixel*pixOffs;

	const unsigned int iHypo = threadIdx.x;

	extern __shared__ float cach[];

	float* values  = (float*) &(cach[0]);

	if (iPixel<NPix) {
		values[iHypo] = hypoImage[iPixelMemOffset+iHypo*3];

		__syncthreads();

		for (unsigned int k = 1; k <= h_max; k *= 2) {
			if ( (iHypo & (k*2-1)) == 0 && iHypo+k < h_max ) {
				float temp = values[iHypo];
				values[iHypo] = temp + values[iHypo+k];
			}
			__syncthreads();
		}
		if (iHypo==0) {
			res[iPixel] = values[0];
		}
	}
}
/*
__global__ void d_expectValue_simple(float* hypoImage, float* xPart, float* yPart, const unsigned int NPix, const float moduloValue, const unsigned int pixOffs, const unsigned int hypoOffset ) {

	const float faktor = 10.0f;

	const unsigned int h_max = blockDim.x;

	const unsigned int iPixel= blockIdx.x+blockIdx.y*gridDim.x;
	const unsigned int iPixelMemOffset= iPixel*pixOffs;

	const unsigned int iHypo = threadIdx.x;

	extern __shared__ float cach[];

	float* values   = (float*) &(cach[0]);
	float* values2  = (float*) &(cach[h_max]);
	float* values3  = (float*) &(cach[2*h_max]);

	if (iPixel<NPix) {

		values[iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset] * faktor;
		values2[iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset+1];
		values3[iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset+2];

		float temp = values2[iHypo];

		values3[iHypo] = floorf(temp/moduloValue) * values[iHypo];
		values2[iHypo] = fmodf(temp,moduloValue) * values[iHypo];

//		xPart[iPixel] = values2[3];
//		yPart[iPixel] = values3[3];

		__syncthreads();

		for (unsigned int k = 1; k <= h_max; k *= 2) {
			if ( ((iHypo & (k*2-1)) == 0) && (iHypo+k < h_max) ) {
				values[iHypo]  = values[iHypo] + values[iHypo+k];
				values2[iHypo] = values2[iHypo] + values2[iHypo+k];
				values3[iHypo] = values3[iHypo] + values3[iHypo+k];
			}
			__syncthreads();
		}
		if (iHypo==0) {
			float umme = values[iHypo];
			if (umme<0.01f) {
				xPart[iPixel] = (moduloValue-1)/2.0f;
				yPart[iPixel] = (moduloValue-1)/2.0f;;
			} else {
				umme = umme + 0.001f;
				float ummex = values2[iHypo] / umme;
				xPart[iPixel] = ummex;
				float ummey = values3[iHypo] / umme;
				yPart[iPixel] = ummey;
			}
		}
	}
}*/
/*
__global__ void d_expectValuNearlyeMax_simple(float* hypoImage, float* xPart, float* yPart, const unsigned int NPix, const float moduloValue, const unsigned int pixOffs, const unsigned int hypoOffset ) {

	const float faktor = 10.0f;

	const unsigned int h_max = blockDim.x;

	const unsigned int iPixel= blockIdx.x+blockIdx.y*gridDim.x;
	const unsigned int iPixelMemOffset= iPixel*pixOffs;

	const unsigned int iHypo = threadIdx.x;

	extern __shared__ float cach[];

	float* values   = (float*) &(cach[0]);
	float* values2  = (float*) &(cach[h_max]);
	float* values3  = (float*) &(cach[2*h_max]);

	if (iPixel<NPix) {

		values[iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset] * faktor;
		values2[iHypo] = hypoImage[iPixelMemOffset+iHypo*hypoOffset+1];
		values3[iHypo] = 0.0f;

		if (iHypo==0) values[iHypo] = values[iHypo] * 20000.0f;
		if (iHypo==1) values[iHypo] = values[iHypo] * 10000.0f;
		if (iHypo==2) values[iHypo] = values[iHypo] *  5000.0f;
		if (iHypo==3) values[iHypo] = values[iHypo] *  2500.0f;

		float temp = values2[iHypo];

		values3[iHypo] = floorf(temp/moduloValue) * values[iHypo];
		values2[iHypo] = fmodf(temp,moduloValue) * values[iHypo];

//		xPart[iPixel] = values2[3];
//		yPart[iPixel] = values3[3];


		__syncthreads();

		for (unsigned int k = 1; k <= h_max; k *= 2) {
			if ( ((iHypo & (k*2-1)) == 0) && (iHypo+k < h_max) ) {
				values[iHypo]  = values[iHypo] + values[iHypo+k];
				values2[iHypo] = values2[iHypo] + values2[iHypo+k];
				values3[iHypo] = values3[iHypo] + values3[iHypo+k];
			}
			__syncthreads();
		}
		if (iHypo==0) {
			float umme = values[iHypo];
			if (umme<0.01f) {
				xPart[iPixel] = (moduloValue-1)/2.0f;
				yPart[iPixel] = (moduloValue-1)/2.0f;;
			} else {
				umme = umme + 0.001f;
				float ummex = values2[iHypo] / umme;
				xPart[iPixel] = ummex;
				float ummey = values3[iHypo] / umme;
				yPart[iPixel] = ummey;
			}
		}
	}
}*/

extern "C" {

void do_sortHypos_simple(float* hypoImage, const unsigned int NPix, const unsigned int h_max, const unsigned int pixOffs, const unsigned int hypoOffset) {
	dim3 dimBlock((int)h_max);
	int bx  = (int)NPix;
	int by = 1;
	while (bx>=65535) {
		by *= 2;
		bx  = iDivUp(bx,2);
	}
	dim3 dimGrid(bx,by);

	d_sortHypos_simple<<<dimGrid, dimBlock, sizeof(float)*3*((int)h_max)>>>(hypoImage, NPix, pixOffs, hypoOffset);

	cudaError_t err = cudaThreadSynchronize();                               \
	if( cudaSuccess != err) {                                                \
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_sortHypos_simple() execution failed\n", __FILE__, __LINE__, cudaGetErrorString( err) );\
		exit(EXIT_FAILURE);                                                  \
	}
}

  void do_normHypos_simple(float* hypoImage, const unsigned int NPix, const unsigned int h_max, const unsigned int pixOffs, float epsilon) {
	dim3 dimBlock((int)h_max);
	int bx  = (int)NPix;
	int by = 1;
	while (bx>=65535) {
		by *= 2;
		bx  = iDivUp(bx,2);
	}
	dim3 dimGrid(bx,by);

	d_normHypos_simple<<<dimGrid, dimBlock, (sizeof(float))* h_max>>>(hypoImage, NPix, pixOffs, epsilon);

	cudaError_t err = cudaThreadSynchronize();                               \
	if( cudaSuccess != err) {                                                \
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_normHypos_simple() execution failed\n", __FILE__, __LINE__, cudaGetErrorString( err) );\
		exit(EXIT_FAILURE);                                                  \
	}

}

void do_sumHypos_simple(float* hypoImage, float* res, const unsigned int NPix, const unsigned int h_max, const unsigned int pixOffs) {
	dim3 dimBlock((int)h_max);
	int bx  = (int)NPix;
	int by = 1;
	while (bx>=65535) {
		by *= 2;
		bx  = iDivUp(bx,2);
	}
	dim3 dimGrid(bx,by);

	d_getHypSum_simple<<<dimGrid, dimBlock, (sizeof(float))* h_max>>>(hypoImage,res, NPix, pixOffs);

	cudaError_t err = cudaThreadSynchronize();                               \
	if( cudaSuccess != err) {                                                \
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_getHypSum_simple() execution failed\n", __FILE__, __LINE__, cudaGetErrorString( err) );\
		exit(EXIT_FAILURE);                                                  \
	}
}
/*
void do_expectedValue_simple(float* hypoImage, float* xPart, float* yPart, const unsigned int NPix, const int moduloValue, const unsigned int h_max, const unsigned int pixOffs, const unsigned int attribOffs) {
	dim3 dimBlock((int)h_max);
	int bx  = (int)NPix;
	int by = 1;
	while (bx>=65535) {
		by *= 2;
		bx  = iDivUp(bx,2);
	}
	dim3 dimGrid(bx,by);

//	d_expectValuNearlyeMax_simple<<<dimGrid, dimBlock, (sizeof(float)*3)* h_max>>>(hypoImage, xPart, yPart, NPix, moduloValue, pixOffs, attribOffs);
	d_expectValue_simple<<<dimGrid, dimBlock, (sizeof(float)*3)* h_max>>>(hypoImage, xPart, yPart, NPix, moduloValue, pixOffs, attribOffs);

	cudaError_t err = cudaThreadSynchronize();                               \
	if( cudaSuccess != err) {                                                \
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_expectValue_simple() execution failed\n", __FILE__, __LINE__, cudaGetErrorString( err) );\
		exit(EXIT_FAILURE);                                                  \
	}
}*/


}
