/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "subsample.hcu"
#include <cuda.h>
#include <stdio.h>
#include <assert.h>
//#include <device_functions.h>
#include <cuda_runtime.h>

#ifndef M_PI
	#define M_PI 3.14159265358979323846
#endif

//
// choose which implementation to choose, enable some additional stuff
// 

// this is a single-threaded (and thus not the fastest) implementation of the
// subsampling routine, where each MT-point has a block of dimension 1 running.
//#define SUBSAMPLE_SINGLETHREAD	

// this is the global activator for the multithreaded implementations. for
// details, look below in the implementation-switch-documentation
#define SUBSAMPLE_MULTITHREAD

// Choose which implementation to use
#ifdef SUBSAMPLE_MULTITHREAD
	// The direct-insert subsampling version is a multithreaded implementation of 
	// the singlethreaded version, where each item is written to its according
	// position in the velocity space matrix.
	// if no atomics are available and this version is used, a simple add will
	// be used (with possible problems)
#   define SUBSAMPLE_DIRECT
//#   define SUBSAMPLE_BONN

		// this switch indicates if atomics shall be used. those will only be
		// activated, if the platform supports it. if the switch is not enabled,
		// or if atomics are not provided on the target platform, the function 
		// will fallback to a simple add
#       define USE_ATOMICS
 
		// this enables an experimental add-heuristic. it will only be used when
		// no atomics are used! the speed comes near (and sometimes even
		// outpassed) the speed of the correct atomical version. the result,
		// though, is not guaranteed to be correct. as tests indicate, this only
		// really works with a thread count <= 32. this may be because of
		// different block-parts running on different processor cores. anyway,
		// the speed is at its best with 32 threads.
#       define USE_ADD_HEURISTIC

	// the broadcast method tries to use the shared memory broadcast to pipe all
	// V1 values to corresponding threads. there are vel_max*2+1 threads running
	// for fast sorting and circumventing the exclusive write problems on G8x
	// cards. Additionally, makes use of coalesced reads (hint: did not check
	// correct constraints, e.g. base-addr of thread 0 divisible by 16)
	//#define SUBSAMPLE_BROADCAST

	// the Bonn version of subsampling that only works properly when atomic
	// operations are available.
	//#define SUBSAMPLE_BONN
	
	// the 'really old' version of a multithreaded subsampling implementation,
	// this is effectively what SUBSAMPLE_DIRECT is, though there are quite a
	// few problems that were regarded in the new DIRECT version
	//#define SUBSAMPLE_OLD


	// -----------------------------------------------------------------------
	// additional options

	// thread count to use. make some benchmarks on the machine the
	// algorithm should run on to set this value. This value is not used in
	// every implementation
	#define SUBSAMPLE_THREAD_COUNT 32
		// ADDITIONAL NOTES for SUBSAMPLE_THREAD_COUNT / VALUES TO BE USED:
		// - 32 for BROADCAST
		// - H_MAX for DIRECT without actually having atomics and not using the
		//     add heuristic will ensure that no data gets lost during
		//     conflicting writes, because h_max hypotheses will be read at
		//     once, starting with the first on each row. as h_max hypotheses
		//     are unique in respect to their (dx,dy) setup, no write will
		//     conflict. if more items are processed concurrently without real
		//     atomics, a lot of garbage will be calculated (compare, for
		//     example, the results with 4 and 32 threads)


	// if enabled, a min(1.0, weight) checking will be done when writing back
	// the h_max hypotheses to global memory
	//#define SUBSAMPLE_WRITEBACK_MIN
#endif


// NOTE: you may encounter lines starting with //{{{ and //}}}. please do not
// remove them, as they are auto-fold markers for vim.

#define DIVUP(a,b)	(((a) + (b) - 1) / (b)) 
#define SCALING		60000.0f


// TODO: possibly move out these functions to another util-file, or to the
// header

__device__ float 
gauss_2d (const float sigma, const int rx, const int ry) 
{
	const float r2 = float(rx*rx+ry*ry);
	return 1.0f/(2.0f*M_PI*sigma*sigma)*exp(-0.5f*r2/(sigma*sigma));
}

template <typename T> __device__ T
clamp (T v, T lo, T hi)
{
	return v < lo ? lo : ( v > hi ? hi : v);
}


template <typename T> __device__ void
reset_cache (T *cache, size_t cache_size, size_t nThreads, T value)
{
	for (size_t i = 0; i < DIVUP(cache_size, nThreads); i++) {
		const size_t n = i * nThreads + threadIdx.x;
		if (n < cache_size)
			cache[n] = value;
	}
	__syncthreads();
}


template <typename T> __device__ void
search_biggest (T *vspace, unsigned int *ucache, int h_max, 
		int vspace_size, size_t nThreads)
{
	// initialize ucache
	for (int i = 0; i < h_max; i++) {
		ucache[threadIdx.x*h_max + i] = 0;	
	}
	__syncthreads();

	// search in parallel
	float min_value = 0.0f;
	int min_ptr = 0;
	for (int i = 0; i < DIVUP(vspace_size, nThreads); i++) {
		const int n = i * nThreads + threadIdx.x;
		if (n < vspace_size) {
			if (vspace[n] >= min_value) {
				ucache[min_ptr + threadIdx.x*h_max] = n;

				min_ptr = 0;
				for (int j = 1; j < h_max; j++) 
					if (vspace[ucache[j + threadIdx.x*h_max]] <= vspace[ucache[min_ptr + threadIdx.x*h_max]])
						min_ptr = j;
				min_value = vspace[ucache[min_ptr + threadIdx.x*h_max]];
			}
		}
	}
	__syncthreads();

	// search overall h-max biggest
	if (threadIdx.x == 0) {
		for (int i = h_max; i < nThreads * h_max; i++) {
			if (vspace[ucache[i]] >= min_value) {
				ucache[min_ptr] = ucache[i];

				min_ptr = 0;
				for (int j = 1; j < h_max; j++) 
					if (vspace[ucache[j]] <= vspace[ucache[min_ptr]])
						min_ptr = j;
				min_value = vspace[ucache[min_ptr]];
			}
		}
	}
	__syncthreads();
}


template <typename T> __device__ void
write_back (float *output, T *vspace, unsigned int *ucache, int h_max,
		int vspace_width, int vel_max, size_t nThreads, int xOutputPos, int yOutputPos,
		int output_width);

template <> __device__ void
write_back<float> (float *output, float *vspace, unsigned int *ucache, int h_max,
		int vspace_width, int vel_max, size_t nThreads, int xOutputPos, int yOutputPos,
		int output_width)
{
	for (int i = 0; i < DIVUP(h_max, nThreads); i++) {
		const int h = i * nThreads + threadIdx.x;
		
		if (h < h_max) {
			const size_t memoffset = (xOutputPos + yOutputPos * output_width) * h_max * 3;
			const size_t idx = ucache[h];
#ifdef SUBSAMPLE_WRITEBACK_MIN
			output[memoffset + h*3 + 0] = min(1.0, vspace[idx]);
#else
			output[memoffset + h*3 + 0] = vspace[idx];
#endif
			output[memoffset + h*3 + 1] = int(idx % vspace_width - vel_max);
			output[memoffset + h*3 + 2] = int(idx / vspace_width - vel_max);
		}
	}
}

template <> __device__ void
write_back<int> (float *output, int *vspace, unsigned int *ucache, int h_max,
		int vspace_width, int vel_max, size_t nThreads, int xOutputPos, int yOutputPos,
		int output_width)
{
	for (int i = 0; i < DIVUP(h_max, nThreads); i++) {
		const int h = i * nThreads + threadIdx.x;

		if (h < h_max) {
			const size_t memoffset = (xOutputPos + yOutputPos * output_width) * h_max * 3;
			const size_t idx = ucache[h];

#ifdef SUBSAMPLE_WRITEBACK_MIN
			output[memoffset + h*3 + 0] = min(1.0, float(vspace[idx] / SCALING));
#else
			output[memoffset + h*3 + 0] = float(vspace[idx] / SCALING);
#endif
			output[memoffset + h*3 + 1] = int(idx % vspace_width - vel_max);
			output[memoffset + h*3 + 2] = int(idx / vspace_width - vel_max);
		}
	}
}


//{{{ SUBSAMPLE_SINGLETHREAD
#ifdef SUBSAMPLE_SINGLETHREAD

__global__ void 
d_subsample_hypos (
		int h_max, int vel_max, float sigma, 
		int v1_window_width,
		float *input, int input_width, int input_height,
		float *output, int output_width, int output_height
		)
{
	// local variable setup
	const int xratio = input_width / output_width;
	const int yratio = input_height / output_height;
	const int xOutputPos = blockIdx.x;
	const int yOutputPos = blockIdx.y;
 
	const int width_vel_space = vel_max * 2 + 1;
	const int size_vel_space  = width_vel_space * width_vel_space;

	extern __shared__ float cache[];
	float *fcache = (float*)&cache[0];
	unsigned int *ucache = (unsigned int*)&cache[size_vel_space + 1];

	reset_cache<float>(cache, (size_vel_space + 1 + h_max), 1, 0.0);

	// load values
	const int shift = (v1_window_width / 2);
	for (int x = 0; x < v1_window_width; x++) {
		const int xInputPos = xOutputPos * xratio + x - shift;

		for (int y = 0; y < v1_window_width; y++) {
			const int yInputPos = yOutputPos * yratio + y - shift;

			if (xInputPos >= 0 && xInputPos < input_width && yInputPos >= 0 && yInputPos < input_height) {
				const size_t inputMemOffset = (xInputPos + yInputPos * input_width) * h_max * 3;	
				
				for (int h = 0; h < h_max; h++) {
					const float w_f = 
						input[inputMemOffset + h*3 + 0] * 
						gauss_2d(sigma, 
								x - shift, 
								y - shift);

					const int dx =
						clamp<int>(input[inputMemOffset + h*3 + 1] + vel_max, 
								0, width_vel_space - 1);

					const int dy =
						clamp<int>(input[inputMemOffset + h*3 + 2] + vel_max, 
								0, width_vel_space - 1);
				
					fcache[dx + dy * width_vel_space] += w_f;
				}
			}
		}
	}

	search_biggest<float>(fcache, ucache, h_max, size_vel_space, 1);

	write_back<float>(output, fcache, ucache, h_max, width_vel_space, vel_max, 1,
			xOutputPos, yOutputPos, output_width);
}


extern "C"
void do_subsample_hypo_image (
		int h_max, int vel_max, float sigma,
		int v1_window_width,
		float* input,  int xRes_input,  int yRes_input,
		float* output, int xRes_output, int yRes_output)
{
	dim3 block(1);
	dim3 grid(xRes_output, yRes_output);
	size_t cache = (h_max + (2 * vel_max + 1) * (2 * vel_max + 1) + 1) * sizeof(float);

	d_subsample_hypos<<<grid, block, cache>>>(
			h_max, vel_max, sigma, v1_window_width,
			input, xRes_input, yRes_input,
			output, xRes_output, yRes_output);

	// TODO
	// do something, but try to avoid stdio.h in .cu files! 
	cudaError_t e = cudaThreadSynchronize(); \
	if (e != cudaSuccess) { \
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_subsample_hypo_image() execution failed\n", __FILE__, __LINE__, cudaGetErrorString(e) ); \
		exit(EXIT_FAILURE); \
	}
}

#endif
//}}} SUBSAMPLE_SINGLETHREAD


//{{{ SUBSAMPLE_MULTITHREAD
#ifdef SUBSAMPLE_MULTITHREAD

//{{{ SUBSAMPLE_BROADCAST
#ifdef SUBSAMPLE_BROADCAST

__global__ void
d_subsample (int h_max, int vel_max, float sigma, int v1_window_width,
		float *input, int input_width, int input_height,
		float *output, int output_width, int output_height)
{
	// variable setup
	const int xratio = input_width / output_width;
	const int yratio = input_height / output_height;
	
	const int xOutputPos = blockIdx.x;
	const int yOutputPos = blockIdx.y;

	const int nThreads = blockDim.x;

	const int vspace_width  = (vel_max * 2)  + 1;
	const int vspace_size   = vspace_width * vspace_width;
	const int cache_size    = vspace_size + 1 + (v1_window_width * h_max * 3);
	const int hypos_per_row = v1_window_width * h_max;
	const int memblock_size = hypos_per_row * 3;
	
	// cache setup
	extern __shared__ float cache[];
	float *vspace, *memblock;
	unsigned int *ucache;
	vspace   = cache;
	memblock = &cache[vspace_size + 1];
	ucache   = (unsigned int *)memblock;

	reset_cache<float>(cache, cache_size, nThreads, 0.0);

	// iterate over all input lines
	const int shift = (v1_window_width / 2);
	const int max_input_per_row = input_width * h_max * 3;

	// load and process all rows
	const int x_input_base_pos = xOutputPos * xratio - shift;
	for (int y = 0; y < v1_window_width; y++) {
		const int y_input_pos = yOutputPos * yratio + y - shift;
		if (y_input_pos >= 0 && y_input_pos < input_height) {
			const int base_offset = (y_input_pos * input_width + x_input_base_pos) * h_max * 3;

			// coalesced read memory
			for (int i = 0; i < DIVUP(memblock_size, nThreads); i++) {
				const int n = i * nThreads + threadIdx.x;
				if (n < memblock_size) {
					const int row_offset = x_input_base_pos * h_max * 3 + n;
					const int offset     = base_offset + n;

					if (offset >= 0 && row_offset >= 0 && row_offset < max_input_per_row) 
						memblock[n] = input[offset];
					else
						memblock[n] = 0.0;
				}
			}
			__syncthreads();

			// postprocess step 1: gauss & clamp
			for (int i = 0; i < DIVUP(hypos_per_row, nThreads); i++) {
				const int h = i * nThreads + threadIdx.x;
				const int x = h / h_max;
				const int offset = h * 3;
				
				memblock[offset + 0] *= gauss_2d(sigma, x-shift, y-shift);
				memblock[offset + 1] = clamp<int>(int(memblock[offset + 1]) + vel_max, 0, vspace_width - 1);
				memblock[offset + 2] = clamp<int>(int(memblock[offset + 2]) + vel_max, 0, vspace_width - 1);
			}
			__syncthreads();


			// postprocess step 2: broadcast items to threads, store them into
			// velocity space. THIS IS THE ABSOLUTE BOTTLENECK!
			for (int i = 0; i < hypos_per_row; i++) {
				const int offset = i * 3;
				const float w        = memblock[offset];
				const int  dx        = memblock[offset + 1];
				const int  dy_offset = memblock[offset + 2] * vspace_width;

				if (((dx + 1) % (threadIdx.x + 1)) == 0) {
					vspace[dx + dy_offset] += w;
				}
			}
		}
		__syncthreads();
	}

	search_biggest<float>(vspace, ucache, h_max, vspace_size, nThreads);

	write_back<float>(output, vspace, ucache, h_max, vspace_width, vel_max, nThreads,
			xOutputPos, yOutputPos, output_width);
}


extern "C" void 
do_subsample_hypo_image (int h_max, int vel_max, float sigma, int v1_window_width,
		float *input, int xRes_input, int yRes_input,
		float *output, int xRes_output, int yRes_output)
{
	// warning: the kernel won't work (due to shared memory restrictions) if the
	// values are bigger than 30/4.
	assert(vel_max <= 30);
	assert(h_max <= 4);

	dim3 grid(xRes_output, yRes_output);
	dim3 block(SUBSAMPLE_THREAD_COUNT,1,1);		
	size_t cache_size = sizeof(float) * ( 
			(vel_max * 2 + 1) * (vel_max * 2 + 1) +
			(v1_window_width * h_max * 3) + 1);

	d_subsample<<<grid,block,cache_size>>>(h_max, vel_max, sigma, v1_window_width,
			input, xRes_input, yRes_input, 
			output, xRes_output, yRes_output);

	cudaError_t e = cudaThreadSynchronize();
	if (e != cudaSuccess) {
		// TODO: move the fprintf out of the cuda file
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_subsample() execution failed\n", __FILE__, __LINE__, cudaGetErrorString(e) ); \
		exit(EXIT_FAILURE);
	}
}

#endif
//}}} SUBSAMPLE_BROADCAST


//{{{ SUBSAMPLE_BONN
#ifdef SUBSAMPLE_BONN


struct short_indexed_weight
{
  unsigned short weight;
  unsigned short index;
  bool operator<( const short_indexed_weight & other ) const
  {
	return weight < other.weight;
  }
};

template<typename T>
struct max_value
{
  __device__ T operator()( const T & v1, const T & v2 ) const
  {
	return v1 < v2 ? v2 : v1;
  }
};

#include "block_reduce.h"

__device__ int clampi( int v, int lo, int hi )
{
  return v < lo ? lo : ( v > hi ? hi : v );
}


__global__ void d_subsample_hypo_image2 
(
 int h_max, int vel_max, float sigma, int v1_window_width,
 float* hypoImageInput, int xRes_input , int yRes_input,
 float* output        , int xRes_output, int yRes_output
 )
{
	const unsigned int width_vel_space = vel_max*2+1;
	const unsigned int size_vel_space  = width_vel_space*width_vel_space;

	const size_t id = threadIdx.x;
	const size_t N  = blockDim.x;

	extern __shared__ short_indexed_weight cachI[];

	//initialize velspace
	for ( unsigned int i = id; i < size_vel_space; i += N )
	  {
		cachI[ i ].weight = 0;
		cachI[ i ].index  = i;
	  }

	//__syncthreads();

	const unsigned int x_output_pos = blockIdx.x;
	const unsigned int y_output_pos = blockIdx.y;
	const unsigned int x_ratio = xRes_input/xRes_output;
	const unsigned int y_ratio = yRes_input/yRes_output;
	const int x_input_pos = int( x_output_pos * x_ratio ) - v1_window_width / 2;
	const int y_input_pos = int( y_output_pos * y_ratio ) - v1_window_width / 2;
	
	const float scaling = 60000.0f;

	// integrate input into full velocity space
	size_t d0 = v1_window_width / 2;
	for ( size_t i = id; i < ( v1_window_width * v1_window_width ); i += N )
	  {
		size_t dx = i % v1_window_width;
		size_t dy = i / v1_window_width;
		int x = x_input_pos + dx;
		int y = y_input_pos + dy;
		if ( x >= 0 && x < xRes_input && y >= 0 && y < yRes_input )
		  {
			size_t offset = ( x + y * xRes_input ) * 3 * h_max;
			for ( size_t h = 0; h < h_max; ++h )
			  {
				float w0 = gauss_2d( sigma, dx - d0, dy - d0 );
				float w =  w0 * hypoImageInput[ offset + 3 * h + 0 ];
				
				int vx = int( hypoImageInput[ offset + 3 * h + 1 ] + 0.5f );
				int vy = int( hypoImageInput[ offset + 3 * h + 2 ] + 0.5f );
				vx = clampi( vx + vel_max, 0, width_vel_space - 1 );
				vy = clampi( vy + vel_max, 0, width_vel_space - 1 );
				vx = 2;
				vy = 0;
				size_t cache_offset = vx + vy * width_vel_space;
#if !defined(CUDA_NO_SM_11_ATOMIC_INTRINSICS)
				atomicAdd( ( int * )  ( cachI + cache_offset ), int( w * scaling ) );
#else
				cachI[ cache_offset ].weight += int( w * scaling );
#endif
			  }
		  }
	  }
	
	// find h_max best hypothesises
	//__syncthreads();

	size_t nvels = size_vel_space;
	size_t n = ( nvels + N - 1 ) / N;
	
	while ( n > h_max )
	  {

		
		for ( size_t i = 0; i < n; ++i )
		  {
			block_reduce( id, 1 + ( nvels - i - 1 ) / n, cachI + i, 
						  n, max_value< short_indexed_weight >(), 1, false );
		  }

		nvels = n;
		n = ( nvels + N - 1 ) / N;

	  }

	//__syncthreads();

	if ( id < nvels )
	  block_reduce( id, nvels, cachI, 1, max_value< short_indexed_weight >(), h_max, false );

	//__syncthreads();

	// write to output array

	if ( id < h_max )
	  { 
		size_t out_off = 3 * ( id + h_max * ( y_output_pos * xRes_output + x_output_pos ) );
		output[ out_off + 0 ] = min( 1.0f, float( cachI[ id ].weight ) / scaling );
		output[ out_off + 1 ] = int( cachI[ id ].index % width_vel_space ) - vel_max;
		output[ out_off + 2 ] = int( cachI[ id ].index / width_vel_space ) - vel_max;
	  }
}


extern "C" 
void do_subsample_hypo_image(
		int h_max, int vel_max, float sigma,
		int v1_window_width,
		float* input, int xRes_input , int yRes_input,
		float* output        , int xRes_output, int yRes_output)
{
	dim3 dimBlock( 32, 1, 1 );
	int bx = xRes_output;
	int by = yRes_output;
	dim3 dimGrid(bx,by);

	size_t n_shared = sizeof(short_indexed_weight) * ( ( 2 * vel_max + 1 ) * ( 2 * vel_max + 1 ) );
						
	d_subsample_hypo_image2<<<dimGrid, dimBlock, n_shared >>>
	  (
	   h_max, vel_max, sigma,v1_window_width,
	   input, xRes_input , yRes_input,
	   output        , xRes_output, yRes_output
	   );


	cudaError_t err = cudaThreadSynchronize();                               \
	if( cudaSuccess != err) {                                                \
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_subsample_hypo_image() execution failed\n", __FILE__, __LINE__, cudaGetErrorString( err) );\
		exit(EXIT_FAILURE);                                                  \
	}
}
#endif
//}}} SUBSAMPLE_BONN


//{{{ SUBSAMPLE_OLD
#ifdef SUBSAMPLE_OLD

#define NO_OF_LINES_SS 6
#define  NO_OF_LINES_LIST 1
#define DIV_UP_MACRO(a,b) (a+b-1)/b

__global__ void d_subsample_hypo_image (
		int h_max, int vel_max, float sigma,
		float* hypoImageInput, int xRes_input , int yRes_input,
		float* output        , int xRes_output, int yRes_output)
{
	const unsigned int x_ratio = xRes_input/xRes_output;
	const unsigned int y_ratio = yRes_input/yRes_output;
	const unsigned int v1_window_width = blockDim.x;
	const unsigned int noOfThreads     = v1_window_width;
	const unsigned int noOfThreadsX   = v1_window_width; //Changed
	const unsigned int noOfThreadsY    = NO_OF_LINES_SS; //Changed
	const unsigned int noOfThreadsTotal    = v1_window_width * NO_OF_LINES_SS; //Changed

	const unsigned int width_vel_space = vel_max*2+1;
	const unsigned int size_vel_space  = width_vel_space*width_vel_space;

	const unsigned int threadPos = threadIdx.x;
	const unsigned int threadPosX = threadIdx.x; //Changed
	const unsigned int threadPosY= threadIdx.y; //Changed
	unsigned int threadPosTotal= threadIdx.x + threadIdx.y*v1_window_width; //Changed

	extern __shared__ unsigned int cachI[];

	const unsigned int thread_loops = (size_vel_space+noOfThreadsX-1)/noOfThreads;
	const unsigned int thread_loops_total = (size_vel_space+noOfThreadsTotal-1)/noOfThreadsTotal;



	//initialize velspace with 0
	for (unsigned int i=0; i<thread_loops_total; ++i) {
		if (threadPosTotal+i*noOfThreadsTotal<size_vel_space) cachI[threadPosTotal+i*noOfThreadsTotal] = 0;
	}
	__syncthreads();

	//printf(" TEST SUBSAMPLING");
	const unsigned int x_output_pos = blockIdx.x;//TODO:CHECK!
	const unsigned int y_output_pos = blockIdx.y;

	const int x_input_pos = (unsigned int)x_output_pos*x_ratio-(unsigned int)(v1_window_width/2)+threadPosX;

	int count_loops = 0;

	//Get all V1 hypos  contributing to one MT position, saved in velocity space
	for (unsigned int i=0; i<h_max; ++i)
	for (unsigned int y=0; y<DIV_UP_MACRO(v1_window_width,noOfThreadsY); ++y) {
		const int y_input_pos = y_output_pos*y_ratio-v1_window_width/2+threadPosY+y*noOfThreadsY;
		if(x_input_pos >= 0 && y_input_pos >= 0 && x_input_pos < xRes_input && y_input_pos < yRes_input)
		{
			const float w_f =    hypoImageInput[((x_input_pos+y_input_pos*xRes_input)*h_max+i)*3+0] * gauss_2d(sigma,threadPosX-v1_window_width/2,(threadPosY+y*noOfThreadsY)-v1_window_width/2);
			//add vel_max to start at array position 0 for smallest (negative!) value
			int vx = (int) hypoImageInput[((x_input_pos+y_input_pos*xRes_input)*h_max+i)*3+1] + vel_max;
			int vy = (int) hypoImageInput[((x_input_pos+y_input_pos*xRes_input)*h_max+i)*3+2] + vel_max;

			vx = min( vx, width_vel_space - 1 );
			vy = min( vy, width_vel_space - 1 );
			vx = max( vx, 0 );
			vy = max( vy, 0 );
			

			unsigned int w_i= (unsigned int) (10000000.0f *w_f);
//#ifndef CUDA_NO_SM_11_ATOMIC_INTRINSICS 
//			atomicAdd(&cachI[vx+vy*width_vel_space],w_i);
//#else
			cachI[vx+vy*width_vel_space] += w_i;
//#endif
			count_loops++;
		}
	}

	__syncthreads(); //TIME: 50ms for YOS!!!

	//artificial hypotheses
	//for(int i = 0; i< size_vel_space; ++i){
	//	cachI[i] = i*10000 ;
	//}

/*
	//just for debug -- show vel_space at one position
	if(threadIdx.x == 0 && blockIdx.x == 4 && blockIdx.y == 4)
	for(int i = 0; i< size_vel_space; ++i)
	{
		output[i] = float(cachI[i]) / 10000000.0f;
	}
	//output[0] = (float) width_vel_space;
	//just for debug ----end
*/


	//COMMENT FOR SPEED UP:
	//The following paragraph could theoretically be computed by more than threadIdx.x threads - 
	//however, so far we could not achieve an implementation that was running for NO_OF_LINES_LIST > 1 
	//(this indicates the number of threads in y-direction used)
	//Note: when the next 10 lines were included in the "if(threadIdx.y < NO_OF_LINES_LIST)" block, 
	//the programme did not work correctly anymore!
	//Note: when NO_OF_LINES_LIST > 0 then the variable threadPos has to be replaced by 
	//threadPosTotal (but: limit the threadIdx first not to exceed memory!)

	unsigned int* current_min_value, *current_min_pointer, *current_find_array;

	//Get biggest 5 hypos:
		current_min_value  = cachI+(size_vel_space+1+threadPos*(h_max+2)+0);//set pointer to memory for current min value
		current_min_pointer= cachI+(size_vel_space+1+threadPos*(h_max+2)+1);//set pointer to pointer for position within current_find_array
		current_find_array = cachI+(size_vel_space+1+threadPos*(h_max+2)+2);//set pointer to first pos. of current_find_array
		(*current_min_value)   = 0;
		(*current_min_pointer) = 0;
		//initialize with zero
		for (unsigned int i=0; i<h_max; ++i) {
			current_find_array[i] = size_vel_space;
		}

		cachI[size_vel_space] = 0;//initialize default position with zero

		if(threadIdx.y < NO_OF_LINES_LIST)
		{
			//analyze whole velspace using the available threads
			for (unsigned int y=0; y<DIV_UP_MACRO(thread_loops,NO_OF_LINES_LIST); ++y)
			{
				const unsigned int currPos = threadPosTotal+y*noOfThreads*NO_OF_LINES_LIST;

				if (currPos < size_vel_space) //make sure not to exceed corresponding memory
				{
					const unsigned int current_value = (unsigned int) cachI[currPos];

					if (current_value>=(*current_min_value)) //current value bigger than smallest one in the list?
					{
						current_find_array[(*current_min_pointer)] =  (currPos);

						//just for debug -- show current_find_arrays at one position
						//if(threadIdx.x == 0 && blockIdx.x == 4 && blockIdx.y == 4)
						//{
						//	output[i] = ((float)cachI[current_find_array[(*current_min_pointer)]])/ 10000000.0f;
						//	++i;
						//}
						//just for debug ----end


						// seach for the new smallest value of hmax biggest
						*current_min_pointer=0;
						*current_min_value  =cachI[current_find_array[0]];
						for (unsigned int i=1; i<h_max; ++i) {
							if (cachI[current_find_array[i]]<=(*current_min_value)) {
								(*current_min_pointer)=i;
								(*current_min_value)  =cachI[current_find_array[i]];
							}
						}
					}
				}
			}
		}
		__syncthreads(); //TIME: 30ms for YOS!!!


		//combine the results of noOfThreads lists with h_max hypotheses each
		if(threadIdx.x == 0 && threadIdx.y == 0)
		{

			for (unsigned int y=1; y<noOfThreads*NO_OF_LINES_LIST; ++y) //start with second current_find_array
			{
					for(unsigned int j=0; j<h_max; ++j)
					{
						if(cachI[current_find_array[y*(h_max+2)+j+2]] >= (*current_min_value))
						{
							current_find_array[(*current_min_pointer)] = current_find_array[y*(h_max+2)+j+2];//pointer to position in velspace

							//just for debug -- show current_find_arrays at one position
							//if(threadIdx.x == 0 && blockIdx.x == 4 && blockIdx.y == 4)
							//{
							//	output[i] = ((float)cachI[current_find_array[(*current_min_pointer)]])/ 10000000.0f;
							//	++i;
							//}
							//just for debug ----end


							// seach for the new smallest value of hmax biggest
							(*current_min_pointer)= 0;
							(*current_min_value)  = cachI[current_find_array[0]];
							for (unsigned int i=1; i<h_max; ++i) {
								if (cachI[current_find_array[i]]<=(*current_min_value)) {
									(*current_min_pointer)=i;
									(*current_min_value)  =cachI[current_find_array[i]];
								}
							}
						}
					}

			}
		}

		__syncthreads();

/*
	//just for debug -- show current_find_array at one position
	if(threadIdx.x == 0 && blockIdx.x == 4 && blockIdx.y == 4)
	for(int i = 0; i< h_max; ++i) //for(int i = 0; i< noOfThreads*(h_max+2)+2; ++i)
	{

		output[i*3 + 0] = cachI[current_find_array[i]]/10000000.0f; //((float)cachI[cachI[size_vel_space+1+i]])/ 10000000.0f;
		output[i*3 + 1] = int(current_find_array[i]/width_vel_space - width_vel_space/2);
		output[i*3 + 2] = int(current_find_array[i]%width_vel_space - width_vel_space/2);
	}
	//just for debug ----end
*/
//

//copy to global memory
	if(threadIdx.x < 1) // && blockIdx.x == 4 && blockIdx.y == 4)
	for(int i = 0; i< h_max; ++i)
	{ //write result into output array
		output[0+3*(i+h_max*(y_output_pos*xRes_output+x_output_pos))] = float(cachI[current_find_array[i]]/10000000.0f);
		output[2+3*(i+h_max*(y_output_pos*xRes_output+x_output_pos))] = int(current_find_array[i]/width_vel_space - vel_max ) ;// - width_vel_space/2 to shift origina to center position
		output[1+3*(i+h_max*(y_output_pos*xRes_output+x_output_pos))] = int(current_find_array[i]%width_vel_space - vel_max ) ;
	}
}


extern "C"
void do_subsample_hypo_image (
		int h_max, int vel_max, float sigma,
		int v1_window_width,
		float* input,  int xRes_input,  int yRes_input,
		float* output, int xRes_output, int yRes_output)
{
	dim3 dimBlock(v1_window_width, NO_OF_LINES_SS);//CHANGED
	int bx = xRes_output;
	int by = yRes_output;
	dim3 dimGrid(bx,by);
	size_t n_shared = 	  (sizeof(float)* ((2*vel_max+1)*(2*vel_max+1) 
										   + (v1_window_width*(h_max+2)*NO_OF_LINES_LIST+1)));


	d_subsample_hypo_image<<<dimGrid, dimBlock, n_shared >>>
	  (
	   h_max, vel_max, sigma,
	   input, xRes_input , yRes_input,
	   output        , xRes_output, yRes_output
	   );

	cudaError_t err = cudaThreadSynchronize();                               \
	if( cudaSuccess != err) {                                                \
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_subsample_hypo_image() execution failed\n", __FILE__, __LINE__, cudaGetErrorString( err) );\
		exit(EXIT_FAILURE);                                                  \
	}
}

#undef NO_OF_LINES_SS
#undef NO_OF_LINES_LIST
#undef DIV_UP_MACRO

#endif
//}}} SUBSAMPLE_OLD


//{{{ SUBSAMPLE_DIRECT
#ifdef SUBSAMPLE_DIRECT

#if defined(CUDA_NO_SM_11_ATOMIC_INTRINSICS) && defined(USE_ATOMICS)
//#		error "You have enabled atomics, but they are not supported by your platform!"
#endif


template <typename T> __device__ void
swap (T *a, T *b)
{
	T tmp = *a;
	*a = *b;
	*b = tmp;
}


__global__ void 
d_subsample (int h_max, int vel_max, float sigma, int v1_window_width,
		float *input, int input_width, int input_height,
		float *output, int output_width, int output_height)
{
	// variable setup
	const int xratio = input_width / output_width;
	const int yratio = input_height / output_height;
	
	const int xOutputPos = blockIdx.x;
	const int yOutputPos = blockIdx.y;

	const int nThreads = blockDim.x;

	const int vspace_width  = (vel_max * 2)  + 1;
	const int vspace_size   = vspace_width * vspace_width;
	const int cache_size    = vspace_size + SUBSAMPLE_THREAD_COUNT * h_max + 1;
	const int hypos_per_row = v1_window_width * h_max;

	// cache setup
	extern __shared__ float cache[];
	float *vspace;
	unsigned int *ucache;
	vspace   = cache;
	ucache   = (unsigned int *)&cache[vspace_size + 1];

	reset_cache<float>(cache, cache_size, nThreads, 0.0);

	// iterate over all input lines
	const int shift = (v1_window_width / 2);
	const int max_input_per_row = input_width * h_max * 3;

	// load and process all rows
	const int x_input_base_pos = xOutputPos * xratio - shift;
	for (int y = 0; y < v1_window_width; y++) {
		const int y_input_pos = yOutputPos * yratio + y - shift;
		if (y_input_pos >= 0 && y_input_pos < input_height) {
			const int base_offset = (y_input_pos * input_width + x_input_base_pos) * h_max * 3;

			for (int i = 0; i < DIVUP(hypos_per_row, nThreads); i++) {
				const int n = i * nThreads + threadIdx.x;
				const int x = n / h_max;

				if (n < hypos_per_row) {
					const int row_offset = (x_input_base_pos * h_max + n) * 3;
					const int offset     = base_offset + n * 3;

					if (offset >= 0 && row_offset >= 0 && row_offset < max_input_per_row) {
		
						float wf = input[offset + 0] * gauss_2d(sigma, x - shift, y - shift);
						int dx   = clamp<int>(int(input[offset + 1]) + vel_max, 0, vspace_width -1);
						int dy   = clamp<int>(int(input[offset + 2]) + vel_max, 0, vspace_width -1);
#ifdef USE_ADD_HEURISTIC
						volatile int dxdy = dx + dy * vspace_width;
#else
						int dxdy = dx + dy * vspace_width;
#endif
		

#if !defined(CUDA_NO_SM_11_ATOMIC_INTRINSICS) && defined(USE_ATOMICS)
						atomicAdd((int *)(cache + dxdy), int(wf * SCALING));
#else
#	ifdef USE_ADD_HEURISTIC
						volatile float tmp;
						do {	
							tmp = vspace[dxdy] + wf;
							vspace[dxdy] = tmp;
						} while (vspace[dxdy] != tmp);
#	else
						vspace[dxdy] += wf;
#	endif
#endif
					}
				}
				//__syncthreads();
			}
		}
	}
	//__syncthreads();

#if !defined(CUDA_NO_SM_11_ATOMIC_INTRINSICS) && defined(USE_ATOMICS)
	search_biggest<int>((int*)&vspace[0], ucache, h_max, vspace_size, nThreads);
	write_back<int>(output, (int*)&vspace[0], ucache, h_max, vspace_width, vel_max, nThreads,
			xOutputPos, yOutputPos, output_width);
#else
	search_biggest<float>(vspace, ucache, h_max, vspace_size, nThreads);
	write_back<float>(output, vspace, ucache, h_max, vspace_width, vel_max, nThreads,
			xOutputPos, yOutputPos, output_width);
#endif
}



extern "C" void 
do_subsample_hypo_image (int h_max, int vel_max, float sigma, int v1_window_width,
		float *input, int xRes_input, int yRes_input,
		float *output, int xRes_output, int yRes_output)
{
	dim3 grid(xRes_output, yRes_output);
	dim3 block(SUBSAMPLE_THREAD_COUNT,1,1);		
	size_t cache_size = sizeof(float) * ( 
			(vel_max * 2 + 1) * (vel_max * 2 + 1) +
			(SUBSAMPLE_THREAD_COUNT * h_max) + 1);

	d_subsample<<<grid,block,cache_size>>>(h_max, vel_max, sigma, v1_window_width,
			input, xRes_input, yRes_input, 
			output, xRes_output, yRes_output);

	cudaError_t e = cudaThreadSynchronize();
	if (e != cudaSuccess) {
		// TODO: move the fprintf out of the cuda file
		fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
				"d_subsample() execution failed\n", __FILE__, __LINE__, cudaGetErrorString(e) ); \
		exit(EXIT_FAILURE);
	}
}

#endif
//}}} SUBSAMPLE_DIRECT

#endif
//}}} SUBSAMPLE_MULTITHREAD

