/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cuda.h>
#include <cutil.h>
//#include "hallo.cu"

#define DISTP (1+1)
#define APRON (2*DISTP)

#define CHUNK_WIDTH  8
#define CHUNK_HEIGHT 6

#define CACH_WIDTH  (2*APRON+CHUNK_WIDTH)
#define CACH_HEIGHT (2*APRON+CHUNK_HEIGHT)

// defines the positions for the zensus value comparison
#define O00 (-APRON-APRON*CACH_WIDTH)
#define O01 (      -APRON*CACH_WIDTH)
#define O02 (+APRON-APRON*CACH_WIDTH)
#define O03 (-DISTP-DISTP*CACH_WIDTH)
#define O04 (      -DISTP*CACH_WIDTH)
#define O05 (+DISTP-DISTP*CACH_WIDTH)
#define O06 (-APRON)
#define O07 (-DISTP)
#define O08 (+DISTP)
#define O09 (+APRON)
#define O10 (-DISTP+DISTP*CACH_WIDTH)
#define O11 (       DISTP*CACH_WIDTH)
#define O12 (+DISTP+DISTP*CACH_WIDTH)
#define O13 (-APRON+APRON*CACH_WIDTH)
#define O14 (       APRON*CACH_WIDTH)
#define O15 (+APRON+APRON*CACH_WIDTH)

// defines the positions for the zensus value within the bit string (to basis 3)
#define F00 1
#define F01 F00*3
#define F02 F01*3
#define F03 F02*3
#define F04 F03*3
#define F05 F04*3
#define F06 F05*3
#define F07 F06*3
#define F08 F07*3
#define F09 F08*3
#define F10 F09*3
#define F11 F10*3
#define F12 F11*3
#define F13 F12*3
#define F14 F13*3
#define F15 F14*3

#define THREADIDX1D threadIdx.x+blockDim.x*threadIdx.y

__device__ unsigned int dd_comp(float& center, float& val, const float& eps) {
	if (center-eps>val)
		return 1;
	if (center+eps<val)
		return 2;
	return 0;
}

__global__ void d_zensus(const int w, const int h, const float* input, unsigned int* output, const float eps) 
{
	const int x_b = blockIdx.x*CHUNK_WIDTH;
	const int y_b = blockIdx.y*CHUNK_HEIGHT;
	
	const int x = x_b + threadIdx.x - APRON;
	const int y = y_b + threadIdx.y - APRON;
	
	const int x_c = max(0,min(w-1,x));
	const int y_c = max(0,min(h-1,y));
	
	extern __shared__ float cach[];
	
	cach[THREADIDX1D] = input[x_c+y_c*w];
	
	__syncthreads();
	
	if (threadIdx.x>=APRON && threadIdx.x-APRON<CHUNK_WIDTH  && x<w &&
		threadIdx.y>=APRON && threadIdx.y-APRON<CHUNK_HEIGHT && y<h ) {
		output[x+y*w]   =    0 + //(F00-1) +
							 F00*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O00],eps) +
		                     F01*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O01],eps) +
		                     F02*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O02],eps) +
		                     F03*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O03],eps) +
		                     F04*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O04],eps) +
		                     F05*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O05],eps) +
		                     F06*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O06],eps) +
		                     F07*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O07],eps) +
		                     F08*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O08],eps) +
		                     F09*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O09],eps) +
		                     F10*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O10],eps) +
		                     F11*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O11],eps) +
		                     F12*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O12],eps) +
		                     F13*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O13],eps) +
		                     F14*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O14],eps) +
		                     F15*dd_comp(cach[THREADIDX1D],cach[THREADIDX1D+O15],eps);
	}
}



extern "C" void do_zensus(const int w, const int h, const float* input, float* output, float eps) {

  dim3 blockDim(CACH_WIDTH,CACH_HEIGHT);
  dim3 gridDim (iDivUp(w,CHUNK_WIDTH),iDivUp(h,CHUNK_HEIGHT));
  
  int cs = CACH_WIDTH * CACH_HEIGHT * sizeof(float);
  
  d_zensus<<<gridDim,blockDim,cs>>>(w,h,input,(unsigned int*)output,eps);


}
