/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cmath>
#include <iostream>

#include <GPUImage.h>
#include "motion-utils.h"

using namespace std;


extern "C" void directionalDerivative(float* output, int width, int height,
									  float theta,
									  float* dxx, float* dyy, float* dxy);


int main(int argc, char* argv[]) {

	// TODO: NOT YET FINISHED!!

	// Check usage
	if (argc != 2)
		{
			cerr << "Usage: " << argv[0] << " input_output.h5" << endl;
			return 1;
		}

	// Parameters
	float sgd = 1.0;
	int ddim = 8;
	int vspace = 5;
	int vdim = 25;
	
	// Open and check the HDF5 file
	// TODO: Don't do it by hand
	H5File file(argv[1], H5F_ACC_RDWR);
	DataSet in_dset = file.openDataSet(arg_in.getValue());
	DataSpace in_dspace = in_dset.getSpace();
	int rank = in_dspace.getSimpleExtentNdims();
	assert(rank == 3);
	hsize_t dims[3];
	in_dspace.getSimpleExtentDims(dims);
	int width = dims[2];
	int height = dims[1];
	int depth = dims[0];
	
	
	// Dataset for the second directional derivatives
	// TODO: Don't do it by hand
	hsize_t dde_dims[4] = {depth, ddim, height, width};
	DataSpace dde_space(4, dde_dims);
	DataSet dde_set = file.createDataSet("dderiv",
										 PredType::IEEE_F32LE, dde_space);
	dde_dims[0] = 1;
	DataSpace dde_mem_space(4, dde_dims);
	// Dataset for the 2nd direction derivative sums
	DataSet dds_set = file.createDataSet("ddsum",
										 PredType::IEEE_F32LE, in_dspace);
	// Datasets for the half detectors
	hsize_t hpos_dims[4] = {depth - 1, vdim, height, width};
	DataSpace hpos_space(4, hpos_dims);
	DataSet hpos_set = file.createDataSet("hpos",
										  PredType::IEEE_F32LE, hpos_space);
	DataSet hneg_set = file.createDataSet("hneg",
										  PredType::IEEE_F32LE, hpos_space);
	DataSet corr_set = file.createDataSet("corr",
										  PredType::IEEE_F32LE, hpos_space);
	hpos_dims[0] = 1;
	DataSpace hpos_mem_space(4, hpos_dims);
	
	// GPU arrays
	GPUData frame(height, width),
		dx(height, width),
		dy(height, width),
		dxy(height, width),
		dxx(height, width),
		dyy(height, width),
		dderiv(ddim, height, width),
		prev_dderiv(ddim, height, width),
		ddsum(height, width),
		ext_dderiv(ddim, height, width),
		ext_dderiv(ddim, height, width);

	// Iterations loop
	for (int t = 0; t < depth; t++)
		{
			cout << "[" << (t + 1) << "/" << depth << endl;
			
			// Load input frame
			load_frame(frame.getData(), in_dset, t);

			// Compute X and Y 2nd derivatives
			// TODO: do this in one step after being checked
			frame.dericheFilter(sgd, 1, 1, dx);
			frame.dericheFilter(sgd, 0, 1, dy);
			dx.dericheFilter(sgd, 1, 1, dxx);
			dx.dericheFilter(sgd, 0, 1, dxy);
			frame.dericheFilter(sgd, 0, 1, dyy);
			
			// Compute the directional derivatives
			for (int d = 0; d < ddim; d++)
				{
					float theta = d * M_PI / ddim;
					directionalDerivative(dderiv[d].getData(), width, height,
										  theta, dxx.getData(),
										  dyy.getData(), dxy.getData());
				}
			
			// Divise normalization
			// TODO: sum in GPU
			// TODO: one blur along each dim for dericheFilter
			ddsum = (dderiv.sum(0).abs().dericheFilter(sgb, 0, 0)
					 .dericheFilter(sgb, 1, 0));
			for (int d = 0; d < ddim; d++)
				dderiv[d] /= (epsilon + ddsum);
			
			// TODO: how do we correlate? surround by 0?
			ext_prev_dderiv = dderiv;
			ext_dderiv.fill(0);

			// Skip the first frame
			if (t == 0)
				continue;

			// Compute the half detectors
			for (int v = 0; v < vdim; v++)
				{
					int vy = vspace[v][0];
					int vx = vspace[v][1];
					
					plus.fill(0);
					minus.fill(0);
					for (int d = 0; d < ddim; d++)
						{
							// Something like that...
							plus += ext_prev_dderiv[d,vy_max:vy_max+height,vx_max:vx_max+width]
								* ext_dderiv[d,vy_max+vy:vy_max+vy+height,vx_max+vx:vx_max+vx+width]
							minus += dderiv[d] * ext_prev_dderiv[d,vy_max+vy:vy_max+vy+height,vx_max+vx:vx_max+vx+width]
						}
					hplus[v] = dericheFilter(plus, sgh);
					hminus[v] = dericheFilter(minus, sgh);
				}
			
			// Combine the half detectors
			hminus.rectify();
			hplus.rectify();
			p0 = (hplus - 0.5 * hminus) / (1.0 + hminus);
			p0.rectify();
		}
}



// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
