/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "CartesianCorrelator.h"
#include "kernels.hpp"
#include <cstdio>
#include <math.h>
#include <limits>

#define ownPI 3.141593f
#define ownOmega 2*ownPI/5

//inline void info(const GPUData& d) {
//	printf("info : min=%f, max=%f\n",d.min(),d.max());
//}

CartesianCorrelator::CartesianCorrelator(int resX, int resY, int maxVel, float subInhibConstA, float divInhibConstA)
	: subInhibConst(subInhibConstA)
	, divInhibConst(divInhibConstA)
	, gauss()
	, maxVelX(maxVel)
	, maxVelY(maxVel)
	, tempImage(resX,resY)
	, tempImage2(resX,resY)
	, tempImage3(resX,resY)
	, tempOplus(resX,resY)
	, tempOminus(resX,resY)
	, fa(0.2)
	, usePrior(false)
	, gauss_4(4.0f)
	, sigma_of_input_features_unsharpnes(0.75f)
	, use_new_prior(-1)
{
	size_t s = (5*2.0); if (s%2==0) s++;
	gauss = GPUData(s);
	fillWithGauss(gauss,2.0f);
}

CartesianCorrelator::~CartesianCorrelator() {
	;
}

extern "C" void inhibitionAndRect(const int N, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor);

static void GPUDataAs3D( GPUData & in, GPUData & out )
{
  int ndim = in.getNoOfDimensions();
  if ( ndim == 3 )
	{
	  out.reference( in );
	}
  else
	{
//	  int outstrides[3] = { 1, in.stride(ndim-2), in.stride(ndim-1) };
//	  int outdims[3] = { 1, in.extent(ndim-2), in.extent(ndim-1) };
//	  for ( int i = 0; i < ndim-2; ++i )
//		{
//		  outstrides[0] = ( i == 0 ) ? in.stride(i) : std::min( in.stride(i), outstrides[0] );
//		  outdims[0] *= in.extent(0);
//		}
//	  out.reference( in.view( 3, outdims, outstrides ) );
	  size_t h_dim = in.getNoOfDimensions()-2;
	  out.reference(in.reduceDimensionality(h_dim));
	}
}

void CartesianCorrelator::correlate(GPUData& images1,
									GPUData& images2,
									int shiftX, int shiftY,
									GPUImage& correlationResultPlus,
									GPUImage& correlationResultMinus)
{
	if ( images1.getNoOfDimensions() != 3 || images2.getNoOfDimensions() != 3 )
		throw GPUDataException("CartesianCorrelator can only handle 3d data");

	const unsigned int noOfImages = (unsigned) images1.getDimensions()[0];

	/* O_plus */
	{
		tempOplus.fill(0.0f);

		for (unsigned int i=0; i<noOfImages; ++i)
			{
				GPUImage im1(images1[i]), im2(images2[i]);
				im2.shiftMulAndAdd( im1, tempOplus, -shiftX, -shiftY );
			}

		tempOplus.convolute1D(gauss,1,tempImage);
		tempImage.convolute1D(gauss,2,tempOplus);
		tempOplus.rect();
	}

	/* O_minus */
	if (shiftX==0 && shiftY==0) {
		tempOminus.copyFrom(tempOplus);
	} else {
		tempOminus.fill(0.0f);

		for (unsigned int i=0; i<noOfImages; ++i)
			{
				GPUImage im1(images1[i]), im2(images2[i]);
				im1.shiftMulAndAdd( im2, tempOminus, -shiftX, -shiftY );
			}

		tempOminus.convolute1D(gauss,1,tempImage);
		tempImage.convolute1D(gauss,2,tempOminus);
		tempOminus.rect();
	}

	
	float sc = 1.0f;
	if (use_new_prior>0.0f) {
#ifndef _WIN32
		const float r  = sqrt(shiftX*shiftX+shiftY*shiftY);
		sc = float( 1.0-use_new_prior*exp(-2.0 * pow(r/sigma_of_input_features_unsharpnes,2.0)) );
#else
		const float r  = sqrt((float)(shiftX*shiftX+shiftY*shiftY));
		sc = float( 1.0-use_new_prior*exp(-2.0f * pow((float)(r/sigma_of_input_features_unsharpnes),2.0f)) );
#endif
	}
	
	inhibitionAndRect(correlationResultPlus.size(),correlationResultPlus.getDataPointer(),tempOplus.getDataPointer(),tempOminus.getDataPointer(),subInhibConst*sc,1.0f,divInhibConst*sc);

	if (shiftX==0 && shiftY==0) {
		correlationResultMinus.copyFrom(correlationResultPlus);
	} else {
		tempOplus.swap(tempImage);
		tempOplus.copyFromShifted(tempOminus, shiftX, shiftY);
		tempOminus.copyFromShifted(tempImage, shiftX, shiftY);

		inhibitionAndRect(correlationResultMinus.size(),correlationResultMinus.getDataPointer(),tempOplus.getDataPointer(),tempOminus.getDataPointer(),subInhibConst*sc,1.0f,divInhibConst*sc);

		tempOplus.swap(tempImage); // swap back
	}
}

void CartesianCorrelator::normalizeInput(GPUData& data) {
	tempImage3.resize(data.extent(2),data.extent(1));
	tempImage2.resize(data.extent(2),data.extent(1));
	for (unsigned int i=0; i<data.extent(0); ++i) {
		tempImage3.copyFrom(data[i]);
		tempImage3.abs();
		if (i==0)
			tempImage2.copyFrom(tempImage3);
		else
			tempImage2.add(tempImage3);
	}
	tempImage2.add(0.000001f);
	gauss_4.apply(tempImage2,tempImage3);
	for (unsigned int i=0; i<data.extent(0); ++i) {
		data[i].div(tempImage3);
	}
}

void CartesianCorrelator::correlateAll(
		GPUData& ims1,
		GPUData& ims2,
		GPUVelActivities& correlationResult,
		bool normalize)
{
	// hack coz below assumes 3d data
	GPUData images1, images2;
	GPUDataAs3D( ims1, images1 );
	GPUDataAs3D( ims2, images2 );

	if (normalize) {
		normalizeInput(images1);
		normalizeInput(images2);
	}

//	correlationResult.getData().fill(0.0f);

	for ( int dy = -maxVelY; dy <= maxVelY; ++dy )
		{
			for ( int dx = 0; dx <= maxVelX; ++dx )
				{
					correlate(images1, images2, dx, dy, tempImage2,tempImage3);
					correlationResult.getOneVelocityAsImage(maxVelX+dx,maxVelY+dy).copyFrom(tempImage2);
					correlationResult.getOneVelocityAsImage(maxVelX-dx,maxVelY-dy).copyFrom(tempImage3);
				}
		}

//	for (int dy=-maxVelY; dy<=maxVelY; ++dy) {
//		for (int dx=-maxVelX; dx<=maxVelX; ++dx)  {
//			correlate(images1, images2, dx, dy, tempImage2);
//			correlationResult.getOneVelocityAsImage(dx+maxVelX,dy+maxVelY).copyFrom(tempImage2);
//		}
//	}

	correlationResult.killBorder(maxVelX,maxVelY);

//	if (normalize) {
//		float maxActivity     = correlationResult.getData().max();
////		std::cout << "max activity : " << maxActivity << std::endl;
//		float meanMaxActivity = fa.average( maxActivity );
////		std::cout << "meanMax activity : " << meanMaxActivity << std::endl;
//		if (meanMaxActivity<=std::numeric_limits<float>::epsilon()) {
//			correlationResult.getData().div( 2*std::numeric_limits<float>::epsilon() );
//		} else {
//			correlationResult.getData().div( meanMaxActivity );
//		}
//	} else {
////		float maxActivity     = correlationResult.getData().max();
////		std::cout << "max activity (but not normalize): " << maxActivity << std::endl;
////		float meanMaxActivity = fa.average( maxActivity );
////		std::cout << "meanMax activity (but not normalize) : " << meanMaxActivity << std::endl;
//	}

	if (usePrior) {
		correlationResult.addToVelSpace(priorToLowVelocities,1.0/(2000*(5*2+1)*(5*2+1)));
		correlationResult.mulToVelSpace(priorToLowVelocities);
	}
}

void CartesianCorrelator::correlateAll(
		GPUData& images1,
		GPUData& images2,
		GPUData& correlationResult,
		bool normalize)
{
	GPUVelActivities temp(correlationResult);
	correlateAll(images1,images2,temp,normalize);
}

GPUImage CartesianCorrelator::getXVelMatrix() const {
	const int xRes = maxVelX*2+1;
	const int yRes = maxVelY*2+1;
	const int size = xRes*yRes;

	GPUImage velX(xRes,yRes);
	float* dataX = new float[size];

	int n=0;
	const int xc = xRes/2;
	for (int y=0; y<yRes; ++y) {
		for (int x=0; x<xRes; ++x) {
			dataX[n] = float(x-xc);
			++n;
		}
	}

	velX.copyFromHostMemory(dataX);
	delete[] dataX;
	return velX;
}

GPUImage CartesianCorrelator::getYVelMatrix() const {
	const int xRes = maxVelX*2+1;
	const int yRes = maxVelY*2+1;
	const int size = xRes*yRes;

	GPUImage velY(xRes,yRes);
	float* dataY = new float[size];

	int n=0;
	const int yc = yRes/2;
	for (int y=0; y<yRes; ++y) {
		for (int x=0; x<xRes; ++x) {
			dataY[n] = float(y-yc);
			++n;
		}
	}

	velY.copyFromHostMemory(dataY);
	delete[] dataY;
	return velY;
}

static inline void createPriorToLowVelocities(GPUImage& res, const GPUImage& xComp, const GPUImage& yComp, const float sigma, const float maxAllowed) {
        res.pythagoras(xComp,yComp);
        GPUImage temp(res.copy());
        res.sqr();
        res.div(-sigma*sigma*2);
        res.exp();
        temp.setIfBetween(maxAllowed+0.00000001f, maxAllowed*100000, 0.0f, res);
        res.div(res.max());
}

void CartesianCorrelator::usePriorToLowVelocities(float sigma) {
	if (sigma==-1.0f) {
		usePrior = false;
	} else {
		const int uRes = maxVelX*2+1;
		const int vRes = maxVelY*2+1;
		GPUImage prior(uRes, vRes);
		createPriorToLowVelocities(prior,getXVelMatrix(),getYVelMatrix(),sigma,maxVelX*sqrt(2.0));
		priorToLowVelocities = prior;
		usePrior = true;
	}
}


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
