/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef CartesianCORRELATOR_H_
#define CartesianCORRELATOR_H_

#include "GPUImage.h"
#include "GPUVelActivities.h"
#include "GPUVelField.h"
#include "SepGaborWaveletFilter.h"
#include "FloatingAverage.h"
#include <vector>

/**
 * This class calculates the correlation for a bunch of different velocities
 * (pixel movements).
 * The function correlateAll() does this for all movements in the range of
 * [-maxVelX , maxVelX] and [-maxVelY , maxVelY].
 *
 * The result of a correlation for one special movement (correlate()) or the
 * whole set of correlations (correlateAll()) is calculated using the following
 * function:
 *
 * \f$ cor = { \left( cor_p - subInhibConst * cor_m \right) \over \left(1 + divInhibConst * cor_m\right) } \f$
 *
 * with
 *
 * \f$ cor_p = < I_1(x,y) | I_2(x+dx,y+dy) > \f$
 *
 * \f$ cor_m = < I_2(x,y) | I_1(x+dx,y+dy) > \f$
 *
 * This results in a likelyhood for a displacement of (dx,dy) from
 * image 1 to image 2.
 */
class CartesianCorrelator {

private:

	float subInhibConst;
	float divInhibConst;
	
	GPUData gauss;

	int maxVelX;
	int maxVelY;

	GPUImage tempImage;
	GPUImage tempImage2;
	GPUImage tempImage3;

	GPUImage tempOplus;
	GPUImage tempOminus;

	FloatingAverage fa;

	bool usePrior;
	GPUImage priorToLowVelocities;

	GaussKernel2 gauss_4;
	void normalizeInput(GPUData& input);

	float sigma_of_input_features_unsharpnes;
	float use_new_prior;
public:

	CartesianCorrelator(int resX, int resY, int maxVel, float subInhibConstA = 0.5f, float divInhibConstA = 1.0f);

	~CartesianCorrelator();

	void correlate(GPUData& images1, GPUData& images2, int shiftX, int shiftY, GPUImage& correlationResultPlus, GPUImage& correlationResultMinus);

	void correlateAll(GPUData& images1, GPUData& images2, GPUVelActivities& correlationResult, bool normalize = false);
	void correlateAll(GPUData& images1, GPUData& images2, GPUData& correlationResult, bool normalize = false);

	GPUImage getXVelMatrix() const;
	GPUImage getYVelMatrix() const;

	void usePriorToLowVelocities(float sigma = 6.0f);

	bool isUsingPrior() { return usePrior; }
	
	float get_sigma_of_input_features_unsharpnes() {
		return sigma_of_input_features_unsharpnes;
	}
	void set_sigma_of_input_features_unsharpnes(float s) {
		sigma_of_input_features_unsharpnes = s;
	}
	void set_use_new_prior(float p) {
		use_new_prior = p;
	}
	float get_use_new_prior() {
		return use_new_prior;
	}
};


#endif

