/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/

#include "CensusClass.h"
#include "CPUData.hpp"

#include <stdio.h>
#include <math.h>
#include <algorithm>
#include <functional>
#include <stdexcept>

#include <boost/format.hpp>

CensusClass::CensusClass(void)
{
	c = 0;
	//iPos = 0;
	iH = 0;
}

CensusClass::CensusClass(unsigned int c0, int iPos0, int iH0, int iCount0, int iFrame0)
{
	c = c0;
	//iPos = iPos0;
	iH=iH0;
}

CensusClass::~CensusClass(void)
{
}

//for greater - sort
static inline bool UDgreaterCensus(CensusClass elem1, CensusClass elem2)
{
  return elem1.c > elem2.c;
}


void build_sorted_census_list( CensusClass * list, const unsigned int * values, int w, int h )
{
  for ( int y = 0, i = 0; y < h; ++y )
	for ( int x = 0; x < w; ++x, ++i )
	  {
		list[i].c = values[i];
		list[i].iPosX = x;
		list[i].iPosY = y;
		list[i].iH = 1;
	  }

  std::sort( list, list + w * h, UDgreaterCensus );


}

/// fills in the iH fields of every entry with the number of occurences
/// of the corresponding census value
/// works by keeping two pointers, one to iterate, the other only moves up when
/// there is a change in census values, filling in the iH along the way
void count_census_occurences( CensusClass * entries, size_t N )
{
  if ( N == 0 ) return;

  // initialize
  size_t istart = 0;
  unsigned int c = entries[0].c;
  // scan
  for ( size_t i = 1; i < N; ++i )
	{
	  unsigned int ci = entries[i].c;
	  // check for change in census value
	  if ( ci != c )
		{
		  // move up second pointer, filling in all iH in the block
		  size_t n_occ = i - istart;
		  for ( ; istart < i; ++istart )
			entries[istart].iH = n_occ;
		  // make the new census value current
		  c = ci;
		}
	}
  // fill in the last block
  size_t n_occ = N - istart;
  for ( ; istart < N; ++istart )
	entries[istart].iH = n_occ;  
}

void CensusClass::createSortedCensusList(unsigned int *currCensusList, int w, int h)
{
  build_sorted_census_list( this, currCensusList, w, h );
  count_census_occurences( this, w * h );
#if 0
  for ( size_t i = 0; i < w * h; ++i )
	{
	  CensusClass & c = this[i];
	  std::cout << str( boost::format("x: %4i y: %4i n: %4i x: %i") % c.iPosX % c.iPosY % c.iH % c.c ) << std::endl;
	}
#endif
}

void CensusClass::computeHyposUnambiguous( CensusClass* pCurr, 
										   unsigned int *currCensusList, 
										   float *censusOcc, 
										   HypoImage & HypoArray1, 
										   int w, int h, int H_MAX_1, int VEL_MAX )
{
  CPUData<float> hostHypoImage;
  computeHyposUnambiguous( pCurr, currCensusList, censusOcc, hostHypoImage, w, h, H_MAX_1, VEL_MAX );
  HypoArray1.getData().copyFromHostMemory( hostHypoImage.data() );
}

void CensusClass::computeHyposUnambiguous( CensusClass* pLast, 
										   unsigned int *currCensusList, 
										   float *censusOcc, 
										   CPUData<float> & hostHypoImage, 
										   int w, int h, int H_MAX_1, int VEL_MAX)
{

	size_t i, j;

	size_t delta = w * h - 1, iend = w * h;

	hostHypoImage.resize( h, w, H_MAX_1 * 3 );
	hostHypoImage.make_contiguous();

	createSortedCensusList( currCensusList, w, h );

	// collect hypo ambiguities
	for( i = 0; i != iend; ++i )
		censusOcc[ pLast[i].iPosX + w * pLast[i].iPosY ] = (float) pLast[i].iH;

	census::match_census_lists( this, pLast, w * h, hostHypoImage, H_MAX_1, VEL_MAX );

}

//GPU computation for hypotheses of unambiguous shifts
extern "C" void do_computeHyposAmbiguous(
			float* censusOccDev,
			float* hypoImageV1,
			float* hypoImageMT,
			float* currCensus,
			float* pastCensus,
			const int w, 
			const int NPix, 
			const int fbf, const float p0, const int df_mt,
			const int VEL_MAX, 
			const int H_MAX_1, 
			const int H_MAX_2);

void CensusClass::computeHyposAmbiguous( float *censusOcc, 
										 HypoImage HypoArrayV1, 
										 HypoImage HypoArrayMT, 
										 GPUImage censusOccDev, 
										 GPUImage currCensus, 
										 GPUImage pastCensus, 
										 int fbf, int H_MAX_2, int VEL_MAX, float p0 )
{
	censusOccDev.copyFromHostMemory(censusOcc);
	do_computeHyposAmbiguous(
			censusOccDev.getDataPointer(),
			HypoArrayV1.getData().getDataPointer(),
			HypoArrayMT.getData().getDataPointer(),
			currCensus.getDataPointer(),
			pastCensus.getDataPointer(),
			HypoArrayV1.getWidth(), 
			HypoArrayV1.getWidth() * HypoArrayV1.getHeight(), 
			fbf, p0, HypoArrayV1.getWidth() / HypoArrayMT.getWidth(),
			VEL_MAX, HypoArrayV1.getHMAX(), H_MAX_2 );

	censusOccDev.copyToHostMemory(censusOcc);
}

void CensusClass::computeHyposAmbiguous( float *censusOcc, 
										 HypoImage HypoArrayV1, 
										 const std::vector<HypoFeed> & feeds ,
										 GPUImage censusOccDev, 
										 GPUImage currCensus, 
										 GPUImage pastCensus, 
										 int H_MAX_2, int VEL_MAX, float p0 )
{
	censusOccDev.copyFromHostMemory(censusOcc);
	do_compute_hypos_ambiguous(
			censusOccDev.getDataPointer(),
			HypoArrayV1.getData().getDataPointer(),
			currCensus.getDataPointer(),
			pastCensus.getDataPointer(),
			HypoArrayV1.getWidth(), 
			HypoArrayV1.getWidth() * HypoArrayV1.getHeight(), 
			p0, feeds.size(), &feeds[0], 
			HypoArrayV1.getHMAX(), H_MAX_2,VEL_MAX );

	censusOccDev.copyToHostMemory(censusOcc);
}



SequentialCensusMatcher::SequentialCensusMatcher( ) 
  : _width(0)
  , _height(0)
  , _vmax(30)
  , _hmax1(4)
  , _hmax2(1000)
  , _have_current(false)
  , _have_last(false)
  , _current_entries( _entries1 )
  , _last_entries( _entries2 )
  , _current_census_image( _census_image1 )
  , _last_census_image( _census_image2 )
{
}

void SequentialCensusMatcher::break_sequence()
{
  _have_current = _have_last = false;
}

void SequentialCensusMatcher::resize( size_t width, size_t height )
{
  if ( _width != width || height != _height )
	{
	  break_sequence();

	  _width = width;
	  _height = height;

	  _entries1.resize( _width * _height );
	  _entries2.resize( _width * _height );

	  _ambiguities.resize( _height, _width, 1 );

	  _matches.resize( _height, _width, 3 * _hmax1 );
	  _gpuhypos.init( _width, _height, _hmax1 );
	  _gpuambiguities.resize( _width, _height );

	  _census_cpu_buffer.resize( _width * _height );
	}
}

void SequentialCensusMatcher::set_params( size_t vmax, 
										  size_t hmax1, 
										  size_t hmax2, 
										  float fbf )
{
  break_sequence();
  _vmax = vmax;
  _hmax1 = hmax1;
  _hmax2 = hmax2;
  _fbf = fbf;
  _matches.resize( _height, _width, 3 * _hmax1 );
  _gpuhypos.init( _width, _height, _hmax1 );
}

void SequentialCensusMatcher::take_census_image( GPUImage & image )
{

  resize( image.getWidth(), image.getHeight() );

  std::swap( _current_entries, _last_entries );
  std::swap( _current_census_image, _last_census_image );
  _have_last = _have_current;
  _have_current = true;

  _current_census_image.reference( image.copy() );  
  _current_census_image.copyToHostMemory( (float*) & _census_cpu_buffer[0] );  

  // build a preprocessed census list ( sorted and ambiguities counted )
  build_sorted_census_list( &_current_entries[0], & _census_cpu_buffer[0], _width, _height );
  count_census_occurences( &_current_entries[0], _width * _height );
}


bool SequentialCensusMatcher::ready()
{
  return _have_last && _have_current;
}

void SequentialCensusMatcher::use_feedback( HypoImage & feedback, float p0 )
{
  do_computeHyposAmbiguous(	_gpuambiguities.data(),
							_gpuhypos.getData().data(),
							feedback.getData().data(),
							_current_census_image.data(),
							_last_census_image.data(),
							_width, _width * _height,
							_fbf, p0, _gpuhypos.getWidth() / feedback.getWidth(),
							_vmax, _hmax1, _hmax2 );  
}

void SequentialCensusMatcher::use_feedback( const std::vector<HypoFeed> & feedbacks, float p0 )
{
  do_compute_hypos_ambiguous( _gpuambiguities.data(),
							  _gpuhypos.getData().data(),
							  _current_census_image.data(),
							  _last_census_image.data(),
							  _width, _width * _height,
							  p0, feedbacks.size(), &feedbacks[0], 
							  _hmax1, _hmax2, _vmax );
}


void SequentialCensusMatcher::match_initial( float velo_power )
{
  if ( !ready() ) 
	throw std::runtime_error("sequential census matcher asked to create match"
							 " without having two images to match");
  census::match_census_lists( &_current_entries[0], &_last_entries[0], 
							  _width * _height, _matches, _hmax1, _vmax, velo_power );

  // collect hypo ambiguities
  BOOST_FOREACH( CensusClass & entry, _current_entries )
	_ambiguities( entry.iPosY, entry.iPosX, 0 ) = entry.iH;

  _gpuhypos.getData().copyFromHostMemory( _matches.data() );
  _gpuambiguities.copyFromHostMemory( _ambiguities.data() );
  
}


const CPUData<float> & SequentialCensusMatcher::initial_matches()
{
  return _matches;
}

const CPUData<float> & SequentialCensusMatcher::initial_ambiguities()
{
  return _ambiguities;
}

const HypoImage & SequentialCensusMatcher::final_matches()
{
  return _gpuhypos;
}
