/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cublas.h>
#include <cmath>
#include "GPUImage.h"
#include "ClassicUUlmNeuralModel.h"
#include "kernels.hpp"

#ifndef PI_f
#define PI_f 3.14159265359f
#endif

ClassicUUlmNeuralModel::ClassicUUlmNeuralModel(GPUVelActivities* inputPtrA,
											   float cA,
											   int   donwSampelFactorA,
											   float simga4DownsamplingA,
											   float simga4VelBlurringA,
											   float sgSpatialInhibitionV1,
											   float sgSpatialInhibitionMT)
	: downSampelFactorE  (donwSampelFactorA)
	, sigma4Downsampling (simga4DownsamplingA)
	, sigma4VelBlurring  (simga4VelBlurringA)
	, sigmaSpatialInhibitionV1(sgSpatialInhibitionV1)
	, sigmaSpatialInhibitionMT(sgSpatialInhibitionMT)
	, cE (cA)
	, tempInput1E(inputPtrA->getYRes(),inputPtrA->getXRes(),inputPtrA->getVRes(),inputPtrA->getURes())
	, tempInput2E(*inputPtrA)
	, feedbackPtrE(0)
	, tempOutputE(tempInput2E.getData(),inputPtrA->getYRes()/downSampelFactorE,inputPtrA->getXRes()/downSampelFactorE,inputPtrA->getVRes(),inputPtrA->getURes())
	, outputE(inputPtrA->getYRes()/downSampelFactorE,inputPtrA->getXRes()/downSampelFactorE,inputPtrA->getVRes(),inputPtrA->getURes())
{
	inputPtrE = inputPtrA;
	init();
}

ClassicUUlmNeuralModel::~ClassicUUlmNeuralModel()
{
	;
}

GPUVelActivities& ClassicUUlmNeuralModel::getMTOutput() {
	return outputE;
}

GPUVelActivities& ClassicUUlmNeuralModel::getV1Output() {
	return tempInput1E;
}

GPUVelActivities& ClassicUUlmNeuralModel::getMTOutputWithInputSize(bool incudePredictionShift) 
{
	if (incudePredictionShift) {
		outputE.computePredictionAndExtrapolate(tempInput1E);
	} else
		outputE.computeExtrapolation(tempInput1E);
	return tempInput1E;
}

void ClassicUUlmNeuralModel::setFeedbackInput(GPUVelActivities& f) 
{
	feedbackPtrE = &f;
}

inline static void createGaussKernel(float sigma, GPUData& dest) 
{
	size_t kSize = size_t( ceil( sigma * 4 ) );

	if ( kSize % 2 == 0 ) kSize++;
	if ( kSize    != dest.size() )
		dest = GPUData(kSize);
	fillWithGauss( dest, sigma );
}

void ClassicUUlmNeuralModel::init() 
{

	useSpaceInhibKernelV1 = sigmaSpatialInhibitionV1!=0.0f;
	useSpaceInhibKernelMT = sigmaSpatialInhibitionMT!=0.0f;

	createGaussKernel(sigma4Downsampling, spaceKernelE);
	createGaussKernel(sigma4VelBlurring, velKernelE);
	if (useSpaceInhibKernelV1)	{
		createGaussKernel(sigmaSpatialInhibitionV1, spaceInhibKernelV1);
	}
	if (useSpaceInhibKernelMT)	{
		createGaussKernel(sigmaSpatialInhibitionMT, spaceInhibKernelMT);
	}

	// creating a down sample view
	int y = tempInput1E.getYRes()/downSampelFactorE;
	int x = tempInput1E.getXRes()/downSampelFactorE;
	int u = tempInput1E.getVRes();
	int v = tempInput1E.getURes();
	size_t uv= u*v;

	int* strides = memdup(tempInput1E.getData().getStrides(),4);
	size_t * sizes   = memdup(tempInput1E.getData().getDimensions(),4);
	int* offset  = new int[4]; for (int i=0; i<4; ++i) offset[i]=0;

	sizes[0] = y;
	sizes[1] = x;

	strides[0] = strides[0]*downSampelFactorE;
	strides[1] = strides[1]*downSampelFactorE;

	GPUData downSampelView = tempInput1E.getData().view(offset,4,sizes,strides);

	tempInput1DownSampelViewE.reInit(downSampelView);

	// create a vector of ones to easily compute the sum via matrix
	// product
	sizes[0] = u*v;
	ones = GPUData(1,sizes);
	ones.fill(1.0f);//(u*v));

	onesSmall = ones.view(&uv);

	// create temp for inhibition
	sizes[0] = tempInput1E.getYRes();
	sizes[1] = tempInput1E.getXRes();

	// Create a temporary array for the spatial inhibition in V1
	tempSpatialInhibV1 = GPUData(sizes[0], sizes[1]);
	tempInhib2DViewE= GPUImage(GPUData(2,sizes));
	tempInhibE = tempInhib2DViewE.reduceDimensionality();

	sizes[0] = tempOutputE.getYRes();
	sizes[1] = tempOutputE.getXRes();

	// Create a temporary array for the spatial inhibition in MT
	tempSpatialInhibMT   = GPUData(sizes[0], sizes[1]);
	tempInhibSmall2DViewE= GPUImage(GPUData(2,sizes));
	tempInhibSmallE = tempInhibSmall2DViewE.reduceDimensionality();

	delete[] strides;
	delete[] sizes;
	delete[] offset;
}

extern "C" void inhibitionSpecial(const int Nx, const int Ny, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor);
extern "C" void inhibitionSpecial2(const int Nx, const int Ny, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor, const float divSelfInhibFactor);
extern "C" void do_fbSpecial(const int N, float* res, float* ff, float* fb, const float c);

void ClassicUUlmNeuralModel::iterateV1()
{
	if (feedbackPtrE!=0 && cE!=0.0f) { // compute feedback
//		printf("maximum in input : max=%f min=%f\n",inputPtrE->getData().max(),inputPtrE->getData().min());
//		printf("maximum in feedb : max=%f min=%f\n",feedbackPtrE->getData().max(),feedbackPtrE->getData().min());
		do_fbSpecial(tempInput1E.getData().size(),tempInput1E.getData().getDataPointer(), (float*) inputPtrE->getData().getDataPointer(),feedbackPtrE->getData().getDataPointer(),cE);
//		printf("maximum in input*(1+%f*feedb) : max=%f min=%f\n",cE,tempInput1E.getData().max(),tempInput1E.getData().min());
	} else {	// no feedback present
//		printf("maximum in input : max=%f min=%f\n",inputPtrE->getData().max(),inputPtrE->getData().min());
//		printf("maximum in feedb : max=%f min=%f\n",0,0);
		tempInput1E.getData().copyFrom(inputPtrE->getData());
//		std::cout << "no feedb : " << std::endl;
//		printf("maximum in input*(1+c*feedb) : max=%f min=%f\n",tempInput1E.getData().max(),tempInput1E.getData().min());
	}

	tempInput1E.getData().sqr();
	tempInput1E.convoluteInUAndV(velKernelE,tempInput2E);

	const int uv = tempInput1E.getURes()*tempInput1E.getVRes();

	/* just convenience views to data*/
	GPUImage in = tempInput1E.get2DView(); // first dim is space, second vel
	GPUImage xv = tempInput2E.get2DView(); // first dim is space, second vel

	/* sum the whole velocity space via matrix one-vector multiplication */
	tempInhibE.matVecMul(in,ones);

	// Add a spatial extent to the inhibition
	if (useSpaceInhibKernelV1) {
		tempInhib2DViewE.convolute1D(spaceInhibKernelV1, 0, tempSpatialInhibV1);
		tempSpatialInhibV1.convolute1D(spaceInhibKernelV1, 1, tempInhib2DViewE);
	}

	const float divConst = 0.01f;
	const float divFakto = 1.0f;
	const float divSelfFactor  = 0.0f;
	const float addSelfFactor  = 1.0f;
	const float subFactor  = 0.5f/uv;

	xv.linFirstOrderODESteayStateSolution(-1*divSelfFactor, in, -1*divFakto ,tempInhibE,-1*divConst,
											 addSelfFactor, in, -1*subFactor,tempInhibE, 0          );
	xv.rect();

	tempInput1E.getData().copyFrom(tempInput2E.getData());

//	printf("maximum in V1 after shunting: %f\n",tempInput1E.getData().max());
	/* end   of V1 */
}

void ClassicUUlmNeuralModel::iterateMT()
{
	/* start of MT*/
//	tempInput1E.getData().mul(cE);
	tempInput1E.getData().sqr();
	tempInput1E.convoluteInXAndY(spaceKernelE,tempInput2E);
	tempInput1E.convoluteInUAndV(velKernelE,tempInput2E);
	tempOutputE.getData().copyFrom(tempInput1DownSampelViewE.getData());
//	tempOutputE.convoluteInUAndV(velKernelE,outputE);

//	printf("maximum in After squaring and subsampling in MT: %f\n",outputE.getData().max());

	const int uv = tempInput1E.getURes()*tempInput1E.getVRes();

	/* just convenience views to data*/
	GPUImage in = tempOutputE.get2DView();  // first dim is space, second vel
	GPUImage xv = outputE.get2DView(); 		// first dim is space, second vel

	tempInhibSmallE.matVecMul(in,onesSmall);

	// Add a spatial extent to the inhibition
	if (useSpaceInhibKernelMT) {
		tempInhibSmall2DViewE.convolute1D(spaceInhibKernelMT, 0, tempSpatialInhibMT);
		tempSpatialInhibMT.convolute1D(spaceInhibKernelMT, 1, tempInhibSmall2DViewE);
	}

	const float divConst = 0.01f;
	const float divFakto = 1.0f;
	const float divSelfFactor  = 0.0f;
	const float addSelfFactor  = 1.0f;
	const float subFactor  = 0.5f/uv;

	xv.linFirstOrderODESteayStateSolution(-1*divSelfFactor, in, -1*divFakto ,tempInhibSmallE,-1*divConst,
											 addSelfFactor, in, -1*subFactor,tempInhibSmallE, 0          );
	xv.rect();
}



// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
