/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "ElabReichardtDetector.h"
#include <cstdio>
#include "kernels.hpp"

//#define maxLLO(a,b) (a>b?a:b)

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

std::vector<GPUImage*>* ElabReichardtDetector::createTempImgagesForPreFiltering() {
	std::vector<GPUImage*>* tempImages = new std::vector<GPUImage*>();
	for (unsigned int i=0; i<allFilters.size(); ++i) {
		tempImages->push_back( new GPUImage((this->tempImage).copy()) );
	}
	return tempImages;
}

ElabReichardtDetector::ElabReichardtDetector(int resX, int resY, int maxVel, int noOfOri)
	: gauss()
	, maxVelX(maxVel)
	, maxVelY(maxVel)
	, allFilters()
	, tempImage(resX,resY)
	, tempImage2(resX,resY)
	, tempOplus(resX,resY)
	, tempOminus(resX,resY)
	, tempImages()
{
	size_t s = (5*5); if (s%2==0) s++;
	gauss = GPUData(1,&s);
	fillWithGauss(gauss,1.0f);

	SepGaborWaveletFilter* gwFilter00 = new SepGaborWaveletFilter(resX,resY, 0, 5.0f, 0.4f);
	allFilters.push_back(gwFilter00);
	for (int i=1; i<noOfOri; ++i) {
		allFilters.push_back(new SepGaborWaveletFilter((*gwFilter00), i*M_PI/noOfOri, 5.0f, 0.4f));
	}

	for (unsigned int i=0; i<allFilters.size(); ++i) {
		switch (i) {
		case 0 : tempImages.push_back(&tempOminus); break;
		case 1 : tempImages.push_back(&tempOplus); break;
		default: tempImages.push_back(new GPUImage(resX,resY)); break;
		}
	}
}

ElabReichardtDetector::~ElabReichardtDetector() {
	for (unsigned int i=0; i<tempImages.size(); ++i) {
		if (i>1) delete tempImages[i];
		delete allFilters[i];
	}
}

void ElabReichardtDetector::preFilter(GPUImage& image,
									  std::vector<GPUImage*>& filterdImages)
{
	for (unsigned int i=0; i<allFilters.size(); ++i) {
		allFilters[i]->filterEven(image,*filterdImages[i]);
		tempImage2.copyFrom(*filterdImages[i]);
		tempImage2.abs();
		if (i==0) {
			tempImage.copyFrom(tempImage2);
		} else {
			tempImage.add(tempImage2);
		}
	}

//	GPUData* gauss = allFilters[0]->getGaussFilter();
//	tempImage.convolute1D(*gauss,1, tempImage2);
//	tempImage2.convolute1D(*gauss,2, tempImage);
//	tempImage.add(0.01f);
//	//tempImage.show();

	tempImage.convolute1D(gauss,1, tempImage2);
	tempImage2.convolute1D(gauss,2, tempImage);
	tempImage.add(0.01f);
	//tempImage.show();

	for (unsigned int i=0; i<filterdImages.size(); ++i) {
		filterdImages[i]->div(tempImage);
	}
}

extern "C" void inhibitionAndRect(const int N, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor);

void ElabReichardtDetector::correlate(std::vector<GPUImage*>& images1,
									  std::vector<GPUImage*>& images2,
									  int shiftX, int shiftY,
									  GPUImage& correlationResult)
{
	tempOminus.fill(0.0f);

	for (unsigned int i=0; i<images1.size(); ++i) {
		images1[i]->shiftMulAndAdd((*images2[i]), tempOminus,-shiftX,-shiftY);
	}

	tempOminus.convolute1D(gauss,1,tempImage);
	tempImage.convolute1D(gauss,2,tempOminus);
	tempOminus.rect();

	//

	tempOplus.fill(0.0f);

	for (unsigned int i=0; i<images1.size(); ++i) {
		images2[i]->shiftMulAndAdd(*(images1[i]), tempOplus,-shiftX,-shiftY);
	}

	tempOplus.convolute1D(gauss,1,tempImage);
	tempImage.convolute1D(gauss,2,tempOplus);
	tempOplus.rect();

	inhibitionAndRect(correlationResult.size(),correlationResult.getDataPointer(),tempOplus.getDataPointer(),tempOminus.getDataPointer(),0.75f,1.0f,1.0f);
}

void ElabReichardtDetector::correlateAll( std::vector<GPUImage*>& images1, 
										  std::vector<GPUImage*>& images2, 
										  GPUVelActivities& correlationResult )
{
	for (int dy=-maxVelY; dy<=maxVelY; ++dy) {
		for (int dx=-maxVelX; dx<=maxVelX; ++dx)  {
			correlate(images1, images2, dx, dy, tempImage2);
			//tempImage2.mul(1.0f/tempImage2.max());
			correlationResult.getOneVelocityAsImage(dx+maxVelX,dy+maxVelY).copyFrom(tempImage2);
		}
	}
//	correlationResult.killBorder(maxVelX,maxVelY);
}

void ElabReichardtDetector::correlateAll( std::vector<GPUImage*>& images1, 
										  std::vector<GPUImage*>& images2, 
										  GPUVelField& velField )
{
	const int xRes = tempImage.getWidth();
	const int yRes = tempImage.getHeight();

	if (tempImageN.size()!=tempImage.size()) {
		// initiate temp obejcts ones befor it could be used.
		tempImageN = GPUImage(xRes,yRes);
	}
	tempImageN.fill(0.0f);

	GPUImage xPart(velField.getXPart()); xPart.fill(0.0f);
	GPUImage yPart(velField.getYPart()); yPart.fill(0.0f);

	int n=0;
	for (int dy=-maxVelY; dy<=maxVelY; ++dy) {
		for (int dx=-maxVelX; dx<=maxVelX; ++dx)  {
			correlate(images1, images2, dx, dy, tempImage2);
			xPart.add(tempImage2,float(dx));
			yPart.add(tempImage2,float(dy));
			//std::cout << tempImage2.min() << " " << tempImage2.max() << std::endl;
			tempImageN.add(tempImage2);
			++n;
		}
	}
	//tempImage2.show();
	//tempImageN.show();
	xPart.div(tempImageN);
	yPart.div(tempImageN);
}

GPUImage ElabReichardtDetector::getXVelMatrix() const 
{
	const int xRes = maxVelX*2+1;
	const int yRes = maxVelY*2+1;
	const int size = xRes*yRes;

	GPUImage velX(xRes,yRes);
	float* dataX = new float[size];

	int n=0;
	const int xc = xRes/2;
	for (int y=0; y<yRes; ++y) {
		for (int x=0; x<xRes; ++x) {
			dataX[n] = float(x-xc);
			++n;
		}
	}

	velX.copyFromHostMemory(dataX);
	delete[] dataX;
	return velX;
}

GPUImage ElabReichardtDetector::getYVelMatrix() const 
{
	const int xRes = maxVelX*2+1;
	const int yRes = maxVelY*2+1;
	const int size = xRes*yRes;

	GPUImage velY(xRes,yRes);
	float* dataY = new float[size];

	int n=0;
	const int yc = yRes/2;
	for (int y=0; y<yRes; ++y) {
		for (int x=0; x<xRes; ++x) {
			dataY[n] = float(y-yc);
			++n;
		}
	}

	velY.copyFromHostMemory(dataY);
	delete[] dataY;
	return velY;
}


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
