/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * FeatureNormalization.cpp
 *
 *  Created on: Fvoid FeatureNormalization::set_sigma(float s)
{
}

eb 3, 2009
 *      Author: jbouecke
 */

#include "FeatureNormalization.hpp"

FeatureNormalization::FeatureNormalization()
	: _enabled(true)
	, _type(epsilon)
	, _features_first(true)
	, _n50(0.25)
	, _relative(true)
	, _epsilon(0.1)
	, _sigma(-1)
	, _gauss(0)
	, _floating_average(0)
{
	;
}

FeatureNormalization::~FeatureNormalization() {
	if (_gauss)
		delete _gauss;
	if (_floating_average)
		delete _floating_average;
}

void FeatureNormalization::set_features_first(bool b)
{
	_features_first = b;
}

void FeatureNormalization::use_naka_rushton(float n50, bool relative)
{
	_type = naka_rushton;
	_relative = relative;
	if (n50>0) _n50 = n50;
}

void FeatureNormalization::use_epsilon(float epsilon_)
{
	_type = epsilon;
	if (_epsilon>0) _epsilon = epsilon_;
}

void FeatureNormalization::use_sum()
{
	_type = sum;
}

void FeatureNormalization::use_max()
{
	_type = max;
}
void FeatureNormalization::use_maxsum()
{
	_type = maxsum;
}

void FeatureNormalization::set_sigma(float s)
{
	if (_gauss!=0 && _gauss->sigma()==s)
		return;
	if (_gauss) delete _gauss;
	if (s>0)
		_gauss = new GaussKernel2(s);
	else
		_gauss = 0;
}

void FeatureNormalization::use_floating_average_for_max(float lambda)
{
	if (lambda<=0.0f && _floating_average!=0) {
		delete _floating_average;
		return;
	}
	if (_floating_average!=0) {
		if (lambda!=_floating_average->getLambda())
			_floating_average->setNewLambda(lambda,false);
	} else {
		_floating_average = new FloatingAverage(lambda);
	}
}

void FeatureNormalization::init(const size_t* size) {
	if (_temp_gpudata.isEmpty() ||
		size[1]!=_temp_gpudata.getWidth() ||
		size[0]!=_temp_gpudata.getHeight())
	{
	  _temp_gpudata .resize( size[1], size[0]  );
	  _temp_gpudata2.resize( size[1], size[0]  );
	}
}

void FeatureNormalization::normalize(GPUData & d)
{
	if (!_enabled) return;
	
	GPUData data(d);
	
	if (!_features_first)
		data = data.shiftDim(-2);

	
	{ /* reduce to 3d view (feature,pixel) */
		size_t di;
		di = data.getNoOfDimensions()-2;
		data = data.reduceDimensionality(di);
		if (data.getNoOfDimensions()!=3)
			std::runtime_error("Normalize on input which could not be reduced to 2 dimensions not yet implemented");
	}

	const int   feature_count = data.extent(0);
	const size_t size[2] = {data.extent(1),data.extent(2)};

	switch(_type) {
	case(naka_rushton) :
		{
			if (_relative) {
				GPUDataException::check_error("befre max");
				float m = data.max();
				GPUDataException::check_error("after max");
				if (_floating_average) {
					if (m==0.0f) {
						m = std::numeric_limits<float>::epsilon();
					} else {
						m = _floating_average->average(m);
					}
				} else {
					if (m==0.0f) {
						m = std::numeric_limits<float>::epsilon();
					}
				}
				data.nakarushton2(_n50*m);
			} else {
				data.nakarushton2(_n50);
			}
		}
		break;
	case(max) :
		{
			float m = data.max();
			if (_floating_average) {
				if (m==0.0f) {
					m = std::numeric_limits<float>::epsilon();
				} else {
					m = _floating_average->average(m);
				}
			} else {
				if (m==0.0f) {
					m = std::numeric_limits<float>::epsilon();
				}
			}
			data.div(m);
		}
		break;
	case(sum) :
		{
			init(size);
			_temp_gpudata.copyFrom(data[0]);
			_temp_gpudata.abs();
			for (unsigned int i=1; i<(unsigned int)feature_count; ++i) {
				_temp_gpudata.add_abs(data[i]);
			}
			if (_gauss) {
				_gauss->apply(_temp_gpudata,_temp_gpudata2);
				_temp_gpudata2.swap(_temp_gpudata);
			}
			for (unsigned int i=0; i<(unsigned int)feature_count; ++i) {
				data[i].div(_temp_gpudata);
			}
		}
		break;
	case(maxsum) :
		{
			init(size);
			_temp_gpudata.copyFrom(data[0]);
			_temp_gpudata.abs();
			for (unsigned int i=1; i<(unsigned int)feature_count; ++i) {
				_temp_gpudata.add_abs(data[i]);
			}
//			if (_gauss) {
//				_gauss->apply(_temp_gpudata,_temp_gpudata2);
//				_temp_gpudata2.swap(_temp_gpudata);
//			}
			float m = _temp_gpudata.max();
			if (_floating_average) {
				if (m==0.0f) {
					m = std::numeric_limits<float>::epsilon();
				} else {
					m = _floating_average->average(m);
				}
			} else {
				if (m==0.0f) {
					m = std::numeric_limits<float>::epsilon();
				}
			}
			data.div(m);
		}
		break;
	case(epsilon) :
		{
			init(size);
			_temp_gpudata.copyFrom(data[0]);
			_temp_gpudata.abs();
			for (unsigned int i=1; i<(unsigned int)feature_count; ++i) {
				_temp_gpudata.add_abs(data[i]);
			}
			_temp_gpudata.add(_epsilon);
			
      if (_gauss) {
				_gauss->apply(_temp_gpudata,_temp_gpudata2);
				_temp_gpudata2.swap(_temp_gpudata);
			}
			
      for (unsigned int i=0; i<(unsigned int)feature_count; ++i) {
				data[i].div(_temp_gpudata);
			}
		}
		break;
	default:
		throw std::runtime_error("this normalization is not implemented");
	}
}

void FeatureNormalization::enable(bool b) {
	_enabled = b;
}


