/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * FeatureNormalization.hpp
 *
 *  Created on: Feb 3, 2009
 *      Author: jbouecke
 */

#ifndef FEATURENORMALIZATION_HPP_
#define FEATURENORMALIZATION_HPP_
#include <GPUData.h>
#include "GPUImage.h"
#include "kernels.hpp"
#include "FloatingAverage.h"
/**
 * This class presents the functionality to normalize image features.
 *
 * To work correct, the position of the image dimension has to be set correct
 * (features first or features last) via set_features_first() method.
 *
 * The normalizer support 5 different types of normalization:
 * - \e sum:  divides each pixels features through their feature sum
 * 			\f$ I_{x y i}=\frac{I_{x y i}}{\sum_j {\| {I_{x y j}}\|}} \f$
 * - \e epsilon: same as \e sum but with the local feature sum inceased by
 * 			     esilon (set via use_epsilon())
 * 			     \f$I_{x y i}=\frac{I_{x y i}}{\epsilon + \sum_j {\|{I_{x y j}}\|}}\f$
 * - \e naka_rusthon: a Naka-Rushton normalization
 * 			\f$ I_{x y i}=\frac{I_{x y i}}{\|{I_{x y i}}\|} \cdot
 *	 			\frac{I_{x y i}^2}{{n50}^2 + I_{x y i}^2} \f$
 *	 		The normalization factor \e n50 could be an absolute value or
 *	 		a value in relation to the global maximum. If the latter one is
 *	 		chosen, the real \e n50 value is \f$ n50 \cdot max_{x,y,j}{\|{I }\|} \f$.
 * - \e max: a division through a global maximum
 * - \e maxsum: a division through a global maximum of local abs feature sum
 *
 * In case of a local normalizations (\e sum, \e epsilon), the normalization
 * value (denominator) could be blured in space via a normalized Gauss function
 * (sum of gauss == 1). The sigma for this is set via set_sigma().
 *
 * The global maximum or the global abs feature maximum, used by \e max,
 * \e maxsum or the relative \e naka_rusthon normalization could be calculated
 * each normalization or averaged with a moving/floating average. This is activated by
 * use_floating_average_for_max().
 * \attention Maximum values of zero are treated special and not used for the
 * floating average calculation!
 */
class FeatureNormalization {
public:
	enum TYPE {sum,epsilon,naka_rushton,max,maxsum};

	FeatureNormalization();
	virtual ~FeatureNormalization();

	void set_features_first(bool b = true);

	void use_naka_rushton(float n50=0.125f, bool relative=true);
	void use_epsilon(float epsilon=-1);
	void use_sum();
	void use_max();
	void use_maxsum();

	/**
	 * Sets the standard deviation for the Gaussian to blur the denominator
	 * for \e sum and \e epsilon normalization.
	 *
	 * To disable the bluring, a value <=0 has to be given to this function.
	 *
	 * @param s sigma for the gaussian or a value<=0 to disable the bluring
	 */
	void set_sigma(float s);

	/**
	 * If a value > 0 is given, the maximum is averaged using a floating average
	 * (see \ref FloatingAverage)
	 * @param lambda
	 */
	void use_floating_average_for_max(float lambda=0.3f);

	/**
	 * does the normalization
	 * @param data data to normalize and also the place to store the result
	 */
	void normalize(GPUData& data);

	/**
	 * calls normalize()
	 */
	void operator() (GPUData& data) { normalize(data); }

	void enable(bool = true);	
	
private:
	bool _enabled;

	TYPE _type;
	bool  _features_first;
	float _n50;
	bool  _relative;
	float _epsilon;

	float _sigma;
	GaussKernel2* _gauss;

	FloatingAverage* _floating_average;

	GPUImage _temp_gpudata;
	GPUImage _temp_gpudata2;

	void init(const size_t* sizes);
};

#endif /* FEATURENORMALIZATION_HPP_ */
