/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * FloatingAverage.h
 *
 *  Created on: Nov 13, 2008
 *      Author: jbouecke
 */

#ifndef FLOATINGAVERAGE_H_
#define FLOATINGAVERAGE_H_

#include <cmath>

#ifdef _WIN32
#define fminf(a,b) (a < b ? a : b)
#define fmaxf(a,b) (a > b ? a : b)
#endif

/**
 * This class implements an floating average (or moving average).
 *
 * The returned average is computed in two different ways, dependent on the
 * number of already seen values. This two phase behavior has been chosen to
 * become an as much as possible stable but still correct initial average.
 *
 * Phase 1: all so far seen values were considered using the same weights
 * Phase 2: each new value is considered with \f$ \lambda \f$ and the
 * 			previous average with \f$ (1-\lambda) \f$.
 *
 * Phase 1 is used initially as long as the first given value would have
 * more impact to the floating mean than 10%. This means for e.g.
 * \f$ \lambda=0.5 \f$ the first 3 values are computed using a normal mean
 * over all so far seen values (up to 3).
 *
 * If a good mean value is known, the first phase could be skipped
 * by setting this mean value via calling setGoodMean().
 *
 * If the mean value should be restricted to a certain range, this could
 * be set by calling setMinimalMeanValue() and setMaximumMeanValue();
 * 
 * To avoid relaxation to zero without setting a fixed minimal mean value, a 
 * fraction of the total maximum seen could be set as via
 * setRelativeMinimumMeanValue();
 */
class FloatingAverage {

private:
	/**
	 * last average value
	 */
	double last_average;

	/**
	 * number of averaging steps currently proceeded
	 */
	int   counter;

	/**
	 * length of initial phase during which all past values have the same
	 * weight to the returned mean
	 */
	int   tenPercentHistory;

	/**
	 * the floating average constant \f$ \lambda \f$
	 */
	float lambda;

	/**
	 * resets the counter to zero and recalculates fivePercentHistory
	 */
	void init();

	bool minimumSet, relativeMinimumSet;
	double minimum;

	double maximalInputValue;
	
	bool maximumSet;
	double maximum;

	inline double assertRange(double d) {
		if (maximumSet) {
			d = fminf(d,maximum);
		}
		if (minimumSet) {
			d = fmaxf(d,minimum);
		} else if (relativeMinimumSet) {
			d = fminf(d,maximalInputValue*minimum);
		}
		return d;
	}

public:
	/**
	 * Creates a default floating average calculator with \f$ \lambda=0.5 \f$
	 */
	FloatingAverage();

	/**
	 * Creates a floating average calculator with the given \f$ \lambda \f$.
	 *
	 * The value should be \f$ \in ]0,1] \f$. A value of 1 would deactivate
	 * the averaging.
	 */
	explicit FloatingAverage(const float factor_);

	/**
	 * resets the alredy calulated average to unknown (counter = 0, restart
	 * with Phase 1)
	 */
	void reset();

	/**
	 * resets the average to unknown (counter = 0, restart with Phase 1) and
	 * sets a new \f$ lambda \f$.
	 */
	void reset(float lambda);

	/**
	 * Sets a new \f$ lambda \f$ but resets the average to
	 * unknown just if \a reinit is set to true.
	 */
	void setNewLambda(float lambda, bool reinit = false);

	/**
	 * Sets a good value for the mean value and thus skipp the phase 1 of
	 * mean calculation.
	 */
	void setGoodMean(double mean_);

	/**
	 * Computes the floating average:
	 * If in Phase 2:
	 * \f[ \bar {v_n} = v_n \, \lambda + \left( 1-\lambda \right) \, \bar {v_{n-1}}  \f]
	 * If in Phase 1:
	 * \f[ \bar {v_n} = {1 over n} \, \sum_{i=1}^{n}{v_{i}}  \f]
	 *
	 * The counter of so far seen values is increased by one after each call
	 * to this method.
	 */
	float average(float value);

	/**
	 * With this function you could assure that the mean value does not fall
	 * below a certain threshold.
	 * \param 	d the minimal mean which should be returned
	 */
	void setMinimumMeanValue(double d);
	
	/**
	 * Via calling this function, a fraction of the maximal (input) value seen
	 * could be set as minimal value. 
	 * @param d the fraction of maximal input under which the average should
	 * 			not fall
	 */
	void setRelativeMinimumMeanValue(double d); 

	/**
	 * With this function you could assure that the mean value does not grow
	 * above a certain threshold.
	 * \param 	d the maximal mean which should be returned
	 */
	void setMaximumMeanValue(double d);

	void unsetMinimumMeanValue();
	
	void unsetMaximumMeanValue();

	float getLambda() const { return lambda; }

	virtual ~FloatingAverage();
};

#endif /* FLOATINGAVERAGE_H_ */
