/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "GPUImage.h"
#include "HostImage.h"
#include <cmath>
#include "kernels.hpp"
#include "downsample.hpp"
#include "cufft.h"
#include "round.hpp"
#include "stdlib.h"
#include "next_power.hpp"

#define lazy_make_kernel_and_return( type, ... )	\
  static type * kernel = 0;							\
  if ( !kernel )									\
	kernel = new type ( __VA_ARGS__ );				\
  return *kernel;

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

DerivativeSimpleKernel1 & GPUImage::centralDiffKernelX()
{
  lazy_make_kernel_and_return( DerivativeSimpleKernel1 , 1 );
}

DerivativeSimpleKernel1 & GPUImage::centralDiffKernelY()
{
  lazy_make_kernel_and_return( DerivativeSimpleKernel1 , 0 );
}

Binom2Kernel1 & GPUImage::bin2KernelX()
{
  lazy_make_kernel_and_return( Binom2Kernel1 , 1 );
}

Binom2Kernel1 & GPUImage::bin2KernelY()
{
  lazy_make_kernel_and_return( Binom2Kernel1, 0 );
}

SobelOperator & GPUImage::sobelKernelX()
{
  lazy_make_kernel_and_return( SobelOperator, 1, 0 );
}

SobelOperator & GPUImage::sobelKernelY()
{
  lazy_make_kernel_and_return( SobelOperator, 0, 1 );
}

#undef lazy_make_kernel_and_return

void GPUImage::affineTransform( const CMatrix33 & inv_transmat )
{
  affineTransform( inv_transmat, *this );
}

void GPUImage::affineTransform( const CMatrix33 & inv_transmat, GPUImage & target )
{
  GPUTextureData2D texture( *this );
  texture.affineTransform( target, inv_transmat );
}

GPUImage GPUImage::scale( float xfact, float yfact )
{
  GPUImage target;
  scale( target, xfact, yfact );
  return target;
}

void GPUImage::scale( GPUImage & target, float xfact, float yfact )
{
  int w = getWidth();
  int h = getHeight();  
  target.resize( w, h );
  GPUTextureData2D texture( *this );
  texture.scale( target, xfact, yfact );
}

GPUImage::GPUImage()
	: GPUFeatureImage()
{
	;
}

GPUImage::GPUImage(const GPUData& a)
	: GPUFeatureImage(a)
{
  if (a.getDimensionsCount()!=2)
	throw GPUDataException("image can only have 2 dimensions");
}

GPUImage::GPUImage(const HostImage& a)
{
  HostImage b(a);
  b.convert_to_type( HostImage::FLOAT );
  b.convert_to_space( HostImage::LUM );
  resize( a.getWidth(), a.getHeight() );
  copyFromHostMemory( b.getPointerToImage() );
}

GPUImage::GPUImage(int xResA, int yResA)
{
  resize( xResA, yResA );
}

GPUImage& GPUImage::operator=(const GPUImage& im)
{
  reference(im);
  return *this;
}

GPUImage& GPUImage::operator=(const GPUData& im)
{
	if (im.rank()!=2)
		throw GPUDataException("image has to be 2-dimensional");
	(*(dynamic_cast<GPUData*>(this))) = im;
	return *this;
}

GPUImage::~GPUImage()
{
	;
}

GPUImage GPUImage::crop( size_t x_pos, size_t y_pos, size_t w_, size_t h_ )
{
  if ( w_ == 0 || h_ == 0 ) return GPUImage();
  const size_t w = getWidth();
  const size_t h = getHeight();
  if ( x_pos + w_ > w || y_pos + h_ > h)
	throw GPUDataException("index excedes image size in GPUImage::crop(...)");
  return range( 1, x_pos, x_pos + w_ - 1 ).range( 0, y_pos, y_pos + h_ - 1 );
}

GPUImage GPUImage::reshape( size_t new_w, size_t new_h, bool alow_new_size)
{
	const size_t  s  = size();
	const size_t ns = new_w*new_h;
	const bool c  = isContiguous();

	if (ns==s && new_w == getWidth() && new_h == getHeight()) return *this;
	if (isEmpty()) {
		if (alow_new_size)
			return GPUImage(new_w,new_h);
		else
			throw GPUDataException("reshape would change number of elements (GPUImage::reshape(...,false))");
	}
	if (!alow_new_size && ns!=s) throw GPUDataException("reshape would change number of elements (GPUImage::reshape(...,false))");
	if (c) {					// data is contiguous
		if (ns<=s)
		  {			// image will become smaller
			int offs[2] = { 0, 0 };
			int stri[2] = { new_w , 1 };
			size_t size[2] = { new_h , new_w };
			return view(offs,2,size,stri);
		  }
		else
		  {				// image will become larger

			const int offs = data() - inner->data;
			if (inner->size-offs >= ns)
			  {
				int offs[2] = { 0, 0 };
				int stri[2] = { new_w , 1 };
				size_t size[2] = { new_h , new_w };
				return view(offs,2,size,stri);
			} else {
				throw GPUDataException("new image is too large for this memory block (GPUImage::reshape(...))");
			}

		}
	} else {
		if (isContiguous(true))
			return this->reshape(new_w,new_h,alow_new_size);
		else
			throw GPUDataException("reshape view (in GPUImage::reshape(...)) not possible of this kind of memory layout");
	}
	return *this;
}

void GPUImage::resize( size_t new_w, size_t new_h )
{
  // daniel 30-12-2009: I kinda killed the choice not to reallocate, to make it simpler, and because that is what reshape is for
  size_t dims[2] = { new_h, new_w };
  GPUData::resize( 2, dims );
}

//void GPUImage::loadImage(const char* fileName) {
//	int x,y;
//	int xRes = getXRes();
//	getImageSize(fileName,x,y);
//	if (getYRes()!=y || getXRes()!=x) {
//		fprintf(stderr,"error: file resolution do not match GPUData resolution");
//	}
//	if (isContiguous()) {
//		float* dataL = new float[size()];
//		readImage(fileName,dataL,0,1,xRes);
//		copyFromHostMemory(dataL);
//		delete[] dataL;
//	} else {
//		readImage(fileName,*this);
//	}
//}
//
//void GPUImage::saveImage(const char* fileName) const {
//	int xRes = getXRes();
//	int yRes = getYRes();
//	if (isContiguous()) {
//		float* dataL = new float[size()];
//		copyToHostMemory(dataL);
//		writeImage(fileName, dataL,0,1,xRes,xRes,yRes,max(),min());
//		delete[] dataL;
//	} else {
//		writeImage(fileName,*this, xRes, yRes, max(),min());
//	}
//}
//
//void GPUImage::show() const {
//	int xRes = getXRes();
//	int yRes = getYRes();
//	if (isContiguous()) {
//		float* dataL = new float[size()];
//		copyToHostMemory(dataL);
//		showImage(dataL,0,1,xRes,xRes,yRes,max(),min());
//		delete[] dataL;
//	} else showImage(*this,xRes, yRes, max(), min());
//}

extern "C" void shiftMulAdd(float* toShift, float* toMul, float* addTo, int shiftX, int shiftY, int Nx, int incY, int Ny);

void GPUImage::shiftMulAndAdd(const GPUImage& toMulWith, GPUImage& addTo, int sx, int sy) 
{
	int xRes = getXRes();
	int yRes = getYRes();
	const size_t size = this->size();
	if (toMulWith.size()!=size || addTo.size()!=size){
		throw GPUDataException("shiftMulAndAdd within GPUImage is not allowed for images with different sizes");
	}
	const int* strides = getStrides();
	if (strides[1]!=1 || toMulWith.getStrides()[1]!=1 || addTo.getStrides()[1]!=1) {
		throw GPUDataException("shiftMulAndAdd within GPUImage is not allowed for incX!=1");
	}
	if (strides[0]!=toMulWith.getStrides()[0] || strides[0]!=addTo.getStrides()[0]) {
		throw GPUDataException("shiftMulAndAdd within GPUImage with different incY is not implemented");
	}
	shiftMulAdd(data(),toMulWith.data(),addTo.data(),sx,sy,xRes,strides[0],yRes);
}

extern "C" void doShift2D(float* dest, float* src, const int nx, const int ny, const int dx, const int dy);

void GPUImage::copyFromShifted(const GPUImage& origin, int dx, int dy) {
	if (origin.size()!=size() || origin.getXRes()!=getXRes() || !origin.isContiguous() || !isContiguous()) {
		throw GPUDataException("copyShift for this type of data not yet implented");
	}
	doShift2D(data(), origin.data(), getXRes(), getYRes(), dx, dy);
}

extern "C" void hinterp2From(float* origin, float* xC, float* yC, const int NxO, const int NyO, const int pitchO,const int NxD,const int NyD,const int pitchD, float* dest);

void GPUImage::interp2From(GPUImage& orig, GPUImage& xOrig, GPUImage& yOrig)
{
	if (getStrides()[1]==1 && orig.getStrides()[1]==1 && size()==xOrig.size() && size()==yOrig.size()) {
		int pitch = getStrides()[0];
		hinterp2From(orig.data(), xOrig.data(), yOrig.data(), orig.getWidth(), orig.getHeight(), orig.getStrides()[0], getWidth(), getHeight(), pitch, data());
	} else {
		throw GPUDataException("interp2 for this type of data not yet implented");
	}
}

extern "C" void haffineInverseWarpFrom(float* origin, const float p1, const float p2, const float p3, const float p4, const float p5, const float p6, const int NxO, const int NyO, const int pitchO,const int NxD,const int NyD,const int pitchD, float* dest, const float xCenter, const float yCenter);

void GPUImage::affineInverseWarpeFrom(GPUImage& orig, const float p1, const float p2, const float p3, const float p4, const float p5, const float p6, float xCenter, float yCenter)
{
	if (getStrides()[1]==1 && orig.getStrides()[1]==1)
	  {
		const int pitchO = orig.getStrides()[0];
		const int pitchD = getStrides()[0];
		if (xCenter == -1)
		  xCenter = orig.getWidth()/2.0f;
		if (yCenter == -1)
		  yCenter = orig.getHeight()/2.0f;
		haffineInverseWarpFrom( orig.data(),
								p1, p2, p3, p4, p5, p6,
								orig.getWidth(), orig.getHeight(), pitchO ,
								getWidth(), getHeight(), pitchD,
								data(), xCenter, yCenter );
	  }
	else
	  {
		throw GPUDataException("affineWarp for this type of data not yet implented");
	  }
}

extern "C" void haffineInverseWarpFrom2(float* origin, float* p1, float*  p2, float* p3, float* p4, float* p5, float* p6, const int NxO, const int NyO, const int pitchO,const int NxD,const int NyD,const int pitchD, float* dest, const float xCenter, const float yCenter);

void GPUImage::affineInverseWarpeFrom(GPUImage& orig, GPUImage& p1, GPUImage& p2, GPUImage& p3, GPUImage& p4, GPUImage& p5, GPUImage& p6, float xCenter, float yCenter)
{
	if (getStrides()[1]==1 && orig.getStrides()[1]==1 && size()==orig.size()) {
		const int pitchO = orig.getStrides()[0];
		const int pitchD = getStrides()[0];
		if (xCenter == -1)
			xCenter = orig.getWidth()/2.0f;
		if (yCenter == -1)
			yCenter = orig.getHeight()/2.0f;
		haffineInverseWarpFrom2(orig.data(), p1.data(),p2.data(),p3.data(),p4.data(),p5.data(),p6.data(), orig.getWidth(), orig.getHeight(), pitchO ,getWidth(),getHeight(),pitchD, data(), xCenter, yCenter);
	} else  {
		throw GPUDataException("affineWarp for this type of data not yet implented");
	}
}

//void GPUImage::copyFrom(const HostImage& im) {
//	copyFromHostMemory(im.getPointerToImage());
//}

//void GPUImage::show() const {
//	HostImage hImage(getWidth(),getHeight());
//	copyToHostMemory(hImage.getPointerToImage());
//	hImage.show();
//}


GPUImage GPUImage::diffX(bool useSobel) {
	GPUImage imT(this->getWidth(),this->getHeight());
	if (useSobel) {
	  sobelKernelX().apply(*this,imT);
	} else {
	  centralDiffKernelX().apply(*this,imT);
	}
	return imT;
}

void GPUImage::diffX(GPUImage& dest, bool useSobel) {
	if (useSobel) {
	  sobelKernelX().apply(*this,dest);
	} else {
	  centralDiffKernelX().apply(*this,dest);
	}
//	return dest;
}

void GPUImage::diffY(GPUImage& dest, bool useSobel) {
	if (useSobel) {
	  sobelKernelY().apply(*this,dest);
	} else {
	  centralDiffKernelY().apply(*this,dest);
	}
//	return dest;
}

GPUImage GPUImage::diffY(bool useSobel) {

	GPUImage imT(this->getWidth(),this->getHeight());
	if (useSobel) {
	  sobelKernelY().apply(*this,imT);
	} else {
	  centralDiffKernelY().apply(*this,imT);
	}
	return imT;
}

GPUImage GPUImage::binom2X() {
	GPUImage imT(this->getWidth(),this->getHeight());
	bin2KernelX().apply(*this,imT);
	return imT;
}

GPUImage GPUImage::binom2Y() {
	GPUImage imT(this->getWidth(),this->getHeight());
	bin2KernelY().apply(*this,imT);
	return imT;
}

GPUImage GPUImage::binom2XY() {
	GPUImage imT(this->getWidth(),this->getHeight());
	GPUImage imT2(this->getWidth(),this->getHeight());
	bin2KernelX().apply(*this,imT);
	bin2KernelY().apply(imT,imT2);
	return imT2;
}

GPUImage GPUImage::sum4PixelsAndSubsample() {
	GPUImage im(this->getWidth()/2,this->getHeight()/2);
	sum4PixelsAndSubsample(im);
	return im;
}

extern "C" void doSubsBy2(float* origin, float* dest, const int NxO, const int NyO, const int pitchO, const int pitchD);

void GPUImage::sum4PixelsAndSubsample(GPUImage& dest)
{
	const size_t w = getWidth();
	const size_t h = getHeight();

	if ( w / 2 > dest.getWidth() ||  h / 2 > dest.getHeight() )
		throw GPUDataException("error in GPUImage::sum4PixelsAndSubsample: destination is not big enought");

	if ( stride(1) != 1 || dest.stride(1) != 1 )
	  {
		throw GPUDataNYIException("GPUImage::sum4PixelsAndSubsample not jet implemented for this kind of storage");
	  }

	const int pitchO = this->getStrides()[0];
	const int pitchD = dest.getStrides()[0];

	doSubsBy2( data(), dest.data(), w, h, pitchO, pitchD);
}

GPUImage GPUImage::calcCircularVarriance(int boxNeighbourhoodRadius) {
	//printf("image : x,y=(%d,%d) min=%f ; max=%f\n",this->getWidth(),this->getHeight(),this->min(),this->max());

	GPUImage imX;
	GPUImage imY;
	GPUImage imL(this->getWidth(),this->getHeight());
	GPUImage imL2(this->getWidth(),this->getHeight());
	GPUImage imW(this->getWidth(),this->getHeight());

	GPUImage imT(this->getWidth(),this->getHeight());

	size_t size = boxNeighbourhoodRadius*2+1;
	GPUData aBoxKernel( size );
	aBoxKernel.fill(1.0f);

	// computes the derivatives in x and y
	imX = this->diffX(true);
	imY = this->diffY(true);

	// computes the norm of the gradient
	imL.pythagoras(imX,imY);

	// compute the radiant of orientations times two
	imW.atan2(imY,imX);
	imW.add(M_PI);
	imW.modulo(M_PI);
	imW.mul(2.0f);

	//imL.sqr();

	// computes the radiant
	imX.cos(imW);imX.mul(imL);
	imY.sin(imW);imY.mul(imL);

	imX.convolute1D(aBoxKernel,1,imT,false);
	imT.convolute1D(aBoxKernel,2,imX,false);// imX.show();

	imY.convolute1D(aBoxKernel,1,imT,false);
	imT.convolute1D(aBoxKernel,2,imY,false);// imY.show();

	imL.convolute1D(aBoxKernel,1,imT,false);
	imT.convolute1D(aBoxKernel,2,imL2,false);// imL2.show();

	//printf("x : min=%f ; max=%f\n",imX.min(),imX.max());
	//printf("y : min=%f ; max=%f\n",imY.min(),imY.max());
	//printf("L2: min=%f ; max=%f\n",imL2.min(),imL2.max());

	imT.pythagoras(imX,imY);
	//imX.sqr();imY.sqr();imT.copyFrom(imX); imT.add(imY); imT.sqrt();

	//printf("t vor norm : min=%f ; max=%f\n",imT.min(),imT.max());

	//imT.show();

	//imTT.copyFrom(imL2);
	//imTT.add(imT,-1.0f);

	//printf("diff : min=%f ; max=%f\n",imTT.min(),imTT.max());
	//imTT.show();
	//imTT.heavisideStepFunction(0,1);
	//printf("subtrahiert!\n");
	//imTT.show();

	imL.setIfBetween(-1,0.0001f,0,imT);
	imL.setIfBetween(-1,0.0001f,1,imL2);

	imT.div(imL2);

	//printf("t nach norm: min=%f ; max=%f\n",imT.min(),imT.max());

	//imT.show();

	imT.add(-1.0f);
	imT.mul(-1.0f);

	//imT.show();

	imL.div(imL.max());
	//imL.show();
	imT.mul(imL);

	return imT;
}

void GPUImage::extrapolate(GPUImage& dest, bool do_downsample)
{
  GPUImage tmp(*this);
  /*
  if ( dest.getWidth() < getWidth() )
	{
	  GPUImage tmp2( getWidth(), getHeight() );
	  GaussKernel1 xsmoother( float( getWidth() ) / float( dest.getWidth() ) / 2.0f, 1 );
	  xsmoother.apply(tmp,tmp2);
	  tmp.reference(tmp2);
	}
  if ( dest.getHeight() < getHeight() )
	{
	  GPUImage tmp2( getHeight(), getHeight() );
	  GaussKernel1 ysmoother( float( getHeight() ) / float( dest.getHeight() ) / 2.0f, 0 );
	  ysmoother.apply(tmp,tmp2);
	  tmp.reference(tmp2);
	}
  */

  // downsampling on gpu is broken :(
  do_downsample = false;

  if ( do_downsample && dest.getWidth() < getWidth() && dest.getHeight() < getHeight() )
	{
	  int xratio = getWidth()  / dest.getWidth();
	  int yratio = getHeight() / dest.getHeight();
	  int ratio  = std::min( xratio, yratio );
	  if ( ratio > 1 )
		{
		  GPUImage tmp2(dest);
		  downsample( tmp, tmp2, ratio, 1 );
		  tmp2.extrapolate( dest );
		  return;
		}
	}

  const float alphaX = float(dest.getWidth()-1)/(tmp.getWidth()-1.0f);
  const float alphaY = float(dest.getHeight()-1)/(tmp.getHeight()-1.0f);
  const float p1 = 1.0f/alphaX - 1.0f;
  const float p4 = 1.0f/alphaY - 1.0f;
  dest.affineInverseWarpeFrom(tmp, p1, 0.0f , 0.0f , p4, 0.0f , 0.0f , 0.0f , 0.0f );

}

void GPUImage::rotate(GPUImage& dest, float rad, float centerX, float centerY) {
	if (dest.getWidth()==getWidth() && dest.getHeight()==getHeight()) {
		const float p1 = +cosf(-rad) -1.0f;
		const float p2 = -sinf(-rad);
		const float p3 = +sinf(-rad);
		const float p4 = +cosf(-rad) -1.0f;
		dest.affineInverseWarpeFrom(*this, p1, p2 , p3 , p4, 0.0f , 0.0f , centerX, centerY);
	} else {
		const float p1 = +cosf(-rad) -1.0f;
		const float p2 = -sinf(-rad);
		const float p3 = +sinf(-rad);
		const float p4 = +cosf(-rad) -1.0f;
		const float p5 = cosf(rad)*(dest.getWidth() -getWidth() ) / -2.0f - sinf(rad)*(dest.getHeight()-getHeight()) / -2.0f;
		const float p6 = sinf(rad)*(dest.getWidth() -getWidth() ) / -2.0f + cosf(rad)*(dest.getHeight()-getHeight()) / -2.0f;
		dest.affineInverseWarpeFrom(*this, p1, p2 , p3 , p4 , p5 , p6 , centerX, centerY);
	}
}

extern "C" void conv2D(float *d_in, int in_x, int in_y, float *d_kernel, int kernel_x, int kernel_y, float *d_erg, bool neumann);
extern "C" void conv2Dstrided(float *d_in, int in_x, int in_y, int in_dx, int in_dy,
							  float *d_kernel, int kernel_x, int kernel_y,
							  float *d_erg, int out_dx, int out_dy,
							  bool neumann);

void GPUImage::convolute2D(const GPUImage& kernel, GPUImage& dest, bool neumann) const
{
  if ( getWidth() == dest.getWidth() && getHeight() == dest.getHeight() )
	{
	  if ( !kernel.isContiguous() || !kernel.isRowMajor() )
		  throw GPUDataException("GPUImage::convolute2D(..): kernel has to be contiguous and row major");
	  if ( kernel.getWidth()%2==0 || kernel.getHeight()%2==0 )
		  throw GPUDataException("GPUImage::convolute2D(..): kernel has to have odd widht and hight");
	  if ( !isContiguous() || !dest.isContiguous() || isRowMajor() != dest.isRowMajor() )
		{
		  std::cout << " strided convolution " << std::endl;
		  conv2Dstrided( data(), extent(1), extent(0), stride(1), stride(0),
						 kernel.data(), kernel.getWidth(), kernel.getHeight(),
						 dest.data(), dest.stride(1), dest.stride(0),
						 neumann);
		}
	  else  // fast path works if both data are contiguous and have the same index ordering
		{
//		  std::cout << " direct convolution " << std::endl;
		  if ( isRowMajor() )
			conv2D( data(), getWidth(), getHeight(),
					kernel.data(), kernel.getWidth(), kernel.getHeight(),
					dest.data(), neumann);
		  else // by switching w/h we get row major again
			conv2D( data(), getHeight(), getWidth(),
					kernel.data(), kernel.getWidth(), kernel.getHeight(),
					dest.data(), neumann);
		}
	} else {
	throw GPUDataException("GPUImage::convolute2D(..) : src and dest does not have the same size ");
  }
}

void GPUImage::killBorder(int xBorder, int yBorder, float val) {
	int offs[2] = {0,0};
	size_t size[2] = {yBorder,getWidth()};
	this->view(offs,size).fill(val);

	offs[0] = getHeight()-yBorder;
	offs[1] = 0;
	this->view(offs,size).fill(val);

	offs[0] = yBorder;
	offs[1] = 0;
	size[0] = getHeight()-2*yBorder;
	size[1] = xBorder;
	this->view(offs,size).fill(val);

	offs[0] = yBorder;
	offs[1] = getWidth()-xBorder;
	size[0] = getHeight()-2*yBorder;
	size[1] = xBorder;
	this->view(offs,size).fill(val);
}

void GPUImage::fft2(GPUHermRedFourierImage& resul) const 
{
	if (!isContiguous())
		throw GPUDataNYIException("fft of non contiguous data not yet implemented");
    
	const int w = getWidth();
	const int h = getHeight();
    
	cufftResult   cufft_return;
	//std::cout << "fft with "<< h << " , "<< w << "  idata:"<< dump_dims() << "  odata:"<< resul.dump_dims()<< std::endl;
    cufftHandle*  plan = get_fft_plan( plan_key(CUFFT_R2C,h,w) );
    
    //cufftHandle* plan;
    //cufft_return = cufftPlan2d( plan, h, w, CUFFT_R2C );
    //if (cufft_return!=CUFFT_SUCCESS)
	//	throw GPUDataException(str(boost::format("error allocating handle in GPUImage::fft2(), error = %d")%cufft_return));
        
	cufftReal	 *idata = (cufftReal*)    data();
	cufftComplex *odata = (cufftComplex*) resul.data();

    
    /* Use the CUFFT plan to transform the signal out of place. */
	cufft_return = cufftExecR2C(*plan, idata, odata);
	if (cufft_return!=CUFFT_SUCCESS)
		throw GPUDataException("error executin cufft in GPUImage::fft2()");
    //cufftDestroy(*plan);
    
}


GPUHermRedFourierImage GPUImage::fft2() const {
	if (!isContiguous())
		throw GPUDataNYIException("fft of non contiguous data not yet implemented");

	const int w = getWidth();
	const int h = getHeight();

	GPUHermRedFourierImage resul(w,h);

	fft2(resul);

	return resul;
}

void GPUImage::ifft2From(const GPUHermRedFourierImage& spec)
{

	if ( spec.extent(0) != getHeight() || spec.getOriginalWidth() != getWidth() || spec.extent(2) != 2 )
		throw GPUDataException("GPUImage::ifft(GPUHermRedFourierImage& spec) sizes do not match");

	const size_t w = getWidth();
	const size_t h = getHeight();

//	GPUData resul(h,w,2);

	cufftResult   cufft_return;
	cufftHandle*  plan = get_fft_plan( plan_key(CUFFT_C2R,h,w) );

	float 		 *odata = data();
	cufftComplex *idata = (cufftComplex*) ((GPUHermRedFourierImage*)&spec)->data();

	/* Use the CUFFT plan to transform the signal out of place. */
	cufft_return = cufftExecC2R(*plan, idata, odata);
	if (cufft_return!=CUFFT_SUCCESS)
		throw GPUDataException("error executin cufft in GPUImage::fft()");

	div(float(w*h));
}

void GPUImage::ifft2Shift() {
	if (!this->isContiguous() || !this->isRowMajor())
		throw GPUDataNYIException("GPUImage::ifft2Shift() for this memory layout not yet implemented");

	const int w = this->getWidth();
	const int h = this->getHeight();
	GPUImage temp( w, h );
	{ /* switch left right */
		int offse_right_read[2]  = {0,w/2};
		int offse_right_write[2] = {0,(w+1)/2};
		int offse_left[2]  = {0,0};
		size_t sizes_right[2] = {h,(w+1)/2};
		size_t sizes_left[2]  = {h,w/2};
		temp.view(offse_left,sizes_right).copyFrom(this->view(offse_right_read,sizes_right));
		temp.view(offse_right_write,sizes_left).copyFrom(this->view(offse_left,sizes_left));
	}
	{ /* switch top down */
		int offse_down_read[2]  = {h/2,0};
		int offse_down_write[2] = {(h+1)/2,0};
		int offse_up[2]   = {0,0};
		size_t sizes_down[2] = {(h+1)/2,w};
		size_t sizes_up[2]   = {h/2,w};
		this->view(offse_up,sizes_down).copyFrom(temp.view(offse_down_read,sizes_down));
		this->view(offse_down_write,sizes_up).copyFrom(temp.view(offse_up,sizes_up));
	}
}

GPUImage GPUImage::locate_template( GPUImage & templ, bool normalize )
{
  GPUImage T( templ.copy() ), I( this->copy() ), I_padded( this->extent(1) * 2, this->extent(0) * 2 );

  // center template
  T.sub( T.mean() );
  //I.sub( I.mean() );

  // pad image
  I_padded.fill( 0.0f );
  I_padded .range( 0, 0, this->extent(0) - 1 )
	       .range( 1, 0, this->extent(1) - 1 )
 	       .copyFrom( I );
  GPUHermRedFourierImage fI_padded( I_padded.fft2() );
    
  // pad template
  GPUImage T_padded( I_padded.extent(1), I_padded.extent(0) );
  int off1 = std::max( 0, int( T.extent(0) - I.extent(0) + 1 ) / 2 );
  int off2 = std::max( 0, int( T.extent(1) - I.extent(1) + 1 ) / 2 );
  int n1 = T.extent(0) - 2 * off1, n2 = T.extent(1) - 2 * off2;
  int N = n1 * n2;
  T_padded.fill( 0.0f );
  int n11 = n1 / 2, n21 = n2 / 2;
  int n12 = float( n1 ) / 2 + 0.5f, n22 = float( n2 ) / 2 + 0.5f;
  int m1 = T_padded.extent(0), m2 = T_padded.extent(1);
  // lower right
  T_padded.range( 0, m1 - n11, m1 - 1 ).range( 1, m2 - n21, m2 - 1 )
	.copyFrom( T.range( 0, off1, off1 + n11 - 1 ).range( 1, off2, off2 + n21 - 1 )  );
  // upper right
  T_padded.range( 0, m1 - n11, m1 - 1 ).range( 1, 0, n22 - 1 )
	.copyFrom( T.range( 0, off1, off1 + n11 - 1 ).range( 1, off2 + n21, off2 + n2 - 1 )  );
  // lower left
  T_padded.range( 0, 0, n12 - 1 ).range( 1, m2 - n21, m2 - 1 )
	.copyFrom( T.range( 0, off1 + n11, off1 + n1 - 1 ).range( 1, off2, off2 + n21 - 1 )  );
  // upper left
  T_padded.range( 0, 0, n12 - 1 ).range( 1, 0, n22 - 1 )
	.copyFrom( T.range( 0, off1 + n11, off1 + n1 - 1 ).range( 1, off2 + n21, off2 + n2 - 1 )  );
  
  // fft transform template, stencil, image, and squared image
  GPUHermRedFourierImage fT_padded( T_padded.fft2() );
  fT_padded.getImagPartView().mul(-1.0f); // reverse the template in fft domain so we do correlation not convolution

  // cross-correlate Image with centered template
  GPUHermRedFourierImage fL_padded = fI_padded.copy();
  fL_padded.mulWithComplexFilter( fT_padded );
  GPUImage L_padded = fL_padded.ifft2();
  
  if ( normalize )
	{
      GPUDataException::check_error( "before normalize" );
      // create template-shaped stencil for local image normalization
	  GPUImage Ones_padded( I_padded.extent(1), I_padded.extent(0)  );
	  Ones_padded.fill( 0.0f );
	  // upper left
	  Ones_padded.range( 0, 0, n11 - 1 )
		.range( 1, 0, n21 - 1 )
		.fill(1);
	  // upper right
	  Ones_padded.range( 0, m1 - n12, m1 - 1 )
		.range( 1, 0, n21 - 1 )
		.fill(1);
	  // lower left
	  Ones_padded.range( 0, 0, n11 - 1 )
		.range( 1, m2 - n22, m2 - 1 )
		.fill(1);
	  // lower right
	  Ones_padded.range( 0, m1 - n12, m1 - 1 )
		.range( 1, m2 - n22, m2 - 1 )
		.fill(1);
	  GPUHermRedFourierImage fOnes_padded( Ones_padded.fft2() );
	  
	  // calculate local centered euclidian norm of the image

	  GPUHermRedFourierImage fI_local_mean_padded = fI_padded.copy();
	  fI_local_mean_padded.mulWithComplexFilter( fOnes_padded );
	  GPUImage I_local_mean_padded = fI_local_mean_padded.ifft2().mul( 1.0f / N );

	  GPUHermRedFourierImage fI2_padded = GPUImage( I_padded.copy().sqr() ).fft2();
	  fI2_padded.mulWithComplexFilter( fOnes_padded );
	  GPUImage I2_local_padded = fI2_padded.ifft2();

	  GPUImage Z = I2_local_padded.sub( I_local_mean_padded.sqr().mul( N ) ).rect().sqrt(); // destroys local mean and local squared mean
	  
	  // calculate centered euclidian norm of the template
	  float ZT =  std::sqrt( T.sqr().sum() ); // destroys T

	  // total norm is product of norms, add epsilon to avoid div by zero
	  Z.mul( ZT );
	  Z.add( Z.mean() * 0.0000001f );

	  L_padded.div( Z );
	}
  
  GPUImage L = L_padded.range( 0, 0, this->extent(0) - 1 ).range( 1, 0, this->extent(1) - 1 ).copy();

  return L;

}



extern "C" void do_zensus(const int w, const int h, const float* input, float* output, float eps);
GPUImage GPUImage::getZensusTransformedImage(const float eps) const {
	GPUImage rtv(getWidth(),getHeight());
	return getZensusTransformedImage(rtv,eps);
}
GPUImage GPUImage::getZensusTransformedImage(GPUImage& rtv, const float eps) const {
	rtv.resize(getWidth(),getHeight());
	
	do_zensus(getWidth(),getHeight(),data(),rtv.data(),eps);
	
	GPUDataException::check_error("do_zensus");
	return rtv;
}

