/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef GPUIMAGE_H_
#define GPUIMAGE_H_
class GPUImage;

#include "HostImage.h"
#include <iostream>
#include <GPUData.h>
#include <GPUFeatureImage.hpp>
#include "GPUHermRedFourierImage.hpp"

#ifndef __kernels_hpp_
class DerivativeSimpleKernel1;
class Binom2Kernel1;
class SobelOperator;
#endif

/**
 * This class extends GPUData object to receive an cleaner interface for
 * 2-dimensional fields of floating point values (gray scale images)
 *
 * It restricts the referenced data layout to 2 dimensions and provides
 * special
 */
class GPUImage : public GPUFeatureImage
{
private:
	
	DerivativeSimpleKernel1 & centralDiffKernelX();
	DerivativeSimpleKernel1 & centralDiffKernelY();
	Binom2Kernel1 & bin2KernelX();
	Binom2Kernel1 & bin2KernelY();
	SobelOperator & sobelKernelX();
	SobelOperator & sobelKernelY();

public:

	GPUImage();
	GPUImage(const GPUData& a);
	GPUImage(const HostImage& a);
	GPUImage(int xRes, int yRes);
	GPUImage& operator=(const GPUImage& im);
	GPUImage& operator=(const GPUData& im);
	virtual ~GPUImage();

	/// get a view to a subimage
	GPUImage crop( size_t x_pos, size_t y_pos, size_t w, size_t h);
	/// change the size of this object
	GPUImage reshape( size_t new_w,  size_t new_h, bool alow_new_size = false);
	void resize( size_t new_w, size_t new_h );

	inline size_t getXRes() const { return extent(1); }
	inline size_t getYRes() const { return extent(0); }

	inline size_t getWidth() const  { return getXRes(); }
	inline size_t getHeight() const { return getYRes(); }

	/// performs linear exprapolation, optionally downsamples first if target is much smaller
	void extrapolate(GPUImage& dest, bool downsample = false );
	void rotate(GPUImage& dest, float rad, float centerX=-1, float centerY=-1);

	GPUImage scale( float xfact, float yfact );
	void scale( GPUImage & target, float xfact, float yfact );

	GPUImage diffX(bool useSobel=false);
	GPUImage diffY(bool useSobel=false);
	void diffX(GPUImage& dest, bool useSobel=false);
	void diffY(GPUImage& dest, bool useSobel=false);

	GPUImage binom2X();
	GPUImage binom2Y();
	GPUImage binom2XY();

	void killBorder(int xBorder, int yBorder, float value = 0.0f);

	GPUImage sum4PixelsAndSubsample();
	void sum4PixelsAndSubsample(GPUImage& dest);

	GPUImage locate_template( GPUImage & templ, bool normalize = false );


	/**
	 *
	 * @param toMulWith
	 * @param addTo
	 * @param sx (means read from this(x-sx , y-sy))
	 * @param sy
	 */
	void shiftMulAndAdd(const GPUImage& toMulWith, GPUImage& addTo, int sx, int sy);

	/**
	 *
	 * @param origin
	 * @param dx means read from this(x-dx,y-dy)
	 * @param dy
	 */
	void copyFromShifted(const GPUImage& origin, int dx, int dy);

	GPUImage calcCircularVarriance(int boxNeighbourhoodRadius=2);

	void interp2From(GPUImage& orig, GPUImage& xOrig, GPUImage& yOrig);

	/**
	 *  											( x )
	 * 						  (1+p1   p3    p5 ) 	(	)
	 * original coordinate =  (                )  * ( y )
	 * 						  ( p2   1+p4   p6 )	( 	)
	 * 												( 1 )
	 */
	void affineInverseWarpeFrom(GPUImage& orig, const float p1, const float p2, const float p3, const float p4, const float p5, const float p6, float xCenter = 0, float yCenter = 0);

	/**
	 *  											( x )
	 * 						  (1+p1   p3    p5 ) 	(	)
	 * original coordinate =  (                )  * ( y )
	 * 						  ( p2   1+p4   p6 )	( 	)
	 * 												( 1 )
	 */
	void affineInverseWarpeFrom(GPUImage& orig, GPUImage& p1, GPUImage& p2, GPUImage& p3, GPUImage& p4, GPUImage& p5, GPUImage& p6, float xCenter = 0, float yCenter = 0);

	typedef float CMatrix33[3][3];

	void affineTransform( const CMatrix33 & inv_transmat );
	void affineTransform( const CMatrix33 & inv_transmat, GPUImage & target );

	void convolute2D(const GPUImage& kernel, GPUImage& dest, bool neumann = true) const;

	GPUHermRedFourierImage fft2() const;

	void fft2(GPUHermRedFourierImage& rtv) const;

	void ifft2From(const GPUHermRedFourierImage& spec);

	void ifft2Shift();
	
	GPUImage getZensusTransformedImage(const float eps = 0.01f) const;
	
	GPUImage getZensusTransformedImage(GPUImage& rtv, const float eps = 0.01f) const;
};

#include "kernels.hpp"

#endif /*GPUIMAGE_H_*/


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
