/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef GPUVELACT_H_
#define GPUVELACT_H_

#include "GPUImage.h"
#include "GPUVelField.h"

#include <vector>

/**
 * This class represents a simple wrapper to a large GPUData object which
 * contains the whole velocity activities.
 * Therein the velocities are organized in 2d fashion.
 * The normal storage is organized as follows:
 * 	i=(1)*u+(sizeU)*v+(sizeU*sizeV)*x+(sizeU*sizeV*sizeX)*y
 */
class GPUVelActivities
{
private:
	GPUData realNDO;
	bool owner;

public:
	GPUVelActivities();
	GPUVelActivities(GPUData data);
	GPUVelActivities(int y, int x, int v, int u);
	GPUVelActivities(GPUData& data, int y, int x, int v, int u);
	virtual ~GPUVelActivities();

	void reInit(GPUData& data);

	bool isImageContiguous();
	bool isVelSpaceContiguous();
	
	GPUData& getData();
	const GPUData& getData() const;

	inline size_t getXRes() const { return realNDO.extent(1); }
	inline size_t getYRes() const { return realNDO.extent(0); }
	inline size_t getURes() const { return realNDO.extent(3); }
	inline size_t getVRes() const { return realNDO.extent(2); }

	void convoluteInXAndY(GPUData& kernel, GPUVelActivities& temp);
	void convoluteInUAndV(GPUData& kernel, GPUVelActivities& temp);
	void convoluteInU(GPUData& kernel, GPUVelActivities& dest, bool cyclic = true);
	void convoluteInV(GPUData& kernel, GPUVelActivities& dest);

	void mulToVelSpace(GPUImage& values);
	void addToVelSpace(GPUImage& values, const float f = 0.001f);

	void shutOfBadPositions(const float maxThreashold=0.00075f);
	void norm_to_sum(const float epsilon=0.000001f);

	GPUImage getOneVelocitySpacesAsImage(size_t x, size_t y);
	GPUImage getOneVelocityAsImage(size_t u, size_t v);

	GPUImage getVelocitySpacesInOneImage(bool normalize=false, int xPos=-1, int yPos=-1, int xSize=-1, int ySize=-1, int xStep=-1, int yStep=-1);
	void 	 getVelocitySpacesInOneImage(GPUImage& dest, bool normalize=false, int xPos=-1, int yPos=-1, int xSize=-1, int ySize=-1, int xStep=-1, int yStep=-1);

	GPUVelField getFlowField();
	void getFlowField(GPUVelField& dest);
	void getFlowField(GPUVelField& dest, float relative_hsum_threshold, float unknown_value=1e10);
//	void getFlowFieldSlice(GPUVelField& dest, int xPos=-1, int yPos=-1, int xSize=-1, int ySize=-1);

	void computePrediction();
	void computePredictionAndExtrapolate(GPUVelActivities& dest);
	void computeExtrapolation(GPUVelActivities& dest);

	void killBorder(int borderSizeX, int borderSizeY);

	/**
	 * this function returns a 2-d view to the data which holds the velocities
	 * in the second (fastes/deepest) dimensions and the position in the other
	 * dimension.
	 */
	GPUImage get2DView() const;

	GPUImage getFeatureSum();
private:
	void copyAllVelocitySpacesIntoOneImage(GPUImage& im);
	void copyAllVelocitySpacesIntoOneImage(int xPos, int xSize, int yPos, int ySize, GPUImage& im, bool normalize=false);
	void copyAllVelocitySpacesIntoOneImage(int xPos, int xStep, int xSize, int yPos, int yStep, int ySize, GPUImage& im, bool normalize=false);

	GPUImage xValues;
	GPUImage yValues;
	GPUData  onesSizeVelSpace;
	GPUImage& getXValues();
	GPUImage& getYValues();
};

#endif /*THE4DIMS_H_*/
