/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "GPUVelField.h"
#include <boost/format.hpp>

int sgn(int x){
  return (x > 0) ? 1 : (x < 0) ? -1 : 0;
}

inline void bresenham(int xstart,int xend,int ystart, int yend, int xSize, int ySize , float* data)
{
	int dx = xend - xstart;
	int dy = yend - ystart;

	int incx = sgn(dx);
	int incy = sgn(dy);

	if(dx<0) dx = -1*dx;
	if(dy<0) dy = -1*dy;

	int pdx,pdy,ddx,ddy,es,el;
    if (dx>dy) {
        pdx=incx;
        pdy=0;
        ddx=incx;
        ddy=incy;
        es =dy;
        el =dx;
    } else {
        pdx=0;
        pdy=incy;
        ddx=incx;
        ddy=incy;
        es =dx;
        el =dy;
    }

    int x = xstart;
    int y = ystart;
    int err = el/2;
    data[x+y*xSize] = 1.0f;

    for(int t=0; t<el; ++t) {
        err -= es;
        if(err<0) {
             err += el;
             x += ddx;
             y += ddy;
        } else {
             x += pdx;
             y += pdy;
        }
        if (x<xSize && y<ySize && x>0 && y>0)
        	data[x+y*xSize] = 1.0f;
    }
}

GPUVelField::GPUVelField()
	: useCartesianInput(true)
{
  realNDO = GPUData(2,1,1);
  //realNDO.fill(0.0f);
  
  xPart   = GPUImage(realNDO[0]);
  yPart   = GPUImage(realNDO[1]);
}

GPUVelField::GPUVelField(const GPUVelField& d)
	: useCartesianInput(d.useCartesianInput)
{
	realNDO = d.realNDO;
	xPart   = GPUImage(realNDO[0]);
	yPart   = GPUImage(realNDO[1]);
}

GPUVelField::GPUVelField(int x, int y)
	: useCartesianInput(true)
{
  realNDO = GPUData(2,y,x);
	//realNDO.fill(0.0f);

	xPart   = GPUImage(realNDO[0]);
	yPart   = GPUImage(realNDO[1]);
}

GPUVelField::GPUVelField(size_t x, size_t y)
	: useCartesianInput(true)
{
  realNDO = GPUData(2,y,x);
  //realNDO.fill(0.0f);  
  xPart   = GPUImage(realNDO[0]);
  yPart   = GPUImage(realNDO[1]);
}


GPUVelField::GPUVelField( const GPUData & data, bool cartesian )
{
  if ( data.rank() != 3 || data.extent(0) != 2 )
	throw GPUDataException( str( boost::format( "cannot construct GPUVelfield from: %s" ) % data.dump_dims() ) );
  useCartesianInput = cartesian;
  realNDO.reference( data );
  realNDO.makeContiguous( true );
  xPart   = GPUImage(realNDO[0]);
  yPart   = GPUImage(realNDO[1]);  
}

GPUVelField::~GPUVelField()
{
	;
}

GPUVelField GPUVelField::operator=(const GPUVelField& d) 
{
  useCartesianInput = d.useCartesianInput;
  realNDO = d.realNDO;
  xPart   = GPUImage(realNDO[0]);
  yPart   = GPUImage(realNDO[1]);
  return *this;
}

const GPUImage GPUVelField::getXPart()const { return xPart;}

const GPUImage GPUVelField::getYPart() const { return yPart; }

GPUImage GPUVelField::getXPart()  {	return xPart;}

GPUImage GPUVelField::getYPart() { return yPart; }

//void GPUVelField::show(HostImage& hi,int subs) {
//
//	if (hi.getHeigth()!=sizeY ||
//		hi.getWidth() !=sizeX    )
//	{
//		HostImage hiNew(sizeX,sizeY);
//
//		const int sx = hi.getWidth();
//		const int sy = hi.getHeigth();
//
//		if (sx*sy < sizeX*sizeY) {
//			throw GPUDataException("interpolation of image for showing of flow filed not yet implemented");
//		}
//
//		const int dx = sx/sizeX;
//		const int dy = sy/sizeY;
//
//		float* dataOrigin = hi.getPointerToImage();
//		float* dataDestin = hiNew.getPointerToImage();
//
//		int n=0;
//		for (int y=0; y<sy; y += dy)
//		for (int x=0; x<sx; x += dx) {
//			dataDestin[n] = dataOrigin[y*sx+x];
//			n++;
//		}
//
//		show(hiNew, subs);
//	} else {
//
//		const int sx = hi.getWidth();
//		const int sy = hi.getHeigth();
//
//		int rad = 4;
//		if (subs>=2) rad = 2;
//		if (subs>=4) rad = 1;
//
//		HostImage im(sizeX*(2*rad+1),sizeY*(2*rad+1));
//
//		int n=0;
//		float* resData = im.getPointerToImage();
//		float* imaData = hi.getPointerToImage();
//		float maxV = imaData[0];
//		//for (int i=0; i<x*y; ++i) resData[i] = 0.0f;
//		for (int yy=0; yy<sy; yy++)
//		for (int xx=0; xx<sx; xx++) {
//			for (int yyy=0; yyy<(2*rad+1); yyy++)
//			for (int xxx=0; xxx<(2*rad+1); xxx++) {
//				resData[xxx+yyy*(2*rad+1)*sx+xx*(2*rad+1)+yy*(2*rad+1)*sx*(2*rad+1)] = imaData[n];
//			}
//			maxV = maxV<imaData[n] ? imaData[n] : maxV;
//			n++;
//		}
//
//		draw(rad,1,subs,im);
//
//		im.show();
//
//	}
//}
//
//void GPUVelField::draw(const int rad, const float lengthFactor, const int subs, HostImage& im) {
//
//	const float factor = rad*2.0f*lengthFactor;
//	const int x = sizeX;
//	const int y = sizeY;
//
//	float* resData = im.getPointerToImage();
//
//	float* xData = new float[sizeX*sizeY]; getXPart().copyToHostMemory(xData);
//	float* yData = new float[sizeX*sizeY]; getYPart().copyToHostMemory(yData);
//
//	int dx,dy;
//	for (int yy=0; yy<sizeY; yy+=subs)
//	for (int xx=0; xx<sizeX; xx+=subs) {
//		dx = (int)(factor*xData[xx+sizeX*yy]);
//		dy = (int)(factor*yData[xx+sizeX*yy]);
//		bresenham(xx*(2*rad+1)+rad+1,xx*(2*rad+1)+rad+1+1,yy*(2*rad+1)+rad+1,yy*(2*rad+1)+rad+1+1,x*(2*rad+1),y*(2*rad+1),resData);
//		bresenham(xx*(2*rad+1)+rad+1,xx*(2*rad+1)+rad+1-1,yy*(2*rad+1)+rad+1,yy*(2*rad+1)+rad+1+1,x*(2*rad+1),y*(2*rad+1),resData);
//		bresenham(xx*(2*rad+1)+rad+1,xx*(2*rad+1)+rad+1+1,yy*(2*rad+1)+rad+1,yy*(2*rad+1)+rad+1-1,x*(2*rad+1),y*(2*rad+1),resData);
//		bresenham(xx*(2*rad+1)+rad+1,xx*(2*rad+1)+rad+1-1,yy*(2*rad+1)+rad+1,yy*(2*rad+1)+rad+1-1,x*(2*rad+1),y*(2*rad+1),resData);
//
//		bresenham(xx*(2*rad+1)+rad+1,xx*(2*rad+1)+rad+1+dx,yy*(2*rad+1)+rad+1,yy*(2*rad+1)+rad+1+dy,x*(2*rad+1),y*(2*rad+1),resData);
//	}
//
//	delete[] xData;
//	delete[] yData;
//}
//
//
//void GPUVelField::show(int subs) {
//	HostImage hi(sizeX,sizeY);
//	hi.initWithZeros();
//	show(hi,subs);
//}

GPUImage GPUVelField::getSpeed() const
{
  GPUImage speed( getWidth(), getHeight() );

  speed.pythagoras(getXPart(),getYPart());
  
  return speed;
}

void GPUVelField::getSpeed(GPUImage& dest) const
{
  if ( dest.getWidth() != getWidth() || dest.getHeight() != getHeight() )
	dest = GPUImage( getWidth(), getHeight() );
  dest.pythagoras(getXPart(),getYPart());
}

GPUImage GPUVelField::getDirection() const
{
	GPUImage dir( getWidth(), getHeight() );

	dir.atan2(xPart,yPart);

	return dir;
}

void GPUVelField::getDirection(GPUImage& dest) const
{
  if ( dest.getWidth() != getWidth() || dest.getHeight() != getHeight() )
	dest = GPUImage(getWidth(), getHeight() );

  dest.atan2(xPart,yPart);

}

GPUData GPUVelField::getAsFeatureImage() const 
{
	size_t dims[3];
	dims[0]=getHeight();
	dims[1]=getWidth();
	dims[2]=2;
	GPUData fid(3,dims);

	int dimStrides[2];
	dimStrides[0] = getWidth()*2;
	dimStrides[1] = 2;

	int offs[3];
	offs[0]=0;
	offs[1]=0;

	offs[2]=0;
	GPUData xView(fid.view(offs,2,dims,dimStrides));
	xView.copyFrom(xPart);

	offs[2]=1;
	GPUData yView(fid.view(offs,2,dims,dimStrides));
	yView.copyFrom(yPart);
	return fid;
}

GPUData GPUVelField::getAsFeatureImageView() 
{
	size_t dims[3];
	dims[0]=getHeight();
	dims[1]=getWidth();
	dims[2]=2;

	int offs[3];
	offs[0] = 0;
	offs[1] = 0;
	offs[2] = 0;

	int dimStrides[3];
	dimStrides[0] = getWidth();
	dimStrides[1] = 1;
	dimStrides[2] = getWidth()*getHeight();

	return realNDO.view(offs,3,dims,dimStrides);
}

float GPUVelField::getMaxSpeed() {
	GPUImage temp1(getWidth(),getHeight());
	GPUImage temp2(getWidth(),getHeight());

	temp1.copyFrom(xPart); temp1.mul(xPart);
	temp2.copyFrom(yPart); temp2.mul(yPart);

	temp1.add(temp2);
	temp1.sqrt();
	return temp1.max();
}

GPUVelField GPUVelField::copy() {
	GPUVelField rtv(getWidth(),getHeight());
	rtv.useCartesianInput = this->useCartesianInput;
	rtv.getData().copyFrom(this->getData());
	return rtv;
}

#include <cstdio>
// Mehtod from Pierre Bayerl (adapted by Jan Bouecke)
void GPUVelField::loadpcm(const char* name)
{
	FILE *file;
	int val[2];
	char magic[3] = {0,0,0};
	int val_c=0;
	int i,n;
	float maxval;
	float *im;
	char line[4097];
	char c;
	int w;
	int h;

	file = fopen(name,"rb");
	if (file==0) {
		printf("file not found");
		exit(1);
	}

	fread(magic,1,2,file);
	while(val_c<2 && !feof(file)) {
		c = fgetc(file);
		if (c=='#') {
			fgets(line,4096,file);
		}
		else if (c>='0' && c<='9') {
			ungetc(c,file);
			fscanf(file,"%d",&val[val_c]);
			val_c++;
		}
	}
	while(val_c<3 && !feof(file)) {
		c = fgetc(file);
		if (c=='#') {
			fgets(line,4096,file);
		}
		else if ((c>='0' && c<='9') || c=='-'|| c=='.') {
			ungetc(c,file);
			fscanf(file,"%f",&maxval);
			val_c++;
		}
	}
	while(1) {
		c = fgetc(file);
		if (c=='\n') break;
	}

	w=val[0];
	h=val[1];

	im = (float*)calloc(2*w*h,sizeof(float));

	if (strcmp(magic,"PC")==0) {
		n=(int)fread((unsigned char*)im,1,2*w*h*sizeof(float),file);
		if (n!=2*w*h*(int)sizeof(float))
		  {
			printf("warning, n=%d 2*w*h=%d -- unsufficient data, filling up with zeros (w=%d h=%d).\n",n,int(2*w*h*sizeof(float)),w,h);
			while (n<2*w*h*(int)sizeof(float))
			  {
				im[n/(int)sizeof(float)]=0;
				n+=(int)sizeof(float);
			  }
			printf("error reading pcm file: %s\n",name);
		  }
//#ifndef __LITTLE_ENDIAN
//		for (i=0;i<2*w*h;i++) {
//			swap_real_endian(&im[i]);
//		}
//#endif
		if (xPart.size()!=size_t(w*h) || !xPart.isContiguous() ||
			yPart.size()!=size_t(w*h) || !yPart.isContiguous())
		{
			printf("error !! dimensions of read flow field does not match dimension of GPUVelField object");
			return;
		}

		float* xd = new float[w*h];
		float* yd = new float[w*h];

		for (i=0;i<w*h;i++)
		  {
			xd[i] = (float)im[2*i+0];
			yd[i] = (float)im[2*i+1];
		  }

		xPart.copyFromHostMemory(xd);
		yPart.copyFromHostMemory(yd);

		delete[] xd;
		delete[] yd;
	}
	else {
		printf("unknown format, only binary PCM file allowed.");
		exit(1);
	}
	fclose(file);
	free(im);
}

// Mehtod from Pierre Bayerl (adapted by Jan Bouecke)
void GPUVelField::savepcm(const char* name) {
	int w = getWidth();
	int h = getHeight();
	float *image_r = new float[w*h];
	float *image_i = new float[w*h];
	xPart.copyToHostMemory(image_r);
	yPart.copyToHostMemory(image_i);
	FILE *file;
	int i;
	float *im;
	float maxval = 0;
	maxval = getMaxSpeed();
	im = (float*)calloc(2*w*h,sizeof(float));
	for (i=0;i<w*h;i++) {
		im[2*i+0] = (float)image_r[i];
		im[2*i+1] = (float)image_i[i];
//#ifndef __LITTLE_ENDIAN
//		swap_real_endian(&im[2*i+0]);
//		swap_real_endian(&im[2*i+1]);
//#endif
	}
	file = fopen(name,"wb");
	if (file==0) {
		printf("cannot write file (savepcm).");
		exit(1);
	}
	fprintf(file,"PC\n%d %d\n%f\n",w,h,maxval);
	fwrite(im,1,2*w*h*sizeof(float),file);
	fclose(file);
	free(im);
	delete [] image_r;
	delete [] image_i;
}
