/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * GaborFeatureExtractor.cpp
 *
 *  Created on: Sep 24, 2008
 *      Author: jbouecke
 */

#include "GaborFeatureExtractor.h"
#include "SepGaborWaveletFilter.h"
#include <boost/format.hpp>

#ifndef PI_f
#define PI_f 3.14159265359f
#endif

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

GaborFeatureExtractor::GaborFeatureExtractor( unsigned image_w, unsigned image_h, 
											  type_t typeA, 
											  unsigned no_of_ori, unsigned no_of_scales, 
											  float lambdaA, float sigma_per_lambda, float scale_faktor, 
											  bool specialRotationShift )
	:	type(typeA)
	,	imageW(image_w)
	,	imageH(image_h)
	,	noOfOri(no_of_ori)
	,	noOfScales(no_of_scales)
	,	scaleFactor(scale_faktor)
	,	lambda(lambdaA)
	,	sigmaPerLambda(sigma_per_lambda)
	,	filters()
	,	useNakaRushton(false)
	,	n50(0.05f)
	,	useScaleBasedNormalizationE(false)
{
	float rat = sigma_per_lambda;
	float offs= specialRotationShift ? -0.5f : 0.0f;
	for ( size_t s=0; s<noOfScales; ++s) {
		std::vector<SepGaborWaveletFilter> filterL;
		lambdaA = lambda;
		for ( size_t i=0; i<s; ++i) lambdaA *= scale_faktor;
		for ( size_t d=0; d<noOfOri; ++d) {
			if (d==0 && s==0)
				filterL.push_back(SepGaborWaveletFilter(imageW,imageH, (0+offs)*M_PI/noOfOri, lambdaA, rat));
			else if (d==0)
				filterL.push_back(SepGaborWaveletFilter(filters[0][0], (d+offs)*M_PI/noOfOri, lambdaA, rat));
			else
				filterL.push_back(SepGaborWaveletFilter(filterL[0]   , (d+offs)*M_PI/noOfOri, lambdaA, rat));
		}
		filters.push_back(filterL);
	}
}

void GaborFeatureExtractor::setUseDCCompensation(bool b) {

	for ( size_t s=0; s<noOfScales; ++s) {
		for ( size_t d=0; d<noOfOri; ++d) {
			filters[s][d].setUseDCCompensation(b);
		}
	}
}
bool GaborFeatureExtractor::getUseDCCompensation() {
	return filters[0][0].getUseDCCompensation();
}

GaborFeatureExtractor::~GaborFeatureExtractor() {
	;
}

void GaborFeatureExtractor::useScaleBasedNormalization(const bool doIt) {
	this->useScaleBasedNormalizationE = doIt;
}

void GaborFeatureExtractor::useNakaRushtonNormalization(const float n50A) {
	if (n50A<=0.0f) {
		useNakaRushton=false;
		return;
	} else {
		useNakaRushton=true;
		n50 = n50A;
		return;
	}
}

//std::vector<GaborFilter2DParameter> GaborFeatureExtractor::getParameters() {
//	//
//}

int GaborFeatureExtractor::getNoOfFeatures() const
{
	int faktor = 1;
	switch (type) {
	case (SIN_AND_COS_AND_minusSIN) 	: faktor = 3; break;
	case (SIN_AND_minusSIN)	:
	case (SIN_AND_COS) 	: faktor = 2; break;
	case (SIN) 			:
	case (COS) 			:
	case (ENERGY)		: faktor = 1; break;
	default:	throw std::runtime_error("error in GaborFeatureExtractor::getNoOfFeatures(): type not specified\n");
	}
	return noOfScales*noOfOri * faktor;
}

std::vector<GPUImage> GaborFeatureExtractor::getFeatureImagesAsVector( GPUImage& input )
{
	if ( input.getWidth() != imageW || input.getHeight() != imageH) 
	  {
		throw std::runtime_error("error in GaborFeatureExtractor::getFeatureImages(const GPUImage&): image size does not match\n");
	  }
	std::vector<GPUImage> rtv;
	const size_t N = getNoOfFeatures();
	for ( size_t i=0; i<N; ++i)
		rtv.push_back(GPUImage(imageW,imageH));

	getFeatureImages(input,rtv);
	return rtv;
}

GPUData GaborFeatureExtractor::getFeatureImages( GPUImage& input )
{
	if (input.getWidth()!=imageW || input.getHeight()!=imageH) {
	  throw std::runtime_error("error in GaborFeatureExtractor::getFeatureImages(const GPUImage&): image size does not match\n");
	}
	const size_t N = getNoOfFeatures();
	size_t dims[3];
	dims[0] = N;
	dims[1] = imageH;
	dims[2] = imageW;
	GPUData rtv(3,dims);

	getFeatureImages(input,rtv);

	return rtv;
}

int GaborFeatureExtractor::getNoOfChannels()
{
  switch (type) {
  case SIN_AND_COS_AND_minusSIN : return 3;
  case SIN_AND_minusSIN : return 2;
  case SIN_AND_COS	: return 2;
  case SIN	: return 1;
  case COS	: return 1;
  case ENERGY	: return 1;
  default: throw std::runtime_error("GaborFeatureExtractor: invalid type");
  }
}

extern "C" void do_nakarushton2(const int N, float* res, float* input,  const float eps, const float maxim);
extern "C" void do_nakarushton2PreservingSign(const int N, float* res, float* input,  const float eps, const float maxim);

void GaborFeatureExtractor::getFeatureImages( GPUImage& input, GPUData& rtv )
{

	size_t shape[5] = { getNoOfChannels(), getNoOfScales(), getNoOfOrientations(), imageH, imageW };
	rtv.resize( 5, shape );

	switch (type) {
	case (SIN_AND_COS_AND_minusSIN)	:
		for ( size_t s=0; s<noOfScales; ++s)
		  {
			for ( size_t d=0; d<noOfOri; ++d)
			  {
  			    GPUImage resE ( rtv[0][s][d] );
  			    GPUImage resO ( rtv[1][s][d] );
  			    GPUImage resO2( rtv[2][s][d] );

				filters[s][d].filter(input,resE,resO);
				resO2.fill(0.0f); resO2.add(resO,-1.0f);
			}
		}
		break;
	case (SIN_AND_COS)	:
		for ( size_t s=0; s<noOfScales; ++s)
		  {
			for ( size_t d=0; d<noOfOri; ++d)
			  {
  			    GPUImage resE ( rtv[0][s][d] );
  			    GPUImage resO ( rtv[1][s][d] );

				filters[s][d].filter(input,resE,resO);
			}
		}
		break;
	case (SIN_AND_minusSIN)	:
		for ( size_t s=0; s<noOfScales; ++s)
		  {
			for ( size_t d=0; d<noOfOri; ++d)
			  {
  			    GPUImage resO ( rtv[0][s][d] );
  			    GPUImage resO2( rtv[1][s][d] );

				filters[s][d].filterOdd(input,resO);
				resO2.fill(0.0f); resO2.add(resO,-1.0f);
			}
		}
		break;
	case (COS) :
		for ( size_t s=0; s<noOfScales; ++s)
		  {
			for ( size_t d=0; d<noOfOri; ++d)
			  {
  			    GPUImage resE ( rtv[0][s][d] );

				filters[s][d].filterEven( input, resE );
			}
		}
		break;
	case (SIN) :
		for ( size_t s=0; s<noOfScales; ++s)
		  {
			for ( size_t d=0; d<noOfOri; ++d)
			  {
  			    GPUImage resO ( rtv[0][s][d] );

				filters[s][d].filterOdd(input, resO);
			}
		}
		break;
	case (ENERGY) :
	  for ( size_t s=0; s<noOfScales; ++s)
		{
			for ( size_t d=0; d<noOfOri; ++d)
			  {
  			    GPUImage resE ( rtv[0][s][d] );

				resE.copyFrom( filters[s][d].getEnergy(input) );
			}
		}
		break;
	default:
	  throw std::runtime_error( "error in GaborFeatureExtractor::getFeatureImages(): type not specified\n" );
	}


	for ( int c = 0; c < getNoOfChannels(); ++c )
	  {
		float sf = 1;
		GPUData datac( rtv[c] );
		for ( int s = 0;  s < noOfScales; ++s, sf *= scaleFactor )
		  {
			GPUData datas( datac[s] );
			for ( size_t d=0; d<noOfOri; ++d)
			  {
  			    GPUImage im ( datas[d] );
				if (useScaleBasedNormalizationE)
					im.div(sf);
				if (useNakaRushton)
					do_nakarushton2PreservingSign( imageW * imageH, im.getDataPointer(), im.getDataPointer(), n50, 1.0f );
			  }
		  }
	  }
}

void GaborFeatureExtractor::getFeatureImages( GPUImage& input, std::vector<GPUImage>& rtv)
{
    GPUData rdata;
	getFeatureImages( input, rdata );
	const size_t  N = getNoOfFeatures();

	rtv.clear();
	rtv.resize(N);

	int i = 0;
	for ( int c = 0; c < getNoOfChannels(); ++c )
	  {
		GPUData datac( rtv[c] );
		for ( int s = 0;  s < noOfScales; ++s )
		  {
			GPUData datas( datac[s] );
			for ( int d = 0; d < noOfOri; ++d )
			  {
				rtv[i++].reference( datas[d] );
			  }
		  }
	  }

}

GPUImage GaborFeatureExtractor::getReconstructedImage(std::vector<GPUImage>& input) {
	GPUImage rtv(imageW,imageH);
	getReconstructedImage(input,rtv);
	return rtv;
}

GPUImage GaborFeatureExtractor::getReconstructedImage(GPUData& input) {
	const size_t si = input.getDimensions()[0];
	std::vector<GPUImage> input_as_vector;
	for ( size_t i=0; i<si; ++i) {
		input_as_vector.push_back(input[i]);
	}
	GPUImage rtv(imageW,imageH);
	getReconstructedImage(input_as_vector,rtv);
	return rtv;
}

void GaborFeatureExtractor::getReconstructedImage(GPUData& input, GPUImage& rtv) {
	const size_t si = input.getDimensions()[0];
	std::vector<GPUImage> input_as_vector;
	for ( size_t i=0; i<si; ++i) {
		input_as_vector.push_back(input[i]);
	}
	getReconstructedImage(input_as_vector,rtv);
}

void GaborFeatureExtractor::getReconstructedImage(std::vector<GPUImage>& input, GPUImage& result) {
	std::vector<GPUImage>::const_iterator feature = input.begin();
	result.fill(0.0f);
	GPUImage temp(imageW,imageH);
	switch (type) {
	case (SIN_AND_COS)	:
		for ( size_t s=0; s<noOfScales; ++s) {
			for ( size_t d=0; d<noOfOri; ++d) {
				result.add(*feature);
				feature++;
				result.add(*feature);
				feature++;
			}
		}
		break;
	case (SIN_AND_minusSIN)	:
		for ( size_t s=0; s<noOfScales; ++s) {
			for ( size_t d=0; d<noOfOri; ++d) {
				temp.copyFrom(*feature);
				temp.rect();
				result.add(temp);
				feature++;
				temp.copyFrom(*feature);
				temp.rect();
				result.add(temp);
				feature++;
			}
		}
		break;
	case (SIN_AND_COS_AND_minusSIN)	:
		for ( size_t s=0; s<noOfScales; ++s) {
			for ( size_t d=0; d<noOfOri; ++d) {
				temp.copyFrom(*feature);
				temp.rect();
				result.add(temp);
				feature++;
				temp.copyFrom(*feature);
				temp.rect();
				result.add(temp);
				feature++;
				temp.copyFrom(*feature);
				temp.rect();
				result.add(temp);
				feature++;
			}
		}
		break;
	case (SIN) :
	case (COS) :
	case (ENERGY) :
		for ( size_t s=0; s<noOfScales; ++s) {
			for ( size_t d=0; d<noOfOri; ++d) {
				if (d==0 && s==0)
					result.copyFrom(*feature);
				else
					result.add(*feature);
				feature++;
			}
		}
		break;
	default:
	  throw std::runtime_error("error in GaborFeatureExtractor::getReconstructedImage(std::vector<GPUImage>& , GPUImage& ):"
							   " type not specified\n");
	}
}
