/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * GaborFeatureExtractor.h
 *
 *  Created on: Sep 24, 2008
 *      Author: jbouecke
 */

#ifndef GABORFEATUREEXTRACTOR_H_
#define GABORFEATUREEXTRACTOR_H_

#include "FeatureExtractor.h"
#include "SepGaborWaveletFilter.h"

class GaborFeatureExtractor: public FeatureExtractor {

public:
	enum type_t {
		SIN_AND_COS,
		SIN_AND_COS_AND_minusSIN,
		COS,
		SIN,
		SIN_AND_minusSIN,
		ENERGY
	};

private:

	type_t type;

	unsigned imageW;
	unsigned imageH;

	unsigned 	noOfOri;
	unsigned 	noOfScales;
	float	scaleFactor;
	float 	lambda;
	float 	sigmaPerLambda;

	std::vector< std::vector<SepGaborWaveletFilter> > filters;

	bool 	useNakaRushton;
	float 	n50;

	bool	useScaleBasedNormalizationE;

public:

	GaborFeatureExtractor( unsigned image_w, unsigned image_h, 
						   type_t type = SIN_AND_COS, 
						   unsigned no_of_ori = 8, 
						   unsigned no_of_scales = 3, 
						   float lambda = 5.0f, 
						   float sigma_per_lambda = 0.3f, 
						   float scale_faktor = 2.0f , 
						   bool specialRotationShift = false );
	virtual ~GaborFeatureExtractor();

	void useNakaRushtonNormalization(const float n50A);
	void useScaleBasedNormalization(const bool doIt);

	/// returns the number of channels (like sin, cos, energy, minusSin)
	int getNoOfChannels();
	/**
	 * Returns the number of features this features
	 * at each position.
	 */
	virtual int getNoOfFeatures() const;

	virtual int getNoOfScales() const {
		return noOfScales;
	}

	virtual int getNoOfOrientations() const {
		return noOfOri;
	}
	
	void setUseDCCompensation(bool b);
	bool getUseDCCompensation();

//	std::vector<GaborFilter2DParameter> getParameters();

	/**
	 * creates for each feature a floating point values image and
	 * returns these images in a vector of
	 * <code> getNoOfFeatures() </code> length.
	 */
	virtual std::vector<GPUImage> getFeatureImagesAsVector( GPUImage& input );

	/**
	 * creates for each feature a floating point values image and
	 * returns these images packed into one 3-d GPUData object thus that
	 * GPUData[i] returns the 2 dimensional image
	 */
	virtual GPUData getFeatureImages( GPUImage& input );

	/**
	 * the same like getFeatureImages(const GPUImage& input) but as in place
	 * function, currently inefficient because data in result is not reused
	 */
	virtual void getFeatureImages( GPUImage& input, std::vector<GPUImage>& result );

	/**
	 * the same like getFeatureImagesAsGPUData(const GPUImage& input) but as in
	 * place function, currently the most efficient if result is already the right size
	 */
	virtual void getFeatureImages( GPUImage& input, GPUData& result );

	/**
	 * recconstructs an image out of its features
	 */
	virtual GPUImage getReconstructedImage(std::vector<GPUImage>& input);

	/**
	 * recconstructs an image out of its features
	 */
	virtual GPUImage getReconstructedImage(GPUData& input);

	/**
	 * same like getReconstructedImage(GPUData& input) but as in place
	 * function
	 */
	virtual void getReconstructedImage(GPUData& input, GPUImage& result);

	/**
	 * same like getReconstructedImage(std::vector<GPUImage>& input) but as in place
	 * function
	 */
	virtual void getReconstructedImage(std::vector<GPUImage>& input, GPUImage& result);
};

#endif /* GABORFEATUREEXTRACTOR_H_ */
