/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "GroupingViaGaussianFilter.hpp"

static GPUImage ifft2Shift(GPUImage input) {
	const int w = input.getWidth();
	const int h = input.getHeight();
	GPUImage temp( w, h );
	GPUImage rtv ( w, h );
	{ /* inverse copy from right to left */
		int offse_from[2] = {0,w/2+1};
		int sizes_from[2] = {h,(w-1)/2};
		int offse_to[2] = {0,0};
		temp.view(offse_from,sizes_from).copyFrom(input.view(offse_to,sizes_from));
	}
	{ /* inverse copy from left to right */
		int offse_from[2] = {0,0};
		int sizes_from[2] = {h,w/2+1};
		int offse_to[2] = {0,(w-1)/2};
		temp.view(offse_from,sizes_from).copyFrom(input.view(offse_to,sizes_from));
	}
	{ /* inverse copy from bottom to top */
		int offse_from[2] = {h/2+1,0};
		int sizes_from[2] = {(h-1)/2,w};
		int offse_to[2] = {0,0};
		rtv.view(offse_from,sizes_from).copyFrom(temp.view(offse_to,sizes_from));
	}
	{ /* inverse copy from top to bottom */
		int offse_from[2] = {0,0};
		int sizes_from[2] = {h/2+1,w};
		int offse_to[2] = {(h-1)/2,0};
		rtv.view(offse_from,sizes_from).copyFrom(temp.view(offse_to,sizes_from));
	}
	return rtv;
}

GroupingViaGaussianFilter::GroupingViaGaussianFilter(
							int   noOfOri_,
							float sigma_small_,
							float sigma_large_)
	: noOfOri(noOfOri_)
	, sigma_large(sigma_large_)
	, sigma_small(sigma_small_)
	, w(-1)
	, h(-1)
{
	;
}

void GroupingViaGaussianFilter::init(int w_, int h_) {
	if (w!=w_ || h!=h_) {
		w=w_;h=h_;
		filterSpecs.clear();
		GPUImage image(w,h);
		GPUHermRedFourierImage filter(w,h);
		for (int i=0; i<noOfOri; ++i) {
			const double phi = (1*M_PI*i)/noOfOri;
			fillWithGauss2Rotated(image,sigma_large,sigma_small,phi);
			ifft2Shift(image).fft2(filter);
			filterSpecs.push_back(filter);
		}
	}
}

void GroupingViaGaussianFilter::getFeatureImages(const GPUImage& input, GPUData& result) {
	const int w_ = input.getWidth();
	const int h_ = input.getHeight();
	init(w_,h_);

	GPUHermRedFourierImage inputSpec = input.fft2();
	GPUHermRedFourierImage tempSpec(w,h);;
	GPUImage			   tempImage;

	const int n = getNoOfFeatures();

	std::vector<GPUHermRedFourierImage>::iterator it = filterSpecs.begin();
	for (int i=0; i<n; ++i) {
		tempSpec.copyFrom(inputSpec);
		tempSpec.mulWithComplexFilter(*it);
		tempImage=result[i];
		tempSpec.ifft2(tempImage);
	}
}

void GroupingViaGaussianFilter::getFeatureImage(const GPUImage& input, int i, GPUImage& result) {
	const int w_ = input.getWidth();
	const int h_ = input.getHeight();
	init(w_,h_);

	GPUHermRedFourierImage inputSpec = input.fft2();

	inputSpec.mulWithComplexFilter(filterSpecs[i]);

	result.ifft2From(inputSpec);
}

GPUData GroupingViaGaussianFilter::getFeatureImages(const GPUImage& input) {
	const int w_ = input.getWidth();
	const int h_ = input.getHeight();
	init(w_,h_);
	GPUData rtv(getNoOfFeatures(),h,w);
	getFeatureImages(input,rtv);
	return rtv;
}

GPUImage GroupingViaGaussianFilter::getFeatureImage(const GPUImage& input, int i) {
	const int w_ = input.getWidth();
	const int h_ = input.getHeight();
	init(w_,h_);
	GPUImage rtv(w,h);
	getFeatureImage(input,i,rtv);
	return rtv;
}

GPUImage GroupingViaGaussianFilter::getReconstructedImage(GPUData& input) {
	if (input.getDimensionsCount()!=3)
		throw GPUDataException("input to GroupingViaGaussian::getReconstructedImage() has to be 3-dimensional");

	GPUImage result(input.extent(2),input.extent(1));
	getReconstructedImage(input,result);
	return result;
}

void GroupingViaGaussianFilter::getReconstructedImage(GPUData& input, GPUImage& result) {
	const int n = getNoOfFeatures();
	for (int i=0; i<n; ++i) {
		if (i==0)
			result.copyFrom(input[i]);
		else
			result.add(input[i]);
	}
}

