/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "HDF5FileAdapter.h"

#include <boost/shared_ptr.hpp>

HDF5FileAdapter::HDF5FileAdapter(const char* fileName, bool overwrite)
	: HDF5MotionFile(fileName,overwrite)
{
	;
}

HDF5FileAdapter::~HDF5FileAdapter()
{
	//this->~HDF5MotionFile();
}

void HDF5FileAdapter::loadDataSet(const char* ds_name) {
	HDF5MotionFile::loadDataSet(ds_name);
}

GPUVelField HDF5FileAdapter::readVelocityField(const char* ds_name) {
	if (ds_name!=0)
		loadDataSet(ds_name);
	if (getDataType()!=HDF5MotionFile::VELOCITY) {
		fprintf(stderr,"error within HDF5FileAdapter::readVelocityField()\n try to load flow field out of non flow field data set\n");
		exit(-1);
	}
	int  ndims = getDataDimension();
	size_t* sizes = getDataSizes();

	if (ndims!=2) {
		fprintf(stderr,"error within HDF5FileAdapter::readVelocityField()\n try to load flow field with more or less than 2 dims\n");
		exit(-1);
	}

	GPUVelField rtv(sizes[1],sizes[0]);

	readVelocityField(rtv);

	delete[] sizes;
	return rtv;
}

void HDF5FileAdapter::readVelocityField(GPUVelField& res, const char* ds_name) {
	if (ds_name!=0)
		loadDataSet(ds_name);
	if (getDataType()!=HDF5MotionFile::VELOCITY) {
		fprintf(stderr,"error within HDF5FileAdapter::readVelocityField()\n try to load flow field out of non flow field data set\n");
		exit(-1);
	}
	int  ndims = getDataDimension();
	size_t* sizes = getDataSizes();

	if (ndims!=2) {
		fprintf(stderr,"error within HDF5FileAdapter::readVelocityField()\n try to load flow field with more or less than 2 dims\n");
		exit(-1);
	}

	int size=1;
	for (int i=0; i<ndims; ++i) {
		size *= sizes[i];
	}

	if (res.getWidth()!=sizes[1] || res.getHeight()!=sizes[0] ) {
		fprintf(stderr,"error within HDF5FileAdapter::readVelocityField()\n size of given return value does not match to data\n");
		exit(-1);
	}

	GPUData rtvDataAsFeatureImage(res.getAsFeatureImage());

	Velocity* hData = new Velocity[size];
	readDataSet((float*)hData);
	rtvDataAsFeatureImage.copyFromHostMemory((float*)hData);

	res.getAsFeatureImageView().copyFrom(rtvDataAsFeatureImage);

	delete[] sizes;
	delete[] hData;
}


void HDF5FileAdapter::writeVelocityField(const char* ds_name, GPUVelField& data) {
	size_t* dimSizes = new size_t[2];
	dimSizes[0] = data.getHeight();
	dimSizes[1] = data.getWidth();

	Velocity* hData = new Velocity[dimSizes[0]*dimSizes[1]];
	data.getAsFeatureImage().copyToHostMemory((float*)hData);

	writeFlowField(ds_name,	hData, 2, dimSizes);

	delete[] dimSizes;
	delete[] hData;
}


void HDF5FileAdapter::writeVelocityActivities(const char* ds_name, GPUVelActivities& data) {
	size_t* dimSizes = new size_t[4];
	dimSizes[0] = data.getYRes();
	dimSizes[1] = data.getXRes();
	dimSizes[2] = data.getVRes();
	dimSizes[3] = data.getURes();

	float* hData = new float[dimSizes[0]*dimSizes[1]*dimSizes[2]*dimSizes[3]];
	data.getData().copyToHostMemory(hData);

	writeFullVelSpaceActivities(ds_name,hData,2,dimSizes,0,dimSizes[2]*dimSizes[3]);

	delete[] dimSizes;
	delete[] hData;
}

GPUImage HDF5FileAdapter::readGrayImage(const char* ds_name) {
	if (ds_name!=0)
		loadDataSet(ds_name);
	if (getDataType()!=HDF5MotionFile::VELOCITY) {
		fprintf(stderr,"error within readGrayImage::readGrayImage(const char*)\n try to load gray image out of non gray image data set (DataType is %d)\n", getDataType());
		exit(-1);
	}
	int  ndims = getDataDimension();
	size_t* sizes = getDataSizes();

	if (ndims!=2) {
		fprintf(stderr,"error within readGrayImage::readGrayImage(const char*)\n try to load image with more or less than 2 dims (%d)\n", ndims);
		exit(-1);
	}

	int size=1;
	for (int i=0; i<ndims; ++i) {
		size *= sizes[i];
	}
	GPUImage rtv(sizes[1],sizes[0]);

	readGrayImage(rtv);

	delete[] sizes;
	return rtv;
}

void HDF5FileAdapter::readGrayImage(GPUImage& res, const char* ds_name) {
	if (ds_name!=0)
		loadDataSet(ds_name);
	if (getDataType()!=HDF5MotionFile::VELOCITY) {
		fprintf(stderr,"error within readGrayImage::readGrayImage(GPUImage&, const char*)\n try to load gray image out of non gray image data set\n");
		exit(-1);
	}
	int  ndims = getDataDimension();
	size_t* sizes = getDataSizes();

	if (ndims!=2) {
		fprintf(stderr,"error within readGrayImage::readGrayImage(GPUImage&, const char*)\n try to load image with more or less than 2 dims\n");
		exit(-1);
	}

	int size=1;
	for (int i=0; i<ndims; ++i) {
		size *= sizes[i];
	}

	if (res.getWidth()!=sizes[1] || res.getHeight()!=sizes[0] ) {
		fprintf(stderr,"error within readGrayImage::readGrayImage(GPUImage&, const char*)\n size of given return value does not match to data\n");
		exit(-1);
	}

	//float* hData = new float[size];
  boost::shared_ptr<float> hData;
  hData.reset( new float[size] );
	readDataSet( hData.get() );
	res.copyFromHostMemory( hData.get() );
	delete[] sizes;
	hData.reset();//delete[] hData;

}

void HDF5FileAdapter::readGrayImage(HostImage& res, const char* ds_name) {
	if (ds_name!=0)
		loadDataSet(ds_name);
	if (getDataType()!=HDF5MotionFile::VELOCITY) {
		fprintf(stderr,"error within readGrayImage::readGrayImage(HostImage&, const char*)\n try to load gray image out of non gray image data set\n");
		exit(-1);
	}
	int  ndims = getDataDimension();
	size_t* sizes = getDataSizes();

	if (ndims!=2) {
		fprintf(stderr,"error within readGrayImage::readGrayImage(HostImage&, const char*)\n try to load image with more or less than 2 dims\n");
		exit(-1);
	}

	int size=1;
	for (int i=0; i<ndims; ++i) {
		size *= sizes[i];
	}

	res.resize(sizes[1],sizes[0]);

	readDataSet(res.getPointerToImage());

	delete[] sizes;
}

void HDF5FileAdapter::writeGrayImage(const char* ds_name, GPUImage& data) {
	const int h = data.getHeight();
	const int w = data.getWidth();

	float* hData = new float[w*h];
	data.copyToHostMemory((float*)hData);

	writeGrayScaleImage(ds_name, hData, w,h);

	delete[] hData;
}
