/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * HypoImage.cpp
 *  
 *  Created on: July 2009
 *      Author: jbouecke, cbeck
 */

#include <stdio.h>
#include "HypoImage.h"
#include <CensusClass.h>
#include <algorithm>
#include <functional>

#include "sorthypos.hcu"
#include "subsample.hcu"

#include <boost/format.hpp>

#ifdef _WIN32
#include <vector>
#endif

class HypoClass 
{

public:
  
  float w;
  float vx;
  float vy;
  
  HypoClass(void)
  {
	w = 0;
	vx = 0;
	vy = 0;
  }
  
  ~HypoClass(void)
  {
  }
  
  HypoClass& operator=(const HypoClass &im)
  {
	w = im.w;
	vx = im.vx;
	vy = im.vy;
	return *this;
  }
  
  bool operator< ( const HypoClass & other ) const
  {
	return w < other.w;
  }

  bool operator==( const HypoClass & other ) const
  {
	return other.vx == vx && other.vy == vy;
  }

};

//for greater - std::sort
bool UDgreaterHyposW(HypoClass elem1, HypoClass elem2)
{
	return elem1.w > elem2.w;
}

//for greater - std::sort
struct UDgreaterHyposVel
{
  UDgreaterHyposVel( int vmax ) : VEL_MAX(vmax) {}
  const int VEL_MAX;
  unsigned int vel2index( int vx, int vy )
  {
	return ( vx + VEL_MAX + ( vy + VEL_MAX ) * ( 2 * VEL_MAX + 1 ) );	
  }
  bool operator()( const HypoClass & elem1, const HypoClass & elem2)
  {
	return vel2index( elem1.vx, elem1.vy ) > vel2index( elem2.vx, elem2.vy );
  }
};

///////////////////////////////////////////////////////////

void HypoImage::init(int w, int h, int max_h) 
{

	this->h_max = max_h;
	this->w = w;
	this->h = h;

	GPUData dataL(h,w,h_max,3);

	this->data = dataL;

	//TODO: ist slice (dim, index) oder (index,dim)
	size_t dim0 = 0;
	weightView = GPUImage( data.slice(3,0).reduceDimensionality( dim0 ) );

}

HypoImage::HypoImage( )
  : w(0), h(0), h_max(0)
{
}

HypoImage::HypoImage( const GPUData & data)
{
  if ( data.rank() == 4 && data.extent(3) == 3 )
	{
	  size_t dim0 = 0;
	  this->w = data.extent(1);
	  this->h = data.extent(0);
	  this->h_max = data.extent(2);
	  this->data.reference( data );
	  this->weightView = GPUImage(data.slice(3,0).reduceDimensionality(dim0));
	}
}

HypoImage::HypoImage(int w, int h, int max_h) 
{
	init (w,h,max_h);
}

HypoImage::~HypoImage() 
{

}

void HypoImage::sqrWeights() 
{
  data.slice(3,0).sqr();
}

GPUImage HypoImage::getHypothesisProperty(int i, int v) 
{
  if ( i >= 0 && i < h_max && v >= 0 && v < 3 ) 
	{
	  GPUImage weightViewL = GPUImage(data.slice(3,v).slice(2,i));
	  return weightViewL;
	} 
  else 
	{
	  throw std::runtime_error("wrong hypothesis index");
	}
}

GPUData HypoImage::getHypothesis(int x_pos, int y_pos) 
{
	return data[y_pos][x_pos];
}

void HypoImage::sortHypos() 
{
  do_sortHypos_simple( data.getDataPointer(), 
					   (unsigned int) ( w * h ), 
					   (unsigned int) h_max, 
					   (unsigned int) ( h_max * 3 ), 
					   (unsigned int) 3         
					   );
}

void HypoImage::setHypo( int i, GPUImage & xVel, GPUImage & yVel, GPUImage & confi ) 
{
  getHypothesisProperty(i,0).copyFrom(confi);
  getHypothesisProperty(i,1).copyFrom(xVel);
  getHypothesisProperty(i,2).copyFrom(yVel);
}

void HypoImage::setHypo(int i, GPUVelField& velField, GPUImage& confi) {
	GPUImage xx(velField.getXPart());
	GPUImage yy(velField.getYPart());
	setHypo(i,xx,yy,confi);
}

void HypoImage::setHypo(int i, int posX, int posY, float xVel, float yVel, float confi){
	GPUData hypo = data[posY][posX][i];
	hypo[0] = confi;
	hypo[1] = xVel;
	hypo[2] = yVel;
}

void HypoImage::printHyposInFile(int H_MAX_1, const char* filename)
{
	FILE *myFile1;
	myFile1 = fopen(filename,"w");

	if(myFile1)
	{
		for(int i=0; i< w*h*H_MAX_1*3 ; i+=3)
		{
			int test_position = i;
			fprintf(myFile1,"%.3f  %.3f  %.3f \n "
					, getData().reduceDimensionality()[test_position].as_float()
					, getData().reduceDimensionality()[test_position + 1].as_float()
					, getData().reduceDimensionality()[test_position + 2].as_float()
					);

		}
		fclose(myFile1);
	}
}

void HypoImage::interpolate(int iFactor, float* BilWeight)
{
//	HostHypoImage HypoMT(w,h,h_max);
//	this->getData().copyToHostMemory(HypoMT.getDataPointer());
//	float *pTmpVec = HypoMT.getDataPointer();
//
//	int i, j, posBilWeights, x, y;
//	int m=0, m2, x1MT, y1MT, x2MT, y2MT, x3MT, y3MT, x4MT, y4MT,k, wMT = w/iFactor, hMT = h/iFactor;
//	HypoClass *pTmpHypo, *begin, *end, *pHypoMT;
//	HypoClass TmpHypoMT[h_max * 4];
//	HostHypoImage expMT(w*iFactor,h*iFactor,h_max);
//
//	pHypoMT = HypoMT;
//
//	for(i=0; i<w*h*iFactor*iFactor; i++)
//	{
//		x = (i% w);
//		y = (int)(i/ w);
//		x1MT = (int)x/ iFactor;
//		y1MT = (int)y/ iFactor;
//		x2MT = x1MT + 1; y2MT = y1MT;
//		x3MT = x1MT;	 y3MT = y1MT+1;
//		x4MT = x1MT + 1; y4MT = y1MT+1;
//		m=0; posBilWeights = ((x% iFactor) + ((y% (iFactor))*(iFactor))) * 4;
//
//
//		int currpos = ((i*h_max+k)*3);
//		//calcuatlate interpolated HypoImage using bilinear interpolation
//		if(x1MT<wMT && y1MT < hMT) for(k=0; k<h_max; k++) {*(TmpHypoMT+m) = (pTmpVec+x1MT + y1MT*wMT)->ListHypo[k]; (*(TmpHypoMT+m)).w *= *(BilWeight + posBilWeights   ); m++;}
//		if(x2MT<wMT && y2MT < hMT) for(k=0; k<h_max; k++) {*(TmpHypoMT+m) = (pTmpVec+x2MT + y2MT*wMT)->ListHypo[k]; (*(TmpHypoMT)).w *= *(BilWeight + posBilWeights +1); m++;}
//		if(x3MT<wMT && y3MT < hMT) for(k=0; k<h_max; k++) {*(TmpHypoMT+m) = (pTmpVec+x3MT + y3MT*wMT)->ListHypo[k]; (*(TmpHypoMT+m)).w *= *(BilWeight + posBilWeights +2); m++;}
//		if(x4MT<wMT && y4MT < hMT) for(k=0; k<h_max; k++) {*(TmpHypoMT+m) = (pTmpVec+x4MT + y4MT*wMT)->ListHypo[k]; (*(TmpHypoMT+m)).w *= *(BilWeight + posBilWeights +3); m++;}
//
//		begin = TmpHypoMT;
//		end = TmpHypoMT + m;
//		sort( begin, end, UDgreaterHypoVel );
//		//add weights of identical velocities
//		pTmpHypo = TmpHypoMT;
//		m2 = m;
//		for(j = 0; j < m-1; j++) //(H_MAX_MT * 4)-1
//		{
//			if((*(pTmpHypo)).vx == (*(pTmpHypo+1)).vx && (*(pTmpHypo)).vy == (*(pTmpHypo+1)).vy) {(*(pTmpHypo+1)).w += (*(pTmpHypo)).w; (*(pTmpHypo)).w = 0.0f; (*(pTmpHypo)).vx = 0.0f; (*(pTmpHypo)).vy = 0.0f; m2--;}
//			pTmpHypo++;
//		}
//		begin = TmpHypoMT;
//		end = TmpHypoMT + m;
//		sort( begin, end, UDgreaterHypoW );
//		pTmpHypo = TmpHypoMT;
//		for(j=0; j<h_max; j++)
//		{
//			(*(expMT + currpos + 1)) = (float)(*pTmpHypo).vx;
//			(*(expMT + currpos + 2)) = (float)(*pTmpHypo).vy;
//			(*(expMT + currpos + 0)) = (float)(*pTmpHypo).w;
//			pTmpHypo++;
//		}
//		currpos+3;
//	}
//
//
//	delete []expMT;
	return;
}


#if 0

void HypoImage::velocityBlur( int VEL_MAX )
{
  CPUData<float> hostHypoImage( h, w, h_max * 3 );
  getData().copyToHostMemory( hostHypoImage.data() );


  // temporary hypothesises under this threshold are discarded immeadiately
  const float c_thresh = 0.01f;

  // 3x3 hamming window
  const float hamming[3][3] = { { 0.25, 0.50, 0.25 }, 
								{ 0.50, 1.00, 0.50 },  
								{ 0.25, 0.50, 0.25 } };
  const size_t window_size   = 3;
  const size_t window_center = window_size / 2;
  
  // tmps needed are maximum number hypos times size of neighborhood
  const size_t n_hypo_tmp =  window_size *  window_size * h_max;
  HypoClass TmpHypo [ n_hypo_tmp ];
  
  for( CPUData<float>::iterator ihypos = hostHypoImage.begin(), iend = hostHypoImage.end(); 
	   ihypos != iend; ++ihypos )
	{
	  size_t n_hypos1 = 0;
	  
	  // generate smoothed hypothesis population
	  // by pushing each source hypo through the hamming window
	  for( float * jhypo = ihypos->begin(), * jend = ihypos->end(); jhypo != jend; jhypo += 3 )
		{
		  const float c_cur  = jhypo[0]; 
		  const float vx_cur = jhypo[1];
		  const float vy_cur = jhypo[2];
		  if( c_cur > c_thresh )
			{
			  
			  for ( int dx = -1; dx <= 1; ++dx )
				for ( int dy = -1; dy <= 1; ++dy, ++n_hypos1 )
				  {															
					HypoClass & h = TmpHypo[ n_hypos1 ];					
					h.vx = vx_cur + dx;						
					h.vy = vy_cur + dy;
					h.w  = c_cur  * hamming[window_center + dx][window_center + dy];
				  }
			}
		}
	  
	  if( n_hypos1 > 0 )
		{

		  // sort hypos by velocity
		  std::sort( TmpHypo, TmpHypo + n_hypos1, UDgreaterHyposVel(VEL_MAX) );		  		  

		  // add weights of identical velocities
		  int n_hypos2 = 1;
		  for( size_t k = 1; k < n_hypos1; ++k )
			{
			  HypoClass & h1 = TmpHypo [ k            ];
			  HypoClass & h2 = TmpHypo [ n_hypos2 - 1 ];
			  if( h2 == h1 )
				h2.w += h1.w; 
			  else
				{
				  h2 = h1;
				  ++n_hypos2;
				}
			}

		  // find best hypos
		  std::make_heap( TmpHypo, TmpHypo + n_hypos2, UDgreaterHyposW );
		  float * jout = ihypos->begin();
		  for ( size_t j = 0, nj = std::min( n_hypos2, int(h_max) ); j < nj; ++j, jout += 3 )
			{
			  HypoClass & h = TmpHypo[0];
			  jout[0] = h.w;
			  jout[1] = h.vx;
			  jout[2] = h.vy;
			  std::pop_heap( TmpHypo, TmpHypo + n_hypos2 - j );
			}

		  // fill rest with zeros
		  for ( float * jend = ihypos->end(); jout != jend; ++jout )
			*jout = 0.0f;

		}
	}
  

  getData().copyFromHostMemory( hostHypoImage.data() );

}

#else

void HypoImage::velocityBlur(  int VEL_MAX)
{
  CPUData<float> hostHypoImage( w, h, 3 * h_max );
  this->getData().copyToHostMemory(hostHypoImage.data());
  float *pTmpVec = hostHypoImage.data();
  
  size_t sizeHF = 3;
  int i, j, s, foundHypos;
  float hanning[9] = {0.25, 0.5, 0.25, 0.5, 1, 0.5,  0.25, 0.5,  
					  0.25};//filter to weight neighbourhood
  HypoClass *begin, *end, *pTmpHypo;
#ifndef _WIN32
  HypoClass TmpHypo[sizeHF * sizeHF * h_max];
#else
  HypoClass * const TmpHypo = (HypoClass*)_alloca(sizeHF * sizeHF * h_max * sizeof(HypoClass));
#endif

  
  for(i=0; i< w*h ; i++)
	{
	  //temporal array to store the sizeHF* sizeHF*  iNrHypo Hypos at each location
	  pTmpHypo = TmpHypo;
	  foundHypos = 0;
	  
	  for(j=0; j<h_max; j++)//take all hypos at each location
		{
		  int currpos = ((i*h_max+j)*3);
		  if((*(pTmpVec + currpos + 0)) > 0.01)
			{
			  foundHypos+=(sizeHF * sizeHF);
			  //Hanning 0
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) - 1;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) - 1;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+0));
			  pTmpHypo++;
			  
			  //Hanning 1
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1))  ;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) - 1;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+1));
			  pTmpHypo++;
			  
			  //Hanning 2
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) + 1 ;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) - 1;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+2));
			  pTmpHypo++;
			  
			  //Hanning 3
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) - 1;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) + 0;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+3));
			  pTmpHypo++;
			  
			  //Hanning 4
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) + 0;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) + 0;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+4));
			  pTmpHypo++;
			  
			  //Hanning 5
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) + 1;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) + 0;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+5));
			  pTmpHypo++;
			  
			  //Hanning 6
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) - 1;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) + 1;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+6));
			  pTmpHypo++;
			  
			  //Hanning 7
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) + 0;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) + 1;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+7));
			  pTmpHypo++;
			  
			  //Hanning 8
			  (*(pTmpHypo)).vx =  (*(pTmpVec + currpos + 1)) + 1;
			  (*(pTmpHypo)).vy = (*(pTmpVec + currpos + 2)) + 1;
			  (*(pTmpHypo)).w = (*(pTmpVec + currpos + 0))  * (*(hanning+8));
			  pTmpHypo++;
			}
		}
	  
	  if(foundHypos)
		{
		  //sortHypos - velocity
		  begin = TmpHypo;
		  end = TmpHypo + foundHypos;
		  std::sort( begin, end, UDgreaterHyposVel( VEL_MAX ) );
		  
		  //add weights of identical velocities
		  pTmpHypo = TmpHypo;
		  for(j = 0; j < foundHypos-1; j++)
			{
			  if((*(pTmpHypo)).vx == (*(pTmpHypo+1)).vx && (*(pTmpHypo)).vy ==  
				 (*(pTmpHypo+1)).vy) {(*(pTmpHypo+1)).w += (*(pTmpHypo)).w;  
				(*(pTmpHypo)).w = 0.0f; (*(pTmpHypo)).vx = 0.0f; (*(pTmpHypo)).vy =  
																   0.0f;}
			  pTmpHypo++;
			}
		  
		  //sort Hypos - weight
		  begin = TmpHypo;
		  end = TmpHypo + foundHypos ;
		  std::sort( begin, end, UDgreaterHyposW );
		  
		  //write hypos with max. weight to position
		  pTmpHypo = TmpHypo;
		  for(j = 0; j < h_max; j++)
			{
			  int currpos = ((i*h_max+j)*3);
			  if(j < foundHypos  && (*pTmpHypo).w > 0 )
				{
				  (*(pTmpVec + currpos + 1)) = (float)(*pTmpHypo).vx;
				  (*(pTmpVec + currpos + 2)) = (float)(*pTmpHypo).vy;
				  (*(pTmpVec + currpos + 0)) = (float)(*pTmpHypo).w;
				}
			  else
				{
				  (*(pTmpVec + currpos + 1)) = 0.0f;
				  (*(pTmpVec + currpos + 2)) = 0.0f;
				  (*(pTmpVec + currpos + 0)) = 0.0f;
				}
			  pTmpHypo++;
			}
		}
	}
  
  //DEBUG:
  //      for(i=0; i< w*h*h_max*3 ; i+=3)
  //      {
  //              int currpos = i;
  //              fprintf(myFile1,"%.3f  %.3f  %.3f \n",  
  //(*(hostHypoImage.data() + currpos + 0)),  
  //	(*(hostHypoImage.data() + currpos + 1)),  
  //	(*(hostHypoImage.data() + currpos + 2)));
  //      }
  //
  //      fclose(myFile1);

  this->getData().copyFromHostMemory(hostHypoImage.data());
  return;
}



#endif

GPUVelField HypoImage::getExpectationValue( ) 
{
  GPUImage foo( w, h );
  return getExpectationValue( foo );
}

GPUVelField HypoImage::getExpectationValue( GPUImage & out_weight_sum ) 
{
  GPUVelField result( w, h );
  out_weight_sum.resize( w, h );
  GPUImage vx = result.getXPart();
  GPUImage vy = result.getYPart();
  vx.fill( 0.0f );
  vy.fill( 0.0f );
  out_weight_sum.fill( 0.0f );
  for ( int i = 0; i < h_max; ++i )
	{
	  GPUData hypo = data.slice( 2, i );
	  GPUData weights( hypo.slice( 2, 0 ) );
	  out_weight_sum.add( weights );
	  vx.add( hypo.slice( 2, 1 ).copy().mul( weights ) );
	  vy.add( hypo.slice( 2, 2 ).copy().mul( weights ) );
	}
  return result;
}

void HypoImage::subsample_hypo_image(int vel_max, float sigma, int v1_window_width, HypoImage& output)
{
  do_subsample_hypo_image( h_max,  vel_max,  sigma,  v1_window_width,
						   getData().getDataPointer(),  w ,  h,
						   output.getData().getDataPointer(),  output.w,  output.h);
}



void HypoImage::normVelSpace(float epsilon) 
{
  do_normHypos_simple(data.getDataPointer(),w*h,h_max,3*h_max,epsilon);
}


void HypoImage::getHypoSum(GPUImage& dest) {
	do_sumHypos_simple(data.getDataPointer(), dest.getDataPointer() ,w*h,h_max,3*h_max);
}

void HypoImage::get_best_hypos( size_t n, HypoImage & output )
{
  n = std::min( n, h_max );
  sortHypos();
  output.init( w, h, n );
  output.getData().assign( getData().range( 2, 0, n - 1 ) );
}

extern "C" void do_log_polar_transform( float * hypos, int width, int height, int depth, int direction );

void HypoImage::cartesian_to_log_polar()
{
  //getData().makeContiguous(true);
  do_log_polar_transform( getData().data(), getWidth(), getHeight(), getHMAX() * 3, 0 );
}

void HypoImage::log_polar_to_cartesian()
{
  //getData().makeContiguous(true);
  do_log_polar_transform( getData().data(), getWidth(), getHeight(), getHMAX() * 3, 1 );
}


extern "C" void do_log_transform( float * hypos, int width, int height, int depth, int direction );

void HypoImage::log_compress()
{
  do_log_transform( getData().data(), getWidth(), getHeight(), getHMAX() * 3, 0 );
}

void HypoImage::log_expand()
{
  do_log_transform( getData().data(), getWidth(), getHeight(), getHMAX() * 3, 1 );
}



extern "C" void do_hypos_mean_shift_smooth
( const float * in_hypos, int in_width, int in_height, int in_depth
  , float * out_hypos,  int out_width, int out_height, int out_depth
  , int window_size, int window_shift
  , float v1min, float v1max, int n1
  , float v2min, float v2max, int n2
  , float sigma_space, float sigma1, float sigma2
  , int iterations	);

extern "C" void do_hypos_gmm_smooth
( const float * in_hypos, int in_width, int in_height, int in_depth
  , float * out_hypos,  int out_width, int out_height, int out_depth
  , int window_size, int window_shift
  , float v1min, float v1max, int n1
  , float v2min, float v2max, int n2
  , float sigma_space, float sigma1, float sigma2
  , int iterations	);

void HypoImage::meanshift_smooth( float sigma_space, float sigma_vel,
								  int window_size, int window_shift
								  ,int nv , int vmax
								  , HypoImage & out, int iterations
								  , bool use_gmm )
{

#if 1
  size_t out_width  = 1 + ( getWidth () - window_size ) / window_shift;
  size_t out_height = 1 + ( getHeight() - window_size ) / window_shift;

  out.init( out_width, out_height, getHMAX() );

  GPUDataException::check_error( "before mean_shift_smooth" );

  log_compress();
  
  vmax = std::log( (float)(1 + vmax) );
  sigma_vel = std::log( 1.0f + sigma_vel );

  if ( use_gmm )
	do_hypos_gmm_smooth( getData().data(), getWidth(), getHeight(), getHMAX(),
						 out.getData().data(), out.getWidth(), out.getHeight(), out.getHMAX(),
						 window_size, window_shift,
						 -vmax, vmax, nv,
						 -vmax, vmax, nv,
						 sigma_space, sigma_vel, sigma_vel,
						 iterations
						 );  
  else
	do_hypos_mean_shift_smooth( getData().data(), getWidth(), getHeight(), getHMAX(),
								out.getData().data(), out.getWidth(), out.getHeight(), out.getHMAX(),
								window_size, window_shift,
								-vmax, vmax, nv,
								-vmax, vmax, nv,
								sigma_space, sigma_vel, sigma_vel,
								iterations
								);
  

  log_expand();

  out.log_expand();

#else
  out = HypoImage( getData().copy() );
#endif

  GPUDataException::check_error( "after mean_shift_smooth" );

}

