/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * HypoImage.h
 *
 *  Created on: July 2009
 *      Author: jbouecke, cbeck
 */

#ifndef HYPOIMAGE_H_
#define HYPOIMAGE_H_

#include <GPUData.h>
#include <GPUImage.h>
#include <GPUVelField.h>
/**
 * This class provides the processing steps for the motion algo of UULM [Bayerl&Neumann, TPAMI, 2007]
 */
class HypoImage {
private:

	int w;
	int h;

	size_t h_max;

	GPUData  data;
	GPUImage weightView;
//	GPUData  weightView1D;

public:

	void init(int w, int h, int max_h);

	HypoImage(int w, int h, int max_h);

	HypoImage();

	HypoImage( const GPUData & data );

	GPUData & getData() { return data; }
	const GPUData & getData() const { return data; }

	GPUImage getWeights() { return weightView; }

	int getWidth() {return w;}

	int getHeight() {return h;}

	int getHMAX() {return h_max;}

	/**
	 * sums all hypo weights at each pixel position
	 * (tested)
	 * @param dest
	 */
	void getHypoSum(GPUImage& dest);

	/**
	 * returns a view to one property of the i'th hypothesis at
	 * each position in the image. The property index is given
	 * by v (0=weight, 1=vx, 2=vy);
	 */
	GPUImage getHypothesisProperty(int i, int v);

	/**
	 * get h_max hypothesis at a given position
	 *
	 * (tested)
	 * @param x_pos
	 * @param y_pos
	 * @return view
	 */
	GPUData getHypothesis(int x_pos, int y_pos);

	/*
	 * square weights of hypotheses
	 */
	void sqrWeights();

	/**
	 * hypotheses are sorted on the GPU (currently not used)
	 */
	void sortHypos();

	virtual ~HypoImage();

	/**
	 * (tested)
	 * @param i hypothesis index
	 * @param xVel velocity part in x direction
	 * @param yVel velocity part in y direction
	 * @param confi weights
	 */
	void setHypo(int i, GPUImage& xVel, GPUImage& yVel, GPUImage& confi);

	/**
	 * see setHypo
	 *
	 * (not tested)
	 */
	void setHypo(int i, GPUVelField& velField, GPUImage& confi);

	GPUVelField getExpectationValue();
	GPUVelField getExpectationValue( GPUImage & out_weight_sums );

	/**
	 * see setHypo
	 *
	 * (not tested)
	 */
	void setHypo(int i, int posX, int posY, float xVel, float yVel, float confi);

	/**
	 * normalize weights of all hypotheses at one  position
	 *
	 * (tested)
	 */
	void normVelSpace(float epsilon = 0.0000000000000001f);

	/**
	 * Print all hypotheses in file (vx, vy and w)
	 *
	 *
	 */
	void printHyposInFile(int H_MAX_1, const char*);

	/**
	 * Subsample hypoimage to get coarser spatial representation - e.g., to compute MT using V1's hypotheses
	 * @param vel_max maximal velocity to detect
	 * @param sigma spatial blur using a gaussian function
	 * @param v1_window_width number of V1 neurons contributing to one MT position in 1D
	 *
	 */
	void subsample_hypo_image(int vel_max, float sigma, int v1_window_width, HypoImage& output);

	/**
	 * Blur in velocity domain at each position
	 * @param vel_max maximal velocity to detect
	 *
	 */
	void velocityBlur( int VEL_MAX );

	/// use mean shift to subsample in space and velocities
	void mean_shift_subsample( float sigma_space, float sigma_vel, 
							   int window_size, int window_shift, int n_scales,
							   int iterations, HypoImage & output );

	void get_best_hypos( size_t n, HypoImage & output );

	/**
	 * Interpolate image to get higher resolution (e.g., to get MT in V1 resolution)
	 * Not yet finished
	 *
	 */
	void interpolate(int iFactor, float* BilWeight);

	void cartesian_to_log_polar();

	void log_polar_to_cartesian();

	void log_compress();

	void log_expand();

	void meanshift_smooth( float sigma_space, float sigma_vel
						   , int window_size, int window_shift
						   ,int nv , int vmax
						   , HypoImage & output
						   , int iterations
						   , bool use_gmm = false
						   );
};

#endif /* HYPOIMAGE_H_ */
