/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * INRIAForm.cpp
 *
 *  Created on: Feb 16, 2009
 *      Author: jbouecke
 */

#include "INRIAForm.hpp"
#include "GPUVelActivities.h"
#include <cmath>

#ifdef _WIN32
#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif
#define fmin(a,b) (a < b ? a : b)
#define fmax(a,b) (a > b ? a : b)
#define fminf(a,b) (a < b ? a : b)
#define fmaxf(a,b) (a > b ? a : b)
#endif // _WIN32

INRIAForm::INRIAForm()
	: ddim( 10 )
	, sgx ( 10.0f )
	, sgtheta ( 2.*M_PI/16 )
	, sgsim ( 0.4f/255.0f )
	, nsz (fmax(3, 3 * sgx))
	, wsize ( 2*nsz+1 )
{
	init_weights();
}

INRIAForm::~INRIAForm() {
	;
}

void INRIAForm::init_weights() {
	size_t sizes[3] = {ddim,wsize,wsize};
	weights.resize(3,sizes);

	for (int d = 0; d < ddim; d++) {
		double weight_sum = 0.0;
		double theta = 2.0 * M_PI * float(d) / float(ddim);
		GPUData weights_d = weights[d];
		for (int q = 0; q<wsize; ++q) {
			int y = q - nsz;
			for (int p = 0; p<wsize; ++p) {
				int x = p - nsz;
				if (x == 0 && y == 0) {
					weights_d[q][p] = 1.0f;
					weight_sum += 1.0;
					continue;
				}

				double rho = sqrt((double)(x*x + y*y));
#ifndef _WIN32
				double theta_xy = norm_angle(atan2(y, x));
#else
				double theta_xy = norm_angle(atan2((double)y, (double)x));
#endif
				double ws = gaussian(rho, sgx);
				double delta_diff = fmin(fabs(theta_xy - theta),
								 2. * M_PI - fabs(theta_xy - theta));
				double wt = gaussian(delta_diff, sgtheta);

				weights_d[q][p] = ws * wt;
				weight_sum += ws * wt;
			}
		}
		weights_d.div(weight_sum);
	}
}

void INRIAForm::set_ddim(int d) {
	if (ddim!=d) {
		ddim = d;
		init_weights();
	}
}

GPUData INRIAForm::phi_for_frame(GPUImage& image) {
	const int h = image.getHeight();
	const int w = image.getWidth();
	size_t sizes[3] = { ddim, h, w };
	phi.resize( 3, sizes );
	phi_for_frame( image, phi );
	return phi;
}
extern "C" void do_phi_function(float* out, float* in, float* weights,
		size_t width, size_t height, int nsz, float sim_sg);

void INRIAForm::phi_for_frame(GPUImage& image, GPUData& rtv) {
	const size_t h = image.getHeight();
	const size_t w = image.getWidth();
	GPUImage frame = image;
	for (int d=0; d<ddim; ++d) {
		GPUData dw = weights[d];
		GPUData dr = rtv[d];
		do_phi_function(dr.getDataPointer(), frame.getDataPointer(), dw.getDataPointer(), w, h, nsz, sgsim);
	}
}

extern "C" void
do_heth_diffusion(float* out, const float* in, const float* phi,
	       float sigma, int width, int height, int ddim);

void INRIAForm::integrate(GPUData& vels_in, GPUData& phi, GPUData& vels_out)
{
	GPUVelActivities in_vel_view(vels_in);
	GPUVelActivities out_vel_view(vels_out);

	// reintepret velocity array as 3d with features first (uv,x,y)
	GPUData in  = in_vel_view.get2DView().shiftDim(1);
	GPUData out = out_vel_view.get2DView().shiftDim(1);

	const size_t h  = in_vel_view.getYRes();
	const size_t w  = in_vel_view.getXRes();
	const size_t wh = w*h;
	const size_t UV = in.extent(0);

	temp1.resize(1,&wh);
	temp2.resize(1,&wh);

	for (size_t i=0; i<UV; ++i) {
		temp1.copyFrom(in[i]);
		do_heth_diffusion(temp2.getDataPointer(), temp1.getDataPointer(), phi.getDataPointer(), sgx, w, h, ddim);
		out[i].copyFrom(temp2);
	}
}

void INRIAForm::iterate(GPUImage& image, GPUData& in, GPUData& out) {
	phi_for_frame(image);
	integrate(in,phi,out);
}
