/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * INRIAForm.hpp
 *
 *  Created on: Feb 16, 2009
 *      Author: jbouecke
 */

#ifndef INRIAFORM_HPP_
#define INRIAFORM_HPP_

#include "GPUImage.h"

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

/**
 * This is a c++ wrapper class to compute and apply the anisotropic diffusion
 * for motion integration, developed by INRIA
 */
class INRIAForm {
private:

	/* BEGIN - model parameters - */

	/// number of directions
	int ddim;

	/// sigma of gaussian in space
	float sgx;

	/// sigma for direction angle theta
	float sgtheta;

	/// sigma for the similarity Gaussian
	float sgsim;

	/// kernel radius (for weights)
	int nsz;

	/// kernel size (for weights)
	int wsize;

	/* END - model parameters - */

	void init_weights();

	/// directional weights
	GPUData weights;

	/// normalize the angle to [0,2*pi]
	inline double norm_angle(double theta) {
		if (theta < 0.0)
			return fmod(theta, 2.0 * M_PI) + 2.0 * M_PI;
		else
			return fmod(theta, 2.0 * M_PI);
	}
	/// returns unnormalized mean free gauss tuned value
	inline double gaussian(double x, double sg) {
		return exp(-x*x/(2*sg*sg));
	}

	/// just temporary space for the current phi function (ori,h,w)
	GPUData phi;

	/// just temporary space
	GPUData temp1,temp2;

	/**
	 * This function calculates the diffusion coefficients at each position for
	 * different directions
	 * @param image the current image
	 * @return a object sharing data with phi (member variable)
	 * 		(direction,height,weight)
	 */
	GPUData phi_for_frame(GPUImage& image);

	/**
	 * see phi_for_frame()
	 */
	void phi_for_frame(GPUImage& image, GPUData& rtv);

	/**
	 * usese the phi values (caluclated by phi_for_frame()) to integrate all
	 * velocity hypothesis
	 * @param vels_in (h,w,vel_y,vel_x)
	 * @param phi
	 * @param vels_out
	 */
	void integrate(GPUData& vels_in, GPUData& phi, GPUData& vels_out);

public:
	INRIAForm();
	virtual ~INRIAForm();

	/// updates model parameter ddim and recalculates weights (if ddim changed)
	void set_ddim(int d);

	/**
	 * This function first computes the phi function for the given image and
	 * than uses this to compute the anisotropic diffusion step.
	 * @param image current image
	 * @param in    velocity activities to diffuse (h,w,vel_y,vel_x)
	 * @param out   destination space for diffused velocities
	 */
	void iterate(GPUImage& image, GPUData& in, GPUData& out);
};

#endif /* INRIAFORM_HPP_ */
