/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * FlowIO.cpp
 *
 *  Created on: Feb 25, 2009
 *      Author: jbouecke
 */

#include "JDBFlowIO.hpp"
#include <Exceptions.hpp>


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

struct CError : public std::exception
{
    CError(const char* msg)                 { strcpy(message, msg); }
    CError(const char* fmt, int d)          { sprintf(message, fmt, d); }
    CError(const char* fmt, float f)        { sprintf(message, fmt, f); }
    CError(const char* fmt, const char *s)  { sprintf(message, fmt, s); }
    CError(const char* fmt, const char *s,
            int d)                          { sprintf(message, fmt, s, d); }
    char message[1024];         // longest allowable message
};


// first four bytes, should be the same in little endian
#define TAG_STRING "PIEI"    // use this when WRITING the file
#define TAG_FLOAT (*((float*)"PIEI"))  // check for this when READING the file

//float JDBFlowIO::UNKNOWN_FLOW_THRESH;
//float JDBFlowIO::UNKNOWN_FLOW;

// return whether flow vector is unknown
bool JDBFlowIO::unknown_flow(float u, float v) {
    return (fabs(u) >  UNKNOWN_FLOW_THRESH)
	|| (fabs(v) >  UNKNOWN_FLOW_THRESH);
}

bool JDBFlowIO::unknown_flow(float *f) {
    return unknown_flow(f[0], f[1]);
}

// read a flow file into 2-band image
float* JDBFlowIO::ReadFlowFile(const char* filename, int& w, int& h)
{
    if (filename == NULL)
	throw CError("ReadFlowFile: empty filename");

    const char *dot = strrchr(filename, '.');
    if (strcmp(dot, ".flow") != 0)
	throw CError("ReadFlowFile (%s): extension .flow expected", filename);

    FILE *stream = fopen(filename, "rb");
    if (stream == 0)
        throw CError("ReadFlowFile: could not open %s", filename);

    int width, height;
    float tag;

    if ((int)fread(&tag,    sizeof(float), 1, stream) != 1 ||
	(int)fread(&width,  sizeof(int),   1, stream) != 1 ||
	(int)fread(&height, sizeof(int),   1, stream) != 1)
	throw CError("ReadFlowFile: problem reading file %s", filename);

    w=width;
    h=height;

    if (tag != TAG_FLOAT) // simple test for correct endian-ness
	throw CError("ReadFlowFile(%s): wrong tag (possibly due to big-endian machine?)", filename);

    // another sanity check to see that integers were read correctly (99999 should do the trick...)
    if (width < 1 || width > 99999)
	throw CError("ReadFlowFile(%s): illegal width %d", filename, width);

    if (height < 1 || height > 99999)
	throw CError("ReadFlowFile(%s): illegal height %d", filename, height);

    int nBands = 3;
    float* img = new float[width*height*nBands];

    //printf("reading %d x %d x 2 = %d floats\n", width, height, width*height*2);
    int n = nBands * width;
    for (int y = 0; y < height; y++) {
		float* ptr = img+y*n;
		if ((int)fread(ptr, sizeof(float), n, stream) != n)
			throw CError("ReadFlowFile(%s): file is too short", filename);
    }

    if (fgetc(stream) != EOF)
	throw CError("ReadFlowFile(%s): file is too long", filename);

    fclose(stream);
    return img;
}

void JDBFlowIO::WriteFlowFile(const GPUVelField& flow, const GPUImage& conf, const char* filename) {
	const int w = flow.getWidth();
	const int h = flow.getHeight();
//	GPUData ddata = flow.getAsFeatureImage();//.swapDimensions(0,1);
	GPUData dddata(h,w,3);
	dddata.slice(2,0).copyFrom(flow.getXPart());
	dddata.slice(2,1).copyFrom(flow.getYPart());
	dddata.slice(2,2).copyFrom(conf);
//	ddata = ddata.copy();
	float* hdata = new float[w*h*3];
	dddata.copyToHostMemory(hdata);
	WriteFlowFile(w,h,hdata,filename);
	delete [] hdata;
}

// write a 3-band image into flow file
void JDBFlowIO::WriteFlowFile(int width, int height, float* img, const char* filename)
{
    if (filename == NULL)
	throw CError("WriteFlowFile: empty filename");

    const char *dot = strrchr(filename, '.');
    if (dot == NULL)
	throw CError("WriteFlowFile: extension required in filename '%s'", filename);

    if (strcmp(dot, ".flow") != 0)
	throw CError("WriteFlowFile: filename '%s' should have extension '.flo'", filename);

    int nBands = 3;

    FILE *stream = fopen(filename, "wb");
    if (stream == 0)
        throw CError("WriteFlowFile: could not open %s", filename);

    // write the header
    fprintf(stream, TAG_STRING);
    if ((int)fwrite(&width,  sizeof(int),   1, stream) != 1 ||
	(int)fwrite(&height, sizeof(int),   1, stream) != 1)
	throw CError("WriteFlowFile(%s): problem writing header", filename);

    // write the rows
    int n = nBands * width;
    for (int y = 0; y < height; y++) {
		float* ptr = img+y*n;
		if ((int)fwrite(ptr, sizeof(float), n, stream) != n)
			throw CError("WriteFlowFile(%s): problem writing data", filename);
    }

    fclose(stream);
}
