/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * LoGFeatureExtractor.cpp
 *
 *  Created on: Sep 22, 2009
 *      Author: jbouecke
 */

#include "LoGFeatureExtractor.hpp"

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

LoGFeatureExtractor::LoGFeatureExtractor()
	: noOfOrientations(8)
	, sigma1(1.5)
	, sigma2(1.5)
	, c1(1.0f)
	, c2(1.0f)
	, gaborLike(true)
	, w(-1)
	, h(-1)
	, _compute_both_derivs(false)
	, _first(true)
{
}

LoGFeatureExtractor::LoGFeatureExtractor(const int n_)
	: noOfOrientations(n_)
	, sigma1(1.5)
	, sigma2(1.5)
	, c1(1.0f)
	, c2(1.0f)
	, gaborLike(true)
	, w(-1)
	, h(-1)
	, _compute_both_derivs(false)
	, _first(true)
{
}

LoGFeatureExtractor::LoGFeatureExtractor(const int n_, const float sigma_)
	: noOfOrientations(n_)
	, sigma1(sigma_)
	, sigma2(sigma_)
	, c1(1.0f)
	, c2(1.0f)
	, gaborLike(true)
	, w(-1)
	, h(-1)
	, _compute_both_derivs(false)
	, _first(true)
{
}

LoGFeatureExtractor::LoGFeatureExtractor(const float sigma_)
	: noOfOrientations(8)
	, sigma1(sigma_)
	, sigma2(sigma_)
	, c1(1.0f)
	, c2(1.0f)
	, gaborLike(true)
	, w(-1)
	, h(-1)
	, _compute_both_derivs(false)
	, _first(true)
{
}

LoGFeatureExtractor::LoGFeatureExtractor(const int w_, const int h_)
	: noOfOrientations(8)
	, sigma1(1.5)
	, sigma2(1.5)
	, c1(1.0f)
	, c2(1.0f)
	, gaborLike(true)
	, w(-1)
	, h(-1)
	, _compute_both_derivs(false)
	, _first(true)
{
}

LoGFeatureExtractor::LoGFeatureExtractor(const int w_, const int h_, const int n_)
	: noOfOrientations(n_)
	, sigma1(1.5)
	, sigma2(1.5)
	, c1(1.0f)
	, c2(1.0f)
	, gaborLike(true)
	, w(-1)
	, h(-1)
	, _compute_both_derivs(false)
	, _first(true)
{
}

LoGFeatureExtractor::LoGFeatureExtractor(const int w_, const int h_, const int n_, const float sigma_)
	: noOfOrientations(n_)
	, sigma1(sigma_)
	, sigma2(sigma_)
	, c1(1.0f)
	, c2(1.0f)
	, gaborLike(true)
	, w(-1)
	, h(-1)
	, _compute_both_derivs(false)
	, _first(true)
{
}



void make_linspace( GPUData & data, size_t n )
{
  int r = n / 2;
  data.resize( n );
  fillLinSpace(data,-r,r);
}

void make_gauss( GPUData & data, float sigma, size_t n )
{
  make_linspace( data, n );
  data.sqr();
  data.div(-2.0f*sigma*sigma);
  data.exp();
  data.mul(1.0/sqrt(2*M_PI)/sigma);  
}

void kernelize( GPUData & envelope, GPUData & temp )
{
#if 0
  GPUData temp2( temp.copy() );
  temp2.sqr();
  temp2.mul( envelope );
  float Z = temp2.sum();
  envelope.mul( temp, 1.0f / Z );
#else
  envelope.mul( temp );
#endif
}

void LoGFeatureExtractor::init (const int w_, const int h_ )
{

	if (_first) 
	  { 
		
		GPUData temp;

		// first derivative

		size_t n1 = 3 * sigma1 + 1;

		make_linspace( temp, n1 );
		temp.mul( -1.0 / ( sigma1 * sigma1 ) ); 

		make_gauss( diffKernel1, sigma1, n1 );
		kernelize( diffKernel1, temp );

		gauss_small1_x = GaussKernel1( sigma1, 1 );
		gauss_small1_y = GaussKernel1( sigma1, 0 );

		// second derivative

		size_t n2 = 3 * sigma2 + 1;

		make_linspace( temp, n2 );
		temp.mul( -1.0 / ( sigma1 * sigma1 ) ); 

		make_gauss( diffKernel2, sigma2, n2 );
		kernelize( diffKernel2, temp );

		make_linspace( temp, n2 );
		temp.sqr();
		temp.div( sigma2 * sigma2 );
		temp.sub( 1 );
		temp.div( sigma2 * sigma2 );

		make_gauss( diffdiffKernel, sigma2, n2 );
		kernelize( diffdiffKernel, temp );

		gauss_small2_x = GaussKernel1( sigma2, 1 );
		gauss_small2_y = GaussKernel1( sigma2, 0 );

	  }
	
	if ( ( w_ == w && h_ == h ) || w_ <= 0 || h_ <= 0 )
	  return;
	
	this->w = w_;
	this->h = h_;
	
	tempImage  = GPUImage(w,h);

	dIx = GPUImage(w,h);
	dIy = GPUImage(w,h);
	
	dIxy = GPUImage(w,h);
	dIxx = GPUImage(w,h);
	dIyy = GPUImage(w,h);
}

LoGFeatureExtractor::~LoGFeatureExtractor() {
	;
}

GPUData LoGFeatureExtractor::getFeatureImages( GPUImage& input ) 
{
  GPUData rtv(noOfOrientations,input.getHeight(),input.getWidth());
  getFeatureImages(input, rtv);
  return rtv;
}

void LoGFeatureExtractor::getFeatureImages( GPUImage& input, GPUData& result ) 
{
	
	init( input.getWidth(), input.getHeight() );

	// dx
	gauss_small1_y.apply(input,tempImage);
	tempImage.convolute1D( diffKernel1, 1, dIx );

	// dy
	gauss_small1_x.apply( input, tempImage );
	tempImage.convolute1D( diffKernel1, 0, dIy );

	// dxx
	gauss_small2_y.apply(input,tempImage);
	tempImage.convolute1D(diffdiffKernel,1,dIxx);

	// dyy
	gauss_small2_x.apply(input,tempImage);
	tempImage.convolute1D(diffdiffKernel,0,dIyy);
		
	// dxy
	input.convolute1D(diffKernel2,1,tempImage);
	tempImage.convolute1D(diffKernel2,0,dIxy);
	
	const int n = noOfOrientations;
	result.resize( _compute_both_derivs ? 2 : 1, n, h, w );
	result.fill( 0.0f );
	for ( int i = 0; i < n; ++i ) 
	  {
		const double phi   = i*M_PI/n;
		const double sinus = sin(phi);
		const double cosin = cos(phi);
		GPUImage r2( result[0][i] );
		r2.add    ( dIxx,     cosin * cosin );
		r2.add    ( dIxy, 2 * sinus * cosin );
		r2.add    ( dIyy,     sinus * sinus );
		r2.mul( c2 );
		if ( _compute_both_derivs )
		  {
			GPUImage r1( result[1][i] );
			r1.add    ( dIx, cosin );
			r1.add    ( dIy, sinus );
			r1.mul( c1 );
		  }
	  }
	
	if ( gaborLike )
	  result.mul( -1 );

	if ( !_compute_both_derivs )
	  result.removeSingletonDimensions();

}

GPUImage LoGFeatureExtractor::getReconstructedImage(GPUData& input) {
	const int w = input.extent(2);
	const int h = input.extent(1);
	GPUImage rtv(w,h);
	getReconstructedImage(input, rtv);
	return rtv;
}

void LoGFeatureExtractor::getReconstructedImage(GPUData& input, GPUImage& result) {
	const int n = input.extent(0);
	result.fill(0.0f);
	for (int i=0; i<n; ++i) { 
		result.add(input[i]);
	}
}
