/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * LoGFeatureExtractor.hpp
 *
 *  Created on: Sep 22, 2009
 *      Author: jbouecke
 */

#ifndef LOGFEATUREEXTRACTOR_HPP_
#define LOGFEATUREEXTRACTOR_HPP_

#include "FeatureExtractor.h"
#include <GPUData.h>
#include "kernels.hpp"

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

/**
 * This class computes second directional derivatives of the image, using a
 * second derivative of a Gaussian function as a convolution kernel.
 * This kernel function is also known as a Laplacian of
 * Gaussian (LoG). 
 * 
 * @note If gaboLike is true (which is the case by default), the negative of a
 * standard LoG is returned.
 * 
 * The distance \f$ d \f$ between both maxima of the resulting Mexican hat
 * function is a multiple of the standard deviation \f$ \sigma \f$ of the used
 * Gaussian. The distance between both maxima correspond approximately to the
 * wavelength of a similar gabor function.
 * 
 * \f[ d = \sqrt{12} \cdot \sigma \f] 
 * 
 * Other parameters are the number of orientations n.
 */
class LoGFeatureExtractor: public FeatureExtractor {
private:
  int noOfOrientations;
  float sigma1, sigma2, c1, c2;
  
  bool  gaborLike;
  
  int w,h;
  
  GPUData 	 diffKernel1;
  GPUData 	 diffKernel2;
  GPUData 	 diffdiffKernel;
  
  GaussKernel1 gauss_small1_x;
  GaussKernel1 gauss_small1_y;
  GaussKernel1 gauss_small2_x;
  GaussKernel1 gauss_small2_y;
  GPUImage	 tempImage,dIx,dIy,dIxx,dIxy,dIyy;
  bool _compute_both_derivs;
  bool _first;
  void init(int w, int h );
  
public:
  
  LoGFeatureExtractor();
  LoGFeatureExtractor(const int n_);
  LoGFeatureExtractor(const float sigma_);
  LoGFeatureExtractor(const int n_, const float sigma_);
  LoGFeatureExtractor(const int w_, const int h_);
  LoGFeatureExtractor(const int w_, const int h_, const int n_);
  LoGFeatureExtractor(const int w_, const int h_, const int n_, const float sigma_);
  virtual ~LoGFeatureExtractor();
  
  void setNoOfOrientations(const int n) { noOfOrientations = n; _first = true; }
  int  getNoOfOrientations() const { return noOfOrientations; };
  
  /// sets both sigmas equal so that the maxima of the second deriv kernel have the specified distance
  void setDistanceBetweenMaxima(float d) 
  {
	sigma1 = sigma2 = d / sqrt(12.0f);
	_first = true;
  }
  
  /// sets both sigmas equal to the specified value
  void setSigma(float s) 
  {
	sigma1 = sigma2 = s;
	_first = true;
  }
  
  /// sets sigma1 and sigma2 such that the kernels look like gabors with the specified lambda
  void setLambda( float lambda ) 
  {
	// these are the sigmas leading to lambda as expected wavelength in the fourirer transforms of the filters
	sigma1 = lambda /  ( 2.0f * std::sqrt( 2.0f * 3.14f ) );
	sigma2 = lambda /  std::sqrt( std::pow( 3.14f, 3.0f ) / 2.0f );
	// these are the scaling factors that make the positive parts of the fourier transforms sum to one
	c1 = std::sqrt( 2.0f * 3.14f ) * sigma1 * sigma1;
	c2 = 2.0f * sigma2 * sigma2;
	_first = true;
  }
  
  float getDistanceBetweenMaxima() 
  {
	return sigma2 * sqrt(12.0f);
  }
  
  float getSigma() 
  {
	return sigma2;
  }
  
  void setGaborlike( bool b ) { gaborLike = b; }
  
  /// if this is set the feature images will contain both first and second derivatives on the first dim
  void setHaveBothDerivatives( bool b ) { _compute_both_derivs = b; }
  
  /**
   * Returns the number of features at each position.
   */
  virtual int getNoOfFeatures() const { return ( _compute_both_derivs ? 2 : 1 ) * noOfOrientations; }
  
  /**
   * creates for each feature a floating one 2-d slice (image) and
   * returns these as a 3-d GPUData object
   * <code> getNoOfFeatures() </code> length.
   */
  virtual GPUData getFeatureImages( GPUImage& input );
  
  /**
   * the same like getFeatureImagesAsGPUData(const GPUImage& input) but as
   * in place function.
   */
  virtual void getFeatureImages( GPUImage& input, GPUData& result );
  
  /**
   * reconstructs an image out of its features
   */
  virtual GPUImage getReconstructedImage(GPUData& input);
  
  /**
   * same like getReconstructedImage(std::vector<GPUImage>& input) but as in place
   * function
   */
  virtual void getReconstructedImage(GPUData& input, GPUImage& result);
};

#endif /* LOGFEATUREEXTRACTOR_HPP_ */
