/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "LucasKanade.h"
#include <GPUVelField.h>

#include <cmath>
#include <kernels.hpp>

#include <boost/format.hpp>
#include <HostImage.h>
#include <MiddleburyFlowIO.hpp>

inline void LucasKanade::computeDerivatives(
		const GPUImage& imPast,
		const GPUImage& im)
{	
	/* input bluring */
	im.convolute1D(kernel_gauss_small,0,temp1,2);
	temp1.convolute1D(kernel_gauss_small,1,temp2,2);

	/* ---------------- */
	/* saliency part 1a */
	/* ---------------- */ 
	if (!useSimpleSaliency) {   // - - just confidence - - //
		float alpha;
		alpha = 0.4f; temp3.mul(alpha); temp3.add(im,1-alpha); // fast
		alpha = 0.6f; temp4.mul(alpha); temp4.add(im,1-alpha); // slow
		temp5.copyFrom(temp3); temp5.sub(temp4); temp5.abs();
		const float m = floatingAverage.average(temp5.max());
//		std::cout << m << std::endl;
		temp5.div(m);
		temp5.heavisideStepFunction(0.05f);//0.075
	}
	
	/* input gradient computation */
	temp2.convolute1D(kernel_d,1,imdx,2);
	temp2.convolute1D(kernel_d,0,imdy,2);

	/* differentiation in time */
	imdt.copyFrom(temp2);

	imPast.convolute1D(kernel_gauss_small,0,temp1,2);
	temp1.convolute1D(kernel_gauss_small,1,temp2,2);

	imdt.sub(temp2);


	/* ---------------- */
	/* saliency part 1b */
	/* ---------------- */ 
	if (!useSimpleSaliency) {   // - - just confidence - - //
		temp1.copyFrom(imdt);
		temp1.abs();
		temp1.div(floatingAverage2.average(temp1.max()));
		temp1.heavisideStepFunction(0.75);
		
		temp4.mul(temp1);
		
		temp1.sub(1); temp1.abs(); temp1.mul(temp3);
		
		temp4.add(temp1);
	}
	
	/* --------------- */
	/* saliency part 2 */
	/* --------------- */ 
	{   // - - just confidence - - //
		temp1.pythagoras(imdx,imdy);
		temp1.div(floatingAverage3.average(temp1.max()));
		temp1.heavisideStepFunction(0.05f);
		if (useSimpleSaliency) 
			temp5.copyFrom(temp1);
		else
			temp5.mul(temp1);
	}
}

inline void LucasKanade::computeMulAndIntegrate(
				GPUImage& res,
				const GPUImage& a,
				const GPUImage& b,
				int noOfScales)
{
	res.copyFrom(a);
	res.mul(b);
	res.convolute1D(kernel_gauss,0,temp1,2);
	temp1.convolute1D(kernel_gauss,1,res,2);
	res.div(noOfScales);
}

inline void LucasKanade::computeMulAndIntegrate(
				GPUImage& res,
				const GPUImage& a,
				int noOfScales)
{
	res.copyFrom(a);
	res.sqr();
	res.convolute1D(kernel_gauss,0,temp1,2);
	temp1.convolute1D(kernel_gauss,1,res,2);
	res.div(noOfScales);
}

//inline void LucasKanade::computeMulAndIntegrateAdditive(
//				GPUImage& res,
//				const GPUImage& a,
//				const GPUImage& b,
//				int noOfScales)
//{
//	temp2.copyFrom(a);
//	temp2.mul(b);
////	temp2.convolute1D(kernel_gauss_large,0,temp1,2);
////	temp1.convolute1D(kernel_gauss_large,1,temp2,2);
//	temp2.convolute1D(kernel_gauss,0,temp1,2);
//	temp1.convolute1D(kernel_gauss,1,temp2,2);
//	res.add(temp2,1.0f/noOfScales);
//}

inline void LucasKanade::computeMul(
				GPUImage& res,
				const GPUImage& a,
				const GPUImage& b)
{
	res.copyFrom(a);
	res.mul(b);
}

LucasKanade::LucasKanade(int w, int h)
	: res(w,h)
	, floatingAverage(0.1f)
	, floatingAverage2(0.1f)
	, floatingAverage3(0.1f)
	, useSaliencyAsSpeedMask(true)
	, useSimpleSaliency(false) 
{
	{
		const float fraction = 0.7f;
		floatingAverage.setGoodMean(0.11f);
		floatingAverage.setRelativeMinimumMeanValue(0.8f);
		floatingAverage2.setRelativeMinimumMeanValue(fraction);
		floatingAverage3.setRelativeMinimumMeanValue(fraction);
	}
	
	/* ----------------------------------------- */
	/* - init kernel for spacial drivative and - */
	/* -     gaussian integration area         - */
	/* ----------------------------------------- */
	{
		int size = 5;
		kernel_d = GPUData(size);
		kernel_d[0]= -1.0f;
		kernel_d[1]=  8.0f;
		kernel_d[2]=  0.0f;
		kernel_d[3]= -8.0f;
		kernel_d[4]=  1.0f;
		kernel_d.div(12.0f);

		// kernel for initial bluring
		size = 19;
		kernel_gauss_small = GPUData(size);
		fillWithGauss(kernel_gauss_small,3.0f);

		// integration kernel
		size = 21;
		kernel_gauss = GPUData(size);
		//fillWithGauss(kernel_gauss,4.0f);
		kernel_gauss.fill(1);
		

		// so far unused kernel
		size = 31;
		kernel_gauss_large = GPUData(size);
		fillWithGauss(kernel_gauss_large,1.5f*2);
	}

	/* ----------------------------------------- */
	/* -       init temporary variables        - */
	/* ----------------------------------------- */
	{
		temp1= GPUImage(w,h);
		temp2= GPUImage(w,h);
		temp3= GPUImage(w,h); temp3.fill(0);
		temp4= GPUImage(w,h); temp4.fill(0);
		temp5= GPUImage(w,h);

		imdx = GPUImage(w,h);
		imdy = GPUImage(w,h);
		imdt = GPUImage(w,h);

		imGxx = GPUImage(w,h);
		imGxy = GPUImage(w,h);
		imGyy = GPUImage(w,h);

		imGxt = GPUImage(w,h);
		imGyt = GPUImage(w,h);

		//res = GPUVelField(w,h);
		res.getData().fill(0.0f);
		resX  = res.getXPart();
		resY  = res.getYPart();
		
		pastIm = GPUImage(w,h); pastIm.fill(0.0f);
	}

	pastCounter = 0;
}

LucasKanade::~LucasKanade()
{
	//std::cout << "LucasKanade freed" << std::endl;;
}

GPUVelField& LucasKanade::getFlowField() { 
	if (useSaliencyAsSpeedMask) {
		res.getXPart().mul(temp5);
		res.getYPart().mul(temp5);
	}
	return res;
}

GPUVelField& LucasKanade::next(const GPUImage& now) {
	if (pastCounter==0) {
		pastIm.copyFrom(now);
		pastCounter = 1;
		temp3.copyFrom(now);
		temp4.copyFrom(now);
		return res;
	}

	computeStep(pastIm,now);

	{
		pastIm.copyFrom(now);
	}

	return res;
}

void LucasKanade::computeStep(const GPUImage& past, const GPUImage& now)
{
	/* ----------- */
	/* small scale */
	/* ----------- */
	{
		computeDerivatives(past,now);

		computeMulAndIntegrate(imGxx,imdx,1);
		computeMulAndIntegrate(imGxy,imdx,imdy,1);
		computeMulAndIntegrate(imGyy,imdy,1);
		computeMulAndIntegrate(imGxt,imdx,imdt,1);
		computeMulAndIntegrate(imGyt,imdy,imdt,1);
	}

	computeMul(temp1,imGxy,imGxy);
	computeMul(temp2,imGxx,imGyy);
	temp1.add(temp2,-1.0f); temp1.add(0.0000001f);
	temp2.fill(0.0f); temp2.add(1.0f);
	temp2.div(temp1);

	computeMul(resX ,imGyy,imGxt);
	computeMul(temp1,imGxy,imGyt);
	resX.add(temp1,-1.0f);

	computeMul(resY ,imGxx,imGyt);
	computeMul(temp1,imGxy,imGxt);
	resY.add(temp1,-1.0f);

	resX.mul(temp2);
	resY.mul(temp2);
}

namespace bo = boost;

void LucasKanade::go(std::string filePattern, int startIdx, int stopIdx) {

	//std::vector<GPUVelField> results;
	
	std::string inputFile = str(bo::format(filePattern) % startIdx);
	
	HostImage h_image(inputFile.c_str());
	
	const int w = h_image.getWidth();
	const int h = h_image.getHeight();
	
	LucasKanade lk(w,h);
	
	lk.next(h_image); // first input (reference frame for first flow)

	for (int i=startIdx+1; i<=stopIdx; ++i) {
		inputFile = str(bo::format(filePattern) % (i));
		GPUVelField& r = lk.next(HostImage(inputFile));
		MiddleburyFlowIO::WriteFlowFile(r,lk.getConfidence(),str(boost::format("flow%05d.flo")%(i-1)));
		//results.push_back(lk.next(HostImage(inputFile)).copy());
	}
}
