/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "LucasKanade.h"
#include "SimpleHostImageSequence.h"
#include "MiddleburyFlowIO.hpp"
#include "JDBFlowIO.hpp"

#define START_WATCH(name) clock_t stop_watch_##name##_start = clock();
#define STOP_WATCH(name) \
  clock_t stop_watch_##name##_stop = clock(); \
  std::cout << "Elapsed time for " #name ": " \
            << (((double)1000.0*(stop_watch_##name##_stop - stop_watch_##name##_start)) / CLOCKS_PER_SEC) \
            << " ms" << std::endl;


enum OUTPUT_DATA { NONE=0, FLOW=1, FLO=1<<2 };

void 
lk (std::string file_pattern, int start, int stop, int out = NONE, bool use_saliency = false, bool use_simple_saliency = false)
{	
	SimpleHostImageSequence imageSequence(file_pattern.c_str(),start,stop);
	
	const int w = imageSequence.getWidth();
	const int h = imageSequence.getHeight();

	GPUImage image(w,h);
	GPUImage imageA(w,h);
	GPUImage imageB(w,h);

	GPUImage* imPtr01 = &imageA;
	GPUImage* imPtr02 = &imageB;

	std::cout << "image resolution: " << w << " x " << h << std::endl;

	int i=0;

	LucasKanade lk(w,h);
				lk.useSaliencyMaskOnSpeed(use_saliency||use_simple_saliency);
				lk.switchToSimpleSaliency(use_simple_saliency);

	/* ---------------------------------- */
	/*        load the first image        */
	/* ---------------------------------- */
	imageSequence.copyNextImage(*imPtr02); imPtr02->div(imPtr02->max());
	lk.next(*imPtr02);

	START_WATCH(lk);

	while (imageSequence.hasNextImage()) {

		printf("#%d\n",i);
		
		/* ------------------------------------ */
		/*         swap images images           */
		/* ------------------------------------ */
		std::swap(imPtr01,imPtr02);

		/* ------------------------------------ */
		/* read new current image               */
		/* ------------------------------------ */
		imageSequence.copyNextImage(*imPtr02); imPtr02->div(imPtr02->max());

		/* ----------------------------------- */
		/* create initial motion estimation    */
		/* ----------------------------------- */
		lk.next(*imPtr02); i++;

		if (out & FLOW) {
			std::string filename = str(boost::format("LK_flow_%05d.flow") % (i-1));
			JDBFlowIO::WriteFlowFile(lk.getFlowField(),lk.getConfidence(),filename);
		}
		if (out & FLO) {
			std::string filename = str(boost::format("LK_flow_%05d.flo") % (i-1));
			if (use_saliency)
				MiddleburyFlowIO::WriteFlowFile(lk.getFlowField(),lk.getConfidence(),filename);
			else
				MiddleburyFlowIO::WriteFlowFile(lk.getFlowField(),filename);
		}		
	}

	STOP_WATCH(lk);

	printf("\n\niterated %d images\n",i);
}


int 
main (int argc, char **argv)
{
	if (argc < 4) {
		std::cout << "usage: LucasKanade file_pattern start_index stop_index [options]" << std::endl;
		std::cout << "  file_pattern    the path and pattern for files, e.g. images/yos_%02d.tif" << std::endl;
		std::cout << "  start_index     starting index for substitution in file_pattern" << std::endl;
		std::cout << "  stop_index      stop index for substitution in file_pattern" << std::endl;
		std::cout << "  [options]       = {flo|flow|use_saliency|use_simple_saliency} " << std::endl;
		return EXIT_FAILURE;
	}
	if (argc > 4) {
		int  output = NONE;
		bool use_saliency = false;
		bool use_simple_saliency = false;
		for (int i=4; i<argc; i++) {
			std::string out(argv[i]);
			if (out.find("flow")!=std::string::npos || out.find("FLOW")!=std::string::npos) {
				output = output | FLOW;
				std::cout << "enabled .flow output"<< std::endl;
				continue;
			}
			if (out.find("flo")!=std::string::npos  || out.find("FLO")!=std::string::npos) { // TODO: this will be found with a 'flow' switch, too...
				output = output | FLO;
				std::cout << "enabled .flo output"<< std::endl;
				continue;
			}
			if (out.find("with")!=std::string::npos || out.find("use_saliency")!=std::string::npos) {
				std::cout << "enabled saliency mask"<< std::endl;
				use_saliency = true;
				continue;
			}
			if (out.find("with_simple")!=std::string::npos || out.find("use_simple_saliency")!=std::string::npos) {
				std::cout << "enabled simple saliency mask"<< std::endl;
				use_simple_saliency = true;				
				continue;
			}
		}
		lk(argv[1], atoi(argv[2]), atoi(argv[3]), output, use_saliency, use_simple_saliency);
	} else {
		lk(argv[1], atoi(argv[2]), atoi(argv[3]));
	}
	return EXIT_SUCCESS;
}

