/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * FlowIO.hpp
 *
 *  Created on: Feb 25, 2009
 *      Author: jbouecke
 */

#ifndef FLOWIO_HPP_
#define FLOWIO_HPP_

#include <GPUVelField.h>

/**
 *
 * read and write simple .flo flow file format
 *
 * ".flo" file format used for optical flow evaluation
 *
 * the following text and code is adapted from example code of:
 * http://vision.middlebury.edu/flow/data/
 *
 * Stores 2-band float image for horizontal (u) and vertical (v) flow components.
 * Floats are stored in little-endian order.
 * A flow value is considered "unknown" if either |u| or |v| is greater than 1e9.
 *
 *  bytes  contents
 *
 *  0-3     tag: "PIEH" in ASCII, which in little endian happens to be the float 202021.25
 *          (just a sanity check that floats are represented correctly)
 *  4-7     width as an integer
 *  8-11    height as an integer
 *  12-end  data (width*height*2*4 bytes total)
 *          the float values for u and v, interleaved, in row order, i.e.,
 *          u[row0,col0], v[row0,col0], u[row0,col1], v[row0,col1], ...
 */
class MiddleburyFlowIO {
private:
	static bool unknown_flow(float u, float v);
    static bool unknown_flow(float *f);
public:
#ifdef _WIN32
	MiddleburyFlowIO();
#endif

	/**
	 * the "official" threshold - if the absolute value of either
	 * flow component is greater, it's considered unknown
	 */
#ifdef _WIN32
	static const float MiddleburyFlowIO::UNKNOWN_FLOW_THRESH;
#else
	static const float UNKNOWN_FLOW_THRESH = 1e9;
#endif
	

	/// value to use to represent unknown flow
#ifdef _WIN32
	static const float MiddleburyFlowIO::UNKNOWN_FLOW;
#else
	static const float UNKNOWN_FLOW = 1e10;
#endif
	

	static void WriteFlowFile(const GPUVelField& flow, const char* filename);
	static void WriteFlowFile(const GPUVelField& flow, const GPUImage& conf, const char* filename, const float confthreshold = 0.5f);
	static void WriteFlowFile(const GPUVelField& flow, const std::string& filename) {
		WriteFlowFile(flow,filename.c_str());
	}
	static void WriteFlowFile(const GPUVelField& flow, const GPUImage& conf, const std::string& filename, const float confthreshold = 0.5f) {
		WriteFlowFile(flow,conf,filename.c_str(),confthreshold);
	}
	static void WriteFlowFile(int width, int height, float* img, const char* filename);
	/**
	 * reads the flow information from the given file name
	 *
	 * The function allocates the needed memory and the calling code has to
	 * care for deleting.
	 *
	 * @param filename name of the .flo file
	 * @param w return the with
	 * @param h return the height
	 * @return the pointer to the read data
	 */
	static float* ReadFlowFile(const char* filename, int& w, int& h);
};

#endif /* FLOWIO_HPP_ */
