/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cublas.h>
#include <cmath>
#include <string.h>

#include "GPUImage.h"
#include "ModelArea.hpp"
#include "kernels.hpp"

ModelArea::ModelArea()
	: _n_dip_ori_neigh(0)
	,_feedback_gain( 0.5f )
	, _divConst( 0.394f )
	, _divFactor( 5.0f )
	, _divSelf( 0.01f )
	, _subFactor( 0.0f )
	, _do_square(true)
	, _use_forward_inhibition(true)
	, _rect_inria_style_input(false)
	, _dt(-1)
	, _sigma_for_spatial_inhibition( 1.0f )
	, _kernel_settings_changed( false )
	, _kernel_type( NONE )
{
	_sigma[0] = _sigma[1] = 0;
	_sigma[2] = _sigma[3] = 0;
	_sigma[4] = 0;
	_typeBorder[0] = _typeBorder[1] = 2;
	_typeBorder[2] = _typeBorder[3] = 2;
	_typeBorder[4] = 2;
	
}

ModelArea::~ModelArea()
{}

void ModelArea::set_shunting_params( float dc, float df, float ds, float sf )
{
	_divConst = dc;
	_divFactor = df;
	_divSelf = ds;
	_subFactor = sf;
}

void ModelArea::set_feedback_gain( float g )
{
	_feedback_gain = g;
}

void ModelArea::set_sigma_for_spatial_inhibition(float sigma)
{
	if ( _sigma_for_spatial_inhibition == sigma ) return;
	_sigma_for_spatial_inhibition = sigma;
	_kernel_settings_changed = true;
}

void ModelArea::set_sigma( int dim, float sigma )
{
	if ( _sigma[ dim ] == sigma ) return;
	_sigma[ dim ] = sigma;
	_kernel_settings_changed = true;
}

void ModelArea::set_typeBorder( int dim, int typeBorder )
{
	if ( _typeBorder[ dim ] == typeBorder ) return;
	_typeBorder[ dim ] = typeBorder;
	_kernel_settings_changed = true;
}

void ModelArea::set_kernel_type( KernelType ktype )
{
	if ( _kernel_type == ktype ) return;
	_kernel_type = ktype;
	_kernel_settings_changed = true;
}

void ModelArea::set_do_square( bool b )
{
	_do_square = b;
}

void ModelArea::set_dt( float dt )
{
	_dt = dt;
}

void ModelArea::set_ori_neigh( unsigned n )
{
	if ( _n_dip_ori_neigh == n ) return;
	_n_dip_ori_neigh = n;
	_kernel_settings_changed = true;
}

void ModelArea::set_use_recurrent_inhibition(bool b) {
	_use_forward_inhibition = !b;
}
void ModelArea::set_use_rectified_simple_input(bool b) {
	_rect_inria_style_input = b;
	if (b && _divFactor!=_subFactor) {
		printf("WARNING: rectified simple input is only possible"
					 "for some special parameters! (error in"
					 "ModelArea::set_use_rectified_simple_input()"
					 ")%f,%f",_divFactor,_subFactor);
	}
}
void ModelArea:: set_INRIA_dynamic( float exitat, float inhib, float leak ) {
	_divConst = leak;
	_divFactor= inhib;
	_subFactor= inhib;
	_divSelf  = exitat;
	set_use_recurrent_inhibition(true);
	set_use_rectified_simple_input(true);
}

void ModelArea::set_modulator( const GPUData & modulator )
{
	if ( _kernel )
		_kernel->set_modulator(modulator);
}

///builds kernels and sets up buffers
void ModelArea::init( GPUData & input )
{
	
	if ( _kernel_settings_changed )
		{
			switch ( _kernel_type )
				{
				case GAUSS:
					{
						MultiKernel * kernel = new MultiKernel;
						int n_sigmas = sizeof( _sigma ) / sizeof( _sigma[0] );
						for ( int i = 0; i < n_sigmas; ++i )
							if ( _sigma[i] > 0 )
								{
									kernel->add_kernel( new GaussKernel1(_sigma[i], i, -1, _typeBorder[i] ) );
								}
						_kernel.reset( kernel );
					}
					break;
				case DIPOLE:
					{
						MultiKernel * kernel = new MultiKernel;
						kernel->add_kernel( new DiPoleKernel(_sigma[0],_sigma[1],_sigma[2],_n_dip_ori_neigh) );
						_kernel.reset( kernel );
					}
					break;
				case MODULATED_DIFFUSION:
					{
						_kernel.reset( new ModulatedDiffusionKernel( _sigma[0], _sigma[1], _sigma[2] ) );
					}
					break;
				case NONE:
					_kernel.reset();
					break;
				default:
					throw ( std::runtime_error( "unsupported diffusion kernel" ) );
				}
			if (_sigma_for_spatial_inhibition>0.0f) {
				MultiKernel* k = new MultiKernel;
				k->add_kernel( new GaussKernel1( _sigma_for_spatial_inhibition, 0 ) );
				k->add_kernel( new GaussKernel1( _sigma_for_spatial_inhibition, 1 ) );
				_kernel2.reset(k);
			} else {
				_kernel2.reset();
			}
			_kernel_settings_changed = false;
		}
	
	const size_t * in_dims = input.getDimensions(), rank = input.getDimensionsCount();
	size_t nf = 1;
	for ( size_t i = 2; i < rank; ++i) nf *= in_dims[i];
	
	// create a vector of ones to easily compute the sum via matrix product
	if ( _ones.size() != nf )
		{
			_ones = GPUData( nf );
			_ones.fill( 1.0f );
		}
	
	// create memory for activity
	if ( _dt>0 && _activity.size()!=input.size() )
		{ // init temporary
			_activity.resize(input.getNoOfDimensions(),input.getSizes());
			_activity.fill(0.0f);
			//	  _activity.copyFrom(input);
			//	  _activity.mul(0.75f);
		}
	
	// create temp for inhibition
	if ( _tmp_inhibit.rank() != 2 || input.extent(0) != _tmp_inhibit.extent(0) || input.extent(1) != _tmp_inhibit.extent(1) )
		_tmp_inhibit  = GPUData( 2, in_dims );
	if ( _tmp_inhibit2.rank() != 2 || input.extent(0) != _tmp_inhibit2.extent(0) || input.extent(1) != _tmp_inhibit2.extent(1) )
		_tmp_inhibit2 = GPUData( 2, in_dims );
}

extern "C" void inhibitionSpecial(const int Nx, const int Ny, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor);
extern "C" void inhibitionSpecial2(const int Nx, const int Ny, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor, const float divSelfInhibFactor);
extern "C" void do_fbSpecial(const int N, float* res, float* ff, float* fb, const float c);

// #include "GPUDataDebugging.hpp"

void ModelArea::iterate( GPUData & input, GPUData & output, GPUData feedback, bool pred_vel_shift_fb )
{
	
	
	bool have_feedback = ( feedback.size() > 0 ) && ( _feedback_gain != 0.0f );
	const size_t * in_dims = input.getDimensions(), rank = input.getDimensionsCount();
	if ( rank < 2 ) throw( std::runtime_error( "ModelArea needs at least 2d data" ) );
	int npixels = in_dims[0] * in_dims[1];
	int nf = 1;
	for ( size_t i = 2; i < rank; ++i ) nf *= in_dims[i];
	
	// lazily prepare buffers and kernel
	init( input );
	
	output.resize( input.rank(), input.getDimensions() );
	
	if ( ! input.is_same_shape_as( output ) )
		throw std::runtime_error("ModelArea: input output have different shape");
	
	
	// diffusion
	if ( _kernel )
		{
			_kernel->apply( input, output );
			input.swap(output);
		}
	
	// feedback
	if ( have_feedback )
		{
			if ( feedback.size() == input.size() )
				{
					do_fbSpecial( input.size(), input.data(), input.data(), feedback.data(), _feedback_gain );
				}
			else if ( feedback.size() < input.size() )
				{
					// TODO: test this and factorize out
					size_t n1 = input.extent(0) / feedback.extent(0);
					size_t n2 = input.extent(1) / feedback.extent(1);
					size_t d1 = ( input.extent(0) - feedback.extent(0) * n1 ) / 2;
					size_t d2 = ( input.extent(1) - feedback.extent(1) * n2 ) / 2;
					if ( pred_vel_shift_fb )
						{
							size_t nu = input.extent(2);
							size_t nv = input.extent(3);
							//TODO: checks
							for ( size_t i = 0; i < n1; ++i )
								for ( size_t j = 0; j < n2; ++j )
									for ( size_t u = 0; u < nu; ++u )
										for ( size_t v = 0; v < nv; ++v )
											{
												GPUData fb_shifted( feedback.slice(3,v).slice(2,u) );
												GPUData in_shifted( input   .slice(3,v).slice(2,u) );
												
												int du = nu / 2 - u;
												int dv = nv / 2 - v;
												// do the mod on a shifted view, cutting edgeds where neccessary
												int    o1 = d1 + i - du;
												int    o2 = d2 + j - dv;
												size_t delta1 = 0;
												size_t delta2 = 0;
												if ( o1 <  0  )
													{
														delta1 = 1 + size_t(-o1) / n1;
														o1  += delta1 * n1;
													}
												if ( o2 <  0  )
													{
														delta2 = 1 + size_t(-o2) / n2;
														o2  += delta2 * n2;
													}
												
												in_shifted = in_shifted.range( 0, size_t(o1), in_shifted.extent(0) - 1 - n1 * delta1, n1 );
												in_shifted = in_shifted.range( 1, size_t(o2), in_shifted.extent(1) - 1 - n2 * delta2, n2 );
												fb_shifted = fb_shifted.range( 0,     delta1, fb_shifted.extent(0) - 1                   );
												fb_shifted = fb_shifted.range( 1,     delta2, fb_shifted.extent(1) - 1                   );
												
												if ( in_shifted.extent(0) > fb_shifted.extent(0) )
													in_shifted = in_shifted.range( 0, 0, fb_shifted.extent(0) - 1 );
												else if ( in_shifted.extent(0) < fb_shifted.extent(0) )
													fb_shifted = fb_shifted.range( 0, 0, in_shifted.extent(0) - 1 );
												
												if ( in_shifted.extent(1) > fb_shifted.extent(1) )
													in_shifted = in_shifted.range( 1, 0, fb_shifted.extent(1) - 1 );
												
												else if ( in_shifted.extent(1) < fb_shifted.extent(1) )
													fb_shifted = fb_shifted.range( 1, 0, in_shifted.extent(1) - 1 );
												
												
												in_shifted.fb_mod( fb_shifted, _feedback_gain );
											}
						}
					else
						{
							for ( size_t i = 0; i < n1; ++i )
								{
									for ( size_t j = 0; j < n2; ++j )
										{
											size_t o1 = size_t( d1 + i );
											size_t o2 = size_t( d2 + j );
											GPUData feedbackrange = feedback;
											GPUData inputrange = input
												.range( 0, o1, input.extent(0) - 1 - d1, n1 )
												.range( 1, o2, input.extent(1) - 1 - d2, n2 );
											
											if ( inputrange.extent(0) > feedbackrange.extent(0) )
												inputrange    = inputrange   .range( 0, 0, feedbackrange.extent(0) - 1 );
											else
												feedbackrange = feedbackrange.range( 0, 0, inputrange   .extent(0) - 1 );
											if ( inputrange.extent(1) > feedbackrange.extent(1) )
												inputrange    = inputrange   .range( 1, 0, feedbackrange.extent(1) - 1 );
											else
												feedbackrange = feedbackrange.range( 1, 0, inputrange   .extent(1) - 1 );
											
											inputrange.fb_mod( feedbackrange, _feedback_gain );
										}
								}
						}
				}
			else
				{
					throw GPUDataException("feedback must be sized <= input");
				}
		}
	
	//sharpening - potentially unstable
	if ( _do_square)
		input.sqr();
	
	size_t sizes[2] = {npixels,nf};
	int strid[2] = {nf,1};
	GPUImage xv(_dt>0 ? _activity.view(2,sizes,strid) : input.view(2,sizes,strid));  // dynamic var as 2d view
	GPUImage inp(input.view(2,sizes,strid)); // input as 2d view
	GPUData  tm(_tmp_inhibit.reduceDimensionality());
	
	// shunting
	/* sum the whole feature space via matrix one-vector multiplication */
	if (_use_forward_inhibition)
		_tmp_inhibit.reference( input.sumContract( 0, 1 ) );
	else
		_tmp_inhibit.reference( (_dt>0?_activity:input).sumContract( 0, 1 ) );
	
	if (_kernel2) 
		{
			_kernel2->apply(_tmp_inhibit,_tmp_inhibit2);
			_tmp_inhibit.swap(_tmp_inhibit2);
			tm = _tmp_inhibit.reduceDimensionality();
		}
	
	/* takes the mean activity */
	tm.mul( 1.0f / nf );
	
	if (_subFactor==_divFactor) { // simpler ode
		if (_dt<=0.0f)
			{
				xv.linFirstOrderODESteayStateSolution(_divConst ,
																							_divSelf, inp,
																							-1*_subFactor, tm,
																							_rect_inria_style_input);
				if (!_rect_inria_style_input) input.rect();
			}
		else
			{
				xv.linFirstOrderODEImplicitEulerStep( _divConst ,
																							//		  xv.linFirstOrderODEEulerStep( _divConst ,
																							_divSelf, inp,
																							-1*_subFactor, tm,
																							_dt,_rect_inria_style_input);
				inp.copyFrom(xv);
			}
	}
	else
		{ // more general ode
			if (_dt<=0.0f) {
				xv.linFirstOrderODESteayStateSolution(-1*_divSelf  , inp,
																							-1*_divFactor, tm,
																							-1*_divConst ,
																							+1.0f        , inp,
																							-1*_subFactor, tm,
																							+0);
				if (_subFactor>0.0f) input.rect();
			}
			else
				{
					xv.linFirstOrderODEImplicitEulerStep( -1*_divSelf  , inp,
																								-1*_divFactor, tm,
																								-1*_divConst ,
																								+1.0f        , inp,
																								-1*_subFactor, tm,
																								+0,
																								_dt);
					inp.copyFrom(xv);
				}
		}
	
	// put output to output (input is trashed...)
	output.swap( input );
}

// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 2
// tab-width: 2
// End:
