/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*
 * ModulatedIntegration.cpp
 *
 *  Created on: Feb 11, 2009
 *      Author: jbouecke
 */

#include "ModulatedIntegration.hpp"

#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif

ModulatedIntegration::ModulatedIntegration()
	: _sigma_small(0.8f)
	, _sigma_large(2.0f)
	, _angle_offset(M_PI/2)
{
	;
}

ModulatedIntegration::ModulatedIntegration(float sigma_large, float sigma_small, float sigma_rest)
	: _sigma_small(sigma_small)
	, _sigma_large(sigma_large)
	, _sigma_rest(sigma_rest)
	, _angle_offset(M_PI/2)
{
	;
}

ModulatedIntegration::~ModulatedIntegration() {
	;
}

void ModulatedIntegration::init(size_t w, size_t h, size_t o)
{
	if ( w != _temp1.getWidth() || h != _temp1.getHeight()) {
		_temp1 = GPUImage(w,h);
		_temp2 = GPUImage(w,h);
		_temp3 = GPUImage(w,h);
		_temp4 = GPUImage(w,h);
		_temp5 = GPUImage(w,h);
	}
	
	if ( o > 0 && o != _filters.size() ) 
	  {
		_filters.clear();
		for ( size_t i = 0; i < o; ++i ) 
		  {
			float angle = _angle_offset + i * M_PI / o;
			size_t w_ = size_t( abs( 5.0f * _sigma_large * cos( angle ) ) );
			size_t h_ = size_t( abs( 5.0f * _sigma_large * sin( angle ) ) );
			if ( w_ % 2 == 0 ) w_++;
			if ( h_ % 2 == 0 ) h_++;
			GPUImage im( w_, h_ );
			fillWithGauss2Rotated( im, _sigma_large, _sigma_small, angle );
			_filters.push_back( im );
		  }
		cirularGauss       = GaussKernel2( _sigma_rest  );
		cirularGauss_large = GaussKernel2( _sigma_large );
	}
}

void ModulatedIntegration::setInput(GPUData & data)
{
	GPUData input(data);
	if (data.getDimensionsCount()!=4) {
		throw std::runtime_error("input has"
				"to be 4 dimensional");
	}
	init(input.extent(1),input.extent(0)); // w,h
	_input =  input;
}

void ModulatedIntegration::setModulator(GPUData & data)
{
	GPUData input(data);
	if (data.getDimensionsCount()>3) {
		size_t d = data.getDimensionsCount() - 2;
		input = input.reduceDimensionality(d);
		if (input.getDimensionsCount()!=3) throw std::runtime_error("input has"
				"to be rectable to 3 dimensions");
	} else if (data.getDimensionsCount()<3) {
		throw std::runtime_error("input has to be at least 3 dimensional");
	}
	init(input.extent(2),input.extent(1),input.extent(0)); // w,h
	_modulator = input;
}


void ModulatedIntegration::work()
{
	const unsigned int noOfOri = _modulator.extent(0);

//	_modulator.abs();
//
//	_temp1.fill(0.0f);
//	_temp2.fill(0.0f);
//	for (unsigned int o = 0; o<noOfOri; ++o) {
//		_temp1.max(_modulator[o]);
//		_temp2.add(_modulator[o]);
//	}
//
//	_temp2.mul(-1); _temp2.add(_temp1,noOfOri); // sum of (I_i-I_max)
//	_temp2.add(0.000001f);
//
//	_modulator.mul(-1);
//	for (unsigned int o = 0; o<noOfOri; ++o) {
//		_modulator[o].add(_temp1);// add maximum
//		_modulator[o].div(_temp2);// div sum of all
//	}
//
//	GPUData data(_input.shiftDim(2));
//	const int V = data.extent(0);
//	const int U = data.extent(1);
//	const int O = _filters.extent(0);
//	for (int v=0; v<V; ++v) {
//		for (int u=0; u<U; ++u) {
//			_temp2.copyFrom(data[v][u]);
//			_temp4.fill(0.0f);
//			for (int o=0; o<O; o++) {
//				GPUImage filter(_filters[o]);
//				_temp2.convolute2D(filter,_temp3);
//				_temp3.mul(_modulator[o]);
//				_temp4.add(_temp3);
//			}
//			_temp4.mul(_temp1);
//			_temp2.mul(_temp5);
//			_temp2.add(_temp4);
//			data[v][u].copyFrom(_temp2);
//		}
//	}

	_modulator.abs();
	_temp1.fill(0.0f);
#if 0
	for (unsigned int o = 0; o<noOfOri; ++o) {
		GPUImage ref(_modulator[o]);
		_temp2.copyFrom(ref);
		cirularGauss_large.apply(_temp2,ref);
		cirularGauss_large.apply(ref,_temp2);
		ref.copyFrom(_temp2);
		_temp1.add(ref);
	}
#else
	for (unsigned int o = 0; o<noOfOri; ++o) {
		_temp1.add(_modulator[o]);
	}
#endif
	_temp1.add(0.000001f);
	for (unsigned int o = 0; o<noOfOri; ++o) {
		_modulator[o].div(_temp1);
	}
	_temp1.div(_temp1.max());
	_temp5.fill(1.0f);
	_temp5.sub(_temp1);

	GPUData data(_input.shiftDim(2));
	const int V = data.extent(0);
	const int U = data.extent(1);
	const int O = _filters.size();
	for (int v=0; v<V; ++v) {
		for (int u=0; u<U; ++u) {
			_temp2.copyFrom(data[v][u]);
			_temp4.fill(0.0f);
			for (int o=0; o<O; o++) {
				GPUImage filter(_filters[o]);
				_temp2.convolute2D(filter,_temp3);
				_temp3.mul(_modulator[o]);
				_temp4.add(_temp3);
			}
#if 1
			cirularGauss.apply(_temp2,_temp3);
			_temp4.mul(_temp1);
			_temp3.mul(_temp5);
			_temp3.add(_temp4);
			data[v][u].copyFrom(_temp3);
#else
			_temp4.mul(_temp1);
			_temp2.mul(_temp5);
			_temp2.add(_temp4);
			data[v][u].copyFrom(_temp2);
#endif
		}
	}
}


