/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "RHOGFeatureExtractor.hpp"

int RHOGFeatureExtractor::getNoOfFeatures() const { return _n_orientations; }

GPUData RHOGFeatureExtractor::getFeatureImages( GPUImage& input )
{
  GPUData result;
  getFeatureImages( input, result );
  return result;
}
  
extern "C" void do_R_HOG_grey( const float * in, int in_n1, int in_n2
							   , float * out, int cell_size, int block_size
							   , int block_overlap, int n_orientations
							   , bool interpolate, float alpha );

extern "C" void do_felzenszwalb_RHOG_grey( const float * in, int in_n1, int in_n2, float * out
										   , int cell_size, int n_orientations, float alpha
										   , bool contrast_sensitive );

extern "C" void do_felzenszwalb_HOG_grey( const float * in, int in_n1, int in_n2, float * out
										  , int cell_size, int n_orientations, float alpha );


extern "C" void do_R_HOG_grey_tex( const cudaArray * in
								   , float * out, int out_n1, int out_n2
								   , int cell_size, int block_size
								   , int block_overlap, int n_orientations
								   , bool interpolate, float alpha
								   , float xscale, float yscale );

void RHOGFeatureExtractor::getFeatureImages( GPUImage& input, GPUData& result )
{
  int in_n1 = input.getYRes();
  int in_n2 = input.getXRes();
  if ( !_felzenszwalb )
	{
	  const int block_stride = ( _block_size - _block_overlap ) * _cell_size;
	  const int block_extent = _block_size * _cell_size;
	  if ( in_n1 < block_extent + 2 ||  in_n2 < block_extent + 2 )
		{
		  result.clear();
		}
	  else
		{
#if 0
		  const int n1_blocks = 1 + ( in_n1 - block_extent - 2 ) / block_stride;
		  const int n2_blocks = 1 + ( in_n2 - block_extent - 2 ) / block_stride;
		  result.resize( n1_blocks * _block_size, n2_blocks * _block_size, _n_orientations );
		  do_R_HOG_grey( input.data(), in_n1, in_n2, 
						 result.data(), _cell_size, _block_size
						 , _block_overlap, _n_orientations
						 , _interpolate, _alpha );
#else
		  in_n1 = int( in_n1 * _yscale + 0.5f );
		  in_n2 = int( in_n2 * _xscale + 0.5f );
		  const int n1_blocks = 1 + ( in_n1 - block_extent - 2 ) / block_stride;
		  const int n2_blocks = 1 + ( in_n2 - block_extent - 2 ) / block_stride;
		  result.resize( n1_blocks * _block_size, n2_blocks * _block_size, _n_orientations );
		  input.move_to_array();
		  do_R_HOG_grey_tex( input.array(), result.data(), result.extent(0), result.extent(1)
							 ,_cell_size, _block_size
							 , _block_overlap, _n_orientations
							 , _interpolate, _alpha
							 , _xscale, _yscale );
#endif
		}
	}
  else
	{
	  const int n1_cells_in = ( in_n1 - 2 ) / _cell_size;
	  const int n2_cells_in = ( in_n2 - 2 ) / _cell_size;
	  if ( n1_cells_in < 4 || n2_cells_in < 4 )
		{
		  result.clear();
		}
	  else
		{
		  if ( _contrast_sensitive )
			{
			  result.resize( n1_cells_in - 2, n2_cells_in - 2, 3 * _n_orientations + 4 );
			  do_felzenszwalb_HOG_grey( input.data(), in_n1, in_n2, result.data(), _cell_size, _n_orientations, _alpha );
			}
		  else
			{
			  result.resize( n1_cells_in - 2, n2_cells_in - 2, _n_orientations + 4 );
			  do_felzenszwalb_RHOG_grey( input.data(), in_n1, in_n2, result.data(), _cell_size, _n_orientations, _alpha, _contrast_sensitive );
			}
		}
	}
}

GPUImage RHOGFeatureExtractor::getReconstructedImage(GPUData& input)
{
  throw std::runtime_error("reconstruction not supported by hog features");
}
  
void RHOGFeatureExtractor::getReconstructedImage(GPUData& input, GPUImage& result)
{
  throw std::runtime_error("reconstruction not supported by hog features");
}
  
RHOGFeatureExtractor::~RHOGFeatureExtractor()
{
}
  
