/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "SepGaborWaveletFilter.h"
#include <cmath>
#include "kernels.hpp"

void SepGaborWaveletFilter::initFilter() {

	float omega = 2*3.14159265359f/lambda;

	dcOffsetFaktor = exp(-0.5*omega*omega*sigma*sigma);

	int size = (int)ceil(sigma*5.01f);
	size = (size%2==0) ? size+1 : size;

	//System.out.print(" -------------------"+size+"------------------- ");
	cosX = EvenGaborKernel1(true,omega*cos(theta),sigma,1,size);
	cosY = EvenGaborKernel1(true,omega*sin(theta),sigma,0,size);
	sinX = OddGaborKernel1( true,omega*cos(theta),sigma,1,size);
	sinY = OddGaborKernel1( true,omega*sin(theta),sigma,0,size);

	gaussX = GaussKernel1(sigma,1,size);
	gaussY = GaussKernel1(sigma,0,size);

}

void SepGaborWaveletFilter::init4Temp(int x, int y) {
	temp1 = GPUImage(x,y);

	temp2 = GPUImage(x,y);

	temp3 = GPUImage(x,y);

	temp4 = GPUImage(x,y);

	isTempShared = false;
}

void SepGaborWaveletFilter::initDCOffsestFactor(bool justCheck) {

	if (!justCheck)
		dcOffsetFaktor = 0;

	const int x = temp1.getWidth();
	const int y = temp1.getHeight();
	GPUImage im1(x,y);
	GPUImage im2(x,y);

	im1.fill(0.0f);
	im1[y/2][x/2]=1.0f;

	filterEven(im1,im2);

	float sumL = im2.sum();

	if (!justCheck) {
		//std::cout << "correct offset with : " << sumL-exp(-0.5*omega*omega*sigma*sigma) << std::endl;
		dcOffsetFaktor = sumL;
	} else {
		//std::cout << "dc offset : " << sumL << std::endl;
		//im2.show();
	}
}

SepGaborWaveletFilter::SepGaborWaveletFilter(const GaborFilter2DParameter& param) {
	ratio = param.sigma/param.lambda;
	lambda= param.lambda;
	theta = param.theta;
	sigma = param.sigma;

	initFilter();
	init4Temp(320,240);
	//initDCOffsestFactor(true);
	initDCOffsestFactor(false);
	//initDCOffsestFactor(true);
}

SepGaborWaveletFilter::SepGaborWaveletFilter(int x, int y) {
	
	useDCCompensation = true;
	
	ratio = 1.0f;
	lambda= 5.0f;
	theta = 0.0/360.0*2*3.14159;
	sigma = ratio*lambda;

	initFilter();
	init4Temp(x,y);
	//initDCOffsestFactor(true);
	initDCOffsestFactor(false);
	//initDCOffsestFactor(true);
}

SepGaborWaveletFilter::SepGaborWaveletFilter( const SepGaborWaveletFilter& partner) {
	*this = partner;
}

//SepGaborWaveletFilter::SepGaborWaveletFilter(SepGaborWaveletFilter& partner) {
//	ratio = 1;
//	lambda= 5.0f;
//	theta = 0.0/360.0*2*3.14159;
//	sigma = ratio*lambda;
//
//	initFilter();
//	sharedTemp(partner);
//	//initDCOffsestFactor(true);
//	initDCOffsestFactor(false);
//	//initDCOffsestFactor(true);
//}

SepGaborWaveletFilter::SepGaborWaveletFilter(int x, int y, double rad, double lambda, float rat) {
	useDCCompensation = true;
	
	ratio = rat;
	this->lambda = lambda;
	theta = rad;
	sigma = ratio*lambda;

	init4Temp(x,y);
	initFilter();
	//initDCOffsestFactor(true);
	initDCOffsestFactor(false);
	//initDCOffsestFactor(true);
}

SepGaborWaveletFilter::SepGaborWaveletFilter(SepGaborWaveletFilter& partner, double rad, double lambda, float rat) {
	useDCCompensation = true;
	
	ratio = rat;
	this->lambda = lambda;
	theta = rad;
	sigma = ratio*lambda;

	sharedTemp(partner);
	initFilter();
	//initDCOffsestFactor(true);
	initDCOffsestFactor(false);
	//initDCOffsestFactor(true);
}

SepGaborWaveletFilter::~SepGaborWaveletFilter() {
	;
}

SepGaborWaveletFilter& SepGaborWaveletFilter::operator=(const SepGaborWaveletFilter& orig) {
	useDCCompensation = orig.useDCCompensation;
	
	ratio = orig.ratio;
	lambda= orig.lambda;
	theta = orig.theta;
	sigma = orig.sigma;

	dcOffsetFaktor = orig.dcOffsetFaktor;

	cosX = orig.cosX;
	cosY = orig.cosY;
	sinX = orig.sinX;
	sinY = orig.sinY;
	gaussX = orig.gaussX;
	gaussY = orig.gaussY;

	//temp1,temp2,temp3,temp4;
	sharedTemp(orig);
	return *this;
}

void SepGaborWaveletFilter::sharedTemp(const SepGaborWaveletFilter& f) {
	temp1 = f.temp1;
	temp2 = f.temp2;
	temp3 = f.temp3;
	temp4 = f.temp4;

	isTempShared = true;
}

GaborFilter2DParameter SepGaborWaveletFilter::getParameters() {
	GaborFilter2DParameter param;
	param.sigma = sigma;
	param.lambda = lambda;
	param.theta = theta;
	return param;
}

void SepGaborWaveletFilter::filterEven( GPUImage& srcA, GPUImage& dest) 
{
	GPUImage src(srcA);
	if ( temp1.size()<=src.size() && (temp1.getWidth() != src.getWidth() || temp1.getHeight() != src.getHeight()) ) {
		GPUImage temp1_new(temp1.reshape(src.getWidth(),src.getHeight(),true));
		GPUImage temp2_new(temp2.reshape(src.getWidth(),src.getHeight(),true));
		filterEven(	src,
					temp1_new,
					temp2_new,
					dest);
	} else if (temp1.size()>src.size()) {
		temp1.resize( src.getWidth(), src.getHeight() );
		temp2.resize( src.getWidth(), src.getHeight() );
		filterEven(src,temp1,temp2, dest);
	} else {
		filterEven(src,temp1,temp2, dest);
	}
}

void SepGaborWaveletFilter::filterEven(GPUImage& srcA, GPUImage& temp1, GPUImage& temp2, GPUImage& dest) 
{
	GPUImage src(srcA);
	cosX.apply(src,temp1); //src.convolute1D(cosX,1, temp1  );
	cosY.apply(temp1,dest);//temp1.convolute1D(cosY,0, dest  );

	sinX.apply(src,temp1);  //src.convolute1D(sinX,1, temp1  );
	sinY.apply(temp1,temp2);//temp1.convolute1D(sinY,0, temp2  );
	dest.add(temp2,-1.0f);

	if (useDCCompensation) {
		gaussX.apply(src,temp2);  //src.convolute1D(gauss,1, temp2  );
		gaussY.apply(temp2,temp1);//temp2.convolute1D(gauss,0, temp1  );
	
		dest.add(temp1,(float)(-1*dcOffsetFaktor));
	}
}


void SepGaborWaveletFilter::filterOdd(GPUImage& src, GPUImage& dest) 
{
	if ( temp1.size()<=src.size() && (temp1.getWidth() != src.getWidth() || temp1.getHeight() != src.getHeight()) ) {
		GPUImage temp1_new(temp1.reshape(src.getWidth(),src.getHeight(),true));
		GPUImage temp2_new(temp2.reshape(src.getWidth(),src.getHeight(),true));
		filterOdd(	src,
					temp1_new,
					temp2_new,
					dest);
	} else if (temp1.size()>src.size()) {
		temp1.resize(src.getWidth(),src.getHeight());
		temp2.resize(src.getWidth(),src.getHeight());
		filterOdd(src,temp1,temp2, dest);
	} else {
		filterOdd(src,temp1,temp2, dest);
	}
}

void SepGaborWaveletFilter::filterOdd( GPUImage& src, GPUImage& temp1, GPUImage& temp2, GPUImage& dest) 
{
	sinX.apply(src,temp1);//src.convolute1D(sinX,1, temp1  );
	cosY.apply(temp1,dest);//temp1.convolute1D(cosY,0, dest  );//Y

	cosX.apply(src,temp1);//src.convolute1D(cosX,1, temp1  );
	sinY.apply(temp1,temp2);//temp1.convolute1D(sinY,0, temp2  );//Y

	dest.add(temp2);
}

void SepGaborWaveletFilter::filter( GPUImage& src, GPUImage& destEven, GPUImage& destOdd) 
{
	if ( temp1.size()<=src.size() && (temp1.getWidth() != src.getWidth() || temp1.getHeight() != src.getHeight()) ) {
		temp1.resize(src.getWidth(),src.getHeight());
		temp2.resize(src.getWidth(),src.getHeight());
	} else if (temp1.size()>src.size()) {
		temp1.resize(src.getWidth(),src.getHeight());
		temp2.resize(src.getWidth(),src.getHeight());
	}
	if ( temp3.size()<=src.size() && (temp3.getWidth() != src.getWidth() || temp3.getHeight() != src.getHeight()) ) {
		temp3.resize(src.getWidth(),src.getHeight());
		temp4.resize(src.getWidth(),src.getHeight());
	} else if (temp3.size()>src.size()) {
		temp3.resize(src.getWidth(),src.getHeight());
		temp4.resize(src.getWidth(),src.getHeight());
	}

	cosX.apply(src,temp1);//src.convolute1D(cosX,1, temp1  );
	sinX.apply(src,temp3);//src.convolute1D(sinX,1, temp3  );

	cosY.apply(temp1,temp2);//temp1.convolute1D(cosY,0, temp2  );
	sinY.apply(temp3,temp4);//temp3.convolute1D(sinY,0, temp4  );

	if (useDCCompensation) {
		gaussX.apply(src,destOdd);//src.convolute1D(gauss,1, destOdd  );
		gaussY.apply(destOdd,destEven);//destOdd.convolute1D(gauss,0, destEven  );
		destEven.mul((float)(-1*dcOffsetFaktor));
		destEven.add(temp2);
		destEven.add(temp4,-1.0f);
	} else {
		destEven.copyFrom(temp2);
		destEven.sub(temp4);
	}

	sinY.apply(temp1,destOdd);//temp1.convolute1D(sinY,0, destOdd  );
	cosY.apply(temp3,temp2);//temp3.convolute1D(cosY,0, temp2  );
	destOdd.add(temp2);
}

GPUImage SepGaborWaveletFilter::getEnergy( GPUImage& src)
{
	GPUImage dest(temp1.copy());
	filterOdd(src,dest);
	dest.sqr();

	GPUImage imTemp(temp1.copy());
	filterEven(src,imTemp);
	imTemp.sqr();

	dest.add(imTemp);
	dest.sqrt();
	return dest;
}

GPUImage SepGaborWaveletFilter::getPhase( GPUImage& src)
{
	GPUImage dest(temp1.copy());
	filterOdd(src,dest);
//	dest.sqr();

	GPUImage imTemp(temp1.copy());
	filterEven(src,imTemp);
//	imTemp.sqr();

	dest.atan2(dest,imTemp);
	return dest;
}
