/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef SEPGABORWAVELETFILTER_H_
#define SEPGABORWAVELETFILTER_H_

#include "GPUImage.h"
#include "kernels.hpp"
#include <cmath>

struct GaborFilter2DParameter {
	/**
	 * wavelength
	 */
	float lambda;
	/**
	 * sigma of gaussian envelope in space
	 */
	float sigma;
	/**
	 * orientation in space
	 * if theta = 0 there is no modulation in the y direction
	 */
	float theta;
};

/**
 * This class represents one 2-dim gabor filter with a special
 * sigma, lambda and theta (orientation).
 *
 * This is done by seperated convolutions.
 *
 * The size of the image to filter is needed to create once the
 * therefor needed temporary images. If there are several gabor
 * filters for images with the same size, they can share their
 * temporary memory via connecting them (see parameter partner).
 *
 * The cosinus part is dc free.
 */
class SepGaborWaveletFilter {

private:
	double ratio;
	double lambda;
	double theta;
	double sigma;

	double dcOffsetFaktor;

	EvenGaborKernel1 cosX,cosY;
	OddGaborKernel1  sinX,sinY;
	GaussKernel1 	 gaussX;
	GaussKernel1 	 gaussY;

	GPUImage temp1,temp2,temp3,temp4;
	bool isTempShared;
	
	bool useDCCompensation;

	void initFilter();

	void init4Temp(int x, int y);

	void initDCOffsestFactor(bool justCheck = true);

	void sharedTemp(const SepGaborWaveletFilter& f);

public:

	/**
	 * this copy constructor performs a copy by reference
	 */
	SepGaborWaveletFilter(const SepGaborWaveletFilter& src);

	/**
	 * This assignment is a copy by reference operator
	 */
	SepGaborWaveletFilter& operator=(const SepGaborWaveletFilter& src);

	double getLambda() {
		return lambda;
	}
	double getSigma() {
		return sigma;
	}
	
	bool getUseDCCompensation() {
		return useDCCompensation;
	}
	
	void setUseDCCompensation(bool b) {
		useDCCompensation = b;
	}

public:
	SepGaborWaveletFilter(int x, int y);
	//explicit SepGaborWaveletFilter(SepGaborWaveletFilter& partner);
	explicit SepGaborWaveletFilter(const GaborFilter2DParameter& param);
	SepGaborWaveletFilter(int x, int y, double rad, double lambda, float rat=1.0f);
	SepGaborWaveletFilter(SepGaborWaveletFilter& partner, double rad, double lambda, float rat=1.0f);

	GaborFilter2DParameter getParameters();

	~SepGaborWaveletFilter();

	void filterEven( GPUImage& src, GPUImage& dest);

	void filterEven( GPUImage& src, GPUImage& temp1, GPUImage& temp2, GPUImage& dest);

	void filterOdd( GPUImage& src, GPUImage& dest);

	void filterOdd( GPUImage& src, GPUImage& temp1, GPUImage& temp2, GPUImage& dest);

	void filter( GPUImage& src, GPUImage& destEven, GPUImage& destOdd);

	GPUImage getEnergy( GPUImage& src);

	GPUImage getPhase( GPUImage& src);
};

#endif /*SEBGWF*/
