/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "census.hpp"
#include "CPUData.hpp"
#include <algorithm>
#include "compute_hypos.hcu"

namespace census {


void
compute_hypos_ambiguous (
		float *censusOcc,
		HypoImage targetArea,
		GPUImage censusOccDev,
		GPUImage currCensus,
		GPUImage pastCensus,
		int fbf,
		int H_MAX_2,
		int VEL_MAX,
		size_t nfeeds,
		HypoFeed *feeds,
		int p0)
{
	censusOccDev.copyFromHostMemory(censusOcc);
	do_compute_hypos_ambiguous (
			censusOccDev.getDataPointer(),
			targetArea.getData().getDataPointer(),
			currCensus.getDataPointer(),
			pastCensus.getDataPointer(),
			targetArea.getWidth(),
			targetArea.getWidth() * targetArea.getHeight(),
			p0, nfeeds, feeds,
			targetArea.getHMAX(), H_MAX_2, VEL_MAX
								);
	censusOccDev.copyToHostMemory(censusOcc);
}



void
compute_hypos_unambiguous (
		CensusValue *current, 
		CensusValue *last,
		unsigned int *current_values,
		float *censusOcc,
		HypoImage &HypoArray1,
		int w, 
		int h, 
		int H_MAX_1, 
		int VEL_MAX)
{
	CPUData<float> hostHypoImage;
	compute_hypos_unambiguous (current, last, current_values, censusOcc, 
			hostHypoImage, w, h, H_MAX_1, VEL_MAX);
	HypoArray1.getData().copyFromHostMemory(hostHypoImage.data());
}


void
match_census_lists (
		const CensusValue *p1, 
		const CensusValue *p2, 
		size_t n,
		CPUData<float> &hostHypoImage,
		int hmax, int vmax, float velo_power)
{
	memset( hostHypoImage.data(), 0, hostHypoImage.size() * sizeof(float) );

	// skip down blocks of identical census values in second list
	// starting at the end of each list
	for (const CensusValue *ip1 = p1 + n - 1, *ip2 = p2 + n - 1; ip2 >= p2; ip2 -= ip2->iH)
	{

		// skip if current entry is ambiguous
		if (ip2->iH > hmax) 
			continue;

		// search downwards in first list ( census-value-block-wise ) to match current entry in second list
		while (ip1 > p1 && ip1->c < ip2->c)
			ip1 -= ip1->iH;

		// skip if found entry is ambiguous or no match was found
		if (ip1->c != ip2->c || ip1->iH > hmax)
			continue;

		// create matches for all pairs if they are closer than vmax
		// by iterating both blocks
		for (size_t i = 0, ni = ip1->iH, nh = 0;  i < ni; ++i) 
		{
			const CensusValue & c1 = ip1[-i]; // since we move downwards indices are negative!
			CPUDataFeatureColumn<float> H = hostHypoImage(c1.iPosY, c1.iPosX);

			for (size_t j = 0, nj = ip2->iH;  j < nj; ++j)
			{
				const CensusValue & c2 = ip2[-j]; // since we move downwards indices are negative!
				int dx = c1.iPosX - c2.iPosX;
				int dy = c1.iPosY - c2.iPosY;
				const int manhattan_speed = std::max( std::abs( dx ), std::abs( dy ) ); 
				if ( manhattan_speed <= vmax )
				{
					float * h = & H[ 3 * nh++ ];
					h[0] = std::pow( (vmax - manhattan_speed) / float(vmax), velo_power );
					h[1] = dx;
					h[2] = dy;
				}
			}
		}
	}
}


void
compute_hypos_unambiguous (
		CensusValue *current,
		CensusValue *last,
		unsigned int *current_values,
		float *censusOcc,
		CPUData<float> &hostHypoImage,
		int w, 
		int h,
		int H_MAX_1,
		int VEL_MAX)
{



	hostHypoImage.resize(h, w, H_MAX_1 * 3);
	hostHypoImage.make_contiguous();
	build_sorted_list(current, current_values, w, h);

	// collect hypo ambiguities
	for (size_t i = 0; i != (w*h); ++i)
		censusOcc[last[i].iPosX + w * last[i].iPosY] = float(last[i].iH);

	match_census_lists(current, last, w*h, hostHypoImage, H_MAX_1, VEL_MAX);
}
	




static inline bool UDgreaterCensus(CensusValue left, CensusValue right)
{
	return left.c > right.c;
}


void fill (CensusValue *list, const unsigned int *values, int w, int h) 
{
	for (int y = 0, i = 0; y < h; ++y) {
		for (int x = 0; x < w; ++x, ++i) {
			list[i].c = values[i];
			list[i].iPosX = x;
			list[i].iPosY = y;
			list[i].iH = 1;
		}
	}
}


void count_occurences (CensusValue *list, size_t n) 
{
	if (n == 0) return;

	// initialize
	size_t start = 0;
	unsigned int c = list[0].c;

	// scan
	for (size_t i = 1; i < n; ++i) {
		unsigned int ci = list[i].c;
		if (ci != c) {
			size_t n_occ = i - start;
			for (; start < i; ++start) 
				list[start].iH = n_occ;
			c = ci;
		}
	}

	// fill last block
	size_t n_occ = n - start;
	for (; start < n; ++start) 
		list[start].iH = n_occ;
}


void build_sorted_list (CensusValue *list, unsigned int *values, int w, int h)
{
	fill(list, values, w, h);
	std::sort(list, list + w*h, UDgreaterCensus);
	count_occurences(list, w*h);
}


};
