/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "downsample.hpp"

extern "C" void h_downsample_cont_rm( float* res, int n1, int n2, int nk, int down_factor, int shift,
									  const float* src, int sx, int mode );

extern "C" void downsample_array( float* dst, int n1, int n2, int shift1, int shift2, const cudaArray* src, int w, int h, int feature_depth, int mode, bool features_first );

void downsample ( GPUData & input, GPUData & output, int n, int mode, int shift )
{

  int n_dims = input.getDimensionsCount();
#ifndef _WIN32
  int offset[n_dims], strides[n_dims];
  size_t sizes[n_dims];
#else
  int * const offset = (int*)_alloca(n_dims * sizeof(int));
  int * const strides = (int*)_alloca(n_dims * sizeof(int));
  size_t * const sizes = (size_t*)_alloca(n_dims * sizeof(size_t));
#endif

  const size_t * in_dims = input.getDimensions();
  const int * in_strides = input.getStrides();

  if ( shift <= 0 ) shift = n;

  std::fill( offset, offset + n_dims, 0 );
  std::copy( in_dims, in_dims + n_dims, sizes );
  std::copy( in_strides, in_strides + n_dims, strides );
  
  sizes[0] = 1 + ( sizes[0] - n ) / shift; 
  sizes[1] = 1 + ( sizes[1] - n ) / shift; 
  
  strides[0] *= shift; 
  strides[1] *= shift;
  output.resize( n_dims, sizes, input.order() );

  std::cout << " downsampling with n="<< n <<" shift="<<shift<<" in = " << input.dump_dims() << " out = " << output.dump_dims() << std::endl;

  if ( mode == 0 )
	{
	  offset[0] = shift/2; offset[1] = shift/2; 
	  output.assign( input.view(offset,n_dims,sizes,strides) );
	}
  else if ( mode >= 1 && mode <= 3 ) // rectangular pooling
	{
	  
	  if ( input .isContiguous(true) && input .isRowMajor() &&
		   output.isContiguous(true) && output.isRowMajor() )
		{
		  size_t one = 0, two = 1;
		  GPUData out3( output.reduceDimensionality( one, two ) );
		  GPUData in3 (  input.reduceDimensionality( one, two ) );
		  out3.fill( 1.0f );
		  h_downsample_cont_rm( out3.data(), out3.extent(0), out3.extent(1), out3.extent(2),
								n, shift, in3.data(), in3.stride(0), mode );
		  GPUDataException::check_error( "contiguous downsampling" );
		}
	  /*else
	  
		
		{
		  size_t n1 = input.extent(0);
		  size_t n2 = input.extent(1);
		  size_t m1 = output.extent(0);
		  size_t m2 = output.extent(1);
		  GPUData layer0( input
						  .range( 0, 0, n1 - 1, shift )
						  .range( 1, 0, n2 - 1, shift ) );
		  output.assign( layer0 );
		  for ( int i = 0; i < n; ++i )
			{
			  offset[0] = i;
			  for ( int j = ( i == 0 ? 1 : 0 ); j < n; ++j )
				{
				  offset[1] = j;
				  GPUData layer( input
								 .range( 0, i, n1 - 1, shift )
								 .range( 1, j, n2 - 1, shift ) );
				  switch(mode)
					{
					case 1: output.add( layer ); break;
					case 2: output.max( layer ); break;
					case 3: output.add( layer ); break;
					}
				}
			}
		  GPUDataException::check_error( "manual downsampling" );
		  if ( mode == 1 )
			output.mul( 1.0f / float( n * n ) );
		  GPUDataException::check_error( "scaling" );
		}
		*/
	}
}



void downsample_on_array( GPUData & input, GPUData & output, int n1, int n2, int shift1, int shift2, int mode, bool features_first )
{
  int n_dims = input.getDimensionsCount();
  size_t * sizes = new size_t[n_dims];
  const size_t * in_dims = input.getDimensions();
  const int * in_strides = input.getStrides();

  std::copy( in_dims, in_dims + n_dims, sizes );	
	
  if ( !input.is_array() )
    input.move_to_array( features_first );
  
  size_t one = 0, two = 1;
  if ( features_first )
    {
      std::cout << "WARNING: using features_first=true in downsample: this mode is implemented but not yet tested! Be careful"<<std::endl;
      one = n_dims-1;
      two = n_dims-2;
    }
  
  size_t feature_depth = input.size() / input.extent(one) / input.extent(two);
  
  sizes[one] = 1 + ( sizes[one] - n1 ) / shift1; 
  sizes[two] = 1 + ( sizes[two] - n2 ) / shift2;
  output.reference( GPUData( n_dims, sizes ) );
  //std::cout << "Downsample mode = " << mode << " n1="<<n1<<" n2="<<n2<<" shift1="<<shift1 <<" shift2="<<shift2<< " out="<<output.dump_dims()<<std::endl;
  
  delete sizes;	
  
  GPUData out = output;
  out.reduceDimensionality( one, two );
  downsample_array( out.data(), n1, n2, shift1, shift2, input.array(), out.extent(two), out.extent(one), feature_depth, mode, features_first );

  //input.move_to_data();
  //input.clear();
  //GPUDataException::check_error( "back to data" );
	
}


