/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <boost/program_options.hpp>
#include <iostream>
#include <iterator>
#include <vector>

#include <SeariseLogger.hpp>
#include <DataWindow.hpp>

#include "ModelArea.hpp"
#include "GaborFeatureExtractor.h"
#include "HDF5RawDataFileAdapter.hpp"

inline std::string imageName(int i) {
	char* str = new char[1024];
	sprintf(str,"input_image_%05d",i);
	std::string rtv(str);
	delete[] str;
	return rtv;
}

inline std::string featureName(int i) {
	char* str = new char[1024];
	sprintf(str,"input_features_%05d",i);
	std::string rtv(str);
	delete[] str;
	return rtv;
}

inline std::string featureNormedName(int i) {
	char* str = new char[1024];
	sprintf(str,"input_features_normed_%05d",i);
	std::string rtv(str);
	delete[] str;
	return rtv;
}

inline std::string featureSumName(int i) {
	char* str = new char[1024];
	sprintf(str,"V2_feature_sum_%05d",i);
	std::string rtv(str);
	delete[] str;
	return rtv;
}

inline GPUImage getFeatureSum(const GPUData& d_) {
	size_t protect_h = d_.getNoOfDimensions()-2;
	GPUData d = d_.reduceDimensionality(protect_h);
	const int n = d.extent(0);
	const int h = d.extent(1);
	const int w = d.extent(2);
	GPUImage rtv(w,h); rtv.fill(0.0f);
	for (int i=0; i<n; ++i) {
		rtv.add(d[i]);
	}
	return rtv;
}
inline GPUImage getFeatureSum2(const GPUData& d_) {
	GPUData d(d_.shiftDim(-2).copy());
	size_t protect_h = d.getNoOfDimensions()-2;
	d = d.reduceDimensionality(protect_h);
	std::cout << "reduced to " << d.getNoOfDimensions() << std::endl;
	const int n = d.extent(0);
	const int h = d.extent(1);
	const int w = d.extent(2);
	GPUImage rtv(w,h); rtv.fill(0.0f);
	for (int i=0; i<n; ++i) {
		rtv.add(d[i]);
	}
	return rtv;
}

int main(int argc, char* argv[]) {

	boost::program_options::options_description desc("Form Runner program options");
	desc.add_options()
		("help,h", "produce help message")
		("input-file,i", boost::program_options::value< std::string >(),
							"Set the png-filename of the input file.")
		("n,n", boost::program_options::value<int>()->default_value(10),
							"Number of iterations")
		("feedback,f", boost::program_options::value<float>()->default_value(10.0f),
							"Feedback factor")
//		("debug-output","Enables output of max-/min-activity values each iteration")
	;

	boost::program_options::positional_options_description p;
	p.add("input-file", 1);

	boost::program_options::variables_map vm;
	boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(desc).positional(p).run(), vm);
	boost::program_options::notify(vm);

	if (vm.count("help")) {
		printf("\nuse:   %s INPUTPFILE \n",argv[0]);
		printf("  example: %s image.png -n 20\n",argv[0]);
	    std::cout << desc << std::endl;
	    return 1;
	}

	/* --- input images ---*/
	std::string inputFile = "lena.png";
	GPUImage image;
	{
//		inputFile = vm["input-file"].as<std::string>();
//		printf("load file %s\n",inputFile.c_str());
		image = GPUImage(HostImage(inputFile.c_str()));
		image.div(image.max());
	}

	/* --- output parameters --- */
	std::string outputName   = "output-file.h5";
	const bool 	debug_output = true;// vm.count("debug-output");

	/* --- algorithm parameters --- */
	const float	fbf				  = vm["feedback"].as<float>();
	const float noOfIterations    = 10;
	const float noOfOri	  = 12;
	const float noOfSca	  = 1;
	const float noOfCha	  = 1;
	const float noOfFeatures	  = noOfOri*noOfSca*noOfCha;
	const float sigmaLarge		  = 4.0f;
	const float sigmaSmall		  = sigmaLarge/2.5f;
	const float sigmaSpace		  = 4.0f;

	{  /* print algorithm parameters parameters to command line */
		std::cout << "noOfFeatures       = " << noOfFeatures << std::endl;
		std::cout << "feedback           = " << fbf << std::endl;
		std::cout << "no of iterations   = " << noOfIterations << std::endl;
		std::cout << "sigma large        = " << sigmaLarge << std::endl;
		std::cout << "sigma small        = " << sigmaSmall << std::endl;
		std::cout << "sigma space        = " << sigmaSpace << std::endl;
	}

	const int w = image.getWidth();
	const int h = image.getHeight();
	std::cout << "image sizes (w,h) = (" << w << "," << h << ")" << std::endl;
	std::cout << "\n" << std::endl;

	DataWindow* image_window = 0;
	DataWindow* v1_window    = 0;
	DataWindow* v2_window    = 0;
	image_window->global_gtk_gl_init();
	image_window = new DataWindow(w,h,"image");
	v1_window = new DataWindow(w,h,"v1");
	v2_window = new DataWindow(w,h,"v2");

	{   /* begin initialization */
		/* file to store the output */
		HDF5RawDataFileAdapter	outFile(outputName.c_str(),true);

		/* space for initial feature extraction */
		GPUData 				inputFeatures(noOfCha,noOfSca,noOfOri,h,w);
		GPUData 				data_1(h,w,noOfCha,noOfSca,noOfOri);
		GPUData 				data_2(h,w,noOfCha,noOfSca,noOfOri);
		GPUData 				data_3(h,w,noOfCha,noOfSca,noOfOri);
								data_3.fill(0.0f);

		/* Gabor feature extractor */
		GaborFeatureExtractor 	gf(w,h,GaborFeatureExtractor::ENERGY,
									noOfOri,noOfSca,
									5.0f,0.25f);
//								gf.useNakaRushtonNormalization(0.025f);

		ModelArea 	model_area_v1;
					model_area_v1.set_do_square(true);
					model_area_v1.set_feedback_gain(fbf);
					model_area_v1.set_kernel_type(ModelArea::NONE);
					model_area_v1.set_sigma_for_spatial_inhibition(sigmaSpace);
					model_area_v1.set_shunting_params(  0.0007536f*(1/0.05f)
														, 1.0f      *(1/0.05f)
														, 0.05f     *(1/0.05f)
														, 0.05f/1.0f*(1/0.05f));

		ModelArea 	model_area_v2;
					model_area_v2.set_do_square(true);
					model_area_v2.set_feedback_gain(0.0f);
					model_area_v2.set_kernel_type(ModelArea::DIPOLE);
					model_area_v2.set_sigma(0,sigmaLarge);
					model_area_v2.set_sigma(1,sigmaSmall);
					model_area_v2.set_sigma(2,sigmaLarge*2);
					model_area_v2.set_sigma_for_spatial_inhibition(sigmaSpace);
					model_area_v2.set_shunting_params(  0.0007536f*(1/0.05f)
														, 1.0f      *(1/0.05f)
														, 0.05f     *(1/0.05f)
														, 0.05f/4.0f*(1/0.05f));

		/* get features */
		gf.getFeatureImages(image,inputFeatures);

		if (debug_output) {
			const float maximum = inputFeatures.max();
			const float minimum = inputFeatures.min();
			const int   size    = inputFeatures.size();
			const float summe   = inputFeatures.sum();
			printf( "\nInitial Feature activity:\n max =%f\n"
					" min =%f\n mean=%f\n",maximum,minimum,summe/size);
		}

		if (image_window) {
			image_window->update(getFeatureSum(inputFeatures));
		}

		/* begin of iteration loop */
		for (int i=0; i<noOfIterations; ++i) {

			std::cout << "iterate " << ++i << std::flush;

			data_1.copyFrom(inputFeatures.shiftDim(2));

			/*
			 * computes V1
			 */
			model_area_v1.iterate(data_1,data_2,data_3);

			if (v1_window) {
				v1_window->update(getFeatureSum2(data_2));
			}

			/*
			 * computes V2
			 */
			model_area_v2.iterate(data_2,data_3);

			if (v2_window) {
				v2_window->update(getFeatureSum2(data_3));
			}

		}
		std::cout << "END" << std::endl;
	}

	if (image_window)
		image_window->wait_until_closed();

	if (image_window) delete image_window;
	if (v1_window) delete v1_window;
	if (v2_window) delete v2_window;

	return 0;
}


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
