/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <boost/program_options.hpp>
#include <boost/regex.hpp>
#include <boost/foreach.hpp>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <iterator>
#include <vector>
#include <cstdlib>

#include <SeariseLogger.hpp>

#if defined(HAVE_GTK)
  #include <DataWindow.hpp>
#endif

#include "ClassicUUlmNeuralModel.h"
#include "GaborFeatureExtractor.h"
#include "LoGFeatureExtractor.hpp"
#include "GradientFeatureExtractor.hpp"
#include "CartesianCorrelator.h"
#include "GPUVelField.h"
#ifdef HAVE_HDF5
	#include "HDF5FileAdapter.h"
	#include "HDF5RawDataFileAdapter.hpp"
#endif
#include "SimpleHostImageSequence.h"
#include "FeatureNormalization.hpp"
#include "MiddleburyFlowIO.hpp"

#include "ModelArea.hpp"
#include "downsample.hpp"
#include "ModulatedIntegration.hpp"
#include "INRIAForm.hpp"

#include "util.hpp"

#ifdef _WIN32
#define fmin(a,b) (a < b ? a : b)
#define fmax(a,b) (a > b ? a : b)
#endif

/**
 * generates simple test sequences
 */
class TestSequence : public HostImageSequence {
private:
	GPUData sequence;
	size_t w;
	size_t h;
	size_t i;

	GaussKernel2 gauss;

public:
	TestSequence(int type=0)
		: w(160)
		, h(120)
		, i(0)
		, gauss(0.75f,0,1)
	{
		int n;

		size_t size[2];
		int offs[2];

		switch (type)
		{
		case 0:
			n=20;
			sequence = GPUData(n,h,w);
			sequence.fill(0.0f);
			size[0] = h/4;
			size[1] = w/5;
			for (int i=0; i<n; ++i) {
				GPUImage image(sequence[i]);
				offs[0] = 10+(h-10-size[0])/n*i;
				offs[1] = 10+(w-10-size[1])/n*i;
				image.view(offs,size).fill(1.0f);
			}
			break;
		case 1:
			n=35;
			sequence = GPUData(n,h,w);
			sequence.fill(0.0f);
			for (int i=0; i<n; ++i) {
				for (int j=0; j<30; ++j) {
					GPUImage image(sequence[i]);
					image[(h/3)+j*2+0][20+(w/3)/n*i*2+int(0.4*(2*j+0))+0] = 1.0f;
					image[(h/3)+j*2+0][20+(w/3)/n*i*2+int(0.4*(2*j+0))+1] = 1.0f;
					image[(h/3)+j*2+1][20+(w/3)/n*i*2+int(0.4*(2*j+1))+0] = 1.0f;
					image[(h/3)+j*2+1][20+(w/3)/n*i*2+int(0.4*(2*j+1))+1] = 1.0f;
				}
			}
			break;
		case 2:
			n=20;
			sequence = GPUData(n,h,w);
			sequence.fill(0.0f);
			size[0] = h/4;
			size[1] = w/5;
			for (int i=0; i<n; ++i) {
				GPUImage image(sequence[i]);
				offs[0] = 10;
				offs[1] = 10+(w-10-size[1])/n*i;
				image.view(offs,size).fill(1.0f);
				offs[0] = 10+size[0]+6;
				offs[1] = 10+(w-10-size[1])/n*(n-1-i);
				image.view(offs,size).fill(1.0f);
			}
			break;
		case 3:
			n=20;
			sequence = GPUData(n,h,w);
			sequence.fill(0.0f);
			size[0] = h/4;
			size[1] = w/5;
			for (int i=0; i<n; ++i) {
				GPUImage image(sequence[i]);
				image.slice(1,w/2).fill(0.5f);
				image[h/2].fill(0.5f);
				offs[0] = 10;
				offs[1] = 10+(w-10-size[1])/n*i;
				image.view(offs,size).fill(1.0f);
				offs[0] = 10+size[0]+6;
				offs[1] = 10+(w-10-size[1])/n*(n-1-i);
				image.view(offs,size).fill(1.0f);
			}
			break;
		case 4:
			n=60;
			sequence = GPUData(n,h,w);
			sequence.fill(0.0f);
			size[0] = h/4;
			size[1] = w/5;
			for (int i=0; i<n; ++i) {
				GPUImage image(sequence[i]);
				offs[0] = 10;
				offs[1] = 10+(w-10-size[1])/n*i;
				image.view(offs,size).fill(1.0f);
				offs[0] = 10+size[0]+6;
				offs[1] = 10+(w-10-size[1])/n*(n-1-i);
				image.view(offs,size).fill(1.0f);
			}
			break;
		default:
			n = 20;
			size[0] = h/3;
			size[1] = w/4;
			for (int i=0; i<n; ++i) {
				GPUImage image(sequence[i]);
				offs[0] = (h-size[0])/n*i;
				offs[1] = (w-size[1])/n*i;
				image.view(offs,size).fill(1.0f);
			}
		}
		for (int i=0; i<n; ++i) {
			GPUImage frame(sequence[i]);
			gauss.apply(frame,frame);
		}
	}

	~TestSequence() {
		;
	}

	HostImage getCurrentImage() {
		HostImage rtv(w,h,HostImage::FLOAT, HostImage::LUM);
		GPUImage image(sequence[i]);
		image.copyToHostMemory(rtv.getPointerToImage());
		return rtv;
	}

	HostImage getNextImage() {
		HostImage rtv(w,h,HostImage::FLOAT, HostImage::LUM);
		GPUImage image(sequence[i++]);
		image.copyToHostMemory(rtv.getPointerToImage());
		return rtv;
	}

	void copyNextImage(GPUImage& gData) {
		gData.resize(w,h);
		gData.copyFrom(sequence[i++]);
	}

	bool hasNextImage() {
		return i<sequence.extent(0);
	}

	int getWidth() { return w; }

	int getHeight() { return h; }

	int getLength() { return sequence.extent(0); }
};

//extern "C" void shiftMulAddSpecial(
//		float* toShiftXX, float* toShiftXY, float* toShiftYY,
//		float* toMulXX, float* toMulXY, float* toMulYY,
//		float* addTo, int shiftX, int shiftY, int Nx, int incY, int Ny);
//
//extern "C" void inhibitionAndRect(const int N, float* res, float* exita, float* inhib, const float subInhibFactor, const float divInhibConstant, const float divInhibFactor);
//
//class LoGCorrelator {
//
//private:
//	GPUData 	 diffKernel;
//	GaussKernel2 gauss_large;
//	GaussKernel2 gauss_small;
//	GPUImage	 tempImage1, tempImage2, tempImage3;
//	GPUImage 	 tempOplus,tempOminus,tempImage;
//	GPUImage	 dIx,dIy;
//	int			 w,h;
//
//	inline void init(int w_, int h_) {
//		if (w_==w && h_==h) return;
//
//		w=w_; h=h_;
//
//		tempImage1 = tempImage1.reshape(w,h,true);
//		tempImage2 = tempImage2.reshape(w,h,true);
//		tempImage3 = tempImage3.reshape(w,h,true);
//
//		tempOplus  = tempOplus.reshape(w,h,true);
//		tempOminus = tempOminus.reshape(w,h,true);
//		tempImage  = tempImage.reshape(w,h,true);
//
//		dIx = dIx.reshape (w,h,true);
//		dIy = dIy.reshape (w,h,true);
////		dI1xy.resize(w,h);
////		dI1xx.resize(w,h);
////		dI1yy.resize(w,h);
////
////		dI1Sum.resize(w,h);
////
////		dI2x.resize (w,h);
////		dI2y.resize (w,h);
////		dI2xy.resize(w,h);
////		dI2xx.resize(w,h);
////		dI2yy.resize(w,h);
////
////		dI2Sum.resize(w,h);
//	}
//
//	inline void correlate(
//			GPUImage& dI1XX,
//			GPUImage& dI2XX,
//			GPUImage& dI1XY,
//			GPUImage& dI2XY,
//			GPUImage& dI1YY,
//			GPUImage& dI2YY,
//			int shiftX, int shiftY,
//			GPUImage& correlationResultPlus,
//			GPUImage& correlationResultMinus
//	) {
//		const int w = dI1XX.getWidth();
//		const int h = dI1XX.getHeight();
//
//		/* O_plus */
//		{
//			const int Nx = w;
//			const int Ny = h;
//			const int incY = Nx;
//			tempOplus.fill(0.0f);
//
//			shiftMulAddSpecial(
//				dI2XX.getDataPointer(), dI2XY.getDataPointer(), dI2YY.getDataPointer(),
//				dI1XX.getDataPointer(), dI1XY.getDataPointer(), dI1YY.getDataPointer(),
//				tempOplus.getDataPointer(), -shiftX, -shiftY, Nx, incY, Ny);
//
//			gauss_small.apply(tempOplus,tempImage);tempImage.swap(tempOplus);
////			tempOplus.convolute1D(gauss,1,tempImage);
////			tempImage.convolute1D(gauss,2,tempOplus);
//			tempOplus.rect();
//		}
//
//		/* O_minus */
//		if (shiftX==0 && shiftY==0) {
//			tempOminus.copyFrom(tempOplus);
//		} else {
//			const int Nx = tempOminus.getWidth();
//			const int Ny = tempOminus.getHeight();
//			const int incY = Nx;
//			tempOminus.fill(0.0f);
//
//			shiftMulAddSpecial(
//					dI1XX.getDataPointer(), dI1XY.getDataPointer(), dI1YY.getDataPointer(),
//					dI2XX.getDataPointer(), dI2XY.getDataPointer(), dI2YY.getDataPointer(),
//					tempOminus.getDataPointer(), -shiftX, -shiftY, Nx, incY, Ny);
//
//			gauss_small.apply(tempOminus,tempImage);tempImage.swap(tempOminus);
////			tempOminus.convolute1D(gauss,1,tempImage);
////			tempImage.convolute1D(gauss,2,tempOminus);
//			tempOminus.rect();
//		}
//
//		inhibitionAndRect(correlationResultPlus.size(),correlationResultPlus.getDataPointer(),tempOplus.getDataPointer(),tempOminus.getDataPointer(),0.5f,1.0f,1.0f);
//
//		if (shiftX==0 && shiftY==0) {
//			correlationResultMinus.copyFrom(correlationResultPlus);
//		} else {
//			tempOplus.swap(tempImage);
//			tempOplus.copyFromShifted(tempOminus, shiftX, shiftY);
//			tempOminus.copyFromShifted(tempImage, shiftX, shiftY);
//
//			inhibitionAndRect(correlationResultMinus.size(),correlationResultMinus.getDataPointer(),tempOplus.getDataPointer(),tempOminus.getDataPointer(),0.5f,1.0f,1.0f);
//
//			tempOplus.swap(tempImage); // swap back
//		}
//	}
//
//public:
//
//	LoGCorrelator()
//		: diffKernel(9)
//		, gauss_large(4.0f)
//		, gauss_small(2.0f)
//		, w(-1)
//		, h(-1)
//	{
//		{ /* init diffKernel */
//			fillLinSpace(diffKernel,-4,4);
//			GPUData temp(diffKernel.copy());
//			diffKernel.sqr();
//			diffKernel.div(-2.0f);
//			diffKernel.exp();
//			diffKernel.mul(temp);
//			diffKernel.mul(-2.0/sqrt(2*M_PI));
//		}
//	}
//
//	GPUData getFeatureImages(GPUImage& im) {
//		GPUData res(3,im.getHeight(),im.getWidth());
//		getFeatureImage(im,res);
//		return res;
//	}
//
//	void getFeatureImage(GPUImage& im, GPUData& res) {
//		init( im.getWidth(), im.getHeight() );
//
//		GPUImage dI1xx(res[0]);
//		GPUImage dI1xy(res[1]);
//		GPUImage dI1yy(res[2]);
//
//		im.convolute1D(diffKernel,1,dIx);
//		im.convolute1D(diffKernel,0,dIy);
//		dIx.convolute1D(diffKernel,0,dI1xy);
//		dIx.convolute1D(diffKernel,1,dI1xx);
//		dIy.convolute1D(diffKernel,0,dI1yy);
//	}
//
//	GPUData getFullFeatureImages(const GPUData& data, const int n=12) {
//		GPUData res(n,data.extent(1),data.extent(2));
//		getFullFeatureImages(data,res,n);
//		return res;
//	}
//
//	void getFullFeatureImages(const GPUData& data, GPUData& res, const int n=12) {
//		GPUImage dxx(data[0]);
//		GPUImage dxy(data[1]);
//		GPUImage dyy(data[2]);
//		res.fill(0.0f);
//		for (int i=0; i<n; ++i) {
//			const double phi   = i*2.0*M_PI/n;
//			const double sinus = sin(phi);
//			const double cosin = cos(phi);
//			GPUImage r(res[i]);
//			r.add(dxx,cosin*cosin);
//			r.add(dxy,sinus*cosin);
//			r.add(dyy,sinus*sinus);
//		}
//	}
//
//	void correlateAll(GPUData& images1, GPUData& images2, GPUData& res) {
//
//		GPUImage dI1xx(images1[0]);
//		GPUImage dI2xx(images2[0]);
//
//		GPUImage dI1xy(images1[1]);
//		GPUImage dI2xy(images2[1]);
//
//		GPUImage dI1yy(images1[2]);
//		GPUImage dI2yy(images2[2]);
//
////		init( image1.getWidth(), image1.getHeight() );
////
////		image1.convolute1D(diffKernel,1,dI1x);
////		image1.convolute1D(diffKernel,0,dI1y);
////		dI1x.convolute1D(diffKernel,0,dI1xy);
////		dI1x.convolute1D(diffKernel,1,dI1xx);
////		dI1y.convolute1D(diffKernel,0,dI1yy);
////
//////		dI1Sum.copyFrom(dI1xx); dI1Sum.abs();
//////		tempImage1.copyFrom(dI1yy); tempImage1.abs();
//////		dI1Sum.add(tempImage1); dI1Sum.mul(M_PI);
//////		gauss_large.apply(dI1Sum,tempImage2); tempImage2.swap(dI1Sum);
//////		dI1Sum.add(0.0001f);
//////		dI1xy.div(dI1Sum); dI1xx.div(dI1Sum); dI1yy.div(dI1Sum);
////
////		image2.convolute1D(diffKernel,1,dI2x);
////		image2.convolute1D(diffKernel,0,dI2y);
////		dI2x.convolute1D(diffKernel,0,dI2xy);
////		dI2x.convolute1D(diffKernel,1,dI2xx);
////		dI2y.convolute1D(diffKernel,0,dI2yy);
////
//////		dI2Sum.copyFrom(dI2xx); dI2Sum.abs();
//////		tempImage1.copyFrom(dI2yy); tempImage1.abs();
//////		dI2Sum.add(tempImage1); dI2Sum.mul(M_PI);
//////		gauss_large.apply(dI2Sum,tempImage2); tempImage2.swap(dI2Sum);
//////		dI2Sum.add(0.0001f);
//////		dI2xy.div(dI2Sum); dI2xx.div(dI2Sum); dI2yy.div(dI2Sum);
//
//		const int maxVelX = res.extent(3)/2;
//		const int maxVelY = res.extent(2)/2;
//
//		for (int dy=-maxVelY; dy<=maxVelY; ++dy) {
//			for (int dx=0; dx<=maxVelX; ++dx)  {
//				correlate(
//					dI1xx,
//					dI2xx,
//					dI1xy,
//					dI2xy,
//					dI1yy,
//					dI2yy, dx, dy, tempImage2,tempImage3);
//				res.shiftDim(2)[maxVelY+dy][maxVelX+dx].copyFrom(tempImage2);
//				res.shiftDim(2)[maxVelY-dy][maxVelX-dx].copyFrom(tempImage3);
//			}
//		}
//	}
//};

//static GaussKernel2 gauss_4(4.0f);
//static GPUImage tempImage1;
//static GPUImage tempImage2;
//
//void normalize(GPUData& d, float eps = 0.01f) {
////	printf("d.ndim=%d\ndims={%d,%d,%d,%d,%d}\n",d.getDimensionsCount(),d.extent(0),d.extent(1),d.extent(2),d.extent(3),d.extent(4));
//	size_t di = d.getDimensionsCount()-2;
//	GPUData data(d.reduceDimensionality(di));
////	printf("data.ndim=%d\ndims={%d,%d,%d}\n",data.getDimensionsCount(),data.extent(0),data.extent(1),data.extent(2));
//	tempImage1.resize(data.extent(2),data.extent(1));
//	tempImage2.resize(data.extent(2),data.extent(1));
//	for (size_t i=0; i<data.extent(0); ++i) {
//		tempImage1.copyFrom(data[i]);
//		tempImage1.abs();
//		if (i==0)
//			tempImage2.copyFrom(tempImage1);
//		else
//			tempImage2.add(tempImage1);
//	}
//	tempImage2.add(eps);
//	gauss_4.apply(tempImage2,tempImage1);
//	for (size_t i=0; i<data.extent(0); ++i) {
//		data[i].div(tempImage1);
//	}
//}

#include <cuda_runtime.h>
#include "opts.hpp"

int main(int argc, char* argv[]) {

//	log().set_log_mask_level(boost::logger::DEFAULT_LEVEL_ERROR);
//	log().set_log_mask_level(boost::logger::DEFAULT_LEVEL_INFO);
//	log().cpumem.off();
//	log().gpumem.on();

	boost::program_options::options_description desc("Motion Runner 2 program options");
	boost::program_options::positional_options_description p;
	setup_opts(desc, p);

	boost::program_options::variables_map vm;
	boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(desc).positional(p).run(), vm);
	boost::program_options::notify(vm);

	if (vm.count("config-file")) {
		std::cout << "Read parametrs from config file ";
		std::cout << vm["config-file"].as<std::string>() << "." << std::endl;
		std::fstream c_file(vm["config-file"].as<std::string>().c_str());
		boost::program_options::store(boost::program_options::parse_config_file(c_file,desc),vm);
		boost::program_options::notify(vm);
	}
//	else if (!vm.count("input.start")) {
//		std::string  s = vm["input.files"].as<std::string>();
//		// parameter "input.files" has to be deleted!
//		boost::regex e(".+\\.config$");
//		if (boost::regex_match(s,e)) {
//			std::cout << "Read parametrs from config file ";
//			std::cout << s << "." << std::endl;
//			std::fstream c_file(s.c_str());
//			boost::program_options::store(boost::program_options::parse_config_file(c_file,desc),vm);
//			boost::program_options::notify(vm);
//		}
//	}

	printf_deviceInfo(-1);


	if (vm.count("help") || argc<=1) {
		printf("\nuse:   %s INPUTPATTERN STARTINDEX STOPINDEX [-o OUTPUTFILE -m MAXIMALVELOCITY -d [all|flow|...]]\n\n",argv[0]);
		printf("  example: %s image%%05d.png 0 100\n",argv[0]);
		printf("           %s image%%05d.png 0 100 -o out.h5\n",argv[0]);
		printf("           %s image%%05d.png 0 100 -o out.h5 -v 7 -d flow\n\n",argv[0]);
	    std::cout << desc << std::endl;
	    return EXIT_FAILURE;
	}

	/* --- input images ---*/
	std::string inputFilePatter;
	int   		start_idx;
	int   		stop_idx;
	HostImageSequence* imageSequence;

	if (!vm.count("input.stop"))
	{ 	/* check weather to chose an artificial example image sequence */
		std::string param = vm["input.files"].as<std::string>();
		if (param == "test" && param.size() == 4) {
			if (vm.count("input.start"))
				start_idx = vm["input.start"].as<int>();
			else
				start_idx = 0;
			TestSequence* t = new TestSequence(start_idx);
			imageSequence = t;
			stop_idx = -1;
			inputFilePatter = "test%d";
			start_idx = 0;
			stop_idx  = t->getLength()-1;
			std::cout << "run test sequence " << vm["input.start"].as<int>() << " ";
			std::cout << "with " << t->getLength() << " frames" << std::endl;
		} else {

			printf("\nuse:   %s INPUTPATTERN STARTINDEX STOPINDEX [-o OUTPUTFILE --max-vel=7 -d=[all|flow|hypo]]\n\n",argv[0]);
			printf("  example: %s image%%05d.png 0 100\n",argv[0]);
			printf("           %s image%%05d.png 0 100 -o out.h5\n",argv[0]);
			printf("           %s image%%05d.png 0 100 -o out.h5 -m 7 -d v1 flow\n\n",argv[0]);
			std::cout << desc << std::endl;
			return EXIT_FAILURE;
		}
	} else {
		inputFilePatter = vm["input.files"].as<std::string>();
		start_idx   = vm["input.start"].as<int>();
		stop_idx    = vm["input.stop"].as<int>();
		if (vm.count("input.crop-input")) {
			std::string crop_string = vm["input.crop-input"].as<std::string>();
			float xo,yo,xs,ys=-1;
			sscanf(crop_string.c_str(),"%f,%f,%f,%f",&xo,&yo,&xs,&ys);
			if (ys==-1) {
				std::cout << "Invalid crop string " << crop_string << "\n";
				std::cout << "Use e.g: 0.1,0.1,0.8,0.8 to cutt of a border.";
				std::cout << std::endl;
				return EXIT_FAILURE;
			}

			float factor = 1;
			
			while (fmax(fmax(xo,yo),fmax(xs,ys))>factor)
				factor *= 10;
			xo/=factor;yo/=factor;xs/=factor;ys/=factor;

			//imageSequence = new SimpleHostImageSequence(inputFilePatter.c_str()
			//											,start_idx,stop_idx
			//											,xo,yo
			//											,xs,ys);
			// TODO: fixme. replaced upper part with lower part, just to get compiled
			imageSequence = new SimpleHostImageSequence(inputFilePatter.c_str()
														,start_idx,stop_idx);
		} else {
			imageSequence = new SimpleHostImageSequence(inputFilePatter.c_str(),start_idx,stop_idx);
		}

		printf("iterate from:\n  ");
		printf(inputFilePatter.c_str(),start_idx);
		printf(" to\n  ");
		printf(inputFilePatter.c_str(),stop_idx);
		printf("\n");
	}

	/* --- output parameters --- */
	std::vector< std::string > op_temp;
	if (vm.count("output.data")) {
		op_temp = vm["output.data"].as< std::vector< std::string > >();
	} else {
		if (vm.count("output.show-results"))
			op_temp = std::vector<std::string>(1,"--non--");
	}
	std::set< std::string > output_data (op_temp.begin(),op_temp.end());
	std::string outputName   = vm["output.file"].as<std::string>();
	const bool 	debug_output = vm.count("output.debug");
	const bool	write_input  = vm.count("output.data")>0;
	const bool  write_init_det = output_data.count("init-features")>=1
								 || output_data.count("all")>=1;
	const bool  write_init_feaN= output_data.count("init-normed-features")>=1
								 || output_data.count("all")>=1;
	const bool  write_init_fea = output_data.count("init-detection")>=1
								 || output_data.count("all")>=1;
	const bool	write_flow =    output_data.count("flow")>=1
							 || output_data.count("all")>=1;
	const bool  write_flo  =    output_data.count("flo")>=1;
	const bool	write_v1   =    output_data.count("v1")>=1
							 || output_data.count("all")>=1;
	const bool	write_v1_n =    output_data.count("v1_normed")>=1
							 || output_data.count("all")>=1;
	const bool	write_v1_n_single =     output_data.count("v1_normed_single")>=1
									 || output_data.count("all")>=1;
	const bool	write_mt   =    output_data.count("mt")>=1
							 || output_data.count("all")>=1;
	const bool	write_mt_n =    output_data.count("mt_normed")>=1
							 || output_data.count("all")>=1;
	const bool	write_mt_n_single =     output_data.count("mt_normed_single")>=1
									 || output_data.count("all")>=1;
	const bool	write_mt_interpolated = output_data.count("mt_interpolated")>=1
									 || output_data.count("all")>=1;
	const float badPosThreshold    = vm["output.threshold-for-visualization"].as<float>();
	{ /* print write parameters to command line */
		if (op_temp.size()>0) {
			std::cout << "write data ( (" << op_temp.size() << ") ";
			for (size_t i=0; i<op_temp.size(); ++i)
				std::cout << op_temp[i] << " ";
			std::cout << ") to " << outputName << std::endl;
		} else {
			std::cout << "do not write data to file" << std::endl;
		}
		std::cout << "threshold-for-visualization      = " << badPosThreshold << std::endl;
	}

	/* --- feature parameters --- */
	{  /* print algorithm parameters parameters to command line */
		std::cout << "feature.norm0                   = " << vm["feature.norm0"].as<std::string>() << std::endl;
		std::cout << "feature.norm0.naka-rushton      = " << vm["feature.norm0.naka-rushton"].as<float>() << std::endl;
		std::cout << "feature.norm0.epsilon           = " << vm["feature.norm0.epsilon"].as<float>() << std::endl;
		std::cout << "feature.norm0.sigma-for-norm    = " << vm["feature.norm0.sigma"].as<float>() << std::endl;
		std::cout << "feature.norm.naka-rushton      = " << vm["feature.norm.naka-rushton"].as<float>() << std::endl;
		std::cout << "feature.norm.epsilon           = " << vm["feature.norm.epsilon"].as<float>() << std::endl;
		std::cout << "feature.norm.sigma-for-norm    = " << vm["feature.norm.sigma"].as<float>() << std::endl;
		std::cout << "feature.norm2             = " << vm["feature.norm2"].as<std::string>() << std::endl;
		std::cout << "feature.type              = " << vm["feature.type"].as<std::string>() << std::endl;
		std::cout << "feature.lambda0           = " << vm["feature.lambda0"].as<float>() << std::endl;
		std::cout << "feature.sigma-per-lambda  = " << vm["feature.sigma-per-lambda"].as<float>() << std::endl;
	}



	/* --- algorithm parameters --- */
	const int   maxVel     = vm["motion.max-vel"].as<int>();
	const int   velSpaceSize = (maxVel*2+1)*(maxVel*2+1);
	      int 	downsample_faktor    = vm["motion.mt-downsampling"].as<int>();
	const float sigma4MTIntegration  = vm["motion.sigma-for-mt-integration"].as<float>();
	      float	sigma4spatialInhibV1 = vm["motion.V1-sigma-for-spatial-inhibition"].as<float>();
	      float	sigma4spatialInhibMT = vm["motion.MT-sigma-for-spatial-inhibition"].as<float>();
	      float	fbf				   = vm["motion.feedback"].as<float>();
	{  /* print algorithm parameters parameters to command line */
		std::cout << "max-vel               = " << maxVel << std::endl;
		std::cout << "feedback              = " << fbf << std::endl;
		std::cout << "   sigma4spatialInhib = " << sigma4spatialInhibV1 << std::endl;
		std::cout << "   div-const          = " << vm["motion.V1-div-const"].as<float>() << std::endl;
		std::cout << "   div-factor         = " << ((vm.count("motion.V1-div-factor-scaled"))?(vm["motion.V1-div-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.V1-div-factor"].as<float>()))  << std::endl;
		std::cout << "   div-self           = " << vm["motion.V1-div-self"].as<float>()  << std::endl;
		std::cout << "   sub-factor         = " << ((vm.count("motion.V1-sub-factor-scaled"))?(vm["motion.V1-sub-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.V1-sub-factor"].as<float>()))  << std::endl;
		std::cout << "mt-downsample         = " << downsample_faktor << std::endl;
		std::cout << "sigma4MTIntegration   = " << sigma4MTIntegration << std::endl;
		std::cout << "MT sigma4spatialInhib = " << sigma4spatialInhibMT  << std::endl;
		std::cout << "   div-const          = " << vm["motion.MT-div-const"].as<float>() << std::endl;
		std::cout << "   div-factor         = " << ((vm.count("motion.MT-div-factor-scaled"))?(vm["motion.MT-div-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.MT-div-factor"].as<float>()))  << std::endl;
		std::cout << "   div-self           = " << vm["motion.MT-div-self"].as<float>()  << std::endl;
		std::cout << "   sub-factor         = " << ((vm.count("motion.MT-sub-factor-scaled"))?(vm["motion.MT-sub-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.MT-sub-factor"].as<float>()))  << std::endl;
		std::cout << "dt = " << vm["motion.dt"].as<float>()  << std::endl;
	}
	const bool  use_form_guided_integration = vm.count("motion.use-form");
	const bool  use_INRIA_Form = use_form_guided_integration &&
					( vm["motion.use-form"].as<std::string>() == "INRIA" ||
					  vm["motion.use-form"].as<std::string>() == "inria");
	const bool  use_UULM_Form = use_form_guided_integration &&
					( vm["motion.use-form"].as<std::string>() == "UULM" ||
					  vm["motion.use-form"].as<std::string>() == "uulm");
	const bool  use_NEW_Form = use_form_guided_integration &&
					( vm["motion.use-form"].as<std::string>() == "NEW" ||
					  vm["motion.use-form"].as<std::string>() == "new");
	if (use_form_guided_integration && !(use_INRIA_Form||use_UULM_Form||use_NEW_Form))
		throw std::runtime_error("invalid form specified\n");
	const float gauss_sigma_base = 0.75f;
	const float sigma4_real_downsampling = gauss_sigma_base*downsample_faktor;
	      float mt_integration_sigma_reduced =
						sqrt(sigma4MTIntegration*sigma4MTIntegration-
								sigma4_real_downsampling*sigma4_real_downsampling)
								/ downsample_faktor;
	if (use_form_guided_integration) {
		std::cout << "use form guided integration" << std::endl;
		std::cout << "       Thus, sigma_large (in V1 resolution) is "
				  << sigma4MTIntegration << std::endl;
		std::cout << "       and sigma_small is " << sigma4_real_downsampling;
		std::cout << ".\n       The rest is blured with a circular gaussian"
				 "with a sigma of " << sigma4_real_downsampling << std::endl;
	} else if (sigma4_real_downsampling>sigma4MTIntegration && downsample_faktor!=1) {
		std::cout << "Error: sigma for mt integration is too small compared to\n";
		std::cout << "       the subsampling rate!";
		return -1;
	} else {
		std::cout << "Because of subsampling with minimal prebluring, the reduced\n";
		std::cout << "mt integration sigma with respect to v1 resolution is "
				  << mt_integration_sigma_reduced * downsample_faktor;
		std::cout << std::endl;
	}

	const int w = imageSequence->getWidth();
	const int h = imageSequence->getHeight();
	std::cout << "image sizes (w,h) = (" << w << "," << h << ")" << std::endl;
	std::cout << "\n" << std::endl;

#if defined(HAVE_GTK)
	DataWindow* init_det_window     = 0;
	DataWindow* init_det_sum_window = 0;
	DataWindow* flow_v1_window    = 0;
	DataWindow* v1_det_sum_window = 0;
	DataWindow* v1_det_window     = 0;
	DataWindow* flow_mt_window    = 0;
	DataWindow* mt_det_window     = 0;
	DataWindow* mt_det_sum_window = 0;
	
	if (vm.count("output.show-results")) {
		flow_v1_window->global_gtk_gl_init();
		flow_v1_window = new DataWindow(w,h,"V1 flow");
		flow_v1_window->setArrowSpacing(10);
		flow_v1_window->setArrowStyle(DataWindow::LINES);
	}
	if (vm.count("output.show-results")) {
		flow_mt_window = new DataWindow(w,h,"MT flow");
		flow_mt_window->setArrowSpacing(10);
		flow_mt_window->setArrowStyle(DataWindow::LINES);
	}
	if (vm.count("output.show-results")) {
		init_det_window  = 0;//new DataWindow(w,h,"initial motion detection");
		init_det_sum_window = new DataWindow(w,h,"initial motion detection feature sum");
		v1_det_window= 0;//new DataWindow(w,h,"v1 features");
		v1_det_sum_window= new DataWindow(w,h,"v1 features sum");
		mt_det_sum_window= new DataWindow(w,h,"mt features sum");
		mt_det_window= 0;//new DataWindow(w/downsample_faktor,h/downsample_faktor,"mt features");
	}
#else
	void* flow_v1_window=0; // these variables are used (among some others) to
	void* flow_mt_window=0; // determine the requirement of calculate a flow from
							// the activity
#endif

	int inPlaceIterations = vm["motion.local-steps"].as<int>();

	GaussKernel2   blur(0.0f);
	GPUImage 	   tempImage(w,h);
	FeatureExtractor* featureExtractor;
	bool		   useLoG = vm["feature.type"].as<std::string>() == "log" ||
							vm["feature.type"].as<std::string>() == "LoG";
	const bool write_each_iteration   = false;
	bool use_precalulated_initial_vel = false; // if there is enough space on the
											// gpu, this could be set to true (speedup)

	{
		/* file to store the output */
#ifdef HAVE_HDF5
		HDF5RawDataFileAdapter	outFile(outputName.c_str(),true);
		size_t size_v1_sequence[4] = {stop_idx-start_idx,h,w,(maxVel*2+1)*(maxVel*2+1)};
		if (write_each_iteration) size_v1_sequence[0]*=(inPlaceIterations+1);
		if (write_v1)			outFile.createDataSet("v1",4,size_v1_sequence);
		if (write_v1_n)			outFile.createDataSet("v1_normed",4,size_v1_sequence);
		size_t size_mt_sequence[4] = {stop_idx-start_idx,h/downsample_faktor,w/downsample_faktor,(maxVel*2+1)*(maxVel*2+1)};
		if (write_each_iteration) size_mt_sequence[0]*=(inPlaceIterations+1);
		if (write_mt)			outFile.createDataSet("mt",4,size_mt_sequence);
		if (write_mt_n)			outFile.createDataSet("mt_normed",4,size_mt_sequence);
		if (write_mt_interpolated) outFile.createDataSet("mt_interpolated",4,size_v1_sequence);
#endif

		/* space for the initial motion estimation*/
		GPUData 		data_1(h,w,maxVel*2+1,maxVel*2+1);
		GPUData 		data_2(h,w,maxVel*2+1,maxVel*2+1);
		GPUData 		data_3(h,w,maxVel*2+1,maxVel*2+1);
						data_3.fill(0.0f);
		GPUData 		data_1s(h/downsample_faktor,w/downsample_faktor,maxVel*2+1,maxVel*2+1);
		GPUData 		data_2s(h/downsample_faktor,w/downsample_faktor,maxVel*2+1,maxVel*2+1);

		/* Gabor feature extractor */
		if (useLoG) {
			LoGFeatureExtractor* tmp = new LoGFeatureExtractor(w,h,12,vm["feature.lambda0"].as<float>()/sqrt(12.0f));
			tmp->setLambda(vm["feature.lambda0"].as<float>());
			featureExtractor = tmp;
		} else {
			GaborFeatureExtractor* 	gf = new GaborFeatureExtractor( w,h,GaborFeatureExtractor::COS,
												12,1,
												vm["feature.lambda0"].as<float>(),
												vm["feature.sigma-per-lambda"].as<float>());
			featureExtractor = gf;
		}
		

		GradientFeatureExtractor grad;
								 grad.extrapolatDirections(8);

		FeatureNormalization    normalizer0;
								normalizer0.set_features_first();
								normalizer0.use_floating_average_for_max(0.2f);
								if (vm["feature.norm0"].as<std::string>()=="true" ||
									vm["feature.norm0"].as<std::string>()=="1")
								{
									if (vm["feature.norm0.sigma"].as<float>()>0)
										normalizer0.set_sigma(vm["feature.norm0.sigma"].as<float>());
									if (vm["feature.norm0.naka-rushton"].as<float>()>0)
										normalizer0.use_naka_rushton(vm["feature.norm0.naka-rushton"].as<float>(),true);
									if (vm["feature.norm0.epsilon"].as<float>()>0)
										normalizer0.use_epsilon(vm["feature.norm0.epsilon"].as<float>());
									if (vm["feature.norm0.naka-rushton"].as<float>()>0 && vm["feature.norm0.epsilon"].as<float>()>0) {
										std::cout << "WARNIG: Feature normalization (0) ";
										std::cout << "is specified ambiguous! epsilon normalization ";
										std::cout << "is applied!" << std::endl;
									}
								} else {
									normalizer0.enable(false);
								}
		FeatureNormalization    normalizer;
								normalizer.set_features_first();
								if (vm["feature.norm.sigma"].as<float>()>0)
									normalizer.set_sigma(vm["feature.norm.sigma"].as<float>());
							    if (vm["feature.norm.naka-rushton"].as<float>()>0)
							    	normalizer.use_naka_rushton(vm["feature.norm.naka-rushton"].as<float>(),true);
							    if (vm["feature.norm.epsilon"].as<float>()>0)
							    	normalizer.use_epsilon(vm["feature.norm.epsilon"].as<float>());
							    if (vm["feature.norm.naka-rushton"].as<float>()>0 && vm["feature.norm.epsilon"].as<float>()>0) {
									std::cout << "WARNIG: Feature normalization ";
									std::cout << "is specified ambiguous! epsilon normalization ";
									std::cout << "is applied!" << std::endl;
							    }
							    if (vm["feature.norm"].as<std::string>()=="true" ||
									vm["feature.norm"].as<std::string>()=="1")
								{
							    	normalizer0.enable(true);
								} else {
									normalizer0.enable(false);
								}
		FeatureNormalization    normalizer2;
								normalizer2.set_features_first();
								normalizer2.use_floating_average_for_max(0.2f);
								if (vm["feature.norm2"].as<std::string>()=="true" ||
									vm["feature.norm2"].as<std::string>()=="1")
								{
									normalizer2.use_maxsum();
								} else {
									normalizer2.enable(false);
								}

		/* correlator */
		CartesianCorrelator     co(w,h,maxVel);//,0.0f,0.0f);

								if (vm["feature.use_old_prior"].as<int>()>0)
									co.usePriorToLowVelocities();
								
								if (vm["feature.use_prior"].as<float>()>0) {
									co.set_use_new_prior(vm["feature.use_prior"].as<float>());
									co.set_sigma_of_input_features_unsharpnes(vm["feature.prior_sigma"].as<float>());
								}


		ModelArea model_area_V1;
				  model_area_V1.set_do_square(vm["motion.V1-square"].as<bool>());
				  model_area_V1.set_feedback_gain(fbf);
//				  model_area_V1.set_shunting_params(  4
//													, 4
//													, 1.0f
//													, 4);
//				  model_area_V1.set_shunting_params(  0.015f
//													, 20.0f
//													, 1.0f
//													, 0.25f);
				  model_area_V1.set_shunting_params(  vm["motion.V1-div-const"].as<float>()
													, (vm.count("motion.V1-div-factor-scaled"))?(vm["motion.V1-div-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.V1-div-factor"].as<float>())
													, vm["motion.V1-div-self"].as<float>()
													, (vm.count("motion.V1-sub-factor-scaled"))?(vm["motion.V1-sub-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.V1-sub-factor"].as<float>()));
				  model_area_V1.set_kernel_type(ModelArea::GAUSS);
#if 1
				  model_area_V1.set_sigma(0,0.75f);
				  model_area_V1.set_sigma(1,0.75f);
#else
				  model_area_V1.set_sigma(2,0.0f);
				  model_area_V1.set_sigma(3,0.0f);
#endif
				  model_area_V1.set_sigma(2,vm["motion.sigma-for-vel-blur"].as<float>());
				  model_area_V1.set_sigma(3,vm["motion.sigma-for-vel-blur"].as<float>());
				  model_area_V1.set_sigma_for_spatial_inhibition(sigma4spatialInhibV1);
				  model_area_V1.set_dt(vm["motion.dt"].as<float>());

				  if (vm.count("motion.recurrent-inhibition"))
					  model_area_V1.set_use_recurrent_inhibition();
				  if (vm.count("motion.rectified-simple-input"))
					  model_area_V1.set_use_rectified_simple_input();
//				  model_area_V1.set_INRIA_dynamic(4*(maxVel*2+1)*(maxVel*2+1),4);

	    GaussKernel2         gauss_for_downsampling(sigma4_real_downsampling,0,1);
	    INRIAForm            inriaForm;
	    ModulatedIntegration mod_integ(sigma4MTIntegration,sigma4_real_downsampling,sigma4_real_downsampling);
	    		//sigma4_real_downsampling);

		ModelArea model_area_MT;
				  model_area_MT.set_do_square(vm["motion.MT-square"].as<bool>());
				  model_area_MT.set_feedback_gain(0.0f);
//				  model_area_MT.set_shunting_params(  4
//													, 2
//													, 1
//													, 2);
//				  model_area_MT.set_shunting_params(  0.086f
//													, 2.0f
//													, 1.0f
//													, 0.25f);
				  model_area_MT.set_shunting_params(  vm["motion.MT-div-const"].as<float>()
													, (vm.count("motion.MT-div-factor-scaled"))?(vm["motion.MT-div-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.MT-div-factor"].as<float>())
													, vm["motion.MT-div-self"].as<float>()
													, (vm.count("motion.MT-sub-factor-scaled"))?(vm["motion.MT-sub-factor-scaled"].as<float>()*velSpaceSize):(vm["motion.MT-sub-factor"].as<float>()));
				  model_area_MT.set_kernel_type(use_NEW_Form ? ModelArea::MODULATED_DIFFUSION : ModelArea::GAUSS);
				  if (use_form_guided_integration) {
					  if (use_NEW_Form) {
						  model_area_MT.set_sigma(0,mt_integration_sigma_reduced);
						  model_area_MT.set_sigma(1,0.1f);
						  model_area_MT.set_sigma(2,0.9f);
					  } else {
						  model_area_MT.set_sigma(0,0);
						  model_area_MT.set_sigma(1,0);
					  }
				  } else {
					  model_area_MT.set_sigma(0,mt_integration_sigma_reduced);
					  model_area_MT.set_sigma(1,mt_integration_sigma_reduced);
					  model_area_MT.set_sigma(2,vm["motion.sigma-for-vel-blur"].as<float>());
					  model_area_MT.set_sigma(3,vm["motion.sigma-for-vel-blur"].as<float>());
				  }

				  model_area_MT.set_sigma_for_spatial_inhibition(sigma4spatialInhibMT);
				  model_area_MT.set_dt(vm["motion.dt"].as<float>());

				  if (vm.count("motion.recurrent-inhibition"))
					  model_area_MT.set_use_recurrent_inhibition();
				  if (vm.count("motion.rectified-simple-input"))
					  model_area_MT.set_use_rectified_simple_input();
//				  model_area_MT.set_INRIA_dynamic(4*(maxVel*2+1)*(maxVel*2+1),4);

		/* two images and their features (t and t-1) */
		GPUImage  currentImage(w,h);
		GPUImage  pastImage(w,h);
				  imageSequence->copyNextImage(pastImage);
				  blur.apply(pastImage,currentImage);
		GPUData   currentFeature = featureExtractor->getFeatureImages(currentImage);
				  normalizer0(currentFeature);
				  normalizer(currentFeature);
				  normalizer2(currentFeature);
		GPUData   pastFeature = featureExtractor->getFeatureImages(currentImage);

		GPUData temp;
		GaussKernel2 blur_new(2.0f); // for new form integration

		GPUData grad_features = grad.getFeatureImages(currentImage);

		if (write_input) {
			//writeGrayImage(outFile,imageName(0),currentImage);
		}

		printf("image values in [%f,%f]\n",currentImage.min(),currentImage.max());

		/* begin of iteration loop */
		int i=-1;
		while (imageSequence->hasNextImage()) {

			std::cout << "iterate " << ++i << std::flush;

			/* the old current will be the new past */
			currentImage.swap(pastImage);
			currentFeature.swap(pastFeature);

			/* get the new current and its features */
			imageSequence->copyNextImage(tempImage);
			blur.apply(tempImage,currentImage);
			featureExtractor->getFeatureImages(currentImage,currentFeature); // TODO: step 1 get gabor features
			{ /* just ouput */
				if (debug_output) {
					const float maximum = currentFeature.max();
					const float minimum = currentFeature.min();
					const int   size    = currentFeature.size();
					const float summe   = currentFeature.sum();
					const int   no      = int (currentFeature.extent(0));
					printf( "\nInitial features (%d) activity:\n max =%f\n"
							" min =%f\n mean=%f\n",no,maximum,minimum,summe/size);
				}
				if (write_init_fea) {
					//writeFeatures(outFile,featureName(i*(inPlaceIterations+1)),currentFeature);
				}
			}
			normalizer0(currentFeature);  // TODO: step 2a normalize features
			{ /* just ouput */
				if (debug_output) {
					const float maximum = currentFeature.max();
					const float minimum = currentFeature.min();
					const int   size    = currentFeature.size();
					const float summe   = currentFeature.sum();
					printf( "\nFeatures after normalization 0:\n max =%f\n"
							" min =%f\n mean=%f\n",maximum,minimum,summe/size);
				}
				if (write_init_feaN) {
					//writeFeatures(outFile,featureNormed0Name(i*(inPlaceIterations+1)),currentFeature);
				}
			}
			normalizer(currentFeature);  // TODO: step 2b normalize features
			{ /* just ouput */
				if (debug_output) {
					const float maximum = currentFeature.max();
					const float minimum = currentFeature.min();
					const int   size    = currentFeature.size();
					const float summe   = currentFeature.sum();
					printf( "\nFeatures after normalization:\n max =%f\n"
							" min =%f\n mean=%f\n",maximum,minimum,summe/size);
				}
				if (write_init_feaN) {
					//writeFeatures(outFile,featureNormedName(i*(inPlaceIterations+1)),currentFeature);
				}
			}
			normalizer2(currentFeature);  // TODO: step 2c normalize features
			{ /* just ouput */
				if (debug_output) {
					const float maximum = currentFeature.max();
					const float minimum = currentFeature.min();
					const int   size    = currentFeature.size();
					const float summe   = currentFeature.sum();
					printf( "\nFeatures after second normalization:\n max =%f\n"
							" min =%f\n mean=%f\n",maximum,minimum,summe/size);
				}
				if (write_init_feaN) {
					//writeFeatures(outFile,featureNormed2Name(i*(inPlaceIterations+1)),currentFeature);
				}
			}

			if (use_form_guided_integration) {
				grad.getFeatureImages(pastImage,grad_features);
			}

			for (int j=0; j<=inPlaceIterations; ++j) {

				std::cout << " . " << std::flush;

				/*
				 * get the initial correlation
				 */
				if (inPlaceIterations>0 && use_precalulated_initial_vel) { // normaly false!!!
					if (j==0) {
						co.correlateAll(pastFeature,currentFeature,data_1);  // TODO: step 3 initial motion detection
						try {
							temp.resize(data_1.getDimensionsCount(),data_1.getDimensions());
							temp.copyFrom(data_1);
						} catch (GPUDataException e) {
							std::cout << "not enough space to permanently store initial motion detection between frames" << std::endl;
							use_precalulated_initial_vel = false;
						}
					} else {
						data_1.copyFrom(temp);
					}
				} else {
					co.correlateAll(pastFeature,currentFeature,data_1); // TODO: step 3 initial motion detection
				}

				{ /* show initial motion detection */
					if (debug_output && j==0) {
						GPUVelActivities p0(data_1);
						const float maximum = data_1.max();
						const float minimum = data_1.min();
						const int   size    = data_1.size();
						const float summe   = data_1.sum();
						const float maxh    = GPUVelActivities(data_1).getFeatureSum().max();
						printf( "\nInitial Motion Estimation activity:\n max =%f\n"
								" min =%f\n mean=%f\n maxh=%f\n",maximum,minimum,summe/size,maxh);

					}
					if (write_init_det && j==0) {
						//writeVelocityActivities(outFile,initDetName(i*(inPlaceIterations+1)+j),GPUVelActivities(data_1));
					}
#if defined(HAVE_GTK)
					if (init_det_sum_window && j==0) {
						GPUVelActivities p0(data_1);
						GPUImage im (p0.getFeatureSum());
						init_det_sum_window->update(im);
						if (!write_each_iteration) {
							if (j==inPlaceIterations) {
								init_det_sum_window->writeToFile(imageName(i)+"_init_det_sum.png");
							}
						} else
							init_det_sum_window->writeToFile(imageName(i*(inPlaceIterations+1)+j)+"_init_det_sum.png");
					}
					if (init_det_window && j==0) {
						GPUVelActivities p0(data_1);
						int sx = w/2-1 > 15 ? 15 : w/2-1;
						int sy = h/2-1 > 15 ? 15 : h/2-1;
						init_det_window->update(p0.getVelocitySpacesInOneImage(false,w/4,h/4,sx,sy));
						if (!write_each_iteration) {
							if (j==inPlaceIterations) {
								init_det_window->writeToFile(imageName(i)+"_init_det.png");
							}
						} else
							init_det_window->writeToFile(imageName(i*(inPlaceIterations+1)+j)+"_init_det.png");
					}
#endif
				}

				/*
				 * computes V1
				 */
				model_area_V1.iterate( data_1,data_2,data_3 ); // TODO: step 4: V1

				{ /* write input and V1 results */
					if (debug_output) {
						const float maximum = data_2.max();
						const float minimum = data_2.min();
						const int   size    = data_2.size();
						const float summe   = data_2.sum();
						const float maxh    = GPUVelActivities(data_2).getFeatureSum().max();
						printf("V1 activity:\n max =%f\n min =%f\n mean=%f\n maxh=%f\n",maximum,minimum,summe/size,maxh);
					}
					if (write_input && j==0) {
						//writeGrayImage(outFile,imageName((i+1)*(inPlaceIterations+1)).c_str(),currentImage);
					}
#if defined(HAVE_GTK)
					if (v1_det_sum_window) {
						GPUVelActivities v1(data_2);
						GPUImage im (v1.getFeatureSum());
						v1_det_sum_window->update(im);
						if (!write_each_iteration) {
							if (j==inPlaceIterations) {
								v1_det_sum_window->writeToFile(imageName(i)+"_v1_det_sum.png");
							}
						} else
							v1_det_sum_window->writeToFile(imageName(i*(inPlaceIterations+1)+j)+"_v1_det_sum.png");
					}
					if (v1_det_window && j==0) {
						GPUVelActivities p1(data_2);
						int sx = w/2-1 > 15 ? 15 : w/2-1;
						int sy = h/2-1 > 15 ? 15 : h/2-1;
						v1_det_window->update(p1.getVelocitySpacesInOneImage(false,2,2,sx,sy));
						if (!write_each_iteration) {
							if (j==inPlaceIterations) {
								v1_det_window->writeToFile(imageName(i)+"_v1_det.png");
							}
						} else
							v1_det_window->writeToFile(imageName(i*(inPlaceIterations+1)+j)+"_v1_det.png");
					}
#endif
					if (write_flow || write_flo || write_v1 || write_v1_n || write_v1_n_single || debug_output || flow_v1_window) {
						GPUVelActivities v1((badPosThreshold<=0 && !write_v1_n) ? data_2 : data_1);
						if (badPosThreshold>0 || write_v1_n) data_1.copyFrom(data_2); // use data_1 as temporary
						if (write_v1 && (j==inPlaceIterations || write_each_iteration)) {
							//writeVelocityActivities(outFile,"v1",v1,write_each_iteration?(i-0)*(inPlaceIterations+1)+j:i);							
						}
						if (write_v1_n || badPosThreshold) {
							float m = v1.getFeatureSum().max();
//							std::cout << m;
							if (m<=0.0f) m += std::numeric_limits<float>::epsilon();
							v1.getData().div(m);
							if (debug_output) {
								printf(" norm by maxh =%f\n",m);
							}
						}
						if (write_v1_n && (j==inPlaceIterations || write_each_iteration)) {
							//writeVelocityActivities(outFile,"v1_normed",v1,write_each_iteration?(i-0)*(inPlaceIterations+1)+j:i);
						}
						if (write_v1_n_single && (j==inPlaceIterations || write_each_iteration)) {
							//writeVelocityActivities(outFile, hypoName((i-0)*(inPlaceIterations+1)+j),v1);
						}
						if (badPosThreshold>0.0f) v1.shutOfBadPositions(badPosThreshold);
						if (write_flow || write_flo || debug_output || flow_v1_window) {
							GPUVelField resFlowFieldV1(v1.getFlowField());
							GPUData d = resFlowFieldV1.getSpeed();
							if (debug_output) {
								const float maximum = d.max();
								const float minimum = d.min();
								const int   size    = d.size();
								const float summe   = d.sum();
								printf(" max_speed =%f\n min_speed =%f\n mean_speed=%f\n",maximum,minimum,summe/size);
							}
#if defined(HAVE_GTK)
							if (flow_v1_window) {
								flow_v1_window->update(pastImage,resFlowFieldV1);
							}
#endif
							if ((write_flow || write_flo) && (j==inPlaceIterations || write_each_iteration)) {
								//writeVelocityField(outFile,flowFieldName(i*(inPlaceIterations+1)+j),resFlowFieldV1);
								if (write_flo) {
									GPUVelActivities(data_2).getFlowField(resFlowFieldV1,0.0001f);
									if (write_each_iteration)
										MiddleburyFlowIO::WriteFlowFile(resFlowFieldV1,flowFieldName(i*(inPlaceIterations+1)+j)+".flo");
									else 
										MiddleburyFlowIO::WriteFlowFile(resFlowFieldV1,flowFieldName(i)+".flo");
								}
							}
						}
					}
				}

				/*
				 * downsampling
				 */
				if (use_form_guided_integration) {
					if (use_UULM_Form) {
						mod_integ.setInput(data_2);
						GPUData temp(grad_features.copy());
						temp.abs();
						mod_integ.setModulator(temp);
						mod_integ.work();
	//					GPUImage im = temp.sumContract(0,1);
	//					DataWindow::show(mod_integ._temp1,"modulator sum");
					} else if (use_INRIA_Form) {
						inriaForm.iterate(pastImage,data_2,data_3);
//						GPUData weights = inriaForm.weights;
//						GPUData phi = inriaForm.phi;
//						DataWindow window(w*3,h*3);
//						for (int i=0; i<phi.extent(0); ++i) {
//							GPUImage im(phi[i]);
//							window.update(im);
//							window.busy_wait(1000);
//						}
						gauss_for_downsampling.apply(data_3,data_2);
					} else if (use_NEW_Form) {
						gauss_for_downsampling.apply(data_2,data_3); data_3.swap(data_2);
						GPUImage diffX = pastImage.diffX(true);
						GPUImage diffY = pastImage.diffY(true);
						diffX.pythagoras(diffX,diffY);
						diffX.div(diffX.max()+0.0000001f);
						
						gauss_for_downsampling.apply(diffX,diffY);
						diffX = GPUImage(w/downsample_faktor,h/downsample_faktor);
						downsample(diffY,diffX,downsample_faktor,0);
/// 						diffX.add(0.0001f);
/// 						GPUImage save=diffY.copy();
/// 						blur_new.apply(diffX,diffY);
/// 						diffX.copyFrom(save);
/// 						diffX.div(diffY);
/// 						diffX.div(diffX.max());
/// //						DataWindow::show(diffX,"befor bluring",true);
/// 						gauss_for_downsampling.apply(diffX,diffY);
/// //						DataWindow::show(diffY,"after bluring",true);
/// 						diffX = GPUImage(w/downsample_faktor,h/downsample_faktor);
/// 						downsample(diffY,diffX,downsample_faktor,0);
/// 						diffX.div(diffX.max());
/// //						diffX.sub(1.0f);
/// //						diffX.mul(-1.0f);
/// //						DataWindow::show(diffX,"after downsampling",true);
						model_area_MT.set_modulator(diffX);
						gauss_for_downsampling.apply(data_2,data_3); data_3.swap(data_2);
					}
				} else {
					gauss_for_downsampling.apply(data_2,data_3); data_3.swap(data_2);
				}
				downsample(data_2,data_1s,downsample_faktor,0);  // TODO: step 5: V1->MT downsampling

				/*
				 * computes MT
				 */
				model_area_MT.iterate(data_1s,data_2s);  // TODO: step 6: MT

				{ /* write MT results */
					if (debug_output) {
						const float maximum = data_2s.max();
						const float minimum = data_2s.min();
						const int   size    = data_2s.size();
						const float summe   = data_2s.sum();
						const float maxh    = GPUVelActivities(data_2s).getFeatureSum().max();
						printf("MT activity:\n max =%f\n min =%f\n mean=%f\n maxh=%f\n",maximum,minimum,summe/size,maxh);
					}
#if defined(HAVE_GTK)
					if (mt_det_sum_window) {
						GPUVelActivities mt(data_2s);
						GPUImage im (mt.getFeatureSum());
						mt_det_sum_window->update(im);
					}
#endif
					if (write_flow || write_flo || write_mt || write_mt_n || write_mt_n_single || debug_output || flow_mt_window) {
						GPUVelActivities mt((badPosThreshold<=0 && !write_mt_n) ? data_2s : data_1s);
						if (badPosThreshold>0 || write_mt_n) data_1s.copyFrom(data_2s); // use data_1 as temporary
						if (write_mt && (j==inPlaceIterations || write_each_iteration)) {
							//writeVelocityActivities(outFile,"mt",mt,write_each_iteration?(i-0)*(inPlaceIterations+1)+j:i);
						}
						if (write_mt_n || badPosThreshold) {
							float m = mt.getFeatureSum().max();
//							std::cout << m;
							if (m<=0.0f) m += std::numeric_limits<float>::epsilon();
							mt.getData().div(m);
						}
						if (write_mt_n && (j==inPlaceIterations || write_each_iteration)) {
							//writeVelocityActivities(outFile,"mt_normed",mt,write_each_iteration?(i-0)*(inPlaceIterations+1)+j:i);
						}
						if (write_mt_n_single && (j==inPlaceIterations || write_each_iteration)) {
							//writeVelocityActivities(outFile,hypo_mtName((i-0)*(inPlaceIterations+1)+j),mt);
						}
						if (badPosThreshold>0.0f) mt.shutOfBadPositions(badPosThreshold);
						if (write_flow || write_flo || debug_output || flow_mt_window) {
							GPUVelField resFlowFieldMT(mt.getFlowField());
							GPUData d = resFlowFieldMT.getSpeed();
							if (debug_output) {
								const float maximum = d.max();
								const float minimum = d.min();
								const int   size    = d.size();
								const float summe   = d.sum();
								printf( " max_speed =%f\n min_speed =%f\n"
										" mean_speed=%f\n",maximum,minimum,summe/size);
							}
#if defined(HAVE_GTK)
							if (flow_mt_window) {
								flow_mt_window->update(pastImage,resFlowFieldMT);
								if (vm.count("output.delay")) {
									DataWindow::busy_wait(vm["output.delay"].as<int>());
								}
							}
							if (mt_det_window && j==0) {
								int sx = (w/downsample_faktor)/2-1 > 6 ? 6 : (w/downsample_faktor)/2-1;
								int sy = (h/downsample_faktor)/2-1 > 6 ? 6 : (w/downsample_faktor)/2-1;
								mt_det_window->update(mt.getVelocitySpacesInOneImage(false,2,2,sx,sy));
								if (!write_each_iteration)
									mt_det_window->writeToFile(imageName(i)+"_mt_det.png");
								else
									mt_det_window->writeToFile(imageName(i*(inPlaceIterations+1)+j)+"_mt_det.png");
							}
#endif
							if ((write_flow || write_flo) && (j==inPlaceIterations || write_each_iteration)) {
								/*writeVelocityField(outFile,
										flowField_mtName(
												i*(inPlaceIterations+1)+j),
										resFlowFieldMT);*/
								if (write_flo) {
									GPUVelActivities(data_2s).getFlowField(resFlowFieldMT,0.0001f);
									if (write_each_iteration)
										MiddleburyFlowIO::WriteFlowFile(resFlowFieldMT,flowField_mtName(i*(inPlaceIterations+1)+j)+".flo");
									else 
										MiddleburyFlowIO::WriteFlowFile(resFlowFieldMT,flowField_mtName(i)+".flo");
								}
							}
						}
					}
				}
					 
				{ // prediction (if end of in place iterations) and extrapolation
					GPUVelActivities mt_res(data_2s);
					GPUVelActivities fb_data(data_3);
					if (j<inPlaceIterations || (write_mt_interpolated && (j==inPlaceIterations || write_each_iteration))) {
						if (downsample_faktor==1) 
							fb_data.getData().copyFrom(mt_res.getData());// TODO: step 7: extrapolate feedback
						else 
							mt_res.computeExtrapolation(fb_data);// TODO: step 7: extrapolate feedback
						if (write_mt_interpolated && (j==inPlaceIterations || write_each_iteration)) {
							if (write_flow) {
								GPUVelField resFlowFieldMTLarge(fb_data.getFlowField());
								/*writeVelocityField(outFile,
													flowField_mt_large_Name(
															i*(inPlaceIterations+1)+j),
													resFlowFieldMTLarge);*/
							}
							//writeVelocityActivities(outFile,"mt_interpolated",fb_data,write_each_iteration?(i-0)*(inPlaceIterations+1)+j:i);
						}
					}
					if (j==inPlaceIterations) {
						mt_res.computePredictionAndExtrapolate(fb_data); // TODO: step 7: extrapolate and predict feedback
					}
				}
			}
			std::cout << std::endl;
		}
		std::cout << "END" << std::endl;
	}

#if defined(HAVE_GTK)
	if (flow_v1_window) delete flow_v1_window;
	if (flow_mt_window) delete flow_mt_window;
	if (init_det_window) delete init_det_window;
#endif

	delete featureExtractor;
	delete imageSequence;

	return EXIT_SUCCESS;
}


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
