/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cmath>

#include <boost/foreach.hpp>

#ifndef __kernels_hpp_
class DerivativeSimpleKernel1;
class Binom2Kernel1;
class SobelOperator;
class SobelOperatorScaled;
#endif
#include "GPUImage.h"
#include "kernels.hpp"

#ifdef _WIN32
#ifndef M_PI
 #define M_PI 3.14159265358979323846264338327950288
#endif
#define fmin(a,b) (a < b ? a : b)
#define fmax(a,b) (a > b ? a : b)
#define fminf(a,b) (a < b ? a : b)
#define fmaxf(a,b) (a > b ? a : b)
#endif // _WIN32

void fillLinSpace(GPUData& d, double start_A, double end_A) {
	float* hData = new float[d.size()];
	const  int s = d.size();

	for (int i=0; i<s;++i) {
		hData[i] = float(start_A+(i*(end_A-start_A))/(s-1.0));
	}

	d.copyFromHostMemory(hData);

	delete[] hData;
}


void fillWithCos(GPUData& d, double omega) {
	float* hData = new float[d.size()];
	const  int s = d.size();

	int center = (d.size()-1)/2;
	for (int i=0; i<s;++i) {
		hData[i] = float(cos((i-center)*omega));
	}

	d.copyFromHostMemory(hData);

	delete[] hData;
}

void fillWithSin(GPUData& d, double omega) {
	float* hData =  new float[d.size()];
	const  int s = d.size();

	int center = (d.size()-1)/2;
	for (int i=0; i<s;++i) {
		hData[i] = float(sin((i-center)*omega));
	}

	d.copyFromHostMemory(hData);

	delete[] hData;
}

void fillWithGauss(GPUData& d, double sigma)
{
  float* hData =  new float[d.size()];
  const  int s = d.size();

  int center = (d.size()-1)/2;
  double sum= 0;

  for (int i=0; i<s;++i) {
    hData[i] = exp(-0.5*((i-center)*(i-center))/(sigma*sigma));
    sum += hData[i];
  }

  for (int i=0; i<d.size();++i) {
    hData[i] = hData[i]/sum;
  }

  d.copyFromHostMemory(hData);

  delete[] hData;
}


void fillWithGauss2(GPUImage& d, double covar_xx, double covar_xy, double covar_yy)
{
  const int nx = d.getWidth();
  const int ny = d.getHeight();

  const int cx = (nx-1)/2;
  const int cy = (ny-1)/2;

  float* hData =  new float[nx*ny];

  float* hDataIt = hData;
  double sum = 0;
  {
	  for (int y=0; y<ny; ++y) {
	  for (int x=0; x<nx; ++x) {
		  const float f = exp( -0.5*( (x-cx)*(x-cx)*covar_xx + 2*(x-cx)*(y-cy)*covar_xy + (y-cy)*(y-cy)*covar_yy ) );
		  (*hDataIt) = f;
		  sum += f;
		  hDataIt++;
	  }
	  }
  }

  d.copyFromHostMemory(hData);
  delete[] hData;
  d.div(sum);
}

void fillWithGauss2Rotated(GPUImage& d, double sigma_x, double sigma_y, double phi)
{
	const double covar_xx = 1.0/(sigma_x*sigma_x);
	const double covar_yy = 1.0/(sigma_y*sigma_y);

	const int nx = d.getWidth();
	const int ny = d.getHeight();

	const int cx = (nx-1)/2;
	const int cy = (ny-1)/2;

	const double R11 =  cos(-phi);
	const double R12 = -sin(-phi);
	const double R21 =  sin(-phi);
	const double R22 =  cos(-phi);

	float* hData =  new float[nx*ny];
	float* hDataIt = hData;

	double sum = 0;
	{
	  for (int y=0; y<ny; ++y) {
	  for (int x=0; x<nx; ++x) {
		  const double x_ = (x-cx)*R11-(y-cy)*R21;
		  const double y_ = (x-cx)*R12-(y-cy)*R22;
		  const float f = exp( -0.5*( pow(x_,2)*covar_xx + pow(y_,2)*covar_yy ) );
		  (*hDataIt) = f;
		  sum += f;
		  hDataIt++;
	  }
	  }
	}

	d.copyFromHostMemory(hData);
	delete[] hData;
	d.div(sum);
}
void fillWithGauss2RotatedShifted(GPUImage& d, double sigma_x, double sigma_y, double phi, double shift)
{
	const double covar_xx = 1.0/(sigma_x*sigma_x);
	const double covar_yy = 1.0/(sigma_y*sigma_y);

	const int nx = d.getWidth();
	const int ny = d.getHeight();

	const int cx = (nx-1)/2;
	const int cy = (ny-1)/2;

	const double R11 =  cos( phi);
	const double R12 =  sin( phi);
	const double R21 = -sin( phi);
	const double R22 =  cos( phi);

	float* hData =  new float[nx*ny];
	float* hDataIt = hData;

	double sum = 0;
	{
	  for (int y=0; y<ny; ++y) {
	  for (int x=0; x<nx; ++x) {
		  const double x_ = (x-cx)*R11 +(y-cy)*R12 - shift;
		  const double y_ = (x-cx)*R21 +(y-cy)*R22;
		  const float f = exp( -0.5*( pow(x_,2)*covar_xx + pow(y_,2)*covar_yy ) );
		  (*hDataIt) = f;
		  sum += f;
		  hDataIt++;
	  }
	  }
	}

	d.copyFromHostMemory(hData);
	delete[] hData;
	d.div(sum);

	//TODO: test
}

Kernel::kernel_pools_t Kernel::_kernel_pools;

Kernel::kernel_pool_t & Kernel::kernel_pool()
{
  Kernel::kernel_pool_t * instance = _kernel_pools.get();
  if ( instance == 0 )
	_kernel_pools.set( instance = new kernel_pool_t() );
  return *instance;
}

// BEGIN Kernel
Kernel::~Kernel() {}

// END Kernel
// BEGIN DiPoleKernel

void MultiKernel::set_modulator( const GPUData & mod )
{
  BOOST_FOREACH( KernelHandle & kernel, _kernels )
	kernel->set_modulator( mod );
}


void DiPoleKernel::init(const unsigned w, const unsigned h, const unsigned noOfOri, const bool force)
{
	if ( _filter.isEmpty()
		 || _filter.extent(0) != noOfOri
		 || force )
	  {
		unsigned k_w = int(2*(2.5 * _sigma_large + _r));
		if ( k_w % 2 == 0 ) k_w++;
		unsigned k_h = k_w;

		_filter = GPUData( noOfOri, k_h, k_w );
		_tex_filters.resize( noOfOri );

		GPUImage im1(k_w,k_h);
		GPUImage im2(k_w,k_h);
		for (int i=0; i<noOfOri; ++i)
		  {
			float phi = M_PI / 2 + ( M_PI * i ) / noOfOri;
			//float phi = ( M_PI * i ) / noOfOri;
			fillWithGauss2RotatedShifted( im1, _sigma_large, _sigma_small, phi, +_r);
			fillWithGauss2RotatedShifted( im2, _sigma_large, _sigma_small, phi, -_r);
			im1.add(im2);
			_filter[i].copyFrom(im1);
			_tex_filters[i] = GPUTextureData2D( im1 );
		  }
	  }
	_tmpimage1.resize(w,h);
	_tmpimage2.resize(w,h);
}

DiPoleKernel::DiPoleKernel(float sigma_large, float sigma_small, float r, unsigned n_ori_neigh )
  : _sigma_large(sigma_large)
  , _sigma_small(sigma_small)
  , _r(r)
  , _orientation_neighborhood(n_ori_neigh)
{
}

void DiPoleKernel::apply( GPUData& input, GPUData& output )
{
	if ( input.getDimensionsCount() > 5 )
		throw GPUDataException("error in DiPoleKernel::apply(..): input has to be 5-or-less-dimensional");

	const int h = input.extent(0);
	const int w = input.extent(1);
	const int C = input.extent(2);
	const int S = input.extent(3);
	const int O = input.extent(4);

	init(w,h,O);

	GPUImage ker;
	GPUImage & inref = _tmpimage1, & outref = _tmpimage2;
	for ( int c = 0; c < C; ++c )
	  {
		GPUData in_c (  input.slice( 2, c ) );
		GPUData out_c( output.slice( 2, c ) );
		for ( int s = 0; s < S; ++s )
		  {
			GPUData in_cs (  in_c.slice( 2, s ) );
			GPUData out_cs( out_c.slice( 2, s ) );
			for ( int o = 0; o < O; ++o )
			  {

				GPUData in_cso (  in_cs.slice( 2, o ) );
				GPUData out_cso( out_cs.slice( 2, o ) );

				if ( in_cso.isContiguous() && in_cso.isRowMajor() )
				  inref = in_cso;
				else
				  inref.copyFrom( in_cso );
				
				if ( out_cso.isContiguous() && out_cso.isRowMajor() )
				  outref = out_cso;
				
				_tex_tmpimage1.writeGPUData( inref );
				_tex_tmpimage1.conv2D( _tex_filters[o], outref );

				if ( ! ( out_cso.isContiguous() && out_cso.isColumnMajor() ) )
				  out_cso.copyFrom( outref );

			  }
			if ( _orientation_neighborhood > 0 )
			  {
				GPUData tmp( out_cs.rank(), out_cs.getDimensions() );
				tmp.fill( 0.0f );
				for ( int o = 0; o < O; ++o )
				  {
					GPUData out_cso( out_cs.slice( 2, o ) );
					for ( int d_o = -int(_orientation_neighborhood); d_o <= int(_orientation_neighborhood); ++d_o )
					  {
						if ( d_o == 0 ) continue;
						float d_rel = std::abs(float(d_o)/float(_orientation_neighborhood+1));
						int o2 = o + d_o;
						while ( o2 < 0 ) o2 += O;
						while ( o2 >= O ) o2 -= O;
						tmp.slice(2,o).add( out_cs.slice( 2, o2 ), 1.0f - d_rel );
					  }
				  }
				out_cs.add( tmp );
			  }
		  }
	  }
}
// END DiPoleKernel
// BEGIN Gauass1

GaussKernel1::GaussKernel1()
  : _sigma( 0 ), _dim( -1 )
{
	;
}

GaussKernel1::GaussKernel1( float sigma, int dim, int size, int type)
  : _sigma( sigma ), _dim( dim ), _type( type )

{
  if (size<1) 	size = std::max( int(1), int(5.0 * sigma) );
  if (size%2==0) 	size++;

	_kernel = GPUData( size );
	fillWithGauss( _kernel, _sigma );

	if (kernel_pool().count(_kernel)>0) {
		_kernel = *kernel_pool().find(_kernel);
	} else {
		_kernel.copyToConstMemory();
		kernel_pool().insert(_kernel);
	}
	_type = type;
}

void GaussKernel1::apply( GPUData & input, GPUData & output)
{
  input.convolute1D( _kernel, _dim, output, _type);
  //output.copyFrom( input );
}

GaussKernel1::~GaussKernel1()
{
	//TODO: should remove the kernel from pool if this makes sense
}

// END Gauss1
// BEGIN GauassLog1

GaussLogKernel1::GaussLogKernel1()
  : _sigma( 0 ), _dim( -1 )
{
	;
}

GaussLogKernel1::GaussLogKernel1( float sigma, int dim, int type)
  : _sigma( sigma ), _dim( dim ), _type( type )
{
	if (sigma<=0.0f) return;
#ifndef _WIN32
	const int x_max_at_five_percen = int(exp(sqrt(2)*sqrt(log(20))*sigma));
	const float x_center = exp(sigma*sigma); // where integral up to this value equals half of total integral
#else
	const int x_max_at_five_percen = int(exp(sqrt(2.0f)*sqrt(log(20.0f))*sigma));
	const float x_center = exp(sigma*sigma); // where integral up to this value equals half of total integral
#endif
//	printf("5 percent border = %d\nx_center = %f\n",x_max_at_five_percen,x_center);
	int size   = 2*x_max_at_five_percen+1;
	_kernel = GPUData( size );
	_kernel.fill(0.0f);
	for (float i=0.0f; i<size; i+=0.1f) {
		const float x = i-(x_max_at_five_percen-x_center);
		if (x<=0)
			_kernel[int(floor(i))].add(0.0f);
		else {
			_kernel[int(floor(i))].add(exp(-0.5f*(log(x)/sigma)*(log(x)/sigma)));
		}
	}
	_kernel.div(_kernel.sum());

	if (kernel_pool().count(_kernel)>0) {
		_kernel = *kernel_pool().find(_kernel);
	} else {
		_kernel.copyToConstMemory();
		kernel_pool().insert(_kernel);
	}
}

void GaussLogKernel1::apply( GPUData & input, GPUData & output)
{
	if (_sigma<=0.0f) return;
    input.convolute1D( _kernel, _dim, output, _type);
}

GaussLogKernel1::~GaussLogKernel1()
{
	//TODO: should remove the kernel from pool if this makes sense
}

// END GaussLog1
// BEGIN Gauass2

GaussKernel2::GaussKernel2()
{
	_sigma = 0;
}

GaussKernel2::GaussKernel2( float sigma, int dim1, int dim2, int size)
{
  if (sigma>0.0f) add_kernel( new GaussKernel1( sigma, dim1, size ) );
  if (sigma>0.0f) add_kernel( new GaussKernel1( sigma, dim2, size ) );
  _sigma = sigma;
}

GaussKernel2::~GaussKernel2()
{
	//TODO: should remove the kernel from pool if this makes sense
}

// END Gauss2
// BEGIN Binom2

Binom2Kernel1::Binom2Kernel1( int dim )
  : _dim( dim )
{
	if (_kernel.size()!=3) {
		_kernel = GPUData( 3 );
		_kernel[0] = 1;
		_kernel[1] = 2;
		_kernel[2] = 1;
	}
}

void Binom2Kernel1::apply( GPUData & input, GPUData & output)
{
  input.convolute1D( _kernel, _dim, output );
}

Binom2Kernel1::~Binom2Kernel1()
{
	;
}

// END Binom2
// BEGIN DerivativeSimple

DerivativeSimpleKernel1::DerivativeSimpleKernel1( int dim )
  : _dim( dim )
{
	if (_kernel.size()!=3) {
		_kernel = GPUData( 3 );
		_kernel[0] =  1;
		_kernel[1] =  0;
		_kernel[2] = -1;
	}
}

void DerivativeSimpleKernel1::apply( GPUData & input, GPUData & output)
{
  input.convolute1D( _kernel, _dim, output );
}

DerivativeSimpleKernel1::~DerivativeSimpleKernel1()
{
	;
}

// END DerivativeSimple

// BEGIN DerivativeSimple

SobelOperator::SobelOperator( int diff_dim, int blur_dim )
  : _diff_kernel( diff_dim )
  , _binom2_kernel( blur_dim )
{
	;
}

void SobelOperator::apply( GPUData & input, GPUData & output)
{
	GPUData temp(input.getDimensionsCount(), input.getDimensions());
	_binom2_kernel.apply(input,temp);
	_diff_kernel.apply(temp,output);
}

SobelOperator::~SobelOperator()
{
	;
}

// END DerivativeSimple

// BEGIN DerivativeSimple Sobel scaled

SobelOperatorScaled::SobelOperatorScaled( int diff_dim, int blur_dim, int size )
  : _diff_kernel( diff_dim )
  , _binom2_kernel( blur_dim )
{

	;
}

void SobelOperatorScaled::apply( GPUData & input, GPUData & output)
{
	GPUData temp(input.getDimensionsCount(), input.getDimensions());
	_binom2_kernel.apply(input,temp);
	_diff_kernel.apply(temp,output);
}

SobelOperatorScaled::~SobelOperatorScaled()
{
	;
}

// END DerivativeSimple


// BEGIN sin gauss

OddGaborKernel1::OddGaborKernel1()
	: _sigma( 0 ), _omega( 0 ), _dim( -1 )
{
	;
}

OddGaborKernel1::OddGaborKernel1(  float lambda, float sigma, int dim, int size )
  : _sigma( sigma ), _omega( 2*M_PI/lambda ), _dim( dim )
{
	if (size<1) 	size = int(5.0 * sigma);
	if (size%2==0) 	size++;

	{ /* kernel creation */
		GPUData gauss_L( size );
		fillWithGauss( gauss_L, _sigma );

		_kernel = GPUData( size );
		fillWithSin( _kernel, _omega );
		_kernel.mul(gauss_L);
	}

	if (kernel_pool().count(_kernel)>0) {
		_kernel = *kernel_pool().find(_kernel);
	} else {
		_kernel.copyToConstMemory();
		kernel_pool().insert(_kernel);
	}
}

OddGaborKernel1::OddGaborKernel1( bool dummy, float omega, float sigma, int dim, int size )
  : _sigma( sigma ), _omega( omega ), _dim( dim )
{
	if (size<1) 	size = int(5.0 * sigma);
	if (size%2==0) 	size++;

	{ /* kernel creation */
		GPUData gauss_L( size );
		fillWithGauss( gauss_L, _sigma );

		_kernel = GPUData( size );
		fillWithSin( _kernel, _omega );
		_kernel.mul(gauss_L);
	}

	if (kernel_pool().count(_kernel)>0) {
		_kernel = *kernel_pool().find(_kernel);
	} else {
		_kernel.copyToConstMemory();
		kernel_pool().insert(_kernel);
	}
}

void OddGaborKernel1::apply( GPUData & input, GPUData & output)
{
  input.convolute1D( _kernel, _dim, output );
}

OddGaborKernel1::~OddGaborKernel1()
{
	//TODO: should remove the kernel from pool if this makes sense
}
// END sin gauss
// BEGIN cos gauss

EvenGaborKernel1::EvenGaborKernel1()
  : _sigma( 0 ), _omega( 0 ), _dim( -1 )
{
	;
}

EvenGaborKernel1::EvenGaborKernel1( float lambda, float sigma, int dim, int size )
  : _sigma( sigma ), _omega( 2*M_PI/lambda ), _dim( dim )
{
	if (size<1) 	size = int(5.0 * sigma);
	if (size%2==0) 	size++;

	{
		GPUData gauss_L( size );
		fillWithGauss( gauss_L, _sigma );

		_kernel = GPUData( size );
		fillWithCos( _kernel, _omega );
		_kernel.mul(gauss_L);
	}

	if (kernel_pool().count(_kernel)>0) {
		_kernel = *kernel_pool().find(_kernel);
	} else {
		_kernel.copyToConstMemory();
		kernel_pool().insert(_kernel);
	}
}

EvenGaborKernel1::EvenGaborKernel1( bool dummy, float omega, float sigma, int dim, int size )
  : _sigma( sigma ), _omega( omega ), _dim( dim )
{
	if (size<1) 	size = int(5.0 * sigma);
	if (size%2==0) 	size++;

	{
		GPUData gauss_L( size );
		fillWithGauss( gauss_L, _sigma );

		_kernel = GPUData( size );
		fillWithCos( _kernel, _omega );
		_kernel.mul(gauss_L);
	}

	if (kernel_pool().count(_kernel)>0) {
		_kernel = *kernel_pool().find(_kernel);
	} else {
		_kernel.copyToConstMemory();
		kernel_pool().insert(_kernel);
	}
}

void EvenGaborKernel1::apply( GPUData & input, GPUData & output)
{
  input.convolute1D( _kernel, _dim, output );
}

EvenGaborKernel1::~EvenGaborKernel1()
{
	//TODO: should remove the kernel from pool if this makes sense
}
// END cos gauss

void ModulatedDiffusionKernel::set_modulator( const GPUData & mod )
{
  _modulator.reference(mod);
}

extern "C" void do_modulated_diffusion( float * d_in, int n1, int n2, int s_in1, int s_in2 ,
										const float * d_mod         , int s_mod1, int s_mod2,
										float * d_out               , int s_out1, int s_out2,
										float gmod, float g0, float sigma         );


void ModulatedDiffusionKernel::apply( GPUData & input, GPUData & output )
{
  int U = input.extent(2);
  int V = input.extent(3);
  int N1 = input.extent(0), N2 = input.extent(1);
  if ( input.rank() > 4 ) throw std::runtime_error(str(boost::format("mod diff only up to rank 4 but rank is %i")
																	 % input.rank() ) );
  output.resize( input.rank(), input.getDimensions() );
  if ( _modulator.empty() ) output.copyFrom(input);
  else if ( _modulator.extent(0) != N1 || _modulator.extent(1) != N2 )
	throw std::runtime_error(str(boost::format("modulated diffusion: mismatch between"
											   " modulator and data, data: %ix%i modulator: %ix%i")
								 % input.extent(0) % input.extent(1) % _modulator.extent(0) % _modulator.extent(1) ) );
  else if ( _modulator.rank() > 2 )
	throw std::runtime_error(str(boost::format("modulated diffusion only allows for a 2d modulator, but modulator is %id")
								 % _modulator.rank() ) );
  else
	{
	  for ( size_t u = 0; u < U; ++u )
		for ( size_t v = 0; v < V; ++v )
		  {
			GPUData in  =  input.slice( 3, v ).slice( 2, u );
			GPUData out = output.slice( 3, v ).slice( 2, u );
			assert(in.extent(0) == N1);
			assert(in.extent(1) == N2);
			assert(out.extent(0) == N1);
			assert(out.extent(1) == N2);
			do_modulated_diffusion( in.data(), N1, N2 , in.stride(0),         in.stride(1),
									_modulator.data(),  _modulator.stride(0), _modulator.stride(1),
									out.data(),         out.stride(0),        out.stride(1),
									_gmod,                   _g0,               _sigma );
			GPUDataException::check_error( str(boost::format("modulated diffusion u=%i v=%i")%u%v));
		  }
	}
}

void ModulatedDiffusionKernel::apply( GPUData & input, GPUData & tmp, GPUData & output )
{
  apply( input, output );
}


void MultiKernel::apply( GPUData & input, GPUData & tmp, GPUData & output )
{
	if ( _kernels.empty() )
	  {
		output.copyFrom( input );
	  }
	else
	  {
		_kernels.front()->apply( input, output );
		for ( size_t i = 1; i < _kernels.size(); ++i )
		  {
			output.swap( tmp );
			_kernels[i]->apply( tmp, output );
		  }
	  }
}

void MultiKernel::apply( GPUData & input, GPUData & output )
{
	if ( _kernels.empty() )
	  {
		output.swap( input );
	  }
	else
	  {
		_kernels.front()->apply( input, output );
		//output.assign( input );
		for ( size_t i = 1; i < _kernels.size(); ++i )
		  {
			input.swap( output );
			_kernels[i]->apply( input, output );
		  }
	  }
}

void MultiKernel::add_kernel( Kernel * kernel )
{
  _kernels.push_back( KernelHandle( kernel ) );
}

void MultiKernel::add_kernel( KernelHandle kernel )
{
  if ( kernel )
	_kernels.push_back( kernel );
}


LaplaceOperator::LaplaceOperator(float sigma_) {
	_sigma = sigma_;
	int n = int(6*sigma_);
	if (n%2==0) n++;
	int c = n/2;

#ifndef _WIN32
	float host_data[n];
#else
	float * const host_data = (float*) _alloca(n * sizeof(float));
#endif
	for (int i=-c; i<=c; ++i) {
		host_data[i+c] = 1.0/(pow(sigma_,3)*sqrt(2*M_PI))*(i*i/pow(sigma_,2)-1)*exp(-i*i/(2*pow(sigma_,2)));
		//printf("%.03f", host_data[i+c]);
	}
	_kernel = GPUData(n);
	_kernel.copyFromHostMemory(host_data);
	_kernel.sub(_kernel.sum()/n);
	_kernel.div(_kernel.absSum());
}


LaplaceOperator::~LaplaceOperator() {
	;
}

void LaplaceOperator::apply(GPUData& input, GPUData& output) {
	if (input.getNoOfDimensions()!=2) throw GPUDataNYIException("LaplaceOperator::apply");

	GaussKernel1 gauss1(_sigma, 1);//New
	GaussKernel1 gauss0(_sigma, 0);//New
	GPUData tmp(output.extent(0), output.extent(1));//New
	GPUData tmp2(output.extent(0), output.extent(1));//New
	tmp2.copyFrom(input);//save input

	gauss1.apply(input,tmp);//New
	input.swap(tmp);//New
	input.convolute1D(_kernel,0,output);//New

	gauss0.apply(tmp2,tmp);//New
	tmp.swap(tmp2);//New
	tmp.convolute1D(_kernel,1,tmp2);//New
	output.add(tmp2);

	//Original code:
	//input.convolute1D(_kernel,0,tmp);
	//input.swap(output);
	//input.convolute1D(_kernel,1,output);


}

LaplaceKernel2::LaplaceKernel2( float sigma,  int dim1, int dim2, int size)
{
	if(size<0) size = 6*sigma;
	if(size%2 == 0) size++;
	GPUData kernels(size, size);
	float r, sigmaSqr = sigma*sigma;

	float* hData =  new float[size*size];
	float* hDataIt = hData;

	double sum = 0;
	{
	  for (int y=0; y<size; ++y) {
	  for (int x=0; x<size; ++x) {
#ifndef _WIN32
		  r = sqrt(x*x + y*y);
#else
		  r = sqrt((float)(x*x + y*y));
#endif
		  const float f = -(1/( M_PI*sigmaSqr )) * ( 1- (r*r/ (2*sigmaSqr )) ) * exp(-r*r/ (2*sigmaSqr));
		  (*hDataIt) = f;
		  sum += f;
		  hDataIt++;
	  }
	  }
	}

	kernels.copyFromHostMemory(hData);
	delete[] hData;
	//kernels.sub(sum/(size*size));
	//kernels.div(sum);


	_dim1 = dim1;
	_dim2 = dim2;
	_kernels = kernels;

}

void LaplaceKernel2::apply( GPUData & input, GPUData & output)
{
	if(input.getNoOfDimensions() < 2)
	{throw GPUDataNYIException("Not yet implemented exception");}

	tmp1.resize(input.extent(0), input.extent(1));
	tmp2.resize(input.extent(0), input.extent(1));

	tmp1.copyFrom(input);
	GPUImage im(_kernels);
	tmp1.convolute2D(im, tmp2);
	output.copyFrom(tmp2);

}

LaplaceKernel2::~LaplaceKernel2() {
	;
}

GaussKernel2Aniso::GaussKernel2Aniso( float sigma1, float sigma2, int dim1, int dim2, int dim3, float num, int size)
{
	if(size<0) size = 5*fmax(sigma1, sigma2);
	if(size%2 == 0) size++;
	GPUData kernels(num, size, size);
	int orderFirstHorThenVer[4], pos;
	orderFirstHorThenVer[0] = 2; orderFirstHorThenVer[1] = 1; orderFirstHorThenVer[2] = 0; orderFirstHorThenVer[3] = 3;

	for(int i=0; i<num; ++i){
		pos = orderFirstHorThenVer[i];
		GPUImage im(kernels[pos]);
		fillWithGauss2Rotated(im, sigma1,sigma2, M_PI/num*i);
	}
	_dim1 = dim1;
	_dim2 = dim2;
	_dim3 = dim3;
	_kernels = kernels;

}

void GaussKernel2Aniso::apply( GPUData & input, GPUData & output)
{
	if(input.getNoOfDimensions() < 3)
	{throw GPUDataNYIException("Not yet implemented exception");}

	tmp1.resize(input.extent(_dim2), input.extent(_dim1));
	tmp2.resize(input.extent(_dim2), input.extent(_dim1));

	for(int i=0; i<input.extent(_dim3); ++i )
	{
		tmp1.copyFrom(input.slice(_dim3, i).removeSingletonDimensions());
		GPUImage im(_kernels[i]);
		tmp1.convolute2D(im, tmp2);
		output.slice(_dim3, i).removeSingletonDimensions().copyFrom(tmp2);

	}
}

GaussKernel2Aniso::~GaussKernel2Aniso() {
	;
}
