/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "GPUImage.h"

#ifndef __kernels_hpp_
#define __kernels_hpp_

#include <boost/shared_ptr.hpp>
#include <boost/unordered_set.hpp>
#include <boost/ptr_container/ptr_map.hpp>
#include "cuda_context_handler.hpp"
#include <vector>

#include "GPUTextureData2D.hpp"

/**
 * Creates a linear index from start_A (including) to end_A (including)
 */
void fillLinSpace(GPUData& d, double start_A, double end_A);

/**
 * Creates a 1-d gaussian with the sum equals to 1.
 * The center is (size-1)/2 divided down.
 */
void fillWithGauss(GPUData& d, double sigma);

/**
 * Creates a 2-d gaussian with given covariances (covar_xx=1/(simga_x^2)).
 * Ths sum is equal to 1.
 * The center is (size-1)/2 divided down.
 */
void fillWithGauss2(GPUImage& d, double covar_xx, double covar_xy, double covar_yy);

/**
 * creates a 2-d gaussian with given sigma for the main axis and rotates it.
 * The rotation is done in clockwise within a normal coordinate system and, thus,
 * appears to bea counter clockwise within an image.
 * The original center is (size-1)/2 divided down.
 * Ths sum is equal to 1.
 */
void fillWithGauss2Rotated(GPUImage& d, double sigma_x, double sigma_y, double phi);

/**
 * creates a 2-d gaussian with given sigma for the main axis, rotates it and shift
 * it along the given direction (orated x-axis).
 * The rotation is done in mathematically positive direction within a normal
 * coordinate system and, thus, appears to be clockwise within an image.
 * Ths sum is equal to 1.
 * The original center is (size-1)/2 divided down.
 */
void fillWithGauss2RotatedShifted(GPUImage& d, double sigma_x, double sigma_y, double phi, double shift);

/**
 * fills with a cosine function.
 * The center is (size-1)/2 divided down.
 */
void fillWithCos(GPUData& d, double omega);

/**
 * fills with a sine function.
 * The center is (size-1)/2 divided down.
 */
void fillWithSin(GPUData& d, double omega);

/**
 * abstract interface to a convolution kernel
 *
 */
class Kernel
{
public:
  typedef boost::unordered_set<GPUData,boost::hash<GPUData>,GPUDataIsEqualTo> kernel_pool_t;
  typedef cuda_context_specific_ptr<kernel_pool_t> kernel_pools_t;
  virtual void apply( GPUData & input, GPUData & output ) = 0;
  virtual void apply( GPUData & input, GPUData & tmp, GPUData & output ) {
	assert(false);input.rank();tmp.rank();output.rank(); // to void unused warnings
  }
  virtual void set_modulator( const GPUData & mod ) { mod.rank();}
  virtual ~Kernel();
protected:
  kernel_pool_t & kernel_pool();
private:
  static kernel_pools_t _kernel_pools;
};

class ModulatedDiffusionKernel : public Kernel
{
public:
  ModulatedDiffusionKernel( float sigma, float g0, float gmod ) 
	: _sigma(sigma), _g0(g0), _gmod(gmod) {}
  virtual void apply( GPUData & input, GPUData & output );
  virtual void apply( GPUData & input, GPUData & tmp, GPUData & output );  
  virtual void set_modulator( const GPUData & mod );
private:
  GPUData _modulator;
  float _sigma, _g0, _gmod;
};

class MultiKernel : public Kernel
{
public:
  typedef boost::shared_ptr<Kernel> KernelHandle;
  virtual void apply( GPUData & input, GPUData & output );
  virtual void apply( GPUData & input, GPUData & tmp, GPUData & output );
  void add_kernel( KernelHandle kernel );
  void add_kernel( Kernel * kernel );
  virtual void set_modulator( const GPUData & mod );
  void clear() { _kernels.clear(); }
private:
  std::vector< KernelHandle > _kernels;
};

class DiPoleKernel : public Kernel
{
private:
  float _sigma_large,_sigma_small,_r;
  GPUData  _filter;
  std::vector<GPUTextureData2D> _tex_filters;
  GPUTextureData2D _tex_tmpimage1;
  GPUImage _tmpimage1, _tmpimage2;
  unsigned _orientation_neighborhood;
  void init(const unsigned w, const unsigned h, const unsigned noOfOri, const bool force=false);
public:
  DiPoleKernel(float sigma_large, float sigma_small, float r, unsigned n_ori_neigh = 0);
  virtual void apply( GPUData& input, GPUData& output );
};

/**
 * A one-dimensional gaussian.
 *
 * The sum is guaranteed to be 1.
 *
 * \f[ \frac{1}{\sqrt[2]{2\pi}\,\sigma} e^{-\frac{x^2}{2\,\sigma^2}} \f]
 */
class GaussKernel1 : public Kernel
{
public:
	GaussKernel1();
	/**
	 * Creates a one dimensional gaussion function with a
	 * sum of 1.
	 *
	 * \param simga the standard deviation of the gaussian
	 * 				distribution
	 *
	 * \param dim	the dimension, where the kernel has to be
	 * 				applied
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of
	 * 				5*sigma is chosen. The function increases
	 * 				the size by one if it is set to a even value.
	 * \param type  the type of convolution (see GPUData::convolute1d())
	 */

	GaussKernel1( float sigma, int dim = 0, int size = -1, int type=2);

	virtual void apply( GPUData & input, GPUData & output);
	virtual ~GaussKernel1();
	
	float getSigma() { return _sigma; }

protected:
	GPUData _kernel;
	float 	_sigma;
	int   	_dim;
	int 	_type;
};

/**
 * A one-dimensional gaussian in log domain.
 *
 * The sum is guaranteed to be 1.
 *
 * \f[ \frac{1}{\sqrt[2]{2\pi}\,\sigma} e^{-\frac{log(x)^2}{2\,\sigma^2}} \f]
 */
class GaussLogKernel1 : public Kernel
{
public:
	GaussLogKernel1();
	/**
	 * Creates a one dimensional gaussion function with a
	 * sum of 1.
	 *
	 * \param simga the standard deviation of the gaussian
	 * 				distribution
	 *
	 * \param dim	the dimension, where the kernel has to be
	 * 				applied
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of
	 * 				5*sigma is chosen. The function increases
	 * 				the size by one if it is set to a even value.
	 * \param type  the type of convolution (see GPUData::convolute1d())
	 */
	GaussLogKernel1( float sigma, int dim = 0, int type=2);
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~GaussLogKernel1();

protected:
	GPUData _kernel;
	float 	_sigma;
	int   	_dim;
	int 	_type;
};


/**
 * A two-dimensional gaussian.
 *
 * The sum is guaranteed to be 1.
 */
class GaussKernel2 : public MultiKernel
{
public:
	GaussKernel2();
	/**
	 * Creates a one dimensional gaussion function with a
	 * sum of 1.
	 *
	 * \param simga the standard deviation of the gaussian
	 * 				distribution
	 *
	 * \param dim1	the first dimension, where the kernel has to be
	 * 				applied
	 *
	 * \param dim2	the second dimension, where the kernel has to be
	 * 				applied
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of
	 * 				5*sigma is chosen. The function increases
	 * 				the size by one if it is set to a even value.
	 */
	GaussKernel2( float sigma, int dim1 = 0, int dim2 = 1, int size = -1);
	virtual ~GaussKernel2();

	float sigma() { return _sigma; }
private:
	float _sigma;
};

/**
 * A two-dimensional gaussian.
 *
 * The sum is guaranteed to be 1.
 */
class GaussKernel2Aniso
{
public:
	/**
	 * Creates a two dimensional anisotropic gaussion function with a
	 * sum of 1.
	 *
	 * \param simga1 the first standard deviation of the gaussian
	 * 				distribution
	 * \param simga2 the second standard deviation of the gaussian
	 * 				distribution
	 *
	 * \param dim1	the first dimension, where the kernel has to be
	 * 				applied
	 *
	 * \param dim2	the second dimension, where the kernel has to be
	 * 				applied
	 *
	 * \param dim3	the third dimension, where the different orientations are stored
	 *
	 * \param num	num of filters that have to be generated, the angle between the filters is computed via 180/num
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of
	 * 				5*sigma is chosen. The function increases
	 * 				the size by one if it is set to a even value.
	 */
	GaussKernel2Aniso( float sigma1, float sigma2, int dim1 = 0, int dim2 = 1, int dim3 = 2, float num = 4, int size = -1);
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~GaussKernel2Aniso();

private:
	int _dim1, _dim2, _dim3;
	GPUImage tmp1, tmp2;
public:	GPUData _kernels;


};

/**
 * A two-dimensional laplacian.
 *
 * The sum is guaranteed to be 1.
 */
class LaplaceKernel2
{
public:
	/**
	 * Creates a two dimensional laplacian function with a
	 * sum of 1.
	 *
	 * \param simga the first standard deviation of the gaussian
	 * 				distribution
	 *
	 * \param dim1	the first dimension, where the kernel has to be
	 * 				applied
	 *
	 * \param dim2	the second dimension, where the kernel has to be
	 * 				applied
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of
	 * 				6*sigma is chosen. The function increases
	 * 				the size by one if it is set to a even value.
	 */
	LaplaceKernel2( float sigma, int dim1 = 0, int dim2 = 1, int size = -1);
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~LaplaceKernel2();

private:
	int _dim1, _dim2;
	GPUImage tmp1, tmp2;
public:	GPUData _kernels;


};


/**
 * represents an odd 1-d gabor function
 *
 * \f[ \sin\left(\frac{2\pi}{\lambda}\,x\right) \cdot \frac{1}{\sqrt[2]{2\pi}\,\sigma}
 * e^{-\frac{x^2}{2\,\sigma^2}} \f]
 *
 */
class OddGaborKernel1 : public Kernel
{
public:
	OddGaborKernel1();
	/**
	 * Creates a one dimensional sin gabor kernel (see class description
	 * for details)
	 *
	 * \param lambda the wavelength \f$\lambda\f$
	 *
	 * \param sigma the standard deviation \f$\sigma\f$ of the gaussian
	 * 				envelope
	 *
	 * \param dim	the dimension, where the kernel has to be applied
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of \f$5\sigma\f$ is
	 * 				chosen. The function increasesthe size by 1 if it
	 * 				is set to a even value.
	 */
	OddGaborKernel1( float lambda, float sigma, int dim, int size = -1 );
	/**
	 * Creates a one dimensional sin gabor kernel (see class description
	 * for details)
	 *
	 * \param omega the angular frequency \f$\frac{2\pi}{\lambda}\f$
	 *
	 * \param sigma the standard deviation \f$\sigma\f$ of the gaussian
	 * 				envelope
	 *
	 * \param dim	the dimension, where the kernel has to be applied
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of \f$5\sigma\f$ is
	 * 				chosen. The function increasesthe size by 1 if it
	 * 				is set to a even value.
	 */
	OddGaborKernel1( bool dummy, float omega, float sigma, int dim, int size = -1 );
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~OddGaborKernel1();
protected:
	GPUData _kernel;
	float 	_sigma;
	float	_omega;
	int   	_dim;
};

/**
 * represents an even 1-d gabor function
 *
 * \f[ \cos\left(\frac{2\pi}{\lambda}\,x\right) \cdot \frac{1}{\sqrt[2]{2\pi}\,\sigma}
 * e^{-\frac{x^2}{2\,\sigma^2}} \f]
 *
 */
class EvenGaborKernel1 : public Kernel
{
public:
	EvenGaborKernel1( );
	/**
	 * Creates a one dimensional cos gabor kernel (see class description
	 * for details)
	 *
	 * \param lambda the wavelength \f$\lambda\f$
	 *
	 * \param sigma the standard deviation \f$\sigma\f$ of the gaussian
	 * 				envelope
	 *
	 * \param dim	the dimension, where the kernel has to be applied
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of \f$5\sigma\f$ is
	 * 				chosen. The function increasesthe size by 1 if it
	 * 				is set to a even value.
	 */
	EvenGaborKernel1( float lambda, float sigma, int dim, int size = -1 );

	/**
	 * Creates a one dimensional cos gabor kernel (see class description
	 * for details)
	 *
	 * \param omega the angular frequency \f$\frac{2\pi}{\lambda}\f$
	 *
	 * \param sigma the standard deviation \f$\sigma\f$ of the gaussian
	 * 				envelope
	 *
	 * \param dim	the dimension, where the kernel has to be applied
	 *
	 * \param size	the size of the kernel. If its value is
	 * 				smaller than 1, a standard size of \f$5\sigma\f$ is
	 * 				chosen. The function increasesthe size by 1 if it
	 * 				is set to a even value.
	 */
	EvenGaborKernel1( bool dummy, float omega, float sigma, int dim, int size = -1 );
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~EvenGaborKernel1();
protected:
	GPUData _kernel;
	float 	_sigma;
	float	_omega;
	int   	_dim;
};

/**
 * second binomial (1,2,1)
 */
class Binom2Kernel1 : public Kernel
{
public:
	Binom2Kernel1( int dim );
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~Binom2Kernel1();
private:
  GPUData _kernel;
	int _dim;
};

/**
 * central difference (-1,0,1)
 */
class DerivativeSimpleKernel1 : public Kernel
{
public:
	DerivativeSimpleKernel1( int dim );
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~DerivativeSimpleKernel1();
private:
  GPUData _kernel;
  int _dim;
};

/**
 * sobel operator
 */
class SobelOperator : public Kernel
{
public:
	SobelOperator( int diff_dim, int blur_dim );
	/**
	 * Take care! This function allocates and frees one temporary data
	 * object of input size each time it is called.
	 */
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~SobelOperator();
private:
	DerivativeSimpleKernel1 _diff_kernel;
	Binom2Kernel1 _binom2_kernel;
};

/**
 * sobel operator with arbitrary size
 */
class SobelOperatorScaled : public Kernel
{
public:
	SobelOperatorScaled( int diff_dim, int blur_dim, int size );
	/**
	 * Take care! This function allocates and frees one temporary data
	 * object of input size each time it is called.
	 */
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~SobelOperatorScaled();
private:
	DerivativeSimpleKernel1 _diff_kernel;
	Binom2Kernel1 _binom2_kernel;
};


/**
 * laplace operator
 */
class LaplaceOperator : public Kernel
{
public:
	LaplaceOperator( float sigma );
	/**
	 * Take care! This function allocates and frees one temporary data
	 * object of input size each time it is called.
	 */
	virtual void apply( GPUData & input, GPUData & output);
	virtual ~LaplaceOperator();
private:
	float _sigma;
	GPUData _kernel;
};

#endif
