/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <boost/program_options.hpp>
#include <boost/format.hpp>
#include <iostream>
#include <iterator>
#include <vector>
#include <cstdlib>

#include <SeariseLogger.hpp>
#include <kernels.hpp>
#include <GPUImage.h>
#include <HostImage.h>
#include <stdio.h>

#include "ClassicUUlmNeuralModel.h"
#include "GaborFeatureExtractor.h"
#include "CartesianCorrelator.h"
#include "GPUVelField.h"
#include "HDF5FileAdapter.h"
#include "HDF5MotionFile.h"
#include "SimpleHostImageSequence.h"

#include "ModelArea.hpp"
#include "downsample.hpp"

#include <CensusClass.h>
#include <HypoImage.h>

#include <functional>


#define START_WATCH(name) clock_t stop_watch_##name##_start = clock();

#define STOP_WATCH(name) \
  clock_t stop_watch_##name##_stop = clock(); \
  std::cout << "Elapsed time for " #name ": " \
            << (((double)1000.0*(stop_watch_##name##_stop - stop_watch_##name##_start)) / CLOCKS_PER_SEC) \
            << " ms" << std::endl;


#include "util.hpp"

int main(int argc, char* argv[]) 
{

	/* --- input images ---*/
	std::string inputFilePatter;
	int   		start_idx = 0, iteration = 0;
	int   		stop_idx = 10;
	HostImageSequence* imageSequence;

	inputFilePatter = argv[1];
	start_idx = atoi( argv[2] );
	stop_idx  = atoi( argv[3] );

	//	inputFilePatter = "../Shell/images/moving_box_large/moving_box_large_%02d.png";

	imageSequence = new SimpleHostImageSequence(inputFilePatter.c_str(),start_idx,stop_idx);
	const int w = imageSequence->getWidth();//get image size
	const int h = imageSequence->getHeight();

	printf("iterate from:\n  ");
	printf(inputFilePatter.c_str(),start_idx);
	printf(" to\n  ");
	printf(inputFilePatter.c_str(),stop_idx);
	printf("\n");

	float fbf = 100, sigma_MT = 5.0f;
	//v1_window_width and VEL_MAX define size of GPU cache used - do not increase these numbers
	int downsample_faktor_mt=5, H_MAX_1 = 4, H_MAX_2 = 1000, VEL_MAX = 30, v1_window_width = 25;

	/* two images and their features (t and t-1) */
	GPUImage  currentImage(w,h);
	GPUImage  pastImage(w,h); pastImage.fill(0.0f);
	imageSequence->copyNextImage(currentImage);
	
	GPUImage  CensusGPU(w,h); CensusGPU.fill(0.0f);//"else"-value
	GPUImage  CensusGPUPast(w,h); CensusGPUPast.fill(0.0f);//"else"-value
	unsigned int currCensusList[w*h];
	
	//list of Census values
	CensusClass* CensusArray1 =  new CensusClass[w*h];
	CensusClass* CensusArray2 =  new CensusClass[w*h];

	//pointer for memory just allocated, used to switch curr and prev
	CensusClass *pCensusArrayCurr = CensusArray1;
	CensusClass *pCensusArrayPrev = CensusArray2;
	
	HypoImage HypoArrayV1(    w, h, H_MAX_1 );
	HypoImage HypoArrayMT( w / downsample_faktor_mt, h / downsample_faktor_mt, H_MAX_1 );

	HypoArrayV1.getData().fill(0.0f);
	HypoArrayMT.getData().fill(0.0f);
	
	//JUST FOR FIRST IMAGE
	//GaussKernel2 blurInput(1,0,1,3);
	//blurInput.apply(currentImage, currentImage);
	CensusGPU.fill(0.0f);//necessary?
	currentImage.getZensusTransformedImage(CensusGPU,currentImage.max()*2/256.0f);
	//copy to host memory
	CensusGPU.copyToHostMemory((float*)&currCensusList);

	//compute sorted Census list
	pCensusArrayCurr->createSortedCensusList(currCensusList, w, h);

	//JUST FOR FIRST IMAGE END*******************************
	HDF5MotionFile MTfile( argv[3], true );
	
	float censusOcc[w*h];
	GPUImage censusOccDev(w,h);
	CPUData<float> hostHypoImage( h, w, 3 * H_MAX_1 );
	
	START_WATCH(calculation);// time measurement

	/* begin of iteration loop */
	while (imageSequence->hasNextImage())
	  {
		HypoArrayV1.getData().fill(0.0f);
		std::cout << "#" << iteration << " "<< std::flush;
		
		imageSequence->copyNextImage(currentImage);/* get the new current image */
		//blurInput.apply(currentImage, currentImage);
		
		//compute Census for current image
		CensusGPU.swap(CensusGPUPast);
		currentImage.getZensusTransformedImage( CensusGPU, currentImage.max() * 2 / 256.0f );
		
		//copy to host memory
		CensusGPU.copyToHostMemory((float*)&currCensusList);
		std::swap( pCensusArrayCurr, pCensusArrayPrev );
		
		//sort CensusClass Array on CPU and generate hypos for count<HMAX1
		pCensusArrayCurr->computeHyposUnambiguous(pCensusArrayPrev, currCensusList, censusOcc, 
												  hostHypoImage, w, h, H_MAX_1, VEL_MAX);

		HypoArrayV1.getData().copyFromHostMemory( hostHypoImage.data() );

		std::cout << " unamibiguous hypos [ " << hostHypoImage.min() 
				  << "," << hostHypoImage.mean() << "," 
				  << hostHypoImage.max() << "]" << std::endl;
		std::cout << " unamibiguous hypos(GPU) [ " << HypoArrayV1.getData().min() 
				  << "," << HypoArrayV1.getData().mean() 
				  << "," << HypoArrayV1.getData().max() << "]" << std::endl;
		
		censusOccDev.fill(1.0f);

		// this is effectively a guided continuation of the correlation
		pCensusArrayCurr->computeHyposAmbiguous(censusOcc,
												HypoArrayV1, 
												HypoArrayMT , 
												censusOccDev, 
												CensusGPU, 
												CensusGPUPast, 
												fbf, H_MAX_2, VEL_MAX);
		
		// this is a model area
		HypoArrayV1.sqrWeights();
		
		HypoArrayV1.velocityBlur( VEL_MAX );
		
		HypoArrayV1.normVelSpace();
		
		char Name1[1024];
		sprintf(Name1, "HypoV1%03d.txt", iteration );
		HypoArrayV1.printHyposInFile(H_MAX_1, Name1);
		
		//MT
		HypoArrayV1.sqrWeights();
		
		HypoArrayV1.subsample_hypo_image(VEL_MAX, sigma_MT, v1_window_width, HypoArrayMT);
		
		HypoArrayMT.velocityBlur( VEL_MAX );

		HypoArrayMT.normVelSpace();
		
		std::string mtdsname = str( boost::format( "MTframe%05i" ) % iteration );
		CPUData<float> MTcpu;
		HypoArrayMT.getData().copyTo( MTcpu, true );

		/** MT FILE not working here
		MTfile.writeSparseVelSpaceActivities( mtdsname.c_str(), 
											  (VelocityHypothesis*)MTcpu.data(),
											  HypoArrayMT.getData().rank(),  
											  HypoArrayMT.getData().dims,
											  HypoArrayMT.getHMAX()
											  );
		**/

		//TODO: predictive shift(interpolate using nearest neighbour included in computeHyposAmbiguous)
		
		char Name2[1024];
		sprintf(Name2, "HypoMT%03d.txt", iteration );
		HypoArrayMT.printHyposInFile(H_MAX_1, Name2);

		
		
		iteration++;
	  }
	
	STOP_WATCH(calculation);
	std::cout << "END" << std::endl;
	
	
	delete imageSequence;
	delete[] CensusArray1;
	delete[] CensusArray2;
	
	return EXIT_SUCCESS;
}

