/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "census.hpp"
#include "SimpleHostImageSequence.h"
#include "ModelArea.hpp"
#include <GPUImage.h>
#include <HostImage.h>
#include "modelsetup.h"
#include "image_information.hpp"
#include "motion_estimator.hpp"
#include "compute_hypos.hcu"
#include "temp.hpp"
#include "JDBFlowIO.hpp"
#include "MiddleburyFlowIO.hpp"

#include <string>
#include <cstdlib>
#include <cstdio>
#include <cstdarg>

#define START_WATCH(name) clock_t stop_watch_##name##_start = clock();
#define STOP_WATCH(name) \
  clock_t stop_watch_##name##_stop = clock(); \
  std::cout << "Elapsed time for " #name ": " \
            << (((double)1000.0*(stop_watch_##name##_stop - stop_watch_##name##_start)) / CLOCKS_PER_SEC) \
            << " ms" << std::endl;


//#define PRINT_DEBUG_OUTPUT				

enum OUTPUT_DATA { NONE=0, FLOW=1, DUMP=1<<1, FLO=1<<2 };
void output (int iteration, MotionEstimator &model, int w, int h, int out = NONE);



void 
motion_algo (std::string file_pattern, int start, int stop, int out = NONE, bool use_saliency = false)
{	
	// PRELUDE ------
	HostImageSequence* imageSequence = new SimpleHostImageSequence(file_pattern.c_str(), start, stop);
	const int w = imageSequence->getWidth();
	const int h = imageSequence->getHeight();
	int nAdditionalScales = 0;

	// images. at the moment one regular ('fine'), one coarse
	ImageInformation future(w, h);
	ImageInformation current(w, h);	
	ImageInformation past(w, h);

	ImageInformation current_coarse(w/FTCDF, h/FTCDF);
	ImageInformation past_coarse(w/FTCDF, h/FTCDF);

	current_coarse.image.fill(0.0f);
	past_coarse.image.fill(0.0f);


	//
	// Model Setup
	//
	nAdditionalScales++;
	MotionEstimator coarse (w / FTCDF, h / FTCDF, DF_MT, 0);
	HypoFeed coarse_to_fine = {coarse.get_MT_data_ptr(), FBF, SSF_DOWNSAMPLE_COMMON, 
			DF_MT * 2, FTCDF };

	nAdditionalScales++;
	MotionEstimator temporal_lookahead (w, h, DF_MT, 0);
	HypoFeed temporal_lookahead_to_fine = {temporal_lookahead.get_MT_data_ptr(), FBF, SSF_DOWNSAMPLE_COMMON, 
			DF_MT, 1.0 };

	nAdditionalScales++;
	MotionEstimator temporal_clamp (w, h, DF_MT, 0);
	HypoFeed temporal_clamp_to_fine = {temporal_clamp.get_MT_data_ptr(), FBF, SSF_DOWNSAMPLE_COMMON,
			DF_MT, 0.5 };

	MotionEstimator fine (w, h, DF_MT, nAdditionalScales);
					fine.saliencyMask(use_saliency);
	fine.setup_external_feeds(
			&coarse_to_fine ,
			&temporal_lookahead_to_fine ,
			&temporal_clamp_to_fine
	);


	// just for first image >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	future.image.fill(0.0f);
	past.image.fill(0.0f);
	past_coarse.image.fill(0.0f);

	current.image.copyFrom( imageSequence->getNextImage().cpu_data_float(true) );
	future.image.copyFrom( imageSequence->getNextImage().cpu_data_float(true) );
	placeholder::simple_downsample(current.image, current_coarse.image, FTCDF);

	current.transform();
	current_coarse.transform();
	future.transform();

	current.sort(fine.current_list);
	current.sort(temporal_clamp.current_list);
	current_coarse.sort(coarse.current_list);
	future.sort(temporal_lookahead.current_list);
	// just for first image <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


	START_WATCH(calculation);
	int iteration = 0;
	while (imageSequence->hasNextImage()) 
	{
		std::cout << "#" << iteration << " " << std::flush;
		
		fine.init_timestep();
		coarse.init_timestep();	
		temporal_lookahead.init_timestep();
		temporal_clamp.init_timestep();

		current.census_gpu.swap(past.census_gpu);
		current_coarse.census_gpu.swap(past_coarse.census_gpu);
		future.census_gpu.swap(current.census_gpu);

		current.copy_image_from(future);
		placeholder::simple_downsample(current.image, current_coarse.image, FTCDF);
		future.get_next_image(imageSequence);

		current.transform();
		current_coarse.transform();
		future.transform();
		
		temporal_lookahead.process_V1(future, current);
		temporal_clamp.process_V1(future, past);
		coarse.process_V1(current_coarse, past_coarse);
		fine.process_V1(current, past, 
				&coarse, 
				&temporal_lookahead,
				&temporal_clamp);
	
		temporal_lookahead.process_MT();
		temporal_clamp.process_MT();
		coarse.process_MT();
		fine.process_MT();

		output(iteration, fine, w, h, out);
		iteration++;
	}
	STOP_WATCH(calculation);
	std::cout << " END" << std::endl;
}



void 
output (int iteration, MotionEstimator &model, int w, int h, int out)
{
#ifdef PRINT_DEBUG_OUTPUT
	out = out | DUMP;
#endif

	if (out & DUMP)
	{
		char name[1024];
		sprintf(name, "HypoV1%03d.txt", iteration);
		model.dump_V1(name);

		sprintf(name, "HypoMT%03d.txt", iteration);
		model.dump_MT(name);
	} 
	if(out & FLOW) {
		char name[1024];
		sprintf(name, "flow_V1_%03d.flow", iteration);
		GPUImage 	weights(w,h);
		GPUVelField velField = model.getV1HypoImage().getExpectationValue(weights);
					weights.add(0.000001f);
					velField.getXPart().div(weights);
					velField.getYPart().div(weights);
		JDBFlowIO::WriteFlowFile(velField,weights,std::string(name));
		
		sprintf(name, "flow_MT_%03d.flow", iteration);
		velField = model.getMTHypoImage().getExpectationValue(weights);
		weights.add(0.000001f);
		// model.getMTHypoImage().getHypoSum(weights);
		velField.getXPart().div(weights);
		velField.getYPart().div(weights);
		JDBFlowIO::WriteFlowFile(velField,weights,std::string(name));
	} 
	if(out & FLO)
	{
		char name[1024];
		sprintf(name, "flow_V1_%03d.flo", iteration);
		GPUImage 	weights(w,h);
		GPUVelField velField = model.getV1HypoImage().getExpectationValue(weights);
					weights.add(0.000001f);
					velField.getXPart().div(weights);
					velField.getYPart().div(weights);
		MiddleburyFlowIO::WriteFlowFile(velField,weights,name);
		
		sprintf(name, "flow_MT_%03d.flo", iteration);
		velField = model.getMTHypoImage().getExpectationValue(weights);
		weights.add(0.000001f);
		// model.getMTHypoImage().getHypoSum(weights);
		velField.getXPart().div(weights);
		velField.getYPart().div(weights);
		MiddleburyFlowIO::WriteFlowFile(velField,weights,name);
	} 
}



void
parse_args (int argc, char **argv, int *output, bool *use_saliency) 
{
	for (int i=4; i<argc; i++) {
		std::string out(argv[i]);
		if ((out.find("flo")!=std::string::npos  || out.find("FLO")!=std::string::npos) 
		&& (out.find("flo") != out.find("flow") || out.find("FLO") != out.find("FLOW"))) {
			*output = *output | FLO;
			std::cout << "enabled .flo output"<< std::endl;
			continue;
		}
		if (out.find("flow")!=std::string::npos || out.find("FLOW")!=std::string::npos) {
			*output = *output | FLOW;
			std::cout << "enabled .flow output"<< std::endl;
			continue;
		}
		if (out.find("dump")!=std::string::npos || out.find("DUMP")!=std::string::npos) {
			*output = *output | DUMP;
			std::cout << "enabled hypo dumping into txt file"<< std::endl;
			continue;
		}
		if (out.find("with")!=std::string::npos || out.find("use_saliency")!=std::string::npos) {
			std::cout << "enabled saliecy mask"<< std::endl;
			*use_saliency = true;
			continue;
		}
	}
}

inline void 
print_help_and_die ()
{
	std::cout << "usage: motionAlgoGPU file_pattern start_index stop_index [options]" << std::endl;
	std::cout << "  file_pattern    the path and pattern for files, e.g. images/yos_%02d.tif" << std::endl;
	std::cout << "  start_index     starting index for substitution in file_pattern" << std::endl;
	std::cout << "  stop_index      stop index for substitution in file_pattern" << std::endl;
	std::cout << "  [options]       = {flo|flow|dump|use_saliency} " << std::endl;
	exit(EXIT_FAILURE);
}


int 
main (int argc, char **argv)
{
	if (argc < 4) {
		print_help_and_die();
	}
	if (argc > 4) {
		int  output = NONE;
		bool use_saliency = false;
		parse_args(argc, argv, &output, &use_saliency);
		motion_algo(argv[1], atoi(argv[2]), atoi(argv[3]), output, use_saliency);
	} else {
		motion_algo(argv[1], atoi(argv[2]), atoi(argv[3]));
	}
	return EXIT_SUCCESS;
}

