/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "motion_estimator.hpp"
#include "modelsetup.h"
#include <cstdarg>

MotionEstimator::
MotionEstimator (const int w, const int h, const int ds_factor, const int n_external_feeds)
	: current_list(new census::CensusValue[w*h])
	, past_list(new census::CensusValue[w*h])
	, _w(w)
	, _h(h)
	, _ds_factor(ds_factor)
	, _nfeeds(n_external_feeds + 1)
	, _feeds(new HypoFeed[_nfeeds])
	, _V1(w, h, _H_MAX_1)
	, _MT(w/ds_factor, h/ds_factor, _H_MAX_1)
	, _hostHypoImage(h, w, 3*_H_MAX_1) // TODO: why h, w, not w, h?
	, _censusOcc(new float[w * h])
	, _censusOccGPU(w, h)
{
	_V1.getData().fill(0.0f);
	_MT.getData().fill(0.0f);
	
	_useSaliencyThreshold = false;
	_saliencyMask = GPUImage(w,h);
	_fast = GPUImage(w,h); _fast.fill(0);
	_slow = GPUImage(w,h); _slow.fill(0);

	_feeds[0].data_ptr = _MT.getData().getDataPointer();
	_feeds[0].weight = FBF;
	_feeds[0].ssf = SSF_DOWNSAMPLE_COMMON;
	_feeds[0].subsampling_factor = DF_MT;
	_feeds[0].speed_scaling = 1;
}


MotionEstimator::
~MotionEstimator ()
{
	delete[] _censusOcc;
	delete[] _feeds;
	delete[] past_list;
	delete[] current_list;
}


void MotionEstimator::
process_V1 (ImageInformation &current, ImageInformation &past, ...)
{
	std::swap(current_list, past_list);

	census::compute_hypos_unambiguous (
			current_list,
			past_list,
			current.census_values,
			_censusOcc,
			_hostHypoImage,
			_w, _h, _H_MAX_1, _VEL_MAX);

	_V1.getData().copyFromHostMemory(_hostHypoImage.data());
	_censusOccGPU.fill(1.0f);

	// reread feed pointers
	_feeds[0].data_ptr = _MT.getData().getDataPointer(); //JDB
	if (_nfeeds > 1) {
		va_list args;
		va_start(args, past);
		for (int i = 1; i < _nfeeds; i++) {
			MotionEstimator *est = va_arg(args, MotionEstimator*);
			_feeds[i].data_ptr = est->get_MT_data_ptr();
		}
		va_end(args);
	}
	
	census::compute_hypos_ambiguous (
			_censusOcc,
			_V1,
			_censusOccGPU,
			current.census_gpu,
			past.census_gpu,
			FBF, _H_MAX_2, _VEL_MAX, 
			_nfeeds, _feeds);

	if (_useSaliencyThreshold) {
		_slow.mul(0.15f); _slow.add(current.image,1-0.15f);
		_fast.mul(0.40f); _fast.add(current.image,1-0.40f);
		_saliencyMask.copyFrom(_fast);
		_saliencyMask.sub(_slow);
		_saliencyMask.abs();
		_saliencyMask.heavisideStepFunction(0.01f);
		const int h_max = _V1.getHMAX();
		for (int i=0; i<h_max; ++i) {
			_V1.getHypothesisProperty(i,0).mul(_saliencyMask);
		}
	}
	_V1.sqrWeights();
	_V1.velocityBlur(_VEL_MAX);
	_V1.normVelSpace();
}


void MotionEstimator::
process_MT ()
{
	// feedforward V1 hypotheses
	_MT.getData().fill(0);
	_V1.subsample_hypo_image(_VEL_MAX, SIGMA_MT, V1_WINDOW_WIDTH, _MT);
	_MT.sqrWeights();
	_MT.velocityBlur(_VEL_MAX);
	_MT.normVelSpace();
}


void MotionEstimator::
setup_external_feeds (HypoFeed *f, ...)
{
	if (_nfeeds <= 1)
		return;

	// first external feed
	copy_hypo_feed(f, &_feeds[1]);

	va_list args;
	va_start(args, f);
	for (int i = 2; i < _nfeeds; i++) {
		HypoFeed *feed = va_arg(args, HypoFeed*);
		copy_hypo_feed(feed, &_feeds[i]);
	}
	va_end(args);
}


void MotionEstimator::
init_timestep ()
{
	_V1.getData().fill(0.0f);
}


float* MotionEstimator::
get_V1_data_ptr ()
{
	return _V1.getData().getDataPointer();
}


float* MotionEstimator::
get_MT_data_ptr ()
{
	return _MT.getData().getDataPointer();
}


void MotionEstimator::
dump_V1 (char *fname)
{
	_V1.printHyposInFile(_H_MAX_1, fname);
}


void MotionEstimator::
dump_MT (char *fname)
{
	_MT.printHyposInFile(_H_MAX_1, fname);
}
