/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/ui/text/TestRunner.h>
#include <GPUImage.h>
#include <GPUHermRedFourierImage.hpp>
#include <GPUComplexImage.hpp>
#include <HostImage.h>
#include <cmath>

class GPUImageTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE(GPUImageTest);
	CPPUNIT_TEST(testCreation);
	CPPUNIT_TEST(testKillBorder);
	CPPUNIT_TEST(testFFT);
	CPPUNIT_TEST(testHermRed);
	CPPUNIT_TEST(testComplexImage);
	CPPUNIT_TEST(testComplexImage2);
	CPPUNIT_TEST(testSobel);
	CPPUNIT_TEST(testZensus);
	CPPUNIT_TEST_SUITE_END();

public:

	void testCreation() {
		GPUImage a( HostImage("lena.png") );
		CPPUNIT_ASSERT(!a.isEmpty());
		CPPUNIT_ASSERT(a.isContiguous());
		CPPUNIT_ASSERT(a.getXRes()==512);
		CPPUNIT_ASSERT(a.getYRes()==512);
	}

	void testKillBorder() {
		GPUImage a( HostImage("lena.png") );
		GPUImage b(a.copy());

//		printf("w=%d,h=%d",a.getWidth(),a.getHeight());

		a.killBorder(2,2);
		b.killBorder(2,2,-2.0f);

		CPPUNIT_ASSERT(a[0][0] == 0.0f);
		CPPUNIT_ASSERT(a[0][1] == 0.0f);
		CPPUNIT_ASSERT(a[2][2] != 0.0f);

		CPPUNIT_ASSERT(a[55][0] == 0.0f);
		CPPUNIT_ASSERT(a[1][0] == 0.0f);
		CPPUNIT_ASSERT(a[2][2] != 0.0f);

		CPPUNIT_ASSERT(a[511][511] == 0.0f);
		CPPUNIT_ASSERT(a[511][511-1] == 0.0f);
		CPPUNIT_ASSERT(a[511-2][511-2] != 0.0f);

		CPPUNIT_ASSERT(a[511][511] == 0.0f);
		CPPUNIT_ASSERT(a[511-1][511] == 0.0f);
		CPPUNIT_ASSERT(a[511-2][511-2] != 0.0f);

		CPPUNIT_ASSERT(b[0][0] ==-2.0f);
		CPPUNIT_ASSERT(b[0][1] ==-2.0f);
		CPPUNIT_ASSERT(b[2][2] !=-2.0f);

		CPPUNIT_ASSERT(b[0][0] ==-2.0f);
		CPPUNIT_ASSERT(b[1][0] ==-2.0f);
		CPPUNIT_ASSERT(b[2][2] !=-2.0f);

		CPPUNIT_ASSERT(b[511][511] ==-2.0f);
		CPPUNIT_ASSERT(b[511][511-1] ==-2.0f);
		CPPUNIT_ASSERT(b[511-2][511-2] !=-2.0f);

		CPPUNIT_ASSERT(b[511][511] ==-2.0f);
		CPPUNIT_ASSERT(b[511-1][511] ==-2.0f);
		CPPUNIT_ASSERT(b[511-2][511-2] !=-2.0f);
	}

	void testFFT() {
		const unsigned w = 8;
		const unsigned h = 8;
		GPUImage image(w,h);

		fillLinSpace(image,-1,1);
		image.exp();

		const float sumL = image.sum();

		GPUImage temp(image.copy());


//		printf("mean=%f\nvari=%f\n",mean,vari);

		GPUHermRedFourierImage res = image.fft2();

		CPPUNIT_ASSERT(res.getDimensionsCount()==3);
		CPPUNIT_ASSERT(res.extent(2)==2);
		CPPUNIT_ASSERT(res.extent(1)==w/2+1);
		CPPUNIT_ASSERT(res.extent(0)==h);

		CPPUNIT_ASSERT(res[0][0][0]==sumL);

//		printf("(0,0) = %f,%f\n",float(res[0][0][0]),float(res[0][0][1]));
//		printf("(0,1) = %f,%f\n",float(res[0][1][0]),float(res[0][1][1]));

		temp.ifft2From(res);
		temp.sub(image);

		CPPUNIT_ASSERT(fabs(temp.sum())<=0.00001);

		res[0][0][0] = 0.0f;

		image.ifft2From(res);

		CPPUNIT_ASSERT(fabs(image.sum())<=0.00001);
	}

	void testHermRed() {
		const int w=18;
		const int h=10;

		GPUHermRedFourierImage im(w,h);
		im.getRealPartView().fill(2.0f);
		im.getImagPartView().fill(3.0f);
//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));

		GPUHermRedFourierImage im2(im.copy());
//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));
		im.getImagPartView().mul(-1.0f);

//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));
//		printf("real(z) = %f\nimag(z) = %f\n\n",float(im2[0][0][0]),float(im2[0][0][1]));

		im.mulWithComplexFilter(im2);

//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));

		CPPUNIT_ASSERT(im[0][0][0]==13.0f);
		CPPUNIT_ASSERT(im[0][0][1]==0.0f);
		CPPUNIT_ASSERT(im[h-1][w/2-2][0]==13.0f);
		CPPUNIT_ASSERT(im[h-1][w/2-2][1]==0.0f);
		CPPUNIT_ASSERT(im[h/2][w/2-1][0]==13.0f);
		CPPUNIT_ASSERT(im[h/2][w/2-1][1]==0.0f);
	}

	void testComplexImage() {
		const int w=18;
		const int h=10;

		GPUComplexImage im(w,h);
		im.getRealPartView().fill(2.0f);
		im.getImagPartView().fill(3.0f);
//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));

		GPUComplexImage im2(im.copy());
//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));
		im.getImagPartView().mul(-1.0f);

//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));
//		printf("real(z) = %f\nimag(z) = %f\n\n",float(im2[0][0][0]),float(im2[0][0][1]));

		im.mulWithComplexFilter(im2);

//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));

		CPPUNIT_ASSERT(im[0][0][0]==13.0f);
		CPPUNIT_ASSERT(im[0][0][1]==0.0f);
		CPPUNIT_ASSERT(im[h-1][w/2-2][0]==13.0f);
		CPPUNIT_ASSERT(im[h-1][w/2-2][1]==0.0f);
		CPPUNIT_ASSERT(im[h/2][w/2-1][0]==13.0f);
		CPPUNIT_ASSERT(im[h/2][w/2-1][1]==0.0f);
	}

	void testComplexImage2() {
		const int w=28;
		const int h=20;

		GPUComplexImage im(w,h);

		im.fill(0.0f);
		im.getRealPartView()[10][10] = 1.0f;
		im.getRealPartView()[10][11] = 1.0f;

		GPUComplexImage im2 = im.fft2();

		im = im2.ifft2();

		//		printf("real(z) = %f\nimag(z) = %f\n",float(im[0][0][0]),float(im[0][0][1]));
//		printf("%f",im.sum());

		CPPUNIT_ASSERT(fabs(im.sum()-2.0f)<0.001f);
	}

	void testSobel() {
		const int w = 512;
		const int h = 512;
		GPUImage a(HostImage("./lena.png"));

		GPUImage b(w,h);

		a.diffX(b,true);

//		printf("\n%f\n",fabs(b.sum()/(w*h)));

		CPPUNIT_ASSERT(fabs(b.sum()/(w*h))<1.0f);
	}

	void testZensus() {
		GPUImage  a ( HostImage("lena.png") );
		GPUImage  b ( a.getWidth(),a.getHeight() ); b.fill(0.0f);
		a.getZensusTransformedImage(b,a.max()*2/256.0f);
		HostImage im( a.getWidth(),a.getHeight(),HostImage::UBYTE,HostImage::RGBA );

		unsigned int mem[a.getWidth()*a.getHeight()];
		b.copyToHostMemory((float*)&mem);
		//
		b.copyToHostMemory((float*)im.getBytePointerToImage());

		b.sub(b.min());
		b.div(b.max());
		HostImage im2( a.getWidth(),a.getHeight(),HostImage::FLOAT,HostImage::LUM );
		b.copyToHostMemory(im2.getPointerToImage());

		im.writeImage("tests/lena-zensus-RGBA.png");
		im2.writeImage("tests/lena-zensus-Gray.png");

//		im.convert_to_space(HostImage::RGB);
//		im.writeImage("tests/lena-zensus.png");

//		const int w = 22;
//		const int h = 18;
//		const int bwd2 = 8;
//		const int bhd2 = 5;
//		GPUImage a(w,h); a.fill(0.0f);
//		GPUImage b(w,h); b.fill(0.0f);
//		for (int y = h/2-bhd2; y<h/2+bhd2; ++y) {
//			for (int x=w/2-bwd2; x<w/2+bwd2; ++x) {
//				a[y][x]=1.0f;
//			}
//		}
//		a.getZensusTransformedImage(b,0.000001f);//0.5f/w);
//		unsigned int hdata[w*h];
//		b.copyToHostMemory((float*)&hdata);
//		std::cout << std::endl;
//		int i=0;
//		for (int y = 0; y<h; ++y) {
//			for (int x = 0; x<w; ++x) {
//				unsigned int val = hdata[i++];
//				for (int i=0; i<16; ++i) {
//					unsigned int tem  = val%3;
//					std::cout << tem;
//					val /= 3;
//				}
//				std::cout << " ";
//			}
//			std::cout << std::endl;
//		}
	}
};

CPPUNIT_TEST_SUITE_REGISTRATION(GPUImageTest);

int main() {
	CppUnit::TextUi::TestRunner runner;
	CppUnit::TestFactoryRegistry& registry = CppUnit::TestFactoryRegistry::getRegistry();
	runner.addTest(registry.makeTest());
	return ! runner.run("", false);
}


// Local Variables:
// mode: c++
// coding: utf-8
// c-basic-offset: 4
// tab-width: 4
// End:
