/*

GPUIMAGEPROCESSING

  THIS MODULE PROVIDES A SIMPLE API FOR IMAGE OPERATIONS ON
  A GRAPHICAL PROCESSING UNIT (GPU) BY MAKING USE OF CUDA 
  TECHNOLOGY.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/ui/text/TestRunner.h>

#include <HostImage.h>
#include <cmath>


static const std::string pic_LUM("tests/test_LUM.png");
static const std::string pic_RGB("tests/test_RGB.png");
static const std::string pic_RGB_BMP("tests/test_RGB.bmp");
static const float eps = 1e-03;

class HostImageTest : public CppUnit::TestFixture {

  CPPUNIT_TEST_SUITE(HostImageTest);

  CPPUNIT_TEST(testLoadLUM1);
  CPPUNIT_TEST(testLoadLUM2);
  CPPUNIT_TEST(testLoadRGB1);
  CPPUNIT_TEST(testLoadRGB2);
//   CPPUNIT_TEST(testLoadRGB_BMP);
  CPPUNIT_TEST(testConvertTYPE);
  CPPUNIT_TEST(testConvertSPACE);
  CPPUNIT_TEST(testConvertSPACE2);
//   CPPUNIT_TEST(testDeleteShadows);
  CPPUNIT_TEST(testCopy);
  CPPUNIT_TEST(testCrop);
  CPPUNIT_TEST(testSaveLoad);

  CPPUNIT_TEST_SUITE_END();

public:
  


  void testLoadLUM1() 
  {
	// Load the pic as UBYTE
	//HostImage pic1( pic_LUM.c_str(), HostImage::UBYTE, HostImage::LUM );
	HostImage pic1( HostImage::UBYTE, HostImage::LUM );
	pic1.read(pic_LUM.c_str());
	int h1, w1;
	pic1.getSize(h1,w1);
	
	std::string save_name = pic_LUM;
	save_name = save_name.substr(0, save_name.find_last_of("."));
	save_name.append("_testLoadLUM1.png");
	pic1.writeImage(save_name.data() );	

	HostImage pic2( save_name.c_str(), HostImage::UBYTE, HostImage::LUM );
	int h2, w2;
	pic2.getSize(h2,w2);
	
	CPPUNIT_ASSERT(h1>0 && w1>0);
	CPPUNIT_ASSERT(h2>0 && w2>0);
	CPPUNIT_ASSERT(h1==h2 && w1==w2);
	
	float err=0.f;
	CPUData<HostImage::ubyte>::iterator iter1 = pic1.cpu_data_ubyte().begin();
	CPUData<HostImage::ubyte>::iterator iter2 = pic2.cpu_data_ubyte().begin();
	for (;iter1!=pic1.cpu_data_ubyte().end() && iter2!=pic2.cpu_data_ubyte().end(); iter1++, iter2++)
	  {
		err += iter1->manhattan_distance( *iter2 );
	  }
	CPPUNIT_ASSERT( err==0.f );
  }
  
  void testLoadLUM2() 
  {
// Load the pic as LUMINANCE FLOAT
	HostImage pic1( pic_LUM.c_str(), HostImage::FLOAT, HostImage::LUM );

	// save pic...
	std::string save_name = pic_LUM;
	save_name = save_name.substr(0, save_name.find_last_of("."));
	save_name.append("_testLoadLUM2.png");

	pic1.writeImage(save_name.data() );
	
	// ..load again as LUMINANCE FLOAT into another container
	HostImage pic2( save_name.c_str(), HostImage::FLOAT, HostImage::LUM );
	
	float err=0.f;
  CPUData<float>::iterator iter1 = pic1.cpu_data_float().begin();
  CPUData<float>::iterator iter2 = pic2.cpu_data_float().begin();
  for (;iter1!=pic1.cpu_data_float().end() && iter2!=pic2.cpu_data_float().end(); iter1++, iter2++)
    {
    err += iter1->manhattan_distance( *iter2 );
    }

	CPPUNIT_ASSERT( err < eps );
  }
  
  
  void testLoadRGB1() 
  {
	// Load the pic as UBYTE
    HostImage pic1( pic_RGB.c_str(), HostImage::UBYTE, HostImage::RGB );
	  int h1, w1;
	  pic1.getSize(h1,w1);
	  
	  std::string save_name = pic_RGB;
	  save_name = save_name.substr(0, save_name.find_last_of("."));
	  save_name.append("_testLoadRGB1.png");
	  pic1.writeImage(save_name.data() );
	  
	  HostImage pic2( save_name.c_str(), HostImage::UBYTE, HostImage::RGB );
	  int h2, w2;
	  pic2.getSize(h2,w2);
	  
	  CPPUNIT_ASSERT(h1>0 && w1>0);
	  CPPUNIT_ASSERT(h2>0 && w2>0);
	  CPPUNIT_ASSERT(h1==h2 && w1==w2);
	  
	  float err=0.f;
	  
	  CPUData<HostImage::ubyte>::iterator iter1 = pic1.cpu_data_ubyte().begin();
	  CPUData<HostImage::ubyte>::iterator iter2 = pic2.cpu_data_ubyte().begin();
	  for (;iter1!=pic1.cpu_data_ubyte().end() && iter2!=pic2.cpu_data_ubyte().end(); iter1++, iter2++)
	    {
		  err += iter1->manhattan_distance( *iter2 );
	    }
	  
	  CPPUNIT_ASSERT( err==0.f );
  }
  

  void testLoadRGB_BMP()
  {
    // Load the pic as LUMINANCE FLOAT
    HostImage pic1( pic_RGB_BMP.c_str(), HostImage::UBYTE, HostImage::RGB );
  
    // save pic...
    std::string save_name = pic_RGB_BMP;
    save_name = save_name.substr(0, save_name.find_last_of("."));
    save_name.append("_testLoadRGB_BMP.bmp");
  
    pic1.writeImage(save_name.data(), false );
    
    // ..load again as LUMINANCE FLOAT into another container
    HostImage pic2( save_name.c_str(), HostImage::UBYTE, HostImage::RGB );
    
    float err = 0.f;
    CPUData<HostImage::ubyte>::iterator iter1 = pic1.cpu_data_ubyte().begin();
    CPUData<HostImage::ubyte>::iterator iter2 = pic2.cpu_data_ubyte().begin();
    for (;iter1!=pic1.cpu_data_ubyte().end() && iter2!=pic2.cpu_data_ubyte().end(); iter1++, iter2++)
      {
        err += iter1->manhattan_distance( *iter2 );
      }
  
    CPPUNIT_ASSERT( err < eps );
  }

  void testLoadRGB2() 
  {
	  // Load the pic as FLOAT
	  HostImage pic1( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
	  int h1, w1;
	  pic1.getSize(h1,w1);
	  
	  std::string save_name = pic_RGB;
	  save_name = save_name.substr(0, save_name.find_last_of("."));
	  save_name.append("_testLoadRGB2.png");
	  pic1.writeImage( save_name.data() );
	  
	  HostImage pic2( save_name.c_str(), HostImage::FLOAT, HostImage::RGB );
	  int h2, w2;
	  pic1.getSize(h2,w2);
	  
	  CPPUNIT_ASSERT(h1>0 && w1>0);
	  CPPUNIT_ASSERT(h2>0 && w2>0);
	  CPPUNIT_ASSERT(h1==h2 && w1==w2);
	  
	  float err=0.f;
    CPUData<float>::iterator iter1 = pic1.cpu_data_float().begin();
    CPUData<float>::iterator iter2 = pic2.cpu_data_float().begin();
    for (;iter1!=pic1.cpu_data_float().end() && iter2!=pic2.cpu_data_float().end(); iter1++, iter2++)
      {
      err += iter1->manhattan_distance( *iter2 );
      }
  
    CPPUNIT_ASSERT( err < eps );
  }
  
  void testConvertTYPE() 
  {
	  // load image as UBYTE
	  HostImage pic1( pic_RGB.c_str(), HostImage::UBYTE, HostImage::RGB );
	  double err = 0.f;
	  
	  // load image as FLOAT and convert
	  HostImage pic2( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
	  pic2.convert_to_type( HostImage::UBYTE );
	  
	  // check that data is the same
	  CPUData<HostImage::ubyte>::iterator iter1 = pic1.cpu_data_ubyte().begin();
    CPUData<HostImage::ubyte>::iterator iter2 = pic2.cpu_data_ubyte().begin();
    for (;iter1!=pic1.cpu_data_ubyte().end() && iter2!=pic2.cpu_data_ubyte().end(); iter1++, iter2++)
      {
        err += iter1->manhattan_distance( *iter2 );
      }
  
    CPPUNIT_ASSERT( err == 0.f );
	  
	  // load image as FLOAT
	  HostImage pic3( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
	  
	  // load image as UBYTE and convert
	  HostImage pic4( pic_RGB.c_str(), HostImage::UBYTE, HostImage::RGB );
	  pic4.convert_to_type( HostImage::FLOAT );
		  
	  err=0.f;
    CPUData<float>::iterator iter3 = pic3.cpu_data_float().begin();
    CPUData<float>::iterator iter4 = pic4.cpu_data_float().begin();
    for (;iter3!=pic3.cpu_data_float().end() && iter4!=pic4.cpu_data_float().end(); iter3++, iter4++)
      {
      err += iter3->manhattan_distance( *iter4 );
      }
  
    CPPUNIT_ASSERT( err == 0.f );

  }
  
  
  void testConvertSPACE() 
  {
	// load RGB image
	HostImage pic1( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
	
	// load RGB image, convert to YCbCr, convert back to RGB
	HostImage pic2( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
	pic2.convert_to_space( HostImage::YCbCr );
	pic2.convert_to_space( HostImage::RGB );
	
	std::string save_name = pic_RGB;
	save_name = save_name.substr(0, save_name.find_last_of("."));
	save_name.append("_testYCbCr.png");
	pic2.writeImage( save_name.data() );
	
	float err1 = 0.f;
	
	CPUData<float>::fast_iterator iter1 = pic1.cpu_data_float().fast_begin(), end1 = pic1.cpu_data_float().fast_end();
	CPUData<float>::fast_iterator iter2 = pic2.cpu_data_float().fast_begin(), end2 = pic2.cpu_data_float().fast_end();
	for ( ; iter1 != end1 && iter2 != end2; ++iter1, ++iter2 )
	  err1 += iter1->manhattan_distance( *iter2 );
	err1 /= pic1.numPixels();
	   	

	CPPUNIT_ASSERT( std::abs(err1) < eps );	
	
  }
  void testConvertSPACE2() 
  {
	// load RGB image
	HostImage pic1( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
	
	// load RGB image, convert to YCbCr, convert to RGBA, convert back to RGB
	HostImage pic2( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
	pic2.convert_to_space( HostImage::YCbCr );
	pic2.convert_to_space( HostImage::RGBA );
	
	std::string save_name = pic_RGB;
	save_name = save_name.substr(0, save_name.find_last_of("."));
	save_name.append("_testRGBA.png");
	pic2.writeImage( save_name.data() );

	pic2.convert_to_space( HostImage::RGB );
	
	float err1 = 0.f;
	
	CPUData<float>::fast_iterator iter1 = pic1.cpu_data_float().fast_begin(), end1 = pic1.cpu_data_float().fast_end();
	CPUData<float>::fast_iterator iter2 = pic2.cpu_data_float().fast_begin(), end2 = pic2.cpu_data_float().fast_end();
	for ( ; iter1 != end1 && iter2 != end2; ++iter1, ++iter2 )
	  err1 += iter1->manhattan_distance( *iter2 );
	err1 /= pic1.numPixels();
	   	

	CPPUNIT_ASSERT( std::abs(err1) < eps );	
	
  }
/*
	void testDeleteShadows()
	{
		// load RGB image, convert to YCbCr, convert back to RGB
		HostImage pic( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
		pic.convert_to_space( HostImage::YCbCr );
		int h, w;
		pic.getSize(h,w);

		CPUData<float> temp;
		temp.resize(w,h,1);


		CPUData<float>::iterator iter1 = pic.cpu_data_float().begin();
		CPUData<float>::iterator iter2 = temp.begin();
		for (;iter1!=pic.cpu_data_float().end() && iter2!=temp.end(); iter1++, iter2++)
		{
			(*iter2)[0] = ( fabs((*iter1)[1]) + fabs((*iter1)[2]) ) / 2.f;
		}


		pic.convert_to_space(HostImage::LUM);
		pic.cpu_data_float().copy_from( temp );
		std::string save_name = pic_RGB;
		save_name = save_name.substr(0, save_name.find_last_of("."));
		save_name.append("_DeleteShadows.png");
		pic.writeImage(save_name.data() );
	}*/

  void testCopy()
  {
    HostImage pic( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );

    HostImage pic1 = pic;
    int h1, w1;
    pic1.getSize(h1,w1);

    HostImage pic2( pic );
    int h2, w2;
    pic2.getSize(h2,w2);

    pic.clear();

    CPPUNIT_ASSERT(h1>0 && w1>0);
    CPPUNIT_ASSERT(h2>0 && w2>0);
    CPPUNIT_ASSERT(h1==h2 && w1==w2);

    float err=0.f;
    int n=0;
    for (int y=0; y<h1; ++y) 
      {
        for (int x=0; x<w1; ++x) 
          {
            err += std::abs(pic1.cpu_data_float().data()[3*n] - pic2.cpu_data_float().data()[3*n]);
            err += std::abs(pic1.cpu_data_float().data()[3*n+1] - pic2.cpu_data_float().data()[3*n+1]);
            err += std::abs(pic1.cpu_data_float().data()[3*n+2] - pic2.cpu_data_float().data()[3*n+2]);
            ++n;
          }
      }
    CPPUNIT_ASSERT( err == 0 );

  }


  void testCrop()
  {


    // Load a full image
    HostImage pic( pic_RGB.c_str(), HostImage::FLOAT, HostImage::RGB );
    int h, w;
    pic.getSize(h,w);
    //pic.writeImage("tests/test_testCrop_full.png");

    // define the rectangular area that will be cropped
    int row_number = w/2;
    int row_count  = w/4;
    int col_number = h/2;
    int col_count  = h/4;

    // Copy image and crop it
    HostImage pic1( pic );
    pic1.crop( row_number, row_count, col_number, col_count );
    //pic1.writeImage("tests/test_testCrop_cropped1.png");

    // Create empty image of cropped size and copy according pixels manually
    HostImage pic2( col_count, row_count, HostImage::FLOAT, HostImage::RGB );
    pic2.initWithZeros();
   
    for (int i = 0; i < row_count; ++i)
      {
        for (int j = 0; j < col_count; ++j)
          {
             pic2.cpu_data_float()(i,j,0) = pic.cpu_data_float()(i+row_number,j+col_number,0);
             pic2.cpu_data_float()(i,j,1) = pic.cpu_data_float()(i+row_number,j+col_number,1);
             pic2.cpu_data_float()(i,j,2) = pic.cpu_data_float()(i+row_number,j+col_number,2);
          }

      }
    //pic2.writeImage("tests/test_testCrop_cropped2.png");


    // Compare cropped image to manually set image
    float err = pic1.cpu_data_float().manhattan_distance( pic2.cpu_data_float() ) / pic1.numPixels();
    CPPUNIT_ASSERT( err == 0 );
  }

  void testSaveLoad()
  {
  HostImage image1( 1000, 1000, HostImage::UBYTE, HostImage::LUM );
  HostImage::BData & bdata1 = image1.cpu_data_ubyte();
  std::fill( bdata1.begin(), bdata1.end(), 0 );
  unsigned oner = 111, onec = 222;
  bdata1(oner,onec,0) = 1;
  image1.writeImage( "testSaveLoad.png" );
  HostImage image2( "testSaveLoad.png", HostImage::UBYTE);
  CPPUNIT_ASSERT( image2.getSpace() == HostImage::LUM );

  HostImage::BData & bdata2 = image2.cpu_data_ubyte();
  CPPUNIT_ASSERT( image2.getWidth() == image1.getWidth() && image2.getHeight() == image1.getHeight() );
  for ( unsigned i = 0; i < image1.getWidth(); ++i )
    for ( unsigned j = 0; j < image1.getHeight(); ++j )
    {
     CPPUNIT_ASSERT( bdata2(i,j,0) == ( ( oner == i && onec == j ) ? 1 : 0 ) );
    }

  }

   
};

CPPUNIT_TEST_SUITE_REGISTRATION(HostImageTest);
/*
  int main() {
  CppUnit::TextUi::TestRunner runner;
  CppUnit::TestFactoryRegistry& registry = CppUnit::TestFactoryRegistry::getRegistry();
  runner.addTest(registry.makeTest());
  return ! runner.run("", false);
  }
*/

