/*

GTKMMBINDINGS

  THIS MODULE PROVIDES AN INTERFACE FOR VISUALISATION OF OPTIC 
  FLOW, SALIENCY MAPS OR IMAGES BY MAKING USE OF GTK.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/

#ifndef _BOXFILTER_KERNEL_H_
#define _BOXFILTER_KERNEL_H_

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "cutil.h"
//#include <cutil_math.h>

//texture< float, 2> tex;
//texture< uchar4, 2, cudaReadModeNormalizedFloat> rgbaTex;
//
//cudaArray* d_array, *d_tempArray;

// RGBA version
// reads from 32-bit uint array holding 8-bit RGBA

// convert floating point rgba color to 32-bit integer
__device__ uint rgbaFloatToInt(float4 rgba)
{
    rgba.x = __saturatef(rgba.x);   // clamp to [0.0, 1.0]
    rgba.y = __saturatef(rgba.y);
    rgba.z = __saturatef(rgba.z);
    rgba.w = __saturatef(rgba.w);
    return (uint(rgba.w*255)<<24) | (uint(rgba.z*255)<<16) | (uint(rgba.y*255)<<8) | uint(rgba.x*255);
}

__device__ float4 rgb( float r, float g, float b )
{
  float4 v;
  v.x = r;
  v.y = g;
  v.z = b;
  v.w = 1;
  return v;
}


__device__ float4 rgbaIntToFloat(uint c)
{
    float4 rgba;
    rgba.x = (c & 0xff) / 255.0f;
    rgba.y = ((c>>8) & 0xff) / 255.0f;
    rgba.z = ((c>>16) & 0xff) / 255.0f;
    rgba.w = ((c>>24) & 0xff) / 255.0f;
    return rgba;
}

inline float __device__ dd_doSomeThing( float h, float q , float p ) 
{
  const float one_halve = 0.5f;
  const float two_third = 2.0f / 3.0f;
  const float one_sixth = 1.0f / 6.0f;

  while ( h >= 1 )
	h -= 1;
  
  while ( h < 0 ) 
	h += 1;
  
  if      ( h < one_sixth )
	return  p + ( q - p ) * 6.0f * h;
  else if ( h < one_halve )
	return q;
  else if ( h < two_third )
	return p + ( q - p ) * 6.0f * ( two_third - h );
  else
	return p;
}

__device__ float4 hlsToRGBFloat( float h, float l, float s )
{
  float4 fVal;
  const float twopi = 2 * M_PI;
  const float one_third = 1.0f / 3.0f;

	l = max( 0.f, min( 1.f, l ) );
	s = max( 0.f, min( 1.f, s ) );
	

  if ( s == 0 )
	{
	  fVal.x = l;
	  fVal.y = l;
	  fVal.z = l;
	}
  else
	{

	  float  q, p;

	  h /= twopi;

	  if ( l <= 0.5f )
		q = l * ( 1 + s );
	  else
		q = l + s - ( l * s );
	  
	  p =  2 * l - q;
	  
	  // actually map to bgra, its faster
	  fVal.z = dd_doSomeThing( h + one_third, q, p );
	  fVal.y = dd_doSomeThing( h            , q, p );
	  fVal.x = dd_doSomeThing( h - one_third, q, p );
	  fVal.w = 1.0f;

	}

  return fVal;
}

// Input should be in interval [0,1]
__device__ float4 grayToJet( const float & grayValue )
{
	// The following array contains the rgb values at equidistant values. Linear interpolation in between.
	                              // r , g , b
	const float colormap_array[9][3] = { { 0 , 0 , .5 },   // gray value 0
	                                     { 0 , 0 , 1  },   // gray value 0.125
	                                     { 0 , .5, 1  },   // gray value 0.25
	                                     { 0 , 1 , 1  },   // gray value 0.375
	                                     { .5, 1 , .5 },   // 0.5
	                                     { 1 , 1 , 0  },   
	                                     { 1 , .5, 0  },
	                                     { 1 , 0 , 0  },
								         { .5, 0 , 0  } };  // 1.0
	
	float4 output;
	int index = int(8*grayValue);
	index = index>0 ? index : 0;
	float dist_to_lower_index = 8*grayValue - index;  // between 0 and 1
	if ( index<8 )
	{
		output.z = (1-dist_to_lower_index)*colormap_array[index][0] + dist_to_lower_index*colormap_array[index+1][0];
		output.y = (1-dist_to_lower_index)*colormap_array[index][1] + dist_to_lower_index*colormap_array[index+1][1];
		output.x = (1-dist_to_lower_index)*colormap_array[index][2] + dist_to_lower_index*colormap_array[index+1][2];
	}
	else
	{
		output.z = colormap_array[8][0];
		output.y = colormap_array[8][1];
		output.x = colormap_array[8][2];
	}
	output.w = 1;
	return output;
}

// Input should be in interval [0,1]
__device__ float4 grayToHot( const float & grayValue )
{
	// The following array contains the rgb values at equidistant values. Linear interpolation in between.
	                              // r , g , b
	const float colormap_array[9][3] = { { 0 ,  0 ,  0 },   // gray value 0
	         				             { 0.33, 0 , 0 },   // gray value 0.125
	                            		 { 0.67, 0 , 0 },   // gray value 0.25
	                               		 { 1 , 0 ,   0 },   // gray value 0.375
	                                     { 1 , 0.33, 0 },   // 0.5
	                               		 { 1 , 0.67, 0 },   
				                         { 1 , 1 ,  0  },
	             				         { 1 , 1 ,  .5 },
										 { 1 , 1 ,  1  } };  // 1.0
	
	float4 output;
	int index = int(8*grayValue);
	index = index>0 ? index : 0;  // make sure index is non-negative
	float dist_to_lower_index = 8*grayValue - index;  // between 0 and 1
	if ( index<8 )
	{
		output.z = (1-dist_to_lower_index)*colormap_array[index][0] + dist_to_lower_index*colormap_array[index+1][0];
		output.y = (1-dist_to_lower_index)*colormap_array[index][1] + dist_to_lower_index*colormap_array[index+1][1];
		output.x = (1-dist_to_lower_index)*colormap_array[index][2] + dist_to_lower_index*colormap_array[index+1][2];
	}
	else
	{
		output.z = colormap_array[8][0];
		output.y = colormap_array[8][1];
		output.x = colormap_array[8][2];
	}
	output.w = 1;
	return output;
}

__global__ void d_imageGrayToRGBAInt( float* imageGray, unsigned int* dest, const int Nx, const int Ny, int colormap = 0 ) 
{
	const int xIdx = blockIdx.x * blockDim.x + threadIdx.x;
	const int yIdx = blockIdx.y * blockDim.y + threadIdx.y;
	const int memIdx = xIdx+yIdx*Nx;
	if (xIdx<Nx && yIdx<Ny) {
		float4 fVal;
		float  grayVal = imageGray[memIdx];
		// you may expect grayVal to be in [0,1]
		switch(colormap)
		{
			case 0: // gray
				fVal.x = fVal.y = fVal.z = grayVal;
				break;

			case 1: // jet
				fVal = grayToJet( grayVal );
				break;

			case 2:// hot
				fVal = grayToHot( grayVal );
				break;

		}
		fVal.w = 1.0f;
		dest[memIdx] = rgbaFloatToInt(fVal);
	}
}

__global__ void d_imageGrayAndDirectionAndLengthToRGBAInt
( float* imageGray, float* direction, float* length, unsigned int* dest, 
  const int Nx, const int Ny, const int Fx, const int Fy, bool brighten,
  bool legend, bool is_image ) 
{
	const int x = blockIdx.x * blockDim.x + threadIdx.x;
	const int y = blockIdx.y * blockDim.y + threadIdx.y;

	const float xfact = float(Fx) / float(Nx);
	const float yfact = float(Fy) / float(Ny);

	const int fx = int( float(x) * xfact );
	const int fy = int( float(y) * yfact );

	if ( x < Nx && y < Ny && fx < Fx && fy < Fy ) 
	{
	  const int imIdx = x + y  * Nx;
	  const int fIdx = fx + fy * Fx;

	  float  grayVal =        imageGray[ imIdx ];

	  float dir, len;

	  if ( legend )
		{
		  int legend_size = Ny / 10;
		  float dx = x - float( legend_size ) / 2;
		  float dy = y - float( legend_size ) / 2;
		  float dr = sqrtf( dx * dx + dy * dy ) / ( 0.5f * legend_size );
		  if ( dr < 1.0f )
			{
			  len = dr;
			  dir = atan2f( dx, dy );
			  grayVal = 1.f;
			}
		  else
			{
			  dir = direction[ fIdx  ];
			  len = length   [ fIdx  ];
			}
		}
	  else
		{
		  dir = direction[ fIdx  ];
		  len = length   [ fIdx  ];
		}

	  float s = fminf( len, 1.0f );
	  float h = M_PI/2 - dir;
	  float l = 0.7f;
	  if ( is_image )
	  	l = brighten ?  ( 0.25f + grayVal / (0.75f) ) : grayVal;
	  
	  float4  rgba = hlsToRGBFloat( h, l, s );
	  rgba.x = 1.f - grayVal * rgba.x;
	  rgba.y = 1.f - grayVal * rgba.y;
	  rgba.z = 1.f - grayVal * rgba.z;
	  dest[imIdx] = rgbaFloatToInt( rgba ) ;
	  
	}

}

__device__ float4 hsvToRGBFloat( float h, float s, float v )
{
  // http://de.wikipedia.org/wiki/HSV-Farbraum#Abgewandelte_Farbmodelle_HSL.2C_HSB.2C_HSI
  int h_i = 3 * h / M_PI;
  float f = 3 * h / M_PI - h_i;
  float p = v * ( 1 - s );
  float q = v * ( 1 - s * f );
  float t = v * ( 1 - s * ( 1 - f ) );
  switch( h_i % 6 )
	{
	case 0:         return rgb( v, t, p );
	case 1:         return rgb( q, v, p );
	case 2:         return rgb( p, v, t );
	case 3:         return rgb( p, q, v );
	case 4:         return rgb( t, p, v );
	case 5:         return rgb( v, p, q );
	}  
  return rgb( 1, 1, 1 );
}


__global__ void d_hsvToRGBAInt( float* imageGray, float* features, unsigned int* dest,
												 const int Nx, const int Ny,
												 const int Fx, const int Sfx, const int Fy, const int Sfy, const int Nf, const int Sff )
{
	const int xIdx = blockIdx.x * blockDim.x + threadIdx.x;
	const int yIdx = blockIdx.y * blockDim.y + threadIdx.y;

	const float xfact = (float)Fx/(float)Nx;
	const float yfact = (float)Fy/(float)Ny;

	const int fxIdx = int( float(xIdx) * xfact );
	const int fyIdx = int( float(yIdx) * yfact );

	if ( xIdx < Nx && yIdx < Ny && fxIdx < Fx && fyIdx < Fy ) 
	  {
		const int fIdx = fxIdx * Sfx + fyIdx * Sfy;
		const int imIdx = xIdx+yIdx*Nx;
		float 
		  h = features[fIdx + 0 * Sff], 
		  s = features[fIdx + 1 * Sff], 
		  v = features[fIdx + 2 * Sff];
		dest[ imIdx ] = rgbaFloatToInt( hsvToRGBFloat( h, s, v ) );
	  }
}

__global__ void d_imageGrayAndFeaturesToRGBAInt( float* imageGray, float* features, unsigned int* dest,
												 const int Nx, const int Ny,
												 const int Fx, const int Sfx, const int Fy, const int Sfy, const int Nf, const int Sff,
												 bool saturate_entropy, bool brighten, bool maxindex )
{
	const int xIdx = blockIdx.x * blockDim.x + threadIdx.x;
	const int yIdx = blockIdx.y * blockDim.y + threadIdx.y;

	const float xfact = (float)Fx/(float)Nx;
	const float yfact = (float)Fy/(float)Ny;

	const int fxIdx = int( float(xIdx) * xfact );
	const int fyIdx = int( float(yIdx) * yfact );

	if ( xIdx < Nx && yIdx < Ny && fxIdx < Fx && fyIdx < Fy ) 
	  {
		const int fIdx = fxIdx * Sfx + fyIdx * Sfy;
		const int imIdx = xIdx+yIdx*Nx;
		const float twopi = 2 * 3.1415926536f;

		float  grayVal = imageGray[imIdx];
		float l;
		float s = 1;
		float h = 0;
		float norm = 0, mpos = 0, H = 0;

		float maxpos = 0, maxval = features[fIdx];
		float ms = 0, mc = 0;
		
		for ( int k = 0; k < Nf; ++k )
		  {
			float fval = features[fIdx + k * Sff];
			float angle = twopi * float(k) / float(Nf);
			norm += fval;
			mpos += k * fval;
			mc += cos( angle ) * fval;
			ms += sin( angle ) * fval;
			if ( fval > maxval )
			  maxpos = (float)k;
		  }

		norm = max( 1e-10f, norm );

		if ( maxindex )
		  mpos = maxpos;
		else
		  mpos = mpos / norm;
		
		float meanangle = 0;
		if ( norm > 0 )
		  {
			mc /= norm;
			ms /= norm;
			meanangle = atan2( ms, mc );
		  }

		//if ( norm < 1.0f ) grayVal *= norm;

		if (brighten) l = 0.25f + grayVal / 2.2f;
		else          l = grayVal / 2.f;

		if ( Nf == 1 || norm == 0 )
		  s = norm > 0.0f ? ( norm < 1.0f ? norm : 1.0f ) : 0.0f;
		else
		  {
			//h = twopi * mpos / Nf;
			h = meanangle;
			 // uses entropy to determine saturation
			if ( saturate_entropy )
			  {
				for ( int k = 0; k < Nf; ++k )
				  {
					float p = features[fIdx + k * Sff] / norm;
					if ( p > 0.0f )
					  H -= p * log(p);
				  }
				s = 1.0f - H / logf( Nf );
			  }
			else
			  {
				s = sqrt( ms * ms + mc * mc );
			  }
		  }

		dest[ imIdx ] = rgbaFloatToInt( hlsToRGBFloat( h, l, s ) );

	  }

}
//
//__global__ void d_imageGrayAndDirectionAndLengthToRGBAInt_with_new_color_code(float* imageGray, float* direction, float* length, unsigned int* dest, const int Nx, const int Ny) {
//	const int bIndexX = blockIdx.x;
//	const int bIndexY = blockIdx.y;
////	const int threadIdx = threadIdx.x;
//	const int xIdx = bIndexX*blockDim.x + threadIdx.x;
//	const int yIdx = bIndexY;
//	const int memIdx = xIdx+yIdx*Nx;
//	if (xIdx<Nx && yIdx<Ny) {
//		float4 fVal;
//		float  grayVal = imageGray[memIdx];
//		float  s = length[memIdx];
//		float  h = (direction[memIdx]/3.1415926536f)/2.0f+1.0f;
//			   h = fmaxf(h,0.0f);
//			   h = fminf(h,0.999999999999999999f);
//		int    hIdx1 = (int)h*55;
//		int    hIdx2 = (hIdx1+1) % 55;
//		float  dH = h-hIdx1;
//
//		h = colorWeel[hIdx1]*(1.0f-dH)+colorWeel[hIdx2]*dH;
//
//		fVal.x = h*s*(0.8f+0.2f*grayVal);
//		fVal.y = h*s*(0.8f+0.2f*grayVal);
//		fVal.z = h*s*(0.8f+0.2f*grayVal);
//
//
//
//		fVal.w = 1.0f;
//
//		dest[memIdx] = rgbaFloatToInt(fVal);
//	}
//}

__global__ void d_arrows(float* xPart, float* yPart, float4* pos, unsigned int w, const unsigned int h, const int overS) 
{
	const int bIndexX = blockIdx.x;
	const int bIndexY = blockIdx.y;
	const int xIdx = bIndexX*blockDim.x + threadIdx.x;
	const int yIdx = bIndexY*blockDim.y + threadIdx.y;
	const int memIdx = xIdx+yIdx*w;
	const float wf = ((float)w);
	const float hf = ((float)h);
	if (xIdx<w && yIdx<h)
	  {
		pos[memIdx*6+0] = make_float4((xIdx+0.6f)/wf,1.0f-(yIdx+0.6f)/hf,0.0f,1.0f);
		pos[memIdx*6+1] = make_float4((xIdx+0.4f)/wf,1.0f-(yIdx+0.4f)/hf,0.0f,1.0f);
		pos[memIdx*6+2] = make_float4((xIdx+0.6f)/wf,1.0f-(yIdx+0.4f)/hf,0.0f,1.0f);
		pos[memIdx*6+3] = make_float4((xIdx+0.4f)/wf,1.0f-(yIdx+0.6f)/hf,0.0f,1.0f);
		pos[memIdx*6+4] = make_float4((xIdx+0.5f)/wf,1.0f-(yIdx+0.5f)/hf,0.0f,1.0f);
		pos[memIdx*6+5] = make_float4((xIdx+0.5f)/wf+xPart[memIdx]/(wf*overS),1.0f-(yIdx+0.5f)/hf-yPart[memIdx]/(hf*overS),-0.15f,1.0f);
	  }
}

__global__ void d_arrows2(float* xPart, float* yPart, float4* pos, unsigned int w, const unsigned int h, const int overS) 
{
	const int bIndexX = blockIdx.x;
	const int bIndexY = blockIdx.y;
	const int xIdx = bIndexX*blockDim.x + threadIdx.x;
	const int yIdx = bIndexY*blockDim.y + threadIdx.y;
	const int memIdx = xIdx+yIdx*w;
	const int aIdx = 6 * memIdx;
	const float wf = ((float)w);
	const float hf = ((float)h);
	if ( xIdx < w && yIdx <h )
	  {
		float dx = xPart[ memIdx ] / ( wf * overS );
		float dy = -yPart[ memIdx ] / ( hf * overS );
		float cx = ( xIdx + 0.5f ) / wf;
		float cy = 1.0f - ( yIdx + 0.5f ) / hf;
		float4 c = make_float4( cx, cy, 0.0f, 1.0f );
		float4 p1 = make_float4( cx + dx    , cy + dy    , 0.0f, 1.0f );
		float4 p2 = make_float4( cx + dx/5  , cy + dy/5  , 0.0f, 1.0f );
		pos[ aIdx + 0 ] = make_float4( cx + dy/4, cy - dx/4, 0.0f, 1.0f );
		pos[ aIdx + 1 ] = p1;
		pos[ aIdx + 2 ] = p2;
		pos[ aIdx + 3 ] = make_float4( cx - dy/4, cy + dx/4, 0.0f, 1.0f );
		pos[ aIdx + 4 ] = p2;
		pos[ aIdx + 5 ] = p1;
	  }
}

//
//extern "C" void initTexture(int width_ToShow, int height_ToShow, void *pImage)
//{
//	int size = width_ToShow * height_ToShow * sizeof(unsigned int);
//
//	// copy image data to array
//	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(8, 8, 8, 8, cudaChannelFormatKindUnsigned);
//	CUDA_SAFE_CALL( cudaMallocArray  ( &d_array, &channelDesc, width_ToShow, height_ToShow ));
//	//CUDA_SAFE_CALL( cudaMemcpyToArray( d_array, 0, 0, pImage, size, cudaMemcpyHostToDevice));
//
//	CUDA_SAFE_CALL( cudaMallocArray  ( &d_tempArray,   &channelDesc, width_ToShow, height_ToShow ));
//
//	// set texture parameters
//	tex.addressMode[0] = cudaAddressModeClamp;
//	tex.addressMode[1] = cudaAddressModeClamp;
//	tex.filterMode = cudaFilterModePoint;
//	tex.normalized = true;
//
//	// Bind the array to the texture
//	CUDA_SAFE_CALL( cudaBindTextureToArray(tex, d_array, channelDesc) );
//}

//extern "C"
//void freeTextures()
//{
//    CUDA_SAFE_CALL(cudaFreeArray(d_array));
//    CUDA_SAFE_CALL(cudaFreeArray(d_tempArray));
//
//}

extern "C"
void copyToPBO(float* imageGray, unsigned int* dest,const int Nx, const int Ny, const int colormap)
 {
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(Nx,dimBlock2d.x),iDivUp(Ny,dimBlock2d.y));
  d_imageGrayToRGBAInt<<<dimGrid2d, dimBlock2d>>>(imageGray,dest,Nx,Ny,colormap);
}

/*
__global__ void d_imageGrayAndDirectionAndLengthToRGBAInt( float* imageGray, float* direction, float* length,
														   unsigned int* dest, 
														   const int Nx, const int Ny, const int Fx, const int Fy, bool brighten ) 
														   */
extern "C"
void copyToPBOColorCode( float* imageGray, float* angle, float* length,
						 unsigned int* dest, const int Nx, const int Ny, 
						 const int Fx, const int Fy, 
						 bool brighten, bool legend, bool is_image) 
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(Nx,dimBlock2d.x),iDivUp(Ny,dimBlock2d.y));
  d_imageGrayAndDirectionAndLengthToRGBAInt<<<dimGrid2d, dimBlock2d>>>
	( imageGray, angle, length, dest, Nx, Ny, Fx, Fy, brighten, legend, is_image );
}

extern "C"
void copyToVBOArrowsLines(float* xPart,float* yPart , float4* pos, const unsigned int w , const unsigned int h, const int overs) {
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(w,dimBlock2d.x),iDivUp(h,dimBlock2d.y));
  d_arrows<<<dimGrid2d, dimBlock2d>>>( xPart, yPart, pos, w, h, overs );
}

extern "C"
void copyToVBOArrowsTriangles(float* xPart,float* yPart , float4* pos, const unsigned int w , const unsigned int h, const int overs) {
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(w,dimBlock2d.x),iDivUp(h,dimBlock2d.y));
  d_arrows2<<<dimGrid2d, dimBlock2d>>>( xPart, yPart, pos, w, h, overs );
}

extern "C"
void copyToPBOFeatures( float* imageGray, float* features, unsigned int* dest,
						const int Nx, const int Ny, 
						const int Fx, const int Sfx, const int Fy, const int Sfy, const int Ff,const int Sff,
						const bool saturate_entropy, const bool brighten, const bool maxindex ) 
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(Nx,dimBlock2d.x),iDivUp(Ny,dimBlock2d.y));
  d_imageGrayAndFeaturesToRGBAInt<<<dimGrid2d, dimBlock2d>>>
	( imageGray, features, dest, 
	  Nx, Ny, Fx, Sfx, Fy, Sfy, Ff, Sff,  
	  saturate_entropy, brighten, maxindex );
}

extern "C"
void copyToPBOHSV( float* imageGray, float* features, unsigned int* dest,
				   const int Nx, const int Ny, 
				   const int Fx, const int Sfx, const int Fy, const int Sfy, const int Ff,const int Sff )
{
  dim3 dimBlock2d = GPUData_dimBlock2d;
  dim3 dimGrid2d(iDivUp(Nx,dimBlock2d.x),iDivUp(Ny,dimBlock2d.y));
  d_hsvToRGBAInt<<<dimGrid2d, dimBlock2d>>>
	( imageGray, features, dest, 
	  Nx, Ny, Fx, Sfx, Fy, Sfy, Ff, Sff );
}

#endif
