/*

GTKMMBINDINGS

  THIS MODULE PROVIDES AN INTERFACE FOR VISUALISATION OF OPTIC 
  FLOW, SALIENCY MAPS OR IMAGES BY MAKING USE OF GTK.

COPYRIGHT (C)

  ULM UNIVERSITY 
  INSTITUTE OF NEURAL INFORMATION PROCESSING (NEUROINFORMATIK)
  89069 ULM
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  INSTITUTE OF NEURAL INFORMATION PROCESSING:
  http://www.uni-ulm.de/en/in/institute-of-neural-information-processing.html

  ULM UNIVERSITY 
  HTTP://WWW.UNI-ULM.DE/

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/

/*
 * DataWindow.cpp
 *
 *  Created on: Sep 30, 2008
 *      Author: jbouecke
 */

#include "DataWindow.hpp"
#include <cuda_runtime.h>
#include <cutil.h>
#include <cuda_gl_interop.h>
#include "CPUData.hpp"

#include <IL/il.h>
#include <IL/ilu.h>
#include <boost/format.hpp>
#include <boost/foreach.hpp>

//cant get this to work...
//#define USE_NEW_MAPPING

#include <boost/format.hpp>

#ifdef USE_NEW_MAPPING
#include <cudaGL.h>
#endif

#ifdef HAVE_GLUT
#include "glut_render_string.hpp"
#endif

#ifndef USE_NEW_MAPPING
# define PREAMBLE(bo,presult) cudaGLRegisterBufferObject(bo); cudaGLMapBufferObject((void**)&presult, bo);
# define POSTAMBLE(bo) cudaGLUnmapBufferObject(bo); cudaGLUnregisterBufferObject(bo);
#else
# define PREAMBLE(bo,presult) unsigned int result_size; CUgraphicsResource resource; \
  cuGraphicsGLRegisterBuffer( &resource, bo, CU_GRAPHICS_MAP_RESOURCE_FLAGS_WRITE_DISCARD ); \
  cuGraphicsResourceGetMappedPointer( (CUdeviceptr*)&presult, &result_size, resource );
# define POSTAMBLE(bo)  cuGraphicsUnregisterResource( resource );
#endif

void check_gl( const std::string & msg )
{
  GLenum err = glGetError();
  if ( err != GL_NONE )
	{
	  throw std::runtime_error(str(boost::format("open gl error: %i msg: \"%s\"")
								   % int(err) % msg ) );
	}
}


DataWindow::DataWindow(int _w, int _h, const char* _title )
  : w( _w ), h( _h )
  , arrow_spacing( 15 )
  , brighten( true )
  , is_image( false )
  , saturate_entropy( true )
  , maxindex( false )
  , updateEveryTime( false )
  , title( _title )
  , style( TRIANGLES )
  , maxSpeed( -1 )
  , vbo(0)
  , pbo(0)
  , texid(0)
  , _have_flow_legend(true)
  , _show_min_max(true)
  , _show_fps(true)
  , _marker_size( 0.05 )
  , _box_size( 0.10 )
  , _colormap(0)
  , _random_colormap(true)
  , _line_width(2)
  , _opacity(0.5)
  , _show_statusbar(true)
#ifdef HAVE_GLUT
  , _render_score(false)
  , _render_score_size( 1 )
  , _render_score_pattern( "%1.2e" )
#endif
{
  initGUI();
}

void DataWindow::set_colormap( int c ) { _colormap = c; }

void DataWindow::set_random_colormap( bool is_random ) { _random_colormap = is_random;};

DataWindow::~DataWindow()
{
  dealloc_vbo();
}

void DataWindow::setMaxSpeed(float s)
{
  maxSpeed = s;
}

void DataWindow::setArrowSpacing( int spacing )
{
  arrow_spacing = spacing;
}

void DataWindow::doubleWindowSize()
{
  int wL = da.get_allocation().get_width();
  int hL = da.get_allocation().get_height();
  resize( wL * 2, hL * 2 );
}

void DataWindow::resize( int nw, int nh )
{
  da.set_size_request( nw, nh );
  vBox.check_resize();
  this->check_resize();
  repaint(true);
}

void DataWindow::alloc_pbo_and_texture()
{
  const int iw = imageLocal.getWidth();
  const int ih = imageLocal.getHeight();

  gl_begin();

  if ( pbo != 0 ) glDeleteBuffers( 1, &pbo );
  glGenBuffersARB(1, &pbo);
  glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
  glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, iw * ih *sizeof(GLubyte)*4, 0, GL_DYNAMIC_DRAW );
  glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

  if ( texid != 0 ) glDeleteTextures(1,&texid);
  glGenTextures(1, &texid);
  glBindTexture(GL_TEXTURE_2D, texid);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, iw, ih, 0, GL_BGRA, GL_UNSIGNED_BYTE, NULL);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glBindTexture(GL_TEXTURE_2D, 0);

  check_gl("DataWindow::alloc_pbo_and_texture");

  gl_end();

}

HostImage DataWindow::dump_to_host()
{

  const int w = da.get_allocation().get_width();
  const int h = da.get_allocation().get_height();

  HostImage image( w, h, HostImage::UBYTE, HostImage::RGB );

  gl_begin();
    glPixelStorei( GL_PACK_ALIGNMENT, 1 );
    glReadBuffer( GL_BACK );
    glReadPixels( 0, 0, w, h, GL_RGB, GL_UNSIGNED_BYTE, image.cpu_data_ubyte().data() );
	check_gl("DataWindow::dump_to_host");
  gl_end();

  return image;

}

void DataWindow::writeToFile(std::string filename)
{
  HostImage image = dump_to_host();
  image.writeImage( filename.c_str(), false );
}


void DataWindow::dealloc_vbo()
{
  gl_begin();
  if ( vbo != 0 ) glDeleteBuffers( 1, &vbo );
	check_gl("DataWindow::dealloc_vbo");
  gl_end();
}

void DataWindow::alloc_vbo()
{

  const int size = xpart.size() * sizeof(float) * 6 * 4;  //six 4d vertices per arrow

  gl_begin();

  if ( vbo != 0 ) glDeleteBuffers( 1, &vbo );
  glGenBuffers( 1, &vbo );
  glBindBuffer( GL_ARRAY_BUFFER, vbo );
  glBufferData( GL_ARRAY_BUFFER, size, 0, GL_DYNAMIC_DRAW );
  glBindBuffer( GL_ARRAY_BUFFER, 0 );
	check_gl("DataWindow::alloc_vbo");

  gl_end();

}

void DataWindow::update( const GPUImage& ref )
{
  imageRef.reference( ref );
  featuresRef.clear();
  ffieldRef.getData().clear();
  update();
}

void DataWindow::update( const GPUImage& ref, const GPUVelField& velFiel )
{
  imageRef.reference( ref );
  ffieldRef = velFiel;
  featuresRef.clear();
  update();
}

void DataWindow::update( const GPUVelField& vf )
{
  GPUImage im( vf.getWidth(), vf.getHeight());
  im.fill(0.5f);
  update( im, vf );
}

void DataWindow::update( const GPUImage& ref, const GPUData& features )
{
  imageRef.reference( ref );
  featuresRef.reference( features );
  ffieldRef.getData().clear();
  hsv = false;
  update();
}

void DataWindow::update( const GPUData& f )
{
  GPUImage im( f.extent(1), f.extent(0) );
  im.fill(1.0f);
  update( im, f );
}

void DataWindow::update( const GPUData& f, bool flag )
{
  hsv = flag;
  GPUImage im( f.extent(1), f.extent(0) );
  im.fill(1.0f);
  imageRef.reference( im );
  featuresRef.reference( f );
  ffieldRef.getData().clear();
  update();
}

void DataWindow::update()
{
  size_t irw = imageRef.getWidth();
  size_t irh = imageRef.getHeight();

  size_t vrw = ffieldRef.getWidth();
  size_t vrh = ffieldRef.getHeight();

  size_t frw = featuresRef.extent(1);
  size_t frh = featuresRef.extent(0);
  size_t frd = featuresRef.extent(2);

  bool have_image = !imageRef.isEmpty();
  bool have_ffield = !ffieldRef.getData().isEmpty();
  bool have_features = !featuresRef.isEmpty();

  bool image_changed = ( irw != imageLocal.getWidth() ) || ( irh != imageLocal.getHeight() );
  bool ffield_changed = ( vrw != xpart.getWidth() ) || ( vrh != xpart.getHeight() );
  //bool features_changed = ( frw != featuresLocal.extent(1) ) || ( frh != featuresLocal.extent(0) ) || ( frd != featuresLocal.extent(2) );

  const size_t w = da.get_allocation().get_width();
  const size_t h = da.get_allocation().get_height();

  std::string statsstr, titlestr;


  double now = double( clock() ) / double(CLOCKS_PER_SEC);
  double delta = now - _last_time;
  _fps = 1.0 / delta;
  _last_time = now;

  if ( !have_image )
	return;

  imageLocal = imageRef.copy();

  if ( have_features )
	{
	  size_t one = 0, two = 1;
	  featuresLocal.reference( featuresRef.copy().reduceDimensionality(one,two) );
	  featuresLocal.abs();
      if ( _show_min_max && _show_statusbar )
      {
        min_feat = featuresRef.min();
        max_feat = featuresRef.max();
        mean_feat = featuresRef.mean();
      }
	  
	}

  if ( image_changed )
	{
	  alloc_pbo_and_texture();
	}

  float minL = 0.f, maxL = 0.f, meanL = 0.f;
  if ( _show_min_max && _show_statusbar )
  {
    minL = imageLocal.min();
    maxL = imageLocal.max();
    meanL= imageLocal.mean();
  }
  /*
  if ( maxL > minL )
	{
	  imageLocal.add( -minL );
	  imageLocal.div( maxL - minL );
	}
  else
	imageLocal.fill(1.0f);
  */

  statsstr += str( boost::format( "i: [%1.1e,%1.1e,%1.1e] " ) % minL % meanL % maxL );
  titlestr += str( boost::format( "%s (%dx%d) im (%dx%d) ") % title % w % h % irw % irh );

  double t0 = clock();

  if ( have_ffield )
	{

	  ffieldRef.getSpeed( speed );
	  ffieldRef.getDirection( direction );

	  float speed_minL = 0.f, speed_maxL = 0.f, speed_meanL = 0.f;
      if ( _show_min_max && _show_statusbar )
      {
        speed_minL = speed.min();
        speed_maxL = speed.max();
        speed_meanL= speed.mean();
      }
  
	  if (maxSpeed > 0)
		speed.div(maxSpeed);
	  else
		speed.div( speed_maxL );

	  speed.clamp(0,1);

	  if ( style != NONE )
		{

		  size_t wa = w / arrow_spacing;
		  size_t ha = h / arrow_spacing;
		  GPUImage xtmp( ffieldRef.getXPart() );
		  GPUImage ytmp( ffieldRef.getYPart() );
		  xpart.resize( wa, ha ); xtmp.extrapolate( xpart, true );
		  ypart.resize( wa, ha ); ytmp.extrapolate( ypart, true );

		  if ( maxSpeed > 0 )
			{
			  xpart.mul( arrow_spacing / maxSpeed / 2 );
			  ypart.mul( arrow_spacing / maxSpeed / 2 );
			}
		  else
			{
			  xpart.mul( arrow_spacing / speed_maxL / 2 );
			  ypart.mul( arrow_spacing / speed_maxL / 2 );
			}

		  if ( ffield_changed || vbo == 0 )
			alloc_vbo();

		  render_arrows();

		} else dealloc_vbo();

	  titlestr += str( boost::format( " vel (%dx%d)" ) % vrw % vrh );
	  statsstr += str( boost::format( " v: [%1.1e,%1.1e,%1.1e]" ) 
					   % speed_minL % speed_meanL % speed_maxL );

	}
  else
	{
	  dealloc_vbo();
	}

  if ( have_ffield )
	{	  
	  render_velocity_image();
	}
  else if ( have_features )
	{
	  titlestr += str( boost::format( " feat (%dx%dx%d)" ) % frw % frh % frd );
	  statsstr += str( boost::format( " f: [%1.1e,%1.1e,%1.1e]" ) % min_feat % mean_feat % max_feat );	  
	  render_feature_image();
	}
  else
	render_plain_image();

  repaint();

  if ( _show_fps )
    statsstr = str( boost::format("fps: %3.1f %s")%_fps%statsstr);

  
  if ( ( _show_fps || _show_min_max ) && _show_statusbar )
    statusBar.push( statsstr.c_str() );
  else
    statusBar.hide();
  set_title( titlestr );

}

void DataWindow::set_range_display( float vmin, float vmax )
{
  std::string statsstr = str( boost::format("[ %1.2e, %1.2e ] ") % vmin % vmax );
  statusBar.push( statsstr.c_str() );
}

//#include "HostImageL.h"
//void DataWindow::writeToFile(const char* fileName) {
//	const int w = da.get_allocation().get_width();
//	const int h = da.get_allocation().get_height();
//
//	unsigned char hData[w*h*4];
//
//	Glib::RefPtr< Gdk::GL::Context  > glcontext  = da.get_gl_context();
//	Glib::RefPtr< Gdk::GL::Drawable > gldrawable = da.get_gl_drawable();
//	gldrawable->make_current(glcontext);
//	if (!gldrawable->gl_begin(glcontext)) g_assert_not_reached ();
//
//	glPixelStorei(GL_PACK_ALIGNMENT,1);
//	glReadBuffer(GL_BACK);
//	glReadPixels(0,0,w,w,GL_RGBA,GL_UNSIGNED_BYTE,hData);
//
//	//HostImageL::saveRGBAImage(fileName, hData, w, h, true);
//
//	gldrawable->gl_end();
//}

extern "C" void copyToPBOColorCode(float* imageGray, float* angel, float* length, 
								   unsigned int* dest,const int Nx, const int Ny, 
								   const int Fx, const int Fy, 
								   bool brighten, bool legend, bool is_image);
extern "C" void copyToVBOArrowsLines(float* xPart,float* yPart , float4* pos, const unsigned int width_ToShow , const unsigned int height_ToShow, const int overs);
extern "C" void copyToVBOArrowsTriangles(float* xPart,float* yPart , float4* pos, const unsigned int width_ToShow , const unsigned int height_ToShow, const int overs);

void DataWindow::render_arrows()
{

  gl_begin();

  float4 *d_f_result;

  PREAMBLE(vbo,d_f_result);
	
  size_t w_arrows = xpart.getWidth();
  size_t h_arrows = xpart.getHeight();
  size_t subs = ( imageLocal.size() > 1 ) ?
	size_t( ( imageLocal.getWidth() + 0.5 ) / w_arrows ) :
	size_t( std::min( (w + 0.5 ) / w_arrows, (h + 0.5) / h_arrows ) );
  
  if ( w_arrows != ypart.getWidth() || h_arrows != ypart.getHeight() )
	throw GPUDataException("x/ypart mismatch while renering arrows");
  
  switch ( style )
	{
	case LINES:copyToVBOArrowsLines( xpart.getDataPointer(),
									 ypart.getDataPointer(),
									 d_f_result, w_arrows, h_arrows, subs ); break;
	case TRIANGLES:copyToVBOArrowsTriangles( xpart.getDataPointer(),
											 ypart.getDataPointer(),
											 d_f_result, w_arrows, h_arrows, subs ); break;
    case NONE: break;
	}
  

  POSTAMBLE(vbo);

  check_gl("DataWindow::render_arrows");

  gl_end();

}

extern "C"
void copyToPBOFeatures( float* imageGray, float* features, unsigned int* dest,
						const int Nx, const int Ny,
						const int Fx, const int Sfx, const int Fy, const int Sfy, const int Ff, const int Sff,
						const bool saturate_entropy, const bool brighten, const bool maxindex );

extern "C"
void copyToPBOHSV( float* imageGray, float* features, unsigned int* dest,
				   const int Nx, const int Ny,
				   const int Fx, const int Sfx, const int Fy, const int Sfy, const int Ff, const int Sff );

extern "C" void copyToPBO(float* imageGray, unsigned int* dest,const int Nx, const int Ny, int colormap);

void DataWindow::render_plain_image()
{

  gl_begin();

  // copy image to pbo
  unsigned int *d_result;

  PREAMBLE(pbo,d_result);

  int iw = imageLocal.getWidth();
  int ih = imageLocal.getHeight();
  copyToPBO( imageLocal.getDataPointer(),d_result, iw, ih, _colormap );

  POSTAMBLE(pbo);

  check_gl("render_plain_image");

  tex_pbo();

  gl_end();
}

void DataWindow::render_velocity_image()
{


  gl_begin();
  unsigned int *d_result = 0;

  PREAMBLE(pbo,d_result);

  size_t iw = imageLocal.getWidth();
  size_t ih = imageLocal.getHeight();
  size_t dw = direction.getWidth();
  size_t dh = direction.getHeight();
  
  //std::cout << " direction [" << direction.min() << "," << direction.mean() << "," << direction.max() << "]" << std::endl;
  //std::cout << " speed [" << speed.min() << "," << speed.mean() << "," << speed.max() << "]" << std::endl;
  
#if 1
  copyToPBOColorCode( imageLocal.getDataPointer(),
					  direction.getDataPointer(),
					  speed.getDataPointer(),
					  d_result,
					  iw, ih, dw, dh, brighten, _have_flow_legend, is_image );
#else
  copyToPBO( imageLocal.getDataPointer(),d_result, iw, ih );
#endif

  check_gl("DataWindow::render_velocity_image");

  POSTAMBLE(pbo);

  tex_pbo();

  gl_end();

}

void DataWindow::render_feature_image()
{

  gl_begin();

  unsigned int *d_result = 0;

  PREAMBLE(pbo,d_result);

  
  int iw = imageLocal.getWidth();
  int ih = imageLocal.getHeight();
  int fw = featuresLocal.extent(1);
  int fh = featuresLocal.extent(0);
  int nf = featuresLocal.extent(2);
  int sfx = featuresLocal.stride(1);
  int sfy = featuresLocal.stride(0);
  int sff = featuresLocal.stride(2);
  //std::cout << str( boost::format( "features size: (%ix%i,%i) strides: (%i,%i,%i)" ) % fw % fh % nf % sfx % sfy % sff ) << std::endl;
  if ( hsv )
	{
	  copyToPBOHSV( imageLocal.data(),
					featuresLocal.data(),
					d_result,
					iw, ih,
					fw, sfx, fh, sfy, nf, sff );
	}
  else
	{
	  copyToPBOFeatures( imageLocal.data(),
						 featuresLocal.data(),
						 d_result,
						 iw, ih,
						 fw, sfx, fh, sfy, nf, sff,
						 saturate_entropy, brighten, maxindex );
	}
  check_gl("DataWindow::render_feature_image");

  POSTAMBLE(pbo);

  tex_pbo();    

  gl_end();

}

void DataWindow::tex_pbo()
{
  if ( pbo == 0 ) throw std::runtime_error("DataWindow::tex_pbo: no pbo!");
  if ( texid == 0 ) throw std::runtime_error("DataWindow::tex_pbo: no texture!");

  glEnable(GL_TEXTURE_2D);
  glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
  glBindTexture(GL_TEXTURE_2D, texid);
  glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, imageLocal.getWidth(), imageLocal.getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, 0);
  glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
  glBindTexture(GL_TEXTURE_2D, 0);
  glDisable(GL_TEXTURE_2D);

  check_gl("DataWindow::tex_pbo");

}

void DataWindow::draw_image()
{

  // load texture from pbo and activate
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, texid);

  glBegin(GL_QUADS);
  glVertex2f(0, 0); glTexCoord2f(0, 0);
  glVertex2f(0, 1); glTexCoord2f(1, 0);
  glVertex2f(1, 1); glTexCoord2f(1, 1);
  glVertex2f(1, 0); glTexCoord2f(0, 1);
  glEnd();

  glDisable(GL_TEXTURE_2D);

  check_gl("draw_image");
}

void DataWindow::draw_arrows()
{
  glBindBuffer(GL_ARRAY_BUFFER, vbo);
  glVertexPointer(4, GL_FLOAT, 0, 0);
  glEnableClientState(GL_VERTEX_ARRAY);
  {
    glColor3f(1.0, 0.0, 0.0);
	switch ( style )
	  {
	  case LINES: glDrawArrays( GL_LINES, 0, 6 * xpart.size() );break;
	  case TRIANGLES: glDrawArrays( GL_TRIANGLES, 0, 6 * xpart.size() );break;
	  case NONE: break;
	  }
  }
  glDisableClientState(GL_VERTEX_ARRAY);

  check_gl("DataWindow::draw_arrows");
}

void DataWindow::draw_markers()
{
  int w       = imageLocal.getWidth();
  int h       = imageLocal.getHeight();
  float width = 1; /// this is needed for lines thick enough to be seen
  float ratio = float(w)/h;
  float M  = _markers.size();
  float dx = _marker_size > 1 ? (_marker_size / 2.f) / h : (_marker_size / 2.f);
  float dy = _marker_size > 1 ? (_marker_size / 2.f) / w : (_marker_size / 2.f) *ratio;
  
  for (int i = M-1; i >= 0; --i)
  {
    marker& p = _markers[i];

    float x  = p.x >  1 ?   float(p.x)/w    :  p.x     ; /// flip the pixel
    float y  = p.y >  1 ?   1-float(p.y)/h  :  1-p.y   ; 
	float c  = p.c;
	float vx = p.vx > 1 ?     float(p.vx) / w  :   p.vx;
    float vy = p.vy > 1 ? 1 - float(p.vy) / h  :  -p.vy;
    
    glLineWidth( width );
    glPointSize( std::max(1.f, width-1.f) );
    glEnable(GL_POINT_SMOOTH);
    
    
    glPushAttrib(GL_CURRENT_BIT); 
    glBegin(GL_LINES);
     // glColor4f( (1-c/M)*0.5 + 0.5, (c/M)*0.5, 0, 0.3 );
	set_gl_color(c);
    glVertex2f( x - dx, y - dy );
    glVertex2f( x + dx, y + dy );
    glVertex2f( x + dx, y - dy );
    glVertex2f( x - dx, y + dy );
    glVertex2f( x     , y      );
    glVertex2f( x + vx, y + vy );
    glEnd();
    glDisable(GL_BLEND);
    glPopAttrib();
  }
  
}

void DataWindow::draw_ellipses()
{
  int w       = imageLocal.getWidth();
  int h       = imageLocal.getHeight();
  float width = h / 400 + 1; /// this is needed for lines thick enough to be seen
  float M     = _ellipses.size();
  int i = -1;

  BOOST_FOREACH( ellipse & e, _ellipses )
  {
	i++;
	float x  = e.x >  1 ?   float(e.x)/w  :  e.x   ; 
	float y  = e.y >  1 ?   1-float(e.y)/h  :  1-e.y   ;
	float c  = e.c <= 0 ?   float(i)/M      :  e.c     ; /// if no color index specified: take array position as colorindex
	
	glLineWidth( width );
	glPointSize( std::max(1.f, width-1.f) );
	glEnable(GL_POINT_SMOOTH);

	glPushAttrib(GL_CURRENT_BIT);
	glBegin(GL_LINE_LOOP);
	set_gl_color(c);

	if ( e.angle < M_PI / 1000 )  // unrotated ellipse or circle
	{
		float radius1 = e.r1 / float(w);
		float radius2 = e.r2 / float(h);

		for (int i=0; i<360; i++)
		{
		  float degInRad = i * M_PI / 180.0f;
		  float x_unrotated = std::cos(degInRad)*radius1;
		  float y_unrotated = std::sin(degInRad)*radius2;
		  glVertex2f( x + x_unrotated, y + y_unrotated );
		}
	}
	else  // rotated ellipse
	{
      float rotation_matrix[] = { std::cos(e.angle), -std::sin(e.angle), std::sin(e.angle), std::cos(e.angle) };
		
	  for (int i=0; i<360; i++)
	  {
	    float degInRad = i * M_PI / 180.0f;
	    float x_unrotated = std::cos(degInRad)*e.r1;
	    float y_unrotated = std::sin(degInRad)*e.r2;
	    float gl_x_rotated = (rotation_matrix[0] * x_unrotated + rotation_matrix[1] * y_unrotated) / w;
	    float gl_y_rotated = (rotation_matrix[2] * x_unrotated + rotation_matrix[3] * y_unrotated) / h;
	  
	    glVertex2f( x + gl_x_rotated, y + gl_y_rotated );
	  }
	  
	}
	glEnd();
	glPopAttrib();
  }

}

void DataWindow::set_gl_color( float c )
{
    float red=0.5, green=0.5, blue=0.5;
    // in first case it is assumed you want to map values in interval [0,1]
    // in second case it is assumed you want to mark objects numbered 0,1,2,3,... with different colors
    if ( (c>=0) && (c<=1) )
    {
        red = 1-c;
        green = c;
        blue = 0;
    }
    else
    {
        if ( _random_colormap )
        {
            red   = std::abs( std::sin( 10*c ) );
            green = std::abs( std::cos( 5*c ) );
            blue  = std::abs( std::sin( 20*c ) );
        }
        else
        {/*
            c = c % g_colormap.size();
            red   = g_colormap[c][0];
            green = g_colormap[c][1];
            blue  = g_colormap[c][2];
        */}
    }
    glColor4f( red, green, blue, _opacity );
}

void DataWindow::draw_trajectories()
{
  int w       = imageLocal.getWidth();
  int h       = imageLocal.getHeight();
  float width = h / 400 + 1; /// this is needed for lines thick enough to be seen
  float ratio = 1;//float(w)/h;
  float M     = _trajectories.size();
 
  glLineWidth( width   );
  glPointSize( std::max(1.f, width-1.f) );
  glEnable(GL_POINT_SMOOTH);


  for (int i = M-1; i >= 0; --i)
  {
    trajectory&  t  = _trajectories[i];

    
    glPushAttrib(GL_CURRENT_BIT); 
    glBegin(GL_LINE_STRIP);

    for ( int j = 0; j < t.points.size(); ++j )
    {

        float x  = t.points[j].x >  1 ?   float(t.points[j].x)/w    :  t.points[j].x     ; /// flip the pixel
        float y  = t.points[j].y >  1 ?   1-float(t.points[j].y)/h  :  1-t.points[j].y   ; 
        float c  = t.points[j].c <= 0 ?   float(i)/M      :  t.points[j].c     ; /// if no color index specified: take array position as colorindex
    
        set_gl_color( c );
        glVertex2f( x, y );
    }
    glEnd();
    glPopAttrib();
  }
  
}


void DataWindow::draw_boxes()
{
  int w       = imageLocal.getWidth();
  int h       = imageLocal.getHeight();
  float width = _line_width; /// this is needed for lines thick enough to be seen
  float ratio = 1;//float(w)/h;
  float M     = _boxes.size();
  float DX    = _box_size > 1 ? (_box_size / 2.f) / h : (_box_size / 2.f);
  float DY    = _box_size > 1 ? (_box_size / 2.f) / w : (_box_size / 2.f);
  
  for (int i = M-1; i >= 0; --i)
  {
    box&  p  = _boxes[i];

    float x  = p.x >  1 ?   float(p.x)/w    :  p.x     ; /// flip the pixel
    float y  = p.y >  1 ?   1-float(p.y)/h  :  1-p.y   ; 

    float c  = p.c <= 0 ?   float(i)/M      :  p.c     ; /// if no color index specified: take array position as colorindex
    float dx = p.w <= 0 ?   DX              :  p.w     ; /// default box size if no size specified
    float dy = p.h <= 0 ?   DY              :  p.h     ; /// default box size if no size specified
    float s  = p.s;
    
    dx       = dx  >  1 ?   (dx / 2.f) / w  :  (dx / 2.f); /// scale if greater than 1
    dy       = dy  >  1 ?   ((dy / 2.f) / h)  * ratio :  (dy / 2.f) * ratio; /// scale if greater than 1   	

	
    glLineWidth( width   );
    glPointSize( std::max(1.f, width-1.f) );
    glEnable(GL_POINT_SMOOTH);

    glPushAttrib(GL_CURRENT_BIT); 
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
    
    glBegin(GL_LINE_LOOP);
    
    
	set_gl_color(c);

    glVertex2f( x - dx, y - dy );
    glVertex2f( x - dx, y + dy );
    glVertex2f( x + dx, y + dy );
    glVertex2f( x + dx, y - dy );
    glEnd();

    glBegin(GL_POINTS);
    glVertex2f( x, y );
    glEnd();
#ifdef HAVE_GLUT
    //glut_render_string( x - dx, y - dy, w, h,  str( boost::format( "%1.2e" ) % c ) );
    if ( _render_score )
        glut_render_string( x - dx, y - dy, w, h,  str( boost::format( _render_score_pattern.c_str() ) % s ), _render_score_size );
#endif

    glDisable(GL_BLEND);
    glPopAttrib();
  }
  
}


/*  old code
void DataWindow::draw_markers()
{
  int w = imageLocal.getWidth();
  int h = imageLocal.getHeight();
  float M = _markers.size();
  //BOOST_FOREACH( const point2_t & point, _markers )
  for (int i = _markers.size()-1; i >= 0; --i)
  {
    point2_t & point = _markers[i];
    float x = point.first / w;
    float y = 1 - point.second / h; // flip the pixel
    float d = _marker_size / 2;
    glPushAttrib(GL_CURRENT_BIT); 
    glBegin(GL_LINE_LOOP);
      glColor4f( (1-i/M)*0.5 + 0.5, (i/M)*0.5, 0, 1 );
      glVertex2f( x - d, y - d );
      glVertex2f( x - d, y + d );
      glVertex2f( x + d, y + d );
      glVertex2f( x + d, y - d );
      //glVertex2f( x + d, y + d );
      //glVertex2f( x + d, y - d );
      //glVertex2f( x - d, y + d );
    glEnd();
    glPopAttrib();
  }
  
}
*/

void DataWindow::draw()
{
  bool have_image  = imageLocal.size() > 1;
  bool have_arrows = xpart.size() > 1;

  if ( have_image )
    draw_image();

  draw_markers();
  draw_boxes();
  draw_ellipses();
  draw_trajectories();

  if ( have_arrows && style != NONE )
	draw_arrows();

}

bool DataWindow::expose_event(GdkEventExpose *event)
{



  if ( updateEveryTime ) update();

  gl_begin();

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  check_gl("expose_event");

  draw();

  Glib::RefPtr<Gdk::GL::Window> glwindow = da.get_gl_window();

  if (glwindow->is_double_buffered ())
	  glwindow->swap_buffers ();
  else
	  glFlush ();

  check_gl("expose_event end");

  gl_end();

  return true;
}

void DataWindow::show(const GPUImage& im, const GPUVelField& vf, bool blocking) {
  DataWindow window( im.getWidth(), im.getHeight() );
  window.update( im, vf );
  if (blocking) {
	window.wait_until_closed();
  } else {
	process_all_pending_gtk_events();
	process_gtk_events(false);process_gtk_events(false);
	process_gtk_events(false);process_gtk_events(false);
  }
}
void DataWindow::show(const GPUImage& im, const GPUVelField& vf, const char* title, bool blocking) {
  DataWindow window( im.getWidth(), im.getHeight(), title );
  window.update( im, vf );
  if (blocking) {
	window.wait_until_closed();
  } else {
	process_all_pending_gtk_events();
	process_gtk_events(false);process_gtk_events(false);
	process_gtk_events(false);process_gtk_events(false);
  }
}
void DataWindow::show(const GPUVelField& vf, bool blocking) {
  GPUImage im(vf.getWidth(),vf.getHeight()); im.fill(0.5f); im[0][0]=0.0f;
  DataWindow window( im.getWidth(), im.getHeight() );
  window.update( im, vf );
  if (blocking) {
	window.wait_until_closed();
  } else {
	process_all_pending_gtk_events();
	process_gtk_events(false);process_gtk_events(false);
	process_gtk_events(false);process_gtk_events(false);
  }
}
void DataWindow::show(const GPUVelField& vf, const char* title, bool blocking) {
  GPUImage im(vf.getWidth(),vf.getHeight()); im.fill(0.5f); im[0][0]=0.0f;
  DataWindow window( im.getWidth(), im.getHeight(), title );
  window.update( im, vf );
  if (blocking) {
	window.wait_until_closed();
  } else {
	process_all_pending_gtk_events();
	process_gtk_events(false);process_gtk_events(false);
	process_gtk_events(false);process_gtk_events(false);
  }
}

void DataWindow::initGUI()
{

  updateEveryTime = false;

  vbo 	= 0;
  pbo 	= 0;
  texid 	= 0;

  //set_border_width(0);
  std::string titleL(title);
  titleL.append(" (%d x %d)");
  char titleLRes[1024];
  sprintf(titleLRes,titleL.c_str(),w,h);
  set_title(titleLRes);

  // window

  signal_window_state_event().connect(sigc::mem_fun(*this, &DataWindow::window_event));
  set_reallocate_redraws(true);

  // vBox

  vBox.set_reallocate_redraws(true);
  vBox.set_spacing(2);

  add(vBox);

  // gl-drawable

  Glib::RefPtr< Gdk::GL::Config > configSPtr
	= Gdk::GL::Config::create ( Gdk::GL::MODE_RGB |
								Gdk::GL::MODE_DEPTH |
								Gdk::GL::MODE_DOUBLE );
  da.set_gl_capability( configSPtr );
  float wL = w;
  float hL = h;
  while (wL<=500/2 && hL<=350/2) {
	wL*=2;
	hL*=2;
  }
  da.set_size_request(wL,hL);
  da.signal_configure_event().connect( sigc::mem_fun( *this, &DataWindow::config_event ) );
  da.signal_expose_event().connect( sigc::mem_fun(*this,&DataWindow::expose_event));

  vBox.pack_start(da,true,true,0);
  da.show();

  initStatusBar();
  vBox.pack_end(statusBar,false,false,0);
  statusBar.show();

  vBox.show();

  show_all();

  busy_wait(100);

}

void DataWindow::gl_end()
{
  Glib::RefPtr< Gdk::GL::Drawable > gldrawable = da.get_gl_drawable();
  gldrawable->gl_end();
}

void DataWindow::gl_begin()
{
  Glib::RefPtr< Gdk::GL::Context  > glcontext  = da.get_gl_context();
  Glib::RefPtr< Gdk::GL::Drawable > gldrawable = da.get_gl_drawable();
  gldrawable->make_current(glcontext);
  if (!gldrawable->gl_begin(glcontext))
	throw std::runtime_error("could not begin gl-drawing");
}

void DataWindow::initStatusBar() {
  //statusBar.set_default_direction()
  statusBar.push("Hallo Welt");
}

void DataWindow::repaint(bool directly) {
  if (directly) {
	da.get_window()->invalidate_rect(da.get_allocation(), false);
	da.get_window()->process_updates(false);
	while (Gtk::Main::events_pending())
	  process_gtk_events(false);
  } else {
	da.queue_draw();
	while (Gtk::Main::events_pending())
	  process_gtk_events(false);
  }
}

bool DataWindow::window_event(GdkEventWindowState* event) {
  switch(event->changed_mask)	{
  case GDK_WINDOW_STATE_WITHDRAWN : closing(); return true; break;
  case GDK_WINDOW_STATE_ICONIFIED : ;
  default:
	//printf("window event was not assigned to a callback yet\n");
  return false;
  }
}

void DataWindow::closing() {
  closed = true;
}


bool DataWindow::config_event(GdkEventConfigure *event) {


  gl_begin();

  const int w = da.get_allocation().get_width();
  const int h = da.get_allocation().get_height();

  { /* reshape */
	glViewport(0, 0, w, h);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
  }

  check_gl("config event");

  gl_end();

  update();

  return true;
}

void DataWindow::show(const GPUImage& im, bool blocking)
{
  DataWindow window( im.getWidth(), im.getHeight(), "show" );
  window.update( im );
  if (blocking) {
	window.wait_until_closed();
  } else {
	process_all_pending_gtk_events();
	process_gtk_events(false);process_gtk_events(false);
	process_gtk_events(false);process_gtk_events(false);
  }
}

void DataWindow::show(const GPUImage& im, const char* title, bool blocking) {
  DataWindow window( im.getWidth(), im.getHeight(), title );
  window.update( im );
  if (blocking) {
	window.wait_until_closed();
  } else {
	process_all_pending_gtk_events();
	process_gtk_events(false);process_gtk_events(false);
	process_gtk_events(false);process_gtk_events(false);
  }
}
