/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <Segmentation.h>
#include <SegmentationModule/RegularizationModule.h>
#include <SegmentationModule/BalloonModule.h>
#include <blitz/array.h>

/**  @example 2D_Initialization.C Mean curvature motion with different initializations
*/
int main(int argc, char* argv[])
{	
	using namespace blitz;
 	using namespace levelset;
 	using namespace segmentation;

	int nx = 300;
  	int ny = 300;
	TinyVector<int, 2> shape(nx, ny);
 
	// initialization of the segmentations objects
        FastMarchingInitializer<2> init;                       // Fast Marching Initializer
	LevelSet<2>                phi(shape, (double) 5.0);   // Level Set of band thickness = 2
	Segmentation<2,float>      segm(&phi, &init);          // Segmentation
 
	RegularizationModule<2>        regul(1.f, true);       // parameters : weight, mean_curv_motion
	segm.add_module(regul);

	// launch the evolution
	timeval _time;

	std::ostringstream name;
	name.str("");
	name << "mkdir -p results/init";
	system (name.str().c_str());
	Globals::initChrono(_time);

	// SPHERE

	std::cout << phi.shape() << std::endl;

	LevelSet<2>::Index center1(150, 150);
	phi.init_sphere(center1, 100, init);
	segm.segment(100, "results/init/mean_curvature_1_");

	phi.thickness() = 10;
	LevelSet<2,float>::Index center2(150, 100);
	phi.init_sphere(center2, 50, init, false);
	segm.segment(100, "results/init/mean_curvature_2_");

	// SMALL SPHERES

	phi.init_small_spheres(20, 50, init);
	segm.segment(100, "results/init/mean_curvature_3_");

	// RECTANGLE

	phi.thickness() = -1;
	LevelSet<2>::Index tl(150, 50);
	LevelSet<2>::Index br(250, 250);
	phi.init_rectangle(tl, br, init, false);
	segm.segment(100, "results/init/mean_curvature_4_");

	// BORDER

	phi.init_border(50, init, false);
	segm.segment(100, "results/init/mean_curvature_5_");

	// IMAGE

	Array<float, 2> im_init(shape);
	im_init = 0.0;
	for (Array<float, 2>::iterator ind=im_init.begin() ; ind!=im_init.end() ; ++ind)
		if(ind.position()(1) < nx/2) *ind = 55;
	phi.init_image(im_init, init, false);

	segm.segment(100, "results/init/mean_curvature_6_");

	Globals::printChrono("\n	=> Mean curvature : ", _time);
	Globals::initChrono(_time); 

        std::cout << "OK" << std::endl;

	return 0;
}
