/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef APPROXIMATORS_H
#define APPROXIMATORS_H

#include <blitz/array.h>

using namespace blitz;

namespace levelset {


	/**
	* @author Jerome Piovano
	* @brief derivatives using 3rd order WENO finite differences
	*/
	template<unsigned DIM, typename Pixel = float>
	class Weno3
	{
	public:	
		typedef          blitz::Array<Pixel, DIM>	   ImageType;
		typedef typename ImageType::T_index          Index;

		/// Forward finite difference in the axis 'axis'
		static Pixel forward(const ImageType & phi, const Index & ind, unsigned axis);
		/// Backward finite difference in the axis 'axis'
		static Pixel backward(const ImageType & phi, const Index & ind, unsigned axis);
	private:
		/// internal function
		static Pixel combine(const Pixel v1, const Pixel v2, const Pixel v3);
	};


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// static functions


	template<unsigned DIM, typename Pixel>
	Pixel Weno3<DIM, Pixel>::forward(const ImageType & phi, const Index & ind, unsigned axis)
	{
		int pos = ind(axis);
		int limit = phi.extent(axis)-1;
		Index ind_tmp = ind;

		ind_tmp(axis)  = (pos+2 > limit) ? limit : pos+2;
		const Pixel u3 = phi(ind_tmp);

		ind_tmp(axis)  = (pos+1 > limit) ? limit : pos+1;
		const Pixel u2 = phi(ind_tmp);

		ind_tmp(axis)  = pos;
		const Pixel u1 = phi(ind_tmp);

		ind_tmp(axis)  = (pos-1 < 0) ? 0 : pos-1;
		const Pixel u0 = phi(ind_tmp);

		return combine(u3-u2,u2-u1,u1-u0);
	}
	
	template<unsigned DIM, typename Pixel>
	Pixel Weno3<DIM, Pixel>::backward(const ImageType & phi, const Index & ind, unsigned axis) 
	{
		int pos   = ind(axis);
		int limit = phi.extent(axis)-1;
		Index ind_tmp = ind;

		ind_tmp(axis) =  (pos+1 > limit) ? limit : pos+1;
		const Pixel u3 = phi(ind_tmp);

		ind_tmp(axis) =  pos;
		const Pixel u2 = phi(ind_tmp);

		ind_tmp(axis) =  (pos-1 < 0) ? 0 : pos-1;
		const Pixel u1 = phi(ind_tmp);

		ind_tmp(axis) =  (pos-2 < 0) ? 0 : pos-2;
		const Pixel u0 = phi(ind_tmp);

		return combine(u1-u0,u2-u1,u3-u2);
	}

	template<unsigned DIM, typename Pixel>
	Pixel Weno3<DIM, Pixel>::combine(const Pixel v1, const Pixel v2, const Pixel v3)
	{
		// replaced Globals::EPS with EPS (because of #define needed in Globals.h
		Pixel s = ( EPS + (v2 - v1)*(v2 - v1) ) / ( EPS + (v3 - v2)*(v3 - v2) );
		s = 1 / (1 + 2*s*s);
		return (v2+v3 - s * (v1 - 2*v2 + v3)) / 2;
	}
}

#endif //APPROXIMATORS_H
