/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef FAST_MARCHING_H
#define FAST_MARCHING_H

#include <algorithm>
#include <limits>

#include <blitz/array.h>
#include <Globals.h>
#include <PriorityQueue.h>

using namespace blitz;

namespace levelset {
  
  /**
   * @author Jerome Piovano
   * @brief Compute the distance function from a set of seed points via the Fast Marching Method
   */
  template<unsigned DIM, typename Pixel = float, int sign = +1>
  class FastMarching : public blitz::Array<unsigned, DIM>
	{
	public:
	
	typedef 	 Array<unsigned, DIM>	             base;
	typedef typename base::T_index                   Index;
	typedef          Array<Pixel, DIM>	             DataType;
	
	typedef          PriorityQueue< Index >          CoordQueue;
	typedef typename CoordQueue::QueuePosition       QueuePosition;
	typedef          Array<QueuePosition, DIM>       ArrayQueuePosition;
	
	/// State of a point in the Fast Marching
	typedef enum { eAlive=0, eTrial=1, eFar=2, eForbidden=3 } eState;
	
	/** @name Constructors - Destructor - Init */
	//@{
	
	/// Default constructor
	FastMarching();
	/// Construct the Fast Marching from an image object
	FastMarching(DataType *_phi);
	/// Destructor
	virtual ~FastMarching();
	
	/// Initialization of the datas
	void init(DataType & _phi);
	
	//@}
	/** @name Access/Modif the Alive-Trial-Forbidden points */
	//@{
	
	/// Number of points in the narrow band
	unsigned nb_alive_points() const;
	/// Get the point of indice i in the narrow band
	const Index& get_alive_point(unsigned i);
	/// Add a alive point (seed point)
	void add_alive_point(const Index& ind);
	/// Add a trial point 
	void add_trial_point(const Index& ind);
	/// Add a forbidden point
	void add_forbidden_point(const Index& ind);
	/// Initialization of the trial points from the alive point
	void init_trial_from_alive();
	/// Remove all the Alive-Trial-Forbidden points, set all points to Far
	void clear();
	
	//@}
	/** @name Launch the fast marching */
	//@{
	
	/// Launch the fast marching with no stop condition (computation on the whole image)
	void run();
	/// Launch the fast marching with stop when the value 'limit' is reached
	void run(Pixel _limit);
	
	//@}
	/** @name Member functions*/
	//@{
	
	protected:
	
	/// Get the value of a point from its coordinate, with test on its state and with the correct sign
	Pixel _get_value(const Index& ind) const;
	/// Compute the value of a point from its neighbors
	virtual Pixel _update_value(const Index& ind) const = 0;
	/// Solve a second degree trinom
	bool _solve_trinome(const Pixel a, const Pixel b, const Pixel c, Pixel &sol_max, Pixel &sol_min) const;
	
	private:
	/// Update a point 
	void __update_point(const Index& ind);
	
	//@}
	/** @name Member datas*/
	//@{
	
	Index *              m_alive_tab;            ///< Tab of the coordinates of the alive points
	unsigned             m_nb_alive;             ///< Number of alive points
	CoordQueue           m_trial_queue;          ///< Queue of the trial points
	ArrayQueuePosition   m_image_queue_position; ///< Image in which each pixel contains the position of the element in the trial queue
	protected:
	Pixel                m_big;                  ///< Maximal value of the type Pixel
	DataType *           m_phi;                  ///< Image of the datas
	
	};
  
  
  /**
   * @author Jerome Piovano
   *
   * @brief Compute the Eikonal Equation from a set of seeds points via the Fast Marching Method
   *
   *    By default , solves the eikonal equation |grad(phi)| = 1
   *    You have the possibility to give an image metric, where each pixel will contain the local metric of the distance function you want to compute
   *    It will then solve |grad(phi)| = 1/F
   *    You can "reset" the metric with set_metric(NULL)
   */
  template <unsigned DIM, typename Pixel, int sign = +1>
	class Eikonal : public FastMarching<DIM, Pixel, sign> 
	{
	public:
	
	typedef FastMarching<DIM, Pixel, sign> base;
	typedef typename base::DataType        DataType;
	typedef typename base::Index           Index;
	
	/// Default constructor
	Eikonal();
	/// Construct the Fast Marching from an image object
	Eikonal(DataType * _phi);
	/// Destructor
	virtual ~Eikonal() {}
	
	/// Set the optional metric of the distance function
	void set_metric(DataType * _metric);
	
	protected:
	/// Update a point
	virtual Pixel _update_value(const Index& ind) const;
	/// Solve of the trinom
	Pixel _solve(const Index& ind, Pixel val[DIM]) const;
	
	private:
	
	DataType * m_metric;  ///< metric of the surface
	};
  
  
  /**  @example 2D_FastMarching.C 2D Fast Marching */
  /**  @example 3D_FastMarching.C 3D Fast Marching */
  /**  @example 2D_MinimalPath.C Minimal path between 2 points in an image + back tracking */
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors - Destructor - Init
  
  
  template<unsigned DIM, typename Pixel, int sign>
	FastMarching<DIM, Pixel, sign>::FastMarching()
	: base()
	, m_alive_tab(NULL)
	, m_nb_alive(0)
	, m_trial_queue()
	, m_image_queue_position()
	, m_big(std::numeric_limits<Pixel>::max())
	, m_phi(NULL)
	{ }
  
  template<unsigned DIM, typename Pixel, int sign>
	FastMarching<DIM, Pixel, sign>::FastMarching(DataType *_phi)
	: base(_phi->shape(), ColumnMajorArray<DIM>()), m_phi(_phi), m_nb_alive(0), m_trial_queue(), m_image_queue_position(_phi->shape()), m_big(std::numeric_limits<Pixel>::max()) 
	{
	  m_alive_tab    = new Index[base::size()];
	  base::operator = (eFar);
	  m_image_queue_position = QueuePosition();
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	FastMarching<DIM, Pixel, sign>::~FastMarching() 
	{
	  if (m_alive_tab != NULL) delete[] m_alive_tab;
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::init(DataType & _phi) 
	{
	  // if not the same shape, resize the datas
	  if (!Globals::same_shape<DIM>(_phi.shape(), base::shape()) ){
		
		base::resize(_phi.shape());
		m_image_queue_position.resize(_phi.shape());
		
		if (m_alive_tab != NULL) delete[] m_alive_tab;
		m_alive_tab = new Index[base::size()];
	  }
	  clear();
	  m_phi = &_phi;
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  Access/Modif the Alive-Trial-Forbidden points
  
  
  template<unsigned DIM, typename Pixel, int sign>
	unsigned FastMarching<DIM, Pixel, sign>::nb_alive_points() const
	{
	  return m_nb_alive;
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	const typename FastMarching<DIM, Pixel, sign>::Index& FastMarching<DIM, Pixel, sign>::get_alive_point(unsigned i)
	{
	  return m_alive_tab[i];
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::add_alive_point(const Index& ind)
	{
	  if (base::operator()(ind) == eFar) {
		base::operator()(ind) = eAlive;
		m_alive_tab[m_nb_alive++] = ind;
	  }
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::add_trial_point(const Index& ind)
	{
	  if (base::operator()(ind) == eFar) {
		base::operator()(ind) = eTrial;
		m_image_queue_position(ind) = m_trial_queue.push(ind, sign*(*m_phi)(ind));
	  }
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::add_forbidden_point(const Index& ind) 
	{
	  if (base::operator()(ind) == eFar) {
		base::operator()(ind) = eForbidden;
	  }
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::init_trial_from_alive()
	{
	  // we first insert the alive point with the lowest value in the trial point heap 
	  // ( to assure that the points inserted will be > to the lowest point in the heap )
	  if (nb_alive_points() > 0){
		Index min_ind = get_alive_point(0);
		Pixel min_val = sign * (*m_phi)(min_ind);
		
		for (unsigned int i = 0 ; i < nb_alive_points() ; i++)
		  {
			const Index& ind = get_alive_point(i);
			Pixel        val = sign * (*m_phi)(ind);
			if (val < min_val){
			  min_ind = ind;
			  min_val = val;
			}	
		  }
		
		m_trial_queue.push(min_ind, min_val);
	  }
	  
	  // we iterate on the list of alive points, and we add their neighborings point to the trial point heap.
	  for (unsigned int j = 0 ; j < nb_alive_points() ; j++)
		{
		  const Index& ind = get_alive_point(j);
		  
		  for ( int i = 0 ; i < int(DIM) ; i++)
			{
			  Index inf = ind;
			  Index sup = ind;
			  inf(i)--;
			  sup(i)++;
			  
			  if ( sup(i) < m_phi->extent(i) ) __update_point(sup);
			  if ( inf(i) >= 0 )               __update_point(inf);
			}
		}
	  
	  // we remove the first alive point inserted
	  if (nb_alive_points() > 0)
		m_trial_queue.pop();
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::clear() 
	{
	  m_nb_alive = 0;
	  base::operator=(eFar);
	  m_trial_queue.clear();
	  m_image_queue_position = QueuePosition();
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  Launch the fast marching
  
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::run() 
	{
	  run( (sign>0) ? m_big : -m_big );
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::run(Pixel _limit) 
	{
	  // while there are still trial points in the heap
	  while(!m_trial_queue.empty()) {
		
		// we get the trial point with the weakest value
		Index ind = m_trial_queue.top();
		
		// If we are at the limit
		if (sign*(*m_phi)(ind)>=sign*_limit) {
		  // we clear the heap, and mark all its point at 'Far'
		  while (!m_trial_queue.empty()) {
			ind = m_trial_queue.top();
			m_trial_queue.pop();
			base::operator()(ind) = eFar;
			m_image_queue_position(ind)   = QueuePosition();
			(*m_phi)(ind)         = _limit;
		  }
		  return;
		}
		
		// we look for the neighborings point of this point, and we update their values
		for (int i=0 ; i<int(DIM) ; i++){
		  Index inf = ind;
		  Index sup = ind;
		  inf(i)--;
		  sup(i)++;
		  
		  if (sup(i)<m_phi->extent(i)) __update_point(sup);
		  if (inf(i)>=0)               __update_point(inf);
		}
		
		// we remove this point from the heap, and add it to the alive points list
		m_trial_queue.pop();	
		m_image_queue_position(ind) = QueuePosition();
		base::operator()(ind) = eAlive;
		m_alive_tab[m_nb_alive++] = ind;
	  }
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  member functions
  
  
  template<unsigned DIM, typename Pixel, int sign>
	Pixel FastMarching<DIM, Pixel, sign>::_get_value(const Index& ind) const 
	{
	  if (base::operator()(ind) <= eTrial)
		return sign*(*m_phi)(ind);
	  return m_big;
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	bool FastMarching<DIM, Pixel, sign>::_solve_trinome(const Pixel a, const Pixel b, const Pixel c, Pixel &sol_max, Pixel &sol_min) const 
	{
	  const Pixel delta = b*b - 4*a*c;
	  if (delta < 0) return false;
	  const Pixel sqrtDelta = std::sqrt(delta);
	  sol_max = (- b + sqrtDelta) / a / 2;
	  sol_min = (- b - sqrtDelta) / a / 2;
	  return true;
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	void FastMarching<DIM, Pixel, sign>::__update_point(const Index& ind) 
	{
	  const unsigned st = base::operator()(ind);
	  if (st == eFar) {
		// if it was a 'Far' point, we set it to 'Trial', we compute its value and we add it to the heap
		base::operator()(ind) = eTrial;
		const Pixel val = _update_value(ind);
		(*m_phi)(ind) = sign*val;
		m_image_queue_position(ind) = m_trial_queue.push(ind, val);
	  }
	  else if (st == eTrial) {
		// if it was a 'Trial' point, we compute its new value
		const Pixel val = _update_value(ind);
		// if its value is smaller, we accept it
		if (val<sign*(*m_phi)(ind)) {
		  // specify the element in the queue, position in the priority queue, and new priority.
		  m_image_queue_position(ind) = m_trial_queue.increase_priority(ind, m_image_queue_position(ind), val);
		  (*m_phi)(ind) = sign*val;
		}
	  }
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Eikonal Equation
  
  
  template<unsigned DIM, typename Pixel, int sign>
	Eikonal<DIM, Pixel, sign>::Eikonal()
	: base(), m_metric(NULL)
	{ }
  
  template<unsigned DIM, typename Pixel, int sign>
	Eikonal<DIM, Pixel, sign>::Eikonal(DataType *_phi)
	: base(_phi), m_metric(NULL)
	{ }
  
  template<unsigned DIM, typename Pixel, int sign>
	void Eikonal<DIM, Pixel, sign>::set_metric(DataType * _metric)
	{
	  m_metric = _metric;
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	Pixel Eikonal<DIM, Pixel, sign>::_update_value(const Index& ind) const 
	{
	  Pixel val[DIM];
	  for ( int i = 0 ; i < int(DIM); i++ )
		{
		  Index inf = ind;
		  Index sup = ind;
		  inf(i)--;
		  sup(i)++;
		  
		  if (sup(i)>=base::extent(i))  val[i] = base::_get_value(inf);
		  else if (inf(i)<0)            val[i] = base::_get_value(sup);
		  else val[i] = std::min(base::_get_value(inf), base::_get_value(sup));
		}
	  
	  return _solve(ind,val);
	}
  
  template<unsigned DIM, typename Pixel, int sign>
	Pixel Eikonal<DIM, Pixel, sign>::_solve(const Index& ind, Pixel val[DIM]) const
	{
	  // we reorder value to have val[DIM-1] max
	  std::sort(val, val+DIM);
	  
	  // we initialise the datas of the trinom to solve
	  Pixel a[DIM];
	  Pixel b[DIM];
	  Pixel c[DIM];
	  
	  a[0] = 1;
	  b[0] = -2*val[0];
	  c[0] = val[0]*val[0] - ((m_metric == NULL)?1:(*m_metric)(ind) * (*m_metric)(ind));
	  
	  for ( int i = 1; i < int(DIM); i++ )
		{
		  a[i] = a[i-1] + 1;
		  b[i] = b[i-1] - 2*val[i];
		  c[i] = c[i-1] + val[i]*val[i];
		}
	  
	  // we test the differents cases
	  Pixel sol_max, sol_min;
	  for (int i=DIM-1 ; i>0 ; i--)
		  // replaced Globals::EPS with EPS (because of #define needed in Globals.h
		if ((val[i]<base::m_big) && _solve_trinome(a[i], b[i], c[i], sol_max, sol_min) && sol_max+EPS>=val[i])
		  return sol_max;
	  
	  return val[0] + ((m_metric == NULL)? 1 : (*m_metric)(ind));
	}
}


#endif // FAST_MARCHING_H
