/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef FLUXES_H
#define FLUXES_H

#include <Globals.h>
#include <Approximators.h>

#include <blitz/array.h>

using namespace blitz;

namespace levelset {


	/**
	* @author Jerome Piovano
	* @brief  Fluxes computation
	*/
	template<unsigned DIM, class Approximator, typename Pixel = float>
		class Fluxes
		{
		public:
			typedef          blitz::Array<Pixel, DIM>	   ImageType;
			typedef typename ImageType::T_index        Index;

			/** @name static functions */
			//@{

			/// Compute the reinitialization equation \f$- sign(\phi)\left( \|\nabla \phi\| - 1 \right)\f$ by finite differences.
			static Pixel reinitialization(const ImageType & phi, const Index & ind);
			/// Compute the advection equation \f$- \langle \nabla u . v \rangle\f$ by finite differences.
			static Pixel advection(const ImageType & u, const Index & ind, const Pixel v[]);

			//@}
		};


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// static functions


	template <unsigned DIM, class Approximator, typename Pixel>
		Pixel Fluxes<DIM, Approximator, Pixel>::reinitialization(const ImageType & phi, const Index & ind)
	{
		const Pixel p = phi(ind);
		const Pixel s = p / sqrt( p*p + DELTA*DELTA );
		Pixel delta = 0;

		if (s > 0) {
		  for ( int i = 0 ; i < int(DIM) ; i++ )
			{
			  const Pixel pf = Approximator::forward(phi,ind,i);
			  const Pixel pb = Approximator::backward(phi,ind,i);
			  if (pf < 0) delta += pf*pf;
			  if (pb > 0) delta += pb*pb;
			}
		}
		else {
		  for (int i = 0 ; i < int(DIM) ; i++)
			{
			  const Pixel pf = Approximator::forward(phi,ind,i);
			  const Pixel pb = Approximator::backward(phi,ind,i);
			  if (pf > 0) delta += pf*pf;
			  if (pb < 0) delta += pb*pb;
			}
		}

		return - s * (sqrt(delta) - 1);
	}

	template <unsigned DIM, class Approximator, typename Pixel>
		Pixel Fluxes<DIM, Approximator, Pixel>::advection(const ImageType & u, const Index & ind, const Pixel v[])
	{
		Pixel result = 0;

		for (int i=0 ; i<DIM ; i++)
			result += v[i-1] * ( (v[i-1]>0) ? Approximator::backward(u,ind,i) : Approximator::forward(u,ind,i) );

		return -result;
	}
}

#endif // FLUXES_H
