/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/


#ifndef GLOBALS_H
#define GLOBALS_H


#include <fstream>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <vector>
#include <string.h>

#ifdef HAVE_CIMG
#include <CImg.h> // In/Out Purpose Only and Deriche Filter
#endif
#include <blitz/array.h>

#ifndef _WIN32
#include <sys/time.h>
#endif

using std::ostream;
using std::vector;
using namespace blitz;

/// Save results each SAVE_EVOLUTION iteration
#define SAVE_EVOLUTION 10

/// Print infos about execution time
// #define PRINT_TIME
#define DIV_SQRT_2PI (1.0f/2.506628274631f)
#define LOG_2PI      (0.79817987f)
#define DELTA        (1.0f)
#define EPS          (1e-6f)

#define M_PI 3.1415926535897931

	
namespace levelset {
		
	/**
	* @author Jerome Piovano
	* @brief set of useful constant and static functions
	*/
	class Globals
	{ 
	public:

		/** @name Constants */
		//@{
 

		//@}
		/** @name Chrono functions */
		//@{


		//@}
		/** @name Scalar functions */
		//@{

		/// gaussian PDE
		template<typename T>
		static float gauss_density(float mean, float var, T val) {
			return (DIV_SQRT_2PI/(var+EPS))*exp(-0.5f*(val-mean)*(val-mean)/ (var+EPS));
		}
					
		/// log likelihood of gaussian PDE
		template<typename T>
		static float log_ll(float mean, float var, T val) {
			return -0.5f*LOG_2PI - log(var+EPS) - 0.5f*sqr(val-mean)/ (var+EPS);
		}
					
		/// regularized dirac
		static float delta(float v, float epsilon=1.0f) {
			if(v> epsilon) return 0;
			if(v<-epsilon) return 0;
			return float(1+cos(float(M_PI)*v/epsilon))/epsilon/2.0;
		}
					
		/// regularized heavyside
		static float heaviside(float v, float epsilon=1.0f){
			if(v>epsilon) return 1;
			if(v<-epsilon) return 0;
			return float(0.5f*(1+v/epsilon+(1/M_PI)*sin(M_PI*v/epsilon)));
		}

		//@}
		/** @name UGLY Blitz & CImg & Image functions & Misc */
		//@{

		/// true if same shape
		template<unsigned DIM>
		static bool same_shape(const TinyVector<int, DIM>& s1, const TinyVector<int, DIM>& s2)
		{
		  for (int i = 0; i < int(DIM) ; i++)
				if (s1(i) != s2(i)) return false;
			return true;
		}

		/// TODO, remove Cimg and write a In/Out image blitz module. (starting by copying/pasting CImg code) 

#ifdef HAVE_CIMG

		/// save an image
		template<typename Pixel>
		static void save(Array<Pixel, 2> image, std::string filename)
		{

			cimg_library::CImg<Pixel> cimage(image.data(), image.extent(0), image.extent(1), 1, 1, true);
			cimage.save(filename.c_str());

			//cimg_library::CImg<Pixel> cimage(image.extent(0), image.extent(1), 1, 1);
			//for (int x=0;  x<cimage.width ; x++)
			//	for (int y=0;  y<cimage.height ; y++)
			//		cimage(x,y) = image(x,y);
			//cimage.save(filename.c_str());
		}

		/// load an image
		template<typename Pixel>
		static void load(Array<Pixel, 2> & image, std::string filename)
		{
			cimg_library::CImg<Pixel> cimage(filename.c_str());
			image.resize(cimage.width, cimage.height);
			for (int x=0;  x<cimage.width ; x++)
				for (int y=0;  y<cimage.height ; y++)
					image(x,y) = cimage(x,y);
		}
/*
		/// blur an image
		template<typename Pixel>
		static void blur(Array<Pixel, 1> image, int sigma)
		{
			cimg_library::CImg<Pixel> cimage(image.data(), image.extent(0), 1, 1, 1, true);
			cimage.blur(sigma);

			//cimg_library::CImg<Pixel> cimage(image.extent(0), image.extent(1), 1, 1);
			//for (int x=0;  x<cimage.width ; x++)
			//	for (int y=0;  y<cimage.height ; y++)
			//		cimage(x,y) = image(x,y);
			//cimage.blur(sigma);
		}

		template<typename Pixel>
		static void blur(Array<Pixel, 2> image, int sigma)
		{
			cimg_library::CImg<Pixel> cimage(image.data(), image.extent(0), image.extent(1), 1, 1, true);
			cimage.blur(sigma);
		}
*/


		template<typename Pixel>
		  static void blur(Array<Pixel, 1> image, int sigma)
		  {
			cimg_library::CImg<Pixel> cimage(image.data(), image.extent(0), 1, 1, 1, true);
			cimage.blur(sigma);
			
			//cimg_library::CImg<Pixel> cimage(image.extent(0), image.extent(1), 1, 1);
			//for (int x=0;  x<cimage.width ; x++)
			//      for (int y=0;  y<cimage.height ; y++)
			//              cimage(x,y) = image(x,y);
			//cimage.blur(sigma);
		  }
		
		template<typename Pixel>
		  static void blur(Array<Pixel, 2> image, int sigma)
		  {
			cimg_library::CImg<Pixel> cimage(image.data(), image.extent(0), image.extent(1), 1, 1, true);
                        cimage.blur(sigma);
		  }

#endif //HAVE_CIMG
			
		/// Compute the min, max, and mean of an image
		template<unsigned DIM, typename Pixel>
		static void stats(const Array<Pixel, DIM>& image, Pixel& min, Pixel& max, float& mean, float* variance=NULL)
		{
			typename Array<Pixel, DIM>::const_iterator ind=image.begin();
			min    = *ind;
			max    = *ind;
			mean   = 0;
						
			while (ind != image.end()){
				if (*ind > max) max = *ind;
				if (*ind < min) min = *ind;
				mean += *ind;
				ind++;
			}
			
			mean /= image.size();
			
			if (variance != NULL){
				for (ind=image.begin() ; ind!=image.end() ; ind++)
				*variance += ( *ind - mean )*( *ind - mean );
				*variance /= image.size();

				*variance = sqrt(*variance);
			}
		}
		
 		/// Normalize an image between min and max.
		template<unsigned DIM, typename Pixel>
		static void normalize(Array<Pixel, DIM>& image, Pixel min, Pixel max)
		{
			Pixel _min, _max, _mean;
			Globals::stats<DIM, Pixel>(image, _min, _max, _mean);
		
			for (typename Array<Pixel, DIM>::iterator ind=image.begin() ; ind!=image.end() ; ++ind )
				*ind =  (Pixel) ((*ind-_min)/(_max-_min)*(max-min)+min);
		}
		
		/// cut an image, between min and max
		template<unsigned DIM, typename Pixel>
		static void cut(Array<Pixel, DIM>& image, Pixel min, Pixel max)
		{
			for (typename Array<Pixel, DIM>::iterator ind=image.begin() ; ind!=image.end() ; ++ind )
				*ind =  (*ind < min)? min : ( (*ind > max)? max : *ind );
		}
			
		//@}
	};
}

#endif // GLOBALS_H
