/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef INITIALIZER_H
#define INITIALIZER_H

#include <LevelSet.h>
#include <FastMarching.h>
#include <Schemes.h>
#include <Globals.h>

namespace levelset {

	template <unsigned DIM, typename Pixel> class LevelSet;
	template <unsigned DIM, typename Pixel> class Schemes;

	/**
	* @author Jerome Piovano
	* @brief (Re)compute the distance map of a Level-Set \ Binary-Image
	*/
	template<unsigned DIM, typename Pixel = float>
	class Initializer
	{
	public:

		typedef  LevelSet<DIM, Pixel>	   LevelSetType;

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor
		Initializer(){}
		/// Destructor
		virtual ~Initializer(){}

		//@}
		/** @name Member Functions */
		//@{

		/// Initiatlization of the band of a level set having no narrow band
		virtual void init_band(LevelSetType & phi) = 0 ;
		/// Reinitiatlization of the band of a level set
		virtual void reinit_band(LevelSetType & phi) = 0 ;
		//@}
	};

	/**
	* @author Jerome Piovano
	* @brief Initilalizer using the Fast marching method
	*/
	template<unsigned DIM, typename Pixel = float>
	class FastMarchingInitializer : public Initializer<DIM, Pixel>
	{
	public:

		typedef           Initializer<DIM, Pixel>       base;
		typedef typename  base::LevelSetType            LevelSetType;
		typedef typename  LevelSetType::Index           Index;
		typedef           Schemes<DIM,Pixel>            SchemesType;

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor
		FastMarchingInitializer(int   _iter = 2,   /// # of iterations of the reinitialization PDE
					float _dt = 0.4    /// time step of the reinitialization PDE
				       );
		/// Destructor
		virtual ~FastMarchingInitializer();

		//@}
		/** @name Member Functions */
		//@{

		/// Initiatlization of the band of a level set having no narrow band
		void init_band(LevelSetType & phi);
		/// Reinitiatlization of the band  of a level set
		void reinit_band(LevelSetType & phi);
		/// # of iterations of the reinitialization PDE
		int& iter();
		/// time step of the reinitialization PDE
		float& dt();

	protected:

		/// Add an alive point to the Fast Marching
		void _add_alive_point(LevelSetType & phi, const Index & ind);
		/// Reinitialization by Fast Marching
		void _reinit_fast_marching(LevelSetType & phi);
	private:

		//@}
		/** @name Data Members */
		//@{

		int                            m_iter;     ///< # of iterations of the reinitialization PDE
		float                          m_dt;       ///< time step of the reinitialization PDE
		Eikonal<DIM, Pixel, +1>      * m_reinit1;  ///< fast marching for the positive side of the band
		Eikonal<DIM, Pixel, -1>      * m_reinit2;  ///< fast marching for the negative side of the band

		//@}
	};


	/**  @example 2D_DistanceMap.C 2D distance map from a binary image */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// FastMarching Initializer 


	template<unsigned DIM, typename Pixel>
	FastMarchingInitializer<DIM, Pixel>::FastMarchingInitializer(int _iter, float _dt)
		: base(), m_iter(_iter), m_dt(_dt)
	{
		m_reinit1 = new Eikonal<DIM, Pixel, +1>();
		m_reinit2 = new Eikonal<DIM, Pixel, -1>();
	}

	template<unsigned DIM, typename Pixel>
	FastMarchingInitializer<DIM, Pixel>::~FastMarchingInitializer()
	{
		delete m_reinit1;
		delete m_reinit2;
	}

	template<unsigned DIM, typename Pixel>
	void FastMarchingInitializer<DIM, Pixel>::init_band(LevelSetType & phi)
	{
		// Initialization of the fast marching datas
		m_reinit1->init(phi);
		m_reinit2->init(phi);

		// clear the band
		phi.clear_band();

		#ifdef PRINT_TIME
		timeval _time;
		Globals::initChrono(_time);
		#endif

		// we add the points neighboring the zero level set to the alive points
		int inf_coords[DIM];
		int sup_coords[DIM];
		for (typename LevelSetType::Iterator ind=phi.begin() ; ind!=phi.end() ; ++ind){

			const Pixel _phi = *ind;
			bool  add        = false;

			// find the coordinates neighboring the point 'ind'
			SchemesType::find_neighbours(phi, ind.position(), inf_coords, sup_coords);

			//for each point 'inf' and 'sup' neighboring 'ind' ...
			for ( int i = 0 ; i < int(DIM) ; i++ )
			  {
				Index inf = ind.position();
				Index sup = ind.position();
				inf(i) = inf_coords[i];
				sup(i) = sup_coords[i];
				if (_phi * phi(inf) <= 0) { add = true; _add_alive_point(phi,inf); }
				if (_phi * phi(sup) <= 0) { add = true; _add_alive_point(phi,sup); }
			  }

			if (add) { _add_alive_point(phi,ind.position());}
		}

		#ifdef PRINT_TIME
		Globals::printChrono("		- points neighboring the zero level set : ", _time);
		Globals::initChrono(_time);
		#endif

		// We add to the band the points alive (for the reinit EDP which is done first).
		// we set the value of the others points to '1' ( border condition ) (2 if alive+trial too)
		for (typename LevelSetType::Iterator ind=phi.begin() ; ind!=phi.end() ; ++ind){
			const Pixel _phi = *ind;
			if (_phi>0) {
				if ((*m_reinit1)(ind.position()) == 2)  *ind =  1;
				else phi.add_point(ind.position());
			}
			else {
				if ((*m_reinit2)(ind.position()) == 2) *ind = -1;
				else phi.add_point(ind.position());
			}
		}

		#ifdef PRINT_TIME
		Globals::printChrono("		- add to the band the points alive: ", _time);
		Globals::initChrono(_time);
		#endif

		// reinitialization PDE (2*iter because first initialisation)
		phi.reinit_PDE(m_iter,m_dt);

		#ifdef PRINT_TIME
		Globals::printChrono("		- reinitialization PDE: ", _time);
		Globals::initChrono(_time);
		#endif

		// init of the trial points from the alive points
		m_reinit1->init_trial_from_alive();
		m_reinit2->init_trial_from_alive();

		#ifdef PRINT_TIME
		Globals::printChrono("		- init of the trial points from the alive points : ", _time);
		Globals::initChrono(_time);
		#endif

		// Fast marching
		_reinit_fast_marching(phi);

		#ifdef PRINT_TIME
		Globals::printChrono("		- FastMarching : ", _time);
		Globals::initChrono(_time);
		#endif

		// we put the other points at the max value
		for (typename LevelSetType::Iterator ind=phi.begin() ; ind!=phi.end() ; ++ind ){
			const Pixel _phi = *ind;
			if      (_phi>0 && (*m_reinit1)(ind.position())==2) *ind =  phi.thickness();
			else if (_phi<0 && (*m_reinit2)(ind.position())==2) *ind = -phi.thickness();
		}

		#ifdef PRINT_TIME
		Globals::printChrono("		- put the other points at the max value : ", _time);
		#endif

	}

	template<unsigned DIM, typename Pixel>
	void FastMarchingInitializer<DIM, Pixel>::reinit_band(LevelSetType & phi)
	{

		#ifdef PRINT_TIME
		timeval _time;
		Globals::initChrono(_time);
		#endif

		// reinitialization EDP 
		phi.reinit_PDE(m_iter,m_dt);

		// Initialization of the fast marching datas
		m_reinit1->init(phi);
		m_reinit2->init(phi);

		#ifdef PRINT_TIME
		Globals::printChrono("		- reinitialization PDE : ", _time);
		Globals::initChrono(_time);
		#endif

		// we add the points neighboring the zero level set to the alive points
		int inf_coords[DIM];
		int sup_coords[DIM];
		for ( int i = 0 ; i < phi.nb_points() ; i++ )
		  {

			const Index& ind = phi.get_point(i);
			const Pixel _phi = phi(ind);
			bool  add        = false;

			// find the coordinates neighboring the point 'ind'
			SchemesType::find_neighbours(phi, ind, inf_coords, sup_coords);

			//for each point 'inf' and 'sup' neighboring 'ind' ...
			for ( int i = 0; i < int(DIM); i++ )
			  {
				Index inf = ind;
				Index sup = ind;
				inf(i) = inf_coords[i];
				sup(i) = sup_coords[i];
				if (_phi * phi(inf) <= 0) { add = true; _add_alive_point(phi,inf); }
				if (_phi * phi(sup) <= 0) { add = true; _add_alive_point(phi,sup); }
			}

			if (add) { _add_alive_point(phi,ind);}
		}

		#ifdef PRINT_TIME
		Globals::printChrono("		- points neighboring the zero level set : ", _time);
		Globals::initChrono(_time);
		#endif

		// trial points are neighbor of alive points
		m_reinit1->init_trial_from_alive();
		m_reinit2->init_trial_from_alive();

		#ifdef PRINT_TIME
		Globals::printChrono("		- init of the trial points from the alive points : ", _time);
		Globals::initChrono(_time);
		#endif

		// we remove from the band the points that are not trial nor alive
		// we put their value at max
		int i = 0;
		while (i<phi.nb_points()){
			const Index& ind = phi.get_point(i);
			const Pixel _phi = phi(ind);

			if (_phi>0 && (*m_reinit1)(ind) == 2){
				phi(ind) = phi.thickness();
				phi.remove_point(ind);
			}
			else if  (_phi<=0 && (*m_reinit2)(ind) == 2) {
				phi(ind) = -phi.thickness();
				phi.remove_point(ind);
			}
			else i++;
		}

		#ifdef PRINT_TIME
		Globals::printChrono("		- remove from the band the points not alive or trial : ", _time);
		#endif

		// Fast marching
		_reinit_fast_marching(phi);

		#ifdef PRINT_TIME
		Globals::printChrono("		- FastMarching : ", _time);
		#endif
	}

	template<unsigned DIM, typename Pixel>
	int& FastMarchingInitializer<DIM, Pixel>::iter()
	{
		return m_iter;
	}

	template<unsigned DIM, typename Pixel>
	float& FastMarchingInitializer<DIM, Pixel>::dt()
	{
		return m_dt;
	}

	template<unsigned DIM, typename Pixel>
	void FastMarchingInitializer<DIM, Pixel>::_add_alive_point(LevelSetType & phi, const Index & ind)
	{
		if (phi(ind)>0) {
			m_reinit1->add_alive_point(ind);
			m_reinit2->add_forbidden_point(ind);
		}
		else {
			m_reinit1->add_forbidden_point(ind);
			m_reinit2->add_alive_point(ind);
		}	
	}

	template<unsigned DIM, typename Pixel>
	void FastMarchingInitializer<DIM, Pixel>::_reinit_fast_marching(LevelSetType & phi) 
	{
		// we launch the two fast marching
		if ( phi.thickness()>0 ) {
#pragma omp parallel sections
{
	#pragma omp section
			m_reinit1->run( phi.thickness());
	#pragma omp section
			m_reinit2->run(-phi.thickness());
}
		}
		else {
#pragma omp parallel sections
{
	#pragma omp section
			m_reinit1->run();
	#pragma omp section
			m_reinit2->run();
}
		}
		
		// we add the alive points to the band
		for (unsigned int i = 0 ; i < m_reinit1->nb_alive_points() ; i++)
			phi.add_point(m_reinit1->get_alive_point(i));

		for (unsigned int i = 0 ; i < m_reinit2->nb_alive_points() ; i++)
			phi.add_point(m_reinit2->get_alive_point(i));
	}
}

#endif // INITIALIZER_H
