/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef LEVELSET_H
#define LEVELSET_H

#include <cmath>
#include <blitz/array.h>
#include <Initializer.h>

#include <Fluxes.h>

#define _USE_MATH_DEFINES 

using namespace blitz;

/**
@namespace levelset
Set of classes which modelize evolution of a Level Set using the Narrow Band Method
*/
namespace levelset {

	/**
	* @author Jerome Piovano
	 * @brief Modelize a Narrow Band Level-Set.
	*/
	template<unsigned DIM, typename Pixel = float>
	class LevelSet : public Array<Pixel, DIM>
	{
	public:

		typedef 	 Array<Pixel, DIM>	           base;
		typedef typename base::T_index                     Index;
		typedef typename base::iterator                    Iterator;

		typedef 	 Initializer<DIM, Pixel>           InitializerType;

		typedef 	 Weno3<DIM, Pixel>                 Approximator;
		typedef 	 Fluxes<DIM, Approximator, Pixel>  FluxesType;
		typedef 	 Schemes<DIM, Pixel>               SchemesType;

		/** @name Constructors - Destructor - Resize */
		//@{

		/// Default constructor
		LevelSet();

		/// Constructor from a given shape
		LevelSet(const Index &  _shape,              ///< shape of the level set
			 float          _thickness = 3.0f    ///< thickness of the narrow band
			 );

		/// Copy constructor
		template <typename Pixel2>
		LevelSet(const LevelSet<DIM, Pixel2>& ls):base(ls, ColumnMajorArray<DIM>()), m_thickness(ls.thickness()), m_mask(ls.shape()), m_nbpts(0), m_pos_to_neg(0), m_neg_to_pos(0)
		{
			m_mask = -1;
			m_band = new Index[ls.size()];
			for (int i=0 ; i<ls.nb_points() ; i++)
				add_point(ls.get_point(i));
		}

		/// affectation operator
		template <typename Pixel2>
		LevelSet operator=(const LevelSet<DIM, Pixel2>& ls)
		{
			if (base::shape() != ls.shape())
				resize(ls.shape());
			
			base::operator=(ls);
			m_thickness = ls.thickness();
			m_nbpts = 0;
			for (int i=0 ; i<ls.nb_points() ; i++)
				add_point(ls.get_point(i));
			return *this;
		}

		/// Destructor
		virtual ~LevelSet();

		using base::operator=;

		void resize(const Index& s);

		//@}
		/** @name Update of the narrow band */
		//@{

		/// Update the level set in the band : phi += dphi * dt.
		void update(const Array<Pixel, DIM> &dphi, const float dt);

		//@}
		/** @name Access/Modif the narrow band */
		//@{

		/// Test if a point is in the narrow band
                bool is_in_band(const Index& ind) const;
		/// Add a point in the narrow band if not already in. return true if the point has been added
		bool add_point(const Index& ind);
		/// remove the point from the narrow_band
		bool remove_point(const Index& ind);
		/// Get the point of indice i in the narrow band
		const Index& get_point(unsigned i) const;
		/// Get the indice in the narrow band of a point
		int get_indice(const Index& ind) const;
		/// Number of points in the narrow band
		int nb_points() const;
		/// Indice [0     ..    pos_to_neg_index()-1] correspond to the points that changed from positive to negative in the last update
		int pos_to_neg_index() const;
		/// Indice [neg_to_pos_index() .. nb_points()-1] correspond to the points that changed from negative to positive in the last update
		int neg_to_pos_index() const;
		/// Thickness of the narrow band
		float & thickness();
		float thickness() const;
		/// Clear the narrow band, remove all points
		void clear_band();

		//@}
		/** @name Initializers */
		//@{

		/// Initialize to a sphere 
		void init_sphere(const Index& center, float radius, InitializerType& init, bool add=true);
		/// Initialize to small spheres, each of them being in a center of a window 
		void init_small_spheres(float radius, float window, InitializerType& init, bool add=true);
		/// Initialize to a rectangle close to the border 
		void init_border(const unsigned border, InitializerType& init, bool add=true);
		/// Initialize to a rectangle 
		void init_rectangle(const Index& tl, const Index& br, InitializerType& init, bool add=true);
		/// Initialize from an image, pix is the value representing the front
		void init_image(const Array<Pixel, DIM>& image, InitializerType& init, bool add=true, const Pixel& pix = 0);

		//@}
		/** @name Auxiliary functions */
		//@{


		/// Reinitialization EDP
		void reinit_PDE(int iter = 2, float dt = 0.4);
		/// Average of a quantity on the surface
		Pixel band_average(const Array<Pixel, DIM>& u, const float epsilon = 2.0) const;
		/// Integral of a quantity on the surface
		Pixel band_integral(const Array<Pixel, DIM>& u, const float epsilon = 2.0) const;
		/// Area of the surface
		Pixel area(const float epsilon = 2.0) const;


		//@}
		/** @name Data Members */
		//@{
	public:
		Array<Pixel, DIM>                m_mask;        ///< Mask of the points of the band
	protected:
		float                            m_thickness;   ///< Thickness of the narrow band	
		Index *                          m_band;        ///< Coords of the points in the band
		unsigned                         m_nbpts;       ///< Number of points in the band
		unsigned                         m_pos_to_neg;  ///< Special indice in 'm_band'
		unsigned                         m_neg_to_pos;  ///< Special indice in 'm_band'

		//@}
	};


	/** @example 2D_GeodesicActiveContour.C Geodesic Active Contour Segmentation of a 2D Image */
	/** @example 2D_Gaussian.C Gaussian Region Segmentation of a 2D Image */
	/** @example 2D_LocalStats.C Local Stats Region Segmentation of a 2D Image */
	/** @example 2D_Parzen.C Parzen Region Segmentation of a 2D Image */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor - Copy

	template<unsigned DIM, typename Pixel>
			LevelSet<DIM, Pixel>::LevelSet()
	: base(), m_thickness(0), m_band(NULL), m_mask(), m_nbpts(0), m_pos_to_neg(0), m_neg_to_pos(0)
	{ }

	template<unsigned DIM, typename Pixel>
			LevelSet<DIM, Pixel>::LevelSet(const Index &  _shape, float _thickness)
	: base(_shape, ColumnMajorArray<DIM>()), m_thickness(_thickness), m_mask(_shape), m_nbpts(0), m_pos_to_neg(0), m_neg_to_pos(0)
	{ 
		m_mask = -1;
		base::operator=( std::min(-m_thickness, -1.f) );
		m_band = new Index[base::size()];
	}

	template<unsigned DIM, typename Pixel>
		LevelSet<DIM, Pixel>::~LevelSet() 
	{ 
		if (m_band != NULL) delete[] m_band;
	}

	template<unsigned DIM, typename Pixel>
		void LevelSet<DIM, Pixel>::resize(const Index& s)
	{
		base::resize(s);
		m_mask.resize(s);
		base::operator=( std::min(-m_thickness, -1.f) );
		m_mask = -1;
		if (m_band != NULL) delete[] m_band;
		m_band = new Index[base::size()];
		m_nbpts = 0;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Update of the narrow band


	// note : after the update, the array 'm_band' is organized as follow :
	//
	//          --------------------------------------------------------------------------------------
	//  m_band  |      pos->neg      |                  other                   |     neg->pos       |
	//          --------------------------------------------------------------------------------------
	//          ^                    ^                                         ^                    ^
	//          0               m_pos_to_neg                             m_neg_to_pos            m_nbpts
	//
	// where,
	//    - pos->neg [0,            m_pos_to_neg-1] : points where the sign of phi changed from positive to negative.
	//    - neg->pos [m_pos_to_neg, m_neg_to_pos-1] : points where the sign of phi didn't change
	//    - other    [m_neg_to_pos,      m_nbpts-1] : points where the sign of phi changed from negative to positive
	// 
	// cf : "The Problem of the Dutch National Flag", Dijkstra

	template<unsigned DIM, typename Pixel>
	void LevelSet<DIM, Pixel>::update(const Array<Pixel, DIM> &dphi, const float dt)
	{
		m_pos_to_neg = 0;
		m_neg_to_pos = m_nbpts;

		unsigned int i = 0;
		while (i < m_neg_to_pos)
		{
			const Index& ind = get_point(i);

			Pixel tmp = base::operator()(ind);

			base::operator()(ind) += dt * dphi(ind);

			// keep the zero level set in the band
			if (m_thickness > 0) {
				base::operator()(ind) = (base::operator()(ind) >  m_thickness) ?  m_thickness : base::operator()(ind);
				base::operator()(ind) = (base::operator()(ind) < -m_thickness) ? -m_thickness : base::operator()(ind);
			}

			// sign changes
			if(tmp * base::operator()(ind) < 0)
			{  
				//pos->neg
				if (tmp>0){
					std::swap(m_band[m_pos_to_neg], m_band[i]);
					std::swap(m_mask(m_band[m_pos_to_neg]), m_mask(m_band[i]));
					m_pos_to_neg++;
					i++;
				}
				//neg->pos
				else if(tmp<0) {
					m_neg_to_pos--;
					std::swap(m_band[m_neg_to_pos], m_band[i]);
					std::swap(m_mask(m_band[m_neg_to_pos]), m_mask(m_band[i]));
				}
			}
			//others
			else i++;
		}
	}
	

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//  Access/Modif of the narrow band 
	

	template<unsigned DIM, typename Pixel>
	bool LevelSet<DIM, Pixel>::is_in_band(const Index& ind) const
	{
		return (m_mask(ind) != -1);
	}
		
	template<unsigned DIM, typename Pixel>
	bool LevelSet<DIM, Pixel>::add_point(const Index& ind)
	{
		if (!is_in_band(ind)){

			// we put the indice of the point in the mask of indice
			m_mask(ind) = m_nbpts;

			// we add the point at the end of m_band
			m_band[m_nbpts++] = ind;
			return true;
		}
		return false;
	}

	template<unsigned DIM, typename Pixel>
	bool LevelSet<DIM, Pixel>::remove_point(const Index& ind)
	{
		if (is_in_band(ind)){
	 		const int position = m_mask(ind);

			// we remove the point from the band
			m_mask(ind) = -1;
			m_nbpts--;

			// we put the last point at its place
			if ( position != int(m_nbpts))
			  {
				m_band[position] = m_band[m_nbpts];
				m_mask(m_band[position]) = position;
			  }
			return true;
		}
		return false;
	}
		
	template<unsigned DIM, typename Pixel>
	const typename LevelSet<DIM, Pixel>::Index& LevelSet<DIM, Pixel>::get_point(unsigned i) const
	{
		return m_band[i];
	}

	template<unsigned DIM, typename Pixel>
	int LevelSet<DIM, Pixel>::get_indice(const Index& ind) const
	{
		return m_mask(ind);
	}

	template<unsigned DIM, typename Pixel>
	int LevelSet<DIM, Pixel>::nb_points() const
	{
		return m_nbpts;
	}	

	template<unsigned DIM, typename Pixel>
	int LevelSet<DIM, Pixel>::pos_to_neg_index() const
	{
		return m_pos_to_neg;
	}	
		
	template<unsigned DIM, typename Pixel>
	int LevelSet<DIM, Pixel>::neg_to_pos_index() const
	{
		return m_neg_to_pos;
	}	
		
	template<unsigned DIM, typename Pixel>
	float & LevelSet<DIM, Pixel>::thickness()
	{
		return m_thickness;
	}
		
	template<unsigned DIM, typename Pixel>
	float LevelSet<DIM, Pixel>::thickness() const
	{
		return m_thickness;
	}

	template<unsigned DIM, typename Pixel>
	void LevelSet<DIM, Pixel>::clear_band()
	{
		m_nbpts = 0;
		m_mask = -1;
	}

	
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Initializers
	

	template<unsigned DIM, typename Pixel>
	void LevelSet<DIM, Pixel>::init_sphere(const Index& center, float radius, InitializerType& init, bool add)
	{
		for (Iterator i=base::begin() ; i!=base::end() ; ++i ){

			//compute the squared distance of the point 'i' to the point 'center'
			float dist = 0;
			for (unsigned dim=0 ; dim<base::dimensions() ; dim++)
				dist += (i.position()(dim)-center(dim))*(i.position()(dim)-center(dim));

			dist = radius - sqrt(dist);

			if (add) *i = ((dist>=0) || (*i>0))? 1 : -1;
			else     *i = ((dist< 0) && (*i>0))? 1 : -1;
		}

		init.init_band(*this);

	}

	template<unsigned DIM, typename Pixel>
	void LevelSet<DIM, Pixel>::init_small_spheres(float radius, float window, InitializerType& init, bool add)
	{
		for (Iterator i=base::begin() ; i!=base::end() ; ++i ){

			//compute the squared distance of the point 'i % windows' to the point of coordinate 'windows / 2'
			float dist = 0;
			for (unsigned dim=0 ; dim<base::dimensions() ; dim++)
				dist += (i.position()(dim)%int(window)-window/2)*(i.position()(dim)%int(window)-window/2);

			dist = radius - sqrt(dist);

			if (add) *i = ((dist>=0) || (*i>0))? 1 : -1;
			else     *i = ((dist< 0) && (*i>0))? 1 : -1;
		}

		init.init_band(*this);
	}
	
	template<unsigned DIM, typename Pixel>
	void LevelSet<DIM, Pixel>::init_border(const unsigned border, InitializerType& init, bool add)
	{
		for (Iterator i=base::begin() ; i!=base::end() ; ++i ){
	
			//inside=true if the point 'i' inside the region
			bool inside = true;
			for (unsigned dim=0 ; dim<base::dimensions() ; dim++)
				inside &= (i.position()(dim) >= border) && (i.position()(dim) < base::extent(dim) - border);

			if (add) *i = ( inside || (*i>0))? 1 : -1;
			else     *i = (!inside && (*i>0))? 1 : -1;
		}

		init.init_band(*this);
	}
	
	template<unsigned DIM, typename Pixel>
	void LevelSet<DIM, Pixel>::init_rectangle(const Index& tl, const Index& br, InitializerType& init, bool add)
	{
		for (Iterator i=base::begin() ; i!=base::end() ; ++i ){

			//inside=true if the point 'i' inside the region
			bool inside = true;
			for (unsigned dim=0 ; dim< base::dimensions() ; dim++)
				inside &= (i.position()(dim) >= tl(dim)) && (i.position()(dim) <= br(dim));

			if (add) *i = ( inside || (*i>0))? 1 : -1;
			else     *i = (!inside && (*i>0))? 1 : -1;
		}

		init.init_band(*this);

	}

	template<unsigned DIM, typename Pixel>		
	void LevelSet<DIM, Pixel>::init_image(const Array<Pixel, DIM>& image, InitializerType& init, bool add, const Pixel& pix)
	{
		if ( !Globals::same_shape<DIM>(base::shape(), image.shape()) )
			resize(image.shape());
		
		for (Iterator i=base::begin() ; i!=base::end() ; ++i ){
			if (add) *i = ((image(i.position()) >  pix) || (*i>0))? 1 : -1;
			else     *i = ((image(i.position()) <= pix) && (*i>0))? 1 : -1;
		}

		init.init_band(*this);
	}
	

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Auxiliary functions


	template<unsigned DIM, typename Pixel>
	void LevelSet<DIM, Pixel>::reinit_PDE(int iter, float dt)
	{
		// note : pb si sur stack et que temp trop gros, foire .. donc alloc/desalloc explicite.
		Pixel* temp = new Pixel[nb_points()];
			
		for (int t=0;t<iter;t++) {	
			#pragma omp parallel for
			for (int i = 0 ; i < nb_points() ; i++){
				const Index& ind = get_point(i);
				temp[i] = FluxesType::reinitialization(*this,ind);
			}

			#pragma omp parallel for
			for (int i = 0 ; i < nb_points() ; i++){
				const Index& ind = get_point(i);
				base::operator()(ind) += dt*temp[i];
			}
		}

		delete[] temp;
	}

	template<unsigned DIM, typename Pixel>
	Pixel LevelSet<DIM, Pixel>::band_average(const Array<Pixel, DIM>& u, const float epsilon) const
	{
		Pixel num = 0;
		Pixel den = 0;

		for (int i = 0 ; i < nb_points() ; i++)
		  {
		    const Index& ind = get_point(i);
		    const Pixel  val = base::operator()(ind);
		    
		    if (fabs(val) < epsilon) 
		      {
			const Pixel tmp = Globals::delta(val, epsilon) * SchemesType::norm_gradient(*this, ind);
			den += tmp;
			num += tmp * u(ind);
		      }
		  }
		
		if (den < EPS) return 0;

		return num / den;
	}

	template<unsigned DIM, typename Pixel>
	Pixel LevelSet<DIM, Pixel>::band_integral(const Array<Pixel, DIM>& u, const float epsilon) const
	{
		Pixel sum = 0;

		for (int i = 0 ; i < nb_points() ; i++){
			const Index& ind = get_point(i);
			const Pixel  val = base::operator()(ind);

			if (fabs(val) < epsilon)
				sum += u(ind) * Globals::delta(val, epsilon) * SchemesType::norm_gradient(*this, ind);
		}

		return sum;
	}

	template<unsigned DIM, typename Pixel>
	Pixel LevelSet<DIM, Pixel>::area(const float epsilon) const
	{
		Pixel sum = 0;
		
		for (int i = 0 ; i < nb_points() ; i++){
			const Index& ind = get_point(i);
			const Pixel  val = base::operator()(ind);

			if (fabs(val) < epsilon) 
				sum += Globals::delta(val, epsilon) * SchemesType::norm_gradient(*this, ind);
		}

		return sum;
	}
}


#endif	// LEVELSET_H
