/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef PRIORITY_QUEUE_H
#define PRIORITY_QUEUE_H

#include <list>
#include <cmath>
#include <iomanip>

namespace levelset {
  
  /**
   * @author Jerome Piovano
   * @brief Priority queue for the Fast Marching algorithm.
   *
   * Implementation of a priority queue containing elements of type Elements.
   * Complexity of insertion and removal in O(1) thanks to an untidy priority queue implementation.
   * see for more details "O(N) Implementation of the Fast Marching Algorithm" - Liron Yatziv, Alberto Bartesaghi, Guillermo Sapiro
   *
   * The only constraint is that you have to insert elements in the queue with priority :
   *               - superior to the lowest priority in the queue
   *               - inferior to (lowest priority in the queue  +  _inc_max)
   */
  
  template <class Element>
	class PriorityQueue
	{
	public:
	  typedef std::list<Element>                              ElementList;
	  typedef typename ElementList::iterator                  ElementListIterator;
	  typedef std::pair<ElementListIterator, int>             QueuePosition;
	  
	  /** @name Constructors - Destructor */
	  //@{
	  
	  /// Constructs an empty priority queue
	  PriorityQueue(unsigned _size    = 2000,   ///< number of quantization level of the priority
					double   _inc_max = 3       ///< max increment over the minimal point in the queue
					);
	  
	  /// Destructs the priority queue
	  ~PriorityQueue();
	  
	  //@}
	  /** @name Access/Modif the queue */
	  //@{
	  
	  
	  /// True if the queue is empty
	  bool empty() const;
	  /// Adds e to the queue, with a priority t, return the position where the elt has been inserted
	  QueuePosition push(const Element &e, double t);
	  /// Remove the element with the lowest priority
	  void pop();
	  /// Returns the element with the lowest priority
	  const Element & top();
	  /// Increases to t_new the priority of the element e
	  QueuePosition increase_priority(const Element& e, QueuePosition& pos, double t_new);
	  /// Clear the queue, remove all elements
	  void clear();
	  /// Number of elements in the queue
	  unsigned size() const;
	  
	  /// print the queue for debugging purpose
	  void print() const;
	  
	private:
	  
	  //@}
	  /** @name Member datas*/
	  //@{
	  
	  ElementList *    m_tab;      ///< Circular array containing the ElementList
	  unsigned         m_size;     ///< Size of the circular array
	  unsigned         m_nb_elem;  ///< Number of elements in the queue
	  double           m_t0;       ///< First level of quantification T0
	  int              m_i0;       ///< Index in the circular array of the first level of quantification T0
	  double           m_delta;    ///< Quantification step
	  double           m_inc_max;  ///< Max increment possible
	  
	  //@}
	};
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors - Destructor
  
  
  template <class Element>
	PriorityQueue<Element>::PriorityQueue(unsigned _size, double _inc_max) 
	: m_size(_size)
	, m_nb_elem(0)
	, m_t0(0)
	, m_i0(0)
	, m_inc_max(_inc_max)
  {
	m_tab   = new ElementList[_size];
	m_delta = _inc_max / static_cast<double>(_size);
  }
  
  template <class Element>
	PriorityQueue<Element>::~PriorityQueue() 
	{
	  delete[] m_tab; 
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Access/Modif the queue
  
  
  template <class Element>
	bool PriorityQueue<Element>::empty() const
	{
	  return (m_nb_elem == 0);
	}
  
  template <class Element>
	typename PriorityQueue<Element>::QueuePosition PriorityQueue<Element>::push(const Element &e, double t)
	{
	  
	  if (t-m_t0 > m_inc_max) 
		{
		//std::cerr << "Warning : Increment " << t-m_t0 << " superior to maximum increment " << m_inc_max << std::endl;
		}
	  else if (t-m_t0 < 0)
	  {
		//std::cerr << "Warning : Priority " << t << " inferior to the lowest priority in the queue " << m_t0 << std::endl;
		}
	  int i = 0;
	  if (empty()) {
		m_t0    = t;
		m_i0    = 0;
	  }
	  else {
		// compute the indice in the circular array.
		i = static_cast<int>(floor((t-m_t0) * m_size / m_inc_max));
		
		// compute the true indice in the regular array.
		i = (i >= int(m_size))? m_size-1 : ((i < 0)? 0 : i);
		i = (i + m_i0) % m_size;
	  }
	  
	  m_nb_elem++;
	  m_tab[i].push_front(e);
	  return QueuePosition(m_tab[i].begin(), i);
	}
  
  template <class Element>
	void PriorityQueue<Element>::pop()
	{
	  if (empty()) return;
	  
	  // shift m_i0 in order to be on a non empty level of quantization
	  while (m_tab[m_i0].empty()){
		m_i0 = (m_i0 + 1) % m_size;
		m_t0 += m_delta;
	  }
	  
	  m_nb_elem--;
	  m_tab[m_i0].pop_back();
	}
  
  template <class Element>
	const Element & PriorityQueue<Element>::top()
	{
	  // shift m_i0 in order to be on a non empty level of quantization
	  while (m_tab[m_i0].empty()){
		m_i0 = (m_i0 + 1) % m_size;
		m_t0 += m_delta;
	  }
	  
	  return m_tab[m_i0].back();
	}
  
  template <class Element>
	typename PriorityQueue<Element>::QueuePosition PriorityQueue<Element>::increase_priority(const Element& e, QueuePosition& pos, double t_new)
	{
	  if (t_new-m_t0 > m_inc_max) 
		std::cerr << "Warning : Increment " << t_new-m_t0 << " superior to maximum increment" << m_inc_max << std::endl;
	  else if (t_new - m_t0 < 0) 
		std::cerr << "Warning : Priority " << t_new << " inferior to the lowest priority in the queue " << m_t0 << std::endl;
	  
	  // compute the indice in the circular array
	  int  i = static_cast<int>(floor((t_new-m_t0) * m_size / m_inc_max));
	  
	  i = (i >= int(m_size))? m_size-1 : ((i < 0)? 0 : i);
	  i = (i + m_i0) % m_size;
	  
	  m_tab[pos.second].erase(pos.first);
	  m_tab[i].push_front(e);
	  return QueuePosition(m_tab[i].begin(), i);
	}
  
  template <class Element>
	void PriorityQueue<Element>::clear()
	{
	  m_nb_elem = 0;
	  m_t0      = 0;
	  m_i0      = 0;
	  for (int i= 0 ; i<int(m_size) ; i++)
		m_tab[i].clear();
	}
  
  template <class Element>
	unsigned PriorityQueue<Element>::size() const
	{
	  return m_nb_elem;
	}
  
  template <class Element>
	void PriorityQueue<Element>::print() const
	{
	  std::cout << "-------------------------------------"  << std::endl;
	  std::cout << "m_t0 = " << m_t0 << std::endl;
	  std::cout << "m_i0 = " << m_i0 << std::endl;
	  std::cout << "size = " << m_nb_elem << std::endl;
	  for (int i = 0; i < m_size; i++){
		if (!m_tab[i].empty()){
		  // index in the circular array, starting at m_i0
		  int j = (i-(int)m_i0 >= 0)? i-(int)m_i0 : i-(int)m_i0+m_size;
		  
		  // print infos concerning the bucket
		  std::cout << "  i=" << std::setw(2) << i 
					<< "| l=" << std::setw(2) << j 
					<< "| p=" << std::setw(5) << m_t0 + j*m_delta 
					<< "->"   << std::setw(5) << m_t0 + j*m_delta + m_delta;
		  // print the elements in the bucket
		  for (typename ElementList::iterator it = m_tab[i].begin() ; it!= m_tab[i].end() ; it++)
			std::cout << " " << *it ;
		  
		  std::cout << std::endl; 
		}
	  }
	}
  
}

#endif // PRIORITY_QUEUE_H
