/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef SCHEMES_H
#define SCHEMES_H

#include <Globals.h>
#include <LevelSet.h>

using namespace blitz;

namespace levelset {

	/**
	* @author Jerome Piovano
	* @brief Compute several mathematical operations using finite differences numerical schemes
	*/
	template<unsigned DIM, typename Pixel = float>
		class Schemes
		{
		public:
			typedef 	 blitz::Array<Pixel, DIM>	  DataType;
			typedef typename DataType::T_index          Index;

			/** @name static functions */
			//@{
	
			/// Fill the array 'sup_coords' and 'inf_coords' with the neighborings coords of 'ind'
			static void find_neighbours(const DataType & phi, const Index & ind, int inf_coords[], int sup_coords[]);
			/// Compute the laplacian \f$\Delta(\phi)\f$ by finite differences.
			static Pixel laplacian(const DataType & phi, const Index & ind);
			/// Compute the mean curvature  \f$\nabla\left(\frac{\nabla \phi}{\|\nabla \phi\|}\right)\f$  by finite differences
			static Pixel mean_curvature(const DataType & phi, const Index & ind);
			/// Compute the normal  \f$\frac{\nabla \phi}{\|\nabla \phi\|}\f$ by finite differences.
			static void normal(const DataType & phi, const Index & ind, Pixel norm[]);
			/// Compute the gradient \f$\nabla \phi\f$  by finite differences.
			static void gradient(const DataType & phi, const Index & ind, Pixel grad[]);
			/// Compute the squared norm of the gradient \f$\|\nabla \phi\|^2\f$ by finite differences.
			static float sq_norm_gradient(const DataType & phi, const Index & ind);
			/// Compute the norm of the gradient \f$\|\nabla \phi\|\f$ by finite differences.
			static float norm_gradient(const DataType & phi, const Index & ind);
			/// Compute the same mean curvature as the previous one, using another kind of discretization
			static Pixel other_mean_curvature(const DataType & phi, const Index & ind);

			//@}
	};


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// static functions


	template<unsigned DIM, typename Pixel>
	void Schemes<DIM, Pixel>::find_neighbours(const DataType & phi, const Index & ind, int inf_coords[], int sup_coords[])
	{
		for (int i=0 ; i<DIM ; i++){
			inf_coords[i] = (ind(i)>0)               ? ind(i) - 1 : ind(i);
			sup_coords[i] = (ind(i)<phi.extent(i)-1) ? ind(i) + 1 : ind(i);
		}
	}

	template <unsigned DIM, typename Pixel>
		Pixel Schemes<DIM, Pixel>::laplacian(const DataType & phi, const Index & ind)
	{
		Pixel result = DIM * phi(ind);
		result *= -2;
	
		int inf_coords[DIM];
		int sup_coords[DIM];
		Schemes::find_neighbours(phi, ind, inf_coords, sup_coords);
		
		Index sup;
		Index inf;
		for (int i=0 ; i<DIM ; i++){
			sup = ind;
			inf = ind;
			sup(i) = sup_coords[i];
			inf(i) = inf_coords[i];
			result += phi(sup) + phi(inf);
		}

		return result;
	}

	template <unsigned DIM, typename Pixel>
		Pixel Schemes<DIM, Pixel>::mean_curvature(const DataType & phi, const Index & ind)
	{
		int inf_coords[DIM];
		int sup_coords[DIM];
		Schemes::find_neighbours(phi, ind, inf_coords, sup_coords);

		if (DIM == 2){

			int x = ind(0);
			int y = ind(1);
			int mx = inf_coords[0];
			int px = sup_coords[0];
			int my = inf_coords[1];
			int py = sup_coords[1];

			const Pixel u0  = phi(x,y);

			const Pixel upx = phi(px,y);
			const Pixel umx = phi(mx,y);
			const Pixel upy = phi(x,py);
			const Pixel umy = phi(x,my);
	
			const Pixel umxmy = phi(mx,my);
			const Pixel upxmy = phi(px,my);
			const Pixel umxpy = phi(mx,py);
			const Pixel upxpy = phi(px,py);


			const Pixel ux = ( upx - umx ) / 2;
			const Pixel uy = ( upy - umy ) / 2;
	
			const Pixel uxx = upx - 2 * u0 + umx;
			const Pixel uyy = upy - 2 * u0 + umy;
			const Pixel uxy = (upxpy + umxmy - upxmy - umxpy) / 4;
	
			const Pixel ux2 = ux * ux;
			const Pixel uy2 = uy * uy;
			const Pixel grad = ux2 + uy2;
	
			if (grad < EPS) return 0;
	
			return (uyy * ux2 - 2 * ux * uy * uxy + uxx * uy2) / grad / sqrt(grad);
		}

		else if (DIM == 3){

			int x = ind(0);
			int y = ind(1);
			int z = ind(2);
			int mx = inf_coords[0];
			int px = sup_coords[0];
			int my = inf_coords[1];
			int py = sup_coords[1];
			int mz = inf_coords[2];
			int pz = sup_coords[2];

			const Pixel u0   = phi(x,y,z);
			const Pixel upx  = phi(px,y,z);
			const Pixel umx  = phi(mx,y,z);
			const Pixel upy  = phi(x,py,z);
			const Pixel umy  = phi(x,my,z);
			const Pixel upz  = phi(x,y,pz);
			const Pixel umz  = phi(x,y,mz);
	
			const Pixel umxmy = phi(mx,my,z);
			const Pixel upxmy = phi(px,my,z);
			const Pixel umxpy = phi(mx,py,z);
			const Pixel upxpy = phi(px,py,z);
			const Pixel umymz = phi(x,my,mz);
			const Pixel upymz = phi(x,py,mz);
			const Pixel umypz = phi(x,my,pz);
			const Pixel upypz = phi(x,py,pz);
			const Pixel umzmx = phi(mx,y,mz);
			const Pixel upzmx = phi(mx,y,pz);
			const Pixel umzpx = phi(px,y,mz);
			const Pixel upzpx = phi(px,y,pz);

			const Pixel ux  = ( upx - umx ) / 2;
			const Pixel uy  = ( upy - umy ) / 2;
			const Pixel uz  = ( upz - umz ) / 2;
	
			const Pixel uxx = upx - 2 * u0 + umx;
			const Pixel uyy = upy - 2 * u0 + umy;
			const Pixel uzz = upz - 2 * u0 + umz;
			const Pixel uxy = (upxpy + umxmy - upxmy - umxpy) / 4;
			const Pixel uyz = (upypz + umymz - upymz - umypz) / 4;
			const Pixel uzx = (upzpx + umzmx - upzmx - umzpx) / 4;
	
			const Pixel ux2 = ux * ux;
			const Pixel uy2 = uy * uy;
			const Pixel uz2 = uz * uz;
			const Pixel grad = ux2 + uy2 + uz2;
	
			if (grad < EPS) return 0;
	
			return (   (uyy+uzz) * ux2
				+ (uxx+uzz) * uy2
				+ (uxx+uyy) * uz2
				- 2*ux*uy*uxy
				- 2*uz*ux*uzx
				- 2*uy*uz*uyz ) / grad / sqrt(grad) / 2;
		}

		else  std::cerr << "Numerical operation not implemented in this dimension." << std::endl;

		return 0;
	}

	template <unsigned DIM, typename Pixel>
		void Schemes<DIM, Pixel>::normal(const DataType & phi, const Index & ind, Pixel norm[])
	{
		Schemes::gradient(phi, ind, norm);	

		float n = 0;

		for (int i=0 ; i<DIM ; i++)
			n += norm[i]*norm[i];

		n = sqrt( n );
		if (n < EPS) n = (Pixel) EPS;

		for (int i=0 ; i<DIM ; i++)
			norm[i] /= n;
	}
	
	template <unsigned DIM, typename Pixel>
		void Schemes<DIM, Pixel>::gradient(const DataType & phi, const Index & ind, Pixel grad[])
	{
		int inf_coords[DIM];
		int sup_coords[DIM];
		Schemes::find_neighbours(phi, ind, inf_coords, sup_coords);

		Index inf;
		Index sup;
		for (int i=0 ; i<int(DIM) ; i++){
			inf = ind;
			sup = ind;
			inf(i) = inf_coords[i];
			sup(i) = sup_coords[i];
 			grad[i] = (phi(sup) - phi(inf)) / 2.0; 
		}
	}

	template <unsigned DIM, typename Pixel>
		float Schemes<DIM, Pixel>::sq_norm_gradient(const DataType & phi, const Index & ind)
	{
		float result = 0;
		Pixel grad[DIM];
		Schemes::gradient(phi, ind, grad);

		for (int i=0 ; i<int(DIM) ; i++)
			result += grad[i]*grad[i];

		return result;
	}

	template <unsigned DIM, typename Pixel>
		float Schemes<DIM, Pixel>::norm_gradient(const DataType & phi, const Index & ind)
	{
		return sqrt( Schemes::sq_norm_gradient(phi, ind));
	}

	template <unsigned DIM, typename Pixel>
		Pixel Schemes<DIM, Pixel>::other_mean_curvature(const DataType & phi, const Index & ind)
	{

		int inf_coords[DIM];
		int sup_coords[DIM];
		Schemes::find_neighbours(phi, ind, inf_coords, sup_coords);

		if (DIM == 2){

			int x = ind(0);
			int y = ind(1);
			int mx = inf_coords[0];
			int px = sup_coords[0];
			int my = inf_coords[1];
			int py = sup_coords[1];
			int ppx = (px<phi.extent(0)-1) ? px+1:px;
			int mmx = (mx>0) ? mx-1:mx;
			int ppy = (py<phi.extent(1)-1) ? py+1:py;
			int mmy = (my>0) ? my-1:my;
	
			const Pixel u0  = phi(x,y);
			const Pixel upx = phi(px,y);
			const Pixel umx = phi(mx,y);
			const Pixel upy = phi(x,py);
			const Pixel umy = phi(x,my);
			const Pixel uppx = phi(ppx,y);
			const Pixel ummx = phi(mmx,y);
			const Pixel uppy = phi(x,ppy);
			const Pixel ummy = phi(x,mmy);
			const Pixel umxmy = phi(mx,my);
			const Pixel upxmy = phi(px,my);
			const Pixel umxpy = phi(mx,py);
			const Pixel upxpy = phi(px,py);
	
			const Pixel upx_x = ( uppx  - u0    ) / 2;
			const Pixel upx_y = ( upxpy - upxmy ) / 2;
			const Pixel umx_x = ( u0    - ummx  ) / 2;
			const Pixel umx_y = ( umxpy - umxmy ) / 2;
			const Pixel upy_x = ( upxpy - umxpy ) / 2;
			const Pixel upy_y = ( uppy  - u0    ) / 2;
			const Pixel umy_x = ( upxmy - umxmy ) / 2;
			const Pixel umy_y = ( u0    - ummy  ) / 2;
	
			const Pixel gradpx = upx_x*upx_x + upx_y*upx_y;
			const Pixel gradmx = umx_x*umx_x + umx_y*umx_y;
			const Pixel gradpy = upy_x*upy_x + upy_y*upy_y;
			const Pixel gradmy = umy_x*umy_x + umy_y*umy_y;
	
			const Pixel normalpx_x = upx_x / sqrt(gradpx + 1e-10);
			const Pixel normalmx_x = umx_x / sqrt(gradmx + 1e-10);
			const Pixel normalpy_y = upy_y / sqrt(gradpy + 1e-10);
			const Pixel normalmy_y = umy_y / sqrt(gradmy + 1e-10);
	
			const Pixel normal_xx = ( normalpx_x - normalmx_x ) / 2;
			const Pixel normal_yy = ( normalpy_y - normalmy_y ) / 2;
	
			return normal_xx + normal_yy;
		}

		else  std::cerr << "Numerical operation not yet implemented in this dimension." << std::endl;

		return 0;
	}
}




#endif // SCHEMES_H
