/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef SEGMENTATION_H
#define SEGMENTATION_H

#include <sstream>
#include <vector>

#include <Globals.h>
#include <Schemes.h>
#include <LevelSet.h>
#include <Initializer.h>

#include <SegmentationModule/SegmentationModule.h>



using namespace blitz;

/**
   @namespace segmentation
   Set of classes which perform segmentation of images thanks to the level set method 
*/
namespace segmentation {
  
using namespace levelset;
  /**
   * @author Jerome Piovano
   *
   * @brief Segmentation of an image thanks to the Level-Set method.
   *
   * The constructor takes 3 objects in parameters :
   *
   *       - LevelSet       : LevelSet that will evolve and segment the image
   *       - Initializer    : Initializer that will reinitilize the LevelSet each 'm_reinit_freq' iterations
   *       - DataType       : optional data term, image to segment
   *
   * The 'modules' are the object that are giving speed term to the level set.
   *
   * @note _dt <= 1 , for stability
   */
  template<unsigned DIM, typename PixelImage, typename Pixel = float>
	class Segmentation 
	{
	public:	
	typedef 	  Array<Pixel, DIM>	                 DataType;
	typedef 	  Array<unsigned char, DIM>	         ResultType;
	typedef           Initializer<DIM, Pixel>                InitializerType;
	typedef           levelset::LevelSet<DIM, Pixel>                   LevelSetType;
	typedef typename  LevelSetType::Index                    Index;
	typedef           SegmentationModule<DIM, Pixel>         SegmentationModuleType;
	typedef           std::vector<SegmentationModuleType *>  Modules;
	
	typedef           Schemes<DIM, Pixel>                    SchemesType;
	
	/** @name Constructors - Destructor*/
	//@{	
	
	/// Constructor
	Segmentation(LevelSetType *          _phi,                   ///< level set that will segment the image
			     InitializerType *       _init,                  ///< Reinitializer of the level-set
			     DataType *              _data         = NULL,   ///< Image to segment (NULL if no data term)
			     Array<bool, DIM> *      _mask         = NULL,   ///< Binary mask to restrict evolution of the level set (NULL if no mask)
			     float                   _dt           = 0.5f,   ///< Time step of the evolution
			     int                     _reinit_freq  = 1       ///< frequence of Reinitialization of the LevelSet
			     );	
	
	/// Destructor
	virtual ~Segmentation();
	
	//@}
	/** @name Access/Modif data terms  */
	//@{
	
	/// Add a module to the list of modules
	void add_module(SegmentationModuleType & _sm);
	/// Remove the module from the list of modules
	void remove_module(SegmentationModuleType & _sm);
	/// remove all modules
	void remove_all_modules();
	/// Set the LevelSet
	void set_level_set(LevelSetType & _phi);
	/// Set the Initializer
	void set_initializer(InitializerType & _init);
	/// Set the data term
	void set_data_term(DataType & _data);
	/// Add a mask to the segmentation process, the level set will evolve only in this mask
	void set_mask(Array<bool, DIM> * _mask);
	/// Time step of the evolution
	float & dt();
	/// Frequence of reinitialization of the band.
	int & reinit_freq();
	
	void check_if_valid();
	
	//@}
	/** @name Results of the segmentation */
	//@{
	
	/// Return an image on which the contour is drawn on the original image
	ResultType get_result();
	/// Return a binary image of the result of the segmentation
	ResultType get_binary_result();
	/// Return an image representing the speed induced by the segmentation module '_module'. i=0 mean the final speed, combined by the module assembler.
	DataType get_speed(int _module = 0);
	/// Compute and return the energy of the segmentation
	float get_energy();
	
	//@}
	/** @name segmentation */
	//@{
	
	/// Start the segmentation. if nbiter < 0, stop at convergence (if no convergence, may not stop).
	void segment(int nbiter = 30);//, RGBPixel neg = RGBPixel(0,0,0), RGBPixel pos = RGBPixel(255,255,255));
	/// Initialize the segmentation (not protected because used in the GUI)
	void init();
	/// Compute one iteration (not protected because used in the GUI)
	void iterate(int iter);
	
	protected:
	
	//@}
	/** @name Member data */
	//@{
	
	LevelSetType *                         m_phi;                ///< Level set 
	DataType                               m_dphi;               ///< Variations of the level set
	InitializerType *                      m_init;               ///< Initializer
	DataType *                             m_data;               ///< Image to segment
	std::vector<SegmentationModuleType *>  m_sm;                 ///< List of segmentation modules
	Array<bool, DIM> *                     m_mask;               ///< Boolean mask where the segmentation is to be applied
	int                                    m_reinit_freq;        ///< Frequence of reinitialization of the band
	float                                  m_dt;                 ///< Time step of the evolution
	
	//@}
	};
  
  
  /** @example 2D_GeodesicActiveContour.C Geodesic Active Contour Segmentation of a 2D Image */
  /** @example 2D_Gaussian.C Gaussian Region Segmentation of a 2D Image */
  /** @example 2D_LocalGaussian.C Local Gaussian Region Segmentation of a 2D Image */
  /** @example 2D_Parzen.C Parzen Region Segmentation of a 2D Image */
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors - Destructor - Initializers
  
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	Segmentation<DIM, PixelImage, Pixel>::Segmentation(LevelSetType * _phi,  InitializerType * _init, DataType * _data, Array<bool, DIM> * _mask, float _dt, int _reinit_freq)
	:m_phi(_phi), m_dphi(_phi->shape()), m_init(_init), m_data(_data), m_mask(_mask), m_reinit_freq(_reinit_freq), m_dt(_dt)
  {
	if (_data != NULL  && !Globals::same_shape<DIM>(_data->shape(), _phi->shape()) ) {
	  std::cerr << "Error : Shape of the data term and shape of the LevelSet don't match"  << std::endl;
	  exit(1);
	}
	if (_mask != NULL && !Globals::same_shape<DIM>(_mask->shape(), _phi->shape()) ) {
	  std::cerr << "Error : Shape of the mask and shape of the LevelSet don't match"  << std::endl;
	  exit(1);
	}
  }
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	Segmentation<DIM, PixelImage, Pixel>::~Segmentation()
	{ }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Modify segmentation information
  
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::add_module(SegmentationModuleType & _sm) 
	{
	  if ( std::find( m_sm.begin(), m_sm.end(), &_sm ) == m_sm.end() )
		m_sm.push_back(&_sm);
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::remove_module(SegmentationModuleType & _sm)	
	{
	  for (typename std::vector<SegmentationModuleType *>::iterator it = m_sm.begin(); it != m_sm.end(); it++ ){
		if (*it == &_sm){
		  m_sm.erase(it);
		  return;
		}
	  }
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::remove_all_modules()	
	{
	  m_sm.clear();
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::set_mask( Array<bool, DIM>* _mask) 
	{
	  if (_mask != NULL && !Globals::same_shape<DIM>(_mask->shape(), m_phi->shape()) ) {
		std::cerr << "Error : Shape of the mask and shape of the LevelSet don't match"  << std::endl;
		exit(1);
	  }
	  m_mask = _mask;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::set_initializer(InitializerType & _init)
	{
	  m_init = &_init;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::set_level_set(LevelSetType & _phi)
	{
	  m_phi = &_phi;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::set_data_term(DataType & _data)
	{
	  if ( !Globals::same_shape<DIM>(_data.shape(), m_phi->shape()) ) {
		std::cerr << "Error : Shape of the data term and shape of the LevelSet don't match"  << std::endl;
		exit(1);
	  }
	  m_data = &_data;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	float & Segmentation<DIM, PixelImage, Pixel>::dt()
	{
	  return m_dt;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	int & Segmentation<DIM, PixelImage, Pixel>::reinit_freq()
	{
	  return m_reinit_freq;
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Results of the segmentation
  
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	typename Segmentation<DIM, PixelImage, Pixel>::ResultType Segmentation<DIM, PixelImage, Pixel>::get_result()
	{
	  ResultType result(m_phi->shape(), ColumnMajorArray<DIM>());
	  
	  if (m_data == NULL)
		result = 255;
	  else
		result = blitz::cast<unsigned char>(*m_data);
	  
#pragma omp parallel for
	  for (int i = 0 ; i < m_phi->nb_points() ; i++){
		const Index& ind = m_phi->get_point(i);
		
		const float _phi = (*m_phi)(ind);
		
		int inf_coords[DIM];
		int sup_coords[DIM];
		SchemesType::find_neighbours(*m_phi, ind, inf_coords, sup_coords);
		
		for (int i=0 ; i<DIM ; i++){
		  Index inf = ind;
		  Index sup = ind;
		  inf(i) = inf_coords[i];
		  sup(i) = sup_coords[i];
		  if ((_phi * (*m_phi)(inf) <= 0) || (_phi * (*m_phi)(sup) <= 0))
			result(ind) = (_phi>0)? 255: 0;
		}
	  }
	  
	  return result;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	typename Segmentation<DIM, PixelImage, Pixel>::ResultType Segmentation<DIM, PixelImage, Pixel>::get_binary_result() 
	{
	  ResultType result(m_phi->shape(), ColumnMajorArray<DIM>());
	  for (typename ResultType::iterator ind=result.begin() ; ind!=result.end() ; ++ind )
		result(ind) = ((*m_phi)(ind) < 0)? 0 : 255;
	  
	  return result;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	typename Segmentation<DIM, PixelImage, Pixel>::DataType Segmentation<DIM, PixelImage, Pixel>::get_speed(int _module)
	{
	  DataType result(m_phi->shape(), ColumnMajorArray<DIM>());
	  result = 0;
	  
	  if (_module < 0 ||  _module > m_sm.size()){
		std::cerr << " Module does not exist !" << std::endl;
		return result;
	  }
	  
	  if (_module == 0){
#pragma omp parallel for
		for (int i = 0 ; i < m_phi->nb_points() ; i++){
		  const Index& ind = m_phi->get_point(i);
		  if (m_mask==NULL || (*m_mask)(ind)==true){
			result(ind) = 0;
			for (int i=0; i<m_sm.size(); i++){
			  const float value  = m_sm[i]->value(*m_phi, ind);
			  const float weight = m_sm[i]->weight(ind);
			  result(ind) += value*weight;
			}
		  }
		}
	  }
	  else {
#pragma omp parallel for
		for (int i = 0 ; i < m_phi->nb_points() ; i++){
		  const Index& ind = m_phi->get_point(i);
		  if (m_mask==NULL || (*m_mask)(ind)==true){
			const float value  = m_sm[_module-1]->value(*m_phi, ind);
			const float weight = m_sm[_module-1]->weight(ind);
			result(ind) = value*weight;
		  }
		}
	  }
	  
	  return result;
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	float Segmentation<DIM, PixelImage, Pixel>::get_energy() 
	{
	  float total_energy = 0;
	  
	  for (typename LevelSetType::Iterator i=m_phi->begin() ; i!=m_phi->end() ; ++i){
		
		Index ind = i.position();
		
		if (m_mask==NULL || (*m_mask)(ind)!=0) 
		  for (int j=0; j<m_sm.size(); j++){
			const float energy = m_sm[j]->energy(*m_phi, ind);
			const float weight = m_sm[j]->weight(ind);
			total_energy += energy*weight;
		  }
	  }
	  
	  return total_energy;
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Launch the segmentation
  
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::segment(int nbiter)
	{
	  // initialization
	  init(); 
	  // main loop
	  for (int i=0 ; i<nbiter ; i++)
		{
		  iterate(i);
		}
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::init()
	{

	  //	std::cout << "Segmentation::init shape: m_dphi="<<m_dphi.shape() << ", m_phi=" << m_phi->shape()<<std::endl;

	  // resize of d_phi
	  if ( !Globals::same_shape<DIM>(m_dphi.shape(), m_phi->shape()) )
		m_dphi.resize(m_phi->shape());
	  
	  m_dphi=0.0f;
	  
	  // initialization of the modules
	  for ( size_t i = 0; i < m_sm.size(); ++i )
    {
		  m_sm[i]->init(*m_phi);
    }
	}
  
  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::check_if_valid()
	{
	/*
	LevelSetType *                         m_phi;                ///< Level set 
	DataType                               m_dphi;               ///< Variations of the level set
	InitializerType *                      m_init;               ///< Initializer
	DataType *                             m_data;               ///< Image to segment
	*/
  
	}

  template<unsigned DIM, typename PixelImage, typename Pixel>
	void Segmentation<DIM, PixelImage, Pixel>::iterate(int iter) 
	{
#ifdef PRINT_TIME
	  timeval _time;
	  Globals::initChrono(_time);
#endif
	  
	  m_dphi=0.0f;
	  
	  // speed computation
#pragma omp parallel for
	  for (int i = 0 ; i < m_phi->nb_points() ; i++)
	  {
		
			const Index& ind = m_phi->get_point(i);
			m_dphi(ind)=0.0f;
		
			if (m_mask==NULL || (*m_mask)(ind)==true)
			{
				m_dphi(ind) = 0;
				for ( int i = 0; i < int(m_sm.size()); i++)
				{
					const float value  = m_sm[i]->value(*m_phi, ind);
					const float weight = m_sm[i]->weight(ind);
					//if ( isnan(value) || isnan(weight) ) throw std::runtime_error("found nan in value");
					m_dphi(ind) += value*weight;
				}
			}
	  }
#ifdef PRINT_TIME
	  Globals::printChrono("	=> compute speed : ", _time);
	  Globals::initChrono(_time);
#endif
	  
	  // phi += dt * d_phi
	  m_phi->update(m_dphi, m_dt);
	  
#ifdef PRINT_TIME
	  Globals::printChrono("	=> update Level set : ", _time);
	  Globals::initChrono(_time);
#endif
	  
	  // update segmentation modules
	  for ( int  i = 0; i < int(m_sm.size()); i++ )
		if (m_sm[i]->adaptive()) m_sm[i]->update_before_reinit(*m_phi);
#ifdef PRINT_TIME
	  Globals::printChrono("	=> total update modules before reinitialization : ", _time);
	  Globals::initChrono(_time);
#endif
	  
	  // level-set reinitialization
	  if(iter%m_reinit_freq == 0) 
		m_init->reinit_band(*m_phi);

	  
#ifdef PRINT_TIME
	  Globals::printChrono("	=> total reinit band : ", _time);
	  Globals::initChrono(_time);
#endif
	  
	  // update segmentation modules
	  for ( int i = 0; i < int(m_sm.size()); i++ )
		if (m_sm[i]->adaptive()) m_sm[i]->update_after_reinit(*m_phi);
		
#ifdef PRINT_TIME
	  Globals::printChrono("	=> total update modules after reinitialization : ", _time);
	  Globals::initChrono(_time);
#endif
	}
}

#endif // SEGMENTATION_H
