/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef GAUSSIANREGIONMODULE_H
#define GAUSSIANREGIONMODULE_H

#include <SegmentationModule/RegionModule.h>
#include <vector>
#include <Schemes.h>

namespace segmentation {

	/**
	* @author Jerome Piovano
	*
	* @brief RegionModule specialization used to drive the LevelSet using a scalar gaussian region evolution
	*
	*/
	template <unsigned DIM, typename PixelImage, typename Pixel = float> 
	class GaussianRegionModule : public RegionModule<DIM, PixelImage, Pixel>
	{
	public:
		
	 	typedef          RegionModule<DIM, PixelImage, Pixel>    base;
		typedef typename base::LevelSetType                      LevelSetType;
		typedef typename base::Index                             Index;
		typedef typename base::HistogramType                     HistogramType;
		typedef typename base::ImageType                         ImageType;

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor
		GaussianRegionModule(ImageType &  _data,             ///< image to segment
				     float        _weight = 1.f,     ///< weight of the module
				     bool         _var=false,        ///< false => don't use the variance (Chan-Vese evolution)
				     bool         _adaptive = true   ///< if the module is adaptive
				     );

		/// Destructor
		virtual ~GaussianRegionModule();

		//@}
		/** @name Modify/Access member data */
		//@{

		/// false => Chan-Vese piecewise constant mumford shah evolution
		void use_variance(bool _var=true);

		//@}
		/** @name Member functions */
		//@{

		/// Value of the module at a point
		float value(const LevelSetType & phi, const Index & ind);
		/// Energy of the module at a point
		float energy(const LevelSetType & phi, const Index & ind);
		/// Initialize the module 
		void init(const LevelSetType & phi);
		/// update the module 
		void update_before_reinit(const LevelSetType & phi);
		/// save results depending on the module data
		void save_result(const LevelSetType & phi, std::string evodir, int iter);


		/// update the parameters of the gaussian (mean, variance) using the histograms, and precompute the log-likelihood of all possible values
		void update_gaussian_params();

		//@}
		/** @name Member data  */
		//@{

	protected:

		std::vector<Pixel>  m_ll_pos;     ///< Log-likelihoods for the positive region
		std::vector<Pixel>  m_ll_neg;     ///< Log-likelihoods for the negative region
		float    	    m_mean_pos;   ///< mean of the positive region
		float   	    m_mean_neg;   ///< mean of the negative region
		float   	    m_var_pos;    ///< variance of the positive region
		float   	    m_var_neg;    ///< variance of the negative region
		bool                m_var;        ///< Use of the variance

		//@}

	};


	/** @example 2D_Gaussian.C Gaussian Region Segmentation of a 2D Image */
	/** @example 3D_ChanVese.C Chan-Vese Region Segmentation of a 3D Image */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <unsigned DIM, typename PixelImage, typename Pixel>
		GaussianRegionModule<DIM, PixelImage, Pixel>::GaussianRegionModule(ImageType & _data, float _weight, bool _var, bool _adaptive)
		:RegionModule<DIM, PixelImage, Pixel>(_data, _weight, _adaptive), m_mean_pos(0), m_mean_neg(0), m_var_pos(0), m_var_neg(0), m_var(_var)
	{ }

	template <unsigned DIM, typename PixelImage, typename Pixel>
		GaussianRegionModule<DIM, PixelImage, Pixel>::~GaussianRegionModule() 
	{ }


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Modify/Access member data 

		
	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GaussianRegionModule<DIM, PixelImage, Pixel>::use_variance(bool _var)
	{
		m_var = _var;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Member functions

		
	template <unsigned DIM, typename PixelImage, typename Pixel>
		float GaussianRegionModule<DIM, PixelImage, Pixel>::value(const LevelSetType & phi, const Index & ind)
	{ 
		PixelImage val = base::m_data(ind);

		//if ( base::m_hist_pos->get_index(val) >= std::min( m_ll_pos.size(), m_ll_neg.size()) || base::m_hist_pos->get_index(val) < 0 ) std::cout << "GaussianRegionModule::value: index out of bounds: index = " << base::m_hist_pos->get_index(val) << " ll_pos.size="<<m_ll_pos.size() << "  ll_neg.size="<<m_ll_neg.size()<<std::endl;

		const Pixel ll_pos = m_ll_pos[base::m_hist_pos->get_index(val)];
		const Pixel ll_neg = m_ll_neg[base::m_hist_neg->get_index(val)];

		//if ( isnan (phi(ind) ) || isnan( ll_pos - ll_neg ) )
		//		std::cout << "GaussianRegionModule::value: nan found: phi(ind)=" << phi(ind) << "  ll_pos="<<ll_pos << "  ll_neg="<<ll_neg<<std::endl;

		return Globals::delta(phi(ind)) * (ll_pos - ll_neg);
		//return  ((m_mean_pos-m_mean_neg > 0)? 1 : -1) * Globals::delta(phi(ind)) * (ll_pos - ll_neg);
	}	

	template <unsigned DIM, typename PixelImage, typename Pixel>
		float GaussianRegionModule<DIM, PixelImage, Pixel>::energy(const LevelSetType & phi, const Index & ind)
	{
		PixelImage val = base::m_data(ind);
		float ll_pos = m_ll_pos[base::m_hist_pos->get_index(val)];
		float ll_neg = m_ll_neg[base::m_hist_neg->get_index(val)];
		float h = Globals::heaviside(phi(ind));

		return -(1-h)*ll_pos - h*ll_neg;
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GaussianRegionModule<DIM, PixelImage, Pixel>::init(const LevelSetType & phi)
	{
		base::init(phi);
		base::init_histograms(phi);
		update_gaussian_params();
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GaussianRegionModule<DIM, PixelImage, Pixel>::update_before_reinit(const LevelSetType & phi)
 	{ 
		base::update_histograms(phi);
		update_gaussian_params();
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GaussianRegionModule<DIM, PixelImage, Pixel>::update_gaussian_params()
	{ 
		// compute the mean and variance of the negative histograms
		base::hist_neg()->compute_statistics(m_mean_neg, m_var_neg, m_var);
		base::hist_pos()->compute_statistics(m_mean_pos, m_var_pos, m_var);

		// if the size of the histogram has changed, resize de log-likelihood vectors
		if ( int(m_ll_neg.size()) != base::hist_neg()->size() ) 
		  m_ll_neg.resize( base::hist_neg()->size() );
		if ( int(m_ll_pos.size()) != base::hist_pos()->size() ) 
		  m_ll_pos.resize( base::hist_pos()->size() );

		if (m_var) {
		  for( int i = 0; i < int(m_ll_neg.size()); i++)
			m_ll_neg[i] = 
			              - 0.5f * log( 2 * M_PI * m_var_neg )
			              - 0.5f * ( i+base::hist_neg()->min() - m_mean_neg) 
			              * ( i+base::hist_neg()->min() - m_mean_neg) / m_var_neg;
		  for( int i = 0; i < int(m_ll_pos.size()); i++ )
			{m_ll_pos[i] = 
			              - 0.5f * log( 2 * M_PI * m_var_pos ) 
			              - 0.5f * ( i+base::hist_pos()->min() - m_mean_pos) 
			              * ( i+base::hist_pos()->min() - m_mean_pos) / m_var_pos;
/*
			if ( isnan(m_ll_pos[i]) )
				{
					std::cout << "m_var_pos = "<< m_var_pos << " base::hist_pos()->min()="<<base::hist_pos()->min()<<" m_mean_pos"<<m_mean_pos << std::endl;
					throw std::runtime_error("Here, I got him!");               
				}
			if ( isnan(m_ll_neg[i]) )
				{
					std::cout << "m_var_neg = "<< m_var_neg << " base::hist_neg()->min()="<<base::hist_neg()->min()<<" m_mean_neg"<<m_mean_neg << std::endl;
					throw std::runtime_error("Here, I got him!");               
				}*/
			}
		}
		else{
		  for( int i = 0; i < int(m_ll_neg.size()); i++ )
			m_ll_neg[i] =  
			  - ( i + base::hist_neg()->min() - m_mean_neg ) 
			  * ( i + base::hist_neg()->min() - m_mean_neg );
		  for( int i = 0; i < int(m_ll_pos.size()); i++ )
			m_ll_pos[i] =  
			  - ( i + base::hist_pos()->min() - m_mean_pos ) 
			  * ( i + base::hist_pos()->min() - m_mean_pos );
		}
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GaussianRegionModule<DIM, PixelImage, Pixel>::save_result(const LevelSetType & phi, std::string evodir, int iter) 
	{ }
}

#endif // GAUSSIANREGIONMODULE_H
