/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef GEODESICACTIVECONTOURMODULE_H
#define GEODESICACTIVECONTOURMODULE_H

#include <SegmentationModule/SegmentationModule.h>

namespace segmentation {


	/**
	* @author Jerome Piovano
	*
	* @brief SegmentationModule specialization used to drive the LevelSet using a geodesic active contour evolution
	*
	* @note default values are : m_v=0; smooth=3.0f
	*/
	template <unsigned DIM, typename PixelImage, typename Pixel = float> 
	class GeodesicActiveContourModule : public SegmentationModule<DIM, Pixel>
	{
	public:

		typedef          SegmentationModule<DIM, Pixel>  base;
		typedef typename base::LevelSetType              LevelSetType;
		typedef typename base::Index                     Index;

		typedef          Array<PixelImage, DIM> 	 ImageType;
		typedef          Array<float, DIM>	         ImageFloat;

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor
		GeodesicActiveContourModule(ImageType *  _data,            ///< image whose metric will be computed
					    float        _weight = 1.f,    ///< weight of the module
					    float        _v = 0.f,         ///< Balloon force
					    float        _smooth = 3.0f,   ///< amount of smoothness on the image
					    float        _cut = 5.0f       ///< percentage of the metric range that will be cut, and set to 0 (in order to prevents leaks)
					    );

		/// Destructor
		virtual ~GeodesicActiveContourModule();
 
		//@}
		/** @name Modify/Access member data */
		//@{

		/// Compute the metric of the riemanian space from the image
		void compute_metric(ImageType * _data, float _smooth = 3.0f, float _cut = 5.0f);
		/// Set the metric of the riemanian space, and optionally its directional derivative
		void set_metric(const ImageFloat & g, ImageFloat * g_deriv = NULL);
		/// Balloon force
		float& balloon();
		/// Metric of the riemanian space
		const ImageFloat & metric();

		//@}
		/** @name Member functions */
		//@{

		/// Value of the module at a point
		float value(const LevelSetType & phi, const Index & ind);
		/// Energy of the module at a point
		float energy(const LevelSetType & phi, const Index & ind);
		/// Initialize the module 
		void init(const LevelSetType & phi);
		/// Save results depending on the module data
		void save_result(const LevelSetType & phi, std::string evodir, int iter);

		//@}

	protected:	

		inline float _g(float v, float p)
		{
			return 1/(1+pow(v,p));
		}

		/** @name Member data  */
		//@{

		ImageFloat   m_g;             ///< Metric of the riemanian space
		ImageFloat   m_g_deriv[DIM];  ///< Directional derivatives of the metric
		float        m_v;             ///< Balloon force

		//@}

	};


	/**  @example 2D_GeodesicActiveContour.C Geodesic Active Contour Segmentation of a 2D Image */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <unsigned DIM, typename PixelImage, typename Pixel>
		GeodesicActiveContourModule<DIM, PixelImage, Pixel>::GeodesicActiveContourModule(ImageType * _data, float _weight, float _v, float _smooth, float _cut)
		:SegmentationModule<DIM, Pixel>(_weight, false), m_g(_data->shape()), m_v(_v)
	{ 
		for (int i=0 ; i<DIM ; i++)
			m_g_deriv[i].resize(_data->shape());

		compute_metric(_data, _smooth, _cut);
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		GeodesicActiveContourModule<DIM, PixelImage, Pixel>::~GeodesicActiveContourModule() 
	{ }


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Modify/Access member data


	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GeodesicActiveContourModule<DIM, PixelImage, Pixel>::compute_metric(ImageType * _data, float _smooth, float _cut)
	{
		// smoothed image 
		ImageType image_smooth(_data->shape(), ColumnMajorArray<DIM>());
		image_smooth = _data->copy();
		Globals::blur(image_smooth, _smooth);

		// metric computation : g = 1 / (1 + sqrt(|grad I|) )
		if (!Globals::same_shape<DIM>(m_g.shape(), _data->shape())) m_g.resize(_data->shape());
		for (typename ImageFloat::iterator ind=m_g.begin() ; ind!=m_g.end() ; ++ind )
			*ind = _g(Schemes<DIM,PixelImage>::norm_gradient(image_smooth, ind.position()), 0.5f);

		// -0.2 to assure a range of value where the metric is 0
		Globals::normalize<DIM, PixelImage>(m_g, 0, 1);
		Globals::cut<DIM, PixelImage>(m_g, _cut/100.0, 1);
		Globals::normalize<DIM, PixelImage>(m_g, 0, 1);

		// directional derivative of the metric
		for (int i=0 ; i<DIM ; i++)
			if (!Globals::same_shape<DIM>(m_g_deriv[i].shape(), _data->shape())) m_g_deriv[i].resize(_data->shape());

		PixelImage grad[DIM];
		for (typename ImageFloat::iterator ind=m_g.begin() ; ind!=m_g.end() ; ++ind ){
			Schemes<DIM, float>::gradient(m_g, ind.position(), grad);
			for (int i=0 ; i<DIM ; i++)
				m_g_deriv[i](ind.position()) = grad[i];
		}

	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GeodesicActiveContourModule<DIM, PixelImage, Pixel>::set_metric(const ImageFloat & g, ImageFloat * g_deriv)
	{
		m_g = g;

		if (g_deriv != NULL)
			for (int i=0 ; i<DIM ; i++)
				m_g_deriv[i] = g_deriv[i];
			
		else {
			// directional derivative of the metric
			for (int i=0 ; i<DIM ; i++)
				if (m_g_deriv[i].shape() != m_g.shape()) m_g_deriv[i].resize(m_g.shape());

			PixelImage grad[DIM];
			for (typename ImageFloat::iterator ind=m_g.begin() ; ind!=m_g.end() ; ++ind ){
				Schemes<DIM, float>::gradient(m_g, ind.position(), grad);
				for (int i=0 ; i<DIM ; i++)
					m_g_deriv[i](ind.position()) = grad[i];
			}
		}
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		float& GeodesicActiveContourModule<DIM, PixelImage, Pixel>::balloon()
	{
		return m_v;
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		const typename GeodesicActiveContourModule<DIM, PixelImage, Pixel>::ImageFloat & GeodesicActiveContourModule<DIM, PixelImage, Pixel>::metric()
	{
		return m_g;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Member functions


	template <unsigned DIM, typename PixelImage, typename Pixel>
		float GeodesicActiveContourModule<DIM, PixelImage, Pixel>::value(const LevelSetType & phi, const Index & ind)
	{
		Pixel grad[DIM];
		Pixel gnorm = 0;

		Schemes<DIM,Pixel>::gradient(phi, ind, grad);
		for (int i=0 ; i<DIM ; i++)
			gnorm += (grad[i])*(grad[i]);
		gnorm = sqrt(gnorm);

		float result = m_g(ind)*gnorm*(Schemes<DIM,Pixel>::mean_curvature(phi,ind) - m_v);
		for (int i=0 ; i<DIM ; i++)
			result += grad[i]*m_g_deriv[i](ind);

		return result;
	}	 

	template <unsigned DIM, typename PixelImage, typename Pixel>
		float GeodesicActiveContourModule<DIM, PixelImage, Pixel>::energy(const LevelSetType & phi, const Index & ind)
	{
		return m_g(ind) * Globals::delta(phi(ind));
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GeodesicActiveContourModule<DIM, PixelImage, Pixel>::init(const LevelSetType & phi)
	{
		base::init(phi);

		if (!Globals::same_shape<DIM>(m_g.shape(), phi.shape())) {
			std::cerr << "Error : Shape of the Level Set and shape of the metric don't match"  << std::endl;
			exit(1);
		}
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GeodesicActiveContourModule<DIM, PixelImage, Pixel>::save_result(const LevelSetType & phi, std::string evodir, int iter)
	{ }

}

#endif // GEODESICACTIVECONTOURMODULE_H
