/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef GRADIENTVECTORFLOWODULE_H
#define GRADIENTVECTORFLOWODULE_H

#include <SegmentationModule/SegmentationModule.h>

namespace segmentation {


	/**
	* @author Jerome Piovano
	*
	* @brief SegmentationModule specialization used to drive the LevelSet using a gradient vector flow evolution
	*
	* @note default values are : m_v=0; smooth=3.0f
	*/
	template <unsigned DIM, typename PixelImage, typename Pixel = float> 
	class GradientVectorFlowModule : public SegmentationModule<DIM, Pixel>
	{
	public:

		typedef          SegmentationModule<DIM, Pixel>  base;
		typedef typename base::LevelSetType              LevelSetType;
		typedef typename base::Index                     Index;

		typedef          Array<PixelImage, DIM> 	 ImageType;
		typedef          Array<float, DIM>	         ImageFloat;


		/** @name Constructors - Destructor */
		//@{

		/// Default constructor
		GradientVectorFlowModule(ImageType *  _data,                ///< image whose metric will be computed
					 float        _weight = 1.f,        ///< weight of the module
					 float        _flow_smooth = 0.2f,  ///< smoothness of the gradient vector flow
					 float        _smooth = 3.0f,       ///< amount of smoothness on the image
					 int          _iter = 100,          ///< # of iterations of the reinitialization PDE
					 float        _dt = 0.2             ///< time step of the reinitialization PDE
					 );

		/// Destructor
		virtual ~GradientVectorFlowModule();

		//@}
		/** @name Modify/Access member data */
		//@{

		/// Compute the gradient vector flow from the image
		void compute_vector_flow(ImageType * _data, float _flow_smooth = 0.2f, float _smooth = 3.0f, int _iter = 2, float _dt = 0.4);
		/// Set the gradient vector flow from the image tab
		void set_vector_flow(ImageFloat * flow_component);
		/// Components of the vector flow
		const ImageFloat & vector_flow_component(int i);

		//@}
		/** @name Member functions */
		//@{

		/// Value of the module at a point
		float value(const LevelSetType & phi, const Index & ind);
		/// Energy of the module at a point
		float energy(const LevelSetType & phi, const Index & ind);
		/// Initialize the module 
		void init(const LevelSetType & phi);
		/// Save results depending on the module data
		void save_result(const LevelSetType & phi, std::string evodir, int iter);

		//@}

	protected:	

		/** @name Member data  */
		//@{

		ImageFloat   m_g;                    ///< Metric of the riemanian space
		ImageFloat   m_grad_g;               ///< Metric of the riemanian space
		ImageFloat   m_g_deriv[DIM];         ///< Directional derivatives of the metric
		ImageFloat   m_vector_flow[DIM];     ///< Gradient vector flow

		//@}

	};


	/**  @example 2D_GradientVectorFlow.C Geodesic Active Contour Segmentation of a 2D Image */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <unsigned DIM, typename PixelImage, typename Pixel>
		GradientVectorFlowModule<DIM, PixelImage, Pixel>::GradientVectorFlowModule(ImageType * _data, float _weight, float _flow_smooth, float _smooth, int _iter, float _dt)
		:SegmentationModule<DIM, Pixel>(_weight, false), m_g(_data->shape()), m_grad_g(_data->shape())
	{ 
		for (int i=0 ; i<DIM ; i++){
			m_g_deriv[i].resize(_data->shape());
			m_vector_flow[i].resize(_data->shape());
		}

		compute_vector_flow(_data, _flow_smooth, _smooth, _iter, _dt);
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		GradientVectorFlowModule<DIM, PixelImage, Pixel>::~GradientVectorFlowModule() 
	{ }


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Modify/Access member data


	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GradientVectorFlowModule<DIM, PixelImage, Pixel>::compute_vector_flow(ImageType * _data, float _flow_smooth, float _smooth, int _iter, float _dt)
	{
		// test CFL
		if (_dt > 0.25/_flow_smooth){
			std::cerr << "CFL condition (_dt < 1/(4*_flow_smooth)) not respected " << std::endl;
			_dt = 0.25/_flow_smooth - 0.05;
			std::cerr << "changing dt to " << _dt << std::endl;
		}
	
		// smoothed image 
		ImageType image_smooth(_data->shape(), ColumnMajorArray<DIM>());
		image_smooth = _data->copy();
		Globals::blur(image_smooth, _smooth);

		// metric computation : g = sq(|grad I|) )
		if (!Globals::same_shape<DIM>(m_g.shape(), _data->shape())) m_g.resize(_data->shape());
		for (typename ImageFloat::iterator ind=m_g.begin() ; ind!=m_g.end() ; ++ind )
			*ind = Schemes<DIM,PixelImage>::sq_norm_gradient(image_smooth, ind.position());

		Globals::normalize<DIM, PixelImage>(m_g, 0, 5);

		// directional derivative of the metric
		PixelImage grad[DIM];
		for (typename ImageFloat::iterator ind=m_g.begin() ; ind!=m_g.end() ; ++ind ){
			Schemes<DIM, float>::gradient(m_g, ind.position(), grad);
			m_grad_g(ind.position()) = 0;
			for (int i=0 ; i<DIM ; i++){
				m_g_deriv[i](ind.position()) = grad[i];
				m_grad_g(ind.position()) += (grad[i])*(grad[i]);
			}
		}

		timeval _time;
		Globals::initChrono(_time);

		int width  = _data->shape()(1);
		int height = _data->shape()(2); 
		int depth  = (DIM >= 3)?_data->shape()(3) : 1;

		// init gradient vector flow
		ImageFloat  temp[DIM];
		for (int i=0 ; i<DIM ; i++){
			temp[i].resize(_data->shape());
			m_vector_flow[i] = 0;
		}

		// gradient vector flow
		for (int t=0 ; t<_iter ; t++){
			for (int i=0 ; i<DIM ; i++){
				for (typename ImageType::iterator ind=_data->begin() ; ind!=_data->end() ; ++ind )
				 	temp[i](ind.position()) = _flow_smooth*Schemes<DIM,PixelImage>::laplacian(m_vector_flow[i], ind.position()) - (m_vector_flow[i](ind.position()) - m_g_deriv[i](ind.position()))*m_grad_g(ind.position());

				for (typename ImageType::iterator ind=_data->begin() ; ind!=_data->end() ; ++ind )
				 	m_vector_flow[i](ind.position()) += _dt* temp[i](ind.position());
			}
		}

		Globals::printChrono("\n	=> Gradient Vector Flow computation : ", _time);
		Globals::initChrono(_time);

		for (int i=0 ; i<DIM ; i++)
			for (typename ImageType::iterator ind=_data->begin() ; ind!=_data->end() ; ++ind )
				m_vector_flow[i](ind.position()) *= 10;
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GradientVectorFlowModule<DIM, PixelImage, Pixel>::set_vector_flow(ImageFloat * flow_component)
	{
		for (int i=0 ; i<DIM ; i++)
			m_vector_flow[i] = flow_component[i];
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		const typename GradientVectorFlowModule<DIM, PixelImage, Pixel>::ImageFloat & GradientVectorFlowModule<DIM, PixelImage, Pixel>::vector_flow_component(int i)
	{
		return m_vector_flow[i];
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Member functions


	template <unsigned DIM, typename PixelImage, typename Pixel>
		float GradientVectorFlowModule<DIM, PixelImage, Pixel>::value(const LevelSetType & phi, const Index & ind)
	{
		Pixel GVF[DIM];

		for (int i=0 ; i<DIM ; i++)
			GVF[i] = m_vector_flow[i](ind);
		
		return LevelSetType::FluxesType::advection(phi, ind, GVF);
	}	 

	template <unsigned DIM, typename PixelImage, typename Pixel>
		float GradientVectorFlowModule<DIM, PixelImage, Pixel>::energy(const LevelSetType & phi, const Index & ind)
	{
		return m_g(ind) * Globals::delta(phi(ind));
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GradientVectorFlowModule<DIM, PixelImage, Pixel>::init(const LevelSetType & phi)
	{
		base::init(phi);

		if (!Globals::same_shape<DIM>(m_g.shape(), phi.shape())) {
			std::cerr << "Error : Shape of the Level Set and shape of the metric don't match"  << std::endl;
			exit(1);
		}
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void GradientVectorFlowModule<DIM, PixelImage, Pixel>::save_result(const LevelSetType & phi, std::string evodir, int iter)
	{ }

}

#endif // GRADIENTVECTORFLOWODULE_H
