/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef HISTOGRAM_H
#define HISTOGRAM_H

#include <boost/format.hpp>
namespace segmentation {

	/**
	* @author Jerome Piovano
	*
	* @brief histogram of scalar regions
	*
	*/
	template <typename Pixel = float> 
	class Histogram
	{
	public:

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor
		Histogram( Pixel     _min = 0,          ///< minimum value of the histogram
			   Pixel     _max = 256,        ///< maximum value of the histogram
			   bool      _adaptive = true,  ///< if the histogram is adaptive
			   int *     _data = NULL       ///< datas of the histogram, of size (_max - _min)
		           );

		/// Copy constructor
		template <typename Pixel2>
		Histogram(const Histogram<Pixel2>& _hist) 
		:m_min(_hist.m_min), m_max(_hist.m_max), m_adaptive(_hist.m_adaptive), m_nb_pix(0), m_hist(NULL)
		{
      copy_histogram( _hist.m_min, _hist.m_max, _hist.data() );
		}

		/// Operator=
		Histogram& operator=(int * _data);

		virtual ~Histogram();

		//@}
		/** @name Access the Histogram datas */
		//@{

		/// Minimum of the histogram
		Pixel  min();
		/// Maximum of the histogram
		Pixel  max();
		/// size of the histogram
		int size();
		/// Array representing the histogram
		int * data() const;
		/// Number of pixel inside the region
		int  nb_pixels();
		/// Return the index in the array of the value _val, bounded by min and max
		int get_index(Pixel _val);
		/// Number of pixels of value _val
		int operator[](Pixel _val);

    /// save allocation of histogram memory, if needed. Attention: values are allocated, but no t initialized!
    void create_histogram( Pixel _min, Pixel _max );
    
    /// save deep-copy. expects src to be of size max-min
    void copy_histogram( Pixel _min, Pixel _max, const int* src );
    
		//@}
		/** @name Modify the Histogram */
		//@{

		/// True if the histogram is adaptative
		bool& adaptative();
		/// Clear the histogram, if adaptative=true
		void clear();
		/// frees histogram memory
    void free();
    /// Resize the histogram, if adaptative=true
		void resize(Pixel _min, Pixel _max);
		/// Add a pixel of value _val, if adaptative=true
		void add_pixel(Pixel _val);
		/// Remove a pixel of value _val, if adaptative=true
		void remove_pixel(Pixel _val);

		//@}
		/** @name Statistics computation */
		//@{

		/// compute the mean, and the variance if compute_variance=true
		void compute_statistics(float& mean, float& var, bool compute_variance=false);

		//@}
		/** @name Member data  */
		//@{

		int  *  m_hist;          ///< Histogram
		bool    m_adaptive;      ///< Adaptive statistics
		Pixel   m_min;	         ///< Minimum of the histogram
		Pixel   m_max;	         ///< Maximum of the histogram
		int     m_nb_pix;        ///< Number of pixel inside the histogram region
	};


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <typename Pixel>
	  Histogram<Pixel>::Histogram( Pixel _min, Pixel _max, bool _adaptive, int * _data)
	  : m_adaptive(_adaptive)
	  , m_min(_min)
	  , m_max(_max)
	  , m_nb_pix(0)
    , m_hist(NULL)
	{
    copy_histogram( _min, _max, _data );
	}

template <typename Pixel>
void Histogram<Pixel>::free()
{
  if ( m_hist )
  {
    delete [] m_hist;
    m_hist = NULL;
  }
}

template <typename Pixel>
void Histogram<Pixel>::create_histogram( Pixel _min, Pixel _max )
{
  if ( _min==m_min && _max==m_max && m_hist )
    return;
    
  free();
  
  size_t size = static_cast<int>(m_max - m_min);
  
  //if ( size == 0 ) throw std::runtime_error( str( boost::format("ERROR in create_histogram: creating zero-sized histogram with min=%f, max=%f, size=%d")%_min%_max%size)  );
  //if ( size < 0  ) throw std::runtime_error( str( boost::format("ERROR in create_histogram: min=%f is greater than max=%f, size=%d !!!")%_min%_max%size) );
  if ( size <= 0 )
  {
    m_hist = NULL;
  }
  else
    m_hist = new int[ size ];
}


template <typename Pixel>
void Histogram<Pixel>::copy_histogram( Pixel _min, Pixel _max, const int* _src )
{
  create_histogram( _min, _max );
  
  m_nb_pix = 0;
  if (_src == NULL)
    for (int i = 0; i < static_cast<int>(m_max-m_min) ; i++)
      m_hist[i] = 0;
  else
    for (int i = 0; i < static_cast<int>(m_max-m_min) ; i++){
        m_hist[i] = _src[i];
        m_nb_pix += _src[i];
      }
  
}

  
  
	template <typename Pixel>
		Histogram<Pixel> & Histogram<Pixel>::operator=(int * _data)
	{
		copy_histogram( m_min, m_max, _data );
	}

	template <typename Pixel>
		Histogram<Pixel>::~Histogram()
	{
		free();
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Access the Histogram datas


	template <typename Pixel>
		Pixel Histogram<Pixel>::min()
	{
		return m_min;
	}

	template <typename Pixel>
		Pixel Histogram<Pixel>::max()
	{
		return m_max;
	}

	template <typename Pixel>
		int Histogram<Pixel>::size()
	{
		return static_cast<int>(m_max-m_min);
	}

	template <typename Pixel>
		int * Histogram<Pixel>::data() const
	{
		return m_hist;
	}
	
	template <typename Pixel>
		int Histogram<Pixel>::nb_pixels()
	{
		return m_nb_pix;
	}

	template <typename Pixel>
		int Histogram<Pixel>::get_index(Pixel _val)
	{
		if      (_val <  m_min  ) return 0;
		else if (_val >= size() ) return size() - 1;
	
		return static_cast<int>(_val - m_min);
	}
	
	 template <typename Pixel>
		int Histogram<Pixel>::operator[](Pixel _val)
	{
		return m_hist[get_index(_val)];
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//Modify the Histogram


	 template <typename Pixel>
		bool& Histogram<Pixel>::adaptative()
	{
		 return m_adaptive;
	}
	
	template <typename Pixel>
		void Histogram<Pixel>::clear()
	{
		if  (m_adaptive){
			m_nb_pix = 0;
			for (int i = 0; i < static_cast<int>(m_max-m_min) ; i++)
				m_hist[i] = 0;
		}
	}
		
	template <typename Pixel>
		void Histogram<Pixel>::resize(Pixel _min, Pixel _max)
	{
		if (m_adaptive){
			create_histogram( _min, _max );
		}
	}
	
	template <typename Pixel>
		void Histogram<Pixel>::add_pixel(Pixel _val)
	{
		if (m_adaptive){
#ifdef DEBUG
      if ( !m_hist )                    throw std::runtime_error("Histogram<Pixel>::add_pixel: m_hist = NULL!");
      if ( get_index(_val) > size()-1 ) throw std::runtime_error( str( boost::format("Histogram<Pixel>::add_pixel: index out of bounds: min=%f, max=%f, size=%d, get_index(%f)=%d ")%m_min %m_max %size() %_val %get_index(_val) ) );
#endif
			m_hist[get_index(_val)]++;
			m_nb_pix++;
		}
	}

	 template <typename Pixel>
		void Histogram<Pixel>::remove_pixel(Pixel _val)
	{
		if (m_adaptive && m_hist[get_index(_val)]>0){
			m_hist[get_index(_val)]--;
			m_nb_pix--;
		}
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//Statistics computation


	template <typename Pixel>
		void Histogram<Pixel>::compute_statistics(float& mean, float& var, bool compute_variance)
	{
#if 1				
		double w0 = 0.001;
		double x20 = 1;
		if ( compute_variance )
		{
			double w = w0, wx = 0, wx2 = w0 * x20;
			for (int i = int(m_min); i < int(m_max) ; i++)
			{
				float n = operator[](i);
				w   += n;
				wx  += i * n;
				wx2 += i * i * n;
			}			
			mean = wx / w;
			var = std::max( 1e-8, wx2 / w - mean * mean );
			if ( var == 0 )
			{
				std::cout << "Error: mean = "<<mean << "  var = "<< var << " wx2 = "<<wx2 << " w = "<< w << " min = "<< int(m_min) << " max = "<< int(m_max)<< " compute variance = "<< ( compute_variance?"true":"false" )<<std::endl;
				throw std::runtime_error("logical error in histogram");
			}
		}
		else
		{
			double w = w0, wx = 0;
			for (int i = int(m_min); i < int(m_max) ; i++)
			{
				float n = operator[](i);
				w   += n;
				wx  += i * n;
			}			
			mean = wx / w;
		}
		
#else
		mean = 0;

		for (int i = static_cast<int>(m_min) ; i < static_cast<int>(m_max) ; i++)
			mean += i * operator[](i);
		mean /= static_cast<float>(m_nb_pix);

		if (compute_variance)
    {
      /// in abcence of data, fall back to prior
      const float n0 = 0.01f;
      const float dx0 = 0.01f;
			var = dx0 * dx0 * n0;
			for (int i = static_cast<int>(m_min) ; i < static_cast<int>(m_max) ; i++)
      {
        float d = i - static_cast<float>(mean);
			  var += operator[](i) * d * d;
      }
			var /= ( static_cast<float>(m_nb_pix) + n0 );
		}
	#endif

	}
}

#endif // HISTOGRAM_H
