/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef LOCALSTATSREGIONMODULE_H
#define LOCALSTATSREGIONMODULE_H

#include <SegmentationModule/RegionModule.h>
#include <Schemes.h>

namespace segmentation {


	/**
	* @author Jerome Piovano
	*
	* @brief RegionModule specialization used to drive the LevelSet using a local stats Mumford Shah region evolution
	*
	*/
	template <unsigned DIM, typename PixelImage, typename Pixel = float>
	class LocalGaussianRegionModule : public RegionModule<DIM, Pixel, PixelImage> 
	{
	public:

		typedef          RegionModule<DIM, Pixel, PixelImage>    base;
		typedef typename base::LevelSetType                      LevelSetType;
		typedef typename base::Index                             Index;
		typedef typename base::ImageType                         ImageType;

		typedef          Array<float, DIM>	                 ImageFloat;

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor
		LocalGaussianRegionModule(ImageType & _data,                    ///< image to segment
				          float       _weight = 1.f,            ///< global weight of the module
				          float       _sigma = 20.0f,           ///< variance of gaussian
				          bool        _var = false              ///< use of local variance
				          );

		/// Destructor
		virtual ~LocalGaussianRegionModule();

		//@}
		/** @name Modify/Access member data */
		//@{

		/// set the variance of the blur
		void set_sigma(float _sigma);
		/// use of local variance
		void use_variance(bool _var=true);

		//@}
		/** @name Point dependent functions */
		//@{

		/// Value of the module at a point
		float value(const LevelSetType & phi, const Index & ind);
		/// Energy of the module at a point
		float energy(const LevelSetType & phi, const Index & ind);
		/// Initialize the module 
		void init(const LevelSetType & phi);
		/// update the module 
		void update_after_reinit(const LevelSetType & phi);
		/// save results depending on the module data
		void save_result(const LevelSetType & phi, std::string evodir, int iter);

		/// update the parameters of the module
		void update_params(const LevelSetType & phi);
	
		//@}
		/** @name Member data  */
		//@{

	protected:

		float          m_sigma;            ///< size of blur

		ImageFloat     m_chi_pos;          ///< smoothed binary mask of internal region
		ImageFloat     m_chi_neg;          ///< smoothed binary mask of external region
		ImageFloat     m_mean_pos;         ///< local mean of the positive region
		ImageFloat     m_mean_neg;         ///< local mean of the negative region
		ImageFloat     m_var_pos;          ///< local variance of the positive region
		ImageFloat     m_var_neg;          ///< local variance of the negative region 
		float          m_image_variance;   ///< global variance (for speed normalization)
		bool           m_var;              ///< true if using local variance

		//@}

	};


	/**  @example 2D_LocalGaussian.C Local Stats Region Segmentation of a 2D Image */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <unsigned DIM, typename Pixel, typename PixelImage>
		LocalGaussianRegionModule<DIM, Pixel, PixelImage>::LocalGaussianRegionModule(ImageType & _data, float _weight, float _sigma, bool _var) 
	  : RegionModule<DIM, Pixel, PixelImage>(_data, _weight, true)
	  , m_sigma(_sigma)
	  , m_chi_pos(_data.shape(), ColumnMajorArray<DIM>())
	  , m_chi_neg(_data.shape(), ColumnMajorArray<DIM>())
	  , m_mean_pos(_data.shape(), ColumnMajorArray<DIM>())
	  , m_mean_neg(_data.shape(), ColumnMajorArray<DIM>())
	  , m_var_pos(_data.shape(), ColumnMajorArray<DIM>())
	  , m_var_neg(_data.shape(), ColumnMajorArray<DIM>())
	  , m_image_variance(-1.0f), m_var(_var)
	{ }

	template <unsigned DIM, typename Pixel, typename PixelImage>
		LocalGaussianRegionModule<DIM, Pixel, PixelImage>::~LocalGaussianRegionModule() 
	{ }


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Modify/Access member data

	
	template <unsigned DIM, typename Pixel, typename PixelImage>
		void LocalGaussianRegionModule<DIM, Pixel, PixelImage>::set_sigma(float _sigma)
	{
		m_sigma = _sigma;
	}

	template <unsigned DIM, typename Pixel, typename PixelImage>
		void LocalGaussianRegionModule<DIM, Pixel, PixelImage>::use_variance(bool _var)
	{
		m_var = _var;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Member functions

	
	template <unsigned DIM, typename Pixel, typename PixelImage>
		float LocalGaussianRegionModule<DIM, Pixel, PixelImage>::value(const LevelSetType & phi, const Index & ind)
	{ 
		PixelImage val = base::m_data(ind);

		float mean_pos = m_mean_pos(ind)/m_chi_pos(ind);
		float mean_neg = m_mean_neg(ind)/m_chi_neg(ind);

		float var_pos = m_var ? m_var_pos(ind)/m_chi_pos(ind) - mean_pos*mean_pos : m_image_variance;
		float var_neg = m_var ? m_var_neg(ind)/m_chi_neg(ind) - mean_neg*mean_neg : m_image_variance;

		var_pos = (var_pos<m_image_variance/2)? m_image_variance/2 : var_pos;
		var_neg = (var_neg<m_image_variance/2)? m_image_variance/2 : var_neg;

		float ll_pos = - (val-mean_pos)*(val-mean_pos)/100.0;
		float ll_neg = - (val-mean_neg)*(val-mean_neg)/100.0;

		// (* signe_evolution) to keep the coherence in the evolution 
		float result =   /* ((mean_pos-mean_neg > 0)? -1 : 1) * */ (ll_pos - ll_neg)/3.0;

		return  result;
	}	

	template <unsigned DIM, typename Pixel, typename PixelImage>
		float LocalGaussianRegionModule<DIM, Pixel, PixelImage>::energy(const LevelSetType & phi, const Index & ind)
	{

		Pixel      _phi = phi(ind);
		PixelImage _val = base::m_data(ind);
		float mean = 0;
		float var  = 0;

		if (_phi>=0){
			mean = m_mean_pos(ind)/m_chi_pos(ind);
			var  = m_var ? m_var_pos(ind)/m_chi_pos(ind) - mean*mean : m_image_variance;
			var = (var<m_image_variance/2)? m_image_variance/2 : var;
		}
		else {
			mean = m_mean_neg(ind)/m_chi_neg(ind);
			var  = m_var ? m_var_neg(ind)/m_chi_neg(ind) - mean*mean : m_image_variance;
			var = (var<m_image_variance/2)? m_image_variance/2 : var;
		}

		return (_val-mean)*(_val-mean);
	}

	template <unsigned DIM, typename Pixel, typename PixelImage>
		void LocalGaussianRegionModule<DIM, Pixel, PixelImage>::init(const LevelSetType & phi)
	{
		base::init(phi);
		update_params(phi);
	}

	template <unsigned DIM, typename Pixel, typename PixelImage>
		void LocalGaussianRegionModule<DIM, Pixel, PixelImage>::update_after_reinit(const LevelSetType & phi)
 	{ 
		update_params(phi);
	}

	 template <unsigned DIM, typename Pixel, typename PixelImage>
		 void LocalGaussianRegionModule<DIM, Pixel, PixelImage>::update_params(const LevelSetType & phi) 
	{
		// computation of the variance of the image
		if(m_image_variance < 0.f) {
			PixelImage min;
			PixelImage max;
			float mean;
			
			Globals::stats<DIM,PixelImage>(base::m_data, min, max, mean, &m_image_variance);
		}

		// computation of m_chi_pos, m_mean_pos, m_mean_neg
		for (typename ImageType::iterator i=base::m_data.begin() ; i!=base::m_data.end() ; ++i ){
			Index ind = i.position();
			const Pixel      _phi = phi(ind);
			const PixelImage _val = *i;

			if (_phi>=0) {
				m_chi_pos(ind)   = 1.0f;
				m_chi_neg(ind)   = 0.0f;
				m_mean_pos(ind)  = _val;
				m_mean_neg(ind) = 0.0f;
				m_var_pos(ind) = _val*_val;
				m_var_neg(ind) = 0.0f;
			}
			else {
				m_chi_pos(ind)   = 0.0f;
				m_chi_neg(ind)   = 1.0f;
				m_mean_pos(ind)  = 0.0f;
				m_mean_neg(ind) = _val;
				m_var_pos(ind) = 0.0f;
				m_var_neg(ind) = _val*_val;
			}
		}

#pragma omp parallel sections
{
	#pragma omp section
		Globals::blur(m_chi_pos , m_sigma);
	#pragma omp section
		Globals::blur(m_chi_neg , m_sigma);
	#pragma omp section
		Globals::blur(m_mean_pos , m_sigma);
	#pragma omp section
		Globals::blur(m_mean_neg , m_sigma);
}

		if(m_var) {
#pragma omp parallel sections
{
	#pragma omp section
		Globals::blur(m_var_pos , m_sigma);
	#pragma omp section
		Globals::blur(m_var_neg , m_sigma);
}
		}
	}

	template <unsigned DIM, typename Pixel, typename PixelImage>
		void LocalGaussianRegionModule<DIM, Pixel, PixelImage>::save_result(const LevelSetType & phi, std::string evodir, int iter)
	{ }
}

#endif // LOCALSTATSREGIONMODULE_H
