/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef PARZENREGIONMODULE_H
#define PARZENREGIONMODULE_H

#include <SegmentationModule/RegionModule.h>
#include <vector>

namespace segmentation {


	/**
	* @author Mikael Rousson
	*
	* @brief RegionModule specialization used to drive the LevelSet using a Parzen region evolution
	*
	*/
	template <unsigned DIM, typename PixelImage, typename Pixel = float> 
	class ParzenRegionModule : public RegionModule<DIM, PixelImage, Pixel>
	{
	public:
	
	 	typedef          RegionModule<DIM, PixelImage, Pixel>    base;
		typedef typename base::LevelSetType                      LevelSetType;
		typedef typename base::Index                             Index;
		typedef typename base::HistogramType                     HistogramType;
		typedef typename base::ImageType                         ImageType;

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor ( with global weight )
		ParzenRegionModule(ImageType & _data,                ///< image to segment
				   float       _weight = 1.f,        ///< weight of the module
				   float       _sigma = 4,           ///< variance of the parzen kernel
				   bool        _adaptive = true      ///< if the module is adaptive
				   );

		/// Destructor
		virtual ~ParzenRegionModule();

		//@}
		/** @name Modify/Access member data */
		//@{

		/// set the variance of the parzen kernel
		void set_sigma(float _sigma);

		//@}
		/** @name Member functions */
		//@{

		/// Value of the module at a point
		float value(const LevelSetType & phi, const Index & ind);
		/// Energy of the module at a point
		float energy(const LevelSetType & phi, const Index & ind);
		/// Initialize the module 
		void init(const LevelSetType & phi);
		/// update the module 
		void update_before_reinit(const LevelSetType & phi);
		/// save results depending on the module data
		void save_result(const LevelSetType & phi, std::string evodir, int iter);

		/// update the parameters, and precompute the log-likelihood of all possible values
		void update_parzen_params();

		//@}
		/** @name Member data  */
		//@{

	protected:

		float                 m_sigma;       ///< Variance of the parzen kernel
		std::vector<Pixel>    m_ll_pos;      ///< Log-likelihoods for the positive region
		std::vector<Pixel>    m_ll_neg;      ///< Log-likelihoods for the negative region
		Array<float, 1>       m_parzen_pos;  ///< Parzen distribution of the positive region
		Array<float, 1>       m_parzen_neg;  ///< Parzen distribution of the negative region
		
		//@}
	};


	/**  @example 2D_Parzen.C Parzen Region Segmentation of a 2D Image */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <unsigned DIM, typename PixelImage, typename Pixel>
		ParzenRegionModule<DIM, PixelImage, Pixel>::ParzenRegionModule(ImageType & _data, float _weight, float _sigma, bool _adaptive)
		:RegionModule<DIM, PixelImage, Pixel>(_data, _weight, _adaptive), m_sigma(_sigma)
	{ }

	template <unsigned DIM, typename PixelImage, typename Pixel>
		ParzenRegionModule<DIM, PixelImage, Pixel>::~ParzenRegionModule() 
	{ }


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Modify/Access member data 


	template <unsigned DIM, typename PixelImage, typename Pixel>
		void ParzenRegionModule<DIM, PixelImage, Pixel>::set_sigma(float _sigma)
	{ 
		m_sigma = _sigma;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Member functions


	template <unsigned DIM, typename PixelImage, typename Pixel>
		float ParzenRegionModule<DIM, PixelImage, Pixel>::value(const LevelSetType & phi, const Index & ind)
	{
		PixelImage val = base::m_data(ind);
		const Pixel ll_pos = m_ll_pos[base::m_hist_pos->get_index(val)];
		const Pixel ll_neg = m_ll_neg[base::m_hist_neg->get_index(val)];

		return  Globals::delta(phi(ind)) * (ll_pos - ll_neg);
	}	

	template <unsigned DIM, typename PixelImage, typename Pixel>
		float ParzenRegionModule<DIM, PixelImage, Pixel>::energy(const LevelSetType & phi, const Index & ind)
	{
		PixelImage val = base::m_data(ind);
		float ll_pos = m_ll_pos[base::m_hist_pos->get_index(val)];
		float ll_neg = m_ll_neg[base::m_hist_neg->get_index(val)];
		float h = Globals::heaviside(phi(ind));

		return -(1-h)*ll_pos - h*ll_neg;
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void ParzenRegionModule<DIM, PixelImage, Pixel>::init(const LevelSetType & phi)
	{
		base::init(phi);
		base::init_histograms(phi);
		update_parzen_params();
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void ParzenRegionModule<DIM, PixelImage, Pixel>::update_before_reinit(const LevelSetType & phi)
 	{ 
		base::update_histograms(phi);
		update_parzen_params();
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void ParzenRegionModule<DIM, PixelImage, Pixel>::update_parzen_params()
	{ 
		// if the size of the histogram has changed, resize parzen densities and log-likelihood
		if (m_parzen_neg.size() != base::hist_neg()->size()){
			m_parzen_neg.resize(base::hist_neg()->size());
			m_ll_neg.resize(base::hist_neg()->size());
		}
		if (m_parzen_pos.size() != base::hist_pos()->size()){
			m_parzen_pos.resize(base::hist_pos()->size());
			m_ll_pos.resize(base::hist_pos()->size());
		}

		// initialize the parzen densities from the histograms
		for(int i=0; i < m_parzen_neg.size() ; i++) m_parzen_neg(i) = base::hist_neg()->data()[i];
		for(int i=0; i < m_parzen_pos.size() ; i++) m_parzen_pos(i) = base::hist_pos()->data()[i];

		// smooth the parzen densities
		Globals::blur(m_parzen_neg, m_sigma);
		Globals::blur(m_parzen_pos, m_sigma);

		// normalize the parzen densities
		float sum_neg=EPS;
		float sum_pos=EPS;
		for(int i=0; i < m_parzen_neg.size() ; i++) sum_neg += m_parzen_neg(i);
		for(int i=0; i < m_parzen_pos.size() ; i++) sum_pos += m_parzen_pos(i);
		for(int i=0; i < m_parzen_neg.size() ; i++) m_parzen_neg(i) /= sum_neg;
		for(int i=0; i < m_parzen_pos.size() ; i++) m_parzen_pos(i) /= sum_pos;

		// log likelihood computation
		for(int i=0; i < m_ll_neg.size(); i++)  m_ll_neg[i] = log(m_parzen_neg(i)+EPS);
		for(int i=0; i < m_ll_pos.size(); i++) 	m_ll_pos[i] = log(m_parzen_pos(i)+EPS);
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void ParzenRegionModule<DIM, PixelImage, Pixel>::save_result(const LevelSetType & phi, std::string evodir, int iter) 
	{ }
}

#endif // PARZENREGIONMODULE_H
