/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef REGIONMODULE_H
#define REGIONMODULE_H

#include <SegmentationModule/SegmentationModule.h>
#include <SegmentationModule/Histogram.h>

namespace segmentation {


	/**
	* @author Jerome Piovano
	*
	* @brief Virtual SegmentationModule specialization used to drive the LevelSet using some region evolution
	*
	* @note This module has an image in member, and gives some functions to estimate histograms inside and outside a given LevelSet
	* @note You can fix the histogram of each region by setting their 'adaptative' attribute to false
	*
	*/
	template <unsigned DIM, typename PixelImage, typename Pixel = float> 
	class RegionModule : public SegmentationModule<DIM, Pixel>
	{
	public:	

		typedef          SegmentationModule<DIM, Pixel>  base;
		typedef typename base::LevelSetType              LevelSetType;
		typedef typename base::Index                     Index;
		typedef typename base::SchemesType               SchemesType;
		typedef          Histogram<PixelImage>           HistogramType;
		typedef          Array<PixelImage, DIM> 	 ImageType;

		/** @name Constructors - Destructor */
		//@{

		/// Constructor
		RegionModule(ImageType &  _data,              ///< image to segment
			     float        _weight = 1.f,      ///< weight of the module
			     bool         _adaptive = true    ///< if the module is adaptive
		             );

		/// Destructor
		virtual ~RegionModule();

		//@}
		/** @name Modify/Access member data */
		//@{

		/// Set the Image to segment
		void set_data_term(ImageType & _data);
		/// Add a mask to restrict the space of computation of the statistics.
		void set_mask(Array<bool, DIM>* _mask);

		/// Histogram of the positive region
		HistogramType * hist_pos();
		/// Histogram of the negative region
		HistogramType * hist_neg();

		//@}
		/** @name Member functions */
		//@{

		/// Value of the module at a point
		virtual float value(const LevelSetType & phi, const Index & ind){ return 0.0f; }
		/// Energy of the module at a point
		virtual float energy(const LevelSetType & phi, const Index & ind){ return 0.0f; }
		/// Initialize the module 
		virtual void init(const LevelSetType & phi);
		/// Update the module before the Level-set reinitialization
		virtual void update_before_reinit(const LevelSetType & phi){ }
		/// Update the module after the Level-set reinitialization
		virtual void update_after_reinit(const LevelSetType & phi){ }
		/// Save results depending on the module data
		virtual void save_result(const LevelSetType & phi, std::string evodir, int iter){ }

		/// compute the histograms of the image, in the area inside and outside the zero level set
		void init_histograms(const LevelSetType & phi);
		/// update the histograms of the image, in the area inside and outside the zero level set
		void update_histograms(const LevelSetType & phi);
		/// save the histograms
		void save_histograms(std::string evodir, int iter);

		//@}
		/** @name Member data  */
		//@{

	protected:	

		ImageType                m_data;       ///< Image to segment
		Array<bool, DIM> *       m_mask;       ///< Boolean mask where the statistics are computed
		HistogramType *          m_hist_pos;   ///< Histogram of the image in the positive region of the level set
		HistogramType *          m_hist_neg;   ///< Histogram of the image in the negative region of the level set

		//@}

	};


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <unsigned DIM, typename PixelImage, typename Pixel>
		RegionModule<DIM, PixelImage, Pixel>::RegionModule(ImageType & _data, float _weight, bool _adaptive)
		:SegmentationModule<DIM, Pixel>(_weight, _adaptive), m_data(_data), m_mask(NULL), m_hist_pos(NULL), m_hist_neg(NULL)
	{ }

	template <unsigned DIM, typename PixelImage, typename Pixel>
		RegionModule<DIM, PixelImage, Pixel>::~RegionModule() 
	{
		if (m_hist_pos != NULL){
			delete m_hist_pos;
			delete m_hist_neg;
		}
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//  Member functions


	template <unsigned DIM, typename PixelImage, typename Pixel>
		void RegionModule<DIM, PixelImage, Pixel>::set_data_term(ImageType & _data)
	{
	  m_data.reference(_data);
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
 		void RegionModule<DIM, PixelImage, Pixel>::set_mask(Array<bool, DIM>* _mask)
	{		
		if (_mask != NULL && _mask->shape() != m_data.shape()) {
			std::cerr << "Error : Shape of the mask and shape of the image don't match"  << std::endl;
			exit(1);
		}
		m_mask = _mask;
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		typename RegionModule<DIM, PixelImage, Pixel>::HistogramType* RegionModule<DIM, PixelImage, Pixel>::hist_pos()
	{
		return m_hist_pos;
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		typename RegionModule<DIM, PixelImage, Pixel>::HistogramType* RegionModule<DIM, PixelImage, Pixel>::hist_neg()
	{
		return m_hist_neg;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Update of the module  


	template <unsigned DIM, typename PixelImage, typename Pixel>
		void RegionModule<DIM, PixelImage, Pixel>::init(const LevelSetType & phi)
	{
		base::init(phi);

		if (!Globals::same_shape<DIM>(m_data.shape(), phi.shape())) {
			std::cerr << "Error : Shape of the data term [" << m_data.shape()<< "] and shape of the LevelSet ["<< phi.shape()<<"] don't match"  << std::endl;
			exit(1);
		}
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void RegionModule<DIM, PixelImage, Pixel>::init_histograms(const LevelSetType & phi)
	{
		if (m_hist_pos == NULL){

			PixelImage min;
			PixelImage max;
			float mean;
			
			Globals::stats<DIM,PixelImage>(m_data, min, max, mean);

			m_hist_pos = new HistogramType(min, max, true);
			m_hist_neg = new HistogramType(min, max, true);
		}

		//computation of the first histograms

		m_hist_pos->clear();
		m_hist_neg->clear();

		for (typename ImageType::iterator ind=m_data.begin() ; ind!=m_data.end() ; ++ind ){
			const Pixel      _phi = phi(ind.position());
			const PixelImage _val = *ind;

			if (m_mask==NULL || (*m_mask)(ind.position())==true){
				if (_phi>=0) m_hist_pos->add_pixel(_val);
				else         m_hist_neg->add_pixel(_val);
			}
		}
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void RegionModule<DIM, PixelImage, Pixel>::update_histograms(const LevelSetType & phi) 
	{	
		if (m_hist_pos == NULL){
			std::cerr << "Error : Histograms not initialized"  << std::endl;
			exit(1);
		}

		//computation of the histograms using previous ones

		if (!m_hist_pos->adaptative() && !m_hist_neg->adaptative()) return;

		for ( int i = 0 ; i < phi.pos_to_neg_index() ; i++){
			const Index& ind = phi.get_point(i);	
			const PixelImage _val = m_data(ind);

			m_hist_neg->add_pixel(_val);
			m_hist_pos->remove_pixel(_val);
		}

		for ( int i = phi.neg_to_pos_index() ; i < phi.nb_points() ; i++){
			const Index& ind = phi.get_point(i);
			const PixelImage _val = m_data(ind);

			m_hist_neg->remove_pixel(_val);
			m_hist_pos->add_pixel(_val);
		}
	}

	template <unsigned DIM, typename PixelImage, typename Pixel>
		void RegionModule<DIM, PixelImage, Pixel>::save_histograms(std::string evodir, int iter) 
	{ }
}

#endif // REGIONMODULE_H
