/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef REGULARIZATIONMODULE_H
#define REGULARIZATIONMODULE_H

#include <SegmentationModule/SegmentationModule.h>

namespace segmentation {


	/**
	* @author Jerome Piovano
	*
	* @brief SegmentationModule specialization used as a term of regularization to move the LevelSet
	*
	* @note can be used as a Mean Curvature flow by setting its attribute '_mean_curv_motion' to true
	* @note weight <= 1 , for stability
	*/
	template <unsigned DIM, typename Pixel = float> 
	class RegularizationModule : public SegmentationModule<DIM, Pixel>
	{
	public:

		typedef SegmentationModule<DIM, Pixel> base;
		typedef typename base::LevelSetType    LevelSetType;
		typedef typename base::Index           Index;
		typedef typename base::SchemesType     SchemesType;

		/** @name Constructors - Destructor */
		//@{

		/// Default constructor ( with global weight )
		RegularizationModule(float _w = 1.f,                   ///< weight of the module
				     bool  _mean_curv_motion = false   ///< true = mean curvature motion
				     );

		/// Destructor
		virtual ~RegularizationModule();


		//@}
		/** @name Modify/Access member data */
		//@{

		/// True = mean curvature motion
		void mean_curv_motion(bool _b=true);

		//@}
		/** @name Member functions */
		//@{

		/// Value of the module at a point
		float value(const LevelSetType & phi, const Index & ind);
		/// Energy of the module at a point
		float energy(const LevelSetType & phi, const Index & ind);

		//@}
		/** @name Member data  */
		//@{

	protected:	

		bool m_mean_curv_motion;   ///< true = Mean Curvature Motion ( * = |grad(phi)| ) ; false = regularisation term ( *= dirac(phi) );

		//@}

	};


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Constructors - Destructor


	template <unsigned DIM, typename Pixel>
		RegularizationModule<DIM, Pixel>::RegularizationModule(float _w, bool _mean_curv_motion)
		:SegmentationModule<DIM, Pixel>(_w, false), m_mean_curv_motion(_mean_curv_motion)
	{ } 

	template <unsigned DIM, typename Pixel>
		RegularizationModule<DIM, Pixel>::~RegularizationModule() 
	{ }


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Modify/Access member data


	template <unsigned DIM, typename Pixel>
		void RegularizationModule<DIM, Pixel>::mean_curv_motion(bool _b)
	{ 
		m_mean_curv_motion = _b;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Member functions


	template <unsigned DIM, typename Pixel>
		float RegularizationModule<DIM, Pixel>::value(const LevelSetType & phi, const Index & ind)
	{
		return  SchemesType::mean_curvature(phi,ind) * ((m_mean_curv_motion) ? SchemesType::norm_gradient(phi,ind) : Globals::delta(phi(ind)));
	}

	template <unsigned DIM, typename Pixel>
		float RegularizationModule<DIM, Pixel>::energy(const LevelSetType & phi, const Index & ind)
	{
		return 0;
		//return  SchemesType::norm_gradient(phi,ind) * Globals::delta(phi(ind));
	}
}

#endif // REGULARIZATIONMODULE_H
