/*
Copyright INRIA 2011

Authors: Jérôme Piovano, Jean Philippe Pons, Mikael Rousson,
         Theodore Papadopoulo <Theodore.Papadopoulo@inria.fr>

This software is a computer program whose purpose is to implement various
segmentation algorithms based on level sets.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef SEGMENTATIONMODULE_H
#define SEGMENTATIONMODULE_H

#include <LevelSet.h>


namespace segmentation {
  
using namespace levelset;
  
  /**
   * @author Jerome Piovano
   *
   * @brief Virtual module for Segmentation, used to compute local speed of a LevelSet.
   *
   * @note By default, the weight is a scalar, global for all points. 
   *       You have the possibility to give a image weight, where each pixel will contain the local weight of the module.
   *       You can set the weight to scalar again, with set_image_weight(NULL)
   *
   */
  template<unsigned DIM, typename Pixel = float>
	class SegmentationModule 
	{
	public:
	
	typedef          LevelSet<DIM, Pixel>   LevelSetType;
	typedef typename LevelSetType::Index    Index;
	typedef          Schemes<DIM,Pixel>     SchemesType;
	
	/** @name Constructors - Destructor */
	//@{
	
	/// Constructor 
	SegmentationModule(float  _weight = 1.f,     ///< weight of the module
					   bool   _adaptive = true   ///< adaptive
					   );
	
	/// Destructor
	virtual ~SegmentationModule();
	
	//@}
	/** @name Access/Modif member data */
	//@{
	
	/// Set the global weight of the module
	void set_weight(float _weight);
	/// Set the local weight of the module
	void set_image_weight(Array<float, DIM> * _weight);
	/// Weight of the module at a point
	float weight(const Index & ind);
	/// true if the module is adaptive (need to be updated)
	bool adaptive();
	
	//@}
	/** @name Virtual functions */
	//@{
	
	/// Value of the module at a point
	virtual float value(const LevelSetType & phi, const Index & ind) {return 0;};
	/// Energy of the module at a point
	virtual float energy(const LevelSetType & phi, const Index & ind) {return 0;};
	/// Initialize the module 
	virtual void init(const LevelSetType & phi);
	/// Update the module before the Level-set reinitialization
	virtual void update_before_reinit(const LevelSetType & phi){ }
	/// Update the module after the Level-set reinitialization
	virtual void update_after_reinit(const LevelSetType & phi){ }
	/// Save results depending on the module data
	virtual void save_result(const LevelSetType & phi, std::string evodir, int iter){ }
	
	//@}
	/** @name Member data  */
	//@{
	
	protected:	
	
	float                       m_weight;        ///< Global weight of the module
	Array<float, DIM> *         m_image_weight;  ///< Local weight of the module
	bool                        m_adaptive;      ///< Adaptive statistics
	
	//@}
	
	};
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors - Destructor
  
  
  template <unsigned DIM, typename Pixel>
	SegmentationModule<DIM, Pixel>::SegmentationModule(float _weight, bool _adaptive)
	:m_weight(_weight), m_image_weight(NULL), m_adaptive(_adaptive) 
  { }
  
  template <unsigned DIM, typename Pixel>
	SegmentationModule<DIM, Pixel>::~SegmentationModule() 
	{ }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  Access/Modif member data
  
  
  template <unsigned DIM, typename Pixel>
	void SegmentationModule<DIM, Pixel>::set_weight(float _weight)
	{
	  m_weight = _weight;
	}
  
  template <unsigned DIM, typename Pixel>
	void SegmentationModule<DIM, Pixel>::set_image_weight(Array<float, DIM> * _weight)
	{
	  m_image_weight = _weight;
	}
  
  template <unsigned DIM, typename Pixel>
	float SegmentationModule<DIM, Pixel>::weight(const Index & ind)
	{
	  if (m_image_weight != NULL)
		return (*m_image_weight)(ind);
	  return m_weight;
	}
  
  template <unsigned DIM, typename Pixel>
	bool SegmentationModule<DIM, Pixel>::adaptive()
	{
	  return m_adaptive;
	}
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  Access/Modif member data
  
  
  template <unsigned DIM, typename Pixel>
	void SegmentationModule<DIM, Pixel>::init(const LevelSetType & phi)
	{
	  if ( (m_image_weight != NULL) && !Globals::same_shape<DIM>(m_image_weight->shape(), phi.shape()) ) {
		std::cerr << "Error : Shape of the image weight and shape of the LevelSet don't match"  << std::endl;
		exit(1);
	  }
	}
}

#endif // SEGMENTATIONMODULE_H
