/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include "math_config.h"
#include <algorithm>
#include <iostream>
#include <cmath>
#include <assert.h>
#include <boost/cstdint.hpp>

#ifdef WIN32
using boost::int32_t;
#endif

#define POLY0(x, c0)                     ( c0                                      )
#define POLY1(x, c0, c1)                 ( c0 + x * POLY0( x, c1                 ) )
#define POLY2(x, c0, c1, c2)             ( c0 + x * POLY1( x, c1, c2             ) )
#define POLY3(x, c0, c1, c2, c3)         ( c0 + x * POLY2( x, c1, c2, c3         ) )
#define POLY4(x, c0, c1, c2, c3, c4)     ( c0 + x * POLY3( x, c1, c2, c3, c4     ) )
#define POLY5(x, c0, c1, c2, c3, c4, c5) ( c0 + x * POLY4( x, c1, c2, c3, c4, c5 ) )

inline float aexp2f( float x )
{
  int32_t ipart;
  float fpart, expipart, expfpart;

  x = std::min( x, 129.0f );
  x = std::max( x, -126.99999f );

  ipart = int32_t( x + ( x > 0 ? 0.5f : -0.5f ) );
  //ipart = int32_t( x - 0.5f );
  fpart = x - float(ipart);  

  ipart = ( ipart + 127 ) << 23;
  expipart = *reinterpret_cast<float*>(&ipart);
   /* minimax polynomial fit of 2**x, in range [-0.5, 0.5[ */
#if EXP_POLY_DEGREE == 5
   expfpart = POLY5(fpart, 9.9999994e-1f, 6.9315308e-1f, 2.4015361e-1f, 5.5826318e-2f, 8.9893397e-3f, 1.8775767e-3f);
#elif EXP_POLY_DEGREE == 4
   expfpart = POLY4(fpart, 1.0000026f, 6.9300383e-1f, 2.4144275e-1f, 5.2011464e-2f, 1.3534167e-2f);
#elif EXP_POLY_DEGREE == 3
   expfpart = POLY3(fpart, 9.9992520e-1f, 6.9583356e-1f, 2.2606716e-1f, 7.8024521e-2f);
#elif EXP_POLY_DEGREE == 2
   expfpart = POLY2(fpart, 1.0017247f, 6.5763628e-1f, 3.3718944e-1f);
#else
#error
#endif   
   return expipart * expfpart;
}

inline float aexpf( float x )
{
  return aexp2f( x * log2e );
}

inline float alog2f( float x )
{
  //  std::cout << " alf " << x << ", " << std::flush;
  int32_t exp = 0x7F800000;
  int32_t mant = 0x007FFFFF;
  float one = 1.0f;
  int32_t ione = *reinterpret_cast<int32_t*>(&one);
  int32_t i = *reinterpret_cast<int32_t*>(&x);
  int32_t iexp = ( ( i & exp ) >> 23 ) - 127;
  int32_t imant = ( i & mant ) | ione;
  float e = float( iexp );
  float m = *reinterpret_cast<float*>(&imant);
  float p;
   /* Minimax polynomial fit of log2(x)/(x - 1), for x in range [1, 2[ */
 #if LOG_POLY_DEGREE == 6
   p = POLY5( m, 3.1157899f, -3.3241990f, 2.5988452f, -1.2315303f,  3.1821337e-1f, -3.4436006e-2f);
#elif LOG_POLY_DEGREE == 5
   p = POLY4(m, 2.8882704548164776201f, -2.52074962577807006663f, 1.48116647521213171641f, -0.465725644288844778798f, 0.0596515482674574969533f);
#elif LOG_POLY_DEGREE == 4
   p = POLY3(m, 2.61761038894603480148f, -1.75647175389045657003f, 0.688243882994381274313f, -0.107254423828329604454f);
#elif LOG_POLY_DEGREE == 3
   p = POLY2(m, 2.28330284476918490682f, -1.04913055217340124191f, 0.204446009836232697516f);
#else
#error
#endif
   /* This effectively increases the polynomial degree by one, but ensures that log2(1) == 0*/
   p *= ( m - 1 );
   return p + e;
}

inline float alogf( float x )
{
  return alog2f( x ) * rlog2e;
}

#undef POLY0
#undef POLY1
#undef POLY2
#undef POLY3
#undef POLY4
#undef POLY5

