/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <iostream>
#include <limits>
#include <typeinfo>
#include <iomanip>

#include <boost/math/policies/error_handling.hpp>
#include <boost/format.hpp>

namespace boost{ namespace math{ namespace policies{

template <class T>
T user_domain_error(const char* function, const char* message, const T& val)
{
  if(function == 0)
    function = "Unknown function with arguments of type %1%";
  if(message == 0)
    message = "Cause unknown with bad argument %1%";
  std::string msg("Error in function ");
  msg += (boost::format(function) % typeid(T).name()).str();
  msg += ": \n";
  int prec = 2 + (std::numeric_limits<T>::digits * 30103UL) / 100000UL;
  msg += (boost::format(message) % boost::io::group(std::setprecision(prec), val)).str();
#if 0  
  std::cerr << msg << std::endl;
#else
  throw std::runtime_error(msg);
#endif
  return std::numeric_limits<T>::quiet_NaN();
}
template <class T>
T user_pole_error(const char* function, const char* message, const T& val)
{
   return user_domain_error(function, message, val);
}
template <class T>
T user_overflow_error(const char* function, const char* message, const T& val)
{
   if(function == 0)
       function = "Unknown function with arguments of type %1%";
   if(message == 0)
       message = "Result of function is too large to represent";

   std::string msg("Error in function ");
   msg += (boost::format(function) % typeid(T).name()).str();
   msg += (boost::format(" val = %f") % val).str();
   msg += ": \n";
   msg += message;

#if 0  
  std::cerr << msg << std::endl;
#else
  throw std::runtime_error(msg);
#endif
   
   // Value passed to the function is an infinity, just return it:
   return val; 
}
template <class T>
T user_underflow_error(const char* function, const char* message, const T& val)
{
   if(function == 0)
       function = "Unknown function with arguments of type %1%";
   if(message == 0)
       message = "Result of function is too small to represent";

   std::string msg("Error in function ");
   msg += (boost::format(function) % typeid(T).name()).str();
   msg += (boost::format(" val = %f") % val).str();

   msg += ": \n";
   msg += message;

#if 0  
  std::cerr << msg << std::endl;
#else
  throw std::runtime_error(msg);
#endif
  
   // Value passed to the function is zero, just return it:
   return val; 
}
template <class T>
T user_denorm_error(const char* function, const char* message, const T& val)
{
   if(function == 0)
       function = "Unknown function with arguments of type %1%";
   if(message == 0)
       message = "Result of function is denormalised";

   std::string msg("Error in function ");
   msg += (boost::format(function) % typeid(T).name()).str();
   msg += (boost::format(" val = %f") % val).str();

   msg += ": \n";
   msg += message;

#if 0  
  std::cerr << msg << std::endl;
#else
  throw std::runtime_error(msg);
#endif
  
   // Value passed to the function is denormalised, just return it:
   return val; 
}
template <class T>
T user_evaluation_error(const char* function, const char* message, const T& val)
{
   if(function == 0)
       function = "Unknown function with arguments of type %1%";
   if(message == 0)
       message = "An internal evaluation error occured with "
                  "the best value calculated so far of %1%";

   std::string msg("Error in function ");
   msg += (boost::format(function) % typeid(T).name()).str();

   msg += ": \n";
   int prec = 2 + (std::numeric_limits<T>::digits * 30103UL) / 100000UL;
   msg += (boost::format(message) % boost::io::group(std::setprecision(prec), val)).str();

#if 0  
  std::cerr << msg << std::endl;
#else
  throw std::runtime_error(msg);
#endif

   // What do we return here?  This is generally a fatal error,
   // that should never occur, just return a NaN for the purposes
   // of the example:
   return std::numeric_limits<T>::quiet_NaN();
}

}}} // namespaces

namespace{

using namespace boost::math::policies;

typedef policy<
   ::boost::math::policies::domain_error<user_error>,
   ::boost::math::policies::pole_error<user_error>,
   ::boost::math::policies::overflow_error<user_error>,
   ::boost::math::policies::underflow_error<user_error>,
   ::boost::math::policies::denorm_error<user_error>,
   ::boost::math::policies::evaluation_error<user_error>
//   ::boost::math::policies::promote_float<   true   >,
//   ::boost::math::policies::digits10<   7   >
> user_error_policy;

BOOST_MATH_DECLARE_SPECIAL_FUNCTIONS(user_error_policy)

} // close unnamed namespace
