/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include <stdexcept>
#include <boost/format.hpp>
#include <boost/math/tools/polynomial.hpp>

namespace chebychev
{
  
  template<typename T>
  class polynomial
  {
	
  public:
	
	typedef typename boost::math::tools::polynomial<T> coeff_poly;
	typedef std::vector<T> storage_t;
	typedef typename storage_t::iterator iterator;
	typedef typename storage_t::const_iterator const_iterator;

	polynomial( T a0 = 1 )
	  : _coeffs(1,a0)
	{}
	
	polynomial( T a0,  T a1 )
	  : _coeffs(2)
	{
	  _coeffs[0] = a0;
	  _coeffs[1] = a1;
	}
	
	void operator*=( const polynomial & b )
	{
	  
	  polynomial a( *this );
	  size_t n1 = a.size();
	  size_t n2 = b.size();
	  size_t nmin = std::min( n1, n2 );
	  size_t nmax = std::max( n1, n2 );
	  _coeffs.resize( n1 + n2 - 1 );
	  
	  _coeffs[0] = a[0] * b[0] * T(0.5);
	  for ( size_t l = 1; l < nmin; ++l )
		_coeffs[0] += a[ l ] * b[ l ];
	  
	  for ( size_t k = 1; k < nmax - 1; ++k )
		{
		  _coeffs[k] = 0;
		  for ( size_t l = std::max( 0, int(k) - int(n1) + 1 ); l < std::min( k, n2 ); ++l )
			_coeffs[k] += a[ k - l ] * b[ l ];
		  if ( k < n2 )
			for ( size_t l = 1; l < std::min( n1, n2 - k ) ; ++l )
			  _coeffs[k] += a[ l ] * b[ l + k ];				
		  if ( k < n1 )
			for ( size_t l = 1; l < std::min( n1 - k, n2 ) ; ++l )
			  _coeffs[k] += a[ l + k ] * b[ l ];
		  _coeffs[k] *= T(0.5);
		}
	  
	  for ( size_t k = nmax; k < size(); ++k )
		{
		  _coeffs[k] = 0;
		  for ( size_t l = k - std::min( nmax, n1 ) + 1; l < std::min( n2, nmax ); ++l )
			_coeffs[k] += a[ k - l ] * b[ l ];
		  _coeffs[k] *= T(0.5);
		}
	  
	}
	
	size_t size() const { return _coeffs.size(); }
	
	T operator[]( size_t i ) const
	{
	  if ( i >= size() )
		throw std::runtime_error( str( boost::format
									   ( "chebychev_polynomial index out of bounds: %i >= %i") 
									   % i % size() ) );
	  return _coeffs[i];
	}
	
	coeff_poly to_coefficient_form( )
	{
	  T x_[2] = { 0, 1 };
	  coeff_poly x( x_, 1 );
	  std::vector<coeff_poly> b( size() + 2, 0 );
	  for ( size_t j = size() - 1; j > 0; --j )
		b[ j ] = 2 * x * b[ j + 1 ] - b[ j + 2 ] + _coeffs[ j ];
	  return x * b[1] - b[2] + _coeffs[0];
	}

	iterator begin() { return _coeffs.begin(); }
	iterator end() { return _coeffs.end(); }
	const_iterator begin() const { return _coeffs.begin(); }
	const_iterator end() const { return _coeffs.end(); }

  private:
	
	std::vector<T> _coeffs;
	
  };
  

template<typename ostream, typename T> ostream & operator<<( ostream & os, const polynomial<T> & p )
{
  os << "{ ";
  for ( size_t i = 0; i < p.size(); ++i )
	os << p[i] << " ";
  os << "}";
  return os;
}
};


