/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once
#include <cmath>
#include <boost/serialization/serialization.hpp>

///   class for exponential moving average
///     usage:  tau : decayfactor [ 0:1 ]
///             operator(): returns actual value, after updating with input
template<typename T> class emwa
{
public:

  emwa( T tau = 1 ) : _tau(tau) { reset(); }

  void reset()
  {
    _N = 0; _Z = 0; _counter = 0;
  }

  size_t counter() const { return _counter; }

  T operator()( T x, T w = 1, T dt = 1 )
  {
    if ( _tau <= 0 )
    {
        _N = x;
        _Z = 1;
    }
    else
    {
        T factor = std::exp( -dt / _tau );
        _N *= factor;
        _Z *= factor;
        _N += x;
        _Z += w;
        _counter += dt;
    }
    return mean();
  }

  T mean() const { return _Z > 0 ? _N / _Z : T(0); }

  void set_tau( T tau ) { _tau = tau; }

  void get_tau( ) const { return _tau; }

  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
  
	ar & boost::serialization::make_nvp( "_N", _N );
	ar & boost::serialization::make_nvp( "_Z", _Z );
	ar & boost::serialization::make_nvp( "_tau", _tau );
	ar & boost::serialization::make_nvp( "_counter", _counter );

  }

private:

  T _N;
  T _Z;
  T _tau;
  T _counter;

};



///   class for exponential moving average with first derivative
///     usage:  tau : decayfactor [ 0:1 ]
///             operator(): updates emwa's
///             mean():     returns (decayed) value
///             diff():     returns (decayed) difference of current and last value
///             absdiff():  returns (decayed) absolute difference
template<typename T> class diff_emwa
{
public:
  diff_emwa( T tau ) : _val_emwa(tau), _diff_emwa(tau), _abs_diff_emwa(tau), _had_last(false) {}

  void operator()( T x )
  {
	_val_emwa(x);
	if ( _had_last )
	  {
		T diff = x - _x_last;
		T abs_diff = std::abs( diff );
		_diff_emwa( diff );
		_abs_diff_emwa( abs_diff );		
	  }
	_x_last = x;
	_had_last = true;
  }

  size_t counter() { return _val_emwa.counter(); }
  size_t diffcounter() { return _diff_emwa.counter(); }

  T mean() { return _val_emwa.mean(); }
  T diff() { return _diff_emwa.mean(); }
  T absdiff() { return _abs_diff_emwa.mean(); }
  void reset() { _val_emwa.reset(); _diff_emwa.reset(); _abs_diff_emwa.reset(); _had_last = false; }

private:

  emwa<T> _val_emwa;
  emwa<T> _diff_emwa;
  emwa<T> _abs_diff_emwa;
  bool _had_last;
  T _x_last;

};

///   class for weighted average
///     usage:  operator(): returns actual value, after updating with input
///             mean():     return actual value without updating
///             var():      computes variance

template<typename T> class wavrg
{
public:

  wavrg( ){ reset(); }
  wavrg( T val, T w = T(1) )
  { 
      reset(); 
      operator()(val, w);
  }


  void reset()
  {
    _X = 0; _Z = 0; _X2 =0 ; _counter = 0;
  }

  size_t counter() const { return _counter; }

  T operator()( T x, T w = T(1) )
  {
    _X  += x;
    _X2 += x*x;
    _Z += w;
    _counter += T(1);
    return mean();
  }

  T mean() const 
  {
    return _Z > T(0) ? _X / _Z : T(0); 
  }
  T var() const
  {
    if ( _Z > T(0) )
      return ( _X2 - _X ) / mean();
    else
      return T(0);
  }

  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
  
	ar & boost::serialization::make_nvp( "_X", _X );
    ar & boost::serialization::make_nvp( "_X2", _X2 );
	ar & boost::serialization::make_nvp( "_Z", _Z );
	ar & boost::serialization::make_nvp( "_counter", _counter );

  }

private:

  T _X;
  T _X2;
  T _Z;
  T _counter;

};

