/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef GAMMA_HPP
#define GAMMA_HPP

#include <cmath>
#include <limits>
#include <complex>
#include <iostream>


#ifndef LIMIT_Z_RANGE
#define LIMIT_Z_RANGE 1
#endif

template<typename T> struct Real
{
    typedef T type;
};

template<typename T> struct Real< std::complex<T> >
{
    typedef T type;
};

template<typename T> T get_real(T x) 
{
    return x;
}

template<typename T> T get_real(std::complex<T> x) 
{
    return x.real();
}

template<typename T> T get_imag(T x)
{
    return 0;
}

template<typename T> T get_imag(std::complex<T> x) 
{
    return x.imag();
}


//
// Author:
//      EAG <special.functions@gmail.com>
//
// Matlab 7.2
// May 17, 2006 v1.0


//could use a fractional precision selector for how many
//iterations to do
//or a selector like 1/8 1/4 1/2 1

/*
function z=fdigamma(z)
% FDIGAMMA: the PSI (Maple)/POLYGAMMA (Mathematica) function 
% for complex arguments.
%
% Accuracy:
%      16 decimal places accurate away from the singularities
%      (z=-1, -2, -3, ...). Full 16 digit accuracy is guaranteed
%      for all z with real(z)>=0. Accuracy decreases as z appro-
%      aches any of the singularities.
%
%      If -n, with n>0, is the nearest singularity to the given
%      argument z, then a minimum of 15 digit accuracy is obtained
%      for abs(z+n)>2 for all abs(z)<=10e5. The accuracy, however,
%      is greater for abs(z)<100 for smaller abs(z+n).
%
% Performance : 
%      About 28 times faster than Matlab's mfun for 100x100 complex
%      matrix arguments. About 12 times faster than mfun for scalar
%      arguments.
%
% Author:
%      EAG <special.functions@gmail.com>
%
% Matlab 7.2
% May 17, 2006 v1.0

sz=size(z);
z=z(:);
for k=1:length(z)
    if (imag(z(k))==0 && -z(k)-floor(abs(z(k)))==0)
        z(k)=NaN; % infinite for negative integers
    else
        if abs(z(k))<=1
            z(k)=digammasmall(z(k));
        elseif real(z(k))>=10 && abs(z(k))>=10
            z(k)=digammalarge(z(k));
        else % compute using distant neighbor formulas
            if (real(z(k))<10 && abs(imag(z(k)))<1)
                if real(z(k))>0
                    n=1:floor(real(z(k)));
                    z(k)=digammasmall(z(k)-floor(real(z(k)))) ...
                        + sum(1./(z(k)-n));
                else
                    n=0:-ceil(real(z(k)))-1;
                    z(k)=digammasmall(z(k)-ceil(real(z(k)))) ...
                        - sum(1./(z(k)+n));
                end
            else
                if real(z(k))>=0
                    n=0:10;
                    z(k)=digammalarge(z(k)+11)-sum(1./(z(k)+n));
                else
                    n=1:11;
                    z(k)=digammalarge(z(k)-11)+sum(1./(z(k)-n));
                end
            end
        end
    end
end
z=reshape(z,sz);
end % fdigamma
*/


template<typename T> T trigam ( T x )

//****************************************************************************80
//
//  Purpose:
//
//    TRIGAM calculates trigamma(x) = d**2 log(gamma(x)) / dx**2
//
//  Modified:
//
//    19 January 2008
//
//  Author:
//
//    BE Schneider
//    Modifications by John Burkardt
//
//  Reference:
//
//    BE Schneider,
//    Algorithm AS 121:
//    Trigamma Function,
//    Applied Statistics, 
//    Volume 27, Number 1, pages 97-99, 1978.
//
//  Parameters:
//
//    Input, T X, the argument of the trigamma function.
//    0 < X.
//
//    Output, int *IFAULT, error flag.
//    0, no error.
//    1, X <= 0.
//
//    Output, T TRIGAM, the value of the trigamma function at X.
//
{
  if (x <= 0) return std::numeric_limits<T>::quiet_NaN();
  
  const T a = 0.0001;
  const T b = 5.0;
  const T b2 =  0.1666666667;
  const T b4 = -0.03333333333;
  const T b6 =  0.02380952381;
  const T b8 = -0.03333333333;

  T value;
  T y;
  T z;

  z = x;

//
//  Use small value approximation if X <= A.
//
  if ( x <= a )
  {
    value = 1.0 / x / x;
    return value;
  }
//
//  Increase argument to ( X + I ) >= B.
//
  value = 0.0;

  while ( z < b )
  {
    value = value + 1.0 / z / z;
    z = z + 1.0;
  }
//
//  Apply asymptotic formula if argument is B or greater.
//
  y = 1.0 / z / z;

  value = value + 0.5 * 
      y + ( 1.0 
    + y * ( b2  
    + y * ( b4  
    + y * ( b6  
    + y *   b8 )))) / z;

  return value;
}

inline float trigamma(float x)
{
  return trigam(x);
}

inline double trigamma(double x)
{
  return trigam(x);
}


#if USE_BOOST_GAMMAS


#include <boost/math/special_functions/gamma.hpp>
#include <boost/math/special_functions/digamma.hpp>
#include "../../Math/include/boost_math_error_handlers.hpp"



#define LGAMMA_F_ARG_MIN 7e-40f
#define LGAMMA_F_ARG_MAX 5e25f
#define DIGAMMA_F_ARG_MIN 3e-39f 
#define DIGAMMA_F_ARG_MAX 3e28f


#define LGAMMA_D_ARG_MIN 7e-40
#define LGAMMA_D_ARG_MAX 5e25
#define DIGAMMA_D_ARG_MIN 3e-39
#define DIGAMMA_D_ARG_MAX 3e28

template<typename T>
inline T clamp( T x, T xmin, T xmax ) 
{
  return std::max( xmin, std::min( xmax, x ) );
}

// TODO: use right guard values for double precision

#if defined(GUARD_GAMMA_FUNCTIONS_BOTH)
# define GUARD(x,xmin,xmax) x = clamp(x,xmin,xmax)
#else
#  if defined(GUARD_GAMMA_FUNCTIONS_BELOW) 
#    define GUARD(x,xmin,xmax) x = std::max(x,xmin)
#  else
#    if defined(GUARD_GAMMA_FUNCTIONS_ABOVE)
#      define GUARD(x,xmin,xmax) x = std::min(x,xmax)
#    else
#      define GUARD(x,xmin,xmax)
#    endif
#  endif
#endif


inline float gammaln(float x)
{
  GUARD( x, LGAMMA_F_ARG_MIN, LGAMMA_F_ARG_MAX );
  return boost::math::lgamma(x,user_error_policy());
}

inline double gammaln(double x)
{
  GUARD( x, LGAMMA_D_ARG_MIN, LGAMMA_D_ARG_MAX );
  return boost::math::lgamma(x,user_error_policy());
}

inline float psi(float x)
{
  GUARD( x, DIGAMMA_F_ARG_MIN, DIGAMMA_F_ARG_MAX );
  return boost::math::digamma(x,user_error_policy());
}

inline double psi(double x)
{
  GUARD( x, DIGAMMA_D_ARG_MIN, DIGAMMA_D_ARG_MAX );
  return boost::math::digamma(x,user_error_policy());
}

#undef GUARD

#else 

#include "gammaApprox.h"

inline float gammaln(float x)
{
  return float(gammaApprox::gammln(x));
}

inline double gammaln(double x)
{
  return gammaApprox::gammln(x);
}

inline float psi(float x)
{
  return float(gammaApprox::digamma(x));
}

inline double psi(double x)
{
  return gammaApprox::digamma(x);
}

#endif

#endif //DIGAMMA_HPP

