/***************************************************************************
 *   Copyright (C) 2009 by Dominik Endres                                  *
 *   dominik.endres@klinikum.uni-tuebingen.de                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#pragma once

#include "math.h"
#include <iostream>
#include <cassert>
#include "float.h"

using namespace std;

namespace gammaApprox {

static const double gammaCof[]={ 1.000000000190015,
								 76.18009172947146,
								 -86.50532032941677,	
								 24.01409824083091,
								 -1.231739572450155,
								 0.1208650973866179e-2,
								 -0.5395239384953e-5
};

static const double cEulerMascheroni=0.577215664901532860606512090082402431042;

#ifndef M_PI
#define M_PI 3.14159265359
#endif

/** \return ln(Gamma(xx)) for xx>0. See Numerical Recipes for details */
inline double gammln(const double &xx) 
{
	double y,tmp,ser;
	y    = xx;
	tmp  = xx + 5.5;
	tmp -= ( xx + 0.5 ) * log(tmp);
	ser  = gammaCof[0];
	for ( int j = 1; j < 6; j++ ) 
	  ser += gammaCof[j] / ++y;
	return log( sqrt( 2.0 * M_PI ) *ser /xx ) - tmp;
}

/** \return Psi(xx) for xx>0, i.e. the dln(Gamma(xx))/dxx. Derivative of code for ln(Gamma(xx)). */
inline double digamma(const double &xx) 
{
	double rv = log( xx + 4.5 ) - 5.0 / ( xx + 4.5 );
	double serNum,serDenom;
	serNum   = 0.0;
	serDenom = gammaCof[0];
	double y = xx - 1.0;
	for(int j=1;j<6;j++) 
	  {
		serDenom += gammaCof[j] /  ++y;
		serNum   += gammaCof[j] / ( y * y );	
	  }
	return rv - serNum / serDenom;
}

/** \return ln(Gamma(xx)) for xx>0. See Numerical Recipes for details */
inline float gammln(const float &xx) 
{
	float y,tmp,ser;
	y    = xx;
	tmp  = xx + 5.5;
	tmp -= ( xx + 0.5 ) * log(tmp);
	ser  = gammaCof[0];
	for ( int j = 1; j < 6; j++ ) 
	  ser += gammaCof[j] / ++y;
	return log( sqrt( 2.0 * M_PI ) *ser /xx ) - tmp;
}

/** \return Psi(xx) for xx>0, i.e. the dln(Gamma(xx))/dxx. Derivative of code for ln(Gamma(xx)). */
inline float digamma(const float &xx) 
{
	float rv = log( xx + 4.5 ) - 5.0 / ( xx + 4.5 );
	float serNum,serDenom;
	serNum   = 0.0;
	serDenom = gammaCof[0];
	float y = xx - 1.0;
	for(int j=1;j<6;j++) 
	  {
		serDenom += gammaCof[j] /  ++y;
		serNum   += gammaCof[j] / ( y * y );	
	  }
	return rv - serNum / serDenom;
}




/** \return Inverse of Psi(xx) for xx>0, i.e. the dln(Gamma(xx))/dxx. Derivative of code for ln(Gamma(xx)). digammaInv(digamma(x))=x to 5 decimal digits for x>1e-4. */
inline double digammaInv(const double &y,const double eps=1e-7) 
{
	// determine initial bracket on x
	double f1,f2,x1,x2;
	if(y<=-1e4) {
		return -1.0/y;
	} else if(y<=20.0) {
		if(y<=0.0) {
			x1=1.0;
			while((f1=digamma(x1)-y)>0.0) x1/=10.0;
			x2=1.47;
			f2=digamma(x2)-y;
		} else {
			x1=exp(y);
			x2=x1*((y>=10.0?exp(0.08):exp(2.6)))-5.0+0.5+1.0;
			x1=max(1.46,x1-5.0+0.5+1.0);
			f1=digamma(x1)-y;
			f2=digamma(x2)-y;
		}
		while(x2-x1>eps) {
			assert(f1<=0.0 && f2>=0.0); // make sure root is bracketed
			const double midx=(x1+x2)/2.0; // bisection step to ensure at least linear convergence
			const double midf=digamma(midx)-y;
			const double rx=(y<=0.0?1.0/(1.0/(x1+4.5)-f1/(f1-f2)*(1.0/(x1+4.5)-(1.0/(x2+4.5))))-4.5:(x1+4.5)*pow((x1+4.5)/(x2+4.5),f1/(f2-f1))-4.5);
			const double rf=digamma(rx)-y; // refined guess for minimum location, based on logarithmic interpolation for y>0 and 1/x interpolation for y<0.
			if((rf<=0 && midf>=0) || (rf>=0 && midf<=0)) {
				x1=min(rx,midx);
				f1=min(rf,midf);
				x2=max(rx,midx);
				f2=max(rf,midf);
			} else if(rf>=0 && midf>=0) {
				x2=min(rx,midx);
				f2=min(rf,midf);
			} else if(rf<=0 && midf<=0) {
				x1=max(rx,midx);
				f1=max(rf,midf);
			}
			if(fabs(f2)<eps) {
				x1=x2;
				break;
			}
			if(fabs(f1)<eps) {
				x2=x1;
				break;
			}
		}
		return (x1+x2)/2.0;
	} else return exp(y)-5.0+0.5;
}


/** Compute posterior parameters of Beta B(p;mean,pseudoCount) approximation to posterior for a likelihood of the form P(D|p) \propto alpha*p+beta and a Beta prior P(p;mean,pseudoCount). Approximation via moment-matching of mean and pseudo-count. If initial pseudo-count<0.1, set to 0.1.
\param alpha: linear term of likelihood
\param beta: constant term of likelihood
\param mean: p(oste)rior mean of Beta distribution
\param pseudoCount: p(poste)rior pseudo-count of Beta distribution.
*/
inline void approxBetaUpdate( const float alpha, 
							  const float beta, 
							  float &mean,float &pseudoCount ) 
 {
   pseudoCount=max(pseudoCount,0.1f);
   if(pseudoCount>1.0f/FLT_EPSILON) return; // no change possible within float accuracy
   const double norm=1.0/(alpha*mean+beta);
   const double elnp = norm * (  alpha * mean * ( digamma( mean * pseudoCount + 1.0 )
												  - digamma( pseudoCount + 1.0 ) )
								 + beta * ( digamma( mean * pseudoCount ) 
											- digamma( pseudoCount ) ) );
   const double eln1mp = norm * ( alpha * mean * ( digamma( ( 1.0 - mean ) * pseudoCount )
												   - digamma( pseudoCount + 1.0 ) )
								  +beta * ( digamma( ( 1.0 - mean ) * pseudoCount ) 
											- digamma( pseudoCount ) ) );
   double pc1 = pseudoCount , f1;
   double stepsize=1.0;
   do {
	 pc1 = max( pc1 - stepsize, 1e-10 );
	 f1  = pc1 -( digammaInv( digamma(pc1) + elnp ) + digammaInv( digamma(pc1) + eln1mp ) );
	 assert( !( f1 >= 0 && pc1 == 1e-10 ) );
	 stepsize*=2;
   } while(f1>=0.0);
   
	double pc2=pseudoCount,f2;
	stepsize=1.0;
	do {
		pc2+=stepsize;
		f2=pc2-(digammaInv(digamma(pc2)+elnp)+digammaInv(digamma(pc2)+eln1mp));
		stepsize *= 2;
#ifndef _WIN32
		if( isinf( pc2 ) ) cout << "ERR " << stepsize << endl;
#endif
	} while(f2<=0.0);
	while(pc2-pc1>1e-4) {
#ifndef NDEBUG
		if( !( f1 <= 0.0 && 0.0 <= f2 ) ) 
		  {
			cout << "ERROR for: " << f1 << "," << pc1 << " and " << f2 << "," << pc2 
				 << ", starting at " << mean << "," << pseudoCount << " with " << alpha << "," << beta << endl;
			assert(false);
		  }
#endif
		double pcm=(pc1+pc2)/2.0;
		double fm=pcm-(digammaInv(digamma(pcm)+elnp)+digammaInv(digamma(pcm)+eln1mp));
		if(fm<=0.0) {
			pc1=pcm;
			f1=fm;
		} else {
			pc2=pcm;
			f2=fm;
		}
	}
	pseudoCount=(pc1+pc2)/2.0;
	mean=(digammaInv(digamma(pseudoCount)+elnp))/pseudoCount;
 }
 
/** Compute posterior parameters of Beta B(p;mean,pseudoCount) approximation to posterior for a likelihood of the form P(D|p) \propto alpha*p+beta and a Beta prior P(p;mean,pseudoCount). Approximation via moment-matching of mean and variance. Is about 500 times faster than approxBetaUpdate().
\param alpha: linear term of likelihood
\param beta: constant term of likelihood
\param mean: p(oste)rior mean of Beta distribution
\param pseudoCount: p(poste)rior pseudo-count of Beta distribution.
*/
 template<typename T1, typename T2>
   inline void approxBetaUpdateVar( float alpha, float beta, T1 &mean, T2 &pseudoCount ) 
{


  pseudoCount = max( pseudoCount, T2(1) );
  const float norm = alpha * mean + beta;
  alpha /= norm;
  beta  /= norm;
  
  //  assert( beta >= 0.0f && beta <= 1.0f && alpha + beta >= 0.0f && alpha + beta <= 1.0f );

  if( alpha >= 1e2 * beta )
	{ // nearly a bernoulli update after observing success
	  mean =  ( mean * pseudoCount + T1(1) ) / ( pseudoCount + T1(1) );
	  pseudoCount += T2(1);
	  return;
	}

  if( 1e-2 * beta > beta + alpha )
	{ // nearly a bernoulli update after observing failure
	  mean =  ( mean * pseudoCount ) / ( pseudoCount + T1(1) );
	  pseudoCount += T2(1);
	  return;
	}
		
  // split at 0.5 to increase numerical stability for large pseudocounts.
  if( mean <= 0.5f ) 
	{
	  const float f1 = ( pseudoCount * mean + 1.0f ) / ( pseudoCount + 1.0f );
	  const float f2 = ( pseudoCount * mean + 2.0f ) / ( pseudoCount + 2.0f );
	  const float n1 = alpha * f1 + beta;
	  const float n2 = ( alpha * f2 + beta ) * f1;
	  mean = max( min( mean * n1, 1.0f - FLT_EPSILON ), FLT_EPSILON );
	  pseudoCount=(n1-n2)/(n2-mean*n1);
	} 
  else 
	{
	  const float n1 =   alpha * pseudoCount * mean / ( pseudoCount + 1.0f ) + beta;
	  const float n2 = ( alpha * pseudoCount * mean / ( pseudoCount + 2.0f ) + beta ) *
		( pseudoCount * ( 1.0 - mean ) + 1.0f ) / ( pseudoCount + 1.0f );
	  mean = max( min( ( 1.0f - mean ) * n1, 1.0f - FLT_EPSILON ), FLT_EPSILON );
	  pseudoCount = ( n1 - n2 ) / ( n2 - mean * n1 );
	  mean = 1.0f - mean;
	}
}
 
}

