/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef MATRIX_OPS_HPP
#define MATRIX_OPS_HPP

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/operation.hpp>
#include "boost/numeric/ublas/lu.hpp"
#include <boost/numeric/ublas/io.hpp>



namespace ublas = boost::numeric::ublas;
typedef ublas::matrix<float> Matrix;
typedef ublas::vector<float> Vector;


// deprecated, but still here for compatibility reasons
// use trans(A) from ublas
template <typename T>
ublas::matrix<T> matrix_transpose (const ublas::matrix<T> & input)
{
	ublas::matrix<T> output ( input.size2(), input.size1() );
	for (size_t i = 0; i<output.size1(); i++)
		for (size_t j = 0; j<output.size2(); j++)
			output(i,j) = input(j,i);
	return output;
}


// deprecated, but still here for compatibility reasons
// use prod(A,B) from ublas
template <typename T>
ublas::matrix<T> matrix_product
	(const ublas::matrix<T> & A, const ublas::matrix<T> & B)
{
	unsigned n = A.size2();
	ublas::matrix<T> output(A.size1(),B.size2());
	output.clear();  // every entry set to 0 ( more precisely: T() )
	for (size_t i = 0; i<output.size1(); i++)
		for (size_t j = 0; j<output.size2(); j++)
			for (size_t k = 0; k<n; k++)
				output(i,j) += A(i,k)*B(k,j);
	return output;
}

// deprecated, but still here for compatibility reasons
// use prod(A,v) from ublas
// multiply matrix with std::vector
template <typename S, typename T>
std::vector<T> matrix_product
	(const ublas::matrix<S> & A, const std::vector<T> & v)
{
	if (A.size2() != v.size())
	{
		std::cout << "Size mismatch! Cannot multiply the matrix and the vector!" << std::endl;
		return std::vector<T>();
	}
	std::vector<T> output(A.size1(),T());
	for (size_t i = 0; i<output.size(); i++)
		for (size_t k = 0; k<v.size(); k++)
				output.at(i) += A(i,k)*v.at(k);
	return output;
}



// Inverts a matrix.
// returns whether inverting the input was successful
// The result is written into the variable inverse
template <typename T>
bool matrix_inverse (const ublas::matrix<T> & input, ublas::matrix<T> & inverse)
{
	using namespace ublas;
	assert( input.size1() == input.size2() );
	inverse.resize( input.size1(), input.size1() );
	typedef permutation_matrix<size_t> pmatrix;
	// matrix A will contain the LU-factorization (L and U in one matrix)
	matrix<T> A(input);
	//A *= 1E10;
	pmatrix pm(A.size1());
	int res = lu_factorize(A,pm);
    if (res != 0)  // lu_factorize returns 0 if successful
    	return false;
    inverse.assign( identity_matrix<T>(A.size1()) );
    //inverse *= 1E10;
    lu_substitute(A, pm, inverse);
 	return true;
}



// add two vectors
template <typename T>
std::vector<T> sum( std::vector<T> a, std::vector<T> b )
{
	std::vector<T> output;
	for (size_t i = 0; i<a.size(); i++)
		output.push_back( a[i]+b[i] );
	return output;

}

// can use A-B instead
template <typename T>
ublas::matrix<T> matrix_difference
	(const ublas::matrix<T> & A, const ublas::matrix<T> & B)
{
	size_t size1 = A.size1();
	size_t size2 = A.size2();
	Matrix output( size1, size2 );
	for ( size_t i = 0; i<size1; i++)
		for ( size_t j = 0; j<size2; j++)
			output(i,j) = A(i,j) - B(i,j);
	return output;
}


// make a matrix symmetric given the upper triangle
template <typename T>
void fillGivenUpperTriangle(boost::numeric::ublas::matrix<T> & A)
{
	for (size_t i = 0; i<A.size1(); i++)
		for (size_t j = 0; j<i; j++)
			A(i,j) = A(j,i);
}


Matrix margin(const Matrix& A)
{
	unsigned size = A.size1()-1;
	Matrix B ( size, size );
	for (unsigned i = 0; i<size; i++)
		for (unsigned j = i; j<size; j++)
			B(i,j) = A(i,j) - A(i,size)/A(size,size) * A(size,j);
	fillGivenUpperTriangle(B);
	return B;
}


#endif
