/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef NORMRHO_HPP
#define NORMRHO_HPP

#include <blitz/array.h>
#include <limits>
#include <sstream>
#include <stdexcept>
#include <boost/foreach.hpp>

#include "blitz_indices.hpp"

#ifdef HAVE_SSE
#include "sse_approx_logexp.h"
#include "sse_math.hpp"
#endif
#include "approx_logexp.h"

#ifdef HAVE_APPROX_LOGEXP
#define fastlogf alogf
#define fastexpf aexpf
#else
using std::log;
using std::exp;
#define fastlogf log
#define fastexpf exp
#endif

inline double alogf( double x )
{
  return alogf( float( x ) );
}

inline double aexpf( double x )
{
  return aexpf( float( x ) );
}

inline float sigmoid( float x, float x0 = 0, float beta = 1 )
{
  return float(1) / ( float(1) + aexpf( - beta * ( x - x0 ) ) );
}

inline double sigmoid( double x, double x0 = 0, double beta = 1 )
{
  return double(1) / ( double(1) + aexpf( - beta * ( x - x0 ) ) );
}

BZ_DECLARE_FUNCTION(alogf);
BZ_DECLARE_FUNCTION(aexpf);
BZ_DECLARE_FUNCTION(sigmoid);

#if defined(NANCHECK) & defined(BZ_HAVE_IEEE_MATH)
#define CHECKNAN(m,s) { \
	if(m.numElements() > 0 && any(blitz_isnan(0*m))) {	\
	std::stringstream ss; \
	ss << __FILE__ << "(" << __LINE__ << "): " << s;	\
	ss << " " << m;										\
	throw(std::runtime_error(ss.str()));\
	}}
#else
#define CHECKNAN(m,s)
#endif

inline bool is_neg_inf(double x)
{
  return -x == std::numeric_limits<double>::infinity();
}

inline bool is_neg_inf(float x)
{
  return -x == std::numeric_limits<double>::infinity();
}

BZ_DECLARE_FUNCTION_RET(is_neg_inf,bool);

template<class T> T kl_divergence(const blitz::Array<T,1> & p1, const blitz::Array<T,1> & p2)
{
  T Z1 = sum(p1), Z2 = sum(p2);
  return sum(where(p1>0,p1*fastlogf(p1/p2),0)) / Z1 + fastlogf(Z2/Z1);
}

template<class T> void normalize(blitz::Array<T,1> & p)
{
  if ( p.numElements() == 0 ) return;

  T Z = sum(p);
  if (Z > 0)
    p *= 1/Z;
  else
    p = 1/p.extent(0); 
}

template<class T, int Rank > 
T bits_kl( const blitz::Array<T,Rank> & m1, const blitz::Array<T,Rank> & m2 )
{
  return sum( m2 * ( safe_log(m2) - safe_log(m1) ) );
}

template<class T> T smooth_message(blitz::Array<T,1> m, T m0, bool do_normalize = true )
{
  blitz::firstIndex i1;
  blitz::secondIndex i2;
  blitz::thirdIndex i3;
  blitz::fourthIndex i4;

  T m_max = std::max( max( m ), m0 );
  m = alogf( aexpf( m - m_max ) + aexpf( m0 - m_max ) );
}

template<class T> T normalize_rho(blitz::Array<T,1> m, bool do_normalize = true )
{
  blitz::firstIndex i1;
  blitz::secondIndex i2;
  blitz::thirdIndex i3;
  blitz::fourthIndex i4;

#if 0
  if (all(is_neg_inf(m)))
	{
	  m = T(1)/T(m.extent(0));
	  return;
	}
#endif

  T z = max(m);

#ifdef HAVE_SSE
  offset_exp_and_sum_op offset_exp_sum( -z );
  sse_op( m.data(), m.data(), m.numElements(), offset_exp_sum );
  if ( do_normalize )
	{
	  sse_mul_const( m.data(), m.data(), m.numElements(), 1.0f / offset_exp_sum.sum() );
	  return 1.0f;
	}
  else
	return offset_exp_sum.sum();
#else
  m = aexpf(m-z);
  T tmp(sum(m));
  if ( do_normalize )
	{
	  m *= 1 / tmp;
	  return T(1);
	}
  else
	return tmp;
#endif
  return 0; //cannot reach, but gcc doesnt grok this..
}


template<class T> 
void normalize_rho ( blitz::Array<T,3> & m, bool do_normalize = true )
{
  BLITZ_INDICES;

  if ( m.numElements() == 0 ) return;

  for ( int i = 0; i < m.extent(0); ++i )
    for ( int j = 0; j < m.extent(1); ++j )
    {
	  normalize_rho( m( i, j, _all ), do_normalize );
    }
}

template<typename T> 
blitz::Array<T,2> squeeze (
  const blitz::Array<T,4> & m )
{
  int r = m.extent(0), c = m.extent(3);
  blitz::Array<T,2> out(r,c);
  for (int i = 0; i < r; ++i)
    for (int j = 0; j < c; ++j)
    {
      out(i,j) = m(i,0,0,j);
    }
    return out;
}

template<typename T> blitz::Array<unsigned int,2> maxindex2(const blitz::Array<T,3> & m)
{
  int r = m.extent(0), c = m.extent(1), K = m.extent(2);
  blitz::Array<unsigned int,2> result(r,c);
  for (int i = 0; i < r; ++i)
    for (int j = 0; j < c; ++j)
    {
      unsigned int imax = 0;
	  T vmax = m(i,j,0), v;
	  for (int k = 1; k < K; ++k)
		if ((v = m(i,j,k)) > vmax)
		  {
			vmax = v;
			imax = k;
		  }
	  result(i,j) = imax;
    }
  return result;
}

template<typename T> blitz::Array<T, 3> frommaxindices2(const blitz::Array<unsigned int, 2> & m, int K)
{
  int r = m.extent(0), c = m.extent(1);
  blitz::Array<T, 3> out(r,c,K);
  out = 0;
  for (int i = 0; i < r; ++i)
    for (int j = 0; j < c; ++j)
  {
    out(i,j,int(m(i,j))) = T(1);
  } 
  return out;
}

template<typename T> void frommaxindices2(const blitz::Array<unsigned int, 2> & m, blitz::Array<T,3> & out )
{
  int r = m.extent(0), c = m.extent(1);
  out = 0;
  for (int i = 0; i < r; ++i)
    for (int j = 0; j < c; ++j)
  {
    out(i,j,int(m(i,j))) = T(1);
  } 
}

template<typename T> blitz::Array<unsigned int, 2> maxindices(const blitz::Array<T, 3> & m)
{
  blitz::Array<unsigned int, 2> result(m.extent(0),m.extent(1));
  if (m.extent(2) < 2)
  {
    result = 0;
  } else {
    for(int i = 0; i < m.extent(0); ++i)
      for(int j = 0; j < m.extent(1); ++j)
      {
        unsigned int kmax = 0;
        T vmax = m(i,j,0);
        const T * pv = &m(i,j,1);
        for(int k = 1; k < m.extent(2); ++k, pv += m.stride(2))
        {
          T v = *pv;
          if (v > vmax) 
          {
            kmax = k;
            vmax = v;
          }
        }
        result(i,j) = kmax;
      }
  }
  return result;
}

template<typename T> blitz::Array<T, 3> transpose(const blitz::Array<T, 3> & m)
{
  blitz::Array<T, 3> out(m.extent(1),m.extent(2),m.extent(0));
  out = m(blitz::thirdIndex(),blitz::firstIndex(),blitz::secondIndex());
  return out;
}

template<typename T> blitz::Array<T, 2> normalize(blitz::Array<T, 3> & m)
{
  BLITZ_INDICES;
  blitz::Array<T, 2> Z(sum(m,i3));
  m = where( Z(i1,i2) > 0, m(i1,i2,i3) / Z(i1,i2), T(1)/T(m.extent(2)) );
  return Z;
}

template<typename T> blitz::Array<T, 3> transpose2(const blitz::Array<T, 3> & m)
{
  blitz::Array<T, 3> out(m.extent(2),m.extent(0),m.extent(1));
  out = m(blitz::secondIndex(),blitz::thirdIndex(),blitz::firstIndex());
  return out;
}

template<typename T> void normalize2(blitz::Array<T, 3> & m)
{
  BLITZ_INDICES;
  blitz::Array<T, 2> Z(sum(m(i3,i1,i2),i3));
  m = where( Z(i2,i3) > 0, m(i1,i2,i3) / Z(i2,i3), T(1)/T(m.extent(0)) );
}

//this beauty of dynamic programming
//calculates the sum over products within all
//elements of the power set of p
//by bottom-up-depth-first traversing the binomial tree
//and reducing at every node.
//thus we reduce time from touching every path in the tree
//to touching every node
template<typename T> T sum_all_products(const blitz::Array<T,1> & p, int inc = 1)
{
	int N = p.numElements();
	T Z = inc^N;
	for (int i = 0; i < p.extent(0); ++i)
	{
		Z += Z * p(i) / N;
	}
	return Z;
}
			


template<typename T> void assign(blitz::Array<T,3> & r, blitz::Array<T,3> & out, int k)
{
  r.reference(out);
}

template<typename T> void assign(blitz::Array<T,3> r, blitz::Array<unsigned int,2> & out, int k)
{
  r.reference(frommaxindices2<T>(out,k));
}

template<class T, class RNG> 
inline unsigned int sample_multinomial(
  blitz::Array<T, 1> p, RNG & rnd
)
{
  T lower = 0, upper = 0;
  T u = rnd();
  for (int i = 0; i < p.extent(0); ++i)
  {
    lower = upper;
    upper += p(i);
    if (u >= lower && u < upper)
      return i;
  }
    //undernormalized p-vector, take highest...
  return p.extent(0) - 1;
}

template<class T, class RNG> inline 
blitz::Array<unsigned int, 2> 
sample_multinomial(
  const blitz::Array<T, 3> & p, RNG & rnd
)
{
  BLITZ_INDICES;
  blitz::Array<unsigned int, 2> result(p.extent(0),p.extent(1));
  for (int i = 0; i < p.extent(0); ++i)
    for (int j = 0; j < p.extent(1); ++j)
      result(i,j) = sample_multinomial(p(i,j,_all),rnd);
  return result;
}

template<class T, class RNG> inline 
void
sample_multinomial_inplace(
  const blitz::Array<T, 3> & p, RNG & rnd
)
{
  BLITZ_INDICES;
  for (int i = 0; i < p.extent(0); ++i)
    for (int j = 0; j < p.extent(1); ++j)
	  {
		int index = sample_multinomial(p(i,j,_all),rnd);
		p(i,j,_all) = where(i1 == index, 1, 0);
	  }
}

template<class T, class RNG> 
void 
sample_binomial(blitz::Array<T,1> p, RNG & rnd, blitz::Array<bool, 1> out)
{
  for (int i = 0; i < p.extent(0); ++i)
  {
    T u = rnd();
    out(i) = u < p(i);
  }
}

template<class T, class RNG> inline 
blitz::Array<unsigned int, 2> 
sample_binomial(
  const blitz::Array<T, 3> & p, RNG & rnd
)
{
  BLITZ_INDICES;
  blitz::Array<unsigned int, 2> result(p.extent(0),p.extent(1));
  for (int i = 0; i < p.extent(0); ++i)
    for (int j = 0; j < p.extent(1); ++j)
      result(i,j) = sample_binomial(p(i,j,_all),rnd);
  return result;
}


template<class T, class RNG> inline 
T sample_gamma(T k, T theta, RNG & rnd )
{
  if (k < 1) 
  {
 // Weibull algorithm
    T c = (1 / k);
    T d = ((1 - k) * std::pow(k, (k / (1 - k))));
    T u, v, z, e, x;
    do {
      u = rnd();
      v = rnd();
      z = -fastlogf(u);
      e = -fastlogf(v);
      x = std::pow(z, c);
    } while ((z + e) < (d + x));
    return (x * theta);
  } 
  else 
  {
 // Cheng's algorithm
    T b = (k - fastlogf(T(4)));
    T c = (k + std::sqrt(2 * k - 1));
    T lam = std::sqrt(2 * k - 1);
    T cheng = (1 + fastlogf(4.5));
    T u, v, x, y, z, r;
    do {
      u = rnd();
      v = rnd();
      y = ((1 / lam) * fastlogf(v / (1 - v)));
      x = (k * std::exp(y));
      z = (u * v * v);
      r = (b + (c * y) - x);
    } while ((r < ((4.5 * z) - cheng)) && (r < fastlogf(z)));
    return (x * theta);
  }
}

inline float binomial_combine(float p, float q)
{
  float pq = p * q;
  float npq = ( 1 - p ) * ( 1 - q );
  return pq / ( pq + npq );
}

inline double binomial_combine(double p, double q)
{
  double pq = p * q;
  double npq = ( 1 - p ) * ( 1 - q );
  return pq / ( pq + npq );
}

inline float binomial_match(float p, float q)
{
	return (p * q) + ((1-p) * (1-q));
}

inline double binomial_match(double p, double q)
{
 return (p * q) + ((1-p) * (1-q));
}

inline float switched_prob(float s, float p)
{
	return (1-s) + s * p;
}

inline double switched_prob(double s, double p)
{
	return (1-s) + s * p;
}

template<typename T> inline T lfac(unsigned x)
{
	T result = 0;
	for (unsigned i = 1; i <= x; ++i)
		result += fastlogf(T(i));
	return result;
}

template<typename T> inline 
blitz::Array<T,2> 
resizeweights(const blitz::Array<T,2> & weights, int w, int h)
{
	blitz::Array<T,2>  result(w,h);
	float cx = T(weights.extent(0))/w;
	float cy = T(weights.extent(1))/h;
	for (int j = 0; j < h; ++j)
	for (int i = 0; i < w; ++i)
	{
		int x1 = int(std::floor(i*cx));
		int y1 = int(std::floor(j*cy));
		int x2 = std::min(weights.extent(0)-1,int(std::ceil((i+1)*cx)));
		int y2 = std::min(weights.extent(1)-1,int(std::ceil((j+1)*cy)));
		blitz::Range rx = blitz::Range(x1,x2);
		blitz::Range ry = blitz::Range(y1,y2);
		blitz::Array<T,2> weightsslice(weights(rx,ry));
    result(i,j) = max(weightsslice);
	}
	return result;
}

inline double safe_log(double x)
{
  return x > 0 ? std::max(-1e20,fastlogf(x)) : -1e20;//-std::numeric_limits<double>::max();
}

inline float safe_log(float x)
{
	return x > 0 ? std::max(-1e10f,fastlogf(x)) : -1e10f;//-std::numeric_limits<double>::max();
}

#define serialize_array(ar,A) for(int i = 0, n = A.numElements(); i < n; ++i) ar & A.data()[i];

#define checked_serialize_array1(ar,m,n1) \
      if (n1 != m.extent(0)) \
        m.resize(n1);\
      serialize_array(ar,m);

#define checked_serialize_array2(ar,m,n1,n2) \
      if (n1 != m.extent(0) || n2 != m.extent(1)) \
        m.resize(n1,n2);\
      serialize_array(ar,m);

#define checked_serialize_array3(ar,m,n1,n2,n3) \
      if (n1 != m.extent(0) || n2 != m.extent(1) || n3 != m.extent(2)) \
        m.resize(n1,n2,n3);\
      serialize_array(ar,m);

#define checked_serialize_array4(ar,m,n1,n2,n3,n4) \
      if (n1 != m.extent(0) || n2 != m.extent(1) || n3 != m.extent(2) || n4 != m.extent(3)) \
        m.resize(n1,n2,n3,n4);\
      serialize_array(ar,m);

#define checked_serialize_array5(ar,m,n1,n2,n3,n4,n5) \
      if (n1 != m.extent(0) || n2 != m.extent(1) || n3 != m.extent(2) || n4 != m.extent(3) || n5 != m.extent(4)) \
        m.resize(n1,n2,n3,n4,n5);\
      serialize_array(ar,m);

BZ_DECLARE_FUNCTION2(binomial_match);
BZ_DECLARE_FUNCTION2(binomial_combine);
BZ_DECLARE_FUNCTION2(switched_prob);
BZ_DECLARE_FUNCTION(safe_log);

template<typename T, int R> bool not_nice( const blitz::Array<T,R> & X )
{
  return mean( X ) >= min( X );
}

template<class T> inline T kl_beta( T alphaP, T betaP, T alphaQ, T betaQ )
{
  T d = gammaln(alphaP + betaP) - gammaln(alphaQ + betaQ)
      - ( gammaln(alphaP) + gammaln(betaP) - gammaln(alphaQ) - gammaln(betaQ)  ) 
      + ( alphaP - alphaQ ) * ( psi(alphaP) - psi(alphaP + betaP) )
      + ( betaP  - betaQ  ) * ( psi(betaP)  - psi(alphaP + betaP) )
  ;
  return d;
}


#endif // NORMRHO_HPP
