/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef RECTANGLE_HPP
#define RECTANGLE_HPP


#include <iostream>
#include <string>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include "CPUData.hpp"
// simple class to describe a rectangle whose sides are parallel to the axes of a 2-dimensional cartesian coordinate system
// x0 is coordinate in up/down direction (often denoted y), x1 in left/right direction

struct rectangle
{
	int x0, x1;  // upper left corner
	int POI0, POI1;  // point of interest. if tracked point is not middle
	size_t size0, size1;  // length of sides
	// the point x0+size0 does not belong to the rectangle (similar with x1+size1)
	
    rectangle()
    : x0(0), x1(0), POI0(0), POI1(0), size0(0), size1(0)
    {}
    rectangle( const rectangle & other )
    : x0(other.x0), x1(other.x1), POI0(other.POI0), POI1(other.POI1), size0(other.size0), size1(other.size1)
    {}
    
	// middle is centroid of rectangle
	std::vector<int> middle()
	{
		std::vector<int> output;
		output.push_back( x0 + int(size0)/2 );
		output.push_back( x1 + int(size1)/2 );
		return output;
	}
	
	
	void newMiddle( int m0, int m1 )
	{
		int diff0 = m0 - middle()[0];
		int diff1 = m1 - middle()[1];
		x0 += diff0;
		x1 += diff1;		
	}
	
	void newPOI( int coord0, int coord1 )
	{
		std::vector<int> shift;
		shift.push_back( coord0 - POI0 );
		shift.push_back( coord1 - POI1 );
		move( shift );
	}
	
	void newPOI( std::vector<int> coordinates)
	{
		newPOI( coordinates[0], coordinates[1] );
	}
	
	void newPOI( boost::numeric::ublas::vector<int> coordinates)
	{
		newPOI( coordinates[0], coordinates[1] );
	}
	
	void newMiddle ( std::vector<int> m )
	{
		if ( m.size() != 2 )
			return;
		newMiddle( m[0], m[1] );
	}
	
	void move( std::vector<int> shift )
	{
		if ( shift.size() != 2 )
			return;
		x0 += shift[0];
		x1 += shift[1];
		POI0 += shift[0];
		POI1 += shift[1];
	}
	
	bool contains( int point_x0, int point_x1 )
	{
		if ( ( (point_x0 < x0 + int(size0)) && (point_x0>=x0) ) 
			&& ( (point_x1 < x1 + int(size1)) && (point_x1>=x1) ) )
			return true;
		else
			return false;
	}
	
    void resize( float factor )
    {
	  resize( factor, factor );
    }

    rectangle rescaled( float factor1, float factor2 ) const
    {
	  rectangle result(*this);
	  result.rescale( factor1, factor2 );
	  return result;
    }

    void rescale( float factor1, float factor2 )
    {
	  size0 *= factor1;
	  size1 *= factor2;
	  x0    *= factor1;
	  x1    *= factor2;
    } 

    void resize( float factor1, float factor2 )
	{
		std::vector<int> old_middle = middle();
		int old_x0 = x0;
		int old_x1 = x1;
		size0 = size_t( size0 * factor1 + 0.5 );
		size1 = size_t( size1 * factor2 + 0.5 );
		newMiddle( old_middle );
		// keep relative position of POI in rectangle:
		POI0 = size_t( (POI0-old_x0)*factor1 + x0 + 0.5 );
		POI1 = size_t( (POI1-old_x1)*factor2 + x1 + 0.5 );
	}
	
	
	// draw rectangle in image, changes image
	void drawIn( CPUData<float> & image, size_t color_channel = 0 )
	{
		size_t color = std::min( image.depth()-1, color_channel );
		for (int i = x0; i<x0+int(size0); i++)
		{
			image(i,x1) = 0;
			image(i,x1,color) = 1;
			image(i,x1+size1-1) = 0;
			image(i,x1+size1-1,color) = 1;
		}
		for (int j = x1; j<x1+int(size1); j++)
		{
			image(x0,j) = 0;
			image(x0,j,color) = 1;
			image(x0+size0-1,j) = 0;
			image(x0+size0-1,j,color) = 1;
		}
	}
	
	void drawInDotted( CPUData<float> & image )
	{
		for (int i = x0; i<x0+int(size0); i++)
			if( i%2 )
			{
				image(i,x1,0) = 1;
				image(i,x1+size1-1,0) = 1;
			}
		for (int j = x1; j<x1+int(size1); j++)
			if( j%2 )
			{
				image(x0,j,0) = 1;
				image(x0+size0-1,j,0) = 1;
			}
	}
	
	
	// make rectangle fit into image having 'rows' rows and 'columns' columns
	void cropToFit( size_t rows, size_t columns )
	{
        x0 = std::max( 0, x0 );
        x1 = std::max( 0, x1 );
        if( x0+int(size0) >= int(rows) )
			size0 = rows-1-x0;
		if( x1+int(size1) >= int(columns) )
			size1 = columns-1-x1;
        /*
		if( x0 < 0 )
		{
			size0 += x0;
			x0 = 0;
		}
		if( x1 < 0 )
		{
			size1 += x1;
			x1 = 0;
		}
        */
	}
	
	bool empty() const
	{
		return ( size0 == 0 || size1 == 0 );
	}
	
	size_t row() const
	{
		return x0;
	}
	
	size_t column() const
	{
		return x1;
	}
    
	size_t rows() const
	{
		return size0;
	}
	
	size_t columns() const
	{
		return size1;
	}
	
	size_t area() const
	{
		return size0*size1;
	}
	
	
	rectangle intersection ( const rectangle & other ) const
	{
		rectangle output;
		// first check if intersection is not empty
		if ( other.x0 >= x0+int(size0) || other.x1 >= x1+int(size1) )
			return output;
		if ( x0 >= other.x0+int(other.size0) || x1 >= other.x1+int(other.size1) )
			return output;
		output.x0 = std::max( x0, other.x0 );
		output.x1 = std::max( x1, other.x1 );
		output.size0 = std::min( x0+size0, other.x0+other.size0 ) - output.x0;
		output.size1 = std::min( x1+size1, other.x1+other.size1 ) - output.x1;
		return output;
	}
		
	
	float overlap_ratio( const rectangle & other ) const
	{
		float overlap = intersection(other).area();
		return overlap / std::min( area(), other.area() );
	}
	
    
    std::string dump_dims()  const
    {
        std::string result("rectangle( upperleft=[ ");
        result.append(boost::lexical_cast<std::string>(x0)+" x " );
        result.append(boost::lexical_cast<std::string>(x1)+" ], extent =[" );
        result.append(boost::lexical_cast<std::string>(size0)+" x " );
        result.append(boost::lexical_cast<std::string>(size1)+" ]" );
        if ( POI0 != 0 || POI1 != 0 )
        {
            result.append( ", POI =[" );
            result.append(boost::lexical_cast<std::string>(POI0)+" x " );
            result.append(boost::lexical_cast<std::string>(POI1)+" ]" );
        }
        result.append(" )");
        return result;
    }
    
};

static std::ostream&
operator << ( std::ostream& os, rectangle box )
{
	os << "Box with data: x0 = " << box.x0 << ", x1 = " << box.x1
		<< ", size0 = " << box.size0 << ", size1 = " << box.size1;
	return os;
}



#endif
