/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __SSE_MATH_HPP__
#define __SSE_MATH_HPP__

#include "emmintrin.h"
#include "sse_approx_logexp.h"
#include "approx_logexp.h"
#include <iostream>
#include <cmath>
#include <algorithm>

#ifdef HAVE_SSE4
#include <smmintrin.h>
#endif

inline bool sse_aligned ( const float * x )
{
  return ( size_t(x) & 15 ) == 0;
}


inline float sse_sum( const float * x, size_t n )
{
  __m128 z4 = _mm_set1_ps( 0.0f );
  size_t n4 = n / 4;
  size_t i = 0;
  if ( sse_aligned( x ) )
	for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	  z4 = _mm_add_ps( _mm_load_ps( x ), z4 );
  else
	for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	  z4 = _mm_add_ps( _mm_loadu_ps( x ), z4 );
  float * pz4 = (float*)&z4;
  float z = pz4[0] + pz4[1] + pz4[2] + pz4[3];
  for ( ; i < n; ++i )
	z += x[i];
  return z;
}

struct exp_op
{  
  __m128 operator()( __m128 x ) const { return expf4( x ); }
  float operator()( float x ) const  { return expf( x ); }
};

struct log_op
{  
  __m128 operator()( __m128 x ) const { return logf4( x ); }
  float operator()( float x ) const  { return logf( x ); }
};

struct add_const_op
{  
  add_const_op( float c ) : _c(c), _c4( _mm_set1_ps( c ) ) {}
  __m128 operator()( __m128 x ) const { return _mm_add_ps( x, _c4 ); }
  float operator()( float x ) const  { return x + _c; }
  float _c;
  __m128 _c4;
};

struct mul_const_op
{  
  mul_const_op( float c ) : _c(c), _c4( _mm_set1_ps( c ) ) {}
  __m128 operator()( __m128 x ) const { return _mm_mul_ps( x, _c4 ); }
  float operator()( float x ) const  { return x * _c; }
  float _c;
  __m128 _c4;
};

struct offset_exp_and_sum_op
{
  offset_exp_and_sum_op( float c ) : _c(c), _sum(0), _c4( _mm_set1_ps( c ) ), _sum4( _mm_set1_ps( 0 ) ) {}
  __m128 operator()( __m128 x ) const { __m128 y = expf4( _mm_add_ps( x, _c4 ) ); const_cast<offset_exp_and_sum_op*>(this)->_sum4 = _mm_add_ps( _sum4, y ); return y; }
  float operator()( float x ) const  { float y = std::exp( x + _c ); const_cast<offset_exp_and_sum_op*>(this)->_sum += y; return y; }
  float sum() const { float * psum4 = (float*)(&_sum4); return psum4[0] + psum4[1] + psum4[2] + psum4[3] + _sum; }
  float _c, _sum;
  __m128 _c4, _sum4;  
};

template<typename OP>
inline void sse_op( const float * x, float * y, size_t n, const OP & op )
{
  size_t n4 = n / 4;
  size_t i = 0;
  bool x_aligned = sse_aligned(x);
  bool y_aligned = sse_aligned(y);
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	{
	  __m128 x4 = x_aligned ? _mm_load_ps( x + i ) : _mm_loadu_ps( x + i );
	  __m128 y4 = op( x4 );
	  if ( y_aligned )
		_mm_store_ps( y + i, y4 );
	  else
		_mm_storeu_ps( y + i, y4 );
	}
  for ( ; i < n; ++i )
	y[i] = op( x[i] );
}

inline void sse_log( const float * x, float * y, size_t n )
{
  sse_op( x, y, n, log_op() );
}

inline void sse_exp( const float * x, float * y, size_t n )
{
  sse_op( x, y, n, exp_op() );
}

inline void sse_mul_const( const float * x, float * y, size_t n, float c )
{
  sse_op( x, y, n, mul_const_op(c) );
}

inline void sse_add_const( const float * x, float * y, size_t n, float c )
{
  sse_op( x, y, n, add_const_op(c) );
}

template<typename OP>
inline void sse_op2( const float * x, float * y, size_t n, const OP & op )
{
  size_t n4 = n / 4;
  size_t i = 0;
  bool x_aligned = sse_aligned(x);
  bool y_aligned = sse_aligned(y);
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	{
	  __m128 x4 = x_aligned ? _mm_load_ps( x + i ) : _mm_loadu_ps( x + i );
	  __m128 y4 = y_aligned ? _mm_load_ps( y + i ) : _mm_loadu_ps( y + i );
	  if ( y_aligned )
		_mm_store_ps( y + i, op( x4, y4 ) );
	  else
		_mm_storeu_ps( y + i, op( x4, y4 ) );
	}
  for ( ; i < n; ++i )
	y[i] = op( x[i], y[i] );
}

inline void sse_add( float * x, const float * y, size_t n )
{
  size_t n4 = n / 4;
  size_t i = 0;
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	_mm_storeu_ps( x + i, _mm_add_ps( _mm_loadu_ps( x + i ), _mm_loadu_ps( y + i ) ) );
  for ( ; i < n; ++i )
	x[i] += y[i];
}

inline void sse_madd( float * x, size_t n, const float s, const float * y )
{
  size_t n4 = n / 4;
  size_t i = 0;
  __m128 s4 = _mm_set1_ps( s );
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	_mm_storeu_ps( x + i, _mm_add_ps( _mm_loadu_ps( x + i ), _mm_mul_ps( s4, _mm_loadu_ps( y + i ) ) ) );
  for ( ; i < n; ++i )
	x[i] += s * y[i];
}

inline void sse_log_madd( float * x, size_t n, const float s, const float * y )
{
  size_t n4 = n / 4;
  size_t i = 0;
  __m128 s4 = _mm_set1_ps( s );
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	_mm_storeu_ps( x + i, _mm_add_ps( _mm_loadu_ps( x + i ), _mm_mul_ps( s4, logf4( _mm_loadu_ps( y + i ) ) ) ) );
  for ( ; i < n; ++i )
	x[i] += s * logf( std::max( 0.000000001f, y[i] ) );
}

#ifdef HAVE_SSE41
inline __m128 sse_dp( const float * x, const float * y, const size_t n )
{
  __m128 dp4 = _mm_set1_ps( 0.0f );
  const int dpmask = 0xff;
  size_t n4 = n / 4;
  size_t i = 0;
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	dp4 = _mm_add_ps( dp4, _mm_dp_ps( _mm_loadu_ps( x + i ), _mm_loadu_ps( y + i ), dpmask ) );
  float dp1 = 0.0f;
  for ( ; i < n; ++i )
	dp1 += x[i] * y[i];
  dp4 = _mm_add_ps( dp4, _mm_set1_ps( dp1 ) );
  return dp4;
}
#endif

inline __m128 loadvec( const float * x, bool aligned = false ) 
{
  if ( aligned )
	return _mm_load_ps( x );
  else
	return _mm_loadu_ps( x );
}

inline float sse_dp1( const float * x, const float * y, const size_t n, bool aligned = false )
{
  size_t n4 = n / 4;
  size_t i = 0;  
  __m128 dp4 = _mm_set1_ps( 0.0f );
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	dp4 = _mm_add_ps( dp4, _mm_mul_ps( loadvec( x + i, aligned ), loadvec( y + i, aligned ) ) );
  float * dp4f = (float*)&dp4;
  float dp1 = dp4f[0] + dp4f[1] + dp4f[2] + dp4f[3];
  for ( ; i < n; ++i )
	dp1 += x[i] * y[i];  
  return dp1;
}

// performs four dotproducts from matrix M
inline __m128 sse_mvp_sub( const float * y, size_t ny, const float * Msub, size_t Mlda, bool aligned = false )
{
  const float * Mi1 = Msub;
  const float * Mi2 = Msub + Mlda;
  const float * Mi3 = Msub + 2 * Mlda;
  const float * Mi4 = Msub + 3 * Mlda;
  // performs and blends four dot products into one 4-vector
#ifdef HAVE_SSE4
  __m128 dp4 = 
	_mm_blend_ps
	( _mm_blend_ps
	  ( _mm_blend_ps
		( sse_dp( Mi1, y, ny )
		  , sse_dp( Mi2, y, ny ), 1 << 1 )
		, sse_dp( Mi3, y, ny ), 1 << 2 )
	  , sse_dp( Mi4, y, ny ), 1 << 3 )
	;
#else
  __m128 dp4;
  float * dp4f = (float*)&dp4;
  dp4f[0] = sse_dp1( Mi1, y, ny, aligned );
  dp4f[1] = sse_dp1( Mi2, y, ny, aligned );
  dp4f[2] = sse_dp1( Mi3, y, ny, aligned );
  dp4f[3] = sse_dp1( Mi4, y, ny, aligned );
  return dp4;
#endif
}

inline __m128 sse_bin_comb( __m128 p, __m128 q )
{
  const __m128 one = _mm_set1_ps( 1.0f );
  __m128 pq = _mm_mul_ps( p, q ); 
  __m128 npq = _mm_mul_ps( _mm_sub_ps( one, p ), _mm_sub_ps( one, q ) );
  return _mm_div_ps( pq, _mm_add_ps( pq, npq ) );  
}

inline float sse_lgauss_nonorm( const float * x, const float * M, const float * Tau, size_t n )
{
  __m128 y4 = _mm_set1_ps( 0.0f );
  size_t n4 = n / 4, i = 0;
  float y = 0.0f;  
  if ( sse_aligned( M ) && sse_aligned( x ) && sse_aligned( Tau ) )
	for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	  {
		__m128 m4   = _mm_load_ps( M   + i );
		__m128 x4   = _mm_load_ps( x   + i );
		__m128 d4 = _mm_sub_ps( x4, m4 );
		__m128 tau4 = _mm_load_ps( Tau + i );
		__m128 d42 = _mm_mul_ps( d4, d4 );
		y4 = _mm_add_ps( y4, _mm_mul_ps( tau4, d42 ) );
	  }
  else
	for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	  {
		__m128 m4   = _mm_loadu_ps( M   + i );
		__m128 x4   = _mm_loadu_ps( x   + i );
		__m128 d4 = _mm_sub_ps( x4, m4 );
		__m128 tau4 = _mm_loadu_ps( Tau + i );
		__m128 d42 = _mm_mul_ps( d4, d4 );
		y4 = _mm_add_ps( y4, _mm_mul_ps( tau4, d42 ) );
	  }
  float * py4 = (float*)&y4;
  y = py4[0] + py4[1] + py4[2] + py4[3];
  for ( ; i < n; ++i )
	{
	  float dx = x[i] - M[i];
	  y += Tau[i] * dx * dx;
	}
  return -0.5f * y;
}

inline float sse_lgauss_nonorm_aligned( const float * x, const float * M, const float * Tau, size_t n )
{
  __m128 y4 = _mm_set1_ps( 0.0f );
  size_t n4 = n / 4, i = 0;
  float y = 0.0f;
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	{
	  __m128 m4   = _mm_load_ps( M   + i );
	  __m128 x4   = _mm_load_ps( x   + i );
	  __m128 d4 = _mm_sub_ps( x4, m4 );
	  __m128 tau4 = _mm_load_ps( Tau + i );
	  __m128 d42 = _mm_mul_ps( d4, d4 );
	  y4 = _mm_add_ps( y4, _mm_mul_ps( tau4, d42 ) );
	}
  float * py4 = (float*)&y4;
  y = py4[0] + py4[1] + py4[2] + py4[3];
  for ( ; i < n; ++i )
	{
	  float dx = x[i] - M[i];
	  y += Tau[i] * dx * dx;
	}
  return -0.5f * y;
}

inline float sse_gauss_conf_sub( const float   x,
								 const float * r, const size_t n,
								 const float * M, const float * Tau, const float * lbias )
{
  __m128 x4 = _mm_set1_ps( x    );
  __m128 p4 = _mm_set1_ps( 0.0f );
  __m128 mo5 = _mm_set1_ps( -0.5f );
  size_t n4 = n / 4, i = 0;
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	{
	  __m128 m4   = _mm_loadu_ps( M   + i );
	  __m128 d4 = _mm_sub_ps( x4, m4 );
	  __m128 tau4 = _mm_mul_ps( _mm_loadu_ps( Tau + i ), mo5 );
	  __m128 r4 = _mm_loadu_ps( r + i );
	  d4 = _mm_mul_ps( tau4, _mm_mul_ps( d4, d4 ) );
	  __m128 b4 = _mm_loadu_ps( lbias + i );
	  p4 = _mm_add_ps( p4, _mm_mul_ps( r4, expf4( _mm_add_ps( d4 , b4 ) ) ) );
	}
  float * pp4 = (float*)&p4;
  float p = pp4[0] + pp4[1] + pp4[2] + pp4[3];
  for ( ; i < n; ++i )
	{
	  float dx = x - M[i];
	  p += r[i] * std::exp( -0.5f * dx * dx * Tau[i] + lbias[i] );
	}
  return p;
}

inline float sse_gauss_conf_madd(  const float * x, const size_t m,
								   const float * r, const size_t k,
								   const float * M, const float * Tau, const float * lbias,
								   const float o, const float s )
{    
  __m128 p4 = _mm_set1_ps( 0.0f );
  __m128 o4 = _mm_set1_ps(   o  );
  __m128 s4 = _mm_set1_ps(   s  );
  size_t n4 = m / 4, i = 0;
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	{

	  __m128 sp4;
	  float * psp4 = (float*)&sp4;
	  size_t o0 = i * m, o1 = ( i + 1 ) * m, o2 = ( i + 2 ) * m, o3 = ( i + 3 ) * m;
	  psp4[0] = sse_gauss_conf_sub( x[i  ], r, k, M + o0, Tau + o0, lbias + o0 );
	  psp4[1] = sse_gauss_conf_sub( x[i+1], r, k, M + o1, Tau + o1, lbias + o1 );
	  psp4[2] = sse_gauss_conf_sub( x[i+2], r, k, M + o2, Tau + o2, lbias + o2 );
	  psp4[3] = sse_gauss_conf_sub( x[i+3], r, k, M + o3, Tau + o3, lbias + o3 );
	  p4 = _mm_add_ps( p4, logf4( _mm_add_ps( o4, _mm_mul_ps( s4, sp4 ) ) ) );
    
	}
  float * pp4 = (float*)&p4;
  float p = pp4[0] + pp4[1] + pp4[2] + pp4[3];
  for ( ; i < m; ++i )
	{
	  size_t j = i * m;
	  float xsub = sse_gauss_conf_sub( x[i  ], r, k, M + j, Tau + j, lbias + j );
	  p += std::log( std::max( 0.000000001f, o + s * xsub ) ); 
	}
  return p;

}

inline float sse_gauss_conf(  const float * x, const size_t m,
							  const float * r, const size_t k,
							  const float * M, const float * Tau, const float * lbias )
{    
  __m128 p4 = _mm_set1_ps( 0.0f );
  size_t n4 = m / 4, i = 0;
  for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
	{

	  __m128 sp4;
	  float * psp4 = (float*)&sp4;
	  size_t o0 = i * m, o1 = ( i + 1 ) * m, o2 = ( i + 2 ) * m, o3 = ( i + 3 ) * m;
	  psp4[0] = sse_gauss_conf_sub( x[i  ], r, k, M + o0, Tau + o0, lbias + o0 );
	  psp4[1] = sse_gauss_conf_sub( x[i+1], r, k, M + o1, Tau + o1, lbias + o1 );
	  psp4[2] = sse_gauss_conf_sub( x[i+2], r, k, M + o2, Tau + o2, lbias + o2 );
	  psp4[3] = sse_gauss_conf_sub( x[i+3], r, k, M + o3, Tau + o3, lbias + o3 );
	  p4 = _mm_add_ps( p4, logf4( sp4 ) );
    
	}
  float * pp4 = (float*)&p4;
  float p = pp4[0] + pp4[1] + pp4[2] + pp4[3];
  for ( ; i < m; ++i )
	{
	  size_t o = i * m;
	  p += std::log( std::max( 0.000000001f, sse_gauss_conf_sub( x[i  ], r, k, M + o, Tau + o, lbias + o ) ) ); 
	}
  return p;

}


template<typename combiner> 
void sse_mvp_generic( float * x, size_t nx, const float * y, size_t ny, const float * M, size_t Mlda, const combiner & comb )
{
  bool aligned = sse_aligned( y ) && sse_aligned( M ) && ( ( Mlda & 3 ) == 0 );
  if ( Mlda == 0 ) Mlda = ny;										   
  size_t nx4 = nx / 4;												   
  size_t i = 0;														   
  for ( size_t i4 = 0; i4 < nx4; ++i4, i += 4 )						   
	{																   
	  __m128 dp4 = sse_mvp_sub( y, ny, M + i * Mlda, Mlda, aligned );			
	  comb.sse_combine( x + i, dp4 );
	}																   
  for ( ; i < nx; ++i )												
	{																   
	  const float * Mi = M + Mlda * i;								
	  float dp1 = sse_dp1(y, Mi, ny, aligned );							   
	  comb.serial_combine( x + i, dp1 );
	}																	
}																		

struct sse_accumulate
{
  void sse_combine( float * y, const __m128 & dp4 ) const { _mm_storeu_ps( y, _mm_add_ps( _mm_loadu_ps( y ), dp4 ) ); }
  void serial_combine( float * y, const float & dp1 ) const { *y += dp1; }
};

struct sse_store
{
  void sse_combine( float * y, const __m128 & dp4 ) const { _mm_storeu_ps( y, dp4 ); }
  void serial_combine( float * y, const float & dp1 ) const { *y = dp1; }
};

struct sse_neg_accumulate
{
  void sse_combine( float * y, const __m128 & dp4 ) const { _mm_storeu_ps( y, _mm_sub_ps( _mm_loadu_ps( y ), dp4 ) ); }
  void serial_combine( float * y, const float & dp1 ) const { *y -= dp1; }
};

template<typename base>
struct sse_log_combiner : base
{
  void sse_combine( float * y, const __m128 & dp4 ) const
  {
	base::sse_combine( y, logf4( dp4 ) );
  }
  void serial_combine( float * y, const float & dp1 ) const
  {
	base::serial_combine( y, std::log( std::max( 0.0000000001f, dp1 ) ) );
  }
};

typedef sse_log_combiner<sse_neg_accumulate> sse_neg_log_accumulate;
typedef sse_log_combiner<sse_accumulate>     sse_log_accumulate;
typedef sse_log_combiner<sse_store>          sse_log_store;

template<typename base>
struct sse_madd_combiner : base
{
  sse_madd_combiner( const float & _o, const float & _s ) 
	: s(_s)
	, o(_o)
	, s4( _mm_set1_ps( _s ) )
	, o4( _mm_set1_ps( _o ) )
  {}
  void sse_combine( float * y, const __m128 & dp4 ) const
  { 
	base::sse_combine( y, _mm_add_ps( o4, _mm_mul_ps( s4, dp4 ) ) );
  }
  void serial_combine( float * y, const float & dp1 ) const
  { 
	base::serial_combine( y, s * dp1 + o );
  }
protected:
  const float s, o;
  const __m128 s4, o4;
};

inline void sse_mvp( float * x, size_t nx, const float * y, size_t ny, const float * M, size_t Mlda = 0 )
{
  sse_mvp_generic( x, nx, y, ny, M, Mlda, sse_store() );
}

inline void sse_log_mvp( float * x, size_t nx, const float * y, size_t ny, const float * M, size_t Mlda = 0 )
{
  sse_mvp_generic( x, nx, y, ny, M, Mlda, sse_log_store() );
}

inline void sse_mvp_accum( float * x, size_t nx, const float * y, size_t ny, const float * M, size_t Mlda = 0 )
{
  sse_mvp_generic( x, nx, y, ny, M, Mlda, sse_accumulate() );
}

inline void sse_log_mvp_accum( float * x, size_t nx, const float * y, size_t ny, const float * M, size_t Mlda = 0 )
{
  sse_mvp_generic( x, nx, y, ny, M, Mlda, sse_log_accumulate() );
}

inline void sse_neg_log_mvp_accum( float * x, size_t nx, const float * y, size_t ny, const float * M, size_t Mlda = 0 )
{
  sse_mvp_generic( x, nx, y, ny, M, Mlda, sse_neg_log_accumulate() );
}

inline void sse_log_mvp_accum_madd( float * x, size_t nx, const float * y, size_t ny, const float * M, float o, float s, size_t Mlda = 0 )
{
  sse_mvp_generic( x, nx, y, ny, M, Mlda, sse_madd_combiner<sse_log_accumulate>(o,s) );
}

inline void sse_neg_log_mvp_accum_madd( float * x, size_t nx, const float * y, size_t ny, const float * M, float o, float s, size_t Mlda = 0 )
{
  sse_mvp_generic( x, nx, y, ny, M, Mlda, sse_madd_combiner<sse_neg_log_accumulate>(o,s) );
}

#undef define_mvp

#endif
