/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "emmintrin.h"
#include "sse_math.hpp"

#pragma once

struct veci4
{
  veci4( int32_t x = 0 )
  {
	v = _mm_set1_epi32( x );
  }

  veci4( int32_t x1, int32_t x2, int32_t x3, int32_t x4 )
  {
	get_entry( 0 ) = x1;
	get_entry( 1 ) = x2;
	get_entry( 2 ) = x3;
	get_entry( 3 ) = x4;
  }

  veci4( const __m128i & w )
  {
	v = w;
  }

  veci4( const int32_t * px, bool aligned = false )
  {
	if ( aligned )
	  v =  _mm_load_si128( (__m128i*) px );
	else
	  v = _mm_loadu_si128( (__m128i*) px );
  }

  int32_t get_entry( size_t i ) const
  {
	return ((int32_t*)&v)[i];
  }

  int32_t & get_entry( size_t i )
  {
	return ((int32_t*)&v)[i];
  }

  int32_t operator[]( size_t i ) const
  {
	return get_entry( i );
  }

  int32_t & operator[]( size_t i ) 
  {
	return get_entry( i );
  }

  int32_t sum() const
  {
	return get_entry( 0 ) + get_entry( 1 ) + get_entry( 2 ) + get_entry( 3 );
  }

  int32_t product() const
  {
	return get_entry( 0 ) * get_entry( 1 ) * get_entry( 2 ) * get_entry( 3 );
  }

  veci4 operator-( const veci4 & other ) const
  {
	return _mm_sub_epi32( v, other.v );
  }

  veci4 operator+( const veci4 & other ) const
  {
	return _mm_add_epi32( v, other.v );
  }

  veci4 operator*( const veci4 & other ) const
  {
	return _mm_mul_epu32( v, other.v );
  }

  veci4 operator==( const veci4 & other ) const
  {
	return _mm_cmpeq_epi32( v, other.v );
  }

  veci4 operator<( const veci4 & other ) const
  {
	return _mm_cmplt_epi32( v, other.v );
  }

  veci4 operator>( const veci4 & other ) const
  {
	return _mm_cmpgt_epi32( v, other.v );
  }

  veci4 operator&( const veci4 & other ) const
  {
	return _mm_and_si128( v, other.v );
  }

  veci4 operator|( const veci4 & other ) const
  {
	return _mm_or_si128( v, other.v );
  }

  veci4 & operator+=( const veci4 & other )
  {
	v = veci4(v) + other;
	return *this;
  }

  veci4 & operator-=( const veci4 & other )
  {
	v = veci4(v) - other;
	return *this;
  }

  veci4 & operator*=( const veci4 & other )
  {
	v = veci4(v) * other;
	return *this;
  }

  operator __m128i() const
  {
	return v;
  }

  void store( int32_t * px, bool aligned = false ) const
  {
	if ( aligned )
	  _mm_storeu_si128( (__m128i*) px, v );
	else
	  _mm_storeu_si128( (__m128i*) px, v );
  }

  

  __m128i v;
};

struct vecf4
{

  vecf4( float x = 0 )
  {
	v = _mm_set1_ps( x );
  }

  vecf4( float x1, float x2, float x3, float x4 )
  {
	get_entry( 0 ) = x1;
	get_entry( 1 ) = x2;
	get_entry( 2 ) = x3;
	get_entry( 3 ) = x4;
  }

  vecf4( const __m128 & w )
  {
	v = w;
  }

  vecf4( const float * px, bool aligned = false )
  {
	if ( aligned )
	  v = _mm_load_ps( px );
	else
	  v = _mm_loadu_ps( px );
  }

  float get_entry( size_t i ) const
  {
	return ((float*)&v)[i];
  }

  float & get_entry( size_t i )
  {
	return ((float*)&v)[i];
  }

  float operator[]( size_t i ) const
  {
	return get_entry( i );
  }

  float & operator[]( size_t i ) 
  {
	return get_entry( i );
  }

  float sum() const
  {
	return get_entry( 0 ) + get_entry( 1 ) + get_entry( 2 ) + get_entry( 3 );
  }

  float product() const
  {
	return get_entry( 0 ) * get_entry( 1 ) * get_entry( 2 ) * get_entry( 3 );
  }

  float max() const
  {
	return std::max( std::max( get_entry( 0 ), get_entry( 1 ) ), std::max( get_entry( 2 ), get_entry( 3 ) ) );
  }

  vecf4 operator-( const vecf4 & other ) const
  {
	return _mm_sub_ps( v, other.v );
  }

  vecf4 operator+( const vecf4 & other ) const
  {
	return _mm_add_ps( v, other.v );
  }

  vecf4 operator*( const vecf4 & other ) const
  {
	return _mm_mul_ps( v, other.v );
  }

  vecf4 operator/( const vecf4 & other ) const
  {
	return _mm_div_ps( v, other.v );
  }

  vecf4 & operator+=( const vecf4 & other )
  {
	v = vecf4(v) + other;
	return *this;
  }

  vecf4 & operator-=( const vecf4 & other )
  {
	v = vecf4(v) - other;
	return *this;
  }

  vecf4 operator&( const veci4 & other ) const
  {
	return _mm_and_ps( v, *((__m128*)&other.v) );
  }

  vecf4 operator&( const vecf4 & other ) const
  {
	return _mm_and_ps( v, other.v );
  }

  vecf4 operator|( const vecf4 & other ) const
  {
	return _mm_or_ps( v, other.v );
  }

  vecf4 & operator*=( const vecf4 & other )
  {
	v = vecf4(v) * other;
	return *this;
  }

  vecf4 & operator/=( const vecf4 & other )
  {
	v = vecf4(v) / other;
	return *this;
  }

  operator __m128() const
  {
	return v;
  }

  vecf4 operator==( const vecf4 & other ) const
  {
	return _mm_cmpeq_ps( v, other.v );
  }

  vecf4 operator<( const vecf4 & other ) const
  {
	return _mm_cmplt_ps( v, other.v );
  }

  vecf4 operator>( const vecf4 & other ) const
  {
	return _mm_cmpgt_ps( v, other.v );
  }

  vecf4 operator>=( const vecf4 & other ) const
  {
	return ( (*this) > other ) | ( (*this) == other );
  }

  vecf4 operator<=( const vecf4 & other ) const
  {
	return ( (*this) < other ) | ( (*this) == other );
  }

  void store( float * px, bool aligned = false ) const
  {
	if ( aligned )
	  _mm_store_ps( px, v );
	else
	  _mm_storeu_ps( px, v );
  }

  __m128 v;
};

inline vecf4 operator*( float x, const vecf4 & y )
{
  return vecf4(x) * y;
}

inline vecf4 operator/( float x, const vecf4 & y )
{
  return vecf4(x) / y;
}

inline vecf4 operator+( float x, const vecf4 & y )
{
  return vecf4(x) + y;
}

inline vecf4 operator-( float x, const vecf4 & y )
{
  return vecf4(x) - y;
}

inline veci4 operator*( int32_t x, const veci4 & y )
{
  return veci4(x) * y;
}

inline veci4 operator+( int32_t x, const veci4 & y )
{
  return veci4(x) + y;
}

inline veci4 operator-( int32_t x, const veci4 & y )
{
  return veci4(x) - y;
}

inline vecf4 max( const vecf4 & x, const vecf4 & y )
{
  return _mm_max_ps( x, y );
}

inline vecf4 min( const vecf4 & x, const vecf4 & y )
{
  return _mm_min_ps( x, y );
}

inline vecf4 sqrt( const vecf4 & x )
{
  return _mm_sqrt_ps( x );
}

inline vecf4 pow2( const vecf4 & x )
{
  return x * x;
}

inline vecf4 exp( const vecf4 & x )
{
  return expf4( x );
}

inline vecf4 log( const vecf4 & x )
{
  return logf4( x );
}

inline vecf4 safe_log( const vecf4 & x )
{
  return logf4( max( 1e-10f, x ) );
}

inline float pow2( float x ) { return x * x; }

