/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
//
 #ifndef INVERT_MATRIX_HPP
 #define INVERT_MATRIX_HPP

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/lu.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/traits.hpp>
#include <string>
#include <algorithm>

namespace ublas = boost::numeric::ublas;

 /* Matrix inversion routine.
    Uses lu_factorize and lu_substitute in uBLAS to invert a matrix */
 template<class T>
 bool invertMatrix (const ublas::matrix<T>& input, ublas::matrix<T>& inverse) {
 	using namespace boost::numeric::ublas;
 	typedef permutation_matrix<std::size_t> pmatrix;
 	// create a working copy of the input
 	matrix<T> A(input);
 	// create a permutation matrix for the LU-factorization
 	pmatrix pm(A.size1());

 	// perform LU-factorization
 	int res = lu_factorize(A,pm);
        if( res != 0 ) return false;

 	// create identity matrix of "inverse"
 	inverse.assign(ublas::identity_matrix<T>(A.size1()));

 	// backsubstitute to get the inverse
 	lu_substitute(A, pm, inverse);

 	return true;
 }


template<class M>
double trace(const M &mat) {
      if(mat.size1()!=mat.size2()) throw std::string("trace: matrix not square");
      double sum=0.0;
      for(unsigned i=0;i<mat.size1();++i) sum+=mat(i,i);
      return sum;
}

// Cholesky factorization, originally by Robbie Vogt. See numerical recipes for details. Return false if decomposition failed.
template<class M>
bool choleskyFactorize (M &m,bool nullifyUpper=false) {
	typedef M matrix_type;
	typedef typename M::size_type size_type;
	typedef typename M::value_type value_type;


	if(m.size1()!=m.size2()) return false; // matrix is not square

#ifdef DEBUG
	if(!(ublas::trans(ublas::triangular_adaptor<M,ublas::upper>(inverse))==ublas::triangular_adaptor<M,ublas::lower>(inverse))) return false; // matrix is not symmetric
#endif
	
	size_type size = m.size1();
	ublas::vector<value_type> d(size); // the diagonal of the decomposition.

	for (size_type i = 0; i < size; ++ i) {
		ublas::matrix_row<M> mri(m, i);
      		for (size_type j = i; j < size; ++ j) {
       			ublas::matrix_row<M> mrj(m, j);

        		value_type elem = m(i,j) - ublas::inner_prod(project(mri,ublas::range(0,i)), project(mrj,ublas::range(0,i)));

        		if (i == j) {
          			if (elem <= 0.0) {
            				// matrix after rounding errors is not positive definite
            				return false;
          			}
          			else {
            				d(i) = ublas::type_traits<value_type>::type_sqrt(elem);
          			}
        		}
        		else {
          			m(j,i) = elem / d(i);
        		}
      		}
    	}
    	// put the diagonal back in
    	for (size_type i = 0; i < size; ++ i) {
      		m(i,i) = d(i);
    	}
    	// decomposition succeeded
	if(nullifyUpper) {
		for (size_type i=0;i<size;++i) {
			ublas::matrix_row<M> mri=(ublas::row(m,i));
			std::fill(mri.begin()+i+1,mri.end(),0.0);
		}
	}
    	return true;
}

// Matrix inversion for square, positive definite matrices. Return true on success.
template<class M>
bool invertMatrixCholesky (const M &mat,M &inverse,float &logdet)
{
	typedef typename M::size_type size_type;
	typedef typename M::value_type value_type;

	size_type size = mat.size1();
	inverse=mat;
	if(!choleskyFactorize(inverse)) return false;
	

	// determine the inverse of the lower triangular matrix
    	for (size_type i = 0; i < size; ++ i) {
		inverse(i,i) = 1.0 / inverse(i,i);

      		for (size_type j = i+1; j < size; ++ j) {
        		value_type elem(0);

        		for (size_type k = i; k < j; ++ k) {
          			elem -= inverse(j,k)*inverse(k,i);
        		}
        		inverse(j,i) = elem / inverse(j,j);
      		}
	}
	
	// compute log of determinant of original matrix
	logdet=0.0;
	for (size_type i = 0; i < size; ++ i) logdet+=log(inverse(i,i));
	logdet*=-2.0;    
	

    	// multiply the upper and lower inverses together
    	inverse = ublas::prod(ublas::trans(ublas::triangular_adaptor<M,ublas::lower>(inverse)), ublas::triangular_adaptor<M,ublas::lower>(inverse));
	return true;
}

template<class M>
M invertMatrixCholesky(const M &mat) {
	M inv;
	float ld;
	bool rv=invertMatrixCholesky(mat,inv,ld);
	if(!rv) throw std::string("invertMatrixCholesky: matrix not symmetric and positive definite");
	return inv;
}

template<class M>
M invertMatrixCholesky(const M &mat,float &logdet) {
	M inv;
	bool rv=invertMatrixCholesky(mat,inv,logdet);
	if(!rv) throw std::string("invertMatrixCholesky: matrix not symmetric and positive definite");
	return inv;
}


template<class M>
double logDet(const M &mat) {
	M cholde=mat;
	bool cdok=choleskyFactorize(cholde);
	if(!cdok) throw std::string("invertMatrixCholesky: matrix not symmetric and positive definite");
	double rv=0.0;
	for(unsigned i=0;i<cholde.size1();++i) rv+=log(cholde(i,i));
	return 2.0*rv;
}



// Cholesky substitution
template<class M, class E>
void cholesky_substitute (const M &m, ublas::vector_expression<E> &e) {
	typedef const M const_matrix_type;
	typedef ublas::vector<typename E::value_type> vector_type;

	ublas::inplace_solve (m, e, ublas::lower_tag ());
	ublas::inplace_solve (trans(m), e, ublas::upper_tag ());
}

template<class M, class E>
void cholesky_substitute (const M &m, ublas::matrix_expression<E> &e) {
	typedef const M const_matrix_type;
	typedef ublas::matrix<typename E::value_type> matrix_type;

	ublas::inplace_solve (m, e, ublas::lower_tag ());
	ublas::inplace_solve (trans(m), e, ublas::upper_tag ());
}

template<class E, class M>
void cholesky_substitute_left (ublas::vector_expression<E> &e, const M &m) {
	typedef const M const_matrix_type;
	typedef ublas::vector<typename E::value_type> vector_type;

	ublas::inplace_solve (trans(m), e, ublas::upper_tag ());
	ublas::inplace_solve (m, e, ublas::lower_tag ());
}


template<class E, class M>
void cholesky_substitute_left (ublas::matrix_expression<E> &e, const M &m) {
	typedef const M const_matrix_type;
	typedef ublas::matrix<typename E::value_type> matrix_type;

	ublas::inplace_solve (trans(m), e, ublas::upper_tag ());
	ublas::inplace_solve (m, e, ublas::lower_tag ());
}



#endif //INVERT_MATRIX_HPP
