/*

MATH

  THIS MODULE INCLUDES GENERIC MATH UTILITIES

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <cassert>
#include "time.h"

#include "invertMatrix.h"
#include <boost/foreach.hpp>

using namespace std;

// check that ,matrix mat is close enough to 0
template<class M>
bool checkZero(M &mat, const double tol=1e-7) {
	bool rv=true;
	for(unsigned i=0;i<mat.size1();++i) {
		const ublas::matrix_row<M> crow (mat, i);
		BOOST_FOREACH( double d, crow ) rv=rv & (fabs(d)<1e-8);
	}
	return rv;
}



int main(int argc, char *argv[])
{
  	const unsigned int msize=200;
	const unsigned int timeruns=10;
  	srand48(time(0));
  	using namespace boost::numeric::ublas;
    	matrix<double> m (msize,msize);
    	for (unsigned i = 0; i < m.size1 (); ++ i)
        	for (unsigned j = 0; j < m.size2 (); ++ j)
            		m (i, j) = (i<=j?(i==j?msize:0.0)+drand48():m(j,i)) ;
    	cout << "Matrix is "<< m << endl;

    	matrix<double> inv (msize,msize);
    	assert(invertMatrix (m,inv));
    	cout << "Inverse, LU decomp "<< inv << endl;
    
	matrix<double> check=ublas::prod(m,inv)-ublas::identity_matrix<double>(msize);
	if(checkZero(check)) cout<<"LU inverse OK "<<endl;
	else cout<<"ERROR: matrix * LU inverse is not 1: "<<prod(m,inv)<<endl;

	clock_t start,dt;
	cout<<"*********TIMING TEST, LU-Decomposition*********"<<endl;
	matrix<double> accu=inv;
	
	dt=0;
	for(unsigned i=0;i<timeruns;++i) {
		start=clock();
		invertMatrix (m,inv);
		if(i%10==0) cout<<i<<" of "<<timeruns<<endl;
		dt+=clock()-start;
		accu+=inv;
	}
	cout<<"This took "<<(1.0*dt)/CLOCKS_PER_SEC/timeruns<<"s per inversion "<<endl;
	cout<<"Ignore this line "<<accu(0,0)<<endl;
	


    	matrix<double> cd=m;
    	assert(choleskyFactorize(cd,true));
    	std::cout<<"Cholesky decomposition "<<cd<<std::endl;

    	matrix<double> invc=invertMatrixCholesky(m);
	cout << "Choleksy inverse "<< invc << endl;
	check=ublas::prod(m,invc)-ublas::identity_matrix<double>(msize);
	if(checkZero(check)) cout<<"Cholesky inverse OK "<<endl;
	else cout<<"ERROR: matrix * Cholesky inverse is not 1: "<<prod(m,inv)<<endl;

	cout<<"*********TIMING TEST, Cholesky-Decomposition*********"<<endl;
	
	dt=0;
	for(unsigned i=0;i<timeruns;++i) {
		start=clock();
		invertMatrixCholesky(m,inv);
		if(i%10==0) cout<<i<<" of "<<timeruns<<endl;
		dt+=clock()-start;
		accu+=inv;
	}
	cout<<"This took "<<(1.0*dt)/CLOCKS_PER_SEC/timeruns<<"s per inversion "<<endl;
	cout<<"Ignore this line "<<accu(0,0)<<endl;
    

  return EXIT_SUCCESS;
}
