/*

RANDOM

  THIS MODULES PROVIDES UTILITIES FOR CREATING RANDOM NUMBERS IN VARIOUS DISTRIBUTIONS
  AND CONCURRENTLY

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "Random.hpp"
#include <boost/nondet_random.hpp>

#define INITSEED(seed)    						\
    _seed( seed )                               \
  , _rng_base( seed )                           \
  , _normal( )                  				\
  , _normalgen( _rng_base, _normal)	            \
	  , _uniform01_factor( T(1) / T( _rng_base.max() - _rng_base.min() ) )

template<typename T>
Random<T>::Random()
  : INITSEED( boost::random_device()() )
{
}

template<typename T>
#ifndef _WIN32
Random<T>::Random(unsigned int seed)
#else
Random<T>::Random(boost::uint32_t seed)
#endif
  : INITSEED( seed )
{
}

template<typename T>
T
Random<T>::beta( T alpha, T beta )
{
  T p_true  = gamma( alpha, 1 );
  T p_false = gamma( beta , 1 );
  return p_true / ( p_true + p_false );
}

template<typename T>
T 
Random<T>::gamma( T k, T theta )
{
  if (k < T(1)) 
	{
	  // Weibull algorithm
	  T c = (1 / k);
	  T d = ((1 - k) * std::pow(k, (k / (1 - k))));
	  T u, v, z, e, x;
	  do {
		u = uniform01();
		v = uniform01();
		z = -std::log(u);
		e = -std::log(v);
		x = std::pow(z, c);
	  } while ((z + e) < (d + x));
	  return (x * theta);
	} 
  else 
	{
	  // Cheng's algorithm
#ifndef _WIN32
	  T b = (k - std::log(4));
#else
	T b = (k - std::log(4.0f));
#endif
	  T c = (k + std::sqrt(2 * k - 1));
	  T lam = std::sqrt(2 * k - 1);
#ifndef _WIN32
	  static const T cheng = (1 + std::log(4.5));
#else
	  static const T cheng = (1 + std::log(4.5f));
#endif
	  T u, v, x, y, z, r;
	  do {
		u = uniform01();
		v = uniform01();
		y = ((1 / lam) * std::log(v / (1 - v)));
		x = (k * std::exp(y));
		z = (u * v * v);
		r = (b + (c * y) - x);
	  } while ((r < ((4.5 * z) - cheng)) && (r < std::log(z)));
	  return (x * theta);
	}
}

template<typename T>
T 
Random<T>::normal()
{
  return _normalgen();
}

template<typename T>
T 
Random<T>::normal( T mu, T sigma )
{
  return normal() * sigma + mu;
}

template<typename T>
T 
Random<T>::uniform01()
{
  T result = T(_rng_base() - _rng_base.min()) * _uniform01_factor;
  //  std::cout << result << std::endl;
  return result;
}

template<typename T>
T
Random<T>::uniformab( T a, T b )
{
  return uniform01() * ( b - a ) + a;
}

template<typename T>
unsigned int 
Random<T>::uniform()
{
  return _rng_base();
}

template<typename T>
unsigned int 
Random<T>::seed()
{
  return _seed;
}

template<typename T>
void Random<T>::seed(unsigned int s)
{
#ifndef _WIN32
  _rng_base.seed( _seed = s );
#else
  _rng_base.seed( _seed = s );
#endif
  
}

template<typename T>
typename Random<T>::static_pointer_t Random<T>::_instance; 

template class Random<float>;
template class Random<double>;
