/*

RANDOM

  THIS MODULES PROVIDES UTILITIES FOR CREATING RANDOM NUMBERS IN VARIOUS DISTRIBUTIONS
  AND CONCURRENTLY

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef RANDOM_HPP_M8MFPBEB
#define RANDOM_HPP_M8MFPBEB

#include <boost/thread/tss.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/foreach.hpp>
#include <boost/random/uniform_01.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/variate_generator.hpp>
#include "approx_logexp.h"

#ifdef HAVE_SSE
#include "rand_sse.hpp"
#endif

/// easy to use random number gen
/// uses a nondet random device as seeder
/// allows controlling the seed
/// has convenience functions to generate draws from the
/// most important distributions
template<class T>
class Random
{
public:
  typedef Random<T> This;  
  /// returns a random number from the gamma distribution
  T gamma( T k, T theta = 1 );
  /// returns a random number from the normal distribution with sigma = 1, mu = 0
  T normal();
  /// returns a random number from the normal distribution with given sigma and mu
  T normal( T mu, T sigma );
  /// returns a random number from the uniform distribution [0,1]
  T uniform01();
  /// returns a random number from the uniform distribution [a,b]
  T uniformab( T a, T b );
  /// returns a number from Beta(alpha,beta)
  T beta( T alpha, T beta );
  /// returns a uint from the base prng
  unsigned int uniform();
  /// makes a prng with a seed from the systems nondeterministic rng
  Random();
  /// makes a prng with a given seed
  Random( unsigned int );
#ifdef _WIN32
  Random( boost::uint32_t );
#endif
  /// returns the seed that started this prng
  unsigned int seed();
  /// seeds and resets the prng
  void seed(unsigned int);
  template<class Container1, class Container2>
  inline void sample_dirichlet( const Container1 & alpha, Container2 & p, T alphamin = 0 )
  {
	T s = 0;
	typename Container1::const_iterator ialpha = alpha.begin(), iend = alpha.end();
	typename Container2::iterator ip = p.begin();
	for ( ; ialpha != iend; ++ialpha, ++iend )
	  s += ( *ip = std::max( alphamin, gamma( *ialpha ) ) );
	BOOST_FOREACH( T & p_i, p )
	  p_i /= s;
  }
  template<class Container>
  inline void sample_dirichlet( T alpha0, Container & p, T alphamin = 0 )
  {
	T s = 0;
	BOOST_FOREACH( T & p_i, p )
	  s += ( p_i = std::max( alphamin, gamma( alpha0 ) ) );
	BOOST_FOREACH( T & p_i, p )
	  p_i /= s;
  }  
  /// container type agnostic multinomial distribution sampler
  template<class Container>
  inline unsigned int sample_multinomial( const Container & p, T Zp = T(1) )
  {
	T upper = 0;
	T u = uniform01() * Zp;
	unsigned int last_valid = 0, count = 0;
	for ( typename Container::const_iterator i = p.begin(), e = p.end(); i != e; ++i )
	  {
		T p_i = *i;
		if ( p_i > 0 )
		  {
			upper += p_i;
			if ( u <= upper)
			  return count;
			last_valid = count;
		  }
		++count;
	  }
	return last_valid;
  }
  /// container type agnostic multinomial distribution sampler
  template<class Container>
  inline unsigned int sample_log_multinomial( const Container & lp )
  {
	T lmax = -1e100;
	for ( typename Container::const_iterator i = lp.begin(), e = lp.end(); i != e; ++i )
	  lmax = std::max( lmax, *i );
	T Z = 0;
	for ( typename Container::const_iterator i = lp.begin(), e = lp.end(); i != e; ++i )
	  Z += aexpf( *i - lmax );
	return sample_log_multinomial( lp, lmax, Z );
  }
  /// container type agnostic multinomial distribution sampler
  template<class Container>
  inline unsigned int sample_log_multinomial( const Container & lp, T l_off, T Z )
  {
	T upper = 0;
	T u = uniform01() * Z;
	unsigned int last_valid = 0, count = 0;
	for ( typename Container::const_iterator i = lp.begin(), e = lp.end(); i != e; ++i )
	  {
		T p_i = aexpf( *i - l_off );
		if ( p_i > 0 )
		  {
			upper += p_i;
			if ( u <= upper)
			  return count;
			last_valid = count;
		  }
		++count;
	  }
	return last_valid;
  }
  //
  inline bool sample_binomial( T p = T(0.5) )
  {
	return p < T(1) ? uniform01() < p : true;
  }
  static Random<T> & get()
  {
	if ( _instance.get() == NULL )
	  _instance.reset( new Random<T>() );
	return *_instance.get();
  }
private:
  typedef boost::mt19937 BaseRNG;
#ifndef _WIN32
  unsigned int _seed;
#else
  boost::uint32_t _seed;
#endif
  typedef boost::mt19937 rng_base_t;
  rng_base_t _rng_base;
  // normal  
  boost::normal_distribution<T> _normal;
  boost::variate_generator< rng_base_t&, boost::normal_distribution<T> > _normalgen;
  // uniform 01
  T _uniform01_factor;
  // TLS instance
  typedef  boost::thread_specific_ptr< Random<T> > static_pointer_t;
  static static_pointer_t _instance; 
};

#endif /* end of include guard: RANDOM_HPP_M8MFPBEB */
