/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "BatchNG.hpp"
#include "blitz_indices.hpp"
#include "metrics.hpp"

#define EUCLID_METRIC 1
#define MANHATTAN_METRIC 2

#define MAX_ULONG ((unsigned long) ~0L)

template< typename T, typename Metric >
BatchNG<T,Metric>::BatchNG( bool do_estimate_variances )
	: _lambda_i(10.)
	, _lambda_f(0.01)
	, _itmax(20)
	, _minIter(0)
	, _maxIter(0)
	, _ctreshold(1e-05)
	, _ready(false)
	, _k(0)
	, _d1(0)
	, _d2(0)
	, _m(0)
	, _initialized(false)
	, _do_estimate_variances(do_estimate_variances)
{

}

template< typename T, typename Metric >
BatchNG<T,Metric>::~BatchNG()
{

}

template<typename T, typename Metric>
void BatchNG<T,Metric>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter(&This::_initialized,false);
  PBase::add_ivar_property_rw("lambda_i", &This::_lambda_i, uninitializer);
  PBase::add_ivar_property_rw("lambda_f", &This::_lambda_f, uninitializer);
  PBase::add_ivar_property_rw("itmax", &This::_itmax, uninitializer);
  PBase::add_ivar_property_rw("threshold", &This::_ctreshold, uninitializer);
  PBase::add_ivar_property_rw("minIter", &This::_minIter, uninitializer);
  PBase::add_ivar_property_rw("maxIter", &This::_maxIter, uninitializer);
  PBase::add_ivar_property_rw("k", &This::_k, uninitializer);
  PBase::add_ivar_property_rw("d1", &This::_d1, uninitializer);
  PBase::add_ivar_property_rw("d2", &This::_d2, uninitializer);
  PBase::add_ivar_property_rw("m", &This::_m, uninitializer);
  PBase::add_ivar_property_rw("do-estimate-variances", &This::_do_estimate_variances, uninitializer);
  
  PBase::add_ivar_property_ro("ready", &This::_ready);
  PBase::add_ivar_property_ro("initialized", &This::_initialized);
  PBase::add_ivar_property_ro("itnum", &This::_it);
  //PBase::add_ivar_property_ro("costs", &This::_costs);
  //PBase::add_ivar_property_ro("lastcosts", &This::_lastcosts);
  
  PBase::add_property( "centers", make_blitz_ivar_getter1( &This::_centers ) );
  PBase::add_property( "rcounts", make_blitz_ivar_getter1( &This::_weightStat ) );
  PBase::add_property( "rcounts3", make_blitz_ivar_getter1( &This::_weightStat3 ) );
  PBase::add_property( "ranks", make_blitz_ivar_getter1( &This::_ranks ) );
}

template< typename T, typename Metric >
blitz::Array<double,1> BatchNG<T,Metric>::get_histogram()
{
	return _weightStat;
}

template< typename T, typename Metric >
blitz::Array<double,3> BatchNG<T,Metric>::get_histogram3()
{
	return _weightStat3;
}

template< typename T, typename Metric >
blitz::Array<T,4> BatchNG<T,Metric>::get_centers()
{
	Matrix4 c ( _centers );
	return c;
}

template< typename T, typename Metric >
blitz::Array<T,4> BatchNG<T,Metric>::get_variances()
{
   BLITZ_INDICES;
   if ( _varianceStat.extent(0) == _k )
	 {
	   Matrix4 variances( _k, _d1, _d2, _m );
	   for ( int k = 0; k < _k; ++k )
		 for ( int m = 0; m < _m; ++m )
		   variances( k, _all, _all, m ) = 
			 _varianceStat( k, _all, _all, m ) /  max( _weightStat3( k, _all, _all ), T(0.0000001) )
			 - pow2( _centers( k, _all, _all, m ) );
	   return variances;
	 }
   else
	 return Matrix4();
}

template< typename T, typename Metric >
blitz::Array<double,4> BatchNG<T,Metric>::get_squareStat()
{
  BLITZ_INDICES;
  return _varianceStat;
}


template< typename T, typename Metric >
blitz::Array<double,4> BatchNG<T,Metric>::get_centerStat()
{
  BLITZ_INDICES;
  return _centerStat;
}


template< typename T, typename Metric >
void BatchNG<T,Metric>::initialize(int k, int d1, int d2, int m)
{
  if (!_initialized || (_k != k) || (_d1 != d1) || (_d2 != d2) || (m != _m) )
	{
		_m = m;
		_k = k;
		_d1 = d1;
		_d2 = d2;

		_centers.resize    ( _k, _d1, _d2, m );
		_centerStat.resize ( _k, _d1, _d2, m );
		_weightStat.resize ( _k );
		_weightStat3.resize( _k, _d1, _d2 );
		if (_do_estimate_variances)
		  _varianceStat.resize( _k, _d1, _d2, _m );
		_ranks.resize( _k );

		_ready = false;
		_it = 0;
		_estimate_variances = false;
		_costs = 0;

	}

	_initialized = true;
}


template< typename T, typename Metric >
void BatchNG<T,Metric>::begin_epoch()
{
  initialize(_k, _d1, _d2, _m);
	
	if ( (_k < 2) || (_d1 < 1) || (_d2 < 1) || (_m < 1) )
	  throw( std::runtime_error( str( boost::format("BatchNG initialzed with k=%d d1=%d d2=%d m=%d") % _k % _d1 % _d2 % _m ) ) );
	
	_lambda = _lambda_i * pow( _lambda_f / _lambda_i, T(_it) / T(_itmax));
	
	for (int i = 0; i < _k; i++)
		_ranks(i) = std::exp( -T(i) / _lambda );
	
	_ranks /= sum( _ranks );

	_centerStat  = 0;
	_weightStat  = 0;
	_weightStat3 = 0;
	_lastcosts = _costs;
	_costs = 0;
	
	if (_estimate_variances)
		_varianceStat = 0;
}


template< typename T, typename Metric >
void BatchNG<T,Metric>::end_epoch()
{
   BLITZ_INDICES;
   _centers = _centerStat(i1, i2, i3, i4) / max( _weightStat3(i1,i2,i3), T(0.0000001) );
	_it++;
	
	if (_lastcosts > 0)
	{
		if (_estimate_variances)
			_ready = true;
		
		T progress = ( _lastcosts - _costs ) / _lastcosts;
		std::cout << "batchNG progress = " << progress << " costs " << _costs << " lastcosts " << _lastcosts << std::endl;
		
		if ( !_ready && ( (_it >= _minIter) && (( progress < _ctreshold ) || (_it >= _maxIter) ) ) )
		{
			if (_do_estimate_variances)
				_estimate_variances = true;
			else 
				_ready = true;
		}
	}
}


template<typename T>
struct validx
{
   T val;
	int idx;
	
	bool operator< (const validx<T> s2) const
	{
		return (val < s2.val);
	}
};


template< typename T, typename Metric >
blitz::Array<T,1> BatchNG<T,Metric>::compute_ranking( const Matrix1 & dists )
{
	Matrix1 ranks;
#ifndef _WIN32
	validx<T> vilist[_k];
#else
	validx<T>* vilist = new validx<T>[_k];
#endif
	
	for (int i = 0; i < _k; i++)
	{
		vilist[i].val = dists(i);
		vilist[i].idx = i;
	}
	
	std::sort(vilist, vilist + _k);
	ranks.resize(_k);

	for (int i = 0; i < _k; i++)
		ranks( vilist[i].idx ) = _ranks(i);
	
	return ranks;
}

template< typename T, typename Metric >
blitz::Array<T,1> BatchNG<T,Metric>::rand_ranking( )
{
	Matrix1 ranks ( _ranks.copy() );
	std::random_shuffle( ranks.data(), ranks.data() + ranks.numElements() );
	
	return ranks;
}


template< typename T, typename Metric >
void BatchNG<T,Metric>::process(const Matrix3 & data)
{
   BLITZ_INDICES;
	// data: [ d x d x m ]
	// _centers: [ k x d x d x m ]
	Matrix1 weights;
	
	if (_it > 0)
	{
	  Matrix1 dists = compute_dists( data );
	  weights.reference( compute_ranking( dists ) );		
	  _costs += sum( weights * pow2( dists ) );
		//std::cout << "costs update = " << sum( weights * pow2( dists ) ) << std::endl;
	}
	else
		weights.reference( rand_ranking() );

	//std::cout << "weights.shape: " << weights.shape() << std::endl;
	
	_centerStat  += weights(i1) * data(i2, i3, i4);
	_weightStat  += weights;
	_weightStat3 += weights(i1) + 0 * i2 * i3;
	
	if (_estimate_variances)
		_varianceStat += weights(i1) * pow2( data(i2,i3,i4) ); 
}


template< typename T, typename Metric >
void BatchNG<T,Metric>::process(const Matrix3 & data, const Matrix2 & saliency)
{
   BLITZ_INDICES;
	// data: [ d x d x m ]
	// _centers: [ k x d x d x m ]
	Matrix1 weights1;
	Matrix3 weights3;

	T fw = max ( saliency );

	if (_it > 0)
	  {
		Matrix1 dists = compute_dists( data, saliency );
		Matrix1 ranking = compute_ranking( dists );
		weights1.reference( Matrix1( fw * ranking ) );
		weights3.reference( Matrix3( saliency(i2,i3) * ranking(i1) ) );
		_costs += sum( ranking * pow2( dists ) );
	  }
	else
	  {
		weights1.reference( Matrix1( fw * rand_ranking( ) ) );
		weights3.resize( _k, _d1, _d2 );
		weights3 = weights1(i1) + i2 * i3 * 0;
	  }
	
	_centerStat += weights3(i1,i2,i3) * data(i2, i3, i4);
	_weightStat  += weights1;
	_weightStat3 += weights3;
	
	if (_estimate_variances)
	  {
		_varianceStat += weights3(i1,i2,i3) * pow2( data(i2, i3, i4) );
	  }
}


template< typename T, typename Metric >
blitz::Array<T,1> BatchNG<T,Metric>::compute_dists( const Matrix3 & data )
{
  BLITZ_INDICES;
  Matrix1 dists(_k);
  for ( int i = 0; i < _k; ++i )
	dists(i) = Metric::dist( data, _centers( i, _all, _all, _all ) );
  return dists;
}

template< typename T, typename Metric >
blitz::Array<T,1> BatchNG<T,Metric>::compute_dists( const Matrix3 & data, const Matrix2 & weights )
{
  BLITZ_INDICES;
  Matrix1 dists(_k);
  for ( int i = 0; i < _k; ++i )
	dists(i) = Metric::dist( data, _centers( i, _all, _all, _all ), weights( i1, i2 ) );
  return dists;
}

/// explicit instatiation of the BatchNG for float data
template class BatchNG< float, manhattan_metric<float> >;
template class BatchNG< float, euclidian_metric<float> >;


