/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include <blitz/range.h>
#include <algorithm>

#define idivup(x,y) (((x)+(y)-1)/(y))

#ifdef HAVE_OPENMP
# include "omp_macros.hpp"
#endif

class ForwardRFIterator
{

public:

  ForwardRFIterator( int n1, int n2, int d1, int d2, int step1, int step2, bool openmp_parallel = false )
	: _n1(n1)
	, _n2(n2)
	, _d1(d1)
	, _d2(d2)
	, _m1( 1 + ( n1 - d1 ) / step1 )
	, _m2( 1 + ( n2 - d2 ) / step2 )
	, _step1(step1)
	, _step2(step2)
  {
	_n1 -= ( _n1 - _d1 ) % _step1;
	_n2 -= ( _n2 - _d2 ) % _step2;
#ifdef HAVE_OPENMP
	if ( openmp_parallel )
	  {
		blitz::Range r1 = WORK_RANGE(_m1);		
		_j10 = r1.first();
		_j20 = 0;
		_j11 = r1.last();
		_j21 = _m2 - 1;
	  }
#endif
	{
	  _j10 = 0;
	  _j20 = 0;
	  _j11 = _m1 - 1;
	  _j21 = _m2 - 1;
	}
	reset_j();
  }  

  int i1, i2, j1, j2;

  bool atend_j() const
  {
	return ( j1 > _j11 );
  }

  void reset_j()
  {
	j1 = 0;
	j2 = 0;
	calc_bounds();	
  }

  size_t size_j() const { return _m1 * _m2; }

  bool atend_i() const
  {
	return i1 > _i11;
  }

  void increment_i()
  {
	if ( ++i2 > _i21 )
	  {
		i2 = _i20;
		++i1;
	  }
  }

  void reset_i()
  {
	i1 = _i10;
	i2 = _i20;
  }

  void increment_j()
  {
	if ( ++j2 > _j21 )
	  {
		j2 = _j20;
		++j1;
	  }
	calc_bounds();
  }

  void move_to( int _j )
  {
	move_to( _j / _m2, _j % _m2 );
  }

  void move_to( int _j1, int _j2 )
  {
	j1 = _j1;
	j2 = _j2;
	calc_bounds();
  }

  int i1_rf() const { int i1rf = i1 - j1 * _step1; assert(i1rf>=0); assert(i1rf<_d1); return i1rf; }

  int i2_rf() const { int i2rf = i2 - j2 * _step2; assert(i2rf>=0); assert(i2rf<_d2); return i2rf; }

  blitz::Range in_range1() const { return blitz::Range( _i10, _i11 ); }

  blitz::Range in_range2() const { return blitz::Range( _i20, _i21 ); }

  float r2_rf() const
  { 
	float d1 = ( 2.0f * i1_rf() - _d1 ) / _d1;
	float d2 = ( 2.0f * i2_rf() - _d2 ) / _d2;
	return d1 * d1 + d2 * d2;
  }

  int j_index() const
  {
	return _j_index;
  }


  void calc_bounds()
  {
	int n1 = idivup( _d1, _step1 );
	int n2 = idivup( _d2, _step2 );
	_i10 = _step1 * j1;
	_i11 = _step1 * j1 + _d1 - 1;
	_i20 = _step2 * j2;
	_i21 = _step2 * j2 + _d2 - 1;
	_j_index = j2 % n2 + n2 * ( j1 % n1 );
	reset_i();
  }

  int _n1, _n2, _d1, _d2, _m1, _m2, _i10, _i11, _i20, _i21, _step1, _step2, _j10, _j11, _j20, _j21, _j_index;
};

#undef idivup
