/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include <blitz/range.h>
#include <algorithm>

#ifdef HAVE_OPENMP
# include "omp_macros.hpp"
#endif

#define idivup(x,y) (((x)+(y)-1)/(y))

class InverseRFIterator
{

public:

  InverseRFIterator( int n1, int n2, int d1, int d2, int step1, int step2, bool openmp_parallel = false )
	: _n1(n1)
	, _n2(n2)
	, _d1(d1)
	, _d2(d2)
	, _m1( 1 + ( n1 - d1 ) / step1 )
	, _m2( 1 + ( n2 - d2 ) / step2 )
	, _step1(step1)
	, _step2(step2)
  {
	_n1 -= ( _n1 - _d1 ) % _step1;
	_n2 -= ( _n2 - _d2 ) % _step2;
#ifdef HAVE_OPENMP
	if ( openmp_parallel )
	  {
		blitz::Range r1 = WORK_RANGE(_n1);
		_i10 = r1.first();
		_i20 = 0;
		_i11 = r1.last();
		_i21 = _n2 - 1;
	  }
	else
#endif
	  {
		_i10 = _i20 = 0;
		_i11 = _n1 - 1;
		_i21 = _n2 - 1;
	  }
	reset_i();
  }  

  int i1, i2, j1, j2;
  blitz::Range jr1, jr2;

  bool atend_j() const
  {
	return ( j1 > _j11 );
  }

  bool atend_i() const
  {
	return i1 > _i11;
  }

  size_t size_i()
  {
    return ( _i11 - _i10 + 1 ) * ( _i21 - _i20 + 1 );
  }

  void move_to( int ii )
  {
    size_t columns = ( _i21 - _i20 + 1 );
    move_to( _i10 + ii / columns, _i20 + ii % columns );
  }

  void move_to( int ii1, int ii2 )
  {
	i1 = ii1; i2 = ii2;
	calc_bounds();
	reset_j();
  }


  void reset_i()
  {
	i1 = _i10;
	i2 = _i20;
	calc_bounds();
  }

  void increment_i()
  {
	if ( ++i2 > _i21 )
	  {
		i2 = 0;
		++i1;
	  }
	calc_bounds();
  }

  void reset_j()
  {
	j1 = _j10;
	j2 = _j20;
  }

  void increment_j()
  {
	if ( ++j2 > _j21 )
	  {
		j2 = _j20;
		++j1;
	  }
  }

  int jj1() const { return j1 - _j10; }

  int jj2() const { return j2 - _j20; }

  int i1_rf() const { int i1rf = i1 - j1 * _step1; assert(i1rf>=0); assert(i1rf<_d1); return i1rf; }

  int i2_rf() const { int i2rf = i2 - j2 * _step2; assert(i2rf>=0); assert(i2rf<_d2); return i2rf; }

  float r2_rf() const
  { 
	float d1 = ( 2.0f * i1_rf() - _d1 ) / _d1;
	float d2 = ( 2.0f * i2_rf() - _d2 ) / _d2;
	return d1 * d1 + d2 * d2;
  }

  void select_j_index( int j_i )
  {
#if 0
	int n1 = idivup( _d1, _step1 );
	int n2 = idivup( _d2, _step2 );
	int j10 = std::max( 0, ( i1 - _d1 ) / _step1 + 1 ); // this is the index of the first parent of this input in the 1st dimension
	int j20 = std::max( 0, ( i2 - _d2 ) / _step2 + 1 ); // this is the index of the first parent of this input in the 2nd dimension
	int ji10 = j10 % np1; // this is the j_index of j0 in the 1st dimension
	int ji11 = j_i / np2; // this is j_index in the 1st dimension
	int ji20 = j20 % n2; // this is the j_index of j0 in the 2nd dimension
	int ji21 = j_i % n2; // this is j_index in the 2nd dimension
	j1 = ( ji11 >= ji10 ) ? ( j10 + ji11 - ji10 ) : ( j10 + np1 - ji10 + ji11 );
	j2 = ( ji21 >= ji20 ) ? ( j20 + ji21 - ji20 ) : ( j20 + np2 - ji20 + ji21 );
#else
	reset_j();
	while( j_index() != j_i )
	  {
		increment_j();
		assert(!atend_j());
	  }
#endif
  }

  int j_index() const
  {
	int n1 = idivup( _d1, _step1 );
	int n2 = idivup( _d2, _step2 );
	return j2 % n2 + n2 * ( j1 % n1 );
  }

  int count_parents() const 
  {
	return ( _j11 - _j10 + 1 ) * ( _j21 - _j20 + 1 );
  }

  //private:

  void calc_bounds()
  {
	_j10 = std::max( 0, idivup( i1 - _d1 + 1, _step1 ) );
	_j11 = std::min( _m1 - 1, i1 / _step1 );
	_j20 = std::max( 0, idivup( i2 - _d2 + 1, _step2 ) );
	_j21 = std::min( _m2 - 1, i2 / _step2 );	
	jr1 = blitz::Range( _j10, _j11 );
	jr2 = blitz::Range( _j20, _j21 );
	reset_j();
  }

  int _n1, _n2, _d1, _d2, _m1, _m2, _j10, _j11, _j20, _j21, _step1, _step2, _i10, _i11, _i20, _i21;
};

#undef idivup
