/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _MIXTUREMODEL2BASE_HPP_
#define _MIXTUREMODEL2BASE_HPP_

#include <blitz/array.h>

#include <boost/tuple/tuple.hpp>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/version.hpp>
#include <boost/serialization/list.hpp>
#include <boost/format.hpp>
#include <boost/array.hpp>

#include "random_permutation.hpp"
#include "SeariseProperties.hpp"
#include "enum_properties.hpp"
#include "blitz_properties.hpp"
#include "mmm2_belief_message_buffer.hpp"
#include "message_component.hpp"
#include "ReceptiveFieldIterator.hpp"

#include "segmentation_tools.hpp"
#include "Random.hpp"
#include "digamma.hpp"
#include "sorting_tools.hpp"

#include "array_utils.hpp"

#include "reduced_rank_covariance.hpp"

#include "omp_macros.hpp"

#include "stopwatch.hpp"

#include "emwa.hpp"

#include "InverseRFIterator.hpp"
#include "ForwardRFIterator.hpp"
#include "masking.cc"
#include "gammaApprox.h"

#ifdef HAVE_SSE
# include "sse_math.hpp"
# include "sse_templates.hpp"
#endif
// #include <boost/accumulators/accumulators.hpp>
// #include <boost/accumulators/statistics/min.hpp>
// #include <boost/accumulators/statistics/mean.hpp>
// #include <boost/accumulators/statistics/median.hpp>
// #include <boost/accumulators/statistics/max.hpp>
// #include <boost/accumulators/statistics/stats.hpp>
// #include <boost/accumulators/statistics/error_of.hpp>
// #include <boost/accumulators/statistics/error_of_mean.hpp>

#include <boost/utility/value_init.hpp>

//namespace acc = boost::accumulators;

#include "message_iterators.hpp"

#define DISPATCH(F) static_cast<Derived*>(this)->F
typedef enum { BACKPROP, LIKELIHOOD, UNLIKELIHOOD, CANONICAL } GradientMode; // this triggers how the input gradient is interpreted

template<class Derived>
class MixtureModel2Base 
  : public HasProperties< SearisePropertyTypes, MixtureModel2Base< Derived > >
  , public message_component< typename TypeTraits<Derived>::num_type >
{
  
public:
  
  typedef typename TypeTraits<Derived>::num_type T;
  //typedef typename Derived::num_type T;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,4> Matrix4;
  
  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<double,4> Matrix4d;
  
  typedef mmm2_belief_message_buffer<T> message_buffer;
  
  typedef blitz::Array<unsigned int,2> IMatrix2;
  typedef blitz::Array<bool,3> BMatrix3;
  typedef blitz::Array<bool,2> BMatrix2;
  typedef blitz::Array<bool,1> BMatrix1;
  typedef MixtureModel2Base< Derived > This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  typedef SeariseProperty PropertyType;
   
  static const size_t num_emission_statistics = TypeTraits<Derived>::num_statistics;
  typedef boost::array<Matrix4 ,num_emission_statistics> emission_parameters_t;
  typedef boost::array<Matrix4d,num_emission_statistics> emission_statistics_t;
  typedef boost::array<Matrix1d,num_emission_statistics> background_statistics_t;
  typedef boost::array<Matrix4 ,num_emission_statistics> emission_gradient_t;
  typedef boost::array<Matrix1 ,num_emission_statistics> background_parameters_t;
  typedef reduced_rank_covariance<T,3> bitwise_emission_gradient_covariance_t;
  typedef reduced_rank_covariance<T,1> prior_gradient_covariance_t;

  typedef enum { NONE, FSD, TSB, FB } BasePrior;
  typedef enum { NO_PARAMETERS, GENERATIVE, VARIATIONAL, DISCRIMINATIVE } ParameterMode;
  typedef enum { PLAIN, LEVENBERG_MARQUARDT, NATURAL_GRADIENT, ADAPTIVE_NATURAL_GRADIENT } GradientDescentStrategy;
  typedef enum { NO_NEIGHBORS, FOUR_NEIGHBORS, EIGHT_NEIGHBORS } Neighborhood;

  MixtureModel2Base()
	: _k(0), _d1(0), _d2(0), _m(0), _step1(0), _step2(0)
	, _s_mrf_iters(3)
	, _alpha0(0.5)
	, _alpha0g(0.5)
	, _mu_m0(0.5)
	, _mu_m_prior(0.5)
	, _mu_b0(0.5)
	, _initcount(1)
	, _Ds(0)
	, _s0(0)
	, _z0(0.5)
	, _gate_W_neighbor(5)
	, _gate_clustering_alpha0(0.5)
	, _delta(1)
	, _tau(1e10)
	, _tau_p(100)
	, _lambda_p(0.1)
	, _anneal_step(0)
	, _s_prior_attenuation(0.5)
	, _convergence_rate(1e-5)
	, _discriminative_generative_coupling(0)
	, _gradient_descent_rate(0.001)
	, _max_gradient_descent_rate(0.01)
	, _gradient_descent_momentum(0.01)
	, _lambda_ng(1e-10)
	, _Lz(0)
	, _Lp(0)
	, _Lmu(0)
	, _L(0)
	, _last_rate(1)
	, _current_rate(1)
	, _linear_annealing(false)
	, _sparsity_annealing(false)
	, _sparsity_annealing_time(0.6)
	, _min_rf_saliency(0)
	, _renormalize_counts(false)
	, _segment(false)
	, _initialized(false)
	, _dirty(false)
	, _do_vb(true)
	, _learning(false)
	, _learning_likelihood(false)
	, _online_update(false)
	, _estimating_hessian(false)
	, _report_gradient(false)
	, _delta_to_random(true)
	, _report_convergence(false)
	, _sequential_update(false)
	, _init_uniform(false)
	, _sampled(false)
	, _converged(true)
	, _s_converged(true)
	, _cache_z(false)
	, _dependent_gates(false)
	, _have_bits(false)
	, _optimize_separability(false)
	, _diagonal_output_jacobian(false)
	, _have_background_model(false)
	, _use_detection_sigmoid(false)
	, _optimize_detection_sigmoid_gain(true)
	, _rescale_detection_likelihoods(false)
	, _continuous_sampling(true)
	, _and_combine_bits_and_gates(true)
	, _learn_background_model(false)
	, _cluster_gates(false)
	, _gamma_mlr(0)
	, _base_prior(NONE)
	, _current_parameter_mode(NO_PARAMETERS)
	, _requested_parameter_mode(GENERATIVE)
	, _gradient_mode(BACKPROP)
	, _gradient_descent_strategy(PLAIN)
	, _gate_neighborhood(FOUR_NEIGHBORS)
	, _epoch_count(0)
	, _gd_count(0)
  {
  }
  
  void set_d( typename PBase::PropertyType v )
  {
	int d = cast<int>( v );
	_d1 = d;
	_d2 = d;
  }
  
  void set_step( typename PBase::PropertyType v )
  {
	int step = cast<int>( v );
	_step1 = step;
	_step2 = step;
  }

  typename PBase::PropertyType get_d( )
  {
	return std::max( _d1, _d2 );
  }

  typename PBase::PropertyType get_separability( )
  {
	return _likelihood_stats_emwa.separability();
  }

  typename PBase::PropertyType get_step( )
  {
    return std::max( _step1, _step2 );
  }

  typename PBase::PropertyType get_mask_coverage( )
  {
	BLITZ_INDICES;
	typedef std::map< int, boost::value_initialized<int> >  histogram_t;
	histogram_t histogram;
	CPUData<float> m( _mask_sample.extent(0), _mask_sample.extent(1), 1 );
	for ( int i = 0; i <  _mask_sample.extent(0); ++i )
	  for ( int j = 0; j <  _mask_sample.extent(1); ++j )
		{
		  int N = sum( where( _mask_sample( i, j, _all ), 1, 0 ) );
		  int N2 = 0;
		  for ( int k = 0; k < _mask_sample.extent(2); ++k )
			if ( _mask_sample( i, j, k ) ) ++N2;
		  if ( N != N2 ) 
			{
			  std::cout << " mask contraction failed blitz: " << N << " manual: " << N2 << std::endl;
			  std::cout << " bits: " << _mask_sample( i, j, _all ) << std::endl;
			  assert(0);
			}
		  histogram[N] += 1;
		  m( i, j, 0 ) = N;
		}
	/*
	std::cout << " mask shape " << _mask_sample.shape() << std::endl;
	std::cout << " mask histogram: " << std::endl;
	for ( histogram_t::iterator i = histogram.begin(), e = histogram.end(); i != e; ++i )
	  std::cout << "  " << i->first << ": " << i->second << std::endl;
	std::cout << " mask coverage " << m.min() << ", " << m.mean() << ", " << m.max() << std::endl;
	*/
	return m;
  }

  typename PBase::PropertyType get_mu_mask( )
  {	
	CPUData<float> cmug( _mu_mask.extent(0),  _mu_mask.extent(1),  _mu_mask.extent(2) );
	for ( int i = 0; i <  _mu_mask.extent(0); ++i )
	  for ( int j = 0; j <  _mu_mask.extent(1); ++j )
		for ( int k = 0; k <  _mu_mask.extent(2); ++k )
		  cmug( i, j, k ) = _mu_mask( i, j, k );
	return cmug;
  }

  typename PBase::PropertyType get_bits( )
  {
	if ( _sampled )
	  {
		CPUData<float> bitsc( _bits_sample.extent(0), _bits_sample.extent(1), 1 );
		std::copy( _bits_sample.begin(), _bits_sample.end(), bitsc.data() );
		return bitsc;
	  }
	else
	  {
		CPUData<float> bitsc( _bits_marginal.extent(0), _bits_marginal.extent(1), 1 );
		std::copy( _bits_marginal.begin(), _bits_marginal.end(), bitsc.data() );
		return bitsc;
	  }
  }
  
  typename PBase::PropertyType get_gate_coverage( )
  {
    if ( !segment() ) return CPUData<float>();
	if ( _sampled )
	  {
		blitz::TinyVector<int,2> s2;
		boost::tie(s2[0],s2[1]) = calc_outsize(_gates_sample.extent(0),_gates_sample.extent(1));
		CPUData<float> coverage( s2[0], s2[1], 1 );
		ForwardRFIterator iter( _gates_sample.extent(0), _gates_sample.extent(1), _d1, _d2, _step1, _step2, true );
		for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		  {
			float c = 0, N = 0;
			for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
			  {
				if ( int( _gates_sample( iter.i1, iter.i2 ) ) == iter.j_index() )
				  c += 1;
				N += 1;
			  }
			coverage( iter.j1, iter.j2, 0 ) = c / N;
		  } 
		return coverage;
	  }
	else
	  {
		CPUData<float> gates = boost::get< CPUData<float> >( get_gates() );
		blitz::TinyVector<int,2> s2;
		boost::tie(s2[0],s2[1]) = calc_outsize( gates.rows(), gates.columns() );
		CPUData<float> coverage( s2[0], s2[1], 1 );
		ForwardRFIterator iter( gates.rows(), gates.columns(), _d1, _d2, _step1, _step2, true );
		for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		  {
			float & c = coverage( iter.j1, iter.j2, 0 );
			c = 0;
			float N = 0;
			for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
			  {
				c += gates( iter.i1, iter.i2, iter.j_index() );
				N += 1;
			  }
			c /= N;
		  } 
		return coverage;
    }
  }

  typename PBase::PropertyType get_gates( )
  {
	BLITZ_INDICES;
	if ( _sampled )
	  {
		int max_n_parents = ( ( _d1 + _step1 - 1 ) / _step1 ) * ( ( _d2 + _step2 - 1 ) / _step2 );
		CPUData<float> cgates( _gates_sample.extent(0), _gates_sample.extent(1), max_n_parents );
		for ( int i = 0; i < _gates_sample.extent(0); ++i )
		  for ( int j = 0; j < _gates_sample.extent(1); ++j )
			{
			  CPUData<float>::ColumnT column = cgates( i, j );
			  column = 0.0f;
			  int gate = _gates_sample( i, j );
			  assert( gate < max_n_parents );
			  column[ gate ] = 1.0f;
			}
		return cgates;
	  }
	else
	  {
		Matrix3 gates( _zbuffer.shape() );
		gates = safe_log( _zbuffer ) + safe_log( _sbuffer );
		if ( _have_bits )
		  gates += safe_log( _bbuffer );
		normalize_rho( gates );	
		CPUData<float> cgates( gates.extent(0), gates.extent(1), gates.extent(2) );
		std::copy( gates.data(), gates.data() + gates.numElements(), cgates.data() );
		return cgates;
	  }
  }

  typename PBase::PropertyType get_sbuffer( )
  {
	BLITZ_INDICES;
	CPUData<float> csbuffer( _sbuffer.extent(0), _sbuffer.extent(1), _sbuffer.extent(2) );
	std::copy( _sbuffer.data(), _sbuffer.data() + _sbuffer.numElements(), csbuffer.data() );
	return csbuffer;
  }

  typename PBase::PropertyType get_zbuffer( )
  {
	BLITZ_INDICES;
	CPUData<float> czbuffer( _zbuffer.extent(0), _zbuffer.extent(1), _zbuffer.extent(2) );
	std::copy( _zbuffer.data(), _zbuffer.data() + _zbuffer.numElements(), czbuffer.data() );
	return czbuffer;
  }

  static void init_properties()
  {
	typename PBase::SideEffect uninitializer = PBase::make_ivar_setter(&This::_initialized,false);
	typename PBase::SideEffect dirtyfier = PBase::make_ivar_setter(&This::_dirty,true);
	//read-write
	PBase::add_ivar_property_rw(std::string("z0" ), &This::_z0   , uninitializer);
	PBase::add_ivar_property_rw(std::string("alpha0"), &This::_alpha0, uninitializer);
	PBase::add_ivar_property_rw(std::string("alpha_g0"), &This::_alpha0g, uninitializer);
	PBase::add_ivar_property_rw(std::string("mu_m0"), &This::_mu_m0, uninitializer);
	PBase::add_ivar_property_rw(std::string("mu_b0"), &This::_mu_b0, uninitializer);
	PBase::add_ivar_property_rw(std::string("mu_m_prior"), &This::_mu_m_prior, uninitializer);
	PBase::add_ivar_property_rw(std::string("delta"), &This::_delta, dirtyfier);      
	PBase::add_ivar_property_rw(std::string("tau"), &This::_tau);                  
	PBase::add_ivar_property_rw(std::string("tau_p"), &This::_tau_p);              
	PBase::add_ivar_property_rw(std::string("lambda_p"), &This::_lambda_p);        
	PBase::add_ivar_property_rw(std::string("anneal-step"), &This::_anneal_step);  
	PBase::add_ivar_property_rw(std::string("initcount"), &This::_initcount);      
	PBase::add_ivar_property_rw(std::string("renormalize-counts"), &This::_renormalize_counts);
	PBase::add_ivar_property_rw(std::string("segment"), &This::_segment);
	PBase::add_ivar_property_rw(std::string("tau-s"), &This::_s_prior_attenuation);
	PBase::add_ivar_property_rw(std::string("Ds"), &This::_Ds);
	PBase::add_ivar_property_rw(std::string("s0"), &This::_s0);
	PBase::add_ivar_property_rw(std::string("detection_sigmoid"), &This::_use_detection_sigmoid);
	PBase::add_ivar_property_rw(std::string("optimize_detection_sigmoid_gain"), &This::_optimize_detection_sigmoid_gain);
	PBase::add_ivar_property_rw(std::string("online_update"), &This::_online_update);
	PBase::add_ivar_property_rw(std::string("have_bits"), &This::_have_bits);
	PBase::add_ivar_property_rw(std::string("optimize_separability"), &This::_optimize_separability );
	PBase::add_ivar_property_rw(std::string("diagonal_output_jacobian"), &This::_diagonal_output_jacobian );
	PBase::add_ivar_property_rw(std::string("dependent_gates"), &This::_dependent_gates);
	PBase::add_ivar_property_rw(std::string("delta_to_random"), &This::_delta_to_random );
	PBase::add_ivar_property_rw(std::string("report_gradient"), &This::_report_gradient );
	PBase::add_ivar_property_rw(std::string("report_convergence"), &This::_report_convergence );
	PBase::add_ivar_property_rw(std::string("sequential_update"), &This::_sequential_update );
	PBase::add_ivar_property_rw(std::string("init_uniform"), &This::_init_uniform );
	PBase::add_ivar_property_rw(std::string("gate_W_neighbor"), &This::_gate_W_neighbor );
	PBase::add_ivar_property_rw(std::string("gate_clustering_alpha0"), &This::_gate_clustering_alpha0 );
	PBase::add_ivar_property_rw(std::string("s_mrf_iters"), &This::_s_mrf_iters );
	PBase::add_ivar_property_rw(std::string("continuous_sampling"), &This::_continuous_sampling );
	PBase::add_ivar_property_rw(std::string("gamma_mlr"), &This::_gamma_mlr );
	PBase::add_ivar_property_rw(std::string("lambda_ng"), &This::_lambda_ng );
	PBase::add_ivar_property_rw(std::string("cache_z"), &This::_cache_z );
	PBase::add_ivar_property_rw(std::string("gradient_descent_rate"), &This::_gradient_descent_rate);
	PBase::add_ivar_property_rw(std::string("convergence_rate"), &This::_convergence_rate);
	PBase::add_ivar_property_rw(std::string("max_gradient_descent_rate"), &This::_max_gradient_descent_rate);
	PBase::add_ivar_property_rw(std::string("gradient_descent_momentum"), &This::_gradient_descent_momentum);
	PBase::add_ivar_property_rw(std::string("do-vb"), &This::_do_vb);
	PBase::add_ivar_property_rw(std::string("background-model"), &This::_have_background_model);
	PBase::add_ivar_property_rw(std::string("learn-background"), &This::_learn_background_model);
	PBase::add_ivar_property_rw(std::string("cluster_gates"), &This::_cluster_gates );
	PBase::add_ivar_property_rw(std::string("min_rf_saliency"), &This::_min_rf_saliency);
	PBase::add_ivar_property_rw(std::string("sparsity_annealing"), &This::_sparsity_annealing);
	PBase::add_ivar_property_rw(std::string("sparsity_annealing_time"), &This::_sparsity_annealing_time);
	PBase::add_ivar_property_rw(std::string("k"), &This::_k, uninitializer );
	PBase::add_property(std::string("gates"), typename PBase::MemFunGetter( &This::get_gates ) );
	PBase::add_property(std::string("separability"), typename PBase::MemFunGetter( &This::get_separability ) );
	PBase::add_property(std::string("gate_coverage"), typename PBase::MemFunGetter( &This::get_gate_coverage ) );
	PBase::add_property(std::string("bits"), typename PBase::MemFunGetter( &This::get_bits ) );
	PBase::add_property(std::string("mask_coverage"), typename PBase::MemFunGetter( &This::get_mask_coverage ) );
	PBase::add_property(std::string("nlog_component_likelihood"), typename PBase::MemFunGetter( &This::get_nlog_component_likelihoods ) );
	PBase::add_property(std::string("mu_mask"), typename PBase::MemFunGetter( &This::get_mu_mask ) );
	PBase::add_property(std::string("sbuffer"), typename PBase::MemFunGetter( &This::get_sbuffer ) );
	PBase::add_property(std::string("zbuffer"), typename PBase::MemFunGetter( &This::get_zbuffer ) );
	PBase::add_property( "pi_init"  , make_blitz_ivar_getter1( &This::_pi_init   ), make_blitz_ivar_setter1( &This::_pi_init   ) );
	PBase::add_property(std::string("d"), 
						typename PBase::MemFunGetter( &This::get_d ), 
						typename PBase::MemFunSetter( &This::set_d, uninitializer ) );
	PBase::add_ivar_property_rw(std::string("d1"), &This::_d1, uninitializer );
	PBase::add_ivar_property_rw(std::string("d2"), &This::_d2, uninitializer );
	PBase::add_ivar_property_rw(std::string("m"), &This::_m, uninitializer );
	PBase::add_property(std::string("step"), 
						typename PBase::MemFunGetter( &This::get_step ), 
						typename PBase::MemFunSetter( &This::set_step, uninitializer ) );
	PBase::add_ivar_property_rw(std::string("step1"), &This::_step1 );
	PBase::add_ivar_property_rw(std::string("step2"), &This::_step2 );
	PBase::add_ivar_property_rw(std::string("discriminative_generative_coupling"), &This::_discriminative_generative_coupling );

	enum_property_access<SearisePropertyTypes, This, GradientDescentStrategy> 
	  gradient_descent_strategy_accessor( &This::_gradient_descent_strategy );
	gradient_descent_strategy_accessor
	  .add_choice( "plain"                     , PLAIN                     )
	  .add_choice( "levenberg_marquardt"       , LEVENBERG_MARQUARDT       )
	  .add_choice( "natural_gradient"          , NATURAL_GRADIENT          )
	  .add_choice( "adaptive_natural_gradient" , ADAPTIVE_NATURAL_GRADIENT )
	  ;
	PBase::add_property( "gradient_descent_strategy", 
						 gradient_descent_strategy_accessor, 
						 gradient_descent_strategy_accessor );

	enum_property_access<SearisePropertyTypes, This, Neighborhood> 
	  gate_neighborhood_accessor( &This::_gate_neighborhood );
	gate_neighborhood_accessor
	  .add_choice( "none" ,    NO_NEIGHBORS )
	  .add_choice( "four" ,  FOUR_NEIGHBORS )
	  .add_choice( "eight", EIGHT_NEIGHBORS )
	  ;
	PBase::add_property( "gate_neighbors", gate_neighborhood_accessor, gate_neighborhood_accessor );

	enum_property_access<SearisePropertyTypes, This, BasePrior> 
	  base_prior_accessor( &This::_base_prior );
	base_prior_accessor
	  .add_choice( "none"     , NONE     )
	  .add_choice( "fsd"      , FSD      )
	  .add_choice( "tsb"      , TSB      )
	  .add_choice( "fixed_bin", FB       )
	  ;
	PBase::add_property( "base_prior", base_prior_accessor, base_prior_accessor );
	
	enum_property_access<SearisePropertyTypes, This, GradientMode> 
	  gradient_mode_accessor( &This::_gradient_mode );
	gradient_mode_accessor
	  .add_choice( "canonical", CANONICAL )
	  .add_choice( "backprop", BACKPROP )
	  .add_choice( "likelihood" , LIKELIHOOD  )
	  .add_choice( "unlikelihood" , UNLIKELIHOOD  )
    ;
	PBase::add_property( "gradient_mode", gradient_mode_accessor, gradient_mode_accessor );
	
	PBase::add_property( std::string("reset_discriminative_parameters"), 
						 typename PBase::ConstGetter(false),
						 typename PBase::MemFunSetter( &This::reset_discriminative_parameters ) );

	//read-only
	PBase::add_property( "S"        , make_blitz_ivar_getter1( &This::_S     ) );
	PBase::add_property( "p_bg"     , make_blitz_ivar_getter1( &This::_p_foreground     ) );
	//PBase::add_property( "discriminative_emission_hessian", make_blitz_ivar_getter1( &This::_discriminative_emission_hessian ) );
	PBase::add_property( "pi"       , make_blitz_ivar_getter1( &This::_pi    ) );
	PBase::add_property( "pid"      , make_blitz_ivar_getter1( &This::_pi_discriminative ) );
	PBase::add_property( "alpha"    , make_blitz_ivar_getter1( &This::_alpha ) );
	PBase::add_property( "wk"       , make_blitz_ivar_getter1( &This::_wk    ) );
	PBase::add_property( "wnx"      , make_blitz_ivar_getter1( &This::_wnx   ) );
	PBase::add_property( "w_ll"     , make_blitz_ivar_getter1( &This::_w_ll    ) );
	PBase::add_property( "w_ll_x"   , make_blitz_ivar_getter1( &This::_w_ll_x  ) );
	PBase::add_property( "w_ll_x2"   , make_blitz_ivar_getter1( &This::_w_ll_x2  ) );
	PBase::add_property( "w_ll_lx"  , make_blitz_ivar_getter1( &This::_w_ll_lx ) );
	PBase::add_ivar_property_ro( "w_lL"     , &This::_w_lL    );
	PBase::add_ivar_property_ro( "w_lL_x"   , &This::_w_lL_x  );
	PBase::add_ivar_property_ro( "w_lL_lx"  , &This::_w_lL_lx );
	PBase::add_ivar_property_ro( "w_lL_x2"  , &This::_w_lL_x2 );
	PBase::add_ivar_property_ro( std::string("L") , &This::_L  );
	PBase::add_ivar_property_ro( std::string("Lz"), &This::_Lz );
	PBase::add_ivar_property_ro( std::string("Lp"), &This::_Lp );
	PBase::add_ivar_property_rw( std::string("initialized"), &This::_initialized);
	PBase::add_ivar_property_rw( std::string("dirty"), &This::_dirty);
	PBase::add_ivar_property_ro(std::string("detection_sigmoid_threshold"), &This::_L_threshold );
	PBase::add_ivar_property_ro(std::string("detection_sigmoid_gain"), &This::_L_beta );
	PBase::add_ivar_property_ro(std::string("discriminative_detection_sigmoid_threshold"), &This::_L_threshold_discriminative );
	PBase::add_ivar_property_ro(std::string("discriminative_detection_sigmoid_gain"), &This::_L_beta_discriminative );
  }
  
  void reset_discriminative_parameters( PropertyType )
  {
	BOOST_FOREACH( Matrix4 & parameters, _discriminative_emission_parameters )
	  parameters.free();
	_likelihood_stats_emwa.reset();
	_l_betas_discriminative.free();
	_l_thresholds_discriminative.free();
  }

  void reset()
  {
  }
  
  void flush()
  {
    _alpha.free();
    _pi.free();
    _initialized = false;
  }
  
  void accumulate_hessian( const Matrix3 & data );
  
  void add_output_hessian( const Matrix3 & delta, GradientMode kind );
  
  void add_output_gradient( const Matrix3 & delta, GradientMode kind );
  
  void calculate_likelihood_detector_gradient( const Matrix3 & data );
  
  void calculate_likelihood_detector_hessian( const Matrix3 & data );

  void perform_gradient_descent( const Matrix3 & data );

  void apply_detection_sigmoid_threshold_gradient( T threshold_gradient, T rate );

  void apply_detection_sigmoid_beta_gradient( T beta_gradient, T rate );

  void apply_prior_gradient( Matrix1 & prior_gradient, T rate );

  // stubs

  void report_relative_gradient_magnitudes()
  {
  }

  // this is used to compare against the gradient to estimate the "learning speed"
  T canonical_parameter_scale( int stats_index )
  {
	// the default simply takes the mean of the parameter itself
	return mean( _discriminative_emission_parameters.at(stats_index) );
  }

  void apply_emission_gradient( emission_gradient_t & parameter_gradient, T rate )
  {
  }
								

  void calculate_emission_gradient( const Matrix3 & x
									, const Matrix1 & output_gradient
									, emission_gradient_t & parameter_gradient )
  {
  } 

  void calculate_emission_hessian( const Matrix3 & x
								   , const Matrix1 & output_hessian
								   , emission_gradient_t & parameter_hessian )
  {
  } 

  // this is also abused for hessian backpropagation to reuse all the higher level infrastructure
  // which is identical to that for error bp
  void backpropagate_error_complete( Matrix3 & data )
  {
	BLITZ_INDICES;

	if ( _requested_parameter_mode != DISCRIMINATIVE ) return;

	DISPATCH(update)();

	OPENMP(barrier);
	OPENMP(master)
	{
	  _backpropagated_error.resize( data.shape() );
	  _backpropagated_error = 0;
	}
	OPENMP(barrier);

	Matrix3 r( message_component<T>::get_belief( ), true );
	Matrix3 out_error( _backpropagated_error, true );

	if ( _estimating_hessian )
	  {
		if ( ( _gradient_mode == LIKELIHOOD || _gradient_mode == UNLIKELIHOOD ) && _output_gradient.numElements() == 0 )
		  calculate_likelihood_detector_hessian( data );
		if ( r.extent(0) != _output_hessian.extent(0) || r.extent(1) != _output_hessian.extent(1) || r.extent(2) != _output_hessian.extent(2) )
		  throw std::runtime_error( str( boost::format( "mm2b::backpropagate error shape %i,%i,%i does not match belief shape %i,%i,%i" )
										 % _output_hessian.extent(0) % _output_hessian.extent(1) % _output_hessian.extent(2) % r.extent(0) % r.extent(1) % r.extent(2) ) );
		Matrix3 output_hessian( _current_output_hessian, true );
		for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2,true,true);!it.atend(); ++it)
		  {
			Matrix3 out_error_rf( out_error( it.in_range1(), it.in_range2(), _all ) );
			DISPATCH(backpropagate_hessian)( it.in()
											 , output_hessian( it.out_x(), it.out_y(), _all )
											 , out_error_rf );
		  }
	  }
	else
	  {
		if ( ( _gradient_mode == LIKELIHOOD || _gradient_mode == UNLIKELIHOOD ) && _output_gradient.numElements() == 0 )
		  calculate_likelihood_detector_gradient( data );
		if ( r.extent(0) != _output_gradient.extent(0) || r.extent(1) != _output_gradient.extent(1) || r.extent(2) != _output_gradient.extent(2) )
		  throw std::runtime_error( str( boost::format( "mm2b::backpropagate error shape %i,%i,%i does not match belief shape %i,%i,%i" )
										 % _output_gradient.extent(0) % _output_gradient.extent(1) % _output_gradient.extent(2) % r.extent(0) % r.extent(1) % r.extent(2) ) );
		Matrix3 output_gradient( _output_gradient, true );
		for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2,true,true);!it.atend(); ++it)
		  {
			Matrix3 out_error_rf( out_error( it.in_range1(), it.in_range2(), _all ) );
			DISPATCH(backpropagate_error)( it.in()
										   , output_gradient( it.out_x(), it.out_y(), _all )
										   , out_error_rf );
		  }
	  }


  }

  void backpropagate_error( Matrix3 & x
							, const Matrix1 & output_gradient
							, Matrix3 & in_error )
  {
  } 
  
  void backpropagate_hessian( Matrix3 & x
							  , const Matrix1 & output_hessian
							  , Matrix3 & in_hessian )
  {
  }

  void request_parameter_mode( ParameterMode mode )
  {
	if ( _learning )
	  throw std::runtime_error( "cannot change parameters while learning" );
	if ( mode == NO_PARAMETERS || mode == VARIATIONAL )
	  throw std::runtime_error( "can only select generative or discriminativ parameters directly" );
	_requested_parameter_mode = mode;
  }

  void begin_discriminative_hessian_estimation()
  {
	if ( _learning ) throw std::runtime_error( "cannot learn and estimate hessian simultaenously" );
	_estimating_hessian = true;
	_hessian_estimation_counter = 0;
	_requested_parameter_mode = DISCRIMINATIVE;
	BOOST_FOREACH( Matrix4d & hessian, _discriminative_emission_hessian )
	  {
		hessian.resize(_k,_d1,_d2,_m);
		hessian = 0;
	  }
	_output_hessian.resize(_k);
	_output_hessian = 0;
	_L_threshold_hessian = 0;
	_L_beta_hessian = 0;
  }

  void end_discriminative_hessian_estimation()
  {
	_estimating_hessian = false;
	BOOST_FOREACH( Matrix4d & hessian, _discriminative_emission_hessian )
	  {
		hessian /= _hessian_estimation_counter;
		std::cout << " hessian " << min(hessian) << ", " << mean(hessian) << ", " << max(hessian) << std::endl;
	  }
	_output_hessian /= _hessian_estimation_counter;
	_L_threshold_hessian /= _hessian_estimation_counter;
	_L_beta_hessian /= _hessian_estimation_counter;
  }

  void begin_discriminative_gradient_descent()
  {	
	_learning = true;
	_requested_parameter_mode = DISCRIMINATIVE;
	check_sigmoid_params();
  }

  void end_discriminative_gradient_descent()
  {
	_learning = false;
	++_gd_count;
  }

  void batchlearn_begin_epoch()
  {
	DISPATCH(update());
	if ( !_online_update || _wk.numElements() == 0 )
	  {
		std::cout << " beginning epoch " << std::endl;
		_wk.resize(_k);
		_nk.resize(_k);
		_wnx.resize(_k,_d1,_d2);
		_wk = 0;
		_nk = 0;
		_wnx = 0;
		BOOST_FOREACH( Matrix4d & stats, _emission_statistics )
		  {
			stats.resize( _k, _d1, _d2, _m );
			stats = 0;
		  }
		if ( _have_background_model && _learn_background_model )
		  {
			BOOST_FOREACH( Matrix1d & stats, _background_statistics )
			  {
				stats.resize( _m );
				stats = 0;
			  }
		  }
		DISPATCH(prepare_statistics());
	  }
	_Lz = 0;
	_Lp = 0;
	_learning = true;
	_requested_parameter_mode = _do_vb ? VARIATIONAL : GENERATIVE;
  }
  
  void set_S( const Matrix2 & S )
  {
	_S.reference(S);
  }
  
  Matrix2 s_prior()
  {
	BLITZ_INDICES;
    Matrix2 prior(_d1,_d2);
    T mid1 = T(_d1-1)/2;
    T mid2 = T(_d2-1)/2;
	T sigma1 = _s_prior_attenuation * mid1;
	T sigma2 = _s_prior_attenuation * mid2;
    prior = T(-0.5) * ( pow2( i1 - mid1 ) + pow2( i2 - mid2 ) )
      / ( sigma1 * sigma2 ) 
      - std::log( T(2.0 * 3.14) * std::sqrt( sigma1 * sigma2 ) );
	return prior;
  }
  
  void begin_mp_frame() 
  {
	// make sure we are up to data
    DISPATCH(update());

	// make sure the prior message component is connected to the buffer
    _base_prior_message.set_buffer( message_component<T>::buffer(), "base_prior" );

	// reset the prior message
	Matrix3 msg( 1, 1, _k );
	msg( 0, 0, blitz::Range::all() ) = _pi;
	_base_prior_message.set_msg( msg, mmm2_belief_state<T>::LINEAR );   

    // if online learning make the neccessary decays/preparations
    if ( _learning && _online_update )
      {
		DISPATCH( decay_statistics() );
		DISPATCH( prepare_statistics() );
      }

    //free buffers
    _S.free();
    _p_foreground.free();
    message_component<T>::get_msg().clear();    
	BOOST_FOREACH( input_likelihood_t & input_likelihood, _input_likelihoods )
	  input_likelihood.clear();
	_output_gradient.free();
	_likelihood_gradient.free();
	_current_output_hessian.free();
  }
  
  void set_input_likelihood( Matrix2 l, size_t channel = 0, int d1 = 1, int s1 = 1, int d2 = 0, int s2 = 0 )
  {
	_input_likelihoods.at(channel).set( l, d1, s1, d2, s2 );
  }

  void end_mp_frame( const Matrix3 & data ) 
  {
	if  ( data.extent(0) < _d1 ||  data.extent(1) < _d2 ) return;
    if ( _learning )
	  {
		if ( _requested_parameter_mode != DISCRIMINATIVE )
		  {			
			DISPATCH( apply_r( data ) );
			if ( _online_update )
			  {
				OPENMP(barrier);
				OPENMP(single)
				{
				  finalize_statistics();
				  DISPATCH( recalc_hyperparams() );
				  recalc_hyperparams();
				}
				DISPATCH(update(true));
			  }
		  }
		else
		  {
			perform_gradient_descent( data );
		  }
	  }
	else if ( _requested_parameter_mode == DISCRIMINATIVE && _estimating_hessian )
	  accumulate_hessian( data );
	else if ( _learning_likelihood )
	  accumulate_likelihoods( data );
  }

  Matrix3 calc_max_component_likelihoods()
  {
	throw std::runtime_error("calc_max_component_likelihoods not implemented");
  }

  Matrix3 calc_component_entropies()
  {
	throw std::runtime_error("calc_max_component_likelihoods not implemented");
  }

  void accumulate_likelihoods( const Matrix3 & data )
  {
	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;
	Matrix3 r( message_component<T>::get_belief(), true );
	T n = r.extent(0) * r.extent(1);
	Matrix2 W, Sw( _S, true );
	bool have_saliency = _S.numElements() > 0;
	T L;
	Matrix3 lmax = DISPATCH(calc_max_component_likelihoods());
	T Lmax = max( sum( sum( lmax( i2, i3, i1 ), i3 ), i2 ) + safe_log( _pi ) );
	for( ForwardRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
		 !iter.atend_j(); iter.increment_j() )
	  {
		Matrix1 rj( r( iter.j1, iter.j2, blitz::Range::all() ), true );
		Matrix3 X( data( iter.in_range1(), iter.in_range2(), blitz::Range::all() ) );
		if ( have_saliency ) { W.reference( Sw( iter.in_range1(), iter.in_range2() ) ); combined_local_likelihood(X,rj,iter.j1, iter.j2, L, W); }
		else                                                                            combined_local_likelihood(X,rj,iter.j1, iter.j2, L   );
		T w = ( have_saliency ? mean( W ) : 1 ) / n;
		L = Lmax - L;
		_w_lL += w;
		_w_lL_x += w * L;
		_w_lL_x2 += w * L * L;
		_w_lL_lx += w * safe_log( L );
		for( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
		  {
			Matrix1 x( data( iter.i1, iter.i2, blitz::Range::all() ) );
			T s = ( have_saliency ? _S( iter.i1, iter.i2 ) : 1 ) / n;
			int i1rf = iter.i1_rf();
			int i2rf = iter.i2_rf();
			for ( int k = 0; k < _k; ++k )
			  {
				T l_max = lmax( i1rf, i2rf, k );
				T l_model = delegate::evaluate_log_upward_at( derived(), x.data(), i1rf, i2rf, k );
				T l = std::max( T(0), l_max - l_model );
				_w_ll   ( i1rf, i2rf, k ) += s                 * rj( k );
				_w_ll_x ( i1rf, i2rf, k ) += s * l             * rj( k );
				_w_ll_x2( i1rf, i2rf, k ) += s * l * l         * rj( k );
				_w_ll_lx( i1rf, i2rf, k ) += s * safe_log( l ) * rj( k );
			  }
		  }
	  }
  }
  
  void decay_statistics()
  {
	T c =  std::exp( - 1.0 / _tau );

	_wk  *= c;
	_wnx *= c;
	_mu_mask = ( c * ( _mu_mask * ( _wg + _alpha0g ) - _mu_m0 * _alpha0g ) + _mu_m0 * _alpha0g ) / ( c * _wg + _alpha0g );
	_wg  *= c;
	
	BOOST_FOREACH( Matrix4d & stats, _emission_statistics )		
	  stats *= c;

	BOOST_FOREACH( Matrix1d & stats, _background_statistics )		
	  stats *= c;

  }
  
  void mean_field_update_gate_messages( InverseRFIterator & iter )
  {

    BLITZ_INDICES;
  // initialize with the prior and the local factor bias
    OPENMP(barrier);
    OPENMP(master)
    {
      _sbuffer.resize( _zbuffer.shape() );
      _lzbias .resize( _zbuffer.shape() );
    }
    OPENMP(barrier);
    bool have_interactions = _gate_neighborhood != NO_NEIGHBORS || _cluster_gates;
    size_t nlines = _sbuffer.extent(0);
    Matrix3 sbufferw( _sbuffer, true );
    Matrix3 bbufferw( _bbuffer, true );
    Matrix3 zbufferw( _zbuffer, true );
    Matrix3 lzbiasw( _lzbias, true );
    Matrix3 sbuffer_local( sbufferw( WORK_RANGE( nlines ), _all, _all ) );
    Matrix3 zbuffer_local( zbufferw( WORK_RANGE( nlines ), _all, _all ) );
    Matrix3 bbuffer_local;
    Matrix3 lzbias_local ( lzbiasw ( WORK_RANGE( nlines ), _all, _all ) );
    if ( _have_bits )
      bbuffer_local.reference( bbufferw( WORK_RANGE( nlines ), _all ) );
  // gaussian spatial prior over gates
    for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
    {
      for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
        sbufferw( iter.i1, iter.i2, iter.j_index() ) = -_s_prior_attenuation * iter.r2_rf();
    }
    if ( have_interactions )
    {
      lzbias_local = safe_log( zbuffer_local );
      if ( _have_bits )
        lzbias_local += safe_log( bbuffer_local );
      sbuffer_local += lzbias_local;
    }
    normalize_rho( sbuffer_local ); 
    if ( !have_interactions )
      return;
  // iterative mean field update of gate distributions, alternates between odd and even rows to speed up mixing
    T lW_neighbor = safe_log( _gate_W_neighbor );
    OPENMP(single)
      _s_converged = false;
    for ( int iteration = 0; iteration < _s_mrf_iters; ++iteration )
    {
      Matrix3 s_old;
      OPENMP(master)
        s_old.reference( _sbuffer.copy() );
      Matrix2 clustering_message;
      if ( _cluster_gates )
      {
        // FIXME: not threadsafe
        // collect dirichlet counts for each output
        blitz::TinyVector<int,2> s2;
        boost::tie(s2[0],s2[1]) = calc_outsize(sbufferw.extent(0),sbufferw.extent(1));
        clustering_message.resize( s2 );
        clustering_message = _gate_clustering_alpha0;
        for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
          for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
            clustering_message( iter.j1, iter.j2 ) += sbufferw( iter.i1, iter.i2, iter.j_index() );
        // use this as multinomial clustering prediction
        clustering_message = safe_log( clustering_message );
      }
      BOOST_FOREACH( int i, random_permutation( iter.size_i() ) )
      {
        iter.move_to( i );
        sbufferw( iter.i1, iter.i2, _all ) = -std::numeric_limits<T>::max(); // "zero"-initialize unused slots
      // collect messages to this gate
        for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
        {
          int j_index = iter.j_index();
          T & lmsg = sbufferw( iter.i1, iter.i2, j_index );
        // bias via the local factor and the prior
          lmsg = lzbiasw( iter.i1, iter.i2, j_index ) - _s_prior_attenuation * iter.r2_rf();
          if ( _cluster_gates )
            lmsg += clustering_message( iter.j1, iter.j2 );
          if ( _gate_neighborhood != NO_NEIGHBORS )
          {
            bool left   = iter.i1_rf() > 0;
            bool top    = iter.i2_rf() > 0;
            bool right  = iter.i1_rf() < ( _d1 - 1 );
            bool bottom = iter.i2_rf() < ( _d2 - 1 );
          // from left
            if ( left                ) lmsg +=  lW_neighbor * _sbuffer( iter.i1 - 1, iter.i2    , j_index );
            if ( right               ) lmsg +=  lW_neighbor * _sbuffer( iter.i1 + 1, iter.i2    , j_index );
            if ( top                 ) lmsg +=  lW_neighbor * _sbuffer( iter.i1    , iter.i2 - 1, j_index );
            if ( bottom              ) lmsg +=  lW_neighbor * _sbuffer( iter.i1    , iter.i2 + 1, j_index );
            if ( _gate_neighborhood == EIGHT_NEIGHBORS )
            {
              if ( left  && top    ) lmsg +=  lW_neighbor * _sbuffer( iter.i1 - 1, iter.i2 - 1, j_index );
              if ( left  && bottom ) lmsg +=  lW_neighbor * _sbuffer( iter.i1 - 1, iter.i2 + 1, j_index );
              if ( right && top    ) lmsg +=  lW_neighbor * _sbuffer( iter.i1 + 1, iter.i2 - 1, j_index );
              if ( right && bottom ) lmsg +=  lW_neighbor * _sbuffer( iter.i1 + 1, iter.i2 + 1, j_index );
            }
          }
        }        
        normalize_rho( sbufferw( iter.i1, iter.i2, _all ) );
      }
      OPENMP(barrier);
      OPENMP(master)
      {
        _s_converged = mean( abs( sbufferw - s_old ) ) < _convergence_rate;
        if ( _report_convergence )
        {			
          std::cout << " s change " << mean( abs( _sbuffer - s_old ) ) << std::endl;
        }
      }
      if ( _s_converged ) break;
    }
    if ( !_s_converged ) throw std::runtime_error("gate mrf did not converge");
  // divide out the local factor to get the leave-one-out-message to it
    sbuffer_local = safe_log( sbuffer_local ) - lzbias_local;
    normalize_rho( sbuffer_local );

  }

  Derived & derived() { return *static_cast<Derived*>(this); }

  void calc_downward_msg_gated( Matrix3 & data, bool logarithmic )
  {
	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;
	Matrix3 & r = message_component<T>::get_belief();
	OPENMP(barrier);
    OPENMP(master)
	{
	  _downward_message.resize( data.shape() );
	  OPENMP(critical)
		if ( _sampled )
		  {
			IMatrix2 samplew( message_component<T>::buffer().get_sample(), true );
			delegate::build_downward_message_sampled( *this, samplew, _downward_message );
		  }
		else
		  delegate::build_downward_message( *this, data, r, _downward_message );
	  if ( logarithmic )
		{
#ifdef HAVE_SSE
		  sse_log( _downward_message.data(), _downward_message.data(), _downward_message.numElements() );
#else
		  _downward_message = safe_log( _downward_message );
#endif
		}
	}
	OPENMP(barrier);
  }

  bool init_sampling( Matrix3 & data, bool cache_parent_likelihoods )
  {
	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;
	bool start = message_component<T>::buffer().get_sample().numElements() == 0 || !_continuous_sampling;
	if ( start )
	  {
		Matrix3 & lr0 = message_component<T>::get_belief( true, mmm2_belief_state<T>::LOG );
		OPENMP(barrier);
		OPENMP(master)
		{
		  _lr0.reference( lr0 );
		  blitz::TinyVector<int,2> s2;
		  boost::tie(s2[0],s2[1]) = calc_outsize(data.extent(0),data.extent(1));
		  IMatrix2 sample( s2 );
		  message_component<T>::buffer().set_sample( sample );
		  _gates_sample.resize( data.extent(0), data.extent(1) );
		  _gates_sample = 0;
		  int max_n_parents = ( ( _d1 + _step1 - 1 ) / _step1 ) * ( ( _d2 + _step2 - 1 ) / _step2 );

		  if ( cache_parent_likelihoods ) _zbuffer.resize( data.extent(0), data.extent(1), max_n_parents );
		  else _zbuffer.free();

		  if ( _have_bits )
			_bits_sample.resize( s2 );
		  else
			_bits_sample.free();
		  
		  if ( _have_background_model && ( _have_bits || _dependent_gates ) )
			_background_likelihoods.resize( data.extent(0), data.extent(1) );
		  else
			_background_likelihoods.free();
		}
		OPENMP(barrier);

		InverseRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
		ForwardRFIterator forward_iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
		
		// this writes directly into the sample of the message buffer
		IMatrix2 latent_samplew( message_component<T>::buffer().get_sample(), true );
		blitz::Range local_range = WORK_RANGE( latent_samplew.extent(0) );
		Matrix3 dataw( data, true );
		Random<T> & rng = Random<T>::get();

		// leave bits on in the beginning
		if ( _have_bits )
		  for ( forward_iter.reset_j(); !forward_iter.atend_j(); forward_iter.increment_j() )
			_bits_sample( forward_iter.j1, forward_iter.j2 ) = true;

		// collect the combined prior (i.e. local prior and externals) and sample latents from it
		Matrix3 lr0w( _lr0, true );
		blitz::Range rows = WORK_RANGE( latent_samplew.extent(0) );
	
		for ( int i = rows.first(); i <= rows.last(); ++i )
		  for ( int j = 0; j < latent_samplew.extent(1); ++j )
			{
			  if ( _lr0.extent(0) > 1 || _lr0.extent(1) > 1 )
				latent_samplew( i, j ) = rng.sample_log_multinomial( _lr0( i, j, _all ) );
			  else
				latent_samplew( i, j ) = rng.sample_log_multinomial( _lr0( 0, 0, _all ) );
			}

		OPENMP(barrier);

		// precalc bg likelihoods if needed
		if ( _background_likelihoods.numElements() > 0 )
		  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
			_background_likelihoods( iter.i1, iter.i2 ) = delegate::background_message( *this, dataw(iter.i1, iter.i2,_all), iter.i1, iter.i2 );

		// sample the gates given the latents without mrf
		delegate::sample_gates( *this, dataw, latent_samplew, iter,
								cache_parent_likelihoods, true /* init */ );

	  }
	return start;
  }

  struct position
  {
	position( int r = 0, int c = 0 ) : row(r), col(c) {}
	int row, col;
  };
  std::vector<position> _rf_sampling_todo;
  std::vector<position> _rf_sampling_workset;
  int _rf_sampling_row_margin, _rf_sampling_col_margin;
  bool _rf_sampling_continues;

  void prepare_rf_sampling( int rows, int cols )
  {
	OPENMP(barrier);
	OPENMP(single)
	{
	  _rf_sampling_workset.resize( OPENMP_NUM_THREADS );
	  _rf_sampling_row_margin = 1 + _d1 / _step1;
	  _rf_sampling_col_margin = 1 + _d2 / _step2;
	  _rf_sampling_todo.resize( rows * cols );
	  typename std::vector<position>::iterator i_item = _rf_sampling_todo.begin();
	  for ( int i = 0; i < rows; ++i )
		for ( int j = 0; j < cols; ++j )
		  {
			*i_item++ = position( i, j );
		  }
	}
  }

  bool next_rf_sampling_position( position & pos )
  {
	OPENMP(barrier);
	OPENMP(single)
	{
	  // somewath stochastic, may yield less than optimal number of threads
	  _rf_sampling_continues = _rf_sampling_todo.size() > 0;
	  int n_picked = 0, retries = 2 * _rf_sampling_workset.size();
	  while ( n_picked < _rf_sampling_workset.size() &&
			  _rf_sampling_todo.size() > 0 &&
			  retries-- > 0 )
		{
		  int propindex = rand() % _rf_sampling_todo.size();
		  position & prop = _rf_sampling_todo[ propindex ];
		  bool ok = true;
		  for ( int i = 0; ok && i < n_picked; ++i )
			{
			  position & other = _rf_sampling_workset[i];
			  if ( std::abs( prop.row - other.row ) < _rf_sampling_row_margin ||
				   std::abs( prop.col - other.col ) < _rf_sampling_col_margin )
				ok = false;
			}
		  if ( ok )
			{
			  _rf_sampling_workset[n_picked++] = prop;
			  // quick pop without preserving order
			  _rf_sampling_todo[propindex] = _rf_sampling_todo.back();
			  _rf_sampling_todo.pop_back();
			}
		}
	  // mark unused work items
	  for ( int i = n_picked; i < _rf_sampling_workset.size(); ++i )
		_rf_sampling_workset[i].row  = -1;
	}
	pos = _rf_sampling_workset[OPENMP_THREAD_NUM];
	return _rf_sampling_continues;
  }

  // receptive field wise sampling
  void sample_receptive_fields( Matrix3 & data, size_t iterations, bool sample_data = false )
  {	
	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;

	if ( _sparsity_annealing )
	  {
		OPENMP(single)
		{
		  _mu_b0_saved = _mu_b0;
		  _mu_b0 = 1;
		}
	  }

	init_sampling( data, true /* cache_parent_likelihoods */ );
	IMatrix2 latent_samplew( message_component<T>::buffer().get_sample(), true );
	Matrix3 dataw( data, true );
	typename delegate::rf_sampling_context context;
	for ( size_t iteration = 0; iteration < iterations; ++iteration )
	  {
		if ( _sparsity_annealing )
		  {
			OPENMP(single)
			{
			  T d = std::min( T(1), T(iteration) / T( _sparsity_annealing_time * std::max(size_t(1),iterations-1) ) );
			  _mu_b0 = 1 - d + d * _mu_b0_saved;
			}
		  }
		// threadsafe
		prepare_rf_sampling( latent_samplew.extent(0), latent_samplew.extent(1) );
		position rf_pos;
		while ( next_rf_sampling_position( rf_pos ) )
		  if ( rf_pos.row >= 0 )
			{
			  delegate::sample_rf( *this, dataw, latent_samplew, rf_pos, context );
			}
	  }
	OPENMP(single)
	{
	  if ( _report_convergence )
		delegate::report( context );
	  _sampled = true;
	  if ( _sparsity_annealing )
		_mu_b0 = _mu_b0_saved;
	}
  }

  void sample_posterior_gated( Matrix3 & data, size_t iterations, bool sample_data = false )
  {
	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;
	
	bool start = init_sampling( data, _cache_z );
	InverseRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
	ForwardRFIterator forward_iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );

	// this writes directly into the sample of the message buffer
	IMatrix2 latent_samplew( message_component<T>::buffer().get_sample(), true );
	blitz::Range local_range = WORK_RANGE( latent_samplew.extent(0) );
	Matrix3 dataw( data, true );
	bool zbuffer_stale = !start;

	if ( sample_data ) delegate::sample_data( *this, dataw, latent_samplew, iter );
	
	stopwatch gates_timer, latents_timer, iteration_timer, data_timer, bits_timer,  masks_timer;

	for ( size_t iteration = 0; iteration < iterations; ++iteration )
	  {
		iteration_timer.tic();

		// 1. sample gates given latents, bits, and masks, iterate to let mrf settle
		int gate_iterations = _gate_neighborhood == NO_NEIGHBORS ? 
		  1 : std::max( 1, _s_mrf_iters );
		for ( int gate_iteration = 0; 
			  gate_iteration < gate_iterations; 
			  ++gate_iteration )
		  {
			gates_timer.tic();
			delegate::sample_gates( *this, dataw, latent_samplew, iter, 
									_cache_z && zbuffer_stale );
			gates_timer.toc();
			zbuffer_stale = false;
			if ( _have_bits )
			  {
				bits_timer.tic();
				delegate::sample_bits( *this, dataw, latent_samplew, forward_iter );
				bits_timer.toc();
			  }
		  }

		OPENMP(barrier);

		latents_timer.tic();

		// 2. sample latents given gates and masks

		delegate::sample_latents( *this, dataw, latent_samplew, _lr0, forward_iter );
		zbuffer_stale = true;

		OPENMP(barrier);
		
		latents_timer.toc();

		OPENMP(barrier);

		// sample data given latents and gates
		if ( sample_data ) 
		  {
			data_timer.tic();
			delegate::sample_data( *this, dataw, latent_samplew, iter );
			data_timer.toc();
		  }

		iteration_timer.toc();

	  }
	/*
	if ( _report_convergence )
	  std::cout << " latent times " << latents_timer << std::endl
				<< " gates times  " << gates_timer << std::endl
				<< " masks times  " << masks_timer << std::endl
				<< " bits times  " << bits_timer << std::endl
				<< " data times   " << data_timer << std::endl
		;
	*/
	OPENMP(single)
	  _sampled = true;

  }

  Matrix2 & calc_foreground_probability( Matrix3 & data )
  {
	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;
	Matrix3 r( message_component<T>::get_belief( false, mmm2_belief_state<T>::LINEAR ), true );
	OPENMP(barrier);
	OPENMP(master)
	{
	  _p_foreground.resize( data.extent(0), data.extent(1) );
	}
	OPENMP(barrier);
	Matrix3 dataw( data, true );
	InverseRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
	for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	  {
		// a pixel is foreground if it is in any of the receptive fields touching it ( probebilistic or )
		T p_bg = T(1);
		Matrix1 x_i( dataw( iter.i1, iter.i2, _all ) );
		for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		  {
			Matrix1 r_j( r( iter.j1, iter.j2, _all ) );
			p_bg *= ( T(1) - delegate::foreground_probability( *this, x_i, r_j, iter.i1, iter.i2, iter.i1_rf(), iter.i2_rf() ) );
		  }
		_p_foreground( iter.i1, iter.i2 ) = T(1) - p_bg;
	  }
	return _p_foreground;
  }

  Matrix2 & calc_foreground_coverage( Matrix3 & data )
  {
	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;
	Matrix3 r( message_component<T>::get_belief( false, mmm2_belief_state<T>::LINEAR ), true );
	OPENMP(barrier);
	OPENMP(master)
	{
	  _n_foreground.resize( r.extent(0), r.extent(1) );
	}
	OPENMP(barrier);
	Matrix3 dataw( data, true );
	ForwardRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
	for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
	  {
		Matrix1 r_j( r( iter.j1, iter.j2, _all ) );
		T p_fg = T(0);
		int n = 0;
		for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
		  {
			Matrix1 x_i( dataw( iter.i1, iter.i2, _all ) );
			p_fg += delegate::foreground_probability( *this, x_i, r_j, iter.i1, iter.i2, iter.i1_rf(), iter.i2_rf() );
			++n;
		  }
		p_fg /= n;
		_n_foreground( iter.j1, iter.j2 ) = p_fg;
	  }
	return _n_foreground;
  }

  // flood update
  // saliency is assumed to be constant across iterations
  // for sequential: update z in-place, recalc r each time
  void update_upward_msg_gated( Matrix3 & data, T rate = 1, bool linear_annealing = false )
  {
	
	//	std::cout << " flood updating upward messages gated " << std::endl;
	
	OPENMP(single)
	{
	  _current_rate = rate;
	  _linear_annealing = linear_annealing;
	}

	BLITZ_INDICES;
	typedef update_gated_msg_delegate<Derived> delegate;
	
	{
	  
	  InverseRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
	  ForwardRFIterator forward_iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );

	  _had_upward = ! message_component<T>::get_msg().empty();
	  
	  OPENMP(barrier);

	  //
	  // !!! do linear annealing in the message buffer on the marginal !!!
	  //

	  Matrix3 old_r;
	  Matrix2 old_bits;
	  if ( _had_upward )
		{
		  // keep marginals copies for convergence monitoring
		  old_r.reference( message_component<T>::get_belief() );
		  old_bits.reference( _bits_marginal.copy() );

		  OPENMP(barrier);
		  OPENMP(master)
		  {
			Matrix3 r( old_r.copy() );
			message_component<T>::get_belief().reference(r); // makes old_r "unique"

		  }
		  OPENMP(barrier);
		}
	  else
		{
		  // first time make some setup
		  blitz::TinyVector<int,3> s;
		  boost::tie(s[0],s[1]) = calc_outsize(data.extent(0),data.extent(1));
		  s[2] = _k;
		  message_component<T>::get_msg().init( s, mmm2_belief_state<T>::LOG );
		  OPENMP(barrier);
		  OPENMP(master)
		  {
			message_component<T>::buffer().touch();
			int max_n_parents = ( ( _d1 + _step1 - 1 ) / _step1 ) * ( ( _d2 + _step2 - 1 ) / _step2 );
			_zbuffer.resize( data.extent(0), data.extent(1), max_n_parents );
			_zbuffer = 0;
			if ( _have_bits )
			  {
				_bbuffer.resize( data.extent(0), data.extent(1), max_n_parents );
				_bbuffer = 0;
				_bits_marginal.resize( s[0], s[1] );
				_bits_marginal = _mu_b0;
			  }
		  }
		  OPENMP(barrier);
		}

	  Matrix3 & msg = message_component<T>::get_msg().data;
	  Matrix3 msgw( msg, true );
	  Matrix3 msg_local( msg( WORK_RANGE( msg.extent(0) ), _all, _all ) );
	  bool do_sequential_update = _had_upward && _sequential_update;

	  if ( do_sequential_update )
		{
		  Matrix3 & r = message_component<T>::get_belief();
		  // ...run a gauss-seidel sweep of parent updates...		  
		  // not threadsafe!
		  size_t n_parents = forward_iter.size_j();
		  BOOST_FOREACH( int j, random_permutation( n_parents ) )
			{
			  forward_iter.move_to( j );
			  delegate::update_parent( *this,  data, r, msg, forward_iter );				  
			}	  
		  OPENMP(barrier);			  

		  // ...let the mrf converge using several gauss-seidel-sweeps...
		  mean_field_update_gate_messages(iter);
		  OPENMP(barrier);			  

		}	  
	  else
		{
		  // construct interaction messages...
		  Matrix3 & r = message_component<T>::get_belief();

		  if ( !_had_upward )
			delegate::init_interactions( *this, data, r, iter );
		  else
			delegate::update_interactions( *this, data, r, iter );

		  OPENMP(barrier);			  
	  
		  // ...relax using mrf...
		  mean_field_update_gate_messages(iter);
		  
		  // ...construct the message to the latent variables, init local region here...
		  msg_local = 0;
		  delegate::build_upward_message_gated( *this, data, msg, rate, iter );	

		  // tell the buffer we changed the message
		  message_component<T>::buffer().touch();

		}
	  
	  OPENMP(single)
		if ( linear_annealing && !do_sequential_update )
		  {
			Matrix3 & r = message_component<T>::get_belief();
			if ( _had_upward )
			  {
				r = ( T(1) - _current_rate ) * old_r + _current_rate * r;
				if ( _have_bits )
				  _bits_marginal = ( T(1) - _current_rate ) * old_bits + _current_rate * _bits_marginal;
			  }
			else // fade to maxent-distrib
			  {
				r = ( T(1) - _current_rate ) * ( T(1) / T(_k) ) + _current_rate * r;
				if ( _have_bits )
				  _bits_marginal = ( T(1) - _current_rate ) * T(0.5) + _current_rate * _bits_marginal;
			  }
		  }

	  // be verbose...
	  if ( _had_upward )
		{
		  OPENMP(barrier);
		  Matrix3 & r = message_component<T>::get_belief();
		  OPENMP(single)
		  {
			old_r = r - old_r;
			old_bits = _bits_marginal - old_bits;
			if ( _report_convergence )
			  {
				std::cout << " <|dr|> " << mean(abs(old_r)) << " max(|dr|) " << max(abs(old_r));
				if ( _have_bits )
				  {
					std::cout << " sparsity " << mean( _bits_marginal ) << " <|db|> " << mean(abs(old_bits)) << " max(|db|) " << max(abs(old_bits));
				  }
				std::cout << std::endl;
			  }
			T convergence_rate =  linear_annealing ? _convergence_rate * rate : _convergence_rate;
			_converged = mean(abs(old_r)) < convergence_rate;
			if ( _have_bits )
			  _converged &= mean(abs(old_bits)) < convergence_rate;
		  }
		}
	  else
		{
		  OPENMP(single)
			_converged = false;
		}
	}
	
	OPENMP(single)
	{
	  _last_rate = _current_rate;
	  _sampled = false;
	}
	
  }

  void end_likelihood_learning()
  {
	_learning_likelihood = false;
	dirty();
  }

  void check_likelihood_stats( bool reset = false )
  {
	if ( reset || _w_ll.numElements() == 0 || _w_ll_x2.numElements() == 0 || _w_ll_lx.numElements() == 0 )
	  {
		BLITZ_INDICES;
		Matrix3 lmax = DISPATCH(calc_max_component_likelihoods());
		Matrix3 h = DISPATCH(calc_component_entropies());
		T Lmax = max( sum( sum( lmax( i2, i3, i1 ), i3 ), i2 ) + safe_log( _pi ) );
		T H = sum( _pi(i1) * h( i2, i3, i1 ) ) - sum( _pi *  safe_log( _pi ) );
		_w_ll.resize( _d1, _d2, _k );
		_w_ll_x.resize( _d1, _d2, _k );
		_w_ll_x2.resize( _d1, _d2, _k );
		_w_ll_lx.resize( _d1, _d2, _k );
		// init with weak prior to regularize, set k == 1, then theta == <x> == <l> == H
		// this is for filtering
		Matrix3 theta = Matrix3( lmax + h );
		_w_ll    = 0.5;
		_w_ll_x  = 0.5 * theta;
		_w_ll_x2 = theta * theta; // <x^2> = var + <x>^2 = k theta^2 + ( k theta )^2 = 2 theta^2
		_w_ll_lx = 0.5 * ( safe_log( theta ) + 0.577216 ); // ln(theta) + euler_gamma
		
		// this is broken in case of likelihood propagation across layers, which is not present but planned...
		// complete likelihood including prior, but not propagated upwards, which is close to impossible to estimate
		// also in case of propagation the origin isn't clear...
		double Theta = Lmax + H;
		std::cout << " Lmax " << Lmax << " H " << H << " Theta: " << Theta << std::endl;
		_w_lL    = 0.5;
		_w_lL    = 0.5 * Theta;
		_w_lL_x2 = Theta * Theta;
		_w_lL_lx = 0.5 * ( safe_log( Theta ) + 0.577216 );
	  }
  }

  void begin_likelihood_learning()
  {	
	BLITZ_INDICES;
	if ( _learning ) throw std::runtime_error( "cannot learn parameters and likelihood distributions simultaenously" );	
	request_parameter_mode( GENERATIVE );
	update();
	check_likelihood_stats(true);
	_learning_likelihood = true;
  }

  bool converged() const { return _converged; }

  template<typename Data>
  void calc_upward_msg( Data & data )
  {
	
	OMP_LOG( " entering upward message update " );
	
	if ( data.dimensions() == 3 && data.depth() != _m )
	  throw std::runtime_error(str(boost::format("mm2b input depth mismatch got %i expected %i")%data.depth()%_m));
	BLITZ_INDICES;
	
	blitz::TinyVector<int,3> s;
	boost::tie(s[0],s[1]) = calc_outsize(data.extent(0),data.extent(1));
	s[2] = _k;
	
	
	OPENMP(single)
	{
	  DISPATCH(update());
	  message_component<T>::get_msg().data.resize(s);
	  message_component<T>::get_msg().domain = mmm2_belief_state<T>::LOG;
	}
	
	Matrix3 msg( message_component<T>::get_msg().data, true );
	bool have_saliency = _S.numElements() > 0;
	Matrix2 wS(_S,true);
	
	typedef update_gated_msg_delegate<Derived> delegate;

	OMP_LOG( " entering kernel " );

	if ( _have_background_model )
	  {
		InverseRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1, _step2, true );
		delegate::build_upward_message_with_saliency( *this, data, msg, iter );
	  }
	else
	  {
		for( ReceptiveFieldIterator<Data,Matrix3> it( data, msg, _d1,_d2, _step1, _step2 ); !it.atend(); ++it )
		  {
			
			Data & X = it.in();
			
			if ( have_saliency )
			  {		
				Matrix2 w( wS( it.in_range1(),it.in_range2() ) );
				DISPATCH(upward_message(X,it.out(),w));		
			  }
			else // "simple" mp
			  {
				DISPATCH(upward_message(X,it.out()));
			  }
		  }	
	  }
	
	OMP_LOG( " leaving upward message update " );
	
	OPENMP(flush);
	CHECKNAN(msg,"mm2base::vmp_up nan in msg\n");
	
  }
  
  template<typename Data>
  void update_upward_msg( Data & data )
  {
    DISPATCH( calc_upward_msg( data ) );
    OPENMP(single)
      message_component<T>::buffer().touch();
  }
  
  Matrix2 calc_local_likelihood( Matrix3 & data, bool marginal = true )
  {
	int n1, n2;
	boost::tie( n1, n2 ) = calc_outsize( data.extent(0), data.extent(1) );
	if ( n1 <= 0 || n2 <= 0 ) return Matrix2();
	Matrix2 p( n1, n2 );
	p = 0;
	Matrix3 & r = message_component<T>::get_belief( );
	
	bool have_saliency = _S.numElements() > 0;
	
	Matrix2 wp(p,true);
	Matrix2 wS(_S,true);
	
	//std::cout << " p " << p.shape() << " data " << data.shape() << " r " << r.shape() << " S " << _S.shape() << " pi "  << _pi.shape() << std::endl;

	for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1, _step2);!it.atend(); ++it)
	  {
		T & p_rf = wp( it.out_x(), it.out_y() );
		if ( have_saliency )
		  {
			Matrix2 w( wS( it.in_range1(),it.in_range2() ) );
			if ( mean(w) < _min_rf_saliency )
			  p_rf = T(-1e10);
			else
			  combined_local_likelihood( it.in(), it.out(), it.out_x(), it.out_y(), p_rf, w, marginal );
		  }
		else // "simple" mp
		  {
			combined_local_likelihood  ( it.in(), it.out(), it.out_x(), it.out_y(), p_rf,    marginal );
		  }		
		p_rf += safe_log( sum( it.out() * _pi ) );
	  }	
	
	CHECKNAN(p,"mm2base::confidence nan in p\n");
	return p;
  }

  Matrix2 calc_self_information( Matrix3 & data )
  {
	
	Matrix2 p( data.extent(0), data.extent(1) );
	p = 0;
	Matrix3 & r = message_component<T>::get_belief( );
	
	bool have_saliency = _S.numElements() > 0;
	
	Matrix2 wp(p,true);
	Matrix2 wS(_S,true);
	
	for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1, _step2);!it.atend(); ++it)
	  {
		Matrix2 p_rf( wp( it.in_range1(), it.in_range2() ) );
		if ( have_saliency )
		  {
			Matrix2 w( wS( it.in_range1(),it.in_range2() ) );
			p_rf += safe_log( w * sum( it.out() * _pi ) + ( T(1) - w ) );
			
			DISPATCH(self_information(it.out(),it.in(),p_rf,w));
		  }
		else // "simple" mp
		  {
			p_rf += safe_log( sum( it.out() * _pi ) );
		    DISPATCH(self_information(it.out(),it.in(),p_rf));
		  }
		
	  }	
	
	CHECKNAN(p,"mm2base::confidence nan in p\n");
	return p;
  }
  
  bool ready() const { return true; }
  
  int downward_message_depth()
  {
	return _m;
  }
  
  void finalize_downward_message( Matrix3 & msg, bool linear = false )
  {
	if ( linear )
	  normalize_rho( msg );
  }
  
  void initialize_downward_message( Matrix3 & msg, bool linear = false )
  {
	if ( linear )
	  msg = 1;
	else 
	  msg = 0;
  }
  
  void calc_downward_msg ( Matrix3 data = Matrix3(), bool logarithmic = true )
  {
	BLITZ_INDICES;

    DISPATCH(update());
	
	bool have_data = data.numElements() > 0 && std::max( _d1, _d2 ) > 1;
    Matrix3 & r = message_component<T>::get_belief( !have_data );
	
	OPENMP(single)
	{
	  _downward_message.resize( _d1+(r.extent(0)-1)*_step1, _d2+(r.extent(1)-1)*_step2, DISPATCH(downward_message_depth()) );
	  DISPATCH( initialize_downward_message( _downward_message, !logarithmic ) );
	}
	OPENMP(flush);
    
	bool have_saliency = _S.numElements() > 0;
	Matrix2 wS(_S,true);
	
    Matrix3 wdata;
	if ( have_data )
	  wdata.reference(data,true);
	
	// TODO: this needs to be parallelized for input regions not output regions    
    
	OMP_LOG("beginning downward message calculation");
	
	if ( have_saliency )
	  {	
		for(ReceptiveFieldIterator< Matrix3, Matrix3 > it(_downward_message,r,_d1,_d2,_step1,_step2, true, true ); !it.atend(); ++it)
		  {
			Matrix2 w( wS( it.in_range1(),it.in_range2() ) );
			if ( have_data )
			  DISPATCH( downward_message ( wdata( it.in_range1(), it.in_range2(), _all ), it.out(), it.in(), w ) );
			else
			  DISPATCH( downward_message ( wdata, it.out(), it.in(), w ) );
		  }
	  }
	else
	  {
		for( ReceptiveFieldIterator< Matrix3, Matrix3 > it( _downward_message, r, _d1,_d2, _step1,_step2, true, true ); !it.atend(); ++it )
		  {
			if ( have_data )
			  DISPATCH( downward_message ( wdata( it.in_range1(), it.in_range2(), _all ), it.out(), it.in() ) );
			else
			  DISPATCH( downward_message ( wdata, it.out(), it.in() ) );
		  }
	  }
	
	OMP_LOG("finished downward message calculation");
	
	DISPATCH( finalize_downward_message( _downward_message, !logarithmic ) );	
	
  	OPENMP(flush);
	
  }
  
  Matrix3 & get_downward_message()
  {
	return _downward_message;
  }
  
  Matrix3 & get_backpropagated_error()
  {
	return _backpropagated_error;
  }
  
#define CHECK_RANGE(var)												\
  if ( !( min(var) >= T(-0.0001) && max(var) <= T(1.0001) ) )			\
	throw std::runtime_error(str(boost::format("%s(%i) variable out of range [%f,%f] : [%f,%f]") \
								 % __FILE__ % __LINE__ % T(-0.0001) % T(1.0001) % min(var) % max(var) ) ); \
  else																	\
    var = max(T(0),min(T(1),var));
  
  
  Matrix2 backtrack( Matrix2 Pout, bool use_saliency = false, bool and_combine = false )
  {
    BLITZ_INDICES;

	use_saliency = use_saliency && _S.numElements() != 0;
	
	if ( std::max( _d1, _d2 ) == 1 && !use_saliency ) return Pout;
	
	Matrix2 wS(_S,true);
    Matrix2 Pin(inshape(Pout.shape()));
    Pin = 0;   
	
	if ( _segment && _sampled )
	  {
		InverseRFIterator iter( Pin.extent(0), Pin.extent(1), _d1, _d2, _step1, _step2, true );
		for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
		  {
			int gate = _gates_sample( iter.i1, iter.i2 );
			iter.select_j_index( gate );
			Pin( iter.i1, iter.i2 ) = Pout( iter.j1, iter.j2 );
		  }
	  }
	else if ( use_saliency )
	  for(ReceptiveFieldIterator<Matrix2,Matrix2> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		{
		  const Matrix2 & w = wS(it.in_range1(),it.in_range2());
		  Matrix2 & pin = it.in();		
#ifdef HAVE_SSE
		  for ( int i = 0; i < _d1; ++i )
			sse_madd( &pin(i,0), _d2, it.out(), &w(i,0) );
#else
		  pin += it.out() * w(i1,i2);
#endif
		}
	else
	  for(ReceptiveFieldIterator<Matrix2,Matrix2> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		it.in() += it.out();
	
    return Pin;
  }
  
  Matrix3 backtrack(Matrix3 Pout, bool use_saliency = false )
  {
    BLITZ_INDICES;
	
	use_saliency = use_saliency && _S.numElements() != 0;

	if ( std::max( _d1, _d2 ) == 1 && !use_saliency ) return Pout;

	Matrix2 wS(_S,true);
    Matrix3 Pin(inshape(Pout.shape()));
    Pin = 0;	
	
	if ( _segment )
	  {
		if ( _sampled )
		  {
			InverseRFIterator iter( Pin.extent(0), Pin.extent(1), _d1, _d2, _step1, _step2, true );
			for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
			  {
				int gate = _gates_sample( iter.i1, iter.i2 );
				iter.select_j_index( gate );
				Pin( iter.i1, iter.i2, _all ) = Pout( iter.j1, iter.j2, _all );
			  }
		  }
		else
		  {
			Matrix3 gates( _zbuffer.shape() );
			gates = safe_log( _zbuffer ) + safe_log( _sbuffer );
			if ( _have_bits )
			  gates += safe_log( _bbuffer );
			normalize_rho( gates );	
			InverseRFIterator iter( Pin.extent(0), Pin.extent(1), _d1, _d2, _step1, _step2, true );
			for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
			  {
				Pin( iter.i1, iter.i2, _all ) = 0;
				for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
				  {
					Pin( iter.i1, iter.i2, _all ) += gates( iter.i1, iter.i2, iter.j_index() ) * Pout( iter.j1, iter.j2, _all );
				  }
			  }
		  }
	  }
	else if ( use_saliency )
	  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		{
		  const Matrix2 & w = wS(it.in_range1(),it.in_range2());
		  Matrix3 pin(Pin( it.in_range1(), it.in_range2(), _all ));
#ifdef HAVE_SSE
		  int K = Pin.extent(2);
		  for ( int i = 0; i < _d1; ++i )
			for ( int j = 0; j < _d2; ++j )
			  sse_madd( &pin(i,j,0), K, w(i,j), it.out().data() );
#else
		  pin += it.out()(i3) * w(i1,i2);
#endif
		}
	else
	  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		it.in() += it.out()(i3);
	
    return Pin;
  }

  void recalc_likelihoods() {}
  
  Matrix3 propagate( const Matrix3 & Pin, bool use_saliency = false, bool and_combine = false  )
  {
    BLITZ_INDICES;
	
	use_saliency = use_saliency && _S.numElements() != 0;
	
	if ( std::max( _d1, _d2 ) == 1 && !use_saliency ) return Pin;
	
	blitz::TinyVector<int,2> sout = outshape(Pin.shape());
	int K = Pin.extent(2);
	Matrix3 Pout( sout[0], sout[1], K );
	Matrix2 wS(_S,true);
	
	if ( use_saliency )
	  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		{
		  const Matrix2 & w = wS(it.in_range1(),it.in_range2());
		  Matrix3 & pin( it.in() );
		  Matrix1 & pout( it.out() );
		  
#ifdef HAVE_SSE
		  pout = 0;
		  for ( int i = 0; i < _d1; ++i )
			for ( int j = 0; j < _d2; ++j )
			  sse_madd( it.out().data(), K, w(i,j), &pin(i,j,0) );
#else
		  pout = sum( sum( pin(i2,i3,i1) * w(i2,i3), i3 ), i2 );
#endif
		}
	else
	  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		it.out() = mean( mean( it.in()(i2,i3,i1), i3 ), i2 );
	
	return Pout;
	
  }
  
  Matrix2 propagate( const Matrix2 & Pin, bool use_saliency = false, bool and_combine = false )
  {
    BLITZ_INDICES;
	
	use_saliency = use_saliency && _S.numElements() != 0;
	
	if ( std::max( _d1, _d2 ) == 1 && !use_saliency ) return Pin;
	
	blitz::TinyVector<int,2> sout = outshape(Pin.shape());
	Matrix2 Pout( sout[0], sout[1] );
	Matrix2 wS(_S,true);
	
	if ( use_saliency )
	  for(ReceptiveFieldIterator<Matrix2,Matrix2> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		{
		  const Matrix2 & w = wS(it.in_range1(),it.in_range2());
		  
		  Matrix2 & pin( it.in() );
		  T & pout( it.out() );
		  
#ifdef HAVE_SSE
		  pout = 0;
		  for ( int i = 0; i < _d1; ++i )
			pout += sse_dp1( &pin(i,0), &w(i,0), _d2 );
#else
		  pout = sum( pin * w );
#endif
		  pout /= max( T(0.000001), sum( w ) );
		}
	else
	  for(ReceptiveFieldIterator<Matrix2,Matrix2> it(Pin,Pout,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		it.out() = mean( it.in() );
	
	return Pout;
	
  }
  
  void recalc_hyperparams( )
  {
	{
	  if ( _wk_prior.numElements() != _wk.numElements() )
		{
		  _wk_prior.resize( _wk.shape() );
		  _wk_prior = 0;
		}
	  if ( _wnx_prior.numElements() != _wnx.numElements() )
		{
		  _wnx_prior.resize( _wnx.shape() );
		  _wnx_prior = 0;
		}	
	  
	  if ( _wk.extent(0) != _k )
		{
		  _wk.resize( _k );
		  _wk = 0;
		  _wk_prior.resize( _k );
		  _wk_prior = 0;		  
		}
	  //std::cout << " mm2b recalc hyperparams alpha " << _alpha.shape() << " wk " << _wk.shape() << " wk_prior " << _wk_prior.shape() << std::endl;
	  
	  _alpha.resize( _k );
	  _alphag.resize( _d1, _d2, _k );
	  if ( _delta_to_random )
		{
		  _alpha  = _alpha0  + _delta * _wk + ( 1 - _delta ) * _wk_prior;
		}
	  else
		{
		  _alpha  = _alpha0  + _delta * _wk;
		}
	  dirty();
	}
  }
  
  Matrix1 batchlearn_end_epoch()
  {
	
	finalize_statistics();
	
    BLITZ_INDICES;
    Matrix1 f;
	
	_nk.reference( _wk.copy() ); // to keep the actual counts before normalization

    T N = sum( _wk );
    Matrix1d p( Matrix1d( _wk / N ) );
    T H = _Lz;
	double Htotal, Hmax;
	if ( message_component<T>::buffer().type() == mmm2_belief_state<T>::BINOMIAL )
	  {
		Hmax = std::log(2.0) * _k;
		Htotal = - sum( p * safe_log( p ) + ( T(1) - p ) * safe_log( T(1) - p ) ) / Hmax;
	  }
	else
	  {
		Hmax = std::log(_k);
		Htotal = - sum( p * safe_log( p ) ) / Hmax;
	  }
	double Hrel = H / ( Hmax *  N );

    if ( _base_prior != NONE && _base_prior != FB && _learning )
	  {
		// this is actually the parameter kl
		_Lz -= ( gammaln( sum(_alpha) ) - gammaln( _k * _alpha0 ) )
		  - ( sum( gammaln(_alpha) ) - _k * gammaln(_alpha0) )
		  + sum( ( _alpha - _alpha0 ) * safe_log( _pi ) );
	  }

    DISPATCH(recalc_likelihoods());
	_likelihood_history.push_back( likelihoods_t( _Lp, _Lmu, _Lz ) );

    if (_renormalize_counts)
	  {
		f.reference(Matrix1(where(_wk > T(1),N/(T(_k)*_wk),T(1))));
		std::cout << " renormalizing f: " << f << std::endl;
#ifdef DEBUG
		std::stringstream ss; ss << _wk;
		CHECKNAN(f,str(boost::format("mm2b::renorm nan in f wk=%s")%ss.str()));
#endif
		_wk = f * _wk;
		_wnx = f(i1) * _wnx;
		_wg = f(i3) * _wg;
		BOOST_FOREACH( Matrix4d & stats, _emission_statistics )
		  stats = stats( i1, i2, i3, i4 ) * f(i1);
	  } 
    else
	  {
		if (any(_wk) < T(1))
		  {
			std::cout << "near dead components: ";
			for(int i = 0; i < _k; ++i)
			  {
				if (_wk(i) < T(1))
				  std::cout << i << "(<n>=" <<_wk(i) << ") ";
			  }
			std::cout << std::endl;
		  }
	  }
	
	recalc_hyperparams();
	
	dirty();
    _learning = false;
		
    std::cout << str( boost::format( "MixtureModel2 ended epoch <n> = %1.2e H = %1.2e <H> = %1.3e Lz/N = %1.3e") 
					   % N % Htotal % Hrel % ( ( _Lz / _delta - H ) / N ) 
					   ) << std::endl;
	
	if ( _delta < 1 && _anneal_step > 0 )
	  {
		T delta_old = _delta;
		_delta = std::min( T(1), _delta + _anneal_step );
		std::cout << str( boost::format( " delta %1.3f -> %1.3f " ) % delta_old % _delta ) << std::endl;
	  }
	
	_requested_parameter_mode = GENERATIVE;
	dirty();

	++_epoch_count;

    return f;
  }
  
  bool is_multinomial() { return false; }			
  
  boost::tuple<int,int> calc_outsize(int in_w, int in_h)
  {
	if ( in_w < _d1 || in_h < _d2 )
	  return boost::tuple<int,int>(0,0);
	return boost::tuple<int,int>(1 + (in_w - _d1) / _step1, 1 + (in_h - _d2) / _step2);
  }
  
  boost::tuple<int,int> calc_insize(int out_w, int out_h)
  {
	if ( out_w == 0 || out_h == 0 ) return boost::tuple<int,int>( 0, 0 );
	return boost::tuple<int,int>( _d1 + ( out_w - 1 ) * _step1,  _d2 + ( out_h - 1 ) * _step2 );
  }
  
  template<int outrank> blitz::TinyVector<int,2> outshape(blitz::TinyVector<int,outrank> inshape)
  {
    blitz::TinyVector<int,2> result;
    boost::tie(result[0],result[1]) = calc_outsize(inshape[0],inshape[1]);
    return result;
  }
  
  typename PBase::PropertyType get_nlog_component_likelihoods() 
  { 
    BLITZ_INDICES;
    CPUData<T> ret;
    
    if ( message_component<T>::get_msg().normalized )
      throw std::runtime_error( "found already normalized message but unnormalized message would be necessary");

    Matrix3 l( message_component<T>::get_msg().data.copy() );
    Matrix3 lmax( DISPATCH( calc_max_component_likelihoods() ) );
    l -= sum( sum( lmax( i3, i4, i5 ), i5 ), i4 );
    
    blitz2cpudata( l, ret );
    
    return ret;
  }

  template<int rank> blitz::TinyVector<int,rank> inshape(blitz::TinyVector<int,rank> outshape)
  {
    blitz::TinyVector<int,rank> result = outshape;
    boost::tie(result[0],result[1]) = calc_insize(outshape[0],outshape[1]);
    return result;
  }
  
  void combined_local_likelihood( blitz::Array<T,3> & x, blitz::Array<T,1> & r, int i1, int i2, T & l_out, bool marginal = true ) 
  {
	DISPATCH( local_likelihood( x, r, l_out, marginal ) );
	int i = 0;
	BOOST_FOREACH( const input_likelihood_t & l, _input_likelihoods )
	  {
		int s1 = l.s1 * _step1;
		int s2 = l.s2 * _step2;
		int d1 = l.d1 + ( _d1 - 1 ) * l.s1;
		int d2 = l.d2 + ( _d2 - 1 ) * l.s2;
		blitz::Range r1( i1 * s1, i1 * s1 + d1 - 1 );
		blitz::Range r2( i2 * s2, i2 * s2 + d2 - 1 );
		if ( l.data.numElements() > 0 )
		  {
			//std::cout << " collecting likelihood " << i << " i: " << r1 << " j: " << r2 << std::endl;
			l_out += sum( l.data( r1, r2 ) );
		  }
		++i;
	  }
  }

  void combined_local_likelihood( blitz::Array<T,3> & x, blitz::Array<T,1> & r, int i1, int i2, T & l_out, blitz::Array<T,2> & w, bool marginal = true ) 
  {
	DISPATCH( local_likelihood( x, r, l_out, w, marginal ) );	
	BOOST_FOREACH( const input_likelihood_t & l, _input_likelihoods )
	  {
		int s1 = l.s1 * _step1;
		int s2 = l.s2 * _step2;
		int d1 = l.d1 + ( _d1 - 1 ) * l.s1;
		int d2 = l.d2 + ( _d2 - 1 ) * l.s2;
		blitz::Range r1( i1 * s1, i1 * s1 + d1 - 1 );
		blitz::Range r2( i2 * s2, i2 * s2 + d2 - 1 );
		if ( l.data.numElements() > 0 )
		  l_out += sum( l.data( r1, r2 ) * w );
	  }
  }

  Matrix2 & get_input_likelihood_gradient( const Matrix3 & data, size_t i )
  {
	input_likelihood_t & input_likelihood = _input_likelihoods[i];
	if ( ( _gradient_mode == LIKELIHOOD || _gradient_mode == UNLIKELIHOOD ) 
		 && input_likelihood.data.numElements() > 0 
		 && input_likelihood.gradient.numElements() == 0 )
	  {
		if ( _output_gradient.numElements() == 0 )
		  calculate_likelihood_detector_gradient( data );
		OPENMP(single)
		{
		  input_likelihood.gradient.resize( input_likelihood.data.shape() );
		  input_likelihood.gradient = 0;
		  int s1 = input_likelihood.s1 * _step1;
		  int s2 = input_likelihood.s2 * _step2;
		  int d1 = input_likelihood.d1 + ( _d1 - 1 ) * input_likelihood.s1;
		  int d2 = input_likelihood.d2 + ( _d2 - 1 ) * input_likelihood.s2;
		  for(ReceptiveFieldIterator<Matrix2,Matrix2> it( input_likelihood.gradient, _likelihood_gradient, d1, d2, s1, s2 );!it.atend(); ++it)
			it.in() += it.out();
		}
	  }
	return input_likelihood.gradient;
  }

  //stubs
#ifdef _WIN32
#define STUB_ERROR throw std::runtime_error(boost::str(boost::format("called unimplemented stub  %s")%typeid(*this).name()));
#else
#define STUB_ERROR throw std::runtime_error(str(boost::format("called unimplemented stub %s on %s")%__func__%typeid(*this).name()));
#endif
  void local_likelihood( blitz::Array<T,3> &, blitz::Array<T,1> &, T &, bool marginal = true ) {STUB_ERROR;}
  void local_likelihood( blitz::Array<T,3> &, blitz::Array<T,1> &, T &, blitz::Array<T,2> &, bool marginal = true ) {STUB_ERROR;}
  // viterbi updates
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> & ) {STUB_ERROR;}
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> &, blitz::Array<T,2> & ) {STUB_ERROR;}
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> &, Matrix3 & ) {STUB_ERROR;}
  void downward_message( const blitz::Array<unsigned int,2> &, int iout, Matrix3 & lrin) {STUB_ERROR;}
  void downward_message( const blitz::Array<unsigned int,2> &, int iout, Matrix3 & lrin, const blitz::Array<T,2> & w) {STUB_ERROR;}
  // baum-welch updates
  void upward_message( const Matrix3 & data, Matrix1 & lrout, bool marginal = false ) {STUB_ERROR;}
  void upward_message( const Matrix3 & data, Matrix1 & lrout, const Matrix2 & w, bool marginal = false ) {STUB_ERROR;}
  void downward_message( const blitz::Array<T,3> &, const blitz::Array<T,1> &, Matrix3 & lrin ) {STUB_ERROR;}
  void downward_message( const blitz::Array<T,3> &,const blitz::Array<T,1> &, Matrix3 &, const blitz::Array<T,2> &) {STUB_ERROR;}
  // (VB-) M-step
  void collect_statistics_gated( Matrix1 & x, Matrix3 & z, Matrix1 & w, T s0, InverseRFIterator & irf ) {STUB_ERROR;}
  void collect_sampled_statistics( Matrix1 & x, int i1rf, int i2rf, int k, T s0 ) {STUB_ERROR;}
  void collect_sampled_background_statistics( Matrix1 & x, T s0 ) {STUB_ERROR;}
  //blitz::Array<T,2> downward_message_components(const blitz::Array<T,1> & , const blitz::Array<unsigned int,2> &, bool) {STUB_ERROR;}
  // integration with a steering markov chain which orchestrates buffered forward-backward message passing
  bool step_backward() { return false; }
  bool step_forward(bool) { return false; }
  Matrix2 learn() { return Matrix2(); }

  // specific hack to get some things from inside of the mcmm2
  Matrix4 get_inner_prediction() { return Matrix4(); }
#undef STUB_ERROR
  
  //protected:
  
  bool initialize(bool force=false)
  {
	
	BLITZ_INDICES;
	
	if ( ( _initialized || !DISPATCH(ready()) ) && !force ) 
	  return false;
	
	OMP_LOG("mm2b initialize");

	OPENMP(single)
	{
	  
	  //std::cout << " initializing " << std::endl;
	  
	  Random<T> & rng = Random<T>::get();
	  
	  _mu_mask.resize(_d1,_d2,_k);  
	  _wg.resize(_d1,_d2,_k);
	  _wg_prior.resize(_d1,_d2,_k);
	  _wk_prior.resize(_k);  _wk_prior = 0;
	  _wnx_prior.resize(_k,_d1,_d2); _wnx_prior = 0;
	  Matrix1d sample( _k );
	  
	  bool have_external = _pi_init.extent(0) == _k;
	  if ( have_external )
		{
		  sample = _pi_init;
		}
	  else
		{
		  switch( _base_prior )
			{
			case TSB:
			case FSD:
			  if ( _init_uniform || _renormalize_counts )
				sample = 1;
			  else
				for( int k = 0; k < _k; ++k )
				  sample(k) = rng.gamma( _alpha0 );
			  if ( sum(sample) < T(1e-10) ) sample = T(1) / T(_k);
			  else sample /= sum( sample );
			  break;
			case NONE:
			  sample = T(1) / T(_k);
			  break;
			case FB:
			  sample = _z0;
			}
		}
	  
	  _wk_prior = _initcount * sample;
	  _wnx_prior = _initcount * sample(i1) + i2 * i3 * 0;
	  _wk.reference( _wk_prior.copy() );
	  _wnx.reference( _wnx_prior.copy() );
	  _wg_prior = _wnx( i3, i1, i2 );
	  _wg = _wnx( i3, i1, i2 );
	  _mu_mask = _mu_m0;

	  recalc_hyperparams();
	  
	  //std::cout << str(boost::format("MixtureModel2 initialized alpha0 = %f initcounts = %f")%_alpha0%_initcount) << std::endl;
	  _gd_count = _epoch_count = 0;
	  _likelihood_history.clear();

	}

	OPENMP(single)
	  _initialized = _dirty = true;
	
	return true;
  }
  
  void break_sequence()
  {
  }
  
  void dirty()
  {
	_dirty = true;
  }
  
  bool update( bool force = false )
  {
	BLITZ_INDICES;
	
	if ( std::max( _d1, _d2 ) <= 0 ) throw std::runtime_error("empty receptive field!");
	if ( (_step1 <= 0) || (_step2 <= 0) ) throw std::runtime_error("receptive field stepping is zero!");
	if ( !( _dirty || force || ( _current_parameter_mode != _requested_parameter_mode ) ) ) return false;
	
	OMP_LOG("mm2b update");

	OPENMP(single)
	{
	  
	  if ( _requested_parameter_mode != DISCRIMINATIVE ||
		   _current_parameter_mode   != DISCRIMINATIVE )
		{
		  _pi.resize(_k);
		  
		  switch ( _base_prior )
			{
			case NONE:
			  _pi = T(1) / T(_k);
			  break;
			case FSD:
			  if ( _requested_parameter_mode  == VARIATIONAL )
				_pi = exp( psi(_alpha) - psi(sum(_alpha)) );
			  else
				_pi = _alpha/sum(_alpha);
			  break;
			case TSB:
			  {
				Matrix1d wx( _alpha - _alpha0 );
				std::vector<int> ranking(_k);
				//compute_order( ranking.begin(), ranking.end(), wx.data() );
				Matrix1d gamma1( 1 + wx );
				Matrix1d gamma2( _k );
				double wx_accum = 0.0;
				for ( int ik = _k - 1; ik >= 0; --ik )
				  {
					//int k = ranking[ik];					
					wx_accum += wx( ik );
					gamma2(ik) = _alpha0 + wx_accum;
				  }
				if ( _requested_parameter_mode  == VARIATIONAL )
				  {
					Matrix1d ElogV1 ( psi( gamma1 ) - psi( gamma1 + gamma2 ) );
					Matrix1d ElogV2 ( psi( gamma2 ) - psi( gamma1 + gamma2 ) );
					double ElogV2_accum  = 0.0;
					for ( int ik = 0; ik < _k; ++ik )
					  {
						//int k = ranking[ik];
						_pi(ik) = std::exp( ElogV2_accum + ElogV1(ik) );
						ElogV2_accum += ElogV2(ik);
					  }	
				  }	  
				else
				  {
					Matrix1d lbeta  ( log( gamma1 ) - log( gamma1 + gamma2 ) );
					Matrix1d nolbeta( log( gamma2 ) - log( gamma1 + gamma2 ) );
					double nolbeta_accum = 0.0;
					for ( int ik = 0; ik < _k; ++ik )
					  {
						//int k = ranking[ik];
						_pi(ik) = std::exp( nolbeta_accum + lbeta(ik) );
						nolbeta_accum += nolbeta(ik);				
					  }
				  }		  
				break;
			  }
			case FB: 
			  {
				_pi = _z0;
			  }
			}

		}

	  _l_betas.free();
	  _l_thresholds.free();

	  // prepare discriminative base prior when switching to discriminative
	  if ( _requested_parameter_mode == DISCRIMINATIVE &&
		   _current_parameter_mode != DISCRIMINATIVE )
		{
		  if ( _pi_discriminative.numElements() == 0 )
			{
			  _pi_discriminative.reference( _pi.copy() );
			  //std::cout << " pi discriminative ( from generative ) " << _pi_discriminative << std::endl;			  
			}
		  // signals the emission model that the generative parameters need updating
		  BOOST_FOREACH( Matrix4 & param, _generative_emission_parameters )
			param.free();
		  // put references in the right places
		  _pi_generative.reference( _pi );
		  _pi.reference( _pi_discriminative );
		}

	  _current_parameter_mode = _requested_parameter_mode;

	}
	
	OPENMP(single)
	{
	  _dirty = false;
	}
	return true;
  }
  
  void check_sigmoid_params()
  {
	BLITZ_INDICES;
	DISPATCH(update());
	OPENMP(single)
	{
	  // prepare sigmoid parameters
	  // here the component likelihood stats collectors are abused as detectors of initialization state
	  // if not discriminative (i.e. generative) is requested, or discriminative is nil, use stats for initialization	  
	  if ( ( _current_parameter_mode != DISCRIMINATIVE || _l_betas_discriminative.numElements() == 0 ) && _l_betas.numElements() == 0 )
		{
		  // initialize thresholds and gains so that the sigmoid's avtive region covers the whole likelihood distribution
		  //if ( _w_ll_x.numElements() > 0 )
		  //throw std::runtime_error( "thresholds are broken for likelihood distrib stats use to initialize sigmoid parameters!!!" );
		  //note: this might end up using the discriminative parameters, though in that case we just started and they are identical
		  check_likelihood_stats(false); // dont force reset
		  Matrix3 lmax = DISPATCH(calc_max_component_likelihoods());
		  T Lmax = max( sum( sum( lmax( i2, i3, i1 ), i3 ), i2 ) + safe_log( _pi ) );
		  T L_mean     = _w_lL_x  / _w_lL;
		  T L_mean2    = _w_lL_x2 / _w_lL;
		  T L_variance = L_mean2 - L_mean * L_mean;
		  _L_beta      = T(1) / std::sqrt( std::max( T(1e-10), L_variance ) );
		  _L_threshold = Lmax - L_mean;
		  std::cout << " threshold " << _L_threshold << " scale " << (1/_L_beta) << std::endl;
		  // makes new matrices, dropping any possible reference to discriminative parameters
		  Matrix3 l_mean    ( _w_ll_x  / _w_ll );
		  Matrix3 l_mean2   ( _w_ll_x2 / _w_ll );
		  Matrix3 l_variance( l_mean2 - l_mean * l_mean );
		  _l_betas.reference( Matrix3( T(1) / sqrt( max( T(1e-10), l_variance ) ) ) );
		  _l_thresholds.reference( Matrix3( lmax - l_mean ) );
		}
	  // if discriminative is requested and discriminative is nil use a copy of generative
	  if ( _current_parameter_mode == DISCRIMINATIVE && _l_betas_discriminative.numElements() == 0 )
		{
		  std::cout << " copying discriminative sigmoid parameters from generative ones" << std::endl;
		  _L_beta_discriminative = _L_beta;
		  _L_threshold_discriminative = _L_threshold;
		  _l_betas_discriminative.reference( _l_betas.copy() );
		  _l_thresholds_discriminative.reference( _l_thresholds.copy() );
		}
	}
  }

  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {

	if ( _learning && !_online_update )
	  {
		throw std::runtime_error("cannot serialize while learning because statistics are invalid!");
	  }


#define IGNORE_NVP(type,member)					\
	{											\
	  type member;								\
	  ar & BOOST_SERIALIZATION_NVP(member);		\
	}

    ar & BOOST_SERIALIZATION_NVP(_k);
	if ( version < 5 )
	  {
		int _d;
		ar & BOOST_SERIALIZATION_NVP(_d);
		_d1 = _d2 = _d;
	  }
	else
	  {
		ar & BOOST_SERIALIZATION_NVP(_d1);
		ar & BOOST_SERIALIZATION_NVP(_d2);
	  }
    ar & BOOST_SERIALIZATION_NVP(_m);
	if ( version > 8 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_step1);
		ar & BOOST_SERIALIZATION_NVP(_step2);
	  }
	else
	  {
		int _step;
		ar & BOOST_SERIALIZATION_NVP(_step);
		_step1 = _step2 = _step;
	  }
    
	ar & BOOST_SERIALIZATION_NVP(_base_prior);
		
    ar & BOOST_SERIALIZATION_NVP(_alpha0);
    ar & BOOST_SERIALIZATION_NVP(_initcount);
    
    ar & BOOST_SERIALIZATION_NVP(_initialized);

	ar & BOOST_SERIALIZATION_NVP(_wk_prior);
	ar & BOOST_SERIALIZATION_NVP(_wnx_prior);
	ar & BOOST_SERIALIZATION_NVP(_wk);
	ar & BOOST_SERIALIZATION_NVP(_wnx);

	array_serializer< Matrix4d, num_emission_statistics >::serialize
	  ( ar
		, _emission_statistics
		, "_emission_statistics" );

	array_serializer< Matrix4d, num_emission_statistics >::serialize
	  ( ar
		, _prior_emission_statistics
		, "_prior_emission_statistics" );

	array_serializer< Matrix4 , num_emission_statistics >::serialize
	  ( ar
		, _discriminative_emission_parameters
		, "_discriminative_emission_parameters" );

	array_serializer< Matrix4d , num_emission_statistics >::serialize
	  ( ar
		, _discriminative_emission_hessian
 		, "_discriminative_emission_hessian" );

	if ( version == 6 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_output_hessian);
		ar & _segment;
	  }

	if ( version > 6 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_output_hessian);
		ar & BOOST_SERIALIZATION_NVP(_segment);
		ar & BOOST_SERIALIZATION_NVP(_gate_W_neighbor);
	  }

	if ( version > 7 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_wg);	
		ar & BOOST_SERIALIZATION_NVP(_wg_prior);
		ar & BOOST_SERIALIZATION_NVP(_alpha0g);
		if ( version > 8 )
		  ar & BOOST_SERIALIZATION_NVP(_mu_m0);
		else
		  IGNORE_NVP(T,_mug0);
		ar & BOOST_SERIALIZATION_NVP(_dependent_gates);
	  }
	else
	  _dependent_gates = false;

	if ( version > 9 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_gate_neighborhood);
	  }

	if ( version > 10 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_mu_m_prior);
		ar & BOOST_SERIALIZATION_NVP(_mu_mask);
	  }

	if ( version > 11 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_have_bits);
		ar & BOOST_SERIALIZATION_NVP(_mu_b0);
	  }

	if ( version > 12 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_likelihood_stats_emwa);
		array_serializer< std::list<bitwise_emission_gradient_covariance_t>, num_emission_statistics >::serialize
		  ( ar
			, _emission_gradient_covariance
			, "_gradient_covariance" );
	  }

	if ( version > 13 )
	  {
		ar & BOOST_SERIALIZATION_NVP( _prior_gradient_covariance );
	  }

	if ( version > 14 )
	  {
		ar & BOOST_SERIALIZATION_NVP( _nk );
		ar & BOOST_SERIALIZATION_NVP( _w_ll );
		ar & BOOST_SERIALIZATION_NVP( _w_ll_x );
		ar & BOOST_SERIALIZATION_NVP( _w_ll_lx );
	  }

	if ( version > 15 )
	  ar & BOOST_SERIALIZATION_NVP( _w_ll_x2 );

	if ( version > 16 )
	  {
		ar & BOOST_SERIALIZATION_NVP( _w_lL );
		ar & BOOST_SERIALIZATION_NVP( _w_lL_x );
		ar & BOOST_SERIALIZATION_NVP( _w_lL_lx );
		ar & BOOST_SERIALIZATION_NVP( _w_lL_x2 );
		ar & BOOST_SERIALIZATION_NVP( _L_beta_discriminative );
		ar & BOOST_SERIALIZATION_NVP( _L_threshold_discriminative );
		ar & BOOST_SERIALIZATION_NVP( _l_betas_discriminative );
		ar & BOOST_SERIALIZATION_NVP( _l_thresholds_discriminative );
		ar & BOOST_SERIALIZATION_NVP( _use_detection_sigmoid );
		ar & BOOST_SERIALIZATION_NVP( _rescale_detection_likelihoods );
	  }

	if ( version > 17 )
	  {
		//std::cout << " pi discriminative ( pre serialization ) " << _pi_discriminative << std::endl;
		ar & BOOST_SERIALIZATION_NVP( _pi_discriminative );
		//std::cout << " pi discriminative ( post serialization ) " << _pi_discriminative << std::endl;
	  }

	if ( version > 18 )
	  {
		ar & BOOST_SERIALIZATION_NVP( _gd_count );
		ar & BOOST_SERIALIZATION_NVP( _epoch_count );
		ar & BOOST_SERIALIZATION_NVP( _likelihood_history );		
		array_serializer< Matrix1d, num_emission_statistics >::serialize
		  ( ar
			, _background_statistics
			, "_background_statistics" );		
		array_serializer< Matrix1d, num_emission_statistics >::serialize
		  ( ar
			, _prior_background_statistics
			, "_prior_background_statistics" );		
	  }

	BOOST_FOREACH( Matrix4 & params, _generative_emission_parameters )
	  params.free();

	if ( !_emission_statistics[0].numElements() == 0 )
	  {
		BOOST_FOREACH( Matrix1 & params, _background_parameters )
		  params.free();
	  }

	recalc_hyperparams();
  }
  
  int _k, _d1, _d2, _m, _step1, _step2, _s_mrf_iters;

  T _alpha0, _alpha0g, _mu_m0, _mu_m_prior, _mu_b0, _mu_b0_saved, _initcount, _Ds, _s0, _z0, _gate_W_neighbor, _gate_clustering_alpha0; 
  T _delta, _tau, _tau_p, _lambda_p, _anneal_step, _s_prior_attenuation, _convergence_rate;
  T _discriminative_generative_coupling, _gradient_descent_rate, _max_gradient_descent_rate, _gradient_descent_momentum, _hessian_estimation_counter, _lambda_ng;
  T _Lz, _Lp, _Lmu, _L;
  T _last_rate, _current_rate;
  bool _linear_annealing, _sparsity_annealing;
  T _sparsity_annealing_time;

  struct likelihood_stats_emwa
  {

	void consume_positive_example( T l, T w = 1, T dt = 1 )
	{
	  _l_emwa( l, w, dt );
	  _l2_emwa( l * l, w, dt );
	  _lp_emwa( l, w, _tp + dt );
	  _lp2_emwa( l * l, w, _tp + dt );
	  _tp = 0;
	  _tn += dt;
	}

	void consume_negative_example( T l, T w = 1, T dt = 1 )
	{
	  _l_emwa( l, w, dt );
	  _l2_emwa( l * l, w, dt );
	  _ln_emwa( l, w, _tn + dt );
	  _ln2_emwa( l * l, w, _tn + dt );
	  _tn = 0;
	  _tp += dt;
	}
	
	T separability() 
	{ 
	  T l_mean   = _l_emwa.mean();
	  T l2_mean  = _l2_emwa.mean();
	  T lp_mean  = _lp_emwa.mean();
	  T lp2_mean = _lp2_emwa.mean();
	  T ln_mean  = _ln_emwa.mean();
	  T ln2_mean = _ln2_emwa.mean();
	  T sigma_lp = lp2_mean - lp_mean * lp_mean;
	  T sigma_ln = ln2_mean - ln_mean * ln_mean;
	  T sigma_l_mean = T(0.5) * ( sigma_lp + sigma_ln );
	  T sigma_l  = l2_mean  - l_mean  * l_mean;
	  T sigma_ratio = sigma_l / sigma_l_mean;
	  return sigma_ratio;
	}

	T calc_descent_factor( bool positive, T l )
	{
	  T l_mean   = _l_emwa.mean();
	  T l2_mean  = _l2_emwa.mean();
	  T lp_mean  = _lp_emwa.mean();
	  T lp2_mean = _lp2_emwa.mean();
	  T ln_mean  = _ln_emwa.mean();
	  T ln2_mean = _ln2_emwa.mean();
	  T sigma_l  = l2_mean  - l_mean  * l_mean;
	  T sigma_lp = lp2_mean - lp_mean * lp_mean;
	  T sigma_ln = ln2_mean - ln_mean * ln_mean;
	  T sigma_l_mean = T(0.5) * ( sigma_lp + sigma_ln );
	  T sigma_ratio = sigma_l / sigma_l_mean;
	  T factor = T(1) / sigma_l;
	  if ( positive )
		factor *= ( T(1) - sigma_ratio ) * l - ( l_mean - sigma_ratio * lp_mean );
	  else
		factor *= ( T(1) - sigma_ratio ) * l - ( l_mean - sigma_ratio * ln_mean );
	  std::cout << " separability stats " << ( positive ? "(pos)" : "(neg)" ) << " l = " 
				<< l_mean  << " (" << (sigma_l)  << ") lp = " 
				<< lp_mean << " (" << (sigma_lp) << ") ln = "
				<< ln_mean << " (" << (sigma_ln) << ") factor = "
				<< factor << std::endl;
	  return factor;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
	  ar & BOOST_SERIALIZATION_NVP( _tp );
	  ar & BOOST_SERIALIZATION_NVP( _tn );
	  ar & BOOST_SERIALIZATION_NVP( _l_emwa );
	  ar & BOOST_SERIALIZATION_NVP( _l2_emwa );
	  ar & BOOST_SERIALIZATION_NVP( _lp_emwa );
	  ar & BOOST_SERIALIZATION_NVP( _ln_emwa );
	  ar & BOOST_SERIALIZATION_NVP( _lp2_emwa );
	  ar & BOOST_SERIALIZATION_NVP( _ln2_emwa );
	}

	size_t counter() const { return _l_emwa.counter(); }

	T get_tau() const { return _l_emwa.get_tau(); }

	void set_tau( T tau )
	{
	  _l_emwa.set_tau( tau );
	  _l2_emwa.set_tau( tau );
	  _lp_emwa.set_tau( tau );
	  _ln_emwa.set_tau( tau );
	  _lp2_emwa.set_tau( tau );
	  _ln2_emwa.set_tau( tau );
	}

	void reset( )
	{
	  _l_emwa.reset( );
	  _l2_emwa.reset( );
	  _lp_emwa.reset( );
	  _ln_emwa.reset( );
	  _lp2_emwa.reset( );
	  _ln2_emwa.reset( );
	}

  private:

	size_t _tp, _tn; // idle counters
	emwa<T> _l_emwa, _l2_emwa, _lp_emwa, _ln_emwa, _lp2_emwa, _ln2_emwa;
	
  } _likelihood_stats_emwa;


  float _min_rf_saliency;

  bool _renormalize_counts;
  bool _segment;
  bool _initialized, _dirty;
  bool _do_vb;  
  bool _learning;
  bool _learning_likelihood;
  bool _online_update;
  bool _estimating_hessian;
  bool _report_gradient;
  bool _delta_to_random;
  bool _report_convergence;
  bool _sequential_update;
  bool _init_uniform;
  bool _sampled;
  bool _converged, _s_converged;
  bool _cache_z;
  bool _dependent_gates;
  bool _have_bits;
  bool _optimize_separability;
  bool _diagonal_output_jacobian;
  bool _have_background_model;
  bool _use_detection_sigmoid;
  bool _optimize_detection_sigmoid_gain;
  bool _rescale_detection_likelihoods;
  bool _continuous_sampling;
  bool _and_combine_bits_and_gates;
  bool _learn_background_model;
  bool _allow_sample_rejection;
  bool _cluster_gates;
  float _gamma_mlr;
  bool _had_upward;

  BasePrior _base_prior;

  // this determnines the currently needed parameter set for inference
  // in discriminative mode and when learning generative parameters are needed as well
  ParameterMode _current_parameter_mode, _requested_parameter_mode;
  GradientMode _gradient_mode;
  GradientDescentStrategy _gradient_descent_strategy;
  Neighborhood _gate_neighborhood;

  Matrix1d _alpha;
  Matrix3d _alphag;
  Matrix1 _pi, _pi_generative, _pi_discriminative;
  Matrix1d _pi_init;
  // these are message contribs for gated bp
  // for a certain gate setting
  Matrix3 _sbuffer, _zbuffer, _bbuffer;
  Matrix3 _sbuffer_last, _zbuffer_last, _bbuffer_last;
  Matrix3 _lzbias;
  IMatrix2 _gates_sample;
  BMatrix3 _mask_sample;
  BMatrix2 _bits_sample;
  Matrix2 _bits_marginal;
  Matrix2 _background_likelihoods;
  Matrix3 _output_error;
  Matrix2 _S;
  Matrix3 _downward_message;
  Matrix3 _backpropagated_error;
  Matrix3 _mu_mask, _lmu_combined, _lnotmu_combined;
  Matrix2 _p_foreground;
  Matrix2 _n_foreground;
  Matrix3 _output_gradient;
  Matrix2 _likelihood_gradient;
  Matrix1 _output_hessian;
  Matrix3 _current_output_hessian;

  Matrix1d _wk, _wk_prior, _nk;
  Matrix3d _wnx, _wnx_prior;
  Matrix3d _wg, _wg_prior;

  Matrix3 _lr0;

  message_component<T> _base_prior_message;

  std::vector< Matrix1d > _wk_local;
  std::vector< Matrix3d > _wnx_local;
  std::vector< Matrix3d > _wg_local;

  emission_statistics_t _emission_statistics;
  emission_statistics_t _prior_emission_statistics;
  std::vector< emission_statistics_t > _local_emission_statistics;
  emission_parameters_t _generative_emission_parameters;
  emission_parameters_t _discriminative_emission_parameters;

  background_statistics_t _prior_background_statistics;
  background_statistics_t _background_statistics;
  std::vector< background_statistics_t > _local_background_statistics;
  background_parameters_t _background_parameters;



  // learning progress
  int _epoch_count;
  int _gd_count;
  struct likelihoods_t
  {
	likelihoods_t( double lp = 0, double lmu = 0, double lz = 0 )
	  : Lp(lp), Lmu(lmu), Lz(lz), L(lz+lmu+lp) 
	{}
	double Lp, Lmu, Lz, L;
	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
	  ar & BOOST_SERIALIZATION_NVP( Lp );
	  ar & BOOST_SERIALIZATION_NVP( Lmu );
	  ar & BOOST_SERIALIZATION_NVP( Lz );
	  ar & BOOST_SERIALIZATION_NVP( L );
	}
  };
  std::vector<likelihoods_t> _likelihood_history;

  struct input_likelihood_t
  {
	Matrix2 data, gradient;
	int d1,d2,s1,s2;
	input_likelihood_t() : d1(1), d2(2), s1(1), s2(2) {}
	void set( Matrix2 & _data, int _d1, int _s1, int _d2, int _s2 )
	{
	  d1 = _d1;
	  d2 = _d2;
	  s1 = _s1;
	  s2 = _s2;
	  data.reference( _data );
	}
	void clear()
	{
	  data.free();
	  gradient.free();
	}	
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
	  ar & BOOST_SERIALIZATION_NVP( d1 );
	  ar & BOOST_SERIALIZATION_NVP( d2 );
	  ar & BOOST_SERIALIZATION_NVP( s1 );
	  ar & BOOST_SERIALIZATION_NVP( s2 );
	}

  };

  boost::array<input_likelihood_t,3> _input_likelihoods;

  emission_gradient_t _last_emission_gradient;
  emission_gradient_t _emission_gradient_flip_counter;
  Matrix1 _last_prior_gradient;
  T _last_L_threshold_gradient;
  T _last_L_beta_gradient;

  emission_statistics_t _discriminative_emission_hessian;
  T _L_threshold_hessian;
  T _L_beta_hessian;

  Matrix3d _w_ll;
  Matrix3d _w_ll_x;
  Matrix3d _w_ll_x2;
  Matrix3d _w_ll_lx;

  Matrix3 _l_betas, _l_thresholds;
  T _L_beta, _L_threshold;

  Matrix3 _l_betas_discriminative, _l_thresholds_discriminative;
  T _L_beta_discriminative, _L_threshold_discriminative;

  double _w_lL, _w_lL_x, _w_lL_lx, _w_lL_x2;

  // we compute gradient covariance for each output_bit independently
  boost::array< std::list<bitwise_emission_gradient_covariance_t>, num_emission_statistics > _emission_gradient_covariance;
  prior_gradient_covariance_t _prior_gradient_covariance;

  // openmp temporaries for reduction
  emission_gradient_t _accumulated_emission_gradient;
  emission_gradient_t _accumulated_emission_hessian;
  Matrix1 _accumulated_prior_gradient;
  T _accumulated_L_threshold_gradient;
  T _accumulated_L_beta_gradient;

  bool input_is_distribution() { return false; }

  bool segment() { return _segment && std::max( _d1, _d2 ) > 1; }

  int j_index_to_j1( int j_index, int i1 ) const
  {
	// count from the first parent up
	int np1 = ( _d1 + _step1- 1 ) / _step1; // number of parents in 1st dim
	int np2 = ( _d2 + _step2- 1 ) / _step2; // number of parents in 2nd dim
	int j0 = std::max( 0, ( i1 - _d1 + _step1 ) / _step1 ); // this is the index of the first parent of this input in the 1st dimension
	int ji0 = j0 % np1; // this is the j_index of j0 in the 1st dimension
	int ji1 = j_index / np2; // this is j_index in the 1st dimension
	// this is j0 plus the modulo distance between the start and the current j_index ( in the 1st dimension )
	if ( ji1 >= ji0 ) return j0 + ji1 - ji0;
	else return j0 + np1 - ji0 + ji1;
  }

  int j_index_to_j2( int j_index, int i2 ) const
  {
	// count from the first parent up
	int np2 = ( _d2 + _step2 - 1 ) / _step2; // number of parents in 2nd dim
	int j0 = std::max( 0, ( i2 - _d2 + _step2 ) / _step2 );  // this is the index of the first parent of this input in the 2nd dimension
	int ji0 = j0 % np2; // this is the j_index of j0 in the 2nd dimension
	int ji1 = j_index % np2; // this is j_index in the 2nd dimension
	// this is the modulo distance between the start and the current j_index ( in the 2nd dimension )
	if ( ji1 >= ji0 ) return j0 + ji1 - ji0;
	else return j0 + np2 - ji0 + ji1;
  }

  void apply_r( const Matrix3 & data, bool update_component_model = true )
  {
	BLITZ_INDICES;			

	size_t thread_num = OPENMP_THREAD_NUM;
	typedef update_gated_msg_delegate<Derived> delegate;

	have_statistics( thread_num );


	bool have_saliency = _S.numElements() > 0;
	Matrix3 r;
	Matrix2 wS(_S,true);

	//#define report(x) std::cout << BOOST_PP_STRINGIZE(x) << " shape " << x.shape() << " min " << min(x) << " mean " << mean(x) << " max " << max(x) << std::endl;
#define report(x)

	Matrix3d wnxw( _wnx_local[thread_num], true );
	if ( _sampled )
	  {
		IMatrix2 & latents_sample = message_component<T>::buffer().get_sample();
		if ( DISPATCH(ready)() )
		  {
			Matrix1d wkw( _wk_local[thread_num], true );
			ForwardRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1,_step2, true );
			for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
			  {
				int k_j = latents_sample( iter.j1, iter.j2 );
				T s0 = 0, n = 0;
				for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
				  {
					if ( int(_gates_sample( iter.i1, iter.i2 )) == iter.j_index() )
					  {
						T s0_i = have_saliency ? _S(iter.i1, iter.i2) : T(1);
						s0 = s0_i;
						n += T(1);
					  }
				  } // iteration over inputs
				// this yields the mean saliency of the attended region
				// without saliency this is always one
				s0 /= std::max( T(1), n );
				wkw( k_j ) += s0;
				// if parent is active and gating distrib is learned collect dependent gating stats here
				if ( _dependent_gates && ( !_have_bits || _bits_sample( iter.j1, iter.j2 ) ) )
				  {
					for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
					  {
						// mask bit gets integrated out, 
						T s0_i = have_saliency ? _S(iter.i1, iter.i2) : T(1);
						T delta_ct = s0_i;
						bool success = int(_gates_sample( iter.i1, iter.i2 )) == iter.j_index();
						OPENMP(critical)
						{
						  T & mu_g = _mu_mask( iter.i1_rf(), iter.i2_rf(), k_j );
						  // if the gate didnt switch here the model predicts the mask
						  T p_m = success ? T(1) : binomial_combine( mu_g, _mu_m_prior );
						  double & ct_g = _wg( iter.i1_rf(), iter.i2_rf(), k_j );
						  T mu_new =  ( mu_g * ( ct_g + _alpha0g ) + delta_ct * p_m ) / ( ct_g + delta_ct + _alpha0g );
						  mu_g = mu_new;
						  ct_g += delta_ct;
						}
					  } // iteration over inputs
				  } // dependent gates
			  } // iteration over outputs
		  }
		if ( update_component_model )
		  {
			InverseRFIterator iter( data.extent(0), data.extent(1), _d1, _d2, _step1,_step2, true );
			Matrix3 dataw( data, true );
			assert( segment() ); // FIXME :)
			for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
			  {
				T s0_i = have_saliency ? _S(iter.i1, iter.i2) : T(1);
				int j_i = _gates_sample( iter.i1, iter.i2 ); iter.select_j_index( j_i );
				Matrix1 x_i( dataw( iter.i1, iter.i2, _all ) );
				int k_j = latents_sample( iter.j1, iter.j2 );
				bool fg = true;
				if ( _have_bits && _and_combine_bits_and_gates )
				  fg &= _bits_sample( iter.j1, iter.j2 );
				if ( fg )
				  DISPATCH( collect_sampled_statistics ) ( x_i, iter.i1_rf(), iter.i2_rf(), k_j, s0_i );
				else if ( _have_background_model && _learn_background_model )
				  DISPATCH( collect_sampled_background_statistics ) ( x_i, s0_i );
				wnxw( k_j, iter.i1_rf(), iter.i2_rf() ) += s0_i;
			  }
		  }
	  } 
	else // not sampled
	  {
	
		Matrix2 pfg;
		if ( _have_background_model )
		  pfg.resize( _d1, _d2 );

		if ( DISPATCH(ready()) )
		  {
			
			Matrix1d wkw( _wk_local[thread_num], true );
			Matrix3d wnxw( _wnx_local[thread_num], true );
			
			r.reference( message_component<T>::get_belief(), true );
			report(r);
			T Lz = 0;

			if ( _have_background_model )
			  {
				for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2);!it.atend(); ++it)
				  {
					Matrix3 & X = it.in();			
					int i0 = it.in_range1().first();
					int j0 = it.in_range2().first();
					for ( int i = 0; i < _d1; ++i )
					  for ( int j = 0; j < _d2; ++j )
						{
						  Matrix1 x( X( i, j, _all ) );
						  pfg( i, j ) = delegate::foreground_probability( *this, x, it.out(), i0 + i, j0 + j, i, j );
						}
					T n = mean(pfg);
					Matrix1 & p = it.out();
					if ( _min_rf_saliency > 0 && n < _min_rf_saliency ) continue;
					Lz += n * ( message_component<T>::buffer().type() == mmm2_belief_state<T>::BINOMIAL  ) 
					  ? ( - sum( p * safe_log( p ) + ( ( T(1) - p ) * safe_log( T(1) - p ) ) ) )
					  : ( - sum( p * safe_log( p ) ) );
					if ( !segment() )
					  wnxw += it.out()(i1) * pfg(i2,i3);
					wkw += it.out()(i1) * n;
				  }
			  }
			else if ( have_saliency )
			  {
				report(_S);
				for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2);!it.atend(); ++it)
				  {
					Matrix2 w( wS( it.in_range1(),it.in_range2() ) );
					T n = mean(w);
					Matrix1 & p = it.out();
					if ( _min_rf_saliency > 0 && n < _min_rf_saliency ) continue;
					Lz += n * ( message_component<T>::buffer().type() == mmm2_belief_state<T>::BINOMIAL  ) 
					  ? ( - sum( p * safe_log( p ) + ( ( T(1) - p ) * safe_log( T(1) - p ) ) ) )
					  : ( - sum( p * safe_log( p ) ) );
					if ( !segment() )
					  wnxw += it.out()(i1) * w(i2,i3);
					wkw += it.out()(i1) * n;
				  }
			  } 
			else
			  {
				Matrix3 p( r( WORK_RANGE( r.extent(0) ), _all, _all ) );
				Lz += ( message_component<T>::buffer().type() == mmm2_belief_state<T>::BINOMIAL  ) 
				  ? ( - sum( p * safe_log( p ) + ( ( T(1) - p ) * safe_log( T(1) - p ) ) ) )
				  : ( - sum( p * safe_log( p ) ) );
				Matrix1 wk( sum( sum( p(i2,i3,i1), i3 ), i2 ) );
				if ( !segment() )
				  wnxw += wk(i1) + 0 * ( i2 + i3 );
				wkw  += wk;
			  }
			report(wnxw);
			report(wkw);
			OPENMP(atomic)
			  _Lz += Lz;
			r.reference( message_component<T>::get_belief( DISPATCH(input_is_distribution()) ), true );
		  }
		else
		  {
			OPENMP(barrier)
			OPENMP(master)
			{
			  blitz::TinyVector<int,2> s2 = outshape( data.shape() );
			  r.resize( s2[0], s2[1], 1 );
			}
			OPENMP(barrier)
		  }
	  
		//update statistics of submodels
		if ( update_component_model )
		  {
			if ( segment() )
			  {				
				Matrix3 dataw( data, true );
				Matrix1 w(_zbuffer.extent(2));
				for ( InverseRFIterator iter( data.extent(0), data.extent(1), _d1, _d2,_step1,_step2); !iter.atend_i(); iter.increment_i() )
				  {
					T s0 = have_saliency ? wS( iter.i1, iter.i2 ) : T(1);
					Matrix1 x( dataw( iter.i1 , iter.i2 , _all ) );
					w = _zbuffer( iter.i1 , iter.i2, _all );
					if ( _sbuffer.numElements() > 0 )
					  w *= _sbuffer( iter.i1 , iter.i2, _all );
					if ( _bbuffer.numElements() > 0 )
					  w *= _bbuffer( iter.i1 , iter.i2, _all );
					T Zw = sum(w);
					if ( Zw > T(1e-10) )
					  {
						DISPATCH(collect_statistics_gated( x, r, w, s0 / Zw, iter ));
					  }
				  }
			  }
			else
			  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2);!it.atend(); ++it)
				{
				  Matrix3 & X = it.in();			
				  if ( _have_background_model )
					{
					  int i0 = it.in_range1().first();
					  int j0 = it.in_range2().first();
					  for ( int i = 0; i < _d1; ++i )
						for ( int j = 0; j < _d2; ++j )
						  {
							Matrix1 x( X( i, j, _all ) );
							pfg( i, j ) = delegate::foreground_probability( *this, x, it.out(), i0 + i, j0 + j, i, j );
						  }
					  DISPATCH(collect_statistics(X,it.out(),pfg));
					}
				  else if ( have_saliency )
					{
					  Matrix2 w( wS( it.in_range1(),it.in_range2() ) );
					  DISPATCH(collect_statistics(X,it.out(),w));
					} 
				  else
					{
					  DISPATCH(collect_statistics(X,it.out()));
					}
				  
				}
		  }
	  }

#undef report

  }
  
  void collect_statistics(const Matrix3 & data, const Matrix1 & r) {}
  
  void collect_statistics(const Matrix3 & data, const Matrix1 & r, const Matrix2 & w2) {}

  void have_statistics( size_t i )
  {
    OPENMP(critical)
    {
      if ( _wk_local.size() <= i )
      {		  
        _wk_local.resize( i + 1 );
        _wnx_local.resize( i + 1 );
        _local_emission_statistics.resize( i + 1, array_initializer<Matrix4d,num_emission_statistics>::make() );
        _local_background_statistics.resize( i + 1, array_initializer<Matrix1d,num_emission_statistics>::make() );
      }
      if ( _wk_local[i].extent(0) == 0 )
      {
        _wk_local[i].resize( _wk.shape() );		 
        _wk_local[i] = 0;
        _wnx_local[i].resize( _wnx.shape() );
        _wnx_local[i] = 0;
        for ( size_t j = 0; j < num_emission_statistics; ++j )
        {
          _local_emission_statistics[i][j].resize( _emission_statistics[j].shape() );
          _local_emission_statistics[i][j] = 0;
          _local_background_statistics[i][j].resize( _background_statistics[j].shape() );
          _local_background_statistics[i][j] = 0;
        }
      }
    }
	OPENMP(barrier);
  }

  void prepare_statistics() 
  {
    OPENMP(single)
    {
      _wk_local.resize( 1 );
      _wnx_local.resize( 1 );
      _local_emission_statistics.resize( 1, array_initializer<Matrix4d,num_emission_statistics>::make() );
      _local_background_statistics.resize( 1, array_initializer<Matrix1d,num_emission_statistics>::make() );

      // fast path for first thread: directly operate on the data
      _wk_local[0].reference( _wk );
      _wnx_local[0].reference( _wnx );
      for ( size_t i = 0; i < num_emission_statistics; ++i )
		{
		  _local_emission_statistics[0][i].reference( _emission_statistics[i] );
		  _local_background_statistics[0][i].reference( _background_statistics[i] );
		}

    }
  }

  void finalize_statistics() 
  {
	{
	  size_t nthreads = _wk_local.size();
	  // skip first thread: operates directly on data
	  for ( size_t i = 1; i < nthreads; ++i )
		{
		  _wk += _wk_local[i];
		  _wnx += _wnx_local[i];
		  for ( size_t j = 0; j < num_emission_statistics; ++j )
			{
			  _emission_statistics[j] += _local_emission_statistics[i][j];
			  _background_statistics[j] += _local_background_statistics[i][j];
			}
		}
	  _wk_local.clear();
	  _wnx_local.clear();
	  _local_emission_statistics.clear();
	  _local_background_statistics.clear();
	}

  }
};

#include "mm2b_impl.cc"

#undef DISPATCH

//templated class versioning
namespace boost { namespace serialization {
	template<typename Derived >
	struct version< MixtureModel2Base< Derived > >
	{
	  BOOST_STATIC_CONSTANT(unsigned int, value = 19);
	};
  }}
#endif /* _MIXTUREMODEL2BASE_HPP_ */
