/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _RECEPTIVEFIELDITERATOR_HPP_
#define _RECEPTIVEFIELDITERATOR_HPP_

#include <blitz/array.h>
#include <iostream>

#include "omp_macros.hpp"

template<typename M> struct ReceptiveFieldIteratorInputTraits
{
};


// FIXME: these need to be made such that they do not use any reference counting
// to avoid threading issues

template<typename T> struct ReceptiveFieldIteratorInputTraits< blitz::Array<T,3> >
{
  typedef blitz::Array<T,3> & OutType;
  OutType get_in( blitz::Array<T,3> & m, int x, int y, int d1, int d2 )
  {
	_tmp.reference(m(blitz::Range(x,x+d1-1),blitz::Range(y,y+d2-1),blitz::Range::all()));
    return _tmp;
  }
  blitz::Array<T,3> _tmp;
};

template<typename T> struct ReceptiveFieldIteratorInputTraits< blitz::Array<T,2> >
{
  typedef blitz::Array<T,2> & OutType;
  OutType get_in( blitz::Array<T,2> & m, int x, int y, int d1, int d2 )
  {
    _tmp.reference(m(blitz::Range(x,x+d1-1),blitz::Range(y,y+d2-1)));
    return _tmp;
  }
  blitz::Array<T,2> _tmp;
};

template<typename M> struct ReceptiveFieldIteratorOutputTraits
{
};

template<typename T> struct ReceptiveFieldIteratorOutputTraits< blitz::Array<T,2> >
{
  typedef T & OutType;
  static OutType get_out(blitz::Array<T,2> & m, int x, int y)
  {
    return m(x,y);
  }
};

template<typename T> struct ReceptiveFieldIteratorOutputTraits< blitz::Array<T,3> >
{
  typedef blitz::Array<T,1> & OutType;
  OutType get_out(blitz::Array<T,3> & m, int x, int y)
  {
    _tmp.reference(m(x,y,blitz::Range::all()));
    return _tmp;
  }
  blitz::Array<T,1> _tmp;
};

template<typename in_t, typename out_t>
class ReceptiveFieldIterator
{

public:
 
  ReceptiveFieldIterator(
	                     const in_t & in, 
						 const out_t & out, 
						 int d1, int d2, int step1, int step2, bool weak = true, 
						 bool mutating_input = false )
	: _in( in, weak )
	, _out( out, weak )
	, _yo(0), _step1(step1), _step2(step2), _d1(d1), _d2(d2)
	, _mutating_input(mutating_input)
  {

#ifdef HAVE_OPENMP
	if ( omp_in_parallel() )
	  {

		if ( !_mutating_input || out.extent(0) / omp_get_num_threads() >= _d1 )
		  {
			int numthreads = omp_get_num_threads();
			_nmax = ( _out.extent(0) + numthreads - 1 ) / numthreads;
			_work_range = WORK_RANGE( out.extent(0) );
		  }
		else
		  {
			_nmax = out.extent(0);
			if ( omp_get_thread_num() == 0 )
			  _work_range = blitz::Range( out.lbound(0), out.ubound(0) );
			else
			  _work_range = blitz::Range( 0, -1 );
		  }
		OMP_LOG(str(boost::format( " parallel rf iterator %i,%i nmax: %i %s") % _work_range.first() % _work_range.last() % _nmax % (_mutating_input?"mutating input":"") ) );
	  }
	else
#endif
	  _work_range = blitz::Range( out.lbound(0), out.ubound(0) );


    _xo = _work_range.first();
 
	OMP_LOG(str(boost::format("xo: %i,%i")%_xo%_work_range.first()));

  }

  size_t n_out_total() const { return ( 1 + ( _in.extent(0) - _d1 ) / _step1 ) * ( 1 + ( _in.extent(1) - _d2 ) / _step2 ); }
 
  void operator++ ()
  {
    ++_yo;
    if ( _yo >= _out.extent(1) )
    {
	  _yo = 0;
	  ++_xo;
#ifdef HAVE_OPENMP
	  if ( _mutating_input && omp_in_parallel() )
		{
		  OMP_LOG(str(boost::format("synchronizing at line %i/%i")%_xo%(_work_range.last()+1)));
		  sync();
		}
#endif
    }
  }  

  bool atend()
  {
    return _xo > _work_range.last();
  }

  typename ReceptiveFieldIteratorOutputTraits<out_t>::OutType
  out()
  {
    return _out_helper.get_out(_out,_xo,_yo);
  }
  
  typename ReceptiveFieldIteratorInputTraits<in_t>::OutType
  in()
  {
    return _in_helper.get_in(_in,_xo*_step1,_yo*_step2,_d1,_d2);
  }

  blitz::Range in_range1()
  {
	return blitz::Range(_xo*_step1,_xo*_step1+_d1-1);
  }
  
  blitz::Range in_range2()
  {
	return blitz::Range(_yo*_step2,_yo*_step2+_d2-1);
  }
  
  int out_x()
  {
	return _xo;
  }
  
  int out_y()
  {
	return _yo;
  }

  ~ReceptiveFieldIterator()
  {
#ifdef HAVE_OPENMP
	if ( _mutating_input && omp_in_parallel() )
	  {
		int nwait = _nmax - ( _xo - _work_range.first() );
		OMP_LOG(str(boost::format("destroying parallel rf iterator xo %i nwait %i ")%_xo%nwait));
		for ( int i = 0; i < nwait; ++i )
		  {
			OMP_LOG(str(boost::format("waiting at barrier %i/%i")%(i+1)%nwait));
			sync();
		  }
	  }
#endif
  }
  
private:

  void sync()
  {
	OPENMP(barrier);
  }

  in_t _in;
  out_t _out;
  ReceptiveFieldIteratorInputTraits<in_t> _in_helper;
  ReceptiveFieldIteratorOutputTraits<out_t> _out_helper;
  int _yo, _xo, _step1, _step2, _d1, _d2;
  blitz::Range _work_range;
  bool _mutating_input;
  int _nmax;
};

#ifdef HAVE_OPENMP_DEFINED
#define HAVE_OPENMP 1
#include "omp_macros_undef.hpp"
#undef HAVE_OPENMP_DEFINED
#endif

#endif /* _RECEPTIVEFIELDITERATOR_HPP_ */
