/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "optimal_hyperparameters.hpp"
#include "bmm2.hpp"
#include "digamma.hpp"
#include "normrho.hpp"

#include <boost/format.hpp>

#ifdef HAVE_SSE
#include "sse_approx_logexp.h"
#endif

#include "enum_properties.hpp"

#define FETCH_PARAMETER(D,N,V,I) \
  const blitz::Array<T,D> N( Base::_learning && Base::_do_vb ? (V) : (I), true );N.extent(0); // to avoid unused warnings

#define FETCH_MU												\
  FETCH_PARAMETER( 4 , mu       , _mudash       , _mu       )	\
  FETCH_PARAMETER( 4 , deltamu  , _deltamudash  , _deltamu  )	\
  FETCH_PARAMETER( 4 , notmu    , _not_mudash   , _not_mu   )

#define FETCH_MUR												   \
  FETCH_PARAMETER( 4 , mur      , _mudashr        , _mur         ) \
  FETCH_PARAMETER( 4 , notmur   , _not_mudashr    , _not_mur     ) \
  FETCH_PARAMETER( 4 , deltamur , _deltamudashr   , _deltamur    ) \
  FETCH_PARAMETER( 4 , notmurinv, _not_mudashr_inv, _not_mur_inv )

#define FETCH_LMU												\
  FETCH_PARAMETER( 4 , lmu      , _lmudash      , _lmu      )	\
  FETCH_PARAMETER( 4 , lnotmu   , _lnotmudash   , _lnotmu   )

#define FETCH_PARAMETERS												\
  FETCH_PARAMETER( 1 , r0       , _r0dash       , _r0       )			\
  FETCH_PARAMETER( 3 , r03      , _r0dash3      , _r03      )			\
  FETCH_PARAMETER( 4 , deltalmu , _deltapsibeta , _deltalmu )			\
  FETCH_MUR																\
  FETCH_MU																\
  FETCH_LMU


template<class T> 
BinomialMixtureModel2<T>::BinomialMixtureModel2 ( )
  : Base()
  , _beta0(0.5)
  , _e_mu(0.5)
  , _learn_beta0_interval(0)
  , _learn_beta0_counter(0)
  , _batchng(false)
  , _greedy_clusterer(false)
  , _pre_clustering_mode(NONE)
{
}

template<typename T>
void BinomialMixtureModel2<T>::init_properties()
{

  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter(&Base::_initialized,false);
  PBase::add_ivar_property_rw("beta0", &This::_beta0, uninitializer);
  PBase::add_ivar_property_rw("e_mu", &This::_e_mu, uninitializer);
  PBase::add_ivar_property_rw("learn-beta0", &This::_learn_beta0_interval);
  PBase::add_property( "mu", make_blitz_ivar_getter1( &This::_mu ) );
  PBase::add_property( "beta", make_blitz_ivar_getter1( &This::_beta ) );
  PBase::add_property( "betahat", make_blitz_ivar_getter1( &This::_betahat4 ) );
  PBase::add_property( "wx", make_blitz_ivar_getter1( &This::_wx ) );
  PBase::add_ivar_property_ro("Lmu", &This::_Lmu);

  PBase::add_property( std::string("greedy_clusterer"), PBase::make_pivar_getter( &This::_greedy_clusterer ) );
  PBase::add_property( std::string("batch_ng"), PBase::make_pivar_getter( &This::_batchng ) );

  enum_property_access<SearisePropertyTypes, This, PreClusteringMode> pre_clustering_accessor( &This::_pre_clustering_mode );
  pre_clustering_accessor.add_choice("none",NONE).add_choice("batch_ng",BATCHNG).add_choice("greedy",GREEDY);
  PBase::add_property( "pre_clustering", pre_clustering_accessor, pre_clustering_accessor );

}

template<class T> void BinomialMixtureModel2<T>::flush()
{
  Base::flush();
  _beta.free();
  _betahat4.free();
  _mu.free();
}

template<class T> void BinomialMixtureModel2<T>::batchlearn_begin_epoch()
{
  BLITZ_INDICES;


  if ( !ready() )
	{
	  if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.init( Base::_d1, Base::_d2, Base::_m );
		}
	  else if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.initialize( Base::_k, Base::_d1, Base::_d2, Base::_m );
		  _batchng.begin_epoch();
		}
	  Base::_learning = true;
	}
  else
	{  
	  Base::batchlearn_begin_epoch();	  
	  if ( !Base::_online_update )
		{
		  _wx.resize(Base::_k,Base::_d1,Base::_d2,Base::_m); _wx = 0;
		}
	}
}

template<typename T> void BinomialMixtureModel2<T>::recalc_likelihoods()
{
  Base::_Lp = sum( _wx * _lmudash ) + sum( ( Base::_wnx(i1,i2,i3) - _wx ) * _lnotmudash );  
  
  _Lmu = 0;
  for ( int k = 0; k < Base::_k; ++k )
	for ( int d1 = 0; d1 < Base::_d1; ++d1 )
	  for ( int d2 = 0; d2 < Base::_d2; ++d2 )
		for ( int m = 0; m < Base::_m; ++m )
		  {
			T alphaP = _beta(k,d1,d2,m);
			T betaP  = _betahat4(k,d1,d2,m) - _beta(k,d1,d2,m);
			T alphaQ = _beta0;
			T betaQ  = _beta0;
			_Lmu -= kl_beta( alphaP, betaP, alphaQ, betaQ );
		  }
  
  Base::_L = Base::_Lz + _Lmu + Base::_Lp;
  
  std::cout << str( boost::format("binomial component model ended epoch Lmu = %1.3e Lp = %1.3e L = %1.3e")
					 % _Lmu % Base::_Lp % Base::_L ) << std::endl;
  
}

template<class T> 
blitz::Array<typename BinomialMixtureModel2<T>::num_type,1> BinomialMixtureModel2<T>::batchlearn_end_epoch ( Matrix1 f_in )
{

  BLITZ_INDICES;

  Matrix1 f;

  if ( !ready() )
	{
	  if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.finish();
		  Base::_wk.reference(  _greedy_clusterer.counts ().copy() );
		  Base::_wnx.reference( _greedy_clusterer.counts3().copy() );
		  _wx.reference( _greedy_clusterer.histograms().copy() );
		  Base::_k = _greedy_clusterer.size();
		}
	  else if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.end_epoch();
		  
		  if (_batchng.ready())
			{
			  Base::_wk.reference( _batchng.get_histogram().copy() );
			  Base::_wnx.reference( _batchng.get_histogram3().copy() );
			  _wx.reference( _batchng.get_centerStat().copy() );
			}
		  else
			{
			  _mu.reference( _batchng.get_centers() );
			  return Matrix1();
			}
		}
	  
	  recalc_hyperparams();
	  Base::recalc_hyperparams();
	  Base::_initialized = true;
	  update();
	  recalc_likelihoods();

	}
  else
	{

	  f.reference( Base::batchlearn_end_epoch() );
	  
	  if (Base::_renormalize_counts)
		{
		  _wx = f(i1) * _wx;
		}
	  
	  recalc_hyperparams();

	  if (_learn_beta0_interval > 0 && ++_learn_beta0_counter >= _learn_beta0_interval)
		{
		  T beta0_old = _beta0;
		  update(true);
		  optimize_beta0(Base::_delta);
		  T deltabeta0 = _beta0 - beta0_old;
		  _beta += deltabeta0;
		  _betahat4 += 2 * deltabeta0;
		}
	  
	  update();
	}

  return f;
}

template<typename T>
void BinomialMixtureModel2<T>::decay_statistics()
{
  Base::decay_statistics();
  _wx *= std::exp( - 1 / Base::_tau );
}

template<class T> void BinomialMixtureModel2<T>::recalc_hyperparams()
{

  BLITZ_INDICES;

  Matrix4d diff( Base::_k, Base::_d1, Base::_d2, Base::_m);
  for ( int m = 0; m < Base::_m; ++m )
	diff(_all,_all,_all,m) = ( Base::_wnx - _wx(_all,_all,_all,m) );
  int wcount = sum( where( diff < 0, 1, 0 ) );
  if (wcount > 0)
	{
	  std::cerr << "WARNING!  batchlearn_end_epoch:  " << wcount << " out of " << diff.numElements()
				<< " values in _wx are bigger than in _wnx." << std::endl
				<< "\t maxdiff = " << max( diff ) << "   meandiff = " << mean( diff ) << std::endl
		;
	  _wx = min ( _wx(i1,i2,i3,i4), Base::_wnx(i1,i2,i3) );
	}
  
  
  T newalpha = _beta0;
  T newbeta = newalpha * (T(1) - _e_mu) / _e_mu;
  
  if ( Base::_delta_to_random )
	_beta.reference( Matrix4( newalpha + Base::_delta * _wx + ( T(1) - Base::_delta ) * _wx_prior ) );
  else
	_beta.reference( Matrix4( newalpha + Base::_delta * _wx ) );
  _betahat4.resize( Base::_k, Base::_d1, Base::_d2, Base::_m );
  if ( Base::_delta_to_random )
	for ( int m = 0; m < Base::_m; ++m )
	  _betahat4( _all, _all, _all, m ) = newalpha + newbeta + Base::_delta * Base::_wnx + ( T(1) - Base::_delta ) * Base::_wnx_prior;
  else
	for ( int m = 0; m < Base::_m; ++m )
	  _betahat4( _all, _all, _all, m ) = newalpha + newbeta + Base::_delta * Base::_wnx;

  Base::dirty();  
}


#ifdef DEBUG
#define CHECK_COUNTS(wx)												\
  if (any(wx < 0) )														\
	{																	\
	  throw( std::runtime_error( str( boost::format("negative occurence counts min: %f") % min(wx) ) ) ); \
 	}
#else
#define CHECK_COUNTS(wx)
#endif


template<class T> void BinomialMixtureModel2<T>::collect_statistics ( const Matrix3 & data, const Matrix1 & r )
{
  OPENMP(single)
  {
  if ( !ready() )
	{
	  if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.process( data );
		}
	  else if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.consume( data );
		}
	}
  else
	{	  
	  BLITZ_INDICES;
	  FETCH_MUR;
	  int n1 = data.extent(0), n2 = data.extent(1);
	  for (int i = 0; i < n1; ++i)
		for (int j = 0; j < n2; ++j)
		  {
			Matrix1 dataij( data( i, j, _all ) );
			Matrix2 muij( mur( i, j, _all, _all ) );
			Matrix2 notmuij( notmur( i, j, _all, _all ) );
			Matrix2d wxij( _wx( _all, i, j, _all ) );
#if 0
			Matrix2 wx    ( r(i1) *    muij(i2,i1) * dataij(i2) );
			Matrix2 wnotx ( r(i1) * notmuij(i2,i1) * dataij(i2) );
			Matrix1 Z ( max( 1e-7f, sum( wx(i2,i1) + wnotx(i2,i1), i2 ) ) );
			/*
			std::cout << "wx " << min(wx) << ", " << mean(wx) << ", " << max(wx) << std::endl;
			std::cout << "wnotx " << min(wnotx) << ", " << mean(wnotx) << ", " << max(wnotx) << std::endl;
			std::cout << "Z " << min(Z) << ", " << mean(Z) << ", " << max(Z) << std::endl;
			*/
			wxij += wx / Z(i2);
#else
			wxij += r(i1) *    muij(i2,i1) * dataij(i2) 
			  / max( 1e-7f, sum( r(i3) * ( muij(i2,i3) * dataij(i2) + notmuij(i2,i3) * ( T(1) - dataij(i2) ) ), i3 ) );
#endif
		  }
	  CHECKNAN( _wx,"mmm2 collect statistics _wx" );
	  CHECK_COUNTS( _wx );
	}
  }
}

template<class T> void BinomialMixtureModel2<T>::collect_statistics ( const Matrix3 & data, const Matrix1 & r, const Matrix2 & w2 )
{
  //  std::cout << " bmm2 collect statistics: data " << data << " r " << r << " weights " << w2 << std::endl;
  OPENMP(single)
  {
  if ( !ready() )
	{
	  if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.process( data, w2 );
		}
	  else if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.consume( data, w2 );
		}
	}
  else
	{
	  BLITZ_INDICES;
	  FETCH_MUR;
	  int n1 = data.extent(0), n2 = data.extent(1);
	  for (int i = 0; i < n1; ++i)
		for (int j = 0; j < n2; ++j)
		  {
			Matrix1 dataij( data( i, j, _all ) );
			Matrix2 muij( mur( i, j, _all, _all ) );
			Matrix2 notmuij( notmur( i, j, _all, _all ) );
			Matrix2 wx    ( r(i1) *    muij(i2,i1) * dataij(i2) );
			Matrix2 wnotx ( r(i1) * notmuij(i2,i1) * dataij(i2) );
			Matrix1 Z (   max( 1e-7f, sum( wx(i2,i1) + wnotx(i2,i1), i2 ) ) );
			Matrix2d wxij( _wx( _all, i, j, _all ) );
			/*
			std::cout << "wx " << min(wx) << ", " << mean(wx) << ", " << max(wx) << std::endl;
			std::cout << "wnotx " << min(wnotx) << ", " << mean(wnotx) << ", " << max(wnotx) << std::endl;
			std::cout << "Z " << min(Z) << ", " << mean(Z) << ", " << max(Z) << std::endl;
			*/
			wxij += wx / Z(i2) * w2(i,j);
		  }
	  CHECKNAN(_wx,"mmm2 collect statistics _wx");
	  CHECK_COUNTS(_wx);
	}
  }
}


  ///////////////////////////////////////////
 ////// message calculators - begin ////////
///////////////////////////////////////////

template<class T> void BinomialMixtureModel2<T>::upward_message ( const Matrix3 & data, Matrix1 & lrout, bool marginal)
{
#ifdef HAVE_SSE
  FETCH_MUR;
  // will only work for float!!!
  // also assumes contiguous data in feature directions!!!
  // does simd over output dimension
  lrout = 0;
  size_t n4 = Base::_k / 4;
  for (int i = 0; i < Base::_d1; ++i)
    for (int j = 0; j < Base::_d2; ++j)
    {
      const float * datap = &data( i, j, 0 );
      for ( int m = 0; m < Base::_m; ++m )
      {
        const float * deltamup = &   mur( i, j, m, 0 );
        const float * notmup   = &notmur( i, j, m, 0 );
		      float * outp     = lrout.data();
        const float   xdata    = datap[m];
        // simd-asisted accumulation loop
        const __m128 xxxxdata         = _mm_set1_ps(        xdata );
        for ( size_t iout = 0, k = 0; iout < n4; ++iout, k += 4 )
        {
          const __m128 notmuv   = _mm_loadu_ps( notmup + k );
          const __m128 deltamuv = _mm_loadu_ps( deltamup + k );
          const __m128 arg =  _mm_add_ps( _mm_mul_ps( xxxxdata, deltamuv ), notmuv );
		  _mm_storeu_ps( outp + k, _mm_add_ps( _mm_loadu_ps( outp + k ), logf4( arg ) ) );
		}
		// serial post-roll
		for ( int k = n4 * 4; k < Base::_k; ++k )
          outp[k] += safe_log( xdata * deltamup[k] + notmup[k] );
      }
    }
#else
  BLITZ_INDICES;
  FETCH_MU;
  lrout = sum(sum(sum( safe_log( data(i2,i3,i4) * mu(i1,i2,i3,i4) + ( T(1) - data(i2,i3,i4) ) * notmu(i1,i2,i3,i4) ) ,i4),i3),i2);
#endif
}

template<class T> void BinomialMixtureModel2<T>::upward_message ( const Matrix3 & data, Matrix1 & lrout, 
																  const Matrix2 & w, bool marginal )
{
#ifdef HAVE_SSE
  FETCH_MUR;
  // will only work for float!!!
  // also assumes contiguous data in feature directions!!!
  // does simd over output dimension
  lrout = 0;
  size_t n4 = Base::_k / 4;
  for (int i = 0; i < Base::_d1; ++i)
    for (int j = 0; j < Base::_d2; ++j)
    {
      const float * datap = &data( i, j, 0 );
	  const float wij = w(i,j);
	  const __m128 s4 = _mm_set1_ps( wij );
	  const __m128 o4 = _mm_set1_ps( ( 1.0f - wij ) / Base::_k );
      for ( int m = 0; m < Base::_m; ++m )
      {
        const float * deltamup = &   mur( i, j, m, 0 );
        const float * notmup = &notmur( i, j, m, 0 );
		float * outp = lrout.data();
        const float xdata = datap[m];
        // simd-asisted accumulation loop
        const __m128 xxxxdata         = _mm_set1_ps(        xdata );
		int k = 0;
        for ( size_t k4 = 0; k4 < n4; ++k4, k += 4 )
        {
          const __m128 notmuv = _mm_loadu_ps( notmup + k );
          const __m128 deltamuv = _mm_loadu_ps( deltamup + k );
		  __m128 arg =  _mm_add_ps( _mm_mul_ps( xxxxdata, deltamuv ), notmuv );
		  arg = _mm_add_ps( o4, _mm_mul_ps( s4, arg ) );
		  _mm_storeu_ps( outp + k, _mm_add_ps( _mm_loadu_ps( outp + k ), logf4( arg ) ) );
		}
		// serial post-roll
		for ( ; k < Base::_k; ++k )
		  //outp[k] += safe_log( xdata * mup[k] + ( 1.0f - xdata ) * notmup[k] );
		  outp[k] += safe_log( wij * xdata * deltamup[k] + wij * ( notmup[k] - 1.0f / Base::_k ) + 1.0f / Base::_k );
      }
    }
#else
  FETCH_MU;
  BLITZ_INDICES;
  int n1 = data.extent(0), n2 = data.extent(1), n3 = lrout.extent(0);
  for (int k = 0; k < n3; ++k)
	{
	  lrout(k) = 0;
	  for (int i = 0; i < n1; ++i)
		for (int j = 0; j < n2; ++j)
		  {
			Matrix1 muslice( mu (k,i,j,_all) );
			Matrix1 notmuslice( notmu (k,i,j,_all) );
			Matrix1 dataslice( data (i,j,_all) );
			lrout(k) += sum( safe_log( w(i,j) * ( dataslice * muslice + ( T(1) - dataslice ) * notmuslice ) + ( T(1) - w(i,j) ) ) );
		  }
	}
#endif
}

template<typename T>
void 
BinomialMixtureModel2<T>::downward_message ( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin, const Matrix2 & w )
{
  FETCH_MUR;
  BLITZ_INDICES;
  int n1 = lrin.extent(0), n2 = lrin.extent(1);
  for (int i = 0; i < n1; ++i)
    for (int j = 0; j < n2; ++j)
	  {
		float s = w(i,j);
		float o = ( T(1) - s ) * T(0.5);
#ifdef HAVE_SSE
		sse_log_mvp_accum_madd    ( &lrin( i, j, 0 ), Base::_m, rout.data(), Base::_k, &mur   ( i, j, 0, 0 ), s, o );
		sse_neg_log_mvp_accum_madd( &lrin( i, j, 0 ), Base::_m, rout.data(), Base::_k, &notmur( i, j, 0, 0 ), s, o );
#else
		int n3 = lrin.extent(2);
		for (int m = 0; m < n3; ++m)
		  {
		     T p_on =  sum(    mur( i, j, m, _all ) * rout );
		     T p_off = sum( notmur( i, j, m, _all ) * rout );
		     lrin( i, j, m ) += safe_log( s * p_on + o ) - safe_log( s * p_off + o );
		  }
#endif
	  }
}

template<typename T>
void 
BinomialMixtureModel2<T>::downward_message ( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin )
{
  FETCH_MUR;
  BLITZ_INDICES;
  int n1 = lrin.extent(0), n2 = lrin.extent(1);
  for (int i = 0; i < n1; ++i)
    for (int j = 0; j < n2; ++j)
      {
#ifdef HAVE_SSE
        sse_log_mvp_accum( &lrin( i, j, 0 ), Base::_m, rout.data(), Base::_k, &mur( i, j, 0, 0 ) );
        sse_neg_log_mvp_accum( &lrin( i, j, 0 ), Base::_m, rout.data(), Base::_k, &notmur( i, j, 0, 0 ) );
#else
        int n3 = lrin.extent(2);
        for (int m = 0; m < n3; ++m)
          {
            T p_on =  sum(    mur( i, j, m, _all ) * rout );
            T p_off = sum( notmur( i, j, m, _all ) * rout );
            lrin( i, j, m ) += safe_log( p_on ) - safe_log( p_off );
          }
#endif
      }
}


template<typename T>
void 
BinomialMixtureModel2<T>::self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin, const Matrix2 & w )
{
  BLITZ_INDICES;
  int n1 = Base::_d1, n2 = Base::_d2, n3 = Base::_m;
  for (int i = 0; i < n1; ++i)
    for (int j = 0; j < n2; ++j)
	  for (int m = 0; m < n3; ++m)
		{
		  pin( i, j    ) += safe_log( w(i,j) * ( sum( rout(i1) * _mu    ( _all, i, j, m ) ) * rin( i, j, m ) +
												 sum( rout(i1) * _not_mu( _all, i, j, m ) ) * ( T(1) - rin( i, j, m ) ) )
									  + ( T(1) - w(i,j) )
									  );
		}
}

template<typename T>
void 
BinomialMixtureModel2<T>::self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin )
{
  BLITZ_INDICES;
  int n1 = Base::_d1, n2 = Base::_d2, n3 = Base::_m;
  for (int i = 0; i < n1; ++i)
    for (int j = 0; j < n2; ++j)
	  for (int m = 0; m < n3; ++m)
		{
		  pin( i, j    ) += safe_log( sum( rout(i1) * _mu    ( _all, i, j, m ) ) * rin( i, j, m ) +
									  sum( rout(i1) * _not_mu( _all, i, j, m ) ) * ( T(1) - rin( i, j, m ) ) );
		}
}

  /////////////////////////////////////////
 ////// message calculators - end ////////
/////////////////////////////////////////

template<class T> void BinomialMixtureModel2<T>::init_sample ( Matrix3 & in )
{
  in = T(0.5);
}

template<typename T>
void BinomialMixtureModel2<T>::sample_point( int irf, int jrf, int k, Matrix1 & data )
{
  BLITZ_INDICES;
  Matrix4 mur2w( _mur2, true );
  Matrix1 mu( mur2w( irf, jrf, k, _all ) );
  Random<T> & rng = Random<T>::get();
  for ( int m = 0; m < Base::_m; ++m )
	data(m) = rng.sample_binomial() ? T(1) : T(0);
}

template<class T> void BinomialMixtureModel2<T>::sample_one(unsigned int out, Matrix3 & in)
{
  BLITZ_INDICES;
  Matrix3 mu(_mu(out,_all,_all,_all));
  in = mu*in/(mu*in+(1-mu)*(1-in));
}

template<class T> void BinomialMixtureModel2<T>::initialize (bool force)
{
  if ( !Base::initialize(force) ) 
	return;

  OPENMP(single)
  {

  _wx_prior.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);

  Random<T> & rng = Random<T>::get();
  
  T newalpha = _beta0;
  T newbeta = newalpha * (T(1) - _e_mu) / _e_mu;
  
  for (int k  = 0; k  < Base::_k; ++k)
	for (int d1 = 0; d1 < Base::_d1; ++d1)
	  for (int d2 = 0; d2 < Base::_d2; ++d2)
		for (int m  = 0; m  < Base::_m; ++m)
		  {
			_wx_prior( k, d1, d2, m ) = Base::_wnx_prior( k, d1, d2 ) * std::max( T(0.01), std::min( T(0.99), rng.beta( newalpha, newbeta ) ) );
		  }
  
  _wx.reference( _wx_prior.copy() );
  
  std::cout << str(boost::format("binomial component model initialized beta0 = %f")%_beta0) << std::endl;

  recalc_hyperparams();

  }

}

template<typename T>
double BinomialMixtureModel2<T>::dL_beta0 ( double la0, void * vparams )
{
  double * params = (double*)vparams;
  double k = params[0];
  double c = params[1];
  double alpha0 = std::exp(la0);
  try 
	{
	  return boost::math::digamma(k*alpha0) - boost::math::digamma(alpha0) + c;
	}
  catch(std::exception & e)
	{
	  std::cerr << " exception evaluating dl beta0 (bmm2): " << e.what() << " type " << typeid(e).name() << std::endl;
	  gsl_error_flag = true;
	  return 0;
	}
}

template<typename T>
void BinomialMixtureModel2<T>::optimize_beta0 ( double delta, double lalphamin, double lalphamax )
{
  typedef std::pair<T,bool> Result;
  double c = mean(_lmudash+_lnotmudash)/2.0;
  double params[2] = {2.0,c};
  gsl_function F = {&dL_beta0, params};	
  gsl_set_error_handler(&gsl_error_handler);	
  Result r = find_root(&F, lalphamin, lalphamax);
  bool converged = r.second;
  if (!converged) 
	{
	  std::cout << "mmm2: hyperparameter optimization did not converge" << std::endl;
	  return;
	}
  double lbeta0 = safe_log(double(_beta0));
  double lbeta0_opt = r.first;
  lbeta0 = (1-double(delta)) * lbeta0 + double(delta) * lbeta0_opt;
  _beta0 = T(std::exp(lbeta0));
  std::cout << "beta0 set to " << _beta0 << " optimal " << std::exp(lbeta0_opt) << std::endl;
}

template<class T> void BinomialMixtureModel2<T>::update(bool force)
{
  BLITZ_INDICES;
  initialize();
  

  if ( Base::update(force) )
	{

	  OPENMP(single)
	  {

		Matrix1  & mu_b = Base::_background_parameters[0];
		mu_b.resize( Base::_m );
		mu_b = _e_mu;
		
	  _not_mu.resize( Base::_k, Base::_d1, Base::_d2, Base::_m );
	  _mu.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);
	  _lmu.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);
	  _lnotmu.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);

	  Matrix4 psibeta( psi( _beta ) );
	  Matrix4 psibetahat( psi( _betahat4 ) );
	  Matrix4 psinotbeta( psi( _betahat4 - _beta ) );
	  
	  _deltalmu.resize(_beta.shape());
	  _deltalmudash.resize(_beta.shape());
	  _deltamu.resize(_beta.shape());
	  _deltamudash.resize(_beta.shape());
	  _lmudash.resize(_beta.shape());
	  _lnotmudash.resize(_beta.shape());
	  _mudash.resize(_beta.shape());
	  _not_mudash.resize(_beta.shape());
	  _r0dash.resize(Base::_k);	
	  _r0dash3.resize(Base::_k,Base::_d1,Base::_d2);
	  _r0.resize(Base::_k);	
	  _r03.resize(Base::_k,Base::_d1,Base::_d2);
	  
	  _r0dash3 = 0;
	  _r03 = 0;
	  
	  CHECKNAN(_beta,"");
	  CHECKNAN(_betahat4,"");
	  	 
	  _deltapsibeta.reference(Matrix4(psibeta - psinotbeta));
	  _mu = _beta/_betahat4;
     
	  int mu_check = sum( where( _mu > 1, 1, 0 ) );
	  if (mu_check > 0)
	  {
		  std::cout << "mu_check = " << mu_check << "  max_div = " << max(_mu) << std::endl;
		  assert( mu_check == 0 );
	  }
	  _not_mu = T(1) - _mu;
	  _lmu = log( _beta ) - log( _betahat4 );
	  _lnotmu = log( _betahat4 - _beta ) - log( _betahat4 );
	  _lmudash = psibeta - psibetahat;
	  _lnotmudash = psinotbeta - psibetahat;
	  _mudash = exp( _lmudash );
	  _not_mudash = exp( _lnotmudash );
	  _r0dash3 = sum(psinotbeta - psibetahat,i4);
	  _r03 = sum(log(_betahat4 - _beta) - log(_betahat4),i4);
	  _r0dash = sum(sum(_r0dash3,i3),i2);

	  _deltalmu = _lmu - _lnotmu;
	  _deltalmudash = _lmudash - _lnotmudash;
	  _deltamu = _mu - _not_mu;
	  _deltamudash = _mudash - _not_mudash;

	  _r0 = sum(sum(_r03,i3),i2);
	  

	  _mur.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _mudashr.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _deltalmur.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _deltalmudashr.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _deltamur.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _deltamudashr.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _mur = _mu(i4,i1,i2,i3);
	  _mudashr = _mudash(i4,i1,i2,i3);
	  _deltalmur = _deltalmu(i4,i1,i2,i3);
	  _deltalmudashr = _deltalmudash(i4,i1,i2,i3);
	  _deltamur = _deltamu(i4,i1,i2,i3);
	  _deltamudashr = _deltamudash(i4,i1,i2,i3);
	  _not_mur.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _not_mur_inv.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _not_mudashr.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _not_mudashr_inv.resize(Base::_d1,Base::_d2,Base::_m,Base::_k);
	  _not_mur = _not_mu(i4,i1,i2,i3);
	  _not_mur_inv = T(1) / _not_mu(i4,i1,i2,i3);
	  _not_mudashr = _not_mudash(i4,i1,i2,i3);
	  _not_mudashr_inv = T(1) / _not_mudash(i4,i1,i2,i3);

	  CHECKNAN(_r0,"_r0");
	  CHECKNAN(_r0dash,"_r0dash");
	  CHECKNAN(_deltalmu,"_deltalmu");
	  CHECKNAN(_deltapsibeta,"_deltapsibeta");
	  CHECKNAN(_r03,"_r03");
	  CHECKNAN(_r0dash3,"_r0dash3");
	  CHECKNAN(_mu,"_mu");
	  CHECKNAN(_lmu,"_lmu");
	  CHECKNAN(_lnotmu,"_lnotmu");
	  CHECKNAN(_lmudash,"_lmudash");
	  CHECKNAN(_lnotmudash,"_lnotmudash");

	  }
	  
  }
}

template class BinomialMixtureModel2<float>;
