/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef BMM2_HPP
#define BMM2_HPP

#ifdef BZ_DEBUG
using namespace std;
#endif

#include <iostream>
#include <stdexcept>
#include <vector>
#include <set>
#include <cmath>
#include <time.h>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/format.hpp>
#include <boost/serialization/version.hpp>

#undef foreach
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH

#include "MixtureModel2Base.hpp"
#include "serialize_blitz.hpp"
#include "greedy_clusterer.hpp"
#include "BatchNG.hpp"

template<typename T> class BinomialMixtureModel2;
template<typename T> struct TypeTraits< BinomialMixtureModel2<T> > 
{	
  typedef T num_type;
  static const int num_statistics = 1;
};

template<class T> class BinomialMixtureModel2
  : public MixtureModel2Base< BinomialMixtureModel2<T> >,
	public HasProperties< SearisePropertyTypes, BinomialMixtureModel2<T> >
{
  
  typedef MixtureModel2Base< BinomialMixtureModel2<T> > Base;
  typedef HasProperties< SearisePropertyTypes, BinomialMixtureModel2<T> > PBase;
  typedef BinomialMixtureModel2<T> This;
  
public:
  
  typedef T num_type;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;

  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<double,4> Matrix4d;

  typedef blitz::Array<unsigned int,2> IMatrix2;
  
  BinomialMixtureModel2();
  
  static void init_properties();
  
  //  void set_property(const std::string & key, PropertyType value) throw(std::runtime_error);
  
  //  PropertyType get_property(const std::string & key) throw(std::runtime_error);
  
  void sample_point( int irf, int jrf, int k, Matrix1 & data );

  void flush();
  
  void batchlearn_begin_epoch();
  
  Matrix1 batchlearn_end_epoch(Matrix1 f_in = Matrix1());
  
  void collect_statistics(const Matrix3 & data, const Matrix1 & r);
  
  void collect_statistics(const Matrix3 & data, const Matrix1 & r, const Matrix2 & w2);
  
  void collect_statistics(const Matrix3 & data, const Matrix1 & r, const Matrix3 & w3);
  
#define STUB_ERROR throw std::runtime_error(boost::str(boost::format("called unimplemented stub on %s")%typeid(*this).name()));
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> & ) 
  {STUB_ERROR;}
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> &, blitz::Array<T,2> & ) 
  {STUB_ERROR;}
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> &, Matrix3 & ) 
  {STUB_ERROR;}
#undef STUB_ERROR

  void upward_message(const Matrix3 & data, Matrix1 & lrout, bool marginal = false);
  
  void upward_message(const Matrix3 & data, Matrix1 & lrout, const Matrix2 & w, bool marginal = false );
  
  void downward_message( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin, const Matrix2 & w);
  
  void downward_message( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin);
 
  void init_sample(Matrix3 & in);
  
  void sample_one(unsigned int out, Matrix3 & in);
  
  void self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin, const Matrix2 & w );

  void self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin );

#if 0
  
  std::vector<int> prune(T threshold);
  
#endif
  
  void reset_input_dimensions(const std::set<unsigned int> & components);
  
  void reset_components(const std::set<unsigned int> & components);
  
  void initialize(bool force=false);
  
  void update(bool force=false);
  
  void initialize_downward_message( Matrix3 & msg, bool linear = false )
  {
	msg = 0;
  }

  void finalize_downward_message( Matrix3 & msg, bool linear = false )
  {
#ifdef HAVE_SSE
	int n = msg.numElements();
	int n4 = n / 4, i = 0;
	__m128 one = _mm_set1_ps( 1.0f );
	if ( linear )
	  {
		for ( int i4 = 0; i4 < n4; ++i4, i += 4 )
		  _mm_storeu_ps( msg.data() + i, _mm_rcp_ps( _mm_add_ps( one, nexpf4( _mm_loadu_ps( msg.data() + i ) ) ) ) );
		for ( ; i < n; ++i )
		  msg.data()[i] = 1.0f / ( 1.0f + std::exp( -msg.data()[i] ) );
	  }
	else
	  {
		for ( int i4 = 0; i4 < n4; ++i4, i += 4 )
		  _mm_storeu_ps( msg.data() + i, nlogf4( _mm_add_ps( one, nexpf4( _mm_loadu_ps( msg.data() + i ) ) ) ) );
		for ( ; i < n; ++i )
		  msg.data()[i] = T(1) / ( T(1) + std::exp( -msg.data()[i] ) );
	  }
#else
	if ( linear )
	  msg = T(1) / ( 1 + exp( -msg ) );
	else
	  msg = -log( T(1) + exp( -msg ) );
#endif
  }

  bool ready() const { return _pre_clustering_mode == NONE || Base::_initialized; }

  void recalc_hyperparams();

  void decay_statistics();

  void recalc_likelihoods();

private:
  
  static double dL_beta0 (double lb0, void * params);
  
  void optimize_beta0(double delta = 1, double lalphamin = -5, double lalphamax = 15);
  
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    BLITZ_INDICES;
	ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(Base);
    ar & BOOST_SERIALIZATION_NVP(_beta0);
	if ( version < 2 )
	  {
		bool _initialized; // dummy for skipping
		ar & BOOST_SERIALIZATION_NVP(_initialized);
	  }
	if ( version > 3 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_wx_prior);
	  }
	else
	  {
		_wx_prior.resize( Base::_k, Base::_d1, Base::_d2, Base::_m );
		_wx_prior = 0;
	  }
    if (Base::_initialized)
	  {
		if (version < 1)
		  {
			Matrix3 _betahat;
			ar & BOOST_SERIALIZATION_NVP(_betahat);
			_betahat4.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);
			for (int m = 0; m < Base::_m; ++m)
			  _betahat4(_all,_all,_all,m) = _betahat;
		  }
		else
		  {
			ar & BOOST_SERIALIZATION_NVP(_betahat4);
		  }
		ar & BOOST_SERIALIZATION_NVP(_beta);
		if ( version > 2 )
		  {
			ar & BOOST_SERIALIZATION_NVP( _greedy_clusterer );
			ar & BOOST_SERIALIZATION_NVP( _pre_clustering_mode );
		  }
	  }
	Base::dirty();
  }
  
  Matrix1 _r0dash, _r0;
  Matrix4 _beta, _mu, _not_mu, _mudash, _not_mudash, _lmu, _lnotmu, _lmudash, _lnotmudash, _deltapsibeta, _deltalmu, _deltalmudash;
  Matrix4 _deltalmur, _deltalmudashr, _not_mur_inv , _not_mudashr_inv, _deltamu, _deltamudash, _deltamur, _deltamudashr;
  Matrix4 _mur, _not_mur, _mudashr, _not_mudashr, _mur2;
  Matrix4d _wx, _wx_prior;
  Matrix4 _betahat4;
  T _beta0;
  T _e_mu;
  T _Lmu;
  Matrix3 _r0dash3, _r03;
  int _learn_beta0_interval, _learn_beta0_counter;

  BatchNG< T, manhattan_metric<T> > _batchng;
  greedy_clusterer< T, manhattan_metric<T> > _greedy_clusterer;
  
  typedef enum { NONE, BATCHNG, GREEDY } PreClusteringMode;
  PreClusteringMode _pre_clustering_mode;

  blitz::firstIndex i1;
  blitz::secondIndex i2;
  blitz::thirdIndex i3;
  blitz::fourthIndex i4;
  blitz::fifthIndex i5;
  
};

//templated class versioning
namespace boost { namespace serialization {
	template<typename T>
	struct version< BinomialMixtureModel2<T> >
	{
	  BOOST_STATIC_CONSTANT(unsigned int, value = 4);
	};
  }}

#endif
