/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "gfm.hpp"
#include "blitz_indices.hpp"
#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <boost/math/tools/polynomial.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <iomanip>

#include "beta_moment_matching.hpp"

#include "chebychev.hpp"

#include "enum_properties.hpp"

#undef HAVE_SSE

#ifdef HAVE_SSE
#include "sse_templates.hpp"
#endif

#define EXCITE_ONLY 0

using boost::math::tools::polynomial;

#ifdef _WIN32
#define throw_not_implemented												\
  throw std::runtime_error( str( boost::format( "%s(%i): method not implented for mcmm2" ) \
								 % __FILE__ % __LINE__ ) )
#else
#define throw_not_implemented												\
  throw std::runtime_error( str( boost::format( "%s(%i): method \"%s\" not implented for mcmm2" ) \
								 % __FILE__ % __LINE__ % __func__ ) )
#endif

/// k is the output space cardinality
/// d is the receptive field size
/// m is the input space cardinality
/// step is the stepping between neighboring receptive fields
/// for the mcmm2 probably d and step need to be 1
template< typename T >
GatedFactorModel<T>::GatedFactorModel( )
  : _n0(0.5)
  , _mu0(0.1)
  , _w0(1)
  , _w0_anneal(1)
  , _p0(0)
  , _in_epsilon(0.2)
  , _ratio_fb(true)	
  , _sequential_update(false)
  , _report_convergence(false)
  , _convergence_limit(10)
  , _convergence_rate(1e-2)
  , _posterior_update_mode(EXACT)
{
}
 
/// this is where to register all parameters accessible in json files
template< typename T >
void GatedFactorModel<T>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter( &Base::_initialized, false );
  PBase::add_ivar_property_ro( "Lmu",         &This::_Lmu      );
  PBase::add_ivar_property_rw( "ratio_fb",    &This::_ratio_fb );
  PBase::add_property(         "mu",          make_blitz_ivar_getter1( &This::_mu ), make_blitz_ivar_setter1( &This::_mu ) );
  PBase::add_property(         "N" ,          make_blitz_ivar_getter1( &This::_N  ), make_blitz_ivar_setter1( &This::_N  ) );
  PBase::add_ivar_property_rw( "N0"      ,    &This::_n0      , uninitializer );
  PBase::add_ivar_property_rw( "mu0"     ,    &This::_mu0     , uninitializer );
  PBase::add_ivar_property_rw( "p0"      ,    &This::_p0       );
  PBase::add_ivar_property_rw( "in_epsilon",  &This::_in_epsilon );
  PBase::add_ivar_property_rw( "w0"      ,    &This::_w0       );
  PBase::add_ivar_property_rw( "w0_anneal",   &This::_w0_anneal);
  PBase::add_ivar_property_rw( "sequential_update",   &This::_sequential_update );
  PBase::add_ivar_property_rw( "report_convergence",   &This::_report_convergence );
  PBase::add_ivar_property_rw( "convergence_rate",   &This::_convergence_rate );
  PBase::add_ivar_property_rw( "convergence_limit",   &This::_convergence_limit );
  enum_property_access<SearisePropertyTypes, This, posterior_update_mode_t> 
	posterior_update_accessor( &This::_posterior_update_mode );
  posterior_update_accessor
	.add_choice( "exact"     , EXACT      )
	.add_choice( "sequential", SEQUENTIAL );
  PBase::add_property( "posterior_update", posterior_update_accessor, posterior_update_accessor );
}

/// upon a call to this the model should forget all it learned and start from scratch
template< typename T >
void GatedFactorModel<T>::flush()
{
  throw_not_implemented;
}

/// here the model should return wether it is ready to produce output
/// i.e. you can return false here if you are running a possibly
/// iterative initialization which needs to see all the data
/// for one or more epochs
template< typename T >
bool GatedFactorModel<T>::ready()
{
  return true;
}

/// not currently used, but here to support pruning
template< typename T >
void GatedFactorModel<T>::reset_input_dimensions(const std::set<unsigned int> & components)
{
  throw_not_implemented;
}

/// not currently used, but here to support pruning
template< typename T >
void GatedFactorModel<T>::reset_components(const std::set<unsigned int> & components)
{
  throw_not_implemented;
}

template<typename T>
void GatedFactorModel<T>::break_sequence()
{
}

/// as the name says: notifies begin of a batch epoch
template< typename T >
void GatedFactorModel<T>::batchlearn_begin_epoch()
{
  Base::batchlearn_begin_epoch();
}

///////////////////////////////////////////////////////////////////////////////////
///// messages: these define the model, they are made up of the conditional //////
///// probabilities of each set of variables given others                  //////
////////////////////////////////////////////////////////////////////////////////

template< typename T >
void GatedFactorModel<T>::calc_upward_msg( IMatrix2 & data )
{
	throw_not_implemented;
}

template< typename T >
void GatedFactorModel<T>::update_s( Matrix3 & data, Matrix3 r )
{
	throw_not_implemented;
}

template< typename T >
void GatedFactorModel<T>::update_s( IMatrix2 & data, Matrix3 r )
{
	throw_not_implemented;
}


template<typename T>
T &  tree_multiply( std::vector<T> & v )
{
  // multiply in a binary tree
  for ( size_t n = v.size(); n > 1; n = ( n + 1 ) / 2 )
	{
	  for ( size_t i = 0, n2 = ( n + 1 ) / 2; i < n2; ++i )
		if ( i + n2 < n )
		  {
			v[i] *= v[n2+i];
		  }
	}
  return v[0];
}

template< typename T >
void GatedFactorModel<T>::match_exact_posterior( const Matrix3 & data )
{

  /// get some bounds
  int D1 = Base::_d1;
  int D2 = Base::_d2;
  int M = Base::_m;
  int S1 = Base::_step1;
  int S2 = Base::_step2;
  int K = Base::_k;

  update_lB();

  Matrix4 muw( _mur, true );
  Matrix3 lB( _lB, true );

  int nout1, nout2;
  boost::tie( nout1, nout2) = calc_outsize(data.extent(0),data.extent(1));
  
  Matrix3 dataw( data, true );

  Matrix3 rout( message_component<T>::get_belief() );

  Matrix1 rout1;

  /// iterate rf
  int D12 = D1 * D2;

  OPENMP(for)
	for ( int d = 0; d < D12; ++d )
	  {
		int d1 = d % D1;
		int d2 = d / D1;
		/// iterate i/o features
		for ( int k = 0; k < K; ++k )
		  {
		  for ( int m = 0; m < M; ++m )
			{
			  /// ploynomial coefficients, initialized with constant term == 1, others == 0
			  
			  T mu = muw( d1, d2, k, m );

			  /// iterate outputs
			  std::vector< polynomial<double> > ppolys;
			  std::vector< polynomial<double> > npolys;
			  int nn = 0, np = 0;
			  for ( int i1 = 0; i1 < nout1; ++i1 )
				for ( int i2 = 0; i2 < nout2; ++i2 )
				  {
					T z = rout( i1, i2, k );
					// fetch input bits to this connection
					int j1 = i1 * S1 + d1;
					int j2 = i2 * S2 + d2;
					T x = dataw( j1, j2, m );
					T lBs = lB( j1, j2, m );
					T A =  ( 1 - 2 * x ) * exp( lBs - safe_log( 1 - z * mu ) );
					T beta  = std::max( T(1e-5), x + A );
					T lp = safe_log( x + A * ( 1 - mu ) ) - safe_log( x + A );
					z *= ( 1 + exp(-lp) );
					T alpha = std::max( T(1e-5) - beta, -A * z );
					T factor = T(1) / std::max( beta , alpha + beta );
					beta *= factor;
					alpha *= factor;
					if ( alpha < 0 ) 
					  { 
						double betaalpha[2] = { alpha + beta, -alpha };
						npolys.push_back( polynomial<double>( betaalpha, 1 ) );
						++nn;
					  }
					else
					  { 
						double betaalpha[2] = { beta, alpha };
						ppolys.push_back( polynomial<double>( betaalpha, 1 ) );
						++np;
					  }
				  } // windows

			  double * pcoeffs_begin = 0, * pcoeffs_end = 0;
			  double * ncoeffs_begin = 0, * ncoeffs_end = 0;
			  if ( !ppolys.empty() )
				{
				  polynomial<double> & pcoeffs = tree_multiply( ppolys );
				  pcoeffs_begin = &pcoeffs[0];
				  pcoeffs_end   = &pcoeffs[0] + pcoeffs.size();
				}
			  if ( !npolys.empty() )
				{
				  polynomial<double> & ncoeffs = tree_multiply( npolys );
				  ncoeffs_begin = &ncoeffs[0];
				  ncoeffs_end   = &ncoeffs[0] + ncoeffs.size();
				}
			  update_beta_simple_with_polynomial_likelihood( pcoeffs_begin, pcoeffs_end
															 , ncoeffs_begin, ncoeffs_end
															 , _mu( k, d1, d2, m )
															 , _N( k, d1, d2, m ) 
															 , Base::_delta );
			  
			} // m
		  //std::cout << std::endl;
		  } // k
		//std::cout << std::endl;
	  } // d
  Base::dirty();
}

template< typename T >
void GatedFactorModel<T>::apply_r( const Matrix3 & data )
{
  Base::apply_r( data, false );

  switch ( _posterior_update_mode )
  {
  case EXACT:
	match_exact_posterior( data );
	break;
  case SEQUENTIAL:
	sequentially_match_posterior( data );
	break;
  }

  _mu = max( T(0.0001), min( T(0.9999), _mu ) );

  Base::_Lp += calc_data_likelihood( data );
}


#if 0
__m128 lB2A4( __m128 x4, __m128 lB4, __m128 z4, __m128 mu4 )
{
  __m128 one = _mm_set1_ps( 1 );
  __m128 two = _mm_set1_ps( 2 );
  __m128 A4 = _mm_mul_ps( _mm_sub_ps
						  ( one, _mm_mul_ps( two, x4 ) ),
						  expf4
						  ( _mm_sub_ps
							( lB4, logf4
							  ( _mm_sub_ps
								( one, _mm_mul_ps
								  ( z4, mu4 ) 
								  ) 
								) 
							  ) 
							) 
						  );
  return A4;
}

__m128 sub_A4_from_z4( __m128 x4, __m128 A4, __m128 z4, __m128 mu4 )
{
  __m128 one  = _mm_set1_ps( 1 );
  __m128 zero = _mm_set1_ps( 0 );
  __m128 mlp4 = _mm_sub_ps( logf4( _mm_add_ps( x4, A4 ) ),
							logf4( _mm_add_ps
								   ( x4, _mm_mul_ps
									 ( A4, _mm_sub_ps( one, mu4 ) ) 
									 )
								   )
							);
  z4 = _mm_max_ps( zero, _mm_min_ps( one, _mm_mul_ps
									 ( z4, _mm_add_ps( one, expf4( mlp4 ) ) 
									   ) ) );
  return z4;
}

__m128 sub_lB_from_z4( __m128 x4, __m128 lB4, __m128 z4, __m128 zb4, __m128 mu4 )
{
  __m128 A4 = lB2A4( x4, lB4, z4, mu4 );
  return sub_A4_from_z4( x4, A4, z4, mu4 );
}

#endif


template< typename T >
void GatedFactorModel<T>::sequentially_match_posterior( const Matrix3 & data )
{

  /// get some bounds
  int D1 = Base::_d1;
  int D2 = Base::_d2;
  int M = Base::_m;
  int S1 = Base::_step1;
  int S2 = Base::_step2;
  int K = Base::_k;

  update_lB();

  Matrix4 muw( _mur2, true );

  int nout1, nout2;
  boost::tie( nout1, nout2) = calc_outsize(data.extent(0),data.extent(1));
  
  Matrix3 dataw( data, true );

  Matrix3 rout( message_component<T>::get_belief() );

  bool have_saliency = Base::_S.numElements() > 0;

   /// iterate rf
  int D12 = D1 * D2;

  OPENMP(for)
	/// iterate outputs
	for ( int i1 = 0; i1 < nout1; ++i1 )
	  for ( int i2 = 0; i2 < nout2; ++i2 )
		for ( int d = 0; d < D12; ++d )
		  {
			int d1 = d % D1;
			int d2 = d / D1;
			int j1 = i1 * S1 + d1;
			int j2 = i2 * S2 + d2;
			/// iterate i/o features
			for ( int m = 0; m < M; ++m )
			  {
				const T lB = _lB( j1, j2, m );
				const T x = _in_epsilon + ( 1 - _in_epsilon ) * dataw( j1, j2, m );
				const T delta = have_saliency ? Base::_delta * Base::_S( j1, j2 ) : Base::_delta;
				int k = 0;
				for ( ; k < K; ++k )
				  {
					T mu_o = muw( d1, d2, m, k );
					T z = rout( i1, i2, k );
#if EXCITE_ONLY
					T A = x * std::exp( std::min( T(1), lB - safe_log( 1 - z * mu_o ) ) );
					T beta  = std::max( T(1e-3), T(0.5) * ( T(1) + x ) - A );
					T alpha = std::max( T(1e-3) - beta, A * z * delta );
#else
					T A = ( 1 - 2 * x ) * std::max( T(-1), std::min( T(1),  exp( lB - safe_log( 1 - z * mu_o ) ) ) );
					T beta  = std::max( T(1e-3), x + A );
					T alpha = std::max( T(1e-3) - beta, -A * z * delta );
#endif
					gammaApprox::approxBetaUpdateVar( alpha, beta 
													  , _mu( k, d1, d2, m )
													  , _N( k, d1, d2, m ) );
				  } // k
			  } // m
		  } // d
  
  Base::dirty();
}


template< typename T >
void GatedFactorModel<T>::calc_upward_msg( Matrix3 & data )
{

  _convergence_monitors.clear();

  size_t iterations = 0;

  T data_likelihood = 0;

  int n1 = data.extent(0);
  int n2 = data.extent(1);
  int K = 1 + ( n1 - Base::_d1 ) / Base::_step1;
  int L = 1 + ( n2 - Base::_d2 ) / Base::_step2;

  OPENMP(single)
  {
	_data.reference( data );
	update();
	blitz::TinyVector<int,3> s;
	boost::tie(s[0],s[1]) = calc_outsize(data.extent(0),data.extent(1));
	s[2] = Base::_k;
	if ( message_component<T>::get_msg().empty() )
	  {
		message_component<T>::get_msg().init( s, mmm2_belief_state<T>::LOG );
		message_component<T>::buffer().touch();
	  }
  }  

  update_lB( K, L );  

  if ( _report_convergence )
	{
	  std::cout << std::endl;
	  data_likelihood = calc_data_likelihood( data );
	  std::cout << " gfm iteration 0 L " << data_likelihood << std::endl;
	}

  do
	{
	  if ( int( iterations ) > _convergence_limit )
		throw std::runtime_error( str( boost::format( "gfm bp failed to converge after %i iterations" )
									   % iterations ) );
	  if ( _sequential_update )
		recalc_upward_msg_sequential( data, ( Base::_learning ? _w0 : T(1) ) );
	  else
		recalc_upward_msg( data, ( Base::_learning ? _w0 : T(1) ) );

	  ++iterations;

	  if ( _report_convergence )
		{
		  T L = calc_data_likelihood( data );
		  Matrix3 msg( message_component<T>::get_msg().data, true );
		  std::cout << " gfm iteration " << iterations << " L " << L ;
		  std::cout << " dL: " << (L - data_likelihood) ;
		  std::cout << " max(|dz|) " << accum::max( _convergence_monitors.back() ); 
		  std::cout << " mean(|dz|) " << accum::mean( _convergence_monitors.back() ); 
		  std::cout << " median(|dz|) " << accum::median( _convergence_monitors.back() ); 
		  std::cout << " msg: " << min( msg ) << ", " << mean( msg ) << ", " << max( msg );
		  std::cout << std::endl;
		  data_likelihood = L;
		}

	  _had_upward_pass = true;
	}
  while( accum::mean( _convergence_monitors.back() ) > _convergence_rate * _w0 );
  
  if ( _report_convergence )
	std::cout << " gfm took " << iterations << " to converge " << std::endl;

  /*
  std::cout << std::fixed << std::setprecision(5);

  std::cout << " gfm convergence stats: " << std::endl;
  std::cout	<< "  min:    ";
  BOOST_FOREACH( const convergence_monitor_t & acc, _convergence_monitors )
	std::cout << accum::min(acc)  << " ";
  std::cout << std::endl;
  std::cout	<< "  mean:    ";
  BOOST_FOREACH( const convergence_monitor_t & acc, _convergence_monitors )
	std::cout << accum::mean(acc)  << " ";
  std::cout << std::endl;
  std::cout	<< "  median:  ";
  BOOST_FOREACH( const convergence_monitor_t & acc, _convergence_monitors )
	std::cout << accum::median(acc)  << " ";
  std::cout << std::endl;
  std::cout	<< "  max:     ";
  BOOST_FOREACH( const convergence_monitor_t & acc, _convergence_monitors )
	std::cout << accum::max(acc)  << " ";
  std::cout << std::endl;
  */

}

template< typename T >
void GatedFactorModel<T>::recalc_upward_msg_sequential( Matrix3 & data, T rate )
{

  Matrix3 rout( message_component<T>::get_belief() );
  Matrix3 msg( message_component<T>::get_msg().data, true );

  /// get some bounds
  int D1 = Base::_d1;
  int D2 = Base::_d2;
  int M = Base::_m;
  int S1 = Base::_step1;
  int S2 = Base::_step2;
  int K = Base::_k;
  //int np = 2 * ( D + S - 1 ) / S * K;

  convergence_monitor_t acc;

  for ( ReceptiveFieldIterator<Matrix3,Matrix3> i( data, msg, D1, D2, S1, S2 ); !i.atend(); ++i )
	{
	  int xoff = i.out_x() * S1;
	  int yoff = i.out_y() * S2;
	  int k = 0;
	  for ( ; k < K; ++k )
		{
		  T & z_out = rout( i.out_x(), i.out_y(), k );
		  const T zold =  z_out;
		  const T old_msg = i.out()(k);
#ifdef HAVE_SSE
		  vecf4 zold4( zold );
		  vecf4 new_msg4( 0.0f );
#endif
		  T new_msg = 0;
		  for ( int i1 = 0; i1 < D1; ++i1 )
			{
			  int i2 = 0, m = 0;
#ifdef HAVE_SSE
			  T * px  = &i.in()( i1, 0, 0 );
			  T * plB = &_lB( xoff + i1, yoff, 0 );
			  T * pmu = &_mu( k, i1, 0, 0 );
			  int i2m = 0;
			  for ( int i2m4 = 0; i2m4 < ( D2 * M ) / 4; i2m += 4, ++i2m4 )
				{
				  const vecf4 x4( _in_epsilon + ( 1.0f - _in_epsilon ) * vecf4(  px + i2m ) ) ) );
				  const vecf4 lB4( plB + i2m );
				  const vecf4 mu4( pmu + i2m );
				  const vecf4 A4 = ( 1.0f - 2.0f * x4 ) *
					exp( min( 0.0f, lB4 - safe_log( 1.0f - zold4 * mu4 ) ) );
				  new_msg4 += safe_log( x4 + A4 * ( 1.0f - mu4 ) ) - safe_log( x4 + A4 );
				}
			  new_msg += new_msg4.sum();
			  i2 = i2m / M;
			  m = i2m % M;
#endif
			  for ( ; i2 < D2; ++i2, m = 0 )
				for ( ; m < M; ++m )
				  {
					T x = _in_epsilon + ( T(1) - _in_epsilon ) * i.in()( i1, i2, m );
					T mu = _mu( k, i1, i2, m );
					T lB = _lB( xoff + i1, yoff + i2, m );
					T A = ( 1 - 2 * x ) * std::exp( std::min( T(0), lB - safe_log( 1 - zold * mu ) ) );
					new_msg += safe_log( x + A * ( 1 - mu ) ) - safe_log( x + A );
				  }
			}
		  // update message and variational distribution
		  if ( rate < T(1) )
			new_msg = old_msg + rate * ( new_msg - old_msg );
		  i.out()(k) = new_msg;
		  const T lzold = safe_log( zold ) - safe_log( T(1) - zold );
		  const T lznew = lzold + new_msg - old_msg;
		  const T znew = T(1) / ( T(1) + std::exp( -lznew ) );
		  z_out = znew;
		  // update "lB"
#ifdef HAVE_SSE
		  vecf4 znew4( znew );
#endif
		  acc( std::abs( znew - zold ) );
		  for ( int i1 = 0; i1 < D1; ++i1 )
			{
			  int i2 = 0, m = 0;
#ifdef HAVE_SSE
			  T * plB = &_lB( xoff + i1, yoff, 0 );
			  T * pmu = &_mu( k, i1, 0, 0 );
			  int i2m = 0;
			  for ( int i2m4 = 0; i2m4 < ( D2 * M ) / 4; i2m += 4, ++i2m4 )
				{
				  const vecf4 mu4( pmu + i2m );
				  vecf4 lB4 = vecf4( plB + i2m )
					+ safe_log( 1.0f - znew4 * mu4 )
					- safe_log( 1.0f - zold4 * mu4 );
				  lB4.store( plB + i2m );
				}
			  i2 = i2m / M;
			  m = i2m % M;
#endif
			  for ( ; i2 < D2; ++i2, m = 0 )
			  for ( ; m < M; ++m )
				{

				  T mu = _mu( k, i1, i2, m );				  
				  _lB( xoff + i1, yoff + i2, m ) += 
					safe_log( 1 - znew * mu ) - 
					safe_log( 1 - zold * mu );
				}
			}
		}
	}
  
  _convergence_monitors.push_back( acc );

}


template< typename T >
void GatedFactorModel<T>::recalc_upward_msg( Matrix3 & data, T rate )
{
  
  Matrix3 rout( message_component<T>::get_belief() );
  Matrix3 msg( message_component<T>::get_msg().data, true );

  /// get some bounds
  int D1 = Base::_d1;
  int D2 = Base::_d2;
  int M = Base::_m;
  int S1 = Base::_step1;
  int S2 = Base::_step2;
  int K = Base::_k;

  Matrix3 dataw( data, true );

  Matrix1 routkl;

  convergence_monitor_t acc;

  for ( ReceptiveFieldIterator<Matrix3,Matrix3> i( data, msg, D1, D2, S1, S2 ); !i.atend(); ++i )
	{
	  routkl.reference( rout( i.out_x(), i.out_y(), blitz::Range::all() ) );
	  int xoff = i.out_x() * S1;
	  int yoff = i.out_y() * S2;
	  for ( int i1 = 0; i1 < D1; ++i1 )
		for ( int i2 = 0; i2 < D2; ++i2 )
		  {
			for ( int m = 0; m < M; ++m )
			  {
				T x = i.in()( i1, i2, m );
				T lB = _lB( xoff + i1, yoff + i2, m );
#if 0//def HAVE_SSE
				__m128 x4  = _mm_set1_ps( x  );
				__m128 lB4 = _mm_set1_ps( lB );
				__m128 one = _mm_set1_ps( 1.0f );
#endif
				int k = 0;
#if 0//def HAVE_SSE
				for ( int k4 = 0, k4max = K / 4; k4 < k4max; ++k4, k += 4 )
				  {
					__m128 mu4 = _mm_loadu_ps( &_mur2( i1, i2, m, k) );
					__m128 z4 = _mm_loadu_ps( &routkl( k ) );
					__m128 A4 = lB2A4( x4, lB4, z4, mu4 );
					_mm_storeu_ps( &i.out()(k), _mm_add_ps
									( _mm_loadu_ps( &i.out()(k) ), _mm_mul_ps
									  ( _mm_set1_ps( rate ), _mm_sub_ps
										( logf4
										  ( _mm_add_ps
											( x4, _mm_mul_ps
											  ( A4, _mm_sub_ps( one, mu4 ) ) 
											  )
											), 
										  logf4( _mm_add_ps( x4, A4 ) )
										  )
										)
									  )
								   );
				  }
#endif
				for ( ; k < K; ++k )
				  {
					T mu = _mur2( i1, i2, m, k);
					T z = routkl( k );
					T A =  std::max( T(-1), std::min( T(1), ( 1 - 2 * x ) * exp( lB - safe_log( 1 - z * mu ) ) ) );
					if ( A < -1 || A > 1 ) 
					  std::cout << " A out of range in upward message: " << A << std::endl;
					T old_msg = i.out()(k);
					T new_msg = ( safe_log( x + A * ( 1 - mu ) ) - safe_log( x + A ) );
					i.out()(k) += rate * ( new_msg - i.out()(k) );
					const T lzold = safe_log( z ) - safe_log( T(1) - z );
					const T lznew = lzold + i.out()(k) - old_msg;
					const T znew = T(1) / ( T(1) + std::exp( -lznew ) );
					acc( znew - z );
					routkl( k ) = znew;
				  }
			  }
		  }

	}

  _convergence_monitors.push_back( acc );

  //  log().info << " upward message in [" << min( msg ) << ", " << max( msg ) << "]" << std::endl;
}

template< typename T >
typename GatedFactorModel<T>::num_type GatedFactorModel<T>::calc_data_likelihood( const Matrix3 & data )
{
  BLITZ_INDICES;
  Matrix3 X( data( _lB.domain() ) );
  T L = sum( safe_log( X + ( T(1) - T(2) * X ) * exp(_lB) ) );  
  return L;
}

template< typename T >
blitz::Array<typename GatedFactorModel<T>::num_type,2> GatedFactorModel<T>::calc_confidence( const Matrix3 & data )
{
  BLITZ_INDICES;
  Matrix3 X( data( _lB.domain() ) );
  Matrix2 confidence( sum( safe_log( X + ( T(1) - T(2) * X ) * exp(_lB) ), i3 ) );
  return confidence;
}

template< typename T >
void GatedFactorModel<T>::update_lB( int K, int L )
{
  
  BLITZ_INDICES;

  Matrix3 rout( message_component<T>::get_belief() );
  Matrix4 mur2( _mur2, true );

  const int M = Base::_m;
  const int S1 = Base::_step1;
  const int S2 = Base::_step2;
  const int D1 = Base::_d1;
  const int D2 = Base::_d2;
  if ( K < 0 || L < 0 )
	{
	  K = rout.extent(0);
	  L = rout.extent(1);
	}
  const int I = D1 + ( K - 1 ) * S1;
  const int J = D2 + ( L - 1 ) * S2;

  _lB.resize( I, J, M );

  Matrix1 routkl;
  
  /// log-probability of the input bits being off given the code bits
  for ( int i = 0; i < I; ++i )
	for ( int j = 0; j < J; ++j )
	  for ( int m = 0; m < M; ++m )
		{
		  T & lB = _lB( i, j, m );
		  lB = safe_log( 1 - _p0 );
		  for ( int k = std::max( 0, i - D1 + S1 ) / S1, kmax = std::min( i / S1, K - 1 ); k <= kmax; ++k )
			for ( int l = std::max( 0, j - D2 + S2 ) / S2, lmax = std::min( j / S2, L - 1 ); l <= lmax; ++l )
			  {
				routkl.reference( rout( k, l, _all ) );
				int di = i - k * S1;
				int dj = j - l * S2;
				int kk = 0;
#ifdef HAVE_SSE
				vecf4 lB4( 0.0f );
				for ( int k4 = 0, k4max = Base::_k / 4; k4 < k4max; ++k4, kk += 4 )
				  {
					vecf4 mu4( &mur2( di, dj, m, kk ) );
					vecf4 z4 = vecf4( &routkl( kk ) );
					lB4 += log( 1.0f - mu4 * z4 );
				  }
				lB += lB4.sum();
#endif
				for ( ; kk < Base::_k; ++kk )
				  {
					T mu = mur2( di, dj, m, kk );
					T z = routkl( kk );
					lB += safe_log( 1 - mu * z );
				  }
			  }
		}
  

}

template< typename T >
void GatedFactorModel<T>::calc_downward_msg( Matrix3 data, bool logarithmic )
{
  BLITZ_INDICES;
  
  //update_lB( );
  OPENMP(single)
  {
	Base::_downward_message.resize( _lB.shape() );
  }
  Matrix3 msg( Base::_downward_message, true );
  msg = T(1) - exp( _lB );
  if ( logarithmic )
	msg = safe_log( msg ) - safe_log( T(1) - msg );

  /*
  log().info << " lB in [" << min( _lB ) << ", " << max( _lB ) << "]" << std::endl;
  log().info << " downward msg in [" << min( msg ) << ", " << max( msg ) << "]" << std::endl;
  */
}

/// here the model should calculate the probability of the latent (output)
/// variable given data (usually probabilities of the input variables)
/// the data is an input receptive field view: (rows x columns x input-depth)
/// the output is a single output "pixel"
/// the output neds to be stored logarithmically
/// variational is for MAP estimation (different parameter set)
template< typename T >
void GatedFactorModel<T>::begin_mp_frame()
{
  Base::begin_mp_frame();  
  message_component<T>::buffer().set_type( mmm2_belief_state<T>::BINOMIAL );
  _had_upward_pass = false;
  if ( Base::_learning )
	decay_statistics();
}


template< typename T >
void GatedFactorModel<T>::end_mp_frame( const Matrix3 & data_ )
{
  Base::end_mp_frame( data_ );
}

template< typename T >
void GatedFactorModel<T>::decay_statistics()
{
  typename Matrix4::iterator imu = _mu.begin();
  typename Matrix4d::iterator iN = _N.begin(), iNend = _N.end();
  typename Matrix4::iterator imu_prior = _mu_prior.begin();
  typename Matrix4d::iterator iN_prior = _N_prior.begin();
  double decay  = std::exp( -Base::_delta / Base::_tau );
  for ( ; iN != iNend; ++iN, ++imu, ++iN_prior, ++imu_prior )
	{
	  double alpha0 =          *imu_prior   * *iN_prior;
	  double beta0  = ( T(1) - *imu_prior ) * *iN_prior;
	  double & N = *iN;
	  T & mu = *imu;
	  double dalpha = decay * (       mu   * N - alpha0 );
	  double dbeta  = decay * ( ( 1 - mu ) * N - beta0  );
	  double alpha = alpha0 + dalpha;
	  double beta  = beta0  + dbeta;
	  mu = alpha / ( alpha + beta );
	  N = alpha + beta;
	}
}

/// this gets called at the end of a e-step epoch
/// f_in is a reweighing factor for the input space
/// and should be applied to any histograms
/// similarly you can return a renormalization factor
/// for the output space which will be propagated
template< typename T >
blitz::Array<typename GatedFactorModel<T>::num_type,1> GatedFactorModel<T>::batchlearn_end_epoch(blitz::Array<T,1> f_in )
{

  BLITZ_INDICES;

  Matrix1 f = Base::batchlearn_end_epoch();
    
  Base::_L = Base::_Lp;
  
  log().info << str( boost::format("gated boolean component model ended epoch Lmu = %1.3e Lp = %1.3e L = %1.3e")
					 % _Lmu % Base::_Lp % Base::_L ) << std::endl;
  
  _w0 = std::pow( _w0, _w0_anneal );

  std::cout << " w0 = " << _w0 << std::endl;

  return f;
  
}

/// this is to initialize all (hyper-) parameters
/// called automatically by base
/// you should call Base::initlialize from here also
/// force means re-initialize, even you already did it before
template< typename T >
void GatedFactorModel<T>::initialize( bool force )
{
  BLITZ_INDICES;
  
  if ( Base::initialize(force) )
	{

	  OPENMP(barrier);
	  OPENMP(single)
	  {

		std::cout << " initializing gfm " << std::endl;

		_mu.resize      ( Base::_k, Base::_d1, Base::_d2, Base::_m );	  
		_mu_prior.resize( Base::_k, Base::_d1, Base::_d2, Base::_m );	  
		_N.resize       ( Base::_k, Base::_d1, Base::_d2, Base::_m );	  
		_N_prior.resize ( Base::_k, Base::_d1, Base::_d2, Base::_m );	  
		_N_prior = _n0 + Base::_initcount;
		Random<T> & rng = Random<T>::get();
		T alpha0 = _n0 *       _mu0  ;
		T beta0  = _n0 * ( 1 - _mu0 );

		std::cout << " iterating _mu: " << _mu.shape() << std::endl;

		//		for ( typename Matrix4::iterator imu = _mu.begin(), iend = _mu.end(); imu != iend; ++imu )
		for ( int j1 = 0; j1 < Base::_k; ++j1 )
		for ( int j2 = 0; j2 < Base::_d1; ++j2 )
		for ( int j3 = 0; j3 < Base::_d2; ++j3 )
		for ( int j4 = 0; j4 < Base::_m; ++j4 )
		  {
			//std::cout << imu.position() << std::endl;
			//T & mu = *imu;
			/*
			std::cout << j1 << ", " 
					  << j2 << ", " 
					  << j3 << ", " 
					  << j4 << ", " 
					  << std::endl;
			*/
			T mu = rng.beta( alpha0, beta0 );
			T alpha = Base::_initcount * (        mu ) + alpha0;
			T beta  = Base::_initcount * ( T(1) - mu ) + beta0;
			_mu_prior( j1, j2, j3, j4 ) = alpha / ( alpha + beta );
		  }

		std::cout << " initialized gfm " << std::endl;

	  }
	  _N = _N_prior;
	  _mu = _mu_prior;
	}
}

/// after this call parameters should be ready for operation
/// you should call initialize if you need to from here
/// also you should call Base::update
/// force as above
template< typename T >
void GatedFactorModel<T>::update(bool force )
{
  BLITZ_INDICES;
  initialize();
  if ( Base::update(force) )
	{
	  OPENMP(barrier);
	  OPENMP(single)
	  {
		_mur.resize ( Base::_d1, Base::_d2, Base::_k, Base::_m );
		_mur2.resize( Base::_d1, Base::_d2, Base::_m, Base::_k );
		_mur  = _mu( i3, i1, i2, i4 );
		_mur2 = _mu( i4, i1, i2, i3 );
	  }
	}
}

template class GatedFactorModel<float>;


