/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef GFM_HPP
#define GFM_HPP

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/serialization/serialization.hpp>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/min.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/accumulators/statistics/median.hpp>
#include <boost/accumulators/statistics/max.hpp>
#include <iostream>

#include "MixtureModel2Base.hpp"

#include "serialize_blitz.hpp"

namespace accum = boost::accumulators;

template<typename T> class GatedFactorModel;
template<typename T> struct TypeTraits< GatedFactorModel<T> > 
{	
  typedef T num_type;
  static const int num_statistics = 0;
};

template<typename T> class GatedFactorModel 
  : public MixtureModel2Base< GatedFactorModel<T> >,
	public HasProperties< SearisePropertyTypes, GatedFactorModel<T> >
{
  
  typedef GatedFactorModel<T> This;
  typedef MixtureModel2Base< This > Base;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  
public:
  
  typedef T num_type;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<double,4> Matrix4d;
  typedef blitz::Array<unsigned int,2> IMatrix2;


  /// k is the output space cardinality
  /// d is the receptive field size
  /// m is the input space cardinality
  /// step is the stepping between neighboring receptive fields
  /// for the mcmm2 probably d and step need to be 1
  GatedFactorModel();
  
  /// this is where to register all parameters accessible in json files
  static void init_properties();
  
  /// upon a call to this the model should forget all it learned and start from scratch
  void flush();

  /// here the model should return wether it is ready to produce output
  /// i.e. you can return false here if you are running a possibly
  /// iterative initialization which needs to see all the data
  /// for one or more epochs
  bool ready();
  
  /// not currently used, but here to support pruning
  void reset_input_dimensions(const std::set<unsigned int> & components);
  
  /// not currently used, but here to support pruning
  void reset_components(const std::set<unsigned int> & components);
  
  /// as the name says: notifies begin of a batch epoch
  void batchlearn_begin_epoch();
  
  ///////////////////////////////////////////////////////////////////////////////////
  ///// messages: these define the model, they are made up of the conditional //////
  ///// probabilities of each set of variables given others                  //////
  ////////////////////////////////////////////////////////////////////////////////

  /// override the complete upward message logic because we process full images at a time here
  void calc_upward_msg( Matrix3 & data );
  void recalc_upward_msg( Matrix3 & data, T rate );
  void recalc_upward_msg_sequential( Matrix3 & data, T rate );
  
  /// override the complete upward message logic because we process full images at a time here
  void calc_upward_msg( IMatrix2 & data );

  /// override the complete downward message logic because we process full images at a time here
  void calc_downward_msg ( Matrix3 data = Matrix3(), bool logarithmic = true );

  Matrix2 calc_confidence( const Matrix3 & data );
  T calc_data_likelihood( const Matrix3 & data );

  void sequentially_match_posterior( const Matrix3 & data );
  void match_exact_posterior( const Matrix3 & data );
  void apply_r( const Matrix3 & data );

  void update_s( Matrix3 & data, Matrix3 r );

  void update_s( IMatrix2 & data, Matrix3 r );

  /// inter-frame callbacks
  void begin_mp_frame();
  void end_mp_frame( const Matrix3 & data );

  /// this gets called at the end of a e-step epoch
  /// f_in is a reweighing factor for the input space
  /// and should be applied to any histograms
  /// similarly you can return a renormalization factor
  /// for the output space which will be propagated
  Matrix1 batchlearn_end_epoch(Matrix1 f_in = Matrix1());

  /// this is to initialize all (hyper-) parameters
  /// called automatically by base
  /// you should call Base::initlialize from here also
  /// force means re-initialize, even you already did it before
  void initialize(bool force=false);
  
  /// after this call parameters should be ready for operation
  /// you should call initialize if you need to from here
  /// also you should call Base::update
  /// force as above
  void update(bool force = false);
  
  /// leave this: means that your input space is multinomial
  bool is_multinomial() { return true; }

  void break_sequence();

  void decay_statistics();
  
private:
 
  void update_lB( int K = -1, int L = -1 );

   /// here you can use boost::serialization to save/load your state
  /// this gets called from above when saving/loading all network state
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
	ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(Base);
	ar & BOOST_SERIALIZATION_NVP( _n0       );
	ar & BOOST_SERIALIZATION_NVP( _N        );
	ar & BOOST_SERIALIZATION_NVP( _N_prior  );
	ar & BOOST_SERIALIZATION_NVP( _mu0      );
	ar & BOOST_SERIALIZATION_NVP( _mu       );
	ar & BOOST_SERIALIZATION_NVP( _mu_prior );
	ar & BOOST_SERIALIZATION_NVP( _Lmu      );
	Base::dirty();
  }


  typedef enum { EXACT, SEQUENTIAL } posterior_update_mode_t;

  T _n0, _mu0, _Lmu, _w0, _w0_anneal, _p0, _in_epsilon;
  Matrix4d _N, _N_prior;
  Matrix4 _mu, _mur, _mur2, _mu_prior;
  
  Matrix3 _lB, _data, _zB;

  bool _ratio_fb, _had_upward_pass, _sequential_update, _report_convergence;

  int _n_iters, _convergence_limit;

  T _convergence_rate;
  
  posterior_update_mode_t _posterior_update_mode;

  typedef accum::accumulator_set< T, accum::stats< 
									   accum::tag::mean, 
									   accum::tag::min, 
									   accum::tag::median, 
									   accum::tag::max
									   > 
								  > convergence_monitor_t;

  std::vector<convergence_monitor_t> _convergence_monitors;

};

#endif
