/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef GMM2_HPP
#define GMM2_HPP

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/binary_object.hpp>
#include <iostream>

#include "MixtureModel2Base.hpp"

#include "digamma.hpp"
#include "normrho.hpp"
#include "fullgammaprior.hpp"

//#include "boost_specfun.hpp"

#ifndef FAST_LOG
#define fast_log log
#endif

#ifdef GSL
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#else
#include <random/gamma.h>
#endif

#include <random/uniform.h>

#undef foreach
#define foreach BOOST_FOREACH


#define parameter(name,type)				\
type get_ ## name() const { return _ ## name; } \
  void set_ ## name(type value) { _ ## name = value; }

template<typename T> class GammaMixtureModel2;
template<typename T> struct TypeTraits< GammaMixtureModel2<T> > 
{	
	typedef T num_type;
};

template<typename T> class GammaMixtureModel2 
: public MixtureModel2Base< GammaMixtureModel2<T> >
{

	typedef MixtureModel2Base< GammaMixtureModel2<T> > Base;
	
public:

  typedef T num_type;
			
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<unsigned int,2> IMatrix2;

  GammaMixtureModel2(int k=0, int d=0, int m=0, int step=0)
    : Base(k,d,m,step),
    _beta0(0.5),
		_full_prior(false),
		_baseline(0.01),
		_betal0(0),
		_fgevalmax(100),
		_initialized(false),
		_dirty(false),
		_Lmu(0)

  {
#ifdef GSL
    _rng = gsl_rng_alloc (gsl_rng_taus);
#endif
  }

  const Matrix4 get_wx()  { initialize(); return _wx; }
  const Matrix4 get_sigma() { update(); return Matrix4(_alphag/_betag); }
  const Matrix4 get_mu() { update(); return Matrix4(_alphag/_betag); }

	void set_property(const std::string & key, PropertyType value) throw(std::runtime_error)
	{
		if (Base::set_base_property(key,value)) return;
    else if (key == "alphap0") { _alphap0 = cast<T>(value); _initialized = false; }
    else if (key == "baseline") { _baseline = cast<T>(value); }
    else if (key == "alphag1") { _alphag1 = cast<T>(value); _initialized = false; }
	  else if (key == "beta0") { _beta0 = cast<T>(value); _initialized = false; }	  
	  else if (key == "betal0") { _betal0 = cast<T>(value); _initialized = false; }	  
    else if (key == "full_gamma_prior") { _full_prior = cast<bool>(value); _initialized = false; }
		else if (key == "fgevalmax") { _fgevalmax = cast<int>(value); }
    else throw(std::runtime_error((boost::format("gmm2::set_property invalid key %s")%key).str()));
	}

	PropertyType get_property(const std::string & key) throw(std::runtime_error)
	{
		bool handled_by_base;
		PropertyType baseresult = Base::get_base_property(key,handled_by_base);
		if (handled_by_base) return baseresult;
	  else if (key == "alphag1") return _alphag1;
	  else if (key == "beta0") return _beta0;
	  else if (key == "betal0") return _betal0;
    else if (key == "full_gamma_prior") return _full_prior;
	  else if (key == "L") return _L;
	  else if (key == "Lmu") return _Lmu;
    else throw(std::runtime_error((boost::format("gmm2::get_property invalid key %s")%key).str()));
	}

  void flush()
  {		
		Base::flush();
    _beta.free();
    _betag.free();
    _initialized = false;
  }

  void batchlearn_begin_epoch()
  {
		Base::batchlearn_begin_epoch();
    update();
    _wx.resize(Base::_k,Base::_d,Base::_d,Base::_m); 
    if (_full_prior)
      _wlx.resize(Base::_k,Base::_d,Base::_d,Base::_m); 
    _wlx = 0;
    _wx = 0;
		CHECKNAN(_lbiasp,"gmm2::bactchlearn_begin_epoch nan in _lbiasp\n");
  }
		
	T expected_log_probability(const Matrix1 & rout, const Matrix3 & data)
	{
    Matrix3 X = prepareData(data,0,0);
    if (!_full_prior && _alphag1 == 1)
      return sum(rout(i1) * _lbiasp) + sum( - X(i2,i3,i4) * _betag * rout(i1) );
    else
      return sum(rout(i1) * _lbiasp) +  sum( rout(i1) * (fast_log(X(i2,i3,i4)) * (_alphag-1) - X(i2,i3,i4) * _betag) );
	}
			
  void upward_message(const Matrix3 & data, Matrix1 & lrout, bool variational = false, bool domax = false)
  {
    if (variational)
      lrout += _lbiasp;
    else
      lrout += _lbias;
    Matrix3 X = prepareData(data,0,0);
    if (!_full_prior && _alphag1 == 1)
      lrout += sum(sum(sum( - X(i2,i3,i4) * _betag ,i4),i3),i2);
    else
      lrout += sum(sum(sum( fast_log(X(i2,i3,i4)) * (_alphag-1) - X(i2,i3,i4) * _betag ,i4),i3),i2);
  }

  void upward_message(const Matrix3 & data, Matrix1 & lrout, const Matrix3 & w, bool variational = false, bool domax = false)
  {
    BLITZ_INDICES;
    Matrix3 X = prepareData(data,0,0);
    if (variational)
      lrout += _lbiasp;
    else
      lrout += _lbias;
    int n1 = data.extent(0), n2 = data.extent(1);
    for (int i = 0; i < n1; ++i)
      for (int j = 0; j < n1; ++j)
        for (int k = 0; k < Base::_k; ++k)
          if (!_full_prior && _alphag1 == 1)
            lrout(k) += w(i,j,k) * sum( - X(i,j,_all) * _betag(k,i,j,_all) );
          else
            lrout(k) += w(i,j,k) * ( (_alphag(k)-1) * sum( fast_log(X(i,j,_all)) ) 
                                - sum( X(i,j,_all) * _betag(k,i,j,_all) ) );
	}
			
  void sample_one(unsigned int out, Matrix3 & X)
  {
    for (typename Matrix3::iterator iX = X.begin(), end = X.end(); iX != end; ++iX)
    {
      blitz::TinyVector<int,3> pos = iX.position();
      *iX = sample_gamma( _alphag(pos),1/_betag(pos),Base::_rnd );
    }
  }
						  
	void collect_statistics(const Matrix3 & data, const Matrix1 & r, T w)
	{
		Matrix3 X = prepareData(data,0,0);
		_wx += r(i1) * X(i2,i3,i4);
    if (_full_prior)
			_wlx += r(i1) * fast_log(X(i2,i3,i4));
		CHECKNAN(_wx,"gmm2::apply_r nan in wx\n");
		CHECKNAN(_wlx,"gmm2::apply_r nan in wlx\n");
	}

	void collect_statistics(const Matrix3 & data, const Matrix1 & r, T w, const Matrix2 & w2)
	{
		Matrix3 X = prepareData(data,0,0);
		_wx += r(i1) * X(i2,i3,i4) * w2(i2,i3);
    if (_full_prior)
			_wlx += r(i1) * fast_log(X(i2,i3,i4)) * w2(i2,i3);
		CHECKNAN(_wx,"gmm2::apply_r nan in wx\n");
		CHECKNAN(_wlx,"gmm2::apply_r nan in wlx\n");
	}

	void collect_statistics(const Matrix3 & data, const Matrix1 & r, T w, const Matrix3 & w3)
	{
		Matrix3 X = prepareData(data,0,0);
		_wx += r(i1) * X(i2,i3,i4) * w3(i1,i2,i3);
    if (_full_prior)
			_wlx += r(i1) * fast_log(X(i2,i3,i4)) * w3(i1,i2,i3);
		CHECKNAN(_wx,"gmm2::apply_r nan in wx\n");
		CHECKNAN(_wlx,"gmm2::apply_r nan in wlx\n");
	}
		
		
	Matrix1 batchlearn_end_epoch(Matrix1 f_in = Matrix1())
  {
		Matrix1 f = Base::batchlearn_end_epoch();
    
    if (Base::_renormalize_counts)
    {
			CHECKNAN(_wx,"gmm2::renorm-pre nan in wx\n");
			CHECKNAN(_wlx,"gmm2::renorm-pre nan in wlx\n");
      _wx = f(i1) * _wx(i1,i2,i3,i4);
      if (_full_prior)
        _wlx = f(i1) * _wlx(i1,i2,i3,i4);
			CHECKNAN(f,"gmm2::renorm nan in f\n");
			CHECKNAN(_wx,"gmm2::renorm nan in wx\n");
			CHECKNAN(_wlx,"gmm2::renorm nan in wlx\n");
    }
    
    _beta = (T(1)-Base::_delta) * _beta + Base::_delta * (_beta0 + _wx);
    if (_full_prior)
      _betal = (T(1)-Base::_delta) * _betal + Base::_delta * (_betal0 + _wlx);
    else
			_alphap = (T(1)-Base::_delta) * _alphap + Base::_delta * (_alphap0 + _alphag1 * Base::_wk);
      
		_L = _Lmu + Base::_Lz + Base::_Lp;
    _wx.free();
		_wlx.free();

    _dirty = true;
    
    return f;
  }

  Matrix3 prepareData(const Matrix3 & data,int x, int y)
  {
		Matrix3 X(_baseline+data(blitz::Range(x,x+Base::_d-1),blitz::Range(y,y+Base::_d-1),blitz::Range::all()));
//    return Matrix3(X*(1/(T(0.1)+sum(X)/data.numElements())));
    return X;
  }

#if 0
  std::vector<int> prune(T threshold)
  {
    if (!_initialized) return;
    Matrix1 score((_alpha - min(_alpha)) / (max(_alpha) - min(_alpha)));
    std::vector<int> survivors;
    for (int k = 0; k < _k; ++k)
      if (score(k) > threshold)
      survivors.push_back(k);
    _k = survivors.size();
    Matrix1 alphanew(_k);
    Matrix4 betanew(_k,_d,_d,_m);
    for (int k = 0; k < _k; ++k)
    {
      alphanew(k) = _alpha(survivors[k]);
			betanew(k,blitz::Range::all(),blitz::Range::all(),blitz::Range::all()) = _beta(survivors[k],blitz::Range::all(),blitz::Range::all(),blitz::Range::all());
    }
    _alpha.reference(alphanew);
    _beta.reference(betanew);
    _dirty = true;
		return survivors;
  }

#endif
			

  void initialize()
  {
	if ( Base::initialize() && _initialized && !force ) 
	  return;

    _beta.resize(Base::_k,Base::_d,Base::_d,Base::_m);
  //draw a few random samples from the prior
    _beta = _beta0;
    Matrix4 sample(_beta.shape());
    ranlib::UniformClosedOpen<T> rnd;
    foreach(T & x, sample) x = _beta0 * rnd.random() + _baseline;
    _beta += Base::_initcount * sample;
    if (_full_prior)
    {
      _betal.resize(Base::_k,Base::_d,Base::_d,Base::_m);
      _betal = _betal0 + Base::_initcount * fast_log(sample);
    } else {
      _alphap.resize(Base::_k);
      _alphap = _alphap0 + _alphag1 * Base::_initcount;
    }
    _initialized = _dirty = true;
  }

  void update()
  {
		Base::update();
		initialize();
    if (_dirty)
    {
      _betag.resize(Base::_k,Base::_d,Base::_d,Base::_m);
      _alphag.resize(Base::_k,Base::_d,Base::_d,Base::_m);
      _lbias.resize(Base::_k);
      _lbiasp.resize(Base::_k);
      if (_full_prior)
      {
        for (int k = 0; k < Base::_k; ++k)
        {
          _lbias(k) = 0; _lbiasp(k) = 0;
        for (int d1 = 0; d1 < Base::_d; ++d1)
        for (int d2 = 0; d2 < Base::_d; ++d2)
        for (int m = 0; m < Base::_m; ++m)
          {
            double alpha, beta, lgalpha, lbeta;
            boost::tie(alpha, lgalpha, beta, lbeta)
            = gamma_prior_expectations(_betal(k,d1,d2,m),_beta(k,d1,d2,m),Base::_alpha(k),Base::_alpha(k),_fgevalmax);
            _lbias(k) += fast_log(beta) * alpha - gammaln(alpha);
            _lbiasp(k) += lbeta * alpha - lgalpha;
            _alphag(k,d1,d2,m) = alpha;
            _betag(k,d1,d2,m) = beta;
          }
        }
				//_lbiasp = _lbias;
      } else {
        _alphag = _alphag1;
        _betag = _alphap(i1) / _beta;
        _lbias = sum(sum(sum(fast_log(_betag) * _alphag1 - gammaln(_alphag1),i4),i3),i2);
        _lbiasp = sum(sum(sum((psi(_alphap(i1))-fast_log(_beta)) * _alphag1 - gammaln(_alphag1),i4),i3),i2);
      }
    }
    _dirty = false;
  }

private:

  friend class boost::serialization::access;
    template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
		Base::serialize(ar,version);
    ar & _alphap0;
    ar & _beta0;
    ar & _betal0;
    ar & _alphag1;
    ar & _full_prior;
    ar & _initialized;
    ar & _baseline;
    if (_initialized)
    {
			checked_serialize_array1(ar,_alphap,Base::_k);
			checked_serialize_array1(ar,_beta,Base::_k,Base::_d,Base::_d,Base::_m);
			checked_serialize_array1(ar,_betal,Base::_k,Base::_d,Base::_d,Base::_m);
    }
    _dirty = true;
  }

  Matrix1 _lbias, _lbiasp, _alphap;
  Matrix4 _beta, _alphag, _betal, _betag, _wx, _wlx;
  T _alphap0, _beta0, _betal0, _alphag1;
  int _fgevalmax;
  bool _full_prior, _initialized, _dirty;			
	T _Lmu, _L;
  T _baseline;

  blitz::firstIndex i1;
  blitz::secondIndex i2;
  blitz::thirdIndex i3;
  blitz::fourthIndex i4;
  blitz::fifthIndex i5;

#ifdef GSL
  gsl_rng * _rng;
#endif

};

#endif
