/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __GREEDY_CLUSTERER_HPP__
#define __GREEDY_CLUSTERER_HPP__

#include <blitz/array.h>
#include <vector>
#include <cmath>
#include <boost/serialization/serialization.hpp>

#include "SeariseProperties.hpp"
#include "blitz_indices.hpp"
#include "blitz_properties.hpp"
#include "serialize_blitz.hpp"
#include "metrics.hpp"

// FIXME: adjusting centers is kinda broken, as histograms are off
//        a possible fix is to have a second run with adjusting centers 
//        ( and maybe growing too ) switched off
 
template<typename T, class Metric> class greedy_clusterer 
  : public HasProperties< SearisePropertyTypes, greedy_clusterer<T,Metric> >
{

public:

  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<double,4> Matrix4d;

  typedef greedy_clusterer<T,Metric> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;

  static void init_properties()
  {
	PBase::add_ivar_property_rw( "d1", &This::_d1 );
	PBase::add_ivar_property_rw( "d2", &This::_d2 );
	PBase::add_ivar_property_rw( "m", &This::_m );
	PBase::add_ivar_property_rw( "max_dist", &This::_max_dist );
	PBase::add_ivar_property_rw( "max_n_centers", &This::_max_n_centers );
	PBase::add_ivar_property_rw( "min_count", &This::_min_count );
	PBase::add_ivar_property_rw( "adjust_centers", &This::_adjust_centers );
	PBase::add_property( "centers", make_blitz_ivar_getter1( &This::_centers ) );
	PBase::add_property( "counts", make_blitz_ivar_getter1( &This::_counts ) );	
  }

  greedy_clusterer( bool estimate_variances = true )
	: _k(0)
	, _d1(1)
	, _d2(1)
	, _m(1)
	, _max_n_centers(-1)
	, _min_count(0)
	, _max_dist(1)
	, _adjust_centers( false )
	, _estimate_variances( estimate_variances )
  {
  }

  size_t size()
  {
	return _k;
  }

  void init( )
  {
	init( _d1, _d2, _m );
  }

  void init( int d1, int d2, int m )
  {
	_d1 = d1;
	_d2 = d2;
	_m = m;
	_k = 0;
	_histograms.free();
	_histograms2.free();
	_centers.free();
	_counts.free();
	_counts3.free();
  }

  void consume( const Matrix3 & data )
  {
	BLITZ_INDICES;

	if ( _k == 0 )
	  spawn_center( data );
	else
	  {
		Matrix1 dists( _k );
		for ( int k = 0; k < _k; ++k )
		  dists( k ) = Metric::dist( data, _centers( k, _all, _all, _all ) );
		int imin = minIndex( dists )[0];
		T dmin = dists( imin );
		if ( dmin > _max_dist )
			spawn_center( data );
		else
		  update_center( imin, data );
	  }
	
  }

  void consume( const Matrix3 & data, const Matrix2 & weights )
  {
	BLITZ_INDICES;

	if ( _k == 0 )
	  spawn_center( data, weights );
	else
	  {
		Matrix1 dists( _k );
		for ( int k = 0; k < _k; ++k )
		  dists( k ) = Metric::dist( data, _centers( k, _all, _all, _all ), weights( i1, i2 ) );
		int imin = minIndex( dists )[0];
		T dmin = dists( imin );
		if ( dmin > _max_dist )
		  spawn_center( data, weights );
		else
		  update_center( imin, data, weights );
	  }
	
  }

  void finish()
  {
	BLITZ_INDICES;
	if ( _min_count > 0 )
	  {
		std::vector<int> survivors;
		for ( int i = 0; i < _k; ++i )
		  if ( _counts(i) >= _min_count )
			survivors.push_back( i );
		select( survivors );
	  }
	
	while ( _max_n_centers > 0 && size() > size_t( _max_n_centers ) )
	  {
		// find smallest center, keep it and pop it
		std::vector<int> entries;
		for ( int i = 0; i < _k; ++i )
		  entries.push_back( i );
		std::sort( entries.begin(), entries.end(), compare_by_lookup( _counts ) );
		size_t l = entries.front();
		Matrix3d center( _centers( l, _all, _all, _all ) );
		T count = _counts( l );
		Matrix2d counts3( _counts3( l, _all, _all ) );
		Matrix3d histogram( _histograms( l, _all, _all, _all ) );
		Matrix3d histogram2;
		if ( _estimate_variances )
		  histogram2.reference( _histograms2( l, _all, _all, _all ) );
		entries.erase( entries.begin() );
		select( entries );
		
		// make closest center consume old smallest center
		Matrix1 dists( _k );
		for ( int k = 0; k < _k; ++k )
		  dists( k ) = Metric::dist( center, _centers( k, _all, _all, _all ) );
		int imin = minIndex( dists )[0];
		//std::cout << "merging " << l << " counts " << count 
		//		  << " into " << imin << " counts " << _counts( imin )
		//		  << std::endl;
		_counts( imin ) += count;
		_counts3( imin, _all, _all ) += counts3;
		_histograms( imin, _all, _all, _all ) += histogram;
		if ( _estimate_variances )
		  _histograms2( imin, _all, _all, _all ) += histogram2;


	  }

	std::cout << " greedy clusterer finished n: " << _k << std::endl;

  }

  const Matrix4d & centers() { return _centers; }
  const Matrix1d & counts()  { return _counts;  }
  const Matrix3d & counts3() { return _counts3; }
  const Matrix4d & histograms() { return _histograms; }
  const Matrix4d & histograms2() { return _histograms2; }

private:
  
  void select( const std::vector<int> & survivors )
  {
	BLITZ_INDICES;

	_k = survivors.size();
	Matrix1d newcounts ( _k );
	Matrix3d newcounts3 ( _k, _d1, _d2 );
	Matrix4d newcenters( _k, _d1, _d2, _m );
	Matrix4d newhistograms( _k, _d1, _d2, _m );
	Matrix4d newhistograms2( _k, _d1, _d2, _m );
	for ( int i = 0; i < _k; ++i )
	  {
		newcounts (i)                = _counts ( survivors[i] );
		newcounts3 ( i, _all, _all ) = _counts3 ( survivors[i], _all, _all );
		newcenters(i,_all,_all,_all) = _centers( survivors[i], _all, _all, _all );
		newhistograms(i,_all,_all,_all) = _histograms( survivors[i], _all, _all, _all );
		if ( _estimate_variances )
		  newhistograms2(i,_all,_all,_all) = _histograms2( survivors[i], _all, _all, _all );		
	  }
	_counts .reference( newcounts  );
	_counts3.reference( newcounts3  );
	_centers.reference( newcenters );
	_histograms.reference( newhistograms );
	if ( _estimate_variances )
	  _histograms2.reference( newhistograms2 );
	
  }

  struct compare_by_lookup
  {
	compare_by_lookup( const Matrix1d & m ) : lut(m) {}
	bool operator() ( int i1, int i2 ) const
	{
	  return lut(i1) < lut(i2);
	}
	const Matrix1d & lut;
  };
  
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & BOOST_SERIALIZATION_NVP(_k);
	if ( version < 1 )
	  {
		int _d;
		ar & BOOST_SERIALIZATION_NVP(_d);
		_d1 = _d2 = _d;
	  }
	else
	  {
		ar & BOOST_SERIALIZATION_NVP(_d1);
		ar & BOOST_SERIALIZATION_NVP(_d2);
	  }
    ar & BOOST_SERIALIZATION_NVP(_m);
    ar & BOOST_SERIALIZATION_NVP(_centers);
    ar & BOOST_SERIALIZATION_NVP(_counts);
    ar & BOOST_SERIALIZATION_NVP(_adjust_centers);
    ar & BOOST_SERIALIZATION_NVP(_max_dist);
  }

  void update_center( int index, const Matrix3 & data )
  {
	BLITZ_INDICES;

	_counts ( index             ) += T(1);
	_counts3( index, _all, _all ) += T(1);
	
	if ( _adjust_centers )
	  {
		Matrix3d center( _centers( index, _all, _all, _all ) );
		double n = _counts( index );
		center = ( n * center + data ) / ( n + 1.0 );
	  }
	
	_histograms( index, _all, _all, _all ) += data;
	if ( _estimate_variances )
	  _histograms( index, _all, _all, _all ) += pow2( data );

  }

  void update_center( int index, const Matrix3 & data, const Matrix2 & weights )
  {
	BLITZ_INDICES;

	_counts ( index             ) += max( weights );
	_counts3( index, _all, _all ) +=      weights;
	
	if ( _adjust_centers )
	  {
		Matrix3d center( _centers( index, _all, _all, _all ) );
		Matrix2d n2 = _counts3( index, _all, _all );
		center = ( n2( i1, i2 ) * center + weights( i1, i2 ) * data ) / max( 0.0001, n2( i1, i2 ) + weights( i1, i2 ) );
	  }
	
	_histograms( index, _all, _all, _all ) += weights(i1,i2) * data;
	if ( _estimate_variances )
	  _histograms( index, _all, _all, _all ) += weights(i1,i2) * pow2( data );

  }

  void spawn_center( const Matrix3 & center )
  {
	BLITZ_INDICES;

	++_k;
	_counts.resizeAndPreserve     ( _k );
	_counts3.resizeAndPreserve    ( _k, _d1, _d2 );
	_centers.resizeAndPreserve    ( _k, _d1, _d2, _m );
	_histograms.resizeAndPreserve ( _k, _d1, _d2, _m );
	_histograms2.resizeAndPreserve( _k, _d1, _d2, _m );

	_counts( _k - 1 ) = T(1);
	_counts3( _k - 1, _all, _all ) = T(1);
	_centers( _k - 1, _all, _all, _all ) = center(i1,i2,i3);
	_histograms( _k - 1, _all, _all, _all ) = center(i1,i2,i3);
	if ( _estimate_variances )
	  _histograms2( _k - 1, _all, _all, _all ) = pow2( center(i1,i2,i3) );

	//std::cout << " greedy clusterer spawned n: " << _k << std::endl;

  }

  void spawn_center( const Matrix3 & center, const Matrix2 & weights )
  {
	BLITZ_INDICES;

	++_k;
	_counts.resizeAndPreserve( _k );
	_centers.resizeAndPreserve( _k, _d1, _d2, _m );
	_counts3.resizeAndPreserve( _k, _d1, _d2 );
	_histograms.resizeAndPreserve( _k, _d1, _d2, _m );
	_histograms2.resizeAndPreserve( _k, _d1, _d2, _m );

	_counts( _k - 1 ) = max( weights );
	_counts3( _k - 1, _all, _all ) = weights(i1,i2);
	_centers( _k - 1, _all, _all, _all ) = center(i1,i2,i3);
	_histograms( _k - 1, _all, _all, _all ) = weights(i1,i2) * center(i1,i2,i3);
	if ( _estimate_variances )
	  _histograms2( _k - 1, _all, _all, _all ) = weights(i1,i2) * pow2( center(i1,i2,i3) );

	std::cout << " greedy clusterer spawned n: " << _k << std::endl;

  }

  int _k, _d1, _d2, _m, _max_n_centers, _min_count;
  T _max_dist;
  bool _adjust_centers, _estimate_variances;
  Matrix1d _counts;
  Matrix3d _counts3;
  Matrix4d _centers;
  Matrix4d _histograms;
  Matrix4d _histograms2;

};

//templated class versioning
namespace boost { namespace serialization {
	template<typename T, class Metric>
	struct version< greedy_clusterer< T, Metric > >
	{
	  BOOST_STATIC_CONSTANT(unsigned int, value = 1);
	};
  }}

#endif
