/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __HEURISTIC_SEQUENCE_CLUSTERING_HPP__
#define __HEURISTIC_SEQUENCE_CLUSTERING_HPP__

#include <blitz/array.h>
#include <boost/format.hpp>
#include <stdexcept>
#include <list>

#include "blitz_indices.hpp"
#include "SeariseProperties.hpp"

template<typename T>
class heuristic_sequence_clusterer
  : public HasProperties< SearisePropertyTypes, heuristic_sequence_clusterer<T> >
{
  
public:
  
  typedef heuristic_sequence_clusterer<T> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<double,4> Matrix4d;
  
  typedef std::list< size_t > state_group;
  typedef std::list< state_group > state_groups;

  void dump_group( const state_group & group )
  {
	std::cout << "( ";
	BOOST_FOREACH( size_t state, group )
	  std::cout << state << " ";
	std::cout << ")";
  }
  
  static void init_properties()
  {
	PBase::add_ivar_property_rw( "top-neighbors", &This::_top_neighbors );
	PBase::add_ivar_property_rw( "max-group-size", &This::_max_group_size );
	PBase::add_ivar_property_rw( "overlapping-groups", &This::_overlapping_groups );
	PBase::add_ivar_property_rw( "large-group-penalty", &This::_lgp );
	PBase::add_ivar_property_rw( "metric", &This::_metric_str );
	PBase::add_property(         "A",          make_blitz_ivar_getter1( &This::_transition_probabilities ) );
	PBase::add_property(         "W",          make_blitz_ivar_getter1( &This::_transition_histogram ) );
	PBase::add_property(         "P",          make_blitz_ivar_getter1( &This::_state_histogram ) );
  }
  
  heuristic_sequence_clusterer() 
	: _top_neighbors(2)
	, _max_group_size(5)
	, _k(0) 
	, _overlapping_groups(false)
	, _lgp(0)
	, _metric_str("numenta")
  {}
  
  void init( size_t k )
  {
	_k = k;
	_transition_histogram.resize( k, k );
	_transition_histogram = 0;
	_state_histogram.resize( k );
	_state_histogram = 0;
	break_sequence();
  }
  
  void break_sequence()
  {
	_last_data.free();
  }
  
  struct descending_score
  {
	const Matrix1 & score;
	descending_score( const Matrix1 & scores ) : score(scores) {}
	bool operator()( size_t i1, size_t i2 )
	{
	  return score(i1) < score(i2);
	}
  };
  
  T groups_numenta_similarity( state_group & group1, state_group & group2 )
  {
	BLITZ_INDICES;
	T A12 = 0, S1 = 0, S2 = 0, N1 = group1.size(), N2 = group2.size();
	BOOST_FOREACH( size_t state1, group1 )
	  BOOST_FOREACH( size_t state2, group2 )
	  A12 += _transition_histogram( int(state1), int(state2) );
	BOOST_FOREACH( size_t state1, group1 )
	  S1 += sum( _transition_histogram( int(state1), _all ) );
	BOOST_FOREACH( size_t state2, group2 )
	  S2 += sum( _transition_histogram( int(state2), _all ) );

	T s =  S1 > T(0) && S2 > T(0) ? A12 / ( ( S1 + S2 ) * std::pow( N1 + N2, _lgp ) ) : T(0);

	return s;
  }

  T groups_my_similarity( state_group & group1, state_group & group2 )
  {
	BLITZ_INDICES;
	T A12 = 0, S1 = 0, S2 = 0;
	BOOST_FOREACH( size_t state1, group1 )
	  BOOST_FOREACH( size_t state2, group2 )
	  A12 += _transition_histogram( int(state1), int(state2) );
	BOOST_FOREACH( size_t state1, group1 )
	  S1 += sum( _transition_histogram( int(state1), _all ) );
	BOOST_FOREACH( size_t state2, group2 )
	  S2 += sum( _transition_histogram( int(state2), _all ) );

	T s =  S1 > T(0) && S2 > T(0) ? A12 / ( std::pow( S1 + S2, _lgp ) ) : T(0);

	return s;
  }

  T groups_max_similarity( state_group & group1, state_group & group2 )
  {
	T d = 0;
	BOOST_FOREACH( size_t state1, group1 )
	  BOOST_FOREACH( size_t state2, group2 )
	  d = std::max( d, _transition_probabilities( int(state1), int(state2) ) );
	return d;
  }
  
  void end_batch()
  {
	// symmetrize, assuming temporal symmetry
	BLITZ_INDICES;
	std::cout << " hsc end batch count " << sum( _transition_histogram ) << std::endl;
	Matrix2d sth( max( T(1.0), T(0.5) * ( _transition_histogram(i1,i2) + _transition_histogram(i2,i1) ) ) );
	_transition_histogram.reference( sth );
	_transition_probabilities.resize( _k, _k );
	_transition_probabilities = _transition_histogram( i1, i2 )  
	  / sum( _transition_histogram( i1, i3 ), i3 );
  }
  
  // hierarchical single linkage clustering up to a level where there are the given amount of groups
  void cluster_AHC( size_t num_groups )
  {
	int metric = -1;
	if ( _metric_str == "numenta" )
	  metric = 0;
	else if ( _metric_str == "min-dist" )
	  metric = 1;
	else if ( _metric_str == "mine" )
	  metric = 2;
	else
	  throw std::runtime_error( str( boost::format( "cluster_AHC invalid metric: %s" ) % _metric_str ) );
	_groups.clear();
	for ( size_t state = 0; state < _k; ++state )
	  _groups.push_back( state_group( 1, state ) );
	//size_t iteration = 0;
	//std::cout << " clustering " << _groups.size() << "/" << num_groups << " groups " << std::endl;
	while ( _groups.size() > num_groups )
	  {
		//std::cout << " iteration " << iteration++ << " with " << _groups.size() << "/" << num_groups << " groups " << std::endl;
		state_groups::iterator g0, g1, g2;
		T d = 0;
		for ( state_groups::iterator i1 = _groups.begin(); i1 != _groups.end(); ++i1 )
		  {
			//std::cout << " g1 ";
			//dump_group(*i1);
			for ( state_groups::iterator i2 = _groups.begin(); i2 != _groups.end(); ++i2 )
			{
			  //std::cout << " g2 ";
			  //dump_group(*i2);
			  if ( i1 != i2 )
				{
				  if ( ( i1->size() + i2->size() ) <= size_t(_max_group_size) )
					{
					  T d12 = -1;
					  switch(metric)
						{
						case 0: d12 = groups_numenta_similarity( *i1, *i2 ); break;
						case 1: d12 = groups_max_similarity( *i1, *i2 ); break;
						case 2: d12 = groups_my_similarity( *i1, *i2 ); break;
						}
					  //std::cout << " d " << d12;
					  if ( d12 > d )
						{
						  d = d12;
						  g1 = i1;
						  g2 = i2;
						}
					}
				  //else
				  //std::cout << " too large! ";
				}
			}
			//std::cout << std::endl;
		  }
		if ( g1 == g0 || g2 == g0 )
		  {
			std::cout << "seq_clustering: stopping prematurely" << std::endl;
			break; // due to the group size limit it is possible that we do not find the required number of groups
		  }
		/*
		std::cout << " merging d = " << d << ": ";
		BOOST_FOREACH( size_t state, *g1 )
		  std::cout << state << " ";
		std::cout << " and ";
		BOOST_FOREACH( size_t state, *g2 )
		  std::cout << state << " ";
		*/
		std::cout << std::endl;
		std::copy( g1->begin(), g1->end(), std::back_inserter( *g2 ) );
		_groups.erase( g1 );
	  }
	report_clustering();
  }
  
  void cluster_heuristic( )
  {
	// set of state__infos where group members are sorted in descending order of abundancy
	state_group unclustered;
	for ( size_t i = 0; i < _k; ++i )
	  unclustered.push_back( i );
	unclustered.sort( descending_score( _state_histogram ) );
	BOOST_FOREACH( size_t state, unclustered )
	  std::cout << " " << _state_histogram(state);
	std::cout << std::endl;
	
	_groups.clear();
	_max_group_size = std::max( 1, _max_group_size );
	
	while ( !unclustered.empty() )
	  {
		state_group current;
		state_group to_add, to_add2;
		Matrix1 scorebuf( _k );
		
		// seed with unclustered state with highest abundance
		to_add.push_back( unclustered.front() );
		
		while ( !to_add.empty() )
		  {
			
			// transfer candidates to group until full
			bool full = false;
			BOOST_FOREACH( size_t state, to_add )
			  {
				if ( ( full = ( current.size() == size_t(_max_group_size) ) ) )
				  break;
				current.push_back( state );
				unclustered.remove( state );
			  }
			if ( full )
			  break;
			
			// find neigbors
			BOOST_FOREACH( size_t state, to_add )
			  {
				
#define contains( container, value ) ( std::find( container.begin(), container.end(), value ) != container.end() )
				
				state_group neighbors;
				get_neighbors( state, neighbors );
				size_t ineighbor = 0;
				
				BOOST_FOREACH( size_t neighbor, neighbors )
				  {
					if ( ++ineighbor > size_t(_top_neighbors) ) 
					  break; 
					if ( ( _overlapping_groups || contains( unclustered, neighbor ) )
						 && !contains( current, neighbor )
						 && !contains( to_add2, neighbor ) 
						 )
					  {
						to_add2.push_back( neighbor );
						scorebuf( int(neighbor) ) = _transition_histogram( int(state), int(neighbor) );
					  }
				  }
#undef contains
				
			  }
			
			// make sure stronger connected neighbors are considered first
			to_add2.sort( descending_score( scorebuf ) );
			
			// prepare recursion
			to_add.clear();
			to_add.swap( to_add2 );
			
		  }
		
		_groups.push_back( current );
		
	  }
	report_clustering();
  }

  void cluster_spectral()
  {
	
  }

  void report_clustering()
  {
	std::cout << "sequence clustering done, got " << _groups.size() << " groups:" << std::endl;
	BOOST_FOREACH( const state_group & group, _groups )
	  {
		double abundance = 0, N = sum( _state_histogram );
		BOOST_FOREACH( size_t state, group )
		  {
			std::cout << " " << state;
			abundance += _state_histogram(state);
		  }		
		std::cout << " abundance " << abundance << " relative " << (abundance/N) << std::endl;
	  }
  }
  
  void collect_statistics( const Matrix3 & data1, const Matrix3 & data2 )
  {
	BLITZ_INDICES;
	
	if ( data1.extent(2) * data2.extent(2) != int(_k) )
	  throw std::runtime_error( str( boost::format( "heuristic_sequence_clustering: "
													"data does not match setup, data depths: %i/%i "
													"expected: %i" ) % data1.extent(2) % data2.extent(2) % _k ) );

	bool in_sequence = _last_data.numElements() > 0 && _last_data2.numElements() > 0;

	if ( in_sequence )
	  {		
		int k1 = data1.extent(2);
		int k2 = data2.extent(2);
		blitz::TinyVector<int,2> sP( k1, k2 );
		blitz::TinyVector<int,4> sT( k1, k2, k1, k2 );
		Matrix4d Tr = reshape( _transition_histogram, sT );
		Matrix2d P  = reshape( _state_histogram, sP );
		
		for ( int i = 0; i < data1.extent(0); ++i )
		  {
			for ( int j = 0; j < data1.extent(1); ++j )
			  {
				Matrix1 x1( _last_data ( i, j, _all ) );
				Matrix1 x2( _last_data2( i, j, _all ) );
				Matrix1 y1( data1( i, j, _all ) );
				Matrix1 y2( data2( i, j, _all ) );
				Tr += x1( i1 ) * x2( i2 ) * y1( i3 ) * y2( i4 );
				P  += y1( i1 ) * y2( i2 )                      ;
			  }		
		  }
	  }
	_last_data.reference( data1.copy() );
	_last_data2.reference( data2.copy() );

  }

  void collect_statistics( const Matrix3 & data )
  {
	BLITZ_INDICES;
	
	if ( data.extent(2) != int(_k) )
	  throw std::runtime_error( str( boost::format( "heuristic_sequence_clustering: "
													"data does not match setup, data depth: %i "
													"expected: %i" ) % data.extent(2) % _k ) );
	bool in_sequence = _last_data.numElements() > 0;
	
	if ( in_sequence && _last_data.numElements() != data.numElements() )
	  throw std::runtime_error( str( boost::format( "heuristic_sequence_clustering: "
													"data(t) has different extents than last data(t-1): "
													"data(t-1): %ix%i data(t): %ix%i" )
									 % _last_data.extent(0) % _last_data.extent(1)
									 % data.extent(0) % data.extent(1) ) );	  
	
	_state_histogram += sum( sum( data(i2,i3,i1), i3 ), i2 );
	
	if ( in_sequence )
	  _transition_histogram += sum( sum( _last_data(i3,i4,i1) * data(i3,i4,i2), i4 ), i3 );
	
	_last_data.reference( data.copy() );
	
  }
  
  const state_groups & groups() { return _groups; } 
  const Matrix2d & transitions() { return _transition_histogram; }
  const Matrix1d & abundances  () { return      _state_histogram; }
  
private:

  template<typename T12,int R1, int R2>
  inline static blitz::Array<T12,R2> reshape( blitz::Array<T12,R1> & X, blitz::TinyVector<int,R2> shape )
  {
	return blitz::Array<T12,R2>( X.data(), shape, blitz::neverDeleteData );
  }
  
  
  inline void get_neighbors( size_t state, state_group & result, bool with_self = false )
  {
	Matrix1 transitions( _transition_histogram( state, blitz::Range::all() ) );	
	result.clear();
	for ( size_t i = 0; i < _k; ++i )
	  if (  i != state || with_self )
		result.push_back(i);
	result.sort( descending_score( transitions ) );
  }

  int _top_neighbors;
  int _max_group_size;
  Matrix3 _last_data;
  Matrix3 _last_data2;
  Matrix2d _transition_histogram;
  Matrix2 _transition_probabilities;
  Matrix1d _state_histogram;
  size_t _k;
  std::list<state_group> _groups;
  bool _overlapping_groups;
  int _num_groups;
  T _lgp;
  std::string _metric_str;

};

#endif
