/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include "masking.hpp"
#include <iostream>
#include "gammaApprox.h"
#include "blitz_indices.hpp"

template<typename T>
struct mu_mask_iter
{
  mu_mask_iter( InverseRFIterator & iter, const blitz::Array<T,3> & mu_mask, const blitz::Array<unsigned int,2> & latents_sample )
	: _iter( iter ), _mu_g( mu_mask ), _sample( latents_sample ) {}
  T next() 
  {
	int i1rf = _iter.i1_rf(), i2rf = _iter.i1_rf(), latent = _sample( _iter.j1, _iter.j2 );
	T mu_j = _mu_g( i1rf, i2rf, latent );
	_iter.increment_j();
	return mu_j;
  }
  InverseRFIterator _iter;
  const blitz::Array<T,3> & _mu_g;
  const blitz::Array<unsigned int,2> & _sample;
};

template<typename T>
struct mu_mask_iter1
{
  mu_mask_iter1( InverseRFIterator & iter, const blitz::Array<T,3> & mu_mask, const blitz::Array<unsigned int,2> & latents_sample, int gate )
	: _iter( iter ), _mu_g( mu_mask ), _sample( latents_sample ), _gate(gate) {}
  T next() 
  {
	if ( _iter.j_index() == _gate ) _iter.increment_j();
	int i1rf = _iter.i1_rf(), i2rf = _iter.i1_rf(), latent = _sample( _iter.j1, _iter.j2 );
	T mu_j = _mu_g( i1rf, i2rf, latent );
	_iter.increment_j();
	return mu_j;
  }
  InverseRFIterator _iter;
  const blitz::Array<T,3> & _mu_g;
  const blitz::Array<unsigned int,2> & _sample;
  int _gate;
};

template<typename T>
struct mu_mask_iter2
{
  mu_mask_iter2( InverseRFIterator & iter, const blitz::Array<T,3> & mu_mask, const blitz::Array<unsigned int,2> & latents_sample, int gate, int parent )
	: _iter( iter ), _mu_g( mu_mask ), _sample( latents_sample ), _gate(gate), _parent(parent) {}
  T next() 
  {
	while ( _iter.j_index() == _gate || _iter.j_index() == _parent ) _iter.increment_j();
	int latent = _sample( _iter.j1, _iter.j2 ), i1rf =  _iter.i1_rf(), i2rf =  _iter.i2_rf();
	T mu_j = _mu_g( i1rf, i2rf, latent );
	_iter.increment_j();
	return mu_j;
  }
  InverseRFIterator _iter;
  const blitz::Array<T,3> & _mu_g;
  const blitz::Array<unsigned int,2> & _sample;
  int _gate, _parent;
};

template<typename T>
T bit_count_distrib( const blitz::Array<T,1> & mu, blitz::Array<T,1> & pN )
{  
  int K = mu.extent(0);
  pN = 0;
  pN(0) = 1; // a zero-bit vector has probability one of not carrying any bits...
  for ( int k = 0; k < K - 1; ++k )
	{
	  T mu_k = mu( k );
	  for ( int n = k + 1; n > 0; --n )
		{
		  // induction: a vector one bit longer induces a mixture of the vector one bit shorter
		  // with zero and one additional bits ( weighed with the new bit's probability of being on or off, respectively )
		  pN(n) = mu_k * pN( n - 1 ) + pN(n) * ( T(1) - mu_k );
		}
	  // the probabilty of no bit being on successively reduces...
	  pN(0) *= ( T(1) - mu(k) );
	}
  return sum( pN );
}

template<typename I, typename T>
T bit_count_distrib_iter( I imu, blitz::Array<T,1> & pN, int K )
{  
  pN(0) = 1; // a zero-bit vector has probability one of not carrying any bits...
  for ( int k = 0; k < K - 1; ++k )
	{
	  pN( k + 1 ) = T( 0 );
	  T mu_k = imu.next();
	  for ( int n = k + 1; n > 0; --n )
		{
		  // induction: a vector one bit longer induces a mixture of the vector one bit shorter
		  // with zero and one additional bits ( weighed with the new bit's probability of being on or off, respectively )
		  pN(n) = mu_k * pN( n - 1 ) + pN(n) * ( T(1) - mu_k );
		}
	  // the probabilty of no bit being on successively reduces...
	  pN(0) *= ( T(1) - mu_k );
	}
  return sum( pN );
}

template<typename T>
T bit_count_distrib_skip_bit( const blitz::Array<T,1> & mu, blitz::Array<T,1> & pN, int skip )
{  
  int K = mu.extent(0);
  pN = 0;
  pN(0) = 1; // a zero-bit vector has probability one of not carrying any bits...
  for ( int k = 0, ik = 0; k < K - 1; ++k, ++ik )
	{
	  if ( k == skip ) ++ik;
	  T mu_k = mu( ik );
	  for ( int n = k + 1; n > 0; --n )
		{
		  // induction: a vector one bit longer induces a mixture of the vector one bit shorter
		  // with zero and one additional bits ( weighed with the new bit's probability of being on or off, respectively )
		  pN(n) = mu_k * pN( n - 1 ) + pN(n) * ( T(1) - mu_k );
		}
	  // the probabilty of no bit being on successively reduces...
	  pN(0) *= ( T(1) - mu_k );
	}
  return sum( pN );
}

template<typename T>
T bit_count_distrib_skip_bits( const blitz::Array<T,1> & mu, blitz::Array<T,1> & pN, int skip1, int skip2 )
{  
  int K = mu.extent(0);
  pN = 0;
  pN(0) = 1; // a zero-bit vector has probability one of not carrying any bits...
  for ( int k = 0, ik = 0; k < K - 1; ++k, ++ik )
	{
	  if ( k == skip1 || k == skip2 ) ++ik;
	  for ( int n = k + 1; n > 0; --n )
		{
		  // induction: a vector one bit longer induces a mixture of the vector one bit shorter
		  // with zero and one additional bits ( weighed with the new bit's probability of being on or off, respectively )
		  pN(n) = mu(ik) * pN( n - 1 ) + pN(n) * ( T(1) - mu(ik) );
		}
	  // the probabilty of no bit being on successively reduces...
	  pN(0) *= ( T(1) - mu(k) );
	}
  return sum( pN );
}

template<typename T>
struct array_iter
{
  typedef T value_type;
  array_iter( const blitz::Array<T,1> & a ) : _array(a), i(0) {}
  bool more() { return i < _array.extent(0); }
  T next() { return _array( i++ ); }
  int i;
  const blitz::Array<T,1> & _array;
};

template<typename T>
struct array_iter_skip1
{
  typedef T value_type;
  array_iter_skip1( const blitz::Array<T,1> & a, int i1 ) : _array(a), i(0), _i1( i1 ) { }
  void check() { while ( i == _i1 ) ++i; }
  bool more() { check(); return i < _array.extent(0); }
  T next() { return _array( i++ ); }
  const blitz::Array<T,1> & _array;
  int i, _i1;
};

template<typename T>
struct array_iter_skip2
{
  typedef T value_type;
  array_iter_skip2( const blitz::Array<T,1> & a, int i1, int i2 ) : _array(a), i(0), _i1( i1 ), _i2(i2) { check(); }
  void check() { while ( i == _i1 || i == _i2 ) ++i; }
  bool more() { return i < _array.extent(0); }
  T next() { T val = _array( i++ ); check(); return val; }
  int i, _i1, _i2;
  const blitz::Array<T,1> & _array;
};

template<typename T> 
struct gate_distrib_kernel
{
  gate_distrib_kernel( T mu ) : _mu( mu ) {}
  T operator()( int i ) const { return _mu / ( 1 + i ); }
  T _mu;
};

template<typename I, typename F> 
typename I::value_type bit_counting_expectation( I imu, const F & f )
{
  typedef typename I::value_type T;
  int i = 0;
  T Ef = f( i++ );
  while ( imu.more() )
	{
	  T mu_i = imu.next();
	  Ef = ( 1 - mu_i ) * Ef + f( i++ ) * mu_i;
	}
  return Ef;
}

template<typename T>
T gating_distrib_new(  const blitz::Array<T,1> & mu, blitz::Array<T,1> & pg )
{
  int K = mu.extent( 0 );
  T Z = 0;
  for ( int k = 0; k < K; ++k )
	{
	  array_iter_skip1<T> mu_iter( mu, k );
	  gate_distrib_kernel<T> kernel( mu( k ) );
	  pg(k) = bit_counting_expectation( mu_iter, kernel );
	  Z += pg(k);
	}
  return Z;
}


template<typename T>
T gating_distrib_fast(  const blitz::Array<T,1> & mu, blitz::Array<T,1> & pN, blitz::Array<T,1> & pg )
{
  int K = mu.extent(0);
  T Z = 0;
  bit_count_distrib( mu, pN );
  for ( int k = 0; k < K; ++k )
	{
	  // pg_k = mu_k * < 1 / ( 1 + count( bits_-k ) ) >
	  T pg_k = 0;
	  T mu_k = mu(k);
	  // reconstruct the counting distrib of bits_-k from that of bits_all
	  T p_n = 0;
	  // this is probably unstable as it is an alternating sum and 1 - mu_k might be very small...
	  // problem is calculating each expectation explicitly would make the algorithm K^3 :-(
	  for ( int n = 0; n < K; ++n )
		{
		  p_n = ( pN(n) - mu_k * p_n ) / ( 1 - mu_k );
		  pg_k += p_n / ( 1 + n );
		}
	  pg(k) = pg_k * mu_k;
	  Z += pg(k);
	}  
  return Z;
}

template<typename T>
T gating_distrib_slow(  const blitz::Array<T,1> & mu, blitz::Array<T,1> & pN, blitz::Array<T,1> & pg )
{
  BLITZ_INDICES;
  int K = mu.extent(0);
  T Z = 0;
  for ( int k = 0; k < K; ++k )
	{
	  // pg_k = mu_k * < 1 / ( 1 + count( bits_-k ) ) >
	  T ZN = bit_count_distrib_skip_bit( mu, pN, k );
	  pg(k) = mu(k) * sum( pN( blitz::Range( 0, K - 1 ) ) / ( i1 + 1 ) ) / ZN;
	  Z += pg(k);
	}  
  return Z;
}

template<typename T1, typename T2>
void update_mask_distrib( const blitz::Array<T1,1> & mu, blitz::Array<T1,1> & pN, int i, int j, T1 & mean, T2 & count )
{
  if ( i == j )
	{
	  // bernoulli success
	  count += T2(1);
	  mean = ( mean * count + T2(1) ) / ( count + T2(1) );
	}
  else
	{
	  int K = mu.extent(0);
	  bit_count_distrib_skip_bits( mu, pN, i, j );
	  T1 c0 = 0, c1 = 0;
	  for ( int n = 0; n < K - 1; ++n )
		{
		  c0 += pN( n ) / ( 1 + n );
		  c1 += pN( n ) / ( 2 + n );
		}
	  gammaApprox::approxBetaUpdateVar( c1 - c0, c0, mean, count );
	}
}
