/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "mcmm2.hpp"
#include "blitz_indices.hpp"
#include <boost/format.hpp>
#include <boost/foreach.hpp>




#define _subchain_num (Base::_k)
#define _input_dim    (Base::_m)

#ifdef _WIN32
#define throw_not_implemented												\
  throw std::runtime_error( str( boost::format( "%s(%i): method not implented for mcmm2" ) \
								 % __FILE__ % __LINE__ ) )
#else
#define throw_not_implemented												\
  throw std::runtime_error( str( boost::format( "%s(%i): method \"%s\" not implented for mcmm2" ) \
								 % __FILE__ % __LINE__ % __func__ ) )
#endif
/// k is the output space cardinality
/// d is the receptive field size
/// m is the input space cardinality
/// step is the stepping between neighboring receptive fields
/// for the mcmm2 probably d and step need to be 1
template< typename T >
MarkovChainMixtureModel2<T>::MarkovChainMixtureModel2( )
  : _subchain_init_uniform(false)
  , _subchain_alpha0(0.5)
  , _subchain_gamma0(0.5)
  , _subchain_gammadiag0(0.5)
  , _subchain_learn_gamma0_interval(0)
  , _subchain_use_pi(false)
  , _subchain_use_pi_inseq(false)
  , _subchain_factorize_transitions(false)
  , _subchain_norm_whole_A(false)
  , _subchain_do_vb(true)
  , _Lmu(0)
{
}

template<typename PropertyTypes, class C, class Container >
struct contained_ptr_getter
{
  contained_ptr_getter( Container C::* _ivar, size_t _index )
	: ivar(_ivar), index(_index) {}
  Container C::* ivar;
  size_t index;
  typedef  HasPropertiesBase<PropertyTypes> Base;
  typedef typename Container::value_type ptr_type;
  typename Base::PropertyType operator()( const Base *b )
  {
	const C * c = dynamic_cast<const C*>(b);
	if ( c == 0 )
	  throw std::runtime_error("could not down cast");
	ptr_type * ptr = const_cast<ptr_type*>( &( (c->* ivar).at(index) ) );
	return ptr;
  }
};

template< typename T >
typename MarkovChainMixtureModel2<T>::PBase::PropertyType MarkovChainMixtureModel2<T>::get_transition_matrices( )
{
  BLITZ_INDICES;
  CPUData<float> tm;
  
  int s1 = _subchains[0].get_transition_matrix().extent(0);
  int s2 = _subchains[0].get_transition_matrix().extent(1);
  
  blitz::Array<float, 3> btm(Base::_k, s1, s2);
  
  for (int k= 0; k < Base::_k; k++)
  {
    const Matrix2 & A = _subchains[k].get_transition_matrix();
    btm(k, _all, _all) = A;
  }
  blitz2cpudata( btm, tm );
  
  return tm;
}

template< typename T >
typename MarkovChainMixtureModel2<T>::PBase::PropertyType MarkovChainMixtureModel2<T>::get_transition_posteriori( )
{
  BLITZ_INDICES;
  CPUData<float> pm;
  
  int s1 = _subchains[0].get_transition_posteriori().extent(0);
  int s2 = _subchains[0].get_transition_posteriori().extent(1);
  
  blitz::Array<float, 3> bpm(Base::_k, s1, s2);
  
  for (int k= 0; k < Base::_k; k++)
  {
    const Matrix2 & gamma = _subchains[k].get_transition_matrix();
    bpm(k, _all, _all) = gamma;
  }
  blitz2cpudata( bpm, pm );
  
  return pm;
}



template< typename T >
typename MarkovChainMixtureModel2<T>::PBase::PropertyType MarkovChainMixtureModel2<T>::get_sub_pi()
{
  BLITZ_INDICES;
  CPUData<float> cpi;
  
  assert( int(_subchains.size()) == Base::_k );
  int s = _subchains[0].get_pi().extent(0);
  
  Matrix2 bpi( Base::_k, s );
  
  for (int k= 0; k < Base::_k; k++)
  {
    const Matrix1 & alpha = _subchains[k].get_pi();
    bpi(k, _all) = alpha;
  }
  blitz2cpudata( bpi, cpi );
  
  return cpi;
}




/// this is where to register all parameters accessible in json files
template< typename T >
void MarkovChainMixtureModel2<T>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter( &This::_initialized, false );
  //read-write
  PBase::add_ivar_property_rw(std::string("init-uniform"),                            &This::_subchain_init_uniform,           uninitializer);
  PBase::add_ivar_property_rw(std::string("Lmu"),                                     &This::_Lmu,                             uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-do-vb"),                          &This::_subchain_do_vb,                  uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-alpha0"),                         &This::_subchain_alpha0,                 uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-gamma0"),                         &This::_subchain_gamma0,                 uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-gammadiag0"),                     &This::_subchain_gammadiag0,             uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-normalize-whole-transition-matrix"),&This::_subchain_norm_whole_A,         uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-learn-gamma0"),                   &This::_subchain_learn_gamma0_interval,  uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-use-pi"),                         &This::_subchain_use_pi,                 uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-use-pi-in-sequence"),             &This::_subchain_use_pi_inseq,           uninitializer);
  PBase::add_ivar_property_rw(std::string("subchain-factorize-joint-statistics"),     &This::_subchain_factorize_transitions,  uninitializer);
  PBase::add_property(std::string("transition-matrices"), typename PBase::MemFunGetter( &This::get_transition_matrices ) );
  PBase::add_property(std::string("transition-posteriori"), typename PBase::MemFunGetter( &This::get_transition_posteriori ) );
  PBase::add_property(std::string("sub-pi"), typename PBase::MemFunGetter( &This::get_sub_pi ) );
  
  /*for ( size_t i = 0; i < 100; ++i )
  {	  
	  typename PBase::Getter getter = contained_ptr_getter< SearisePropertyTypes, This, std::vector<Subchain_Type> >( &This::_subchains, i );
	  PBase::add_property( str( boost::format( "subchain%i" ) % i ), getter );
  }*/

  // DANIEL: die folgenden kannst du von der Base klasse nehmen ... nagut!
  // PBase::add_ivar_property_rw(std::string("delta"), &This::_delta);
  // PBase::add_ivar_property_rw(std::string("initcount"), &This::_initcount);
  // PBase::add_ivar_property_rw(std::string("renormalize_counts"), &This::_renormalize_counts);
}

/// upon a call to this the model should forget all it learned and start from scratch
template< typename T >
void MarkovChainMixtureModel2<T>::flush()
{
  throw_not_implemented;
}

/// here the model should return wether it is ready to produce output
/// i.e. you can return false here if you are running a possibly
/// iterative initialization which needs to see all the data
/// for one or more epochs
template< typename T >
bool MarkovChainMixtureModel2<T>::ready()
{
  return true;
}

/// not currently used, but here to support pruning
template< typename T >
void MarkovChainMixtureModel2<T>::reset_input_dimensions(const std::set<unsigned int> & components)
{
  throw_not_implemented;
}

/// not currently used, but here to support pruning
template< typename T >
void MarkovChainMixtureModel2<T>::reset_components(const std::set<unsigned int> & components)
{
  throw_not_implemented;
}

/// as the name says: notifies begin of a batch epoch
template< typename T >
void MarkovChainMixtureModel2<T>::batchlearn_begin_epoch()
{
  _Lmu = 0;
  update();
  BOOST_FOREACH( Subchain_Type & subchain, This::_subchains )
	  subchain.batchlearn_begin_epoch( );
  Base::batchlearn_begin_epoch();
}

///////////////////////////////////////////////////////////////////////////////////
///// messages: these define the model, they are made up of the conditional //////
///// probabilities of each set of variables given others                  //////
////////////////////////////////////////////////////////////////////////////////

template< typename T >
void MarkovChainMixtureModel2<T>::calc_upward_msg( IMatrix2 & data )
{
  throw_not_implemented;
}

template< typename T >
void MarkovChainMixtureModel2<T>::break_sequence()
{
  BOOST_FOREACH( Subchain_Type & subchain, This::_subchains )
	subchain.reset();
}

template< typename T >
void MarkovChainMixtureModel2<T>::calc_upward_msg( Matrix3 & data )
{
  BLITZ_INDICES;
  
  update();

  _data_buffer.front().data.reference( data );
  _data_buffer.front().saliency.reference( Base::_S );

  Matrix3 msg(data.extent(0), data.extent(1), _subchains.size());

  for ( size_t i = 0; i < _subchains.size(); ++i )
  {
    const Matrix3 & submsg = _switch2subchain_messages[i].get_belief(true);
    
    

    if ( submsg.extent(0) == 1 && submsg.extent(1) == 1 )
        msg(_all,_all,int(i)) = safe_log( sum( data( i1, i2, i3 ) * submsg( 0, 0, _all )(i3), i3) );
    else
	    msg(_all,_all,int(i)) = safe_log( sum( data( i1, i2, i3 ) * submsg( i1, i2, i3 ), i3) );

  }
  
  set_msg( msg );
}


/// here the model should calculate the probability of the latent (output)
/// variable given data (usually probabilities of the input variables)
/// the data is an input receptive field view: (rows x columns x input-depth)
/// the output is a single output "pixel"
/// the output neds to be stored logarithmically
/// variational is for MAP estimation (different parameter set)

template< typename T >
void MarkovChainMixtureModel2<T>::begin_mp_frame()
{
  update();
  Base::begin_mp_frame();
  
  unsigned int history_capacity = message_component<T>::buffer().get_history_capacity();
  
  if (history_capacity != _data_buffer.size() )
    _data_buffer.resize( history_capacity );

  _data_buffer.push_front( data_sal_pair_t<T>( Matrix3(), Matrix2() ) ) ;

  BOOST_FOREACH( Subchain_Type & subchain, This::_subchains )
  {
    subchain.buffer().set_history_capacity( history_capacity );
    
    subchain.begin_mp_frame( );
  }
}


template< typename T >
void MarkovChainMixtureModel2<T>::update_chain_messages(  )
{
  BLITZ_INDICES;
   
  Matrix3 mu_s_f3( message_component<T>::get_belief(true) );
  Matrix3 S_belief;

  if ( Base::_learning )
  {
    message_component<T>::get_belief(true).free();
    S_belief.reference( message_component<T>::get_belief() );
  }
  
  int active_state = message_component<T>::buffer().active_state();
  Matrix3 y( _data_buffer[active_state].data, true );
  
  //bool have_saliency = Base::_S.extent(0) == y.extent(0) && Base::_S.extent(1) == y.extent(1);

  for ( int k = 0; k < Base::_k; ++k )
  {
	  Matrix2 mu_s_f3_k = mu_s_f3( _all, _all, k );
	
  	Matrix3 msg;
	  
	  if ( mu_s_f3_k.extent(0) == 1 && mu_s_f3_k.extent(1) == 1 )
	  {
		  msg.reference( Matrix3 ( safe_log( ( y( i1, i2, i3 ) * mu_s_f3_k( 0, 0 ) ) +   // subchain active contrib.
										   ( ( T(1) - mu_s_f3_k( 0, 0 ) ) / T(Base::_k) ) ) )  );
	  }
	  else if ( mu_s_f3_k.numElements() > 0 )
	  {
		  msg.reference( Matrix3 ( safe_log( ( y( i1, i2, i3 ) * mu_s_f3_k( i1, i2 ) ) +   // subchain active contrib.
										   ( ( T(1) - mu_s_f3_k( i1, i2 ) ) / T(Base::_k) ) ) ) );
	  }
	  
    _switch2subchain_messages[k].set_msg(msg);
    
    Base::_Lp += cast<T>( _subchains[k].get_property("Lp") );
    _Lmu += cast<T>( _subchains[k].get_property("Lmu") );
	
  }
  Base::_L = Base::_Lz + _Lmu + Base::_Lp;
}


template< typename T >
void MarkovChainMixtureModel2<T>::update_switch_message(  )
{
  int active_state = buffer().active_state();
  calc_upward_msg( _data_buffer[active_state].data );
}


template< typename T >
typename MarkovChainMixtureModel2<T>::Matrix2
MarkovChainMixtureModel2<T>::learn( )
{
  BLITZ_INDICES;
  int active_state = buffer().active_state();

  Matrix2 saliency( _data_buffer[active_state].saliency, true );
  
  Matrix3 S_belief;
  message_component<T>::get_belief(true).free();
  S_belief.reference( message_component<T>::get_belief() );
  

  for ( int k = 0; k < Base::_k; ++k )
  {	
    if ( (saliency.extent(0) > 0) && (saliency.extent(1) > 0) )
      _subchains[k].set_saliency( Matrix2( S_belief( _all, _all, k ) * saliency )  );
    else
      _subchains[k].set_saliency( Matrix2( S_belief( _all, _all, k ) )  );
    _subchains[k].apply_r();
    _subchains[k].set_saliency( Matrix2() );
  }
  
  return saliency;
}


template< typename T >
blitz::Array<T,4> MarkovChainMixtureModel2<T>::get_inner_prediction( )
{
  BLITZ_INDICES;
  
  if (Base::_k == 0)
    return Matrix4(); 
  
  const Matrix3 & first = _subchains[0].get_msg().data;
  Matrix4 ret( Base::_k, first.extent(0), first.extent(1), first.extent(2) );

  for ( int k = 0; k < Base::_k; ++k )
  {
    ret(k, _all, _all, _all) = _subchains[k].get_msg().data(_all, _all, _all);
  }

  return ret;
}

template< typename T >
blitz::Array<T,3> MarkovChainMixtureModel2<T>::calc_max_component_likelihoods( )
{
  Matrix3 ret( Base::_k, 1, 1 );

  for (int k=0; k < Base::_k; k++)
  {
    ret(k,0,0) = safe_log( max( _subchains[k].get_transition_matrix() ) );
  }
  
  return ret;
}


template< typename T >
bool MarkovChainMixtureModel2<T>::step_forward( bool front )
{

  if ( front )
    update_chain_messages();
  for ( int k = 0; k < Base::_k; ++k )
  {	
    _subchains[k].step_forward();
  }  
  if ( !front )
  {
    update_chain_messages();
    update_switch_message();
  }

  return true;
}

template< typename T >
bool MarkovChainMixtureModel2<T>::step_backward( )
{
  for ( int k = 0; k < Base::_k; ++k )
  {	
    _subchains[k].step_backward();
  }
  update_chain_messages();
  update_switch_message();
  
  return true;
}



template< typename T >
void MarkovChainMixtureModel2<T>::end_mp_frame( const Matrix3 & data_ )
{
  BLITZ_INDICES;

  _data_buffer.front().data.reference( data_ );
  _data_buffer.front().saliency.reference( Base::_S );
  _data_buffer.front().data.makeUnique();
  _data_buffer.front().saliency.makeUnique();    

  for ( int k = 0; k < Base::_k; ++k )
  {
    _subchains[k].end_mp_frame(true);
  }

  Base::end_mp_frame( data_ );  // calls apply_r
  
}

template< typename T >
blitz::Array<typename MarkovChainMixtureModel2<T>::num_type,3> MarkovChainMixtureModel2<T>::calc_downward_msg ( Matrix3 data, bool logarithmic )
{
  BLITZ_INDICES;
  // This::_sub_r: R x [ W x H x M ]
  //            r: [ W x H x R ]
  //      out_msg: [ W x H x M ]
  
  This::update();
    
  Matrix3 & s = message_component<T>::get_belief();
  Matrix3 out_msg( s.extent(0), s.extent(1), Base::_m );
  out_msg = 0;
  
  //  bool have_saliency = This::_S.extent(0) == out_msg.extent(0) && This::_S.extent(1) == out_msg.extent(1);

#if 0  
  if ( have_saliency )
  {	
    throw_not_implemented;
  }
  else
#endif
  {
    for ( int i = 0; i < _subchain_num; ++i )
    {
      Matrix2 si = s( _all, _all, i );
      Matrix3 & sub_prediction = _switch2subchain_messages[i].get_belief(true);
      out_msg += sub_prediction( i1, i2, i3 ) * si( i1, i2 );
    }
  }
	
  if ( logarithmic )
    out_msg = safe_log( out_msg );
	
  return out_msg;
}

template< typename T >
blitz::Array<typename MarkovChainMixtureModel2<T>::num_type,2> MarkovChainMixtureModel2<T>::calc_self_information( Matrix3 & data )
{
  BLITZ_INDICES;
  // data(rin) : W x H x M
 	Matrix3 & master_belief = message_component<T>::get_belief( true ); // W x H x R

  Matrix2 sinf( data.extent(0), data.extent(1));
  sinf = 0;

  for ( int i = 0; i < _subchain_num; ++i )
  {
    Matrix2 mbelief_slice = master_belief( _all, _all, i );  // W x H
    Matrix3 & sub_prediction = _switch2subchain_messages[i].get_belief(true);  // W x H x M
    sinf += sum( sub_prediction( i1, i2, i3 ) * data( i1, i2, i3 ), i3 ) * mbelief_slice( i1, i2 );
  }

  sinf = safe_log( sinf );
  return sinf;
}


/// this gets called at the end of a e-step epoch
/// f_in is a reweighing factor for the input space
/// and should be applied to any histograms
/// similarly you can return a renormalization factor
/// for the output space which will be propagated
template< typename T >
blitz::Array<typename MarkovChainMixtureModel2<T>::num_type,1> MarkovChainMixtureModel2<T>::batchlearn_end_epoch(blitz::Array<T,1> f_in )
{
  Base::_Lp = 0;
  //Base::_Lz = 0;
  Base::_Lmu = 0;
  Matrix1 f = Base::batchlearn_end_epoch();
  for ( int k = 0; k < Base::_k; ++k )
  {	
	T f_out = f.numElements() == Base::_k ? f( k ) : 1.0f;
	_subchains[k].batchlearn_end_epoch( f_in, f_out );
    Base::_Lp  += cast<T>( _subchains[k].get_property( "Lp"  ) );  
    Base::_Lmu += cast<T>( _subchains[k].get_property( "Lmu" ) );  
    Base::_Lz  += cast<T>( _subchains[k].get_property( "Lz"  ) );  
  }
  Base::_L = Base::_Lp + Base::_Lmu + Base::_Lz;
  //std::cout << boost::str( boost::format("mcmm2 ended epoch Lmu = %1.3e Lp = %1.3e L = %1.3e Lz = %1.3e") % Base::_Lmu % Base::_Lp % Base::_L % Base::_Lz ) << std::endl;
  return f;
}

/// this is to initialize all (hyper-) parameters
/// called automatically by base
/// you should call Base::initlialize from here also
/// force means re-initialize, even you already did it before
template< typename T >
void MarkovChainMixtureModel2<T>::initialize( bool force )
{
  bool base_initialized = Base::initialize( force );
  
  if ( base_initialized )
  {
    
    This::_subchains.resize(Base::_k);
    This::_switch2subchain_messages.resize(Base::_k);
	// connect a switch2subchain slot to each subchain, to send messages from the swithc to each subchain
    for ( int i = 0; i < Base::_k; ++i )
      _switch2subchain_messages[i].set_buffer( _subchains[i].buffer(),
          str( boost::format("_switch2subchain[%i]")%i ));
    
    BOOST_FOREACH( Subchain_Type & subchain, This::_subchains )
    {
      // propagate settings to all subchains
      
      subchain.set_property( "init-uniform",               This::_subchain_init_uniform );
      subchain.set_property( "do-vb",                      This::_subchain_do_vb );
      subchain.set_property( "alpha0",                     This::_subchain_alpha0 );
      subchain.set_property( "gamma0",                     This::_subchain_gamma0 );
      subchain.set_property( "gammadiag0",                 This::_subchain_gammadiag0 );
      subchain.set_property( "learn-gamma0",               This::_subchain_learn_gamma0_interval );
      subchain.set_property( "factorize-joint-statistics", This::_subchain_factorize_transitions );
      subchain.set_property( "use-pi",                     This::_subchain_use_pi );
      subchain.set_property( "use-pi-in-sequence",         This::_subchain_use_pi_inseq );
      
      subchain.set_property( "k",                          Base::_m );
      subchain.set_property( "initcount" ,                 Base::_initcount );
      subchain.set_property( "delta"     ,                 Base::_delta );
      subchain.set_property( "owns-state",                 true         );
      
      // call the subchain initialize funtion -> obviously private!
      subchain.initialize(force);
    }
  }
}

/// after this call parameters should be ready for operation
/// you should call initialize if you need to from here
/// also you should call Base::update
/// force as above
template< typename T >
void MarkovChainMixtureModel2<T>::update(bool force )
{
  initialize();
  if ( Base::update( force ) )
  {
    BOOST_FOREACH( Subchain_Type & subchain, This::_subchains )
      subchain.update(force);

    float diffsum = 0;
    for (int i = 1; i < Base::_k; i++ )
    {
      for (int j = 0; j < i; j++ )
      {
         diffsum += sum( abs( _subchains[j].get_transition_matrix() - _subchains[i].get_transition_matrix() ) );
      }
    }
    std::cout << "diffsum " << diffsum << std::endl;
  }
}


/// explicit instatiation of the mcmm2 for float data
template class MarkovChainMixtureModel2<float>;


