/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef MCMM2_HPP
#define MCMM2_HPP

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/serialization/serialization.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <iostream>

#include "MixtureModel2Base.hpp"
#include "serialize_blitz.hpp"
#include "tfmm.hpp"


template<typename T>
struct data_sal_pair_t
{
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  
  Matrix3 data;
  Matrix2 saliency;

  data_sal_pair_t( )
  {
  }

  data_sal_pair_t( const Matrix3 &d, const Matrix2 &s )
  {
    data.reference( d );
    saliency.reference( s );
  }

  data_sal_pair_t & operator=( const data_sal_pair_t & other )
  {
    data.reference( other.data );
    saliency.reference( other.saliency );
    return *this;
  }
};




template<typename T> class MarkovChainMixtureModel2;
template<typename T> struct TypeTraits< MarkovChainMixtureModel2<T> > 
{	
  typedef T num_type;
  static const int num_statistics = 0;
};

template<typename T> class MarkovChainMixtureModel2 
  : public MixtureModel2Base< MarkovChainMixtureModel2<T> >,
	public HasProperties< SearisePropertyTypes, MarkovChainMixtureModel2<T> >
{
  
public:

  typedef MarkovChainMixtureModel2<T> This;
  typedef MixtureModel2Base< This > Base;
  typedef HasProperties< SearisePropertyTypes, MarkovChainMixtureModel2<T> > PBase;
  
  
  typedef T num_type;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<unsigned int,2> IMatrix2;
  typedef TemporalForwardMarkovModel<T> Subchain_Type;


  /// k is the output space cardinality
  /// d is the receptive field size
  /// m is the input space cardinality
  /// step is the stepping between neighboring receptive fields
  /// for the mcmm2 probably d and step need to be 1
  MarkovChainMixtureModel2();
  
  /// this is where to register all parameters accessible in json files
  static void init_properties();
  
  /// upon a call to this the model should forget all it learned and start from scratch
  void flush();

  /// here the model should return wether it is ready to produce output
  /// i.e. you can return false here if you are running a possibly
  /// iterative initialization which needs to see all the data
  /// for one or more epochs
  bool ready();
  
  /// not currently used, but here to support pruning
  void reset_input_dimensions(const std::set<unsigned int> & components);
  
  /// not currently used, but here to support pruning
  void reset_components(const std::set<unsigned int> & components);
  
  /// as the name says: notifies begin of a batch epoch
  void batchlearn_begin_epoch();
  
  ///////////////////////////////////////////////////////////////////////////////////
  ///// messages: these define the model, they are made up of the conditional //////
  ///// probabilities of each set of variables given others                  //////
  ////////////////////////////////////////////////////////////////////////////////

  /// this if for models allowing overlap, so probably not for this
  void segmentation_message2 ( const Matrix1 & rout, const IMatrix2 & data, Matrix2 & lp ) {}
  
  /// this if for models allowing overlap, so probably not for this
  void segmentation_message2 ( const Matrix1 & rout, const Matrix3 & data, Matrix2 & lp ) {}


  /// override the complete upward message logic because we process full images at a time here
  void calc_upward_msg( Matrix3 & data );
  
  /// override the complete upward message logic because we process full images at a time here
  void calc_upward_msg( IMatrix2 & data );


  /// override the complete downward message logic because we process full images at a time here
  Matrix3 calc_downward_msg ( Matrix3 data = Matrix3(), bool logarithmic = true );
  
  /// overrode the complete calc_self_information because we process full images at a time here
  Matrix2 calc_self_information( Matrix3 & data );
  
  
  /// inter-frame callbacks
  void begin_mp_frame();
  void end_mp_frame( const Matrix3 & data );

  /// this gets called at the end of a e-step epoch
  /// f_in is a reweighing factor for the input space
  /// and should be applied to any histograms
  /// similarly you can return a renormalization factor
  /// for the output space which will be propagated
  Matrix1 batchlearn_end_epoch(Matrix1 f_in = Matrix1());

  /// this is to initialize all (hyper-) parameters
  /// called automatically by base
  /// you should call Base::initlialize from here also
  /// force means re-initialize, even you already did it before
  void initialize(bool force=false);
  
  /// after this call parameters should be ready for operation
  /// you should call initialize if you need to from here
  /// also you should call Base::update
  /// force as above
  void update(bool force = false);
  
  /// leave this: means that your input space is multinomial
  bool is_multinomial() { return true; }
  
  void break_sequence();
  
  bool step_forward(bool front = true);
  bool step_backward();
  Matrix2 learn( );
  
  Matrix4 get_inner_prediction();
  Matrix3 calc_max_component_likelihoods();

private:
  
  mmm2_belief_message_buffer<T> & buffer()
  {
    return message_component<T>::buffer();
  }
  
  typename PBase::PropertyType get_transition_matrices( );
  typename PBase::PropertyType get_transition_posteriori( );
  typename PBase::PropertyType get_sub_pi();

   
  void update_chain_messages(  );
  void update_switch_message(  );
 
   /// here you can use boost::serialization to save/load your state
  /// this gets called from above when saving/loading all network state
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
	ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(Base);
	ar & BOOST_SERIALIZATION_NVP(_subchains);
	ar & BOOST_SERIALIZATION_NVP(_subchain_alpha0);
	ar & BOOST_SERIALIZATION_NVP(_subchain_gamma0);
    ar & BOOST_SERIALIZATION_NVP(_subchain_gammadiag0);
    ar & BOOST_SERIALIZATION_NVP(_subchain_learn_gamma0_interval);
    ar & BOOST_SERIALIZATION_NVP(_subchain_use_pi);
    ar & BOOST_SERIALIZATION_NVP(_subchain_do_vb);
    ar & BOOST_SERIALIZATION_NVP(_subchain_factorize_transitions);
    if ( _switch2subchain_messages.size() != _subchains.size() )
        _switch2subchain_messages.resize( _subchains.size() );
    for ( size_t i = 0; i < _switch2subchain_messages.size(); ++i )
    	_switch2subchain_messages.at(i).set_buffer( _subchains.at(i).buffer(), 
          str( boost::format("_switch2subchain[%i]")%i ));
  }

  
  boost::ptr_vector<Subchain_Type> _subchains;
  std::vector< message_component<T> > _switch2subchain_messages;
  
  // statistics
  
  // subchain vars
  bool _subchain_init_uniform;
  T _subchain_alpha0;
  T _subchain_gamma0;
  T _subchain_gammadiag0;
  int _subchain_learn_gamma0_interval;
  bool _subchain_use_pi;
  bool _subchain_use_pi_inseq;
  bool _subchain_factorize_transitions;
  bool _subchain_norm_whole_A;
  bool _subchain_do_vb;
  T _Lmu;
  //Matrix3 switch_prediction();
  
  boost::circular_buffer< data_sal_pair_t<T> > _data_buffer;
};

#endif
