/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once
#define clamp(x,xlo,xhi) std::min( T(xhi), std::max( T(xlo), T(x) ) )

template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type >::sample_data( MixtureModel2Base<Derived> & model, Matrix3 & data, IMatrix2 & latents_sample, InverseRFIterator & iter )
{
  BLITZ_INDICES;
  IMatrix2 gates_sample( model._gates_sample, true );
  Derived & derived = model.derived();
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  int gate_sample = gates_sample( iter.i1, iter.i2 );
	  int j1 = model.j_index_to_j1( gate_sample, iter.i1 );
	  int j2 = model.j_index_to_j2( gate_sample, iter.i2 );
	  int i1rf = iter.i1 - j1 * model._step1;
	  int i2rf = iter.i2 - j2 * model._step2;
	  int latent_sample = latents_sample( j1, j2 );
	  Matrix1 data_point( data( iter.i1, iter.i2, _all ) );
	  derived.sample_point( i1rf, i2rf, latent_sample, data_point );
	}
}

template<typename Derived> void get_current_factor( MixtureModel2Base<Derived> & model, 
													typename Derived::upward_message_iterator_t & p,
													typename TypeTraits<Derived>::num_type & current_factor )
{
  const bool have_temperature = !model._linear_annealing && model._current_rate != 1;
  current_factor = have_temperature 
	? aexpf( model._current_rate * p.evaluate_log() ) 
	: p.evaluate_linear();
}

template<typename Derived> void get_last_factor( MixtureModel2Base<Derived> & model, 
												 typename Derived::upward_message_iterator_t & p, 
												 typename TypeTraits<Derived>::num_type & current_factor,
												 typename TypeTraits<Derived>::num_type & last_factor )
{
  const bool have_temperature = !model._linear_annealing && model._current_rate != 1;
  last_factor = have_temperature && model._current_rate != model._last_rate
	? aexpf( model._last_rate * p.evaluate_log() ) 
	: last_factor;
}

#ifdef HAVE_SSE
template<typename Derived> void get_current_factor_sse( MixtureModel2Base<Derived> & model, 
														typename Derived::upward_message_iterator_t & p, 
														vecf4 & current_factor )
{
  const bool have_temperature = !model._linear_annealing && model._current_rate != 1;
  current_factor = have_temperature 
	? exp( model._current_rate * p.evaluate_log_sse() ) 
	: p.evaluate_linear_sse();
}

template<typename Derived> void get_last_factor_sse( MixtureModel2Base<Derived> & model, 
													 typename Derived::upward_message_iterator_t & p, 
													 vecf4 & current_factor, vecf4 & last_factor )
{
  const bool have_temperature = !model._linear_annealing && model._current_rate != 1;
  last_factor = have_temperature && model._current_rate != model._last_rate
	? exp( model._last_rate * p.evaluate_log_sse() ) 
	: current_factor;
}
#endif

// this is for sequential update in gauss-seidel style where the parents are factorized ( mean-field )
template<typename Derived>
 void update_gated_msg_delegate< 
  Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::update_parent( MixtureModel2Base<Derived> & model, Matrix3 & data, 
					Matrix3 & r, Matrix3 & msg, ForwardRFIterator & iter )
{
  
  Derived & derived = model.derived();
  BLITZ_INDICES;
  
  // === pass1: update r given gates ===
  Matrix3 rw( r, true );
  Matrix3 zbufferw( model._zbuffer, true );
  Matrix3 sbufferw( model._sbuffer, true );
  Matrix3 bbufferw( model._bbuffer, true );
  Matrix3 dataw( data, true );
  Matrix1 r1( rw( iter.j1, iter.j2, _all ) );
  Matrix1 r1_old;
  Matrix3 msgw( msg, true );
  Matrix1 sbuffer, zbuffer, bbuffer;
  Matrix1 msg1( msgw( iter.j1, iter.j2, _all ) );
#ifdef HAVE_SSE
  bool msg_aligned = sse_aligned( msg1.data() );
  bool r_aligned = sse_aligned( r1.data() );
#endif
  bool have_saliency = model._S.numElements() > 0;
  T s,z,b,gain,bit_llr,f0;
  T old_bit, new_bit;
  Matrix2 bit_messages;
  Matrix3 latent_messages;
  const bool linear_annealing =  model._linear_annealing;
  if ( linear_annealing )
	r1_old.reference( r1.copy() );
  else
	latent_messages.resize( model._d1, model._d2, model._k );
  if ( model._have_bits )
	{
	  bool have_temperature = !model._linear_annealing && model._current_rate != 1;
	  bit_llr = ( have_temperature ? model._current_rate : 1 ) * safe_log( model._mu_b0 ) - safe_log( T(1) - model._mu_b0 );
	  if ( linear_annealing  )
		old_bit = model._bits_marginal( iter.j1, iter.j2 );
	  else
		bit_messages.resize( model._d1, model._d2 );
	}
  // ...calculate the new message...
  r1 = safe_log( r1 ) - msg1; // go to log and divide out the old message ..
  msg1 = 0;
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  T s0 = have_saliency ? std::max( T(0), std::min( T(1), model._S( iter.i1, iter.i2 ) ) ) : 1;
	  Matrix1 x_i( dataw( iter.i1, iter.i2, _all ) );
	  zbuffer.reference( zbufferw( iter.i1, iter.i2, _all ) );
	  sbuffer.reference( sbufferw( iter.i1, iter.i2, _all ) );
	  upward_message_iterator_t p = derived.upward_message_iterator( iter.i1_rf(), iter.i2_rf(), x_i.data() );
	  s = sbuffer( iter.j_index() );
	  z = zbuffer( iter.j_index() );
	  if ( model._have_bits )
		{
		  bbuffer.reference( bbufferw( iter.i1, iter.i2, _all ) );
		  b = bbuffer( iter.j_index() );
		  f0 = sum( sbuffer * zbuffer * bbuffer ) - s * z * b;
		  T bit_message = safe_log( f0 + s * z ) - safe_log( f0 );
		  bit_llr += bit_message;
		  if ( !linear_annealing )
			bit_messages( iter.i1_rf(), iter.i2_rf() ) = bit_message;
		  gain = s * b;
		}
	  else
		{
		  f0 = sum( sbuffer * zbuffer ) - s * z;
		  gain = s;
		}
#ifdef HAVE_SSE
	  for ( int k4 = 0, k4max = model._k / 4; k4 < k4max; ++k4, p.advance(4) )
		{
		  vecf4 f_k4, msg_k4;
		  get_current_factor_sse( model, p, f_k4 );
		  msg_k4 = safe_log( s0 * ( s * f_k4 + f0 ) + 1.0f - s0 );
		  ( vecf4( msg1.data() + p.k, msg_aligned ) + msg_k4 ).store( msg1.data() + p.k, msg_aligned );
		  if ( !linear_annealing )
			msg_k4.store( &latent_messages( iter.i1_rf(), iter.i2_rf(), p.k ) );
		}
#endif
	  for ( ; p.k < model._k; p.advance(1) )
		{
		  T f_k, msg_k;
		  get_current_factor( model, p, f_k );
		  msg_k = safe_log( s0 * ( s * f_k + f0 ) + 1.0f - s0 );
		  msg1(p.k) += msg_k;
		  if ( !linear_annealing )
			latent_messages( iter.i1_rf(), iter.i2_rf(), p.k ) = msg_k;
		}
	}	
  r1 += msg1; // ...multiply in the new message...
  normalize_rho( r1 ); // ..and  go back to linear
  if ( linear_annealing )
	{
	  r1 = ( 1 - model._current_rate ) *r1_old + model._current_rate * r1;
	}
  if ( model._have_bits )
	{
	  new_bit = sigmoid( bit_llr );
	  if ( linear_annealing )
		new_bit = ( 1 - model._current_rate ) * old_bit + model._current_rate * new_bit;
	  model._bits_marginal( iter.j1, iter.j2 ) = new_bit;
	}
  
  // === pass2: update gates given r ===
  
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  Matrix1 x_i( dataw( iter.i1, iter.i2, _all ) );
	  upward_message_iterator_t p = derived.upward_message_iterator( iter.i1_rf(), iter.i2_rf(), x_i.data() );
#ifdef HAVE_SSE
	  vecf4 z4 = 0.0f, Z4 = 0.0f;
	  for ( int k4 = 0, k4max = model._k / 4; k4 < k4max; ++k4, p.advance(4) )
		{
		  vecf4 f_k4;
		  get_current_factor_sse( model, p, f_k4 );
		  vecf4 r4( r1.data() + p.k, r_aligned );
		  if ( !linear_annealing )
			{
			  vecf4 msg_k4( &latent_messages( iter.i1_rf(), iter.i2_rf(), p.k ) );
			  r4 = exp( safe_log( r4 ) - msg_k4 );
			}
		  z4 += r4 * f_k4;
		  Z4 += r4;
		}
	  float z = z4.sum();
	  T Z = Z4.sum();
#else
	  T z = 0;
	  T Z = 0;
#endif
	  for ( ; p.k < model._k; p.advance(1) )
		{
		  T f_k, r, msg_k;
		  get_current_factor( model, p, f_k );
		  msg_k = latent_messages( iter.i1_rf(), iter.i2_rf(), p.k );
		  r = r1(p.k);
		  if ( !linear_annealing )
			r = aexpf( safe_log( r ) - msg_k );
		  Z += r;
		  z += r1(p.k) * f_k;
		}
	  zbufferw( iter.i1, iter.i2, iter.j_index() ) = z / Z;
	}

  if ( model._have_bits )
	{
	  if ( linear_annealing )
		bbufferw( iter.in_range1(), iter.in_range2(), iter.j_index() ) = new_bit;
	  else
		bbufferw( iter.in_range1(), iter.in_range2(), iter.j_index() ) = sigmoid( bit_llr - bit_messages );
	}
}


template<typename Derived>
typename TypeTraits<Derived>::num_type update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::foreground_probability( MixtureModel2Base<Derived> & model, Matrix1 & data, Matrix1 & r, int i, int j, int irf, int jrf )
{
  if ( model._have_background_model )
	{
	  T lp_b1 = background_message( model, data, i, j );
	  T lp_b0 = foreground_message( model, data, r, irf, jrf );
	  // todo: subtract the corresponding expected kl
	  return T(1) / ( T(1) + aexpf( lp_b1 - lp_b0 ) );
	}
  else if ( model._S.numElements() > 0 )
	{
	  return model._S( i, j );
	}
  else
	return 1;
}

template<typename Derived>
typename TypeTraits<Derived>::num_type update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::
background_message( MixtureModel2Base<Derived> & model, const Matrix1 & data, int i, int j ) 
{
  Derived & derived = model.derived();
  upward_message_iterator_t p = derived.background_message_iterator( data.data() );
  T lp_bg = p.evaluate_log();
  if ( model._S.numElements() > 0 )
	{
	  T s0 = std::max( T(0), std::min( T(1), model._S( i, j ) ) );
	  T ls0 = safe_log( s0 ) - safe_log( 1 - s0 );	  
	  lp_bg -= ls0;
	}
  return lp_bg;
}

template<typename Derived>
typename TypeTraits<Derived>::num_type update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::foreground_message( MixtureModel2Base<Derived> & model, Matrix1 & data, Matrix1 & r, int irf, int jrf ) 
{
  BLITZ_INDICES;
  Derived & derived = model.derived();
  upward_message_iterator_t p = derived.upward_message_iterator( irf, jrf,data.data() );
  return safe_log( eval_likelihood( p, r ) );
}

template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::build_downward_message( MixtureModel2Base<Derived> & model, Matrix3 & data, Matrix3 & r, Matrix3 & msg )
{
  BLITZ_INDICES;
  bool have_saliency = model._S.numElements() > 0;
  Matrix3 rw( r, true );
  Matrix3 zbufferw( model._zbuffer, true );
  Matrix3 sbufferw( model._sbuffer, true );
  Matrix3 bbufferw( model._bbuffer, true );
  Matrix3 dataw( data, true );
  Matrix3 msgw( msg, true );
  Derived & derived = model.derived();
  Matrix1 r_loo(model._k), bbuffer;
  T gain, z, s, b;
  for ( InverseRFIterator iter( msg.extent(0), msg.extent(1), 
								model._d1, model._d2, 
								model._step1, model._step2 );
		!iter.atend_i(); iter.increment_i() )
	{
	  Matrix1 x_i( dataw( iter.i1, iter.i2, _all ) );
	  Matrix1 msg_i( msgw( iter.i1, iter.i2, _all ) );
	  Matrix1 zbuffer( zbufferw( iter.i1, iter.i2, _all ) );
	  Matrix1 sbuffer( sbufferw( iter.i1, iter.i2, _all ) );
	  if ( model._have_bits )
		{
		  bbuffer.reference( bbufferw( iter.i1, iter.i2, _all ) );
		}

	  T Z_i = 0;
	  msg_i = 0;
	  
	  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		{
		  Matrix1 r_j( rw( iter.j1, iter.j2, _all ) );
		  s = sbuffer( iter.j_index() );
		  z = zbuffer( iter.j_index() );
		  if ( model._have_bits )
			{
			  b = bbuffer( iter.j_index() );
			  gain = s * b * z;
			}
		  else
			{
			  gain = s * z;
			}
		  Matrix2 mu_j = derived.get_mur( iter.i1_rf(), iter.i2_rf() );
		  for ( int m = 0; m < model._m; ++m )
			{							
#ifdef HAVE_SSE
			  msg_i(m) += gain * sse_dp1( &mu_j( m, 0 ), r_j.data(), model._k, sse_aligned( r_j.data() ) && sse_aligned( &mu_j( m, 0 ) ) );
#else
			  for ( int k = 0; k < model._k; ++k ) msg_i(m) += gain * mu_j( m, k ) * r_j(k);
#endif
			}
		  Z_i += gain * sum( r_j );
		} // parents iteration
	  msg_i *= T(1) / Z_i;
	}
  
  
}

/// belporp message init for gated network
template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type >::init_interactions
( MixtureModel2Base<Derived> & model, Matrix3 & data_, Matrix3 & latent_prior, InverseRFIterator & iter )	
{
  BLITZ_INDICES;
  Matrix3 data( data_, true );
  Derived & derived = model.derived();
  Matrix3 zbufferw( model._zbuffer, true );
  Matrix3 bbufferw( model._bbuffer, true );
  Matrix1 zbuffer, bbuffer, x_i;

  // calculate interactions ( into zbuffer ) and old messages
  T * p_latent_prior = latent_prior.data();
  const bool latent_prior_singular = ( ( latent_prior.extent(0) == 1 ) && ( latent_prior.extent(1) == 1 ) );
  const bool have_latent_prior = latent_prior.numElements() > 0;
  const bool latent_prior_complete = !latent_prior_singular && have_latent_prior;
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  x_i.reference( data( iter.i1, iter.i2, _all ) );
	  zbuffer.reference( zbufferw( iter.i1, iter.i2, _all ) );
	  if ( model._have_bits )
		bbuffer.reference( bbufferw( iter.i1, iter.i2, _all ) );

	  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		{
		  if ( latent_prior_complete ) p_latent_prior = &latent_prior( iter.j1, iter.j2, 0 );
		  upward_message_iterator_t p = derived.upward_message_iterator( iter.i1_rf(), iter.i2_rf(), x_i.data() );

		  // z is the marginalization of mu_(z->f) * f over the latent (z) state
		  // a prior is always assumed, in absence it is set to 1/k
		  // b contains the messages from the bits
		  // z * b * s makes the gate distrib ( after normalization )

#ifdef HAVE_SSE
		  vecf4 z4 = 0.0f, f4;
		  if ( have_latent_prior )
			{
			  bool latent_prior_aligned = sse_aligned( p_latent_prior );
			  for ( int k4 = 0, k4max = model._k / 4; k4 < k4max; ++k4, p.advance(4) )
				{
				  get_current_factor_sse( model, p, f4 );
				  z4 += vecf4( p_latent_prior + p.k, latent_prior_aligned ) * f4;
				}
			}
		  else
			for ( int k4 = 0, k4max = model._k / 4; k4 < k4max; ++k4, p.advance(4) )
				{
				  get_current_factor_sse( model, p, f4 );
				  z4 += ( 1.0f / model._k ) * f4;
				}
		  float z = z4.sum();
#else
		  T z = 0;
#endif
		  T f;
		  if ( have_latent_prior )
			for ( ; p.k < model._k; p.advance(1) )
			  {
				get_current_factor( model, p, f );
				z += p_latent_prior[p.k] * f;
			  }
		  else
			for ( ; p.k < model._k; p.advance(1) )
			  {
				get_current_factor( model, p, f );
				z += f * ( T(1) / model._k );
			  }
		  zbuffer( iter.j_index() ) = z;

		  // at init there is not contrib of the bits beyond the prior, does this make a difference?
		  if ( model._have_bits )
			bbuffer = model._mu_b0;

		}
	}
}

/// belprop message update for gated network
template<typename Derived>
void update_gated_msg_delegate<
  Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::update_interactions( MixtureModel2Base<Derived> & model, Matrix3 & data, Matrix3 & r, InverseRFIterator & iter )
{
  BLITZ_INDICES;
  bool have_saliency = model._S.numElements() > 0;
  Matrix3 rw( r, true );
  Matrix3 zbufferw( model._zbuffer, true );
  Matrix3 sbufferw( model._sbuffer, true );
  Matrix3 bbufferw( model._bbuffer, true );
  Matrix3 dataw( data, true );
  Matrix1 zbuffer, sbuffer, bbuffer, x_i;
  Derived & derived = model.derived();
  T Z_old, z_old, s_old, b_old, f0_old, old_gain;
  bool linear_annealing = model._linear_annealing && model._current_rate != 1;
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{

	  // z is the marginalization of mu_(z->f) * f over the latent (z) state
	  // a prior is always assumed, in absence it is set to 1/k
	  // b contains the messages from the bits
	  // z * b * s makes the gate distrib ( after normalization )
	  T s0 = have_saliency ? clamp( model._S( iter.i1, iter.i2 ), 0, 1 ) : 1;
	  x_i.reference( dataw( iter.i1, iter.i2, _all ) );
	  zbuffer.reference( zbufferw( iter.i1, iter.i2, _all ) );
	  sbuffer.reference( sbufferw( iter.i1, iter.i2, _all ) );
	  if ( model._have_bits )
		{
		  bbuffer.reference( bbufferw( iter.i1, iter.i2, _all ) );
		  Z_old = sum( zbuffer * sbuffer * bbuffer );
		}
	  else
		Z_old = sum( sbuffer * zbuffer );

	  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		{
		  Matrix1 r_j( rw( iter.j1, iter.j2, _all ) );			
		  T * p_r = r_j.data();
		  z_old = zbuffer( iter.j_index() );
		  s_old = sbuffer( iter.j_index() );
		  if ( model._have_bits )
			{
			  b_old = bbuffer( iter.j_index() );
			  f0_old = std::max( T(0), Z_old - b_old * z_old * s_old ); 
			  old_gain = b_old * s_old;
			}
		  else
			{
			  f0_old = std::max( T(0), Z_old - z_old * s_old ); 
			  old_gain = s_old;
			}
		  T Zr = 0;
		  upward_message_iterator_t p = derived.upward_message_iterator( iter.i1_rf(), iter.i2_rf(), x_i.data() );
#ifdef HAVE_SSE
		  bool r_aligned = sse_aligned( p_r );
		  vecf4 z4 = 0.0f;
		  for ( int k4 = 0, k4max = model._k / 4; k4 < k4max; ++k4, p.advance(4) )
			{
			  vecf4 f_k4, f_k4_old;
			  get_current_factor_sse( model, p, f_k4 );
			  get_last_factor_sse( model, p, f_k4, f_k4_old );
			  vecf4 r4( p_r + p.k, r_aligned );
			  if ( !linear_annealing )
				r4 = exp( safe_log( r4 ) - safe_log( s0 * ( old_gain * f_k4_old + f0_old ) + T(1) - s0 ) );
			  z4 += f_k4 * r4;
			  Zr += r4.sum();
			}
		  float z = z4.sum();
#else
		  T z = 0;
#endif
		  for ( ; p.k < model._k; p.advance(1) )
			{
			  T f_k, f_k_old;
			  get_current_factor( model, p, f_k );
			  get_last_factor( model, p, f_k, f_k_old );
			  T r = p_r[p.k];
			  if ( !linear_annealing )				
				r = aexpf( safe_log( r ) - safe_log( s0 * ( old_gain * f_k_old + f0_old ) + T(1) - s0 ) );
			  Zr += r;
			  z += r * f_k;
			}
		  assert( Zr > T(1e-10) );
		  
#define update_message(mu,mu_new)										\
		  {																\
			T & mu_ref = mu;											\
			T mu_new_value = mu_new;									\
			mu_ref = mu_new_value;										\
		  }

		  // update mu(z_j->f)
		  update_message( zbuffer( iter.j_index() ), z / Zr );

		  if ( model._have_bits )
			{
			  T b = model._bits_marginal( iter.j1, iter.j2 );
			  T b_llr = safe_log( b ) - safe_log( T(1) - b );			  
			  T b_local_msg = safe_log( f0_old + z_old * s_old ) - safe_log( f0_old );

			  // update mu(b_j->f)
			  update_message( bbuffer( iter.j_index() ), sigmoid( b_llr - b_local_msg ) );

			}
		  
#undef update_message

		}
	}
}

template<typename Derived>
void update_gated_msg_delegate<
  Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::build_upward_message_gated( MixtureModel2Base<Derived> & model, Matrix3 & data, Matrix3 & msg, T rate, InverseRFIterator & iter )	
{
  BLITZ_INDICES;
  Matrix3 zbufferw( model._zbuffer, true );
  Matrix3 sbufferw( model._sbuffer, true );
  Matrix3 bbufferw( model._bbuffer, true );
  Matrix3 dataw( data, true );
  Matrix1 zbuffer, sbuffer, bbuffer, x_i;
  Matrix1 zbuffer_old, sbuffer_old, bbuffer_old;
  T Z, f0, gain, z, s, b;
  Derived & derived = model.derived();
  const bool have_saliency = model._S.numElements() > 0;
  Matrix2 bits_msg;
  bool have_temperature = !model._linear_annealing && model._current_rate != 1;
  if ( model._have_bits )
	{
	  T mu_l0 = ( have_temperature ? model._current_rate : 1 ) * ( safe_log( model._mu_b0 ) - safe_log( T(1) - model._mu_b0 ) );
	  bits_msg.resize( msg.extent(0), msg.extent(1) );
	  bits_msg = mu_l0;
	}

  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{

	  // fetch
	  T s0 = have_saliency ? clamp( model._S( iter.i1, iter.i2 ), 0, 1 ) : 1;
	  x_i.reference( dataw( iter.i1, iter.i2, _all ) );
	  zbuffer.reference( zbufferw( iter.i1, iter.i2, _all ) );
	  sbuffer.reference( sbufferw( iter.i1, iter.i2, _all ) ); 

	  // calc normalization and handle bits
	  if ( model._have_bits )
		{
		  bbuffer.reference( bbufferw( iter.i1, iter.i2, _all ) );
		  Z = sum( zbuffer * sbuffer * bbuffer );
		}
	  else
		{
		  Z = sum( sbuffer * zbuffer );
		}


	  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		{
		  upward_message_iterator_t p = derived.upward_message_iterator( iter.i1_rf(), iter.i2_rf(), x_i.data() );
		  Matrix1 msg_j( msg( iter.j1, iter.j2, _all ) );
		  z = zbuffer( iter.j_index() );
		  s = sbuffer( iter.j_index() );
		  if ( model._have_bits )
			{
			  b = bbuffer( iter.j_index() );
			  f0 = std::max( T(0), Z - b * z * s ); 
			  gain = b * s;
			  // mu(f->b) update
			  bits_msg( iter.j1, iter.j2 ) += safe_log( f0 + z * s ) - safe_log( f0 );
			}
		  else
			{
			  f0 = std::max( T(0), Z - z * s ); 
			  gain = s;
			}

		  // functional-fu excersize: formulate the loop body in a single lambda and fire sse and scalar loops with that

		  // mu(f->z) update
#ifdef HAVE_SSE
		  bool msg_aligned = sse_aligned( msg_j.data() );
		  for ( int k4 = 0, k4max = model._k / 4; k4 < k4max; ++k4, p.advance(4) )
			{			  
			  vecf4 f_k4;
			  get_current_factor_sse( model, p, f_k4 );
			  float * p_msg_k4 = msg_j.data() + p.k;
			  vecf4 local_msg = s0 * ( gain * f_k4 + f0 ) + 1.0f - s0;
			  ( vecf4( p_msg_k4, msg_aligned ) + safe_log( local_msg ) ).store( p_msg_k4, msg_aligned );
			}
#endif
		  for ( ; p.k < model._k; p.advance(1) )
			{
			  T f_k;
			  get_current_factor( model, p, f_k );
			  T local_msg = s0 * ( gain * f_k + f0 ) + T(1) - s0;
			  msg_j(p.k) += safe_log( local_msg );
			}
		}
	}  

  if ( model._have_bits )
	model._bits_marginal = sigmoid( bits_msg );
}

#ifdef HAVE_SSE
static inline vecf4 smooth_message1(vecf4 x,vecf4 s) 
{ 
  vecf4 m = max( x, s );
  return safe_log( exp( x - m ) + exp( s - m ) ) + m;
}
#endif

static inline float smooth_message1(float x, float s)
{
  float m = std::max( x, s );
  return safe_log( aexpf( x - m ) + aexpf( s - m ) ) + m;
}

static inline double smooth_message1(double x, double s)
{
  double m = std::max( x, s );
  return safe_log( aexpf( x - m ) + aexpf( s - m ) ) + m;
}


template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type >::build_upward_message_with_saliency( MixtureModel2Base<Derived> & model, Matrix3 & data, Matrix3 & msg, InverseRFIterator & iter )	
{
  BLITZ_INDICES;
  Derived & derived = model.derived();
  msg = 0;
  
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  Matrix1 x_i( data( iter.i1, iter.i2, _all ) );
	  T lp_bg = background_message( model, x_i, iter.i1, iter.i2 );
#ifdef HAVE_SSE
	  vecf4 lp_bg4( lp_bg );
#endif
	  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		{
		  upward_message_iterator_t p = derived.upward_message_iterator( iter.i1_rf(), iter.i2_rf(), x_i.data() );
		  Matrix1 msg_j( msg( iter.j1, iter.j2, _all ) );
#ifdef HAVE_SSE
		  bool msg_aligned = sse_aligned( msg_j.data() );
		  for ( int k4 = 0, k4max = model._k / 4; k4 < k4max; ++k4, p.advance(4) )
			smooth_message1( p.evaluate_log_sse(), lp_bg4 ).store( msg_j.data() + p.k, msg_aligned );
#endif
		  for ( ; p.k < model._k; p.advance(1) )
			msg_j(p.k) = smooth_message1( p.evaluate_log(), lp_bg );
		}
	}  
}

template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type >::build_downward_message_sampled( MixtureModel2Base<Derived> & model, IMatrix2 & latents_sample, Matrix3 & msg )
{
  BLITZ_INDICES;
  IMatrix2 gates_sample( model._gates_sample, true );
  Derived & derived = model.derived();
  Matrix3 msgw( msg, true );
  InverseRFIterator iter( msg.extent(0), msg.extent(1), model._d1, model._d2, model._step1, model._step2 );
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  Matrix1 msg_i( msgw( iter.i1, iter.i2, _all ) );
	  int gate_sample = gates_sample( iter.i1, iter.i2 );
	  iter.select_j_index( gate_sample );
	  unsigned int latent_sample = latents_sample( iter.j1, iter.j2 );
	  Matrix2 mu_j = derived.get_mur( iter.i1_rf(), iter.i2_rf() );
	  Matrix1 mu_jk( mu_j( _all, latent_sample ) );
	  msg_i = mu_jk;
	}
}

template<typename Derived>
typename TypeTraits<Derived>::num_type
evaluate_linear_upward_at( Derived & derived
						, const typename TypeTraits<Derived>::num_type * x
						, int i1rf, int i2rf, int k )
{
  typename Derived::upward_message_iterator_t p = derived.upward_message_iterator( i1rf, i2rf, x );
  p.advance( k );
  return std::max( typename TypeTraits<Derived>::num_type(1e-10), p.evaluate_linear() );
}

template<typename Derived>
typename TypeTraits<Derived>::num_type
update_gated_msg_delegate< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type >
::evaluate_log_upward_at( Derived & derived
							 , const typename TypeTraits<Derived>::num_type * x
							 , int i1rf, int i2rf, int k )
{
  typename Derived::upward_message_iterator_t p = derived.upward_message_iterator( i1rf, i2rf, x );
  p.advance( k );
  return p.evaluate_log();
}

template<typename Derived>
bool
check_gate_neighbor( IMatrix2 & gates_sample
					 , MixtureModel2Base<Derived> & model
					 , int ioff, int joff
					 , InverseRFIterator & iter )
{
  int n1 = gates_sample.extent(0);
  int n2 = gates_sample.extent(1);
  int i1 = iter.i1 + ioff;
  int i2 = iter.i2 + joff;
  // is the neighbor in the input range ?
  if ( ( i1 < 0 ) || ( i2 < 0 ) || ( i1 >= n1 ) || ( i2 >= n2 ) ) return false;
#if 1
  unsigned sample = gates_sample( i1, i2 );     
  int j1 = model.j_index_to_j1( sample, i1 );
  int j2 = model.j_index_to_j2( sample, i2 );
  int i1rf = iter.i1 - j1 * model._step1;
  int i2rf = iter.i2 - j2 * model._step2;
  //are we in the neighbors receptive field?
  if ( i1rf < 0 || i2rf < 0 || i1rf >= model._d1 || i2rf >= model._d2 ) return false;
#endif
  return true;
}

template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::sample_bits( MixtureModel2Base<Derived> & model, 
				  Matrix3 & data,
				  IMatrix2 & latents_sample,
				  ForwardRFIterator & iter )
{
  BLITZ_INDICES;
  Random<T> & rng = Random<T>::get();
  Derived & derived = model.derived();
  IMatrix2 gates_sample( model._gates_sample, true );
  BMatrix2  bits_sample( model. _bits_sample, true );
  const bool have_saliency = model._S.numElements() > 0;
  // seems wasteful, more linear would be to sweep data space and distribute llrs
  // would be tricky with threads though, since the write pattern is random over the 
  // "inverse RF"
  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
	{
	  T llr = safe_log( model._mu_b0 ) - safe_log( 1 - model._mu_b0 );			
	  int k_j = latents_sample( iter.j1, iter.j2 );
	  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
		{
		  // messages from datapoints
		  if ( int(gates_sample( iter.i1, iter.i2 )) == iter.j_index() )
			{
			  if ( model._have_background_model )
				{
				  Matrix1 x_i( data( iter.i1, iter.i2, _all ) );
				  T s0 = have_saliency ? 
					std::max( T(0), std::min( T(1), model._S( iter.i1, iter.i2 ) ) ) : 1;
				  llr += evaluate_log_upward_at( derived, x_i.data(), 
												 iter.i1_rf(), iter.i2_rf(), 
												 k_j ) * s0; // foreground-term
				  llr -=model._background_likelihoods( iter.i1, iter.i2 );
				}
			  else
				llr -= T(-1e100); // force on
			}
		}
	  //FIXME: monitor stuff, cannot control sparsity even with bg
	  bits_sample( iter.j1, iter.j2 ) = rng.sample_binomial( sigmoid( llr ) );
	}
  OPENMP(barrier);
}

template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::sample_mask( MixtureModel2Base<Derived> & model
				  , Matrix3 & data, IMatrix2 & latents_sample
				  , InverseRFIterator & iter ) 
{
  // FIXME: rethink this...
  BLITZ_INDICES;
  IMatrix2 gates_sample( model._gates_sample, true );
  BMatrix3 mask_sample ( model._mask_sample , true );
  BMatrix2 bits_sample ( model._bits_sample , true );
  Matrix3 mu_mask( model._mu_mask, true );
  Random<T> & rng = Random<T>::get();
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  int gate = gates_sample( iter.i1, iter.i2 );
	  BMatrix1 m_i( mask_sample( iter.i1, iter.i2, _all ) );
	  // count active masks of inactive parents
	  m_i( gate ) = true;
	  int N_i = sum( where( m_i(i1), 1, 0 ) );
	  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
		{
		  // check if the corresponding parent is seeing data from here...
		  int parent = iter.j_index(); if ( parent == gate ) continue;
		  // ...switch this bit off...
		  bool m_old = m_i( parent ), m_new;
		  if ( m_old ) --N_i;
		  if ( model._have_bits && !bits_sample( iter.j1, iter.j2 ) && !model._and_combine_bits_and_gates )
			{
			  m_new = false;
			}
		  else
			{
			  // ...fetch its prior from the parent latent...
			  int latent = latents_sample( iter.j1, iter.j2 );
			  T mu_jk = binomial_combine( mu_mask( iter.i1_rf(), iter.i2_rf(), latent ), model._mu_m_prior );
			  T mu_ij = mu_jk * T( N_i ) / ( T( 1 + N_i ) -  mu_jk );
			  // ...flip it according to the rescaled prior
			  m_new = rng.sample_binomial( mu_ij ); 
			}
		  if ( m_new ) ++N_i;
		  m_i( parent ) = m_new;
		}
	}
  OPENMP(barrier);
}

template<typename Derived>
 void update_gated_msg_delegate
< Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
  >::sample_gates( MixtureModel2Base<Derived> & model, Matrix3 & data
				   ,IMatrix2 & latents_sample, InverseRFIterator & iter
				   ,bool cache_z, bool init )
{
  IMatrix2 gates_sample( model._gates_sample, true );
  BMatrix2 bits_sample ( model._bits_sample , true );
  int max_n_parents = ( ( model._d1 + model._step1 - 1 ) / model._step1 ) 
	* ( ( model._d2 + model._step2 - 1 ) / model._step2 );
  Matrix1 s( max_n_parents );
  BLITZ_INDICES;
  // I think we can ignore saliency because its constant for the gate...
  // ...not quite: the influence of the parents is smoothed over non-salienc locations
  bool have_zbuffer = ( !cache_z ) && ( model._zbuffer.numElements() > 0 );
  Matrix3 dataw( data, true );
  Matrix3 zbuffer( model._zbuffer, true );
  Random<T> & rng = Random<T>::get();
  double z_accum = 0, z_count = 0;
  Matrix2 clustering_message;
  if ( model._cluster_gates && !init )
  {
      // FIXME: not threadsafe
      // collect dirichlet counts for each output
      blitz::TinyVector<int,2> s2;
      boost::tie(s2[0],s2[1]) = model.calc_outsize(gates_sample.extent(0),gates_sample.extent(1));
      clustering_message.resize( s2 );
      clustering_message = model._gate_clustering_alpha0;
      for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
      {
		    int old_gate = gates_sample( iter.i1, iter.i2 );
        iter.select_j_index( old_gate );
        clustering_message( iter.j1, iter.j2 ) += 1;
      }
      clustering_message = safe_log( clustering_message );
  }
  // visit each gate in sequence, first odd then even rows with barrier
  // this vistation pattern ensures neighboring lines are not being updated concurrently
  for ( int parity1 = 0; parity1 <= 1; ++parity1 )
	{
  for ( int parity2 = 0; parity2 <= 1; ++parity2 )
	{
	  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
		{
		  // reduce sampling order influence and avoid bordering with other threads
		  if ( iter.i1 & parity1 ) continue;
		  if ( iter.i2 & parity2 ) continue;
		  
		  int old_gate = gates_sample( iter.i1, iter.i2 );
		  T l0 = -std::numeric_limits<T>::max();

		  // collect data likelihood under each parent
		  if ( have_zbuffer )
			{
			  s = zbuffer( iter.i1, iter.i2, _all );
			}
		  else
			{
			  s = l0;
			  bool found_old_gate = false;
			  bool bit = true;			  
			  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
				{					
				  int j_index = iter.j_index();
				  if ( model._have_bits ) bit = model._bits_sample( iter.j1, iter.j2 );
				  if ( j_index == old_gate ) found_old_gate = true;
				  s( j_index ) = parent_likelihood( model, latents_sample( iter.j1, iter.j2 ), bit, dataw, iter );
				}
			  if ( cache_z )
				zbuffer( iter.i1, iter.i2, _all ) = s;
			}
			
			// include the clustering dirichlet message
			if ( model._cluster_gates && !init )
			  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
          s(iter.j_index()) += clustering_message( iter.j1, iter.j2 );
			  
		  normalize_rho( s );

		  // incorporate gate_mrf
		  T Zs = 1;
		  if ( model._gate_neighborhood != MixtureModel2Base<Derived>::NO_NEIGHBORS && !init )
			{
#define handle_neighbor( ioff, joff )									\
			  if ( check_gate_neighbor( gates_sample, model, ioff, joff, iter ) ) \
				{ int j = gates_sample( iter.i1 + ioff, iter.i2 + joff ); \
				  T new_s = s(j) * model._gate_W_neighbor;				\
				  Zs += new_s - s(j); s(j) = new_s; }
			  handle_neighbor( -1,  0 );
			  handle_neighbor(  1,  0 );
			  handle_neighbor(  0, -1 );
			  handle_neighbor(  0,  1 );
			  if ( model._gate_neighborhood == MixtureModel2Base<Derived>::EIGHT_NEIGHBORS )
				{
				  handle_neighbor( -1,  1 );
				  handle_neighbor( -1, -1 );
				  handle_neighbor(  1,  1 );
				  handle_neighbor(  1, -1 );
				}
#undef handle_neighbor
			}
		  // sample from normalized collected likelihood vector
		  int g = rng.sample_multinomial( s, Zs );
		  z_accum += s(g) / Zs;
		  z_count += 1;
		  //std::cout << " g " << g << " p " << s(g) << ", " << " p in " << min(s) << ", " << mean(s) << ", " << max(s);
		  gates_sample( iter.i1, iter.i2 ) = g;
		}
	}
  // sync when switching line parirty to avoid threads accessing neighboring lines
	  OPENMP(barrier);
	}
  if ( model._report_convergence ) std::cout << " <s(g)>/#(g) = " << ( z_accum / z_count ) << std::endl;
}

template<typename Derived>
void update_gated_msg_delegate< 
  Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
>::sample_latents( MixtureModel2Base<Derived> & model, 
				   Matrix3 & data, IMatrix2 & latents_sample, 
				   Matrix3 lr0, ForwardRFIterator & iter )
{
  BLITZ_INDICES;
  IMatrix2 gates_sample( model._gates_sample, true );
  Derived & derived = model.derived();
  const bool have_saliency = model._S.numElements() > 0;
  Matrix3 lmu_combined( model._lmu_combined, true );
  Matrix3 lnotmu_combined( model._lnotmu_combined, true );
  Matrix1 r( model._k );
  Random<T> & rng = Random<T>::get();
  Matrix3 lr0w( lr0, true );
  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
	{
	  // build multinomial distrib of this latent given its neighbors and the gates
	  if ( lr0.numElements() == 0 )
		r = 0;
	  else if ( ( lr0.extent(0) == 1 ) && ( lr0.extent(1)  == 1 ) )
		r = lr0w( 0, 0, _all );
	  else
		r = lr0w( iter.j1, iter.j2, _all );
	  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
		{
		  int gate = gates_sample( iter.i1, iter.i2 ), parent = iter.j_index();
		  if ( gate == parent )
			{
			  bool bit = model._have_bits ? model._bits_sample( iter.j1, iter.j2 ) : true;
			  for ( int k = 0; k < model._k; ++k )
				r(k) = parent_likelihood( model, k, bit, data, iter );
			}
		}
	  T Zr = normalize_rho( r, false );
	  // sample from it
	  latents_sample( iter.j1, iter.j2 ) = rng.sample_multinomial( r, Zr );
	}
}


// new sampling strategy:
// step:
// visit an RF at random
// for each latent setting ( including off state if having bits ) sample associated auxiliaries
// record likelihood change for each 
// sample from original state and each new state using multinomial loaded with llrs
// need a snapshop of each setting:
// - latent
// - bit
// - gates

template<typename Derived, typename Iter>
typename TypeTraits<Derived>::num_type 
parent_likelihood( MixtureModel2Base<Derived> & model, 
				   int k_j, bool bit, Matrix3 & data, 
				   Iter & iter )
{
  BLITZ_INDICES;
  typedef typename TypeTraits<Derived>::num_type T;
  typedef update_gated_msg_delegate<Derived> delegate;
  T l_bg = model._have_background_model ? model._background_likelihoods( iter.i1, iter.i2 ) : -std::numeric_limits<T>::max();
  bool fg = true;
  Matrix1 x_i( data( iter.i1, iter.i2, _all ) );
  if ( model._have_bits && ( !model._dependent_gates || model._and_combine_bits_and_gates ) )
	fg &= bit;
  if ( fg )
	{
	  bool have_saliency = model._S.numElements() > 0;
	  T s0 = have_saliency ? std::max( T(0), std::min( T(1), model._S( iter.i1, iter.i2 ) ) ) : 1;
	  T l_data = delegate::evaluate_log_upward_at( model.derived(), x_i.data(), 
											  iter.i1_rf(), iter.i2_rf(), k_j ) * s0;
	  if ( model._dependent_gates )
		{
		  T mu_jk = binomial_combine( model._mu_mask( iter.i1_rf(), iter.i2_rf(), k_j ), model._mu_m_prior );
		  T l_mask = safe_log( mu_jk );
		  return l_data + l_mask;
		}
	  else
		return l_data;
	}
  else
	return l_bg;
}

/// samples one receptive field, i.e. the latent, the bit, and contained gates
/// done using one mcmc-step
/// one proposal is generated for each setting of the latent and the bit
/// by block-gibbs-sampling the gates for this setting
/// the llr of each proposal against the original settings is recoreded
/// from the proposal llrs and the original setting ( llr := 0 ) a multinomial
/// is generated and the new setting is sampled from that
/// sampling the original setting is equivalent to rejection
/// question: do I need rejection, and in any case is this a valid mcmc move?
///           anwer: no, because correct conditionals are used in the proposal
template<typename Derived>
 void update_gated_msg_delegate< 
  Derived, typename boost::enable_if< has_upward_message_iterator<Derived> >::type 
>::sample_rf( MixtureModel2Base<Derived> & model, 
			  Matrix3 & data, IMatrix2 & latents_sample, 
			  typename MixtureModel2Base<Derived>::position pos, 
			  rf_sampling_context & context )
{
  // todo: masks
  BLITZ_INDICES;

  Random<T> & rng = Random<T>::get();

  stopwatch init_timer, proposal_timer;

  ForwardRFIterator iter( data.extent(0), data.extent(1), model._d1, model._d2, model._step1, model._step2, true );
  iter.move_to( pos.row, pos.col );
  int j_index = iter.j_index();
  blitz::Range r1 = iter.in_range1();
  blitz::Range r2 = iter.in_range2();
  Matrix3 zbuffer( model._zbuffer, true );
  IMatrix2 gates( model._gates_sample, true );

  context.init_timer.tic();

  context.init( model._k, model._d1, model._d2, model._have_bits );
  rf_samples_t & rf_samples = context.samples;
  Matrix1 & llrs = context.llrs;

  context.init_timer.toc();

  // latent prior
  Matrix3 lr0w( model._lr0, true );
  Matrix1 latent_prior;
  if ( ( lr0w.extent(0) == 1 ) && ( lr0w.extent(1)  == 1 ) )
    latent_prior.reference( lr0w( 0, 0, _all ) );
  else if ( lr0w.numElements() > 0 )
    latent_prior.reference( lr0w( pos.row, pos.col, _all ) );

  // todo: do not switch gates to non-existing parents or those with their bit off

  context.proposal_timer.tic();

  int k_end = model._have_bits ? ( model._k + 1 ) : ( model._k );

  for ( int k = 0; k < k_end; ++k ) // one extra for the off state
	{
	  T & llr = llrs(k);
	  llr = 0;
	  rf_sample & current_sample = rf_samples[k];
	  current_sample.bit = k < model._k;
	  current_sample.latent = current_sample.bit ? k : 0;
	  // priors
	  if ( model._have_bits )
		{
		  llr += current_sample.bit ? safe_log( model._mu_b0 ) : safe_log( 1 - model._mu_b0 );
		}
	  if ( current_sample.bit && latent_prior.extent(0) > 0 )
		llr += latent_prior(k);
	}
  
  for ( iter.reset_i(); !iter.atend_i(); iter.increment_i() )
	{
	  Matrix1 l( zbuffer( iter.i1, iter.i2, _all ) );
	  l( j_index ) = -std::numeric_limits<T>::max(); // disallow the "here" state as this is being reconsidered anyhow
	  T l_off = max( l );
	  T Z, l_other;
	  if ( l_off > -std::numeric_limits<T>::max() )
		{
		  Z = sum( where( l - -std::numeric_limits<T>::max() > l_off, aexpf( l - l_off ), T(0) ) ); // be safe as float_min - x for x > 0 overflows
		  l_other = safe_log( Z ) + l_off;
		}
	  else
		{
		  Z = 0;
		  l_other = l_off;
		}
	  // sample a gate proposal
	  int gate_proposal, old_gate = gates( iter.i1, iter.i2 );
	  if ( l_off == -std::numeric_limits<T>::max() )
		{
		  gate_proposal = 0;
		}
#if 0 // this path may not be to great for mixing
	  else if ( old_gate != j_index )
		{
		  gate_proposal = old_gate;
		}
#endif
	  else
		{
		  do
			gate_proposal = rng.sample_log_multinomial( l, l_off, Z );
		  while ( l( gate_proposal ) == -std::numeric_limits<T>::max() );
		}
	  
	  for ( int k = 0; k < k_end; ++k ) // one extra for the off state
		{
		  rf_sample & current_sample = rf_samples[k];		  
		  T & llr = llrs(k);
		  if ( llr == -std::numeric_limits<T>::max() ) continue;
		  T l_here = parent_likelihood( model, current_sample.latent, current_sample.bit, data, iter );
		  T l_new;
		  int new_gate;
		  // pre-branch: sample wether the gate goes here or "away"
		  if ( current_sample.bit && ( l_other == -std::numeric_limits<T>::max() ||
		                               rng.sample_binomial( sigmoid( l_here - l_other ) ) )
			   )
			{
			  new_gate = j_index;
			  l_new = l_here;
			}
		  else if ( l_other != -std::numeric_limits<T>::max() )
			{
			  new_gate = gate_proposal;
			  l_new = zbuffer( iter.i1, iter.i2, new_gate );
			}
		  else
			{
			  llr = -std::numeric_limits<T>::max();
			}
		  
		  if ( llr != -std::numeric_limits<T>::max() )
			{
			  int i1_rf = iter.i1_rf();
			  int i2_rf = iter.i2_rf();
			  llr += l_new;
			  current_sample.gates( i1_rf, i2_rf ) = new_gate;
			  current_sample.parent_likelihoods( i1_rf, i2_rf ) = l_here;
			}
		}
	}

  context.proposal_timer.toc();

  context.choice_timer.tic();

  // apply the choice
  // std::cout << " float min " << -std::numeric_limits<T>::max() << std::endl;
  int choice;
  //std::cout << llrs << std::endl;
  assert( max( llrs ) != -std::numeric_limits<T>::max() );
  do 
    choice = rng.sample_log_multinomial( llrs );
  while ( llrs(choice) == -std::numeric_limits<T>::max() );
  rf_sample & chosen_sample = rf_samples[choice];
  latents_sample( pos.row, pos.col ) = chosen_sample.latent;
  if ( model._have_bits )
	model._bits_sample( pos.row, pos.col ) = chosen_sample.bit;
  gates( r1, r2 ) = chosen_sample.gates;
  zbuffer( r1, r2, j_index ) = chosen_sample.parent_likelihoods;

  context.choice_timer.toc();

}

#undef clamp

#define INSTANTIATE_GATED_MESSAGE_DELEGATE(Derived)			\
  template struct update_gated_msg_delegate				\
  < Derived, boost::enable_if< has_upward_message_iterator< Derived > >::type >
