/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
template<typename Derived>
void MixtureModel2Base<Derived>::accumulate_hessian( const Matrix3 & data )
{
  BLITZ_INDICES;
  Matrix2 Sw( _S, true );
  Matrix2 s;
  Matrix3 r( message_component<T>::get_belief( ), true );
  if ( ( _gradient_mode == LIKELIHOOD || _gradient_mode == UNLIKELIHOOD ) && _current_output_hessian.numElements() == 0 )
	calculate_likelihood_detector_hessian( data );
  Matrix3 output_hessian( _current_output_hessian, true );
  emission_gradient_t emission_hessian;
  BOOST_FOREACH( Matrix4 & hessian, emission_hessian )
	{
	  hessian.resize( _k, _d1, _d2, _m );
	  hessian = 0;
	}
  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1, _step2);!it.atend(); ++it)
	{
	  DISPATCH(calculate_emission_hessian)( it.in(), output_hessian( it.out_x(), it.out_y(), _all ), emission_hessian );
	}
  OPENMP(single)
	_hessian_estimation_counter += r.extent(0) * r.extent(1);
  OPENMP(critical)
  {
	for ( size_t i = 0; i < emission_hessian.size(); ++i )
	  _discriminative_emission_hessian[i] += emission_hessian[i];
	_output_hessian += sum( sum( output_hessian, i3 ), i2 );
  }
}

template<typename Derived>
void MixtureModel2Base<Derived>::add_output_hessian( const Matrix3 & delta, GradientMode kind )
{
  Matrix3 r( message_component<T>::get_belief( ), true );
  if ( _current_output_hessian.numElements() == 0 )
	{
	  _current_output_hessian.resize( r.shape() );		
	  _current_output_hessian = 0;
	}
  BLITZ_INDICES;
#define kronecker(i1,i2) where(i1==i2,T(1),T(0))
  switch( kind )
	{
	case CANONICAL: // canonical effectively means in terms of the argument of the transfer function (sigmoid/softmax), i.e. fall-through
	  _current_output_hessian += delta;
	  break;
	case BACKPROP: // in terms of the output, pull through transfer function jacobian
	  if ( message_component<T>::buffer().type() == mmm2_belief_state<T>::MULTINOMIAL && !_diagonal_output_jacobian )
		_current_output_hessian += sum( delta(i1,i2,i4) * pow2( kronecker(i3,i4) * r(i1,i2,i3) - r(i1,i2,i3) * r(i1,i2,i4) ), i4 ); 
	  // proper jacobian backprop respects the inhibitive interactions
	  else
		_current_output_hessian += delta * pow2( r - pow2( r ) );
	  break;
	case LIKELIHOOD: 
	case UNLIKELIHOOD: 
	  {
	    Matrix2 delta_l( delta( _all, _all, 0 ) );
	    _current_output_hessian += delta_l(i1,i2) * pow2( r );
	  }
	  break;
	}
#undef kronecker
  
}

template<typename Derived>
void MixtureModel2Base<Derived>::add_output_gradient( const Matrix3 & delta, GradientMode kind )
{
  Matrix3 r( message_component<T>::get_belief( ), true );
  if ( _output_gradient.numElements() == 0 )
	{
	  _output_gradient.resize( r.shape() );		
	  _output_gradient = 0;
	}
  BLITZ_INDICES;
  switch( kind )
	{
	case CANONICAL:
	  _output_gradient += delta;
	  break;
	case BACKPROP:
	  if ( message_component<T>::buffer().type() == mmm2_belief_state<T>::MULTINOMIAL && !_diagonal_output_jacobian )
		_output_gradient += r * ( delta - sum( delta(i1,i2,i4) * r(i1,i2,i4), i4 ) ); // proper jacobian backprop respects the inhibitive interactions
	  else
		_output_gradient += delta * ( r - pow2( r ) );
	  break;
	case LIKELIHOOD:
	  {
	    Matrix2 delta_l( delta( _all, _all, 0 ) );
	    _output_gradient += delta_l(i1,i2) * r( i1, i2, i3 );
	  }
	  break;
	case UNLIKELIHOOD:
	  throw std::runtime_error("unlikelihood is invalid as mode for incoming gradient, use likelihood and proper signs");
	  break;
	}	
}

template<typename Derived>
void MixtureModel2Base<Derived>::calculate_likelihood_detector_gradient( const Matrix3 & data )
{
  BLITZ_INDICES;
  bool have_saliency = _S.numElements() > 0;
  Matrix3 r( message_component<T>::get_belief( ), true );
  blitz::Range R( WORK_RANGE( r.extent(0) ) );
  T t = _gradient_mode == LIKELIHOOD ? 1 : -1;
  OPENMP(barrier);
  OPENMP(single)
  {
	_output_gradient.resize( r.shape() );
	_likelihood_gradient.resize( r.extent(0), r.extent(1) );
  }
  Matrix3 output_gradient( _output_gradient, true );
  Matrix2 likelihood_gradient( _likelihood_gradient, true );
  T L_threshold_gradient = 0;
  T L_beta_gradient = 0;
  T N = r.extent(0) * r.extent(1);;
  if ( _use_detection_sigmoid )
	{
	  Matrix2 Sw( _S, true );
	  Matrix2 s;
	  Matrix3 r( message_component<T>::get_belief( ), true );
	  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		{
		  Matrix3 & x = it.in();
		  Matrix1 & z = it.out();		
		  if ( have_saliency ) s.reference( Sw( it.in_range1(), it.in_range2() ) );
		  T l = 0;
		  if ( have_saliency ) combined_local_likelihood( x, z, it.out_x(), it.out_y(), l, s, false );
		  else 			     combined_local_likelihood( x, z, it.out_x(), it.out_y(), l   , false );
		  
		  T dl = l - _L_threshold;
		  T arg = _L_beta_discriminative * dl;
		  T sigmoid_gradient;
		  if ( _gamma_mlr > 0 )
			{
			  // Zhang 2003: for gamma -> infinity this becomes hinge-loss (svm)
			  sigmoid_gradient = -t * sigmoid( _gamma_mlr * ( -t * arg - 1 ) );
			}
		  else
			{				
			  sigmoid_gradient = -t * sigmoid( -t * arg ); // == ( y - t )
			}
		  L_threshold_gradient += - _L_beta_discriminative * sigmoid_gradient;
		  L_beta_gradient += arg * sigmoid_gradient;
		  output_gradient( it.out_x(), it.out_y(), _all ) = z * sigmoid_gradient;
		  likelihood_gradient( it.out_x(), it.out_y() ) = sigmoid_gradient;
		  
		}
	}
  else
	{
	  output_gradient( R, _all, _all ) = -t * r( R, _all, _all );
	  likelihood_gradient( R, _all ) = -t;
	}
  L_threshold_gradient *= T(1) / N;
  L_beta_gradient *= T(1) / N;
  
  OPENMP(single)
  {
	_accumulated_L_threshold_gradient = _accumulated_L_beta_gradient = 0;
  }
  
  OPENMP(critical)
  {
	_accumulated_L_threshold_gradient += L_threshold_gradient;
	_accumulated_L_beta_gradient += L_beta_gradient;
  }

}

template<typename Derived>
void MixtureModel2Base<Derived>::calculate_likelihood_detector_hessian( const Matrix3 & data )
{
  BLITZ_INDICES;
  bool have_saliency = _S.numElements() > 0;
  Matrix3 r( message_component<T>::get_belief( ), true );
  blitz::Range R( WORK_RANGE( r.extent(0) ) );
  OPENMP(barrier);
  OPENMP(single)
  {
	_current_output_hessian.resize( r.shape() );
  }
  Matrix3 output_hessian( _output_gradient, true );
  T L_threshold_hessian = 0;
  T L_beta_hessian = 0;
  T N = 0;
  if ( _use_detection_sigmoid )
	{
	  Matrix2 Sw( _S, true );
	  Matrix2 s;
	  Matrix3 r( message_component<T>::get_belief( ), true );
	  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2);!it.atend(); ++it)
		{
		  Matrix3 & x = it.in();
		  Matrix1 & z = it.out();		
		  if ( have_saliency ) s.reference( Sw( it.in_range1(), it.in_range2() ) );
		  T l = 0;
		  if ( have_saliency ) combined_local_likelihood( x, z, it.out_x(), it.out_y(), l, s, false );
		  else 			     combined_local_likelihood( x, z, it.out_x(), it.out_y(), l   , false );
		  
		  T dl = l - _L_threshold;
		  T arg = _L_beta_discriminative * dl;
		  T y = sigmoid( arg );
		  T sigmoid_hessian = y * ( 1 - y );
		  T likelihood_hessian = _L_beta_discriminative * _L_beta_discriminative * sigmoid_hessian;
		  T beta_hessian = dl * dl * _L_beta_discriminative * sigmoid_hessian;
		  output_hessian( it.out_x(), it.out_y(), _all ) = z * ( 1 - z ) * likelihood_hessian;
		  L_threshold_hessian += likelihood_hessian;
		  L_beta_hessian += beta_hessian;
		}
	  N += 1;
	}
  else
	output_hessian( R, _all, _all ) = r( R, _all, _all ) * ( 1 - r( R, _all, _all ) );
  
  OPENMP(critical)
  {
	_L_threshold_hessian += L_threshold_hessian;
	_L_beta_hessian += L_beta_hessian;
  }
}


template<typename Derived>
void MixtureModel2Base<Derived>::perform_gradient_descent( const Matrix3 & data )
{
  BLITZ_INDICES;
  Matrix3 r( message_component<T>::get_belief( ), true );
  if ( ( _gradient_mode == LIKELIHOOD || _gradient_mode == UNLIKELIHOOD ) && _output_gradient.numElements() == 0 )
	calculate_likelihood_detector_gradient( data );
  Matrix3 output_gradient( _output_gradient, true );
  T N = r.extent(0) * r.extent(1);
  if ( r.extent(0) != output_gradient.extent(0) || r.extent(1) != output_gradient.extent(1) || r.extent(2) != output_gradient.extent(2) )
	{
	  throw std::runtime_error( str( boost::format( "mm2b::gradient_descent error shape %i,%i,%i does not match belief shape %i,%i,%i" )
									 % output_gradient.extent(0) % output_gradient.extent(1) % output_gradient.extent(2) % r.extent(0) % r.extent(1) % r.extent(2) ) );
	}
  
  if ( _report_gradient )
	{
	  OPENMP(single)
	  {
		// todo: report average ratio of gradient to parameters, but on canonical scale (i.e. log for bit expectations and precisions)
		std::cout << "- gradient report -" << std::endl;
		int i = 0;
		std::cout << "  - parameters (pre) " << std::endl;
		BOOST_FOREACH( Matrix4 & parameters, _discriminative_emission_parameters )
		  std::cout << "   - " << i++ << ") "
					 <<  min( parameters ) << ", " 
					 <<  mean( parameters) << ", " 
					 <<  max( parameters) << std::endl;
		std::cout << "    - prior " << min(_pi_discriminative) << ", " << mean(_pi_discriminative) << ", " << max(_pi_discriminative) << std::endl;
		std::cout << "    - beta " << _L_beta_discriminative << " threshold " << _L_threshold_discriminative << std::endl;
		std::cout << "    - output gradient " << _output_gradient.shape() 
			   << " [ " << min(_output_gradient) << ", " << mean(_output_gradient) << ", " << max(_output_gradient) << "] "
			   << std::endl;
	  }
	}
  
  
  Matrix1 lpi_gradient( _k );
  emission_gradient_t emission_gradient;
  bool ready = true;
  
  // calculate average gradient for this frame
  BOOST_FOREACH( Matrix4 & gradient, emission_gradient )
	{
	  gradient.resize(_k,_d1,_d2,_m);
	  gradient = 0;
	}
  lpi_gradient = 0;
  
  // collect local gradients
  for(ReceptiveFieldIterator<Matrix3,Matrix3> it(data,r,_d1,_d2,_step1,_step2);!it.atend(); ++it)
	{
	  DISPATCH(calculate_emission_gradient)( it.in(), output_gradient( it.out_x(), it.out_y(), _all ), emission_gradient );
	  lpi_gradient += output_gradient( it.out_x(), it.out_y(), _all );
	}
  
  // normalize
  if ( N > 0 )
	{
	  BOOST_FOREACH( Matrix4 & gradient, emission_gradient )
		gradient *= T(1) / N;
	  lpi_gradient *= T(1) / N;
	}
  
  // merge thread contributions
  OPENMP(critical)
  {
	if ( N > 0 )
	  {
		if ( _accumulated_emission_gradient[0].numElements() == 0 )
		  {
			for ( size_t i = 0; i < emission_gradient.size(); ++i )
			  _accumulated_emission_gradient[i].reference( emission_gradient[i] );
			_accumulated_prior_gradient.reference( lpi_gradient );
		  }
		else
		  {
			for ( size_t i = 0; i < emission_gradient.size(); ++i )
			  _accumulated_emission_gradient[i] += emission_gradient[i];
			_accumulated_prior_gradient += lpi_gradient;
		  }
	  }
  }
  OPENMP(barrier);
  OPENMP(single)
  {	  
	for ( size_t i = 0; i < emission_gradient.size(); ++i )
	  emission_gradient[i].reference( _accumulated_emission_gradient[i] );	  
	lpi_gradient.reference( _accumulated_prior_gradient );
	T L_threshold_gradient = _accumulated_L_threshold_gradient;
	T L_beta_gradient = _accumulated_L_beta_gradient;
	if ( _report_gradient )
	  {
		std::cout << " gradients (scaled) " << std::endl;
		DISPATCH(report_relative_gradient_magnitudes());
		/*
		BOOST_FOREACH( Matrix4 & gradient, emission_gradient )
		  {
			std::cout << " gradients " << std::endl;
			std::cout << "  - emission " 
					   <<  min(abs(gradient)) << ", " 
					   <<  mean(abs(gradient)) << ", " 
					   <<  max(abs(gradient)) << std::endl;	
		  }
		*/
		std::cout << "  - prior " << min(lpi_gradient) << ", " << mean(lpi_gradient) << ", " << max(lpi_gradient) << std::endl;
		std::cout << "  - beta " << L_beta_gradient << " threshold " << L_threshold_gradient << std::endl;
	  }
	
	// apply gradient descent strategy
	if ( _gradient_descent_strategy == LEVENBERG_MARQUARDT )
	  {
		if ( _discriminative_emission_hessian[0].numElements() > 0 )
		  {
			// this limits the effective learning rate at 0.1
			T epsilon = _gradient_descent_rate / _max_gradient_descent_rate;
			for ( size_t i = 0; i < emission_gradient.size(); ++i )
			  emission_gradient[i] *= T(1) / ( epsilon + _discriminative_emission_hessian[i] );
			lpi_gradient *= T(1) / ( epsilon + _output_hessian );
			L_threshold_gradient *= T(1) / ( epsilon + _L_threshold_hessian );
			L_beta_gradient *= T(1) / ( epsilon + _L_beta_hessian );
		  }
	  }
	else if ( _gradient_descent_strategy == NATURAL_GRADIENT || _gradient_descent_strategy == ADAPTIVE_NATURAL_GRADIENT )
	  {
		for ( size_t i = 0; i < emission_gradient.size(); ++i )
		  { 
			if ( int(_emission_gradient_covariance[i].size()) < _k )
			  {
				_emission_gradient_covariance[i].resize( _k, bitwise_emission_gradient_covariance_t( 10, 10, std::exp( -1 / _tau ) ) );
			  }
			typename std::list<bitwise_emission_gradient_covariance_t>::iterator icov = _emission_gradient_covariance[i].begin();
			for ( int k = 0; k < _k; ++k, ++icov )
			  {
				Matrix3 gradient( emission_gradient[i]( k, _all, _all, _all ) );
				bitwise_emission_gradient_covariance_t & gradient_covariance = *icov;
				gradient_covariance( gradient );
				if ( gradient_covariance.ready() )
				  gradient_covariance.product_with_precision( gradient, _lambda_ng );
				else
				  ready = false;
			  }
		  }		
		_prior_gradient_covariance.set_gamma( std::exp( -1 / _tau ) );
		_prior_gradient_covariance( lpi_gradient );
		if ( _prior_gradient_covariance.ready() )
		  _prior_gradient_covariance.product_with_precision( lpi_gradient, _lambda_ng );
		if ( _gradient_descent_strategy == ADAPTIVE_NATURAL_GRADIENT )
		  {
			T epsilon = _gradient_descent_rate / _max_gradient_descent_rate;
			for ( size_t i = 0; i < emission_gradient.size(); ++i )
			  emission_gradient[i] *= T(1) / ( epsilon + _output_hessian(i1) );
			lpi_gradient *= T(1) / ( epsilon + _output_hessian );
		  }
		
	  }
	else if ( _gradient_descent_strategy == PLAIN && _gradient_descent_momentum > 0 )
	  {
		
		// prepare momentum term
		if ( _last_emission_gradient[0].numElements() == 0 )
		  {
			for ( size_t i = 0; i < _last_emission_gradient.size(); ++i )
			  {
				_last_emission_gradient[i].resize(_k,_d1,_d2,_m);
				_last_emission_gradient[i] = 0;
				_emission_gradient_flip_counter[i].resize(_k,_d1,_d2,_m);
				_emission_gradient_flip_counter[i] = 0;
			  }
			_last_prior_gradient.resize( _k );
			_last_prior_gradient = 0;
			_last_L_threshold_gradient = 0;
			_last_L_beta_gradient = 0;
		  }
		
		// apply momentum
		for ( int i = 0; i < int(num_emission_statistics); ++i )
		  {
			_last_emission_gradient[i] =  _last_emission_gradient[i] * _gradient_descent_momentum + emission_gradient[i];
			emission_gradient[i].reference( _last_emission_gradient[i] );
		  }
		_last_prior_gradient = _last_prior_gradient * _gradient_descent_momentum + lpi_gradient;
		lpi_gradient.reference( _last_prior_gradient );
		if ( ( _gradient_mode == LIKELIHOOD || _gradient_mode == UNLIKELIHOOD ) && ( _use_detection_sigmoid ) )
		  {
			_last_L_threshold_gradient = _last_L_threshold_gradient * _gradient_descent_momentum + L_threshold_gradient;
			_last_L_beta_gradient = _last_L_beta_gradient * _gradient_descent_momentum + L_beta_gradient;
		  }
		
		
	  }
	
	if ( !ready )  goto end_perform_gradient_descent;
	
	if ( _report_gradient )
	  {
		std::cout << " gradients (scaled) " << std::endl;
		DISPATCH(report_relative_gradient_magnitudes());
		/*
		BOOST_FOREACH( Matrix4 & gradient, emission_gradient )
		  {			
			std::cout << "  - emission " 
					   << _gradient_descent_rate * min( gradient ) << ", " 
					   << _gradient_descent_rate * mean(gradient) << ", " 
					   << _gradient_descent_rate * max(gradient) << std::endl;	
		  }
		*/
		std::cout << "  - prior " << _gradient_descent_rate * min(lpi_gradient) << ", " << _gradient_descent_rate * mean(lpi_gradient) << ", " << _gradient_descent_rate * max(lpi_gradient) << std::endl;
		std::cout << "  - beta " << _gradient_descent_rate * L_beta_gradient << " threshold " << _gradient_descent_rate * L_threshold_gradient << std::endl;
	  }
	
	// apply gradients
	DISPATCH(apply_emission_gradient)( emission_gradient, _gradient_descent_rate );
	apply_prior_gradient( lpi_gradient, _gradient_descent_rate );
	if ( ( _gradient_mode == LIKELIHOOD || _gradient_mode == UNLIKELIHOOD ) && ( _use_detection_sigmoid ) )
	  {
		apply_detection_sigmoid_threshold_gradient( L_threshold_gradient, _gradient_descent_rate );
		if ( _optimize_detection_sigmoid_gain )
		  apply_detection_sigmoid_beta_gradient( L_beta_gradient, _gradient_descent_rate );
	  }
	
	if ( _report_gradient )
	  {
		std::cout << "  - parameters (post) " << std::endl;
		int i = 1;
		BOOST_FOREACH( Matrix4 & parameters, _discriminative_emission_parameters )
		  std::cout << "    - " << i++ << ") "
					 <<  min( parameters ) << ", " 
					 <<  mean( parameters) << ", " 
					 <<  max( parameters) << std::endl;
		std::cout << "    - prior " << min(_pi_discriminative) << ", " << mean(_pi_discriminative) << ", " << max(_pi_discriminative) << std::endl;
		std::cout << "    - beta " << _L_beta_discriminative << " threshold " << _L_threshold_discriminative << std::endl;
		std::cout << "    - output gradient " << _output_gradient.shape() 
			   << " [ " << min(_output_gradient) << ", " << mean(_output_gradient) << ", " << max(_output_gradient) << "] "
			   << std::endl;
	  }
	
	// if we got this far the parameters have changed
	dirty();
	for ( size_t i = 0; i < emission_gradient.size(); ++i )
	  _accumulated_emission_gradient[i].free();
	
	// bail out point, neccesary because openmp won't let us return out of a block, so instead we just jump to the end
  end_perform_gradient_descent:;
	
  }
}

template<typename Derived>
void MixtureModel2Base<Derived>::apply_detection_sigmoid_threshold_gradient( T threshold_gradient, T rate )
{
  _L_threshold_discriminative -= rate * threshold_gradient;
}

template<typename Derived>
void MixtureModel2Base<Derived>::apply_detection_sigmoid_beta_gradient( T beta_gradient, T rate )
{
  _L_beta_discriminative = std::exp( safe_log( _L_beta_discriminative ) - rate * beta_gradient );
}

template<typename Derived>
void MixtureModel2Base<Derived>::apply_prior_gradient( Matrix1 & prior_gradient, T rate )
{
  if ( _base_prior == NONE ) return;
  _pi_discriminative = safe_log( _pi_discriminative ) - rate * prior_gradient;
  normalize_rho( _pi_discriminative );
  //std::cout << " pi discriminative " << _pi_discriminative << std::endl;
}

