/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include <boost/format.hpp>
#include <boost/foreach.hpp>

#include "mmm2.hpp"
#include "digamma.hpp"
#include "normrho.hpp"
#include "optimal_hyperparameters.hpp"
#include "Random.hpp"

#ifdef HAVE_SSE
#include "sse_approx_logexp.h"
#include "sse_math.hpp"
#include "sse_templates.hpp"
#endif

#include "enum_properties.hpp"

#include "message_iterators.cc"

#define FETCH_PARAMETERS Matrix4 mu( _mu, true );Matrix4 mur( _mur, true );Matrix4 mur2( _mur2, true );

template<typename T>
MultinomialMixtureModel2<T>::MultinomialMixtureModel2 () 
  : _greedy_clusterer(false)
  , _batchng(false)
  , _pre_clustering_mode(NONE)
  , _beta0(0.5)
  , _learn_beta0_interval(0)
  , _learn_beta0_counter(0)
  , _Lmu(0)
  , _diagonal_parameter_jacobian(false)
  , _exact_downward_message(false)
{
}

template<typename T>
SeariseProperty MultinomialMixtureModel2<T>::get_Hmu()
{
  CPUData<double> Hmu;
  blitz2cpudata( Base::_discriminative_emission_hessian[0], Hmu );
  return Hmu;
}

template<typename T>
void MultinomialMixtureModel2<T>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter(&Base::_initialized,false);
  
  PBase::add_ivar_property_rw( "beta0",       &This::_beta0,  uninitializer);
  PBase::add_ivar_property_rw( "learn-beta0", &This::_learn_beta0_interval);
  PBase::add_property(         "mu",   typename PBase::MemFunGetter( &This::get_mu  ) );
  PBase::add_property(         "wx",   typename PBase::MemFunGetter( &This::get_wx  ) );
  PBase::add_property(         "mu_g",        typename PBase::MemFunGetter( &This::get_mu_g ) );
  PBase::add_property(         "mu_d",        typename PBase::MemFunGetter( &This::get_mu_d ) );
  PBase::add_property(         "beta",        make_blitz_ivar_getter1( &This::_beta ) );
  PBase::add_ivar_property_ro( "Lmu",         &This::_Lmu);
  PBase::add_ivar_property_rw( "diagonal_parameter_jacobian", &This::_diagonal_parameter_jacobian );
  PBase::add_ivar_property_rw( "exact_downward_message", &This::_exact_downward_message );
  PBase::add_property( "Hmu", typename PBase::MemFunGetter( &This::get_Hmu ) );
  enum_property_access<SearisePropertyTypes, This, PreClusteringMode> pre_clustering_accessor( &This::_pre_clustering_mode );
  pre_clustering_accessor
	.add_choice( "none"     , NONE     )
	.add_choice( "greedy"   , GREEDY   )
	.add_choice( "batch_ng" , BATCHNG  )
	.add_choice( "sequence" , SEQUENCE );
  PBase::add_property( "pre_clustering", pre_clustering_accessor, pre_clustering_accessor );

  PBase::add_property( std::string("sequence_clusterer"), PBase::make_pivar_getter( &This::_seq_clusterer ) );
  PBase::add_property( std::string("greedy_clusterer"), PBase::make_pivar_getter( &This::_greedy_clusterer ) );
  PBase::add_property( std::string("batch_ng"), PBase::make_pivar_getter( &This::_batchng ) );
  
}

template<typename T>
SeariseProperty MultinomialMixtureModel2<T>::get_mu_d()
{
  update();
  Matrix4 mu_d( Base::_discriminative_emission_parameters[0].copy() );
  CPUData<float> result;
  blitz2cpudata( mu_d, result );
  return result;
}

template<typename T>
SeariseProperty MultinomialMixtureModel2<T>::get_wx()
{
  Matrix4d wx( Base::_emission_statistics[0].copy() );
  CPUData<double> result;
  blitz2cpudata( wx, result );
  return result;
}

template<typename T>
SeariseProperty MultinomialMixtureModel2<T>::get_mu_g()
{
  update();
  Matrix4 mu_g( Base::_generative_emission_parameters[0].copy() );
  CPUData<float> result;
  blitz2cpudata( mu_g, result );
  return result;
}

template<typename T>
void MultinomialMixtureModel2<T>::break_sequence()
{
  if ( !ready() && _pre_clustering_mode == SEQUENCE )
	_seq_clusterer.break_sequence();
}

template<typename T>
double MultinomialMixtureModel2<T>::dL_beta0 ( double la0, void * vparams )
{
  double * params = (double*)vparams;
  double k = params[0];
  double c = params[1];
  double alpha0 = std::exp(la0);
  try {
	return boost::math::digamma(k*alpha0) - boost::math::digamma(alpha0) + c;
  }	catch(std::exception & e) {
	std::cerr << " exception evaluating dl beta0 (mmm2): " << e.what() << " type " << typeid(e).name() << std::endl;
	gsl_error_flag = true;
	return 0;
  }
}

template<typename T>
typename MultinomialMixtureModel2<T>::upward_message_iterator_t MultinomialMixtureModel2<T>::background_message_iterator( const T * x )
{
  return upward_message_iterator_t( Base::_m, 1, x, Base::_background_parameters[0].data(), Base::_background_parameters[0].data() );
}

template<typename T>
typename MultinomialMixtureModel2<T>::upward_message_iterator_t MultinomialMixtureModel2<T>::upward_message_iterator( int irf, int jrf, const T * x ) 
{
  FETCH_PARAMETERS;
  return upward_message_iterator_t( Base::_m, Base::_k, x, &mur( irf, jrf, 0, 0 ), &mur2( irf, jrf, 0, 0 ) );
}


template<typename T>
void MultinomialMixtureModel2<T>::optimize_beta0 ( double delta, double lalphamin, double lalphamax )
{
  typedef std::pair<T,bool> Result;
  double c = mean( safe_log( _mu ) );
  double params[2] = {double(Base::_m),c};
  gsl_function F = {&dL_beta0, params};	
  gsl_set_error_handler(&gsl_error_handler);	
  Result r = find_root(&F, lalphamin, lalphamax);
  bool converged = r.second;
  if (!converged) 
	{
	  std::cout << "mmm2: hyperparameter optimization did not converge" << std::endl;
	  return;
	}
  double lbeta0 = safe_log(double(_beta0));
  double lbeta0_opt = r.first;
  lbeta0 = (1-double(delta)) * lbeta0 + double(delta) * lbeta0_opt;
  _beta0 = T(std::exp(lbeta0));
  std::cout << "beta0 set to " << _beta0 << " optimal " << std::exp(lbeta0_opt) << std::endl;
}

template<typename T>
void MultinomialMixtureModel2<T>::batchlearn_begin_epoch()
{

  if ( !ready() )
	{
	  if ( _pre_clustering_mode == SEQUENCE )
		{
		  _seq_clusterer.init( Base::_m );
		}
	  else if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.init( Base::_d1, Base::_d1, Base::_m );
		}
	  else if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.initialize( Base::_k, Base::_d1, Base::_d2, Base::_m );
		  _batchng.begin_epoch();
		}
	  Base::_learning = true;
	}
  else
	{
	  Base::batchlearn_begin_epoch();
	}
}			

template<typename T>
void MultinomialMixtureModel2<T>::flush()
{
  _beta.free();
  _mu.free();
}

  ///////////////////////////////////////////
 ////// message calculators - begin ////////
///////////////////////////////////////////

template<typename T>
void MultinomialMixtureModel2<T>::downward_message ( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin, const Matrix2 & w )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  int n1 = lrin.extent(0), n2 = lrin.extent(1);
#ifdef HAVE_SSE
  for (int i = 0; i < n1; ++i)
	for (int j = 0; j < n2; ++j)
	  {
		// TODO: divide out upward message
		// need to know the corresponding input distrib
		sse_log_mvp_accum_madd( &lrin(i,j,0), Base::_m, rout.data(), Base::_k, 
								&mur( i, j, 0, 0 ), ( 1.0f - w(i,j) ) / Base::_m, w(i,j), mur.stride(2) );	  
	  }
#else
  int n3 = lrin.extent(2);
  for (int i = 0; i < n1; ++i)
	for (int j = 0; j < n2; ++j)
	  for (int m = 0; m < n3; ++m)
		lrin(i,j,m) += safe_log( w(i,j) * sum( rout * mu(_all,i,j,m) ) + ( T(1) - w(i,j) ) / Base::_m );
#endif
}

template<typename T>
void MultinomialMixtureModel2<T>::downward_message ( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  if ( _exact_downward_message && data.extent(0) > 0 )
	{
	  lrin += safe_log( sum( exp( safe_log( rout(i4) ) - safe_log( sum( data( i1, i2, i5 ) * mu( i4, i1, i2, i5 ), i5 ) ) ) * mu(i4,i1,i2,i3) ,i4) );
	}
  else
	{
#ifdef HAVE_SSE
	  int n1 = lrin.extent(0), n2 = lrin.extent(1);
	  for (int i = 0; i < n1; ++i)
		for (int j = 0; j < n2; ++j)
		  {
			sse_log_mvp_accum( &lrin(i,j,0), Base::_m, rout.data(), Base::_k, 
							   &mur( i, j, 0, 0 ), mur.stride(2) );
		  }
#else
	lrin += safe_log( sum( rout(i4) * mu(i4,i1,i2,i3) ,i4) );
#endif
	}
}

template<typename T>
void MultinomialMixtureModel2<T>::downward_message ( const IMatrix2 & data, int iout, Matrix3 & lrin )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  lrin += safe_log( mu(iout, _all, _all, _all) );
}

template<typename T>
void MultinomialMixtureModel2<T>::downward_message ( const IMatrix2 & data, int iout, Matrix3 & lrin, const Matrix2 & w )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  lrin += safe_log( w(i1,i2) * mu(iout,_all,_all,_all) + ( T(1) - w(i1,i2) ) / Base::_m );
}


#define report(x) std::cout << BOOST_PP_STRINGIZE(x) << " shape " << x.shape() << " min " << min(x) << " mean " << mean(x) << " max " << max(x) << std::endl;

template<class T> 
void MultinomialMixtureModel2<T>::local_likelihood
( blitz::Array<T,3> & X, blitz::Array<T,1> & r, T & l, bool marginal )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  l = sum( safe_log( sum( X( i1, i2, i3 ) * sum( r( i4 ) * mur( i1, i2, i3, i4 ), i4 ), i3 ) ) );
}

template<class T> 
void MultinomialMixtureModel2<T>::local_likelihood
( blitz::Array<T,3> & X, blitz::Array<T,1> & r, T & l, blitz::Array<T,2> & w, bool marginal )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  T W = mean( w );
  if ( W <= Base::_min_rf_saliency ) l = T(-1e10);
  else
	l = sum( w( i1, i2 ) * safe_log( w( i1, i2 ) * sum( X( i1, i2, i3 ) * sum( r( i4 ) * mur( i1, i2, i3, i4 ), i4 ), i3 ) ) ) / W;
}

#undef report

template<typename T>
void MultinomialMixtureModel2<T>::self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & p )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  p -= sum( sum( rout( i3 ) * rin( i1, i2, i4 ) * safe_log( mu( i3, i1, i2, i4 ) ), i4 ), i3 );
}

template<typename T>
void MultinomialMixtureModel2<T>::self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & p, const Matrix2 & w )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;  
  p -= sum( sum( rin( i1, i2, i4 ) * w( i1, i2 ) * rout( i3 ) * safe_log( mu( i3, i1, i2, i4 ) ), i4 ), i3 );
}

template<typename T>
void MultinomialMixtureModel2<T>::self_information (int iout, IMatrix2 & iin, Matrix2 & p )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  for ( int i = 0; i < Base::_d1; ++i )
	for ( int j = 0; j < Base::_d2; ++j )
	  p(i,j) -= safe_log( mur( i, j, int(iin(i,j)), iout ) );
}

template<typename T>
  void MultinomialMixtureModel2<T>::self_information ( int iout, IMatrix2 & iin,Matrix2 & p, const Matrix2 & w )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  for ( int i = 0; i < Base::_d1; ++i )
	  for ( int j = 0; j < Base::_d2; ++j )
	    p(i,j) -= safe_log( w(i,j) * mur( i, j, int(iin(i,j)), iout ) + ( T(1) - w(i,j) ) );
}

template<typename T>
bool MultinomialMixtureModel2<T>::upward_message1( blitz::Array<T,1> & in, blitz::Array<T,1> & out, size_t i_rf, size_t j_rf )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  blitz::Array<T,2> muij( mur2( i_rf, j_rf, _all, _all ) );
#ifdef HAVE_SSE
  sse_mvp( out.data(), Base::_k, in.data(), Base::_m, muij.data(), muij.stride(0) );
#else
  out = sum( muij(i1,i2) * in(i2), i2 );
#endif
  return false;
}

template<typename T>
void MultinomialMixtureModel2<T>::upward_message ( const Matrix3 & data, Matrix1 & lrout, bool marginal )
{
  FETCH_PARAMETERS;
#ifdef HAVE_SSE
  lrout = 0;
  float * outp = lrout.data();
  for (int i = 0; i < Base::_d1; ++i)
    for (int j = 0; j < Base::_d2; ++j)
	  sse_log_mvp_accum( outp, Base::_k, &data( i, j, 0 ), 
						 Base::_m, &mur2( i, j, 0, 0 ), 
						 mur2.stride(2) );	  
#else
  BLITZ_INDICES;
  lrout = sum(sum( safe_log( sum( data(i2,i3,i4) * mu(i1,i2,i3,i4) ,i4) ) ,i3),i2);
#endif
}

template<typename T>
void MultinomialMixtureModel2<T>::upward_message ( const Matrix3 & data, Matrix1 & lrout, const Matrix2 & w, bool marginal )
{
  FETCH_PARAMETERS;
#ifdef HAVE_SSE
  lrout = 0;
  float * outp = lrout.data();
  for (int i = 0; i < Base::_d1; ++i)
    for (int j = 0; j < Base::_d2; ++j)
	  sse_log_mvp_accum_madd( outp, Base::_k, &data( i, j, 0 ), Base::_m, &mur2( i, j, 0, 0 ),
							  (1.0f - w(i,j)) / Base::_m, w(i,j), mur2.stride(2) );	  
#else
  BLITZ_INDICES;
  lrout = sum(sum( safe_log( w(i2,i3) * sum( data(i2,i3,i4) * mu(i1,i2,i3,i4) ,i4) + T(1) / T(Base::_m) * ( T(1) - w(i2,i3) ) ) ,i3) ,i2);
#endif
}

template<typename T>
void MultinomialMixtureModel2<T>::upward_message ( const IMatrix2 & data, Matrix1 & lrout )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  int n1 = data.extent(0), n2 = data.extent(1);
  for (int i = 0; i < n1; ++i)
    for (int j = 0; j < n2; ++j)
	  lrout += safe_log( mu(_all,i,j,int(data(i,j))) );
}

template<typename T>
void MultinomialMixtureModel2<T>::upward_message ( const IMatrix2 & data, Matrix1 & lrout, const Matrix2 & w )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  int n1 = data.extent(0), n2 = data.extent(1);
  for (int i = 0; i < n1; ++i)
    for (int j = 0; j < n2; ++j)
	    lrout += safe_log( w(i,j) * mu(_all,i,j,int(data(i,j))) + ( T(1) - w(i,j) ) / Base::_m );
}

template<typename T>
void MultinomialMixtureModel2<T>::calculate_emission_hessian( const Matrix3 & x
															  , const Matrix1 & output_hessian
															  , typename Base::emission_gradient_t & hessian )
{
  Matrix4 & mu_d = Base::_discriminative_emission_parameters[0];
  Matrix4 & mu_g = Base::_generative_emission_parameters[0];
  Matrix4 & h_lmu = hessian[0];
  T coupling = Base::_discriminative_generative_coupling;
#ifdef HAVE_SSE
  if ( coupling > 0 )
	{
	  for ( int i = 0; i < Base::_d1; ++i )
		for ( int j = 0; j < Base::_d2; ++j )
		  for ( int k = 0; k < Base::_k; ++k )
			{
			  float Zinv = 1.0f / std::max( 1e-5f, sse_dp1( &x(i,j,0), &mu_d(k,i,j,0), Base::_m ) );
			  float C = output_hessian(k);
			  vecf4 vC( C );
			  int m = 0;
			  float * px = (float*)&x(i,j,0);
			  float * pm = &mu_d(k,i,j,0);
			  float * ph = &h_lmu(k,i,j,0);			  
			  float * pmg = (float*)&mu_g(k,i,j,0);
			  if ( _diagonal_parameter_jacobian )
				{
				  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vh( ph + m ), vmg( pmg + m );
					  ( vh + C * pow2( ( vm - pow2( vm ) ) * Zinv * vx ) + pow2( coupling * ( log( vm ) - log( vmg ) ) ) ).store( ph + m );
					}
				  for ( ; m < Base::_m; ++m )
					ph[m] += C * pow2( ( pm[m] - pow2( pm[m] ) ) * Zinv * px[m] ) + pow2( coupling * ( safe_log( pm[m] ) - safe_log( pmg[m] ) ) );
				}
			  else
				{
				  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vh( ph + m ), vmg( pmg + m );
					  ( vh + C * pow2( vm * ( Zinv * vx - 1.0f ) ) + pow2( coupling * ( log( vm ) - log( vmg ) ) ) ).store( ph + m );
					}
				  for ( ; m < Base::_m; ++m )
					ph[m] += C * pow2( pm[m] * ( Zinv * px[m] - 1.0f ) ) + pow2( coupling * ( safe_log( pm[m] ) - safe_log( pmg[m] ) ) );
				}
			}
	}
  else
	{
	  for ( int i = 0; i < Base::_d1; ++i )
		for ( int j = 0; j < Base::_d2; ++j )
		  for ( int k = 0; k < Base::_k; ++k )
			{
			  float Zinv = 1.0f / std::max( 1e-5f, sse_dp1( &x(i,j,0), &mu_d(k,i,j,0), Base::_m ) );
			  float C = output_hessian(k);
			  vecf4 vC( C );
			  int m = 0;
			  float * px = (float*)&x(i,j,0);
			  float * pm = &mu_d(k,i,j,0);
			  float * ph = &h_lmu(k,i,j,0);			  
			  if ( _diagonal_parameter_jacobian )
				{
				  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vh( ph + m );
					  ( vh + C * pow2( ( vm - pow2( vm ) ) * Zinv * vx ) ).store( ph + m );
					}
				  for ( ; m < Base::_m; ++m )
					ph[m] += C * pow2( ( pm[m] - pow2( pm[m] ) ) * Zinv * px[m] );
				}
			  else
				{
				  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vh( ph + m );
					  ( vh + C * pow2( vm * ( Zinv * vx - 1.0f ) ) ).store( ph + m );
					}
				  for ( ; m < Base::_m; ++m )
					ph[m] += C * pow2( pm[m] * ( Zinv * px[m] - 1.0f ) );
				}
			}
	}
#else
  BLITZ_INDICES;
  if ( _diagonal_parameter_jacobian )
	{
	  if ( coupling > 0 )
		h_lmu += pow2( ( mu_d - pow2( mu_d ) ) * ( x(i2,i3,i4) / max( T(1e-5), sum( x(i2,i3,i5) * mu_d(i1,i2,i3,i5), i5 ) ) ) ) * output_hessian( i1 ) + pow2( coupling * ( safe_log( mu_d ) - safe_log( mu_g ) ) );
	  else
		h_lmu += pow2( ( mu_d - pow2( mu_d ) ) * ( x(i2,i3,i4) / max( T(1e-5), sum( x(i2,i3,i5) * mu_d(i1,i2,i3,i5), i5 ) ) ) ) * output_hessian( i1 );
	}
  else
	{
	  if ( coupling > 0 )
		h_lmu += pow2( mu_d * ( x(i2,i3,i4) / max( T(1e-5), sum( x(i2,i3,i5) * mu_d(i1,i2,i3,i5), i5 ) ) - T(1) ) ) * output_hessian( i1 ) + pow2( coupling * ( safe_log( mu_d ) - safe_log( mu_g ) ) );
	  else
		h_lmu += pow2( mu_d * ( x(i2,i3,i4) / max( T(1e-5), sum( x(i2,i3,i5) * mu_d(i1,i2,i3,i5), i5 ) ) - T(1) ) ) * output_hessian( i1 );
	}
#endif
}


template<typename T>
void MultinomialMixtureModel2<T>::calculate_emission_gradient( const Matrix3 & x
															   , const Matrix2 & input_gradient
															   , const Matrix1 & output_gradient
															   , typename Base::emission_gradient_t & gradient )
{
  Matrix4 & mu_d = Base::_discriminative_emission_parameters[0];
  Matrix4 & mu_g = Base::_generative_emission_parameters[0];
  Matrix4 & d_lmu = gradient[0];
  T coupling = Base::_discriminative_generative_coupling;
  BLITZ_INDICES;
  d_lmu += mu_d * ( x(i2,i3,i4) / max( T(1e-10), sum( x(i2,i3,i5) * mu_d(i1,i2,i3,i5), i5 ) ) - T(1) )
	* input_gradient(i2,i3)
	* output_gradient( i1 )
	+ coupling * ( safe_log( mu_d ) - safe_log( mu_g ) );
}

template<typename T>
void MultinomialMixtureModel2<T>::calculate_emission_gradient( const Matrix3 & x
															   , const Matrix1 & output_gradient
															   , typename Base::emission_gradient_t & gradient )
{
  Matrix4 & mu_d = Base::_discriminative_emission_parameters[0];
  Matrix4 & mu_g = Base::_generative_emission_parameters[0];
  Matrix4 & d_lmu = gradient[0];
  T coupling = Base::_discriminative_generative_coupling;
#ifdef HAVE_SSE
  for ( int i = 0; i < Base::_d1; ++i )
	for ( int j = 0; j < Base::_d2; ++j )
	  for ( int k = 0; k < Base::_k; ++k )
		{
		  float Zinv = 1.0f / std::max( 1e-10f, sse_dp1( &x(i,j,0), &mu_d(k,i,j,0), Base::_m ) );
		  float C = output_gradient(k);
		  int m = 0;
		  float * px = (float*)&x(i,j,0);
		  float * pm = &mu_d(k,i,j,0);
		  float * pg = &d_lmu(k,i,j,0);			  
		  if ( coupling == 0 )
			{
			  if ( _diagonal_parameter_jacobian )
				{
				  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vg( pg + m );
					  ( vg + C * ( vm - pow2( vm ) ) * Zinv * vx ).store( pg + m );
					}
				  for ( ; m < Base::_m; ++m )
					pg[m] += C * ( pm[m] - pow2( pm[m] ) ) * Zinv * px[m];
				}
			  else
				{
				  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vg( pg + m );
					  ( vg + C * vm * ( Zinv * vx - 1.0f ) ).store( pg + m );
					}
				  for ( ; m < Base::_m; ++m )
					pg[m] += C * pm[m] * ( Zinv * px[m] - 1.0f );
				}
			}
		  else
			{
			  T * pmg = (float*)&mu_g(k,i,j,0);
			  if ( _diagonal_parameter_jacobian )
				{
				  for ( int m4 = 0; m4 < Base::_m / 4; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vg( pg + m ), vmg( pmg + m );
					  ( vg + C * ( vm - pow2( vm ) ) * Zinv * vx + coupling * ( log( vm ) - log( vmg ) ) ).store( pg + m );
					}				  
				  for ( ; m < Base::_m; ++m )
					pg[m] += C * ( pm[m] - pow2( pm[m] ) ) * Zinv * px[m] + coupling * ( safe_log( pm[m] ) - safe_log( pmg[m] ) );;
				}
			  else
				{
				  for ( int m4 = 0; m4 < Base::_m / 4; ++m4, m += 4 )
					{
					  vecf4 vx( px + m ), vm( pm + m ), vg( pg + m ), vmg( pmg + m );
					  ( vg + C * vm * ( Zinv * vx - 1.0f ) + coupling * ( log( vm ) - log( vmg ) ) ).store( pg + m );
					}				  
				  for ( ; m < Base::_m; ++m )
					pg[m] += C * pm[m] * ( Zinv * px[m] - 1.0f ) + coupling * ( safe_log( pm[m] ) - safe_log( pmg[m] ) );;
				}
			}
		}
#else
  BLITZ_INDICES;
  if ( _diagonal_parameter_jacobian )
	d_lmu += ( mu_d - pow2( mu_d ) ) * ( x(i2,i3,i4) / max( T(1e-10), sum( x(i2,i3,i5) * mu_d(i1,i2,i3,i5), i5 ) ) ) * output_gradient( i1 )
	  + coupling * ( safe_log( mu_d ) - safe_log( mu_g ) );
  else
	d_lmu += mu_d * ( x(i2,i3,i4) / max( T(1e-10), sum( x(i2,i3,i5) * mu_d(i1,i2,i3,i5), i5 ) ) - T(1) )* output_gradient( i1 )
	  + coupling * ( safe_log( mu_d ) - safe_log( mu_g ) );
#endif
}

template<typename T>
void MultinomialMixtureModel2<T>::apply_emission_gradient( typename Base::emission_gradient_t & parameter_gradient, T rate )
{

  BLITZ_INDICES;
  Matrix4 & d_lmu = parameter_gradient[0];
  Matrix4 & mu_d = Base::_discriminative_emission_parameters[0];

  mu_d = safe_log( mu_d ) - rate * d_lmu;
  for ( int k = 0; k < Base::_k; ++k )
	{
	  Matrix3 mu_d_k( mu_d( k, _all, _all, _all ) );
	  normalize_rho( mu_d_k );
	}
}

template<typename T>
void MultinomialMixtureModel2<T>::backpropagate_error( Matrix3 & x
													   , const Matrix1 & output_gradient
													   , Matrix3 & in_error )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
#ifdef HAVE_SSE
  for ( int i = 0; i < Base::_d1; ++i )
	for ( int j = 0; j < Base::_d2; ++j )
	  for ( int k = 0; k < Base::_k; ++k )
		{
		  float * px = &x(i,j,0);
		  float * pm = &mu(k,i,j,0);
		  float * pe = &in_error(i,j,0);			  
		  float Z = std::max( 1e-10f, sse_dp1( px, pm, Base::_m ) );
		  float C = output_gradient(k) / Z;
		  vecf4 vC( C );
		  int m = 0;
		  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
			{
			  vecf4 vx( px + m ), vm( pm + m ), ve( pe + m );
			  ( ve + vC * vm * ( vm - vm * vm ) ).store( pe + m );
			}
		  for ( ; m < Base::_m; ++m )
			pe[m] += C * pm[m] * ( pm[m] - pm[m] * pm[m] );
		}  
#else
  in_error += sum( mu(i4,i1,i2,i3) / max( T(1e-10), sum( x(i1,i2,i5) * mu(i4,i1,i2,i5), i5 ) ) * output_gradient( i4 ), i4 );
#endif
}

template<typename T>
void MultinomialMixtureModel2<T>::backpropagate_hessian( Matrix3 & x
														 , const Matrix1 & output_hessian
														 , Matrix3 & in_error )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
#ifdef HAVE_SSE
  for ( int i = 0; i < Base::_d1; ++i )
	for ( int j = 0; j < Base::_d2; ++j )
	  for ( int k = 0; k < Base::_k; ++k )
		{
		  float * px = &x(i,j,0);
		  float * pm = &mu(k,i,j,0);
		  float * pe = &in_error(i,j,0);			  
		  float Z = std::max( 1e-5f, sse_dp1( px, pm, Base::_m ) );
		  float C = output_hessian(k) / ( Z * Z );
		  vecf4 vC( C );
		  int m = 0;
		  for ( int m4 = 0, m4max = Base::_m / 4; m4 < m4max; ++m4, m += 4 )
			{
			  vecf4 vx( px + m ), vm( pm + m ), ve( pe + m );
			  ( ve + vC * pow2( vm * ( vm - vm * vm ) ) ).store( pe + m );
			}
		  for ( ; m < Base::_m; ++m )
			pe[m] += C * pow2( pm[m] * ( pm[m] - pm[m] * pm[m] ) );
		}  
#else
  in_error += blitz::pow2( sum( mu(i4,i1,i2,i3) / max( T(1e-5), sum( x(i1,i2,i5) * mu(i4,i1,i2,i5), i5 ) ) * output_hessian( i4 ) ) );
#endif
}

  /////////////////////////////////////////
 ////// message calculators - end ////////
/////////////////////////////////////////

#ifdef DEBUG
#define CHECK_COUNTS(wx)												\
  if (any(wx < 0) )														\
	{																	\
	  throw(std::runtime_error(str(boost::format("negative occurence counts range: [%f,%f]")%min(wx)%max(wx)))); \
 	}
#else
#define CHECK_COUNTS(wx)
#endif

template<typename T>
void MultinomialMixtureModel2<T>::collect_statistics ( const Matrix3 & data, const Matrix1 & r )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  if ( !ready() )
	{
	  OPENMP(single)
	  {
		if ( _pre_clustering_mode == BATCHNG )
		  _batchng.process( data );
		else if ( _pre_clustering_mode == GREEDY )
		  _greedy_clusterer.consume( data );
	  }
	}
  else
	{

	  Matrix4d wxw( Base::_local_emission_statistics[OPENMP_THREAD_NUM][0], true );
	  int n1 = data.extent(0), n2 = data.extent(1);
	  for (int i = 0; i < n1; ++i)
		for (int j = 0; j < n2; ++j)
		  {
			Matrix1 dataij( data( i, j, _all ) );
			Matrix2 muij( mur( i, j, _all, _all ) );
			Matrix2d wxij( wxw( _all, i, j, _all ) );
			wxij += r(i1) * muij(i2,i1) * dataij(i2) / sum( r(i3) * muij(i4,i3) * dataij(i4) );
		  }
	}
  
}

template<typename T>
void MultinomialMixtureModel2<T>::collect_statistics ( const Matrix3 & data, const Matrix1 & r, const Matrix2 & w2 )
{
  BLITZ_INDICES;
  FETCH_PARAMETERS;
  if ( !ready() )
	{
	  OPENMP(single)
	  {
		if ( _pre_clustering_mode == BATCHNG )
			_batchng.process( data, w2 );
		else if ( _pre_clustering_mode == GREEDY )
			_greedy_clusterer.consume( data, w2 );
	  }
	}
  else
  {
	Matrix4d wxw( Base::_local_emission_statistics[OPENMP_THREAD_NUM][0], true );
	int n1 = data.extent(0), n2 = data.extent(1);
	
	for (int i = 0; i < n1; ++i)
	  for (int j = 0; j < n2; ++j)
		{
		  T w = w2(i,j);
		  Matrix1 dataij( data( i, j, _all ) );
		  Matrix2 muij( mur( i, j, _all, _all ) );
		  Matrix2 wx ( r(i1) * muij(i2,i1) * dataij(i2) );
		  Matrix2d wxij( wxw( _all, i, j, _all ) );
		  wxij += wx * ( T(1) / sum( wx ) ) * w;
		}
  }

}

template<typename T>
void MultinomialMixtureModel2<T>::collect_statistics_gated( Matrix1 & x, Matrix3 & z, Matrix1 & w, T s0, InverseRFIterator & iter )
{
  BLITZ_INDICES;
  FETCH_PARAMETERS;
  if ( !ready() )
	throw std::runtime_error("gated statistics can not be collected with pre-clustering enabled (yet?)");
  // here I would need a normalized kjm tensor to avoid factorization...
  Matrix4d wxw ( Base::_local_emission_statistics[OPENMP_THREAD_NUM][0], true );
  Matrix3d wnxw( Base::                _wnx_local[OPENMP_THREAD_NUM]   , true );
  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
	{
	  Matrix2d  wxw_ij( wxw ( _all, iter.i1_rf(), iter.i2_rf(), _all ) );
	  Matrix1d wnxw_ij( wnxw( _all, iter.i1_rf(), iter.i2_rf()       ) );
	  Matrix1     z_j ( z   (       iter.j1     , iter.j2     , _all ) );
	  T w_ij = w( iter.j_index() );
	  wxw_ij  += z_j(i1) * w_ij *       x(i2)   * s0;
	  wnxw_ij += z_j(i1) * w_ij                 * s0;
	}  
}

template<typename T>
void MultinomialMixtureModel2<T>::collect_sampled_statistics( Matrix1 & x, int i1rf, int i2rf, int k, T s0 )
{
  BLITZ_INDICES;
  Base::_local_emission_statistics[OPENMP_THREAD_NUM][0]( k, i1rf, i2rf, _all ) += s0 * x;
}

template<typename T>
void MultinomialMixtureModel2<T>::collect_sampled_background_statistics( Matrix1 & x, T s0 )
{
  Base::_local_background_statistics[OPENMP_THREAD_NUM][0] += s0 * x;
}

template<typename T>
void MultinomialMixtureModel2<T>::apply_r( const Matrix3 & data )
{
  if ( !ready() && _pre_clustering_mode == SEQUENCE )
	_seq_clusterer.collect_statistics( data );
  else
	Base::apply_r( data );
}

template<typename T>
void MultinomialMixtureModel2<T>::recalc_likelihoods()
{
  Base::_Lp = sum( Base::_emission_statistics[0] * safe_log( _mu ) );
	  
  T lCbeta0 = Base::_k * Base::_d1 * Base::_d2 * ( gammaln( Base::_m *_beta0 ) - Base::_m * gammaln( _beta0 ) );
  T lCbeta  = sum( gammaln( sum( _beta, i4 ) ) ) - sum( gammaln( _beta ) );
  
  T Elp = sum( ( _beta0 - 1 ) * safe_log( _mu ) );
  T Elq = sum( ( _beta  - 1 ) * safe_log( _mu ) );
  
  if ( Base::_have_background_model && Base::_learn_background_model )
	{
	  lCbeta0 += gammaln( Base::_m *_beta0 ) - Base::_m * gammaln( _beta0 );
	  lCbeta += gammaln( sum( _beta_bg ) ) - sum( gammaln( _beta_bg ) );
	  Elp += sum( ( _beta0    - 1 ) * safe_log( _mu_bg ) );
	  Elq += sum( ( _beta_bg  - 1 ) * safe_log( _mu_bg ) );
	}

  _Lmu = Elp - Elq + lCbeta0 - lCbeta;
  
  Base::_L = Base::_Lz + _Lmu + Base::_Lp;
  
  std::cout << str( boost::format("multinomial component model ended epoch Lmu = %1.3e Lp = %1.3e L = %1.3e")
					 % _Lmu % Base::_Lp % Base::_L ) << std::endl;
}

template<typename T>
void MultinomialMixtureModel2<T>::recalc_hyperparams( )
{
  Matrix4d & wx = Base::_emission_statistics[0];
  Matrix4d & wx_prior = Base::_prior_emission_statistics[0];
  Matrix1d & wx_bg = Base::_background_statistics[0];
  Matrix1d & wx_bg_prior = Base::_prior_background_statistics[0];
  if ( wx_prior.numElements() != wx.numElements() )
	{
	  wx_prior.resize( wx.shape() );
	  wx_prior = 0;
	}
  if ( Base::_delta_to_random )
	_beta.reference( Matrix4d( _beta0 + Base::_delta * wx + ( 1 - Base::_delta ) * wx_prior ) );
  else
	_beta.reference( Matrix4d( _beta0 + Base::_delta * wx ) );
  if ( Base::_have_background_model && Base::_learn_background_model )
	{
	  if ( wx_bg_prior.numElements() != wx_bg.numElements() )
		{
		  wx_bg_prior.resize( wx_bg.shape() );
		  wx_bg_prior = 0;
		}
	  if ( Base::_delta_to_random )
		_beta_bg.reference( Matrix1d( _beta0 + Base::_delta * wx_bg + ( 1 - Base::_delta ) * wx_bg_prior ) );
	  else
		_beta_bg.reference( Matrix1d( _beta0 + Base::_delta * wx_bg ) );
	}
  Base::dirty();
}

template<typename T>
blitz::Array<typename MultinomialMixtureModel2<T>::num_type,1>  MultinomialMixtureModel2<T>::batchlearn_end_epoch ( Matrix1 f_in )
{
  BLITZ_INDICES;

  Matrix4d & wx = Base::_emission_statistics[0];
  Matrix4d & wx_prior = Base::_prior_emission_statistics[0];
  Matrix1d & wx_bg = Base::_background_statistics[0];

  Matrix1 f;

  if ( !ready() )
	{

	  if ( _pre_clustering_mode == SEQUENCE )
		{

		  std::cout << " applying sequence clustering " << std::endl;
		  _seq_clusterer.end_batch();
		  _seq_clusterer.cluster_AHC( Base::_k );
		  
		  typedef typename heuristic_sequence_clusterer<T>::state_groups state_groups;
		  typedef typename heuristic_sequence_clusterer<T>::state_group state_group;

		  const state_groups & groups = _seq_clusterer.groups();

		  Base::_k = groups.size();
		  Base::_d1 = Base::_d2 = 1;
		  Base::_wk.resize( Base::_k ) ; Base::_wk  = 0;
		  Base::_wk_prior.resize( Base::_wk.shape() ); Base::_wk_prior = 0;
		  Base::_wnx.resize( Base::_k, 1, 1 ); Base::_wnx = 0;
		  wx.resize( Base::_k, 1, 1, Base::_m ); wx = 0;
		  wx_prior.resize( wx.shape() ); wx_prior = 0;

		  size_t igroup = 0;
		  const Matrix1d & abundances  = _seq_clusterer.abundances();
		  BOOST_FOREACH( const state_group & group, groups )
			{
			  BOOST_FOREACH( size_t state, group )
				{
				  Base::_wk( igroup ) += abundances( int(state) );
				  Base::_wnx( igroup, _all, _all ) += abundances( int(state) );
				  wx ( int(igroup), 0, 0, int(state) ) += abundances( int(state) );
				}
			  ++igroup;
			}
		  
		  Base::_emission_statistics[0].reference( wx );
		  Base::_prior_emission_statistics[0].reference( wx_prior );

		  Base::_initialized = true;
		  Base::_learning = false;
		  Base::dirty();

		  Base::recalc_hyperparams();
		  recalc_hyperparams();
		  update();
		  recalc_likelihoods();
		  
		}
	  else if ( _pre_clustering_mode == GREEDY )
		{
		  std::cout << "finishing gredy clustering " << std::endl;
		  _greedy_clusterer.finish();
		  Base::_k = _greedy_clusterer.size();
		  Base::_wk.reference(  _greedy_clusterer.counts ().copy() );
		  Base::_wnx.reference( _greedy_clusterer.counts3().copy() );
		  wx.reference( _greedy_clusterer.histograms().copy() );
		  Base::recalc_hyperparams();
		  recalc_hyperparams();
		  Base::_initialized = true;
		  update();
		  recalc_likelihoods();
		  Base::_learning = false;
		}
	  else if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.end_epoch();
		  
		  if (_batchng.ready())
			{
			  Base::_wk.reference( _batchng.get_histogram().copy() );
			  Base::_wnx.reference( _batchng.get_histogram3().copy() );
			  wx.reference( _batchng.get_centerStat().copy() );
			  Base::recalc_hyperparams();
			  recalc_hyperparams();
			  Base::_initialized = true;
			  update();
			  recalc_likelihoods();
			  Base::_learning = false;
			}
		  else
			{
			  _mu.reference( _batchng.get_centers() );
			  return Matrix1();
			}
		}

	}
  else
	{
	  
	  f.reference( Base::batchlearn_end_epoch() );
	  
	  if (f_in.numElements() > 0)
		{
		  wx = f_in(i4) * wx;
		  if ( Base::_have_background_model && Base::_learn_background_model )
			wx_bg = f_in(i4) * wx_bg;			
		}
  
	  if (_learn_beta0_interval > 0 && ++_learn_beta0_counter >= _learn_beta0_interval)
		{
		  T beta0_old = _beta0;
		  update(true);
		  optimize_beta0(Base::_delta);
		  T deltabeta0 = _beta0 - beta0_old;
		  _beta += deltabeta0;
		}

  
	  CHECKNAN(_beta,"nan in beta after epoch");
  
	  recalc_hyperparams();
	  	  
	}
  return f;
  
}

template<typename T>
void MultinomialMixtureModel2<T>::init_sample ( Matrix3 & in )
{
  in = T(1)/T(Base::_k);
}

template<typename T>
void MultinomialMixtureModel2<T>::sample_point( int irf, int jrf, int k, Matrix1 & data )
{
  BLITZ_INDICES;
  Matrix4 mur2w( _mur2, true );
  Matrix1 mu( mur2w( irf, jrf, k, _all ) );
  int m = Random<T>::get().sample_multinomial( mu );
  data = T(0);
  data(m) = T(1);
}

template<typename T>
void MultinomialMixtureModel2<T>::sample_one ( unsigned int out, Matrix3 & in )
{
  BLITZ_INDICES;
  in *= _mu(out,_all,_all,_all);
  normalize(in);
}

template<typename T>
void MultinomialMixtureModel2<T>::update( bool force )
{
  if ( !ready() ) return;

  BLITZ_INDICES;
  initialize();
  
  if ( Base::update(force) ) 
	{

	  OPENMP(barrier)
	  OPENMP(single)
	  {

		Matrix1  & mu_b = Base::_background_parameters[0];
		mu_b.resize( Base::_m );
		mu_b = T(1) / Base::_m;

		_mu.free();
		_mu.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);

		if ( Base::_requested_parameter_mode == Base::VARIATIONAL )
		  {
			_mu = exp( psi(_beta) - psi(sum(_beta(i1,i2,i3,i5),i5)) );
			Base::_generative_emission_parameters[0].reference( _mu );
			//std::cout << "mmm2 made variational parameters" << std::endl;
		  }
		else if ( Base::_requested_parameter_mode == Base::GENERATIVE ||
				  ( Base::_requested_parameter_mode == Base::DISCRIMINATIVE &&
					Base::_generative_emission_parameters[0].numElements() == 0 ) )
		  {
			_mu = _beta / sum( _beta( i1, i2, i3, i5 ), i5 );
			Base::_generative_emission_parameters[0].reference( _mu );
			//std::cout << "mmm2 made generative parameters" << std::endl;
		  }			  

		if ( Base::_have_background_model )
		  {
			if ( Base::_requested_parameter_mode == Base::VARIATIONAL 
				 && Base::_learn_background_model )
			  _mu_bg = exp( psi( _beta_bg ) - psi( sum( _beta_bg ) ) );
			else
			  _mu = _beta / sum( _beta );
			Base::_background_parameters[0].reference( _mu_bg );
		  }

		if ( Base::_requested_parameter_mode == Base::DISCRIMINATIVE )
		  {
			if ( Base::_discriminative_emission_parameters[0].numElements() == 0 )
			  {
				Base::_discriminative_emission_parameters[0].reference( Base::_generative_emission_parameters[0].copy() );
				//std::cout << "mmm2 copied discriminative from generative parameters" << std::endl;
			  }
			_mu.reference( Base::_discriminative_emission_parameters[0] );
		  }		
		
		// FIXME: are all these neccessary? overhead?
		_mur.resize( Base::_d1, Base::_d2, Base::_m, Base::_k );
		_mur = _mu(i4,i1,i2,i3);
		_mur2.resize( Base::_d1, Base::_d2, Base::_k, Base::_m );
		_mur2 = _mu(i3,i1,i2,i4);

	  }

	}

}

template<typename T>
void MultinomialMixtureModel2<T>::initialize ( bool force )
{

  if ( !ready() ) return;

  if ( Base::initialize(force) )
  {

	OPENMP(barrier)
	OPENMP(single)
	{
  
	  Matrix4d & wx = Base::_emission_statistics[0];
	  Matrix4d & wx_prior = Base::_prior_emission_statistics[0];
	  Matrix4d & wx_bg = Base::_emission_statistics[0];
	  Matrix4d & wx_bg_prior = Base::_prior_emission_statistics[0];

	  wx_prior.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);
	  Random<T> & rng = Random<T>::get();
	  Matrix4 sample(Base::_k,Base::_d1,Base::_d2,Base::_m);
	  if ( Base::_init_uniform )
		sample = 1;
	  else
		for ( typename Matrix4::iterator i = sample.begin(), e = sample.end(); i != e; ++i )
		  *i = std::max( T(1e-3) * Base::_k, rng.gamma( _beta0 ) );
	  wx_prior = Base::_wnx_prior(i1,i2,i3) * sample / sum(sample(i1,i2,i3,i5),i5);

	  std::cout << str(boost::format("multinomial component model initialized beta0 = %f")%_beta0) << std::endl;

	  wx.reference( wx_prior.copy() );

	  // background: always init uniform
	  if ( Base::_have_background_model )
		{
		  wx_bg_prior.resize(Base::_m);
		  wx_bg_prior = sum( Base::_wk_prior ) / Base::_m;
		  wx_bg.reference( wx_bg_prior.copy() );
		}

	  recalc_hyperparams();

	}

  }
}

template class MultinomialMixtureModel2<float>;
INSTANTIATE_GATED_MESSAGE_DELEGATE( MultinomialMixtureModel2<float> );
