/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef MMM2_HPP
#define MMM2_HPP

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <iostream>
#include <boost/serialization/version.hpp>

#include "heuristic_sequence_clustering.hpp"
#include "greedy_clusterer.hpp"
#include "BatchNG.hpp"
#include "metrics.hpp"
#include "serialize_blitz.hpp"
#include "MixtureModel2Base.hpp"

#ifdef HAVE_SSE
#include "sse_math.hpp"
#include "sse_templates.hpp"
#endif

template<typename T> class MultinomialMixtureModel2;

template<typename T> struct has_upward_message_iterator< MultinomialMixtureModel2<T> >
 : public boost::true_type {};

template<typename T> struct TypeTraits< MultinomialMixtureModel2<T> > 
{	
  typedef T num_type;
  static const int num_statistics = 1;
};

template<typename T> class MultinomialMixtureModel2 
  : public MixtureModel2Base< MultinomialMixtureModel2<T> >,
	public HasProperties< SearisePropertyTypes, MultinomialMixtureModel2<T> >
{
  
  typedef MultinomialMixtureModel2<T> This;
  typedef MixtureModel2Base< This > Base;
  typedef HasProperties< SearisePropertyTypes, MultinomialMixtureModel2<T> > PBase;
  
public:
  
  typedef T num_type;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,4> Matrix4;

  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<double,4> Matrix4d;

  typedef blitz::Array<unsigned int,2> IMatrix2;
  
  MultinomialMixtureModel2();
  
  static void init_properties();
  
  SeariseProperty get_mu() 
  {
    update();
    CPUData<float> mu;
    blitz2cpudata( _mu, mu );
    return mu;
  }

  struct upward_message_iterator_t
  {
	upward_message_iterator_t( int _M, int _K, const T * _x, T * _mu, T * _mu2 )
	  : M(_M), K(_K), x(_x), mu(_mu), mu2(_mu2), k(0) 
	{}
	void advance( int n ) { k += n; }
	T evaluate_linear( ) 
	{
	  T y = 0;
    size_t offset = M * k;
#ifdef HAVE_SSE
    y = sse_dp1( x, mu2 + offset, M, sse_aligned( x ) && sse_aligned( mu2 + offset ) );
#else    
	  for ( int m = 0; m < M; ++m ) y += x[m] * mu2[offset+m];
#endif    
	  return y;
	}
	T evaluate_log() 
	{
	  return safe_log( evaluate_linear() );
	}
#ifdef HAVE_SSE
	vecf4 evaluate_linear_sse( )
	{
	  vecf4 y = 0.0f;
	  for ( int m = 0, im = 0; m < M; ++m, im += K ) 
      y += x[m] * vecf4( mu + im + k, sse_aligned( mu + im + k ) );
	  return y;
	}
	vecf4 evaluate_log_sse() 
	{
	  return safe_log( evaluate_linear_sse() );
	}
#endif
	const int M;
	const int K;
	const T * x;
	const T * mu, * mu2;
	int k;
  };
  
  upward_message_iterator_t upward_message_iterator( int irf, int jrf, const T * x );
  
  upward_message_iterator_t background_message_iterator( const T * x );

  Matrix2 get_mur( int irf, int jrf ) { BLITZ_INDICES; Matrix4 Mrw( _mur, true ); return Mrw( irf, jrf, _all, _all ); }

  void sample_point( int irf, int jrf, int k, Matrix1 & data );

  SeariseProperty get_Hmu();

  void flush();

  // overridden to intercept in case of greedy pre-clustering
  void apply_r( const Matrix3 & data );

  bool ready() 
  { 
	return _pre_clustering_mode == NONE || Base::_initialized;
  }
  
  Matrix3 calc_component_entropies() 
  {
	Matrix3 H( -sum( _mur2 * safe_log( _mur2 ), blitz::fourthIndex() ) );
	return H;
  }

  Matrix3 calc_max_component_likelihoods() 
  { 
	Matrix3 lmax( 
				 max( 
					 safe_log( _mur2 ),
					 blitz::fourthIndex()
					  )
				  );
	return lmax;
  }

  void reset_input_dimensions(const std::set<unsigned int> & components);
  
  void reset_components(const std::set<unsigned int> & components);
  
  void batchlearn_begin_epoch();
  
  void downward_message( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin, const Matrix2 & w );
  
  void downward_message( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin );
  
  void downward_message( const IMatrix2 & data, int iout, Matrix3 & lrin );
  
  void downward_message( const IMatrix2 & data, int iout, Matrix3 & lrin, const Matrix2 & w );
  
  void upward_message( const Matrix3 & data, Matrix1 & lrout, bool marginal = false );
  
  void upward_message( const Matrix3 & data, Matrix1 & lrout, const Matrix2 & w, bool marginal = false );
  
  void upward_message( const IMatrix2 & data, Matrix1 & lrout );
  
  void upward_message( const IMatrix2 & data, Matrix1 & lrout, const Matrix2 & w );
  
  bool upward_message1( blitz::Array<T,1> & in, blitz::Array<T,1> & out, size_t i_rf, size_t j_rf );

  void self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin, const Matrix2 & w );

  void self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin );

  void self_information ( int iout, IMatrix2 & iin,Matrix2 & p );

  void self_information ( int iout, IMatrix2 & iin,Matrix2 & p, const Matrix2 & w );

  void local_likelihood( Matrix3 &, Matrix1 &, T &, bool marginal = true );

  void local_likelihood( Matrix3 &, Matrix1 &, T &, Matrix2 &, bool marginal = true );

  void collect_statistics( const Matrix3 & data, const Matrix1 & r );
  
  void collect_statistics( const Matrix3 & data, const Matrix1 & r, const Matrix2 & w2 );
  
  void collect_statistics( const Matrix3 & data, const Matrix1 & r, const Matrix3 & w3 );
  
  void collect_statistics_gated( Matrix1 & x, Matrix3 & z, Matrix1 & Z, T s0, InverseRFIterator & irf );

  void collect_sampled_statistics( Matrix1 & x, int i1rf, int i2rf, int k, T s0 );

  void collect_sampled_background_statistics( Matrix1 & x, T s0 );

  Matrix1 batchlearn_end_epoch(Matrix1 f_in = Matrix1());
  
  void init_sample(Matrix3 & in);
  
  void sample_one(unsigned int out, Matrix3 & in);
  
  void update(bool force = false);
  
  void initialize(bool force=false);
  
  bool is_multinomial() { return true; }
  
  void break_sequence();

  void recalc_hyperparams();

  void calculate_emission_gradient( const Matrix3 & x
									, const Matrix1 & output_gradient
									, typename Base::emission_gradient_t & gradient );

  void report_relative_gradient_magnitudes()
  {
	Matrix4 & d_lmu = Base::_accumulated_emission_gradient[0];
	Matrix4 & mu_d = Base::_discriminative_emission_parameters[0];
	Matrix4 g_rel( d_lmu / safe_log( mu_d ) );
	T x_mean = mean(g_rel), x2_mean = mean(g_rel*g_rel), x_max = max(g_rel), x_min = min(g_rel);
	T x_std = std::sqrt( std::max( T(0), x2_mean - x_mean * x_mean ) );
	std::cout << "   mu: [ " << x_min << ", " << x_mean << " (" << x_std << "), " << x_max << "]" << std::endl;
  }

  T canonical_parameter_scale( int stats_index )
  {
	// the canonical parameter is the log of the parameter
	return mean( safe_log( Base::_discriminative_emission_parameters.at(stats_index) ) );
  }

  void calculate_emission_gradient( const Matrix3 & x
									, const Matrix2 & input_gradient
									, const Matrix1 & output_gradient
									, typename Base::emission_gradient_t & gradient );

  void calculate_emission_hessian( const Matrix3 & x
								   , const Matrix1 & output_hessian
								   , typename Base::emission_gradient_t & parameter_hessian );

  void apply_emission_gradient( typename Base::emission_gradient_t & parameter_gradient, T rate );

  void backpropagate_error( Matrix3 & x
							, const Matrix1 & output_gradient
							, Matrix3 & in_error );

  void backpropagate_hessian( Matrix3 & x
							  , const Matrix1 & output_hessian
							  , Matrix3 & in_hessian );

  void recalc_likelihoods();
  
private:

  SeariseProperty get_mu_g();

  SeariseProperty get_wx();

  SeariseProperty get_mu_d();

  static double dL_beta0 (double lb0, void * params);
  
  void optimize_beta0(double delta = 1, double lalphamin = -5, double lalphamax = 15);
  
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
	ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(Base);
    ar & BOOST_SERIALIZATION_NVP(_beta0);
	recalc_hyperparams();
  }

  heuristic_sequence_clusterer<T>  _seq_clusterer;
  greedy_clusterer< T, manhattan_metric<T> >  _greedy_clusterer;
  BatchNG< T, manhattan_metric<T> >  _batchng;
  
  typedef enum { NONE, BATCHNG, GREEDY, SEQUENCE } PreClusteringMode;
  PreClusteringMode _pre_clustering_mode;
  
  Matrix4d _beta;
  Matrix4 _mu, _mur, _mur2;
  Matrix1d _beta_bg;
  Matrix1 _mu_bg;

  T _beta0;
  int _learn_beta0_interval, _learn_beta0_counter;
  T _Lmu;
  
  bool _diagonal_parameter_jacobian;
  bool _exact_downward_message;

  blitz::firstIndex i1;
  blitz::secondIndex i2;
  blitz::thirdIndex i3;
  blitz::fourthIndex i4;
  blitz::fifthIndex i5;
  blitz::sixthIndex i6;
  
};

//templated class versioning
namespace boost { namespace serialization {
	template<typename T>
	struct version< MultinomialMixtureModel2<T> >
	{
	  BOOST_STATIC_CONSTANT(unsigned int, value = 2);
	};
  }}
#endif
