/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _MMM2_BELIEF_MESSAGE_BUFFER_HPP_
#define _MMM2_BELIEF_MESSAGE_BUFFER_HPP_


#include <stdexcept>
#include <boost/format.hpp>

#include <blitz/array.h>
#include "blitz_indices.hpp"
#include "Random.hpp"
#include "normrho.hpp"
#include <algorithm>
#include <set>
#include <boost/circular_buffer.hpp>

#include "omp_macros.hpp"

#ifdef HAVE_SSE
#include "sse_approx_logexp.h"
#include "sse_math.hpp"
#endif
#include "approx_logexp.h"

template<typename T>
struct mmm2_belief_state
{
  enum {LOG = 0,LINEAR = 1};
  typedef enum { MULTINOMIAL = 0, BINOMIAL = 1 } distribution_type;

  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<unsigned,2> IMatrix2;
  typedef blitz::Array<T,3> Matrix3;

  int domain;
  bool normalized;
  Matrix3 data;

  mmm2_belief_state & operator=( const mmm2_belief_state & other )
  {
    domain = other.domain;
    normalized = other.normalized;
    data.reference( other.data );
    _type = other._type;
    return *this;
  }

  mmm2_belief_state( distribution_type type = MULTINOMIAL ) : domain(LINEAR), normalized(false) , _type(type) 
  {
  }

  mmm2_belief_state( const mmm2_belief_state & other ) 
  : domain(other.domain), normalized(other.normalized), data(other.data), _type(other._type)
  {
  }

  ~mmm2_belief_state()
  {
  }

  bool empty() const { return data.numElements() == 0; }

  void init(const blitz::TinyVector<int,3> & s, int dom = LOG)
  {

	OMP_LOG("state init");

	OPENMP(barrier);
	OPENMP(master)
	{
	  data.resize(s);
	}
	OPENMP(barrier);

    BLITZ_INDICES;
    blitz::Range R( WORK_RANGE( data.extent(0) ) );
    if ( !R.empty() )
      {
	Matrix3 wdata( data, true );
	
	switch(dom)
	  {
	  case LOG:    wdata(R,_all,_all) = 0; break;
	  case LINEAR: wdata(R,_all,_all) = _type == BINOMIAL ? T(0.5) : 1; break;
	  default: throw(std::runtime_error("invalid domain"));
	  }	
      }

	OPENMP(barrier);
    OPENMP(single)
    {
      domain = dom;
    }

	OMP_LOG("state init end");

  }

  void clear()
  {
	data.free();
	domain = LINEAR;
	normalized = false;
  }

  void set(Matrix3 dat, int dom, bool norm = false)
  {
    data.reference(dat);
    domain = dom;
    normalized = norm;
  }

  blitz::TinyVector<int, 3> shape()
  {
    return data.shape();
  }

  void to_linear(bool norm = true)
  {

	OMP_LOG("state to linear");

	OPENMP(barrier);
	
	BLITZ_INDICES;
	
	blitz::Range R( WORK_RANGE( data.extent(0) ) );
	
	if (!R.empty() )
	  {
		
		Matrix3 wdata( data, true );
		Matrix3 subdata( wdata( R, _all, _all ) ); 
		if (domain == LOG)
		  {
			switch( _type )
			  {
			  case MULTINOMIAL: 
				normalize_rho( subdata, norm ); 
				break;
			  case BINOMIAL: subdata = T(1) / ( T(1) + aexpf( -subdata ) ); break;
			  }
		  }
		else if (norm && !normalized)
		  {
			switch( _type )
			  {
			  case MULTINOMIAL: 
				normalize(subdata);
				break;
			  case BINOMIAL: break;
			  }
		  }
	  }
	
	OPENMP(barrier);
	OPENMP(single)
	  {	  
		if ( domain == LOG || ( norm && !normalized && _type == MULTINOMIAL ) )
		  {
			normalized = ( domain == LINEAR && normalized ) || norm;
			domain = LINEAR;
		  }
	  }

	OMP_LOG("state to linear end");

  }

  void to_log()
  {

	OMP_LOG("state to log");

    BLITZ_INDICES;

	blitz::Range R( WORK_RANGE( data.extent(0) ) );

	OPENMP(barrier);

    if (domain == LINEAR)
    {	 

      if ( !R.empty() )
	{

	  Matrix3 wdata( data, true );
	  Matrix3 subdata( wdata( R, _all, _all ) ); 
	  
	  switch( _type )
	    {
	    case MULTINOMIAL:
	      {
#ifdef HAVE_SSE
		size_t n = subdata.numElements();
		size_t n4 = n / 4;
		size_t i = 0;
		for ( size_t i4 = 0; i4 < n4; ++i4, i += 4 )
		  _mm_storeu_ps( subdata.data() + i, logf4( _mm_loadu_ps( subdata.data() + i ) ) );
		for ( i = 4 * n4; i < n; ++i )
		  subdata.data()[i] = safe_log( subdata.data()[i] );
#else
		const T epsilon = 1e-10;
		const T lepsilon = std::log(epsilon);
		subdata = where(subdata>epsilon,log(subdata),lepsilon);
#endif
	      }
	      break;
	    case BINOMIAL: subdata = safe_log( subdata ) - safe_log( T(1) - subdata ); break;
	    }
	  
	}
      
      

    }
	
	OPENMP(barrier);
	OPENMP(single)
      {
		domain = LOG;
      }

	OMP_LOG("state to log end");

  }

  template<class ostream> void do_stream_out(ostream & os)
  {			
    std::string domainstr = ( domain == LOG ? "LOG" : ( domain == LINEAR ? "LINEAR" : "???" ) );
    std::string nstr = normalized ? "normalized" : "unnormalized";
    os << data.shape() << " " << domainstr << " " << nstr;
  }

  void set_type( distribution_type type ) { _type = type; }

private:
  
  distribution_type _type;

};

template <class ostream, typename T>
ostream & operator<< (ostream& os, mmm2_belief_state<T> & belief)
{
  belief.do_stream_out(os);
  return os;
}


///factorization of the 2d multinomial sheet propblem
///where a node state is defined via product of messages, normalized
template<typename T> 
class mmm2_belief_message_buffer
{

public:

  typedef blitz::Array<unsigned int,2> IMatrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef mmm2_belief_state<T> belief_state;

  mmm2_belief_message_buffer( )
	: _states( 1 )
	, _sampling(false)
	, _type( belief_state::MULTINOMIAL )
	, _active_state(0)
	, _w(1)
  {
	advance();
  }

  ~mmm2_belief_message_buffer()
  {
  }


  size_t size() { return state().messages.size(); }

  void resize(int n)
  {
	state().resize(n);
  }

  template<class ostream> void do_stream_out(ostream & os)
  {
    os << " message buffer size " << size() << " dirty: " << (dirty()?"true":"false") 
       << " history " << get_history_size() << "/" << get_history_capacity()
       << " selected " << active_state()
       << std::endl;
    for ( slot_map_t::iterator i = _used_slots.begin(), iend = _used_slots.end();
        i != iend;
        ++i)
    {
      size_t index = i->first;
      std::string & name = i->second;
//      if (message(index).empty()) continue;
	      os << "   " << name << ") " << message(index) << std::endl;
    }
  }

  void set_msg(int slot, const Matrix3 & msg, int domain = belief_state::LOG)
  {
    message(slot).set(msg, domain);
    touch();
  }

  belief_state & message(int i,bool allow_resize = true) 
  {
    if (i >= int(size()))
      {
	if ( allow_resize )
	  resize(i+1);
	else
	  throw std::runtime_error("requesting message at index beyond bounds and resize is forbidden");
      }
    return state().messages.at(i); 
  }

  belief_state & belief() { return state().belief; }

  template<int N>
  static bool empty(const blitz::TinyVector<int, N> & s)
  {
    for (int i = 0; i < N; ++i) if ( s[i] == 0 ) return true;
    return false;
  }

  template<int N>
  static bool different(const blitz::TinyVector<int, N> & s1, const blitz::TinyVector<int, N> & s2)
  {
    for (int i = 0; i < N; ++i) if ( s1[i] != s2[i] ) return true;
    return false;
  }

  blitz::Array<T,3> & sample( ::Random<T> & rng, int exclude = -1, bool force_resample = false )
  {
    BLITZ_INDICES;
    if ( dirty() || force_resample || exclude != state().exclude )
    {
      Matrix3 r ( update(exclude), true );
      OPENMP(flush);
      OPENMP(single)
		state().sample.resize(r.extent(0), r.extent(1));
      for ( int i = 0; i < r.extent(0); ++i )
        for ( int j = 0; j < r.extent(1); ++j )
	  state().sample( i, j ) = rng.sample_multinomial( r( i, j, _all ) );
      belief().set( frommaxindices2<T> ( state().sample, r.extent(2) ), belief_state::LINEAR, true );
    }
    return belief().data;
  }

  blitz::Array<unsigned,2> & get_maxima( )
  {
	return state().maxima;
  }

  void update_maxima( int ignore = -1 )
  {
	update( ignore );	
    if ( state().maxima_dirty )
      {
	OPENMP(flush);
	OPENMP(single)
	  state().maxima.reference( maxindex2( belief().data ) );
    }
  }

  blitz::Array<T,3> & sample( int exclude = -1, bool force_resample = false)
  {
    Random<T> & rng = Random<T>::get();
    return sample( rng, exclude, force_resample );
  }

  void set_sample( IMatrix2 & sample )
  {
	state().set_sample( sample );
  }

  inline void release_buffer()
  {
	state().release_buffer();
  }

  inline Matrix3 & update(int exclude = -1, bool force_linear = true)
  {    
	OMP_LOG("buffer update");
    
    BLITZ_INDICES;
    blitz::TinyVector<int,3> s = check_shape(exclude);
    
    state_t & S = state();
    
	if ( exclude < 0 && state().sampled && S.dirty )
	  {
		OPENMP(barrier);
		OPENMP(master)
		{
		  belief().data.resize( get_sample().extent(0), get_sample().extent(1), s[2] );
		}
		OPENMP(barrier);
		blitz::Range R( WORK_RANGE( get_sample().extent(0) ) );
		Matrix3 wbelief( belief().data, true );
		Matrix3 subbelief( wbelief( R, _all, _all ) );
		IMatrix2 samplew( get_sample(), true );
		IMatrix2 subsample( samplew( R, _all ) );
		frommaxindices2( subsample, subbelief );
		return belief().data;
	  }	  
	else if ( ( S.dirty || exclude != S.exclude) && !empty( s ) )
      {

		OPENMP(barrier)
		  OPENMP( single )
		  {
			S.dirty = false;
			S.maxima_dirty = true;
			S.exclude = exclude;
		  }
		
		belief().init(s,belief_state::LOG);
		
		blitz::Range R( WORK_RANGE( belief().data.extent(0) ) );
		Matrix3 wbelief( belief().data, true );
		Matrix3 subbelief( wbelief( R, _all, _all ) );
		
		for (int i = 0; i < int( size() ); ++i)
		  {
			// no resize will happen here since i < size()
			belief_state & msg = message(i,false);
			if  ( i == exclude || msg.empty() ) continue;				
			
			
			msg.to_log();
			
			Matrix3 wmsg( msg.data, true );
			
			if ( !R.empty() )
			  {		
				if ( wmsg.extent(0) == 1 && wmsg.extent(1) == 1 )  // broadcasting in space			
				  {
					subbelief += wmsg(0,0,_all)(i3);
				  }
				else
				  {
					Matrix3 submsg( wmsg( R, _all, _all ) );
					subbelief += submsg;
				  }
			  }
		  }
		
		if ( _w >= 0 && _w < 1 )
		  subbelief *= _w;
		
		
		if ( !empty( state().observed.shape() ) )
		  {
			belief().to_linear(false);
			Matrix3 wobs( state().observed, true );
			Matrix3 subobs( wobs( R, _all, _all ) );
			subbelief = where( subobs >= 0, subobs * max( 1e-10f, subbelief ), subbelief );
			//subbelief = where( subobs >= 0, subobs, subbelief );
			belief().to_linear(true);
		  }
		else if ( force_linear || _sampling )
		  belief().to_linear(true);
		
		
		if ( _sampling )
		  {
			Matrix3 & r = belief().data;
			Matrix3  & rsub = subbelief;
			
			OPENMP(single)
			{
			  state().sample.resize(r.extent(0), r.extent(1));
			  belief().normalized = true;
			  belief().domain = mmm2_belief_state<T>::LINEAR;
			}
			
			if ( !R.empty() )
			  {
				
				IMatrix2 wsmp( state().sample, true );
				IMatrix2 subsmp( wsmp( R, _all ) );
				
				for ( int i = 0; i < rsub.extent(0); ++i )
				  for ( int j = 0; j < rsub.extent(1); ++j )
					subsmp( i, j ) = Random<T>::get().sample_multinomial( rsub( i, j, _all ) );
				
				frommaxindices2( subsmp, subbelief );		 
			  }
			
		  }
		
      }
    
    OPENMP(barrier);
    
    return belief().data;
  }
  
  void set_sampling( bool b ) { _sampling = b; }
  
  blitz::TinyVector<int,3> check_shape( int exclude = -1 )
  {
    blitz::TinyVector<int,3> s(0,0,0);
    for (int i = -1; i < int( size() ); ++i)
      {
		if ( exclude >= 0 && exclude == i ) continue;
		const Matrix3 & current = i >= 0 ? message(i).data : state().observed;		
		blitz::TinyVector<int,3> si = current.shape();
		if ( empty( si ) ) continue;
		if ( empty( s ) )
		  s = si;
		else if ( different( si, s ) && 
				  ! ( ( ( si[0] == 1 && si[1] == 1 ) || ( s[1] == 1 && s[1] == 1 ) ) && si[2] == s[2] ) )
		  throw std::runtime_error(str(boost::format("message/belief shape mismatch"
													 " current shape: (%ix%ix%i) "
													 " _messages[%i].shape(): (%ix%ix%i)")
									   % s[0] % s[1] % s[2]
									   % i % si[0] % si[1] % si[2]
									   ));
#ifndef _WIN32
		if(any(blitz_isnan(current)))
		  {
			throw std::runtime_error( str( boost::format("message buffer nan in message %i: %s")%i%_used_slots[i]));
		  }
#endif
	  }
	
	OMP_LOG("buffer update end");
	
    return s;
	
  }
  
  void set_observed(Matrix3 observed = Matrix3())
  {
    state().observed.reference(observed);
  }
  
  void set_update_factor( T w )
  {
    _w = w;
  }
  
  Matrix3 & observed()
  {
    return state().observed;
  }
  
  void touch() 
  {
    state().touch();
  }
  
  bool dirty()
  {
    return state().dirty;
  }
  
  void clear( bool do_clear_history = false )
  {
    _sampling = false;
    if ( do_clear_history )
      clear_history();
    else
      state().clear();
  }
  
  bool slot_in_use( size_t i )
  {
    return _used_slots.find( i ) != _used_slots.end();
  }
  
  void unoccupy_slot( size_t i )
  {
    _used_slots.erase( i );
  }
  
  void occupy_slot( size_t i, std::string name = "incognito" )
  {
    if ( slot_in_use( i ) )
      throw std::runtime_error( str( boost::format("message buffer slot %i already in use")%i));
    _used_slots[i] = name;
    if ( i >= size() )
      resize( i + 1 );
  }
  
  size_t free_slot()
  {
    size_t i = 0;
    while ( slot_in_use( i ) ) ++i;
    return i;
  }
  
  blitz::Array<unsigned,2> & get_sample() 
  {
    return state().sample;
  }
  
  void clear_history()
  {
    _states.clear();
    advance();
  }
  
  void set_history_capacity( size_t n )
  {
    if ( ( n + 1 ) != _states.capacity() )
    {
	    _states.clear();
	    _states.set_capacity( n + 1 );
	    advance();
    }
  }
  
  int get_history_capacity( )
  {
    return (_states.capacity()-1);
  }
  
  int get_history_size( )
  {
    return (_states.size()-1);
  }
 
  int active_state()
  {
    return _active_state;
  }

  void advance_history()
  {
    advance();
  }
  
  bool history_full()
  {
    return _states.size() == _states.capacity();
  }
  
  void select_history( size_t t_back )
  {
    if ( t_back >= _states.size() )
      throw std::runtime_error( str( boost::format("mmm2_belief_message_buffer select_history selected: %i max: %i")
				     % t_back % ( _states.size() - 1 ) ) );
    _active_state = t_back;
  }
  
  typename belief_state::distribution_type type()
  {
    return _type;
  }
  
  void set_type( typename belief_state::distribution_type type )
  {
    if ( type != _type )
      {
	_type = type;
	BOOST_FOREACH( state_t & state, _states )
	  state.set_type( type );
      }
  }
private:
  
  typedef std::map<size_t,std::string> slot_map_t;

  //--->
  struct state_t
  {
    
    state_t( typename belief_state::distribution_type type_ = belief_state::MULTINOMIAL ) 
      : dirty(true), maxima_dirty(true), sampled(false), exclude(-1)
    {
      set_type( type_ );
    }

    state_t & operator=( const state_t & other )
    {
      messages.resize( other.messages.size() );
      std::copy( other.messages.begin(), other.messages.end(), messages.begin() );
      sample.reference( other.sample );
      maxima.reference( other.maxima );
      belief = other.belief;
      observed.reference( other.observed );
      dirty = other.dirty;
      maxima_dirty = other.maxima_dirty;
      exclude = other.exclude;
      type = other.type;
      return *this;
    }
    
    void clear()
    {
      BOOST_FOREACH( belief_state & message, messages )
		message.clear();
      belief.clear();
      exclude = -1;
      dirty = false;
	  sampled = false;
      observed.free();
      sample.free();
      maxima.free();
    }
    
    void touch()
    {
      maxima_dirty = dirty = true; 
    }
    
	void set_sample( IMatrix2 & isample )
	{
	  sample.reference( isample );
	  sampled = dirty = true;
	}

    void set_type( typename belief_state::distribution_type type_ )
    {
      type = type_;
      belief.set_type( type );
      BOOST_FOREACH( belief_state & message, messages )
		message.set_type( type );
      dirty = true;
    }
    
    void resize( size_t n )
    {
      messages.resize(n);
      set_type( type );
    }
    
	void release_buffer()
	{
	  belief.clear();
	  touch();
	}

    std::vector<belief_state> messages; //well...conditioned beliefs...
    blitz::Array<unsigned,2> sample;
    blitz::Array<unsigned,2> maxima;
    belief_state belief;
    Matrix3 observed;
    bool dirty;
    bool maxima_dirty;
	bool sampled;
    int exclude;
    typename belief_state::distribution_type type;
  };
  //<---
  
  state_t & state()
  {
    return _states[_active_state];
  }
  
  boost::circular_buffer<state_t> _states;
  
  void advance()
  {
    _states.push_front( state_t( _type ) );
  }
  
  slot_map_t _used_slots;
  bool _sampling;
  typename belief_state::distribution_type _type;
  size_t _active_state;
  T _w;
};


template <class ostream, typename T>
ostream & operator<< (ostream& os, mmm2_belief_message_buffer<T> & buffer)
{
  buffer.do_stream_out(os);
  return os;
}

#ifdef HAVE_OPENMP_DEFINED
#define HAVE_OPENMP 1
#include "omp_macros_undef.hpp"
#undef HAVE_OPENMP_DEFINED
#endif

#endif /* _MMM2_BELIEF_MESSAGE_BUFFER_HPP_ */
