/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "mmm_compound_layer.hpp"
#include "normrho.hpp"
#include "enum_properties.hpp"
#include <boost/ref.hpp>

template<typename T>
message_component<T> & mmm_compund_layer_base<T>::topdown_component( size_t i )
{
  size_t buffer_size = _topdown_buffers.size();
  if ( i >= buffer_size )
    {
      OPENMP(barrier)
	OPENMP(single)
      {
	_topdown_buffers.resize( i + 1 );
	for ( size_t ibuffer = buffer_size; ibuffer <= i; ++ibuffer )
	  {
	    _topdown_buffers.at(ibuffer).set_buffer( buffer(),
          str( boost::format("topdown%i")%ibuffer) );
	  }
      }
    }
  return _topdown_buffers.at(i);
}

template <class emission_model_type>
mmm_compund_layer<emission_model_type>::mmm_compund_layer(int k, int d, int m, int s) 
  : _emission_model( )
  , _tfmm(k)
  , _mrf(k)
  , _do_mrf(false)
  , _do_tfmm(false)
  , _n_iters(0)
  , _k(k)
  , _new_data(false)
  , _use_background_model(false)
  , _annealing(NO_ANNEALING)
  , _linear_annealing_rate(0.1)
  , _max_annealing_temperature(100)
{
  _emission_model.set_buffer( buffer(), "emission" );
  _mrf.set_buffer( buffer(), "mrf" );
  _tfmm.set_buffer( buffer(), "tfmm" );
  _tfmm.stepped_forward.connect( boost::bind( &emission_model_type::step_forward, boost::ref(_emission_model), _1) ); 
  _tfmm.stepped_backward.connect( boost::bind( &emission_model_type::step_backward, boost::ref(_emission_model)) ); 
  _tfmm.learned.connect( boost::bind( &emission_model_type::learn, boost::ref(_emission_model)) ); 
  //_tfmm.stepped_forward.connect( _emission_model.step_forward ); 
  //_tfmm.stepped_backward.connect( _emission_model.step_backward ); 
}
template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::update()
{
  if ( _do_tfmm )
	_tfmm.update();
  if ( _do_mrf )
	_mrf.update();
  _emission_model.update();
}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::init_properties()
{
  PBase::add_ivar_property_rw(std::string("do-mrf"), &This::_do_mrf);
  PBase::add_ivar_property_rw(std::string("do-tfmm"), &This::_do_tfmm);
  PBase::add_ivar_property_rw(std::string("linear-annealing-rate"), &This::_linear_annealing_rate);
  PBase::add_ivar_property_rw(std::string("max-annealing-temperature"), &This::_max_annealing_temperature);
  PBase::add_ivar_property_rw(std::string("annealing-steps"), &This::_annealing_steps);
  PBase::add_ivar_property_rw(std::string("use-background-model"), &This::_use_background_model );
  PBase::add_ivar_property_rw(std::string("n-iters"), &This::_n_iters);
  PBase::add_property( std::string("k"), 
					   PBase::make_ivar_getter( &This::_k ), 
					   typename PBase::MemFunSetter( &This::set_k ) );

  enum_property_access<SearisePropertyTypes, This, annealing_t> 
	annealing_accessor( &This::_annealing );
  annealing_accessor
	.add_choice( "none"                     , NO_ANNEALING                    )
	.add_choice( "linear"                   , LINEAR_ANNEALING                )
	.add_choice( "ramped-linear"            , RAMPED_LINEAR_ANNEALING         )
	.add_choice( "classical"                , CLASSICAL_ANNEALING             )
	;
  PBase::add_property( "annealing", annealing_accessor, annealing_accessor );

  //read-only
  PBase::add_property( std::string("emission"), PBase::make_pivar_getter( &This::_emission_model ) );
  PBase::add_property( std::string("tfmm"),     PBase::make_pivar_getter( &This::_tfmm           ) );
  PBase::add_property( std::string("mrf"),      PBase::make_pivar_getter( &This::_mrf            ) );
  PBase::add_property(std::string("generator"), 
					  make_blitz_ivar_getter1( &This::_generator ) ,
					  make_blitz_ivar_setter1( &This::_generator, true ) );
  PBase::add_ivar_property_ro(std::string("L"), &This::_L);
  PBase::add_ivar_property_ro(std::string("Lz"), &This::_Lz);
  PBase::add_ivar_property_ro(std::string("Lp"), &This::_Lp);
  PBase::add_ivar_property_ro(std::string("Lmu"), &This::_Lmu);

}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::set_k( property_type pk )
{
  _emission_model.set_property( "k", pk );
  _tfmm.set_property( "k", pk );
  _mrf.set_property( "k", pk );
  _k = cast<int>( pk );
}


template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::begin_mp_frame()
{
  _saliency.free();
  _data.free();
  _idata.free();
  _new_data = false;
  update();
  if ( _emission_model.ready() )
    {
      if ( _do_tfmm )
		_tfmm.begin_mp_frame();
      else
		buffer().clear();
    }
  _emission_model.begin_mp_frame();
  if ( _generator.numElements() > 0 )
    {
      buffer().set_observed( _generator );
    }
}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::break_sequence()
{
  _emission_model.break_sequence();
  if ( _emission_model.ready() )
	{
	  if ( _do_tfmm ) 
		_tfmm.reset();
	}  
}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::end_mp_frame()
{
  if ( _data.numElements() == 0 ) return;
  
  _emission_model.end_mp_frame( _data );

  if ( _emission_model.ready() )
	{
	  if ( _do_mrf ) 
		_mrf.end_mp_frame();
	  if ( _do_tfmm )
        {
            // this has the effect that the ssaliency is used by the embedded markov chain only for learning
            // i.e. the chain keeps making predictions in the dark but does not learn on those
            _tfmm.set_saliency( _saliency );
		    _tfmm.end_mp_frame();
            _tfmm.set_saliency( blitz::Array<T,2>() );
        }
	}
  OPENMP(barrier);
}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::flush()
{
  _emission_model.flush();
  _tfmm.flush();
  _mrf.flush();
}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::do_mp ( bool sample, bool rf_sampling )
{	
  //stay safe, and allow the emission model to consistenize its internal state
  _emission_model.update();

  if ( !_emission_model.ready() ) return;
  
  //fetch and interpret relevant parameters
  blitz::TinyVector<int,2> outshape = _emission_model.outshape ( _data.shape() );
  bool segment = _emission_model.segment();
  bool anneal = _annealing != NO_ANNEALING;
  bool linear_annealing = _annealing == LINEAR_ANNEALING || _annealing == RAMPED_LINEAR_ANNEALING;


  CHECKNAN(_data,"data in mmm_compund_layer::do_mp");

  if ( sample )
	{
	  assert(segment);
	  if ( rf_sampling )
		_emission_model.sample_receptive_fields( _data, _n_iters );
	  else
		_emission_model.sample_posterior_gated( _data, _n_iters );
	}
  else
	{

	  T max_temperature = _max_annealing_temperature;

	  //1. update upward messages
	  if ( _new_data )
		{
		  T rate = linear_annealing ? _linear_annealing_rate : 1 / max_temperature;
		  if ( segment )
			_emission_model.update_upward_msg_gated ( _data, rate, linear_annealing );
		  else
			{
			  if ( _data.numElements() > 0 )
				_emission_model.update_upward_msg ( _data );
			  else if  ( _idata.numElements() > 0 )
				_emission_model.update_upward_msg ( _idata );
			}
		  OPENMP(barrier);
		  OPENMP(single)
			_new_data = false;		  
		}
	  
	  
	  //2. iteratively update lateral messages
	  if ( _do_mrf || segment )
		{
		  bool have_temperature = _annealing == CLASSICAL_ANNEALING;
		  for ( int anneal_step = 0, anneal_steps = ( have_temperature ? _annealing_steps : 1 ); anneal_step < anneal_steps; ++anneal_step )
			{
			  T temperature = anneal 
				? T( 1 + ( max_temperature - 1 ) * ( anneal_steps - anneal_step - 1 ) / ( std::max( 1, anneal_steps - 1 ) ) ) 
				: T(1);
			  for ( int iteration = 0; iteration < _n_iters; ++iteration )
				{
				  if ( _do_mrf )  
					{ 
					  _mrf.update_msg();
					  OPENMP(barrier);
					}
				  if ( segment ) 
					{
					  T rel_iter = float(iteration) / std::max( _n_iters - 1, 1 );
					  T rate = linear_annealing 
						? ( _annealing == RAMPED_LINEAR_ANNEALING ? ( rel_iter + ( 1 - rel_iter ) * _linear_annealing_rate ) : _linear_annealing_rate )
						: 1 / temperature;
					  _emission_model.update_upward_msg_gated ( _data, rate, linear_annealing );
					  OPENMP(barrier);
					}
				}
			}
		}
	}

  if ( !_emission_model.converged() ) 
	std::cout << "emission model did not converge!" << std::endl;
  
  //CHECKNAN(buffer().update(),"beliefs in mmm_compund_layer::do_mp");

}

template <class emission_model_type>
const blitz::Array<typename emission_model_type::num_type,3> &
mmm_compund_layer<emission_model_type>::data()
{
  return _data;
}

template <class emission_model_type>
void
mmm_compund_layer<emission_model_type>::set_data( const blitz::Array<T,3> & data )
{
  _data.reference(data);
  _idata.free();
  _new_data = true;
}

template <class emission_model_type>
const IMatrix2 &
mmm_compund_layer<emission_model_type>::idata()
{
  return _idata;
}

template <class emission_model_type>
void
mmm_compund_layer<emission_model_type>::set_idata( const IMatrix2 & data )
{
  _idata.reference(data);
  _data.free();
  _new_data = true;
}

template <class emission_model_type>
void
mmm_compund_layer<emission_model_type>::calc_downward_message( bool maxinfer, bool logarithmic )
{
  if ( !_emission_model.ready() ) return;  
  if ( _emission_model.segment() )
	_emission_model.calc_downward_msg_gated ( _data, logarithmic );
  else
	_emission_model.calc_downward_msg ( _data, logarithmic );
}

template <class emission_model_type>
blitz::Array<typename emission_model_type::num_type,3> & 
mmm_compund_layer<emission_model_type>::downward_message()
{
  return _emission_model.get_downward_message();
}


template <class emission>
blitz::Array<typename emission::num_type,4> mmm_compund_layer<emission>::inner_prediction()
{
  return _emission_model.get_inner_prediction();
}


template <class emission_model_type>
blitz::Array<typename emission_model_type::num_type,1>
mmm_compund_layer<emission_model_type>::batchlearn_end_epoch( Matrix1 f_in)
{

	blitz::Array<T,1> f = _emission_model.batchlearn_end_epoch(f_in);

	if ( !_emission_model.ready() ) return blitz::Array<T,1>();  
	
	_L   = cast<T>( _emission_model.get_property("L"  ) );
	_Lp  = cast<T>( _emission_model.get_property("Lp" ) );
	_Lmu = cast<T>( _emission_model.get_property("Lmu") );
	if (_do_tfmm) 
	  {
		_tfmm.batchlearn_end_epoch(f);
		_L   += cast<T>( _tfmm.get_property("L"  ) );
	  }
	if (_do_mrf) 
	  {
		_mrf.batchlearn_end_epoch(f);
		_L   += cast<T>( _mrf.get_property("L"  ) );
		_Lp  += cast<T>( _mrf.get_property("Lp" ) );
		_Lmu += cast<T>( _mrf.get_property("Lmu") );
	  }
		
	return f;
}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::batchlearn_begin_epoch()
{
	_emission_model.batchlearn_begin_epoch();

	if ( _emission_model.ready() )
	  {  
		int k = cast<int>( _emission_model.get_property("k") );
		if ( _do_tfmm )
		  {
			if ( cast<int>( _tfmm.get_property("k") ) != k )
			  _tfmm.set_property("k",k);
			_tfmm.batchlearn_begin_epoch();
		  }
		if ( _do_mrf )
		  {
			if ( cast<int>( _mrf.get_property("k") ) != k )
			  _mrf.set_property("k",k);
			_mrf.batchlearn_begin_epoch();
		  }
	  }
}


template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::serialize(boost::archive::polymorphic_iarchive & ar, const unsigned int file_version)
{
  serialize_impl( ar, file_version );
}

template <class emission_model_type>
void mmm_compund_layer<emission_model_type>::serialize(boost::archive::polymorphic_oarchive & ar, const unsigned int file_version)
{
  serialize_impl( ar, file_version );
}

template <class emission_model_type>
blitz::Array<typename emission_model_type::num_type,3> mmm_compund_layer<emission_model_type>::backtrack( blitz::Array<T,3> & Pin, bool use_saliency )
{
  if ( !_emission_model.ready() ) return blitz::Array<T,3>();
  else return _emission_model.backtrack( Pin, use_saliency );
}

template <class emission_model_type>
blitz::Array<typename emission_model_type::num_type,2> mmm_compund_layer<emission_model_type>::backtrack( blitz::Array<T,2> & Pin, bool use_saliency )
{
  if ( !_emission_model.ready() ) return blitz::Array<T,2>();
  else return _emission_model.backtrack( Pin, use_saliency );
}

template <class emission_model_type>
blitz::Array<typename emission_model_type::num_type,3> mmm_compund_layer<emission_model_type>::propagate( blitz::Array<T,3> & Pin, bool use_saliency  )
{
  if ( !_emission_model.ready() ) return blitz::Array<T,3>();  
  else return _emission_model.propagate( Pin, use_saliency );
}

template <class emission_model_type>
blitz::Array<typename emission_model_type::num_type,2> mmm_compund_layer<emission_model_type>::propagate( blitz::Array<T,2> & Pin, bool use_saliency )
{
  if ( !_emission_model.ready() ) return blitz::Array<T,2>();
  else return _emission_model.propagate( Pin, use_saliency );
}


template <class emission_model_type>
template<class Archive>
void mmm_compund_layer<emission_model_type>::serialize_impl(Archive & ar, const unsigned int version)
{
  ar & boost::serialization::make_nvp( "emission-model", _emission_model );
  ar & boost::serialization::make_nvp( "tfmm", _tfmm );
  ar & boost::serialization::make_nvp( "mrf", _mrf );
  ar & boost::serialization::make_nvp( "do-mrf", _do_mrf );
  ar & boost::serialization::make_nvp( "do-tfmm", _do_tfmm );
  if ( version < 3 )
	{
	  int dummy;
	  ar & boost::serialization::make_nvp( "n-iters", dummy );
	}
  // avoid nans since seriliaztion cannot handle them
  ar & boost::serialization::make_nvp( "k", _k );

#define killnan(v) if ( std::isnan( v ) || std::isinf( v ) ) v = 0;
  killnan( _L );
  killnan( _Lz );
  killnan( _Lmu );
  killnan( _Lp);
#undef killnan

  ar & boost::serialization::make_nvp( "L", _L );
  ar & boost::serialization::make_nvp( "Lz", _Lz );
  ar & boost::serialization::make_nvp( "Lmu", _Lmu );
  ar & boost::serialization::make_nvp( "Lp", _Lp );

  if ( version == 1 )
	{
	  bool dummy;
	  ar & boost::serialization::make_nvp( "estimate-s", dummy );
	}
  _emission_model.update();
  if ( _do_tfmm )
    _tfmm.update();
  if ( _do_mrf )
    _mrf.update();
}

template <class emission>
blitz::Array<typename emission::num_type,2> mmm_compund_layer<emission>::latent_self_information()
{
  BLITZ_INDICES;
  blitz::Array<T,2> I;
  Matrix3 prediction  = _emission_model.get_belief(true);
  _emission_model.get_msg().to_linear();
  Matrix3 measurement =   _emission_model.get_msg().data;
  // FIXME: different for binomial, this is for multinomial
  I = safe_log( sum( prediction * measurement, i3 ) );
  return I;
}

template <class emission>
blitz::Array<typename emission::num_type,2> mmm_compund_layer<emission>::foreground_probability()
{
  return _emission_model.calc_foreground_probability( _data );
}

template <class emission>
blitz::Array<typename emission::num_type,2> mmm_compund_layer<emission>::foreground_coverage()
{
  return _emission_model.calc_foreground_coverage( _data );
}

template <class emission>
blitz::Array<typename emission::num_type,2> mmm_compund_layer<emission>::emission_self_information()
{
  if ( _data.numElements() == 0 )
    return blitz::Array<T,2>();
  else
    return _emission_model.calc_self_information( _data );
}

template <class emission>
void mmm_compund_layer<emission>::set_input_likelihood( Matrix2 l, size_t channel, int d1, int s1, int d2, int s2 )
{
  _emission_model.set_input_likelihood( l, channel, d1, s1, d2, s2 );
}

template <class emission>
blitz::Array<typename emission::num_type,2> & mmm_compund_layer<emission>::get_input_likelihood_gradient( size_t i )
{
  return _emission_model.get_input_likelihood_gradient( _data, i );
}

template <class emission>
blitz::Array<typename emission::num_type,2> mmm_compund_layer<emission>::local_likelihood(bool marginal)
{
  if ( _data.numElements() == 0 )
    return Matrix2();
  else
    return _emission_model.calc_local_likelihood( _data, marginal );
}

template <class emission>
void mmm_compund_layer<emission>::select_parameter_set( ParameterSet params )
{
  switch( params )
	{
	case GENERATIVE:
	  _emission_model.request_parameter_mode( emission::GENERATIVE );
	  break;
	case DISCRIMINATIVE:
	  _emission_model.request_parameter_mode( emission::DISCRIMINATIVE );
	  break;
	}
}

template <class emission>
void mmm_compund_layer<emission>::begin_likelihood_learning()
{
  _emission_model.begin_likelihood_learning();
}

template <class emission>
void mmm_compund_layer<emission>::end_likelihood_learning()
{
  _emission_model.end_likelihood_learning();
}

template <class emission>
void mmm_compund_layer<emission>::begin_discriminative_gradient_descent()
{
  _emission_model.begin_discriminative_gradient_descent();
}

template <class emission>
void mmm_compund_layer<emission>::end_discriminative_gradient_descent()
{
  _emission_model.end_discriminative_gradient_descent();
}

template <class emission>
void mmm_compund_layer<emission>::begin_discriminative_hessian_estimation()
{
  _emission_model.begin_discriminative_hessian_estimation();
}

template <class emission>
void mmm_compund_layer<emission>::end_discriminative_hessian_estimation()
{
  _emission_model.end_discriminative_hessian_estimation();
}

template <class emission>
void  mmm_compund_layer<emission>::add_output_hessian( const Matrix3 & delta, GradientMode kind )
{
  _emission_model.add_output_hessian( delta, (GradientMode)kind );
}

template <class emission>
void  mmm_compund_layer<emission>::add_output_gradient( const Matrix3 & delta, GradientMode kind )
{
  _emission_model.add_output_gradient( delta, (GradientMode)kind );
}

template <class emission>
void mmm_compund_layer<emission>::set_saliency( const blitz::Array<T,2> & saliency)
{
  _saliency.reference( saliency );
  _emission_model.set_S( _saliency );
}

template <class emission>
void mmm_compund_layer<emission>::backpropagate_error()
{
  _emission_model.backpropagate_error_complete( _data );
}

template <class emission>
blitz::Array<typename emission::num_type,3> & mmm_compund_layer<emission>::backpropagated_error()
{
  return _emission_model.get_backpropagated_error();
}

#include "mmm2.hpp"
#include "nmm2.hpp"
#include "bmm2.hpp"
#include "mcmm2.hpp"
//#include "tmm2.hpp"
//#include "gfm.hpp"

template class mmm_compund_layer_base<float>;

template class mmm_compund_layer< MultinomialMixtureModel2<float> >;
template class mmm_compund_layer< NormalMixtureModel2<float> >;
template class mmm_compund_layer< BinomialMixtureModel2<float> >;
template class mmm_compund_layer< MarkovChainMixtureModel2<float> >;
//template class mmm_compund_layer< TransitionMixtureModel2<float> >;
//template class mmm_compund_layer< GatedFactorModel<float> >;
