/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _MMM_COMPUND_LAYER_HPP_
#define _MMM_COMPUND_LAYER_HPP_

#include <boost/archive/polymorphic_iarchive.hpp>
#include <boost/archive/polymorphic_oarchive.hpp>

#include "MixtureModel2Base.hpp"
#include "mrf2.hpp"
#include "tfmm.hpp"
#include "Random.hpp"

///this combines all parts for a "compund layer", i.e. a complete S- or C-layer
///this includes the emission model, the markov chain, and the markov random field
///slot layout of the emission model's message buffer:
///	slot 0: botton->up message
/// slot 1: forward message
/// slot 2: lateral message
/// all other slots can take arbitrary projections (i.e. "top-down")


  typedef enum { GENERATIVE, DISCRIMINATIVE } ParameterSet;

///functions as a fully functional base class for polymorphic access, regardless of the emission model
template <typename T>
class mmm_compund_layer_base
  : public HasProperties< SearisePropertyTypes, mmm_compund_layer_base<T> >
{

public:

  typedef mmm_compund_layer_base<T> This;
  typedef HasProperties<SearisePropertyTypes, This> PBase;
  typedef TemporalForwardMarkovModel<T> tfmm_type;
  typedef MarkovRandomField<T> mrf_type;
  typedef mmm2_belief_message_buffer<T> message_buffer;
  typedef typename message_buffer::belief_state belief_state;  

  typedef blitz::Array<unsigned int,2> IMatrix2;
 
  virtual void select_parameter_set( ParameterSet ) = 0;
  virtual void begin_discriminative_gradient_descent() = 0;
  virtual void end_discriminative_gradient_descent() = 0;
  virtual void begin_discriminative_hessian_estimation() = 0;
  virtual void end_discriminative_hessian_estimation() = 0;
  virtual void begin_likelihood_learning() = 0;
  virtual void end_likelihood_learning() = 0;

  virtual void  add_output_hessian( const Matrix3 & delta, GradientMode kind ) = 0;
  virtual void  add_output_gradient( const Matrix3 & delta, GradientMode kind ) = 0;

  ///access the message-buffer
  message_buffer & buffer() { return _buffer; }
  ///register active bottom-up data
  virtual void set_data( const blitz::Array<T,3> & data ) = 0;
  ///register active bottom-up data (indexed)
  virtual void set_idata( const IMatrix2 & data ) = 0;
  ///pre-timestep setup
  virtual void end_mp_frame() = 0;
  ///post-timestep cleanup
  virtual void begin_mp_frame() = 0;
  ///triggers the intra-layer schedule, indicate is the data has been seen before, so one update may be spared
  virtual void do_mp( bool sample = false, bool rf_sampling = false ) = 0;
  ///forget everything you learned...
  virtual void flush() = 0;
  virtual void backpropagate_error() = 0;
  virtual blitz::Array<T,3> & backpropagated_error() = 0;
  ///convenience method to perform downward message, does not currently do any leave-one-out!
  virtual void calc_downward_message(bool maxinfer = false, bool logarithmic = true) = 0;
  virtual blitz::Array<T,3> & downward_message() = 0;
  ///does all the neccessary stuff with all involved components
  virtual void batchlearn_begin_epoch() = 0;
  ///does all the neccessary stuff with all involved components
  typedef blitz::Array<T,1> Matrix1;
  virtual blitz::Array<T,1> batchlearn_end_epoch( Matrix1 f_in = Matrix1() ) = 0;
  ///get data
  virtual const blitz::Array<T,3> & data() = 0;
  ///get data
  virtual const IMatrix2 & idata() = 0;
  ///propagate a multinomial probability distribution using segmentation units
  virtual blitz::Array<T,3> propagate( blitz::Array<T,3> & Pin, bool use_saliency = false ) = 0;
  ///backtrack a multinomial probability distribution using segmentation units
  virtual blitz::Array<T,3> backtrack( blitz::Array<T,3> & Pin, bool use_saliency = false  ) = 0;
  ///backtrack a binomial probability distribution using segmentation units
  virtual blitz::Array<T,2> backtrack( blitz::Array<T,2> & Pin, bool use_saliency = false  ) = 0;
  ///propagate a binomial probability distribution using segmentation units
  virtual blitz::Array<T,2> propagate( blitz::Array<T,2> & Pin, bool use_saliency = false  ) = 0;
  ///returns the self-information in the latent variables
  virtual blitz::Array<T,2> latent_self_information() = 0;
  ///returns the self-information in the input data
  virtual blitz::Array<T,2> emission_self_information() = 0;
  ///returns the probability for each input that it is modeled as foreground
  virtual blitz::Array<T,2> foreground_probability() = 0;
  virtual blitz::Array<T,2> foreground_coverage() = 0;
  ///sets the pixel-based saliency map used for weighing learning and recognition
  virtual void set_saliency( const blitz::Array<T,2> & saliency) = 0;
  ///breaks any inter-frame predictions and marks the start of a new sequence
  virtual void break_sequence() = 0;
  
  // return the inner prediction of the model ( mcmm2 )
  virtual  blitz::Array<T,4> inner_prediction() = 0;
  
  // update all components
  virtual void update() = 0;

  //stub, no properties here
  static void init_properties() {}
  ///yields auxiliary message components by index for topdown messages, components are automatically spawned as needed
  message_component<T> & topdown_component( size_t i );
  ///returns the probability of the current data in the mixture model per output code
  virtual blitz::Array<T,2> local_likelihood(bool marginal = true) = 0;
  ///set likelihood contributions of the input (i.e. from externel models)
  virtual void set_input_likelihood( Matrix2 l, size_t channel = 0, int d1 = 1, int s1 = 1, int d2 = 0, int s2 = 0 ) = 0;
  ///get the detection ce-error gradient w.r.t. corresponding input likelihood contribution
  virtual blitz::Array<T,2> & get_input_likelihood_gradient( size_t i ) = 0;

  virtual void serialize(boost::archive::polymorphic_iarchive & ar, const unsigned int file_version) = 0;
  virtual void serialize(boost::archive::polymorphic_oarchive & ar, const unsigned int file_version) = 0;
  
  ::Random<T> & rng() { return _rng; }

private:

	
  std::vector< message_component<T> > _topdown_buffers;
  message_buffer _buffer;
  ::Random<T> _rng;

};

template <class emission_model_type>
class mmm_compund_layer
  : public mmm_compund_layer_base< typename TypeTraits<emission_model_type>::num_type >,
	public HasProperties< SearisePropertyTypes, mmm_compund_layer< emission_model_type > >
{
	
public:
	
  typedef SeariseProperty property_type;
  typedef mmm_compund_layer< emission_model_type > This;
  typedef HasProperties< SearisePropertyTypes , This> PBase;
  typedef typename emission_model_type::num_type T;
  typedef TemporalForwardMarkovModel<T> tfmm_type;
  typedef MarkovRandomField<T> mrf_type;
  typedef typename emission_model_type::message_buffer message_buffer;
  typedef typename message_buffer::belief_state belief_state;
  typedef mmm_compund_layer_base< typename TypeTraits<emission_model_type>::num_type > Base;

  mmm_compund_layer(int k = 100, int d = 3, int m = 0, int s = 3);
  
  static void init_properties();
  
  message_buffer & buffer() { return Base::buffer(); }
  ///pre-timestep setup
  virtual void end_mp_frame();
  //set the current data
  virtual void set_data( const blitz::Array<T,3> & data );
  //set the current data (indexed)
  virtual void set_idata( const IMatrix2 & data );
  ///post-timestep cleanup
  virtual void begin_mp_frame();
  ///triggers the intra-layer schedule, indicate is the data has been seen before, so one update may be spared
  virtual void do_mp( bool sample = false, bool rf_sampling = false );
  ///forget everything you learned...
  virtual void flush();
  ///does all the neccessary stuff with all involved components
  virtual void batchlearn_begin_epoch();
  ///does all the neccessary stuff with all involved components
  virtual blitz::Array<T,1> batchlearn_end_epoch(Matrix1 f_in = Matrix1());
  ///
  virtual blitz::Array<T,3> & backpropagated_error();
  ///get the downward projection
  virtual void calc_downward_message( bool maxinfer = false, bool logarithmic = true);
  virtual blitz::Array<T,3> & downward_message();
  ///get data
  virtual const blitz::Array<T,3> & data();
  ///get data (indexed)
  virtual const IMatrix2 & idata();
  ///read from archive
  virtual void serialize(boost::archive::polymorphic_iarchive & ar, const unsigned int file_version);
  ///write to archive
  virtual void serialize(boost::archive::polymorphic_oarchive & ar, const unsigned int file_version);
  ///backtrack a multinomial probability distribution using segmentation units
  virtual blitz::Array<T,3> backtrack( blitz::Array<T,3> & Pin, bool use_saliency = false  );
  ///backtrack a binomial probability distribution using segmentation units
  virtual blitz::Array<T,2> backtrack( blitz::Array<T,2> & Pin, bool use_saliency = false  );
  ///propagate a multinomial probability distribution using segmentation units
  virtual blitz::Array<T,3> propagate( blitz::Array<T,3> & Pin, bool use_saliency = false  );
  ///propagate a binomial probability distribution using segmentation units
  virtual blitz::Array<T,2> propagate( blitz::Array<T,2> & Pin, bool use_saliency = false  );
  virtual blitz::Array<T,2> latent_self_information();
  virtual blitz::Array<T,2> emission_self_information();
  virtual blitz::Array<T,2> foreground_probability();
  virtual blitz::Array<T,2> foreground_coverage();
  ///returns the probability of the current data in the mixture model per output code
  virtual blitz::Array<T,2> local_likelihood( bool marginal = true );
  ///set likelihood contributions of the input (i.e. from externel models)
  virtual void set_input_likelihood( Matrix2 l, size_t channel = 0, int d1 = 1, int s1 = 1, int d2 = 0, int s2 = 0 );
  ///get the detection ce-error gradient w.r.t. corresponding input likelihood contribution
  virtual blitz::Array<T,2> & get_input_likelihood_gradient( size_t i );
  // return the inner prediction of the model ( mcmm2 )
  virtual  blitz::Array<T,4> inner_prediction();
  
  // update all components
  virtual void update();
  ///sets the pixel-based saliency map used for weighing learning and recognition
  virtual void set_saliency( const blitz::Array<T,2> & saliency);
  virtual void break_sequence();
  virtual void select_parameter_set( ParameterSet );
  virtual void begin_discriminative_gradient_descent();
  virtual void end_discriminative_gradient_descent();
  virtual void begin_discriminative_hessian_estimation();
  virtual void end_discriminative_hessian_estimation();
  virtual void begin_likelihood_learning();
  virtual void end_likelihood_learning();
  virtual void  add_output_hessian( const Matrix3 & delta, GradientMode kind );
  virtual void  add_output_gradient( const Matrix3 & delta, GradientMode kind );
  virtual void backpropagate_error();

private:

  friend class boost::serialization::access;
  template<class Archive>
  void serialize_impl(Archive & ar, const unsigned int version);

  void set_k( property_type pk );

  emission_model_type _emission_model;

  std::vector< blitz::Array<T,3> >  _datas;
  std::vector<IMatrix2> _idatas;

  tfmm_type _tfmm;
  mrf_type _mrf;
  bool _do_mrf, _do_tfmm;
  int _n_iters, _k;
  T _L, _Lz, _Lmu, _Lp;
  blitz::Array<T,3> _data, _generator;
  IMatrix2 _idata;
  blitz::Array<T,2> _saliency;
  bool _new_data, _use_background_model;
  typedef enum { NO_ANNEALING, LINEAR_ANNEALING, RAMPED_LINEAR_ANNEALING, CLASSICAL_ANNEALING } annealing_t;
  annealing_t _annealing;
  T _linear_annealing_rate, _max_annealing_temperature, _annealing_steps;

};
//templated class versioning
namespace boost { namespace serialization {
	template <class emission_model_type>
	struct version< mmm_compund_layer<emission_model_type> >
	{
	  BOOST_STATIC_CONSTANT(unsigned int, value = 3);
	};
  }}

#endif
