/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "mrf2.hpp"
#include "digamma.hpp"
#include "normrho.hpp"
#include "optimal_hyperparameters.hpp"
#include "Random.hpp"

#include <boost/format.hpp>

#ifdef HAVE_OPENMP
#include <omp.h>
#endif


#ifdef HAVE_SSE
#include "sse_math.hpp"
#endif

#undef foreach
#define foreach BOOST_FOREACH

#include "trigamma.hpp"

template<class T>  MarkovRandomField<T>::MarkovRandomField(int k,int d) 
  : _k(k)
  ,_d(d)
  ,_initcount(100)
  ,_initialized(false)
  ,_dirty(false)
  ,_in_batch(false)
  ,_init_uniform(true)
  ,_keep_messages(false)
  ,_learn_beta0_interval(0)
  ,_learn_beta0_counter(0)
  ,_iterations(1)
  ,_learning(false)
  ,_factorize_transitions(true)
  ,_do_vb(true)
  ,_beta0(0.5)
  ,_delta(1)
  ,_Lmu(0)
  ,_Lp(0)
  ,_L(0)
{
}

template<class T>  const blitz::Array<T,3>  MarkovRandomField<T>::get_r() { return _r; }
template<class T>  const blitz::Array<T,4> MarkovRandomField<T>::get_beta()  { initialize(); return _beta; }
template<class T>  const blitz::Array<T,4> MarkovRandomField<T>::get_mu()  { update(); return _mu; }

//does all the complicated business of iterating lateral connections
struct MRFIter
{
  MRFIter(int d, int R, int C) 
	: _d(d), _R(R), _C(C), _i(0), _j(0)
  {
#ifdef HAVE_OPENMP
    if ( omp_in_parallel() )
    {
      _istep = omp_get_num_threads();
      _i     = omp_get_thread_num();
    } 
    else 
    {
#endif
      _istep = 1;
      _i = 0;
#ifdef HAVE_OPENMP
    }
#endif
    start_region();
  }
  
  int _d, _R, _C, _i, _j, _i_msg, _ii, _jj, _iimax, _jjmin, _jjmax, _istep;
  
  /* 
     halve connection pattern of all unique connections:
         /
        * -
        | \
     the full pattern emerges by forward and backward passing and scanning
  */
  
  void start_region()
  {
    _iimax = std::max( 0, std::min( _d - 1, _R - _i - 1 ) );
    _jjmin = std::min( 0, std::max( 1 - _d, -_j ) );
    _jjmax = std::max( 0, std::min( _d - 1, _C - _j - 1 ) );
    if ( _j == ( _C - 1 ) ) //nothing to the right
	  {
		  _ii = 1;
		  _jj = _jjmin;
	  }
    else
	  {
		  _ii = 0;
		  _jj = 1; //skip empy first halve row
	  }
	//    std::cout << str( boost::format("region@(%i,%i) in (%i,%i) (%i,%i) to (%i,%i) ii=%i jj=%i\n")%i%j%R%C%0%jjmin%iimax%jjmax%ii%jj );
  }
  
  bool region_end()
  {
    return _ii >= _iimax && _jj >= _jjmax;
  }
  
  int i_region()
  {
    return _ii + _d - 1;
  }
  
  int j_region()
  {
    return _jj + _d - 1;
  }
  
  int msg_index_in()
  {
    return i_region() + j_region() * 2 * ( _d - 1 ) - 1; // omit origin
  }
  
  int msg_index_out()
  {
    // the receivers perspective is point reflected around the origin : p_o = - ( p_i - c ) + c = 2 * c - p
    int i_region_out = 2 * ( _d - 1 ) - i_region();
    int j_region_out = 2 * ( _d - 1 ) - j_region();
    
    return i_region_out + j_region_out * 2 * ( _d - 1 );
  }
  
  //iterator always expands the halve-star "downwards"
  void advance()
  {
    _jj++;
    if ( _jj > _jjmax )
      {
	_ii++;
	if ( _ii > _iimax )
	  {
	    _j++;
	    if ( _j >= _C )
	      {
		_j = 0;
		_i += _istep;
	      }
	    start_region();
	  }
	else
	  _jj = _jjmin;
      }
    //    std::cout << str(boost::format("->(%i,%i)(%i,%i)\n")%i%j%ii%jj);
  }
  
  int i()
  {
	return _i;
  }

  int j()
  {
	return _j;
  }

  int ito()
  {
    return _i + _ii;
  }
  
  int jto()
  {
    return _j + _jj;
  }
  
  int irf()
  {
    return _ii;
  }
  
  int jrf()
  {
    return _jj + _d - 1;
  }
  
  bool atend()
  {
    //nothing to do in the lower right corner
    return _i >= _R || ( _i == ( _R - 1 ) && _j >= ( _C - 2 ) );
  }
  
};

template<class T>  void MarkovRandomField<T>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter(&This::_initialized,false);
  //read-write
  PBase::add_ivar_property_rw(std::string("factorize-joint-statistics"), &This::_factorize_transitions );
  PBase::add_ivar_property_rw(std::string("beta0"), &This::_beta0, uninitializer);
  PBase::add_ivar_property_rw(std::string("delta"), &This::_delta);
  PBase::add_ivar_property_rw(std::string("init-uniform"), &This::_init_uniform);
  PBase::add_ivar_property_rw(std::string("initcount"), &This::_initcount);
  PBase::add_ivar_property_rw(std::string("iterations"), &This::_iterations);
  PBase::add_ivar_property_rw(std::string("keep-messages"), &This::_keep_messages);
  PBase::add_ivar_property_rw("learn-beta0", &This::_learn_beta0_interval);
  PBase::add_ivar_property_rw(std::string("k"), &This::_k, uninitializer );
  PBase::add_ivar_property_rw(std::string("d"), &This::_d, uninitializer );
  //read-only
  PBase::add_ivar_property_ro(std::string("initialized"), &This::_initialized);
  PBase::add_ivar_property_ro(std::string("dirty"), &This::_dirty);
  PBase::add_ivar_property_ro(std::string("in_batch"), &This::_in_batch);
  PBase::add_ivar_property_ro("L", &This::_L);
  PBase::add_ivar_property_ro("Lp", &This::_Lp);
  PBase::add_ivar_property_ro("Lmu", &This::_Lmu);
}

template<class T>  void MarkovRandomField<T>::flush()
{
  
  _beta.free();
  _mu.free();
  _lmudash.free();
  _initialized = false;
}

template<class T>  void MarkovRandomField<T>::batchlearn_begin_epoch()
{
  BLITZ_INDICES;
  update();
  _wx.resize(_k,_d,2*_d-1,_k);
  _wx = 0;
  _learning = _in_batch = true;
  _Lp = 0;
}


template<typename T>
double MarkovRandomField<T>::dL_beta0 ( double la0, void * vparams )
{
  double * params = (double*)vparams;
  double k = params[0];
  double c = params[1];
  double alpha0 = std::exp(la0);
  try {
	return boost::math::digamma(k*k*alpha0) - boost::math::digamma(alpha0) + c;
  }	catch(std::exception & e) {
    std::cerr  << " exception evaluating dl beta0 (mrf2): " << e.what() << " type " << typeid(e).name() << std::endl;
	gsl_error_flag = true;
	return 0;
  }
}

template<typename T>
void MarkovRandomField<T>::optimize_beta0 ( double delta, double lalphamin, double lalphamax )
{
  BLITZ_INDICES;
  typedef std::pair<T,bool> Result;
  double c = sum( where( ( i2 == 0 ) & ( i3 < _d ), 0, _lmudash ) ) / ( _lmudash.numElements() - _d ) ;
  double params[2] = {double(_k),c};
  gsl_function F = {&dL_beta0, params};	
  gsl_set_error_handler(&gsl_error_handler);	
  Result r = find_root(&F, lalphamin, lalphamax);
  bool converged = r.second;
  if (!converged) 
	{
	  std::cout << "mrf2: hyperparameter optimization did not converge" << std::endl;
	  return;
	}
  double lbeta0 = std::log(double(_beta0));
  double lbeta0_opt = r.first;
  lbeta0 = (1-double(delta)) * lbeta0 + double(delta) * lbeta0_opt;
  _beta0 = T(std::exp(lbeta0));
  std::cout << "beta0mrf set to " << _beta0 << " optimal " << std::exp(lbeta0_opt) << std::endl;
}

static double mrf_dL_beta0_covariant (double lb0, void * vparams)
{
  blitz::Array<double,4> n(*static_cast<blitz::Array<double,4>*>(vparams));
  double dL = 0;
  try {
	int K = n.extent(0) * n.extent(1);
	int D = n.extent(1);
	double beta0 = std::exp(lb0);
	double dL0 = boost::math::digamma( beta0 )-boost::math::digamma( K * beta0 );
	for (int j = 0; j < 2*D-1; ++j) 
	  {
		for (int i = (j<D?1:0); i < D; ++i)
		  {
			double N = 0;
			double dLik = 0;
			for (int k1 = 0; k1 < K; ++k1)
			  for (int k2 = 0; k2 < K; ++k2)
				{
				  double n_k = n(k1,i,j,k2);
				  N += n_k;
				  double beta_k = beta0 + n_k;
				  dLik -=  boost::math::digamma( beta_k ) + n_k * searise::math::trigamma( beta_k );
				}
			double beta = K * beta0 + N;
			dLik = dLik / K + boost::math::digamma( beta ) + N * searise::math::trigamma( beta );
			dL += dL0 + dLik;
		  }
	  }
  }	catch(std::exception & e) {
    std::cerr  << " exception evaluating dl beta0 (mrf2): " << e.what() << " type " << typeid(e).name() << std::endl;
	gsl_error_flag = true;
	return 0;
  }
  return dL;
}

template<typename T> void
MarkovRandomField<T>::optimize_beta0_covariant( double delta, double lalphamin, double lalphamax )
{
  typedef std::pair<T,bool> Result;
  gsl_function F = {&mrf_dL_beta0_covariant, &_wx};	
  gsl_set_error_handler(&gsl_error_handler);	
  T dL0 = mrf_dL_beta0_covariant(lalphamin,&_wx);
  T dL1 = mrf_dL_beta0_covariant(lalphamax,&_wx);
  std::cout << " dL(" << std::exp(lalphamin) << ")=" << dL0;
  std::cout << " dL(" << std::exp(lalphamax) << ")=" << dL1;
  std::cout << std::endl;
  Result r = find_root(&F, lalphamin, lalphamax);
  bool converged = r.second;
  if (!converged) 
	{
	  std::cout << "mrf2: hyperparameter optimization did not converge" << std::endl;
	  return;
	}
  double lbeta0 = std::log(double(_beta0));
  double lbeta0_opt = r.first;
  lbeta0 = (1-double(delta)) * lbeta0 + double(delta) * lbeta0_opt;
  _beta0 = T(std::exp(lbeta0));
  std::cout << "beta0mrf set to " << _beta0 << " optimal " << std::exp(lbeta0_opt) << std::endl;
}

//Gibbs-sample the MRF (inplace)

template<class T>  void MarkovRandomField<T>::sample ( IMatrix2 & ir, Matrix3 & ptop, int n )
{
  BLITZ_INDICES;
  if (n <= 0) return;
  update();
  
  Random<T> & rng = Random<T>::get();

  int n1 = ir.extent(0);
  int n2 = ir.extent(1);
  for (int iter = 0; iter < n; ++iter)
	{      
	  for (int i = 0; i < n1; ++i)
		for (int j = 0; j < n2; ++j)
		  {
			Matrix1 p(safe_log(ptop(i,j,_all)));
			//from left
			for (int k = std::max(0,i-_d+1); k <= i; ++k)
			  {
				for (int l = std::max(0,j-_d+1); l < (k==i?j-1:std::min(n2,j+_d-1)); ++l)
				  p += _lmu(int(ir(k,l)),i-k,j-l+_d-1,_all);
			  }
			//from right
			for (int k = i; k < std::min(n1,i+_d-1); ++k)
			  for (int l = (k==i?j+1:std::max(0,j-_d+1)); l < std::min(n2,j+_d); ++l)
				if (!(k == i && l <= j))
				  p += _lmu(_all,k-i,l-j+_d-1,int(ir(k,l)));
			normalize_rho(p);    
			ir(i,j) = rng.sample_multinomial( p );
		  }   
	}
}

template<class T>  
T  MarkovRandomField<T>::calc_Lp ( const Matrix3 & r )
{
  BLITZ_INDICES;
  update();
  int R = r.extent(0);
  int C = r.extent(1);
  T Lp = 0;
  
  for (MRFIter I(_d,R,C); !I.atend(); I.advance())
	{
	  Matrix1 r1( r( I.i(), I.j(), _all ) );
	  Matrix1 r2( r( I.ito(), I.jto(), _all ) );
	  Matrix2 mu2( _mudash( _all, I.irf(), I.jrf(), _all ) );
	  Lp += safe_log( sum( mu2(i1,i2) * r1(i1) * r2(i2) ) );
	}
  
  return Lp;
}

template<class T>  
blitz::Array<T,2>  MarkovRandomField<T>::calc_lp ( const Matrix3 & r )
{
  BLITZ_INDICES;
  Matrix3 lr( safe_log(r) );
  Matrix3 msg(iterate(lr,1));
  return Matrix2(safe_log(sum(r*exp(msg),i3)));
}


// FIXME: this needs to be done right:
//       - store all messages separately (in a 4d array (r,c,k,direction))
//       - calculate messages properly by using the leave-one-out belief
template<class T>  
blitz::Array<T,3>  MarkovRandomField<T>::do_mp_up ( const Matrix3 & r, bool oneway )
{
  BLITZ_INDICES;
  update();
  Matrix3 msg(r.shape());	msg = 0;
  int R = r.extent(0);
  int C = r.extent(1);
  
  Matrix4 lmu( _learning && _do_vb ? _lmudash : _lmu );
  Matrix4 mu(  _learning && _do_vb ? _mudash  : _mu );
  
  for ( MRFIter I( _d, R, C ); ! I.atend(); I.advance() )
	{
	  
	  Matrix1 r1( r( I.i(), I.j(), _all ) );
	  Matrix1 r2( r( I.ito(), I.jto(), _all ) );
	  Matrix2 mu2( mu( _all, I.irf(), I.jrf(), _all ) );
	  
	  msg(   I.i()  , I.j()  , _all ) += safe_log( sum( mu2( i1, i2 ) * r2(i2), i2 ) );
	  
	  if ( !oneway )
		msg( I.ito(), I.jto(), _all ) += safe_log( sum( mu2( i2, i1 ) * r1(i2), i2 ) );

	  
	}
  return msg;
}

template<class T>  
void MarkovRandomField<T>::update_msg()
{
  if ( _iterations > 1 || _keep_messages )
    set_msg( iterate(  get_belief( true, mmm2_belief_state<T>::LOG    ), _iterations ) );
  else
    set_msg( do_mp_up( get_belief( true, mmm2_belief_state<T>::LINEAR )              ) );
}

template<class T>  
blitz::Array<T,3>  MarkovRandomField<T>::iterate ( const Matrix3 & lr_, size_t n_iters )
{
  BLITZ_INDICES;
  update();
  int R = lr_.extent(0);
  int C = lr_.extent(1);
  int K = lr_.extent(2);
  int N = ( 2 * _d - 1 ) * ( 2 * _d - 1 ) - 1; // number of messages = 2 * number of factors
  
  Matrix4 old_messages_( R, C, N, K ), new_messages_( R, C, N, K );

  // weak refs
  Matrix4 old_messages( old_messages_, true );
  Matrix4 new_messages( new_messages_, true );
  Matrix4 mur1 (  _learning && _do_vb ? _murdash1 : _mur1, true );
  Matrix4 mur2 (  _learning && _do_vb ? _murdash2 : _mur2, true );
  Matrix3 lr( lr_, true );

  old_messages = 0; new_messages = 0;
  

  for ( size_t iteration = 0; iteration < n_iters; ++iteration )
	{
	  
#ifdef HAVE_OPENMP
#pragma omp parallel
#endif
	  for ( MRFIter I( _d, R, C ); ! I.atend(); I.advance() )
		{
		  
		  int msg_index_in  = I.msg_index_in ();
		  int msg_index_out = I.msg_index_out();    
		  
		  // combine all incoming messages except the one evaluating and used for propagating
		  // and vice versa
		  
		  Matrix1 lr1( lr( I.i(), I.j(), _all ) );
		  Matrix1 lr2( lr( I.ito(), I.jto(), _all ) );
		  Matrix2 mu21( mur1 ( I.irf(), I.jrf(), _all, _all ) );
		  Matrix2 mu22( mur2( I.irf(), I.jrf(), _all, _all ) );
		  Matrix2 old_messages_in ( old_messages( I.i()  , I.j()  , _all, _all ) );
		  Matrix2 old_messages_out( old_messages( I.ito(), I.jto(), _all, _all ) );
		  Matrix2 new_messages_in ( new_messages( I.i()  , I.j()  , _all, _all ) );
		  Matrix2 new_messages_out( new_messages( I.ito(), I.jto(), _all, _all ) );
		  
		  Matrix1 r1( lr1.copy() );
		  Matrix1 r2( lr2.copy() );
		  
		  if ( iteration > 0 )
			{
			  // calculate basis for message propagation (leave-one-out)
			  for ( int n = 0; n < N; ++n )
				{
				  if ( n != msg_index_in  )
#ifdef HAVE_SSE
					sse_add( r1.data(), &old_messages( n, 0 ), _k );
#else
					r1 += old_messages_in ( n, _all );
#endif
				  if ( n != msg_index_out )
#ifdef HAVE_SSE
					sse_add( r2.data(), &old_messages( n, 0 ), _k );
#else
					r2 += old_messages_in ( n, _all );
#endif
				}
			}
		  
		  normalize_rho( r1 );
		  normalize_rho( r2 );
		  
#ifdef HAVE_SSE
		  sse_log_mvp( &new_messages_out( msg_index_out, 0 ), _k, r1.data(), _k, mu22.data() );
		  sse_log_mvp( &new_messages_out( msg_index_in , 0 ), _k, r2.data(), _k, mu21.data() );
#else
		  new_messages_out( msg_index_out, _all ) = safe_log( sum( mu22( i1, i2 ) * r1(i2), i2 ) );
		  new_messages_in ( msg_index_in , _all ) = safe_log( sum( mu21( i1, i2 ) * r2(i2), i2 ) );
#endif
		  
		}
	  swap( old_messages, new_messages);
	}
  _messages.reference( old_messages );
  return Matrix3( sum( _messages( i1, i2, i4, i3 ), i4 ) );
}

// this is gibbs-sampling
// FIXME: this needs to be done right:
//       - store all messages separately (in a 4d array (r,c,k,direction))
//       - calculate messages properly by using the leave-one-out belief, maximized
template<class T>  
blitz::Array<T,3>  MarkovRandomField<T>::do_mp_up ( const IMatrix2 & r, bool oneway )
{
  update();
  Matrix3 msg( r.extent(0), r.extent(1), _k ); msg = 0;
  int R = r.extent(0);
  int C = r.extent(1);
  
  Matrix4 lmu (  _learning && _do_vb ? _lmudash : _lmu, true );
  
#ifdef HAVE_OPENMP
#pragma omp parallel
#endif
  for ( MRFIter I( _d, R, C ); ! I.atend(); I.advance() )
	{
	  unsigned int i1 = r( I.i(), I.j() );
	  unsigned int i2 = r( I.ito(), I.jto() );
	  Matrix2 lmu2( lmu( blitz::Range::all(), I.irf(), I.jrf(), blitz::Range::all() ) );	  
	  msg( I.i(),   I.j(),   blitz::Range::all() ) += lmu2( blitz::Range::all(), i2                  );
	  if ( !oneway )
		msg( I.ito(), I.jto(), blitz::Range::all() ) += lmu2( i1,                  blitz::Range::all() );
	}
  return msg;
}

//just a hack useful for visualization right now

template<class T>  blitz::Array<T,2>  MarkovRandomField<T>::border_potentials( Matrix3 & r )
{
  
  BLITZ_INDICES;
  update();
  Matrix2 lp(_d*r.shape());
  lp = 0;
  
  int R = r.extent(0);
  int C = r.extent(1);
  
  Matrix4 lmu (  _learning && _do_vb ? _lmudash : _lmu, true );
  
#ifdef HAVE_OPENMP
#pragma omp parallel
#endif
  for (MRFIter I(_d,R,C); !I.atend(); I.advance())
	{
	  Matrix1 r1( r( I.i()  , I.j()  , _all ) );
	  Matrix1 r2( r( I.ito(), I.jto(), _all ) );
	  Matrix2 lmu2( lmu( _all, I.irf(), I.jrf(), _all ) );
	  int ilp = 2 * I.i() + I.irf();
	  int jlp = 2 * I.j() + std::abs( ( _d - 1 ) - I.jrf() );
	  lp( ilp, jlp ) += sum( r1(i1) * lmu2 * r2(i2) );
	}
  
  return lp;
}

template<class T>  
void MarkovRandomField<T>::end_mp_frame()
{
  apply_r( message_component<T>::get_belief(!_factorize_transitions) );
}

template<class T>  
void MarkovRandomField<T>::apply_r ( Matrix3 & r, const Matrix2 * weights )
{
  BLITZ_INDICES;
  int R = r.extent(0);
  int C = r.extent(1);	
  Matrix4 mur (  _learning && _do_vb ? _murdash1 : _mur1, true );
  for (MRFIter I(_d,R,C); !I.atend(); I.advance())
	{
	  Matrix1 r1( r( I.i()  , I.j()  , _all ) );
	  Matrix1 r2( r( I.ito(), I.jto(), _all ) );
	  if ( weights )
		{
		  T w1 = (*weights)( I.i()  , I.j()   );
		  T w2 = (*weights)( I.ito(), I.jto() );
		  if ( _factorize_transitions )
			_wx( _all, I.irf(), I.jrf(), _all ) += r1(i1) * r2(i2) * w1 * w2;
		  else
			{
			  Matrix2 mu2( mur ( I.irf(), I.jrf(), _all, _all ) );
			  Matrix2 wx( mu2(i1,i2) * r1(i1) * r2(i2) );			  
			  _wx( _all, I.irf(), I.jrf(), _all ) += wx / sum(wx) * w1 * w2;
			}
		}
	  else
		{
		  if ( _factorize_transitions )
			_wx( _all, I.irf(), I.jrf(), _all ) += r1(i1) * r2(i2);
		  else
			{
			  Matrix2 mu2( mur ( I.irf(), I.jrf(), _all, _all ) );
			  Matrix2 wx( mu2(i1,i2) * r1(i1) * r2(i2) );			  
			  _wx( _all, I.irf(), I.jrf(), _all ) += wx / sum(wx);
			}
		}
	}
  
  //_Lp += calc_Lp(r);
  
}

template<class T> 
boost::tuple<T,T,T> MarkovRandomField<T>::calculate_Lmu(bool report)
{
  BLITZ_INDICES;
  
  int N = _d * (2*_d-2);
  
  T lCbeta0 = N * ( gammaln( _k * _k * _beta0 ) - _k * _k * gammaln( _beta0 ) ); //log(C(beta))
  T lCbeta = sum( where( ( i1 == 0 ) & ( i2 < _d ), 
						 0, 
						 gammaln( sum( sum( _beta( i4, i1, i2, i3 ), i4 ), i3 ) ) )
				  - sum( sum( gammaln( _beta( i4, i1, i2, i3 ) ), i4 ), i3 ) );
  
  T Elp = sum( where( ( i2 == 0 ) & ( i3 < _d ), 0, ( _beta0 - 1 ) * _lmudash ) ) + lCbeta0;
  T Elq = sum( where( ( i2 == 0 ) & ( i3 < _d ), 0, ( _beta  - 1 ) * _lmudash ) ) + lCbeta; 
  
  T Lmu = Elq - Elp;
  
  std::cout << " mrf2: ln(Z_p) = " << lCbeta0 << " ln(Z_q) = " << lCbeta << " <ln(p)>_q = " << Elp << " <ln(q)>_q = " << Elq << std::endl;

  return boost::make_tuple(Lmu,Elp,Elq);
}

template<class T> void MarkovRandomField<T>::batchlearn_end_epoch(Matrix1 f_in)
{
  BLITZ_INDICES;    
  
  using boost::tuples::ignore;
  
  _Lp = sum( where( ( i2 == 0 ) & ( i3 < _d ), 0, _wx * _lmudash ) );

  if (f_in.numElements() > 0)
	{
	  _wx = _wx * f_in(i1) * f_in(i4);
	}
  
  // Matrix2 Nk(sum(sum(_wx(i3,i1,i2,i4),i4),i3));
  // std::cout << "mrf Nk:" << Nk;
  
  _beta = (T(1)-_delta) * _beta + _delta * (_beta0 + _wx);
  
  _dirty = true;
  
  if (_learn_beta0_interval > 0 && ++_learn_beta0_counter >= _learn_beta0_interval)
	{
	  T beta0_old = _beta0;
	  update(true);
	  optimize_beta0_covariant(_delta);
	  T deltabeta0 = _beta0 - beta0_old;
	  _beta += deltabeta0;
	}	
  
  update(true);
  
  boost::tie(_Lmu,ignore,ignore) = calculate_Lmu(true);

  _L = _Lmu + _Lp;

  std::cout << str( boost::format("markov random field ended epoch Lmu = %1.3e Lp = %1.3e L = %1.3e")
					 % _Lmu % _Lp % _L ) << std::endl;

  
  _learning = _in_batch = false;
  _wx.free();
    
}


template<class T>  blitz::Array<unsigned int, 2>  MarkovRandomField<T>::maxinfer(Matrix3 & data)
{
  update();
  return _ri;
}


template<class T>  void MarkovRandomField<T>::reset()
{
  _r.free();
}  


template<class T>  void MarkovRandomField<T>::update(bool force)
{
  BLITZ_INDICES;
  if (!_initialized) initialize();
  if (force || _dirty)
	{
	  int d2 = 2 * _d - 1;
	  _mu.resize     ( _k, _d, d2, _k );
	  _mudash.resize ( _k, _d, d2, _k );
	  _lmu.resize    ( _k, _d, d2, _k );
	  _lmudash.resize( _k, _d, d2, _k );

	  _mu = _beta / sum(sum(_beta(i6,i2,i3,i5),i6),i5);
	  _lmu = log(_beta) - log(sum(sum(_beta(i6,i2,i3,i5),i6),i5));
	  _lmudash = psi(_beta) - psi(sum(sum(_beta(i6,i2,i3,i5),i6),i5));
	  _mudash = exp(_lmudash);

	  _mur1.resize    ( _d, d2, _k, _k );
	  _murdash1.resize( _d, d2, _k, _k );
	  _mur2.resize    ( _d, d2, _k, _k );
	  _murdash2.resize( _d, d2, _k, _k );
	  _mur1 =         _mu(i3,i1,i2,i4);
	  _murdash1 = _mudash(i3,i1,i2,i4);
	  _mur2 =         _mu(i4,i1,i2,i3);
	  _murdash2 = _mudash(i4,i1,i2,i3);
	  
	}
  _dirty = false;
}


template<class T>  void MarkovRandomField<T>::initialize(bool force)
{
  BLITZ_INDICES;
  if (_initialized&&!force) return;
  _beta.resize( _k, _d, 2 * _d - 1, _k ); _beta = _beta0;
  if ( _init_uniform )
	_beta += _initcount / ( _k * _k );
  else
	{
	  Random<T> & rng = Random<T>::get();
	  Matrix2 sample( _k, _k );
      for ( int ii = 0; ii < _beta.extent(0); ++ii )
	    for ( int jj = 0; jj < _beta.extent(1); ++jj )
	  	  {
		    rng.sample_dirichlet( _beta0, sample, 0.001 * _k );            
		    _beta( _all, ii, jj, _all ) += _initcount * sample;
		  }
	}
  _initialized = _dirty = true;

  std::cout << str(boost::format("markov random field initialized beta0 = %f initcounts = %f mode = %s")
                                  %_beta0%_initcount%(_init_uniform?"uniform":"random")) << std::endl;
}

template class MarkovRandomField<float>;
