/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef MRF2_HPP
#define MRF2_HPP
#undef BZ_THREADSAFE

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/foreach.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/binary_object.hpp>
#include <boost/tuple/tuple.hpp>
#include <iostream>
#include <stdexcept>

#include "SeariseProperties.hpp"
#include "message_component.hpp"

template<class T> class MarkovRandomField
: public HasProperties< SearisePropertyTypes, MarkovRandomField<T> >
, public message_component<T>
{

public:

  typedef T num_type;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,1> Matrix1;

  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<double,4> Matrix4d;

  typedef blitz::Array<unsigned int,2> IMatrix2;
  typedef MarkovRandomField<T> This;
  typedef SeariseProperty PropertyType;
  typedef HasProperties< SearisePropertyTypes, This > PBase;

  MarkovRandomField(int k =0,int d = 2);

  static void init_properties();

  const blitz::Array<T,3> get_r();
  const Matrix4 get_beta();
  const Matrix4 get_mu();

//	void set_property(const std::string & key, PropertyType value) throw(std::runtime_error);

//  PropertyType get_property(const std::string & key) throw(std::runtime_error);

  void flush();

  void batchlearn_begin_epoch();

	Matrix2  calc_lp ( const Matrix3 & r );
			
  //Gibbs-sample the MRF (inplace)
  void sample(IMatrix2 & ir, Matrix3 & ptop, int n);

  // proper belief propagation
  Matrix3 iterate ( const Matrix3 & lr, size_t n_iters );

  Matrix3 do_mp_up(const Matrix3 & r, bool oneway = false);

  Matrix3 do_mp_up(const IMatrix2 & r, bool oneway = false);

  //just a hack useful for visualization right now
  Matrix2 border_potentials(Matrix3 & r);

  void apply_r( Matrix3 & r, const Matrix2 * weights = 0);

  void end_mp_frame();

  void batchlearn_end_epoch(Matrix1 f_in = Matrix1());

  IMatrix2 maxinfer(Matrix3 & data);

  void reset();

  void update(bool force = false);

  void update_msg();

private:

  boost::tuple<T,T,T> calculate_Lmu(bool report = false);
  
  T calc_Lp (const Matrix3 & r);
  
  static double dL_beta0 (double lb0, void * params);
  
  void optimize_beta0_covariant( double delta, double lalphamin = -5, double lalphamax = 15 );
  void optimize_beta0(double delta, double lalphamin = -5, double lalphamax = 15 );

  void initialize(bool force=false);

  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & BOOST_SERIALIZATION_NVP(_k);
    ar & BOOST_SERIALIZATION_NVP(_beta0);
    ar & BOOST_SERIALIZATION_NVP(_initialized);
    if (_initialized)
    {
	    ar & BOOST_SERIALIZATION_NVP(_beta);
    }
    _dirty = true;
  }

  int _k, _d, _initcount;
  bool _initialized, _dirty, _in_batch;
  bool _init_uniform, _keep_messages;
  int _learn_beta0_interval, _learn_beta0_counter, _iterations;
  bool _learning, _factorize_transitions, _do_vb;
  
  T _beta0;
  T _delta;
  T _Lmu, _Lp, _L;

  Matrix4 _beta, _lmudash, _mudash, _mu, _lmu;
  Matrix4 _mur1, _murdash1;
  Matrix4 _mur2, _murdash2;
  Matrix4d _wx;
  Matrix3 _r;
  IMatrix2 _ri;
  Matrix4 _messages;
  
};
#endif
