/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "nmm2.hpp"
#include "digamma.hpp"
#include "normrho.hpp"
#include "optimal_hyperparameters.hpp"
#include "Random.hpp"

#include <sstream>
#include <stdexcept>


#include <boost/format.hpp>

#include "enum_properties.hpp"

#include "message_iterators.cc"

#define FETCH_PARAMETER(D,N,V) blitz::Array<T,D> N(  V, true ); N.extent(0); // to avoid unused warnings

#define FETCH_PARAMETERS									   \
  FETCH_PARAMETER( 1 , lbias   , _lbias      )				   \
  FETCH_PARAMETER( 3 , lbias3  , _lbias3     )				   \
  FETCH_PARAMETER( 4 , Tau     , _Tau        )				   \
  FETCH_PARAMETER( 4 , Taur    , _Taur       )				   \
  FETCH_PARAMETER( 4 , Taur2   , _Taur2      )				   \
  FETCH_PARAMETER( 4 , M       , _M          )				   \
  FETCH_PARAMETER( 4 , Mr      , _Mr         )				   \
  FETCH_PARAMETER( 4 , Mr2     , _Mr2        )				   \
  FETCH_PARAMETER( 4 , lbias4  , _lbias4     )				   \
  FETCH_PARAMETER( 4 , lbias4r , _lbias4r    )


template<class T> 
NormalMixtureModel2<T>::NormalMixtureModel2 ( )
  :_M0(0.0)
  ,_alphag0(2)
  ,_betag0(2)
  ,_lambdag0(0.5)
  ,_Lmu(0)
  ,_learn_beta0_interval(0)
  ,_learn_beta0_counter(0)
  ,_batchng(true)
  ,_greedy_clusterer(true)
  ,_pre_clustering_mode(NONE)
{
}

template<typename T>
double NormalMixtureModel2<T>::calc_alphag(double lambdag)
{
  return _alphag0 + ( lambdag - _lambdag0 ) / T(2);
}

template<typename T>
blitz::Array<double,3> NormalMixtureModel2<T>::calc_alphag()
{
  Matrix3d alphag(_lambdag.shape());
  for ( typename Matrix3d::iterator ialphag = alphag.begin(), ilambdag = _lambdag.begin(), iend = alphag.end(); 
        ialphag != iend; 
        ++ialphag )
    *ialphag = calc_alphag( *ilambdag );
  return   alphag;
}

template<typename T>
void NormalMixtureModel2<T>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter(&Base::_initialized,false);
  PBase::add_ivar_property_rw("betag0", &This::_betag0, uninitializer);
  PBase::add_ivar_property_rw("alphag0", &This::_alphag0, uninitializer);
  PBase::add_ivar_property_rw("lambda0", &This::_lambdag0, uninitializer);
  PBase::add_ivar_property_rw("M0", &This::_M0, uninitializer);
  PBase::add_ivar_property_rw("learn-beta0", &This::_learn_beta0_interval);

  enum_property_access<SearisePropertyTypes, This, PreClusteringMode> pre_clustering_accessor( &This::_pre_clustering_mode );
  pre_clustering_accessor.add_choice("none",NONE).add_choice("batch_ng",BATCHNG).add_choice("greedy",GREEDY);
  PBase::add_property( "pre_clustering", pre_clustering_accessor, pre_clustering_accessor );


  // these two need to update first...
  PBase::add_property(         "mu",   typename PBase::MemFunGetter( &This::get_mu  ) );
  PBase::add_property(         "taug", typename PBase::MemFunGetter( &This::get_tau ) );

  PBase::add_ivar_property_ro("Lmu", &This::_Lmu);
  
  PBase::add_property( std::string("batch_ng"), PBase::make_pivar_getter( &This::_batchng ) );
  PBase::add_property( std::string("greedy_clusterer"), PBase::make_pivar_getter( &This::_greedy_clusterer ) );

  PBase::add_property( "M_init"  , make_blitz_ivar_getter1( &This::_M_init   ), make_blitz_ivar_setter1( &This::_M_init   ) );
  PBase::add_property( "Tau_init", make_blitz_ivar_getter1( &This::_Tau_init ), make_blitz_ivar_setter1( &This::_Tau_init ) );
}

template<class T> void NormalMixtureModel2<T>::flush()
{
  Base::flush();
  _lambdag.free();
  _M.free();
  _betag.free();
  _lbias.free();
}

static double dL_lambdag0 (double ll0, void * vparams)
{
  double * params = (double*)vparams;
  double c = params[0];
  double lambda0 = std::exp(ll0);
  try {
	return 1/lambda0 + safe_log(1+lambda0) - boost::math::digamma((lambda0+1)/2) + c;
  }	catch(std::exception & e) {
	std::cerr << " exception evaluating dl beta0 (mmm2): " << e.what() << " type " << typeid(e).name() << std::endl;
	gsl_error_flag = true;
	return 0;
  }
}

template<class T> void NormalMixtureModel2<T>::optimize_hyperparameters 
( double delta, double llambdamin, double llambdamax )
{
  BLITZ_INDICES;
  update();
  typedef std::pair<T,bool> Result;
  
  Matrix3d alphag( calc_alphag() );
  
  double phi1bar = mean( alphag(i1,i2,i3) * _M(i1,i2,i3,i4) / _betag(i1,i2,i3,i4) );
  double phi2bar = mean( alphag(i1,i2,i3) / _betag(i1,i2,i3,i4) );
  double twoAbar = mean( 1 / _lambdag(i1,i2,i3)
						+ alphag(i1,i2,i3) * pow2( _M(i1,i2,i3,i4) )/ _betag(i1,i2,i3,i4)
						+ log(_betag(i1,i2,i3,i4))
						- psi(alphag(i1,i2,i3))
						);
  double c = -(twoAbar - (phi1bar * phi1bar) / phi2bar + safe_log(2*phi2bar));
  double params[1] = {c};
  gsl_function F = {&dL_lambdag0, params};	
  gsl_set_error_handler(&gsl_error_handler);	
  Result r = find_root(&F, llambdamin, llambdamax);
  bool converged = r.second;
  if (!converged) 
	{
	  std::cout << "nmm2: hyperparameter optimization did not converge" << std::endl;
	  return;
	}
  double llambda0opt = r.first;
  double lbetag0opt = safe_log((std::exp(llambda0opt)+1)/(2*phi2bar));
  double M0opt = phi1bar/phi2bar;
  std::cout << str(boost::format("nmm2: optimized hyperparameters old values: lambda0=%f M0=%f beta0=%f\n")%_lambdag0%_M0%_betag0);
  _lambdag0 = log_update(_lambdag0,llambda0opt,delta);
  _M0 = lin_update(_M0,M0opt,delta);
  _betag0 = log_update(_betag0,lbetag0opt,delta);
  std::cout << str(boost::format("                                new values: lambda0=%f M0=%f beta0=%f\n")%_lambdag0%_M0%_betag0);
  Base::dirty();
}

template<class T> void NormalMixtureModel2<T>::batchlearn_begin_epoch()
{
  
  if ( !ready() )
	{
	  if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.init( Base::_d1, Base::_d2, Base::_m );
		}
	  else if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.initialize( Base::_k, Base::_d1, Base::_d2, Base::_m );
		  _batchng.begin_epoch();
		}
	  Base::_learning = true;
	}
  else
	{
	  Base::batchlearn_begin_epoch();
	}
}

  ///////////////////////////////////////////
 ////// message calculators - begin ////////
///////////////////////////////////////////


template<class T> void NormalMixtureModel2<T>::upward_message ( const Matrix3 & data, Matrix1 & lrout, bool marginal)
{
  FETCH_PARAMETERS;
#ifdef HAVE_SSE
  for (int k = 0; k < Base::_k; ++k)
	{
	  lrout(k) = 0;
	  for (int i = 0; i < Base::_d1; ++i)
  		lrout(k) += sse_lgauss_nonorm( &data( i, 0, 0 ), &_M( k, i, 0, 0 ), &_Tau( k, i, 0, 0 ), Base::_m * Base::_d2 );
    assert( lrout(k) <= 0 );
	}
  lrout += lbias;
#else
  BLITZ_INDICES;
  lrout = lbias + sum(sum(sum( T(-0.5) * Tau * pow2( data(i2,i3,i4) - M ) ,i4),i3),i2);
#endif
}

// fixme: saliency should fade over to the prior data likelihood

template<class T> void NormalMixtureModel2<T>::upward_message 
( const Matrix3 & data, Matrix1 & lrout, const Matrix2 & w, bool marginal )
{
  FETCH_PARAMETERS;
  lrout = 0;
#ifdef HAVE_SSE
  int K4 = Base::_k / 4;
  for (int i = 0; i < Base::_d1; ++i)
	for (int j = 0; j < Base::_d2; ++j)
	  {
		int k = 0;
		__m128 s = _mm_set1_ps( w(i,j) );
		__m128 o = _mm_set1_ps( ( 1.0f - w(i,j) ) / Base::_k );
		for (int k4 = 0; k4 < K4; ++k4, k += 4 )
		  {
			__m128 arg4;
			__m128 lbias4 = _mm_loadu_ps( &lbias3( i, j, k ) );
			float * arg4p = (float*)&arg4;
			arg4p[0] = sse_lgauss_nonorm( &data( i, j, 0 ), &_M( k + 0, i, j, 0 )
										  , &_Tau( k + 0, i, j, 0 ), Base::_m );
			arg4p[1] = sse_lgauss_nonorm( &data( i, j, 0 ), &_M( k + 1, i, j, 0 ), &_Tau( k + 1, i, j, 0 ), Base::_m );
			arg4p[2] = sse_lgauss_nonorm( &data( i, j, 0 ), &_M( k + 2, i, j, 0 ), &_Tau( k + 2, i, j, 0 ), Base::_m );
			arg4p[3] = sse_lgauss_nonorm( &data( i, j, 0 ), &_M( k + 3, i, j, 0 ), &_Tau( k + 3, i, j, 0 ), Base::_m );
			__m128 lp = logf4( _mm_add_ps( _mm_mul_ps( s, expf4( _mm_add_ps( lbias4, arg4 ) ) ), o ) );
			_mm_storeu_ps( lrout.data() + k, _mm_add_ps( _mm_loadu_ps( lrout.data() + k ), lp ) );
		  }
		for ( ; k < Base::_k; ++k )
		  {
			float p = std::exp( lbias3(i,j,k) + sse_lgauss_nonorm( &data( i, j, 0 ), &_M( k, i, j, 0 ), &_Tau( k, i, j, 0 ), Base::_m ) );
			lrout(k) += safe_log( w(i,j) * p + ( 1.0f - w(i,j) ) / Base::_k );
		  }
	  }
#else
  BLITZ_INDICES;
  for (int i = 0; i < data.extent(0); ++i)
	  for (int j = 0; j < data.extent(1); ++j)
	    for (int k = 0; k < Base::_k; ++k)
		    lrout(k) += safe_log( w(i,j) * exp( lbias3(i,j,k) + T(-0.5) * sum( Tau(k,i,j,_all) * pow2( data(i,j,_all) - M(k,i,j,_all) ) ) )
								  + ( T(1) - w(i,j) ) / Base::_k );
#endif
}

template<typename T>
void NormalMixtureModel2<T>::finalize_downward_message( Matrix3 & msg, bool linear )
{
  BLITZ_INDICES;
  if ( !linear ) throw std::runtime_error("nmm2 can only generate linear downward message");
  Matrix3 mu( msg( _all, _all, blitz::Range( 0, Base::_m - 1 ) ) );
  Matrix3 tau( msg( _all, _all, blitz::Range( Base::_m, 2 * Base::_m - 1 ) ) );
  msg.free();
  msg.resize( mu.shape() );
  msg = mu / tau;
}

template<typename T>
void NormalMixtureModel2<T>::initialize_downward_message( Matrix3 & msg, bool linear )
{
  if ( !linear ) throw std::runtime_error("nmm2 can only generate linear downward message");
  msg.resize( msg.extent(0), msg.extent(1), 2 * Base::_m );
  msg = 0;
}

template<typename T>
void NormalMixtureModel2<T>::downward_message( const blitz::Array<T,3> & x, const blitz::Array<T,1> & z, Matrix3 & msg )
{
  BLITZ_INDICES;
  Matrix3 mu( msg( _all, _all, blitz::Range( 0, Base::_m - 1 ) ) );
  Matrix3 tau( msg( _all, _all, blitz::Range( Base::_m, 2 * Base::_m - 1 ) ) );
  mu += sum( _Taur * _Mr * z(i4), i4 );
  tau += sum( _Taur * z(i4), i4 );
}

template<typename T>
void NormalMixtureModel2<T>::calculate_emission_hessian( const Matrix3 & x
								 , const Matrix1 & output_hessian
								 , typename Base::emission_gradient_t & hessian )
{
  Matrix4 & M_d = Base::_discriminative_emission_parameters[0];
  Matrix4 & M_g = Base::_generative_emission_parameters[0];
  Matrix4 & Tau_d = Base::_discriminative_emission_parameters[1];
  Matrix4 & Tau_g = Base::_generative_emission_parameters[1];
  T coupling = Base::_discriminative_generative_coupling;
  BLITZ_INDICES;
#ifdef HAVE_SSE
  for ( int k = 0; k < Base::_k; ++k )
	for ( int i = 0; i < Base::_d1; ++i )
	  {
		if ( coupling <= 0 )
		  {
			float h = output_hessian( k );
			vecf4 vh( h );
			float * px = (float*)&x( i, 0, 0 );
			float * pTau_d = &Tau_d( k, i, 0, 0 );
			float * pM_d = &M_d( k, i, 0, 0 );
			float * ph0 = &hessian[0]( k, i, 0, 0 );
			float * ph1 = &hessian[1]( k, i, 0, 0 );
			int njm = Base::_m * Base::_d2;
			int jm = 0;
			for ( int jm4 = 0, jm4max = njm / 4; jm4 < jm4max; ++jm4, jm += 4 )
			  {
				vecf4 vx( px + jm ), vTau_d( pTau_d + jm ), vM_d( pM_d + jm ), vh0( ph0 + jm ), vh1( ph1 + jm );
				( vh0 + vh * vTau_d * pow2( ( vx - vM_d )                ) ).store( ph0 + jm );
				( vh1 + vh * pow2( 0.5f * ( 1.0f - ( pow2( vx ) - pow2( vM_d ) ) )  )   ).store( ph1 + jm );
			  }
			for ( ; jm < njm; ++jm )
			  {
				ph0[jm] += h * pTau_d[jm] * pow2( ( px[jm] - pM_d[jm] ) );
				ph1[jm] += h * pow2( 0.5f * ( 1.0f - ( pow2( px[jm] ) - pow2( pM_d[jm] ) ) ) );
			  }
		  }
		else
		  {
			float * pTau_g = &Tau_g( k, i, 0, 0 );
			float * pM_g = &M_g( k, i, 0, 0 );
			vecf4 vc( coupling );
			float h = output_hessian( k );
			vecf4 vh( h );
			float * px = (float*)&x( i, 0, 0 );
			float * pTau_d = &Tau_d( k, i, 0, 0 );
			float * pM_d = &M_d( k, i, 0, 0 );
			float * ph0 = &hessian[0]( k, i, 0, 0 );
			float * ph1 = &hessian[1]( k, i, 0, 0 );
			int njm = Base::_m * Base::_d2;
			int jm = 0;
			for ( int jm4 = 0, jm4max = njm / 4; jm4 < jm4max; ++jm4, jm += 4 )
			  {
				vecf4 vx( px + jm ), vTau_d( pTau_d + jm ), vM_d( pM_d + jm ), vTau_g( pTau_g + jm ), vM_g( pM_g + jm ), vh0( ph0 + jm ), vh1( ph1 + jm );
				( vh0 + vh * vTau_d * pow2( ( vx - vM_d )    ) + pow2( vc * ( sqrt( vTau_d ) * vM_d  - sqrt( vTau_g ) * vM_g     ) ) ).store( ph0 + jm );
				( vh1 + vh * pow2( 0.5f * ( 1.0f - pow2( vx ) - pow2( vM_d ) ) ) + pow2( vc * ( safe_log( vTau_d ) - safe_log( vTau_g ) ) ) ).store( ph1 + jm );
			  }
			for ( ; jm < njm; ++jm )
			  {
				ph0[jm] += h * pTau_d[jm] * pow2( ( px[jm] - pM_d[jm] ) ) + pow2( coupling * ( sqrt( pTau_d[jm] ) * pM_d[jm] - sqrt( pTau_g[jm] ) * pM_g[jm] ) );
				ph1[jm] += h * pow2( 0.5f * ( 1.0f - pow2( px[jm] ) - pow2( pM_d[jm] ) ) ) + pow2( coupling * ( safe_log( pTau_d[jm] ) - safe_log( pTau_g[jm] ) ) );
			  }
		  }
	}
#else
  if ( coupling > 0 )
	{
	  hessian[0] += output_hessian( i1 ) * Tau_d * pow2( ( x( i2, i3, i4 ) - M_d ) ) + pow2( coupling * ( sqrt( Tau_d ) * M_d - sqrt( Tau_g ) * M_g ) );
	  hessian[1] += output_hessian( i1 ) * pow2( T(0.5) * ( T(1) - pow2( x( i2, i3, i4 ) ) - pow2( M_d ) ) ) + pow2( coupling * ( safe_log( Tau_d ) - safe_log( Tau_g ) ) );
	}
  else
	{
	  hessian[0] += output_hessian( i1 ) * Tau_d * pow2( Tau_d * ( x( i2, i3, i4 ) - M_d ) );
	  hessian[1] += output_hessian( i1 ) * pow2( T(0.5) * ( T(1) - pow2( x( i2, i3, i4 ) ) - pow2( M_d ) ) );
	}
#endif
}
								

template<typename T>
void NormalMixtureModel2<T>::calculate_emission_gradient( const Matrix3 & x
														  , const Matrix2 & input_gradient
														  , const Matrix1 & output_gradient
														  , typename Base::emission_gradient_t & gradient )
{
  Matrix4 & M_d = Base::_discriminative_emission_parameters[0];
  Matrix4 & M_g = Base::_generative_emission_parameters[0];
  Matrix4 & Tau_d = Base::_discriminative_emission_parameters[1];
  Matrix4 & Tau_g = Base::_generative_emission_parameters[1];
  T coupling = Base::_discriminative_generative_coupling;
  BLITZ_INDICES;
  if ( coupling > 0 )
	{
	  gradient[0] += input_gradient( i2, i3 ) * output_gradient( i1 ) * sqrt( Tau_d ) * ( x( i2, i3, i4 ) - M_d ) + coupling * ( sqrt( Tau_d ) * M_d - sqrt( Tau_g ) * M_g );
	  gradient[1] += output_gradient( i1 ) * T(0.5) * ( T(1) - Tau_d * ( pow2( x( i2, i3, i4 ) ) - pow2( M_d ) ) ) + coupling * ( safe_log( Tau_d ) - safe_log( Tau_g ) );
	}
  else
	{
	  gradient[0] += input_gradient( i2, i3 ) * output_gradient( i1 ) * sqrt( Tau_d ) * ( x( i2, i3, i4 ) - M_d );
	  gradient[1] += input_gradient( i2, i3 ) * output_gradient( i1 ) * T(0.5) * ( T(1) - Tau_d * ( pow2( x( i2, i3, i4 ) ) - pow2( M_d ) ) );
	}
}

template<typename T>
void NormalMixtureModel2<T>::calculate_emission_gradient( const Matrix3 & x
														  , const Matrix1 & output_gradient
														  , typename Base::emission_gradient_t & gradient )
{
  using std::sqrt;
  Matrix4 & M_d = Base::_discriminative_emission_parameters[0];
  Matrix4 & M_g = Base::_generative_emission_parameters[0];
  Matrix4 & Tau_d = Base::_discriminative_emission_parameters[1];
  Matrix4 & Tau_g = Base::_generative_emission_parameters[1];
  T coupling = Base::_discriminative_generative_coupling;
  BLITZ_INDICES;
#ifdef HAVE_SSE
  for ( int k = 0; k < Base::_k; ++k )
	for ( int i = 0; i < Base::_d1; ++i )
	  {
		float g = output_gradient( k );
		vecf4 vg( g );
		float * px = (float*)&x( i, 0, 0 );
		float * pTau_d = &Tau_d( k, i, 0, 0 );
		float * pM_d = &M_d( k, i, 0, 0 );
		float * pg0 = &gradient[0]( k, i, 0, 0 );
		float * pg1 = &gradient[1]( k, i, 0, 0 );
		int njm = Base::_m * Base::_d2;
		if ( coupling > 0 )
		  {
        float * pTau_g = &Tau_g( k, i, 0, 0 );
        float * pM_g = &M_g( k, i, 0, 0 );
        vecf4 vc( coupling );
        int jm = 0;
        for ( int jm4 = 0, jm4max = njm / 4; jm4 < jm4max; ++jm4, jm += 4 )
          {
            vecf4 vx( px + jm ), vTau_d( pTau_d + jm ), vM_d( pM_d + jm ), vTau_g( pTau_g + jm ), vM_g( pM_g + jm ), vg0( pg0 + jm ), vg1( pg1 + jm );
            ( vg0 + vg * sqrt( vTau_d ) *( vx - vM_d )                            + vc * (  sqrt( vTau_d ) * vM_d     -  sqrt( vTau_g ) * vM_g     ) ).store( pg0 + jm );
            ( vg1 + vg * 0.5f * ( 1.0f - vTau_d * ( pow2( vx ) - pow2( vM_d ) ) ) + vc * ( safe_log( vTau_d ) - safe_log( vTau_g ) ) ).store( pg1 + jm );
          }
        for ( ; jm < njm; ++jm )
          {
            pg0[jm] += g * sqrt( pTau_g[jm] ) * ( px[jm] - pM_d[jm] ) + coupling * ( sqrt( pTau_d[jm] ) * pM_d[jm] - sqrt( pTau_g[jm] ) * pM_g[jm] );
            pg1[jm] += g * 0.5f * ( 1.0f - pTau_d[jm] * ( pow2( px[jm] ) - pow2( pM_d[jm] ) ) ) + coupling * ( safe_log( pTau_d[jm] ) - safe_log( pTau_g[jm] ) );
          }
		  }
		else
		  {
        int jm = 0;
        for ( int jm4 = 0, jm4max = njm / 4; jm4 < jm4max; ++jm4, jm += 4 )
          {
            vecf4 vx( px + jm ), vTau_d( pTau_d + jm ), vM_d( pM_d + jm ), vg0( pg0 + jm ), vg1( pg1 + jm );
            ( vg0 + vg * sqrt( vTau_d ) * ( vx - vM_d ) ).store( pg0 + jm );
            ( vg1 + vg * vTau_d * ( 1 - ( pow2( vx ) - pow2( vM_d ) ) ) ).store( pg1 + jm );
          }
        for ( ; jm < njm; ++jm )
          {
            pg0[jm] += g * sqrt( pTau_d[jm] ) * ( px[jm] - pM_d[jm] );
            pg1[jm] += g * pTau_d[jm] * ( 1.0f - ( pow2( px[jm] ) - pow2( pM_d[jm] ) ) );
          }
		  }
	}
#else
  if ( coupling > 0 )
	{
	  gradient[0] += output_gradient( i1 ) * sqrt( Tau_d ) * ( x( i2, i3, i4 ) - M_d ) + coupling * ( sqrt( Tau_d ) *  M_d - sqrt( Tau_g ) * M_g );
	  gradient[1] += output_gradient( i1 ) * T(0.5) * ( T(1) - Tau_d * ( pow2( x( i2, i3, i4 ) ) - pow2( M_d ) ) ) + coupling * ( safe_log( Tau_d ) - safe_log( Tau_g ) );
	}
  else
	{
	  gradient[0] += output_gradient( i1 ) * sqrt( Tau_d ) * ( x( i2, i3, i4 ) - M_d );
	  gradient[1] += output_gradient( i1 ) * T(0.5) * ( T(1) - Tau_d * ( pow2( x( i2, i3, i4 ) ) - pow2( M_d ) ) );
	}
#endif
}

template<typename T>
void NormalMixtureModel2<T>::apply_emission_gradient( typename Base::emission_gradient_t & parameter_gradient, T rate )
{
  Matrix4 & d_M_scaled = parameter_gradient[0];
  Matrix4 & d_lTau = parameter_gradient[1];
  Matrix4 & M_d = Base::_discriminative_emission_parameters[0];
  Matrix4 & Tau_d = Base::_discriminative_emission_parameters[1];

  M_d   = M_d - rate * d_M_scaled / sqrt( Tau_d );
  // this tends to explode, better limit it here
  Tau_d = exp( max( T(-10), min( T(10), safe_log( Tau_d ) - rate * d_lTau ) ) );
}

template<class T> void NormalMixtureModel2<T>::local_likelihood
( blitz::Array<T,3> & X, blitz::Array<T,1> & r, T & l, bool marginal )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  if ( marginal )
	// this is the expected likelihood of the gaussian
	l = sum( r(i1) *
			 ( 
			  _lbias_ml4(i1,i2,i3,i4) -
			  ( _alphag0 + ( _lambdag(i1,i2,i3) - _lambdag0 + 1 ) * T(0.5) ) *
			  safe_log( _betag + T(0.5) * _lambdag(i1,i2,i3) / ( _lambdag(i1,i2,i3) + 1 ) * pow2( X(i2,i3,i4) - _M ) )
			   )
			 );
  else
	// this is the likelihood of the expected gaussian
	l = sum( r(i1) *
			 ( 
			  _lbias4(i1,i2,i3,i4) -
			  T(0.5) * Tau(i1,i2,i3,i4) 
			  * pow2( X(i2,i3,i4) - M(i1,i2,i3,i4) ) 
			   )
			 );
}

template<class T> void NormalMixtureModel2<T>::local_likelihood
( blitz::Array<T,3> & X, blitz::Array<T,1> & r, T & l, blitz::Array<T,2> & w, bool marginal )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  T W = sum( w );
  if ( W <= 0 ) l = T(-1e10);
  else
	{
	  if ( marginal )
		// this is the expected likelihood of the gaussian
		l = sum( w(i2,i3) * r(i1) *
				 ( 
				  _lbias_ml4(i1,i2,i3,i4) -
				  ( _alphag0 + ( _lambdag(i1,i2,i3) - _lambdag0 + 1 ) * T(0.5) ) *
				  safe_log( _betag + T(0.5) * _lambdag(i1,i2,i3) / ( _lambdag(i1,i2,i3) + 1 ) * pow2( X(i2,i3,i4) - _M ) )
				   )
				 ) / W;
	  else
		// this is the likelihood of the expected gaussian
		l = sum( w(i2,i3) * r(i1) *
				 ( 
				  _lbias4(i1,i2,i3,i4) -
				  T(0.5) * Tau(i1,i2,i3,i4) 
				  * pow2( X(i2,i3,i4) - M(i1,i2,i3,i4) ) 
				   )
				 ) / W;
	}
}

template<class T> void NormalMixtureModel2<T>::self_information
( const Matrix1 & rout, Matrix3 & X, Matrix2 & pin )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  for (int i = 0; i < X.extent(0); ++i)
	  for (int j = 0; j < X.extent(1); ++j)
#if 0 //def HAVE_SSE
	    pin(i,j) += sse_gauss_conf( &X(i,j,0), Base::_m, rout.data(), Base::_k,
									&Mr(i,j,0,0), &Taur(i,j,0,0), &lbias4r(i,j,0,0) );
#else
  pin(i,j) -= sum( rout(i2) *
				   ( 
					_lbias4r(i,j,_all,_all)(i1,i2) +
					T(-0.5) * Taur(i,j,_all,_all)(i1,i2) 
					 * pow2( X(i,j,_all)(i1) - Mr(i,j,_all,_all)(i1,i2) ) 
					 )
				   );
#endif
}

template<class T> void NormalMixtureModel2<T>::self_information
( const Matrix1 & rout, Matrix3 & X, Matrix2 & pin, const Matrix2 & w )
{
  FETCH_PARAMETERS;
  BLITZ_INDICES;
  for (int i = 0; i < X.extent(0); ++i)
  	for (int j = 0; j < X.extent(1); ++j)
#if 0 //def HAVE_SSE
	    pin(i,j) += sse_gauss_conf_madd( &X(i,j,0), Base::_m, rout.data(), Base::_k,
										 &Mr(i,j,0,0), &Taur(i,j,0,0), &lbias4r(i,j,0,0),
										 T(1) - w(i,j), w(i,j) );
#else
  pin(i,j) += w(i,j) * sum( rout(i2) *
							( 
							 _lbias4r(i,j,_all,_all)(i1,i2) +
							 T(-0.5) * Taur(i,j,_all,_all)(i1,i2) 
							  * pow2( X(i,j,_all)(i1) - Mr(i,j,_all,_all)(i1,i2) ) 
							  )
							);
#endif
}

template<typename T>
void NormalMixtureModel2<T>::collect_sampled_statistics( Matrix1 & x, int i1rf, int i2rf, int k, T s0 )
{
  BLITZ_INDICES;
  Base::_local_emission_statistics[OPENMP_THREAD_NUM][0]( k, i1rf, i2rf, _all ) += s0 * x;
  Base::_local_emission_statistics[OPENMP_THREAD_NUM][1]( k, i1rf, i2rf, _all ) += s0 * x * x;
}

template<typename T>
void NormalMixtureModel2<T>::collect_sampled_background_statistics( Matrix1 & x, T s0 )
{
  BLITZ_INDICES;
  Base::_local_background_statistics[OPENMP_THREAD_NUM][0] += s0 * x;
  Base::_local_background_statistics[OPENMP_THREAD_NUM][1] += s0 * x * x;
}


template<class T> void NormalMixtureModel2<T>::collect_statistics 
( const Matrix3 & X, const Matrix1 & r )
{
  BLITZ_INDICES;
  
  if ( !ready() )
	{
	  if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.process( X );
		}
	  else if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.consume( X );
		}
	}
  else
	{	  
	  Base::_local_emission_statistics[OPENMP_THREAD_NUM][0] += r(i1) *       X(i2,i3,i4);
	  Base::_local_emission_statistics[OPENMP_THREAD_NUM][1] += r(i1) * pow2( X(i2,i3,i4) );
	}

}

template<class T> void NormalMixtureModel2<T>::collect_statistics 
( const Matrix3 & X, const Matrix1 & r, const Matrix2 & w2 )
{
  BLITZ_INDICES;

  if ( !ready() )
	{
	  OPENMP(single)
	  {
		if ( _pre_clustering_mode == BATCHNG )
		  _batchng.process( X, w2 );
		else if ( _pre_clustering_mode == GREEDY )
		  _greedy_clusterer.consume( X, w2 );
	  }
	}
  else
	{	  
  	  
	  Base::_local_emission_statistics[OPENMP_THREAD_NUM][0] += r(i1) *       X(i2,i3,i4)   * w2(i2,i3);
	  Base::_local_emission_statistics[OPENMP_THREAD_NUM][1] += r(i1) * pow2( X(i2,i3,i4) ) * w2(i2,i3);

	}
}

template<class T> void NormalMixtureModel2<T>::collect_statistics_gated
( Matrix1 & x, Matrix3 & z, Matrix1 & w, T s0, InverseRFIterator & iter )
{
  BLITZ_INDICES;
  if ( !ready() )
	throw std::runtime_error("gated statistics can not be collected with pre-clustering enabled (yet?)");
  // the old code was gruesome, this probably is a sort of factorization, but hey...
  // I think what is missing is a full normalized jk tensor
  Matrix4d wxw ( Base::_local_emission_statistics[OPENMP_THREAD_NUM][0], true );
  Matrix4d wx2w( Base::_local_emission_statistics[OPENMP_THREAD_NUM][1], true );
  Matrix3d wnxw( Base::                _wnx_local[OPENMP_THREAD_NUM]   , true );
  for ( iter.reset_j(); !iter.atend_j(); iter.increment_j() )
	{
	  Matrix2d  wxw_ij( wxw ( _all, iter.i1_rf(), iter.i2_rf(), _all ) );
	  Matrix2d wx2w_ij( wx2w( _all, iter.i1_rf(), iter.i2_rf(), _all ) );
	  Matrix1d wnxw_ij( wnxw( _all, iter.i1_rf(), iter.i2_rf()       ) );
	  Matrix1     z_j ( z   (       iter.j1     , iter.j2     , _all ) );
	  T w_ij = w( iter.j_index() );
	  wxw_ij  += z_j(i1) * w_ij *       x(i2)   * s0;
	  wx2w_ij += z_j(i1) * w_ij * pow2( x(i2) ) * s0;
	  wnxw_ij += z_j(i1) * w_ij                 * s0;
	}  
}


double kl_normal_gamma( double alphaP, double betaP, double lambdaP, double muP, double alphaQ, double betaQ, double lambdaQ, double muQ )
{
  double dmu = muP - muQ;
  double d = 
	// gamma part
	gammaln(alphaQ) - gammaln(alphaP) + alphaP * std::log(betaP) - alphaQ * std::log(betaQ)
	+ ( alphaP - alphaQ ) * ( psi(alphaP) - std::log(betaP) )
	+ ( betaQ - betaP ) / betaP
	// normal part
	+ double(0.5) * ( std::log(lambdaQ) - std::log(lambdaP) + ( lambdaP + dmu * dmu ) / lambdaQ - double(1) );
#ifndef _WIN32
  if ( std::isnan(d) || std::isinf(d) )
	std::cout << str( boost::format("nan in kl-normal-gamma alpha(%e->%e) beta(%e->%e) lambda(%e->%e) mu(%e->%e)\n")
					  % alphaP % betaP % lambdaP % muP % alphaQ % betaQ % lambdaQ % muQ );
#endif
  return d;
}

template<class T> void NormalMixtureModel2<T>::recalc_likelihoods ( )
{
  BLITZ_INDICES;
  
  Matrix4d xmean( Base::_k,  Base::_d1,  Base::_d2,  Base::_m );
  Matrix4d ws   ( Base::_k,  Base::_d1,  Base::_d2,  Base::_m );

  Matrix4d & wx = Base::_emission_statistics[0];
  Matrix4d & wx2 = Base::_emission_statistics[1];

  for ( int m = 0; m < Base::_m; ++m )
	{
	  xmean( _all, _all, _all, m ) = wx( _all, _all, _all, m ) / max( 1e-10, Base::_wnx );
	  ws( _all, _all, _all, m ) = max( 0, wx2( _all, _all, _all, m ) - Base::_wnx * pow2( xmean( _all, _all, _all, m ) ) );
	}
   
  Matrix1d & wk = Base::_wk;
  Matrix3d & wn = Base::_wnx;
  Matrix1  & lb = _lbias;
  Matrix3d & lg = _lambdag;
  float m = Base::_m;

  #define report(x)
  //#define report(x) std::cout << BOOST_PP_STRINGIZE(x) << " shape " << x.shape() << " min " << min(x) << " mean " << mean(x) << " max " << max(x) << std::endl;

  report(lb);
  report(wk);
  report(_Tau);
  report(ws);
  report(wn);
  report(wx);
  report(wx2);
  report(xmean);
  report(_M);
  report(lg);

#undef report

  Base::_Lp = sum( wk * lb ) - T(0.5) * ( sum( _Tau * ( ws + wn(i1,i2,i3) * pow2( xmean - _M )  ) ) 
										  + sum( m * wn / lg ) );

  _Lmu = 0;
  for (int k = 0; k < Base::_k; ++k)
	for (int i = 0; i < Base::_d1; ++i)
	  for (int j = 0; j < Base::_d2; ++j)
		for (int m = 0; m < Base::_m; ++m)
		  {
			T lambda = _lambdag(k,i,j);
			double alpha = calc_alphag(lambda);
			T beta = _betag(k,i,j,m);
			T M = _M(k,i,j,m);
			_Lmu -= kl_normal_gamma(alpha,beta,lambda,M,_alphag0,_betag0,_lambdag0,_M0);
		  }
  
  Base::_L = Base::_Lz + _Lmu + Base::_Lp;
}

template<class T> 
blitz::Array<typename NormalMixtureModel2<T>::num_type ,1> NormalMixtureModel2<T>::batchlearn_end_epoch ( Matrix1 f_in )
{
  BLITZ_INDICES;
  
  Matrix1 f;

  if ( !ready() )
	{


	  if ( _pre_clustering_mode == GREEDY )
		{
		  _greedy_clusterer.finish();
		  Base::_wk.reference(  _greedy_clusterer.counts ().copy() );
		  Base::_wnx.reference( _greedy_clusterer.counts3().copy() );
		  Base::_emission_statistics[0].reference( _greedy_clusterer.histograms().copy() );
		  Base::_emission_statistics[1].reference( _greedy_clusterer.histograms2().copy() );
		  Base::_k = _greedy_clusterer.size();
		  Base::_dirty = Base::_initialized = true;
		  Base::recalc_hyperparams();
		  recalc_hyperparams();
		  update(true);
		  recalc_likelihoods();
		}
	  else if ( _pre_clustering_mode == BATCHNG )
		{
		  _batchng.end_epoch();
		  
		  if (_batchng.ready())
			{
			  std::cout << " ====== BatchNG done! ====== " << std::endl;
			  Base::_wk.reference( _batchng.get_histogram().copy() );			  
			  Base::_wnx.reference( _batchng.get_histogram3().copy() );
			  Base::_emission_statistics[0].reference( _batchng.get_centerStat().copy() );
			  Base::_emission_statistics[1].reference( _batchng.get_squareStat().copy() );
			  Base::_dirty = Base::_initialized = true;
			  Base::recalc_hyperparams();
			  recalc_hyperparams();
			  update(true);
			  recalc_likelihoods();
			}
		  else
			{
			  _M.reference( _batchng.get_centers() );
			  _Tau.reference( _batchng.get_variances() );
			}
		}

	  Base::_learning = false;

	  return Matrix1();

	}
  else
	{

	  f.reference( Base::batchlearn_end_epoch() );
	  
	  if (_learn_beta0_interval > 0 && ++_learn_beta0_counter == _learn_beta0_interval)
		{
		  optimize_hyperparameters(Base::_delta);
		}
	  
	  recalc_hyperparams();
	  
	  double ELp = Base::_Lp / ( sum( Base::_wnx ) * Base::_m );

	  std::cout << str(boost::format("normal component model ended epoch Lmu = %1.3e Lp = %1.3e <Lp> = %1.3e L = %1.3e")
						% _Lmu % Base::_Lp % ELp % Base::_L ) << std::endl;
	  
	}

  return f;
}

template<class T> void NormalMixtureModel2<T>::reset_input_dimensions ( const std::set<unsigned int> & dimensions)
{
  throw std::runtime_error( "NormalMixtureModel2<T>::reset_input_dimensions not implemented");
}

template<class T> void NormalMixtureModel2<T>::reset_components 
( const std::set<unsigned int> & components )
{
  throw std::runtime_error( "NormalMixtureModel2<T>::reset_components not implemented");
}

template<class T> void NormalMixtureModel2<T>::sample_point( int irf, int jrf, int k, Matrix1 & data )
{
  BLITZ_INDICES;
  Matrix4 Mur2w ( _Mr2, true );
  Matrix4 Taur2w( _Taur2, true );
  Matrix1 mu (  Mur2w( irf, jrf, k, _all ) );
  Matrix1 tau( Taur2w( irf, jrf, k, _all ) );
  Random<T> & rng = Random<T>::get();
  for ( int m = 0; m < Base::_m; ++m )
	{
	  data(m) = rng.normal( mu(k), T(1) / std::sqrt( tau(k) ) );
	}
}

template<class T> void NormalMixtureModel2<T>::sample_one ( unsigned int out, Matrix3 & X )
{
  FETCH_PARAMETERS;

  Random<T> & rng = Random<T>::get();

  for (typename Matrix3::iterator iX = X.begin(), end = X.end(); iX != end; ++iX)
	{
	  blitz::TinyVector<int,3> pos = iX.position();
	  *iX = M(pos) + rng.normal( M(pos), T(1) / std::sqrt(Tau(pos)) );
	}
}

template<class T> void NormalMixtureModel2<T>::initialize()
{

  if ( !ready() ) return;

  if ( !Base::initialize() ) 
	return;

  OPENMP(single)
  {
	
	Random<T> & rng = Random<T>::get();
	
	Matrix4d & wx = Base::_emission_statistics[0];
	Matrix4d & wx2 = Base::_emission_statistics[1];
	Matrix4d & wx_prior = Base::_prior_emission_statistics[0];
	Matrix4d & wx2_prior = Base::_prior_emission_statistics[1];

	wx_prior.resize ( Base::_k, Base::_d1, Base::_d2, Base::_m );
	wx2_prior.resize( Base::_k, Base::_d1, Base::_d2, Base::_m );
  
	T tau0 = _alphag0 / _betag0;
	
	bool have_external = _M_init.extent(0) == Base::_k && _M_init.extent(1) == Base::_d1 && _M_init.extent(2) == Base::_d2 && _M_init.extent(3) == Base::_m
	  && _Tau_init.extent(0) == Base::_k && _Tau_init.extent(1) == Base::_d1 && _Tau_init.extent(2) == Base::_d2 && _Tau_init.extent(3) == Base::_m;

	if ( have_external )
	  std::cout << " gauss model init from external " << std::endl;

	for (int k  = 0; k  < Base::_k; ++k)
	  for (int d1 = 0; d1 < Base::_d1; ++d1)
		for (int d2 = 0; d2 < Base::_d2; ++d2)
		  for (int m  = 0; m  < Base::_m; ++m)
			{
			  // first we draw a gaussian, then we draw samples from this gaussian
			  T tau, M;
			  if ( have_external )
				{
				  tau = _Tau_init( k, d1, d2, m );
				  M = _M_init( k, d1, d2, m );
				}
			  else
				{
				  tau = tau0;
				  M = Base::_init_uniform ? _M0 : rng.normal( _M0, T(1) / sqrt( _lambdag0 * tau ) );
				}
			  wx_prior( k, d1, d2, m ) = Base::_wnx_prior( k, d1, d2 ) * M;
			  wx2_prior( k, d1, d2, m ) = Base::_wnx_prior( k, d1, d2 ) * ( T(1) / tau + M * M );
			}
	
	wx.reference( wx_prior.copy() );
	wx2.reference( wx2_prior.copy() );

	recalc_hyperparams();	  
	
	//std::cout << str(boost::format("normal component model initialized alphag0 = %f betag0 = %f M0 = %f")%_alphag0%_betag0%_M0) << std::endl;

  }
  OPENMP(flush);

}


template<class T> void NormalMixtureModel2<T>::recalc_hyperparams ()
{
  BLITZ_INDICES;
  
  Matrix4d & wx = Base::_emission_statistics[0];
  Matrix4d & wx2 = Base::_emission_statistics[1];
  Matrix4d & wx_prior = Base::_prior_emission_statistics[0];
  Matrix4d & wx2_prior = Base::_prior_emission_statistics[1];
  Matrix4  & M = _Mg;
	
	M.free();
	M.resize(Base::_k,Base::_d1,Base::_d2,Base::_m);

	_lambdag.resize( Base::_k, Base::_d1, Base::_d2           );
	_betag.resize  ( Base::_k, Base::_d1, Base::_d2, Base::_m );
	
	if ( wx_prior.numElements() == 0 )
	  {
		wx_prior.resize( wx.shape() );
		wx2_prior.resize( wx2.shape() );
		wx_prior = 0;
		wx2_prior = 0;
	  }
	
	if ( Base::_delta_to_random )
	  {
		_lambdag = _lambdag0 + Base::_delta * Base::_wnx + ( 1 - Base::_delta ) * Base::_wnx_prior;
		M = ( _lambdag0 * _M0 + Base::_delta * wx + ( 1 - Base::_delta ) * wx_prior ) / ( _lambdag(i1,i2,i3) );	  	  
	  }
	else
	  {
		_lambdag = _lambdag0 + Base::_delta * Base::_wnx;
		M = ( _lambdag0 * _M0 + Base::_delta * wx ) / ( _lambdag(i1,i2,i3) );	  	  
	  }
	
	for (int k = 0; k < Base::_k; ++k)
	  for (int i = 0; i < Base::_d1; ++i)
		for (int j = 0; j < Base::_d2; ++j)
		  {
			T n = Base::_delta * Base::_wnx(k,i,j) + ( 1 - Base::_delta ) * Base::_wnx_prior(k,i,j);
			if (n == 0)
			  {
				_betag(k,i,j,_all) = _betag0;
			  }
			else
			  for (int m = 0; m < Base::_m; ++m)
				{
				  T xm  = Base::_delta_to_random ?
					( Base::_delta * wx (k,i,j,m) + ( 1 - Base::_delta ) * wx_prior (k,i,j,m) ) / n :
					( Base::_delta * wx (k,i,j,m)                                             ) / n ;
				  T x2m = Base::_delta_to_random ?
					( Base::_delta * wx2(k,i,j,m) + ( 1 - Base::_delta ) * wx2_prior(k,i,j,m) ) / n :
					( Base::_delta * wx2(k,i,j,m)                                             ) / n ;
				  T S2 = std::max( T(0), x2m - xm * xm );
				  T D2 = ( xm - _M0 ) * ( xm - _M0 );
				  T betanew = _betag(k,i,j,m) = _betag0 + T(0.5) * n * ( S2 +  _lambdag0 * D2 / ( _lambdag0 + n ) );
				  _betag(k,i,j,m) = betanew;
				}
		  }
	
	Base::dirty();
	
}

template<class T> void NormalMixtureModel2<T>::update( bool force )
{
  BLITZ_INDICES;

  if ( !ready() ) return;

  initialize();
  
  if ( Base::update(force) )
	{

	  OPENMP(single)
	  {
		
		T log2pi = safe_log(2*M_PI);

		Matrix1  & Mb = Base::_background_parameters[0];
		Matrix1  & Taub = Base::_background_parameters[1];
		Mb.resize( Base::_m );
		Taub.resize( Base::_m );
		Mb = _M0;
		Taub = _alphag0 / _betag0;
		_lbias_b = T(0.5) * sum( safe_log( Taub ) - log2pi );

		_lbias.resize  ( Base::_k ); _lbias = 0;
		_lbias3.resize ( Base::_d1, Base::_d2, Base::_k); _lbias3 = 0;
		_lbias4 .resize( Base::_k, Base::_d1, Base::_d2, Base::_m ); 		
		_lbias_ml4.resize( Base::_k, Base::_d1, Base::_d2, Base::_m ); 		
		_Tau.resize    ( Base::_k, Base::_d1, Base::_d2, Base::_m ); 
		
		
		if ( Base::_requested_parameter_mode == Base::VARIATIONAL )
		  {
			for (int k = 0; k < Base::_k; ++k)
			  for (int i = 0; i < Base::_d1; ++i)
				for (int j = 0; j < Base::_d2; ++j)
				  {
					T lambda = _lambdag(k,i,j);
					T alphag = calc_alphag( lambda );
					T psialphag = psi(alphag);
					for (int m = 0; m < Base::_m; ++m)
					  {
						T betag = _betag(k,i,j,m);
						T logbetag = safe_log(betag);
						T lbias  = T(0.5) * ( psialphag - logbetag - log2pi );
						_Tau(k,i,j,m)     = alphag / betag;
						_lbias(k)        += lbias;
						_lbias3(i,j,k)   += lbias;
						_lbias4(k,i,j,m)  = lbias;
						_lbias_ml4(k,i,j,m)  = T(0.5) * ( safe_log( lambda ) - safe_log( lambda + 1 ) - log2pi ) + gammaln( alphag + 0.5 ) - gammaln( alphag ) + alphag * safe_log( betag );
					  }
				  }
			Base::_generative_emission_parameters[0].reference( _Mg );
			Base::_generative_emission_parameters[1].reference( _Tau );
			//			std::cout << "nmm2 made variational parameters and normalization" << std::endl;
		  }
		else if ( Base::_requested_parameter_mode == Base::GENERATIVE ||
				  ( Base::_requested_parameter_mode == Base::DISCRIMINATIVE &&
					Base::_generative_emission_parameters[0].numElements() == 0 ) )
		  {
			for (int k = 0; k < Base::_k; ++k)
			  for (int i = 0; i < Base::_d1; ++i)
				for (int j = 0; j < Base::_d2; ++j)
				  {
					double alphag = calc_alphag( _lambdag(k,i,j) );
					for (int m = 0; m < Base::_m; ++m)
					  {
						T betag = _betag(k,i,j,m);
						_Tau(k,i,j,m)     = alphag / betag;
					  }
				  }
			Base::_generative_emission_parameters[0].reference( _Mg );
			Base::_generative_emission_parameters[1].reference( _Tau );
			//			std::cout << "nmm2 made generative parameters" << std::endl;
		  }

		if ( Base::_requested_parameter_mode == Base::DISCRIMINATIVE )
		  {
			if ( Base::_discriminative_emission_parameters[0].numElements() == 0 )
			  {
				Base::_discriminative_emission_parameters[0].reference( Base::_generative_emission_parameters[0].copy() );
				Base::_discriminative_emission_parameters[1].reference( Base::_generative_emission_parameters[1].copy() );
				//				std::cout << "nmm2 copied discriminative from generative parameters" << std::endl;
			  }
			_M.reference( Base::_discriminative_emission_parameters[0] );
			_Tau.reference( Base::_discriminative_emission_parameters[1] );
		  }
		else
		  {
			_M.reference( _Mg );
			_Tau.reference( Base::_generative_emission_parameters[1] );
			//			std::cout << "nmm2 restored generative M parameter" << std::endl;
		  }

		if ( Base::_requested_parameter_mode != Base::VARIATIONAL )
		  {
			for (int k = 0; k < Base::_k; ++k)
			  for (int i = 0; i < Base::_d1; ++i)
				for (int j = 0; j < Base::_d2; ++j)
				  {
					T lambda = _lambdag(k,i,j);
					T alphag = calc_alphag( lambda );					
					for (int m = 0; m < Base::_m; ++m)
					  {
						T betag = _betag(k,i,j,m);
						T lbias  = T(0.5) * ( safe_log( _Tau(k,i,j,m) ) - log2pi );
						_lbias(k)        += lbias;
						_lbias3(i,j,k)   += lbias;
						_lbias4(k,i,j,m)  = lbias;
						_lbias_ml4(k,i,j,m)  = T(0.5) * ( safe_log( lambda ) - safe_log( lambda + 1 ) - log2pi ) + gammaln( alphag + 0.5 ) - gammaln( alphag ) + alphag * safe_log( betag );
					  }
				  }
		  }


		_Taur.resize( Base::_d1, Base::_d2, Base::_m, Base::_k ); 
		_Mr.resize  ( Base::_d1, Base::_d2, Base::_m, Base::_k ); 
		_Taur2.resize( Base::_d1, Base::_d2, Base::_k, Base::_m );
		_Mr2.resize  ( Base::_d1, Base::_d2, Base::_k, Base::_m );
		_lbias4r.resize( Base::_d1, Base::_d2, Base::_m, Base::_k); 
		_Taur = _Tau(i4,i1,i2,i3);
		_Mr   = _M  (i4,i1,i2,i3);
		_Taur2 = _Tau(i3,i1,i2,i4);
		_Mr2   = _M  (i3,i1,i2,i4);
		_lbias4r  = _lbias4 (i4,i1,i2,i3);


	  }
	}
  OPENMP(flush);
}

template class NormalMixtureModel2<float>;
INSTANTIATE_GATED_MESSAGE_DELEGATE( NormalMixtureModel2<float> );
