/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef NMM2_HPP
#define NMM2_HPP

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/binary_object.hpp>
#include <iostream>

#include <time.h>

#include "MixtureModel2Base.hpp"
#include "BatchNG.hpp"
#include "metrics.hpp"
#include "greedy_clusterer.hpp"

#ifdef HAVE_SSE
#include "sse_approx_logexp.h"
#include "sse_math.hpp"
#include "sse_templates.hpp"
#endif
#include "approx_logexp.h"

#define foreach BOOST_FOREACH

template<typename T> class NormalMixtureModel2;

template<typename T> struct has_upward_message_iterator< NormalMixtureModel2<T> >
 : public boost::true_type {};

template<typename T> struct TypeTraits< NormalMixtureModel2<T> > 
{	
  typedef T num_type;
  static const int num_statistics = 2;
};

template<class T> class NormalMixtureModel2 
  : public MixtureModel2Base< NormalMixtureModel2<T> >,
	public HasProperties< SearisePropertyTypes, NormalMixtureModel2<T> >
{
  
  typedef NormalMixtureModel2<T> This;
  typedef MixtureModel2Base< This > Base;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  
public:
  
  typedef T num_type;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,4> Matrix4;

  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<double,4> Matrix4d;

  typedef blitz::Array<unsigned int,2> IMatrix2;
  
  NormalMixtureModel2();
  
  SeariseProperty get_mu() 
  {
    update();
    CPUData<float> mu;
    blitz2cpudata( _M, mu );
    return mu;
  }

  SeariseProperty get_tau() 
  {
    update();
    CPUData<float> tau;
    blitz2cpudata( _Tau, tau );
    return tau;
  }

  struct upward_message_iterator_t
  {
	inline upward_message_iterator_t( int _M, int _K, const T * _x, T * _tau, T * _mu, T * _taur, T * _mur, T * _bias )
	  : M(_M), K(_K), x(_x), tau(_tau), mu(_mu), taur(_taur), mur(_mur), bias(_bias), k(0) 
	{}
	inline void advance( int n ) { k += n; }
	static T pow2(T x) { return x * x; }
	inline T evaluate_log( ) 
	{
	  T y = bias[k];
	  size_t k_off = k * M;
	  int m = 0;
#if 0//def HAVE_SSE
	  vecf4 dy4(0);
	  for ( ; ( m + 3 ) < M; m += 4 ) 
		{
		  dy4 += vecf4( taur + k_off + m ), * pow2( vecf4( x + m ) - vecf4( mur + k_off + m ) );
		}
	  y += dy4.sum();
#endif
	  for ( ; m < M; ++m ) 
		{
		  const T dy = taur[k_off+m] * pow2( x[m] - mur[k_off+m] );
		  y -= dy;
		}
	  return y;
 	}
	inline T evaluate_linear() 
	{
	  T p = aexpf( evaluate_log() );
	  return p;
	}
#ifdef HAVE_SSE
	static vecf4 pow2(vecf4 x) { return x * x; }
	inline vecf4 evaluate_log_sse( )
	{
	  vecf4 y = vecf4( bias + k, sse_aligned( bias + k ) );
	  for ( int m = 0, im = 0; m < M; ++m, im += K ) 
      y -= vecf4( tau + im + k, sse_aligned( tau + im + k ) ) 
          * pow2( x[m] - vecf4( mu + im + k, sse_aligned( mu + im + k )) );
	  return y;
	}
	inline vecf4 evaluate_linear_sse() 
	{
	  return exp( evaluate_log_sse() );
	}
#endif
	const int M;
	const int K;
	const T * x;
	const T * tau;
	const T * mu;
	const T * taur;
	const T * mur;
	const T * bias;
	int k;
  };
  
  inline upward_message_iterator_t background_message_iterator( const T * x )
  {
	T * M   = Base::_background_parameters[0].data();
	T * Tau = Base::_background_parameters[1].data();
	return upward_message_iterator_t( Base::_m, 1, x, Tau, M, Tau, M, &_lbias_b );
  }
  
  inline upward_message_iterator_t upward_message_iterator( int irf, int jrf, const T * x ) 
  {
	return upward_message_iterator_t( Base::_m, Base::_k, x
									  , &_Taur( irf, jrf, 0, 0 ), &_Mr( irf, jrf, 0, 0 )
									  , &_Taur2( irf, jrf, 0, 0 ), &_Mr2( irf, jrf, 0, 0 )
									  , &_lbias3( irf, jrf, 0 ) );
  }


  
  Matrix2 get_mur( int irf, int jrf ) { BLITZ_INDICES; Matrix4 Mrw( _Mr, true ); return Mrw( irf, jrf, _all, _all ); }

  void sample_point( int irf, int jrf, int k, Matrix1 & data );

  static void init_properties();
  
  void flush();
  
  //  std::vector<int> prune(T threshold);
  
  void batchlearn_begin_epoch();
  
  void reset_input_dimensions(const std::set<unsigned int> & components);
  
  void reset_components(const std::set<unsigned int> & components);

#ifdef _WIN32
#define STUB_ERROR throw std::runtime_error(boost::str(boost::format("called unimplemented stub  %s")%typeid(*this).name()));
#else
#define STUB_ERROR throw std::runtime_error(str(boost::format("called unimplemented stub %s on %s")%__func__%typeid(*this).name()));
#endif
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> & ) 
  {STUB_ERROR;}
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> &, blitz::Array<T,2> & ) 
  {STUB_ERROR;}
  void upward_message( blitz::Array<unsigned int,2> &, blitz::Array<T,1> &, Matrix3 & ) 
  {STUB_ERROR;}
  void downward_message( const Matrix3 & data, const Matrix1 & rout, Matrix3 & lrin, const Matrix2 & w)
  {STUB_ERROR;}    
  void downward_message( const IMatrix2 & data, int iout, Matrix3 & lrin, const Matrix2 & w)
  {STUB_ERROR;}    
  void downward_message( const IMatrix2 & data, int iout, Matrix3 & lrin, const Matrix3 & w)
  {STUB_ERROR;}    
#undef STUB_ERROR

  Matrix3 calc_max_component_likelihoods() { return _lbias3.transpose(2,0,1).copy(); }

  Matrix3 calc_component_entropies() { return Matrix3( - _lbias3.transpose(2,0,1) + T(0.5 * Base::_m) ); }

  void finalize_downward_message( Matrix3 & msg, bool linear = false );

  void initialize_downward_message( Matrix3 & msg, bool linear = false );

  void upward_message ( const Matrix3 & X, Matrix1 & lrout, bool marginal = false);
  
  void upward_message ( const Matrix3 & X, Matrix1 & lrout, const Matrix2 & w, bool marginal = false);
  
  void downward_message( const blitz::Array<T,3> &, const blitz::Array<T,1> &, Matrix3 & lrin );

  void self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin, const Matrix2 & w );

  void self_information ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin );

  void local_likelihood( Matrix3 &, Matrix1 &, T &, bool marginal = true );

  void local_likelihood( Matrix3 &, Matrix1 &, T &, Matrix2 &, bool marginal = true );

  void collect_statistics( const Matrix3 & X, const Matrix1 & r);
  
  void collect_statistics( const Matrix3 & X, const Matrix1 & r, const Matrix2 & w2);
  
  void collect_statistics_gated( Matrix1 & x, Matrix3 & z, Matrix1 & w, T s0, InverseRFIterator & irf );

  void collect_sampled_statistics( Matrix1 & x, int i1rf, int i2rf, int k, T s0 );

  void collect_sampled_background_statistics( Matrix1 & x, T s0 );
  
  Matrix1 batchlearn_end_epoch(Matrix1 f_in = Matrix1());
  
  void recalc_hyperparams ();
  
  void sample_one(unsigned int out, Matrix3 & X);
  
  void initialize();
  
  void update(bool force = false);

  bool ready()
  {
    return _pre_clustering_mode == NONE || Base::_initialized;
  }
  
  void recalc_likelihoods();

  bool input_is_distribution() { return false; }

  void apply_emission_gradient( typename Base::emission_gradient_t & parameter_gradient
								, T rate );
  void report_relative_gradient_magnitudes()
  {
	Matrix4 & d_M_scaled = Base::_accumulated_emission_gradient[0];
	Matrix4 & d_lTau = Base::_accumulated_emission_gradient[1];
	//Matrix4 & M_d = Base::_discriminative_emission_parameters[0];
	Matrix4 & Tau_d = Base::_discriminative_emission_parameters[1];
	Matrix4 g_rel_mu( d_M_scaled / sqrt(Tau_d) ); // we just take the scale, the mean magnitude does not have much meaning here
	Matrix4 g_rel_tau( d_lTau / safe_log( Tau_d ) );
	{
	  T x_mean = mean(g_rel_mu), x2_mean = mean(g_rel_mu*g_rel_mu), x_max = max(g_rel_mu), x_min = min(g_rel_mu);
	  T x_std = std::sqrt( std::max( T(0), x2_mean - x_mean * x_mean ) );
	  std::cout << "   mu: [ " << x_min << ", " << x_mean << " (" << x_std << "), " << x_max << "]" << std::endl;
	}
	{
	  T x_mean = mean(g_rel_tau), x2_mean = mean(g_rel_tau*g_rel_tau), x_max = max(g_rel_tau), x_min = min(g_rel_tau);
	  T x_std = std::sqrt( std::max( T(0), x2_mean - x_mean * x_mean ) );
	  std::cout << "   tau: [ " << x_min << ", " << x_mean << " (" << x_std << "), " << x_max << "]" << std::endl;
	}
  }


  T canonical_parameter_scale( int stats_index )
  {
	Matrix4 & M = Base::_discriminative_emission_parameters.at(0);
	Matrix4 & Tau = Base::_discriminative_emission_parameters.at(1);
	switch( stats_index )
	  {
	// the canonical mean parameter is the mean relative to the standard deviation
	  case 0: return mean( sqrt(Tau) * M );
	// the canonical precision parameter is the log of the parameter
	  case 1: return mean( safe_log( Tau ) );
	  }
	throw std::runtime_error("admissible parameter indices are 0 or 1");
	return 0;
  }


  void calculate_emission_gradient( const Matrix3 & x
				    , const Matrix1 & output_gradient
				    , typename Base::emission_gradient_t & gradient );
  
  void calculate_emission_gradient( const Matrix3 & x
				    , const Matrix2 & input_gradient
				    , const Matrix1 & output_gradient
				    , typename Base::emission_gradient_t & gradient );

  void calculate_emission_hessian( const Matrix3 & x
								   , const Matrix1 & output_gradient
								   , typename Base::emission_gradient_t & gradient );
								
private:
  
  //static double dL_lambda0 (double lb0, void * params);  
  void optimize_hyperparameters (double delta = 1, double llambdamin = -5, double llambdamax = 15);
  
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(Base);
    ar & BOOST_SERIALIZATION_NVP(_betag0);
    ar & BOOST_SERIALIZATION_NVP(_lambdag0);
    recalc_hyperparams();
  }
  
  double calc_alphag(double lambdag);
  Matrix3d calc_alphag();

  T _M0, _alphag0, _betag0, _lambdag0; //priors
  T _Lmu, _lbias_b;

  int _learn_beta0_interval,_learn_beta0_counter;

  Matrix3d _lambdag;
  Matrix4 _M, _Mg, _Mr, _Mr2;
  Matrix1 _lbias;
  Matrix4 _Tau, _Taur, _Taur2;
  Matrix4d _betag;
  Matrix4 _Tau_init, _M_init; // for external initialization
  
  Matrix3 _r, _lbiasp3, _lbias3;

  Matrix4 _lbias4, _lbias4r, _lbias_ml4;
  
  BatchNG< T, euclidian_metric<T> > _batchng;
  greedy_clusterer< T, euclidian_metric<T> > _greedy_clusterer;

  typedef enum { NONE, BATCHNG, GREEDY } PreClusteringMode;
  PreClusteringMode _pre_clustering_mode;

};

//templated class versioning
namespace boost { namespace serialization {
	template<typename T>
	struct version< NormalMixtureModel2<T> >
	{
	  BOOST_STATIC_CONSTANT(unsigned int, value = 4);
	};
}}
#endif
