/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef OPTIMAL_HYPERPARAMETERS_MULTINOMIAL
#define OPTIMAL_HYPERPARAMETERS_MULTINOMIAL

#include <utility>
#include <stdexcept>
#include <iostream>
#include <sstream>

#include <boost/math/special_functions.hpp>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_sf.h>

#include <blitz/array.h>

extern bool gsl_error_flag;

template<typename T> struct 
optimal_hyperparameter_multinomial_params
{
	int K;
	int N;
	T * p;
	T * Nk;
	T dL0;
	T cdiag;
};

inline void 
gsl_error_handler (	const char * reason,
										const char * file,
										int line,
										int gsl_errcode      )
{
  std::stringstream ss;
    ss << "gsl error " << reason 
      << " in " << file << "(" << line << ")" 
      << " error: " << gsl_strerror(gsl_errcode);
    std::cout << ss.str();
  gsl_error_flag = true;
}

inline std::pair<double,bool> //root, converged
find_root( gsl_function * F, 
					 double xlo, double xhi,
					 double abstol=0, double reltol=0.001,
					 int maxiter = 35,
					 const gsl_root_fsolver_type * type = gsl_root_fsolver_brent )
{
	//set up gsl
	gsl_root_fsolver * s = gsl_root_fsolver_alloc(type);
	gsl_root_fsolver_set (s, F, xlo, xhi);

	//iterate root finder
	int status=-1, iter=0;
  double r = (xlo+xhi)/2, x_lo, x_hi;
	try
	{
    do {
      status = gsl_root_fsolver_iterate (s); //status unused!
			if (status != GSL_SUCCESS) break;
      r = gsl_root_fsolver_root (s);
      x_lo = gsl_root_fsolver_x_lower (s);
      x_hi = gsl_root_fsolver_x_upper (s);
      status = gsl_root_test_interval (x_lo, x_hi, abstol, reltol);
    } while (status == GSL_CONTINUE && ++iter < maxiter);
	} catch(std::runtime_error & e) {
		std::cerr << "runtime error finding root: " << e.what() << std::endl;
	} catch(std::domain_error & e) {
		std::cerr << "domain error finding root: " << e.what() << std::endl;
	} catch(std::exception & e) {
		std::cerr << "exception finding root: " << e.what() << std::endl;
	} catch(...) {
		std::cerr << "...something... finding root" << std::endl;
	}
	
  bool converged = (status == GSL_SUCCESS) && (gsl_error_flag == false);
	
  gsl_error_flag = false;

	//TODO: error handling
	return std::make_pair(r,converged);
}

inline double log_update(double oldval, double loptval, double delta)
{
  double lold = std::log(oldval);
  double lnew = (1-delta)*lold + delta*loptval;
  return std::exp(lnew);
}

inline double lin_update(double oldval, double optval, double delta)
{
  return (1-delta)*oldval + delta*optval;
}

template<typename T> double 
optimal_hyperparameter_uniform_multinomial_dL (double la0, void * vparams)
{
	typedef optimal_hyperparameter_multinomial_params<T> ParamT;
	ParamT * params = (ParamT*) vparams;
	T dL = params->dL0;
	T * pNk = params->Nk;
	T * pp = params->p;
	T a0 = std::exp(la0);
	for (int i = 0; i < params->N; ++i, pNk += params->K, pp += params->K)
	{
		T N = 0;
		for (int j = 0; j < params->K; ++j)
		{
			N += pNk[j];
			dL -= boost::math::digamma(a0 + pNk[j]);
		}
		dL += params->K * boost::math::digamma(params->K * a0 + N);
	}
	return double(dL);
}

template<typename T> std::pair<T,bool> 
optimal_hyperparameter_uniform_multinomial (int K, int N, T * p, T * Nk, T lalphamin, T lalphamax)
{
	optimal_hyperparameter_multinomial_params<T> params = {K,N,p,Nk,0,0};
  gsl_function F = {&optimal_hyperparameter_uniform_multinomial_dL<T>, &params};
	gsl_set_error_handler(&gsl_error_handler);

	//calculate constant offset
	T * pp = p;
	for (int i = 0; i < N; ++i, pp += K)
		for (int j = 0; j < K; ++j)
		{
			params.dL0 += (pp[j]);
		}
	
	return find_root(&F, lalphamin, lalphamax);
}

template<typename T> double optimal_hyperparameter_fattened_markov_transitions_dL (double lg0, void * vparams)
{
	//unpack parameters
	typedef blitz::Array<T,2> M2;
	typedef std::pair<M2,T> Params;
	Params params = *static_cast<Params*>(vparams);
	M2 n(params.first);
	int K = n.extent(0);
	T gamma0 = std::exp(lg0);
	T c = params.second;
	T f = c - 1;
	T KK = T(K) + f;
	T dL = 0;
	//get going
	for (int i = 0; i < K; ++i)
	{
		T Ni = 0;
		for (int k = 0; k < K; ++i)
		{
			T nki = n(i,k);
			T cki = (k==i)?c:T(1);
			Ni += nki;
			dL += cki*(boost::math::digamma(nki+gamma0*cki) - boost::math::digamma(gamma0*cki));
		}
		dL += KK * (boost::math::digamma(KK*gamma0)-boost::math::digamma(Ni+KK*gamma0));
	}
	return double(dL);
}

template<typename T> std::pair<T,bool> optimal_hyperparameter_fattened_markov_transitions (
	blitz::Array<T,2> n, T c, T lalphamin = T(-5), T lalphamax = T(15) )
{
	typedef blitz::Array<T,2> M2;
	typedef std::pair<M2,T> Params;
	Params params(n,c);
  gsl_function F = {&optimal_hyperparameter_fattened_markov_transitions_dL<T>, &params};
	gsl_set_error_handler(&gsl_error_handler);
	return find_root(&F, lalphamin, lalphamax);
}

#endif
