/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "reduced_rank_covariance.hpp"
#include "sorting_tools.hpp"

#include <vector>


template<typename T, int R> 
reduced_rank_covariance<T,R>::reduced_rank_covariance( int b, int k, T gamma )
	: _initialized( false ), _ready( false ) 
	, _b(b), _k(k), _batch_pos(0)
	, _G_workspace(0), _G(0), _G_evecs(0), _G_evals(0)
	, _gamma( gamma ), _w( 0 ), _wx2( 0 )
  {
  }
template<typename T, int R> 
void reduced_rank_covariance<T,R>::product_with_precision( data_t & x, T lambda )
{
  if ( !ready() ) throw std::runtime_error("trying to use reduced rank covariance before statistics have been accumulated");
  // using the woodbury identity C in low rank approximation becomes
  // ( lambda * I + X * X' )^-1 = lambda^-1 * ( I - lambda^-1* X * ( I + X' * ( lambda^-1 I ) * X )^-1 * X' ) )
  //                            = lambda^-1 * ( I -            X * ( lambda + X' * X              )^-1 * X' ) )
  // parameter _lambda is the sqrt of lambda acutally, i.e. the soft threshold on the gradient standard deviation
  //T pre_norm = std::sqrt( mean( pow2( x ) ) );
  T delta = std::sqrt( _wx2 / _w );
  // subtraction of the projections can be done in sequence since the projection directions are orthogonal
  // thus we can do this in-place, saving memory and bandwidth
  for ( size_t k = 0; k < _k; ++k )
	{
	  T Zk = _X_norms(k) / _w;
	  if ( Zk > 0 )
		{		  
		  T lambda_scaled = lambda / Zk * delta;
		  T coeff = sum( blitz_slice( _X, 0, k ) * x ) / ( T(1) + lambda_scaled * lambda_scaled );
		  x -= coeff * blitz_slice( _X, 0, k );
		}
	}  
  //T post_norm = std::sqrt( mean( pow2( x ) ) );
  //std::cout << " reduced_rank_convariance regularization " << post_norm / std::max( T(1e-10), pre_norm ) << " |x| pre " << pre_norm << std::endl;
}

template<typename T, int R> 
void reduced_rank_covariance<T,R>::reduce()
{
  
  //std::cout << "reducing " << _X_large.shape() << " min " << min( _X_large ) << " mean " << mean( _X_large ) << " max " << max( _X_large );
  //std::cout << " pre scale " << _scaling;

  // keep a rescaled X with absmax 1
  // this leaves the eigenvectors invariant


  if ( _ready )
	{
	  // decay old entries
	  _X_norms( blitz::Range( 0, _k - 1 ) ) *= std::pow( _gamma, _b / T(2) );
	  // normalize and decay new entries
	  for ( int i = _k; i < ( _k + _b ); ++i )
		{
		  data_t Xi( blitz_slice( _X_large, 0, i ) );
		  T scale = T(1) / std::max( T(1e-15), max( abs( Xi ) ) );
		  _X_norms( i ) = std::sqrt( sum( pow2( scale * Xi ) ) ) / scale;
		  if ( _X_norms( i ) > 0 )
			Xi *= T(1) / _X_norms(i);
		  _X_norms( i ) *= std::pow( _gamma, ( _k + _b - i - 1 ) / T(2) );
		}
	}
  else
	{
	  for ( int i = 0; i < ( _k + _b ); ++i )
		{
		  data_t Xi( blitz_slice( _X_large, 0, i ) );
		  T scale = T(1) / std::max( T(1e-15), max( abs( Xi ) ) );
		  _X_norms( i ) = std::sqrt( sum( pow2( scale * Xi ) ) ) / scale;
		  if ( _X_norms( i ) > 0 )
			Xi *= T(1) / _X_norms(i);
		  _X_norms( i ) *= std::pow( _gamma, ( _k + _b - i - 1 ) / T(2) );
		}
	}

  // calculate the gram-matrix
  // rescale to avoid instability ( leaves eigenvectors intact )
  T maxnorm = std::max( T(1e-15), max( _X_norms ) );
  int r = _k + _b;
  for ( int r1 = 0; r1 < r; ++r1 )
	{
	  for ( int r2 = 0; r2 <= r1; ++r2 )
		{
		  double g12 = sum( blitz_slice( _X_large, 0, r1 ) * blitz_slice( _X_large, 0, r2 ) ) 
			* ( _X_norms( r1 ) / maxnorm ) * ( _X_norms( r2 ) / maxnorm );
		  if ( r1 != r2 )
			{
			  gsl_matrix_set( _G, r2,r1, g12 );
			  gsl_matrix_set( _G, r1,r2, g12 );
			}
		  else
			{
			  g12 = std::max( 1e-10, g12 );
			  gsl_matrix_set( _G, r1,r2, g12);
			}
		}
	}

  // decompose the gram-matrix using gsl
  gsl_eigen_symmv( _G, _G_evals, _G_evecs, _G_workspace );

  // keep the k eigenvectors with the largest eigenvalues ( most variant directions )
  std::vector<size_t> order( r );
  std::vector<float> evals( r );
  for ( int i = 0; i < r; ++i )
	evals[i] = gsl_vector_get( _G_evals, i );
  covariance_t X_new( _X.shape() );
  X_new = 0;
  compute_order( order.begin(), order.end(), evals );

  double variance_captured = 0;
  for ( int k = 0; k < int(_k); ++k )
	variance_captured += evals[  order[ r - k - 1 ] ];
  double variance_lost = 0;
  for ( int k = 0; k < int(_b); ++k )
	variance_lost += evals[ order[ k ] ];
  //std::cout << " reduction lost " << ( variance_lost / ( variance_lost + variance_captured ) * 100 ) << "% of the variance" << std::endl;
  //std::cout << " smallest kept std eigenvector " << ( std::sqrt( evals[ order[ _b ] ] ) * maxnorm )
  //			<< " mean std " << std::sqrt( _wx2 / _w ) << std::endl;
  
  // form new X from old X linear combination given by the gram matrix eigenvectors
  for ( int k = 0; k < int(_k); ++k )
	{
	  size_t ev_index = order[ r - k - 1 ];
	  data_t X_new_k = blitz_slice( X_new, 0, k );
	  for ( int l = 0; l < r; ++l )
		{
		  X_new_k += blitz_slice( _X, 0, l ) * gsl_matrix_get ( _G_evecs, l, ev_index ) * _X_norms( l );
		}
	}

  _X = X_new;

  // renormalize
  for ( int i = 0; i < _k; ++i )
	{
	  data_t Xi( blitz_slice( _X, 0, i ) );
	  T scale = T(1) / std::max( T(1e-15), max( abs( Xi ) ) );
	  _X_norms( i ) = std::sqrt( sum( pow2( scale * Xi ) ) ) / scale;
	  if ( _X_norms( i ) > 0 )
		Xi *= T(1) / _X_norms(i);
	}

  _batch_pos = _k;
  _ready = true;

}

template class reduced_rank_covariance<float,3>;
template class reduced_rank_covariance<float,1>;
