/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include "blitz/array.h"
#include "blitz_utils.hpp"
#include <gsl/gsl_eigen.h>
#include <stdexcept>
#include <iostream>
#include <boost/serialization/serialization.hpp>

//FIXME: need a slice function for blitz arrays like slice( X, dim, offset ), then also do range( X, dim, lo, hi )

template<typename T, int R> class reduced_rank_covariance
{

public:

  reduced_rank_covariance( int b = 10, int k = 10, T gamma = 0.999 );

  ~reduced_rank_covariance()
  {
	clear();
  }

  bool ready() { return _ready; }

  void clear()
  {
    _X.free();
    _X_large.free();
    _X_norms.free();
    _batch_pos = 0;
	free_gsl_structs();
    _ready = _initialized = false;
  }

  typedef blitz::Array<T,R> data_t;
  typedef blitz::Array<T,1> vector_t;
  typedef blitz::Array<T,R+1> covariance_t;

  void operator()( const data_t x )
  {
    check_initialize(x);
    blitz_slice( _X_large, 0, _batch_pos++ ) = x;
    T scale = T(1) / std::max( T(1e-15), max( abs( x ) ) );
    _wx2 = _gamma * _wx2 + mean( pow2( x * scale ) ) / ( double( scale ) * double( scale ) );
    _w = _gamma * _w + 1;
    if ( _batch_pos == _b + _k )
      reduce();
  }

  void product_with_precision( data_t & x, T lambda );

  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
	ar & BOOST_SERIALIZATION_NVP(_initialized);
	ar & BOOST_SERIALIZATION_NVP(_ready);
	ar & BOOST_SERIALIZATION_NVP(_b);
	ar & BOOST_SERIALIZATION_NVP(_k);
	ar & BOOST_SERIALIZATION_NVP(_batch_pos);
	ar & BOOST_SERIALIZATION_NVP(_X);
	ar & BOOST_SERIALIZATION_NVP(_X_large);
	ar & BOOST_SERIALIZATION_NVP(_X_norms);
	ar & BOOST_SERIALIZATION_NVP(_gamma);
	ar & BOOST_SERIALIZATION_NVP(_w);
	ar & BOOST_SERIALIZATION_NVP(_wx2);
	alloc_gsl_structs(); // playing safe
	clear(); // crashes after deserialization
  }

  void set_gamma( T gamma ) { _gamma = gamma; }

private:

  void reduce();

  void free_gsl_structs()
  {
    if ( _G_evals ) gsl_vector_free( _G_evals ); 
    if ( _G ) gsl_matrix_free( _G );
    if ( _G_evecs ) gsl_matrix_free( _G_evecs );
    if ( _G_workspace ) gsl_eigen_symmv_free( _G_workspace );
	_G_evals = NULL;
	_G = NULL;
	_G_evecs = NULL;
	_G_workspace = NULL;
  }

  void alloc_gsl_structs()
  {
	free_gsl_structs();
    int r = _k + _b;
    _G_evals = gsl_vector_alloc( r );
    _G = gsl_matrix_alloc( r, r );
    _G_evecs = gsl_matrix_alloc( r, r );
    _G_workspace = gsl_eigen_symmv_alloc( r );
  }

  void check_initialize( const data_t & x )
  {
    if ( _initialized ) return;
    clear();
    int r = _k + _b;
	alloc_gsl_structs();
    // the additional dimension is inserted on the left
    // X_large keeps the low-rank covariance root and the additional data vectors
    blitz::TinyVector<int,R+1> large_shape;
    large_shape[0] = r;
    for ( size_t dim = 0; dim < R; ++dim )
      large_shape[dim+1] = x.extent(dim);
    _X_large.resize( large_shape );
    // X is just a view into the first k rows of X_large
    blitz::RectDomain<R+1> small_domain = _X_large.domain();
    small_domain.ubound( 0 ) = _k - 1;
    _X.reference( _X_large( small_domain ) );
    _X_norms.resize( r );
    _initialized = true;
    //std::cout << " initialized reduced rank covariance " << std::endl;
  }

  bool _initialized, _ready;
  size_t _b, _k, _batch_pos;
  // X_large accumulates b data vectors on top of X
  covariance_t _X, _X_large;
  vector_t _X_norms;
  gsl_eigen_symmv_workspace * _G_workspace;
  gsl_matrix * _G, * _G_evecs;
  gsl_vector * _G_evals;
  T _gamma; 
  double _w, _wx2;
};
