/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "segmentation_tools.hpp"
#include <set>
#include <algorithm>
#include <boost/foreach.hpp>

template<typename T>
void resizeweights ( 
					const blitz::Array<T,2> & weights, 
					blitz::Array<T,2> & result, 
					int w, int h )
{
  result.resize(w,h);
  T cx = T(weights.extent(0))/w;
  T cy = T(weights.extent(1))/h;
  for (int j = 0; j < h; ++j)
    for (int i = 0; i < w; ++i)
	  {
		int x1 = int(std::floor(i*cx));
		int y1 = int(std::floor(j*cy));
		int x2 = std::min(weights.extent(0)-1,int(std::ceil((i+1)*cx)));
		int y2 = std::min(weights.extent(1)-1,int(std::ceil((j+1)*cy)));
		blitz::Range rx = blitz::Range(x1,x2);
		blitz::Range ry = blitz::Range(y1,y2);
		blitz::Array<T,2> weightsslice(weights(rx,ry));
		result(i,j) = max(weightsslice);
	  }
}

template<class T>
void generate_label_segmentation (
								  const blitz::Array<T,3> & labels,
								  int step, int D, 
								  blitz::Array<T,3> & s,
								  const blitz::Array<T,2> & weights,
								  T lmaxlo, T lmaxhi, T suppression_factor )
{
  BLITZ_INDICES;
  
  int ni = labels.extent(0);
  int nj = labels.extent(1);
  if (s.numElements() == 0)
	{
	  s.resize(ni,nj,4);
	  s = T(-1);
	}
  blitz::Array<T,1> h(labels.extent(2));
  for (int i = 0; i + D <= ni; i += step)
    for (int j = 0; j + D <= nj; j += step)
	  {
		blitz::Range ri(i,i+D-1), rj(j,j+D-1);
		blitz::Array<T,3> labelscut(labels(ri,rj,_all));
		
		//compute an optionally weighed histogram and find value and position of maximum
		if (weights.numElements()>0)
		  {
			blitz::Array<T,2> weightscut(weights(ri,rj));
			for (int ii = 0; ii < D; ++ii)
			  for (int jj = 0; jj < D; ++jj)
				h += weightscut(ii,jj)*labelscut(ii,jj,_all);
		  }
		else
		  {
			h = sum(sum(labelscut(i2,i3,i1),i3),i2);
		  }
		
		h *= T(1)/sum(h);
		int imax = maxIndex(h)(0);
		
		//heuristically determine "winner on a border"
		//if there is one pre-segment, keep normalized
		//    if (h(imax) > lmaxlo && h(imax) < lmaxhi)
		{
		  blitz::Array<T,3> scut(s(ri,rj,_all));
		  
		  blitz::Array<T,2> lmax(labelscut(_all,_all,imax));
		  
#if 0
		  //assign the others (1-lmax) / 3
		  scut = 1; 
		  for (int i=0; i<4; ++i)
			scut(_all,_all,i) -= lmax;
		  scut *= T(0.33333);
		  
		  //assign the winner lmax
		  blitz::Range r1(0,D/2-1), r2;
		  for (int is = 0; is < 4; ++is)
			{
			  switch(is)
				{
				case 0:r2=r1;break;
				case 1:r1=r1+D/2;break;
				case 2:r2=r2+D/2;break;
				case 3:r1=r1-D/2;break;
				}
			  scut(r1,r2,is) = lmax(r1,r2);
			}
#else
		  //just supress the outliers, depend on later renormalization
		  blitz::Range r1(0,D/2-1), r2;
		  for (int is = 0; is < 4; ++is)
			{
			  switch(is)
				{
				case 0:r2=r1;break;
				case 1:r1=r1+D/2;break;
				case 2:r2=r2+D/2;break;
				case 3:r1=r1-D/2;break;
				}
			  for (int ii = r1.first();  ii <= r1.last(); ++ii)
				for (int jj = r2.first();  jj <= r2.last(); ++jj)
				  if (lmax(ii,jj)<lmaxlo)
					{
					  //							scut(ii,jj,_all) = (T(1)-suppression_factor)/T(4);
					  scut(ii,jj,is) = suppression_factor;
					}
			}
#endif
		}
	  }
}

template<typename T> 
void convertlabels ( 
					const blitz::Array<unsigned int,2>  & labels, 
					int w, int h,
					blitz::Array<T,3> & out, 
					std::vector<int> & outindices,
					blitz::Array<T,2> & weights,
					T bgweight,
					std::map<int,int> & labelmap
					 )
{
  BLITZ_INDICES;
  std::set<unsigned int> labelset(labels.begin(),labels.end());
  bool have_labelmap = labelmap.size() > 0;
  outindices.clear(); 
  outindices.resize(labelset.size()); 
  std::vector<int>::iterator inditer = outindices.begin();
  BOOST_FOREACH(int label,labelset)
	{
	  *inditer++ = label;
	}
  blitz::Array<T,3> result(w,h,outindices.size()); 
  result = 0;
  T cx = T(labels.extent(0))/w;
  T cy = T(labels.extent(1))/h;
  for (int j = 0; j < h; ++j)
	{
	  for (int i = 0; i < w; ++i)
		{
		  int x1 = int(std::floor(i*cx));
		  int y1 = int(std::floor(j*cy));
		  int x2 = std::min(labels.extent(0)-1,int(std::ceil((i+1)*cx)));
		  int y2 = std::min(labels.extent(1)-1,int(std::ceil((j+1)*cy)));
		  blitz::Range rx = blitz::Range(x1,x2);
		  blitz::Range ry = blitz::Range(y1,y2);
		  blitz::Array<unsigned int,2> labelslice(labels(rx,ry));
		  for (int k = 0; k < int(outindices.size()); ++k)
			{
			  result(i,j,k) = count(labelslice==unsigned(outindices[k]));
			}
		}
	}
  normalize(result);
  out.reference(result);
  weights.resize(w,h);
  weights = 0;
  for (size_t l = 0; l < outindices.size(); ++l)
	{
	  int label = outindices[l];
	  if (have_labelmap)
		label = labelmap[label];
	  if (label==0)
		weights += bgweight * out(_all,_all,l);
	  else if (label>0)
		weights += out(_all,_all,l);
	}
}

template<class T>
void segment_propagate ( const blitz::Array<T,2> & in, 
						 const blitz::Array<T,3> & s,
						 blitz::Array<T,2> & out,
						 int step, int D )
{
  BLITZ_INDICES;
  
  int ni = in.extent(0);
  int nj = in.extent(1);

  if ( ni < D || nj < D ) 
    return;

  int mi = 1 + ( ni - D ) / step;
  int mj = 1 + ( nj - D ) / step;
  
  out.resize(mi,mj);

  blitz::Array<T,2> nout(out.shape());	

  out = 0;
  nout = 0;
 
  Matrix3 sw(s,true);
  Matrix2 inw(in,true);
  Matrix2 outw(out,true); 
  Matrix2 noutw(nout,true); 

  SEGMENTATION_RANGES(rs,D)

    int iend = ni - D + 1;

#ifdef HAVE_OPENMP
#pragma omp parallel for
#endif
  for (int i = 0; i < iend; i += step)
    for (int j = 0; j + D <= nj; j += step)
	  {
		T & outref ( outw ( i / step, j / step ) );
		T & noutref( noutw( i / step, j / step ) );
		blitz::Range ri( i, i + D - 1 ), rj( j, j + D - 1 );
		blitz::Array<T,2> incut( inw( ri, rj       ) );
		blitz::Array<T,3> scut ( sw ( ri, rj, _all ) );

		for (int is = 0; is < 4; ++is)
		  {
			for (int ii = 0; ii < D / 2; ++ii)
			  for (int jj = 0; jj < D / 2; ++jj)
				{
				  T svalue  =  scut( rs1[is].first() + ii, rs2[is].first() + jj, is );
				  T invalue = incut( rs1[is].first() + ii, rs2[is].first() + jj );
				  if ( svalue < 0 ) svalue = T(0.25);
				  outref  += svalue *       invalue;
				  noutref += svalue * ( 1 - invalue );
				}			
		  }

		outref = ( ( outref > 0 ) || ( noutref > 0 ) ) ? outref / ( outref + noutref ) : T(0.5);

	  }
  
}

template<class T>
void segment_propagate ( const blitz::Array<T,3> & in, 
						 const blitz::Array<T,3> & s,
						 blitz::Array<T,3> & out,
						 int step, int D )
{
  BLITZ_INDICES;
  
  int ni = in.extent(0);
  int nj = in.extent(1);
  int mi = ni>=D?(1+(ni-D)/step):0;
  int mj = nj>=D?(1+(nj-D)/step):0;
  if (mi == 0 || mj == 0) 
    return;
  
  out.resize(mi,mj,in.extent(2));
  out = 0;
  
  Matrix3 sw(s,true);
  Matrix3 inw(in,true);
  Matrix3 outw(out,true); 

  int iend = ni - D + 1;

  SEGMENTATION_RANGES(rs,D)

    
#ifdef HAVE_OPENMP
#pragma omp parallel for
#endif
    for (int i = 0; i < iend; i += step)
      for (int j = 0; j + D <= nj; j += step)
	  {
		blitz::Array<T,1> outcolumn(outw(i/step,j/step,_all));
		blitz::Range ri( i, i + D - 1 ), rj( j, j + D - 1 );
		blitz::Array<T,3> labelscut( inw ( ri, rj, _all ) );
		blitz::Array<T,3> scut     ( sw  ( ri, rj, _all ) );

		for (int is = 0; is < 4; ++is)
		  {
			for ( int ii = 0; ii < D / 2; ++ii )
			  for ( int jj = 0; jj < D / 2; ++jj )
				{
				  T svalue =                           scut  ( rs1[is].first() + ii, rs2[is].first() + jj, is     );
				  blitz::Array<T,1> labelscolumn ( labelscut ( rs1[is].first() + ii, rs2[is].first() + jj, _all ) );
				  if (svalue < 0) svalue = T(0.25);
				  outcolumn += svalue * labelscolumn;
				}			
		  }
		normalize(outcolumn);
	  }
  
}

template<typename T>
blitz::Array<T,3> resize_multinomial_array (const blitz::Array<T,3> & in, int w, int h, const blitz::Array<T,2> * weights )
{
  BLITZ_INDICES;
  int M = in.extent(2);
  blitz::Array<T,3>  result(w,h,M);
  float cx = T(in.extent(0))/w;	
  float cy = T(in.extent(1))/h;
  for (int j = 0; j < h; ++j)
    for (int i = 0; i < w; ++i)
	  {
		int x1 = int(std::floor(i*cx));
		int y1 = int(std::floor(j*cy));
		int x2 = std::min(in.extent(0)-1,int(std::ceil((i+1)*cx)));
		int y2 = std::min(in.extent(1)-1,int(std::ceil((j+1)*cy)));
		blitz::Range rx = blitz::Range(x1,x2);
		blitz::Range ry = blitz::Range(y1,y2);
		blitz::Array<T,3> inslice(in(rx,ry));
		if (weights)
		  {
			blitz::Array<T,2> weightsslice((*weights)(rx,ry));
			for (int m = 0; m < M; ++m)
			  result(i,j,m) = sum(weightsslice*inslice(_all,_all,m));
		  }
		else
		  result(i,j,_all) = sum(sum(inslice(i2,i3,i1),i3),i2);
	  }
  normalize(result);
  return result;
}

template<typename T>
blitz::Array<T,3> 
expand_indexed ( 
				const blitz::Array<T,3> & out, 
				const std::vector<int> & outindices, int K )
{
  BLITZ_INDICES;
  blitz::Array<T,3> tmp(out.extent(0),out.extent(1),K);
  tmp = 0;
  for (int i = 0; i < int(outindices.size()); ++i)
    tmp(_all,_all,outindices[i]) = out(_all,_all,i);
  return tmp;
}

template<typename T>
blitz::Array<T,3> apply_labelmap ( 
								  const blitz::Array<T,3> & out, const std::map<int,int> & labelmap, 
								  const std::vector<int> & outindices, int K,
								  T invalid_threshold )
{
  BLITZ_INDICES;
  blitz::Array<T,3> tmp(out.extent(0),out.extent(1),K);
  tmp = 0;
  blitz::Array<T,2> invalids;
  typedef std::pair<int,int> MappingT;
  BOOST_FOREACH(MappingT mapping, labelmap)
	{
	  //support sparse label-distribution
	  if (outindices.size() > 0)
		{
		  std::vector<int>::const_iterator pos = std::find( outindices.begin(), outindices.end(), mapping.first );
		  if (pos != outindices.end())
			mapping.first = pos - outindices.begin();
		  else
			continue;
		}
	  //handle invalids
	  if (mapping.second == -1)
		{
		  if (invalids.numElements() == 0)
			{
			  invalids.resize(out.extent(0),out.extent(1));
			  invalids = 0;
			}
		  invalids += out(_all,_all,mapping.first);
		}
	  else
		tmp(_all,_all,mapping.second) += out(_all,_all,mapping.first);
	}
  normalize(tmp);
  if (invalids.numElements() > 0)
    tmp = where(invalids(i1,i2)>invalid_threshold,-1,tmp);
  return tmp;
}

//instantiate
template void convertlabels<> ( 
							   const blitz::Array<unsigned int,2>  & labels, 
							   int w, int h,
							   blitz::Array<floattype,3> & outdata, 
							   std::vector<int> & outindices,
							   blitz::Array<floattype,2> & weights,
							   floattype bgweight,
							   std::map<int,int> &
								);

template blitz::Array<floattype,3> resize_multinomial_array (
															 const blitz::Array<floattype,3> & in, int w, int h, 
															 const blitz::Array<floattype,2> * weights);

template blitz::Array<floattype,3> apply_labelmap ( 
												   const blitz::Array<floattype,3> & out, const std::map<int,int> & labelmap, 
												   const std::vector<int> & outindices, int K,
												   floattype invalid_threshold );

template blitz::Array<floattype,3> expand_indexed ( 
												   const blitz::Array<floattype,3> & out, 
												   const std::vector<int> & outindices, int K );

template void resizeweights ( 
							 const blitz::Array<floattype,2> & weights, 
							 blitz::Array<floattype,2> & result, 
							 int w, int h );

template void generate_label_segmentation (
										   const blitz::Array<floattype,3> & labels,
										   int step, int D, 
										   blitz::Array<floattype,3> & s,
										   const blitz::Array<floattype,2> & weights,
										   floattype lmaxlo, floattype lmaxhi, 
										   floattype suppression_factor );

template void segment_propagate (
								 blitz::Array<floattype,3> const& in, 
								 blitz::Array<floattype,3> const& s,
								 blitz::Array<floattype,3> & out,
								 int step, int D);

template void segment_propagate (
								 blitz::Array<floattype,2> const& in, 
								 blitz::Array<floattype,3> const& s,
								 blitz::Array<floattype,2> & out,
								 int step, int D);
