/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef __SFA2_HPP__
#define __SFA2_HPP__

#include <blitz/array.h>
#include <gsl/gsl_eigen.h>
#include <vector>
#include <utility>
#include <algorithm>

#include "blitz_indices.hpp"


template<typename T>
class sfa2
{

  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,3> Matrix3;

public:

  sfa2()
  {
	init();
  }

  void init()
  {
	_k = 0;
	_x_last.free();
	_n = 0;
	_W_dt_x2.free();
	_W.free();
	_V.free();
  }

  void collect_statistics( const Matrix3 & x )
  {
	BLITZ_INDICES;
	if ( all( x.shape() == _x_last.shape() ) )
	  {
		if ( _k == 0 )
		  {
			_k = x.extent(2);
			_W_dt_x2.resize( _k, _k );
		  }
		else if ( _k != size_t( x.extent(2) ) )
		  throw std::runtime_error( "sfa2: depth mismatch" );
		for ( size_t i = 0; i < size_t( x.extent(0) ); ++i )
		  for ( size_t j = 0; j < size_t( x.extent(0) ); ++j )
			{
			  Matrix1 dt_x( x( int(i), int(j), _all ) - _x_last( int(i), int(j), _all ) );
			  _W_dt_x += dt_x;
			  _W_dt_x2 += dt_x(i1) * dt_x(i2);
			  _n += 1;
			}
	  }
	_x_last.reference( x );
  }

	template<typename P> static bool compare_first( const P & p1, const P & p2 ) { return p1.first < p2.first; }

  void do_sfa()
  {
	BLITZ_INDICES;
	gsl_eigen_symmv_workspace * esw = gsl_eigen_symmv_alloc( _k );
	gsl_matrix * C = gsl_matrix_alloc( _k, _k );
	gsl_matrix * evecs = gsl_matrix_alloc( _k, _k );
	gsl_vector * evals = gsl_vector_alloc( _k );

	Matrix1 E_dt_x ( _W_dt_x  / _n );
	Matrix2 E_dt_x2( _W_dt_x2 / _n );

	Matrix2 Cb( E_dt_x2 - E_dt_x(i1) * E_dt_x(i2) );

	for ( size_t i = 0; i < _k; ++i )
	  for ( size_t j = 0; j < _k; ++j )
		gsl_matrix_set( C, i, j, Cb( int(i), int(j) ) );

	gsl_eigen_symmv( C, evals, evecs, esw );

	typedef std::pair< double, int > ev_entry;
	std::vector<ev_entry> ev_entries;
	for ( size_t i = 0; i < _k; ++i )
	  ev_entries[i] = ev_entry( gsl_vector_get( evals, i ), i );
	std::sort( ev_entries.begin(), ev_entries.end(), compare_first<ev_entry> );

	_V.resize( _k );
	_W.resize( _k, _k );

	for ( size_t i = 0; i < _k; ++i )
	  {
		_V(i) = ev_entries[i].first;
		size_t evindex = ev_entries[i].second;
		for ( size_t j = 0; j < _k; ++j )
		  {
			_W( int(i), int(j) ) = gsl_matrix_get( evecs, j, evindex );
		  }
	  }

	gsl_eigen_symmv_free( esw );
	gsl_matrix_free( C );
	gsl_matrix_free( evecs );
	gsl_vector_free( evals );
  }

  const Matrix2 & get_W() const
  {
	return _W;
  }

  const Matrix1 & get_V() const
  {
	return _V;
  }

private:
  
  Matrix3 _x_last;
  Matrix1 _W_dt_x;
  Matrix2 _W_dt_x2, _W;
  Matrix1 _V;
  T _n;
  size_t _k;

};

#endif

