/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "tfmm.hpp"
#include "Random.hpp"

#include <boost/format.hpp>


template<typename T>
static blitz::Array<T,2> dummy_saliency_signal() { return blitz::Array<T,2>() ; }



template<typename T>
TemporalForwardMarkovModel<T>::TemporalForwardMarkovModel  ( int k ) 
  : _initcount(100)
  , _k(k)
  , _learn_gamma0_interval(0)
  , _learn_gamma0_count(0)
  , _initialized(false)
  , _dirty(false)
  , _in_batch(false)
  , _use_pi(false)
  , _use_pi_inseq(false)
  , _fix_transitions(false)
  , _init_uniform(true)
  , _factorize_transitions(false)
  , _learning(false)
  , _do_vb(true)
  , _owns_state(false)
  , _online_update(false)
  , _alpha0(0.5)
  , _gamma0(0.5)
  , _gammadiag0(0.5)
  , _delta(1)
  , _Lmu(0)
  , _Lp(0)
  , _L(0)
  , _tau(100)
{
  learned.connect( dummy_saliency_signal<T> );
}


template<typename T>
void TemporalForwardMarkovModel<T>::set_history_size( PropertyType p )
{
  int history_size = cast<int>( p );
  message_component<T>::buffer().set_history_capacity( std::max(history_size,1) );
}


template<typename T>
typename TemporalForwardMarkovModel<T>::PropertyType TemporalForwardMarkovModel<T>::get_history_size( )
{
  return message_component<T>::buffer().get_history_capacity();
}

template<typename T>
void TemporalForwardMarkovModel<T>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter( &This::_initialized, false );
  //read-write
  PBase::add_ivar_property_rw(std::string("do-vb"), &This::_do_vb );
  PBase::add_property(std::string("history-size"), typename PBase::MemFunGetter( &This::get_history_size ), typename PBase::MemFunSetter( &This::set_history_size ) );
  PBase::add_ivar_property_rw(std::string("normalize-whole-transition-matrix"), &This::_normalize_whole_transition_matrix );
  PBase::add_ivar_property_rw(std::string("init-uniform"), &This::_init_uniform );
  PBase::add_ivar_property_rw(std::string("owns-state"), &This::_owns_state );
  PBase::add_ivar_property_rw(std::string("k"), &This::_k, uninitializer );
  PBase::add_ivar_property_rw(std::string("alpha0"), &This::_alpha0, uninitializer);
  PBase::add_ivar_property_rw(std::string("tau"), &This::_tau );
  PBase::add_ivar_property_rw(std::string("online_update"), &This::_online_update );
  PBase::add_ivar_property_rw(std::string("gamma0"), &This::_gamma0, uninitializer);
  PBase::add_ivar_property_rw(std::string("gammadiag0"), &This::_gammadiag0, uninitializer);
  PBase::add_ivar_property_rw(std::string("delta"), &This::_delta);
  PBase::add_ivar_property_rw(std::string("initcount"), &This::_initcount);
  PBase::add_ivar_property_rw(std::string("fix-transitions"), &This::_fix_transitions);
  PBase::add_ivar_property_rw("learn-gamma0", &This::_learn_gamma0_interval);
  PBase::add_ivar_property_rw(std::string("use-pi"), &This::_use_pi);
  PBase::add_ivar_property_rw(std::string("use-pi-in-sequence"), &This::_use_pi_inseq);
  PBase::add_ivar_property_rw(std::string("factorize-joint-statistics"), &This::_factorize_transitions);
  PBase::add_ivar_property_rw(std::string("initialized"), &This::_initialized);

  //read-only
  PBase::add_property( "lpi"    , make_blitz_ivar_getter1( &This::_lpi     ) );
  PBase::add_property( "lA"     , make_blitz_ivar_getter1( &This::_lA      ) );
  PBase::add_property( "A"      , make_blitz_ivar_getter1( &This::_A       ) );
  PBase::add_property( "gamma"  , make_blitz_ivar_getter1( &This::_gamma   ) );
  PBase::add_property( "lpidash", make_blitz_ivar_getter1( &This::_lpidash ) );
  PBase::add_property( "lAdash" , make_blitz_ivar_getter1( &This::_lAdash  ) );
  PBase::add_property( "wA"     , make_blitz_ivar_getter1( &This::_wA      ) );
  //PBase::add_ivar_property_ro(std::string("initialized"), &This::_initialized);
  PBase::add_ivar_property_ro(std::string("dirty"), &This::_dirty);
  PBase::add_ivar_property_ro(std::string("in-batch"), &This::_in_batch);
  PBase::add_ivar_property_ro( "L"  , &This::_L   );
  PBase::add_ivar_property_ro( "Lp" , &This::_Lp  );
  PBase::add_ivar_property_ro( "Lz" , &This::_Lz  );
  PBase::add_ivar_property_ro( "Lmu", &This::_Lmu );
}

template<typename T>
void TemporalForwardMarkovModel<T>::calc_forward_message( )
{	

  BLITZ_INDICES;
  
  update();
    
  Matrix1 lpi( _learning && _do_vb ? _lpidash : _lpi, true );
  Matrix2 A( _learning && _do_vb ? _Adash : _A, true );
  Matrix2 S( _S, true );
  
  bool have_saliency = _S.numElements() > 0;
  
  // FIXME: support inference from index array
  //bool have_ri = ( ( _ri.extent(0) == n1 ) && ( _ri.extent(1) == n2 ) );
  
  Matrix3 & rlast = _backward_message.get_belief( true ); //leaves out backward message
  
  _lp.resize( rlast.shape() );
  int n1 = rlast.extent(0);
  int n2 = rlast.extent(1);
  
  Matrix3 lpweak( _lp, true );
  Matrix3 rlweak( rlast, true );
  
  lpweak = T(0);
  
  T p0 = T(1) / _k;

  OPENMP(for)
	for (int x = 0; x < n1; ++x)
	  for (int y = 0; y < n2; ++y)
        {
          Matrix1 lpslice( lpweak( x, y, _all ) );
          Matrix2 Aslice( A( _all, _all ) );
          Matrix1 rslicelast( rlweak( x, y, _all ) );
          if ( have_saliency )
            lpslice += safe_log( S(x,y) * sum( rslicelast( i2 ) * Aslice( i2, i1 ), i2 ) + ( 1 - S(x,y) ) * p0 );
          else
            lpslice += safe_log( sum( rslicelast( i2 ) * Aslice( i2, i1 ), i2 ) );
        }
    
}

template<typename T>
void TemporalForwardMarkovModel<T>::calc_backward_message( )
{	

  BLITZ_INDICES;
  
  update();  
  
  Matrix1 lpi( _learning && _do_vb ? _lpidash : _lpi, true );
  Matrix2 A( _learning && _do_vb ? _Adash : _A, true );
  Matrix2 S( _S, true );
  
  bool have_saliency = _S.numElements() > 0;
  
  // FIXME: support inference from index array
  //bool have_ri = ( ( _ri.extent(0) == n1 ) && ( _ri.extent(1) == n2 ) );
  
  Matrix3 & rnext = message_component<T>::get_belief( true ); //TODO: leave out backward message
  
  _lp.resize( rnext.shape() );
  int n1 = rnext.extent(0);
  int n2 = rnext.extent(1);
  
  Matrix3 lpweak( _lp, true );
  Matrix3 rnweak( rnext, true );
  
  
  lpweak = T(0);
  
  T p0 = T(1) / _k;

  OPENMP(for)
	for (int x = 0; x < n1; ++x)
	  for (int y = 0; y < n2; ++y)
        {
      	  Matrix1 lpslice( lpweak( x, y, _all ) );
    		  // this slice is backwards in space
          blitz::Array<T,1> rslicenext( rnweak( x, y, _all ) );
          if ( have_saliency )
            lpslice += safe_log( S(x,y) * sum( rslicenext( i2 ) * A( i1, i2 ), i2 ) + ( 1 - S(x,y) ) * p0 );
          else
            lpslice += safe_log( sum( rslicenext( i2 ) * A( i1, i2 ), i2 ) );
        }
  
}

template<typename T>
void TemporalForwardMarkovModel<T>::step_backward( )
{	

  calc_backward_message();
  
  int state = buffer().active_state();
  buffer().select_history( state + 1 );
   
  //std::cout << "lp_shape: " << _lp.shape() << std::endl;
  _backward_message.set_msg( _lp );
  
  if ( _use_pi_inseq )
    prior_in_buffer();
  
  stepped_backward();
}

template<typename T>
void TemporalForwardMarkovModel<T>::step_forward( )
{	

  calc_forward_message( );
  
  int state = buffer().active_state();
  bool front; 
  if ( state == 0 )
  {
    buffer().advance_history(); // this pops the oldest history item ( moving to future )
    front = true;
  }
  else
  {
    buffer().select_history( state - 1 );
    front = false;
  }

  message_component<T>::set_msg( _lp );
  
  if ( _use_pi_inseq )
    prior_in_buffer();
  
  stepped_forward(front);
}





template<typename T>
void TemporalForwardMarkovModel<T>::flush()
{
  _alpha.free();
  _lA.free();
  _lAdash.free();
  _lpi.free();
  _lpidash.free();
  _initialized = false;
}

template<typename T>
void TemporalForwardMarkovModel<T>::batchlearn_begin_epoch()
{
  update();
  if ( !_online_update )
	{
	  _wA = 0;
	  _wpi = 0;
	}
  _learning = _in_batch = true;
  _Lz = 0;
}

template<typename T>
void TemporalForwardMarkovModel<T>::apply_r( )
{

  BLITZ_INDICES;

  int active_state = buffer().active_state();
  
  std::cout << "buffer content:\n" << buffer() << std::endl;

  const Matrix3 & r = message_component<T>::get_belief( true );  
  Matrix3 rweak( r, true );
  Matrix2d wA( _wA, true );
  Matrix2 A( _do_vb ? _Adash : _A, true );
  Matrix2 S( _S, true );
  
  bool inSequence = ( active_state < buffer().get_history_size() ); 

  bool have_saliency = S.numElements() > 0;
  //if ( have_saliency ) std::cout << S << std::flush; 
  if ((!inSequence) || _use_pi_inseq )
  {
	  
    Matrix1 wpi;
	  if ( have_saliency )
		  wpi.reference( Matrix1( sum(sum(r(i2,i3,i1)*S(i2,i3),i3),i2) ) );
	  else
		  wpi.reference( Matrix1( sum(sum(r(i2,i3,i1),i3),i2) ) );
	  OPENMP(critical)
		_wpi += wpi;
    } 
  
  if (inSequence && (!_fix_transitions)) 
	{
	  buffer().select_history( active_state+1 );
	  const Matrix3 & rlast = _backward_message.get_belief( true );
	  //std::cout << rlast.shape() << std::endl;
	  buffer().select_history( active_state );
	  Matrix3 rlweak( rlast, true );
	  
	  int n1 = r.extent(0);
	  int n2 = r.extent(1);
	  for (int x = 0; x < n1; ++x)
		for (int y = 0; y < n2; ++y)
		  {
			Matrix1 rlastslice(rlweak(x,y,blitz::Range::all()));
			Matrix1 rslice( rweak( x, y, _all ) );
			Matrix2d wAslice( wA( _all, _all ) );
			if ( _factorize_transitions )
			  {
				if ( have_saliency )
				  wAslice += S(x,y) * rlastslice(i1) * rslice(i2);
				else
				  wAslice += rlastslice(i1) * rslice(i2);
			  }
			else
			  {
				
          Matrix2 wA( A.shape() );
        
          if ((rslice.extent(0) == 1) && (rslice.extent(1) == 1))
            wA = rlastslice(i1) * A(i1,i2) * rslice(i2);
          else
            wA = rlastslice(i1) * A(i1,i2) * rslice(i2);
				
        OPENMP(critical)
				  if ( have_saliency )
					_wA += S(x,y) * wA / max( T(1e-10), sum( wA( i1, i3 ), i3 ) );
				  else
					_wA += wA / max( T(1e-10), sum( wA( i1, i3 ), i3 ) );
			  }
		  }
	}
  
    if ( _owns_state )
      {
    		T Lz = 0;
        if ( have_saliency )
          Lz -= sum( S( i1, i2 ) * sum( r * safe_log( r ), i3 ) );
        else
          Lz -= sum( r * safe_log( r ) );
   		OPENMP(critical)
		  _Lz += Lz;
      };
  
  learned();

  if ( _online_update )
	{
	  recalc_hyperparams();
	  update(true);
	}

}

template<typename T>
void TemporalForwardMarkovModel<T>::recalc_hyperparams()
{
  BLITZ_INDICES;
  _alpha.resize( _k );
  _gamma.resize( _k, _k );
  _alpha = _alpha0 + _wpi;
  _gamma = where(i1 == i2,_gammadiag0,_gamma0) + _wA;
  _dirty = true;
			
  if (_learn_gamma0_interval > 0 && ++_learn_gamma0_count >= _learn_gamma0_interval)
	{
	  optimize_beta0(_delta);
	}
  
}

template<typename T>
void TemporalForwardMarkovModel<T>::batchlearn_end_epoch(Matrix1 f_in, T fout )
{
  BLITZ_INDICES;
  _Lmu = sum((where(i1==i2,_gammadiag0,_gamma0)-_gamma) * _lAdash);
  _Lmu += _k*(gammaln(_gammadiag0+(_k-1)*_gamma0) - (_k-1)*gammaln(_gamma0)-gammaln(_gammadiag0));
  _Lmu -= sum(gammaln(sum(_gamma,i2)))  - sum(gammaln(_gamma));
  if (_use_pi)
    {
	  _Lmu += sum( ( _alpha0 - _alpha ) * _lpidash );
	  _Lmu += gammaln( _k * _alpha0 )   - _k * gammaln( _alpha0 );
	  _Lmu -= gammaln( sum( _alpha ) )  - sum( gammaln( _alpha ) );
	}

  _Lp = sum( _wA * _lAdash );

  _wA *= fout;
  _wpi *= fout;

  double n_eff = sum( _wA );
  recalc_hyperparams();
  
  _learning = _in_batch = false;
  _L = _Lmu + _Lp;
  
  std::cout << str( boost::format("markov chain ended epoch Lmu = %1.3e Lp = %1.3e L = %1.3e Lz = %1.3e %1.3e effective events")
					 % _Lmu % _Lp % _L % _Lz % n_eff ) << std::endl;
}



template<typename T>
void TemporalForwardMarkovModel<T>::prior_in_buffer( )
{
  Matrix3 prior_msg( 1, 1, _k );
  if ( _use_pi )
  {
    prior_msg( 0, 0, blitz::Range::all() ) = ( ( _learning && _do_vb ) ? _lpi : _lpidash );
  }
  else
  {
    prior_msg( 0, 0, blitz::Range::all() ) = T(1) / _k;
  }
  _prior_message.set_msg( prior_msg );
}  



template<typename T>
void TemporalForwardMarkovModel<T>::reset()
{
  _lp.free();
  message_component<T>::buffer().clear(true);
  
  update();
  prior_in_buffer();
}  


template<typename T>
void TemporalForwardMarkovModel<T>::update( bool force )
{
  BLITZ_INDICES;

  initialize();

  if ( _dirty || force )
    {
      _lpi.resize(_k);
      _lpidash.resize(_k);
      _lA.resize(_k,_k);
      _lAdash.resize(_k,_k);
      _A.resize(_k,_k);
      _Adash.resize(_k,_k);

	  _lA = log( _gamma ) - log( sum( _gamma(i1,i3), i3 ) );
	  _A = _gamma / sum( _gamma(i1,i3), i3 );
	  _lpi = log( _alpha ) - log( sum( _alpha ) );
	  _lpidash = psi( _alpha ) - psi( sum( _alpha ) );
	  if ( _fix_transitions )
		_lAdash.reference( _lA );
	  else
		_lAdash = psi( _gamma ) - psi( sum( _gamma(i1,i3), i3 ) );
      _Adash = exp( _lAdash );
    }
  _dirty = false;
}


template<typename T>
double TemporalForwardMarkovModel<T>::dL_beta0 ( double la0, void * vparams )
{
  double * params = (double*)vparams;
  double k = params[0];
  double c = params[1];
  double c_diag = params[2];
  double alpha0 = std::exp(la0);
  try {
	//properly handle diagonal fattening
	return ((k-1+c_diag)*boost::math::digamma((k-1+c_diag)*alpha0) 
            - (k-1)*boost::math::digamma(alpha0) + c_diag * boost::math::digamma(c_diag*alpha0))/k
	  + c;
  }	catch(std::exception & e) {
	std::cerr << " exception evaluating dl gamma0 (tfmm2): " << e.what() << " type " << typeid(e).name() << std::endl;
	gsl_error_flag = true;
	return 0;
  }
}

template<typename T>
void TemporalForwardMarkovModel<T>::optimize_beta0 ( double delta, double lalphamin, double lalphamax )
{
  typedef std::pair<T,bool> Result;
  double c = mean(_lAdash);
  double c_diag = _gammadiag0/_gamma0;
  double params[3] = {_k,c,c_diag};
  gsl_function F = {&dL_beta0, params};	
  gsl_set_error_handler(&gsl_error_handler);	
  Result r = find_root(&F, lalphamin, lalphamax);
  bool converged = r.second;
  if (!converged) 
    {
      std::cout << "tfmm2: hyperparameter optimization did not converge" << std::endl;
      return;
    }
  double lbeta0 = safe_log(double(_gamma0));
  double lbeta0_opt = r.first;
  lbeta0 = (1-double(delta)) * lbeta0 + double(delta) * lbeta0_opt;
  _gamma0 = T(std::exp(lbeta0));
  _gammadiag0 = c_diag * _gamma0;
  std::cout << "gamma0 set to " << _gamma0 << " optimal " << std::exp(lbeta0_opt) << std::endl;
  _dirty = true;
}


template<typename T>
void TemporalForwardMarkovModel<T>::initialize( bool force )
{

  if ( _initialized && !force ) return;

  BLITZ_INDICES;

  _wpi.resize( _k );
  _wA.resize( _k, _k );
  

  if ( _init_uniform )
	{
	  T cdiag = _gammadiag0 / ( _gammadiag0 + ( _k - 1 ) * _gamma0 );
	  T coffdiag = ( 1 - cdiag ) / ( _k - 1 );
	  for ( int k1 = 0; k1 < _k; ++k1 )
			for ( int k2 = 0; k2 < _k; ++k2 )
			  _wA( k1, k2 ) = ( k1 == k2 ? cdiag : coffdiag ) * _initcount / ( _d * _d );
	  _wpi = _initcount / _k;
	}
  else // initialize hyperparams by accumulating samples from the dirichlet priors
	{
	  Random<T> & rng = Random<T>::get();
	  Matrix1 sample(_k);
	  rng.sample_dirichlet( _alpha0, sample );
	  _wpi = _initcount * sample;
	  Matrix1 sample1(_wA.extent(1), _wA.extent(2), _wA.extent(3));
	  for ( int k1 = 0; k1 < _k; ++k1 )
		{
			  for ( int k2 = 0; k2 < _k; ++k2 )
  				sample1(k2) = rng.gamma( ( k1 == k2 ? _gammadiag0 : _gamma0 ) );
		    _wA( k1, _all ) = T(_initcount) * sample1 / sum(sample1);
		}	  
	}
  std::cout << str(boost::format("markov chain initialized alpha0 = %f k = %d gamma0 = %f gammadiag0 = %f initcounts = %f mode = %s")
					%_alpha0%_k%_gamma0%_gammadiag0%_initcount%(_init_uniform ? "uniform":"random") ) << std::endl;
  std::cout << " wA " << _wA.shape() << std::endl;
  _initialized = true;
  recalc_hyperparams();
}

template class TemporalForwardMarkovModel<float>;
