/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef TFMM_HPP
#define TFMM_HPP
#undef BZ_THREADSAFE

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/foreach.hpp>
#include <boost/signal.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/binary_object.hpp>
#include <iostream>
#include "digamma.hpp"
#include "normrho.hpp"
#include "optimal_hyperparameters.hpp"
#include "SeariseProperties.hpp"
#include "blitz_properties.hpp"
#include "serialize_blitz.hpp"
#include "message_component.hpp"

#undef foreach
#define foreach BOOST_FOREACH


struct bool_default_false_combiner
{
  typedef bool result_type;

  template<typename InputIterator>
  bool operator()(InputIterator first, InputIterator last) const
  {
    // If there are no slots to call, just return the
    // default-constructed value
    if (first == last)
      return false;

    bool cur_value = *first++;
    while (first != last) {
      cur_value = cur_value || (*first);
      ++first;
    }
  
    return cur_value;
  }
};


template<class T> class TemporalForwardMarkovModel 
: public HasProperties< SearisePropertyTypes, TemporalForwardMarkovModel<T> >
, public message_component<T>
{

public:

  typedef T num_type;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<T,1> Matrix1;

  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<double,4> Matrix4d;

  typedef blitz::Array<unsigned int,2> IMatrix2;
  typedef TemporalForwardMarkovModel<T> This;
  typedef SeariseProperty PropertyType;
  typedef HasProperties< SearisePropertyTypes, This > PBase;

  TemporalForwardMarkovModel(int k =0 );
  
  static void init_properties();
  void flush();
  void batchlearn_begin_epoch();
  
  void batchlearn_end_epoch(Matrix1 f_in = Matrix1(), T fout = T(1));
  void reset();
  
  mmm2_belief_message_buffer<T> & buffer()
  {
    return message_component<T>::buffer();
  }
  
  
  void begin_mp_frame()
  {
    _backward_message.set_buffer( buffer(), "backward" );
    _prior_message.set_buffer( buffer(), "prior" );

    if ( buffer().get_history_capacity() < 1 )
      buffer().set_history_capacity( 1 );
    
    if ( buffer().get_history_size() == 0 )
      reset();
    
	if ( _learning && _online_update )
	  decay_statistics();
  }

  void decay_statistics()
  {
	_wA  *= std::exp( - 1.0 / _tau );	  
	_wpi *= std::exp( - 1.0 / _tau );	  
  }
  
  void end_mp_frame( bool slave=false )
  {
    int history_size = buffer().get_history_size();
    int history_capacity = buffer().get_history_capacity();

    if (! slave)
    {
      bool at_prior = history_size == history_capacity - 1;
      bool in_sequence = history_size == history_capacity;
      if ( _learning && ( in_sequence || at_prior ) )
      {
        if ( in_sequence )
        {
          for ( int i = 0; i < ( history_size - 1 ); ++i )
            step_backward();
        }
        else // if ( at_prior )
        {
          for ( int i = 0; i < history_size; ++i )
            step_backward();
        }
        apply_r();
        while ( buffer().active_state() > 0 )
          step_forward();
      }
      step_forward();
    }
    
    _S.free();
  }

  void initialize( bool force = false );
  void update( bool force = false );
  void apply_r( );
  
  const Matrix2 & get_transition_matrix() const { return _A; }
  const Matrix2 & get_transition_posteriori() const { return _gamma; }
  const Matrix1 & get_pi() const { return _alpha; }

  void set_saliency( const Matrix2 & S )
  {
    _S.reference( S );
  }
  
  typedef boost::signal<bool (bool), bool_default_false_combiner> fwd_step_signal_t;
  typedef boost::signal<bool (), bool_default_false_combiner> bck_step_signal_t;
  typedef boost::signal<Matrix2 ()> Matrix2_signal_t;

  bck_step_signal_t stepped_backward;
  fwd_step_signal_t stepped_forward;
  void step_backward( );
  void step_forward(  );

  Matrix2_signal_t learned;
  

private:

  void calc_forward_message( );
  void calc_backward_message( );
  void recalc_hyperparams( );
  void prior_in_buffer( );

  void set_history_size( PropertyType p );
  PropertyType get_history_size( );
  static double dL_beta0 ( double la0, void * vparams );
  void optimize_beta0 ( double delta, double lalphamin=-20, double lalphamax=20);
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
	BLITZ_INDICES;
	ar & BOOST_SERIALIZATION_NVP(_k);
	ar & BOOST_SERIALIZATION_NVP(_alpha0);
	ar & BOOST_SERIALIZATION_NVP(_gamma0);  
	ar & BOOST_SERIALIZATION_NVP(_gammadiag0);
	ar & BOOST_SERIALIZATION_NVP(_initialized);
	if (_initialized)
	  {
		ar & BOOST_SERIALIZATION_NVP(_alpha);   
    
    if ( (version <= 2) || (version >= 5) )
		  {
	  		ar & BOOST_SERIALIZATION_NVP(_wA);
		  	ar & BOOST_SERIALIZATION_NVP(_gamma);
		  }
		else
		  { // backward compatibility for the idea of diffusive predictions...
        int _d;
	  		Matrix4 gamma;
	  		Matrix4d wA;
		  	ar & boost::serialization::make_nvp( "_wA", wA );
    		ar & BOOST_SERIALIZATION_NVP(_d);
		  	ar & boost::serialization::make_nvp( "_gamma", gamma );
        _gamma.reference( gamma( _all, _d, _d, _all ).copy() );
        _wA.reference( wA( _all, _d, _d, _all ).copy() );
		  }
	  }
    if ( version > 3 )
	  {
		ar & BOOST_SERIALIZATION_NVP(_wpi);		
	  }
	else
	  {
		_wpi.resize( _alpha.shape() );
		_wpi = _alpha - _alpha0;
	  }
    if ( version > 0 )
        ar & BOOST_SERIALIZATION_NVP(_owns_state);
	if ( version > 1 )
	  {
        ar & BOOST_SERIALIZATION_NVP(_tau);		
	  }
	_dirty = true;
  }

  float _initcount;
  int _k, _learn_gamma0_interval, _learn_gamma0_count, _d;
  bool _initialized, _dirty, _in_batch, _use_pi, _use_pi_inseq;
  bool _fix_transitions, _init_uniform, _factorize_transitions;
  bool _learning, _do_vb, _owns_state, _normalize_whole_transition_matrix, _online_update;

  T _alpha0, _gamma0, _gammadiag0;
  T _delta, _Lmu, _Lp, _L, _Lz, _tau;

  Matrix2 _S;

  Matrix2 _gamma, _lA, _lAdash, _A, _Adash, _Ar, _Adashr;
  Matrix2d _wA;

  Matrix1 _alpha, _lpi, _lpidash;
  Matrix1d _wpi;
		
  Matrix3 _lp;
  IMatrix2 _ri;

  message_component<T> _backward_message;
  message_component<T> _prior_message;
};

//templated class versioning
namespace boost { namespace serialization {
	template<typename T>
	struct version< TemporalForwardMarkovModel<T> >
	{
	  BOOST_STATIC_CONSTANT(unsigned int, value =  5);
	};
  }}

#endif
