/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "tmm2.hpp"
#include "blitz_indices.hpp"
#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include "enum_properties.hpp"

#ifdef _WIN32
#define throw_not_implemented												\
  throw std::runtime_error( str( boost::format( "%s(%i): not implented for tmm2" ) \
								 % __FILE__ % __LINE__ ) )
#else
#define throw_not_implemented												\
  throw std::runtime_error( str( boost::format( "%s(%i): method \"%s\" not implented for tmm2" ) \
								 % __FILE__ % __LINE__ % __func__ ) )
#endif

/// k is the output space cardinality
/// d is the receptive field size
/// m is the input space cardinality
/// step is the stepping between neighboring receptive fields
/// for the mcmm2 probably d and step need to be 1
template< typename T >
TransitionMixtureModel2<T>::TransitionMixtureModel2(  )
  : _beta0(0.5)
  , _pre_clustering_mode(NONE)
  , _normalize_transitions(false)

{
}
 
/// this is where to register all parameters accessible in json files
template< typename T >
void TransitionMixtureModel2<T>::init_properties()
{
  typename PBase::SideEffect uninitializer = PBase::make_ivar_setter( &This::_initialized, false );
  PBase::add_property(         "mu",           make_blitz_ivar_getter1( &This::_mu ) );
  PBase::add_property(         "beta",         make_blitz_ivar_getter1( &This::_beta ) );
  PBase::add_property(         "wxy",          make_blitz_ivar_getter1( &This::_wxy ) );
  
  PBase::add_ivar_property_ro( "Lmu",         &This::_Lmu);
  PBase::add_ivar_property_rw( "beta0",       &This::_beta0,  uninitializer); 
  PBase::add_ivar_property_rw( "normalize-transitions", &This::_normalize_transitions,  uninitializer); 

  enum_property_access<SearisePropertyTypes, This, PreClusteringMode> pre_clustering_accessor( &This::_pre_clustering_mode );
  pre_clustering_accessor
	.add_choice( "none"     , NONE     )
	.add_choice( "sequence" , SEQUENCE );
  PBase::add_property( "pre_clustering", pre_clustering_accessor, pre_clustering_accessor );

  PBase::add_property( std::string("sequence_clusterer"), PBase::make_pivar_getter( &This::_seq_clusterer ) );

}

/// upon a call to this the model should forget all it learned and start from scratch
template< typename T >
void TransitionMixtureModel2<T>::flush()
{
  throw_not_implemented;
}

/// not currently used, but here to support pruning
template< typename T >
void TransitionMixtureModel2<T>::reset_input_dimensions(const std::set<unsigned int> & components)
{
  throw_not_implemented;
}

/// not currently used, but here to support pruning
template< typename T >
void TransitionMixtureModel2<T>::reset_components(const std::set<unsigned int> & components)
{
  throw_not_implemented;
}

template<typename T>
void TransitionMixtureModel2<T>::break_sequence()
{
  if ( !ready() && _pre_clustering_mode == SEQUENCE )
	_seq_clusterer.break_sequence();
  _rlast.free();
}

/// as the name says: notifies begin of a batch epoch
template< typename T >
void TransitionMixtureModel2<T>::batchlearn_begin_epoch()
{
  if ( !ready() && _pre_clustering_mode == SEQUENCE )
	{
	  _seq_clusterer.init( Base::_k * Base::_k );
	  Base::_learning = true;
	}
  else
	{
	  Base::batchlearn_begin_epoch();
	  _wxy.resize(Base::_k,Base::_m,Base::_m); 
	  _wxy = 0;  
	}
}

///////////////////////////////////////////////////////////////////////////////////
///// messages: these define the model, they are made up of the conditional //////
///// probabilities of each set of variables given others                  //////
////////////////////////////////////////////////////////////////////////////////

template< typename T >
blitz::Array<typename TransitionMixtureModel2<T>::num_type,3> TransitionMixtureModel2<T>::calc_upward_msg( IMatrix2 & data )
{
	throw_not_implemented;
}


template< typename T >
blitz::Array<typename TransitionMixtureModel2<T>::num_type,3> TransitionMixtureModel2<T>::calc_upward_msg( blitz::Array<T,3> & data )
{
  BLITZ_INDICES;
  
  update();

  Matrix3 msg(data.extent(0), data.extent(1), Base::_k);
  Matrix3 mu( Base::_learning && Base::_do_vb ? _mu : _mudash );

  if ( _rlast.numElements() > 0 )
	msg = safe_log( sum( sum( _rlast(i1,i2,i4) * mu(i3,i4,i5) * data(i1,i2,i5), i5 ), i4 ) );
  else
	msg = safe_log( T(1) / T(Base::_m) * sum( sum( mu(i3,i4,i5) * data(i1,i2,i5), i5 ), i4 ) );

  return msg;
}


/// here the model should calculate the probability of the latent (output)
/// variable given data (usually probabilities of the input variables)
/// the data is an input receptive field view: (rows x columns x input-depth)
/// the output is a single output "pixel"
/// the output neds to be stored logarithmically
/// variational is for MAP estimation (different parameter set)
template< typename T >
void TransitionMixtureModel2<T>::begin_mp_frame()
{
  Base::begin_mp_frame();  
}


template< typename T >
void TransitionMixtureModel2<T>::end_mp_frame( const Matrix3 & data_ )
{
  Base::end_mp_frame( data_ );
  _rlast.reference( data_.copy() );  
}

template< typename T >
blitz::Array<typename TransitionMixtureModel2<T>::num_type,3> TransitionMixtureModel2<T>::calc_downward_msg ( Matrix3 data, bool logarithmic )
{
  BLITZ_INDICES;
  Matrix3 & rout = message_component<T>::get_belief(true);
  Matrix3 mu( Base::_learning && Base::_do_vb ? _mu : _mudash );
  Matrix3 out_msg(rout.extent(0), rout.extent(1), Base::_m );
  if ( logarithmic )
	if ( _rlast.numElements() > 0 )
	  out_msg = safe_log( sum( sum( _rlast(i1,i2,i4) * mu(i5,i4,i3) * rout(i1,i2,i5), i5 ), i4 ) );
	else
	  out_msg = safe_log( ( T(1) / T(Base::_m) ) * sum( sum( mu(i5,i4,i3) * rout(i1,i2,i5), i5 ), i4 ) );	  
  else
	if ( _rlast.numElements() > 0 )
	  out_msg = sum( sum( _rlast(i1,i2,i4) * mu(i5,i4,i3) * rout(i1,i2,i5), i5 ), i4 );
	else
	  out_msg = ( T(1) / T(Base::_m) ) * sum( sum( mu(i5,i4,i3) * rout(i1,i2,i5), i5 ), i4 );
  return out_msg;
}

/// here, given output and input distribution you should calculate
/// the probability of this configuration given the model and store
/// it in pin in linear space
/// not strictliy neccesary for operation
template< typename T >
void TransitionMixtureModel2<T>::confidence ( const blitz::Array<T,1> & rout, blitz::Array<T,3> & rin, blitz::Array<T,2> & pin )
{
  throw_not_implemented;
}

/// as above, with weights
template< typename T >
void TransitionMixtureModel2<T>::confidence ( const blitz::Array<T,1> & rout, blitz::Array<T,3> & rin, blitz::Array<T,2> & pin, const blitz::Array<T,2> & w )
{
  throw_not_implemented;
}

/// as above, with indexed input and output
template< typename T >
void TransitionMixtureModel2<T>::confidence ( int iout, blitz::Array<unsigned,2> & iin,blitz::Array<T,2> & p )
{
  throw_not_implemented;
}

/// as above, with indexed input and output and weights
template< typename T >
void TransitionMixtureModel2<T>::confidence ( int iout, blitz::Array<unsigned,2> & iin,blitz::Array<T,2> & p, const blitz::Array<T,2> & w )
{
  throw_not_implemented;
}

template< typename T >
void TransitionMixtureModel2<T>::recalc_likelihoods()
{
  BLITZ_INDICES;

  Base::_Lp = sum( _wxy * log( _mudash ) );
  T lCbeta0 = Base::_k * Base::_m * ( gammaln( Base::_m *_beta0 ) - Base::_m * gammaln( _beta0 ) );
  T lCbeta  = sum( gammaln( sum( _beta(i1,i3,i2), i3 ) ) ) - sum( gammaln( _beta ) );
  
  T Elp = sum( ( _beta0 - 1 ) * log( _mudash ) ) + lCbeta0;
  T Elq = sum( ( _beta  - 1 ) * log( _mudash ) ) + lCbeta;
  
  _Lmu = Elp - Elq;	  
  
  Base::_L = Base::_Lz + _Lmu + Base::_Lp;
  
  log().info << str( boost::format("transition component model ended epoch Lmu = %1.3e Lp = %1.3e L = %1.3e")
					 % _Lmu % Base::_Lp % Base::_L ) << std::endl;
  
  
}

template< typename T >
void TransitionMixtureModel2<T>::recalc_hyperparams()
{
  _beta.reference( Matrix3d( _beta0 + Base::_delta * _wxy ) );
  Base::dirty();  
}

/// this gets called at the end of a e-step epoch
/// f_in is a reweighing factor for the input space
/// and should be applied to any histograms
/// similarly you can return a renormalization factor
/// for the output space which will be propagated
template< typename T >
blitz::Array<typename TransitionMixtureModel2<T>::num_type,1> TransitionMixtureModel2<T>::batchlearn_end_epoch(blitz::Array<T,1> f_in )
{

  BLITZ_INDICES;

  Matrix1 f;

    if ( !ready() && _pre_clustering_mode == SEQUENCE )
	{
		  std::cout << " applying sequence clustering " << std::endl;
		  _seq_clusterer.end_batch();
		  _seq_clusterer.cluster_AHC( Base::_k );
		  
		  typedef typename heuristic_sequence_clusterer<T>::state_groups state_groups;
		  typedef typename heuristic_sequence_clusterer<T>::state_group state_group;

		  const state_groups & groups = _seq_clusterer.groups();

		  Base::_k = groups.size();
		  Base::_d1 = Base::_d2 = 1;
		  Base::_wk.resize( Base::_k ) ; Base::_wk  = 0;
		  Base::_wnx.resize( Base::_k, 1, 1 ); Base::_wnx = 0;
		  _wxy.resize( Base::_k, Base::_m, Base::_m ); _wxy = 0;
		  
		  size_t igroup = 0;
		  const Matrix1d & abundances  = _seq_clusterer.abundances();
		  BOOST_FOREACH( const state_group & group, groups )
			{
			  BOOST_FOREACH( size_t state, group )
				{
				  Base::_wk( igroup ) += abundances( int(state) );
				  Base::_wnx( igroup, _all, _all ) += abundances( int(state) );
				  int state_from = state / Base::_m;
				  int state_to   = state % Base::_m;
				  _wxy ( int(igroup), state_from, state_to ) += abundances( int(state) );
				}
			  ++igroup;
			}
		  
	  recalc_hyperparams();
	  Base::recalc_hyperparams();
	  Base::_initialized = true;
	  update();
	  recalc_likelihoods();

	}
	else
	  {
		
		f.reference( Base::batchlearn_end_epoch() );
		
		if (Base::_renormalize_counts)
		  {
			_wxy = f(i1) * _wxy;
		  }
		
		if (f_in.numElements() > 0)
		  {
			_wxy = f_in(i2) * f_in(i3) * _wxy;
		  }
		

		recalc_hyperparams();
		
		CHECKNAN(_beta,"nan in beta after epoch");
				
		update();
  
	  }

  return f;
  
}

template< typename T >
void TransitionMixtureModel2<T>::apply_r( const Matrix3 & data )
{
	//std::cout << "apply_r: mean = " << mean( data ) << " std = " << std::sqrt( mean( pow2( data - mean( data ) ) ) ) << std::endl;
	
	BLITZ_INDICES;

	if ( _pre_clustering_mode == SEQUENCE && !ready() )
	  {
		_seq_clusterer.collect_statistics( _rlast, data );
		return;
	  }

	Base::apply_r( data, false );
	if ( _rlast.numElements() > 0 )
	{
		  {
			Matrix3 & r = message_component<T>::get_belief(true);
			for ( int i = 0; i < r.extent(0); ++i )
			  for ( int j = 0; j < r.extent(1); ++j )
				{
				  Matrix1 rlast1( _rlast( i, j, _all ) );
				  Matrix1 r_in1( data( i, j, _all ) );
				  Matrix1 r1( r( i, j, _all ) );
				  Matrix3d wxy( r1(i1) * rlast1(i2) * r_in1(i3) * _mudash );
				  double Z = sum( wxy );
				  
				  _wxy += wxy / std::max( 0.000000001, Z );
				}
		  }
	}
}

/// this is to initialize all (hyper-) parameters
/// called automatically by base
/// you should call Base::initlialize from here also
/// force means re-initialize, even you already did it before
template< typename T >
void TransitionMixtureModel2<T>::initialize( bool force )
{

  if ( !ready() ) return;

  BLITZ_INDICES;

  if ( Base::initialize(force) )
	{
	  _beta.resize(Base::_k,Base::_m,Base::_m);

	  //draw a few random samples from the prior
	  _beta = _beta0;
	  
	  Random<T> & rng = Random<T>::get();
	  Matrix3 sample( _beta.shape() );
	  BOOST_FOREACH(T & x, sample) 
	    x = rng.gamma( _beta0 );
	  if ( _normalize_transitions )
		  _beta += Base::_initcount * sample / max( T(0.00001), sum( sum( sample(i1,i4,i5), i5), i4 ) );
	  else
		  _beta += Base::_initcount * sample / max( T(0.00001), sum( sample( i1, i2, i4 ), i4 ) ); 
	  
	  log().info << str(boost::format("transition component model initialized beta0 = %f")%_beta0) << std::endl;
	}
}

/// after this call parameters should be ready for operation
/// you should call initialize if you need to from here
/// also you should call Base::update
/// force as above
template< typename T >
void TransitionMixtureModel2<T>::update(bool force )
{

  if ( !ready() ) return;

  BLITZ_INDICES;
  initialize();
  if ( Base::update( force ) )
	{
	  _mu.resize( Base::_k, Base::_m, Base::_m );
	  _mudash.resize( Base::_k, Base::_m, Base::_m );
	  if ( _normalize_transitions )
	  {
		  _mu = _beta / sum( sum( _beta( i1, i4, i5 ), i5 ), i4 );
		  _mudash = exp( psi( _beta ) - psi( sum( sum( _beta( i1, i4, i5 ), i5 ), i4 ) ) );
	  }
	  else
	  {
		  _mu = _beta / sum( _beta( i1, i2, i4 ), i4 );
		  _mudash = exp( psi( _beta ) - psi( sum( _beta( i1, i2, i4 ), i4 ) ) );
	  }
	}
}


/// explicit instatiation of the mcmm2 for float data
template class TransitionMixtureModel2<float>;


