/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef TMM2_HPP
#define TMM2_HPP

#include <set>
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include <boost/serialization/serialization.hpp>
#include <iostream>

#include "MixtureModel2Base.hpp"
#include "serialize_blitz.hpp"
#include "heuristic_sequence_clustering.hpp"

template<typename T> class TransitionMixtureModel2;
template<typename T> struct TypeTraits< TransitionMixtureModel2<T> > 
{	
  typedef T num_type;
  static const int num_statistics = 1;
};

template<typename T> class TransitionMixtureModel2 
  : public MixtureModel2Base< TransitionMixtureModel2<T> >,
	public HasProperties< SearisePropertyTypes, TransitionMixtureModel2<T> >
{
  
  typedef TransitionMixtureModel2<T> This;
  typedef MixtureModel2Base< This > Base;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  
public:
  
  typedef T num_type;
  
  typedef blitz::Array<T,1> Matrix1;
  typedef blitz::Array<double,1> Matrix1d;
  typedef blitz::Array<T,2> Matrix2;
  typedef blitz::Array<double,2> Matrix2d;
  typedef blitz::Array<T,3> Matrix3;
  typedef blitz::Array<double,3> Matrix3d;
  typedef blitz::Array<T,4> Matrix4;
  typedef blitz::Array<unsigned int,2> IMatrix2;


  /// k is the output space cardinality
  /// d is the receptive field size
  /// m is the input space cardinality
  /// step is the stepping between neighboring receptive fields
  /// for the mcmm2 probably d and step need to be 1
  TransitionMixtureModel2();
  
  /// this is where to register all parameters accessible in json files
  static void init_properties();
  
  /// upon a call to this the model should forget all it learned and start from scratch
  void flush();

  /// not currently used, but here to support pruning
  void reset_input_dimensions(const std::set<unsigned int> & components);
  
  /// not currently used, but here to support pruning
  void reset_components(const std::set<unsigned int> & components);
  
  /// as the name says: notifies begin of a batch epoch
  void batchlearn_begin_epoch();
  
  ///////////////////////////////////////////////////////////////////////////////////
  ///// messages: these define the model, they are made up of the conditional //////
  ///// probabilities of each set of variables given others                  //////
  ////////////////////////////////////////////////////////////////////////////////

  /// this if for models allowing overlap, so probably not for this
  void segmentation_message2 ( const Matrix1 & rout, const IMatrix2 & data, Matrix2 & lp ) {}
  
  /// this if for models allowing overlap, so probably not for this
  void segmentation_message2( const Matrix1 & rout, const Matrix3 & data, Matrix2 & lp ) {}

  /// override the complete upward message logic because we process full images at a time here
  blitz::Array<T,3> calc_upward_msg( blitz::Array<T,3> & data );
  
  /// override the complete upward message logic because we process full images at a time here
  blitz::Array<T,3> calc_upward_msg( IMatrix2 & data );


  void apply_r( const Matrix3 & data );

  /// override the complete downward message logic because we process full images at a time here
  Matrix3 calc_downward_msg ( Matrix3 data = Matrix3(), bool logarithmic = true );

  /// here, given output and input distribution you should calculate
  /// the probability of this configuration given the model and store
  /// it in pin in linear space
  /// not strictliy neccesary for operation
  void confidence ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin );

  /// as above, with weights
  void confidence ( const Matrix1 & rout, Matrix3 & rin, Matrix2 & pin, const Matrix2 & w );

  /// as above, with indexed input and output
  void confidence ( int iout, IMatrix2 & iin,Matrix2 & p );

  /// as above, with indexed input and output and weights
  void confidence ( int iout, IMatrix2 & iin,Matrix2 & p, const Matrix2 & w );
  
  
  /// inter-frame callbacks
  void begin_mp_frame();
  void end_mp_frame( const Matrix3 & data );

  /// this gets called at the end of a e-step epoch
  /// f_in is a reweighing factor for the input space
  /// and should be applied to any histograms
  /// similarly you can return a renormalization factor
  /// for the output space which will be propagated
  Matrix1 batchlearn_end_epoch(Matrix1 f_in = Matrix1());

  /// this is to initialize all (hyper-) parameters
  /// called automatically by base
  /// you should call Base::initlialize from here also
  /// force means re-initialize, even you already did it before
  void initialize(bool force=false);
  
  /// after this call parameters should be ready for operation
  /// you should call initialize if you need to from here
  /// also you should call Base::update
  /// force as above
  void update(bool force = false);
  
  /// leave this: means that your input space is multinomial
  bool is_multinomial() { return true; }

  void break_sequence();
  
  bool ready() { return Base::_initialized || _pre_clustering_mode == NONE; }

  void recalc_hyperparams();

  void recalc_likelihoods(); 

private:

   /// here you can use boost::serialization to save/load your state
  /// this gets called from above when saving/loading all network state
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    BLITZ_INDICES;
	 ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(Base);
    ar & BOOST_SERIALIZATION_NVP(_beta0);
	 ar & BOOST_SERIALIZATION_NVP(_normalize_transitions);
	 ar & BOOST_SERIALIZATION_NVP(_Lmu);
    if (Base::_initialized)
		 ar & BOOST_SERIALIZATION_NVP(_beta);
	 Base::dirty();
  }


  T _beta0, _Lmu;
  Matrix3d _wxy, _beta;
  Matrix3 _mu, _mudash, _rlast;
  
  heuristic_sequence_clusterer<T>  _seq_clusterer;
  typedef enum { NONE, BATCHNG, GREEDY, SEQUENCE } PreClusteringMode;
  PreClusteringMode _pre_clustering_mode;

  bool _normalize_transitions;
  
};

#endif
