/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

#include <blitz/array.h>
#include <vector>
#include "normrho.hpp"
#include "digamma.hpp"

template<typename T>
class tsb1
{  

public:

  tsb( int T, int alpha0 )
	: _T(T)
	, _alpha0(alpha0)
	, _s10(1)
	, _s20(1/alpha0)
	, _gamma1( T )
	, _gamma2( T )
	, _wq(T)
	, _pi(T)
	, _pidash(T)
  {
	reset_statistics();
	update_distribs();
  }

  void reset_statistics()
  {
	_wq = 0;
  }

  void increment_statistics( const blitz::Array<T,1> & q )
  {
	_wq += q;
  }

  void update_gamma()
  {
	_gamma1 = T(1) + _wq;
	_gamma2 = _alpha0;
	// ascending cumsum
	T cum = 0;
	std::vector<int> order;
	compute_order( order.begin(), order.end(), _wq.data() );
	for ( int i = 0; i < _T; ++i )
	  {
		int j = order[_T - i - 1];
		_gamma2[j] = _alpha0 + cum;
		cum += _wq(j);	
	  }
  }

  void update_alpha0()
  {
	T w1 = _s10 + _T - 1;
	blitz::Array<T,1> ElogV2( psi( _gamma2 ) - psi( _gamma1 + _gamma2 ) );
	T w2 = _s20 - ( ElogV2.sum() - ElogV2.min() );
	_alpha0 = w1 / w2;
  }

  void update_pi()
  {
	blitz::Array<T,1> beta( exp( safe_log( _gamma1 ) - safe_log( _gamma1 + _gamma2 ) ) );
	// descending cumprod
	T cum = 1;
	std::vector<int> order;
	compute_order( order.begin(), order.end(), _wq.data() );
	for ( int i = 0; i < _T; ++i )
	  {
		int j = order[i];
		_pi[j] = cum * beta(j);
		cum *= beta(j);
	  }
	normalize_rho( _pidash );	
  }

  void update_pidash()
  {
	blitz::Array<T,1> ElogV1( psi( _gamma1 ) - psi( _gamma1 + _gamma2 ) );
	blitz::Array<T,1> ElogV2( psi( _gamma2 ) - psi( _gamma1 + _gamma2 ) );
	// descending cumsum
	T cum = 0;
	std::vector<int> order;
	compute_order( order.begin(), order.end(), _wq.data() );
	for ( int i = 0; i < _T; ++i )
	  {
		int j = order[i];
		_pidash[j] = ElogV1(j) + cum;
		cum += ElogV2(j);
	  }
	normalize_rho( _pidash );
  }

  const blitz::Array<T,1> & pi() const { return _pi; }
  const blitz::Array<T,1> & pidash() const { return _pidash; }

  void update_distribs()
  {
	update_gamma();
	update_alpha();
	update_gamma();
	update_pi();
	update_pidash();
  }

private:

  int _T;
  T _alpha0, _s10, _s20;
  blitz::Array<T,1> _gamma1, _gamma2, _wq, _pi, _pidash;

};
