/*

SHAPELEARNING

  THIS MODULE PROVIDES A FRAMEWORK FOR CONVOLUTIONAL BAYESIAN NETWORKS
  BY PROVIDING MODULES TO BUILD THE LAYERS, THE COMPOSITION NEEDS TO
  BE HANDLED BY THE USER

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#include "Random.hpp"
#include "masking.cc"
#include <boost/fusion/support/void.hpp>
#include <boost/function.hpp>
#include <boost/format.hpp>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/min.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/accumulators/statistics/median.hpp>
#include <boost/accumulators/statistics/max.hpp>
#include <vector>
#include <map>
#include <cmath>
#include <fstream>
#include "blitz_indices.hpp"

//#include <boost/svg_plot/svg_2d_plot.hpp>

typedef float num_type;

typedef boost::function< blitz::Array<num_type,1> ( blitz::Array<num_type,1> ) > distrib_function_t;

blitz::Array<num_type,1> draw_mu( size_t K, num_type alpha, num_type beta )
{
  Random<num_type>& rng = Random<num_type>::get();
  blitz::Array<num_type,1> mu( K );
  for ( int k = 0; k < K; ++k )
	mu(k) = rng.beta( alpha, beta );
  return mu;
}

blitz::Array<num_type,1> calc_gating_distrib_slow( blitz::Array<num_type,1> & mu )
{
  int K = mu.extent(0);
  blitz::Array<num_type,1> pN( K + 1 ), pg(K);
  num_type Z = gating_distrib_slow( mu, pN, pg );
  pg /= Z;
  return pg;
}

blitz::Array<num_type,1> calc_gating_distrib_new( blitz::Array<num_type,1> & mu )
{
  int K = mu.extent(0);
  blitz::Array<num_type,1> pg(K);
  num_type Z = gating_distrib_new( mu, pg );
  pg /= Z;
  return pg;
}

blitz::Array<num_type,1> calc_gating_distrib_approx1( blitz::Array<num_type,1> & mu )
{
  return blitz::Array<num_type,1>( mu / sum( mu ) );
}

typedef blitz::Array<num_type,1> datavec;

blitz::Array<num_type,1> calc_gating_distrib_approx2( blitz::Array<num_type,1> & mu )
{
  BLITZ_INDICES;
  num_type Z = sum( mu );
  datavec p( mu(i1) / ( 1 + max( num_type(0), Z - mu(i1) ) ) );
  p /= sum(p);
  return p;
}

num_type kl( const datavec & p, const datavec & q )
{
  return sum( where( p > 0, p * ( log( p ) - log( max( num_type( 1e-10 ), q ) ) ), num_type( 0 ) ) );
}

std::vector< std::vector<num_type> >
run_test( int K, int N, num_type alpha, num_type beta,
		  distrib_function_t f1,
		  std::vector< distrib_function_t > f2
		  )
{
  int n10 = N / 10;
  std::vector< std::vector<num_type> > errors( f2.size() );
  for ( int i = 0; i < N; ++i )
	{
	  blitz::Array<num_type,1> mu = draw_mu( K, alpha, beta );
	  mu = max( mu, 0.0001f );
	  blitz::Array<num_type,1> pg0 = f1( mu );
	  for ( size_t j = 0; j < f2.size(); ++j )
		{
		  blitz::Array<num_type,1> pg1 = f2[j]( mu );
		  num_type error = sum( abs( pg0 - pg1 ) );
		  //num_type error = kl( pg0, pg1 );
		  errors[j].push_back( error );
		}
	  if ( i > 0 && i % n10 == 0 )
		std::cout << "." << std::flush;
	}
  std::cout << std::endl;
  return errors;
}

void dump_csv( std::map< int, std::vector< std::vector<num_type> > > & data, const std::string filename )
{
  std::ofstream ofile( filename.c_str() );
  for( std::map< int, std::vector< std::vector<num_type> > >::iterator j = data.begin(), e = data.end();
	   j != e; ++j )
	{
	  std::vector< std::vector<num_type> > jdata = j->second;
	  for ( size_t k = 0; k < jdata.size(); ++k )
		{
		  for ( size_t i = 0; i < jdata[k].size(); ++i )
			{
			  if ( i > 0 ) ofile << ";";
			  ofile << jdata[k][i];
			}
		  ofile << std::endl;
		}
	}
}

int main( int argc, const char ** argv )
{
  int n_samples = 1000;
  num_type alpha = 1, beta = 1;
  if ( argc > 1 ) n_samples = atoi( argv[1] );
  if ( argc > 2 ) alpha     = atof( argv[2] );
  if ( argc > 3 ) beta      = atof( argv[3] );
  std::cout << str( boost::format( " running tests with alpha = %f beta = %f for %i samples " ) % alpha % beta % n_samples ) << std::endl;
  distrib_function_t exact( calc_gating_distrib_slow );
  std::vector< distrib_function_t > approximations;

  approximations.push_back( calc_gating_distrib_approx1 );
  //approximations.push_back( calc_gating_distrib_approx2 );
  //approximations.push_back( calc_gating_distrib_new );

  std::map< int, std::vector< std::vector<num_type> > > results;
  int vK[] = { 5, 10, 20, 50 };
  int nK = sizeof(vK) / sizeof(vK[0]);
  for ( int i = 0; i < nK; ++i )
	{
	  int K = vK[i];
	  std::cout << str( boost::format( " K = %4i : " ) % K ) << std::flush;
	  results[K] = run_test( K, n_samples, alpha, beta, exact, approximations );
	}
  std::string param_str = str( boost::format( "alpha_%1.2f_beta_%1.2f" ) % alpha % beta );
  dump_csv( results, param_str + ".csv" );
  return 0;
}
