/*
Copyright INRIA 2011

Authors: Neil D.B. Bruce <Neil.Bruce@inria.fr>
Contact: Pierre Kornprobst <Pierre.Kornprobst@inria.fr>

This software is a computer program whose purpose is to estimate saliency.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#include "Fusion.h"
#include <cstdio>
#include <limits>
#include <fstream>
#ifdef HAVE_HDF5
#include <HDF5FloatFile.h>
#endif

using namespace blitz;

Fusion::Fusion(int w_in, int h_in, int feats_in, int type_in, int scale_in) // Still need to pass in weighting factor and decay factor
{
  init(w_in,h_in,feats_in,type_in,scale_in);
}

Fusion::Fusion()
{
}


void Fusion::init(int w_in, int h_in, int feats_in, int type_in, int scale_in)
{

  currentData.resize(w_in,h_in,feats_in); // Holds the current feature responses for the frame under consideration
  combinedFusion.resize(w_in,h_in); // Sum of marginal Fusion across features
  currentSlice.resize(w_in,h_in);

  minVals.resize(feats_in);
  maxVals.resize(feats_in);
  weights.resize(feats_in);
  scalingType = scale_in; // Generally this should be 2, but might be made a free parameter for "special circumstances" - if a channel is almost silent and this interval is used for learning, this can create noise
  w=w_in;
  h=h_in;
  f=feats_in;
  type = type_in;
  minVals = 10e10;
  maxVals = -10e10;
  rangeDefined = 0;
  weights = 1.0; // A priori equal weighting unless set
}

// Currently the functionality for loading predefined ranges is not there, it is unclear if this is needed


Fusion::~Fusion() {
    ;
}


void Fusion::setScaleType(int _in_type)
{
  scalingType = _in_type;
}

void Fusion::setWeights(blitz::Array<float,1> newWeights)
{
  weights.reference(newWeights);
}


CPUData<float> Fusion::Fuse(CPUData<float> newData)
{

  // Copy the new representation to the current one
  currentData = Array<float,3>(newData.data(), shape(w,h,f), neverDeleteData);


  // These values store intermediate results
  blitz::Array<float,2>  leftArgument;
  blitz::Array<float,2>  rightArgument;
  leftArgument.resize(w,h);
  rightArgument.resize(w,h);


  if (!scalingType == 0)
    scaleCurrentData(); // 0 - no scaling, 1 - saliency maps normalized to [0,1] individually, 2 - normalization based on values across all feature maps globally

  if (type == 1)
    {// sort the data
      sortedData.resize(w,h,f);
      for (int ii=0; ii<w; ii++)
	for (int jj=0; jj<h; jj++)
	  {
	    // Create a temporary vector for the current location
	    float *tempvec;
	    tempvec = (float *)malloc(f * sizeof(float));
	    tempvec = currentData(ii,jj,Range::all()).data();
	    std::sort(tempvec,tempvec+f,std::greater<float>());
	    sortedData(ii,jj,Range::all()) = Array<float,1>(tempvec, shape(f), neverDeleteData);    
	  }
    }

  combinedFusion = currentData(Range::all(),Range::all(),0);

  if (type < 2)
    {//weight the first feature map
      combinedFusion = combinedFusion*weights(0);
    }

  if (type == 10)
    {
      // Contrast adjustment
      combinedFusion = blitz::pow(combinedFusion,weights(0));
    }

  for (int i=1; i<f; i++) { // starting from the second feature map, do whatever is needed in order
      currentSlice = Array<float,2>(currentData(Range::all(),Range::all(),i));

      switch (type) {

	case 0:
	  // Weighted Average - deferred to case 1 as only difference is sort
	case 1: 
	  // Ordered Weighted Average
	  combinedFusion = combinedFusion+currentSlice*weights(i);
	break;
	case 2:
	  // Yager norm
	  leftArgument   = blitz::pow(1-combinedFusion,weights(0));
	  rightArgument  = blitz::pow(1-currentSlice,weights(0));
	  combinedFusion = blitz::max(1-blitz::pow(leftArgument+rightArgument,(float)(1.0/weights(0))),0.0);
	break;
	case 3:
	  // Yager co-norm
	  leftArgument =  blitz::pow(combinedFusion,weights(0));
	  rightArgument = blitz::pow(currentSlice,weights(0));
	  combinedFusion = blitz::min(blitz::pow(leftArgument+rightArgument,(float)(1.0/weights(0))),1.0); 
	break;
	case 4:
	  // Hamacher norm
	  combinedFusion = combinedFusion*currentSlice/(weights(0)+(1-weights(0))*(combinedFusion+currentSlice-combinedFusion*currentSlice));
	break;
	case 5:
	  // Hamacher co-norm
	  combinedFusion = (combinedFusion+currentSlice-combinedFusion*currentSlice-(1-weights(0))*combinedFusion*currentSlice)/(1-(1-weights(0))*combinedFusion*currentSlice);
	break;
	case 6:
	  // Schweizer and Sklar norm
	  leftArgument = blitz::pow(1-combinedFusion,weights(0))+blitz::pow(1-currentSlice,weights(0));
	  rightArgument = blitz::pow(1-combinedFusion,weights(0))*blitz::pow(1-currentSlice,weights(0));
	  combinedFusion = 1-blitz::pow(leftArgument+rightArgument,(float)(1.0/weights(0)));
	break;
	case 7:
	  leftArgument = blitz::pow(combinedFusion,weights(0))+blitz::pow(currentSlice,weights(0));
	  rightArgument = blitz::pow(combinedFusion,weights(0))*blitz::pow(currentSlice,weights(0));
	  combinedFusion = blitz::pow(leftArgument+rightArgument,(float)(1.0/weights(0)));
	  // Schweizer and Sklar co-norm
	break;
	case 8:
	  combinedFusion = blitz::max((combinedFusion+currentSlice-1+weights(0)*combinedFusion*currentSlice)/(1+weights(0)),0.0);
	  // Weber-Sugeno norm
	break;
	case 9:
	  combinedFusion = blitz::min((combinedFusion+currentSlice+weights(0)*combinedFusion*currentSlice),1.0);
	  // Weber-Sugeno co-norm
	break;
	case 10:
	  combinedFusion = combinedFusion+blitz::pow(currentSlice,(float)weights(0)); // Contrast adjustment
	break;
	case 11:
	  // simple weighted product
	  combinedFusion = combinedFusion * currentSlice * weights(i);
	break;
	default:
	  throw std::runtime_error( "Unknown Type!" );
      }

  }
  return CPUData<float>(combinedFusion.data(),w,h,1);

}

void Fusion::readWeights(const char* fileName)
{

  ifstream input(fileName);
  std::cerr << "The weights: -->"; 
  for(int i=0;i<f;i++)
    {
      input >> weights(i);
      std::cerr << weights(i) << " ";
    }
  std::cerr << ".\n";
}

void Fusion::scaleCurrentData()
{
  int ii=0;
  // If range is unspecified, grow the limits of the density estimate for each channel to the limits of observed values
  // Note that it may be better to define this according to say 99.9 percentile values so that outliers don't cost precision
  if (rangeDefined == 0) // It's MUCH better to load the range and save this computation!
    {
      float tempMin=0;
      float tempMax=0;
      for (ii=0; ii<f; ii++)
	{
	  tempMin = min(currentData(Range::all(),Range::all(),ii));
	  tempMax = max(currentData(Range::all(),Range::all(),ii));  // Rescaling without defined range of limits happens here
	  if (tempMin < minVals(ii))
	    minVals(ii) = tempMin;
	  if (tempMax > maxVals(ii))
	    maxVals(ii) = tempMax;
	  if (scalingType == 3)
	    {
	      currentData(Range::all(),Range::all(),ii)-=tempMin; // This rescaling is very important!
	      if (tempMin != 0 || tempMax != 0)
		currentData(Range::all(),Range::all(),ii)/=(tempMax-tempMin); // This difference could be pre-computed in loaded case for speed 

	    }
	}
    }

  if (maxVals(0) != minVals(0))
    {

      if (scalingType == 1) // Scaling is based on a known range for each filter type
	{
	  // If rangeDefined is non-zero, these values may be used, scaling is based on the above operation - for speed/robustness could potentially normalize across all feature output
	  for (ii=0; ii<f; ii++)
	    {
	      currentData(Range::all(),Range::all(),ii)-=minVals(ii); // This rescaling is very important!
	      if (minVals(ii) != 0 || maxVals(ii) != 0)
		currentData(Range::all(),Range::all(),ii)/=(maxVals(ii)-minVals(ii)); // This difference could be pre-computed in loaded case for speed 
	    }
	}
      else if (scalingType == 2)// Note no else if required since this function only gets called if scalingType is not 0
	{
	  // Alternatively! - Scaling is based on the response across all channels (assumes response range is the same for each channel -- read below)
	  currentData -= min(minVals);
	  currentData /= (max(maxVals)-min(minVals)); // In the case that a channel is relatively silent over training, we don't want to amplify this                                        
	  // This can improve robustness in scaling provided all filters have a similar response in general
	}
      else
	{
	  // Scaling happens based on the values of the current frame only
	  // This means it already happened above
	}

    }
}
