/*
Copyright INRIA 2011

Authors: Neil D.B. Bruce <Neil.Bruce@inria.fr>
Contact: Pierre Kornprobst <Pierre.Kornprobst@inria.fr>

This software is a computer program whose purpose is to estimate saliency.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/

#include "Surprise.h"
#include <cstdio>
#include <limits>
#include <fstream>
#ifdef HAVE_HDF5
#include <HDF5FloatFile.h>
#endif



using namespace blitz;

Surprise::Surprise(int w_in, int h_in, int feats_in, int bins_in, int type_in, float scale_in, float decay_in, int sw_in, float limitscale_in, CPUData<float> weightMatrix_in) // Still need to pass in weighting factor and decay factor
{
  init(w_in,h_in,feats_in,bins_in,type_in,scale_in,decay_in,sw_in,limitscale_in,weightMatrix_in);
  cout << "Initialized constructor!" << w << h << f << "!\n";
}

Surprise::Surprise()
{
}


void Surprise::init(int w_in, int h_in, int feats_in, int bins_in, int type_in, float scale_in, float decay_in, int sw_in, float limitscale_in, CPUData<float> weightMatrix_in)
{

  probDensities.resize(w_in,h_in,feats_in,bins_in); // Stores the probability densities corresponding to each location
  normedDensities.resize(w_in,h_in,feats_in,bins_in); // Stores a normalized version of the above for quick lookup
  priorDensities.resize(w_in,h_in,feats_in,bins_in); // A pre-existing representation of the densities is stored here (read from a file) used in type 3 (learnedSurprise) and type 4 (decaying surprise) 
  currentData.resize(w_in,h_in,feats_in); // Holds the current feature responses for the frame under consideration
  catData.resize(w_in,h_in,feats_in); // Holds a cumulative data representation for the computation of categorical surprise
  normCatData.resize(w_in,h_in,feats_in); // Normalized version of categorical representation
  indices.resize(w_in,h_in,feats_in); // A mapping from the normedDensities to bin indices
  catCounter.resize(w_in,h_in); // A scaling factor for categories
  marginalSurprise.resize(w_in,h_in,feats_in); // Topographic surprise at the level of features, may be stored to allow weighting later on
  combinedSurprise.resize(w_in,h_in); // Sum of marginal surprise across features
  simpleDensities.resize(feats_in,bins_in); // A simpler overall probability density for each feature type for fast surprise computation
  minVals.resize(feats_in);
  maxVals.resize(feats_in);
  sw = sw_in;
  scalingType = 0; // Generally this should be 0, but might be made a free parameter for "special circumstances" - if a channel is almost silent and this interval is used for learning, this can create noise
  priorDefined = 0; // This becomes 1 if the loadPriorDensities function is called
  weightMatrix.resize(2*sw+1,2*sw+1);
  scaleFactor = scale_in;
  decayFactor = decay_in;
  oneMinusDecay = 1-decayFactor;
  oneMinusScale = 1-scaleFactor;
  limitscale = limitscale_in;  // Binning is based on where the observed values sit within a range. Given the shape of the distributions and the binsizes, sometimes
  // scaling based on max and min values results in a poor representation of the distribution
  // For this reason, it is sensible to instead rescale based on some fraction of maximum and minimum values
  // this divisive factor is too big if density accumulates in the "end" bins for a Generalized Gaussian Distribution


  w=w_in;
  h=h_in;
  f=feats_in;
  type = type_in;
  bins=bins_in;
  minVals = 10e10;
  maxVals = -10e10;
  rangeDefined = 0;
  simpleDensities=0.00000001;
  probDensities  =0.00000001; // Make non-zero to start, to avoid divide by zeros, this has no effect over time
  // In temporal case, it is also fine to use 0 as "Surprise" is computed after an observation is seen
  // guaranteeing a non-zero likelihood
  normedDensities=1.0/bins;     // A priori everything is equal in its expectation - if a priorDensities is loaded, normedDensities becomes priorDensities

  catCounter = 0;
  catData = 0;

  // If no weight matrix is specified use a uniform spatial distribution
  weightMatrix = 1.0/((2*sw+1)*(2*sw+1));

  if (!weightMatrix_in.empty())
    {
      if (weightMatrix_in.rows() != 2*sw+1 || weightMatrix_in.columns() != 2*sw+1)
	throw std::runtime_error("Size of weight matrix given in slot 2 does not match sw (support window width/height).");   

      weightMatrix = Array<float,2>(weightMatrix_in.data(), shape(2*sw+1,2*sw+1), neverDeleteData);
    }

}


Surprise::~Surprise() {
    ;
}

void Surprise::setMask(CPUData<float> newMask)
{
  if ( newMask.empty() )
    catMask.free();
  else
    catMask.reference( Array<float,2>( newMask.data(), shape(w,h), neverDeleteData ) );
}



void Surprise::setCurrentData(CPUData<float> newData)
{
  // Whenever new data is observed, the marginal surprise and combined surprise are updated according to the method specified in type

  // Copy the new representation to the current one
  currentData = Array<float,3>(newData.data(), shape(w,h,f), neverDeleteData);

  // Update density here as prescribed by the method that is specified
  switch (type) {

    case 0:
      scaleCurrentData();
      updateDensitiesSimple();      // Simple global feature estimate
      computeMarginalSurpriseSimple();
    break;

    case 1: 
      scaleCurrentData();
      updateDensitiesAccumulate();  // Local estimate that accumulates continuously
      computeMarginalSurprise();
    break;

    case 2:
      scaleCurrentData();
      updateDensitiesWeighted();    // Local estimate updated over time according to a weighted average
      computeMarginalSurprise();
    break;

    case 3:
      scaleCurrentData();
      // Do nothing beyond this
      // the characteristic probabilities have been loaded, all that remains is computing the marginal/combined surprise
      if (priorDefined == 1)
	{} // Do nothing - normedDensities has been set to priorDensities and will not change
      else
	{
	  std::cout << "Asking to use prior distribution in computation, but it has not been loaded.\n";
	  exit(1);
	}
      computeMarginalSurprise();
    break;

    case 4:
      scaleCurrentData();
      if (priorDefined == 1)
	updateDensitiesDecay();
      else
	{
	  std::cout << "Asking to use prior distribution for decay type update, but prior has not been loaded.\n";
	  exit(1);
	}
      computeMarginalSurprise();
    break;
    case 6:
    case 7:
      computeCatSurprise();
      updateDensitiesCat();
    break;

    default:

      throw std::runtime_error( "Unknown Type!" );

  }
  // Why was this removed?
  computeCombinedSurprise();
}


void Surprise::updateDensitiesCat()
{

  if ( catMask.numElements() > 0 )
    {
      catData    = oneMinusDecay * catData + scaleFactor * currentData * catMask( firstIndex(), secondIndex() );
      catCounter = oneMinusDecay * catCounter + scaleFactor * catMask;
    }
  else
    {
      catData    = oneMinusDecay * catData + scaleFactor * currentData;
      catCounter = oneMinusDecay * catCounter + scaleFactor;
    }

}


void Surprise::computeCatSurprise()
{
  firstIndex i;
  secondIndex j;
  thirdIndex k;

  // Normalize to average bits
  // Add a tiny delta in case of 0 probabilities
  float delta = 0.5f;
  normCatData = ( catData + delta ) / ( catCounter( i, j ) + delta );

  if (type == 6)
    {
      if ( catMask.numElements() > 0 )	  
	combinedSurprise = sum( catMask( i, j ) * currentData * -log( normCatData ), k );
      else
	combinedSurprise = sum( currentData * -log( normCatData ), k );
    }
  else if (type == 7)
    {
      if ( catMask.numElements() > 0 )	  
	combinedSurprise = sum( catMask( i, j ) * currentData * log( currentData / normCatData ), k );
      else
	combinedSurprise = sum( currentData * log( max( 1e-5, currentData ) / normCatData ), k );
    }
}


// 



void Surprise::scaleCurrentData()
{
  int ii=0;
  // If range is unspecified, grow the limits of the density estimate for each channel to the limits of observed values
  // Note that it may be better to define this according to say 99.9 percentile values so that outliers don't cost precision
  if (rangeDefined == 0) // It's MUCH better to load the range and save this computation!
    {
      float tempMin;
      float tempMax;
      for (ii=0; ii<f; ii++)
	{
	  tempMin = min(currentData(Range::all(),Range::all(),ii))/limitscale;
	  tempMax = max(currentData(Range::all(),Range::all(),ii))/limitscale;  // Rescaling without defined range of limits happens here
	  if (tempMin < minVals(ii))
	    minVals(ii) = tempMin;
	  if (tempMax > maxVals(ii))
	    maxVals(ii) = tempMax;
	}



    }



  if (scalingType == 0) // Scaling is based on a known range for each filter type
    {
      // If rangeDefined is non-zero, these values may be used, scaling is based on the above operation - for speed/robustness could potentially normalize across all feature output
      for (ii=0; ii<f; ii++)
	{
	  currentData(Range::all(),Range::all(),ii)-=minVals(ii); // This rescaling is very important!
	  currentData(Range::all(),Range::all(),ii)/=(maxVals(ii)-minVals(ii)); // This difference could be pre-computed in loaded case for speed 
	}
    }
  else
    {
      // Alternatively! - Scaling is based on the response across all channels (assumes response range is the same for each channel -- read below)
      currentData -= min(minVals);
      currentData /= (max(maxVals)-min(minVals)); // In the case that a channel is relatively silent over training, we don't want to amplify this                                        
      // This can improve robustness in scaling provided all filters have a similar response in general
    }

  // Convert from likelihoods to bins
  indices = floor(currentData*bins); // currentData lies between 0 and 1 - this operation maps the observed values to the bin they belong in
  indices = max(indices,0);
  indices = min(indices,bins-1); // Stick outliers on the outermost bins - this requires range to be chosen carefully, since confidence may accumulate in the 0,f-1 bins if range is narrow

}


void Surprise::updateDensitiesSimple()
{

  for (int i=0; i<w; i++)
    for (int j=0; j<h; j++)
      for (int k=0; k<f; k++){
	  simpleDensities(k,indices(i,j,k))+=1;
      }

  // Divide each dimension k by sum to get a pdf - This is a special case and thus doesn't use the normalize function
  for (int k=0; k<f; k++){
      simpleDensities(k,Range::all()) /= sum(simpleDensities(k,Range::all()));
      //
  }
}


void Surprise::updateDensitiesAccumulate()
{
  // This function updates the probability density map and controls the type of behavior observed
  // Typically using a temporal window that may be of varying extent


  // Can build in a spatial support to accumulate information faster here and provide smoother densities
  // can also extend this to arbitrary support

  updateDensitiesCommon();

  updateNormalizedDensityApprox();
}

void Surprise::updateDensitiesCommon()
{
  for ( int i = sw; i < w - sw; i++ )
    for ( int j = sw; j< h - sw; j++ )
      for ( int k = 0; k < f; k++ )       
	{
	  blitz::Array<float,1> pijk( probDensities( i, j, k, blitz::Range::all() ) );
	  blitz::Array<int,2> indijk( indices( blitz::Range( i - sw, i + sw ), 
		blitz::Range( j - sw, j + sw ), 
		k ) );
	  blitz::Array<int,2>::iterator iind = indijk.begin();
	  blitz::Array<float,2>::iterator iw = weightMatrix.begin(), iwend = weightMatrix.end();

	  for ( ; iw != iwend; ++iw, ++iind )
	    pijk( *iind ) += (*iw);
	}
}

void Surprise::updateDensitiesWeighted()
{
  updateDensitiesCommon();

  probDensities *= scaleFactor;

  normedDensities = normedDensities*oneMinusScale + probDensities;

}

void Surprise::updateDensitiesDecay()
{
  // There is a decay rate and an updating rate here - uses updateDensitiesWeighted for second step

  // Force Normalized Density to Decay towards prior loaded representation
  // Both Normalized and Prior Density should already be normalized
  normedDensities=decayFactor*priorDensities+oneMinusDecay*normedDensities;

  // Update with new observation according to some weighting
  updateDensitiesWeighted();

}


void Surprise::computeMarginalSurpriseSimple()
{

  // look values up from density map
  for (int i=0; i<w; i++)
    for (int j=0; j<h; j++)
      for (int k=0; k<f; k++){
	  marginalSurprise(i,j,k)=-log(simpleDensities(k,indices(i,j,k)));
      }

}

void Surprise::computeMarginalSurprise()
{

  float oneOverBins = 1.0/bins;

  // look values up from density map
  for (int i=sw; i<w-sw; i++)
    for (int j=sw; j<h-sw; j++)
      for (int k=0; k<f; k++){
	  marginalSurprise(i,j,k)=-log(normedDensities(i,j,k,indices(i,j,k))); 

      }

}


void Surprise::computeCombinedSurprise()
{
  firstIndex i;
  secondIndex j;
  thirdIndex k;
  // Sum across features
  combinedSurprise=sum(marginalSurprise,k);
}



CPUData<float> Surprise::returnMarginalSurprise()
{

  return CPUData<float>(marginalSurprise.data(),w,h,f);

}



CPUData<float> Surprise::returnCombinedSurprise()
{
  return CPUData<float>(combinedSurprise.data(),w,h,1);
}

blitz::Array<float,4>& Surprise::returnNormedDensities()
{
  return normedDensities;
}

void Surprise::updateNormalizedDensityApprox()
{
  // In fact, this is only approximate for some types, otherwise it is exact - but surprisingly speed-up is not enormous
  float mult = (w-(2*sw))*(h-(2*sw))*f;
  float totalsum = sum(probDensities);
  float div = mult/totalsum;

  normedDensities = probDensities*div;
}


// In the case of range values, these are currently based on maximum and minimum observations
// Precision might be higher using less bins if percentile values are specified in the file loaded (e.g. 0.01 and 99.9)
// to be robust to outliers

void Surprise::writeRange(const char* fileName)
{
  ofstream output(fileName);

  for(int i=0;i<f;i++)
    {
      output<< minVals(i) <<endl;
      output<< maxVals(i) <<endl;
    }

}

void Surprise::readRange(const char* fileName)
{
  rangeDefined = 1;
  // Set minVals, maxVals to values specified in file - should happen before setCurrentData gets called the first time
  // Unless one wishes to determine the range of values "on the fly" from the data

  ifstream output(fileName);
  float tempFloat;
  for(int i=0;i<f;i++)
    {
      output >> minVals(i);
      output >> maxVals(i);
    }

  // If range is read, rescaling of end limits happens here
  minVals(Range::all())/=limitscale;
  maxVals(Range::all())/=limitscale;


}

// Read/Write observed probability density functions for locations/features to disk

void Surprise::writePriorDensities(const char* fileName)
{
#ifdef HAVE_HDF5
  // Writes the normalized prior densities to an HDF5 file
  HDF5FloatFile outFile = HDF5FloatFile(fileName); // Crashes if file doesn't exist... fix this
  const size_t fdim[4] = {h,w,f,bins};
  outFile.writeDataSet("NormalizedDensities",normedDensities.data(),4,fdim);
#else
  throw std::runtime_error("persistence for Surprise needs hdf5");
#endif
}

void Surprise::readPriorDensities(const char* fileName)
{
#ifdef HAVE_HDF5
  priorDefined = 1;
  // Read density from H5 file into priorDensities
  HDF5FloatFile outFile = HDF5FloatFile(fileName); // Crashes if file doesn't exist... fix this
  const size_t fdim[4] = {h,w,f,bins};
  outFile.readDataSet(priorDensities.data(),"NormalizedDensities");
  // Set the existing normalized density representation to the prior
  normedDensities = priorDensities;

  // Check if there is a zero in loaded bins to avoid division by 0
  if (min(priorDensities)==0)
    cout << "***** WARNING (void Surprise::readPriorDensities(const char* fileName)): There is a zero value in the priorDensity that has been loaded!!! *****" << std::endl;
#else
  throw std::runtime_error("persistence for Surprise needs hdf5");
#endif

}

blitz::Array<int,3> Surprise::returnIndex()
{
  return indices;
}

// computes the jensen-shannon-divergence between the local and a foreign histogram
CPUData<float> Surprise::discriminantSaliency(blitz::Array<float,4> otherNormedDensities)
{
  blitz::Array<float,4> avgDist;
  avgDist.resize(w,h,f,bins);
  avgDist=0;
  avgDist=(normedDensities+otherNormedDensities)/2.0;

  float P1,P2,Q;
  float epsilon=10e-6;
  CPUData<float> temp(w,h,1);
  std::fill(temp.begin(),temp.end(),0.0);
  for (int i = 0; i<w; i++)
    for (int j=0; j<h; j++)
      for (int k=0; k<f ; k++)
	for (int l=0; l<bins; l++)
	  {
	    P1 = normedDensities(i,j,k,l);
	    P2 = otherNormedDensities(i,j,k,l);
	    Q = avgDist(i,j,k,l);
	    if (P1>epsilon && P2 >epsilon && Q> epsilon)
	      temp(i,j,0)+=0.5*P1*log(P1/Q)+0.5*P2*log(P2/Q);
	  }
  return temp;
}

