/*
Copyright INRIA 2011

Authors: Neil D.B. Bruce <Neil.Bruce@inria.fr>
Contact: Pierre Kornprobst <Pierre.Kornprobst@inria.fr>

This software is a computer program whose purpose is to estimate saliency.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/
#include "SurpriseMod.h"
#include <cstdio>
#include <limits>
#include <fstream>
#ifdef HAVE_HDF5
	#include <HDF5FloatFile.h>
#endif
#include <boost/format.hpp>

//#define FAST_MODE

// To add:
// Functions to read/write probabilistic representations to disk
// Tracking of min/max or percentile values of features for normalization of feature inputs
// Allowing type to be specified in the constructor to distinguish between global surprise, accumulating surprise, learned characteristic surprise, updating surprise, decaying surprise, buffered surprise
// Constructors that allow characteristic surprise or range of filter responses to be specified a priori
// Difference metric of the Gao form that acts on "this" and another Surprise module
// Can also do spatiotemporal updates on the global mode for fast non location specific temporal modeling



using namespace blitz;

SurpriseMod::SurpriseMod(int w_in, int h_in, int feats_in, int bins_in, int type_in, float scale_in, float decay_in, int sw_in, int step_x_in, int step_y_in, float limitscale_in, int metric_in, CPUData<float> weightMatrix_in) 
{
  init(w_in,h_in,feats_in,bins_in,type_in,scale_in,decay_in,sw_in, step_x_in, step_y_in,limitscale_in, metric_in,  weightMatrix_in);
}

SurpriseMod::SurpriseMod()
{
}

SurpriseMod::~SurpriseMod()
{
  deinit();
}

void SurpriseMod::deinit()
{
  type          = 0;
  metric        = 0;
  scalingType   = 0; 									/// Generally this should be 0, but might be made a free parameter for "special circumstances" - if a channel is almost silent and this interval is used for learning, this can create noise
  priorDefined  = 0; 									/// This becomes 1 if the loadPriorDensities function is called
  set_sw( 0 );                    /// sets sw_extent aswell
  step_x        = 1; 
  step_y        = 1;
  w             = 0;
  h             = 0;
  f             = 0;
  h_max         = 0;
  bins          = 1;
  scaleFactor   = 1;
  decayFactor   = 0;
  rangeDefined  = 0;
  globalCounter = 0;
  priorGlobalCounter = 0;
  limitscale    = 1;
  _do_learn     = true;
  _do_detect    = true;
  
  minVals.free();
  maxVals.free();
  histogram.free();
  counter.free();
  indices.free();
  marginalSurprise.free();
  combinedSurprise.free();
  globalHistogram.free();
  priorHistogram.free();
  priorGlobalHistogram.free();
  weightMatrix.free();
}

void SurpriseMod::init(int w_in, int h_in, int feats_in, int bins_in, int type_in, float scale_in, float decay_in, int sw_in, int step_x_in, int step_y_in, float limitscale_in, int metric_in, CPUData<float> weightMatrix_in)
{
  //std::cout << "SurpriseMod::init: w = "<< w_in << " h = "<< h_in << " f = " << feats_in << " bins = "<< bins_in << std::endl;
  /// init scalar values
  type          = type_in;
  metric        = metric_in;
  scalingType   = 0; 									/// Generally this should be 0, but might be made a free parameter for "special circumstances" - if a channel is almost silent and this interval is used for learning, this can create noise
  priorDefined  = 0; 									/// This becomes 1 if the loadPriorDensities function is called
  set_sw( sw_in );                    /// sets sw_extent aswell
  step_x        = step_x_in;
  step_y        = step_y_in;
  w             = w_in;
  h             = h_in;
  if (step_x_in < 1 || step_x_in >= w_in || step_y_in < 1 || step_y_in >= h_in)
    throw std::runtime_error(str( boost::format("invalid step values: x=%d  y=%d, w=%d, h=%d")%step_x_in%step_y_in%w_in%h_in ));
  
  f             = feats_in;
  bins          = bins_in;
  scaleFactor   = scale_in;
  decayFactor   = decay_in;
  rangeDefined  = 0;
  globalCounter = 0;
  priorGlobalCounter = 0;
  limitscale    = limitscale_in;  							/// Binning is based on where the observed values sit within a range. Given the shape of the distributions and the binsizes, sometimes
                                     						/// scaling based on max and min values results in a poor representation of the distribution
                                     						/// For this reason, it is sensible to instead rescale based on some fraction of maximum and minimum values
                                     						/// this divisive factor is too big if density accumulates in the "end" bins for a Generalized Gaussian Distribution
  _do_learn  = true;
  _do_detect = true;
  
  initDefaultMatrices(weightMatrix_in);

}
 

  
void SurpriseMod::initDefaultMatrices( CPUData<float> weightMatrix_in )
{
  minVals.resize              (         f           );
  maxVals.resize              (         f           );
  histogram.resize            ( w,  h,  f,  bins    );  /// Stores a normalized version of the above for quick lookup
  counter.resize              ( w,  h,  f           );  /// Stores a normalized version of the above for quick lookup
  indices.resize              ( w,  h,  f           );  /// A mapping from the histogram to bin indices
  marginalSurprise.resize     ( w,  h,  f           );  /// Topographic surprise at the level of features, may be stored to allow weighting later on
  combinedSurprise.resize     ( w,  h               );  /// Sum of marginal surprise across features
  globalHistogram.resize      (         f,  bins    );
  //priorHistogram.resize       ( w,  h,  f,  bins    );
  //priorGlobalHistogram.resize ( w,  h,  f,  bins    );
  weightMatrix.resize         ( sw_extent, sw_extent);
  
  globalHistogram=0.00000001;
  globalCounter = 0.5f;
  indices = 0;
  switch ( metric )
   {
     case 0 : // si
     case 2 : // modsi
     case 4 : // kl
     case 5 : // ikl
     case 6 : // modkl
      histogram = 1.0/bins;
      break;
     case 1 : // bsi
     case 3 : // bmodsi
     case 7 : // bkl
     case 8 : // bikl
     case 9 : // bsymkl
      histogram = 0.5f;
      break;
   }
  counter   = 0.5f;
  
  if (weightMatrix_in.empty())
  {
    /// If no weight matrix is specified use a uniform spatial distribution
    weightMatrix = 1.0/( ( sw_extent ) * ( sw_extent ) );
  }
  else
  {
    if (weightMatrix_in.rows() != sw_extent || weightMatrix_in.columns() != sw_extent )
      throw std::runtime_error("Size of weight matrix given in slot 2 does not match sw (support window width/height).");   
    
    weightMatrix = Array<float,2>(weightMatrix_in.data(), shape(sw_extent,sw_extent), neverDeleteData);
  }
}


void SurpriseMod::processFeatureData( CPUData<float> newData, CPUData<float> confidence )
{
  
  /// Whenever new data is observed, the marginal surprise and combined surprise are updated according to the method specified in type
  /// change representation of input
  cpudata2blitz( newData,    currentData );
  cpudata2blitz( confidence, currentMask );
  
  check_metric( false );
  
  quantizeData( );      /// create indices and delete original data (reference) as it is no longer needed
  
  switch (type) 
  {
    case 0:   /// "global"
      if ( _do_detect )    dispatch_compute_global();
      if ( _do_learn  )    updateDensitiesGlobal();
      break;
    case 1:   /// "local"
      if ( _do_detect )    dispatch_compute_local();
      if ( _do_learn  )    updateDensitiesLocal();
      break;
    default:
      throw std::runtime_error( "Unknown Type!" );
  }
}
void SurpriseMod::processHistogramData( CPUData<float> newData, CPUData<float> confidence)
{
  
  /// Whenever new data is observed, the marginal surprise and combined surprise are updated according to the method specified in type
  /// change representation of input
  cpudata2blitz( newData,    currentData );
  cpudata2blitz( confidence, currentMask );
  
  check_metric( true );
  
  indices.free( );
  
  switch (type) 
  {
    case 0:   /// "global"
      if ( _do_detect )    dispatch_compute_global();
      if ( _do_learn  )    updateDensitiesGlobal();
      break;
    case 1:   /// "local"
      if ( _do_detect )    dispatch_compute_local();
      if ( _do_learn  )    updateDensitiesLocal();
      break;
    default:
      throw std::runtime_error( "Unknown Type!" );
  }
}

void SurpriseMod::processIndexedData( CPUData<unsigned> newData, CPUData<float> confidence )
{
  cpudata2blitz( newData, indices );
  cpudata2blitz( confidence, currentMask );
  
#ifdef DEBUG
  check_if_indices_valid();
#endif

  switch (type) 
  {
    case 0:   /// "global"
      if ( _do_detect )    dispatch_compute_global();
      if ( _do_learn  )    updateDensitiesGlobal();
      break;
    case 1:   /// "local"
      if ( _do_detect )
#ifndef FAST_MODE
        dispatch_compute_local();
#else
        {
          for (int i=sw; i<w-sw; i++)
            for (int j=sw; j<h-sw; j++)
                for (int k=0; k<f; k++)
                    marginalSurprise(i,j,k) = - /*p*/currentMask( i, j ) * my_log( /*h*/histogram(i,j,k,int(indices(i,j,k))) / /*Z*/counter(i,j) );
        }
#endif
      if ( _do_learn  )    updateDensitiesLocal();
      
      break;
    default:
      throw std::runtime_error( "Unknown Type!" );
  }
}

void SurpriseMod::check_if_indices_valid()
{
  for ( int i = 0; i < w ; i++ )
    for ( int j = 0; j< h ; j++ )
      for ( int k = 0; k< f ; k++ )
        if ( indices( i, j, k ) >= bins )
          throw std::runtime_error( str( boost::format( "indices out of bounds at [ %d , %d , %d ] = %d, binsize = %d. in case of 2d vel-data: Check if indexing module has same max_vel as SurpriseMod" )%i%j%k%indices( i, j, k )%bins ) );
}


void SurpriseMod::processHypoData( CPUData<float> hypoData )
{
  cpudata2blitz( hypoData,    currentHypoData );
  h_max = currentHypoData.shape()[2];
  check_metric( false );
  if ( type !=1 ) throw std::runtime_error("processing hypooImages only possible with type=1 (local)!");
  quantizeHypoData();
  
  if ( _do_learn  )    updateHypoDensitiesLocal();
  if ( _do_detect )    dispatch_compute_local();
}

void SurpriseMod::do_filter( const CPUData<float>& input, blitz::Array<float,3>& output, const blitz::Array<float,2>& weights )
{
  std::cout << "starting filter 3d..."<<std::endl;
  size_t extent = weights.shape()[0];
  size_t sw     = (extent-1) / 2;
  CPUData<float> w;
  blitz2cpudata( weights, w );
  
  sw_accum_w core( input.depth(), w );

  sliding_window_iterator<sw_accum_w> switer( input, extent, input.depth(), core );
  
  for ( int i = sw; i < output.shape()[0] - sw; ++i )
    for ( int j = sw; j < output.shape()[1] - sw; ++j )
    {
      CPUDataFeatureColumn<float> in = (*switer);
      for (int k = 0; k < in.size(); ++k )
        output( i, j, k ) = in[k];
      switer.increment();
    }
    std::cout << "finished filter 3d"<<std::endl;
}

void SurpriseMod::do_filter( const CPUData<float>& input, blitz::Array<float,2>& output, const blitz::Array<float,2>& weights )
{
  std::cout << "starting filter 2d..."<<std::endl;
  if (input.depth() != 1)
    throw std::runtime_error( "SurpriseMod::do_filter: cannot filter from 3D to 2D!" );
  
  size_t extent = weights.shape()[0];
  size_t sw     = (extent-1) / 2;
  CPUData<float> w;
  blitz2cpudata( weights, w );
  
  sw_accum_w core( input.depth(), w );

  sliding_window_iterator<sw_accum_w> switer( input, extent, input.depth(), core );
  
  for ( int i = sw; i < output.shape()[0] - sw; ++i )
    for ( int j = sw; j < output.shape()[1] - sw; ++j )
    {
      output( i, j ) = (*switer)[0];
      switer.increment();
    }
  std::cout << "finished filter 2d"<<std::endl;
}

void SurpriseMod::dispatch_compute_global()
{
  switch (metric)
  {
    case 0:
      computeMarginalSurpriseGlobal< metric_si >();
      break;
    case 1:
      computeMarginalSurpriseGlobal< metric_bsi >();
      break;
    case 2:
      computeMarginalSurpriseGlobal< metric_modsi >();
      break;
    case 3:
      computeMarginalSurpriseGlobal< metric_bmodsi >();
      break;
    case 4:
      computeMarginalSurpriseGlobal< metric_kl >();
      break;
    case 5:
      computeMarginalSurpriseGlobal< metric_ikl >();
      break;
    case 6:
      computeMarginalSurpriseGlobal< metric_symkl >();
      break;
    case 7:
      computeMarginalSurpriseGlobal< metric_bkl >();
      break;
    case 8:
      computeMarginalSurpriseGlobal< metric_bikl >();
      break;
    case 9:
      computeMarginalSurpriseGlobal< metric_bsymkl >();
      break;
    case 10:
      computeMarginalSurpriseGlobal< metric_js >();
      break;
    case 11:
      computeMarginalSurpriseGlobal< metric_bjs >();
      break;
    default:
      throw std::runtime_error( "Unknown metric!" );
  }
}

/// prefix "b" stands for binomial interpretation, else multinomial is expected
void SurpriseMod::dispatch_compute_local()
{
  switch (metric)
  {
    case 0:
      computeMarginalSurpriseLocal< metric_si >();
      break;
    case 1:
      computeMarginalSurpriseLocal< metric_modsi >();
      break;
    case 2:
      computeMarginalSurpriseLocal< metric_bsi >();
      break;
    case 3:
      computeMarginalSurpriseLocal< metric_bmodsi >();
      break;
    case 4:
      computeMarginalSurpriseLocal< metric_kl >();
      break;
    case 5:
      computeMarginalSurpriseLocal< metric_ikl >();
      break;
    case 6:
      computeMarginalSurpriseLocal< metric_symkl >();
      break;    
    case 7:
      computeMarginalSurpriseLocal< metric_bkl >();
      break;
    case 8:
      computeMarginalSurpriseLocal< metric_bikl >();
      break;
    case 9:
      computeMarginalSurpriseLocal< metric_bsymkl >();
      break;
    case 10:
      computeMarginalSurpriseLocal< metric_js >();
      break;
    case 11:
      computeMarginalSurpriseLocal< metric_bjs >();
      break;
    default:
      throw std::runtime_error( "Unknown metric!" );
  }
}

/// binomial metrics should be applied to quantized data and not indexed ones and vise versa
bool SurpriseMod::is_metric_applicable( bool is_quantized )
{ 
  switch (metric)
  {
    case 0:
    case 4:
	  return true;
    case 5:
    case 6:
    case 1:
    case 2:
    case 3:
    case 7:
    case 8:
    case 9:
    case 10:
    case 11:
	  return is_quantized;
    default:
      throw std::runtime_error(str(boost::format("SurpriseMod: unknown metric:%i")%metric ));
  }
}

void SurpriseMod::check_metric ( bool is_quantized )
{
  if ( !is_metric_applicable(is_quantized) )
  {
    throw std::runtime_error(str(boost::format("SurpriseMod: metric:%i is not compatible with input which is %s quantized!")%metric %(!is_quantized?"not":"interpreted as") ));
  }
}

/// checks data range and quantizes after rescaling
void SurpriseMod::quantizeData(  )
{
  int ii=0;
  // If range is unspecified, grow the limits of the density estimate for each channel to the limits of observed values
  // Note that it may be better to define this according to say 99.9 percentile values so that outliers don't cost precision
  if (rangeDefined == 0) // It's MUCH better to load the range and save this computation!
  {
    float tempMin;
    float tempMax;
    for (ii=0; ii<f; ii++)
    {
      tempMin = min(currentData(Range::all(),Range::all(),ii))/limitscale;
      tempMax = max(currentData(Range::all(),Range::all(),ii))/limitscale;  // Rescaling without defined range of limits happens here
      if (tempMin < minVals(ii))
        minVals(ii) = tempMin;
      if (tempMax > maxVals(ii))
        maxVals(ii) = tempMax;
    }           
  }

  if (scalingType == 0) // Scaling is based on a known range for each filter type
	{
	  // If rangeDefined is non-zero, these values may be used, scaling is based on the above operation - for speed/robustness could potentially normalize across all feature output
	  for (ii=0; ii<f; ii++)
		{
		  currentData(Range::all(),Range::all(),ii)-=minVals(ii); // This rescaling is very important!
		  currentData(Range::all(),Range::all(),ii)/=(maxVals(ii)-minVals(ii)); // This difference could be pre-computed in loaded case for speed 
		}
	}
  else
	{
	  // Alternatively! - Scaling is based on the response across all channels (assumes response range is the same for each channel -- read below)
	  currentData -= min(minVals);
	  currentData /= (max(maxVals)-min(minVals)); // In the case that a channel is relatively silent over training, we don't want to amplify this                                        
	  // This can improve robustness in scaling provided all filters have a similar response in general
	}
  
  
  // Convert from likelihoods to bins
  indices = floor(currentData*bins); // currentData lies between 0 and 1 - this operation maps the observed values to the bin they belong in
  indices = max(indices,0);
  indices = min(indices,bins-1); // Stick outliers on the outermost bins - this requires range to be chosen carefully, since confidence may accumulate in the 0,f-1 bins if range is narrow
  currentData.free();
}


/// checks data range and quantizes after rescaling
void SurpriseMod::quantizeHypoData(  )
{
  int ii=0;
  // If range is unspecified, grow the limits of the density estimate for each channel to the limits of observed values
  // Note that it may be better to define this according to say 99.9 percentile values so that outliers don't cost precision
  if (rangeDefined == 0) // It's MUCH better to load the range and save this computation!
  {
    float tempMin;
    float tempMax;
    if ( f != 2 ) throw std::runtime_error( str( boost::format("Error in quantizeHypoData: with hypoImages, featuecount must be 2, but it is = %d")%f ) );
    for (ii=0; ii<f; ii++)
    {
      tempMin = min(currentHypoData(Range::all(),Range::all(),Range::all(), ii))/limitscale;
      tempMax = max(currentHypoData(Range::all(),Range::all(),Range::all(), ii))/limitscale;  // Rescaling without defined range of limits happens here
      if (tempMin < minVals(ii))
        minVals(ii) = tempMin;
      if (tempMax > maxVals(ii))
        maxVals(ii) = tempMax;
    }           
  }

  if (scalingType == 0) // Scaling is based on a known range for each filter type
	{
	  // If rangeDefined is non-zero, these values may be used, scaling is based on the above operation - for speed/robustness could potentially normalize across all feature output
	  for (ii=0; ii<f; ii++)
		{
		  currentHypoData(Range::all(),Range::all(),Range::all(),ii)-=minVals(ii); // This rescaling is very important!
		  currentHypoData(Range::all(),Range::all(),Range::all(),ii)/=(maxVals(ii)-minVals(ii)); // This difference could be pre-computed in loaded case for speed 
		}
	}
  else
	{
	  // Alternatively! - Scaling is based on the response across all channels (assumes response range is the same for each channel -- read below)
	  for (ii=0; ii<f; ii++)
	  {
	    currentHypoData(Range::all(),Range::all(),Range::all(),ii) -= min(minVals);
	    currentHypoData(Range::all(),Range::all(),Range::all(),ii) /= (max(maxVals)-min(minVals)); // In the case that a channel is relatively silent over training, we don't want to amplify this                                        
    }
	  // This can improve robustness in scaling provided all filters have a similar response in general
	}
  
  
  // Convert from likelihoods to bins
  
  if ( hypoIndices.numElements() == 0 )
  {
    hypoIndices.resize( w, h, h_max, f );
  }
  else
  {
    
    /// check if hypoImages are of same size as inputdata
     if (    ( hypoIndices.shape()[0] != currentHypoData.shape()[0] )    /// w
          || ( hypoIndices.shape()[1] != currentHypoData.shape()[1] )    /// h
          || ( hypoIndices.shape()[2] != currentHypoData.shape()[2] )    /// h_max
          || ( hypoIndices.shape()[3] != currentHypoData.shape()[3]-1 )  /// features = 2
      )
    throw std::runtime_error( str(boost::format("hypoIndices=%s and   Data=%s do not match!")%blitz_dump_dims_str(hypoIndices)%blitz_dump_dims_str(currentHypoData) ) );
  
  }
  
  hypoIndices = floor(currentHypoData(Range::all(),Range::all(),Range::all(),Range( 0, f-1 )) * bins); // currentData lies between 0 and 1 - this operation maps the observed values to the bin they belong in
  hypoIndices = max(hypoIndices,0);
  hypoIndices = min(hypoIndices,bins-1); // Stick outliers on the outermost bins - this requires range to be chosen carefully, since confidence may accumulate in the 0,f-1 bins if range is narrow
  
  
}

void SurpriseMod::updateHypoDensitiesLocal()
{
  
  // There is a decay rate and an updating rate here - uses updateDensitiesWeighted for second step  
  // Force Normalized Density to Decay towards prior loaded representation
  // Both Normalized and Prior Density should already be normalized
  float c_prior = decayFactor;
  float c_old   = ( 1 - decayFactor );
  float c_new   = scaleFactor;
  float maskval = 0;
  histogram *= c_old;
  counter   *= c_old;
  
  /// todo: use spatial support, slidingwindow_iterator
  if ( hypoIndices.numElements() > 0 )
  { /// indexed data case
    for ( int i = sw; i < w - sw-1; i++ )
      for ( int j = sw; j< h - sw-1; j++ )
        for ( int k = 0; k < f; k++ )       
        {
          blitz::Array<float,1> pijk( histogram( i, j, k, blitz::Range::all() ) );
          float & cij = counter( i, j );
          
          
          for (int n = 0; n < h_max; n++ ) // consider all hypothesises
          {
            
            blitz::Array<unsigned,2>   index_window_k( indices(  blitz::Range( i - sw, i + sw ), 
                                                            blitz::Range( j - sw, j + sw ), 
                                                            n, /// take n-th hypothesis
                                                            k ) );            
            blitz::Array<float,2> confidence( currentHypoData(  blitz::Range( i - sw, i + sw ), 
                                                                blitz::Range( j - sw, j + sw ),
                                                                n, /// take n-th hypothesis
                                                                2 ) ); /// confidence
            
            blitz::Array<float,2>::iterator im = confidence.begin();
            blitz::Array<unsigned,2>::iterator iind = index_window_k.begin();
            blitz::Array<float,2>::iterator iw = weightMatrix.begin(), iwend = weightMatrix.end();
            
            for ( ; iw != iwend; ++iw, ++iind, ++im )
            {
              if ( *iind < 0 || *iind >= bins )
                throw std::runtime_error( str( boost::format("index out of bounds: %d not in [0, %d]")%(*iind)%(bins-1) ) );
              pijk( *iind ) += c_new * (*iw) * (*im);
              cij += (*im) * (*iw);
            }
            
          }
          
        }
  }
  else if ( currentData.numElements() > 0 )
       throw std::runtime_error( " updateDensitiesLocal: no hypoimage data! " );
  else throw std::runtime_error( " updateDensitiesLocal: no data! " );
  
  if ( c_prior > 0 && priorHistogram.numElements() == histogram.numElements() && priorCounter.numElements() == counter.numElements() )
  {
      std::cout << "prior       : "<<std::flush;
      blitz_dump_dims( priorHistogram );
      std::cout << "histogram   : "<<std::flush;
      blitz_dump_dims( histogram );
      std::cout << "priorCounter: "<<std::flush;
      blitz_dump_dims( priorCounter );
      std::cout << "counter     : "<<std::flush;
      blitz_dump_dims( counter );
    
  }


}





void SurpriseMod::updateDensitiesGlobal()
{  
  float c_prior = decayFactor;
  float c_old   = ( 1 - decayFactor );
  float c_new   = scaleFactor;

  globalHistogram *= c_old;
  globalCounter   *= c_old;
  
  if ( indices.numElements() > 0 )
  { /// indexed data case
    for (int i = 0; i < w; i++ )
      for (int j = 0; j < h; j++ )
        for (int k = 0; k < f; k++ )
          {
            globalHistogram( k, int(indices( i, j, k )) ) += c_new;
            globalCounter += 1;
          }
  
  }
  else if ( currentData.numElements() > 0 )
  { /// quantized data case
    //globalDensities += sum( sum( currentData( i2, i3, i1 ), i3 ), i2 );
    for (int i = 0; i < w; i++ )
      for (int j = 0; j < h; j++ )
        for (int b = 0; b < bins; b++ )
        {
          globalHistogram( 0, b ) += c_new * currentData( i, j, b );
          globalCounter           += 1;
        }
    
  }
  else throw std::runtime_error( " updateDensitiesGlobal: no data! " );
  
  if ( c_prior > 0 && priorGlobalHistogram.numElements() == globalHistogram.numElements() )
  {
    globalHistogram += c_prior * priorGlobalHistogram;
    globalCounter   += c_prior * priorGlobalCounter;
  }
  
}

void SurpriseMod::updateDensitiesLocal()
{
  // There is a decay rate and an updating rate here - uses updateDensitiesWeighted for second step  
  // Force Normalized Density to Decay towards prior loaded representation
  // Both Normalized and Prior Density should already be normalized
  float c_prior = decayFactor;
  float c_old   = ( 1 - decayFactor );
  float c_new   = scaleFactor;
  float maskval = 0;
  histogram *= c_old;
  counter   *= c_old;
  
  /// todo: use spatial support, slidingwindow_iterator
  if ( indices.numElements() > 0 )
  { /// indexed data case
#ifndef FAST_MODE
    for ( int i = sw; i < w - sw; i++ )
      for ( int j = sw; j< h - sw; j++ )
        for ( int k = 0; k < f; k++ )       
      {
        blitz::Array<float,1> pijk( histogram( i, j, k, blitz::Range::all() ) );
        float & cij = counter( i, j );
        
        blitz::Array<unsigned,2>   index_window_k( indices(  blitz::Range( i - sw, i + sw ), 
                                                        blitz::Range( j - sw, j + sw ), 
                                                        k ) );
        blitz::Array<unsigned,2>::iterator iind = index_window_k.begin();
        blitz::Array<float,2>::iterator iw = weightMatrix.begin(), iwend = weightMatrix.end();
        

        if ( currentMask.numElements() > 0 )
        {
          blitz::Array<float,2> mask_window( currentMask( blitz::Range( i - sw, i + sw ), 
                                                          blitz::Range( j - sw, j + sw ) ) );
          blitz::Array<float,2>::iterator im = mask_window.begin();
          for ( ; iw != iwend; ++iw, ++iind, ++im )
          {
            pijk( *iind ) += c_new * (*iw) * (*im);
            cij += (*im);
          }
        }
        else
        {
          cij += ( sw_extent ) * ( sw_extent );
          for ( ; iw != iwend; ++iw, ++iind )
            pijk( *iind ) += c_new * (*iw);
        }

      }
#else
    for ( int i = sw; i < w - sw; i++ )
      for ( int j = sw; j< h - sw; j++ )
        for ( int k = 0; k < f; k++ )       
      {
          histogram( i, j, k, int(indices(i, j, k)) ) += c_new * currentMask( i, j );
          counter (  i, j )                           +=         currentMask( i, j );
      }
#endif
  }
  else if ( currentData.numElements() > 0 )
  { /// quantized data case
    for ( int i = sw; i < w - sw; i++ )
      for ( int j = sw; j< h - sw; j++ )
      {
        float & cij = counter( i, j );
        blitz::Array<float,1> pij( histogram( i, j, 0, blitz::Range::all() ) );
        blitz::Array<float,3> data_window( currentData( blitz::Range( i - sw, i + sw ), 
                                          blitz::Range( j - sw, j + sw ), blitz::Range::all() ) );

        if ( currentMask.numElements() > 0 )
        {        
          blitz::Array<float,2> mask_window( currentMask( blitz::Range( i - sw, i + sw ), 
                                                          blitz::Range( j - sw, j + sw ) ) );
          pij += c_new * sum( sum( mask_window( i2, i3 ) * weightMatrix( i2, i3 ) * data_window( i2, i3, i1 ), i3 ), i2 );
          cij += c_new * sum( mask_window * weightMatrix );
        }
        else
        {
          
          pij += c_new * sum( sum( weightMatrix( i2, i3 ) * data_window( i2, i3, i1 ), i3 ), i2 );
          cij += ( sw_extent ) * ( sw_extent );
        }
       
      }
    
  }
  else throw std::runtime_error( " updateDensitiesLocal: no data! " );
  
  //std::cout << "c_prior = "<< c_prior << " phist.size = "<< priorHistogram.numElements() << "  pcount.size = " << priorCounter.numElements()<<std::endl;
  if ( c_prior > 0 && priorHistogram.numElements() == histogram.numElements() && priorCounter.numElements() == counter.numElements() )
  {
      std::cout << "prior       : "<<std::flush;
      blitz_dump_dims( priorHistogram );
      std::cout << "histogram   : "<<std::flush;
      blitz_dump_dims( histogram );
      std::cout << "priorCounter: "<<std::flush;
      blitz_dump_dims( priorCounter );
      std::cout << "counter     : "<<std::flush;
      blitz_dump_dims( counter );
      
    if ( any(priorHistogram.shape() != histogram.shape()) )
    {
      std::cout << "prior is strange : "<<std::flush;
      blitz_dump_dims( priorHistogram );
      std::cout << "base histogram   : "<<std::flush;
      blitz_dump_dims( histogram );
    }
    else
    {
      histogram += c_prior * priorHistogram;
      counter   += c_prior * priorCounter;
    }
  }


}






void SurpriseMod::updateDensitiesBuffered()
{
  // This requires keeping a buffer of probability densities over some range of frames
  // This might be computed using recursive filter design selectively choosing frames to ensure a smooth shape
  // Alternatively, the data itself may be saved to save memory at the expense of more computation
  // There will be significant overhead with this approach and as such it will only be implemented if it appears necessary.
  // This can allow the shape of the temporal contribution to the likelihood to be altered unlike the weighted case
  
  throw std::runtime_error("Buffered Density is Not Implemented.");

}








void SurpriseMod::computeCombinedSurprise()
{
  /// Sum across features
  combinedSurprise = sum( marginalSurprise, i3 );
}

CPUData<float> SurpriseMod::returnMarginalSurprise()
{
  CPUData<float> result;
  blitz2cpudata( marginalSurprise, result );
  return result;
}

CPUData<float> SurpriseMod::returnCombinedSurprise()
{
  computeCombinedSurprise();

  CPUData<float> result;
  blitz2cpudata( combinedSurprise, result );
  return result;
}

CPUData<float> SurpriseMod::returnHistogram()
{
  CPUData<float> result;
  blitz2cpudata( histogram, result );
  return result;
}
CPUData<float> SurpriseMod::returnCounter()
{
  CPUData<float> result;
  blitz2cpudata( counter, result );
  return result;
}



void SurpriseMod::set_learn(bool do_learn)
{
  _do_learn = do_learn;
}

void SurpriseMod::set_detect(bool do_detect)
{
  _do_detect = do_detect;
}

void SurpriseMod::update_prior()
{
  //priorHistogram = histogram;
  //priorCounter   = counter;
  
  CPUData<float> tmp;
  blitz2cpudata( histogram, tmp );
  tmp.make_unique();
  cpudata2blitz( tmp, priorHistogram );
  tmp.clear();
  
  blitz2cpudata( counter, tmp );
  tmp.make_unique();
  cpudata2blitz( tmp, priorCounter );
  //std::cout << "update_prior to h="<< priorHistogram.numElements()<< "  c="<<priorCounter.numElements()<<std::endl;
}

void SurpriseMod::reset_prior()
{
  priorHistogram.free();
  priorCounter.free();
}

// In the case of range values, these are currently based on maximum and minimum observations
// Precision might be higher using less bins if percentile values are specified in the file loaded (e.g. 0.01 and 99.9)
// to be robust to outliers

void SurpriseMod::writeRange(const char* fileName)
{
  ofstream output(fileName);
  
  for(int i=0;i<f;i++)
    {
	  output<< minVals(i) <<endl;
	  output<< maxVals(i) <<endl;
    }
  
}

void SurpriseMod::readRange(const char* fileName)
{
  rangeDefined = 1;
  // Set minVals, maxVals to values specified in file - should happen before setCurrentData gets called the first time
  // Unless one wishes to determine the range of values "on the fly" from the data
  
  ifstream output(fileName);
  for(int i=0;i<f;i++)
	{
	  output >> minVals(i);
	  output >> maxVals(i);
	}
  
      // If range is read, rescaling of end limits happens here
           minVals(Range::all())/=limitscale;
           maxVals(Range::all())/=limitscale;


}

// Read/Write observed probability density functions for locations/features to disk

void SurpriseMod::writePriorDensities(const char* fileName)
{
#ifdef HAVE_HDF5	
  HDF5FloatFile outFile;
  try
  {
    outFile.open(fileName, true); 
  }
  catch (...)
  {
    std::cerr << "SurpriseMod::writePriorDensities: failed to write to \""<<fileName<<"\"..."<<std::endl;
    return;
  }
  const size_t fdim[4] = {w,h,f,bins};
  const size_t cdim[2] = {w,h};
  
  outFile.writeDataSet("histogram", histogram.data(), 4, fdim);
  outFile.writeDataSet("counter",   counter.data(),   2, cdim);
  
  outFile.close();
#else
  throw std::runtime_error("persistence for SurpriseMod needs hdf5");
#endif
  
}

void SurpriseMod::readPriorDensities(const char* fileName)
{
#ifdef HAVE_HDF5
  priorDefined = 1;
  
  HDF5FloatFile outFile;
  try
  {
    outFile.open(fileName); 
  }
  catch ( ... )
  {
    std::cerr << "SurpriseMod::readPriorDensities: failed to read to \""<<fileName<<"\"..."<<std::endl;
    return;
  }
  
  int     ndim  = outFile.getDataDimension("histogram");
  size_t* dims  = outFile.getDataSizes("histogram");
  int     ncdim = outFile.getDataDimension("counter");
  size_t* cdims = outFile.getDataSizes("counter");
  
  if ( ndim != 4 || ncdim != 2 || dims[0] != cdims[0] || dims[1] != cdims[1])
    throw std::runtime_error(str(boost::format("SurpriseMod::readPriorDensities: data in \"%s\" has strange dimensionality of %i (4 is expected).")%fileName %ndim ));
  
  
  if ( priorHistogram.numElements() != dims[0]*dims[1]*dims[2]*dims[3] )
    priorHistogram.resize( dims[0], dims[1], dims[2], dims[3] );
  if ( priorCounter.numElements() != cdims[0] * cdims[1])
    priorCounter.resize( cdims[0], cdims[1] );
  
  outFile.readDataSet(priorHistogram.data(),"histogram");
  outFile.readDataSet(priorCounter.data(),"counter");
  outFile.close();
  
  
  if (min(priorHistogram)==0)
    cout << "***** WARNING (void SurpriseMod::readPriorDensities(const char* fileName)): There is a zero value in the priorDensity that has been loaded!!! *****" << std::endl;
#else
  throw std::runtime_error("persistence for Surprise needs hdf5");
#endif
}

CPUData<float> SurpriseMod::discriminantSaliency( blitz::Array<float,4> otherHistogram )
{
  blitz::Array<float,4> avgDist;
  avgDist.resize(w,h,f,bins);
  avgDist=0;
  //cout << "A" << sum(avgDist) << avgDist.shape() << endl;
  //cout << "B" << sum(histogram) << histogram.shape() << endl;
  //cout << "C" << sum(otherNormedDensities) << otherNormedDensities.shape() << endl;
  avgDist = ( histogram + otherHistogram ) / 2.0;
  
  //cout << "D" << sum(avgDist) << endl;
  float P1,P2,Q;
  float epsilon=10e-6;
  CPUData<float> temp(w,h,1);
  std::fill(temp.begin(),temp.end(),0.0);
  for (int i = 0; i<w; i++)
	for (int j=0; j<h; j++)
	  for (int k=0; k<f ; k++)
        for (int l=0; l<bins; l++)
		  {
			P1 = histogram(i,j,k,l);
			P2 = otherHistogram(i,j,k,l);
			Q = avgDist(i,j,k,l);
			if (P1>epsilon && P2 >epsilon && Q> epsilon)
			  temp(i,j,0)+=0.5*P1*log(P1/Q)+0.5*P2*log(P2/Q);
			// temp(i,j,0)=histogram(i,j,24,0);
			//else
			//std::cout << P1 << P2 << Q << "\n";   
		  }
  return temp;
}

