/*
Copyright INRIA 2011

Authors: Neil D.B. Bruce <Neil.Bruce@inria.fr>
Contact: Pierre Kornprobst <Pierre.Kornprobst@inria.fr>

This software is a computer program whose purpose is to estimate saliency.

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/
#include "CPUData.hpp"
#include "GPUImage.h"

#include "sliding_window_iterator.hpp"

//#include "ProbabilityDensity.h"
#include <blitz/array.h>
#include <blitz/tinyvec-et.h>
#include "serialize_blitz.hpp"
#include <blitz_properties.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/binary_object.hpp>

#include "time_logger.hpp"

class SurpriseMod {
  
private:
  
  /// scalar Values
  int w, h, f, bins, type, metric ,rangeDefined, priorDefined, scalingType, sw, step_x, step_y, h_max; /// h_max is for hypoimages
  unsigned int sw_extent;
  float scaleFactor, decayFactor, limitscale;
  bool _do_learn, _do_detect;
  
  /// default Matrices
  blitz::Array<float,4>       histogram;         /// learned features
  blitz::Array<float,2>       counter;           /// counts appearances of features
  blitz::Array<float,2>       globalHistogram;   /// used by mode "global"
  unsigned long               globalCounter;
  
  blitz::Array<unsigned,3>    indices;           /// list of indices from feature to bin
  blitz::Array<unsigned,4>    hypoIndices;       /// list of indices for HpyoImages. has dims H x W x K x 2
  blitz::Array<float,3>       currentData;       /// raw data used when data is interpreted as a histogram
  blitz::Array<float,2>       currentMask;
  blitz::Array<float,4>       currentHypoData;   /// HypoImage of dims H x W x K x 3
  
  
  blitz::Array<float,1>       minVals;           /// min for each feature
  blitz::Array<float,1>       maxVals;           /// max for each feature
  blitz::Array<float,2>       weightMatrix;      /// weighting for spatial support

  /// optional Matrices
  blitz::Array<float,4>       priorHistogram;    /// used by mode "decay" and "recursive"
  blitz::Array<float,2>       priorCounter;
  blitz::Array<float,2>       priorGlobalHistogram;    /// used by mode "decay" and "recursive"
  unsigned long               priorGlobalCounter;
  
  
  /// surprise matrices
  blitz::Array<float,3>       marginalSurprise;  /// surprise for each feature channel
  blitz::Array<float,2>       combinedSurprise;  /// combined surprise over all feature channels


  blitz::firstIndex i1;
  blitz::secondIndex i2;
  blitz::thirdIndex i3;
  blitz::fourthIndex i4;
  blitz::fifthIndex i5;

  TimeLogger _timer;
  
/// basic functions
public: 
                          SurpriseMod();
                          SurpriseMod  ( int w, int h, int feats, int bins, int type, 
                                      float scale_in = 0.0, float decay_in = 0.0, int sw_in = 0,int step_x_in = 1,int step_y_in = 1, float limitscale_in = 1.0,
                                      int metric_in = 0,
                                      CPUData<float> weightMatrix_in = CPUData<float>());
                          ~SurpriseMod();
  void                    init      ( int w_in, int h_in, int feats_in, int bins_in, int type_in, 
                                      float scale_in = 0.0, float decay_in = 0.0, int sw_in = 0,int step_x_in = 1,int step_y_in = 1, float limitscale_in = 1.0,
                                      int metric_in = 0,
                                      CPUData<float> weightMatrix_in = CPUData<float>());
  void                    deinit();
  void                    initDefaultMatrices( CPUData<float> weightMatrix_in = CPUData<float>() );
  void                    initMetric( int metric );
  void                    freeTemporaryData();
  void                    processFeatureData(  CPUData<float> newData,
                                        CPUData<float> confidence );
  void                    processHistogramData(  CPUData<float> newData,
                                        CPUData<float> confidence );
  void                    processHypoData( CPUData<float> hypoData );
  void                    processIndexedData( CPUData<unsigned> newData,
                                        CPUData<float> confidence );

/// general internal functionality
private:

  void                    quantizeData();
  void                    quantizeHypoData();
  void                    dispatch_compute_global();               /// dispatches global compute according to   metric
  void                    dispatch_compute_local();                /// dispatches local compute according to   metric
  void                    computeCombinedSurprise();
  bool                    is_metric_applicable( bool is_quantized );
  void                    check_metric ( bool is_quantized );
  void                    check_if_indices_valid();
  void                    do_filter( const CPUData<float>& input, blitz::Array<float,3>& output, const blitz::Array<float,2>& weights );
  void                    do_filter( const CPUData<float>& input, blitz::Array<float,2>& output, const blitz::Array<float,2>& weights );


/// type specific internal functionality
  void                    updateDensitiesAccumulate();
  void                    updateDensitiesGlobal();
  void                    updateDensitiesWeighted();
  void                    updateDensitiesDecay();
  void                    updateDensitiesBuffered();
  void                    updateNormalizedDensityApprox();
  void                    updateDensitiesLocal();
  void                    updateHypoDensitiesLocal();
  

/// external optional functionality
public:
  CPUData<float>          returnCombinedSurprise();
  CPUData<float>          returnMarginalSurprise();
  CPUData<float>          returnHistogram();
  CPUData<float>          returnCounter();
  blitz::Array<unsigned,3> returnIndex();
  CPUData<float>          discriminantSaliency(blitz::Array<float,4> otherSurprise);
  void                    setScaleType(int scaleType){scalingType = scaleType;};
  int                     getScaleType() const {return scalingType;};
  void                    set_learn(bool do_learn);
  bool                    get_learn() const {return _do_learn;};
  void                    set_detect(bool do_detect);
  bool                    get_detect() const {return _do_detect;};
  void                    set_sw(int sw_in){ sw=sw_in; sw_extent= 2 * sw + 1; };
  int                     get_sw(){return sw;};
  void                    set_decay( float decay ){ decayFactor = decay; };
  float                   get_decay() const{ return decayFactor; };
  void                    update_prior();
  void                    reset_prior();


/// save/load functionality
  void                    writeRange(const char* fileName);
  void                    readRange(const char* fileName);
  void                    writePriorDensities(const char* fileName);
  void                    readPriorDensities(const char* fileName);



/// boost serialization code, for automatic saving/loading
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & BOOST_SERIALIZATION_NVP( type );
    ar & BOOST_SERIALIZATION_NVP( metric );
    ar & BOOST_SERIALIZATION_NVP( w );
    ar & BOOST_SERIALIZATION_NVP( h );
    ar & BOOST_SERIALIZATION_NVP( f );
    ar & BOOST_SERIALIZATION_NVP( bins );
    ar & BOOST_SERIALIZATION_NVP( rangeDefined );
    ar & BOOST_SERIALIZATION_NVP( priorDefined );
    ar & BOOST_SERIALIZATION_NVP( scalingType );
    ar & BOOST_SERIALIZATION_NVP( sw );
    ar & BOOST_SERIALIZATION_NVP( sw_extent );
    ar & BOOST_SERIALIZATION_NVP( step_x );
    ar & BOOST_SERIALIZATION_NVP( step_y );
    ar & BOOST_SERIALIZATION_NVP( scaleFactor );
    ar & BOOST_SERIALIZATION_NVP( decayFactor );
    ar & BOOST_SERIALIZATION_NVP( limitscale );
    ar & BOOST_SERIALIZATION_NVP( _do_learn );
    ar & BOOST_SERIALIZATION_NVP( _do_detect );
    
    ar & BOOST_SERIALIZATION_NVP( histogram );
    ar & BOOST_SERIALIZATION_NVP( counter );
    ar & BOOST_SERIALIZATION_NVP( indices );
    ar & BOOST_SERIALIZATION_NVP( minVals );
    ar & BOOST_SERIALIZATION_NVP( maxVals );
    ar & BOOST_SERIALIZATION_NVP( weightMatrix );
    ar & BOOST_SERIALIZATION_NVP( priorHistogram );
    ar & BOOST_SERIALIZATION_NVP( globalHistogram );
    ar & BOOST_SERIALIZATION_NVP( globalCounter );
    ar & BOOST_SERIALIZATION_NVP( priorGlobalHistogram );
    ar & BOOST_SERIALIZATION_NVP( priorGlobalCounter );
    ar & BOOST_SERIALIZATION_NVP( priorCounter );
    ar & BOOST_SERIALIZATION_NVP( marginalSurprise );
    ar & BOOST_SERIALIZATION_NVP( combinedSurprise );
  }






/// metrics
private:
/// abstract base struct for metrics
/*struct metric_base
{
  static float compute( const float& h, const float& p, const float& Z ) const;
};
*/

static float my_log( float x )
{
  return std::log( std::max( 1e-10f, x ) );
}

/// wrapper for self information
struct metric_si
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    return - p * my_log( h / Z );
  }
};

/// wrapper for self information
struct metric_js
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
	float q = h / Z;
	float m = ( p + q ) / 2;
    return  0.5 * ( p * ( my_log( p) - my_log( m ) ) +  q * ( my_log( q) - my_log( m ) ) );
  }
};

/// wrapper for self information
struct metric_bjs
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
	float q = h / Z;
	float m = ( p + q ) / 2;
    return  0.5 * ( p     * ( my_log( p ) - my_log( m ) ) +  q     * ( my_log( q ) - my_log( m ) ) ) + 
	        0.5 * ( (1-p) * ( my_log(1-p) - my_log(1-m) ) +  (1-q) * ( my_log(1-q) - my_log(1-m) ) );
  }
};

/// wrapper for modified self information
struct metric_modsi //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    return - ( p - h / Z ) * my_log( h / Z );
  }
};

struct metric_bsi
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z <= 0 ) return 0;
    float p0 = h / Z;
    return - p * my_log( p0 ) - ( 1 - p ) * my_log( 1 - p0 );
  }
};
/// wrapper for modified self information
struct metric_bmodsi //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    return - ( p - h / Z ) * ( my_log( h / Z ) - my_log( 1 - h / Z ) );
  }
};

/// wrapper for zero-enforcing kl-divergence
struct metric_kl //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    return p * ( my_log( p ) - my_log( h / Z ) );
  }
};

/// wrapper for zero-enforcing kl-divergence
struct metric_bkl //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    float q = 1 - p;
    float p0 = h / Z;
    float q0 = 1 - p0;
    return p * ( my_log( p ) - my_log( p0 ) ) +
           q * ( my_log( q ) - my_log( q0 ) );
  }
};
/// wrapper for zero-avoiding kl-divergence
struct metric_ikl //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    return (h / Z) * ( my_log( h / Z ) - my_log( p ) );
  }
};
/// wrapper for zero-enforcing kl-divergence
struct metric_bikl //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    float q = 1 - p;
    float p0 = h / Z;
    float q0 = 1 - p0;
    return p0 * ( my_log( p0 ) - my_log( p ) ) +
           q0 * ( my_log( q0 ) - my_log( q ) );
  }
};
/// wrapper for modified self information
struct metric_symkl //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    return -0.5f * ( p - h / Z ) * ( my_log( p ) - my_log( h / Z ) );
  }
};
/// wrapper for zero-enforcing kl-divergence
struct metric_bsymkl //: metric_base
{
  static float compute( const float& h, const float& p, const float& Z )
  {
    if ( Z == 0 ) return 0;    
    float q = 1 - p;
    float p0 = h / Z;
    float q0 = 1 - p0;
    return -0.5f * ( ( p - p0 ) * ( my_log( p ) - my_log( p0 ) ) +
                     ( q - q0 ) * ( my_log( q ) - my_log( q0 ) )  );
  }
};

template< typename Metric >
void computeMarginalSurpriseGlobal(  )
{
  marginalSurprise = 0.f;
  if ( indices.numElements() > 0 )
  { /// indexed data case
    for (int i=0; i<w; i++)
      for (int j=0; j<h; j++)
        for (int k=0; k<f; k++)
          marginalSurprise(i,j,k) = Metric::compute( globalHistogram(k,int(indices(i,j,k))), 1, globalCounter ); //  -log(globalDensities(k,indices(i,j,k)));
  }
  else if ( currentData.numElements() > 0 )
  { /// quantized data case
    for (int i=0; i<w; i++)
      for (int j=0; j<h; j++)
        for (int b=0; b<bins; b++)
          marginalSurprise(i,j,0) += Metric::compute( globalHistogram(0,b), currentData(i,j,b) , globalCounter );   //  globalDensities(0,indices(i,j,k)), 1, counter(i,j) );
  }
  else throw std::runtime_error( " computeMarginalSurpriseGlobal: no data! " );
  
  if ( currentMask.numElements() > 0 )
    marginalSurprise *= currentMask( i1, i2 );
}

template< typename Metric >
void computeMarginalSurpriseLocal(  )
{
  //CPUData<float> test;
  //blitz2cpudata( marginalSurprise, test );
  //std::cout << "  marginalSurprise before: mean="<< test.mean() << std::flush;
  marginalSurprise = 0.f;
  //blitz2cpudata( marginalSurprise, test );
  //std::cout << "  marginalSurprise  after: mean="<< test.mean() << std::endl;
  
  if ( indices.numElements() > 0 )
  { /// indexed data case
    for (int i=sw; i<w-sw; i++)
      for (int j=sw; j<h-sw; j++)
        for (int k=0; k<f; k++)
          marginalSurprise(i,j,k) = Metric::compute( histogram(i,j,k,int(indices(i,j,k))), 1, counter(i,j) ); //  -log(histogram(i,j,k,indices(i,j,k)) / counter(i,j)) : 0;
  }
  else if ( currentData.numElements() > 0 )
  { /// quantized data case

    for (int i=sw; i<w-sw; i++)
      for (int j=sw; j<h-sw; j++)
        for (int b=0; b<bins; b++)
        {
          marginalSurprise(i,j,0) += Metric::compute( histogram(i,j,0,b), currentData(i,j,b), counter(i,j) ); //  -log(histogram(i,j,k,indices(i,j,k)) / counter(i,j)) : 0;
          //for (int dx = i-sw; dx < i+sw; ++dx)
          //  for (int dy = j-sw; dy < j+sw; ++dy)
          //    marginalSurprise(i,j,0) += Metric::compute( histogram(dx,dy,0,b), currentData(dx,dy,b), counter(dx,dy) ); //  -log(histogram(i,j,k,indices(i,j,k)) / counter(i,j)) : 0;
        }
  }
  else throw std::runtime_error( " computeMarginalSurpriseGlobal: no data! " );
  
  if ( currentMask.numElements() > 0 )
    marginalSurprise *= currentMask( i1, i2 );
}
 
};
