/*

COMMON

  THIS MODULE INCLUDES GENERAL SUPPORT FILES THAT DID NOT FIND THEIR OWN PLACE
  THIS INCLUDES HELPERS FOR DEALING WITH BLITZ ARRAS, CONCURRENCY, PROFILING
  GSL, UBLAS, SERIALIZATION, AND A SIMPLE YET EFFECTIVE MIXIN CLASS FOR PROVIDING
  ARBITRARY C++ CLASSES WITH ATTRIBUTES THAT CAN BE QUERIED IN TEXT FORM TO 
  EASE SCRIPTING

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef BLITZ_PROPERTIES_HPP
#define BLITZ_PROPERTIES_HPP

#include "SeariseProperties.hpp"
#include <blitz/array.h>
#include <iostream>

template< typename T, int R >
bool is_flat_and_contiguous( const blitz::Array<T, R> & bdata )
{
  int expected_stride = 1;
  for ( int dim = R-1; dim >= 0; --dim )
	{
	  if ( bdata.stride(dim) != expected_stride )
		return false;
	  expected_stride *= bdata.extent(dim);
	}
  return true;
}

template< typename T, int R >
bool blitz_is_nan( const blitz::Array<T, R> & bdata )
{
  BOOST_FOREACH( const T& val, bdata )
  {
    if ( isnan( val ) )
      return true;
  }
  return false;
}

template< typename T, int R >
void blitz_dump_dims( const blitz::Array<T, R> & bdata, bool newline = true )
{
  blitz::TinyVector<int,R> shape = bdata.shape();
  std::cout << "[ ";
  for (int r = 0; r < R-1; ++r)
  {
    std::cout << shape[r] << " x ";
  }
  if ( newline )
    std::cout << shape[R-1] <<" ]"<< std::endl;
  else
    std::cout << shape[R-1] <<" ]"<< std::flush;
}

template< typename T, int R >
std::string blitz_dump_dims_str( const blitz::Array<T, R> & bdata, bool newline = false )
{
  std::string result("[ ");
  blitz::TinyVector<int,R> shape = bdata.shape();
  for (int r = 0; r < R-1; ++r)
  {
    result = result + boost::lexical_cast<std::string>(shape[r]) + " x ";
  }
  if ( newline )
    result = result + boost::lexical_cast<std::string>(shape[R-1]) +" ]\n";
  else
    result = result + boost::lexical_cast<std::string>(shape[R-1]) +" ]";
  return result;
}

/*
template< typename T, int R >
bool operator==( const blitz::TinyVector<T,R> & left, const blitz::TinyVector<T,R> & right )
{
  for (int r = 0; r < R; ++r)
    if ( left[r] != right[r] );
      return false;
  return true;
}

template< typename T, int R >
bool operator!=( const blitz::TinyVector<T,R> & left, const blitz::TinyVector<T,R> & right )
{
  for (int r = 0; r < R; ++r)
    if ( left[r] != right[r] );
      return true;
  return false;
}
*/
template< typename T, int rank >
inline void blitz2cpudata ( const blitz::Array<T, rank> & bdata, CPUData<T> & cdata, bool copy = true )
{
  if ( bdata.numElements() == 0 )
	{
	  cdata.clear();
	  return;
	}

  int n1 = rank > 0 ? bdata.extent(0) : 1;
  int n2 = rank > 1 ? bdata.extent(1) : 1;
  int n3 = 1;
  std::vector<size_t> column_shape( rank > 2 ? rank - 2 : 0 );
  for ( int i = 2; i < rank; ++i )
	{
	  n3 *= bdata.extent(i);
	  column_shape[i-2] = bdata.extent(i);
	}

  if ( !copy && is_flat_and_contiguous( bdata ) )
	{
	  cdata = CPUData<T>( const_cast< CPUDataBlock<T> * >( bdata.block() ), n1, n2, n3 );
	}
  else
	{
	  cdata.resize( n1, n2, n3 );
	  cdata.make_contiguous();
	  std::copy( bdata.begin(), bdata.end(), cdata.data() );
	}
  if ( rank > 3 )
	cdata.set_column_shape( column_shape );
}

template< typename T1, typename T2, int rank >
inline void blitz2cpudata ( const blitz::Array<T1, rank> & bdata, CPUData<T2> & cdata )
{
  int n1 = rank > 0 ? bdata.extent(0) : 1;
  int n2 = rank > 1 ? bdata.extent(1) : 1;
  int n3 = 1;
  std::vector<size_t> column_shape( rank > 2 ? rank - 2 : 0 );
  for ( int i = 2; i < rank; ++i )
	{
	  n3 *= bdata.extent(i);
	  column_shape[i-2] = bdata.extent(i);
	}

	{
	  cdata.resize( n1, n2, n3 );
	  cdata.make_contiguous();
	  std::copy( bdata.begin(), bdata.end(), cdata.data() );
	}

  if ( rank > 3 )
	cdata.set_column_shape( column_shape );
}

template<class C, int rank, typename T>
struct BlitzIVarGetter1
{
  typedef HasPropertiesBase<SearisePropertyTypes> Base;
  typedef blitz::Array<T, rank>  Matrix;
  BlitzIVarGetter1( Matrix C::* _ivar )
	: ivar(_ivar) 
  {}
  Matrix C::* ivar;
  SeariseProperty operator()( const Base *c )
  {
	const Matrix & m = dynamic_cast<const C*>(c)->* ivar;
	CPUData<float> cdata;
	blitz2cpudata( m, cdata );
	return cdata;
  }
};

template<class C, int rank, typename T>
static BlitzIVarGetter1<C, rank, T>
make_blitz_ivar_getter1( blitz::Array<T, rank> C::* ivar )
{
  return BlitzIVarGetter1<C, rank, T>( ivar );
}

template<typename T, int rank>
static void cpudata2blitz ( CPUData<T> cdata, blitz::Array<T, rank> & bdata, 
							bool may_rearrange_cdata = false, bool copy_data = true )
{
  if ( cdata.empty() )
	bdata.free();
  else
	{
	  if ( !cdata.contiguous() ) 
		  cdata.make_contiguous();
	  if ( copy_data )
		cdata.make_unique();
	  blitz::TinyVector<int,rank> shape;
	  shape[0] = cdata.rows();
	  if ( rank > 1 )
		{
		  shape[1] = cdata.columns();
		  if ( rank == 2 )
			shape[1] *= cdata.depth();
		  else
			{
			  if ( rank <= 3 )
				shape[2] = cdata.depth();
			  else
				{
				  std::cout << "cpudata2blitz rank = " << rank << " : " << cdata.dump_dims() << std::endl;
				  typename CPUData<T>::ColumnShapeT column_shape = cdata.column_shape();
				  if ( column_shape.size() == 0 )
					{
					  shape[2] = cdata.depth();
					  for ( int i = 3; i < rank; ++i )
						shape[i] = 1;
					}
				  else
					{
					  for ( int i = 2; i < std::min( rank, int( column_shape.size() + 2 ) ); ++i )
						shape[i] = column_shape[i-2];
					  for ( int i = std::min( rank, int( column_shape.size() + 2 ) ); i < rank; ++i )
						shape[i] = 1;
					  if ( column_shape.size() > rank )
						for ( int i = rank; i < int( column_shape.size() + 2 ); ++i )
						  shape[rank-1] *= column_shape[i-2];
					}
				}
			}
		}
	  bdata.reference( blitz::Array<T,rank>( cdata.block(), shape, cdata.weak() ) );
	}
}

template<class C, typename T, int rank>
struct BlitzIVarSetter1
{
  typedef HasPropertiesBase<SearisePropertyTypes> Base;
  typedef blitz::Array<T, rank>  Matrix;
  typedef CPUData<T> CData;
  BlitzIVarSetter1( Matrix C::* _ivar, bool _copy = false )
	: ivar(_ivar), copy(_copy) {}
  Matrix C::* ivar;
  bool copy;
  boost::function1<void,Base*> sideffect;
  void operator()( Base * b, SeariseProperty v)
  {
	if (b)
      {
		CData cdata = cast<CData>( v );
		C * c = dynamic_cast<C*>(b);
		Matrix & matrix = c->* ivar;
		cpudata2blitz( cdata, matrix, false, copy );
	  }
  }
};
template<class C,typename T, int rank>
static BlitzIVarSetter1<C,T,rank> make_blitz_ivar_setter1 ( blitz::Array<T, rank> C::* ivar, bool copy = false )
{
  return BlitzIVarSetter1<C,T,rank>(ivar,copy);
}

#endif
