/*

COMMON

  THIS MODULE INCLUDES GENERAL SUPPORT FILES THAT DID NOT FIND THEIR OWN PLACE
  THIS INCLUDES HELPERS FOR DEALING WITH BLITZ ARRAS, CONCURRENCY, PROFILING
  GSL, UBLAS, SERIALIZATION, AND A SIMPLE YET EFFECTIVE MIXIN CLASS FOR PROVIDING
  ARBITRARY C++ CLASSES WITH ATTRIBUTES THAT CAN BE QUERIED IN TEXT FORM TO 
  EASE SCRIPTING

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#pragma once

// from http://www.justsoftwaresolutions.co.uk/threading/implementing-a-thread-safe-queue-using-condition-variables.html
// extended by supporting maximum load on the queue, using a second coindition variable
// for the necce ssary comminication from output to input



#include <queue>
#include <boost/optional.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/condition_variable.hpp>

template<typename Data>
class concurrent_queue
{

private:

  std::queue< boost::optional<Data> > _queue;
  mutable boost::mutex _mutex;
  boost::condition_variable _cond_in;
  boost::condition_variable _cond_out;
  size_t _max_load;
  bool _should_terminate;

public:

  concurrent_queue( size_t max_load = 1 )
	: _max_load( max_load )
	, _should_terminate(false)
  {
  }

  void terminate_out()
  {
    push( boost::none );
  }

  // to use by the output to make the input terminate
  void terminate_in()
  {	
	boost::mutex::scoped_lock lock(_mutex);
	_should_terminate = true;
	lock.unlock();
	_cond_out.notify_one();
  }

  bool push( boost::optional<Data> data)
  {
	boost::mutex::scoped_lock lock(_mutex);
	
	// max load handling
	if ( _max_load > 0 )
	  while( _queue.size() >= _max_load && !_should_terminate )
		_cond_out.wait(lock);

	//termination handling
	if ( _should_terminate )
	  return false;

	//the actual work
	_queue.push(data);
	lock.unlock();
	_cond_in.notify_all();
	return true;
  }
  
  bool empty() const
  {
	boost::mutex::scoped_lock lock(_mutex);
	return _queue.empty();
  }
  
  bool pop( Data& popped_value, bool wait = true )
  {
	boost::mutex::scoped_lock lock(_mutex);
	bool retval = true;

	while( wait && _queue.empty() && !_should_terminate )
	  _cond_in.wait(lock);
	
	if ( _queue.empty() || !_queue.front() )
	  {
		retval = false;
	  }
	else
	  {
		popped_value = *_queue.front();
		_queue.pop();
	  }
	
	lock.unlock();
	_cond_out.notify_one();

	return retval;
  }
  
  
  void clear()
  {
    boost::mutex::scoped_lock lock(_mutex);
    _should_terminate = true;

    lock.unlock();
    _cond_out.notify_all();
    _cond_in.notify_all();
  }
  
  size_t size() const
  {
    boost::mutex::scoped_lock lock(_mutex);
    return _queue.size();
  }
  
  size_t capacity() const
  {
    boost::mutex::scoped_lock lock(_mutex);
    return _max_load;
  }
};
