#ifndef NUMPY_SUPPORT
#define NUMPY_SUPPORT

#include "Python.h"

extern "C" {
#include<numpy/arrayobject.h>
}

#include "CPUData.hpp"
#include "SeariseProperties.hpp"
#include "boost/format.hpp"

struct PyCPUDataFloatKeeper 
{
  PyObject_HEAD
  CPUData<float> cpudata;
  static PyTypeObject type;
  PyObject * py_new( CPUData<float> & data )
  {
	PyCPUDataFloatKeeper * self = new PyCPUDataFloatKeeper(data);
	self->ob_refcnt = 0;
	self->ob_type = &type;
#ifdef Py_TRACE_REFS
	self->_ob_next = 0;
	self->_ob_prev = 0;
#endif
	return (PyObject *)self;
  }
private:
  PyCPUDataFloatKeeper( CPUData<float> & data ) 
	: cpudata(data)
  {}
};

struct python_gil_manager;

struct python_object_keeper
{
  python_object_keeper( PyObject * po, python_gil_manager * accessor = 0 );
  python_object_keeper( const python_object_keeper & other );
  ~python_object_keeper();
private:
  python_object_keeper() {}
  PyObject * _pyobject;
  python_gil_manager * _accessor;
};



struct numpy_to_CPUData_converter
{
  template< typename T, typename T2 >
  static CPUData<T2> make_cpudata( const T * data, size_t rank, const npy_intp * dims, const npy_intp * strides )
  {
	// extract cpudata shape information
	size_t n1 = rank > 0 ? dims[0] : 1;
	size_t n2 = rank > 1 ? dims[1] : 1;
	size_t n3 = rank > 2 ? dims[2] : 1;
	for ( size_t dim = 3; dim < rank; ++dim )
	  n3 *= dims[dim];
	std::vector<size_t> column_shape( std::max( 0, int(rank) - 2 ) );
	if ( rank > 0 )
	  std::copy( dims + 2, dims + rank, column_shape.begin() );
	// create the cpudata object
	CPUData<T2> result( n1, n2, n3 );
	result.set_column_shape( column_shape );
	// iterate the data and copy it into the cpudata object
	ptrdiff_t pos[rank];
	T2 * target = result.data();
	if ( rank == 0 )
	  *target = *data;
	else
	  {
		std::fill(pos, pos + rank, 0);
		while( pos[0] < int(n1) )
		  {
			ptrdiff_t offset = 0;
			for ( size_t dim = 0; dim < rank; ++dim )
			  {
				offset += pos[dim] * ( strides[dim] / sizeof(T) );
			  }
			*target++ = data[offset];
			for ( ptrdiff_t dim = rank - 1; dim >= 0; --dim )
			  {
				++pos[dim];
				if ( pos[dim] >= dims[dim] && dim > 0 )
				  pos[dim] = 0;
				else
				  break;
			  }		  
		  }
	  }
	// done...
	return result;
  }

  static bool representable_as_cpudata( const PyArrayObject * a )
  {
	npy_intp ndims = a->nd;
	npy_intp * dims = a->dimensions;
	npy_intp * strides = a->strides;
	npy_intp expected_stride = 1;
	// all but the left-most dim must be row-major-contiguous
	for ( npy_intp dim = ndims - 1; dim > 0; --dim )
	  {
		npy_intp stride =  strides[dim];
		npy_intp extent = dims[dim];
		if ( stride != expected_stride ) return false;
		expected_stride *= extent;
	  }
	return true;
  }

  template<typename T> static CPUData<T> cpudata( PyObject * c, python_gil_manager * manager = 0 )
  {
	PyArrayObject * a = (PyArrayObject*) c;
	
	switch( PyArray_TYPE( a ) )
	  {
		// handles all the real (i.e. non-complex) numeric types
#define HANDLE_TYPE( typenum, target_type, output_type )							\
		case typenum: return make_cpudata<target_type,output_type>( (target_type*) a->data, a->nd, a->dimensions, a->strides )
		
		HANDLE_TYPE( NPY_BOOL, bool, float );
		HANDLE_TYPE( NPY_BYTE, char, float );
		HANDLE_TYPE( NPY_UBYTE, unsigned char,float );
		HANDLE_TYPE( NPY_SHORT, short, float );
		HANDLE_TYPE( NPY_USHORT, unsigned short, float );
		HANDLE_TYPE( NPY_INT, int, float );
		HANDLE_TYPE( NPY_UINT, unsigned int, float );
		HANDLE_TYPE( NPY_LONG, long, float );
		HANDLE_TYPE( NPY_ULONG, unsigned long, float );
		HANDLE_TYPE( NPY_LONGLONG, long long, double );
		HANDLE_TYPE( NPY_ULONGLONG, unsigned long long, double );
		//HANDLE_TYPE( NPY_FLOAT, float );
		//HANDLE_TYPE( NPY_DOUBLE, double );

	  case NPY_FLOAT:
		{
		  return make_cpudata<float,float>( (float*) a->data, a->nd, a->dimensions, a->strides );
		}

	  case NPY_DOUBLE:
		{
		  return make_cpudata<double,double>( (double*) a->data, a->nd, a->dimensions, a->strides );
		}

	  }

#undef HANDLE_TYPE

	throw std::runtime_error("unsupported numarray type");
  }

  static SeariseProperty execute( PyObject * c, python_gil_manager * manager = 0 )
  {
	PyArrayObject * a = (PyArrayObject*) c;
	
	switch( PyArray_TYPE( a ) )
	  {
		// handles all the real (i.e. non-complex) numeric types
#define HANDLE_TYPE( typenum, target_type, output_type )							\
		case typenum: return make_cpudata<target_type,output_type>( (target_type*) a->data, a->nd, a->dimensions, a->strides )
		
		HANDLE_TYPE( NPY_BOOL, bool, float );
		HANDLE_TYPE( NPY_BYTE, char, float );
		HANDLE_TYPE( NPY_UBYTE, unsigned char,float );
		HANDLE_TYPE( NPY_SHORT, short, float );
		HANDLE_TYPE( NPY_USHORT, unsigned short, float );
		HANDLE_TYPE( NPY_INT, int, float );
		HANDLE_TYPE( NPY_UINT, unsigned int, float );
		HANDLE_TYPE( NPY_LONG, long, float );
		HANDLE_TYPE( NPY_ULONG, unsigned long, float );
		HANDLE_TYPE( NPY_LONGLONG, long long, double );
		HANDLE_TYPE( NPY_ULONGLONG, unsigned long long, double );
		//HANDLE_TYPE( NPY_FLOAT, float );
		//HANDLE_TYPE( NPY_DOUBLE, double );

	  case NPY_FLOAT:
		{
		  return make_cpudata<float,float>( (float*) a->data, a->nd, a->dimensions, a->strides );
		}

	  case NPY_DOUBLE:
		{
		  return make_cpudata<double,double>( (double*) a->data, a->nd, a->dimensions, a->strides );
		}

	  }

#undef HANDLE_TYPE

	throw std::runtime_error("unsupported numarray type");
  }
};

// helpers from hippodraw
#ifdef _MSC_VER
template <typename T>  inline PyArray_TYPES get_enum ();
#else
template<typename T> inline PyArray_TYPES get_enum(void)
{
  throw std::runtime_error(str(boost::format("no numpy mapping available for type") % typeid(T).name() ) );
}
#endif

template<> inline PyArray_TYPES get_enum<float>(void) { return PyArray_FLOAT; }
template<> inline PyArray_TYPES get_enum<double>(void) { return PyArray_DOUBLE; }
template<> inline PyArray_TYPES get_enum<unsigned>(void) { return PyArray_UINT; }

template<typename T>
inline PyObject* cpudata2python(  CPUData<T> & cdata )
{
  if ( cdata.empty() ) return NULL;
  size_t ndims = 2 + std::max( size_t(1), cdata.column_shape().size() );
  npy_intp dims[ndims];
  
  dims[0] = cdata.rows();
  dims[1] = cdata.columns();
  if ( ndims > 3 )
    for ( size_t i = 0; i < ( ndims - 2 ); ++i )
      dims[i+2] = cdata.column_shape()[i];
  else
    dims[2] = cdata.depth();
  
  // reduce dimensionality
  /*
    for ( int i = ndims - 1; i >= 0; --i )
    {
    if ( dims[i] == 1 )
    {
    for ( size_t j = i; j < (ndims - 1); ++j )
    dims[j] = dims[j+1];
    --ndims;
    }
    }
  */
  PyObject * result = PyArray_New( &PyArray_Type, ndims, dims, get_enum<T>(), NULL, NULL, sizeof(T), NULL, NULL );
  T * datap = (T*) PyArray_BYTES( result );
  if ( cdata.contiguous() )
    std::copy( cdata.data(), cdata.data_end(), datap );
  else
    {
      // rows are contiguous, so we copy a row at a time
      T * datap = cdata.data();
      size_t n_row = cdata.depth() * cdata.columns();
      for ( size_t row = 0; row < cdata.rows(); ++row )
	{
	  T * startp = &cdata( row, 0, 0 );
	  std::copy( startp, startp + n_row, datap );
	  datap += n_row;
	}
    }
  return result;
}

#endif
