/*

COMMON

  THIS MODULE INCLUDES GENERAL SUPPORT FILES THAT DID NOT FIND THEIR OWN PLACE
  THIS INCLUDES HELPERS FOR DEALING WITH BLITZ ARRAS, CONCURRENCY, PROFILING
  GSL, UBLAS, SERIALIZATION, AND A SIMPLE YET EFFECTIVE MIXIN CLASS FOR PROVIDING
  ARBITRARY C++ CLASSES WITH ATTRIBUTES THAT CAN BE QUERIED IN TEXT FORM TO 
  EASE SCRIPTING

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/* CircularFifo.h 
* Not any company's property but Public-Domain
* Do with source-code as you will. No requirement to keep this
* header if need to use it/change it/ or do whatever with it
*
* Note that there is No guarantee that this code will work 
* and I take no responsibility for this code and any problems you
* might get if using it. The code is highly platform dependent!
*
* Code & platform dependent issues with it was originally 
* published at http://www.kjellkod.cc/threadsafecircularqueue
* 2009-11-02
* @author Kjell Hedström, hedstrom@kjellkod.cc */

#ifndef CIRCULARFIFO_H_
#define CIRCULARFIFO_H_

#include <vector>
#include <string>

/** Circular Fifo (a.k.a. Circular Buffer) 
* Thread safe for one reader, and one writer */
template<typename Element>
class CircularFifo {
public:

  typedef std::vector<Element> storage_t;
  typedef typename storage_t::iterator iterator;
  typedef typename storage_t::const_iterator const_iterator;

   CircularFifo(size_t size = 0) 
   {
     init( size );
   }
   virtual ~CircularFifo() {}

   bool push();
   bool pop();
   
   Element & get_tail() { return elements[tail]; }
   Element & get_head() { return elements[head]; }

   iterator begin() { return elements.begin(); }
   iterator end()   { return elements.end(); }

   bool empty() const { return (head == tail); }
   bool full() const { return ((tail+1) % elements.size() == head); }
   
   size_t space_available()
   {
     size_t t = tail, h = head;
     if ( h > t ) return h - t;
     if ( h < t ) return elements.size() - ( t - h );
     return elements.size();
   }
   
   std::string dump_status( char filled_char = 'X', char empty_char = '-' )
   {
       int cnt_empty = space_available();
       std::string empty( cnt_empty, empty_char );
       std::string filled( size()-cnt_empty, filled_char );
       return std::string( "[" ).append(filled).append(empty).append("]");
   }
   
   std::string dump( )
   {
        char result[size()+2];
        for (int i = 0; i < size()+2; ++i)
            result[i] = '-';
        result[    0     ] = '[';
        result[ size()+1 ] = ']';
        result[   head+1 ] = 'H';
        result[   tail+1 ] = 'T';
        if ( head == tail ) result[   head+1 ] = '0';
        return std::string( result );
   }
    
   void init( size_t n )
   {
     elements.clear();
     elements.resize( n );
     tail = head = 0;
   }
   
   size_t size()
   {
        return elements.size();
   }

   void consumer_busy_wait( size_t n_micros = 5000 )
   {
      while( empty() ) usleep( n_micros );
   }
      
   void producer_busy_wait( size_t n_micros = 5000 )
   {
      while( full() ) usleep( n_micros );
   }
   
private:

   volatile size_t tail; // input index
   volatile size_t head; // output index
   std::vector<Element> elements;

   size_t increment(size_t idx_) const
   {
     idx_ = (idx_+1) % elements.size();
     return idx_;
   }

};




/** Producer only: Adds item to the circular queue. 
* If queue is full at 'push' operation no update/overwrite
* will happen, it is up to the caller to handle this case
*
* \param item_ copy by reference the input item
* \return whether operation was successful or not */
template<typename Element>
bool CircularFifo<Element>::push()
{
   size_t nextTail = increment(tail);
   if(nextTail != head)
   {
      tail = nextTail;
      return true;
   }

   // queue was full
   return false;
}

/** Consumer only: Removes and returns item from the queue
* If queue is empty at 'pop' operation no retrieve will happen
* It is up to the caller to handle this case
*
* \param item_ return by reference the wanted item
* \return whether operation was successful or not */
template<typename Element>
bool CircularFifo<Element>::pop()
{
   if(head == tail)
      return false;  // empty queue

   head = increment(head);
   return true;
}


#endif /* CIRCULARFIFO_H_ */
