/*

COMMON

  THIS MODULE INCLUDES GENERAL SUPPORT FILES THAT DID NOT FIND THEIR OWN PLACE
  THIS INCLUDES HELPERS FOR DEALING WITH BLITZ ARRAS, CONCURRENCY, PROFILING
  GSL, UBLAS, SERIALIZATION, AND A SIMPLE YET EFFECTIVE MIXIN CLASS FOR PROVIDING
  ARBITRARY C++ CLASSES WITH ATTRIBUTES THAT CAN BE QUERIED IN TEXT FORM TO 
  EASE SCRIPTING

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef TIME_LOGGER_H
#define TIME_LOGGER_H

#include <iostream>
#include <string>
#include <fstream>
#include "get_secs.hpp"

/// this class provides a logging system to write timings and compare them
/// functioning:
///   constructor takes a filename and an index, needed to separate tasks/ tracked events, as default with absolute time-of-day values 

///   start(): begins a timing event, meant to be invoked directly BEFORE function to be timed
///   end():   ends to time an event, meant to be invoked directly AFTER  function to be timed
///   NOTE: setting the flag simple results in output of form <time>,<dt> 

///   tic():   writes the time since last tic


class TimeLogger
{
public:
    TimeLogger( )
    : _use_start_end( true )
    , _start_time( get_secs() )
    , _cur_time( get_secs() )
    , _last_time( get_secs() )
    , _running(false)
    , _tic_eps( 1e-05 )
    {
	  reset_stats();
    }
    
    TimeLogger( const TimeLogger & src )
    : _filename( src._filename )
    , _use_start_end( src._use_start_end )
    , _start_time( src._start_time )
    , _cur_time( src._cur_time )
    , _last_time( src._last_time )
    , _running( src._running )
    , _tic_eps( src._tic_eps )
    {
	  reset_stats();
        if ( _filename != "" )
            _file.open( _filename.c_str() );
    }
    
    TimeLogger( const std::string& filename, bool use_start_end = true )
    {
	  init ( filename, use_start_end );
    }
    
    
    void init( const std::string& filename, bool use_start_end = true )
    {
        _filename = filename;
        _use_start_end = use_start_end;
        _start_time =  get_secs();
        _cur_time = get_secs();
        _running = false;
		reset_stats();
        
        if ( _filename != "" )
        {
            _file.open( _filename.c_str() );
            _file.precision( 24 );
        }
    }
    
    double start()
    {
        if (  !_use_start_end ) 
            throw std::runtime_error("error in TimeLogger: wrong mode!");
        
        _running = true;
        
        _cur_time = get_secs();
        _last_time = _cur_time;
        
        return _cur_time;
    }
    
    double end()
    {
        if ( !_running || !_use_start_end ) 
            throw std::runtime_error("error in TimeLogger: double end or wrong mode!");
        else
            _running = false;
        
        _cur_time = get_secs();
        
        double dt = _cur_time - _last_time;
        if (_file.is_open() ) _file << _last_time << "," << dt << std::endl;
            
		accumulate_dt( dt );

        return dt;
    }

    bool running() const
    {
	  return _running;
    }

    double_t std() const
    {
	  return std::sqrt( std::max( double_t(0), _sum_dt2 / std::max( double_t(1), _n ) - mean() * mean() ) );
    }    

    double_t mean() const
    {
	  return _sum_dt / std::max( double_t(1), _n );
    }

    double tic()
    {	  
      if ( _use_start_end )
      {
          if (_file.is_open() ) _file << get_secs() << "," << _tic_eps << std::endl;
          if ( !_running )
          {
              start();
              return 0;
          }
          else
          {
			  double_t dt = get_secs() - _last_time;
			  accumulate_dt( dt );
			  return dt;
          }
      }
      else
      {
          
          if (!_running)
          {
              _cur_time = get_secs();
              _running = true;
              return 0;
          }
          else
          {
              double new_time = get_secs();
              double dt = new_time - _cur_time;
              _cur_time = new_time;
              if (_file.is_open() ) _file << dt << std::endl;
			  accumulate_dt( dt );
              return dt;
          }
      }
    }

    double_t sum() const
    {
	  return _sum_dt;
    }
    
    ~TimeLogger()
    {
        _file.close();
    }
    
    void reset_stats()
    {
	  _sum_dt = _sum_dt2 = _n = 0;
    }

private:

    std::string     _filename;
    std::ofstream   _file;
    bool            _use_start_end;
    double          _start_time;
    double          _cur_time;
    double          _last_time;
    bool            _running;
    double_t        _tic_eps;
    double_t        _sum_dt;
    double_t        _sum_dt2;
    double_t        _n;

    void accumulate_dt( double_t dt )
    {
		_n += 1;
		_sum_dt += dt;
		_sum_dt2 += dt * dt;
    }

    
};

#endif
