/*

COMMON

  THIS MODULE INCLUDES GENERAL SUPPORT FILES THAT DID NOT FIND THEIR OWN PLACE
  THIS INCLUDES HELPERS FOR DEALING WITH BLITZ ARRAS, CONCURRENCY, PROFILING
  GSL, UBLAS, SERIALIZATION, AND A SIMPLE YET EFFECTIVE MIXIN CLASS FOR PROVIDING
  ARBITRARY C++ CLASSES WITH ATTRIBUTES THAT CAN BE QUERIED IN TEXT FORM TO 
  EASE SCRIPTING

COPYRIGHT (C)

  FRAUNHOFER FIT-LIFE
  SCHLOSS BIRLINGHOVEN
  53754 ST. AUGUSTIN
  GERMANY

CONTACT INFORMATION AVAILABLE AT:

  Fraunhofer-Institut für Angewandte Informationstechnik FIT 
  http://fit.fraunhofer.de

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the 
CeCILL-C license as circulated by CEA, CNRS and INRIA at the following 
URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
#ifndef _VARIANT_CASTS_HPP_
#define _VARIANT_CASTS_HPP_

#include <boost/variant.hpp>
#include "CPUData.hpp"

//cast support

template < typename Types > struct HasPropertiesBase;

template<class T, class U> 
struct cast_visitor_helper
{
  static T cast(const U & u)
  {
    return T(u);
  }
};

template<class T> 
struct cast_visitor_helper< T, std::string >
{
  static T cast(const std::string & u)
  {
	throw std::runtime_error("invalid cast to string");
    return T();
  }
};

template<> 
struct cast_visitor_helper< std::string, std::string >
{
  static std::string cast(const std::string & u)
  {
    return u;
  }
};

template<typename U>
struct cast_visitor_helper< std::string, U >
{
  static std::string cast(const U & u)
  {
    std::stringstream ss;
    ss << u;
    return ss.str();
  }
};

template<typename num_T>
struct cast_visitor_helper< CPUData<float>, num_T >
{
  static CPUData<float> cast( num_T const & number )
  {
	CPUData<float> result(1,1,1);
	result(0,0,0) = float(number);
	return result;
  }
};

template<typename num_T>
struct cast_visitor_helper< CPUData<double>, num_T >
{
  static CPUData<double> cast( num_T const & number )
  {
	CPUData<double> result(1,1,1);
	result(0,0,0) = double(number);
	return result;
  }
};

template<>
struct cast_visitor_helper< CPUData<double>, CPUData<float> >
{
  static CPUData<double> cast( CPUData<float> const & data )
  {
	CPUData<double> result( data.rows(), data.columns(), data.depth() );
	std::copy( data.begin(), data.end(), result.begin() );
	result.set_column_shape( data.column_shape() );
	return result;
  }
};

template<>
struct cast_visitor_helper< CPUData<float>, CPUData<double> >
{
  static CPUData<float> cast( CPUData<double> const & data )
  {
	CPUData<float> result( data.rows(), data.columns(), data.depth() );
	std::copy( data.begin(), data.end(), result.begin() );
	result.set_column_shape( data.column_shape() );
	return result;
  }
};

template<>
struct cast_visitor_helper< CPUData<double>, CPUData<double> >
{
  static CPUData<double> cast( CPUData<double> const & data )
  {
	return data;
  }
};

template<>
struct cast_visitor_helper< CPUData<float>, CPUData<float> >
{
  static CPUData<float> cast( CPUData<float> const & data )
  {
	return data;
  }
};


template<typename U>
struct cast_visitor_helper< CPUData<double>, U * >
{
  static CPUData<double> cast( U * const & u )
  {
	throw std::runtime_error("cannot cast pointer to CPUData<double>!");
  }
};


template<typename U>
struct cast_visitor_helper< CPUData<float>, U * >
{
  static CPUData<float> cast( U * const & u )
  {
	throw std::runtime_error("cannot cast pointer to CPUData<float>!");
  }
};


template<>
struct cast_visitor_helper< CPUData<float>, std::string >
{
  static CPUData<float> cast( std::string const & number )
  {
	throw std::runtime_error("cannot cast string to CPUData<float>!");
  }
};

template<>
struct cast_visitor_helper< CPUData<double>, std::string >
{
  static CPUData<double> cast( std::string const & number )
  {
	throw std::runtime_error("cannot cast string to CPUData<double>!");
  }
};

#define FORBID_CPUDATA_CAST(T)											\
  template<typename data_T>												\
  struct cast_visitor_helper< T, CPUData<data_T> >						\
  {																		\
	static T cast(CPUData<data_T> const & )								\
	{																	\
	  throw std::runtime_error("cannot cast CPUData<data_T> to scalar!"); \
	}																	\
  }


FORBID_CPUDATA_CAST(float);
FORBID_CPUDATA_CAST(bool);
FORBID_CPUDATA_CAST(int);

#define FORBID_POINTER_CAST(T)											\
  template<typename U>													\
  struct cast_visitor_helper< T, U * >									\
  {																		\
	static T cast(U * const & u)										\
	{																	\
	  throw std::runtime_error("cannot cast pointer to scalar");		\
	}																	\
  };

FORBID_POINTER_CAST(int)
FORBID_POINTER_CAST(float)

template<class T> 
struct cast_visitor 
  : public boost::static_visitor<T>
{
  template<class U> T operator()(const U & u) const
  {
    return cast_visitor_helper<T,U>::cast(u);
  }
};

template<class T, class Variant>
T cast(Variant v)
{
  return boost::apply_visitor(cast_visitor<T>(),v);
}

template<class T, class Variant>
void cast(Variant v, T & t)
{
  t = boost::apply_visitor(cast_visitor<T>(),v);
}

// template< typename T, class Types >
// struct cast_visitor_helper< T, HasPropertiesBase< Types > * >
// {
//   typedef HasPropertiesBase< Types > * PBaseP;
//   static T cast( const PBaseP & )
//   {
//     throw std::runtime_error("cannot cast HasProperties ptr to anything else!");
//     return T();
//   }
// };
// 
template< typename U, class Types >
struct cast_visitor_helper< HasPropertiesBase< Types > *, U >
{
  typedef HasPropertiesBase< Types > * PBaseP;
  static PBaseP cast( const U & )
  {
    throw std::runtime_error("cannot cast to HasProperties ptr from anything else!");
    return 0;
  }
};


template< class Types >
struct cast_visitor_helper< HasPropertiesBase< Types > *, HasPropertiesBase< Types > * >
{
  typedef HasPropertiesBase< Types > * PBaseP;
  static PBaseP cast( const PBaseP & p )
  {
    return p;
  }
};

#endif
