// -*- C++ -*-
/***************************************************************************
 * blitz/memblock.h      MemoryBlock<T> and MemoryBlockReference<T>
 *
 * $Id: memblock.h,v 1.20 2008/02/21 03:21:53 julianc Exp $
 *
 * Copyright (C) 1997-1999 Todd Veldhuizen <tveldhui@oonumerics.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Suggestions:          blitz-dev@oonumerics.org
 * Bugs:                 blitz-bugs@oonumerics.org
 *
 * For more information, please see the Blitz++ Home Page:
 *    http://oonumerics.org/blitz/
 *
 ***************************************************************************/

#ifndef BZ_MEMBLOCK_H
#define BZ_MEMBLOCK_H

#include <blitz/blitz.h>

#include <stddef.h>     // ptrdiff_t
#include <stdexcept>
#include "CPUDataBlock.hpp"

BZ_NAMESPACE(blitz)

enum preexistingMemoryPolicy { 
  duplicateData, 
  deleteDataWhenDone, 
  neverDeleteData 
};

template<typename P_type>
class MemoryBlockReference {

public:
    typedef P_type T_type;

protected:
    T_type * restrict data_;
    // oberhoff 2009: allow weak references
  bool weak_;

public:

#if 0
  typedef MemoryBlock<T_type> block_t;
#else
  typedef CPUDataBlock<T_type> block_t;
#endif

  block_t * block_;
    MemoryBlockReference()
    {
        block_ = 0;
        addReference();
        data_ = 0;
        weak_ = false;
    }

    bool weak() const
    {
   	    return weak_;
	}

    MemoryBlockReference( MemoryBlockReference<T_type>& ref, size_t offset=0, bool weak = false)
    {
	    weak_ = weak || ref.weak();
        block_ = ref.block_;
        addReference();
        data_ = ref.data_ + offset;
    }

    MemoryBlockReference( block_t * block, size_t offset=0, bool weak = false)
    {
	  weak_ = weak && block;
	  block_ = block;
	  if ( block )
		{
		  addReference();
		  data_ = block->data() + offset;
		}
	  else
		data_ = 0;
    }

    MemoryBlockReference(size_t length, T_type* data, preexistingMemoryPolicy deletionPolicy)
    {
        // Create a memory block using already allocated memory. 

        // Note: if the deletionPolicy is duplicateData, this must
        // be handled by the leaf class.  In MemoryBlockReference,
        // this is treated as neverDeleteData; the leaf class (e.g. Array)
        // must duplicate the data.

        weak_ = false;

        if ((deletionPolicy == neverDeleteData) 
            || (deletionPolicy == duplicateData)) {
// in this case, we do not need a MemoryBlock to ref-count the data
            block_ = 0;
        }
        else if (deletionPolicy == deleteDataWhenDone) {
            block_ = new block_t( length, data );

#ifdef BZ_DEBUG_LOG_ALLOCATIONS
            cout << "MemoryBlockReference: created MemoryBlock at "
                 << ((void*)block_) << endl;
#endif
        }
        addReference();

        data_ = data;
    }

    explicit MemoryBlockReference(size_t items)
    {
        weak_ = false;

        block_ = new block_t(items);
        addReference();
        data_ = block_->data();

#ifdef BZ_DEBUG_LOG_ALLOCATIONS
    cout << "MemoryBlockReference: created MemoryBlock at "
         << ((void*)block_) << endl;
#endif

    }

   ~MemoryBlockReference()
    {
        blockRemoveReference();
    }


  block_t * block() { return block_; }

  const block_t * block() const { return block_; }

protected:

    int numReferences() const
    {
	    if ( weak_ ) return 2; // so that it does not look unique
        if (block_) 
            return block_->references();
#ifdef BZ_DEBUG_LOG_REFERENCES
        cout << "Invalid reference count for data at "<< data_ << endl;
#endif
        return -1;      
    }

    bool lockReferenceCount(bool lockingPolicy) const
    {
        if (block_)
            return block_->doLock(lockingPolicy);
        // if we have no block, consider request successful
#ifdef BZ_DEBUG_LOG_REFERENCES
        cout << "No reference count locking for data at "<< data_ << endl;
#endif
        return true;    
    }

    void changeToNullBlock()
    {
        blockRemoveReference();
        block_ = 0;
        addReference();
        data_ = 0;
    }

    void changeBlock(MemoryBlockReference<T_type>& ref, size_t offset=0, bool weak = false)
    {
        blockRemoveReference();
        weak_ = weak || ref.weak();
        block_ = ref.block_;
        addReference();
        data_ = ref.data_ + offset;
    }

    void newBlock(size_t items)
    {
        blockRemoveReference();
        block_ = new block_t(items);
        addReference();
        data_ = block_->data();
		weak_ = false;

#ifdef BZ_DEBUG_LOG_ALLOCATIONS
    cout << "MemoryBlockReference: created MemoryBlock at "
         << ((void*)block_) << endl;
#endif
    }

private:
    void blockRemoveReference()
    {
        if ( weak_ ) return;
        int refcount = removeReference();
        if (refcount == 0)
        {
#ifdef BZ_DEBUG_LOG_ALLOCATIONS
    cout << "MemoryBlock: no more refs, delete MemoryBlock object at "
         << ((void*)block_) << endl;
#endif

            delete block_;
        }
    }

    void addReference() const 
    {
        if (block_ && !weak_ ) {
            block_->addReference();
        }
        else {
#ifdef BZ_DEBUG_LOG_REFERENCES
            cout << "Skipping reference count for data at "<< data_ << endl;
#endif
        }
    };

    int removeReference() const 
    {
        if ( weak_ ) return 1;
        if (block_)
            return block_->removeReference();
#ifdef BZ_DEBUG_LOG_REFERENCES
        cout << "Skipping reference count for data at "<< data_ << endl;
#endif
        return -1;      
    };
  
    void operator=(const MemoryBlockReference<T_type>&)
    { }
};


BZ_NAMESPACE_END

#include <blitz/memblock.cc>

#endif // BZ_MEMBLOCK_H
