#pragma once
#include <set>
#include <map>
#ifdef USE_POOL
# include <boost/pool/pool.hpp>
# include <boost/ptr_container/ptr_map.hpp>
#endif
#include <boost/thread/tss.hpp>
#include <iostream>

namespace blitz{

  struct memblock_manager
  {

	struct counter_t
	{
	  counter_t() : alloc_count(0), free_count(0) {}
	  size_t alloc_count, free_count;
	};

	memblock_manager( bool tb ) : thread_bound(tb), live_count(0)
	{
	  std::cout << " new allocator " << this << std::endl;
	}

	typedef std::map<size_t,counter_t> histogram_t;
	typedef std::set<char*> registry_t;

#ifdef USE_POOL
	typedef boost::pool<> pool_t;
	typedef boost::ptr_map<size_t,pool_t> pools_t;
	pools_t pools;
#endif

	histogram_t histogram;
	registry_t registry;
	bool thread_bound;	
	size_t live_count;

	bool alive() const { return live_count > 0; }
	
	void detach() { thread_bound = false; }

	char * alloc( size_t n )
	{
	  if ( n == 0 ) return 0;
#ifdef USE_POOL
	  pool_t * pool;
	  pools_t::iterator ipool = pools.find( n );
	  if ( ipool == pools.end() )
		{
		  // make n divisible by 0x0f so thatpointer bumping yields sse-able memory
		  size_t n_real = 16 * ( ( n + 15 ) / 16 );
		  pool = new pool_t( n_real );
		  pools.insert( n, pool );
		}
	  else
		pool = ipool->second;
	  char * mem = (char*)pool->malloc();
#else
	  char * mem = new char[n];
#endif
	  histogram[n].alloc_count++;
	  registry.insert( mem );
	  ++live_count;
	  return mem;
	}

	void free( char * p, size_t n )
	{
	  if ( n == 0 || p == 0 ) return;
	  histogram[n].free_count++;
	  --live_count;
	  if ( registry.find(p) == registry.end() )
		std::cout << " warning: size " << n << " freed by a different manager than it was allocated from " << std::endl;
	  else
		registry.erase(p);
#ifdef USE_POOL
	  pools.at( n ).free( p );
#else
	  delete[] p;
#endif
	}

	void dump_histogram() const
	{
	  for ( histogram_t::const_iterator begin = histogram.begin(), end = histogram.end(); begin != end; ++begin )
		{
		  // fixme: delegate to a list of desctructors
		  std::cout << " memblock@"<< this<< " size " << begin->first << " allocs: " << begin->second.alloc_count << " frees: " << begin->second.free_count << std::endl;
		}
	}
  
	~memblock_manager()
	{
	  std::cout << " delete allocator " << this << " live count " << live_count << std::endl;
	  //dump_histogram();
	  if ( live_count > 0 )
		detach();
	}


  };

  extern boost::thread_specific_ptr< memblock_manager > memblock_manager;

}; //ns blitz
