#pragma once

#include "SeariseProperties.hpp"
#include "variational_model.hpp"
#include <algorithm>
#include <iostream>
#include "float.h"
#include <boost/numeric/ublas/matrix.hpp>
#include "time.h"

#include "Random.hpp"

template< class float_t=float, class data_t = blitz::Array<float_t,1> >
class beta_binomial : public variationalModel< data_t, float_t >,
  public HasProperties< SearisePropertyTypes, beta_binomial< float_t > >
{
  typedef float_t float_type;
  typedef data_t data_type;
  typedef typename data_type::T_index shape_type;
  typedef variationalModel<data_type,float_type> baseT;
  typedef typename variationalModel<data_type,float_type>::floatVecT floatVecT;
  
  typedef beta_binomial< float_type, data_type > This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;

  typedef Random<float_type> float_rng_t;
  typedef Random<double> double_rng_t;

  static const float_type epsilon = FLT_EPSILON;

  template<typename X>
  float_type log_likelihood( const X & x, bool variational = false ) const
  {
	// fixme: variational not properly done
	return std::log( max( epsilon, float_type( sum( mMuVar * X + ( 1 - mMuVar ) * ( 1 - X ) ) ) ) );
  }

  static void init_properties()
  {
    PBase::add_ivar_property_rw( "nu0", &This::_nu );
    PBase::add_ivar_property_rw( "mu0", &This::_mu );
	Base::add_property( "mu"        , make_blitz_ivar_getter1( &This::mMuVar ) );
	PBase::add_ivar_property_rw( "N", &This::mParentPostAcc );
  }

  template<typename X>
  void learn( const X & x, const float_type parentPost) 
  {
	if( parentPost < epsilon ) return;
	mDataAcc       += parentPost * x;
	mParentPostAcc += parentPost;
  }

  void setShape( shape_type shape )
  {
	_shape = shape;
  }

  void setPrior( float_type add_pseudodata = 0, bool randomize_pseudodata = false ) 
  {
	mMu.resize( _shape );
	mMuVar.resize( _shape );
	mNu.resize( _shape );
	mNuVar.resize( _shape );
	mMu = _mu;
	mMuVar = _mu;
	mNu = _nu;
	mNuVar = _nu;
	if ( add_pseudodata > 0 )
	  {
		data_type X( _shape );
		if ( randomize_pseudodata )
		  {
			float_rng_t & rng = float_rng_t::get();
			BOOST_FOREACH( float_type & x, X )
			  x = rng.beta( _mu * _nu, ( 1 - _mu ) * _nu );
		  }
		else
		  X = _mu;
		
		baseT::beginLearning();
		baseT::setData(X);
		baseT::learn( add_pseudodata );
		baseT::updateParameterPosterior();
		baseT::posteriorToPrior();
	  }
  }

  /** Accumulate sufficient statistics. Implement in derived classes.
  \param parentPost: parent posterior, e.g. responsibilities in a mixture model. If length==0, assume full responsibility for the data.
  */
  void doLearn(const float_type parentPost) 
  {
	learn(baseT::mData,parentPost);
  }
  
  /** Do inference (E-step) given parentPosterior. Implement in derived classes.
  */
  void doInference() 
  {
    // No hidden variables. Nothing to do.
  };
  

  /** Update variational posterior with current accumulated sufficient statistics and prior (M-step). Implement in derived classes.
  */
  bool doVariationalParameterUpdate()
  {
      if ( mParentPostAcc / mNu < FLT_EPSILON ) return false; //not responsible for these data
      mBetaVar = mBeta + mParentPostAcc;
      mNuVar   = mNu   + mParentPostAcc;
	  mMuVar   = mMu * ( mNu / mNuVar ) + mDataAcc / mNuVar;
      return true;
  };
  
private:
  
  data_type mMu, mMuVar; // current (pseudo-) data statistical mean
  data_type mDataAcc;
  float_type mNu, mNuVar; // current (pseudo-) data statistical weight
  float_type mParentPostAcc;

  float_type _nu, _mu;
  shape_type _shape;
};
