/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/

#pragma once


#include <stdlib.h>
#include "math.h"
#include <cassert>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/foreach.hpp>
#include <boost/shared_ptr.hpp>
#include <algorithm>
#include <iostream>
#include <vector>
#include <limits>
#include "Random.hpp"
#include "omp.h"

#include "variational_model.hpp"

using namespace std;


template<class component_container_type> class variationalComponentModel; // varmodel with 1 type of components
template<class component_container_type1,class component_container_type2> class variationalComponentModel2; // varmodel with 2 types of components

template<class varmodel_type> class bottomComponentContainer; // bottom-level component container. Components in this container have no shared sub-components.
template<class component_varmodel_type> class componentContainer; // holds components with 1 type of shared sub-component
template<class components_varmodel_type> class component2Container; // holds components with 2 types of shared sub-components. If this was dynamically typed, it wouldn't have to be redefined.


/** Bottom-level container for variational_model components. Useful for models with sub-components, e.g. DPMM or sequence_model. Also efficient for component sharing across models --
likelihoods of sub-models need to be evaluated only once for each new datapoint. Can buffer datapoints (useful for sequence learning) */
template<class varmodel_type> 
class bottomComponentContainer {
public:
  
  typedef varmodel_type componentT;
  typedef typename componentT::dataT dataT;
  typedef typename componentT::pdataT pdataT;
  typedef boost::shared_ptr<componentT> pComponentT;
  typedef ublas::matrix<pComponentT> componentsT;
  typedef typename componentT::floatVecT floatVecT;
  typedef typename componentT::floatMatT floatMatT;
  typedef typename componentT::floatT floatT;
  typedef std::vector<int> intVecT;
  
  typedef bottomComponentContainer<varmodel_type> This;
  
  /** Construct an empty, non-shared container */
  bottomComponentContainer() 
  : mbShared(false)
  , mbSharedOpNow(false)
  {
    mPrototype.setPrior(0);
    setNumThreads(1);
  };
 
  /** Copy-construct container with components that share the parameters with the source */
  bottomComponentContainer(const This &source) : mPrototype(source.mPrototype)
  {
    mPrototype.unshare();
    doShareWith(source);
  }
  
  /** Make parameter-sharing copies of components in source and stick into this container */
  void shareWith(const This& source)
  {
     if(mbShared && !mbSharedOpNow) return;
     doShareWith(source);
  }
  
  /** Make shared parameters private. Make mComponent(seg,comp)=mComponent(0,comp) */
  void unshare()
  {
    if(mbShared && !mbSharedOpNow) return;
    for(int comp=0;comp<components();++comp)
    {
      mComponents(0,comp)->unshare();
      for(int seg=1;seg<segments();++seg)
      {
	mComponents(seg,comp)=pComponentT( new componentT(*mComponents(0,comp)) );
	mComponents(seg,comp)->setIndex(comp);
	mComponents(seg,comp)->setSegment(seg);
      }
    }
  }
  
  
  /** set prior. Add pseudodata for each component */
  void setPrior(floatVecT pseudodata=floatVecT()) {
    //if(mbShared && !mbSharedOpNow) return;
    RERR( (mComponents.size1()==0 || mComponents.size2()==0), (boost::format("Error setting prior: components not initialized (%d,%d)") %  mComponents.size1() % mComponents.size2()).str() );
    RERR( (pseudodata.size()!=mComponents.size2()) , (boost::format("Error setting prior: got %d pseudcounts, have %d components") % pseudodata.size() % mComponents.size2()).str() );
    for(int i=0;i<components();++i) 
    {
      mComponents(0,i)->setPrior(pseudodata[i]); // setting prior is only necessary for first instance of every component.
      for(int j=1;j<segments();++j) mComponents(j,i)->updateAuxiliaryVariables(); // others just need to update internal states.
    }
  }
  
  /** \param shared: if true, mark this container as shared */
  void setShared(bool shared) {mbShared=shared;};
  
  /** \return true if this container is shared */
  bool isShared() const {return mbShared;};
  
  /** \param prototype: the prototype to which new components of this container are initialized. Clears components. */
  void setPrototype(const varmodel_type & prototype) {
    mPrototype=prototype; 
    mComponents.resize(0,0);
    mLogLikelihoods.resize(0,0);
  };
  
  /** \return reference to prototype */
  componentT & getPrototype(){return mPrototype;};
  
  /** \return shared_ptr to segment seg, component comp */
  pComponentT getComponent(unsigned seg,unsigned comp) 
  {
    RERR( (seg>=mComponents.size1() || comp>=mComponents.size2()) , (boost::format("Error accessing segment %d, component %d : have only %d , %d") % seg % comp % mComponents.size1() % mComponents.size2()).str() );
    return mComponents(seg,comp);
  }
  
  
  
  /** \return matrix of components(segment,component) */
  const componentsT & getComponents() const {return mComponents;};
  
  /** Resize container to newsize components. Number of segments from prototype. Init new components to prototype. */
  int resize(int newsize,bool keepOld=false) 
  {
    if(!keepOld) mComponents.resize(0,0);
    int oldsize=components();
    mComponents.resize(mPrototype.segments(),newsize,true); // keep old components.
    if(mPrototype.segments()==0) return oldsize;
    for(int i=0;i<newsize;++i) {
      if(i>=oldsize) {
	mComponents(0,i)=pComponentT( new varmodel_type(mPrototype) ); // only init new components, keep old ones
	mComponents(0,i)->unshare(); // detach from protoype
      }
      mComponents(0,i)->setIndex(i);
      mComponents(0,i)->setSegment(0);
      for(int j=1;j<mPrototype.segments();++j) {
	mComponents(j,i)=pComponentT( new componentT(*mComponents(0,i)) ); // share parameters across segments
	mComponents(j,i)->setIndex(i);
	mComponents(j,i)->setSegment(j);
      }
    }
   
    mLogLikelihoods.resize(segments(),components());
    return oldsize;
  }
  
  /** \return number of components */
  int components() const {return mComponents.size2();};
  
  /** \return number of segments */
  int segments() const { 
    if(mComponents.size1()<1) return 1;
    return mComponents.size1();
  };
  
  /** set number of threads for next run of learn/inference. */
  void setNumThreads(int nt) {
    mNumThreads=std::min(std::max(nt,1),8);
  }
  
  /** Run inference for all components in container. If shared, run only if setSharedOpNow(true) has been called before.
  \param data_ptr: shared_ptr to data (of segment, if segmented).
  \param variational: if true, compute variational expectations.
  */
  void inference(const pdataT & data_ptr,bool variational=true) 
  {
    if(!mbShared || mbSharedOpNow) {
      RERR( (mComponents.size1()!=mLogLikelihoods.size1() || mComponents.size2()!=mLogLikelihoods.size2()),
	    (boost::format("Error in inference: have (%d,%d) components, but (%d,%d) log-likelihoods. Please resize bottomComponentContainer") % mComponents.size1() % mComponents.size2() % mLogLikelihoods.size1() % mLogLikelihoods.size2() ).str());
      
      omp_set_num_threads(mNumThreads);
      const int numsegs=segments();
     
      #pragma omp parallel for 
      for(int segment=0;segment<numsegs;++segment) {
	pdataT pCurSeg=mPrototype.getDataSegment(data_ptr,segment);
	for(int comp=0;comp<components();++comp) {
	  mComponents(segment,comp)->inference(pCurSeg,variational);
	  mLogLikelihoods(segment,comp)=mComponents(segment,comp)->getLogLikelihood();
	}
      }
    }
  }
  
  /** \return likelihoods of components in given segment */
  const floatVecT getComponentLikelihoods(unsigned segment) const
  {
    RERR( (segment>=mLogLikelihoods.size1()) , (boost::format("can't get likelihood for segment %d of %d") % segment % mLogLikelihoods.size1()).str() );
    return ublas::matrix_row<const floatMatT>(mLogLikelihoods,segment);
  }
  
  /** \return log-likelihoods(segment,component) */
  const floatMatT & getComponentLikelihoods() const {return mLogLikelihoods;};
  
  /** Learn for all components in container. If shared, run only if setSharedOpNow(true) has been called before.
  \param data_ptr: shared_ptr to data (of segment, if segmented).
  \param parentPosterior: matrix(segment,component) of component posteriors (responsibilities).
  */
  void learn(const pdataT & data_ptr,const floatMatT & parentPosteriors) 
  {
    if(!mbShared || mbSharedOpNow) {
      RERR( (parentPosteriors.size1()!=mComponents.size1() || parentPosteriors.size2()!=mComponents.size2() ),
	    (boost::format("Error learning in bottomComponentContainer: got (%d,%d) parent posteriors, have (%d,%d)") % parentPosteriors.size1() % parentPosteriors.size2() % mComponents.size1() % mComponents.size2() ).str() );
   
      omp_set_num_threads(mNumThreads);
      const int numsegs=segments();
      
      #pragma omp parallel for
      for(int segment=0;segment<numsegs;++segment) {
	pdataT pCurSeg=mPrototype.getDataSegment(data_ptr,segment);
	for(int comp=0;comp<components();++comp) {
	  mComponents(segment,comp)->learn(pCurSeg,parentPosteriors(segment,comp));
	}
      }
    }
  }
  
  
  /**  Prediction pass, i.e. send posteriors computed by parents back to children. If shared, run only if setSharedOpNow(true) has been called before.
  \param parentPosterior: matrix(segment,component) of component posteriors (responsibilities).
  */
  void predictPass(const floatMatT & parentPosteriors) 
  {
    if(!mbShared || mbSharedOpNow) {
      RERR( (parentPosteriors.size1()!=mComponents.size1() || parentPosteriors.size2()!=mComponents.size2() ),
	    (boost::format("Error in prediction pass in bottomComponentContainer: got (%d,%d) parent posteriors, have (%d,%d)") % parentPosteriors.size1() % parentPosteriors.size2() % mComponents.size1() % mComponents.size2() ).str() );
      for(int segment=0;segment<segments();++segment) {
	for(int comp=0;comp<components();++comp) {
	  mComponents(segment,comp)->predictPass(parentPosteriors(segment,comp));
	}
      }
    }
  }
  
  
  /** Do variational posterior update for all components with current accumulated sufficient statistics.
  If shared, run only if setSharedOpNow(true) has been called before.
  */
  void updateParameterPosterior()
  {
    callMethodForAll(&componentT::updateParameterPosterior);
  }
  
  /** Reset variational parameter posterior of components to prior.
  If shared, run only if setSharedOpNow(true) has been called before.
  */
  void resetParameterPosteriorToPrior()
  {
    callMethodForAll(&componentT::resetParameterPosteriorToPrior);
  }
  
  /** Do variational posterior update for all components with current accumulated sufficient statistics.
  If shared, run only if setSharedOpNow(true) has been called before.
  */
  void posteriorToPrior()
  {
    callMethodForAll(&componentT::posteriorToPrior,true);
  }
  

  /** collect variational bound for all components. If components are shared, run only if setSharedOp(true) has been called before, else return 0*/
  floatT collectVarBound() 
  {
    return collectFromAll(&componentT::getAccumulatedVarBound);
  }
  
  /** Sample.
  \param compo: component from which to sample in each segment.
  \return shared ptr to datapoint containing the sample(s).
  */
  pdataT sample(const intVecT & compo)
  {
    RERR( (segments()!=(int)compo.size()) , (boost::format("Error sampling: got %d segments, need %d") % compo.size() % segments()).str() );
    pdataT samp( new dataT );
    for(unsigned seg=0;seg<compo.size();++seg) {
      RERR( (compo[seg]<0) , (boost::format("Error sampling segment %d: no component specified (%d)") % seg % compo[seg]).str() );
      pdataT csamp=mComponents(seg,compo[seg])->sample();
      *mPrototype.getDataSegment(samp,seg) = *csamp;
    }
    return samp;
  }
  
  /** Decay certainty of posteriors, used to escape local minima. 
  \param tau: typically, certainty should be reduced by a factor exp(-1/tau)
  */
  void decayParameters(floatT tau, bool update_immediately = true) 
  {
     if(!mbShared || mbSharedOpNow) {
      for(int cmp=0;cmp<components();++cmp) 
      {
	mComponents(0,cmp)->decayParameters(tau,update_immediately);
      }
    }
  }
  
  /** reset hidden variables of components to prior. If shared, run only if setSharedOpNow(true) has been called before */
  void resetToPrior() {
    callMethodForAll(&componentT::resetToPrior);
  }
  
  /** reset variational bound accumulators of all components. If shared, run only if setSharedOpNow(true) has been called before */
  void resetVarBound() {
    callMethodForAll(&componentT::resetVarBound);
  }
  
  /** begin learning in all components. If shared, run only if setSharedOpNow(true) has been called before */
  void beginLearning(bool accumulateVarBound=false) {
    if(!mbShared || mbSharedOpNow) {
      for(int seg=0;seg<segments();++seg) 
	for(int comp=0;comp<components();++comp)
	  mComponents(seg,comp)->beginLearning(accumulateVarBound);
    }
  }
  
  /** freeze some or all parameters of all components during learning. Useful for layer-wise training etc. NOT a hierarchical call.
  \param level: freeze-level. 0 (default) should mean no parameters are frozen
  */
  void freeze(int level=0) {
    callIntMethodForAll(&componentT::freeze,level,true);
  }
  
  /** set sequence learning phase. for meaning, \see variationalModel::seqPhaseT */
  void setSequenceLearningPhase(int phase=0)
  {
    callIntMethodForAll(&componentT::setSequenceLearningPhase,phase);
  }
  
  
  /** collect KL divergence of all components, one segment. If shared, run only if setSharedOpNow(true) has been called before */
  floatT collectKL() {
    floatVecT weights=ublas::scalar_vector<floatT>(components(),1.0);
    return collectFromAll(&componentT::getKL,weights);
  }
  
  /** get entropy of components, weighted by parent posterior.If shared, run only if setSharedOpNow(true) has been called before 
  \param parentPosts: (accumulated) posteriors of parents
  */
  floatT collectEntropy(const floatVecT & parentPosts)
  {
    return collectFromAll(&componentT::getEntropy,parentPosts);
  }
  
  
  /** get expected log^2(p) of components, weighted by parent posterior.If shared, run only if setSharedOpNow(true) has been called before 
  \param parentPosts: (accumulated) posteriors of parents
  */
  floatT collectEntropy2(const floatVecT & parentPosts)
  {
    return collectFromAll(&componentT::getEntropy2,parentPosts);
  }
 
  /** Write to stream. Only write the first component of each segment.
  If this container is not shared, just write all components and sub-components. If it is shared, write only if setSharedOp(true) has been called before
  \param os  reference to ostream to write into.
  \param cc  reference to component container
  \return ostream after container has been inserted
  */
  friend ostream &operator<<(ostream &os,const This &cc) 
  {
	if(cc.mbShared && !cc.mbSharedOpNow) return os;
        os<<"Bottom-layer component container"<<endl;
	os<<"Number of components: "<<cc.components()<<endl;
	os<<"Share type: "<<(cc.isShared()?"shared":"not_shared")<<endl;
	os<<"Prototype"<<endl;
	os<<cc.mPrototype;
	if(cc.segments()>0)
	  for(int comp=0;comp<cc.components();++comp) os<<*(cc.mComponents(0,comp)); // save only 1st. segment. others are copies.
	
        return os;
  }
  
  /** read from stream is. If this container is not shared, read all components and sub-components. If it is shared, do nothing.
          \param is  reference to stream to read from.
          \param varmod reference to variationalModel to read into
          \return stream is after variationalModel has been read.
  */
  friend istream &operator>>(istream &is,This &cc) 
  {
        if(cc.mbShared && !cc.mbSharedOpNow) return is;
        std::string name;
	getline(is,name);
	if(name!="Bottom-layer component container") throw std::runtime_error("Error loading bottom-layer component container: got "+name);
	unsigned numcmp;
	is>>eatString(3)>>numcmp>>ws;
	is>>eatString(2)>>name>>ws;
	cc.setShared(name=="shared");
	getline(is,name);
	is>>cc.mPrototype>>ws;
	cc.resize(numcmp);
	if(cc.segments()<1) return is;
	
	for(int comp=0;comp<cc.components();++comp)
	  {
	    is>>*cc.mComponents(0,comp);
	    cc.mComponents(0,comp)->setIndex(comp);
	    cc.mComponents(0,comp)->setSegment(0);
	    for(int seg=1;seg<cc.segments();++seg)
	    {
	      cc.mComponents(seg,comp)=This::pComponentT( new This::componentT(*cc.mComponents(0,comp)) );
	      cc.mComponents(seg,comp)->setIndex(comp);
	      cc.mComponents(seg,comp)->setSegment(seg);
	    }
	  }
	
        return is;
  }
  
  
  friend bool operator==(const This & first,const This & second) 
  {
   return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
      if(!(first.mbShared==second.mbShared)) return false;
      if( (first.mbShared && !first.mbSharedOpNow) || (second.mbShared && !second.mbSharedOpNow) ) return true; // to avoid multiply comparisons in a shared hierarchy
      if(first.isShared()!=second.isShared()) return false;
      if(!(first.mPrototype==second.mPrototype)) return false;
      if(first.components()!=second.components()) return false;
      if(first.segments()!=second.segments()) return false;
      for(int seg=0;seg<first.segments();++seg) {
	for(int comp=0;comp<first.components();++comp) {
	  if (!(*first.mComponents(seg,comp)==*second.mComponents(seg,comp))) return false;
	} 
      }
      return true;
  }
 
 
 // for use by serialization and inference etc. 
  void setSharedOpNow(bool sn) {mbSharedOpNow=sn;};
  bool getSharedOpNow() const {return mbSharedOpNow;};
  
protected:
  varmodel_type mPrototype; //!< prototype for components. Init to this prototype when generating new components.
  componentsT mComponents; //!< components held by this container. mComponents(segment,component) accesses instance "segment" of component.
  bool mbShared; //!< true if this is a container of shared components. When serializing, such a container needs only to be serialzed once.
  bool mbSharedOpNow; //!< if true, serialize a shared container, or run inference on a shared container etc.     
  floatMatT mLogLikelihoods; //!< log-likelihood buffer(segment,component) (to speed up segment-wise access).
  int mNumThreads; //!< number of threads used during learning and inference.
  
  /** Call a method of all components, all segments.
  \param firstOnly: if true, call only on 1st segment. Default: false
  If shared, run only if setSharedOpNow(true) has been called before.
  */
  void callMethodForAll(void (componentT::*themethod)(),bool firstOnly=false)
  {
    if(!mbShared || mbSharedOpNow) {
      RERR( (segments()==0) , ("Error in callMethodForAll: no segments"));
      for(int cmp=0;cmp<components();++cmp) 
	for(int seg=0;seg<(firstOnly?1:segments());++seg)
	  (mComponents(seg,cmp).get()->*themethod)();
    }
  }
  
  /** Call a method with an int argument for all components, all segments 
  \param ipar: parameter passed to method
  \param onlyThisContainer: if true, call independent of share status. Otherwise, call only if container is not shared or if shared op. is done now.
  */
  void callIntMethodForAll(void (componentT::*themethod)(int), int ipar,bool onlyThisContainer=false)
  {
    if(!mbShared || mbSharedOpNow || onlyThisContainer) {
      for(int seg=0;seg<segments();++seg)
	for(int comp=0;comp<components();++comp)
	  (mComponents(seg,comp).get()->*themethod)(ipar);
    }
  }
  
  
  /** Collect a float value from a method of all components
  \param weights: if supplied, weight values by these weights and call only on 1st segment.
  If shared, run only if setSharedOpNow(true) has been called before.
  \return (weighted) sum of all collected values
  */
  floatT collectFromAll(floatT (componentT::*themethod)(),floatVecT weights=floatVecT())
  {
    floatT retval=0.0;
    if(!mbShared || mbSharedOpNow) {
      RERR( (segments()==0) , ("Error in collectFromAll: no segments"));
      if(weights.size()==0) {
	for(int seg=0;seg<segments();++seg)
	  for(int cmp=0;cmp<components();++cmp)
	    retval+=(mComponents(seg,cmp).get()->*themethod)();
	return retval;
      }
	  
      RERR( (weights.size()!=(unsigned)components()) , (boost::format("Error in collectFromAll: got %d weights, have %d components") % weights.size() % components()).str() );
      for(int cmp=0;cmp<components();++cmp) retval+=weights[cmp]*(mComponents(0,cmp).get()->*themethod)();
    }  
    return retval;
 }
 
  /** make parameter sharing copies of components. Used both by shareWith() and copy-constructor */
  void doShareWith(const This & source) 
  {
    mComponents.resize(0,0);
    mComponents.resize(source.mComponents.size1(),source.mComponents.size2());
    for(int seg=0;seg<segments();++seg)
      for(int comp=0;comp<components();++comp) {
	mComponents(seg,comp)=pComponentT( new componentT( *source.mComponents(seg,comp) ) ); // make a parameter-sharing copy of the components
	mComponents(seg,comp)->setIndex(comp);
	mComponents(seg,comp)->setSegment(seg);
      }
      
    mbShared=source.mbShared;
    mbSharedOpNow=source.mbSharedOpNow;
    mLogLikelihoods=ublas::zero_matrix<floatT>(source.mLogLikelihoods.size1(),source.mLogLikelihoods.size2());       
    mNumThreads=source.mNumThreads;
  }
 
};
  
  

template<class component_varmodel_type>
class componentContainerBase : public bottomComponentContainer<component_varmodel_type>
{
public:
  
  
  typedef bottomComponentContainer<component_varmodel_type> baseT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::floatT floatT;
  
  typedef typename baseT::componentT componentT;
  typedef typename baseT::pComponentT pComponentT;
  typedef typename baseT::componentsT componentsT;
   
  typedef typename componentT::dataT dataT;
  typedef typename baseT::pdataT pdataT;
  
  typedef componentContainerBase<component_varmodel_type> This;
 
   /** Construct an empty, non-shared container */
  componentContainerBase() : baseT()
  {};

  
protected:
  
  /**  Set component[comp]s' sub-component-containers to psubcon (for all segments), if the latter is shared. 
  Otherwise, initialize new sub-component container for this component.
  Set index of component[index] to index.
  \param comp: component index
  \param psubcon: container of sub-components
  \param setter: pointer to member which sets sub-component containers of the components.
  */
  template<class subcc,class comp>
  void setCompSub(int compnr,boost::shared_ptr<subcc> & psubcon, void (comp::*setter)(boost::shared_ptr<subcc> &) ) 
  {
    assert(compnr>=0 && (unsigned)compnr<This::components());
    for(int seg=0;seg<This::segments();++seg)
    {
      pComponentT & curcon=This::mComponents(seg,compnr);
      if(psubcon->isShared()) (curcon.get()->*setter)(psubcon);
      else {
	boost::shared_ptr<subcc> newsub=boost::shared_ptr<subcc>( new subcc() );
	(curcon.get()->*setter)(psubcon);
      }
      This::mComponents(seg,compnr)->setIndex(compnr);
      This::mComponents(seg,compnr)->setSegment(seg);
    }
  }
  
  /** Do inference on sub-components, if they are shared. Otherwise, do nothing.
  \param psubcon: shared_ptr to container of sub-components
  \param data_ptr: data
  \param variational: if true, compute variational expectations
  */
  template<class subcc,class pdat>
  void subInference(boost::shared_ptr<subcc> & psubcon,const pdat & data_ptr,bool variational=true) 
  {
    if(psubcon->isShared()) 
    {
      psubcon->setSharedOpNow(true);
      psubcon->inference(data_ptr,variational);
      psubcon->setSharedOpNow(false);
    }
  }
  
  /** Do learning of sub-components, if they are shared. otherwise, do nothing. 
  \param psubcon: shared_ptr to container of sub-components
  \param data_ptr: pointer to data
  \param subPost: posterior responsibilities of sub-components
  */
  template<class subcc,class pdat>
  void subLearn(boost::shared_ptr<subcc> & psubcon,const pdat & data_ptr,const floatMatT & subPost) 
  {
    if(psubcon->isShared()) 
    {
	psubcon->setSharedOpNow(true);
	psubcon->learn(data_ptr,subPost);
	psubcon->setSharedOpNow(false);
    }
  }
   
  
  
  /** Execute an operation on sub-componentents, if they are shared. otherwise, do nothing. Operation has no parameters and no return value
  \param psubcon: shared_ptr to container of sub-components
  */
  template<class subcc>
  void subExec(boost::shared_ptr<subcc> & psubcon,void (subcc::*operation)()) 
  {
    if(psubcon->isShared()) 
    {
	psubcon->setSharedOpNow(true);
	(psubcon.get()->*operation)();
	psubcon->setSharedOpNow(false);
    }
  }
  
  
  /** Execute an operation on sub-componentents, if they are shared. otherwise, do nothing. Operation has no parameters and returns a floatT.
  \param psubcon: shared_ptr to container of sub-components
  */
  template<class subcc>
  floatT subExecF(boost::shared_ptr<subcc> & psubcon, floatT (subcc::*operation)()) 
  {
    if(psubcon->isShared()) 
    {
	psubcon->setSharedOpNow(true);
	floatT retval=(psubcon.get()->*operation)();
	psubcon->setSharedOpNow(false);
	return retval;
    } else return 0.0;
  }
  
  /** Execute an operation on sub-componentents, if they are shared. otherwise, do nothing. Operation has an int parameter and returns nothing.
  \param psubcon: shared_ptr to container of sub-components
  \param iprm: the parameter passed to each sub-component
  \param operation: pointer-to-member of sub-components
  */
  template<class subcc>
  void subExecI(boost::shared_ptr<subcc> & psubcon, void (subcc::*operation)(int), int iprm) 
  {
    if(psubcon->isShared()) 
    {
	psubcon->setSharedOpNow(true);
	(psubcon.get()->*operation)(iprm);
	psubcon->setSharedOpNow(false);
    }
  }
  
  /** Execute an operation on sub-componentents, if they are shared. otherwise, do nothing. Operation takes a floatVecT as parameter and returns a floatT.
  \param psubcon: shared_ptr to container of sub-components
  \param param: vector(component) of parameters
  */
  template<class subcc>
  floatT subExecF_V(boost::shared_ptr<subcc> & psubcon, floatT (subcc::*operation)(const floatVecT &),const floatVecT &param) 
  {
    if(psubcon->isShared()) 
    {
	psubcon->setSharedOpNow(true);
	floatT retval=(psubcon.get()->*operation)(param);
	psubcon->setSharedOpNow(false);
	return retval;
    } else return 0.0;
  }
  
  /** Execute an operation on sub-componentents, if they are shared. otherwise, do nothing. Operation takes a floatMatT as parameter and returns a nothing.
  \param psubcon: shared_ptr to container of sub-components
  */
  template<class subcc>
  void subExec_M(boost::shared_ptr<subcc> & psubcon, void (subcc::*operation)(const floatMatT &),const floatMatT &param) 
  {
    if(psubcon->isShared()) 
    {
	psubcon->setSharedOpNow(true);
	(psubcon.get()->*operation)(param);
	psubcon->setSharedOpNow(false);
    };
  }
  
  /** Execute an operation on sub-componentents, if they are shared. otherwise, do nothing. Operation takes a bool as parameter and returns nothing
  \param psubcon: shared_ptr to container of sub-components
  */
  template<class subcc>
  void subExec_B(boost::shared_ptr<subcc> & psubcon, void (subcc::*operation)(bool),bool param) 
  {
    if(psubcon->isShared()) 
    {
	psubcon->setSharedOpNow(true);
	(psubcon.get()->*operation)(param);
	psubcon->setSharedOpNow(false);
    };
  }
  
  /** Decay sub-component posteriors, if they are shared. otherwise, do nothing. 
  \param psubcon: shared_ptr to container of sub-components
  \param tau: decay constant
  */
  template<class subcc>
  void subDecayParameters(boost::shared_ptr<subcc> & psubcon,floatT tau, bool update_immediately = true) 
  {
    if(psubcon->isShared())
    {
      psubcon->setSharedOpNow(true);
      psubcon->decayParameters(tau,update_immediately);
      psubcon->setSharedOpNow(false);
    }
  }
  
  /** collect responsibilities from all components, multiply with parent posterior. 
  Used during learning and prediction.
  \param parentPosteriors: posterior probabilities (responsibilities) of the components' parents
  \param postGetter: pointer to member of component which retrieves posterior of subordinates given component and segment
  \return responsibilities of the sub-components. If this is an empty vector, no significant responsibilites could be found.
  */
  template<class compT>
  floatMatT collectSubComponentResponsibilities(const floatMatT & parentPosteriors,const floatMatT & (compT::*postGetter)() const) 
  {
    floatMatT respo;
    RERR( ((int)parentPosteriors.size1()!=This::segments() && (int)parentPosteriors.size2()!=This::components() ) 
	, (boost::format("Error collecting responsibilities: got (%d,%d) parent posteriors, but have (%d,%d) segments/components") % parentPosteriors.size1() % parentPosteriors.size2() % This::segments() % This::components() ).str() );
    for(int pseg=0;pseg<This::segments();++pseg)
      for(int pcomp=0;pcomp<This::components();++pcomp)
      {
	if(parentPosteriors(pseg,pcomp)<numeric_limits<floatT>::epsilon()) continue;
	const floatMatT & cpost=(This::mComponents(pseg,pcomp).get()->*postGetter)();
	if(respo.size1()==0) respo=parentPosteriors(pseg,pcomp)*cpost;
	else respo+=parentPosteriors(pseg,pcomp)*cpost;
      }
    return respo;
  }
};



/** Container for variational_model components. Useful for models with sub-components, e.g. DPMM or sequence_model. Also efficient for component sharing across models --
likelihoods of sub-models need to be evaluated only once for each new datapoint. Can hold another component_container of sub-subordinate components, thereby enabling
the building of tree-structured hierarchies. If the data are segmented (e.g. image patches), the prototype controls number of patches. */
template<class component_varmodel_type> 
class componentContainer : public componentContainerBase<component_varmodel_type> 
{
public: 
  
  typedef componentContainerBase<component_varmodel_type> baseT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::floatT floatT;
  
  typedef typename baseT::componentT componentT;
  typedef typename baseT::pComponentT pComponentT;
  typedef typename baseT::componentsT componentsT;
  
  typedef typename componentT::componentContainerT subComponentContainerT;
  typedef typename componentT::pComponentContainerT pSubContainerT;
  typedef typename subComponentContainerT::componentT subComponentT;
  typedef typename subComponentContainerT::pComponentT pSubComponentT;
  
  typedef typename componentT::dataT dataT;
  typedef typename baseT::pdataT pdataT;
  
  typedef componentContainer<component_varmodel_type> This;
  

  /** Construct an empty, non-shared container */
  componentContainer() 
  : baseT()
  , mSubComponents(new subComponentContainerT())
  {
  };
  
  /** Copy-construct. Parameters will be shared, hidden variables not. */
  componentContainer(const This &source) : baseT(source){
    mSubComponents=pSubContainerT( new subComponentContainerT( *source.mSubComponents ) );
    doShareWith(source);    
  }
  
  /** share parameters with components (and sub-components) in source */
  void shareWith(const This &source) 
  {
     if(This::mbShared && !This::mbSharedOpNow) return;
     mSubComponents->setSharedOpNow(true);
     mSubComponents->shareWith(*source.mSubComponents);
     mSubComponents->setSharedOpNow(false);
     doShareWith(source);
  }
  
  
  /** make parameters of components private */
  void unshare()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::unshare();
    mSubComponents->setSharedOpNow(true);
    mSubComponents->unshare();
    mSubComponents->setSharedOpNow(false);
  }
   

  /** \return shared_ptr to container of sub-components */
  pSubContainerT getSubContainer() {return mSubComponents;};
  
  /** Set sub-container 
  \param shared_ptr to a sub-container.
  */
  void setSubContainer(pSubContainerT & subc) {
    mSubComponents=subc;
    for(int seg=0;seg<This::segments();++seg)
      for(int comp=0;comp<This::components();++comp)
	This::mComponents(seg,comp)->notifyResize(mSubComponents->segments(),mSubComponents->components());
  };
  
 
  /** Resize container to newsize. If newsize>current size, keep existing components. Init new components to prototype. Segments from prototype
  \param newsize: new number of components. Segments from prototype.
  */
  int resize(int newsize,bool keepOld=false) 
  {
    int oldsize=baseT::resize(newsize,keepOld);
    for(int i=0;i<newsize;++i) This::setCompSub(i,mSubComponents,&componentT::setComponentContainer);
    return oldsize;
  }
  
  /** \param prototype: the prototype to which new components of this container are initialized. Important for segmented data, because the prototype does the segmenting.
  Clears components.
  */
  void setPrototype(const component_varmodel_type & prototype) {
    This::mPrototype=prototype; 
    This::mComponents.resize(0,0);
  };


  
  /** Run inference for all components in container. If shared, run only when setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call inference(data_ptr) for each component.
  If sub-components are shared: run inference(data_ptr) on sub-components. Then, run inference(null_ptr) on all components, which has to have the effect that the components do not call inference on their component containers again.
  \param data_ptr: shared_ptr to data (of segment, if segmented)
  \param variational: if true, compute variational expectations.
  */
  void inference(const pdataT & data_ptr,bool variational=true,unsigned segment=0) 
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    subInference(mSubComponents,data_ptr,variational); 
    baseT::inference(data_ptr,variational);
  }
  
  /** Learn for all components in container. If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call learn() for each component.
  If sub-components are shared: Call learn() for each component. Collect component posteriors (responsibilities) from all components, then run learn for sub-components.
  \param data_ptr: shared_ptr to data.
  \param parentPosterior: matrix(segment,component) of parent posteriors.
  */
  void learn(const pdataT & data_ptr,const floatMatT & parentPosteriors) 
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::learn(data_ptr,parentPosteriors);
   
    if(mSubComponents->isShared()) 
    {
      floatMatT subpo=collectSubComponentResponsibilities(parentPosteriors,&componentT::getComponentPosteriors);
      RERR( ((int)subpo.size1()!=mSubComponents->segments() || (int)subpo.size2()!=mSubComponents->components())
	, (boost::format("Error in learning shared subcomponents: got (%d,%d) responsibilities, but have (%d,%d) components!") % subpo.size1() % subpo.size2() % mSubComponents->segments() % mSubComponents->components()).str() );
      subLearn(mSubComponents,data_ptr,subpo);
      
    }
  }
  
  /** set sequence learning phase. for meaning, \see variationalModel::seqPhaseT */
  void setSequenceLearningPhase(int phase=0) {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::setSequenceLearningPhase(phase);
    subExecI(mSubComponents,&subComponentContainerT::setSequenceLearningPhase,phase);
  }
  
  /** Predicition pass for all components in container. If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call predictPass() for each component.
  If sub-components are shared: Call predictPass() for each component. Collect component posteriors (responsibilities) from all components, then run predictPass for sub-components.
  \param parentPosterior: matrix(segment,component) of parent posteriors.
  */
  void predictPass(const floatMatT & parentPosteriors) 
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::predictPass(parentPosteriors);
   
    if(mSubComponents->isShared()) 
    {
      floatMatT subpo=collectSubComponentResponsibilities(parentPosteriors,&componentT::getComponentPosteriors);
      RERR( ((int)subpo.size1()!=mSubComponents->segments() || (int)subpo.size2()!=mSubComponents->components())
	, (boost::format("Error in prediction pass for shared subcomponents: got (%d,%d) responsibilities, but have (%d,%d) components!") % subpo.size1() % subpo.size2() % mSubComponents->segments() % mSubComponents->components()).str() );
      subExec_M(mSubComponents,&subComponentContainerT::predictPass,subpo);      
    }
  }


  /** \return accumulated variational bound from components */
  floatT collectVarBound()
  {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    return baseT::collectVarBound()+subExecF(mSubComponents,&subComponentContainerT::collectVarBound);
  }
    
  /** Do variational posterior parameter update for all components in container. 
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call updateParameterPosterior() for each component.
  If sub-components are shared: also call for sub-components.
  */
  void updateParameterPosterior()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::updateParameterPosterior();
    subExec(mSubComponents,&subComponentContainerT::updateParameterPosterior);
  }
  
  /** Reset variational parameter posterior to prior for all components.
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call resetParameterPosteriorToPrior() for each component.
  If sub-components are shared: also call for sub-components.
  */
  void resetParameterPosteriorToPrior()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::resetParameterPosteriorToPrior();
    subExec(mSubComponents,&subComponentContainerT::resetParameterPosteriorToPrior);
  }
  
  
  /** Make current posterior new prior.
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call updateParameterPosterior() for each component.
  If sub-components are shared: also call for sub-components.
  */
  void posteriorToPrior()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::posteriorToPrior();
    subExec(mSubComponents,&subComponentContainerT::posteriorToPrior);
  }
  
  /** Decay posterior certainty. Useful for escaping local minima etc.
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call updateParameterPosterior() for each component.
  If sub-components are shared: also call for sub-components.
  \param tau: decay parameter, typically certainty is reduced by factor exp(-1.0/tau)
  */
  void decayParameters(floatT tau,bool update_immediately=true) 
  {
     if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::decayParameters(tau,update_immediately);
     subDecayParameters(mSubComponents,tau,update_immediately);
  }
  
  
  /** reset hidden variables of components to prior. If shared, run only if setSharedOpNow(true) has been called before */
  void resetToPrior() {
     if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::resetToPrior();
     subExec(mSubComponents,&subComponentContainerT::resetToPrior);
  }
  
  /** reset variational bound accumulators of all components. If shared, run only if setSharedOpNow(true) has been called before */
  void resetVarBound() {
     if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::resetVarBound();
     subExec(mSubComponents,&subComponentContainerT::resetVarBound);
  }
  
  /** begin learning in all components. If shared, run only if setSharedOpNow(true) has been called before */
  void beginLearning(bool accumulateVarBound=false) {
    if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::beginLearning(accumulateVarBound);
     subExec_B(mSubComponents,&subComponentContainerT::beginLearning,accumulateVarBound);
  }
  
  /** collect KL divergence of all components. If shared, run only if setSharedOpNow(true) has been called before */
  floatT collectKL() {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    return baseT::collectKL()+subExecF(mSubComponents,&subComponentContainerT::collectKL);
  }
  
  /** get entropy of components, weighted by parent posterior.If shared, run only if setSharedOpNow(true) has been called before 
  \param parentPosts: (accumulated) posteriors of parents, one per component (components have same conditional entropy across segments)
  */
  floatT collectEntropy(const floatVecT & parentPosts)
  {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    floatT retval=0.0;
    if(mSubComponents->isShared()) 
    {
	RERR( ((int)parentPosts.size()!=This::components()) , (boost::format("Error collecting entropy: got %d components, need %d") % parentPosts.size() % This::components()).str());
	floatMatT parpo=ublas::matrix<floatT>(This::segments(),parentPosts.size());
	for(unsigned seg=0;seg<parpo.size1();++seg) ublas::matrix_row<floatMatT>(parpo,seg)=parentPosts;
	// all segments have same conditional entropy. Thus, just collect posteriors across segments
	floatMatT subpo=collectSubComponentResponsibilities(parpo,&componentT::getComponentPosteriors);
	floatVecT sr=ublas::matrix_row<floatMatT>(subpo,0);
	for(unsigned seg=1;seg<subpo.size1();++seg) sr+=ublas::matrix_column<floatMatT>(subpo,seg);
	
	retval=subExecF_V(mSubComponents,&subComponentContainerT::collectEntropy,sr);
    }
    return retval;
  }
  
  
  /** get expected log^2(p) of components, weighted by parent posterior.If shared, run only if setSharedOpNow(true) has been called before 
  \param parentPosts: (accumulated) posteriors of parents
  */
  floatT collectEntropy2(const floatVecT & parentPosts)
  {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    floatT retval=0.0;
    if(mSubComponents->isShared()) 
    {
      	RERR( ((int)parentPosts.size()!=This::components()) , (boost::format("Error collecting entropy: got %d components, need %d") % parentPosts.size() % This::components()).str());
	floatMatT parpo=ublas::matrix<floatT>(mSubComponents->segments(),parentPosts.size());
	for(unsigned seg=0;seg<parpo.size1();++seg) ublas::matrix_row<floatMatT>(parpo,seg)=parentPosts;
	// all segments have same conditional entropy. Thus, just collect posteriors across segments
	floatMatT subpo=collectSubComponentResponsibilities(parpo,&componentT::getComponentPosteriors);
	floatVecT sr=ublas::matrix_row<floatMatT>(subpo,0);
	for(unsigned seg=1;seg<subpo.size1();++seg) sr+=ublas::matrix_column<floatMatT>(subpo,seg);
	retval=subExecF_V(mSubComponents,&subComponentContainerT::collectEntropy2,sr);
    }
    return retval;
  }
  
  
  
    
 
  /** Write to stream. 
  If this container is not shared, just write all components and sub-components. If it is shared, do nothing.
  \param os  reference to ostream to write into.
  \param varmod  reference to component container
  \return ostream after varmod has been inserted
  */
  friend ostream &operator<<(ostream &os,componentContainer<component_varmodel_type> &cc) 
  {
	if(cc.mbShared && !cc.mbSharedOpNow) return os;
	
	os<<"Component container"<<endl;
 	os<<"Serializing sub-container..."<<endl;
 	cc.mSubComponents->setSharedOpNow(true);
 	os<<*cc.mSubComponents;
 	cc.mSubComponents->setSharedOpNow(false);
 	os<<"Serializing sub-container done!"<<endl;
	
	os<<"Number of components: "<<cc.components()<<endl;
	os<<"Share type: "<<(cc.isShared()?"shared":"not_shared")<<endl;
	os<<"Prototype"<<endl;
	os<<cc.mPrototype;
	if(cc.segments()>0)
	  for(int comp=0;comp<cc.components();++comp) os<<*(cc.mComponents(0,comp));

        return os;
  }
  
  /** read from stream is. If this container is not shared, read all components and sub-components. If it is shared, do nothing.
          \param is  reference to stream to read from.
          \param varmod reference to variationalModel to read into
          \return stream is after variationalModel has been read.
  */
  friend istream &operator>>(istream &is,This &cc) 
  {
	std::string name;
        if(cc.mbShared && !cc.mbSharedOpNow) return is;
	
	getline(is,name);
	if(name!="Component container") throw std::runtime_error("Error loading component container: got "+name);
 	getline(is,name);
 	cc.mSubComponents->setSharedOpNow(true);
 	is>>*cc.mSubComponents;
 	cc.mSubComponents->setSharedOpNow(false);
 	getline(is,name);
	
	int numcmp;
	is>>eatString(3)>>numcmp>>ws;
	
	is>>eatString(2)>>name>>ws;
	cc.setShared(name=="shared");
	getline(is,name);
	is>>cc.mPrototype>>ws;
	
	cc.resize(numcmp);
	
	if(cc.segments()>0)
	  for(int comp=0;comp<cc.components();++comp)
	  {
	    This::pComponentT pcomp=cc.mComponents(0,comp);
	    if(cc.mSubComponents->isShared()) pcomp->setComponentContainer(cc.mSubComponents);
	    is>>*pcomp;
	    pcomp->setSegment(0);
	    pcomp->setIndex(comp);
	    for(int seg=1;seg<cc.segments();++seg)
	    {
	      This::pComponentT pcomp2=(cc.mComponents(seg,comp)=This::pComponentT( new This::componentT(*pcomp) ) );
	      pcomp2->setSegment(seg);
	      pcomp2->setIndex(comp);
	    }
	  }
        return is;
  }
  
  
  friend bool operator==(const This & first,const This & second) 
  {
   return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {      
      if(!(first.mbShared==second.mbShared)) return false;
      if( (first.mbShared && !first.mbSharedOpNow) || (second.mbShared && !second.mbSharedOpNow) ) return true; // to avoid multiply comparisons in a shared hierarchy
      
	
      first.mSubComponents->setSharedOpNow(true);
      second.mSubComponents->setSharedOpNow(true);
      bool rv=(*first.mSubComponents==*second.mSubComponents);
      first.mSubComponents->setSharedOpNow(false);
      second.mSubComponents->setSharedOpNow(false);
      
      rv &= baseT::cmp(first,second);
      
      return rv;
      
      
  }
    
protected:  
  pSubContainerT mSubComponents; //!< shared components subordinate to the components in this container 
  
  
  /** share parameters with components (and sub-components) in source */
  void doShareWith(const This &source) 
  {
   baseT::doShareWith(source);
   if(mSubComponents->isShared()) 
      for(int seg=0;seg<This::segments();++seg)
	for(int comp=0;comp<This::components();++comp)
	  This::mComponents(seg,comp)->setComponentContainer(mSubComponents);
    else assert(false); // sub-components should always be shared.
  }
};



/** Like componentContainer, but expects component_varmodel_type to implement the variationalComponentModel2 interface, which has 2 types of components, e.g. sequences and labels */
template<class component_varmodel_type> 
class componentContainer2 : public componentContainerBase<component_varmodel_type>
{
public: 
  
  typedef componentContainerBase<component_varmodel_type> baseT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::floatT floatT;
  
  typedef typename baseT::componentT componentT;
  typedef typename baseT::pComponentT pComponentT;
  typedef typename baseT::componentsT componentsT;
  
  typedef typename component_varmodel_type::componentContainerT1 subComponentContainerT1;
  typedef boost::shared_ptr<subComponentContainerT1> pSubContainerT1;
  typedef typename subComponentContainerT1::componentT subComponentT1;
  typedef typename subComponentContainerT1::pComponentT pSubComponentT1;
  
  typedef typename component_varmodel_type::componentContainerT2 subComponentContainerT2;
  typedef boost::shared_ptr<subComponentContainerT2> pSubContainerT2;
  typedef typename subComponentContainerT2::componentT subComponentT2;
  typedef typename subComponentContainerT2::pComponentT pSubComponentT2;
  
  typedef componentContainer2<component_varmodel_type> This;
  
  
  typedef typename componentT::dataT dataT;
  typedef typename componentT::pdataT pdataT;
  

  /** Construct an empty, non-shared container */
  componentContainer2() 
  : baseT()
  , mSubComponents1(new subComponentContainerT1())
  , mSubComponents2(new subComponentContainerT2())
  {};
  
  
  /** Copy-construct. Parameters will be shared, hidden variables not. */
  componentContainer2(const This &source) : baseT(source){
    mSubComponents1=pSubContainerT1( new subComponentContainerT1( source->mSubComponentContainer1 ) );
    mSubComponents2=pSubContainerT2( new subComponentContainerT2( source->mSubComponentContainer2 ) );
    doShareWith(source);
  }
  
  /** share parameters with components (and sub-components) in source */
  void shareWith(const This &source)
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    mSubComponents1->setSharedOpNow(true);
    mSubComponents2->setSharedOpNow(true);
    mSubComponents1->shareWith(*source.mSubComponents1);
    mSubComponents2->shareWith(*source.mSubComponents2);
    mSubComponents1->setSharedOpNow(false);
    mSubComponents2->setSharedOpNow(false);
    doShareWith(source); 
  }
  

  
  /** make parameters of components private */
  void unshare()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::unshare();
    mSubComponents1->setSharedOpNow(true);
    mSubComponents2->setSharedOpNow(true);
    mSubComponents1()->unshare();
    mSubComponents2()->unshare();
    mSubComponents1->setSharedOpNow(false);
    mSubComponents2->setSharedOpNow(false);
  }

  
  /** \param prototype: the prototype to which new components of this container are initialized. Important for segmented data, because the prototype does the segmenting. 
  Resizes the sub-container to match prototype's number of segments. Clears components.
  */
  void setPrototype(const component_varmodel_type & prototype) {
    This::mPrototype=prototype; 
    mSubComponents1->resize(mSubComponents1->size(),This::prototype.segments());
    mSubComponents2->resize(mSubComponents2->size(),This::prototype.segments());
    This::mComponents.resize(0,0);
  };
   
  
  /** \return shared_ptr to container of sub-components, 1st type */
  pSubContainerT1 getSubContainer1() {return mSubComponents1;};
  
  /** Set sub-container, 1st type.
  \param shared_ptr to a sub-container.
  */
  void setSubContainer1(pSubContainerT1 & subc) {
    mSubComponents1=subc;
    for(int seg=0;seg<This::segments();++seg) 
      for(int comp=0;comp<This::components();++comp)
	This::mComponents(seg,comp)->notifyResize(mSubComponents1->segments(),mSubComponents1->components());
  };
  
  
  /** \return shared_ptr to container of sub-components, 2nd type */
  pSubContainerT2 getSubContainer2() {return mSubComponents2;};
  
  /** Set sub-container, 2nd type.s
  \param shared_ptr to a sub-container.
  */
  void setSubContainer2(pSubContainerT2 & subc) {
    mSubComponents2=subc;
    for(int seg=0;seg<This::segments();++seg) 
      for(int comp=0;comp<This::components();++comp)
	This::mComponents(seg,comp)->notifyResize(mSubComponents2->segments(),mSubComponents2->components());
  };
  
  

  
  /** Resize container to newsize. If newsize>current size, keep existing components. Init new components to prototype.
  \param newsize: new number of components. Segments from prototype. */
  int resize(int newsize,bool keepOld=false) 
  {
    int oldsize=baseT::resize(newsize,keepOld);
    for(int i=0;i<newsize;++i) {
      This::setCompSub(i,mSubComponents1,&componentT::setComponentContainer1);
      This::setCompSub(i,mSubComponents2,&componentT::setComponentContainer2);
    }
    RERR( (mSubComponents1->segments()!=mSubComponents2->segments() || mSubComponents1->components()!=mSubComponents2->components())
	, (boost::format("Error resizing componentContainer2: found (%d,%d) tpye 1 and (%d,%d) type 2 components") % mSubComponents1->segments() % mSubComponents1->components() % mSubComponents2->segments() % mSubComponents2->components()).str() );
    return oldsize;
  }
  
  /** Run inference for all components in container.
  If sub-components are not shared: just call inference(data_ptr) for each component.
  If sub-components are shared: run inference(data_ptr) on sub-components. Then, run inference(null_ptr) on all components, which has to have the effect that the components do not call inference on their component containers again.
  \param data_ptr: shared_ptr to data.
  \param variational: if true, compute variational expectations.
  */
  void inference(const pdataT & data_ptr,bool variational=true) 
  {
      if(This::mbShared && !This::mbSharedOpNow) return;
      
      subInference(mSubComponents1,data_ptr->first,variational);
      subInference(mSubComponents2,data_ptr->second,variational);
      
      baseT::inference(data_ptr,variational);
  }
  
  /** Learn for all components in container. If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call learn() for each component.
  If sub-components are shared: Call learn() for each component. Collect component posteriors (responsibilities) from all components, then run learn for sub-components.
  \param data_ptr: shared_ptr to data.
  \param parentPosterior: vector of component posteriors (responsibilities). One per component.
  */
  void learn(const pdataT & data_ptr,const floatMatT & parentPosteriors) 
  {
    
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::learn(data_ptr,parentPosteriors);
   
    if(mSubComponents1->isShared() || mSubComponents2->isShared()) 
    {
      floatMatT subpo=collectSubComponentResponsibilities(parentPosteriors,&componentT::getComponentPosteriors);
      RERR( (subpo.size1()!=mSubComponents1->segments() || subpo.size2()!=mSubComponents1->components())
	, (boost::format("Error in learning shared subcomponents type 1: got (%d,%d) responsibilities, but have (%d,%d) components!") % subpo.size1() % subpo.size2() % mSubComponents1->segments() % mSubComponents1->components()).str() );
      RERR( (subpo.size1()!=mSubComponents2->segments() || subpo.size2()!=mSubComponents2->components())
	, (boost::format("Error in learning shared subcomponents type 2: got (%d,%d) responsibilities, but have (%d,%d) components!") % subpo.size1() % subpo.size2() % mSubComponents2->segments() % mSubComponents2->components()).str() );
      subLearn(mSubComponents1,data_ptr->first);
      subLearn(mSubComponents2,data_ptr->first);
    }    
  }
  
  /** set sequence learning phase. for meaning, \see variationalModel::seqPhaseT */
  void setSequenceLearningPhase(int phase=0) {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::setSequenceLearningPhase(phase);
    subExecI(mSubComponents1,&subComponentContainerT1::setSequenceLearningPhase,phase);
    subExecI(mSubComponents2,&subComponentContainerT2::setSequenceLearningPhase,phase);
  }
  
  
  /** Predicition pass for all components in container. If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call predictPass() for each component.
  If sub-components are shared: Call predictPass() for each component. Collect component posteriors (responsibilities) from all components, then run predictPass for sub-components.
  \param parentPosterior: matrix(segment,component) of parent posteriors.
  */
  void predictPass(const floatMatT & parentPosteriors) 
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::predictPass(parentPosteriors);
   
    if(mSubComponents1->isShared() || mSubComponents2->isShared()) 
    {
      floatMatT subpo=collectSubComponentResponsibilities(parentPosteriors,&componentT::getComponentPosteriors);
      RERR( (subpo.size1()!=mSubComponents1->segments() || subpo.size2()!=mSubComponents1->components())
	, (boost::format("Error in prediction pass for shared subcomponents type 1: got (%d,%d) responsibilities, but have (%d,%d) components!") % subpo.size1() % subpo.size2() % mSubComponents1->segments() % mSubComponents1->components()).str() );
      RERR( (subpo.size1()!=mSubComponents2->segments() || subpo.size2()!=mSubComponents2->components())
	, (boost::format("Error in prediction pass for shared subcomponents type 2: got (%d,%d) responsibilities, but have (%d,%d) components!") % subpo.size1() % subpo.size2() % mSubComponents2->segments() % mSubComponents2->components()).str() );
      subExec_M(mSubComponents1,&subComponentContainerT1::predictPass,subpo);      
      subExec_M(mSubComponents2,&subComponentContainerT2::predictPass,subpo);      
    }
  }
  
 
  /** \return accumulated variational bound from components */
  floatT collectVarBound()
  {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    return baseT::collectVarBound()
	   +subExecF(mSubComponents1,&subComponentContainerT1::collectVarBound)
	   +subExecF(mSubComponents2,&subComponentContainerT2::collectVarBound);
  }

 
  /** Do variational posterior parameter update for all components in container. 
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call updateParameterPosterior() for each component.
  If sub-components are shared: also call for sub-components.
  */
  void updateParameterPosterior()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::updateParameterPosterior();
    subExec(mSubComponents1,&subComponentContainerT1::updateParameterPosterior);
    subExec(mSubComponents2,&subComponentContainerT2::updateParameterPosterior);
  }
  
  /** Reset variational parameter posterior to prior for all components.
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call resetParameterPosteriorToPrior() for each component.
  If sub-components are shared: also call for sub-components.
  */
  void resetParameterPosteriorToPrior()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::resetParameterPosteriorToPrior();
    subExec(mSubComponents1,&subComponentContainerT1::resetParameterPosteriorToPrior);
    subExec(mSubComponents2,&subComponentContainerT2::resetParameterPosteriorToPrior);
  }
  
  
  
  /** Make current posterior new prior.
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call updateParameterPosterior() for each component.
  If sub-components are shared: also call for sub-components.
  */
  void posteriorToPrior()
  {
    if(This::mbShared && !This::mbSharedOpNow) return;
    baseT::posteriorToPrior();
    subExec(mSubComponents1,&subComponentContainerT1::posteriorToPrior);
    subExec(mSubComponents2,&subComponentContainerT2::posteriorToPrior);
  }
  
  /** Decay posterior certainty. Useful for escaping local minima etc.
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call updateParameterPosterior() for each component.
  If sub-components are shared: also call for sub-components.
  \param tau: decay parameter, typically certainty is reduced by factor exp(-1.0/tau)
  */
  void decayParameters(floatT tau,bool update_immediately=true) 
  {
     if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::decayParameters(tau,update_immediately);
     subDecayParameters(mSubComponents1,tau,update_immediately);
     subDecayParameters(mSubComponents2,tau,update_immediately);
  }
  
  
  /** reset hidden variables of components to prior. If shared, run only if setSharedOpNow(true) has been called before */
  void resetToPrior() {
     if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::resetToPrior();
     subExec(mSubComponents1,&subComponentContainerT1::resetToPrior);
     subExec(mSubComponents2,&subComponentContainerT2::resetToPrior);
  }
  
  /** reset variational bound accumulators of all components. If shared, run only if setSharedOpNow(true) has been called before */
  void resetVarBound() {
     if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::resetVarBound();
     subExec(mSubComponents1,&subComponentContainerT1::resetVarBound);
     subExec(mSubComponents2,&subComponentContainerT2::resetVarBound);
  }
  
  /** begin learning in all components. If shared, run only if setSharedOpNow(true) has been called before */
  void beginLearning(bool accumulateVarBound=false) {
    if(This::mbShared && !This::mbSharedOpNow) return;
     baseT::beginLearning(accumulateVarBound);
     subExec(mSubComponents1,&subComponentContainerT1::beginLearning,accumulateVarBound);
     subExec(mSubComponents2,&subComponentContainerT2::beginLearning,accumulateVarBound);
  }
  
  /** collect KL divergence of all components. If shared, run only if setSharedOpNow(true) has been called before */
  floatT collectKL() {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    return baseT::collectKL()+subExecF(mSubComponents1,&subComponentContainerT1::collectKL)
	    +subExecF(mSubComponents2,&subComponentContainerT2::collectKL);
  }
  
  /** get entropy of components, weighted by parent posterior.If shared, run only if setSharedOpNow(true) has been called before 
  \param parentPosts: (accumulated) posteriors of parents
  */
  floatT collectEntropy(const floatVecT & parentPosts)
  {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    floatT retval=0.0;
    if(mSubComponents1->isShared() || mSubComponents2->isShared()) 
    {
	// all segments have same conditional entropy. Thus, just collect posteriors across segments
	floatMatT subpo=collectSubComponentResponsibilities(parentPosts,&componentT::getComponentPosteriors);
	if(subpo.size1()==0 || subpo.size2()==0) return retval;
	floatVecT sr=ublas::matrix_column<floatT>(subpo,0);
	for(int col=1;col<subpo.size1();++col) sr+=ublas::matrix_column<floatT>(subpo,col);
	retval=subExecF_V(mSubComponents1,&subComponentContainerT1::collectEntropy,sr)+subExecF_V(mSubComponents2,&subComponentContainerT2::collectEntropy,sr);
    }
    return retval;
    
  }
  
  
  /** get expected log^2(p) of components, weighted by parent posterior.If shared, run only if setSharedOpNow(true) has been called before 
  \param parentPosts: (accumulated) posteriors of parents
  */
  floatT collectEntropy2(const floatVecT & parentPosts)
  {
    if(This::mbShared && !This::mbSharedOpNow) return 0.0;
    floatT retval=0.0;
    if(mSubComponents1->isShared() || mSubComponents2->isShared()) 
    {
	// all segments have same conditional entropy. Thus, just collect posteriors across segments
	floatMatT subpo=collectSubComponentResponsibilities(parentPosts,&componentT::getComponentPosteriors);
	if(subpo.size1()==0 || subpo.size2()==0) return retval;
	floatVecT sr=ublas::matrix_column<floatT>(subpo,0);
	for(int col=1;col<subpo.size1();++col) sr+=ublas::matrix_column<floatT>(subpo,col);
	retval=subExecF_V(mSubComponents1,&subComponentContainerT1::collectEntropy2,sr)+subExecF_V(mSubComponents2,&subComponentContainerT2::collectEntropy2,sr);
    }
    return retval;
  }
  
 
  /** Write to stream. 
  If this container is not shared, just write all components and sub-components. If it is shared, do nothing.
  \param os  reference to ostream to write into.
  \param varmod  reference to component container
  \return ostream after varmod has been inserted
  */
  friend ostream &operator<<(ostream &os,componentContainer2<component_varmodel_type> &cc) 
  {
	if(cc.mbShared && !cc.mbSharedOpNow) return os;

	os<<"Component container, 2 subordinate types"<<endl;
	
 	os<<"Serializing sub-container 1..."<<endl;
 	cc.mSubComponents1->setSharedOpNow(true);
 	os<<*cc.mSubComponents1;
 	cc.mSubComponents1->setSharedOpNow(false);
 	os<<"Serializing sub-container done!"<<endl;
	
	os<<"Serializing sub-container 2..."<<endl;
 	cc.mSubComponents2->setSharedOpNow(true);
 	os<<*cc.mSubComponents2;
 	cc.mSubComponents2->setSharedOpNow(false);
 	os<<"Serializing sub-container done!"<<endl;
	
	os<<"Number of components: "<<cc.size()<<endl;
	os<<"Share type: "<<(cc.isShared()?"shared":"not_shared")<<endl;
	os<<"Prototype"<<endl;
	os<<cc.mPrototype;
	if(cc.segments()>0) 
	  for(unsigned comp=0;comp<cc.size();++comp) os<<*(cc.mComponents[comp][0]);
	

        return os;
  }
  
  /** read from stream is. If this container is not shared, read all components and sub-components. If it is shared, do nothing.
          \param is  reference to stream to read from.
          \param varmod reference to variationalModel to read into
          \return stream is after variationalModel has been read.
  */
  friend istream &operator>>(istream &is,componentContainer2<component_varmodel_type> &cc) 
  {
	
	std::string name;
        if(cc.mbShared && !cc.mbSharedOpNow) return is;
        
	getline(is,name);
	if(name!="Component container, 2 subordinate types") throw std::runtime_error("Error loading component container with 2 subordinate types: got "+name);
	
 	getline(is,name);
 	cc.mSubComponents1->setSharedOpNow(true);
 	is>>*cc.mSubComponents1;
 	cc.mSubComponents1->setSharedOpNow(false);
 	getline(is,name);
	
	getline(is,name);
 	cc.mSubComponents2->setSharedOpNow(true);
 	is>>*cc.mSubComponents2;
 	cc.mSubComponents2->setSharedOpNow(false);
 	getline(is,name);
	
	int numcmp;
	is>>eatString(3)>>numcmp>>ws;
	is>>eatString(2)>>name>>ws;
	cc.setShared(name=="shared");
	getline(is,name);
	is>>cc.mPrototype>>ws;
	
	cc.resize(numcmp);
	if(cc.segments()>0)
	  for(int comp=0;comp<cc.components();++comp)
	  {
	    This::pComponentT pc=cc.mComponents[comp][0];
	    if(cc.mSubComponents1->isShared()) pc->setComponentContainer1(cc.mSubComponents1);
	    if(cc.mSubComponents2->isShared()) pc->setComponentContainer2(cc.mSubComponents2);
	    is>>*pc;
	    pc->setSegment(0);
	    pc->setIndex(comp);
	    
	    for(int seg=1;seg<cc.segments();++seg)
	    {
	      This::pComponentT pc2=(cc.mComponents[comp][seg]=This::pComponentT( new This::componentT(*pc) ) );
	      pc2->setSegment(seg);
	      pc2->setIndex(comp);
	    }
	}
	
        return is;
  }
  
  friend bool operator==(const This & first,const This & second) 
  {
   return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
      if(!(first.mbShared==second.mbShared)) return false;
      if( (first.mbShared && !first.mbSharedOpNow) || (second.mbShared && !second.mbSharedOpNow) ) return true; // to avoid multiply comparisons in a shared hierarchy
      
      first.mSubComponents1->setSharedOpNow(true);
      second.mSubComponents1->setSharedOpNow(true);
      first.mSubComponents2->setSharedOpNow(true);
      second.mSubComponents2->setSharedOpNow(true);
      bool rv = (*first.mSubComponents1==*second.mSubComponents1);
      rv &= (*first.mSubComponents2==*second.mSubComponents2);
      first.mSubComponents1->setSharedOpNow(false);
      second.mSubComponents1->setSharedOpNow(false);
      first.mSubComponents2->setSharedOpNow(false);
      second.mSubComponents2->setSharedOpNow(false);
      
      if(!rv) return false;
    
      return baseT::cmp(first,second);
      
  }
    
protected:  
  pSubContainerT1 mSubComponents1; //!< shared components subordinate to the components in this container, type 1
  pSubContainerT2 mSubComponents2; //!< shared components subordinate to the components in this container, type 2
  
  /** share parameters with components (and sub-components) in source */
  void doShareWith(const This & source) 
  {
    RERR( (mSubComponents1->isShared()!=mSubComponents2->isShared())
    , "Error copy-constructing componentContainer2: share status does not match between sub-containers");
    if(mSubComponents1->isShared() ) 
      for(int seg=0;seg<This::segments();++seg)
	for(int comp=0;comp<This::components();++comp)
	{
	  This::mComponents(seg,comp)->setComponentContainer1(mSubComponents1);
	  This::mComponents(seg,comp)->setComponentContainer2(mSubComponents2);
	}   
    else assert(false); // sub-components should always be shared. 
  }
  
};



/** Base class for variational models with components, e.g. mixtures or HMMs. Knows its index and segment number in the parent container, if it has one. 
Do not instantiate.
*/
template<class data_type,typename float_type, class data_handle = boost::shared_ptr<data_type> >
class variationalComponentModelBase : public variationalModel<data_type,float_type,data_handle>
{
public:
  typedef variationalModel<data_type,float_type,data_handle> baseT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::floatT floatT;
  typedef typename baseT::pdataT pdataT;
  typedef typename baseT::dataT dataT;
  typedef typename baseT::intVecT intVecT;
  
  
  typedef variationalComponentModelBase<data_type,float_type> This;
  typedef typename baseT::varmodelBaseT varmodelBaseT;
  
  variationalComponentModelBase():baseT(),mComponentPosteriors() {};
  
  variationalComponentModelBase(const This & source ):baseT(source) 
  {
    doShareWith(source,false);
  }
  
  
  virtual int subSegments() const {
    if(mComponentPosteriors.size1()<1) return 1;
    return mComponentPosteriors.size1();
  };
  
  virtual int components() const {return mComponentPosteriors.size2();};
  
  
  /** \return component posterior vector for segment */
  const floatVecT getComponentPosteriors(unsigned segment) const
  {
    RERR( (segment>=mComponentPosteriors.size1()) , (boost::format("error getting component posteriors for data segment %d of %d") % segment % mComponentPosteriors.size1()).str());
    return ublas::matrix_row<const floatMatT>(mComponentPosteriors,segment);
  }
  
  /**  \return component posteriors(segment,component) */
  const floatMatT & getComponentPosteriors() const {return mComponentPosteriors;};
  
  
    /** Make current posterior the new prior. Call posteriorToPriorComponents() in derived implementations. */
  void posteriorToPrior()
  {
    baseT::posteriorToPrior();
    posteriorToPriorComponents(); // call this method in derived classes!
  }
  
  void freezeAll(int level=0)
  {
   This::freeze(level);
   freezeComponents(level);
  }
 
  /** Sample from posterior. Overwrite in derived classes. Call resetToPrior() if you want to sample from hidden var. prior. */
  pdataT sample()
  {
    Random<float> & rng = Random<float>::get();
    intVecT compos;
    RERR( (components()<1) , ("Error sampling from variationalComponentModel: no components found"));
    for(int segment=0;segment<subSegments();++segment) {
      compos.push_back( rng.sample_multinomial( ublas::matrix_row<floatMatT>(mComponentPosteriors,segment) ) );
    }
    return sampleComponents(compos);
  }
  
  /** Decay posterior certainty. Useful for escaping local minima etc. Call decayParametersComponents() in derived implementations.
  If shared, run only if setSharedOpNow(true) has been called before.
  If sub-components are not shared: just call updateParameterPosterior() for each component.
  If sub-components are shared: also call for sub-components.
  \param tau: decay parameter, typically certainty is reduced by factor exp(-1.0/tau)
  */
  void decayParameters(floatT tau,bool update_immediately=true) 
  {
     decayParametersComponents(tau,update_immediately);
  }
  
  
  /** \return total variational bound, i.e. accumulated hidden variable part and KL-div. If components are not shared, also include subordinates */
  floatT getVarBound()
  {
   return baseT::getAccumulatedVarBound()+getAccumulatedVarBoundComponents()-getKL(); 
  }
  
  /** reset hidden vars to prior. If components are not shared, also reset subordinates. */
  void resetToPrior() {
     baseT::resetToPrior();
     mComponentPosteriors=ublas::scalar_matrix<floatT>(subSegments(),components(),1.0/components());
     resetToPriorComponents();
  }
  
  /** reset accumulated variational bound to 0. If components are not shared, also reset subordinates */
  void resetVarBound() {
     baseT::resetVarBound();
     resetVarBoundComponents();
  }
  
  
  /** \return KL-divergence between variational parameter posterior and prior. If not shared, also include subordinates' KL-div */
  floatT getKL() {
    return baseT::getKL()+getKLComponents();
  }
  

  /** Notify derived classes that number of sub-components have changed. Base class method resizes component posterior.*/
  virtual void notifyResize(int segments,int components) {
    mComponentPosteriors=ublas::zero_matrix<floatT>(segments,components);
  };
  
  /** prediction pass */
  void predictPass(const floatT parentPost=1.0) {
   baseT::predictPass(parentPost);
   predictPassComponents(parentPost); // call this in derived classes
  }
  
  void setSequenceLearningPhase(int phase=0) {
    baseT::setSequenceLearningPhase(phase);
    setSequenceLearningPhaseComponents(phase);
  }
  
  
  
protected:  
  floatMatT mComponentPosteriors; //!< component posterior probabilites, e.g. responsibilities. One vector per segment.
  
  
  /** \return marginal posterior (across segments) of subordinates. */
  floatVecT getMarginalPosterior()
  {
    floatVecT marginalPost=ublas::zero_vector<floatT>(mComponentPosteriors.size2());
    for(unsigned seg=0;seg<mComponentPosteriors.size1();++seg) marginalPost+=ublas::matrix_row<floatMatT>(mComponentPosteriors,seg);
    return marginalPost;
  }

  virtual void posteriorToPriorComponents()=0;
  
  /** \return a sample from subordinate components */
  virtual pdataT sampleComponents(const intVecT &compos)=0;
  
  virtual void decayParametersComponents(floatT tau,bool update_immediately)=0;
  
  virtual floatT getAccumulatedVarBoundComponents()=0;
  
  virtual void resetToPriorComponents()=0;
  
  virtual void resetVarBoundComponents()=0;
  
  virtual void beginLearningComponents()=0;
  
  virtual floatT getKLComponents()=0;
  
  virtual void resetParameterPosteriorToPriorComponents()=0;
  
  virtual void predictPassComponents(const floatT parentPost)=0;
  
  virtual void freezeComponents(int level=0)=0;
  
  virtual void setSequenceLearningPhaseComponents(int phase=0)=0;
  
  void doShareWith(const This & source,bool shareBase=false) {
   if(shareBase) baseT::doShareWith(source,shareBase);
   mComponentPosteriors=source.mComponentPosteriors; 
  }
  
  void doBeginLearning() {
    beginLearningComponents();
  }

  
  void doResetParameterPosteriorToPrior() {
   resetParameterPosteriorToPriorComponents(); 
  }
  
  
  virtual void doInferenceComponents(const pdataT & data_ptr,bool variational)=0;
  
  /** \return likelihood of each (segment,component) for data from last doInferenceComponents() call.*/
  virtual floatMatT getLikelihoodsComponents() const=0;
  
  /** This base class takes full responsibility of the component with the same index as itself. If there is no such component, deny all responsibility 
  \return log-likelihood of *data_ptr
  */
  float_t doInference(const pdataT & data_ptr,bool variational) 
  {
    doInferenceComponents(data_ptr,variational);  
    
    // by default, deny all responsibility
    mComponentPosteriors=ublas::zero_matrix<floatT>(subSegments(),components());
    
    // take full responsibility of subordinate component with same index, if there is one
    if(baseT::mIndex>=0 && (unsigned)baseT::mIndex<components()) 
    {
	  floatT llhood=0.0;
	  for(int segment=0;segment<subSegments();++segment) 
	  {
	    mComponentPosteriors(segment,baseT::mIndex)=1.0;
	    llhood+=getLikelihoodsComponents()(segment,baseT::mIndex);
	  }
	  return llhood;
    }
	
    // otherwise, pretend we've not seen any data.
    return 0.0;
  }
  
  
  /** Accumulate variational bound for this object. */
  void accumulateVarBound(const pdataT & data_ptr, floatT parentPost=1.0) 
  {
   // this base class has no hidden vars and is not connected to the data. 
   // Thus, it has no var bound contribution other than through its subordinates, which are collected during learning.
  }
 
  
  
  virtual void doLearnComponents(const pdataT & data_ptr, const floatT parentPost)=0;
    
  /** Base class just learns components. */
  void doLearn(const pdataT & data_ptr, const floatT parentPost) 
  {
    doLearnComponents(data_ptr,parentPost);
  };
  
  virtual void variationalParameterUpdateComponents()=0;
  
  /** Do update of variational posterior parameters. Implement in derived classes */
  bool doVariationalParameterUpdate()
  {
    variationalParameterUpdateComponents(); // call this method in derived classes!
    return true;
  }
  

  

};


/** Subtype of variational model with components. e.g. mixtures or HMMs. Components are held in a component container type.
Knows its index in the parent container (if it has one). 
This base class takes full responsibility of the component with the same index as itself.

To instantiate a (derived) model of this type, do
\code
  variationalComponentModel<comp_container_type> model;
  model.set_property("alpha0",1.0);
  .
  .
  .
  model.resize(50); // for a model which does not share its components
  model.setComponentContainer(shared_ptr_to_container); // for a model which shares components
  
  model.setPrior(35.4); // init prior to a pseudocount of 35.4. Component container MUST be set before calling setPrior()!!!
  
\endcode

*/
template<class component_container_type>
class variationalComponentModel : public variationalComponentModelBase< typename component_container_type::componentT::dataT, typename component_container_type::componentT::floatT, typename component_container_type::componentT::pdataT >
{
public:
  
  typedef variationalComponentModelBase< typename component_container_type::componentT::dataT, typename component_container_type::componentT::floatT, typename component_container_type::componentT::pdataT > baseT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::floatT floatT;
  typedef typename baseT::intVecT intVecT;
  
  typedef typename component_container_type::componentT::dataT dataT;
  typedef typename baseT::pdataT pdataT;
  typedef component_container_type componentContainerT;
  typedef boost::shared_ptr<componentContainerT> pComponentContainerT;
  typedef typename componentContainerT::componentT componentT;
  typedef typename componentContainerT::pComponentT pComponentT;
  
  typedef variationalComponentModel<componentContainerT> This;
  typedef typename baseT::varmodelBaseT varmodelBaseT;
  
  variationalComponentModel()
  : baseT()
  , mComponents( new componentContainerT() )
  {
    This::mComponentPosteriors.resize(0,0);
  };
  
  variationalComponentModel(const This &source):baseT(source) 
  {
    if(source.mComponents->isShared()) mComponents=source.mComponents;
    else mComponents=pComponentContainerT(new componentContainerT( *source.mComponents ));
    doShareWith(source,false);
  }
  
    
  void unshare() {
    baseT::unshare();
    mComponents->unshare();
  }
  
  
  /** \return reference to component container */
  pComponentContainerT getComponentContainer() {return mComponents;};
  
  /** Set component container. Resize posterior.*/
  void setComponentContainer(pComponentContainerT & comptr) {
    mComponents=comptr;
    RERR( (mComponents->segments()<1), (boost::format("Error setting component container with sizes (%d,%d)") % mComponents->segments() % mComponents->components()).str() );
    This::mComponentPosteriors=ublas::zero_matrix<floatT>(mComponents->segments(),mComponents->components());
    this->notifyResize(mComponents->segments(),mComponents->components());
  };
  
  /** Set prior pseudocount to members of pseudodata.
  \param pseudodata: vector of pseudodata for each component
  */
  void setPrior(floatT pseudodata=0.0) 
  {
    baseT::setPrior(pseudodata);
    mComponents->setPrior(floatVecT(mComponents->components(),pseudodata));
  }
  
  /** Resize component container */
  void resize(int newsize)
  {
    RERR( (newsize<1) , (boost::format("Error resizing to %d") % newsize).str() );
    mComponents->resize(newsize);
    setComponentContainer(mComponents);
  }
    
  /** \return reference to component */
  pComponentT getComponent(int segment,int component)
  {
    return mComponents->getComponent(segment,component);
  }
  
  friend bool operator==(const This & first,const This & second) 
  {
   return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
    if(!(*first.mComponents==*second.mComponents)) return false;
    return baseT::cmp(first,second);
  }
  
  /** \return conditional entropy of data */
  floatT getEntropy()
  {
    // Since this model is not connected directly to the data, just collect entropy of subordinates weighted by current posteriors.
    return mComponents->collectEntropy(This::getMarginalPosterior()); 
  }
  
  /** \return E(log(P)^2) for variance of entropy */
  floatT getEntropy2()
  {
    // Since this model is not connected directly to the data, just collect entropy of subordinates weighted by current posteriors.
    return mComponents->collectEntropy2(This::getMarginalPosterior()); 
  }
 
  
protected:
  pComponentContainerT mComponents; //!< reference to container with components. Non-shared by default.

  
  void posteriorToPriorComponents()
  {
    mComponents->posteriorToPrior();
  }
    
  pdataT sampleComponents(const intVecT &compo)
  {
    RERR( ((int)compo.size() != mComponents->segments()), (boost::format("Error sampling components in variationalComponentModel: got %d, need %d") % compo.size() % mComponents->segments()).str());
    return mComponents->sample(compo);
  }
  
  void decayParametersComponents(floatT tau, bool update_immediately=true)
  {
    mComponents->decayParameters(tau,update_immediately);
  }
  
  
  floatT getAccumulatedVarBoundComponents() 
  {
   return mComponents->collectVarBound(); 
  }
  
  void resetVarBoundComponents() 
  {
   mComponents->resetVarBound();
  }
  
  void resetToPriorComponents() 
  {
   mComponents->resetToPrior();
  }
  
  
  void beginLearningComponents() 
  {
   mComponents->beginLearning(This::mbAccumulateVarBound);
  }
  
  void freezeComponents(int level=0)
  {
   mComponents->freeze(level); 
  }
 
  
  /** \return accumulated KL-div from components */
  floatT getKLComponents() 
  {
   return mComponents->collectKL(); 
  }
  
  
  void writeToStream(ostream &os) const 
  {
        os<<"Variational COMPONENT model base class"<<endl;
	os<<*mComponents;
	baseT::writeToStream(os);
  };
  
  
  void readFromStream(istream &is) 
  {
        string dummy; 
        getline(is,dummy);
	if(dummy!="Variational COMPONENT model base class") {
	  throw std::runtime_error("Error loading variational component model base class: got "+dummy);
	}
	is>>*mComponents;
	baseT::readFromStream(is);
	setComponentContainer(mComponents);
  }
  
  /** Do inference for components.
  \param data_ptr: datapoint
  \param variational: if true, compute variational expectations
  */
  void doInferenceComponents(const pdataT & data_ptr,bool variational) 
  {
    if(mComponents->isShared() && !mComponents->getSharedOpNow()) return;
    mComponents->inference(data_ptr,variational);
  }
  
  /** \return likelihood of components */
  floatMatT getLikelihoodsComponents() const
  {
    return mComponents->getComponentLikelihoods();
  }
  
  
  void doLearnComponents(const pdataT & data_ptr, const floatT parentPost)
  {
    if(parentPost<numeric_limits<floatT>::epsilon()) return;
    if(mComponents->isShared() && !mComponents->getSharedOpNow()) return;
    floatMatT chpost=parentPost*This::getComponentPosteriors();
    mComponents->learn(data_ptr,chpost);
  }
  
  void predictPassComponents(const floatT parentPost)
  {
    if(mComponents->isShared() && !mComponents->getSharedOpNow()) return;
    floatMatT chpost=parentPost*This::getComponentPosteriors();
    mComponents->predictPass(chpost);
  }
    
  void variationalParameterUpdateComponents()
  {
    mComponents->updateParameterPosterior();
  }
  
  void resetParameterPosteriorToPriorComponents() 
  {
    mComponents->resetParameterPosteriorToPrior();
  }
  
  void doShareWith(const This & source,bool shareBase=false) {
     if(shareBase) {
       baseT::doShareWith(source,shareBase);
       mComponents->shareWith(*source.mComponents);
     }
     this->setComponentContainer(mComponents);
  }
  
  void setSequenceLearningPhaseComponents(int phase=0)
  {
    mComponents->setSequenceLearningPhase(phase);
  }
  
};


/** Subtype of variational model with 2 types of components. Components are held in component container types. Both component containers should have the same size, this is why
there is only 1 resize() method. If their size differs, minimum size will determine which components are considered. 
*/
template<class component_container_type1,class component_container_type2>
class variationalComponentModel2 : public variationalComponentModelBase< std::pair<typename component_container_type1::componentT::pdataT, typename component_container_type2::componentT::pdataT >, typename component_container_type1::floatT >
{
public:
  // first component type
  typedef component_container_type1 componentContainerT1;
  typedef boost::shared_ptr<componentContainerT1> pComponentContainerT1;
  typedef typename componentContainerT1::componentT componentT1;
  typedef typename componentContainerT1::pComponentT pComponentT1;
  
  // second component type
  typedef component_container_type2 componentContainerT2;
  typedef boost::shared_ptr<componentContainerT2> pComponentContainerT2;
  typedef typename componentContainerT2::componentT componentT2;
  typedef typename componentContainerT2::pComponentT pComponentT2;
  
  // compound data type 
  typedef typename componentT1::dataT dataT1;
  typedef typename componentT1::pdataT pdataT1;
  typedef typename componentT2::dataT dataT2;
  typedef typename componentT2::pdataT pdataT2;
  
  typedef std::pair<pdataT1,pdataT2> dataT;
  typedef boost::shared_ptr<dataT> pdataT;
  
  typedef variationalComponentModelBase< dataT, typename component_container_type1::floatT > baseT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::floatT floatT;
  typedef typename baseT::intVecT intVecT;
  typedef variationalComponentModel2<componentContainerT1,componentContainerT2> This;
  typedef typename baseT::varmodelBaseT varmodelBaseT;
  
  
  variationalComponentModel2()
  : baseT()
  , mComponents1( new componentContainerT1() )
  , mComponents2( new componentContainerT2() )
  {
    This::mComponentPosteriors.resize(0,0);
  };
  
  
  variationalComponentModel2(const This &source):baseT(source) 
  {
    RERR( (source.mComponents1->isShared()!=source.mComponents2->isShared()) , (boost::format("Error copy-constructing variationalComponentModel2: found containers with share status %d,%d") % source.mComponents1->isShared() % source.mComponents2->isShared()).str() );
    if(source.mComponents1->isShared()) {
     mComponents1=source.mComponents1;
     mComponents2=source.mComponents2;
    } else {
      mComponents1=pComponentContainerT1(new componentContainerT1( *source.mComponents1 ));
      mComponents2=pComponentContainerT2(new componentContainerT2( *source.mComponents2 ));
    }
    doShareWith(source,false);
  }
  
  
  void unshare() {
    baseT::unshare();
    mComponents1->unshare();
    mComponents2->unshare();
  }

  
  /** Set prior pseudocount to members of pseudodata.
  \param pseudodata: vector of pseudodata for each component
  */
  void setPrior(floatT pseudodata=0) 
  {
    baseT::setPrior(pseudodata);
    mComponents1->setPrior(floatVecT(mComponents1->components(),pseudodata));
    mComponents2->setPrior(floatVecT(mComponents2->components(),pseudodata));
  }
  
  /** \return reference to 1st component container */
  pComponentContainerT1 getComponentContainer1() {return mComponents1;};
  
  /** \return reference to 1st component container */
  pComponentContainerT2 getComponentContainer2() {return mComponents2;};

  /** Set 1st component container. */
  void setComponentContainer1(pComponentContainerT1 & comptr) {
    mComponents1=comptr;
    RERR( (mComponents1->segments()<1), (boost::format("Error setting component container type 1 with sizes (%d,%d)") % mComponents1->segments() % mComponents1->components()).str() );
    This::mComponentPosteriors=ublas::zero_matrix<floatT>(mComponents1->segments(),mComponents1->components());
    this->notifyResize(mComponents1->segments(),mComponents1->components());
  };
  
    /** Set 2nd component container. */
  void setComponentContainer2(pComponentContainerT2 & comptr) {
    RERR( (mComponents2->segments()<1), (boost::format("Error setting component container type 2 with sizes (%d,%d)") % mComponents2->segments() % mComponents2->components()).str() );
    This::mComponentPosteriors=ublas::zero_matrix<floatT>(mComponents2->segments(),mComponents2->components());
    this->notifyResize(mComponents2->segments(),mComponents2->components());
  };

  
  /** Resize component containers */
  void resize(int newsize)
  {
    mComponents1->resize(newsize);
    mComponents2->resize(newsize);
    setComponentContainer1(mComponents1);
    setComponentContainer2(mComponents2);
  }
  
  
  
  /** \return reference to component, 1st type */
  pComponentT1 getComponent1(int segment,int component)
  {
    RERR( (segment<0 || segment>=mComponents1->segments() || component<0 || component>=mComponents1->components()),
	  (boost::format("Error getting component type 1: got (%d,%d), have (%d,%d)") % segment % component % mComponents1->segments() % mComponents1->components()).str() );
    return mComponents1->getComponents(segment,component);
  }
  
  /** \return reference to component, 2nd type */
  pComponentT2 getComponent2(int segment, int component)
  {
    RERR( (segment<0 || segment>=mComponents2->segments() || component<0 || component>=mComponents2->components()),
	  (boost::format("Error getting component type 1: got (%d,%d), have (%d,%d)") % segment % component % mComponents2->segments() % mComponents2->components()).str() );
    return mComponents2->getComponents(segment,component);
  }
   
  friend bool operator==(const This & first,const This & second) 
  {
   return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
    if(!(*first.mComponents1==*second.mComponents1)) return false;
    if(!(*first.mComponents2==*second.mComponents2)) return false;
    return baseT::cmp(first,second);
  }
  
  floatT getEntropy()
  {
    // Since this model is not connected directly to the data, just collect entropy of subordinates weighted by current posteriors.
    floatVecT marginalPost=This::getMarginalPosterior();
    
    return mComponents1->collectEntropy(marginalPost)
	  +mComponents2->collectEntropy(marginalPost);
  }
  
  floatT getEntropy2()
  {
    // Since this model is not connected directly to the data, just collect entropy of subordinates weighted by current posteriors.
    floatVecT marginalPost=This::getMarginalPosterior();
    
    return mComponents1->collectEntropy2(marginalPost)
	  +mComponents2->collectEntropy2(marginalPost);
  }
  
    
protected:
  pComponentContainerT1 mComponents1; //!< reference to 1st container with components. Non-shared by default.
  pComponentContainerT2 mComponents2; //!< reference to 2nd container with components. Non-shared by default.
  floatMatT mLLComp; //!< joint log-likelihood of both component types
  
    
  /** draw sample from components compo, one per segment */
  pdataT sampleComponents(const intVecT & compo)
  {
    RERR( ((int)compo.size() != mComponents1->segments()), (boost::format("Error sampling components in variationalComponentModel2: got %d for type 1, need %d") % compo.size() % mComponents1->segments()).str());
    RERR( ((int)compo.size() != mComponents2->segments()), (boost::format("Error sampling components in variationalComponentModel2: got %d for type 2, need %d") % compo.size() % mComponents2->segments()).str());
    pdataT1 psamp1=mComponents1->sample(compo);
    pdataT2 psamp2=mComponents2->sample(compo);
    pdataT psamp=pdataT( new dataT(psamp1,psamp2) );
    return psamp;
  }
  
  /** decay component posteriors */
  void decayParametersComponents(floatT tau,bool update_immediately=true)
  {
      mComponents1->decayParameters(tau,update_immediately);
      mComponents2->decayParameters(tau,update_immediately);
  }
  
  /** \return accumulated variational bound from components */
  floatT getAccumulatedVarBoundComponents() 
  {
   return mComponents1->collectVarBound()+mComponents2->collectVarBound(); 
  }
  
   /** reset components to prior */
  void resetToPriorComponents() 
  {
   mComponents1->resetToPrior();
   mComponents2->resetToPrior();
  }
  
  /** reset components var bound to 0 */
  void resetVarBoundComponents() 
  {
   mComponents1->resetVarBound();
   mComponents2->resetVarBound();
  }
  
  /** begin learning of components */
  void beginLearningComponents() 
  {
   mComponents1->beginLearning(This::mbAccumulateVarBound);
   mComponents2->beginLearning(This::mbAccumulateVarBound);
  }
 
 void freezeComponents(int level=0)
 {
   mComponents1->freeze(level); 
   mComponents2->freeze(level); 
 }
  
  /** \return accumulated KL-div from components */
  floatT getKLComponents() 
  {
   return mComponents1->collectKL()+mComponents2->collectKL(); 
  }
  
  
  void writeToStream(ostream &os) const 
  {
    os<<"Variational COMPONENT model, 2 types, base class"<<endl;
    os<<*mComponents1;
    os<<*mComponents2;
    baseT::writeToStream(os);
  };
  
  
  void readFromStream(istream &is) 
  {
    string dummy; 
    getline(is,dummy);
    if(dummy!="Variational COMPONENT model, 2 types, base class") throw std::runtime_error("Error loading variational component model, 2 types, base class: got "+dummy);
    is>>*mComponents1;
    is>>*mComponents2;
    baseT::readFromStream(is);
 
    setComponentContainer1(mComponents1);
    setComponentContainer2(mComponents2);
  }
  
  
  /** Do inference for components.
  \param data_ptr: datapoint
  \param variational: if true, compute variational expectations
  */
  void doInferenceComponents(const pdataT & data_ptr,bool variational) 
  {
    RERR( (mComponents1->isShared()!=mComponents2->isShared()), "variationalComponentModel2 must not contain shared and non-shared components!");
    if(mComponents1->isShared() && !mComponents1->getSharedOpNow()) return;
    mComponents1->inference(data_ptr->first,variational);
    mComponents2->inference(data_ptr->second,variational);
  }
  
  /** \return log-likelihood of pairs of components. Default behaviour: add log-likelihoods pairwise. */
  floatMatT getLikelihoodsComponents() const
  {
    floatMatT mLLComp=mComponents1->getComponentLikelihoods()+mComponents2->getComponentLikelihoods();
    return mLLComp;
  }
  
  
  void doLearnComponents(const pdataT & data_ptr, const floatT parentPost)
  {
    if(parentPost<numeric_limits<floatT>::epsilon()) return;
    RERR( (mComponents1->isShared()!=mComponents2->isShared()), "variationalComponentModel2 must not contain shared and non-shared components!");
    if(mComponents1->isShared() && !mComponents1->getSharedOpNow()) return;
    floatMatT chpost=parentPost*This::getComponentPosteriors();
    RERR( ((int)chpost.size1()!=mComponents1->segments() || (int)chpost.size2()!=mComponents1->components()) 
      , (boost::format("error learning components, type 1: expected (%d,%d), got (%d,%d)") % mComponents1->segments() % mComponents1->components() % chpost.size1() % chpost.size2()).str()) ;
    RERR( ((int)chpost.size1()!=mComponents2->segments() || (int)chpost.size2()!=mComponents2->components()) 
      , (boost::format("error learning components, type 2: expected (%d,%d), got (%d,%d)") % mComponents2->segments() % mComponents2->components() % chpost.size1() % chpost.size2()).str()) ;

    mComponents1->learn(data_ptr->first,chpost);
    mComponents2->learn(data_ptr->second,chpost);
  }
  
  void predictPassComponents(const floatT parentPost)
  {
    floatMatT chpost=parentPost*This::getComponentPosteriors();
    if(mComponents1->isShared() && !mComponents1->getSharedOpNow()) return;
    mComponents1->predictPass(chpost);
    mComponents2->predictPass(chpost);
  }

  
  
  void variationalParameterUpdateComponents()
  {
    mComponents1->updateParameterPosterior();
    mComponents2->updateParameterPosterior();
  }
  
 
  void posteriorToPriorComponents()
  {
    mComponents1->posteriorToPrior();
    mComponents2->posteriorToPrior();
  }
  
  void resetParameterPosteriorToPriorComponents() 
  {
    mComponents1->resetParameterPosteriorToPrior();
    mComponents2->resetParameterPosteriorToPrior();
  }
  
  void doShareWith(const This & source,bool shareBase=false) {
     if(shareBase) {
       baseT::doShareWith(source,shareBase);
       mComponents1->shareWith(*source.mComponents1);
       mComponents2->shareWith(*source.mComponents2);
     }
     this->setComponentContainer1(mComponents1);
     this->setComponentContainer2(mComponents2);
  }
  
  void setSequenceLearningPhaseComponents(int phase=0)
  {
    mComponents1->setSequenceLearningPhase(phase);
    mComponents2->setSequenceLearningPhase(phase);
  }
};


/** Selector class for variational component models. Allows you to derive from the base type depending on the number of template arguments.
Example: 


  template<class container_type1,class container_type2=int>
  class derived : public varmodSelector<container_type1,container_type2>::baseT
  {...};
  
  then, if you write
  
  derived<container_type1> derv;
  
  you get a derived object which has variationalComponentModel as base class.
  If you write
  
  derived<container_type1,container_type2> derv;
  
  you get a derived object which has variationalComponentModel2 as base class.
*/
template<class component_container_type1,class component_container_type2=int>
class varmodSelector 
{
public:
  typedef variationalComponentModel2<component_container_type1,component_container_type2> baseT;
};
 
template<class component_container_type1>
class varmodSelector<component_container_type1,int> 
{
public:
   typedef variationalComponentModel<component_container_type1> baseT;
};

