#pragma once

#include "blitz_properties.hpp"
#include "SeariseProperties.hpp"
#include "variational_model.hpp"
#include "sorting_tools.hpp"
#include "normrho.hpp"
#include "Random.hpp"
#include <list>
#include <blitz/array.h>
#include <boost/shared_ptr.hpp>
#include "gamma.hpp"

// should be useful for all kinds of things
// for example "segmentation" or blob finding
// could also be a particle filter if equipped with a transition model on the components
// new 29.09.10: store list of boost::shared_ptr to components for sharing components across dpmms

template<typename C>
struct dpmm : public variationalModel< typename C::dataT, typename C::floatT, typename C::pdataT >,
  public HasProperties< SearisePropertyTypes, dpmm<C> >
{
  typedef C component_t;
  typedef dpmm<C> this_t;
  typedef typename component_t::floatT float_type;
  typedef typename component_t::dataT data_type;
  typedef typename component_t::pdataT data_ptr;
  typedef variationalModel< data_type, float_type, data_ptr > base_t;
  typedef boost::shared_ptr<component_t> component_p;
  typedef std::list< component_p > components_t;
  typedef typename components_t::iterator components_iterator;
  typedef typename components_t::const_iterator const_components_iterator;
  typedef blitz::Array<float_type,1> vector_t;

  typedef dpmm<C> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  typedef SeariseProperty PropertyType;

  typedef Random<float_type> float_rng_t;
  typedef Random<double> double_rng_t;
  
  dpmm() 
	: base_t()
	,_T(40)
	,_alpha0(1.0)
	,_is_finite(false)
  {}

  static void init_properties()
  {
    PBase::add_ivar_property_rw( "alpha0", &This::_alpha0 );	
    PBase::add_ivar_property_rw( "T", &This::_T );
    PBase::add_ivar_property_rw( "finite", &This::_is_finite );
    PBase::add_property( "prototype", PBase::make_pivar_getter( &This::_prototype ) );
    PBase::add_property( "pi", typename PBase::MemFunGetter( &This::get_pi ) );
    PBase::add_property( "r"       ,  make_blitz_ivar_getter1( &This::_r   ) );
    PBase::add_property( "total_psum"       , make_blitz_ivar_getter1( &This::_total_psum   ) );
    
  }

  PropertyType get_pi()
  {
    vector_t pi( blitz::exp( _lpi ) );
    CPUData<float> cpi;
    blitz2cpudata( pi, cpi );
    return cpi;
  }

  // this is part of the variationalModel interface
  void setPrior(  float add_pseudodata = 0)
  {
    base_t::setPrior();
    initialize();
    
    if ( add_pseudodata > 0 )
    {
      _wk_prior += blitz::exp( _lpi ) * add_pseudodata;
      size_t i = 0;
      BOOST_FOREACH( component_p & component, components() )
      {
        component->setPrior( add_pseudodata * std::exp( _lpi(i) ) );
        ++i;
      }
	  base_t::updateParameterPosterior();
    }
  }

  data_ptr sample() 
  {
    float_rng_t & rng = float_rng_t::get();
    vector_t pi( blitz::exp( _lpi ) );
    int k = rng.sample_multinomial( pi );
    return component(k).component_t::sample();
  }

 

  components_t & components() {return _components; }
  const components_t & components() const {return _components; }

  component_t & component(size_t i)
  {
    if ( i >= size() )
      throw std::runtime_error(boost::str(boost::format("invalid component index %i (valid: [%i,%i])")%i%0%(size()-1)));
    components_iterator icomp = components().begin();
    for ( ; i > 0; --i, ++icomp );
    return **icomp;
  }
  
  // returns a hard assigment of the last datapoint by argmax( r )
  int hard_assignment() const
  {
	return maxIndex( _r )[0];
  }

  float_type & responsibility(size_t i) 
  {
    if ( i >= size() )
      throw std::runtime_error(boost::str(boost::format("invalid component index %i for responsibility (valid: [%i,%i])")%i%0%(size()-1)));
    return _r(i);
  }

  size_t size() const { return _T; }

  void initialize( size_t T = 0 )
  {
    if ( T > 0 )
      _T = T;
    _components.clear();
    for ( int t = 0; t < _T; ++t )
    {
      component_p nc( new component_t(_prototype) );
	  nc->unshare();
      _components.push_back( nc );
      _components.back()->setPrior();
    }
    _gamma1.resize( _T );
    _gamma2.resize( _T );
    _lpi.resize( _T );
    _lpidash.resize( _T );
    _wk_prior.resize( _T ); _wk_prior = 0;
    _wk_posterior.resize( _T ); _wk_posterior = 0;
    _r.resize( _T );
    _total_psum.resize( _T ); _total_psum=0.0;
    update_gamma();
    update_pi();
  }
  
  void grow() {
    base_t::posteriorToPrior();
    size_t delta = ( _T + 1 ) / 2;
    _wk_prior.resizeAndPreserve( _T + delta );
    _wk_posterior.resizeAndPreserve( _T + delta );
    for ( size_t k = _T; k < _T + delta; ++k )
    {
      component_p nc( new component_t(_prototype) );
      _components.push_back( nc );
      _components.back()->setPrior();
      _wk_prior(k) = 0;
      _wk_posterior(k) = 0;
    }
    _T += delta;
    _r.resize(_T);
    _total_psum.resize(_T,0.0); 
    this_t::updateParameterPosterior();
  }

  void adapt_size( float_type growing_threshold = 0.001, float_type shrinking_threshold = 0.0001, size_t min_size = 10 )
  {
    return;
	base_t::posteriorToPrior();
    if ( std::exp( min( _lpi ) ) >= growing_threshold ) grow();
    else
    {
      int T_new = _T;
      std::vector<int> order;
      compute_order( order.begin(), order.end(), _lpi.data() );
      while( T_new > min_size && std::exp( _lpi( order[ _T - T_new ] ) ) < shrinking_threshold )
        -- T_new;
      if ( T_new < _T ) 
      {
        vector_t wk_new( T_new );
        for ( size_t k = 0; k < T_new; ++k )
          wk_new(k) = _wk_prior( order[ _T - 1 - k ] );
        _wk_prior.reference( wk_new );
      // remove components by collecting iterators and then erasing them one by one
      // warning: this only works with iterators not invalidating under removal of 
      // elements from the container, i.e. with a list
        std::list<typename components_t::iterator> to_remove;
        for ( size_t k = 0; k < _T - T_new; ++k )
        {
        // get components at indices by walking up the list for each one
          typename components_t::iterator ic = _components.begin();
          size_t index = order[k];
          for ( size_t ik = 0; ik < index; ++ik, ++ic )
            ;
          to_remove.push_back( ic );
        }
        BOOST_FOREACH( typename components_t::iterator ic, to_remove )
          _components.erase( ic );
        this_t::updateParameterPosterior();
      }
    }
    _r.resize(_T);
    _total_psum.resize(_T,0.0); // TODO: when shrinking, erase correct components
  }

  const vector_t & latent() { return _r; }

  
  virtual void decayParameters(float_type tau, bool update_immediately = true)
  {
    _wk_prior *= std::exp( -1 / tau );
    if ( update_immediately )
    {
      update_gamma();
      update_pi();
    }
    BOOST_FOREACH( component_p & component, _components )
      component->decayParameters( tau, update_immediately );
  }
  
  int getNumUsed() 
  {
    int numused=0;
    for(int i=0;i<_total_psum.size();++i) if(_total_psum(i)>1) ++numused; // a component has been used if it has seen at least 1 full datapoint
    return numused;
  }
  
  void resetToPrior() {
    _r = 1.0*_lpi;
    normalize_rho(_r);
  }
  

  float_type getVarBound()
  {
    float_type rv=This::getAccumulatedVarBound() - This::getKL();
    BOOST_FOREACH( component_p & component, _components ) rv+=component->getVarBound();
    return rv;
  }

protected:
  
  virtual void accumulateVarBound(const float_type parentPost) 
  {
      for(unsigned k=0;k<_components.size();++k) 
	base_t::mAccumulatedVarBound += parentPost * (_r(k)*_lpidash(k) - plogp(_r(k))); 
  }
  
  virtual typename base_t::floatT doInference( const typename base_t::pdataT & x,bool variational = false )
  {
    vector_t l(_T);
    components_iterator ic = _components.begin();
    for ( int k = 0; k < _T; ++k, ++ic ) 
	  {
		(*ic)->inference( x, variational );
		l(k) = (*ic)->getLogLikelihood( );
	  }
    if ( variational )
	  l += _lpidash;
	else
	  l += _lpi;
    float_type L = max(l);
	_r = blitz::exp( l - L );
	float_type z = sum( _r );
	L += safe_log( z );
	_r *= 1 / std::max( float_type( 1e-20 ), z );
	return L;
  }
  
  void doLearn( const data_ptr & x, float_type w ) 
  {
  // apply
    _wk_posterior += _r * w;
    components_iterator ic = _components.begin();
    for ( int k = 0; k < _T; ++k, ++ic )
      (*ic)->learn( x, _r(k) * w );
  }
  
  
  void update_gamma()
  {
    _gamma1.resize( _T );
    _gamma2.resize( _T );
    // For an infinte DP, use B(1,alpha0) as prior stick breaking components
    // For a finite Dirichlet, use B(alpha0,(T-1-k)*alpha0) as prior stick breaking components for uniform prior
    // TODO: shouldn't the last component in the truncated strick breaking construction have gamma2->0.0 so that the expected prob. is 1.0 ? Not strictly necessary.
    
    if(_is_finite) 	_gamma1 = _wk_prior + _wk_posterior + _alpha0;
    else 		_gamma1 = _wk_prior + _wk_posterior + 1;
  // _gamma2 = _alpha0 + ascending_cumsum( _wk );
    float_type accumulator = 0;
    for ( int k = 0; k < _T; ++k )
    {
      size_t index = _T - 1 - k;
      if(_is_finite) 	_gamma2(index) = ((k+0.01) * _alpha0) + accumulator; // add 0.1 to avoid overflows in update_pi() at k=0
      else		_gamma2(index) = _alpha0 + accumulator;
      accumulator += _wk_prior(index) + _wk_posterior(index);

    }
  }
  
  
  float_type computeKL()
  {
    float_type rv=0.0;
    float_type prior_acc=0.0;
    for( int k = 0; k<_T; ++k ) {
      size_t index = _T - 1 - k;
      prior_acc += _wk_prior(index);
      
      float_type beta,alpha;
      if(_is_finite) {
	beta=_alpha0+_wk_prior(index);
	alpha=(k+0.01)*_alpha0 + prior_acc;
      } else {
	beta=1.0+_wk_prior(index);
	alpha=_alpha0+prior_acc;
      }
      
      float_type psisum=psi(_gamma1(index)+_gamma2(index));
      
      rv += (gammaln(_gamma1(index)+_gamma2(index))-gammaln(alpha+beta)) - (gammaln(_gamma2(index))-gammaln(alpha)) - (gammaln(_gamma1(index))-gammaln(beta)) 
	      +(_gamma1(index)-beta)*(psi(_gamma1(index))-psisum) + (_gamma2(index)-alpha)*(psi(_gamma2(index))-psisum);
    }
    return rv;
  }
  
  void update_pi()
  {
    _lpi.resize( _T );
    _lpidash.resize( _T );
    float_type ElogV2_accum = 0, notlbeta_accum = 0;
    for ( int k = 0; k < _T; ++k )
    {
      size_t index = k;
      float_type ElogV1    = psi( _gamma1(index) ) -  psi( _gamma1(index) + _gamma2(index) );
      float_type ElogV2    = psi( _gamma2(index) ) -  psi( _gamma1(index) + _gamma2(index) );
      float_type lbeta     = log( _gamma1(index) ) -  log( _gamma1(index) + _gamma2(index) );
      float_type notlbeta  = log( _gamma2(index) ) -  log( _gamma1(index) + _gamma2(index) );
      _lpidash(index) = ElogV2_accum   + ElogV1;
      _lpi    (index) = notlbeta_accum + lbeta;
      notlbeta_accum += notlbeta;
      ElogV2_accum += ElogV2;
    }
  }


  virtual void doPosteriorToPrior()
  {
    _wk_prior += _wk_posterior;
    _wk_posterior = 0;
    BOOST_FOREACH( component_p & component, _components )
      component->posteriorToPrior();
  }
  
  virtual bool doVariationalParameterUpdate()
  {
    update_gamma();
    update_pi();
    BOOST_FOREACH( component_p & component, _components )
      component->updateParameterPosterior();
  //std::cout << " lpi " << _lpi << std::endl;
  //std::cout << " lpidash " << _lpidash << std::endl;
    return true;
  }

  virtual void doBeginLearning()
  {
    _wk_posterior = 0;
    BOOST_FOREACH( component_p & component, _components )
      component->beginLearning(This::mbAccumulateVarBound);
  }



  void precomputeExpensive() {
      // do nothing here.
  }
  
  
  
  /** write to stream.
  \param os:  reference to ostream to write into.
  */
  void writeToStream(ostream &os) const 
  {
    os<<"Dirichlet process mixture model"<<endl;
    os<<"Truncation level: "<<_T<<endl;
    os<<"Prior type: "<<_is_finite<<endl;
    os<<"Concentration parameter: "<<_alpha0<<endl;
    unsigned i=0;
    os<<"Number of components: "<<components().size()<<endl;
    BOOST_FOREACH( const component_p & component, _components )
    {
      os<<"Component: "<<i<<endl;
      os<<*component;
      ++i;
    }
    os<<"Component prototype:"<<endl;
    os<<_prototype<<endl;
    os<<"P(oste)rior Beta parameters:"<<endl;
    os<<_gamma1<<_gamma2;
    os<<"P(oste)rior accumulated counts:"<<endl;
    os<<_wk_prior<<_wk_posterior;
    os<<"Log-probabilities of clusters, (non-)variational"<<endl;
    os<<_lpi<<_lpidash;
    
  }  
	
    
  void readFromStream(istream &is) 
  {
    std::string name;
    getline(is,name);
    if(name!="Dirichlet process mixture model") throw std::runtime_error("Error loading DPMM: got "+name);
    is>>eatString(2)>>_T;
    is>>eatString(2)>>name;
    _is_finite=(name=="True");
    is>>eatString(2)>>_alpha0;
    unsigned numcomp=0;
    is>>eatString(3)>>numcomp>>ws;
    for(unsigned j=0;j<numcomp;++j) {
      component_p nc(new component_t(_prototype));
      _components.push_back(nc);
      getline(is,name);
      is>>*nc;
    }
    
    is>>eatString(2);
    is>>_prototype;
    is>>eatString(3);
    is>>_gamma1>>_gamma2;
    is>>eatString(3);
    is>>_wk_prior>>_wk_posterior;
    is>>eatString(4);
    is>>_lpi>>_lpidash;
    _r.resize(_lpidash.shape());
    resetToPrior();

  }

  component_t & prototype() { return _prototype; }

private:

  int _T; // current truncation level
  components_t _components; //!< list of boost::shared_ptr to the components, to make components shareable across dpmms
  component_t _prototype; // used for parameter settings
  float_type _alpha0; // Stick-breaking proportions are drawn from Beta(1,alpha0) for inifinte DP, and Beta(alpha0,(_T-1-k)*alpha0) for finite Dirichlet
  bool _is_finite; // True if using finite dirichlet prior up to _T (and infinte stick-breaking progression beyond that)
  vector_t _gamma1, _gamma2, _wk_prior, _wk_posterior, _lpi, _lpidash;
  vector_t _r,_total_psum;

};
