
/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres and Daniel Oberhoff              */
/*    dominik.endres@gmail.com                                              */
/*    danieloberhoff@googlemail.com                                         */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/


#pragma once

#include "component_container.hpp"
#include "blitz_properties.hpp"
#include "SeariseProperties.hpp"
#include "component_container.hpp"
#include "Random.hpp"
#include "gamma.hpp"

/** Dirichlet process mixture model. Takes a component_container or component_container2 as a template parameter.
Freeze level 0: learn cluster assignment prior
Freeze level 1: cluster assignement prior fixed.
*/
template<class component_container_type1,class component_container_type2=int>
struct dpmmc : public varmodSelector<component_container_type1,component_container_type2>::baseT,
  public HasProperties< SearisePropertyTypes, dpmmc<component_container_type1,component_container_type2> >
{
public:  
  typedef typename varmodSelector<component_container_type1,component_container_type2>::baseT baseT;
  
  typedef typename baseT::floatT floatT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::intVecT intVecT;
  typedef typename baseT::dataT dataT;
  typedef typename baseT::pdataT pdataT;
    
  typedef dpmmc<component_container_type1,component_container_type2> This;
  typedef typename baseT::varmodelBaseT varmodelBaseT;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  typedef SeariseProperty PropertyType;

  typedef Random<floatT> float_rng_t;
  
  typedef boost::shared_ptr<floatT> pfloatT;
  typedef boost::shared_ptr<floatVecT> pfloatVecT;
  typedef boost::shared_ptr<floatMatT> pfloatMatT;
  typedef boost::shared_ptr<bool> pboolT;
 
  
  dpmmc():baseT()
  , _alpha0_init(1.0)
  , _is_finite_init(false)
  {
    _alpha0=pfloatT( new floatT(_alpha0_init) );
    _is_finite=pboolT( new bool(_is_finite_init) );
    _wk_prior=pfloatVecT( new floatVecT() );
    _wk_posterior=pfloatVecT( new floatVecT() );
    _gamma1=pfloatVecT( new floatVecT() );
    _gamma2=pfloatVecT( new floatVecT() );
    _lpi=pfloatVecT( new floatVecT() );
    _lpidash=pfloatVecT( new floatVecT() );
    _total_psum=pfloatVecT( new floatVecT() );
    _wk_accu_total=pfloatVecT( new floatVecT() );
  }
  
  dpmmc(const This &source):baseT(source)
  {
    _alpha0_init=source._alpha0_init;
    _is_finite_init=source._is_finite_init;
    doShareWith(source,false);
  }
  
  
  /** Make all parameters private to this instance */
  void unshare() {
    baseT::unshare();
    _alpha0=pfloatT( new floatT(*_alpha0) );
    _is_finite=pboolT( new bool(*_is_finite) );
    _wk_prior=pfloatVecT( new floatVecT(*_wk_prior) );
    _wk_posterior=pfloatVecT( new floatVecT(*_wk_posterior) );
    _gamma1=pfloatVecT( new floatVecT(*_gamma1) );
    _gamma2=pfloatVecT( new floatVecT(*_gamma2) );
    _lpi=pfloatVecT( new floatVecT(*_lpi) );
    _lpidash=pfloatVecT( new floatVecT(*_lpidash) );
    _total_psum=pfloatVecT( new floatVecT(*_total_psum) );
    _wk_accu_total=pfloatVecT( new floatVecT(*_wk_accu_total) );
  };

  static void init_properties()
  {
    PBase::add_ivar_property_rw( "alpha0", &This::_alpha0_init );	
    PBase::add_ivar_property_rw( "finite", &This::_is_finite_init );
    PBase::add_property( "pi", typename PBase::MemFunGetter( &This::get_pi ) );
    PBase::add_property( "r"       ,  typename PBase::MemFunGetter( &This::get_post   ) );
    PBase::add_property( "total_psum"       , typename PBase::MemFunGetter( &This::get_psum ) );
    
  }

  PropertyType get_pi()
  {
    floatVecT pi( exp( *_lpi ) );
    CPUData<floatT> cpi(&(*pi.begin()),pi.size(),1,1);
    return cpi;
  }
  
  PropertyType get_post()
  {
    floatVecT cp=This::getMarginalPosterior();
    CPUData<floatT> ppost(&(*cp.begin()),cp.size(),1,1);
    return ppost;
  }
  
  PropertyType get_psum()
  {
    CPUData<floatT> psum(&(*_total_psum->begin()),_total_psum->size(),1,1);
    return psum;
  }

  void setPrior(floatT add_pseudodata)
  {
    initialize();
    if(add_pseudodata<numeric_limits<floatT>::epsilon()) return;
    
    Random<float> & rng = Random<float>::get();
    
    // add random stick-pieces to prior in proportion to Beta-samples from prior
    float stickLen=1.0;
    for(unsigned comp=0;comp<_wk_prior->size();++comp) {
      float prop=rng.beta((*_gamma1)(comp),(*_gamma2)(comp));
      (*_wk_prior)(comp)+=stickLen*prop*add_pseudodata;
      stickLen*=(1.0-prop);
    }
    *_wk_posterior = ublas::zero_vector<floatT>(_wk_posterior->size());
    baseT::updateExpensiveForAll(true);
    resetToPrior();
  }


  void initialize( )
  {
    *_alpha0=_alpha0_init;
    *_is_finite=_is_finite_init;
    unsigned T=This::components();
    _gamma1->resize(T);
    _gamma2->resize(T);
    _lpi->resize(T);
    _lpidash->resize(T);
    _wk_prior->resize(T); *_wk_prior = ublas::scalar_vector<floatT>(T,0.0);
    _wk_posterior->resize(T); *_wk_posterior = ublas::scalar_vector<floatT>(T,0.0);
    _wk_accu.resize(T); _wk_accu=ublas::scalar_vector<floatT>(T,0.0);
    _wk_accu_total->resize(T); *_wk_accu_total=ublas::scalar_vector<floatT>(T,0.0);
    _total_psum->resize(T); *_total_psum= ublas::scalar_vector<floatT>(T,0.0);
    baseT::updateExpensiveForAll(true);
  }
  
  
  void grow() {
    This::posteriorToPrior();
    unsigned oldsize= baseT::components();
    unsigned delta = ( oldsize + 1 ) / 2;
    {
      baseT::guardT( *This::mpUpdateLock );
      _wk_prior->resize( oldsize + delta );
      _wk_posterior->resize( oldsize + delta );
      _wk_accu.resize( oldsize + delta );
      _wk_accu_total->resize( oldsize+delta );
      _total_psum->resize(oldsize+delta,0.0); 
      for ( unsigned k = oldsize; k < oldsize + delta; ++k )
      {
	(*_wk_prior)[k] = 0;
	(*_wk_posterior)[k] = 0;
	_wk_accu[k]=0;
	(*_wk_accu_total)[k]=0;
	(*_total_psum)[k]=0;
      }
    }  
    baseT::resize(oldsize+delta);
    This::updateParameterPosterior();
  }

  
  void decayParameters(floatT tau, bool update_immediately = true)
  {
    *_wk_prior *= std::exp( -1 / tau );
    if ( update_immediately ) baseT::updateExpensiveForAll(true);
    This::decayParametersComponents(tau,update_immediately);
  }
  
  
  
  int getNumUsed() 
  {
    int numused=0;
    for(int i=0;i<_total_psum->size();++i) if(*_total_psum(i)>1) ++numused; // a component has been used if it has seen at least 1 full datapoint
    return numused;
  }
  
  void resetToPrior() {
   baseT::resetToPrior();
   floatVecT cp=floatVecT(*_lpi);
   normalize(cp);
   for(int seg=0;seg<This::subSegments();++seg) ublas::matrix_row<floatMatT>(This::mComponentPosteriors,seg)=cp;
   This::resetToPriorComponents();
  }
  
  
  floatT getKL()
  {
    floatT rv=0.0;
    floatT prior_acc=0.0;
    
    floatVecT &g1=*_gamma1;
    floatVecT &g2=*_gamma2;
      
    for( int k = 0; k<This::components(); ++k ) {
      int index = This::components() - 1 - k;
      
      floatT beta,alpha;
      if(*_is_finite) {
	beta=*_alpha0+(*_wk_prior)(index);
	alpha=(k+0.01)*(*_alpha0) + prior_acc;
      } else {
	beta=1.0+(*_wk_prior)(index);
	alpha=(*_alpha0)+prior_acc;
      }
      
      prior_acc += (*_wk_prior)(index);
      
      floatT psisum=psi(g1(index)+g2(index));
      
      rv += (gammaln(g1(index)+g2(index))-gammaln(alpha+beta)) - (gammaln(g2(index))-gammaln(alpha)) - (gammaln(g1(index))-gammaln(beta)) 
	      +(g1(index)-beta)*(psi(g1(index))-psisum) + (g2(index)-alpha)*(psi(g2(index))-psisum);
    }
    return rv+This::getKLComponents();
  }
  
  void posteriorToPrior()
  {
    {
      typename baseT::guardT g( *This::mpUpdateLock );
      *_wk_prior += *_wk_posterior;
      *_total_psum+= *_wk_posterior;
      *_wk_posterior = ublas::zero_vector<floatT>(_wk_posterior->size());
    }
    baseT::posteriorToPrior();
  }
  
  
  void notifyResize(int segments,int components)
  {
    baseT::notifyResize(segments,components);
   _wk_posterior->resize(components,true);
   _wk_prior->resize(components,true);
   _wk_accu.resize(components,true);
   _wk_accu_total->resize(components,true);
   _gamma1->resize(components,true);
   _gamma2->resize(components,true);
   _lpi->resize(components,true);
   _lpidash->resize(components,true);
   _total_psum->resize(components,true);
   baseT::updateExpensiveForAll(true);
  }
  
  
  /** comparison operator. compares contents, not pointers */
  friend bool operator==(const This & first,const This & second) 
  {
   return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
    if(!(*first._alpha0==*second._alpha0)) return false;
    if(!(*first._is_finite==*second._is_finite)) return false;
    if(!(*first._wk_prior==*second._wk_prior)) return false;
    if(!(*first._wk_posterior==*second._wk_posterior)) return false;
    if(!(*first._gamma1==*second._gamma1)) return false;
    if(!(*first._gamma2==*second._gamma2)) return false;
    if(!(*first._lpi==*second._lpi)) return false;
    if(!(*first._lpidash==*second._lpidash)) return false;
    if(!(*first._total_psum==*second._total_psum)) return false;
    return baseT::cmp(first,second);
  }


protected:
  
  void doShareWith(const This & source,bool shareBase=false)
  {
    if(shareBase) baseT::doShareWith(source,shareBase);
    _alpha0=source._alpha0;
    _is_finite=source._is_finite;
    _wk_prior=source._wk_prior;
    _wk_posterior=source._wk_posterior;
    _gamma1=source._gamma1;
    _gamma2=source._gamma2;
    _lpi=source._lpi;
    _lpidash=source._lpidash;
    _total_psum=source._total_psum;
    _wk_accu=ublas::zero_vector<floatT>(_wk_posterior->size());
    _wk_accu_total=source._wk_accu_total;
    resetToPrior();
  }
  

  
  floatT doInference( const pdataT & data_ptr, bool variational )
  {
    doInferenceComponents(data_ptr,variational); 
    floatT cumll=0.0;
    This::mComponentPosteriors=This::getLikelihoodsComponents();
    const floatVecT & pri=(variational?*_lpidash:*_lpi);
    for(int segment=0;segment<This::subSegments();++segment) 
    {
      ublas::matrix_row<floatMatT> cseg(This::mComponentPosteriors,segment);
      cseg+=pri;
      cumll+=normalize(cseg);
    }
    return cumll;
  }
    
  void accumulateVarBound(floatT parentPost=1.0) 
  {
    for(int seg=0;seg<This::subSegments();++seg)
    {
      const floatVecT &po=ublas::matrix_row<floatMatT>(This::mComponentPosteriors,seg);
      This::mAccumulatedVarBound+=parentPost*(ublas::inner_prod(po,*_lpidash)-plogp(po));
    }
  }
  
  void doLearn( const pdataT & data_ptr, floatT parentPost ) 
  {
    _wk_accu+=This::getMarginalPosterior() * parentPost;
    //*_wk_posterior += This::getMarginalPosterior() * parentPost;
    doLearnComponents(data_ptr,parentPost);
  }
  
  
  void update_gamma()
  {
    _gamma1->resize( This::components() );
    _gamma2->resize( This::components() );
    // For an infinte DP, use B(1,alpha0) as prior stick breaking components
    // For a finite Dirichlet, use B(alpha0,(T-1-k)*alpha0) as prior stick breaking components for uniform prior
    // TODO: shouldn't the last component in the truncated strick breaking construction have gamma2->0.0 so that the expected prob. is 1.0 ? Not strictly necessary.
    
    if(*_is_finite) 	*_gamma1 = *_wk_prior + *_wk_posterior + ublas::scalar_vector<floatT>(_gamma1->size(),*_alpha0);
    else 		*_gamma1 = *_wk_prior + *_wk_posterior + ublas::scalar_vector<floatT>(_gamma1->size(),1.0);
  // _gamma2 = _alpha0 + ascending_cumsum( _wk );
    floatT accumulator = 0;
    for ( int k = 0; k < This::components(); ++k )
    {
      int index = This::components() - 1 - k;
      if(*_is_finite) 	(*_gamma2)(index) = ((k+0.01) * (*_alpha0) ) + accumulator; // add 0.01 to avoid overflows in update_pi() at k=0
      else		(*_gamma2)(index) = *_alpha0 + accumulator;
      accumulator += (*_wk_prior)(index) + (*_wk_posterior)(index);

    }
  }
  
  
  void update_pi()
  {
    _lpi->resize( This::components() );
    _lpidash->resize( This::components() );
    floatT ElogV2_accum = 0, notlbeta_accum = 0;
    for ( int k = 0; k < This::components(); ++k )
    {
      int index = k;
      floatT ElogV1    = psi( (*_gamma1)(index) ) -  psi( (*_gamma1)(index) + (*_gamma2)(index) );
      floatT ElogV2    = psi( (*_gamma2)(index) ) -  psi( (*_gamma1)(index) + (*_gamma2)(index) );
      floatT lbeta     = log( (*_gamma1)(index) ) -  log( (*_gamma1)(index) + (*_gamma2)(index) );
      floatT notlbeta  = log( (*_gamma2)(index) ) -  log( (*_gamma1)(index) + (*_gamma2)(index) );
      (*_lpidash)(index) = ElogV2_accum   + ElogV1;
      (*_lpi)    (index) = notlbeta_accum + lbeta;
      notlbeta_accum += notlbeta;
      ElogV2_accum += ElogV2;
    }
  }

  
  bool doVariationalParameterUpdate()
  {
    if(This::mFreezeLevel<1) {
      *_wk_accu_total+=_wk_accu;
      *_wk_posterior=*_wk_accu_total;
    }
    This::variationalParameterUpdateComponents();
    return true;
  }
  
  /** Reset variational parameter posterior to prior */
  void doResetParameterPosteriorToPrior() 
  {
   *_wk_posterior=ublas::zero_vector<floatT>(_wk_posterior->size()); 
  }

  void doBeginLearning()
  {
    _wk_accu=ublas::zero_vector<floatT>(_wk_posterior->size());
    *_wk_accu_total=ublas::zero_vector<floatT>(_wk_posterior->size());
    This::beginLearningComponents();
  }


  void precomputeExpensive() {
    update_gamma();
    update_pi();
  }
  
  void copyExpensive(const varmodelBaseT & source)
  {
    // nothing to copy -- everything is shared in current impl.
  }
  
  
  
  /** write to stream.
  \param os:  reference to ostream to write into.
  */
  void writeToStream(ostream &os) const 
  {
    os<<"Dirichlet process mixture model"<<endl;
    os<<"Prior type: "<<(*_is_finite?"finite":"infinite")<<endl;
    os<<"Concentration parameter: "<<*_alpha0<<endl;
    os<<"Posterior Beta parameters:"<<endl;
    os<<*_gamma1<<*_gamma2;
    os<<"P(oste)rior accumulated counts:"<<endl;
    os<<*_wk_prior<<*_wk_posterior;
    os<<"Log-probabilities of clusters, (non-)variational"<<endl;
    os<<*_lpi<<*_lpidash;
    os<<"Probabilities of clusters, (non-)variational"<<endl;
    floatVecT temp=*_lpi;
    normalize(temp);
    os<<temp;
    temp=*_lpidash;
    normalize(temp);
    os<<temp;
    baseT::writeToStream(os);
  }  
	
    
  void readFromStream(istream &is) 
  {
    std::string name;
    getline(is,name);
    if(name!="Dirichlet process mixture model") throw std::runtime_error("Error loading DPMM: got "+name);
    is>>eatString(2)>>name;
    *_is_finite=(name=="finite");
    is>>eatString(2)>>*_alpha0;
    is>>eatString(3);
    is>>*_gamma1>>*_gamma2;
    is>>eatString(3);
    is>>*_wk_prior>>*_wk_posterior;
    is>>eatString(4);
    is>>*_lpi>>*_lpidash;
    is>>eatString(4);
    floatVecT temp;
    is>>temp;
    is>>temp;
    baseT::readFromStream(is);
    {
     typename baseT::guardT g( *This::mpUpdateLock );
     _wk_accu=ublas::zero_vector<floatT>(_wk_posterior->size());
     *_wk_accu_total=ublas::zero_vector<floatT>(_wk_posterior->size());
    }
    resetToPrior();
    This::updateExpensiveForAll(true);

  }
  



private:
  floatT _alpha0_init;
  bool _is_finite_init;
  pfloatT _alpha0; //!< Stick-breaking proportions are drawn from Beta(1,alpha0) for inifinte DP, and Beta(alpha0,(_T-1-k)*alpha0) for finite Dirichlet
  pboolT _is_finite; //!< True if using finite dirichlet prior up to truncation level (and infinte stick-breaking progression beyond that)
  floatVecT _wk_accu; //!< accumulated counts since last call to beginLearning() for this instance
  pfloatVecT _wk_accu_total; //!< accumulated counts since last (collective) call to beginLearning() across all instances
  pfloatVecT _wk_prior, _wk_posterior; //!< accumulated counts, prior and posterior.
  pfloatVecT _gamma1, _gamma2; //!< parameters of stick-breaking Betas
  pfloatVecT _lpi, _lpidash; //!< log-probabilities of cluster assignement
  pfloatVecT _total_psum; //!< total responsibility assigned to a cluster. useful for usage checking.
};
