
/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres and Daniel Oberhoff              */
/*    dominik.endres@gmail.com                                              */
/*    danieloberhoff@googlemail.com                                         */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/


#pragma once

#include "dpmmc.hpp"
#include "blitz_properties.hpp"
#include "SeariseProperties.hpp"

/** Exchangeable sequence, based on Dirichlet process mixture model. Takes a component_container or component_container2 as a template parameter.
Can accumulate likelihood during inference, thereby connecting points in time.
*/
template<class component_container_type1,class component_container_type2=int>
struct exSequence : public dpmmc<component_container_type1,component_container_type2>,
  public HasProperties< SearisePropertyTypes, exSequence<component_container_type1,component_container_type2> >
{
  
  typedef dpmmc<component_container_type1,component_container_type2> baseT;
  
  typedef typename baseT::floatT floatT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  typedef typename baseT::intVecT intVecT;
  typedef typename baseT::dataT dataT;
  typedef typename baseT::pdataT pdataT;
    
  typedef exSequence<component_container_type1,component_container_type2> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  typedef SeariseProperty PropertyType;

  typedef Random<floatT> float_rng_t;
  
  typedef boost::shared_ptr<floatT> pfloatT;
  typedef boost::shared_ptr<floatVecT> pfloatVecT;
  typedef boost::shared_ptr<floatMatT> pfloatMatT;
  typedef boost::shared_ptr<bool> pboolT;


public:
  
  exSequence():baseT(),mBufLikelihoods()
  {
  };
  
  exSequence(const This &source):baseT(source)
  {
    mBufLikelihoods.clear();
    doShareWith(source,false);
  }
  
  void unshare()
  {
    baseT::unshare();
    mBufLikelihoods.clear();
  }
  
  static void init_properties()
  {
    //TODO: property for accumulation getter/setter
  }
  
  
  
  void resetToPrior() {
   baseT::resetToPrior();
   if(This::mSequenceLearningPhase==This::FORWARD) mBufLikelihoods.push_back(0.0);
   if(This::mSequenceLearningPhase==This::BACKWARD) mBufLikelihoods.pop_back();
  }
  
  
protected:
  
  std::vector<floatT> mBufLikelihoods; //!< likelihoods of sequences buffered during forward pass through sequences.
  
  void doBeginLearning() 
  {
    baseT::doBeginLearning();
    mBufLikelihoods.clear();
    mBufLikelihoods.push_back(0.0);
  }
  

  
    
  floatT doInference( const pdataT & data_ptr, bool variational )
  {
    floatT cll=baseT::doInference(data_ptr,variational);
    
    if(This::mSequenceLearningPhase==This::FORWARD) {
      mBufLikelihoods.back()=This::mLogLikelihood+cll;
      return mBufLikelihoods.back();
    }
    if(This::mSequenceLearningPhase==This::BACKWARD) {
      return mBufLikelihoods.back();
    }
    
    return This::mLogLikelihood+cll;
    
  }
  
    /** write to stream.
  \param os:  reference to ostream to write into.
  */
  void writeToStream(ostream &os) const 
  {
    os<<"Exchangeable sequence model"<<endl;
    baseT::writeToStream(os);
  }
  
  void readFromStream(istream &is) 
  {
    std::string dummy;
    getline(is,dummy);
    if(dummy!="Exchangeable sequence model") throw std::runtime_error("Error loading exchangeable sequence model: got "+dummy);
    mBufLikelihoods.clear();
    baseT::readFromStream(is);
  }
    
};
  