#pragma once

#include "variational_model.hpp"
#include "gsl_blitz.hpp"
#include <blitz/array.h>
#include "gamma.hpp"
#include "blitz_properties.hpp"
#include "Random.hpp"

//TODO: implement computeKL()

template<typename T>
blitz::Array<T,1> sample_multivariate_normal_from_root
  ( const blitz::Array<T,1> & mu, const blitz::Array<T,2> & A, Random<T> & rng )
{
  // draw D samples from N(0,1)
  int D = mu.extent(0);
  blitz::Array<T,1> x(mu.copy());
  for ( int d = 0; d < D; ++d )
  {
    x += A(d,blitz::Range::all()) * rng.normal();
  }
  return x;
}

template<typename T>
blitz::Array<T,1> sample_multivariate_normal_from_covariance
  ( const blitz::Array<T,1> & mu, const blitz::Array<T,2> & Sigma, Random<T> & rng )
{
  blitz::Array<T,2> A( Sigma.shape() );
  cholesky_decomp( Sigma, A );
  return sample_multivariate_normal_from_root( mu, A, rng );
}

template <class floatT = float >
class full_covariance_gaussian_model : public variationalModel< blitz::Array<floatT,1>, floatT, blitz::Array<floatT,1> >,
  public HasProperties< SearisePropertyTypes, full_covariance_gaussian_model<floatT> >

{

public:

  typedef floatT float_type;
  typedef blitz::Array<float_type,1> data_type;
  typedef blitz::Array<float_type,2> data2T;
  typedef blitz::Array<double,1> stats1T;
  typedef blitz::Array<double,2> stats2T;

  typedef blitz::Array<float_type,1> vector_t;

  typedef variationalModel< data_type, float_type, data_type > base_t;
  typedef typename base_t::pdataT data_ptr;

  typedef full_covariance_gaussian_model<floatT> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;

  typedef Random<float_type> float_rng_t;
  typedef Random<double> double_rng_t;

  static void init_properties()
  {
    PBase::add_ivar_property_rw( "beta0", &This::_beta0 );	
    PBase::add_ivar_property_rw( "S0", &This::_S0 );	
    PBase::add_ivar_property_rw( "m0", &This::_m0 );	
    PBase::add_ivar_property_rw( "n0", &This::_n0 );	
    PBase::add_ivar_property_rw( "D", &This::_D );	
    PBase::add_property( "mu"        , make_blitz_ivar_getter1( &This::_m     ) );
    PBase::add_property( "tau"       , make_blitz_ivar_getter1( &This::_Tau   ) );
  }

  void setPrior(  float add_pseudodata = 0 )
  {
    initialize();
    if ( add_pseudodata > 0 )
    {
      double_rng_t & rng = double_rng_t::get();
      // for the variance draw n0 samples from a N(0,S0) and use their scatter matrix over n0 scaled
      //  - the prior is diagonal, so we just draw from D univariates
      stats2T Tau(_D,_D);
      stats2T Sigma(_D,_D);
      stats2T Sigma_mu(_D,_D);
      stats1T mu(_D);
      data_type x(_D);

      Tau = 0;
      mu = _m0;

      int n_int = int( _n0 + 0.5f );
      for ( int n = 0; n < n_int; ++n )
      {
        for ( int d = 0; d < _D; ++d )
          x(d) = rng.normal( 0, 1 / ( _S0 * _n0 ) ); // S0 = 1 / Tau0 = 1 / ( n0 * W0 ) ==> W0 = 1 / ( n0 * S0 )
        Tau += x(i1) * x(i2);
      }
      Tau *= _n0 / n_int;
      invert_symmetric( Tau, Sigma );

      // for the mean draw from N(m,beta*Tau) (lambda is the variance sampled above) and scale that
      Sigma_mu = Sigma / _beta0;
      mu = sample_multivariate_normal_from_covariance( mu, Sigma_mu, rng );
      
      // synthesize statistics from that
      _w_prior = add_pseudodata;
      _wx_prior = mu * add_pseudodata;
      _wx2_prior += ( Sigma + mu(i1) * mu(i2) ) * add_pseudodata;
      base_t::updateParameterPosterior();
    }
  }

  data_ptr sample() 
  {
    float_rng_t & rng = float_rng_t::get();
    return sample_multivariate_normal_from_root( _m, _A, rng );
  }

  void initialize()
  {
    _w_prior = _w_posterior = 0;
    _wx_prior.resize( _D ); _wx_prior = 0;
    _wx_posterior.resize( _D ); _wx_posterior = 0;
    _wx2_posterior.resize( _D, _D );_wx2_posterior = 0;
    _wx2_prior.resize( _D, _D ); _wx2_prior = 0;
    base_t::updateParameterPosterior();
  }

  
  
  void decayParameters(float_type tau, bool update_immediately = true)
  {
    _w_prior   *=  std::exp( -1 / tau );
    _wx_prior  *=  std::exp( -1 / tau );
    _wx2_prior *=  std::exp( -1 / tau );
    if ( update_immediately )
      base_t::updateParameterPosterior();
  }
  
  
protected:
  
  void precomputeExpensive() {
      // do nothing here.
  }
  
  float_type computeLogLikelihood(const data_type & x, bool variational) 
  {
    // non-variational needs the multivariate students-t
    return ( variational ? _lCdash : _lC ) - float_type(0.5) * sum( ( x( i1 ) - _m( i1 ) ) * _Tau * ( x( i2 ) - _m( i2 ) ) );
  };
  
  
  float_type computeKL() 
  {
        return FLT_MAX/10.0; // not implemented yet
  };
  
  void doBeginLearning() 
  {
    _w_posterior = 0;
    _wx_posterior = 0;
    _wx2_posterior = 0;
  }
  
  virtual typename base_t::floatT doInference( const typename base_t::pdataT & data_ptr,bool variational = false ) 
  {
  }
  
  
  void doLearn( const data_type & x, float_type w = float_type(1) ) 
  {
    _w_posterior += w;
    _wx_posterior += w * x;
    _wx2_posterior += w * ( x(i1) * x(i2) );
  
  }
  
  bool doVariationalParameterUpdate()
  {
  // 10.60 - 10.63 in prml
    float_type lw;
    data_type xm(_D); // mean
    data2T S(_D,_D);  // scatter matrix (sample covariance)
    float_type N =  _w_prior + _w_posterior;

    bool didUpdate = false;
    
    if ( N > 0 )
    {
      xm = ( _wx_prior + _wx_posterior ) / N;
      S = ( _wx2_prior + _wx2_posterior ) / N - xm( i1 ) * xm( i2 );
      didUpdate = true;
    }
    else
    {
      xm = 0;
      S = 0;
      N = 0;
    }

    _m.resize( _wx_prior.shape() );
    _W.resize( _D, _D );
    _A.resize( _D, _D );
    _Tau.resize( _D, _D );

    // wishart update
    _beta = _beta0 + N;
    _m = ( 1 / _beta ) * ( _beta0 * _m0 + N * xm );
    _W = _n0 * where( i1 == i2, _S0, 0 ) + N * S + _beta0 * N / ( _beta0 + N ) * ( xm( i1 ) - _m0 ) * ( xm( i2 ) - _m0 );
    _n = _n0 + N;
    cholesky_decomp( _W, _W );
    _A = _W / std::sqrt( _n ); // save the cholesky of the covariance for sampling
    cholesky_invert( _W, _W, lw );

    _Tau = _W * _n;

    assert( _n > ( _D - 1 ) );
    // original: _lCdash = float_type(0.5) * ( _D * std::log( float_type(2.0) ) + lw - _D / _beta );
    _lCdash = float_type(0.5) * (- _D * std::log( float_type(M_PI) ) + lw - _D / _beta );
    for ( int i = 0; i < _D; ++i )
    {
      _lCdash += float_type(0.5) * psi( ( _n - i ) * float_type(0.5) );
    }
    _lC = float_type(0.5) * ( lw + _D * std::log( _n ) );

    return didUpdate;
  }

  
  void doPosteriorToPrior() 
  {
      base_t::updateParameterPosterior();
      _w_prior += _w_posterior; _w_posterior = 0;
      _wx_prior += _wx_posterior; _wx_posterior = 0;
      _wx2_prior += _wx2_posterior; _wx2_posterior = 0;
  }
    
    
  
  /** write to stream. Implement in derived classes.
          \param os:  reference to ostream to write into.
  */
  virtual void writeToStream(ostream &os) const 
  {
        os<<"Multivariate Gaussian observation model, Daniel's implementation. Serialization not implemented yet"<<endl;
  };
  
  /** read from stream is. Implement in derived classes.
          \param is : reference to stream to read from.
  */
  virtual void readFromStream(istream &is) 
  {
        string dummy; 
        is>>dummy; // read line and discard
  }
  
  /** check if all points in mData have the correct shape. Implement in derived classes. */
  virtual bool checkDataShape(const data_type & data) {return true;};
  
  
  

private:

  blitz::firstIndex i1;
  blitz::secondIndex i2;

  int _D;

  double _w_prior;
  stats1T _wx_prior;
  stats2T _wx2_prior;

  double _w_posterior;
  stats1T _wx_posterior;
  stats2T _wx2_posterior;

  float_type _beta0;
  float_type _m0;
  float_type _S0;
  float_type _n0;

  float_type _beta;
  data_type _m;
  data2T _W, _Tau, _A;
  float_type _n;
  float_type _lC, _lCdash;

};
