/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/

#pragma once

#include "SeariseProperties.hpp"
#include "CPUData.hpp"
#include "variational_model.hpp"
#include "ublas_inversion.hpp"
#include "ublas_storage_adaptors.hpp"
#include <algorithm>
#include <iostream>
#include "float.h"
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/shared_ptr.hpp>
#include <limits>
#include "time.h"

using namespace std;
using namespace boost::numeric::ublas;


/** Type-casting templates. Generic implementation assumes that data_t  can be cast onto float_vec_t and vice versa */
template<class data_t,class data_ptr_t,class float_vec_t>
struct vectorCaster { 
 typedef data_t dataT;
 typedef data_ptr_t pdataT;
 typedef float_vec_t floatVecT;
 
 static const floatVecT dataToFloatVec(const pdataT & data_ptr) {return *data_ptr;};
 static pdataT floatVecToData(const floatVecT & data) 
 {
   pdataT rv=pdataT( new dataT(data) );
   return rv;
 }
};

template<>
struct vectorCaster<CPUData<float>,boost::shared_ptr<CPUData<float> >,boost::numeric::ublas::vector<float> > {
 typedef CPUData<float> dataT;
 typedef boost::shared_ptr<CPUData<float> > pdataT;
 typedef boost::numeric::ublas::vector<float> floatVecT;
  
 static const floatVecT dataToFloatVec(const pdataT & data_ptr)
 {
   return make_vector_from_pointer<float>(data_ptr->size(),data_ptr->data());
 }
 
 static pdataT floatVecToData(const floatVecT & data)
 {
    pdataT rv=pdataT( new dataT(data.size(),1,1) );
    rv->make_contiguous();
    std::copy(data.begin(),data.end(),rv->data());
    return rv;
 }
 
};


template<>
struct vectorCaster<CPUData<double>,boost::shared_ptr<CPUData<double> >,boost::numeric::ublas::vector<double> > {
 typedef CPUData<double> dataT;
 typedef boost::shared_ptr<CPUData<double> > pdataT;
 typedef boost::numeric::ublas::vector<double> floatVecT;
  
 static const floatVecT dataToFloatVec(const pdataT & data_ptr)
 {
   return make_vector_from_pointer<double>(data_ptr->size(),data_ptr->data());
 }
 
 static pdataT floatVecToData(const floatVecT & data)
 {
    pdataT rv=pdataT( new dataT(data.size(),1,1) );
    std::copy(data.begin(),data.end(),rv->data());
    return rv;
 }
 
};


/** Gaussian observation model with Gauss-Wishart prior on mean and covariance. Members with suffix "Var" are variational proposal parameters,
members with suffix "Acc" are accumulated sufficient statistics. 
Properties:

\param D: dimensionality of input vectors
\param m0: prior mean (of all components) of input vector
\param beta0: concentration parameter of Gaussian prior on mean
\param n0: concentration parameter of Wishart prior on precision matrix
\param S0: initial variance on diagonal of covariance matrix. all off-diagonal entries will be 0.0
\param N: accumulated parent posterior since last call to beginLearning()
\param mu: read-only current mean.

Freeze level 0: learn mean and covariance.
Freeze level 1: learn mean, covariance fixed.
Freeze level 2: all parameters fixed.

*/

template<class data_t=ublas::vector<float> , class float_t=float >
class gaussWishart : public variationalModel< data_t, float_t >,
  public HasProperties< SearisePropertyTypes, gaussWishart< data_t, float_t > >
{
public:
  
  typedef variationalModel<data_t,float_t> baseT;
  typedef data_t dataT;
  typedef boost::shared_ptr<dataT> pdataT;
  typedef typename baseT::floatT floatT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::floatMatT floatMatT;
  

  typedef gaussWishart<data_t,float_t> This;
  typedef typename baseT::varmodelBaseT varmodelBaseT;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  
  typedef boost::shared_ptr<floatT> pfloatT;
  typedef boost::shared_ptr<floatVecT> pfloatVecT;
  typedef boost::shared_ptr<floatMatT> pfloatMatT;
  


  gaussWishart()
    : mLowerCVCurrent(false)
    , _dim(0)
    , _mu(0)
    , _beta(0.01)
    , _nu(1.0)
    , _ivar(1.0)
  {
    // init sharable params
    mNu=pfloatT( new floatT(0.0));
    mNuVar=pfloatT( new floatT(0.0));
    mBeta=pfloatT( new floatT(0.01));
    mBetaVar=pfloatT( new floatT(0.01));
    mMu=pfloatVecT( new floatVecT() );
    mMuVar=pfloatVecT( new floatVecT() );
    mInvScale=pfloatMatT( new floatMatT() );
    mInvScaleVar=pfloatMatT( new floatMatT() );
    mParentPostAccTotal=pfloatT(new floatT(0.0));
    mDataAccTotal=pfloatVecT(new floatVecT());
    mData2AccTotal=pfloatMatT(new floatMatT());
  }
  
  /** create parameter-sharing copy of source. If you want private copies of the params, call unshare() afterwards. */
  gaussWishart(const This &source) : baseT(source)
  {
    _dim=source._dim;
    _mu=source._mu;
    _beta=source._beta;
    _nu=source._nu;
    _ivar=source._ivar;
    doShareWith(source,false);
  }
  

  
  /** Make all parameters private to this instance */
  void unshare() 
  {
	if ( baseT::shared() )
	  {
		baseT::unshare();
		mNu=pfloatT( new floatT(*mNu) );
		mNuVar=pfloatT( new floatT(*mNuVar) );
		mBeta=pfloatT( new floatT(*mBeta) );
		mBetaVar=pfloatT( new floatT(*mBetaVar) );
		mMu=pfloatVecT( new floatVecT(*mMu) );
		mMuVar=pfloatVecT( new floatVecT(*mMuVar) );
		mInvScale=pfloatMatT( new floatMatT(*mInvScale) );
		mInvScaleVar=pfloatMatT( new floatMatT(*mInvScaleVar) );
		mParentPostAccTotal=pfloatT(new floatT(*mParentPostAccTotal));
		mDataAccTotal=pfloatVecT(new floatVecT(*mDataAccTotal));
		mData2AccTotal=pfloatMatT(new floatMatT(*mData2AccTotal));    
	  }
  };

  static void init_properties()
  {
    PBase::add_ivar_property_rw( "beta0", &This::_beta );
    PBase::add_ivar_property_rw( "S0", &This::_ivar );
    PBase::add_ivar_property_rw( "m0", &This::_mu );
    PBase::add_ivar_property_rw( "n0", &This::_nu );
    PBase::add_ivar_property_rw( "D", &This::_dim );
    PBase::add_property( "mu", typename PBase::MemFunGetter( &This::get_mu ) );
    PBase::add_ivar_property_rw( "N", &This::mParentPostAcc );
  }

  typename PBase::PropertyType get_mu()
  {
    CPUData<floatT> mu( _dim, 1, 1 );
    std::copy( mMuVar->begin(), mMuVar->end(), mu.data() );
    return mu;
  } 

  
  /** Set prior. Call after setting properties. Call only for 1 instance of a parameter-sharing group to avoid overwrites
  \param add_pseudodata: number of sampled datapoints for initialisation. Average will be rescaled to match property n0 and beta0
  */
  void setPrior( floatT add_pseudodata = 0, bool randomize_pseudodata = false ) 
  {
    baseT::setPrior();
    *mMu 	= _mu*floatVecT(_dim,1.0);
    *mMuVar	= *mMu;
    mDataAcc = ublas::zero_vector<floatT>(_dim);
    *mDataAccTotal=ublas::zero_vector<floatT>(_dim) ;
    *mBeta = *mBetaVar = max(_beta,0.01f);
    
    mParentPostAcc = 0.0;
    *mParentPostAccTotal=0.0;
    _nu = max( _nu, _dim + 1.1f );
    *mNu = *mNuVar = _nu;
    *mInvScale = _ivar * ( _nu - _dim - 1.0 ) *  ublas::identity_matrix<floatT>(_dim,_dim); // 
    mScaleVar = *mInvScaleVar = *mInvScale;
    mData2Acc = ublas::zero_matrix<floatT>(_dim,_dim);
    *mData2AccTotal = ublas::zero_matrix<floatT>(_dim,_dim);
    
	// broken?
    if ( add_pseudodata > 0.0 ) 
	  {
		// dream up some pseudo-data
		This::beginLearning();
		for(unsigned i=0;i<add_pseudodata;++i) This::learn(sample());
		This::updateParameterPosterior();
		posteriorToPrior();
		
		// rescale
		*mBeta = *mBetaVar = _beta;
		(*mInvScale)*=( _nu - _dim - 1.0 ) / (*mNu - _dim - 1.0);
		*mInvScaleVar = *mInvScale;
		*mNu = *mNuVar = _nu;
	  }    
    
    This::updateExpensiveForAll(true);
    posteriorToPrior();
    doBeginLearning(); // erase accumulators
  };
  
  
  void setMean(const floatVecT & mean)
  {
   RERR((mean.size() != mMu->size()), (boost::format("Error setting mean in Gauss-Wishart: got %d components, need %d") % mean.size() % mMu->size()).str());
   *mMu=mean;
   *mMuVar=mean;
  }

  
  
      
  /** \return KL-divergence between variational distributions.Useful for convergence checking. */
  friend floatT getKLto(const This & first,const This & second) 
  {
    return gaussWishart<floatT>::KLhelper(*first.mBetaVar,*second.mBetaVar,*first.mMuVar,*second.mMuVar
					      ,*first.mNuVar,*second.mNuVar,*first.mInvScaleVar,*second.mInvScaleVar);
  }
  
  /** comparison operator */
  friend bool operator==(const This & first,const This & second) 
  {
    return cmp(first,second);
  };
  
  static bool cmp(const This & first, const This &second)
  {
    bool rv=(*first.mNu==*second.mNu) && (*first.mNuVar==*second.mNuVar) ;
    rv &= (*first.mBeta==*second.mBeta) && (*first.mBetaVar==*second.mBetaVar);
    rv &= (first.mParentPostAcc==second.mParentPostAcc);
    rv &= equalp(*first.mMu,*second.mMu) && equalp(*first.mMuVar,*second.mMuVar);
    rv &= equalp(*first.mInvScale,*second.mInvScale) && equalp(*first.mInvScaleVar,*second.mInvScaleVar);
    rv &= equalp(first.mLogLikeConst,second.mLogLikeConst);
    rv &= (first.mDataAcc==second.mDataAcc) && (first.mData2Acc==second.mData2Acc);
    // compare scale matrices via effect on inverse scale matrices
//     ublas::matrix<float_t> imat=ublas::identity_matrix<float_t>(first.mMu->size(),first.mMu->size());
//     ublas::matrix<float_t> pmat=prod(first.mScaleVar,*first.mInvScaleVar);
//     rv &= equalp(pmat,imat,1e-4f);
//     ublas::matrix<float_t> pmat2=prod(second.mScaleVar,*second.mInvScaleVar);
//     rv &= equalp(pmat2,imat,1e-4f);

    rv &= baseT::cmp(first,second);
    return rv;
  }
  
  
  
  /** draw a sample */
  pdataT sample() 
  {
    if(mMu->size()<1) return pdataT( new dataT() );
    if(!mLowerCVCurrent) {
      mLowerCV=*mInvScaleVar/(*mNuVar-mMu->size()-1.0);
      choleskyFactorize(mLowerCV,true);
      mLowerCVCurrent=true;
    }
   
    
    return vectorCaster<dataT,pdataT,floatVecT>::floatVecToData( ublas::prod(mLowerCV,gaussVariates(mMuVar->size()))+*mMuVar );
   
  };
  
  /** Make current variational posterior the new prior. Completes learning */
  void posteriorToPrior() 
  {
    typename baseT::guardT g( *This::mpUpdateLock );
    *mBeta=*mBetaVar;
    *mMu=*mMuVar;
    *mNu=*mNuVar;
    *mInvScale=*mInvScaleVar;
    doBeginLearning();
    baseT::posteriorToPrior();
  }


  /** \return KL-divergence between current variational parameter distribution and prior. */
  floatT getKL() 
  {
    floatT brat=*mBetaVar/(*mBeta);
    floatVecT mudiff=*mMuVar-*mMu;
    int dim=mMu->size();
    floatT klgauss=0.5*(dim*log(brat)+dim*(1.0/brat-1.0)+*mNuVar*(*mBeta)*ublas::inner_prod(mudiff,ublas::prod(invertMatrixCholesky(*mInvScaleVar),mudiff)));
    assert(klgauss>=-1e-8);
    floatT rv=klgauss+KLWishart(*mNuVar,*mNu,*mInvScaleVar,*mInvScale);
    return rv;
  };
  
  virtual void decayParameters(floatT tau, bool update_immediately = true)
  {
	floatT c = std::exp( -1 / tau );
	*mInvScale *= c;
	*mNu *= c;
	*mBeta *= c;
	variationalModel<data_t,float_t>::notifyParametersChanged();
  }
  
  void transform( const floatMatT & transmat )
  {
	*mMu = ublas::prod( transmat, *mMu );
	*mInvScale = ublas::prod( *mInvScale, ublas::trans( transmat ) );
	*mInvScale = ublas::prod( transmat, *mInvScale );
	//*mInvScale = ublas::prod( transmat, ublas::prod( *mInvScale, ublas::trans( transmat ) ) );
  }

  void increaseVariance( const floatVecT & amount )
  {
	// add amount to the diagonal
	for ( int i = 0; i < _dim; ++i )
	  (*mInvScale)( i, i ) += *mNuVar * amount[i];
  }  

protected:
  // variational params are in shared_ptrs so they can be shared across instances.
  pfloatT mNu,mNuVar; //!< concentration parameter of Wishart density on covariance matrix.
  pfloatT mBeta,mBetaVar; //!< concentration parameter of Gaussian density on mean.
  pfloatVecT mMu,mMuVar; //!< mean of Gaussian density on mean.
  pfloatMatT mInvScale,mInvScaleVar; //!< inverse scale matrix of wishart density.
  pfloatT mParentPostAccTotal; //!< accumulated parent posterior (=responsibilities) across all instances
  pfloatVecT mDataAccTotal; //!< accumulated data weighted by parent posterior across all instances
  pfloatMatT mData2AccTotal; //!< accumulated data*data^T weighted by parent posterior across all instances
  
  // every instance has one copy of these params.
  floatMatT mScaleVar; //!< scale matrix of wishart density. Expected precision of Gaussian = mNuVar*mScaleVar
  floatMatT mLowerCV; //!< lower triangular part of expected covariance matrix.  Useful for sampling.
  bool mLowerCVCurrent; //!< true if mLowerCV is up-to-date
  floatT mLogLikeConst; //!< constant summand in log likelihood.
  floatT mParentPostAcc; //!< accumulated parent posterior (=responsibilities)
  floatVecT mDataAcc; //!< accumulated data weighted by parent posterior.
  floatMatT mData2Acc; //!< accumulated data*data^T weighted by parent posterior.
  
  /** Share parameters with source */
  void doShareWith(const This &source,bool shareBase=false)
  {
    if(shareBase) baseT::doShareWith(source,shareBase);
    mNu=source.mNu;
    mNuVar=source.mNuVar;
    mBeta=source.mBeta;
    mBetaVar=source.mBetaVar;
    mMu=source.mMu;
    mMuVar=source.mMuVar;
    mInvScale=source.mInvScale;
    mInvScaleVar=source.mInvScaleVar;
    mParentPostAccTotal=source.mParentPostAccTotal;
    mDataAccTotal=source.mDataAccTotal;
    mData2AccTotal=source.mData2AccTotal;
    
    doBeginLearning();
    copyExpensive(source);
   
    
    
  }
  

  
  
  /** \return vector of gauss variates with mean 0 and variance 1 */
  floatVecT gaussVariates(unsigned dim) 
  {
    
 	floatVecT rr=ublas::zero_vector<floatT>(dim);
 	Random<float> & rng = Random<float>::get();
 	for(unsigned i=0;i<dim;++i) rr[i]=rng.normal();
 	return rr;
	
/*	floatVecT rv(dim+(dim % 2));
	for(unsigned i=0;i<rv.size();i+=2) {
	  floatT u1=drand48();
	  floatT u2=drand48();
	  rv[i]=rv[i+1]=sqrt(-2.0*log(u1));
	  rv[i]*=cos(2.0*M_PI*u2);
	  rv[i+1]*=sin(2.0*M_PI*u2);
	}
	rv.resize(dim);
	return rv;*/
  }
  
  /** compute & store scale matrix, the log of its determinant and the constant part of the log-likelihood */
  void precomputeExpensive()
  {
      floatT logdet=0.0;
	  // boost the diagonal until inversion works
      while( !invertMatrixCholesky(*mInvScaleVar,mScaleVar,logdet) )
		{
		  *mInvScaleVar+=ublas::identity_matrix<floatT>(mInvScaleVar->size1(),mInvScaleVar->size2());
		  cout<<"BOOST DIAG"<<endl;
		}
      mLowerCVCurrent=false;
      int dim=mMu->size();
      mLogLikeConst=dLogMultiGamma(*mNuVar,dim)+0.5*(-logdet-dim*log(M_PI)-dim/(*mBetaVar));  
  }
  
  
  void copyExpensive(const varmodelBaseT & source)
  {
    mScaleVar=((const This *) & source)->mScaleVar;
    mLowerCVCurrent=false;
    mLogLikeConst=((const This *) & source)->mLogLikeConst;
  }  
  
  
  
  
  static floatT KLWishart(const floatT & nu0, const floatT & nu1, const floatMatT & invScale0, const floatMatT & invScale1) 
  {
      floatT nudiff=nu0-nu1;
      int dim=invScale0.size1();
      assert(invScale0.size1()==invScale0.size2());
      assert(invScale1.size1()==invScale1.size2());
      assert(invScale1.size1()==invScale0.size1());
      floatT klwishart=0.5*(nudiff*(-logDet(invScale0)+2.0*dLogMultiGamma(nu0,dim))
                                +nu0*logDet(invScale0)-nu1*logDet(invScale1)+2.0*(logMultiGamma(nu1,dim)-logMultiGamma(nu0,dim))
                                +nu0*(trace(ublas::prod(invScale1,invertMatrixCholesky(invScale0)))-dim));
				
      //assert(klwishart>=-nu0*10.0*sqrt(FLT_EPSILON)); // roundoff in inversion goes with square root ?
      if(klwishart<=-nu0*sqrt(FLT_EPSILON)) cout<<"Warning: Wishart KL negative: "<<klwishart<<endl;
      return klwishart;    
  }

  static floatT KLhelper(const floatT &beta0, const floatT &beta1,const typename baseT::floatVecT &mu0,const typename baseT::floatVecT &mu1,
		      const floatT & nu0, const floatT & nu1, const floatMatT & invScale0, const floatMatT & invScale1) {
     
      floatVecT mudiff=mu0-mu1;
      int dim=mu0.size();
      floatT klgauss=0.5*(-logDet(invScale0)+logDet(invScale1)+2.0*(dLogMultiGamma(nu0,dim)-dLogMultiGamma(nu1,dim))
	  +nu1/(nu0-dim-1.0)*(1.0+1.0/beta0)*trace(prod(invertMatrixCholesky(invScale0),invScale1))+dim*(1.0/beta1-1.0)
	  +nu1*inner_prod(mudiff,prod(invScale1,mudiff))
      );
      
      assert(klgauss>=-1e-8);
      return klgauss+KLWishart(nu0,nu1,invScale0,invScale1);
  };
  


  
  
  /** Reset accumulators to begin learning. Implement in derived classes. */
  void doBeginLearning() 
  {
      mParentPostAcc=0.0;
      mDataAcc=ublas::zero_vector<floatT>(mMu->size());
      mData2Acc=ublas::zero_matrix<floatT>(mInvScale->size1(),mInvScale->size2());
      
      *mParentPostAccTotal=0.0;
      *mDataAccTotal=ublas::zero_vector<floatT>(mMu->size());
      *mData2AccTotal=ublas::zero_matrix<floatT>(mInvScale->size1(),mInvScale->size2());
  }
  
  /** Do inference (E-step) given parentPosterior. Computes log-likelihood
  */
  floatT doInference(const pdataT & data_ptr,bool variational) 
  {
    
      if(data_ptr.get()==0) return 0.0; // unobserved;
    
      RERR( (vectorCaster<dataT,pdataT,floatVecT>::dataToFloatVec(data_ptr).size()!=mMuVar->size()) 
      , (boost::format("Error in gauss-wishart inference: got vector of size %d, need %d") % vectorCaster<dataT,pdataT,floatVecT>::dataToFloatVec(data_ptr).size() % mMuVar->size()).str());
         
      floatVecT xmm = vectorCaster<dataT,pdataT,floatVecT>::dataToFloatVec(data_ptr) - *mMuVar;
      return mLogLikeConst - 0.5 * (*mNuVar) * ( ublas::inner_prod( xmm, ublas::prod(mScaleVar,xmm) ) ); // TODO: could be accelerated by a factor 2 via chol-desc. of the scale matrix
  };
  
  /** Accumulate sufficient statistics. Implement in derived classes.
  \param parentPost: parent posterior, e.g. responsibilities in a mixture model.
  */
  void doLearn(const pdataT & data_ptr, const floatT parentPost) 
  {
      if(parentPost<numeric_limits<floatT>::epsilon() || data_ptr.get()==0) return; // unobserved or not responsible
      
      floatVecT cd=vectorCaster<dataT,pdataT,floatVecT>::dataToFloatVec(data_ptr);
      
      RERR( (cd.size()!=mMuVar->size()) , (boost::format("Error in gauss-wishart inference: got vector of size %d, need %d") % cd.size() % mMuVar->size()).str());
           
      mDataAcc       += parentPost * cd;
      mParentPostAcc += parentPost;
      mData2Acc      += parentPost * outer_prod( cd, cd );
  }

  /** Update variational posterior with current accumulated sufficient statistics and prior (M-step). 
  */
  bool doVariationalParameterUpdate()
  {
	if( mParentPostAcc / (*mBeta) < FLT_EPSILON && mParentPostAcc / (*mNu) < FLT_EPSILON )
	  return false; //not responsible for these data
	*mParentPostAccTotal+=mParentPostAcc;
	*mDataAccTotal+=mDataAcc;
	*mData2AccTotal+=mData2Acc;
	
	if(This::mFreezeLevel<2) 
      {
		*mBetaVar=*mBeta+*mParentPostAccTotal;
		*mMuVar=(*mDataAccTotal+*mBeta*(*mMu))/(*mBetaVar);
      }
	
	if(This::mFreezeLevel<1)
      {
		*mNuVar=*mNu+*mParentPostAccTotal;
		const floatVecT dmean=(*mDataAccTotal)/(*mParentPostAccTotal+FLT_MIN);
		const floatVecT mudiff=dmean-*mMu;
		*mInvScaleVar = *mInvScale + *mData2AccTotal + *mParentPostAccTotal *( -outer_prod(dmean, dmean) + *mBeta / (*mBetaVar) * outer_prod( mudiff, mudiff ) );
      }
	
	return true;
  };
  
  /** Reset variational parameter posterior to prior */
  void doResetParameterPosteriorToPrior() 
  {
    *mBetaVar=*mBeta;
    *mNuVar=*mNu;
    *mMuVar=*mMu;
    *mInvScaleVar=*mInvScale;
  }  
  
   /** Accumulate likelihood and hidden variable part of the variational bound.
  \param parentPost: responsibility for this datapoint
  */
  void accumulateVarBound(floatT parentPost=1.0) 
  {
    This::mAccumulatedVarBound+=This::getLogLikelihood()*parentPost; // there are no hidden variables and no children (i.e. this component is connected to the data)
  }
  
  
  /** write to stream.
  \param os:  reference to ostream to write into.
  */
  void writeToStream(ostream &os) const 
  {
	unsigned oldpc=os.precision(7);
	os<<"Gaussian emission model with Gauss-Wishart prior"<<endl;
	os<<"Prior mean: "<<*mMu;
	os<<"Variational posterior mean: "<<*mMuVar;
	os<<"Concentration parameters on mean: "<<*mBeta<<" "<<*mBetaVar<<endl;
	os<<"Inverse scale matrix, prior: "<<endl;
	os<<*mInvScale;
	os<<"Inverse scale matrix, variational posterior: "<<endl;
	os<<*mInvScaleVar;
	os<<"Wishart concentration parameters: "<<*mNu<<" "<<*mNuVar<<endl;
	os<<"Expected covariance matrix (variational)"<<endl;
	floatMatT ecv=*mInvScaleVar/(*mNuVar-mMu->size()-1.0);
	os.precision(oldpc);
	os<<ecv;
	baseT::writeToStream(os);
  };
  
  /** read from stream is. Implement in derived classes.
	  \param is : reference to stream to read from.
  */
  void readFromStream(istream &is) 
  {
	string name; 
	getline(is,name);
	if(name!="Gaussian emission model with Gauss-Wishart prior") throw runtime_error("Error loading Gauss-Wishart: got "+name);
	is>>eatString(2)>>*mMu;
	is>>eatString(3)>>*mMuVar;
	mDataAcc=ublas::zero_vector<floatT>(mMu->size());
	is>>eatString(4)>>*mBeta>>*mBetaVar;
	mParentPostAcc=0.0;
	is>>eatString(4);
	is>>*mInvScale;
	is>>eatString(5);
	is>>*mInvScaleVar;
	mData2Acc=ublas::zero_matrix<floatT>(mInvScale->size1(),mInvScale->size2());
	is>>eatString(3)>>*mNu>>*mNuVar;
	is>>eatString(4);
	floatMatT dummat;
	is>>dummat;
	baseT::readFromStream(is);
	{
	  typename baseT::guardT g( *This::mpUpdateLock );
	  doBeginLearning();
	}
	This::updateExpensiveForAll(true);
	
  }
  
  bool checkDataShapes(const pdataT & data_ptr) {
    if(data_ptr.get()==0) return true;
    return data_ptr->size()==mMu.size();
  };
  
private:

  int _dim;
  floatT _mu;
  floatT _beta; 
  floatT _nu;
  floatT _ivar;
      
};


