/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/

#pragma once

#include "SeariseProperties.hpp"
#include "CPUData.hpp"
#include "variational_model.hpp"
#include "gauss_wishart.hpp"
#include <algorithm>
#include <iostream>
#include "float.h"
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/shared_ptr.hpp>
#include <limits>
#include "ublas_storage_adaptors.hpp"

using namespace std;
using namespace boost::numeric::ublas;





/** Gaussian optic flow feature with gauss-wishart prior. Can segment flow patch into features. */
template<typename float_type=float>
class gaussWishartFlowFeature : public gaussWishart< CPUData<float_type> , float_type >,
  public HasProperties< SearisePropertyTypes, gaussWishartFlowFeature<float_type> >
{
public:
  
  typedef gaussWishart< CPUData<float_type> , float_type > baseT;
  typedef typename baseT::varmodelBaseT varmodelBaseT;
  typedef typename baseT::dataT dataT;
  typedef typename baseT::pdataT pdataT;
  typedef typename baseT::floatT floatT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::intVecT intVecT;
  typedef typename baseT::floatMatT floatMatT;  
  
  typedef boost::shared_ptr<int> pint;
  
  typedef gaussWishartFlowFeature<float_type> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  
  
  /** Default init: feature size 10x10, 2 optic flow components, 2x2 features per patch */
  gaussWishartFlowFeature()
  : mFX(5)
  , mFY(5)
  , mFF(1)
  , mPX(2)
  , mPY(2)
  {}
  
  /** create parameter-sharing copy of source. If you want a private copy of the feature, call unshare() afterwards. */
  gaussWishartFlowFeature(const This &source) : baseT(source)
  {
    mFX=source.mFX;
    mFY=source.mFY;
    mFF=source.mFF;
    mPX=source.mPX;
    mPY=source.mPY;
  }
  
  
  
  static void init_properties()
  {
    PBase::add_ivar_property_rw( "featureX", &This::mFX );
    PBase::add_ivar_property_rw( "featureY", &This::mFY );
    PBase::add_ivar_property_rw( "flowComponents", &This::mFF );
    PBase::add_ivar_property_rw( "segmentsX", &This::mPX );
    PBase::add_ivar_property_rw( "segmentsY", &This::mPY );
  }
  
  /** Set prior. Call after setting properties. 
  \param add_pseudodata: number of sampled datapoints for feature initialisation. Average will be rescaled to match feature properties n0 and beta0
  */
  void setPrior( float_type add_pseudodata = 0 ) 
  {
    RERR( (mFX<1) , (boost::format("Error initializing gaussWishartFlowFeature: feature size X=%d") % mFX).str());
    RERR( (mFY<1) , (boost::format("Error initializing gaussWishartFlowFeature: feature size Y=%d") % mFY).str());
    RERR( (mFF<1) , (boost::format("Error initializing gaussWishartFlowFeature: number of optic flow components=%d") % mFF).str());
    RERR( (mPX<1) , (boost::format("Error initializing gaussWishartFlowFeature: segments in X direction=%d") % mPX).str());
    RERR( (mPY<1) , (boost::format("Error initializing gaussWishartFlowFeature: segments in Y direction=%d") % mPY).str());
    int featureSize=mFX*mFY*mFF;
    This::set_property("D",featureSize);
    baseT::setPrior(add_pseudodata);
  }
  

  
  friend bool operator==(const This & first,const This & second) 
  {
   return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
    bool rv= 	(first.mFX==second.mFX) && (first.mFY==second.mFY) && (first.mFF==second.mFF)
	  &&	(first.mPX==second.mPX) && (first.mPY==second.mPY);
    if(!rv) return false;
    return baseT::cmp(first,second); 
  }
  
  pdataT sample() {
    pdataT samp=baseT::sample();

    RERR( ((int)samp->rows()!=mFX*mFY*mFF || (int)samp->columns()!=1 || (int)samp->depth()!=1)
    , (boost::format("Error sampling gaussWishartFlowFeature: cannot reshape CPUData(%d,%d,%d) into CPUData(%d,%d,%d)") % samp->rows() % samp->columns() % samp->depth() % mFX % mFY % mFF).str());
    
    pdataT rv=pdataT( new dataT(mFX,mFY,mFF) );
    rv->make_contiguous();
    std::copy(samp->data(),samp->data()+samp->size(),rv->data()); //FIXME: reshaping would avoid copy. Ask Daniel if this is possible with CPUData ?
    
    return rv;
  };
  
  
  /** \return segments X * segments Y */
  int segments() const {return mPX*mPY;};
  
  /** \return patch size in x direction */
  int patchSizeX() const {return mPX*mFX;};
  
  /** \return patch size in y direction */
  int patchSizeY() const {return mPY*mFY;};
  
  /** Get data segment */
  pdataT getDataSegment(const pdataT &data_ptr,unsigned segment=0) const {
    int sx,sy;
    getPosFromSegment(sx,sy,segment);
    
    const int xext=data_ptr->rows();
    const int yext=data_ptr->columns();
    
    const int xpos=sx+xext/2;
    const int ypos=sy+yext/2;
    
    if(xpos<0 || xpos+mFX>xext || ypos<0 || ypos+mFY>yext) {
      // cout<<(boost::format("UNOBSERVED SEGMENT %d AT (%d,%d) with size (%d,%d), total area (%d,%d)") % segment % xpos % ypos % mFX % mFY % xext % yext).str()<<endl;
      return pdataT(); // part of the segment is outside of the patch. treat as unobserved
    }
    
    RERR( ((int)data_ptr->depth() != mFF) , (boost::format("Error getting data segment: got %d optic flow components, need %d") % data_ptr->depth() % mFF).str());
        
    pdataT seg=pdataT( new dataT(*data_ptr,xpos,mFX,ypos,mFY,true) ); 
    seg->make_contiguous();
    return seg;
    
  };
  
  /** Set data in segment of data_ptr from data_source. data_ptr must point to allocated memory. Assumes that an assigment operator for these data exists. */
  void setDataSegment(pdataT &data_ptr,pdataT &data_source,unsigned segment=0) const {
    int sx,sy;
    getPosFromSegment(sx,sy,segment);
    
    sx+=mFX*mPX/2; // transform back to coordinates relative to top-left corner
    sy+=mFY*mPY/2;
    
    RERR( (sx+mFX>(int)data_ptr->rows() || sy+mFY>(int)data_ptr->columns()) 
    , (boost::format("Error setting data segment at (%d,%d) with sizes (%d,%d): data has size (%d,%d)") % sx % sy % mFX % mFY % data_ptr->rows() % data_ptr->columns()).str() ); 
    RERR( ((int)data_ptr->depth() != mFF) , (boost::format("Error getting data segment: got %d optic flow components, need %d") % data_ptr->depth() % mFF).str());
    RERR( ((int)data_source->depth() != mFF || (int)data_source->rows() != mFX || (int)data_source->columns() != mFY)
    , (boost::format("Error setting data with shapes (%d,%d,%d): need (%d,%d,%d)") % data_source->rows() % data_source->columns() % data_source->depth() % mFX % mFY % mFF).str() );
    
   
    for(int x=0;x<mFX;++x)
      for(int y=0;y<mFY;++y)
	(*data_ptr)(x+sx,y+sy)=(*data_source)(x,y);
      
    
  }
  
  /** \return a new datapoint, initalized to the correct size */
  pdataT newData() const {pdataT nd=pdataT( new dataT(mFX*mPX,mFY*mPY,mFF) );std::fill(nd->data(),nd->data()+nd->size(),0.0); return nd;}
  
  void getTotalDataSize(int &xs, int &ys, int &fd) const {
    xs=mFX*mPX;
    ys=mFY*mPY;
    fd=mFF;
  }
  
protected:
  int mFX,mFY; //!< feature dimensions in x and y direction
  int mFF; //!< number of flow components
  int mPX,mPY; //!< features per patch in x and y direction
  
  /** compute x and y positions (top left corner) of feature from segment number. Coordinates are relative to center.*/
  void getPosFromSegment(int & sx,int &sy,int segment) const
  {
    const int numseg=mPX*mPY;
    const int xcenter=mPX*mFX/2;
    const int ycenter=mPY*mFY/2;
    
    RERR( (segment>=numseg) , (boost::format("Error getting data segment %d of %d in gaussWishartFlowFeature") % segment % numseg).str() );
    
    sx=(segment / mPY)*mFX-xcenter;
    sy=(segment % mPY)*mFY-ycenter;
  }
  
  

  void writeToStream(ostream &os) const
  {
    os<<"Gauss-Wishart optic flow feature"<<endl;
    os<<"Feature size X: "<<mFX<<endl;
    os<<"Feature size Y: "<<mFY<<endl;
    os<<"Optic flow components: "<<mFF<<endl;
    os<<"Features per patch, X: "<<mPX<<endl;
    os<<"Features per patch, Y: "<<mPY<<endl;
    baseT::writeToStream(os);
  }
  
  void readFromStream(istream &is) 
  {
    string dummy;
    getline(is,dummy);
    is>>eatString(3)>>mFX>>ws;
    is>>eatString(3)>>mFY>>ws;
    is>>eatString(3)>>mFF>>ws;
    is>>eatString(4)>>mPX>>ws;
    is>>eatString(4)>>mPY>>ws;
    baseT::readFromStream(is);
  }
    
  

  bool checkDataShape(const pdataT & data_ptr) const {
    if(data_ptr.get()==0) return true;
    RERR( ((int)data_ptr->rows()!=mFX || (int)data_ptr->columns()!=mFY || (int)data_ptr->depth()!=mFF)
    , (boost::format("Wrong data shape (%d,%d,%d) in gaussWishartFlowFeature: need (%d,%d,%d)") % data_ptr->rows() % data_ptr->columns() % data_ptr->depth() % mFX % mFY % mFF).str() );
    return true;
  }
  
};

