/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/

#pragma once


#include <stdlib.h>
#include <cassert>
#include <string>
#include "variational_model.hpp"
#include "SeariseProperties.hpp"

//using namespace std;



/** Label observation model. Useful for learning and prediciting class labels.
*/
template<class label_type=std::string,class float_type=float>
class labelObservation: public variationalModel<label_type,float_type>, public HasProperties< SearisePropertyTypes, labelObservation<label_type,float_type> >
{

public:
  
  typedef variationalModel<label_type,float_t> baseT;
  typedef typename baseT::dataT dataT;
  typedef typename baseT::floatT floatT;
  typedef typename baseT::pdataT pdataT;
  
  typedef labelObservation<label_type,float_t> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  
  
  labelObservation() 
  : baseT()
  ,mLabel()
  {}
  
  static void init_properties()
  {
    PBase::add_ivar_property_rw( "label", &This::mLabel );
  }

  
  
  void setLabel(const label_type &label) {
    mLabel=label;
  }
  
  pdataT sample() {return pdataT( new dataT(mLabel) ) ;};
    
  friend floatT getKLto(const labelObservation<label_type,floatT> & first,const labelObservation<label_type,floatT> & second) 
  {
        if(first.mLabel==second.mLabel) return 0.0;
        return FLT_MAX/1e3;
  }
  
  bool friend operator<(const labelObservation &l1,const labelObservation &l2) {
    return l1.mLabel<l2.mLabel;
  }
  
  
  const label_type & getLabel() const {return mLabel;};
  
  void posteriorToPrior() {};
  
  floatT getKL() 
  {
    return 0.0;
  }
  
  friend bool operator==(const This & first,const This & second) 
  {
    return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
   if(first.mLabel!=second.mLabel) return false;
   return baseT::cmp(first,second);
  }  
    
protected:
  
  label_type mLabel; //!< label of this label observation model.
  
  
  // These methods don't need to be implemented here, because there's nothing to learn
  void precomputeExpensive(){};
  void doBeginLearning() {};
  void doLearn(const pdataT & data_ptr,const floatT parentPost) {};
  bool doVariationalParameterUpdate(){return false;}
  
  
  floatT doInference(const pdataT & data_ptr,bool variational)  
  {
    if (data_ptr.get()==0) return 0.0; // unobserved label
    if (*data_ptr==mLabel) return 0.0;
    return -1e10;
  }
  
  floatT doPredict(const pdataT & data_ptr) { return doInference(data_ptr,true); };
  
  void writeToStream(ostream &os) const 
  {
    os<<"Label observation model"<<endl;
    os<<mLabel<<endl;
  }
  void readFromStream(istream &is) {
    std::string name;
    getline(is,name);
    if(name!="Label observation model") throw std::runtime_error("Error loading label observation model: got "+name);
    is>>mLabel>>ws;
  }
  bool checkDataShapes(const pdataT & data_ptr) const {return true;};
  
};


template<>
labelObservation<std::string,float>::labelObservation()
  :variationalModel<std::string,float>()
  ,mLabel("@@an*unlikely*label!!") {};
  
template<>
labelObservation<int,float>::labelObservation()
  :variationalModel<int,float>()
  ,mLabel(-1) {};
  
template<>
labelObservation<float,float>::labelObservation()
  :variationalModel<float,float>()
  ,mLabel(-1.0) {};

