/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/


#include "ex_sequence.hpp"
#include "gauss_wishart_flow_feature.hpp"
#include "label_observation.hpp"
#include "CPUData.hpp"
#include "SeariseProperties.hpp"
#include "varutils.h"
#include <string>
#include <set>
#include <algorithm>
#include <map>
#include <functional>

using namespace std;



class sequenceClassifier : public HasProperties< SearisePropertyTypes, sequenceClassifier >
{
public:

  typedef gaussWishartFlowFeature<> featureT;		// full cov. gauss feature observation model for optic flow
  typedef bottomComponentContainer<featureT> featureBagT;	// a salient region is a patch of features
  typedef bottomComponentContainer<featureT>::componentsT pfeaturesT;
  typedef featureBagT::pComponentT pfeatureT;
	
  typedef labelObservation<string> labelT;		// the class label observation model
  typedef bottomComponentContainer<labelT> labelBagT;	// and its container for all labels
  typedef bottomComponentContainer<labelT>::componentsT plabelsT;
  typedef labelBagT::pComponentT plabelT;
	
  typedef exSequence<featureBagT> patchSeqT;			// feature patches are concatenated to sequences
  typedef componentContainer<patchSeqT> sequenceBagT;	// which are held in this container
  typedef componentContainer<patchSeqT>::componentsT psequencesT;
  typedef patchSeqT::pComponentContainerT pfeatureBagT;
  typedef sequenceBagT::pComponentT psequenceT;
	
  typedef dpmmc<sequenceBagT,labelBagT> seqLabAssocT;	// the possible associations between labels and sequences are learned by a DP mixture
  typedef seqLabAssocT::pComponentContainerT1 psequenceBagT;
  typedef seqLabAssocT::pComponentContainerT2 plabelBagT;
  
  
  

  typedef sequenceClassifier This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  typedef SeariseProperty PropertyType;
  
  sequenceClassifier() 
  : mModel()
  , mFX(5)
  , mFY(5)
  , mFF(2)
  , mPX(3)
  , mPY(3)
  , mPatchX(5*3)
  , mPatchY(5*3)
  , mMaxLabels(30)
  , mNumFeatures(10)
  , mSeqLen(5)
  , mKnownLabels()
  , mLastGuess("unknown")
  , mLastGuessProb(0.0)
  , mNumFramesSinceReset(0)
  , mbLearning(false)
  {
    initModel();
  }
  
  sequenceClassifier(const This &source)
  :PBase(source)
  ,mModel(source.mModel)
  , mFX(source.mFX)
  , mFY(source.mFY)
  , mFF(source.mFF)
  , mPX(source.mPX)
  , mPY(source.mPY)
  , mPatchX(source.mPatchX)
  , mPatchY(source.mPatchY)
  , mMaxLabels(source.mMaxLabels)
  , mNumFeatures(source.mNumFeatures)
  , mSeqLen(source.mSeqLen)
  , mKnownLabels(source.mKnownLabels)
  , mLastGuess(source.mLastGuess)
  , mLastGuessProb(source.mLastGuessProb)
  , mNumFramesSinceReset(source.mNumFramesSinceReset)
  , mbLearning(source.mbLearning)
  
  {
    mModel.unshare();
  }
  
  static void init_properties()
  {
    // properties which can only be changed before call to initModel()
    PBase::add_ivar_property_rw( "featureX", &This::mFX );
    PBase::add_ivar_property_rw( "featureY", &This::mFY );
    PBase::add_ivar_property_rw( "flowComponents", &This::mFF );
    PBase::add_ivar_property_rw( "segmentsX", &This::mPX );
    PBase::add_ivar_property_rw( "segmentsY", &This::mPY );
    PBase::add_ivar_property_rw( "maxNumLabels", &This::mMaxLabels );
    PBase::add_ivar_property_rw( "numFeatures", &This::mNumFeatures );
    
    // properties which can be changed at any time
    PBase::add_property( "numThreads", 0, PBase::MemFunSetter( &This::setNumThreads ) );
    PBase::add_property( "sequenceLength", PBase::make_ivar_getter( &This::mSeqLen ), PBase::MemFunSetter( &This::setSequenceLength ) );
    PBase::add_property( "freezeLevel", 0, PBase::MemFunSetter( &This::setFreeze ) );
  }

  
  void initModel() 
  {
    psequenceBagT sequences=mModel.getComponentContainer1();
    pfeatureBagT features=sequences->getSubContainer();
    featureT & fproto=features->getPrototype();
    
    // init feature sizes and segments
    features->setShared(true);
    fproto.set_property("featureX",mFX);
    fproto.set_property("featureY",mFY);
    fproto.set_property("flowComponents",mFF);
    fproto.set_property("segmentsX",mPX);
    fproto.set_property("segmentsY",mPY);
    fproto.set_property("m0",0.0);
    fproto.set_property("S0",1.0);
    fproto.set_property("n0",mFX*mFY+10.0);
    fproto.setPrior(0);
    mPatchX=fproto.patchSizeX();
    mPatchY=fproto.patchSizeY();
    features->resize(mNumFeatures);
    
    
    // resize sequence-label associations
    mModel.set_property("finite",true);
    mModel.set_property("alpha0",0.1);
    mModel.resize(mMaxLabels);
    mModel.setPrior(0);
    
    
    // init label names
    plabelBagT labels=mModel.getComponentContainer2();
    plabelsT plab=labels->getComponents();
    for(int cl=0;cl<labels->components();++cl)
    {
     plab(0,cl)->set_property("label",unknownLabel(cl)); 
     plab(0,cl)->setPrior(0); 
    }
    
    
    // init sequences
    sequences->getPrototype().set_property("finite",true);
    psequencesT pseq=sequences->getComponents();
    for(int seqno=0;seqno<sequences->components();++seqno) {
	 pseq(0,seqno)->set_property("alpha0",0.1);
    }
    
    mbLearning=false;
  }
  
  void setPrior(float pseudodata=10.0)
  {
    psequenceBagT sequences=mModel.getComponentContainer1();
    pfeatureBagT features=sequences->getSubContainer();
    
    features->setPrior(scalar_vector<float>(features->components(),pseudodata));
    sequences->setPrior(scalar_vector<float>(sequences->components(),pseudodata));
    
    mbLearning=false;
       
  }
  
  
  void addLabel(string slabel)
  {
    if(mKnownLabels.count(slabel)==0) // new label
    {
      RERR( ((int)mKnownLabels.size()==mMaxLabels) , (boost::format("Error creating new label: maximal number of labels %d reached") % mMaxLabels).str());
      string ulab=unknownLabel(mKnownLabels.size());
      plabelBagT labels=mModel.getComponentContainer2();
      plabelsT plab=labels->getComponents();
      bool found=false;
      for(int cl=0;cl<labels->components();++cl)
      {
	if(plab(0,cl)->getLabel()==ulab) {
	 plab(0,cl)->setLabel(slabel);
	 mKnownLabels.insert(slabel);
	 found=true;
	 break;  
	}
      }
      RERR( (!found) , "Error creating new label: could not find "+ulab );
    } 
  }
  
  /** Set number of threads used in learn/recognize calls */
  void setNumThreads(PropertyType nt) {
    psequenceBagT sequences=mModel.getComponentContainer1();
    pfeatureBagT features=sequences->getSubContainer();
    
    features->setNumThreads(cast<int>(nt));
  
  };
  
  /** Set sequence length after which sequence model hidden vars. will be reset to prior. Resets model to prior. */
  void setSequenceLength(PropertyType sl) {
    reset();
    mSeqLen=std::max(1,cast<int>(sl));
  }
  

  
  string addLabel(const CPUData<float> & label)
  {
    string slabel=CPUDataToString(label);
    addLabel(slabel);
    return slabel;
  }
  
  void beginLearning(bool accumulateVarBound=false)
  {
   mModel.beginLearning(accumulateVarBound); 
   mbLearning=true;
  }
  
  
  void learn(CPUData<float> & flow,CPUData<float> & label)
  {
    RERR( (!mbLearning) , "Error: call beginLearning() before learn");
    
    string slabel=addLabel(label);
    const CPUData<float> fl=rightsizeFlow(flow);
    seqLabAssocT::pdataT ldat=seqLabAssocT::pdataT( new seqLabAssocT::dataT() );
    ldat->first=seqLabAssocT::pdataT1( new seqLabAssocT::dataT1(fl) );
    ldat->second=seqLabAssocT::pdataT2( new seqLabAssocT::dataT2(slabel) );
    
    mModel.inference(ldat);
    mModel.learn(ldat);
    
  }
  
  void updateParameterPosterior() {
    mModel.updateParameterPosterior();
    mbLearning=false;
  }
  
  void posteriorToPrior() {
    mModel.posteriorToPrior();
    mbLearning=false;
  }
  
  float getVarBound() {return mModel.getVarBound();}
  
  bool isLearning() {return mbLearning;}
  
  
  void recognize(const CPUData<float> & flow) {
    
    if(mNumFramesSinceReset==mSeqLen) {
      mModel.resetToPrior();
      mNumFramesSinceReset=0;
    }
    
    const CPUData<float> fl=rightsizeFlow(flow);
    seqLabAssocT::pdataT ldat=seqLabAssocT::pdataT( new seqLabAssocT::dataT() );
    ldat->first=seqLabAssocT::pdataT1( new seqLabAssocT::dataT1(fl) );
    ldat->second=seqLabAssocT::pdataT2(); // label unobserved during recognition
    
    mModel.inference(ldat);
    mNumFramesSinceReset++;
   
    if(mNumFramesSinceReset>=mSeqLen/2) {
     mModel.predictPass();
     map<float,string,greater<float> > probLab;
     plabelBagT labels=mModel.getComponentContainer2();
     plabelsT plab=labels->getComponents();
     for(int cl=0;cl<labels->components();++cl) {
       probLab[plab(0,cl)->getParentPost()]=plab(0,cl)->getLabel(); 
     }
     mLastGuess=probLab.begin()->second;
     mLastGuessProb=probLab.begin()->first;
     if(isUnknown(mLastGuess)) mLastGuess="unknown";
    } 
  }
  
  void clearBuffer() {mModel.clearBuffer();};
  
  void bufferData(const CPUData<float> & flow,const CPUData<float> &label,bool resetPoint=false) {
  
    const CPUData<float> fl=rightsizeFlow(flow);
    
    seqLabAssocT::pdataT ldat=seqLabAssocT::pdataT( new seqLabAssocT::dataT() );
    ldat->first=seqLabAssocT::pdataT1( new seqLabAssocT::dataT1(fl) );
    
    if(label.rows()==0) {
      ldat->second=seqLabAssocT::pdataT2(); // no label. learn unsupervised
    } else {
      string slabel=addLabel(label);
      ldat->second=seqLabAssocT::pdataT2( new seqLabAssocT::dataT2(slabel) ); // label provided. supervised learning.
    }
    
    mModel.bufferData(ldat,resetPoint);
  };
  
  
  void learnBuffer() {
    mModel.inferBuffer();
    mModel.learnBuffer();
  };
  
  
  void reset() {
   mNumFramesSinceReset=0;
   mModel.resetToPrior();
   mLastGuess="unknown";
   mLastGuessProb=0.0;
  }
  
  /** \param size: size of CPUData for string. If label is shorter than size, pad with 0.0
  \return CPUData representation of label
  */
  CPUData<float> guessLabel(int size=-1) {
    return stringToCPUData(mLastGuess,size);
  }
  
  float guessConfidence() {
    return mLastGuessProb;
  }
  
  
  void setFreeze(PropertyType frl)
  {
    int fr=cast<int>(frl);
    freeze(fr);
  }
  
  
  void freeze(int level=0) {
    int featureLevel=level % 10;
    int sequenceLevel=(level/10) % 10;
    int assocLevel=(level/100) % 10;
    
    psequenceBagT sequences=mModel.getComponentContainer1();
    pfeatureBagT features=sequences->getSubContainer();
    
    mModel.freeze(assocLevel);
    sequences->freeze(sequenceLevel);
    features->freeze(featureLevel);
    
  }
  
  void sample(CPUData<float> & flow, CPUData<float> & label)
  {
    seqLabAssocT::pdataT samp=mModel.sample();
    flow=*(samp->first);
    label=stringToCPUData(*(samp->second));
  }
  
  friend bool operator==(const This & left,const This & right)
  {
    if( (left.mFX!=right.mFX) || (left.mFY!=right.mFY) || (left.mFF!=right.mFF)
       || (left.mPX!=right.mPX) || (left.mPY!=right.mPY)
       || (left.mMaxLabels!=right.mMaxLabels) || (left.mNumFeatures!=right.mNumFeatures) || (left.mSeqLen!=right.mSeqLen) 
       || (left.mKnownLabels!=right.mKnownLabels) )
      return false;
    
    return left.mModel==right.mModel;  
  }
  
  
  friend ostream & operator<<(ostream &os,const This & lc) {
    os<<"Sequence classifier"<<endl;
    os<<"Sequence length: "<<lc.mSeqLen<<endl;
    os<<"Number of known labels: "<<lc.mKnownLabels.size()<<endl;
    os<<"Known labels:"<<endl;
    BOOST_FOREACH(const string & lab,lc.mKnownLabels)
      os<<lab<<endl;
    os<<lc.mModel;  
    return os; 
  }
  
  friend istream & operator>>(istream &is,This & lc) {
   string dummy;
   getline(is,dummy);
   RERR( (dummy!="Sequence classifier") , "Error loading sequence classifier: got "+dummy);
   is>>eatString(2)>>lc.mSeqLen>>ws;
   int numlab;
   is>>eatString(4)>>numlab>>ws;
   getline(is,dummy);
   for(int i=0;i<numlab;++i) {
    is>>dummy>>ws;
    lc.mKnownLabels.insert(dummy);
   }
   is>>lc.mModel;
   
   This::psequenceBagT sequences=lc.mModel.getComponentContainer1();
   This::pfeatureBagT features=sequences->getSubContainer();
   This::featureT & fproto=features->getPrototype();
   This::plabelBagT labels=lc.mModel.getComponentContainer2();
   
   lc.mFX=cast<int>(fproto.get_property("featureX"));
   lc.mFY=cast<int>(fproto.get_property("featureY"));
   lc.mFF=cast<int>(fproto.get_property("flowComponents"));
   lc.mPX=cast<int>(fproto.get_property("segmentsX"));
   lc.mPY=cast<int>(fproto.get_property("segmentsY"));
   
   lc.mMaxLabels=labels->components();
   lc.mNumFeatures=features->components();
   
   return is;
  }
  
  static CPUData<float> stringToCPUData(const string & s,int size=-1)
  {
   if(size<0) size=s.size();
   CPUData<float> rv=CPUData<float>(size,1,1);
   std::fill(rv.data(),rv.data()+size,0.0f);
   for(unsigned i=0;i<std::min((size_t)size,s.size());++i) rv(i,0,0)=(char)s[i];
   return rv;
  }
  
  static string CPUDataToString(const CPUData<float> & d)
  {
    string rv;
    RERR( (d.columns()!=1 || d.depth()!=1) , (boost::format("Error converting CPUData to string: need colums()==1,depth()==1, got colums()=%d, depth()=%d") % d.columns() % d.depth()).str());
    for(unsigned i=0;i<d.rows();++i) {
      if(d(i,0,0)==0.0) break;
      rv+=(char)d(i,0,0);
    } 
    return rv;
  }
  
  
protected:
  seqLabAssocT mModel; //!< the generative sequence classifier
  int mFX,mFY,mFF; //!< feature size in x,y and optic flow components
  int mPX,mPY; //!< features per patch in x,y direction
  int mPatchX,mPatchY; //!< patch sizes in x,y direction
  int mMaxLabels; //!< maximal number of labels
  int mNumFeatures; //!< number of features
  int mSeqLen; //!< sequence length after which label posterior is reset to prior
  set<string> mKnownLabels; //!< set of known labels
  string mLastGuess; //!< last label guess
  float mLastGuessProb; //!< probability of last label guess
  int mNumFramesSinceReset; //!< number of frames since last reset to prior
  bool mbLearning; //!< true if beginLearning() was called. false if updateParameterPosterior() or posteriorToPrior() was called.
  
  
  
  
  string unknownLabel(int num) {return (boost::format("_unknown%04d")  % num ).str();}
  
  bool isUnknown(const string &s) {return s.substr(0,8)=="_unknown";}
  
  
  
  const CPUData<float> rightsizeFlow(const CPUData<float> & flow)
  {
    RERR ( ((int)flow.depth()!=mFF) , (boost::format("Error rightsizing flow in sequenceClassifier: need %d flow components, got %d") % mFF % flow.depth()).str() );

    
    return CPUData<float>(flow);
    
    int sx=(flow.rows()-mPatchX)/2;
    int sy=(flow.columns()-mPatchY)/2;
    
    if(mPatchX==(int)flow.rows() && mPatchY==(int)flow.columns()) return flow;
    if(mPatchX>(int)flow.rows() || mPatchY>(int)flow.columns()) {
      CPUData<float> rv(flow);
      rv.make_unique();
      rv.upsample_atleast(mPatchX,mPatchY);
      return CPUData<float>(rv,sx,mPatchX,sy,mPatchY);
    }
    return CPUData<float>(flow,sx,mPatchX,sy,mPatchY);
  }
  
};


