#pragma once

#include "SeariseProperties.hpp"
#include "variational_model.hpp"
#include "component_container.hpp"
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_expression.hpp>
#include <boost/format.hpp>
#include "gamma.hpp"
#include <vector>
#include <algorithm>
#include <numeric>
#include <boost/concept_check.hpp>


using namespace boost::numeric::ublas;
using namespace boost;


/** Sequence model, basically a restrictable HMM for faster memorization of single examples. Restriction is feasible by using a (finite) stick-breaking
prior on the transition matrix. This way, it is e.g. possible to specify that only the current state and its next neighbour in time have non-zero probability with high a-priori uncertainty,
whereas all the other states are (almost certainly) unreachable. Uses the variationalModel interface for data buffering. Data can be in segments, in which case this model implements a "bag of sequences".

If you want to learn in batches, you might have to reset the hidden variable posterior whenever a new sequence begins:

\code
for n in number-of-EM-steps or until converged:
  sequenceModel.beginLearning(); 			// resets state dist. to prior. Erases buffer.
  for x in data:  					// x must be a shared_ptr to data
    if(new sequence) sequenceModel.resetToPrior() 	// reset state dist to prior if new sequence begins.
    sequenceModel.infer(x)				// update state distribution
    sequenceModel.bufferData(x)				// buffer datapoint
  sequenceModel.learnBuffer(parentPost)			// learn from buffered data. Assume that there is only 1 parent post at the moment.
  sequenceModel.updateVariationalPosterior() 		// reestimate var. post. of params. erases buffer (state dists. will change when params change -> buffer invalidated)
sequenceModel.posteriorToPrior() 			// forget history, just remember results.
\endcode

*/


template<class component_container_type1,class component_container_type2=int>
class sequenceModel : public varmodSelector<component_container_type1,component_container_type2>::baseT,
  public HasProperties< SearisePropertyTypes, sequenceModel<component_container_type1,component_container_type2> >
{

public:

  typedef typename varmodSelector<component_container_type1,component_container_type2>::baseT baseT;
  
  typedef typename baseT::floatT floatT;
  typedef typename baseT::floatVecT floatVecT;
  typedef typename baseT::dataT dataT;
  typedef boost::shared_ptr<dataT> pdataT;
    
  typedef sequenceModel<component_container_type1,component_container_type2> This;
  typedef HasProperties< SearisePropertyTypes, This > PBase;
  typedef SeariseProperty PropertyType;

  typedef Random<floatT> float_rng_t;
  
  
  typedef typename ublas::matrix<floatT> floatMatT;
  
  
  
  
  sequenceModel() 
  : baseT<dataT,floatT>()
  , mMaxBufSize(0)
  , mPriorType("forward")
  
  {
  };
  

  static void init_properties()
  {
    PBase::add_ivar_property_rw( "maxBufSize", &This::mMaxBufSize );
    PBase::add_ivar_property_rw( "priorType", &This::mPriorType );
    PBase::add_ivar_property_rw( "alpha0", &This::mAlpha0 );
  }
  
  /** Init buffers and accumulators. Also init emissions, which are then owned by this object. */
  void initialize() 
  {
    
    RERR( (This::size()==0) , ("Cannot initialize sequence model before number of states have been set") );
    
    numStates=This::size();
    
    // suff.stat. of the transition matrix, prior and var.post.
    mGamma1=mGamma2=mGamma1Var=mGamma2Var=ublas::scalar_matrix<floatT>(numStates,numStates,1.0);
    
    // expected log of transition matrix under variational posterior
    mLogTrans=ublas::scalar_matrix<floatT>(mNumStates,mNumStates,-log(numStates));
    
    // suff.stat of state prior, and var.post.
    mDelta1=mDelta2=mDelta1Var=mDelta2Var=ublas::scalar_vector<floatT>(numStates,1.0);
    
    // expected log of state prior under var. post.
    mLogStatePrior=ublas::scalar_vector<floatT>(numStates,-log(numStates));
   
    // log-likelihood of data up to (excluding) previous point in time
    mPreviousLogDataProb=0.0;
    // log-likelihoods of previous states and current states. Forward-probs of a HMM
    mPreviousLogProbs=mCurrentLogProbs=mCurrentLogLikelihoodData=std::vector<floatVecT(This::getNumSegments(),ublas::scalar_vector<floatT>(mNumStates,-log(mNumStates)));
    
    clearBuffers();
      
  }
  
  
  /** 
        \param add_pseudodata: amount of pseudodata to add to the prior to simulate an initial posterior. Useful for sampling and randomization of initial conditions
  */
  void setPrior( float add_pseudodata = 0 ) 
  {
        variationalModel<dataT,floatT>::setPrior(add_pseudodata);
	initialize();
	
	
	const int maxs=mNumStates-1;
	const floatT cg1=100.0; // Beta(x,cg1,cg2) should represent p(x=1) \approx 1
	const floatT cg2=0.01;
	
	// make earlier states more probable. Should break symmetry.
	for(int state=0;state<maxs;++state) {
	  mDelta1(state)=1.0;
	  mDelta2(state)=mAlpha0;
	}
	mDelta1(maxs)=cg1;
	mDelta2(maxs)=cg2;
	
	// Transition matrix stores transMat(to,from). Assuming row-major storage, this should lead to faster inference.
	// all transitions are extremely improbable...
	mGamma1=ublas::scalar_matrix<floatT>(mNumStates,mNumStates,cg2);
	mGamma2=ublas::scalar_matrix<floatT>(mNumStates,mNumStates,cg1);
	// this makes the stick-breaking prior (approximately) truncated at mNumStates, i.e last state is absorbing
	for(int state=0;state<mNumStates;++state) 
	{
	  mGamma1(maxs,state)=cg1;
	  mGamma2(maxs,state)=cg2;
	}
	
	  
	
	if(mPriorType=="forwardOnly" || mPriorType=="forwardOnlyCircular") {
	  // ...except for the transition from state->state+1
	  for(int state=0;state<mNumStates-1;++state) {
	    mGamma1(state+1,state)=cg1;
	    mGamma2(state+1,state)=cg2;
	  }
	  if(mPriorType=="forwardOnlyCircular") 
	  {
	    // wrap around from last state to first state
	    mGamma1(maxs,maxs)=cg2;
	    mGamma2(maxs,maxs)=cg1;
	    mGamma1(0,maxs)=cg1;
	    mGamma2(0,maxs)=cg2;
	  }
	} else if (mPriorType=="forward" || mPriorType=="forwardCircular") {
	  // ...except for transitions into same state or next one
	  for(int state=0;state<mNumStates-1;++state) {
	    mGamma1(state,state)=1.0;
	    mGamma2(state,state)=mAlpha0;
	    mGamma1(state+1,state)=cg1;
	    mGamma2(state+1,state)=cg2;
	  }
	  if(mPriorType=="forwardCircular") 
	  {
	    // wrap around from last state to first state
	    mGamma1(maxs,maxs)=cg1;
	    mGamma2(maxs,maxs)=cg2;
	    mGamma1(0,maxs)=mAlpha0;
	    mGamma2(0,maxs)=1.0;
	  } 
	} else if (mPriorType=="full") {
	  // A DP-like prior, but truncated at mNumStates
	  mGamma1=ublas::scalar_matrix<floatT>(mNumStates,mNumStates,1.0);
	  mGamma2=ublas::scalar_matrix<floatT>(mNumStates,mNumStates,mAlpha0);
	  // this makes the stick-breaking prior (approximately) truncated at mNumStates, i.e last state is absorbing
	  for(int state=0;state<mNumStates;++state) 
	  {
	    mGamma1(maxs,state)=cg1;
	    mGamma2(maxs,state)=cg2;
	  }	
	  
	} else if (mPriorType=="flat") {
	  // A flat Dirichlet prior with baseline mAlpha0
	  const floatT al=mAlpha0/mNumStates;
	  for(int state=0;state<mNumStates;++state)
	    for(int targetState=0;state<maxs;++targetState) 
	    {
	      mGamma1(targetState,state)=al;
	      mGamma2(targetState,state)=al*(maxs-targetState);
	    }
	} else throw std::runtime_error("Unknown prior type: "+mPriorType);
	
	mGamma1Var=mGamma1;
	mGamma2Var=mGamma2;
	mDelta1Var=mDelta1;
	mDelta2Var=mDelta2;
	
	precomputeExpensive();
	
  };
  
  
  /** Store Datapoint and parent-post for later posterior update with updateVariationalPosterior(). Copies data.*/
  void learn(const dataT & data ,floatT parentPost = 1.0) {
      if(!mbLearning) throw std::runtime_error("Call beginLearning() before trying to learn");
      doLearn(data,parentPost);
      pDataT ndp=pDataT( new dataT(data) );
      mData.push_back(ndp);
      
  }
  
  /** Store reference to datapoint and parent-post for later posterior update with updateVariationalPosterior(). Does not copy data. */
  void learn(const pDataT dpointer,floatT parentPost = 1.0) {
    if(!mbLearning) throw std::runtime_error("Call beginLearning() before trying to learn");
    doLearn(*dpointer,parentPost);
    mData.push_back(dpointer);  
  }
  
  /** Reset hidden variables to their prior values. */
  void resetToPrior() {
    mCurrentLogProbs=mLogStatePrior;
    mCurrentLogLikelihoodData=ublas::zero_vector<floatT>(mNumStates);
    this->mLogLikelihood=0.0;
    if(mbLearning && mPriorReset.size()<=mData.size()) mPriorReset.push_back(true); // remeber that the prior was reset, but only once.
  
  };
  
  /** Go back to previous state. Works only once (because only 1 previous state is buffered) */
  void rewind() 
  {
    // TODO
  }
  
private:
    int	mNumStates; //!< number of emission states
    int	mMaxBufSize; //!< maximal buffer size
      
protected:
 
  floatMatT mGamma1,mGamma2,mGamma1Var,mGamma2Var; //!< Parameters of stick-breaking priors on the transition matrix, both prior and variational posterior.
  floatMatT mTransMatAcc; //!< Accumulated pseudocounts on transition matrix
  floatMatT mLogTrans; //!< expected log(transition probs) under current variational posterior
  floatVecT mDelta1,mDelta2,mDelta1Var,mDelta2Var; //!< Parameters of stick-breaking priors on state prior, both prior and variational posterior.
  floatVecT mStatePriorAcc; //!< Accumulated pseudocounts on state prior
  floatVecT mLogStatePrior; //!< expected log(state prior probs) under current variational posterior
  
  std::vector<floatVecT> mPreviousLogProbs,mCurrentLogProbs; //!< previous and current forward-probabilities, i.e. log p(state(now-1),sequence up to (now-1)) and log p(state(now),sequence up to now). one per segment.
  std::vector<floatVecT> mCurrentLogLikelihoodData; //!< log-likelihoods of current data, one per segment.
  floatT mPreviousLogDataProb; //!< log p(sequences up to (now - 1)). log p(sequences up to now) = mLogLikelihood.
  
  
  probBufferT mForwardLogProbs; //!< buffered log-probabilities of p(state(t),sequence up to t) for buffered sequence. One per data segment
  probBufferT mLogLikelihoodData; //!< buffered log-likelihoods of the data, for use in backward pass.
  std::vector<floatT> mParentPosts; //!< parent posteriors for each datapoint.
  std::vector<bool> mPriorReset; //!< true if prior has been reset at corresponding time point in forward iteration. Backward pass has to be restarted at these points
  
  std::string mPriorType; //!< name of prior type. Currently implemented: forward,forwardCircular,forwardOnly,forwardOnlyCircular,full (a DP-like prior, but truncated), flat (a flat Dirichlet prior)
  floatT mAlpha0; //!< stick-breaking alpha for prior.
  
  
  /** 
  \return log of sum of exp(vector components). */
  inline floatT logAdd(const floatVecT & lp) const
  {
    floatT mv=*std::max_element(lp.begin(),lp.end());
    floatT sum=0.0;
    BOOST_FOREACH(const floatT &v,lp) sum+=exp(v-mv);
    return mv+log(sum);
  }
  
  /** 
  \return log of sum of exp(vector1+vector2). */
  inline floatT logAdd(const floatVecT & v1,const floatVecT & v2) const
  {
    const floatVecT vsum=v1+v2;
    floatT mv=*std::max_element(vsum.begin(),vsum.end());
    floatT sum=0.0;
    BOOST_FOREACH(const floatT &v,vsum) sum+=exp(v-mv);
    return mv+log(sum);
  }
  
  /** \param lp: vector with log probabilities to be normalized. Will contain probabilities after call. */
  inline void normalize(floatVecT &lp) const 
  {
    lp-=ublas::scalar_vector<floatT>(mNumStates,logAdd(lp));
    BOOST_FOREACH(floatT &v,lp) v=exp(v);
  }
  
  /** store res[j]=log( \sum_i exp(vector1[i]+vector2[i]+matrix(j,i) ). 
  \param rowwise if true, sum over rows, i.e. matrix(i,j). Else sum over columns, i.e. matrix(j,i)
  */
  inline void logAdd(const floatVecT & v1,const floatVecT & v2, floatMatT & tm, floatVecT & res, bool rowwise=true) const
  {
    assert(v1.size()==v2.size() && res.size()==v2.size() && tm.size1()==v2.size() && tm.size2()==v2.size());
    for(unsigned i=0;i<res.size();++i) {
      floatVecT vsum;
      if(rowwise) vsum=v1+v2+ublas::matrix_row<floatMatT>(tm,i);
      else vsum=v1+v2+ublas::matrix_column<floatMatT>(tm,i);
      floatT mv=*std::max_element(vsum.begin(),vsum.end());
      floatT sum=0.0;
      BOOST_FOREACH(const floatT &v,vsum) sum+=exp(v-mv);
      res[i]=mv+log(sum);
    }
  }
  
  /** clear learning buffers */
  void clearBuffers() 
  {
    // clear buffers for learning
    This::mBufData.clear();   
    mForwardLogProbs.clear();
    mLogLikelihoodData.clear();
    mPriorReset.clear();
    mParentPosts.clear();
    mTransMatAcc=ublas::zero_matrix<floatT>(mNumStates,mNumStates);
    mStatePriorAcc=ublas::zero_vector<floatT>(mNumStates);
  }

  
  /** pre-compute expected log transition probs and state prior probs. called whenever variational parameters change. */
  void precomputeExpensive()
  {  
     mLogTrans=ublas::zero_matrix<floatT>(mNumStates,mNumStates);
     mLogStatePrior=ublas::zero_vector<floatT>(mNumStates);
     
     floatT pacc=0.0;
     for(int s=0;s<mNumStates;++s) 
     {
       floatT acc=0.0;
       floatT psisum=0.0;
       for(int t=0;t<mNumStates;++t) 
       {
	  psisum=psi(mGamma1(t,s)+mGamma2(t,s));
	  mLogTrans(t,s)=acc+psi(mGamma1(t,s))-psisum;
	  acc+=psi(mGamma2(t,s))-psisum;
       }
       
       psisum=psi(mDelta1[s]+mDelta2[s]);
       mLogStatePrior[s]=pacc+psi(mDelta1[s])-psisum;
       pacc+=psi(mDelta2[s])-psisum;
     }
  }
  
  
  /** \return log-likelihood of supplied data, given recent past.  Does NOT advance the sequence.   */
  floatT computeLogLikelihood(const dataT & data, bool variational) 
  {
	for(int i=0;i<mNumStates;++i) mCurrentLogLikelihoodData[i]=mEmissions[i]->getLogLikelihood(data,variational);
	return logAdd(mCurrentLogLikelihoodData,mCurrentLogProbs);
  };
  
  /** \return KL-divergence between current variational distribution and prior. Implement in derived classes. */
  floatT computeKL() 
  {
        assert(false); // IMPLEMENT ME !!       
        return 0.0;
  };
  
  /** Reset accumulators to begin learning. Implement in derived classes. */
  void doBeginLearning() 
  {
	mbLearning=true;
	clearBuffers();
	resetToPrior();
  }
  
  /** Do inference (E-step) given parentPosterior. Advances the sequence.
  */
  void doInference(const dataT & data) 
  {
	// move old values to previous state
	mPreviousLogDataProb=this->mLogLikelihood;
	mPreviousLogProbs=mCurrentLogProbs;
	
	// do forward iteration, store in mCurrentLogProbs.
	for(int i=0;i<mNumStates;++i) mCurrentLogLikelihoodData[i]=mEmissions[i]->getLogLikelihood(data,variational);
	logAdd(mPreviousLogProbs,mCurrentLogLikelihoodData,mLogTrans,mCurrentLogProbs,true);
	
	this->mLogLikelihood=logAdd(mCurrentLogProbs);
  }
  
  /** Learning remebers forward-probabilities and whether states have been reset to prior.
  \param data: reference to current datapoint
  \param parentPost: parent posterior, e.g. responsibilities in a mixture model.
  */
  void doLearn(const dataT & data, const floatT parentPost) 
  {
	if(mPriorReset.size()<=mForwardLogProbs.size()) mPriorReset.push_back(false); // remember that the prior has not been reset, unless it has been reset.	 
	mForwardLogProbs.push_back(mPreviousLogProbs); // assuming infer has been called before learn, this stores log p(data0....data,states)
	mLogLikelihoodData.push_back(mCurrentLogLikelihoodData);
	mParentPosts.push_back(parentPost);
	if(mParentPosts.size()>unsigned(mMaxBufSize)) throw std::runtime_error( (format("Maximal learning buffer size of %d exceeded!") % mMaxBufSize).str() );
  };
  
  
  /** Update variational posterior with current accumulated sufficient statistics and prior (M-step). Implement in derived classes.
  \return true if parameters have been changed
  */
  bool doVariationalParameterUpdate()
  {
	mbLearning=false;
	if(std::accumulate(mParentPosts.begin(),mParentPosts.end(),0.0)<1e-3) return false;
	
	// do backward pass. update trans.mat and state prior. Call infer & learn for all emissions.
	if(mParentPosts.size() != mForwardLogProbs.size()) 
	  throw std::runtime_error( (format( "Learn has been called %d times, but infer only %d times. Unable to update") % mParentPosts.size() % mForwardLogProbs.size()).str() );
	
	const floatVecT ones=ublas::scalar_vector<floatT>(mNumStates,1.0);
	const floatVecT zeros=ublas::zero_vector<floatT>(mNumStates);
	
	floatVecT backwardLogProbs=zeros;
	
	for(int t=mParentPosts.size()-1;t>=0;--t) // travese backwards through the buffered data
	{
	  if(mParentPosts[t]<1e-3) continue; // nothing to accumulate if not responsible
	  floatVecT stateProbs=backwardLogProbs+mForwardLogProbs[t];
	  normalize(stateProbs);
	  stateProbs*=mParentPosts[t]; // weighting by responsibility
	  
	  if(mPriorReset[t]) 
	  {
	    mStatePriorAcc+=stateProbs; // accumulate state prior updates
	    backwardLogProbs=zeros; // restart backward pass
	  }
	  else
	  {
	    // accumulate transition matrix updates
	    floatMatT tmup=mLogTrans+ublas::outer_prod(backwardLogProbs+mLogLikelihoodData[t],ones)+ublas::outer_prod(ones,mForwardLogProbs[t]);
	    
	    for(int i=0;i<mNumStates;++i) 
	    {
	      floatT mv=-1e30;
	      floatT rowsum=0.0;
	      for(int j=0;j<mNumStates;++j) mv=std::max(mv,tmup(i,j)); // find maximal element
	      for(int j=0;j<mNumStates;++j) rowsum+=(tmup(i,j)=exp(tmup(i,j)-mv)); // compute normalization constant
	      for(int j=0;j<mNumStates;++j) tmup(i,j)/=rowsum; // normalize
	    }
	    tmup*=mParentPosts[t]; // weighting by responsibility
	    mTransMatAcc+=tmup;
	  }
	  
	  // accumulate in emissions
	  for(int state=0;state<mNumStates;++state) {
	    mEmissions[state]->inference(*mData[t]); // need to re-call infer for emissions, in case they have latent variables. This is not necessary if they are sequence_models and do their own buffering!!!!!
	    mEmissions[state]->learn(*mData[t],stateProbs[state]);
	  }
	  
	  // compute new backward probs, i.e. backward-pass in a HMM
	  floatVecT newBackwardLogProbs=ublas::zero_vector<floatT>(mNumStates);
	  logAdd(backwardLogProbs,mLogLikelihoodData[t],mLogTrans,newBackwardLogProbs,false);
	  backwardLogProbs=newBackwardLogProbs;
	  
	}
	
	
	BOOST_FOREACH(pEmissionT & em,mEmissions) em->updateParameterPosterior();
	return true;
  };
  
  /** Make current variational posterior the new prior. Completes learning */
  void doPosteriorToPrior() 
  {
        mGamma1=mGamma1Var;
	mGamma2=mGamma2Var;
	mDelta1=mDelta1Var;
	
  }
  
  /** \return predictive probability of supplied datapoint */
  virtual floatT doPredict(const dataT & datapoint) { return std::exp( getLogLikelihood(datapoint,true)) ; }
    
    
  
  /** write to stream.
          \param os:  reference to ostream to write into.
  */
  void writeToStream(ostream &os) const 
  {
      os<<"Sequence model"<<endl;
      os<<"Number of states: "<<mNumStates<<endl;
      os<<"Prior type: "<<mPriorType<<endl;
      os<<"Prior alpha0: "<<mAlpha0<<endl; 
      os<<"Maximal buffer size: "<<mMaxBufSize<<endl;
      os<<"State prior exponents"<<endl;
      os<<mDelta1;
      os<<mDelta2;
      os<<"Variational state posterior exponents"<<endl;
      os<<mDelta1Var;
      os<<mDelta2Var;
      os<<"Expected log(state prior) "<<endl;
      os<<mLogStatePrior;
      os<<"Transition matrix prior exponents"<<endl;
      os<<mGamma1;
      os<<mGamma2;
      os<<"Variational transition matrix posterior exponents"<<endl;
      os<<mGamma1Var;
      os<<mGamma2Var;
      os<<"Expected log(transition matrix)"<<endl;
      os<<mLogTrans;
      os<<"Emission models:"<<endl;
      BOOST_FOREACH(const pEmissionT &pem,mEmissions) os<<*pem;  

    
  };
  
  /** read from stream is.
          \param is : reference to stream to read from.
  */
  void readFromStream(istream &is) 
  {
      string name; 
      std::getline(is,name);
      if(name!="Sequence model") throw std::runtime_error("Error reading sequence model: got "+name);
      is>>eatString(3)>>mNumStates>>ws;
      is>>eatString(2)>>mPriorType>>ws;
      is>>eatString(2)>>mAlpha0>>ws;
      is>>eatString(3)>>mMaxBufSize>>ws;
      std::getline(is,name);
      is>>mDelta1;
      is>>mDelta2;
      std::getline(is,name);
      is>>mDelta1Var;
      is>>mDelta2Var;
      std::getline(is,name);
      is>>mLogStatePrior;
      std::getline(is,name);
      is>>mGamma1;
      is>>mGamma2;
      std::getline(is,name);
      is>>mGamma1Var;
      is>>mGamma2Var;
      std::getline(is,name);
      is>>mLogTrans;
      std::getline(is,name);
      for(int i=0;i<mNumStates;++i) 
      {
	pEmissionT pem=pEmissionT( new emission_type() );
	is>>*pem;
	mEmissions.push_back(pem);
      }

      mPreviousLogProbs=ublas::zero_vector<floatT>(mNumStates);
      mPreviousLogDataProb=0.0;

      mbLearning=false;
      clearBuffers();
      resetToPrior();
      precomputeExpensive();

  }
  
  /** check if all points in mData have the correct shape. needs to call emission model? */
  bool checkDataShape(const dataT & data) {return true;};

};
