/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/

# pragma once
#include "varutils.h"
#include <boost/thread/mutex.hpp>
#include <boost/thread/locks.hpp>

namespace ublas = boost::numeric::ublas;
namespace bmath = boost::math;

using namespace std;

template<class data_type,class float_type=float, class data_handle = boost::shared_ptr<data_type> > class variationalModel;

/** Base class for variational models. Derive emission models, mixtures, HMMs etc. from this class.
  Keeps track of whether computed values for likelihoods, KL-divergences etc. are still up-to-date.
  Does not maintain data.
  Template parameters: dataType: type of data digested by this model
  floatType: type of float used for internal computations. Default float.

  the concept is like this:
  
  All data have to be wrapped in boost::shared_ptr. data.use_count()==0 indicates that no data have been observed.

  there are effectively three parameter sets:

  the base prior, before any data has been seen
  one "prior" data dependent parameter set
  one "posterior" data dependent parameter set

  calling setPrior selects the base prior, and resets the other two
  optionally initialises the data dependent prior with some base-prior-drawn or base-prior-expected pseudodata

  calling infer infers the hidden state given the supplied data
  calling learn updates the sufficient statistics cumulant, given supplied data and current internal states of hidden variables
  calling beginLearning resets the sufficient statistics cumulant
  calling updateParameterPosterior updates the posterior from the base prior, the data prior, and the sufficient statistics cumulant
  calling posteriorToPrior makes the current data posterior the data prior

  this leaves a lot of learning options:

  batch learning:

  setPrior
  until converged or number-of-EM-steps > max-number-of-EM-steps:
    beginlearning
    for datapoint in data:
      infer(datapoint)
      learn(datapoint)
    updateParameterPosterior
  posteriorToPrior

  onlinelearning:

  setPrior
  while there is a new datapoint :
    until converged or number-of-EM-steps > max-number-of-EM-steps:
      beginlearning
      inference(datapoint)
      learn(datapoint)
      updateParameterPosterior
    posteriorToPrior
 
  onlinelearning with minibatches and/or several em-steps:

  setPrior
  for minibatch in batches:
    until converged or number-of-EM-steps > max-number-of-EM-steps:
      beginlearning
      for datapoint in minibatch:
        inference(datapoint)
        learn(datapoint)
      updateParameterPosterior
    posteriorToPrior
    
    
  Furthermore, learn() accepts a 2nd parameter, the parent's posterior probability. This is useful for hierarchical modelling.
  Currently, only tree-structured models can be build this way. During infer(), the parents have to pass the data down to their children,
  observation models for data have to be at leaf nodes. Children compute their likelihoods, which is queried by the parents etc.
  During learn(), parents pass data and posteriors (responsibilities for data) down to the children, every node performs a suff.stat.cumulant update.
  inference(datapoint) also computes & stores log-likelihood of data. Can then be queried with getLogLikelihood or log_likelihood(). 
  Calling inference(datapoint) before requesting the likelihood is necessary if the inference in the model is not exact, because the log-likelihood is then an expectation
  under the variational posterior of the hidden variables. If inference is exact, log-likelihood is the normalization constant of the posterior and needs to be
  computed in inference(datapoint) as well. So there will be no significant overhead by calling inference(datapoint) first.
  When inference() is called without data, it is assumed that the data are unobserved.
  To pass parent posteriors to children without learning (e.g. for prediction), call predictPass(float parentPost).
  
  For sequence learning, the model can buffer data via bufferData().
  Data are buffered with bufferData(datapoint,resetToPrior). If resetToPrior is true, call resetToPrior() before datapoint during inference & learning.
  This effectively starts a new sequence. In this way, the buffer can contain several sequences for batch-like learning.
  
 
  sequence-learning with minibatches and/or several em-steps the looks like:

  setPrior()
  clearBuffer()
  for datapoint,resetPoint in (sequence_data,resetPoints):
    bufferData(datapoint,resetPoint)
    
  until converged or number-of-EM-steps > max-number-of-EM-steps:
    beginLearning()
    inferBuffer()
    learnBuffer(parentPost)
    updateParameterPosterior()
  posteriorToPrior()
  
  if you are worried about wasting time during the initial buffering of the data, you can also
    
  setPrior()
  clearBuffer()
  beginLearning()
  for datapoint,resetPoint in (sequence_data,resetPoints):
    if(resetPoint) resetToPrior()
    inference(datapoint)
    bufferData(datapoint,resetPoint)
  learnBuffer(parentPost)
  updateParameterPosterior() # complete the 1st EM steps
  
  until converged or number-of-EM-steps > max-number-of-EM-steps:
    beginLearning()
    inferBuffer()
    learnBuffer(parentPost)
    updateParameterPosterior()
  posteriorToPrior()
  
  
 A model can also share its parameters with another instance of its type. Parameter sharing is accomplished through copy-construction. If you want to make a sharing instances
 "private", call unshare(). Hidden variables and (local) accumulators of sufficient statistics are not shared, i.e. inference and learning can be parallelized.
  
 
*/
template<class data_type,class float_type, class data_handle>
class variationalModel 
{

public:

  typedef data_type dataT;
  typedef data_handle pdataT;
  typedef float_type floatT;
  typedef ublas::vector<float_type> floatVecT;
  typedef ublas::matrix<float_type> floatMatT;  
  typedef std::vector<int> intVecT;

  typedef typename std::vector<pdataT> dataBufferT;
  typedef boost::shared_ptr<int> pintT;
  
  typedef variationalModel<data_type,float_type, data_handle> This;
  typedef variationalModel<data_type,float_type, data_handle> thisT; // do we need both?
  typedef std::vector<This*> thisVecT;
  typedef boost::shared_ptr<thisVecT> pthisVecT;
  typedef variationalModel<data_type,float_type, data_handle> varmodelBaseT; // for upcasting in virtual methods that take a this-pointer as a parameter

  typedef boost::mutex lockT;
  typedef boost::shared_ptr<lockT> pLockT;
  typedef boost::lock_guard<lockT> guardT;
  
  typedef enum {NONE, FORWARD, BACKWARD} seqPhaseT;
  

  
  variationalModel() 
    : mLogLikelihood(0.0)
    , mParentPost(1.0)
    , mAccumulatedVarBound(0.0)
    , mbAccumulateVarBound(false)
    , mbNeedUpdate(false)
    , mMaxBufSize(1024)
    , mBufData()
    , mSequenceLearningPhase(0)
    , mbResetPoint()
    , mIndex(-1)
    , mSegment(-1)
    , mFreezeLevel(0)
    
  {
    mpUpdateLock=pLockT( new lockT() );
    mpNeighbours=pthisVecT( new thisVecT() );
    mpNeighbours->push_back(this);
  };
  
  /** On copy construction, share parameters with source. Index and Segment need to be set afterwards (if segmented and if component in a container) */
  variationalModel(const This &source) {
    mLogLikelihood=source.mLogLikelihood;
    mParentPost=source.mParentPost;
    mAccumulatedVarBound=source.mAccumulatedVarBound;
    mbAccumulateVarBound=source.mbAccumulateVarBound;
    mMaxBufSize=source.mMaxBufSize;
    mBufData.clear();
    mbResetPoint.clear();
    mSequenceLearningPhase=0;
    
    mIndex=source.mIndex;
    mSegment=source.mSegment; 
    mFreezeLevel=source.mFreezeLevel;
    mbNeedUpdate=source.mbNeedUpdate;
    
    doShareWith(source,false);
  }
  
  This & operator=(const This & source) {
    mLogLikelihood=source.mLogLikelihood;
    mParentPost=source.mParentPost;
    mAccumulatedVarBound=source.mAccumulatedVarBound;
    mbAccumulateVarBound=source.mbAccumulateVarBound;
    mMaxBufSize=source.mMaxBufSize;
    mBufData.clear();
    mbResetPoint.clear();
    mSequenceLearningPhase=0;
    
    mIndex=source.mIndex;
    mSegment=source.mSegment; 
    mFreezeLevel=source.mFreezeLevel;
    mbNeedUpdate=source.mbNeedUpdate;
    
    doShareWith(source,true);
    return *this;
  }
  
  
  /** Make private copies of all parameters. */
  virtual void unshare() 
  {
    if( !shared() ) return; // non-shared already
    {
      guardT g( *mpUpdateLock );
      typename thisVecT::iterator pself=find(mpNeighbours->begin(),mpNeighbours->end(),this);
      RERR( (pself==mpNeighbours->end()), "Error unsharing variationalModel: cannot find self in mpNeighbours");
      mpNeighbours->erase(pself); 
    }
    mpUpdateLock=pLockT( new lockT() );
    mpNeighbours=pthisVecT(new thisVecT());
    mpNeighbours->push_back(this);
  };
  


  /** Implement to set prior parameters.
  \param add_pseudodata: amount of pseudodata to add to the prior to simulate an initial posterior. Useful for sampling and randomization of initial conditions
  */
  virtual void setPrior( float add_pseudodata = 0 ) {};
  
  
  /** init accumulators */
  virtual void initialize() {};
  
  
  /** \return log-likelihood of data which were used in last inference(data) call.
  */
  floatT getLogLikelihood() {
    return mLogLikelihood;
  }
  
  
  /** start learning from scratch, i.e. reset any accumulators from previous learn steps, does not update the parameter distribution 
  Clear data buffer for sequence learning.
  \param accumulateVarBound: if true, accumulate var bound during this learning run
  **/
  void beginLearning(bool accumulateVarBound=false)
  {
    resetVarBound();
    mLogLikelihood=0.0;
    mbAccumulateVarBound=accumulateVarBound;
    if(haveParametersChanged()) updateExpensiveForAll();
    resetToPrior();
    guardT g( *mpUpdateLock );
    doBeginLearning();
  }
  
  /** freeze some or all parameters during learning. Useful for layer-wise training etc. Not hierachical!
  \param level: freeze-level. 0 (default) should mean no parameters are frozen
  */
  
  void freeze(int level=0) 
  {
    mFreezeLevel=level;
  }
  
  /** freeze hierarchical model to given level */
  virtual void freezeAll(int level=0)
  {
    freeze(level);
  }
  
        
  /** Infer state of hidden variables, given current data. E-step of an EM-algorithm, leaves-to-roots pass of a belief propagation algorithm.
  \param data_ptr: boost::shared_ptr to the data
  \param variational: if true (default), compute variational expected log-likelihood. otherwise, compute log-likelihood.
  */
  void inference(const pdataT & data_ptr,bool variational=true) 
  {
    checkDataShape(data_ptr);
    if(haveParametersChanged()) updateExpensiveForAll();
    mLogLikelihood=doInference(data_ptr,variational);
  }
 
  /** Set posterior responsibilities for data, which are computed by the parents. roots-to-leaves pass of a belief propagation algorithm. */
  virtual void predictPass(const floatT parentPost=1.0)
  {
    mParentPost=parentPost;
  }
 
  /** \return predicitive log-probability of pdataT based on current responsibility */ 
  floatT predict(const pdataT & data_ptr)
  {
    if(haveParametersChanged()) updateExpensiveForAll();
    return doInference(data_ptr,true)+log(mParentPost<1e-100?1e-100:mParentPost);
  }
  
  /** \return posterior responsibility received from parent in last predictPass */
  floatT getParentPost() {return mParentPost;};
  
  
  /** Do variational parameter update based on current data and parent posterior probabilities given these data. variational M-step.
  \param parentPost: posterior of parent nodes in a hierarchical model. For example, responsibilities for mixture components.
  \param data_ptr: boost::shared_ptr to the data
  */
  void learn(const pdataT & data_ptr,floatT parentPost = 1.0)
  {
    checkDataShape(data_ptr);
    if(haveParametersChanged()) updateExpensiveForAll();;
    if(mbAccumulateVarBound) accumulateVarBound(parentPost);
    predictPass(parentPost);
    doLearn(data_ptr,parentPost);
  }
  
  /** Clear data buffer used for sequence learning */
  void clearBuffer()
  {
   mBufData.clear();
   mbResetPoint.clear();
   resetToPrior();
  }
  
  /** Buffer data for sequence learning.
  \param data_ptr: boost::shared_ptr to the data
  \param resetBefore: if true, reset hidden vars to prior before data point (i.e. start new sequence)
  */
  void bufferData(const pdataT & data_ptr,bool resetBefore=false)
  {
    checkDataShape(data_ptr);
    mBufData.push_back(data_ptr);
    mbResetPoint.push_back(resetBefore);
  }
  
  /** Call inference() on all data in the buffer. Useful for forward-iteration over buffered sequences.
  \param variational: if true, compute variational expecatations.
  */
  void inferBuffer(bool variational=true)
  {
    setSequenceLearningPhase(FORWARD);
    for(unsigned pt=0;pt<mBufData.size();++pt)
    {
      if(mbResetPoint[pt]) resetToPrior();
      inference(mBufData[pt],variational);
    }
    setSequenceLearningPhase(NONE);
  }
  
  /** Call learn() on all data in the buffer in reverse order. Backward iteration over buffered sequences. Assumes that inferBuffer has been called just before.
  \param parentPost: parent posterior prob (responsibility) for all data in the buffer.
  */
  void learnBuffer(floatT parentPost=1.0)
  {
    setSequenceLearningPhase(BACKWARD);
    for(int pt=(int)mBufData.size()-1;pt>=0;--pt)
    {
      learn(mBufData[pt],parentPost);
      if(mbResetPoint[pt]) resetToPrior();
      if(pt>0) inference(mBufData[pt-1]);
    }    
    setSequenceLearningPhase(NONE);
  }
  
  /** Set sequence learning phase. 0: non-sequential learning (i.e. datapoints are not connected in time), 1: forward inference pass, 2: backward inference pass & learning */
  virtual void setSequenceLearningPhase(int seqlearnPhase=0)
  {
    mSequenceLearningPhase=seqlearnPhase;
  }
  
  
  /** use the information from the learn steps to update the parameter distribution, this completes an M-step
  **/
  void updateParameterPosterior() 
  {
    guardT g( *mpUpdateLock );
    if(doVariationalParameterUpdate()) {
      if( !shared() ) precomputeExpensive();
      else {
		BOOST_FOREACH(This * neighbour,*mpNeighbours) // tell all neighbours that shared params have changed
		  {
			neighbour->notifyParametersChanged(); 
		  }
      }	  
    }
  }	
  
  bool shared() const { return mpUpdateLock.use_count() > 1; }
  
  /** \return a sample from this model given current variational posteriors */
  virtual pdataT sample() {return pdataT( );};
  
  /** \return expectation of data given current variational posteriors */
  virtual pdataT getExpectation() {return pdataT( );};

  /** forget current prior and make current variational posterior the new prior. Call base class method in derived classes. 
  If parallelized, make sure only one thread in a sharing group does the update at any given time.**/
  void posteriorToPrior() 
  {
    mLogLikelihood=0.0;
    resetVarBound();
  }
  
  
  /** for online learning with fogetting **/
  virtual void decayParameters(floatT tau, bool update_immediately = true) {}
  
  /** \return KL-divergence between prior and current variational param posterior */
  virtual floatT getKL() 
  {
    return 0.0;
  }
  
  /** \return conditional entropy contribution */
  virtual floatT getEntropy() {return 0.0;}
  
  /** \return expectation contribution of log^2(p) for conditional variance of conditional entropy */
  virtual floatT getEntropy2() {return 0.0;}


  /** \return variational lower bound on marginal log likelihood of data seen since last call to beginLearning() */
  virtual floatT getVarBound()
  {
    return getAccumulatedVarBound() - getKL();
  }
 
  
  /** Reset accumulated variational bound */
  void resetVarBound() {mAccumulatedVarBound=0.0;};  
  
  
  /** \return accumulated hidden-variable and likelihood part of variational bound. Subtract getKL() to get bound. */
  floatT getAccumulatedVarBound() {return mAccumulatedVarBound;};
  
  /** write to stream.
  \param os  reference to ostream to write into.
  \param varmod  reference to variationalModel to write
  \return ostream after varmod has been inserted
  */
  friend ostream &operator<<(ostream &os,const This &varmod) 
  {
    varmod.writeToStream(os);
    return os;
  }
  
  /** read from stream is.
          \param is  reference to stream to read from.
          \param varmod reference to variationalModel to read into
          \return stream is after variationalModel has been read.
  */
  friend istream &operator>>(istream &is, This &varmod) 
  {
    varmod.readFromStream(is);
    varmod.mSequenceLearningPhase=0;
    varmod.clearBuffer();
    varmod.mbNeedUpdate=false;
    return is;
  }
  
  
  /** Reset hidden variables to their prior values. Reset likelihood. Call base class method in derived classes. */
  virtual void resetToPrior() 
  {
    mLogLikelihood=0.0;
    mParentPost=0.0;
  };
  
  /** Reset variational parameter posterior to prior. Call before updating parameter posterior */
  void resetParameterPosteriorToPrior()
  {
    guardT g( *mpUpdateLock );
    doResetParameterPosteriorToPrior();
    BOOST_FOREACH(This * neighbour,*mpNeighbours) // tell all neighbours that shared params have changed
    {
      neighbour->notifyParametersChanged(); 
    }
  }
  
  
  friend bool operator==(const This & first,const This & second) 
  {
    return cmp(first,second);
  }
  
  static bool cmp(const This & first, const This &second)
  {
    return (first.mIndex==second.mIndex) && (first.mSegment==second.mSegment) && (first.segments()==second.segments());
  }
  
  virtual ~variationalModel()
  { 
    // remove myself from notification list
    {
      guardT g( *mpUpdateLock );
      typename thisVecT::iterator pself=find(mpNeighbours->begin(),mpNeighbours->end(),this);
      RERR( (pself==mpNeighbours->end()), "Error destroying variationalModel: cannot find self in mpNeighbours");
      mpNeighbours->erase(pself); 
    }  
  }
  
  // Methods for segmented data and models with components
  
   /** Set index of component */
  void setIndex(int index) {mIndex=index;};
  
  /** Get index of component */
  int getIndex() const {return mIndex;};
  
  /** Set segment of component */
  void setSegment(int segment) {mSegment=segment;};
  
  /** Get segment of component */
  int getSegment() const {return mSegment;};
  
  /** \return number of segments. overwrite in derived classes. */
  virtual int segments() const {return 1;};
  
  /** Get data segment. Implement in derived classes if working on segmented data (e.g. image patches) */
  virtual pdataT getDataSegment(const pdataT &data_ptr,unsigned segment=0) const {return data_ptr;};
  
  /** \return a new datapoint, initalized to the correct size */
  //virtual pdataT newData() {pdataT nd=pdataT( new dataT() ); return nd;};
  
  // Methods for segmented data and models with components END
  
  virtual void updateAuxiliaryVariables() {precomputeExpensive();};
  
    
protected:
  
  
  floatT mLogLikelihood; //!< log-likelihood for current datapoint
  floatT mParentPost; //!< posterior probability of this model, computed usually by parent. E.g. responsibility in a mixture.
  floatT mAccumulatedVarBound; //!< accumulated part of var bound for latent variables (i.e. one per datapoint)
  bool mbAccumulateVarBound; //!< if true, accumulate varational bound during learning. incurs significant computations.
  
  pLockT mpUpdateLock; //!< for concurrent parameter update synchronisation
  lockT mChangeLock; //!< for concurrent access to mNeedUpdate
  bool mbNeedUpdate; //!< true if current parameters need update via call to precomputeExpensive()
  pthisVecT mpNeighbours; //!< pointers to copies which share parameters 
  
  
  
  unsigned mMaxBufSize; //!< maximal buffer size.
  dataBufferT mBufData; //!< buffered data for sequence learning
  int mSequenceLearningPhase; //!< 0: non-sequential learning (i.e. datapoints are not connected in time), 1: forward inference pass, 2: backward inference & learn pass.
  std::vector<bool> mbResetPoint; //!< true if resetToPrior() was called before a datapoint was buffered.
  
  
  int mIndex; //!< index of component in its parent container (if it has one)
  int mSegment; //!< segment of data which this instance is responsible for (if data are segmented, e.g. image in patches or so)
  int mFreezeLevel; //!< level of parameter-freezing during learning.
  
  
  /** pre-compute some expensive quantities, e.g. matrix inverses or so.*/
  virtual void precomputeExpensive()
  {
     // IMPLEMENT ME!!
  }
  
  /** copy expensive-to-precompute local variables (like inverses of large matrices) from source */
  virtual void copyExpensive(const varmodelBaseT & source)
  {
    // IMPLEMENT ME!!!
  }
  
  /** do expensive precomputes, if necessary. Copy them to all instances which share with this instance
  \param always: if true, do it. if not true, do only if my parameters are still not uptodate after I acquire the update lock.
  */
  void updateExpensiveForAll(bool always=false) {
    guardT g( *mpUpdateLock );
    if(!always && !haveParametersChanged()) return; // someone else has done it while I waited for the update lock
    precomputeExpensive();
    notifyParametersChanged(false);
    
    BOOST_FOREACH(This * neighbour,*mpNeighbours)
    {
      if(neighbour!=this) neighbour->copyExpensive(*this); // copy to neighbour
      neighbour->notifyParametersChanged(false); // tell neighbour that it's up-to-date now
    }
  }
  
   /** Share parameters with the source model. Overwrite in derived classes, and call base class method. */
  virtual void doShareWith(const This &source, bool shareBase=false) 
  {
    mpUpdateLock=source.mpUpdateLock;
    mpNeighbours=source.mpNeighbours;
    {
      guardT g( *mpUpdateLock );
      typename thisVecT::iterator pself=find(mpNeighbours->begin(),mpNeighbours->end(),this);
      if(pself==mpNeighbours->end()) {
	mpNeighbours->push_back(this); 
      } 
    }
  };
  
  /** notification of parameter change */
  void notifyParametersChanged(bool parChange=true)
  {
    mChangeLock.lock();
    mbNeedUpdate=parChange;
    mChangeLock.unlock();
  }
  
  /** update expensive-to-precompute variables for self and copy to all sharing instances */
  
  
  /** check if parameters have changed. Reset mbNeedUpdate=false. */
  bool haveParametersChanged() 
  {
    mChangeLock.lock();
    bool rv=mbNeedUpdate;
    mChangeLock.unlock();
    return rv;
  }
  
  
  
  
  
  /** Reset accumulators to begin learning. Implement in derived classes. */
  virtual void doBeginLearning() 
  {
	throw std::runtime_error( "CALLED unimplemented doBeginLearning()" );
    //assert(false); // IMPLEMENT ME!!
  }
  
  /** Do inference (E-step) given parentPosterior. Compute log-likelihood. Implement in derived classes.
  \param data_ptr: shared_ptr to data
  \param variational: whether to compute variational expectation.
  \return log-likelihood, which is almost a by-product of the inference step.
  */
  virtual floatT doInference( const pdataT & data_ptr,bool variational = false )
  {
	throw std::runtime_error( "CALLED unimplemented doInference()" );
    return 0.0;
  }
  
    /** Accumulate likelihood and hidden variable part of the variational bound, based on data used for last inference() call.
  Only needs to be implemented if var bound computation will be done.
  \param parentPost: responsibility for this datapoint
  */
  virtual void accumulateVarBound(floatT parentPost=1.0) 
  {
    mAccumulatedVarBound=0.0;
  }

  
  

  
  /** Do learning (accumulation of sufficient statistics) given parentPosterior. Implement in derived classes.
  \param data: reference to current datapoint
  \param parentPost: parent posterior, e.g. responsibilities in a mixture model.
  */
  virtual void doLearn(const pdataT & data_ptr, const floatT parentPost) 
  {
    cerr<<"Calling unimplemented variationalModel::doLearn()"<<endl; // IMPLEMENT ME
  };
  
  
  /** Update variational posterior with current accumulated sufficient statistics and prior (M-step). Implement in derived classes.
  \return true if parameters have been changed.
  */
  virtual bool doVariationalParameterUpdate()
  {
    assert(false); // IMPLEMENT ME!!!
    return false;
  };
  
  /** Reset variational parameter posterior to prior. Implement in derived classes */
  virtual void doResetParameterPosteriorToPrior()
  {
   RERR( (true), "Error in variationalModel: called unimplemented  doResetParameterPosteriorToPrior()");
  };
  
  
  /** write to stream. Implement in derived classes.
          \param os:  reference to ostream to write into.
  */
  virtual void writeToStream(ostream &os) const 
  {
    os<<"Variational model base class"<<endl;
    os<<"Index: "<<mIndex<<endl;
    os<<"Segment: "<<mSegment<<endl;    
    os<<"Freeze level: "<<mFreezeLevel<<endl;
    os<<"Number of segments (info only):"<<segments()<<endl;
    
  };
  
  /** read from stream is. Implement in derived classes.
          \param is : reference to stream to read from.
  */
  virtual void readFromStream(istream &is) 
  {
    string dummy; 
    getline(is,dummy);
    RERR( (dummy!="Variational model base class"), (boost::format("Error loading variationalModel base class: got %s") % dummy).str() );
    is>>eatString(1)>>mIndex>>ws;
    is>>eatString(1)>>mSegment>>ws;
    is>>eatString(2)>>mFreezeLevel>>ws;
    getline(is,dummy);
    mLogLikelihood=mParentPost=0.0;
    mAccumulatedVarBound=0.0;
    
  }
  
  /** check if all points in mData have the correct shape. Implement in derived classes. */
  virtual bool checkDataShape(const pdataT & data_ptr) const {return true;};
  
};

