/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/

#pragma once


#include <stdlib.h>
#include "math.h"
#include <cassert>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/foreach.hpp>
#include <boost/math/special_functions/gamma.hpp>
#include <boost/math/special_functions/digamma.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>
#include <algorithm>
#include <limits>
#include <iostream>
#include <vector>
#include <boost/concept_check.hpp>

#ifndef NDEBUG
#define RERR( cond, msg) { if(cond) throw std::runtime_error(msg);}
#else
#define RERR(msg) {}
#endif


namespace ublas = boost::numeric::ublas;
namespace bmath = boost::math;

using namespace std;

static float TINY=1e-30;

const double M_LNPI=log(M_PI);


/** underflow-save p*log(p) for entropy calculation */
inline float plogp(float p) { return p<1e-30?0.0:p*log(p); };
inline double plogp(double p) { return p<1e-30?0.0:p*log(p); };

/** Logarithm of the multivatriate gamma function of x with k degrees of freedom. */
inline double logMultiGamma(double x,unsigned k) {
  double rv=k*(k-1)/4.0*M_LNPI;
  for(unsigned i=0;i<k;++i) rv+=bmath::lgamma(0.5*(x-i));
  return rv;
};


/** Derivative (w.r.t. x) of the logarithm of the multivatriate gamma function of x with k degrees of freedom. */
inline double dLogMultiGamma(double x,unsigned k) {
  double rv=0.0;
  for(unsigned i=0;i<k;++i) rv+=bmath::digamma(0.5*(x-i));
  return rv*0.5;
}       

/** Sum ublas matrix */
template<class value_type>
value_type sum(const ublas::matrix<value_type> &mat) 
{
  value_type rv=0.0;
  for(unsigned i=0;i<mat.size1();++i) rv+=sum(ublas::matrix_row<const ublas::matrix<value_type> >(mat,i));
  return rv;
};



/** Consume strings from an istream */
class eatString {
public:
  unsigned numStrings; //!< number of strings to consume
  eatString(unsigned ns=1):numStrings(ns) {};

  friend istream & operator>>(istream &is,const eatString &es) {
    string dummy;
    for(unsigned i=0;i<es.numStrings;++i) is>>dummy>>ws;
    return is;
  };
};

/** Write ublas::vector to ostream */
template<class float_type>
ostream & operator<<(ostream &os,const ublas::vector<float_type> &vec) {
  os<<vec.size();
  for(unsigned i=0;i<vec.size();++i) os<<" "<<vec[i];
  os<<endl;
  return os;
};

/** Read ublas::vector from istream */
template<class float_type>
istream & operator>>(istream &is,ublas::vector<float_type> &vec) {
  int dim;
  is>>dim>>ws;
  vec.resize(dim);
  for(int i=0;i<dim;++i) is>>vec[i]>>ws;
  return is;
};


template<class float_type>
bool equalp(const float_type &x,const float_type &y,const float_type cmp_eps=1e-6) {
  return fabs(x-y)<max(TINY,max(fabs(x),fabs(y)))*cmp_eps;
}



/** comparsion function for ublas::vector with limitable precision */
template<class float_type>
bool equalp(const ublas::vector<float_type> &vec1,const ublas::vector<float_type> &vec2,const float_type cmp_eps=1e-6) {
  if(vec1.size()!=vec2.size()) return false;
  float_type maxval=TINY;
  for(unsigned i=0;i<vec1.size();++i) maxval=max(maxval,max(fabs(vec1[i]),fabs(vec2[i])));
  maxval*=cmp_eps;
  for(unsigned i=0;i<vec1.size();++i)
    if(fabs(vec1[i]-vec2[i])>=maxval) return false;
  return true;
}


/** Comparison op for ublas::vector */
template<class float_type>
bool operator==(const ublas::vector<float_type> &vec1,const ublas::vector<float_type> &vec2) {
  return equalp(vec1,vec2);
}


/** Write ublas::matrix to ostream */
template<class float_type>
ostream & operator<<(ostream &os,const ublas::matrix<float_type> &mat) {
  os<<mat.size1()<<" "<<mat.size2()<<endl;
  for(unsigned i=0;i<mat.size1();++i) {
    for(unsigned j=0;j<mat.size2();++j)
      os<<mat(i,j)<<" ";
    os<<endl;
  };
  return os;
};

/** Read ublas::matrix from istream */
template<class float_type>
istream & operator>>(istream &is,ublas::matrix<float_type> &mat) {
  int dim1,dim2;
  is>>dim1>>ws>>dim2>>ws;
  mat.resize(dim1,dim2);
  for(int i=0;i<dim1;++i) 
    for(int j=0;j<dim2;++j) is>>mat(i,j)>>ws;
  return is;
};

/** comparison func. for ublas::matrix with limitable precision */
template<class float_type>
bool equalp(const ublas::matrix<float_type> &mat1,const ublas::matrix<float_type> &mat2,const float_type cmp_eps=1e-6) {
  if(mat1.size1()!=mat2.size1() || mat1.size2()!=mat2.size2()) return false;
  float_type maxval=TINY;
  for(unsigned i=0;i<mat1.size1();++i)
    for(unsigned j=0;j<mat1.size2();++j)
      maxval=max(maxval,max(fabs(mat1(i,j)),fabs(mat2(i,j))));
  maxval*=cmp_eps;
  for(unsigned i=0;i<mat1.size1();++i)
    for(unsigned j=0;j<mat1.size2();++j)
      if(fabs(mat1(i,j)-mat2(i,j))>maxval) return false;
       
    return true;
}



/** Comparison op for ublas::matrix */
template<class float_type>
bool operator==(const ublas::matrix<float_type> &mat1,const ublas::matrix<float_type> &mat2) {
  return equalp(mat1,mat2);
}


        

/** Exponentiate and normalize every row of matrix logp. Type of logp should be ublas::matrix.
\return exponentiated and row-normalized version of logp
*/
template<class M>
M expNorm( const M &logp )
{
  M retval=logp;
  for(unsigned r=0;r<logp.size1();++r) 
  {
    ublas::matrix_row<const M> crow(retval,r);
    if(crow.size()==0) continue;
    const double maxcr=*max_element(crow.begin(),crow.end());
    double sum=numeric_limits<typename M::value_type>::min();
    for(unsigned i=0;i<crow.size();++i)  sum+=(crow[i]=exp(crow[i]-maxcr));
    crow/=sum;
  }
  return retval;
 }
 
/** Exponentiate and normalize vector logp.
\param logp: reference to vector of log-probabilities.
\return log of sum of components, i.e. normalization constant
*/
template<class vec>
inline typename vec::value_type normalize(vec & logp) 
{
  if(logp.size()==0) return 0.0;
  typedef typename vec::value_type floatT;
  floatT maxcomp=*max_element(logp.begin(),logp.end());
  floatT sum=0.0;
  BOOST_FOREACH(floatT &vc,logp) 
	sum += ( vc = exp( vc - maxcomp ) );
  logp /= sum;
  return log( max( sum, numeric_limits<floatT>::min() ) ) + maxcomp;
}

/** Exponentiate vector logp */
template<typename float_type>
inline ublas::vector<float_type> exp(const ublas::vector<float_type> &v) {
  ublas::vector<float_type> retval=v;
  BOOST_FOREACH(float_type &vv,retval) vv=exp(vv);
  return retval;
}

/** Compute \sum p*log(p) for vector */
template<class vec>
inline typename vec::value_type plogp(const vec &v) {
  typename vec::value_type retval=0.0;
  BOOST_FOREACH(const typename vec::value_type &vv,v) 
	retval += ( vv > 1e-20 ? vv * log(vv) : 0.0 );
  return retval;
}


