/****************************************************************************/
/*    Copyright (C) 2010 by Dominik Endres                                  */
/*    dominik.endres@klinikum.uni-tuebingen.de                              */ 
/*                                                                          */
/*   This program is free software; you can redistribute it and#or modify   */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation; either version 2 of the License, or     */
/*   (at your option) any later version.                                    */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*   You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                         */
/*    Free Software Foundation, Inc.,                                       */
/*    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             */
/****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <cassert>
#include "time.h"
#include <iostream>
#include <vector>
#include <boost/shared_ptr.hpp>


#include "variational_model.hpp"
#include "component_container.hpp"
#include "gauss_wishart.hpp"
#include "label_observation.hpp"
#include "gauss_wishart_flow_feature.hpp"

#include "dpmmc.hpp"
#include "ex_sequence.hpp"

#include "sequence_classifier.hpp"

// #include "sequence_model.hpp"


#include <fstream>
#include <string>


using namespace std;

typedef gaussWishartFlowFeature<> emissionT;

typedef bottomComponentContainer<emissionT> botT;
	
typedef labelObservation<string> labelT;
typedef bottomComponentContainer<labelT> labelConT;
	
typedef exSequence<botT> patchSeqT;
typedef componentContainer<patchSeqT> sequenceBagT;
	
typedef dpmmc<sequenceBagT,labelConT> seqLabAssocT;

const int dim=10; // dimensionality of input
const unsigned nruns=100; // number of runs
const unsigned nsamp=1000; // number of samples per run


void testGaussWishart() {
	gaussWishart<> gw,gengw,gw2;
	
	cout<<"***** TESTING: Gauss-wishart emission model *********"<<endl;
	// init learner
	gw.set_property("D",dim);
	// int ddim=cast<int>(gw.get_property("D"));
	gw.set_property("beta0",0.01);
	gw.set_property("m0",0.0);
	gw.set_property("n0",dim+1.0);
	gw.set_property("S0",1.0);
	gw.setPrior(0);
	
	
	// init teacher. means=3, cv-matrix diagonal with 2.0 on diag.
	gengw.set_property("D",dim);
	gengw.set_property("beta0",10000.0);
	gengw.set_property("m0",3.0);
	gengw.set_property("n0",10000.0);
	gengw.set_property("S0",2.0);
	
 	gengw.setPrior(0);
	
	// test serialisation
	ofstream of("gw.tmp");
	of<<gw;
	of.close();
	
	ifstream iff("gw.tmp");
	iff>>gw2;
	iff.close();
	
	if(!(gw==gw2)) {
	    ofstream of("gw2.tmp");
	    of<<gw2;
	    of.close();
	    throw runtime_error("GW models not equal after load. Maybe roundoff error. Please check gw.tmp and gw2.tmp");
	}
	
	
	// do learning. var. bound should increase on average.
	srand48(12345);
	
	std::vector<gaussWishart<>::pdataT > samples;
	for(unsigned j=0;j<nsamp;++j)
	  samples.push_back(gengw.sample());
	  
	  
	for(unsigned i=0;i<nruns;++i) {
	  cout<<"EM-iteration "<<i<<endl;
	      gw.beginLearning(true);
	      for(unsigned j=0;j<nsamp;++j) {
		gaussWishart<>::pdataT &curex=samples[j];
		gw.inference(curex);
		gw.learn(curex);
	      }
	      cout<<"Var. bound before parameter update: "<<gw.getVarBound()<<endl;
	      gw.updateParameterPosterior();  
	      
	      
	  gw.posteriorToPrior();
	}
	
	cout<<gw<<endl;  
}



template<class varmodel,class sourcemodel>
void runModel(varmodel & model,sourcemodel &source, string name="mymodel",unsigned numBatches=10,unsigned batchSize=100,unsigned maxIter=200) 
{
  // test load/save
  cout<<"Start save..."<<endl;
  ofstream os((name+".dat").c_str());
  os<<model<<flush;
  os.close();
  cout<<"Save ok. start load"<<endl;
  
  ifstream is((name+".dat").c_str());
  varmodel model2;
  is>>model2;
  is.close();
  cout<<"Load ok. start save2"<<endl;
	
  ofstream oi((name+"2.dat").c_str());
  oi<<model2;
  oi.close();
  cout<<"Load/save check done!"<<endl;
  
  if(model==model2) cout<<"Models equal after load/save"<<endl;
  else cerr<<"MODELS NOT EQUAL AFTER LOAD/SAVE"<<endl;
  
  // test copy-construction
  cout<<"Start copy"<<endl;
  varmodel cm2(model);
  cout<<"copy ok"<<endl;
  cm2.unshare();
  cout<<"unshare ok"<<endl;
  if(model==cm2) cout<<"Models equal after copy construction"<<endl;
  else cerr<<"MODELS NOT EQUAL AFTER COPY CONSTRUCTION"<<endl;
  ofstream oi2((name+"3.dat").c_str());
  oi2<<cm2;
  oi2.close();
  
 
  
  std::vector<typename varmodel::pdataT> dvec;
  for(unsigned i=0;i<numBatches*batchSize;++i) {
    dvec.push_back(source.sample());
  }
  ofstream oi4((name+".source.dat").c_str());
  oi4<<source;
  oi4.close();
  
  cout<<"Sampled "<<dvec.size()<<" data points"<<endl;
  
  model.posteriorToPrior();
  model.resetToPrior();
  cout<<"Initial values"<<endl;
  cout<<"entropy "<<model.getEntropy()<<endl;
  cout<<"var.bound "<<model.getVarBound()<<endl;
  cout<<"kl-div "<<model.getKL()<<endl;
  cout<<"likelihood "<<model.getLogLikelihood()<<endl;
  
//  cout<<"Incremental init of variational posterior for symmetry breaking.."<<endl; // utter rubbish. doesn't help.
  // TODO: a subsample should be enough for this ?
//   for(unsigned bs=0;bs<dvec.size();bs+=1) {
//     if(bs>dvec.size()) break;
//     model.beginLearning();
//     for(unsigned dp=0;dp<bs;++dp) {
//      model.inference(dvec[dp]);
//      model.learn(dvec[dp]);
//     }
//     model.updateParameterPosterior();
//   }
//  model.freezeAll(0);
  

//  cout<<"Incremental init of variational posterior for symmetry breaking DONE."<<endl;


  cout<<"Testing batch learning..."<<endl;
  
  clock_t start=clock();
  double elapsed=0;
  unsigned EMsteps=0;
  for(unsigned batch=0;batch<numBatches;++batch) {
    cout<<"At batch "<<batch<<endl;
    if(batch>2) model.freezeAll(0);
    float oldVB=-1e100,VB=0.0;
    for(EMsteps=0;EMsteps<maxIter;++EMsteps)
    {
      model.beginLearning(true);
      for(unsigned d=batch*batchSize;d<(batch+1)*batchSize;++d)
      {
	model.inference(dvec[d]);
	model.learn(dvec[d]);
      }
      elapsed+=clock()-start;
      VB=model.getVarBound();
      cout<<" iteration "<<EMsteps<<", var.bound: "<<VB<<endl;
      start=clock();
      model.updateParameterPosterior();
      if(fabs(VB/oldVB-1.0)<1e-4) break;
      oldVB=VB;
    }
    model.posteriorToPrior();
    cout<<endl;
  }
  cout<<"Learning took "<<(elapsed/CLOCKS_PER_SEC)/EMsteps/dvec.size()<<"s per example, not counting var bound eval"<<endl;
  
  cout<<"Testing inference..."<<endl;
  start=clock();
  for(unsigned d=0;d<dvec.size();++d)
  {
    model.resetToPrior();
    model.inference(dvec[d]);
    if(d%100==0) {cout<<".";cout.flush();}
  }
  cout<<endl;
  elapsed=clock()-start;
  cout<<"Inference took "<<elapsed/CLOCKS_PER_SEC/dvec.size()<<"s per example."<<endl;
  
  model.predictPass(1.0);
  
  ofstream resf;
  resf.open((name+".learn.dat").c_str());
  resf<<model;
  resf.close();
 
  
}



void runClassify(seqLabAssocT & model,string name="classifier",int numDataPoints=1000,int maxIter=100, int seqLen=5) 
{  
  // test copy-construction
  
  seqLabAssocT teacher(model);
  teacher.unshare();
  if(model==teacher) cout<<"Models equal after copy construction"<<endl;
  else cerr<<"MODELS NOT EQUAL AFTER COPY CONSTRUCTION"<<endl;
  ofstream oi2((name+".teacher.dat").c_str());
  oi2<<teacher;
  oi2.close();
  
  teacher.setPrior(10);
  teacher.resetToPrior();
  
  typedef std::vector< seqLabAssocT::pdataT> cldataT;
  cldataT training;
  for(int i=0;i<numDataPoints;++i) {
    training.push_back(teacher.sample());
  }
  cout<<"Sampled "<<training.size()<<" data points for training"<<endl;
    

  
  
  model.posteriorToPrior();
  model.resetToPrior();
  
  cout<<"Incremental init of variational posterior for symmetry breaking.."<<endl;
  for(unsigned bs=10;bs<=training.size();bs+=10) {
    if(bs>training.size()) break;
    model.beginLearning();
    for(unsigned dp=0;dp<bs;++dp) {
     model.inference(training[dp]);
     model.learn(training[dp]);
    }
    model.updateParameterPosterior();
  }
  cout<<"Incremental init of variational posterior for symmetry breaking DONE."<<endl;


  cout<<"Testing batch learning..."<<endl;
  
  clock_t start=clock();
  double elapsed=0;
  float oldVB=-1e100,VB=0.0;
  int EMsteps=0;
  for(EMsteps=0;EMsteps<maxIter;++EMsteps)
  {
    model.beginLearning(true);
    for(unsigned d=0;d<training.size();++d)
    {
      model.inference(training[d]);
      model.learn(training[d]);
    }
    elapsed+=clock()-start;
    VB=model.getVarBound();
    cout<<"iteration "<<EMsteps<<", var.bound: "<<VB<<endl;
    if(fabs(VB/oldVB-1.0)<1e-5) break;
    oldVB=VB;
    start=clock();
    model.updateParameterPosterior();
  }
  model.posteriorToPrior();
  cout<<"Learning took "<<(elapsed/CLOCKS_PER_SEC)/EMsteps/training.size()<<"s per example, not counting var bound eval"<<endl;
  
  
   seqLabAssocT::pComponentContainerT2 labels=model.getComponentContainer2();
  int numClasses=labels->components();
  cout<<"Testing classification, number of classes: "<<numClasses<<endl;
  
  std::map<string,cldataT> testdata;
  int numpts=0;
  while(numpts<numClasses*numDataPoints) {
   seqLabAssocT::pdataT example=teacher.sample();
   string clname=*(example->second);
   if(testdata.count(clname)==0) testdata[clname]=cldataT();
   if((int)testdata[clname].size()<numDataPoints) {
    example->second=boost::shared_ptr<string>();
    testdata[clname].push_back(example);
    numpts++;
   }
  }
  
  
  std::map<string,cldataT>::iterator clit;
  for(clit=testdata.begin();clit!=testdata.end();++clit)
  {
    string curclass=clit->first;
    cldataT & data=clit->second;
    cout<<"Testing class "<<curclass<<" on "<<data.size()<<" points"<<endl;
    std::map<string,float> guesses;
    for(int pn=0;pn<(int)data.size();pn+=seqLen) {
      if(pn>=(int)data.size()) break;
      model.resetToPrior();
      for(int s=0;s<seqLen;++s) model.inference(data[pn+s]);
      model.predictPass();
      for(int lab=0;lab<labels->components();++lab) {
	string ccl=labels->getComponent(0,lab)->getLabel();
	float prob=labels->getComponent(0,lab)->getParentPost();
	if(guesses.count(ccl)==0) guesses[ccl]=prob*seqLen;
	else guesses[ccl]+=prob*seqLen;
      }
    }
   std::map<string,float>::iterator rit;
   float sum=0.0;
   for(rit=guesses.begin();rit!=guesses.end();++rit) {
     cout<<rit->first<<": "<<rit->second/data.size()<<endl;
     sum+=rit->second/data.size();
   }
   cout<<"Sum: "<<sum<<endl<<endl;
  }
  
  ofstream resf;
  resf.open((name+".learn.dat").c_str());
  resf<<model;
  resf.close();
 
  
}

typedef std::pair<CPUData<float>, CPUData<float> > clpT;

std::vector<clpT> drawSample(sequenceClassifier & seqcl,int avgSeqLen,int numSeqs) {
    
    std::vector<clpT> sample;
    
    Random<float> & rngf = Random<float>::get();
    seqcl.reset();
    CPUData<float> flow;
    CPUData<float> label;
    seqcl.sample(flow,label);
    string clabel=sequenceClassifier::CPUDataToString(label);
    const float pchange=1.0/avgSeqLen;
    
    for(int tp=0;tp<numSeqs*avgSeqLen;++tp)
    {
      if(rngf.uniform01()<pchange) // change point. draw a new label.
      {
	seqcl.sample(flow,label);
	clabel=sequenceClassifier::CPUDataToString(label);
      }
      
      string slabel="";
      while(slabel!=clabel)
      {
	  seqcl.sample(flow,label);
	  slabel=sequenceClassifier::CPUDataToString(label);
      }
      
      sample.push_back(clpT(flow,label));
            
    }

    return sample;
  
}


int main(int argc, char *argv[]) {
  
	cout<<"***************** Testing optic flow classifier *****************"<<endl;
	
	sequenceClassifier motionClassifier;
	cout<<"CONSTRUICT OK"<<endl;
	motionClassifier.set_property("featureX",3);
	motionClassifier.set_property("featureY",3);
	motionClassifier.set_property("flowComponents",1);
	motionClassifier.set_property("segmentsX",5);
	motionClassifier.set_property("segmentsY",5);
	motionClassifier.set_property("sequenceLength",10);
	cout<<"PROP OK"<<endl;
	motionClassifier.initModel();
	cout<<"INIT OK"<<endl;
	motionClassifier.setPrior(10);
	
	cout<<"Testing save/load"<<endl;
	ofstream os("flowclass.dat");
	os<<motionClassifier;
	os.close();
	cout<<"Save ok. Start load"<<endl;
	ifstream is("flowclass.dat");
	sequenceClassifier generator;
	is>>generator;
	is.close();
	cout<<"Load ok. Start save2"<<endl;
	ofstream os2("flowclass2.dat");
	os2<<generator;
	os2.close();
	
	if(motionClassifier==generator) cout<<"Models equal after load/save"<<endl;
	else cout<<"MODELS NOT EQUAL AFTER LOAD/SAVE"<<endl;
	
	generator.set_property("maxNumLabels",3);
	generator.initModel();
	generator.setPrior(30);
	generator.addLabel("run");
	generator.addLabel("brawl");
	generator.addLabel("wave");
	generator.reset();
	
	
	cout<<"Saving generator"<<endl;
	ofstream osg("flowclass.generator.dat");
	osg<<generator;
	osg.close();
	
	cout<<"Testing copy construction"<<endl;
	double cbeg=omp_get_wtime();
	sequenceClassifier mc2(motionClassifier);
	cout<<"copy ok. Took "<<omp_get_wtime()-cbeg<<"s. Saving"<<endl;
	ofstream osc("flowclass3.dat");
	osc<<mc2;
	osc.close();
	
	
	if(motionClassifier==mc2) cout<<"Models equal after copy construction"<<endl;
	else cout<<"MODELS NOT EQUAL AFTER COPY CONSTRUCTION"<<endl;
	
	
	
	const int numBatches=3;
	const int avgSeqLen=cast<int>(motionClassifier.get_property("sequenceLength"))*4;
	const int testSequences=100;
	const int learnSequences=10;
	
	// learn in batches
	motionClassifier.set_property("numThreads",2);
	for(int batch=0;batch<numBatches;++batch)
	{
	 cout<<"AT BATCH "<<batch<<endl;
	 motionClassifier.clearBuffer();
	 std::vector<clpT> sample=drawSample(generator,avgSeqLen,learnSequences);
	 
 	 if(batch==0) motionClassifier.freeze(1+10);
	 else if(batch==1) motionClassifier.freeze(10);
	 else motionClassifier.freeze(0);
	  
	  std::string olab="";
	  for(int dp=0;dp<sample.size();++dp) {
	    std::string slab=sequenceClassifier::CPUDataToString(sample[dp].second);
	    bool newSeq=(slab!=olab) && (dp>0);
	    if(newSeq) {
	      olab=slab;
	    }
	    motionClassifier.bufferData(sample[dp].first,sample[dp].second,newSeq);
	  }
	  
	 
	 
	 
	 
	 float oldVB=-1e30,VB=0.0;
	 for(int EMsteps=0;EMsteps<50;++EMsteps) {
	    motionClassifier.beginLearning(true);
	    motionClassifier.learnBuffer();
	    //for(int dp=0;dp<flows.size();++dp) motionClassifier.learn(flows[dp],labels[dp]);
	    VB=motionClassifier.getVarBound();
	    cout<<"EM-step "<<EMsteps<<", var. bound: "<<VB<<endl;
	    motionClassifier.updateParameterPosterior();
	    if(fabs(fabs(VB)/fabs(oldVB)-1.0)<1e-3) break;
	    oldVB=VB;
	 }
	 motionClassifier.posteriorToPrior();
	}
	
	CPUData<float> flow;
	CPUData<float> label;
	
	cout<<"FORCING UPDATE FOR ORIG"<<endl;
	generator.sample(flow,label);
	motionClassifier.recognize(flow);
	cout<<"ORIG DONE"<<endl;
	
	cout<<"Saving learned model"<<endl;
	ofstream osl("flowclass.learn.dat");
	osl<<motionClassifier;
	osl.close();
	
	// test classification performance
	Random<float> & rngf = Random<float>::get();
	motionClassifier.reset();
	
	generator.sample(flow,label);
	string clabel=sequenceClassifier::CPUDataToString(label);
	const float pchange=1.0/avgSeqLen;
	std::map<pair<string,string>,float> classres;
	
	cout<<"START COPY FOR CLASSIFY"<<endl;
	std::vector<sequenceClassifier> mocl2;
	mocl2.resize(20,motionClassifier);
	cout<<"DONE COPY FOR CLASSIFY"<<endl;
	if (mocl2[0]==motionClassifier) cout<<"COPIES EQUAL"<<endl;
	else cout<<"COPIES NOT EEQUAL"<<endl;
	
	cout<<"FORCING UPDATE"<<endl;
	generator.sample(flow,label);
	mocl2[0].recognize(flow);
	cout<<"COPY DONE"<<endl;
	
	
	
	
	cout<<"Saving copy of learned model"<<endl;
	ofstream osl2("flowclass.learn2.dat");
	osl2<<mocl2[0];
	osl2.close();
	
	mocl2[0].reset();
	
	double start=omp_get_wtime();
	int numex=0;
	for(int tp=0;tp<testSequences*avgSeqLen;++tp)
	{
	  if(rngf.uniform01()<pchange) // change point. draw a new label.
	  {
	    generator.sample(flow,label);
	    clabel=sequenceClassifier::CPUDataToString(label);
	  }
	  
	  string slabel="";
	  while(slabel!=clabel)
	  {
	     generator.sample(flow,label);
	     slabel=sequenceClassifier::CPUDataToString(label);
	  }
	  
	  mocl2[0].recognize(flow);
	  CPUData<float> guess=mocl2[0].guessLabel(50);
	  string sguess=sequenceClassifier::CPUDataToString(guess);
	  
	  std::pair<string,string> llp(slabel,sguess);
	  if(classres.find(llp)==classres.end()) classres[llp]=0.0;
	  classres[llp]+=1.0;
	  
	  ++numex;
	}
	double ctime=omp_get_wtime()-start;
	
	// evaluate classification performance
	std::map<pair<string,string>,float>::iterator cliter;
	float norm=0.0,clrate=0.0;
	for(cliter=classres.begin();cliter!=classres.end();++cliter) norm+=cliter->second;
	for(cliter=classres.begin();cliter!=classres.end();++cliter) {
	  cliter->second/=norm;
	  std::string label=cliter->first.first;
	  std::string glabel=cliter->first.second;
	  if(label==glabel) clrate+=cliter->second;
	  cout<<"ACT: "<<label<<", GUESS: "<<glabel<<", prob: "<<cliter->second<<endl;
	}
	cout<<"CLASSIFICATION RATE "<<clrate<<endl;
	cout<<"CLASSIFICATION TOOK "<<ctime/numex<<" PER EXAMPLE"<<endl;
	
  
  
	cout<<"***************** Testing optic flow classifier END *****************"<<endl;
	return 0;
  
	Random<float> & rng = Random<float>::get();
	rng.seed(260671);
  
	const int fx=3;
	const int fy=3;
	const int fc=1;
	const int segX=1;
	const int segY=1;
	
	const int numClusters=10;
	const int numLabels=5;
	
	
	cout<<"************ Gauss-wishart flow feature start ***********"<<endl;
	
	emissionT gwflow;
	
	gwflow.set_property("featureX",fx);
	gwflow.set_property("featureY",fy);
	gwflow.set_property("flowComponents",fc);
	gwflow.set_property("segmentsX",segX);
	gwflow.set_property("segmentsY",segY);
	
	gwflow.setPrior(0);
	
	emissionT gwfs=gwflow;
	gwfs.unshare();
	gwfs.setPrior(40);
	
	
	
	runModel(gwflow,gwfs,"gwflow");
	
	
	cout<<"************ Gauss-wishart flow feature end ***********"<<endl;
	
	
	cout<<"*********** mixture model start ************"<<endl;
	
	typedef dpmmc<botT> patchT;
	
	rng.seed(260671);
	patchT patch;
 	boost::shared_ptr<botT> ppatchCon=patch.getComponentContainer();
 	
	
	patch.getComponentContainer()->getPrototype().set_property("featureX",fx);
	patch.getComponentContainer()->getPrototype().set_property("featureY",fy);
	patch.getComponentContainer()->getPrototype().set_property("flowComponents",fc);
	patch.getComponentContainer()->getPrototype().set_property("segmentsX",segX);
	patch.getComponentContainer()->getPrototype().set_property("segmentsY",segY);
	patch.getComponentContainer()->getPrototype().setPrior(0);
	
  	patch.resize(numClusters);
	rng.seed(time(0));
 	for(int comp=0;comp<patch.components();++comp) {
 	  botT::componentsT pcc=ppatchCon->getComponents();
	  pcc(0,comp)->set_property("S0",1.0);
	  pcc(0,comp)->set_property("m0",1.0);
 	  pcc(0,comp)->setPrior(10);
 	}
	
	patch.set_property("alpha0",4.0);
	patch.set_property("finite",true);
	patch.setPrior(20);
	
	patch.beginLearning();
	
	
	patchT ps=patch;
	ps.unshare();
	ps.set_property("alpha0",2.0);
	ps.set_property("finite",true);
	ps.setPrior(0);
	for(int comp=0;comp<ps.components();++comp) {
 	  botT::componentsT pcc=ps.getComponentContainer()->getComponents();
	  boost::shared_ptr<emissionT> pco=pcc(0,comp);
	  pco->set_property("S0",0.01*(comp+1));
 	  pco->setPrior(0);
	  emissionT::floatVecT mean(fx*fy*fc);
	  std::fill(mean.begin(),mean.end(),0.0);
	  for(unsigned i=0;i<mean.size();++i) 
	    if(i%ps.components() == (unsigned)comp) mean(i)=10.0;
	  pco->setMean(mean);
 	}
	
	patch.freeze(1);
	ppatchCon->freeze(1);
	
 	runModel(patch,ps,"mixture");
	
	cout<<"*********** mixture model END **************"<<endl<<endl;
	
	
/*  
	cout<<"********* Mixture of mixtures start **************"<<endl;
	
 	
	typedef dpmmc<patchBagT> topT;
	
	
	topT seqbag;
	 topT::pComponentContainerT patches=seqbag.getComponentContainer();
	 patchBagT::pSubContainerT features=patches->getSubContainer();
	
	
	// resize & init emission models
	features->setShared(true);
	features->getPrototype().setPrior(2);
	features->resize(20);
	for(int comp=0;comp<features->components();++comp) {
 	  botT::componentsT pcc=features->getComponents();
 	  pcc(0,comp)->set_property("D",inputDim);
 	  pcc(0,comp)->setPrior(2);
 	}
 	
 	// resize & init patch mixtures
	patches->getPrototype().set_property("alpha0",1.0);
	
	seqbag.set_property("alpha0",4.0);
	
	seqbag.resize(4);
	seqbag.setPrior(7);
	
	
	runModel(seqbag,"bagofsequences");
	
	cout<<"*************** mixture of mixtures END ********************"<<endl<<endl;
 */
	
	cout<<"**************** Sequence-label associations ***************"<<endl;
	
	seqLabAssocT seqclass;
	 seqLabAssocT::pComponentContainerT1 sequences=seqclass.getComponentContainer1();
	 sequenceBagT::pSubContainerT emissions=sequences->getSubContainer();
	
	 seqLabAssocT::pComponentContainerT2 labels=seqclass.getComponentContainer2();
	
	// resize & init emission models
	emissions->setShared(true);
	emissions->getPrototype().set_property("featureX",fx);
	emissions->getPrototype().set_property("featureY",fy);
	emissions->getPrototype().set_property("flowComponents",fc);
	emissions->getPrototype().set_property("segmentsX",segX);
	emissions->getPrototype().set_property("segmentsY",segY);
	emissions->getPrototype().setPrior(0);
	emissions->resize(numClusters);
	
	for(int comp=0;comp<emissions->components();++comp) {
 	  botT::componentsT pcc=emissions->getComponents();
	  pcc(0,comp)->set_property("m0",10.0*comp);
	  pcc(0,comp)->set_property("S0",1.0);
	  pcc(0,comp)->set_property("n0",100.0);
 	  pcc(0,comp)->setPrior(10);
 	}
	
	// resize sequence-label associations
	seqclass.set_property("finite","true");
	seqclass.resize(numLabels);
	seqclass.setPrior(0);
	
	// init label names
	labelConT::componentsT plab=labels->getComponents();
	for(int labno=0;labno<labels->components();++labno) {
	  plab(0,labno)->set_property("label",(boost::format("Label%02d") % labno).str() );
	  plab(0,labno)->setPrior(0);
	}
	
	// init sequences
	sequenceBagT::componentsT pseq=sequences->getComponents();
	for(int seqno=0;seqno<sequences->components();++seqno) {
	 pseq(0,seqno)->set_property("finite",0);
	 pseq(0,seqno)->set_property("alpha0",4.0);
	 pseq(0,seqno)->setPrior(100);
	}
	
	seqclass.beginLearning();
	
	
	runClassify(seqclass,"labelclass");
	
	
	cout<<"**************** Sequence-label associations END ***************"<<endl<<endl;
	
	
	
	 
	
	
	
	
	
	
	
	
	
/*
	
	// 1st. hidden: frames = bags of patches
 	typedef componentContainer<dpmmc<botT> > layer1T;
	

	// 2nd. hidden: sequences of frames (put in sequence model instead of dpmmc once it's finished)
 	typedef componentContainer<dpmmc<layer1T> > layer2T;
	
	
	// bottom of label pathway
	typedef labelObservation<> labelT;
	typedef bottomComponentContainer<labelT> labelConT;
	
	// mixtures of labels for association with sequence type
	typedef dpmmc<labelConT> labelDistT;
	typedef componentContainer<labelDistT> labelDistConT;
	
	// for classification: associate mixtures of labels with layer2T, which are the sequences
	typedef dpmmc<layer2T,labelDistConT> topT;
	
	
	//shared hierarchy example
	
 	//instantiate. Top level can be static, subordinates are in shared_ptr.
 	topT bosmod;
 	
	// below top level, branch into label path and image-sequence path
	topT::pComponentContainerT1 sequences=bosmod.getComponentContainer1();
	topT::pComponentContainerT2 labelHistograms=bosmod.getComponentContainer2();
	
	// (observable) bottom of label path
	labelDistT::pComponentContainerT labels=labelHistograms->getSubContainer();
	
	// salient region is described by bag of patches. A patch is a collection of gaussian features with full covariance
	layer2T::pSubContainerT salientRegion=sequences->getSubContainer();
	layer1T::pSubContainerT patches=salientRegion->getSubContainer();
	
	//lower layers contain shared components, top layer doesnt.
	labels->setShared(true);
		
	patches->setShared(true);
	salientRegion->setShared(true);

	
	
	labels->resize(10);
	int lnum=0;
	BOOST_FOREACH( labelConT::pComponentT pc,labels->getComponents() )
	{
	  pc->set_property("label",(boost::format("Label%03d") % lnum).str());
	  pc->setPrior(10);
	  ++lnum;
	}
	
	BOOST_FOREACH( labelDistConT::pComponentT pc, labelHistograms->getComponents() )
	  pc->setPrior(10);
	  
	
	cout<<"Label init ok"<<endl;
	
	patches->resize(200);
	BOOST_FOREACH( botT::pComponentT pc,patches->getComponents() )
	{
	  pc->set_property("D",inputDim);
	  pc->setPrior(10);
	}
	
	cout<<"Patch init ok"<<endl;
	
	salientRegion->resize(100);
	BOOST_FOREACH(layer1T::pComponentT pc,salientRegion->getComponents() )
	  pc->setPrior(10);
	
	
	bosmod.resize(20);
	cout<<"Resize bosmod ok"<<endl;
	bosmod.setPrior(10);
	BOOST_FOREACH(layer2T::pComponentT pc,sequences->getComponents() )
	  pc->setPrior(10);
	
	  //TODO: should setPrior() operate hierarchically?
	
	cout<<"INIT OK"<<endl;
	
	runModel(bosmod);
	return 0;
	
/*	
	//non-shared hierarchy example
	
	//instantiate. Containers are not shared by default.
	layer2T layer2ns;
	layer2ns.resize(3);
	layer2T::componentsT l2c=layer2ns.getComponents();
	
	
	//resize 1st layer components.
	int botcount=0;
	for(unsigned i=0;i<l2c.size();++i) {
	  l2c[i]->resize(i+3);
	  layer2T::pSubContainerT pl1cont=l2c[i]->getComponentContainer();
	  layer1T::componentsT l1c=pl1cont->getComponents();
	  //resize bottom layer components
	  for(unsigned j=0;j<l1c.size();++j) {
	    l1c[j]->resize(j+3);
	    layer1T::pSubContainerT pbotcont=l1c[j]->getComponentContainer();
	    //init bottom layer components.
	    botT::componentsT botc=pbotcont->getComponents();
	    botcount+=botc.size();
	    for(unsigned k=0;k<botc.size();++k) {
	      botc[k]->set_property("D",inputDim);
	      botc[k]->setPrior(10);
	    }
	  }
	}  
	
	cout<<"NON-SHARED CONSTRUCTION DONE"<<endl;
	
	ofstream os2("nonSharedHierarchy.dat");
	os2<<layer2ns;
	os2.close();
	
	ifstream is2("nonSharedHierarchy.dat");
	layer2T l2nsnew;
	is2>>l2nsnew;
	is2.close();
	
	ofstream oi2("nonSharedHierarchy2.dat");
	oi2<<l2nsnew;
	oi2.close();
	
	assert(l2nsnew==layer2ns);
	
	
	
	cout<<"NOT-SHARED INFERENCE, bottom level component count "<<botcount<<endl;
	layer2ns.inference(pdat->first);
	l2nsnew.inference(pdat->first);
	layer2ns.learn(pdat->first,layer2T::floatVecT(layer2ns.size(),0.8));
	l2nsnew.learn(pdat->first,layer2T::floatVecT(layer2ns.size(),0.8));
	*/
	

	
	//testGaussWishart();
	return 0; 
/*  
	sequenceModel<labelObservation<std::string> > seq;
	
	seq.set_property("alpha0",0.5);
	seq.set_property("numStates",10);
	seq.set_property("maxBufSize",200);
	
	
	seq.set_property("priorType",std::string("forwardCircular"));
	cout<<seq.get_property("priorType")<<endl;
	
	seq.setPrior(0);
	seq.resetToPrior();
	
	
	ofstream outf("seqmod.dat");
	outf<<seq;
	outf.close();
	
	ifstream inf("seqmod.dat");
	
	sequenceModel<labelObservation<std::string> > seq2;
	inf>>seq2;
	inf.close();
	
	cout<<seq2<<endl;
	*/

	return 0;
};
