#include <boost/python.hpp>

#undef NO_IMPORT_ARRAY
extern "C" {
#include<numpy/arrayobject.h>
}

#ifndef HAVE_NUMPY
# define HAVE_NUMPY
#endif

#include "numpy_support.hpp"
#include "python_property_access.hpp"
#include "dpmm.hpp"
#include "full_covariance_gaussian_model.hpp"
#include "SeariseProperties.hpp"
#include "blitz_properties.hpp"
#include "gauss_wishart.hpp"
#include <fstream>
#include "ublas_storage_adaptors.hpp"
#include "label_observation.hpp"

using namespace boost::python;
using boost::numeric::ublas::readonly_array_adaptor;

typedef full_covariance_gaussian_model< float > gaussian;
typedef gaussWishart< > gaussian2;


//inject python property adaptors into the component type
struct gaussian_python : public gaussian
{
  void set( std::string key, PyObject * pyvalue )
  {
    SeariseProperty value = property_from_pyobject( pyvalue );
    gaussian::set_property( key, value );
  }

  PyObject * get( std::string key )
  {
    return pyobject_from_property( gaussian::get_property( key ) );
  }
  
};

typedef dpmm< gaussian_python > gdpmm;

//inject python property adaptors into the component type
struct gaussian2_python : public gaussian2
{
  void set( std::string key, PyObject * pyvalue )
  {
    SeariseProperty value = property_from_pyobject( pyvalue );
    gaussian2::set_property( key, value );
  }

  PyObject * get( std::string key )
  {
    return pyobject_from_property( gaussian2::get_property( key ) );
  }
  
  
};

typedef dpmm< gaussian2_python > gdpmm2;

//inject handles for critical methods to be useful for python
struct gdpmm_python : public gdpmm
{

  void set( std::string key, PyObject * pyvalue )
  {
    SeariseProperty value = property_from_pyobject( pyvalue );
    gdpmm::set_property( key, value );
  }

  PyObject * get( std::string key )
  {
    return pyobject_from_property( gdpmm::get_property( key ) );
  }

  static PyObject * blitz2py( blitz::Array<float,1> x )
  {
    CPUData<float> data;
    blitz2cpudata( x, data );
    return cpudata2python( data );
  }

  static blitz::Array<float,1> py2blitz( PyObject * x )
  {
    CPUData<float> data = numpy_to_CPUData_converter::cpudata<float>( x );
    blitz::Array<float,1> bdata;
    cpudata2blitz( data, bdata );	
	return bdata;
  }

  PyObject * py_sample()
  {
    return blitz2py( gdpmm::sample() );
  }
  
  void infer( PyObject * x )
  {
    gdpmm::inference( py2blitz( x ) );
  }
  
  void learn( PyObject * x )
  {
    gdpmm::learn( py2blitz( x ) );
  }

  float logLikelihood( PyObject * x )
  {
    return gdpmm::getLogLikelihood( );
  }
  
  void accumulateVarBound(PyObject * x) 
  {
    gdpmm::accumulateVarBound( 1.0 );
  }

  gaussian_python & component( int i )
  {
    return gdpmm::component( i );
  }
  
  void save(PyObject * name) 
  {
    std::string filename=PyString_AsString(name); 
    std::ofstream ofile(filename.c_str());
    if(!ofile.is_open()) throw std::runtime_error("Cannot open file"+filename+" for writing");
    writeToStream(ofile);
    ofile.close();
  }
  
  void load(PyObject * name) 
  {
    std::string filename=PyString_AsString(name); 
    std::ifstream ifile(filename.c_str());
    if(!ifile.is_open()) throw std::runtime_error("Cannot open file"+filename+" for reading");
    readFromStream(ifile);
    ifile.close();
  }

};

//inject handles for critical methods to be useful for python
struct gdpmm_python2 : public gdpmm2
{

  CPUData<float> _tmp;
  typedef ublas::vector< const float, readonly_array_adaptor<float> > vector_adaptor;
  typedef ublas::vector<float_type> vector_t;

  void set( std::string key, PyObject * pyvalue )
  {
    SeariseProperty value = property_from_pyobject( pyvalue );
    gdpmm2::set_property( key, value );
  }

  PyObject * get( std::string key )
  {
    return pyobject_from_property( gdpmm2::get_property( key ) );
  }

  boost::shared_ptr<vector_t> py2ublas( PyObject * x )
  {
    _tmp = numpy_to_CPUData_converter::cpudata<float>( x );
    vector_adaptor bdata( make_vector_from_pointer( _tmp.size(), _tmp.data() ) );
	boost::shared_ptr<vector_t> result( new vector_t( bdata ) );
	return result;
  }

  PyObject * ublas2py( ublas::vector<float> x )
  {
	npy_intp dims[1] = { x.size() };
	PyObject * result = PyArray_New( &PyArray_Type, 1, dims, PyArray_FLOAT, NULL, NULL, sizeof(float), NULL, NULL );
	float * datap = (float*) PyArray_BYTES( result );
	std::copy( x.begin(), x.end(), datap );
	return result;
  }

  PyObject * py_sample()
  {
    return ublas2py( *gdpmm2::sample() );
  }
  
  void infer( PyObject * x )
  {
    gdpmm2::inference( py2ublas( x ) );
  }
  
  void learn( PyObject * x )
  {
    gdpmm2::learn( py2ublas( x ) );
  }

  float logLikelihood( PyObject * x )
  {
    return gdpmm2::getLogLikelihood( );
  }
  
  void accumulateVarBound(PyObject * x) 
  {
    gdpmm2::accumulateVarBound( 1.0 );
  }
  

  gaussian2_python & component( int i )
  {
    return gdpmm2::component( i );
  }
  
  void save(PyObject * name) 
  {
    std::string filename=PyString_AsString(name); 
    std::ofstream ofile(filename.c_str());
    if(!ofile.is_open()) throw std::runtime_error("Cannot open file"+filename+" for writing");
    writeToStream(ofile);
    ofile.close();
  }
  
  void load(PyObject * name) 
  {
    std::string filename=PyString_AsString(name); 
    std::ifstream ifile(filename.c_str());
    if(!ifile.is_open()) throw std::runtime_error("Cannot open file"+filename+" for reading");
    readFromStream(ifile);
    ifile.close();
  }

};



BOOST_PYTHON_MODULE(varmod)
{
  import_array();

  class_< gaussian_python > ( "gaussian", "" )
  .def( "set", &gaussian_python::set, args( "key","value") )
  .def( "get", &gaussian_python::get, args( "key" ) )
  ;

  class_< gdpmm_python > ( "gdpmm", "" )
  .def( "set", &gdpmm_python::set, args( "key","value") )
  .def( "get", &gdpmm_python::get, args( "key" ) )
  .def( "setPrior", &gdpmm::setPrior )
  .def( "posteriorToPrior", &gdpmm::posteriorToPrior )  
  .def( "beginLearning", &gdpmm::beginLearning )
  .def( "updateParameterPosterior", &gdpmm::updateParameterPosterior )
  .def( "infer", &gdpmm_python::infer, args( "data" ) )
  .def( "learn", &gdpmm_python::learn, args( "data" ) )
  .def( "logLikelihood", &gdpmm_python::logLikelihood, args( "data" ) )
  .def( "sample", &gdpmm_python::py_sample )
  .def( "component", &gdpmm_python::component, return_internal_reference<>() )
  .def( "save", &gdpmm_python::save, args("data") )
  .def( "load", &gdpmm_python::load, args("data") )
  .def( "getKL", &gdpmm_python::getKL )
  .def( "getVarBound", &gdpmm_python::getVarBound )
  .def( "accumulateVarBound", &gdpmm_python::accumulateVarBound, args("data"))
  .def( "getNumUsed", &gdpmm_python::getNumUsed )
  ;

  class_< gaussian2_python > ( "gaussian2", "" )
  .def( "set", &gaussian2_python::set, args( "key","value") )
  .def( "get", &gaussian2_python::get, args( "key" ) )
  ;

  class_< gdpmm_python2 > ( "gdpmm2", "" )
  .def( "set", &gdpmm_python2::set, args( "key","value") )
  .def( "get", &gdpmm_python2::get, args( "key" ) )
  .def( "setPrior", &gdpmm2::setPrior )
  .def( "posteriorToPrior", &gdpmm2::posteriorToPrior )  
  .def( "beginLearning", &gdpmm2::beginLearning )
  .def( "updateParameterPosterior", &gdpmm2::updateParameterPosterior )
  .def( "infer", &gdpmm_python2::infer, args( "data" ) )
  .def( "learn", &gdpmm_python2::learn, args( "data" ) )
  .def( "logLikelihood", &gdpmm_python2::logLikelihood, args( "data" ) )
  .def( "sample", &gdpmm_python2::py_sample )
  .def( "component", &gdpmm_python2::component, return_internal_reference<>() )
  .def( "save", &gdpmm_python2::save, args("data") )
  .def( "load", &gdpmm_python2::load, args("data") )
  .def( "getKL", &gdpmm_python2::getKL )
  .def( "getVarBound", &gdpmm_python2::getVarBound )
  .def( "accumulateVarBound", &gdpmm_python2::accumulateVarBound, args("data"))
  .def( "getNumUsed", &gdpmm_python2::getNumUsed )
  ;
  
  
}
