#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys

sys.path.append(".")
sys.path.append("..")

import numpy
import varmod
import time
import matplotlib.pyplot as plt

NDIM=2

def make_model( n_init = 0, mtype = varmod.gdpmm, truncationLimit=5, sig0=2.0, beta0=0.01, nu0=2*NDIM+1, finite=False ):
  global NDIM
  model = mtype()
  model.set("alpha0",5)
  model.set("T",truncationLimit)
  model.set("prototype.D",NDIM)
  model.set("prototype.m0",0)
  model.set("prototype.n0",nu0)
  model.set("prototype.beta0",beta0)
  model.set("prototype.S0",sig0)
  model.set("finite",finite)
  model.setPrior(n_init)
  return model

def normrho(lp):
  p = numpy.exp( lp - lp.max() )
  p /= p.sum()
  return p

def plot_distrib(distrib,X,Y):
  lp = numpy.zeros((len(X),len(Y)))
  for i,x in enumerate(X):
    for j,y in enumerate(Y):
      lp[j,i] = distrib.logLikelihood(numpy.array((x,y)))
  p = normrho(lp)
  plt.imshow( lp[::-1], extent = (X.min(),X.max(),Y.min(),Y.max()) )

def draw_samples(distrib,N):
  global NDIM
  samples = numpy.zeros((N,NDIM))
  for i in range(N):
    samples[i] = model.sample().ravel()
  return samples

def learn_data(model,data,iters=5,batchsizes=[(20,2000),(100,1000000)],checkVarBound=False):
  # learn in (mini-)batches. a batchsize descriptor is a tuple (batchsize,maxND), meaning that up to the maxND-th datapoint, learn in batches of size maxND
  # the default setting does online-learning up to 500 examples, to break the prior symmetry in the right direction
  # after that, learning can proceed in larger steps, thus avioding costly matrix inversions (or suchlike) in updateParameterPosterior for every example.
  
  def getBatchSize(dpnum):
    batchsize=0
    for bs,udp in batchsizes:
      batchsize=bs
      if dpnum<udp: break
    return batchsize
      
  
  
  currentBatch=[]
  dataPointNumber=0
  cbs=getBatchSize(dataPointNumber)
  
  startt=time.clock()
  

  
  for x in data:
    currentBatch+=[x]
    dataPointNumber+=1
    if len(currentBatch)<cbs: continue
    oldVB=-1e100
    for i in range(iters):
      model.beginLearning(False)
      for dp in currentBatch:
	model.infer(dp)
	model.learn(dp)
	if checkVarBound: model.accumulateVarBound(dp)
      if checkVarBound: 
	curVB=model.getVarBound()
	if (curVB-oldVB)/numpy.abs(oldVB)<-0.1: print "WARNING: variational Bound decreased from ",oldVB,"to",curVB,"iteration",i
	oldVB=curVB
      model.updateParameterPosterior()
    model.posteriorToPrior() 
    
    cbs=getBatchSize(dataPointNumber)

    currentBatch=[]
   
  return time.clock()-startt
    
def log_likelihood(model,data):
  llh = 0.0
  for x in data:
    model.infer(x)
    llh += model.logLikelihood(x)
  return llh / len(data)
    
  
model  = make_model( 10, mtype = varmod.gdpmm, finite=False )
model2 = make_model( 0, mtype = varmod.gdpmm, truncationLimit=40, sig0=0.1 )
model3 = make_model( 0, mtype = varmod.gdpmm2, truncationLimit=40, sig0=0.1 )

print "made models"

samples = draw_samples(model,2000)
samples_test = draw_samples(model,2000)
ofile=open("samples.dat","w")
for sample in samples:
  print >> ofile,sample
ofile.close()

print "drew samples"

print type(model)

genll=log_likelihood(model,samples_test)

print "evaluated ll"

print "Daniel impl. takes ",learn_data( model2, samples , 10)
print "Log likelihood of test data: ",log_likelihood(model2,samples_test)," should be ",genll
print "Number of used components",model2.getNumUsed()

dtime=learn_data( model3, samples , 10, checkVarBound=True)
print "Dominiks impl. takes",dtime,"with var. bound calculation"
print "Log likelihood of test data: ",log_likelihood(model3,samples_test)," should be ",genll
print "Number of used components",model3.getNumUsed()

if False:
  print "Checking save/load"
  
  model3.save("dominik.dat")
  model3load=varmod.gdpmm2()
  model3load.load("dominik.dat")
  print "Load done"
  print "Log likelihood after load, Dominik: ",log_likelihood(model3load,samples_test)
  

if NDIM!=2: sys.exit()

x = numpy.linspace(samples[:,0].min(),samples[:,0].max())
y = numpy.linspace(samples[:,1].min(),samples[:,1].max())
print x.min(), x.max()

show_every=1

plt.subplot(2,2,1)
plot_distrib(model,x,y)
plt.plot(samples[::show_every,0],samples[::show_every,1],'.')
plt.title("Generator")
plt.subplot(2,2,2)
plot_distrib(model2,x,y)
plt.plot(samples[::show_every,0],samples[::show_every,1],'.')
plt.title("Daniel")
plt.subplot(2,2,3)
plot_distrib(model3,x,y)
plt.plot(samples[::show_every,0],samples[::show_every,1],'.')
plt.title("Dominik")

plt.show()
